#include "blitting.h"


static void BlitBufTilePixel8(void *tp, const void *sp);
static void BlitBufTilePixel15(void *tp, const void *sp);
static void BlitBufTilePixel16(void *tp, const void *sp);
static void BlitBufTilePixel24(void *tp, const void *sp);
static void BlitBufTilePixel32(void *tp, const void *sp);

static void BlitBufTileArea(
	unsigned int bpp,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int offset_x, int offset_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	float zoom,
	float magnification,
	void (*pixel_op_func)(void *, const void *)
);
void BlitBufTile(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int offset_x, int offset_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	float zoom,			/* 0.1 to 1.0 supported */
	float magnification		/* 1.0 to 255.0 supported */
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)

#define WARPTOZERO(a,b)	(((a) >= (b)) ? 0 : (a))


static void BlitBufTilePixel8(void *tp, const void *sp)
{
	*(u_int8_t *)tp = *(const u_int8_t *)sp;
}

static void BlitBufTilePixel15(void *tp, const void *sp)
{
	*(u_int16_t *)tp = *(const u_int16_t *)sp;
}

static void BlitBufTilePixel16(void *tp, const void *sp)
{
	*(u_int16_t *)tp = *(const u_int16_t *)sp;
}

static void BlitBufTilePixel24(void *tp, const void *sp)
{
	u_int8_t *tp8 = (u_int8_t *)tp;
	const u_int8_t *sp8 = (const u_int8_t *)sp;
	tp8[0] = sp8[0];
	tp8[1] = sp8[1];
	tp8[2] = sp8[2];
}

static void BlitBufTilePixel32(void *tp, const void *sp)
{
	*(u_int32_t *)tp = *(const u_int32_t *)sp;
}


/*
 *	Copy/tiles the source buffer to the target buffer.
 */
static void BlitBufTileArea(
	unsigned int bpp,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int offset_x, int offset_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	float zoom,
	float magnification,
	void (*pixel_op_func)(void *, const void *)
)
{
	int	src_x_start, src_y_start,
		/* tar_x, */ tar_y,
		src_x, src_y,
		tar_width2, tar_height2;
	unsigned int src_x_inc, src_y_inc;
	u_int8_t *tar_ptr, *tar_edge, *tar_end;
	const u_int8_t *src_ptr;

	if((tar_buf == NULL) || (src_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (src_width == 0) || (src_height == 0) ||
	   (zoom <= 0.0f)
	)
	    return;

	if(zoom > 1.0f)
	    zoom = 1.0f;

	if(magnification < 1.0f)
	    magnification = 1.0f;

	/* Calculate zoom applyed target width and height */
	tar_width2 = (int)((int)tar_width / zoom);
	tar_height2 = (int)((int)tar_height / zoom);

	/* Calculate the starting tile position within the source
	 * buffer
	 */
	src_x_start = (int)(
	    ((offset_x / zoom) - ((int)tar_width2 / 2)) /
	    magnification
	);
	src_y_start = (int)(
	    ((offset_y / zoom) + ((int)tar_height2 / 2)) /
	    magnification
	);
	/* Modulate the source starting positions */
	if(src_x_start < 0)
	    src_x_start = (src_x_start % (int)src_width) + (int)src_width;
	else
	    src_x_start = src_x_start % (int)src_width;

	if(src_y_start < 0)
	    src_y_start = (src_y_start % (int)src_height) * -1;
	else
	    src_y_start = (int)src_height - (src_y_start % (int)src_height);

	/* Tile the source starting positions so that they fall within
	 * the buffer
	 */
	if(src_x_start < 0)
	    src_x_start = (int)src_width - 1; 
	else if(src_x_start >= (int)src_width)
	    src_x_start = 0;
	if(src_y_start < 0)
	    src_y_start = (int)src_height - 1;
	else if(src_y_start >= (int)src_height)
	    src_y_start = 0;

	/* Set the source starting position */  
	src_x = src_x_start;
	src_y = src_y_start;


	/* Convert all source units to * 256
	 *
	 * With the exception of src_bpl
	 */
	src_x_start <<= 8;
	src_y_start <<= 8;
	src_width <<= 8;
	src_height <<= 8;
	src_x <<= 8;
	src_y <<= 8;

	src_x_inc = (unsigned int)MAX(
	    (256 / magnification / zoom),
	    1
	);
	src_y_inc = (unsigned int)MAX(
	    (256 / magnification / zoom),
	    1
	);

	/* Set the target starting position */
/*	tar_x = 0; */
	tar_y = 0;


	/* Get the starting target and source buffer pointers */
	tar_ptr = tar_buf;
	tar_edge = tar_ptr + (tar_width * bpp);
	tar_end = tar_ptr + (tar_bpl * tar_height);

	src_ptr = src_buf +
	    ((src_y >> 8) * (int)src_bpl) +
	    ((src_x >> 8) * (int)bpp);


	/* Begin blitting */
	while(tar_ptr < tar_end)
	{
	    /* Copy this pixel */
	    pixel_op_func(tar_ptr, src_ptr);

	    /* Seek to the next target pixel */
	    tar_ptr += bpp;

	    /* Increment the x positions */
/*	    tar_x++; */
	    src_x = WARPTOZERO(
		(src_x + (int)src_x_inc),
		(int)src_width
	    );

	    /* Target new line? */
	    if(tar_ptr >= tar_edge)
	    {
		tar_y++;
		tar_ptr = tar_buf + (tar_y * (int)tar_bpl);
		tar_edge = tar_ptr + (tar_width * bpp);

		src_x = src_x_start;
		src_y = WARPTOZERO(
		    (src_y + (int)src_y_inc),
		    (int)src_height
		);
	    }

	    /* Get the next source pixel */
	    src_ptr = src_buf +
		((src_y >> 8) * (int)src_bpl) +
		((src_x >> 8) * (int)bpp);
	}
}

void BlitBufTile(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int offset_x, int offset_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	float zoom,			/* 0.1 to 1.0 supported */
	float magnification		/* 1.0 to 255.0 supported */
)
{
	void (*pixel_op_func)(void *, const void *);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;
	if(src_bpl == 0)
	    src_bpl = src_width * bpp;

	switch(bpp)
	{
	  case 1:
	    pixel_op_func = BlitBufTilePixel8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		pixel_op_func = BlitBufTilePixel15;
		break;
	      case 16:
		pixel_op_func = BlitBufTilePixel16;
		break;
              default:
                return;
                break;
	    }
	    break;
	  case 3:
            pixel_op_func = BlitBufTilePixel24;
            break;
	  case 4:
	    switch(depth)
	    {
	      case 24:
		pixel_op_func = BlitBufTilePixel24;
		break;
	      case 32:
		pixel_op_func = BlitBufTilePixel32;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufTileArea(
	    bpp,
	    tar_buf, src_buf,
	    offset_x, offset_y,
	    tar_width, tar_height, tar_bpl,
	    src_width, src_height, src_bpl,
	    zoom,
	    magnification,
	    pixel_op_func
	);
}
