#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>

#include "../include/disk.h"
#include "../include/tga.h"

#include "blitting.h"
#include "xsw.h"


static void BridgeSSGetAndBlit(win_t src_w, image_t *tar_img);
void BridgeScreenShot(
	xsw_bridge_win_struct *bw,
	const char *save_dir
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Macro to get image from w and put it on tar_img.
 */
static void BridgeSSGetAndBlit(win_t src_w, image_t *tar_img)
{
	int x, y;
	unsigned int width, height;
	image_t *img;

	if((src_w == None) || (tar_img == NULL))
	    return;

	OSWGetWindowGeometry(
	    src_w, &x, &y, &width, &height, NULL
	);
	img = OSWGetImage(src_w, 0, 0, width, height);
	if(img != NULL)
	{
	    BlitBufAbsolute(
		tar_img->bytes_per_pixel, tar_img->depth,
		(u_int8_t *)tar_img->data,
		(const u_int8_t *)img->data,
		x, y,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,
		0, 0,
		img->width, img->height,
		img->bytes_per_line,
		img->width, img->height,
		1, 1
	    );
	    OSWDestroyImage(&img);
	}
}

/*
 *	Takes a screen shot of the bridge window.
 *
 *	The save_dir is the directory to save the new screen shot
 *	image file to.
 *
 *	Currently the coordinate inputs are not used.
 */
void BridgeScreenShot(
	xsw_bridge_win_struct *bw,
	const char *save_dir
)
{
	int i, status;
	const char *s;
	char *dsave_dir, *path;
	struct stat stat_buf;
	image_t *final_image;
	win_attr_t top_wattr;
	tga_data_struct *td;

	if((bw == NULL) || STRISEMPTY(save_dir))
	    return;

	dsave_dir = STRDUP(save_dir);
	s = PathSubHome(dsave_dir);
	if(s != NULL)
	{
	    char *tmp_dir = STRDUP(s);
	    free(dsave_dir);
	    dsave_dir = tmp_dir;
	}
	if(dsave_dir == NULL)
	    return;

	/* Make sure that the directory to save at exists */
	if(stat(dsave_dir, &stat_buf))
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		char *s = (char *)malloc(
		    STRLEN(dsave_dir) + STRLEN(error_msg) + 80
		);
		*error_msg = toupper(*error_msg);
		sprintf(
		    s,
"%s: %s.",
		    dsave_dir, error_msg
		);
		XSWPrintWarning(s);
		free(s);
	    }
	    free(error_msg);
	    free(dsave_dir);
	    return;
	}
	if(!S_ISDIR(stat_buf.st_mode))
	{
	    char *s = (char *)malloc(STRLEN(dsave_dir) + 80);
	    sprintf(
		s,
"%s: Not a directory.",
		dsave_dir
	    );
	    XSWPrintWarning(s);
	    free(s);
	    free(dsave_dir);
	    return;
	}


	/* Begin generate the file name in which to save to */

	s = PrefixPaths(dsave_dir, "xsw");
	if(s == NULL)
	{
	    free(dsave_dir);
	    return;
	}

	path = (char *)malloc(STRLEN(s) + 40 + 40 + 80);
	for(i = 0; i < 100000; i++)
	{
	    sprintf(
		path,
		"%s%i.tga",
		s, i
	    );
	    /* Check if this file does not exist */
	    if(stat(path, &stat_buf))
		break;
	}
	if(i >= 100000)
	{
	    XSWPrintWarning("Out of tempory names.");
	    free(path);
	    free(dsave_dir);
	    return;
	}


	/* Get attributes of toplevel */
	OSWGetWindowAttributes(bw->toplevel, &top_wattr);
	if((top_wattr.width == 0) ||
	   (top_wattr.height == 0)
	)
	{
	    free(path);
	    free(dsave_dir);
	    return;
	}

	/* Create final image buffer (same size as toplevel) */
	if(OSWCreateImage(&final_image, top_wattr.width, top_wattr.height))
	{
	    XSWPrintWarning(
"Unable to create image buffer for screen shot."
	    );
	    free(path);
	    free(dsave_dir);
	    return;
	}


	/* Get images from windows */
	BridgeSSGetAndBlit(bw->pan_p1, final_image);
	BridgeSSGetAndBlit(bw->pan_p2, final_image);
	BridgeSSGetAndBlit(bw->pan_p3, final_image);
	BridgeSSGetAndBlit(bw->pan_p4, final_image);
	BridgeSSGetAndBlit(bw->viewscreen, final_image);
	BridgeSSGetAndBlit(bw->scanner, final_image);
	BridgeSSGetAndBlit(bw->pan_s1, final_image);
	BridgeSSGetAndBlit(bw->pan_s2, final_image);
	BridgeSSGetAndBlit(bw->pan_s3, final_image);
	BridgeSSGetAndBlit(bw->mesg_box, final_image);



	/* Save to tga image */
	td = (tga_data_struct *)calloc(
	    1, sizeof(tga_data_struct)
	);
	td->fp = NULL;

	td->id_field_len = 0;
	td->cmap_type = TgaColormapNone;
	td->img_type = TgaImageDataURGB;

	td->cmap_first_color_index = 0;
	td->cmap_total_colors = 0;
	td->cmap_entry_size = 24;

	td->x = 0;
	td->y = 0;
	td->width = final_image->width;
	td->height = final_image->height;
	td->depth = 24;
	td->flags = (tga_image_flags)0;

	td->file_size = 0l;
	td->data_size = 0l;

	td->cur_load_pixel = -1;
	td->header_data = NULL;
	td->data = (u_int8_t *)final_image->data;
	td->data_depth = final_image->depth;
	td->data_bits_per_pixel = final_image->bits_per_pixel;

	status = TgaWriteToFile(
	   path,
	   td,
	   24		/* 24 bits on file */
	);
	free(td);
	if(status != TgaSuccess)
	{
	    char *s = (char *)malloc(STRLEN(path) + 80);
	    sprintf(
		s,
"An error occured while saving the screen shot to the file: %s",
		path
	    );
	    XSWPrintError(s);
	    free(s);
	}
	else
	{
	    char *s = (char *)malloc(STRLEN(path) + 80);
	    sprintf(
		s,
"Saved screen shot to file: %s",
		path
	    );
	    XSWPrint(s);
	    free(s);
	}

	OSWDestroyImage(&final_image);
	free(path);
	free(dsave_dir);

	return;
}
