#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/disk.h"
#include "../include/cfgfmt.h"

#include "gctl.h"
#include "keymaps.h"
#include "vsmenu.h"
#include "sound.h"
#include "net.h"
#include "xsw.h"
#include "xswaction.h"
#include "cfgfile.h"


static int CfgSetColor(const char *s, WColorStruct *c);
int CfgFileOpen(const char *path, xsw_option_struct *opt);
int CfgFileOpenVSMenu(
	const char *path,
	xsw_vsmenu_item_struct ***list, int *total_rtn
);

static void CfgPrintColorString(
	FILE *fp, const char *label, const WColorStruct *c
);
int CfgFileSave(
	const char *path,
	xsw_option_struct *opt,
	xsw_vsmenu_struct *vsmenu
);


#define RETURN_MEMORY_ALLOCATION_ERROR(_fp_)	{	\
 char *s = (char *)malloc(				\
  STRLEN(path) + 40 + STRLEN(parm) + 80			\
 );							\
 sprintf(						\
  s,							\
"%s: Line %i: %s: Warning: Memory allocatino error.",	\
  path, lines_read, parm				\
 );							\
 XSWPrintWarning(s);					\
 free(s);						\
							\
 if((_fp_) != NULL)					\
  fclose(_fp_);						\
							\
 return(-3);						\
}

#define WARN_NO_SUCH_PARM(_parm_)		{	\
 char *s = (char *)malloc(				\
  STRLEN(path) + 40 + STRLEN(_parm_) + 80		\
 );							\
 sprintf(						\
  s,							\
"%s: Line %i: Unsupported parameter `%s'.",		\
  path, lines_read, (_parm_)				\
 );							\
 XSWPrintWarning(s);					\
 free(s);						\
}

#define WARN_NO_SUCH_FILE(_path_,_error_code_)	{	\
 const int error_code = (_error_code_);			\
 char *s, *error_msg = STRDUP(strerror(error_code));	\
 if(error_msg == NULL)					\
  error_msg = strdup("no such file");			\
 *error_msg = toupper(*error_msg);			\
							\
 s = (char *)malloc(					\
  STRLEN(path) + 40 + STRLEN(parm) + STRLEN(_path_) +	\
  STRLEN(error_msg) + 80				\
 );							\
 sprintf(						\
  s,							\
"%s: Line %i: %s: Warning: %s: %s.",			\
  path, lines_read, parm, (_path_), error_msg		\
 );							\
 XSWPrintWarning(s);					\
 free(s);						\
							\
 free(error_msg);					\
}

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Macro to parse standard color string and set
 *	values into WColorStruct c.
 */
static int CfgSetColor(const char *s, WColorStruct *c)
{
	if((s == NULL) || (c == NULL))
	    return(-1);

	c->a = 0x00;

	return(
	    StringParseStdColor(
		s,
		&(c->r),
	        &(c->g),
	        &(c->b)
	    )
	);
}

/*
 *	Opens the configuration from the specified file.
 *
 *	This function should only be called once at start up, because
 *	certain resources should not get re-added (ie the universe
 *	items).
 */
int CfgFileOpen(const char *path, xsw_option_struct *opt)
{
	int lines_read = 0;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	unsigned long file_size = 0l;
	struct stat stat_buf;
	xsw_sound_struct *snd = &sound;
	xsw_net_parms_struct *net = &net_parms;

	if(STRISEMPTY(path) || (opt == NULL))
	    return(-2);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for reading");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    XSWPrintWarning(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;
	}

#ifdef JS_SUPPORT
	/* Delete all the joystick maps */
	JSMapsListClear(xsw_jsmap, total_jsmaps);
	xsw_jsmap = NULL;
	total_jsmaps = 0;
#endif	/* JS_SUPPORT */

	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';


	    /* VersionMajor */
	    if(!strcasecmp(parm, "VersionMajor"))
	    {
		opt->cfg_version_major = atoi(val);
	    }
	    /* VersionMinor */
	    else if(!strcasecmp(parm, "VersionMinor"))
	    {
		opt->cfg_version_minor = atoi(val);
	    }
	    /* VersionRelease */
	    else if(!strcasecmp(parm, "VersionRelease"))
	    {
		opt->cfg_version_release = atoi(val);
	    }

	    /* LocalToplevelDir */
	    else if(!strcasecmp(parm, "LocalToplevelDir"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(dname.ltoplevel, val, sizeof(dname.ltoplevel));
		}
		else
		{
		    const char *s = PrefixPaths(dname.home, val);
		    strncpy(
			dname.ltoplevel, (s != NULL) ? s : val,
			sizeof(dname.ltoplevel)
		    );
		}
		dname.ltoplevel[sizeof(dname.ltoplevel) - 1] = '\0';

		/* Path does not exist? */
		if(stat(dname.ltoplevel, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(dname.ltoplevel, errno);
		}
	    }
	    /* ToplevelDir */
#if !defined(__WIN32__)
	    else if(!strcasecmp(parm, "ToplevelDir"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(dname.toplevel, val, sizeof(dname.toplevel));
		}
		else
		{
		    const char *s = PrefixPaths(
			XSWCurrentDirectory(), val
		    );
		    strncpy(
			dname.toplevel, (s != NULL) ? s : val,
			sizeof(dname.toplevel)
		    );
		}
		dname.toplevel[sizeof(dname.toplevel) - 1] = '\0';

		/* Path does not exist? */
		if(stat(dname.toplevel, &stat_buf)) 
		{
		    WARN_NO_SUCH_FILE(dname.toplevel, errno);
		}
	    }
#endif
	    /* ImagesDir */
	    else if(!strcasecmp(parm, "ImagesDir"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(dname.images, val, sizeof(dname.images));
		}
		else
		{
		    const char *s = PrefixPaths(dname.toplevel, val);
		    strncpy(
			dname.images, (s != NULL) ? s : val,
			sizeof(dname.images)
		    );
		}
		dname.images[sizeof(dname.images) - 1] = '\0';

		/* Path does not exist? */
		if(stat(dname.images, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(dname.images, errno);
		}
	    }
	    /* SoundsDir */
	    else if(!strcasecmp(parm, "SoundsDir"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(dname.sounds, val, sizeof(dname.sounds));
		}
		else
		{
		    const char *s = PrefixPaths(dname.toplevel, val);
		    strncpy(
			dname.sounds, (s != NULL) ? s : val,
			sizeof(dname.sounds)
		    );
		}
		dname.sounds[sizeof(dname.sounds) - 1] = '\0';

#if 0
/* Do not check if sounds dir exists or not */
		/* Path does not exist? */
		if(stat(dname.sounds, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(dname.sounds, errno);
		}    
#endif
	    }
	    /* ScriptsDir */
	    else if(!strcasecmp(parm, "ScriptsDir")) 
	    {
		char *s = PathSubHome(val);
		if(s != NULL)   
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(dname.scripts, val, sizeof(dname.scripts));
		}
		else
		{   
		    const char *s = PrefixPaths(dname.home, val);
		    strncpy(
			dname.scripts, (s != NULL) ? s : val,
			sizeof(dname.scripts)
		    );
		}
		dname.scripts[sizeof(dname.scripts) - 1] = '\0';

		/* Path does not exist? */
		if(stat(dname.scripts, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(dname.scripts, errno);
		}
	    }
	    /* DownloadsDir */ 
	    else if(!strcasecmp(parm, "DownloadsDir"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(dname.downloads, val, sizeof(dname.downloads));
		}
		else
		{
		    const char *s = PrefixPaths(dname.home, val);
		    strncpy(
			dname.downloads, (s != NULL) ? s : val,
			sizeof(dname.downloads)
		    );
		}
		dname.downloads[sizeof(dname.downloads) - 1] = '\0';

		/* Path does not exist? */
		if(stat(dname.downloads, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(dname.downloads, errno);
		}
	    }
	    /* StarChartDir */
	    else if(!strcasecmp(parm, "StarChartDir"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(dname.starchart, val, sizeof(dname.starchart));
		}
		else
		{
		    const char *s = PrefixPaths(dname.home, val);
		    strncpy(
			dname.starchart, (s != NULL) ? s : val,
			sizeof(dname.starchart)
		    );
		}
		dname.starchart[sizeof(dname.starchart) - 1] = '\0';

		/* Path does not exist? */
		if(stat(dname.starchart, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(dname.starchart, errno);
		}
	    }
	    /* EtcDir */
#if !defined(__WIN32__)
	    else if(!strcasecmp(parm, "EtcDir"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(dname.etc, val, sizeof(dname.etc));
		}
		else
		{
		    const char *s = PrefixPaths(dname.toplevel, val);
		    strncpy(
			dname.etc, (s != NULL) ? s : val,
			sizeof(dname.etc)
		    );
		}
		dname.etc[sizeof(dname.etc) - 1] = '\0';

		/* Path does not exist? */
		if(stat(dname.etc, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(dname.etc, errno);
		}
	    }
#endif
	    /* NetBandwidth */
	    else if(!strcasecmp(parm, "NetBandwidth") ||
		    !strcasecmp(parm, "NetDeviceMaxLoad")
	    )
	    {
		opt->net_bandwidth = ATOL(val);
	    }
	    /* NetIntervalAutoTune */
	    else if(!strcasecmp(parm, "NetIntervalAutoTune"))
	    {
		auto_interval_tune.state = StringIsYes(val);
	    }
	    /* NetIntervalTuneInterval */
	    else if(!strcasecmp(parm, "NetIntervalTuneInterval"))
	    {
		auto_interval_tune.interval = ATOL(val);
	    }
	    /* DefaultServerAddress */
	    else if(!strcasecmp(parm, "DefaultServerAddress"))
	    {
		if(!(net->flags & XSW_CON_FLAG_ADDRESS_SET))
		{
		    free(net->address);
		    net->address = STRDUP(val);
		}
	    }
	    /* DefaultServerPort */
	    else if(!strcasecmp(parm, "DefaultServerPort"))
	    {
		if(!(net->flags & XSW_CON_FLAG_ADDRESS_SET))
		{
		    net->port = ATOI(val);
		    if(net->port < 0)
			net->port = 0;
		}
	    }
	    /* DefaultLoginName */
	    else if(!strcasecmp(parm, "DefaultLoginName"))
	    {
		if(!(net->flags & XSW_CON_FLAG_ADDRESS_SET))
		{
		    free(net->login_name);
		    net->login_name = STRDUP(val);
		}
	    }
	    /* DefaultLoginPassword */
	    else if(!strcasecmp(parm, "DefaultLoginPassword"))
	    {
		if(!(net->flags & XSW_CON_FLAG_ADDRESS_SET))
		{
		    free(net->login_password);
		    net->login_password = STRDUP(val);
		}
	    }
	    /* ThrottleScopeMode */
	    else if(!strcasecmp(parm, "ThrottleScopeMode"))
	    {
		if(strcasepfx(val, "i"))
		    opt->throttle_mode = XSW_THROTTLE_MODE_INCREMENTAL;
		else if(strcasepfx(val, "b"))
		    opt->throttle_mode = XSW_THROTTLE_MODE_BIDIRECTIONAL;
		else
		    opt->throttle_mode = XSW_THROTTLE_MODE_NORMAL;
	    }
	    /* Local updates */
	    else if(!strcasecmp(parm, "LocalUpdates"))
	    {
		opt->local_updates = StringIsYes(val);
	    }
	    /* AutoMapEconomyWindow */
	    else if(!strcasecmp(parm, "AutoMapEconomyWindow"))
	    {
		opt->auto_map_eco_win = StringIsYes(val);
	    }
	    /* AutoViewscreenZoom */
	    else if(!strcasecmp(parm, "AutoViewscreenZoom"))
	    {
		opt->auto_zoom = StringIsYes(val);
	    }
	    /* ScannerOrientation */
	    else if(!strcasecmp(parm, "ScannerOrientation"))
	    {
		if(strcasepfx(val, "vessel"))
		{
		    opt->def_scanner_orient = XSW_SCANNER_ORIENT_PLAYER;
		}
		else if(strcasepfx(val, "galactic_core"))
		{
		    opt->def_scanner_orient = XSW_SCANNER_ORIENT_GC;
		}
		else
		{
		    char *s = (char *)malloc(
			STRLEN(path) + 40 + STRLEN(parm) +
			STRLEN(val) + (2 * 80)
		    );
		    sprintf(
			s,
"%s: Line %i: %s: Warning: Unsupported value `%s'.\
 Supported values are; galactic_core, vessel",
			path, lines_read, parm, val
		    );
		    XSWPrintWarning(s);
		    free(s);
		}
	    }
	    /* ShowViewscreen */
	    else if(!strcasecmp(parm, "ShowViewscreen"))
	    {
		opt->show_viewscreen_flags = atoi(val);
	    }
	    /* ShowLensFlares */
	    else if(!strcasecmp(parm, "ShowLensFlares"))
	    {
		opt->show_lens_flares = StringIsYes(val);
	    }
	    /* ShowStrobeGlow */
	    else if(!strcasecmp(parm, "ShowStrobeGlow"))
	    {
		opt->show_strobe_glow = StringIsYes(val);
	    }
	    /* ShowNebulaGlow */
	    else if(!strcasecmp(parm, "ShowNebulaGlow"))
	    {
		opt->show_nebula_glow = StringIsYes(val);
	    }
	    /* ShowFires */
	    else if(!strcasecmp(parm, "ShowFires"))
	    {
		opt->show_fires = StringIsYes(val);
	    }
	    /* ShowSmokeTrails */
	    else if(!strcasecmp(parm, "ShowSmokeTrails"))
	    {
		opt->show_smoke_trails = StringIsYes(val);
	    }
	    /* MaxParticles */  
	    else if(!strcasecmp(parm, "MaxParticles"))
	    {
		opt->max_particles = MAX(atoi(val), 0);
	    }
	    /* FormalNameLabels */
	    else if(!strcasecmp(parm, "FormalNameLabels"))
	    {
		opt->show_formal_label = ATOI(val);
	    }
	    /* NotifyScannerContacts */
	    else if(!strcasecmp(parm, "NotifyScannerContacts"))
	    {
		opt->notify_scanner_contacts = StringIsYes(val);
	    }
	    /* NetTimeout */
	    else if(!strcasecmp(parm, "NetTimeout"))
	    {
		opt->net_timeout = ATOL(val);
	    }
	    /* NetObjectOutdatedTime */
	    else if(!strcasecmp(parm, "NetObjectOutdatedTime"))
	    {
		opt->net_object_outdated_time = ATOL(val);
	    }     
	    /* ShowNetErrors */
	    else if(!strcasecmp(parm, "ShowNetErrors"))
	    {
		opt->show_net_errors = StringIsYes(val);
	    }
	    /* ShowServerErrors */
	    else if(!strcasecmp(parm, "ShowServerErrors"))
	    {
		opt->show_server_errors = StringIsYes(val);
	    }
	    /* SoundServerType */
	    else if(!strcasecmp(parm, "SoundServerType"))
	    {
		snd->server_type = MAX(atoi(val), SOUND_SERVER_TYPE_NONE);
	    }
	    /* Sounds */
	    else if(!strcasecmp(parm, "Sounds"))
	    {
		opt->sounds = atoi(val);

		if(opt->sounds < XSW_SOUNDS_NONE)
		    opt->sounds = XSW_SOUNDS_NONE;
	    }
	    /* Music */
	    else if(!strcasecmp(parm, "Music"))
	    {
		opt->music = StringIsYes(val);
	    }
	    /* Controller */
	    else if(!strcasecmp(parm, "Controller") ||
		    !strcasecmp(parm, "ControllerType") ||
		    !strcasecmp(parm, "ControlType")
	    )
	    {
		if(!opt->cmd_line_set_controller)
		{
		    /* Pointer */
		    if(strcasepfx(val, "p"))
		    {
			opt->controller = XSW_CONTROLLER_POINTER;
		    }
		    /* Keyboard */
		    else if(strcasepfx(val, "k"))
		    {
			opt->controller = XSW_CONTROLLER_KEYBOARD;
		    }
#ifdef JS_SUPPORT
		    /* Joystick */
		    else if(strcasepfx(val, "j"))
		    {
		        opt->controller = XSW_CONTROLLER_JOYSTICK;
		    }
#endif	/* JS_SUPPORT */
		    else
		    {
			char *s = (char *)malloc(
			    STRLEN(path) + 40 + STRLEN(parm) +
			    STRLEN(val) + (2 * 80)
			);
		        sprintf(
			    s,
"%s: Line %i: %s: Warning: Unsupported value `%s'.\
 Supported values are; keyboard, joystick, pointer",
			    path, lines_read, parm, val
			);
			XSWPrintWarning(s);
			free(s);
		    }
		    /* Set and initialize the new game controller type */
		    GCtlDelete(xsw_gctl);
		    xsw_gctl = GCtlNew(opt->controller);
		    if(xsw_gctl == NULL)
		    {
                        char *s = (char *)malloc(
                            STRLEN(path) + 40 + STRLEN(parm) + (2 * 80)
                        );
                        sprintf(
                            s,
"%s: Line %i: %s: Warning: Unable to initialize the game controller.",
			    path, lines_read, parm
                        );
                        XSWPrintWarning(s);
                        free(s);
		    }
		}
	    }
	    /* CPU Priority */
	    else if(!strcasecmp(parm, "CPUPriority"))
	    {
		opt->cpu_priority = atoi(val);
	    }
	    /* Units */
	    else if(!strcasecmp(parm, "Units"))
	    {
		if(strcasepfx(val, "XSW"))
		{
		    opt->units = XSW_UNITS_XSW;
		}
		else if(strcasepfx(val, "Metric") ||
			strcasepfx(val, "AstroMetric")
		)
		{
		    opt->units = XSW_UNITS_METRIC;
		}
		else if(strcasepfx(val, "English"))
		{
		    opt->units = XSW_UNITS_ENGLISH;
		}
		else if(strcasepfx(val, "Parsecs"))
		{
		    opt->units = XSW_UNITS_PARSECS;
		}
		else
		{
		    char *s = (char *)malloc(
			STRLEN(path) + 40 + STRLEN(parm) +
			STRLEN(val) + (2 * 80)
		    );
		    sprintf(
			s,
"%s: Line %i: %s: Warning: Unsupported value `%s'.\
 Supported values are; XSW, Metric, English, Parsecs",
			path, lines_read, parm, val
		    );
		    XSWPrintWarning(s);
		    free(s);
		}
	    }
	    /* AsyncImageLoading */
	    else if(!strcasecmp(parm, "AsyncImageLoading"))
	    {
		opt->async_image_loading = StringIsYes(val);
	    }
	    /* AsyncImageLoadPixels */
	    else if(!strcasecmp(parm, "AsyncImageLoadPixels"))
	    {
		opt->async_image_pixels = atoi(val);
	    } 
	    /* SoundServerStartCommand */
	    else if(!strcasecmp(parm, "SoundServerStartCommand"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}
		free(snd->start_cmd);
		snd->start_cmd = STRDUP(val);
	    }

#ifdef JS_SUPPORT
	    /* JSCalibrationFile */
	    else if(!strcasecmp(parm, "JSCalibrationFile") ||
		    !strcasecmp(parm, "JSCaliberationFile")	/* Typo */
	    )
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(!ISPATHABSOLUTE(val))
		{
		    const char *home_dir = getenv("HOME");
		    if(home_dir == NULL)
			home_dir = "/";

		    strncpy(
			fname.js_calib,
			PrefixPaths(home_dir, val),
			sizeof(fname.js_calib)
		    );
		}
		else
		{
		    strncpy(fname.js_calib, val, sizeof(fname.js_calib));
		}
		fname.js_calib[sizeof(fname.js_calib) - 1] = '\0';

		/* Path does not exist? */
		if(stat(fname.js_calib, &stat_buf)) 
		{
		    WARN_NO_SUCH_FILE(fname.js_calib, errno);
		}
	    }
#endif	/* JS_SUPPORT */
	    /* MainPageFile */
	    else if(!strcasecmp(parm, "MainPageFile"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(
			fname.main_page, val,
			sizeof(fname.main_page)
		    );
		}
		else
		{
		    const char *s = PrefixPaths(dname.etc, val);
		    strncpy(
			fname.main_page, (s != NULL) ? s : val,
			sizeof(fname.main_page)
		    );
		}
		fname.main_page[sizeof(fname.main_page) - 1] = '\0';

		/* Path does not exist? */
		if(stat(fname.main_page, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(fname.main_page, errno);
		}
	    }
	    /* DestroyedPageFile */
	    else if(!strcasecmp(parm, "DestroyedPageFile"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(
			fname.destroyed_page, val,
			sizeof(fname.destroyed_page)
		    );
		}
		else
		{
		    const char *s = PrefixPaths(dname.etc, val);
		    strncpy(
			fname.destroyed_page, (s != NULL) ? s : val,
			sizeof(fname.destroyed_page)
		    );
		}
		fname.destroyed_page[sizeof(fname.destroyed_page) - 1] = '\0';

		/* Path does not exist? */
		if(stat(fname.destroyed_page, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(fname.destroyed_page, errno);
		}     
	    }
	    /* OCSNamesFile */
	    else if(!strcasecmp(parm, "OCSNamesFile"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(
			fname.ocsn, val, sizeof(fname.ocsn)
		    );
		}
		else
		{
		    const char *s = PrefixPaths(dname.etc, val);
		    strncpy(
			fname.ocsn, (s != NULL) ? s : val,
			sizeof(fname.ocsn)
		    );
		}
		fname.ocsn[sizeof(fname.ocsn) - 1] = '\0';

		/* Path does not exist? */
		if(stat(fname.ocsn, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(fname.ocsn, errno);
		}
	    }
	    /* ImgRefsFile */
	    else if(!strcasecmp(parm, "ImgRefsFile") ||
		    !strcasecmp(parm, "ISRFile")
	    )
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(fname.imgrefs, val, sizeof(fname.imgrefs));
		}
		else
		{
		    const char *s = PrefixPaths(dname.images, val);
		    strncpy(
			fname.imgrefs, (s != NULL) ? s : val,
			sizeof(fname.imgrefs)
		    );
		}
		fname.imgrefs[sizeof(fname.imgrefs) - 1] = '\0';

		/* Path does not exist? */
		if(stat(fname.imgrefs, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(fname.imgrefs, errno);
		}
	    }
	    /* SndRefsFile */
	    else if(!strcasecmp(parm, "SndRefsFile") ||
		    !strcasecmp(parm, "SoundSchemeFile")
	    )
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(
			fname.sndrefs, val,
			sizeof(fname.sndrefs)
		    );
		}
		else
		{
		    const char *s = PrefixPaths(dname.sounds, val);
		    strncpy(
			fname.sndrefs, (s != NULL) ? s : val,
			sizeof(fname.sndrefs)
		    );
		}
		fname.sndrefs[sizeof(fname.sndrefs) - 1] = '\0';

		/* Do not check for existance */
	    }
	    /* SoundServerConnectionArgument */
	    else if(!strcasecmp(parm, "SoundServerConnectionArgument"))
	    {
		free(snd->con_arg);
		snd->con_arg = STRDUP(val);
	    }
	    /* LogFile */
	    else if(!strcasecmp(parm, "LogFile"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(fname.log, val, sizeof(fname.log));
		}
		else
		{
		    const char *s = PrefixPaths(dname.ltoplevel, val);
		    strncpy(
			fname.log, (s != NULL) ? s : val,
			sizeof(fname.log)
		    );
		}
		fname.log[sizeof(fname.log) - 1] = '\0';
	    }
	    /* UniverseListFile */
	    else if(!strcasecmp(parm, "UniverseListFile"))
	    {
		char *s = PathSubHome(val);
		if(s != NULL)
		{
		    strncpy(val, s, sizeof(val));
		    val[sizeof(val) - 1] = '\0';
		}

		if(ISPATHABSOLUTE(val))
		{
		    strncpy(
			fname.universes, val,
			sizeof(fname.universes)
		    );
		}
		else
		{
		    const char *s = PrefixPaths(dname.ltoplevel, val);
		    strncpy(
			fname.universes, (s != NULL) ? s : val,
			sizeof(fname.universes)
		    );
		}
		fname.universes[sizeof(fname.universes) - 1] = '\0';

#if 0
		/* Path does not exist? */
		if(stat(fname.universes, &stat_buf))
		{
		    WARN_NO_SUCH_FILE(fname.universes, errno);   
		}
#endif
	    }


	    /* *************************************************** */

	    /* ColorStandardText */
	    else if(!strcasecmp(parm, "ColorStandardText"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.standard_text_cv
		);
	    }
	    /* ColorBoldText */
	    else if(!strcasecmp(parm, "ColorBoldText")) 
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bold_text_cv
		);
	    }
	    /* ColorWithdrawnText */
	    else if(!strcasecmp(parm, "ColorWithdrawnText"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.withdrawn_text_cv
		);
	    }


	    /* ColorBPStandardText */
	    else if(!strcasecmp(parm, "ColorBPStandardText"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_standard_text_cv
		);
	    }
	    /* ColorBPBoldText */
	    else if(!strcasecmp(parm, "ColorBPBoldText"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_bold_text_cv
		);
	    }
	    /* ColorBPWithdrawnText */
	    else if(!strcasecmp(parm, "ColorBPWithdrawnText"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_withdrawn_text_cv
		);
	    }
	    /* ColorBPLightOutline */
	    else if(!strcasecmp(parm, "ColorBPLightOutline"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_light_outline_cv
		);
	    }
	    /* ColorBPNormalOutline */
	    else if(!strcasecmp(parm, "ColorBPNormalOutline"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_normal_outline_cv
		);
	    }
	    /* ColorBPDarkOutline */
	    else if(!strcasecmp(parm, "ColorBPDarkOutline"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_dark_outline_cv
		);
	    }


	    /* ColorBPWarning */
	    else if(!strcasecmp(parm, "ColorBPWarning"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_warning_cv
		);
	    }
	    /* ColorBPDanger */
	    else if(!strcasecmp(parm, "ColorBPDanger"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_danger_cv
		);
	    }
	    /* ColorBPCritical */
	    else if(!strcasecmp(parm, "ColorBPCritical"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_critical_cv
		);
	    }

	    /* ColorBPFriendly */
	    else if(!strcasecmp(parm, "ColorBPFriendly"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_friendly_cv
		);
	    }
	    /* ColorBPUnknown */
	    else if(!strcasecmp(parm, "ColorBPUnknown"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_unknown_cv
		);
	    }
	    /* ColorBPHostile */
	    else if(!strcasecmp(parm, "ColorBPHostile"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bp_hostile_cv
		);
	    }

	    /* ColorBPHull */
	    else if(!strcasecmp(parm, "ColorBPHull"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bpol_hull_cv
		);
	    }
	    /* ColorBPPower */
	    else if(!strcasecmp(parm, "ColorBPPower"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bpol_power_cv
		);
	    }
	    /* ColorBPVis */
	    else if(!strcasecmp(parm, "ColorBPVis"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bpol_vis_cv
		);
	    }
	    /* ColorBPShields */
	    else if(!strcasecmp(parm, "ColorBPShields"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bpol_shields_cv
		);
	    }
	    /* ColorBPDmgCtl */
	    else if(!strcasecmp(parm, "ColorBPDmgCtl"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bpol_dmgctl_cv
		);
	    }
	    /* ColorBPThrottle */
	    else if(!strcasecmp(parm, "ColorBPThrottle"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bpol_throttle_cv
		);
	    }
	    /* ColorBPThrottleRev */
	    else if(!strcasecmp(parm, "ColorBPThrottleRev"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.bpol_throttle_rev_cv
		);
	    }

	    /* ColorVSLabelBg */
	    else if(!strcasecmp(parm, "ColorVSLabelBg"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.vs_label_bg_cv
		);
	    }
	    /* ColorVSLabelFg */
	    else if(!strcasecmp(parm, "ColorVSLabelFg"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.vs_label_fg_cv
		);
	    }

	    /* ColorScMarkUnknown */
	    else if(!strcasecmp(parm, "ColorScMarkUnknown"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.scmark_unknown_cv
		);
	    }
	    /* ColorScMarkLocked */
	    else if(!strcasecmp(parm, "ColorScMarkLocked"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.scmark_locked_cv
		);
	    }
	    /* ColorScMarkWeapon */
	    else if(!strcasecmp(parm, "ColorScMarkWeapon"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.scmark_weapon_cv
		);
	    }
	    /* ColorScMarkHome */
	    else if(!strcasecmp(parm, "ColorScMarkHome"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.scmark_home_cv
		);
	    }
	    /* ColorScMarkArea */
	    else if(!strcasecmp(parm, "ColorScMarkArea"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.scmark_area_cv
		);
	    }


	    /* ColorKeymapQueryBg */
	    else if(!strcasecmp(parm, "ColorKeymapQueryBg"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.keymap_query_bg_cv
		);
	    }
	    /* ColorKeymapQueryFg */
	    else if(!strcasecmp(parm, "ColorKeymapQueryFg"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.keymap_query_fg_cv
		);
	    }


	    /* ColorHeadingArrow */
	    else if(!strcasecmp(parm, "ColorHeadingArrow"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.heading_arrow
		);
	    }
	    /* ColorLockArrow */
	    else if(!strcasecmp(parm, "ColorLockArrow"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.lock_arrow
		);
	    }
	    /* ColorLockCursor */
	    else if(!strcasecmp(parm, "ColorLockCursor"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.lock_cursor
		);
	    }
	    /* ColorStreamGreen */
	    else if(!strcasecmp(parm, "ColorStreamGreen"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.stream_green
		);
	    }
	    /* ColorStreamOrange */
	    else if(!strcasecmp(parm, "ColorStreamOrange"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.stream_orange
		);
	    }
	    /* ColorStreamPurple */
	    else if(!strcasecmp(parm, "ColorStreamPurple"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.stream_purple
		);
	    }
	    /* ColorStreamYellow */
	    else if(!strcasecmp(parm, "ColorStreamYellow"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.stream_yellow
		);
	    }
	    /* ColorShieldBlue */
	    else if(!strcasecmp(parm, "ColorShieldBlue"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.shield_blue
		);
	    }
	    /* ColorVisibilityMarker */
	    else if(!strcasecmp(parm, "ColorVisibilityMarker"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.visibility_marker
		);
	    }
	    /* ColorStarGlow */
	    else if(!strcasecmp(parm, "ColorStarGlow"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.star_glow
		);
	    }

	    /* ColorChartBackground */
	    else if(!strcasecmp(parm, "ColorChartBackground"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.chart_bg_cv
		);
	    }
	    /* ColorChartGrid */
	    else if(!strcasecmp(parm, "ColorChartGrid"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.chart_grid_cv
		);
	    }
	    /* ColorChartSectorGrid */
	    else if(!strcasecmp(parm, "ColorChartSectorGrid"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.chart_sector_grid_cv
		);
	    }
	    /* ColorChartCrossHairs */
	    else if(!strcasecmp(parm, "ColorChartCrossHairs"))
	    {
		CfgSetColor(
		    val,
		    &xsw_color.chart_cross_hairs_cv
		);
	    }


	    /* **************************************************** */
	    /* BeginKeyMap */
	    else if(!strcasecmp(parm, "BeginKeyMap"))
	    {
		const int m = KeymapsGetTotal();
		bool_t got_match;
		int i;
		keymap_struct *km;

		/* Begin reading this block */
		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';


		    /* Match the keymap by the parameter */
		    got_match = False;
		    for(i = 0; i < m; i++)
		    {
			km = KeymapGetPtr(i);
			if(km == NULL)
			    continue;

			if(!strcasecmp(parm, km->cfg_name))
			{
			    km->keycode = ATOI(val);
			    got_match = True;
			}
		    }
		    if(got_match)
			continue;

		    /* EndKeyMap */
		    if(!strcasecmp(parm, "EndKeyMap"))
		    {
			break;
		    }
		    /* Unknown parameter */
		    else
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }
	    /* **************************************************** */
	    /* Bridge Window */
	    /* BridgeWinX */
	    else if(!strcasecmp(parm, "BridgeWinX"))
	    {
		opt->bridge_x = atoi(val);
	    }
	    /* BridgeWinY */
	    else if(!strcasecmp(parm, "BridgeWinY"))
	    {
		opt->bridge_y = atoi(val);
	    }
	    /* BridgeWinWidth */
	    else if(!strcasecmp(parm, "BridgeWinWidth"))
	    {
		opt->bridge_width = (unsigned int)MAX(atoi(val), 0);
	    }
	    /* BridgeWinHeight */
	    else if(!strcasecmp(parm, "BridgeWinHeight"))
	    {
		opt->bridge_height = (unsigned int)MAX(atoi(val), 0);
	    }
	    /* **************************************************** */
	    /* Message Window */
	    /* MsgWinX */
	    else if(!strcasecmp(parm, "MsgWinX"))
	    {
		opt->msgw_x = atoi(val);
	    }
	    /* MsgWinY */
	    else if(!strcasecmp(parm, "MsgWinY"))
	    {
		opt->msgw_y = atoi(val);
	    }
	    /* MsgWinWidth */
	    else if(!strcasecmp(parm, "MsgWinWidth"))
	    {
		opt->msgw_width = (unsigned int)MAX(atoi(val), 0);
	    }
	    /* MsgWinHeight */
	    else if(!strcasecmp(parm, "MsgWinHeight"))
	    {
		opt->msgw_height = (unsigned int)MAX(atoi(val), 0);
	    }
	    /* **************************************************** */
	    /* Universe list window position */
	    /* AutoMapUniverseListWindow */
	    else if(!strcasecmp(parm, "AutoMapUniverseListWindow"))
	    {
		opt->auto_map_univ_list_win = StringIsYes(val);
	    }
	    /* UnivListWinX */
	    else if(!strcasecmp(parm, "UnivListWinX"))
	    {
		opt->ulw_x = atoi(val);
	    }
	    /* UnivListWinY */
	    else if(!strcasecmp(parm, "UnivListWinY"))
	    {
		opt->ulw_y = atoi(val);
	    }
	    /* UnivListWinWidth */
	    else if(!strcasecmp(parm, "UnivListWinWidth"))
	    {
		opt->ulw_width = (unsigned int)MAX(atoi(val), 0);
	    }
	    /* UnivListWinHeight */
	    else if(!strcasecmp(parm, "UnivListWinHeight"))
	    {
		opt->ulw_height = (unsigned int)MAX(atoi(val), 0);
	    }

	    /* **************************************************** */
	    /* Economy window position */
	    /* EcoWinX */
	    else if(!strcasecmp(parm, "EcoWinX"))
	    {
		opt->ecow_x = atoi(val);
	    }
	    /* EcoWinY */
	    else if(!strcasecmp(parm, "EcoWinY"))
	    {
		opt->ecow_y = atoi(val);   
	    }
	    /* EcoWinWidth */
	    else if(!strcasecmp(parm, "EcoWinWidth"))
	    {
		opt->ecow_width = (unsigned int)MAX(atoi(val), 0);
	    }
	    /* EcoWinHeight */
	    else if(!strcasecmp(parm, "EcoWinHeight"))
	    {
		opt->ecow_height = (unsigned int)MAX(atoi(val), 0);
	    }

	    /* **************************************************** */
	    /* StarChart position */
	    /* StarChartWinX */
	    else if(!strcasecmp(parm, "StarChartWinX"))
	    {
		opt->scht_x = atoi(val);
	    }
	    /* StarChartWinY */
	    else if(!strcasecmp(parm, "StarChartWinY"))
	    {
		opt->scht_y = atoi(val);
	    }
	    /* StarChartWinWidth */
	    else if(!strcasecmp(parm, "StarChartWinWidth"))
	    {
		opt->scht_width = (unsigned int)MAX(atoi(val), 0);
	    }
	    /* StarChartWinHeight */
	    else if(!strcasecmp(parm, "StarChartWinHeight"))
	    {
		opt->scht_height = (unsigned int)MAX(atoi(val), 0);
	    }
	    /* ******************************************************** */
	    /* BeginVSMenuItem */
	    else if(!strcasecmp(parm, "BeginVSMenuItem"))
	    {
		/* Begin reading this block but do not load */
		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* EndVSMenuItem */
		    if(!strcasecmp(parm, "EndVSMenuItem"))
		    {
			break;
		    }
		}
	    }
#ifdef JS_SUPPORT
	    /* ******************************************************** */
	    /* BeginJSMap */
	    else if(!strcasecmp(parm, "BeginJSMap"))
	    {
		xsw_jsmap_struct *jsm;
		int i = MAX(total_jsmaps, 0);
		total_jsmaps = i + 1;
		xsw_jsmap = (xsw_jsmap_struct **)realloc(
		    xsw_jsmap,
		    total_jsmaps * sizeof(xsw_jsmap_struct *)
		);
		if(xsw_jsmap == NULL)
		{
		    total_jsmaps = 0;
		    free(s);
		    RETURN_MEMORY_ALLOCATION_ERROR(fp);
		}
		xsw_jsmap[i] = jsm = JSMapNew();
		if(jsm == NULL)
		{
		    free(s);
		    RETURN_MEMORY_ALLOCATION_ERROR(fp);
		}

		/* Begin reading this block */
		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* DeviceName */
		    if(!strcasecmp(parm, "DeviceName"))
		    {
			free(jsm->device_name);
			jsm->device_name = STRDUP(val);
		    }
		    /* BeginAxis */
		    else if(!strcasecmp(parm, "BeginAxis"))
		    {
			int n = jsm->total_axises;
			xsw_jsmap_axis_struct *axis;
			jsm->total_axises = n + 1;
			jsm->axis = (xsw_jsmap_axis_struct **)realloc(
			    jsm->axis,
			    jsm->total_axises * sizeof(xsw_jsmap_axis_struct *)
			);
			if(jsm->axis == NULL)
			{
			    jsm->total_axises = 0;
			    free(s);
			    RETURN_MEMORY_ALLOCATION_ERROR(fp);
			}
			jsm->axis[n] = axis = (xsw_jsmap_axis_struct *)calloc(
			    1, sizeof(xsw_jsmap_axis_struct)
			);
			if(axis == NULL)
			{
			    free(s);
			    RETURN_MEMORY_ALLOCATION_ERROR(fp);
			}

			/* Set initial valuse */
			axis->op = JSMAP_AXIS_OP_NONE;

			/* Begin reading this block */
			while(1)
			{
			    /* Get the next line */
			    free(s);
			    s = FReadNextLineAllocCount(
				fp, UNIXCFG_COMMENT_CHAR, &lines_read
			    );
			    if(s == NULL)
				break;

			    /* Get the parameter from the line */
			    s2 = StringCfgParseParm(s);
			    if(s2 == NULL)
				continue;
			    strncpy(parm, s2, sizeof(parm));
			    parm[sizeof(parm) - 1] = '\0';

			    /* Get the value from the line */
			    s2 = StringCfgParseValue(s);
			    if(s2 == NULL)
				s2 = "0";
			    strncpy(val, s2, sizeof(val));
			    val[sizeof(val) - 1] = '\0';

			    /* OpCode */
			    if(!strcasecmp(parm, "OpCode"))
			    {
				axis->op = ATOI(val);
			    }
			    /* EndAxis */
			    else if(!strcasecmp(parm, "EndAxis"))
			    {
				break;
			    }
			    /* Unknown parameter */
			    else
			    {
				WARN_NO_SUCH_PARM(parm);
			    }
			}
		    }
		    /* BeginButton */
		    else if(!strcasecmp(parm, "BeginButton"))
		    {
			int n = jsm->total_buttons;
			xsw_jsmap_button_struct *btn;
			jsm->total_buttons = n + 1;
			jsm->button = (xsw_jsmap_button_struct **)realloc(
			    jsm->button,
			    jsm->total_buttons * sizeof(xsw_jsmap_button_struct *)
			);
			if(jsm->button == NULL)
			{
			    jsm->total_buttons = 0;
			    free(s);
			    RETURN_MEMORY_ALLOCATION_ERROR(fp);
			}
			jsm->button[n] = btn = (xsw_jsmap_button_struct *)calloc(
			    1, sizeof(xsw_jsmap_button_struct)
			);
			if(btn == NULL)
			{
			    free(s);
			    RETURN_MEMORY_ALLOCATION_ERROR(fp);
			}

			/* Set initial values */
			btn->keycode = 0;
			btn->key_state = 0l;
			btn->state = False;

			/* Begin reading this block */
			while(1)
			{
			    /* Get the next line */
			    free(s);
			    s = FReadNextLineAllocCount(
				fp, UNIXCFG_COMMENT_CHAR, &lines_read
			    );
			    if(s == NULL)
				break;

			    /* Get the parameter from the line */
			    s2 = StringCfgParseParm(s);
			    if(s2 == NULL)
				continue;
			    strncpy(parm, s2, sizeof(parm));
			    parm[sizeof(parm) - 1] = '\0';

			    /* Get the value from the line */
			    s2 = StringCfgParseValue(s);
			    if(s2 == NULL)
				s2 = "0";
			    strncpy(val, s2, sizeof(val));
			    val[sizeof(val) - 1] = '\0';


			    /* KeyCode */
			    if(!strcasecmp(parm, "KeyCode"))
			    {
				btn->keycode = atoi(val);
			    }
			    /* KeyState */
			    else if(!strcasecmp(parm, "KeyState"))
			    {
				btn->key_state = ATOI(val);
			    }
			    /* EndButton */
			    else if(!strcasecmp(parm, "EndButton"))
			    {
				break;
			    }
			    /* Unknown parameter */
			    else
			    {
				WARN_NO_SUCH_PARM(parm);
			    }
			}
		    }
		    /* EndJSMap */
		    else if(!strcasecmp(parm, "EndJSMap"))
		    {
			break;
		    }
		    /* Unknown parameter */
		    else
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }
#endif	/* JS_SUPPORT */
	    /* Unknown parameter */
	    else
	    {
		WARN_NO_SUCH_PARM(parm);
	    }
	}
	free(s);

	/* Close the file */
	fclose(fp);

	return(0);
}


/*
 *	Opens the Viewscreen Menu configuration from the specified
 *	file.
 *
 *	This is treated as the same file as the main configuration file.
 */
int CfgFileOpenVSMenu(
	const char *path,
	xsw_vsmenu_item_struct ***list, int *total_rtn
)
{
	int lines_read = 0;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	unsigned long file_size;
	struct stat stat_buf;

	if(STRISEMPTY(path) || (list == NULL) ||
	   (total_rtn == NULL)
	)
	    return(-2);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for reading");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    XSWPrintWarning(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;
	}

	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';

	    /* BeginVSMenuItem */
	    if(!strcasecmp(parm, "BeginVSMenuItem"))
	    {
		xsw_vsmenu_item_struct *item;
		int i = MAX(*total_rtn, 0);
		*total_rtn = i + 1;
		*list = (xsw_vsmenu_item_struct **)realloc(
		    *list,
		    (*total_rtn) * sizeof(xsw_vsmenu_item_struct *)
		);
		if(*list == NULL)
		{
		    *total_rtn = 0;
		    free(s);
		    RETURN_MEMORY_ALLOCATION_ERROR(fp);
		}
		(*list)[i] = item = VSMenuItemNew();
		if(item == NULL)
		{
		    free(s);
		    RETURN_MEMORY_ALLOCATION_ERROR(fp);
		}

		/* Begin reading this block */
		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* Type */
		    if(!strcasecmp(parm, "Type"))
		    {
			item->type = ATOI(val);
		    }
		    /* Text */
		    else if(!strcasecmp(parm, "Text"))
		    {
			free(item->text);
			item->text = STRDUP(val);
		    }
		    /* Action */
		    else if(!strcasecmp(parm, "Action"))
		    {
			item->action = ATOI(val);
		    }
		    /* EndVSMenuItem */
		    else if(!strcasecmp(parm, "EndVSMenuItem"))
		    {
			break;
		    }
		    /* Unknown parameter */
		    else
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }
	    /* Unknown parameter */
	    else
	    {
		/* Ignore and do not warn */
	    }
	}
	free(s);

	/* Close the file */
	fclose(fp);

	return(0);
}


/*
 *	Writes a `parameter = value' string.
 */
static void CfgPrintColorString(
	FILE *fp, const char *label, const WColorStruct *c
)
{
	if((fp == NULL) || (label == NULL) || (c == NULL))
	    return;

	fprintf(
	    fp,
"%s = #%.2x%.2x%.2x\n",
	    label,
	    c->r,
	    c->g,
	    c->b
	);
}

/*
 *	Saves the configuration to the specified file.
 */
int CfgFileSave(
	const char *path,
	xsw_option_struct *opt,
	xsw_vsmenu_struct *vsmenu
)
{
	const int nkeymaps = KeymapsGetTotal();
	int i, n;
	FILE *fp;
	keymap_struct *km;
	xsw_sound_struct *snd = &sound;
	xsw_net_parms_struct *net = &net_parms;

	if(STRISEMPTY(path) || (opt == NULL))
	    return(-2);

	/* Open the file for writing */
	fp = fopen(path, "wb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for writing");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    XSWPrintError(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}


	/* Header */
	fprintf(
	    fp,
"# %s Version %s Configuration File\n\
#\n",
	    PROG_NAME_FULL,
	    PROG_VERSION
	);

	/* Version */
	fprintf(
	    fp,
"# Configuration Version:\n\
#\n"
	);
	fprintf(
	    fp,
	    "VersionMajor = %i\n",
	    PROG_VERSION_MAJOR
	);
	fprintf(
	    fp,
	    "VersionMinor = %i\n",
	    PROG_VERSION_MINOR
	);
	fprintf(
	    fp,
	    "VersionRelease = %i\n",
	    PROG_VERSION_RELEASE
	);

	fprintf(fp, "\n\n");


	/* Directories & Files */
	fprintf(
	    fp,
"# Directories and files:\n\
#\n"
	);

	/* LocalToplevelDir */
	if(!STRISEMPTY(dname.ltoplevel))
	{
	    char *path = STRDUP(dname.ltoplevel);
	    StripParentPath(path, dname.home);
	    fprintf(
		fp,
		"LocalToplevelDir = %s\n",
		path
	    );
	    free(path);
	}

	/* ToplevelDir */
	if(ISPATHABSOLUTE(dname.toplevel))
	{
	    fprintf(fp, "ToplevelDir = %s\n", dname.toplevel);
	}
	else
	{
	    /* ToplevelDir must be absolute, use default if not */
	    fprintf(fp, "ToplevelDir = %s\n", XSW_DEF_GLOBAL_DATA_DIR);
	}

	/* EtcDir */
	if(!STRISEMPTY(dname.etc))
	{
	    char *path = STRDUP(dname.etc);
	    StripParentPath(path, dname.toplevel);
	    fprintf(
		fp,
		"EtcDir = %s\n",
		path
	    );
	    free(path);
	}

	/* ImagesDir */
	if(!STRISEMPTY(dname.images))
	{
	    char *path = STRDUP(dname.images);
	    StripParentPath(path, dname.toplevel);
	    fprintf(
		fp,
		"ImagesDir = %s\n",
		path
	    );
	    free(path);
	}

	/* SoundsDir */
	if(!STRISEMPTY(dname.sounds))
	{
	    char *path = STRDUP(dname.sounds);
	    StripParentPath(path, dname.toplevel);
	    fprintf(
		fp,
		"SoundsDir = %s\n",
		path
	    );
	    free(path);
	}

	/* ScriptsDir */
	if(!STRISEMPTY(dname.scripts))
	{
	    char *path = STRDUP(dname.scripts);
	    StripParentPath(path, dname.home);
	    fprintf(
		fp,
		"ScriptsDir = %s\n",
		path
	    );
	    free(path);
	}

	/* DownloadsDir */
	if(!STRISEMPTY(dname.downloads))
	{
	    char *path = STRDUP(dname.downloads);
	    StripParentPath(path, dname.home);
	    fprintf(
		fp,
		"DownloadsDir = %s\n",
		path
	    );
	    free(path);
	}

	/* StarChartDir */
	if(!STRISEMPTY(dname.starchart))
	{
	    char *path = STRDUP(dname.starchart);
	    StripParentPath(path, dname.home);
	    fprintf(
		fp,
		"StarChartDir = %s\n",
		path
	    );
	    free(path);
	}

	/* SoundServerStartCommand */
	if(!STRISEMPTY(snd->start_cmd))
	    fprintf(
		fp,
		"SoundServerStartCommand = %s\n",
		snd->start_cmd
	    );

#ifdef JS_SUPPORT
	/* Joystick calibration file */
	fprintf(
	    fp,
	    "JSCalibrationFile = %s\n",
	    fname.js_calib
	);
#endif /* JS_SUPPORT */

	/* MainPageFile */
	if(!STRISEMPTY(fname.main_page))
	{
	    char *path = STRDUP(fname.main_page);
	    StripParentPath(path, dname.etc);
	    fprintf(
		fp,
		"MainPageFile = %s\n",
		path
	    );
	    free(path);
	}

	/* DestroyedPageFile */
	if(!STRISEMPTY(fname.destroyed_page))
	{
	    char *path = STRDUP(fname.destroyed_page);
	    StripParentPath(path, dname.etc);
	    fprintf(
		fp,
		"DestroyedPageFile = %s\n",
		path
	    );
	    free(path);
	}

	/* OCSNamesFile */
	if(!STRISEMPTY(fname.ocsn))
	{
	    char *path = STRDUP(fname.ocsn);
	    StripParentPath(path, dname.etc);
	    fprintf(
		fp,
		"OCSNamesFile = %s\n",
		path
	    );
	    free(path);
	}

	/* ImgRefsFile */
	if(!STRISEMPTY(fname.imgrefs))
	{
	    char *path = STRDUP(fname.imgrefs);
	    StripParentPath(path, dname.images);
	    fprintf(
		fp,
		"ImgRefsFile = %s\n",
		path
	    );
	    free(path);
	}

	/* SndRefsFile */
	if(!STRISEMPTY(fname.sndrefs))
	{
	    char *path = STRDUP(fname.sndrefs);
	    StripParentPath(path, dname.sounds);
	    fprintf(
		fp,
		"SndRefsFile = %s\n",
		path
	    );
	    free(path);
	}

	/* LogFile */
	if(!STRISEMPTY(fname.log))
	{
	    char *path = STRDUP(fname.log);
	    StripParentPath(path, dname.ltoplevel);
	    fprintf(
		fp,
		"LogFile = %s\n",
		path
	    );
	    free(path);
	}

	/* UniverseListFile */
	if(!STRISEMPTY(fname.universes))
	{
	    char *path = STRDUP(fname.universes);
	    StripParentPath(path, dname.ltoplevel);
	    fprintf(
		fp,
		"UniverseListFile = %s\n",
		path
	    );
	    free(path);
	}

	fprintf(fp, "\n\n");


	/* Network */
	fprintf(
	    fp,
	    "# Network\n\n"
	);

	/* NetBandwidth */
	fprintf(
	    fp,
	    "NetBandwidth = %ld\n",
	    opt->net_bandwidth
	);

	/* NetIntervalAutoTune */
	fprintf(fp,
	    "NetIntervalAutoTune = %s\n",
	    auto_interval_tune.state ? "yes" : "no"
	);

	/* NetIntervalTuneInterval */
	fprintf(fp,
	    "NetIntervalTuneInterval = %i\n",
	    (int)auto_interval_tune.interval
	);

	/* DefaultServerAddress */
	if(!STRISEMPTY(net->address))
	    fprintf(
		fp,
		"DefaultServerAddress = %s\n",
		net->address
	    );

	/* DefaultServerPort */
	if(net->port > 0)
	    fprintf(
		fp,
		"DefaultServerPort = %i\n",
		net->port
	    );

	/* DefaultLoginName */
	if(!STRISEMPTY(net->login_name))
	    fprintf(
		fp,
		"DefaultLoginName = %s\n",
		net->login_name
	    );

	/* DefaultLoginPassword */
	if(!STRISEMPTY(net->login_password))
	    fprintf(
		fp,
		"DefaultLoginPassword = %s\n",
		net->login_password
	    );

	/* NetTimeout */
	fprintf(
	    fp,
	    "NetTimeout = %ld\n",
	    opt->net_timeout
	);

	/* NetObjectOutdatedTime */
	fprintf(
	    fp,
	    "NetObjectOutdatedTime = %ld\n",
	    opt->net_object_outdated_time
	);

	fprintf(fp, "\n\n");


	/* ********************************************************* */
	/* General */
	fprintf(fp,
	    "# General\n\n"
	);

	/* AutoMapEconomyWindow */
	fprintf(fp,
	    "AutoMapEconomyWindow = %s\n",
	    opt->auto_map_eco_win ? "yes" : "no"
	);

	/* AutoViewscreenZoom */
	fprintf(fp,
	    "AutoViewscreenZoom = %s\n",
	    opt->auto_zoom ? "on" : "off"
	);

	/* ScannerOrientation */
	if(xsw_bridge_win != NULL)
	{
	    switch(xsw_bridge_win->scanner_orient)
	    {
	      case XSW_SCANNER_ORIENT_PLAYER:
		fprintf(
		    fp,
		    "ScannerOrientation = vessel\n"
		);
		break;
	      default:	/* XSW_SCANNER_ORIENT_GC */
		fprintf(
		    fp,
		    "ScannerOrientation = galactic_core\n"
		);
		break;
	    }
	}

	/* ThrottleScopeMode */
	switch(opt->throttle_mode)
	{
	  case XSW_THROTTLE_MODE_INCREMENTAL:
	    fprintf(fp, "ThrottleScopeMode = incremental\n");
	    break;

	  case XSW_THROTTLE_MODE_BIDIRECTIONAL:
	    fprintf(fp, "ThrottleScopeMode = bidirectional\n");
	    break;

	  default:
	    fprintf(fp, "ThrottleScopeMode = normal\n");
	    break;
	}

	/* LocalUpdates */
	fprintf(fp,
	    "LocalUpdates = %s\n",
	    (opt->local_updates) ? "on" : "off"
	);

	/* ShowViewscreen */
	fprintf(
	    fp,
	    "ShowViewscreen = %i\n",
	    opt->show_viewscreen_flags
	);

	/* ShowLensFlares */
	fprintf(
	    fp,
	    "ShowLensFlares = %s\n",
	    (opt->show_lens_flares) ? "yes" : "no"
	);

	/* ShowStrobeGlow */
	fprintf(
	    fp,
	    "ShowStrobeGlow = %s\n",
	    (opt->show_strobe_glow) ? "yes" : "no"
	);

	/* ShowNebulaGlow */
	fprintf(
	    fp,
	    "ShowNebulaGlow = %s\n",
	    (opt->show_nebula_glow) ? "yes" : "no"
	);

	/* FormalNameLabels */
	fprintf(
	    fp,
	    "FormalNameLabels = %i\n",
	    opt->show_formal_label
	);

	/* NotifyScannerContacts */
	fprintf(
	    fp,
	    "NotifyScannerContacts = %s\n",
	    (opt->notify_scanner_contacts) ? "on" : "off"
	);

	/* ShowNetErrors */
	fprintf(fp,
	    "ShowNetErrors = %s\n",
	    (opt->show_net_errors) ? "yes" : "no"
	);

	/* ShowServerErrors */
	fprintf(fp,
	    "ShowServerErrors = %s\n",
	    (opt->show_server_errors) ? "yes" : "no"
	);

	/* SoundServerType */
	fprintf(
	    fp,
	    "SoundServerType = %i\n",
	    snd->server_type
	); 

	/* Sounds */
	fprintf(
	    fp,
	    "Sounds = %i\n",
	    opt->sounds
	);

	/* Music */
	fprintf(
	    fp,
	    "Music = %s\n",
	    (opt->music) ? "on" : "off"
	);

	/* SoundServerArgument */
	if(snd->con_arg != NULL)
	    fprintf(
		fp,
		"SoundServerConnectionArgument = %s\n",
		snd->con_arg
	    );

	/* CPU Priority */
	fprintf(
	    fp, 
"# CPU Priority:\n\
#\n\
#\t0 = Highest\n\
#\t1 = High\n\
#\t2 = Medium\n\
#\t3 = Lowest\n\
#\n\
CPUPriority = %i\n",
	    opt->cpu_priority
	);

	/* Units */
	switch(opt->units)
	{
	  case XSW_UNITS_METRIC:
	    fputs("Units = Metric\n", fp);
	    break;
	  case XSW_UNITS_ENGLISH:
	    fputs("Units = English\n", fp);
	    break;
	  case XSW_UNITS_PARSECS:
	    fputs("Units = Parsecs\n", fp);
	    break;
	  default:	/* XSW_UNITS_XSW */
	    fputs("Units = XSW\n", fp);
	    break;
	}

	/* AsyncImageLoading */
	fprintf(
	    fp,
	    "AsyncImageLoading = %s\n",
	    opt->async_image_loading ? "yes" : "no"
	);

	/* AsyncImageLoadPixels */
	fprintf(
	    fp,
	    "AsyncImageLoadPixels = %i\n",
	    opt->async_image_pixels
	);


	fprintf(fp, "\n\n");


	/* ********************************************************* */
	/* Bridge window position */

	fprintf(
	    fp,
	    "BridgeWinX = %i\n",
	    opt->bridge_x
	);
	fprintf(
	    fp,
	    "BridgeWinY = %i\n",
	    opt->bridge_y
	);
	fprintf(
	    fp,
	    "BridgeWinWidth = %i\n",
	    opt->bridge_width
	);
	fprintf(
	    fp,
	    "BridgeWinHeight = %i\n",
	    opt->bridge_height
	);

	fprintf(fp, "\n\n");


	/* ********************************************************* */
	/* Messages Window */

	fprintf(
	    fp,
	    "MsgWinX = %i\n",
	    opt->msgw_x
	);
	fprintf(
	    fp,
	    "MsgWinY = %i\n",
	    opt->msgw_y
	);
	fprintf(
	    fp,
	    "MsgWinWidth = %i\n",
	    opt->msgw_width
	);
	fprintf(
	    fp,
	    "MsgWinHeight = %i\n",
	    opt->msgw_height
	);

	fprintf(fp, "\n\n");


	/* ********************************************************* */
	/* Universes List */

	fprintf(
	    fp,
	    "AutoMapUniverseListWindow = %s\n",
	    opt->auto_map_univ_list_win ? "yes" : "no"
	);
	fprintf(
	    fp,
	    "UnivListWinX = %i\n",
	    opt->ulw_x
	);
	fprintf(
	    fp,
	    "UnivListWinY = %i\n",
	    opt->ulw_y
	);
	fprintf(
	    fp,
	    "UnivListWinWidth = %i\n",
	    opt->ulw_width
	);
	fprintf(
	    fp,
	    "UnivListWinHeight = %i\n",
	    opt->ulw_height
	);
	fprintf(fp, "\n\n");


	/* ******************************************************** */
	/* Economy Window */
	fprintf(
	    fp,
	    "EcoWinX = %i\n",
	    opt->ecow_x
	);
	fprintf(
	    fp,
	    "EcoWinY = %i\n",
	    opt->ecow_y
	);
	fprintf(
	    fp,
	    "EcoWinWidth = %i\n",
	    opt->ecow_width
	);
	fprintf(
	    fp,
	    "EcoWinHeight = %i\n",
	    opt->ecow_height
	);

	fprintf(fp, "\n\n");

	/* ******************************************************** */
	/* StarChart */
	fprintf(
	    fp,
	    "StarChartWinX = %i\n",
	    opt->scht_x
	);
	fprintf(
	    fp,
	    "StarChartWinY = %i\n",
	    opt->scht_y
	);
	fprintf(
	    fp,
	    "StarChartWinWidth = %i\n",
	    opt->scht_width
	);
	fprintf(
	    fp,
	    "StarChartWinHeight = %i\n",
	    opt->scht_height
	);
	fprintf(
	    fp,
	    "\n\n"
	);

	/* ********************************************************* */
	/* Viewscreen Menu Items */
	if(vsmenu != NULL)
	{
	    const int total_items = vsmenu->total_items;
	    xsw_vsmenu_item_struct **items_list = vsmenu->item;
	    xsw_vsmenu_item_struct *item;

	    fprintf(
		fp,
"# Viewscreen Menu Items:\n\n"
	    );

	    for(i = 0; i < total_items; i++)
	    {
		item = items_list[i];
		if(item == NULL)
		    continue;

		fprintf(
		    fp,
		    "BeginVSMenuItem\n"
		);
		fprintf(
		    fp,
		    "\tType = %i\n",
		    item->type
		);
		if(!STRISEMPTY(item->text))
		    fprintf(
			fp,
			"\tText = %s\n",
			item->text
		    );
		if(item->action != XSW_ACTION_NONE)
		    fprintf(
			fp,
			"\tAction = %i\n",
			item->action
		    );
		fprintf(
		    fp,
		    "EndVSMenuItem\n\n"
		);
	    }

	    fprintf(fp, "\n");
	}


	/* Key Mappings */
	if(nkeymaps > 0)
	{
	    fprintf(
		fp,
"# Key Mappings:\n"
	    );

	    /* BeginKeyMap */
	    fprintf(fp, "BeginKeyMap\n");

	    for(i = 0; i < nkeymaps; i++)
	    {
		km = KeymapGetPtr(i);
		if(km == NULL)
		    continue;

		fprintf(
		    fp,
		    "\t%s = %i\n",
		    km->cfg_name, km->keycode
		);
	    }

	    /* EndKeyMap */
	    fprintf(fp, "EndKeyMap\n\n\n");
	}


#ifdef JS_SUPPORT
	/* Joystick mappings */
	if(total_jsmaps > 0)
	{
	    xsw_jsmap_axis_struct *axis;
	    xsw_jsmap_button_struct *btn;
	    xsw_jsmap_struct *jsm;

	    fprintf(
		fp,
		"# Joystick mappings:\n\n"
	    );

	    /* Iterate through jsmaps list */
	    for(i = 0; i < total_jsmaps; i++)
	    {
		jsm = xsw_jsmap[i];
		if(jsm == NULL)
		    continue;

		/* BeginJSMap */
		fprintf(fp, "# Joystick %i\n", i);
		fprintf(fp, "BeginJSMap\n");

		/* DeviceName */
		if(jsm->device_name != NULL)
		    fprintf(
		        fp,
		        "\tDeviceName = %s\n",
		        jsm->device_name
		    );

		/* Axises */
		for(n = 0; n < jsm->total_axises; n++) 
		{
		    axis = jsm->axis[n];
		    if(axis == NULL)
			continue;

		    /* BeginAxis */
		    fprintf(fp, "#\tAxis %i.\n", n);
		    fprintf(fp, "\tBeginAxis\n");

		    /* OpCode */
		    fprintf(
			fp,
			"\t\tOpCode = %i\n",
			axis->op
		    );

		    fprintf(fp, "\tEndAxis\n");
		}

		/* Buttons */ 
		for(n = 0; n < jsm->total_buttons; n++)
		{
		    btn = jsm->button[n];
		    if(btn == NULL)
			continue;

		    /* BeginButton */
		    fprintf(fp, "#\tButton %i.\n", n);
		    fprintf(fp, "\tBeginButton\n");

		    /* KeyCode */
		    fprintf(
			fp,
			"\t\tKeyCode = %i\n",
			btn->keycode
		    );
		    /* KeyState */
		    fprintf(
			fp, 
			"\t\tKeyState = %ld\n",
			btn->key_state
		    );

		    /* EndButton */
		    fprintf(fp, "\tEndButton\n");
		}

		/* EndJSMap */
		fprintf(fp, "EndJSMap\n\n");
	    }
	}

	fprintf(fp, "\n\n");
#endif  /* JS_SUPPORT */


	/* ControlType (must be specified after joystick mappings) */
	fprintf(
	    fp,
"# Controller type (should be specified after keys and joystick mappings\n\n"
	);
	switch(opt->controller)
	{
#ifdef JS_SUPPORT
	  case XSW_CONTROLLER_JOYSTICK:
	    fprintf(fp, "ControlType = joystick\n");
	    break;
#endif /* JS_SUPPORT */ 

	  case XSW_CONTROLLER_POINTER:
	    fprintf(fp, "ControlType = pointer\n");
	    break;
	 
	  default:
	    fprintf(fp, "ControlType = keyboard\n");
	    break;
	}
	fprintf(fp, "\n\n");


	/* ********************************************************* */
	/* Colors */

	/* Colors: General text */

	CfgPrintColorString(
	    fp, "ColorStandardText", &xsw_color.standard_text_cv
	);
	CfgPrintColorString(
	    fp, "ColorBoldText", &xsw_color.bold_text_cv
	);
	CfgPrintColorString(
	    fp, "ColorWithdrawnText", &xsw_color.withdrawn_text_cv
	);

	fprintf(fp, "\n");


	/* Colors: Bridge panel */
	fprintf(fp, "# Colors: Bridge Panel\n");


	CfgPrintColorString(
	    fp, "ColorBPStandardText", &xsw_color.bp_standard_text_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPBoldText", &xsw_color.bp_bold_text_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPWithdrawnText", &xsw_color.bp_withdrawn_text_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPLightOutline", &xsw_color.bp_light_outline_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPNormalOutline", &xsw_color.bp_normal_outline_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPDarkOutline", &xsw_color.bp_dark_outline_cv
	);

	fprintf(fp, "\n");

	CfgPrintColorString(
	    fp, "ColorBPWarning", &xsw_color.bp_warning_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPDanger", &xsw_color.bp_danger_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPCritical", &xsw_color.bp_critical_cv
	);

	fprintf(fp, "\n");

	CfgPrintColorString(
	    fp, "ColorBPFriendly", &xsw_color.bp_friendly_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPUnknown", &xsw_color.bp_unknown_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPHostile", &xsw_color.bp_hostile_cv
	);

	fprintf(fp, "\n");

	/* Colors: Bridge panel outline */
	fprintf(fp, "# Colors: Bridge Panel Outline\n");

	CfgPrintColorString(
	    fp, "ColorBPHull", &xsw_color.bpol_hull_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPPower", &xsw_color.bpol_power_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPVis", &xsw_color.bpol_vis_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPShields", &xsw_color.bpol_shields_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPDmgCtl", &xsw_color.bpol_dmgctl_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPThrottle", &xsw_color.bpol_throttle_cv
	);
	CfgPrintColorString(
	    fp, "ColorBPThrottleRev", &xsw_color.bpol_throttle_rev_cv
	);

	fprintf(fp, "\n");


	/* Colors: Viewscreen labels */
	fprintf(fp, "# Colors: Viewscreen Labels\n");

	CfgPrintColorString(
	    fp, "ColorVSLabelBg", &xsw_color.vs_label_bg_cv
	);

	CfgPrintColorString(
	    fp, "ColorVSLabelFg", &xsw_color.vs_label_fg_cv
	);

	fprintf(fp, "\n");


	/* Colors: Scanner marks */
	fprintf(fp, "# Colors: Scanner Markings\n");

	CfgPrintColorString(
	    fp, "ColorScMarkUnknown", &xsw_color.scmark_unknown_cv
	);

	CfgPrintColorString(
	    fp, "ColorScMarkLocked", &xsw_color.scmark_locked_cv
	);

	CfgPrintColorString(
	    fp, "ColorScMarkWeapon", &xsw_color.scmark_weapon_cv
	);

	CfgPrintColorString(
	    fp, "ColorScMarkHome", &xsw_color.scmark_home_cv
	);

	CfgPrintColorString(
	    fp, "ColorScMarkArea", &xsw_color.scmark_area_cv
	);

	fprintf(fp, "\n");


	/* Colors: Keymap query dialog */
	fprintf(fp, "# Colors: Keymap Query Dialog\n");

	CfgPrintColorString(
	    fp, "ColorKeymapQueryBg", &xsw_color.keymap_query_bg_cv
	);
	CfgPrintColorString(
	    fp, "ColorKeymapQueryFg", &xsw_color.keymap_query_fg_cv
	);

	fprintf(fp, "\n");


	/* Colors: Viewscreen graphics */
	fprintf(fp, "# Colors: Viewscreen Graphics\n");
	CfgPrintColorString(
	    fp, "ColorHeadingArrow", &xsw_color.heading_arrow
	);
	CfgPrintColorString(
	    fp, "ColorLockArrow", &xsw_color.lock_arrow
	);
	CfgPrintColorString(
	    fp, "ColorLockCursor", &xsw_color.lock_cursor
	);
	CfgPrintColorString(
	    fp, "ColorStreamGreen", &xsw_color.stream_green
	);
	CfgPrintColorString(
	    fp, "ColorStreamOrange", &xsw_color.stream_orange
	);
	CfgPrintColorString(
	    fp, "ColorStreamPurple", &xsw_color.stream_purple
	);
	CfgPrintColorString(
	    fp, "ColorStreamYellow", &xsw_color.stream_yellow
	);
	CfgPrintColorString(
	    fp, "ColorShieldBlue", &xsw_color.shield_blue
	);
	CfgPrintColorString(
	    fp, "ColorVisibilityMarker", &xsw_color.visibility_marker
	);
	CfgPrintColorString(
	    fp, "ColorStarGlow", &xsw_color.star_glow
	);
	fprintf(fp, "\n");

	/* Colors: Starchart */
	fprintf(fp, "# Colors: Starchart\n");
	CfgPrintColorString(
	    fp, "ColorChartBackground", &xsw_color.chart_bg_cv
	);
	CfgPrintColorString(
	    fp, "ColorChartGrid", &xsw_color.chart_grid_cv
	);
	CfgPrintColorString(
	    fp, "ColorChartSectorGrid", &xsw_color.chart_sector_grid_cv
	);
	CfgPrintColorString(
	    fp, "ColorChartCrossHairs", &xsw_color.chart_cross_hairs_cv
	);

	fprintf(fp, "\n\n");


	/* Close the file */
	fclose(fp);


	return(0);
}
