#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fnmatch.h>

#include "../include/string.h"

#include "gctl.h"
#include "sound.h"
#include "sndrefs.h"
#include "net.h"
#include "xsw.h"


static int CMDSETPARMMATCH(const char *s, const char *pattern);
int CmdSet(const char *arg);


#define THIS_CMD_NAME		"set"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Macro to perform parameter name matching.
 */
static int CMDSETPARMMATCH(const char *s, const char *pattern)
{
	return(!fnmatch(s, pattern, 0));
}


int CmdSet(const char *arg)
{
	const char *s, *s_end, *this_parm;
	char *parm = NULL, *val = NULL;
	xsw_option_struct *opt = &option;

	/* Parse argument */
	if(arg == NULL)
	    arg = "*";

	/* Get the parameter */
	s = arg;
	s_end = (s != NULL) ? strchr(s, '=') : NULL;
	if(s_end != NULL)
	{
	    int len = s_end - s;
	    if(len > 0)
	    {
		parm = (char *)malloc((len + 1) * sizeof(char));
		memcpy(parm, s, len);
		parm[len] = '\0';
	    }

	    s = s_end + 1;
	}
	else
	{
	    if(STRISEMPTY(s))
		parm = STRDUP("*");
	    else
		parm = STRDUP(s);
	    s = NULL;
	}
	if(parm == NULL)
	    parm = strdup("");
	strstrip(parm);

	/* Get the value */
	val = STRDUP(s);
	if(val == NULL)
	    val = strdup("");
	strstrip(val);


	/* Auto Interval */
	this_parm = "auto_interval";
	if(CMDSETPARMMATCH(parm, this_parm) ||
	   CMDSETPARMMATCH(parm, "aint")
	)
	{
	    char *s;
	    if(!STRISEMPTY(val))
	        auto_interval_tune.state = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		auto_interval_tune.state ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Auto Map Economy Window */
	this_parm = "auto_map_eco_win";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->auto_map_eco_win = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->auto_map_eco_win ? "yes" : "no"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Auto Zoom */
	this_parm = "auto_zoom";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->auto_zoom = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->auto_zoom ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Controller */
	this_parm = "controller";
	if(CMDSETPARMMATCH(parm, this_parm) ||
	   CMDSETPARMMATCH(parm, "control_type") ||
	   CMDSETPARMMATCH(parm, "control")
	)
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
#if 0
/* Pointer support not available yet */
		if(strcasepfx(val, "p"))
		    opt->controller = XSW_CONTROLLER_POINTER
#endif
		if(strcasepfx(val, "k"))
		    opt->controller = XSW_CONTROLLER_KEYBOARD;
#ifdef JS_SUPPORT
		else if(strcasepfx(val, "j"))
		    opt->controller = XSW_CONTROLLER_JOYSTICK;
#endif	/* JS_SUPPORT */
		else
#ifdef JS_SUPPORT
		    XSWPrint(
"Available control_type values: keyboard  joystick"
		    );
#else
		    XSWPrint(
"Available control_type values: keyboard"
		    );
#endif	/* JS_SUPPORT */

		/* Reinitialize controller */
		GCtlDelete(xsw_gctl);
		xsw_gctl = GCtlNew(opt->controller);
		if(xsw_gctl == NULL)
		{
		    XSWPrintWarning(
"An error occured while initializing the game controller."
		    );
		}
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    switch(opt->controller)
	    {
	      case XSW_CONTROLLER_POINTER:
		sprintf(s, "%s = pointer", this_parm);
		break;
	      case XSW_CONTROLLER_KEYBOARD:
		sprintf(s, "%s = keyboard", this_parm);
		break;
#ifdef JS_SUPPORT
	      case XSW_CONTROLLER_JOYSTICK:
		sprintf(s, "%s = joystick", this_parm);
		break;
#endif
	       default:
		sprintf(s, "%s = *badvalue*", this_parm);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	}

	/* CPU Priority */
	this_parm = "cpu_priority";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		if(isdigit(*val))
		{
		    opt->cpu_priority = atoi(val);
		}
		else
		{
		    if(!strcasecmp(val, "highest"))
			opt->cpu_priority = XSW_CPU_PRIORITY_HIGHEST;
		    else if(strcasepfx(val, "h"))
			opt->cpu_priority = XSW_CPU_PRIORITY_HIGH;
		    else if(strcasepfx(val, "m"))
			opt->cpu_priority = XSW_CPU_PRIORITY_MEDIUM;
		    else
			opt->cpu_priority = XSW_CPU_PRIORITY_LOWEST;
		}
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    switch(opt->cpu_priority)
	    {
	      case XSW_CPU_PRIORITY_HIGHEST:
		sprintf(s, "%s = highest", this_parm);
		break;
	      case XSW_CPU_PRIORITY_HIGH:
		sprintf(s, "%s = high", this_parm);
		break;
	      case XSW_CPU_PRIORITY_MEDIUM:
		sprintf(s, "%s = medium", this_parm);
		break;
	      default:		/* XSW_CPU_PRIORITY_LOWEST */
		sprintf(s, "%s = lowest", this_parm);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	}

	/* Energy Saver Mode */
	this_parm = "energy_saver_mode";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->energy_saver_mode = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->energy_saver_mode ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Local Updates */
	this_parm = "local_updates";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->local_updates = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->local_updates ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Log Client */
	this_parm = "log_client";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->log_client = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->log_client ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Log Net */
	this_parm = "log_net";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->log_net = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->log_net ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Log Errors */
	this_parm = "log_errors";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->log_errors = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->log_errors ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Login Name */
	this_parm = "login_name";
	if(CMDSETPARMMATCH(parm, this_parm) ||
	   CMDSETPARMMATCH(parm, "lname")
	)
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		free(net_parms.login_name);
		net_parms.login_name = strdup(val);

		switch(DBValidateObjectName(net_parms.login_name))
		{
		  case -1:
		    s = (char *)malloc(
			STRLEN(net_parms.login_name) + 80
		    );
		    sprintf(
			s,
"%s: Invalid name.",
			net_parms.login_name
		    );
		    XSWPrintWarning(s);
		    free(s);
		    break;
		  case 1:
		    s = (char *)malloc(
			STRLEN(net_parms.login_name) + 80
		    );  
		    sprintf(
			s,
"%s: Name too long.",
			net_parms.login_name
		    );
		    XSWPrintWarning(s);
		    free(s);
		    break;
		  case 2:
		    s = (char *)malloc(
			STRLEN(net_parms.login_name) + 80
		    );  
		    sprintf(
			s,
"%s: Name too short.",
			net_parms.login_name
		    );
		    XSWPrintWarning(s);
		    free(s);
		    break;
		  case 3:
		    s = (char *)malloc(
			STRLEN(net_parms.login_name) + 80
		    );  
		    sprintf(
			s,
"%s: Name contains invalid character(s).",
			net_parms.login_name
		    );
		    XSWPrintWarning(s);
		    free(s);
		    break;
		}
	    }

	    s = (char *)malloc(
		STRLEN(this_parm) +
		STRLEN(net_parms.login_name) + 80
	    );
	    sprintf(
		s,
		"%s = `%s'",
		this_parm,
		(net_parms.login_name != NULL) ?
		    net_parms.login_name : ""
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Login Password */
	this_parm = "login_password";
	if(CMDSETPARMMATCH(parm, this_parm) ||
	   CMDSETPARMMATCH(parm, "login_passwd") ||
	   CMDSETPARMMATCH(parm, "login_pass") ||
	   CMDSETPARMMATCH(parm, "lpassword") ||
	   CMDSETPARMMATCH(parm, "lpasswd") ||
	   CMDSETPARMMATCH(parm, "lpass")
	)
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		free(net_parms.login_password);
		net_parms.login_password = strdup(val);
	    }
	    s = (char *)malloc(
		STRLEN(this_parm) +
		STRLEN(net_parms.login_password) + 80
	    );
	    sprintf(
		s,
		"%s = `%s'",
		this_parm,
		(net_parms.login_password != NULL) ?
		    net_parms.login_password : ""
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Music */
	this_parm = "music";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		opt->music = StringIsYes(val);
		if(opt->music)
		    XSWUpdateBackgroundMusic();
		else
		    SoundStopBackgroundMusic(&sound);
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->music ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Net Interval */
	this_parm = "net_interval";
	if(CMDSETPARMMATCH(parm, this_parm) ||
	   CMDSETPARMMATCH(parm, "net_int") ||
	   CMDSETPARMMATCH(parm, "interval") ||
	   CMDSETPARMMATCH(parm, "int")
	)
	{
	    char *s;
	    if(!STRISEMPTY(val))
		net_parms.net_int = MAX(atol(val), 0l);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %ld ms",
		this_parm,
		net_parms.net_int
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Net Bandwidth */
	this_parm = "net_bandwidth";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		opt->net_bandwidth = MAX(atol(val), 0l);
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %ld byte%s",
		this_parm,
		opt->net_bandwidth,
		(opt->net_bandwidth == 1l) ? "" : "s"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Net Timeout */
	this_parm = "net_timeout";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->net_timeout = MAX(atol(val), 0l);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %ld ms",
		this_parm,
		opt->net_timeout
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Notify Scanner Contacts */
	this_parm = "notify_scanner_contacts";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->notify_scanner_contacts = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->notify_scanner_contacts ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Scanner Limiting */
	this_parm = "scanner_limiting";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->scanner_limiting = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->scanner_limiting ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Show Formal Labels */
	this_parm = "show_formal_labels";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		int vi = XSW_SHOW_FORMAL_LABEL_NEVER;
		if(isdigit(*val))
		{
		    vi = atoi(val);
		}
		else
		{
		    if(strcasepfx(val, "never"))
			vi = XSW_SHOW_FORMAL_LABEL_NEVER;
		    else if(strcasepfx(val, "as_needed") ||
			    strcasepfx(val, "asneeded")
		    )
			vi = XSW_SHOW_FORMAL_LABEL_AS_NEEDED;
		    else if(strcasepfx(val, "always"))
			vi = XSW_SHOW_FORMAL_LABEL_ALWAYS;
		    else
			XSWPrintWarning(
"Available show_formal_label values; never, as_needed, always"
			);
		}
		opt->show_formal_label = vi;
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    switch(opt->show_formal_label)
	    {
	      case XSW_SHOW_FORMAL_LABEL_ALWAYS:
		sprintf(s, "%s = always", this_parm);
		break;
	      case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
		sprintf(s, "%s = as_needed", this_parm);
		break;
	      case XSW_SHOW_FORMAL_LABEL_NEVER:
		sprintf(s, "%s = never", this_parm);
		break;
	      default:
		sprintf(s, "%s = *badvalue*", this_parm);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	}

	/* Show Lens Flares */
	this_parm = "show_lens_flares";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->show_lens_flares = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->show_lens_flares ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Show Nebula Glow */
	this_parm = "show_nebula_glow";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->show_nebula_glow = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->show_nebula_glow ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Show Network Errors */
	this_parm = "show_net_errors";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->show_net_errors = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->show_net_errors ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Show Strobe Glow */
	this_parm = "show_strobe_glow";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
		opt->show_strobe_glow = StringIsYes(val);
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		opt->show_strobe_glow ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Show Viewscreen Markings */
	this_parm = "show_viewscreen_markings";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		if(StringIsYes(val))
		    opt->show_viewscreen_flags |=
			XSW_SHOW_VIEWSCREEN_MARKINGS;
		else
		    opt->show_viewscreen_flags &=
			~XSW_SHOW_VIEWSCREEN_MARKINGS;
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		(opt->show_viewscreen_flags &
		    XSW_SHOW_VIEWSCREEN_MARKINGS) ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Show Viewscreen Netstats */
	this_parm = "show_viewscreen_netstats";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		if(StringIsYes(val))
		    opt->show_viewscreen_flags |=
			XSW_SHOW_VIEWSCREEN_NETSTATS;
		else
		    opt->show_viewscreen_flags &=
			~XSW_SHOW_VIEWSCREEN_NETSTATS;
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		(opt->show_viewscreen_flags & 
		    XSW_SHOW_VIEWSCREEN_NETSTATS) ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);    
	}

	/* Show Viewscreen Object Labels */
	this_parm = "show_viewscreen_object_labels";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		if(StringIsYes(val))
		    opt->show_viewscreen_flags |=
			XSW_SHOW_VIEWSCREEN_OBJECT_LABELS;
		else
		    opt->show_viewscreen_flags &=
			~XSW_SHOW_VIEWSCREEN_OBJECT_LABELS;
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		(opt->show_viewscreen_flags &
		    XSW_SHOW_VIEWSCREEN_OBJECT_LABELS) ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Show Viewscreen Selected Weapon */
	this_parm = "show_viewscreen_selected_weapon";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		if(StringIsYes(val))
		    opt->show_viewscreen_flags |=
			XSW_SHOW_VIEWSCREEN_SELECTED_WEAPON;
		else
		    opt->show_viewscreen_flags &=
			~XSW_SHOW_VIEWSCREEN_SELECTED_WEAPON;
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    sprintf(
		s,
		"%s = %s",
		this_parm,
		(opt->show_viewscreen_flags &
		    XSW_SHOW_VIEWSCREEN_MARKINGS) ? "on" : "off"
	    );
	    XSWPrint(s);
	    free(s);
	}

	/* Sound Server Type */
	this_parm = "sound_server_type";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		if(strcasepfx(val, "none"))
		    sound.server_type = SOUND_SERVER_TYPE_NONE;
		else if(strcasepfx(val, "yiff"))
		    sound.server_type = SOUND_SERVER_TYPE_YIFF;
		else if(strcasepfx(val, "esound"))
		    sound.server_type = SOUND_SERVER_TYPE_ESOUND;
		else if(strcasepfx(val, "mikmod"))
		    sound.server_type = SOUND_SERVER_TYPE_MIKMOD;
		else
		    XSWPrintWarning(
"Available sound_server_type values; none, yiff, esound, mikmod"
		    );
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    switch(sound.server_type)
	    {
	      case SOUND_SERVER_TYPE_NONE:
		sprintf(s, "%s = none", this_parm);
		break;
	      case SOUND_SERVER_TYPE_YIFF:
		sprintf(s, "%s = yiff", this_parm);
		break;
	      case SOUND_SERVER_TYPE_ESOUND:
		sprintf(s, "%s = esound", this_parm);
		break;
	      case SOUND_SERVER_TYPE_MIKMOD:
		sprintf(s, "%s = mikmod", this_parm);
		break;
	      default:
		sprintf(s, "%s = *badvalue*", this_parm);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	}

	/* Sounds */
	this_parm = "sounds";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    xsw_sound_struct *snd = &sound;
	    if(!STRISEMPTY(val))
	    {
		int vi = XSW_SOUNDS_NONE;
		if(isdigit(*val))
		{
		    vi = atoi(val);
		}
		else
		{
		    if(strcasepfx(val, "none"))
			vi = XSW_SOUNDS_NONE;
		    else if(strcasepfx(val, "events"))
			vi = XSW_SOUNDS_EVENTS;
		    else if(strcasepfx(val, "engine"))
			vi = XSW_SOUNDS_ENGINE;
		    else if(strcasepfx(val, "all"))
			vi = XSW_SOUNDS_ALL;
		    else
			XSWPrintWarning(
"Available sounds values; none, events, engine, all"
			);
		}
		/* Initialize or shut down the sound server as
		 * needed
		 */
		if(snd->server_type > SOUND_SERVER_TYPE_NONE)
		{
		    if((opt->sounds > XSW_SOUNDS_NONE) &&
		       (vi <= XSW_SOUNDS_NONE)
		    )
		    {
			/* Sound was on, now turned off */

			/* Delete all sndrefs */
			SndRefDeleteAll(&xsw_sndref, &total_sndrefs);

			/* Shutdown sound */
			SoundShutdown(snd);
		    }
		    else if((opt->sounds <= XSW_SOUNDS_NONE) &&
			    (vi > XSW_SOUNDS_NONE) &&
			    (snd->con_data == NULL)
		    )
		    {
			/* Sound was off, now turned on */

			/* Initialize sound */
			if(SoundInit(snd, snd->server_type))
			{
			    /* Could not initialize sound */
			    XSWDialog(
				"Sound Initialization Failed",
				XSW_IMG_DIALOG_ICON_ERROR,
"Unable to establish a connection to the sound server\n\
or unable to start the sound server. Please make sure\n\
that the sound server is running and that the specified\n\
address is correct.",
				(xsw_bridge_win != NULL) ?
				    xsw_bridge_win->toplevel : None
			    );
			}
			else
			{
			    /* Successfully initialized sound */

			    /* Open the sndrefs from file */
			    XSWOpenSndRefs(fname.sndrefs);
			}
		    }
		    else
		    {
			/* No change */
		    }
		    opt->sounds = vi;
		}
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    switch(opt->sounds)
	    {
	      case XSW_SOUNDS_NONE:
		sprintf(s, "%s = off", this_parm);
		break;
	      case XSW_SOUNDS_EVENTS:
		sprintf(s, "%s = events", this_parm);
		break;
	      case XSW_SOUNDS_ENGINE:
		sprintf(s, "%s = engine", this_parm);
		break;
	      case XSW_SOUNDS_ALL:
		sprintf(s, "%s = all", this_parm);
		break;
	      default:
		sprintf(s, "%s = *badvalue*", this_parm);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	}

	/* Throttle Mode */
	this_parm = "throttle_mode";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		if(strcasepfx(val, "i"))
		    opt->throttle_mode = XSW_THROTTLE_MODE_INCREMENTAL;
		else if(strcasepfx(val, "b"))
		    opt->throttle_mode = XSW_THROTTLE_MODE_BIDIRECTIONAL;
		else if(strcasepfx(val, "n"))
		    opt->throttle_mode = XSW_THROTTLE_MODE_NORMAL;
		else
		    XSWPrintWarning(
"Available throttle_mode values; incremental, bidirectional, normal"
		    );
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    switch(opt->throttle_mode)
	    {
	      case XSW_THROTTLE_MODE_INCREMENTAL:
		sprintf(s, "%s = incremental", this_parm);
		break;
	      case XSW_THROTTLE_MODE_BIDIRECTIONAL:
		sprintf(s, "%s = bidirectional", this_parm);
		break; 
	      case XSW_THROTTLE_MODE_NORMAL:
		sprintf(s, "%s = normal", this_parm);
		break;
	      default:
		sprintf(s, "%s = *badvalue*", this_parm);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	}

	/* Units */
	this_parm = "units";
	if(CMDSETPARMMATCH(parm, this_parm))
	{
	    char *s;
	    if(!STRISEMPTY(val))
	    {
		if(strcasepfx(val, "xsw"))
		    opt->units = XSW_UNITS_XSW;
		else if(strcasepfx(val, "metric"))
		    opt->units = XSW_UNITS_METRIC;
		else if(strcasepfx(val, "english"))
		    opt->units = XSW_UNITS_ENGLISH;
		else if(strcasepfx(val, "parsecs"))
		    opt->units = XSW_UNITS_PARSECS;
		else
		    XSWPrintWarning(
"Available units values: XSW, Metric, English, Parsecs"
		    );
	    }
	    s = (char *)malloc(STRLEN(this_parm) + 40 + 80);
	    switch(opt->units)
	    {
	      case XSW_UNITS_XSW:
		sprintf(s, "%s = XSW", this_parm);
		break;
	      case XSW_UNITS_METRIC:
		sprintf(s, "%s = Metric", this_parm);
		break;
	      case XSW_UNITS_ENGLISH:
		sprintf(s, "%s = English", this_parm);
		break;
	      case XSW_UNITS_PARSECS:
		sprintf(s, "%s = Parsecs", this_parm);
		break;
	      default:
		sprintf(s, "%s = *badvalue*", this_parm);
		break;
	    }
	    XSWPrint(s);
	    free(s);
	}

	/* Check if no value was given */
	if(STRISEMPTY(val))
	{
	    /* No value given, print usage */
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s [property][=<value>]'",
		THIS_CMD_NAME
	    );
	    XSWPrint(s);
	    free(s);
	}
	else
	{
	    /* Value is specified */

	    /* Redraw the bridge window */
	    BridgeWinDrawAll(xsw_bridge_win);
	}

	free(parm);
	free(val);

	return(0);
}
