#include <stdlib.h>
#include <string.h>

#include "../include/unvmath.h"

#include "xsw.h"


void DBInRangeUpdate(
	const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int DBInRangeAdd(int ref_obj_num, int tar_obj_num, char check_range);
void DBInRangeDelete(xsw_object_struct *obj);
void DBInRangeClear(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Updates the inrange objects list by first clearing it and
 *	then adding only the objects that are in the same sector as
 *	the object specified by obj_num.
 */
void DBInRangeUpdate(
	const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, n, type;
	xsw_object_struct *ref_obj, *obj;

	/* First clear the inrange objects list */
	DBInRangeClear();

	/* Get the pointer to the referance object */
	if(DBIsObjectGarbage(obj_num))
	    return;

	ref_obj = xsw_object[obj_num];


	/* Begin fetching objects that are in range */

	/* Get home and area objects first */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    /* Skip if not these type of objects */
	    type = obj->type;
	    if((type != SW_OBJ_TYPE_HOME) &&
	       (type != SW_OBJ_TYPE_AREA) &&
	       (type != SW_OBJ_TYPE_WORMHOLE) &&
	       (type != SW_OBJ_TYPE_ELINK)
	    )
		continue;

	    /* Check if object is in the same sector as referance
	     * object and that it's recently updated
	     */
	    if(!SWMath3DInSameSectorPtr(obj, ref_obj))
		continue;

	    /* Object not updated recently? */
	    if((obj->last_updated + option.net_object_outdated_time)
		< cur_ms
	    )
	    {
		/* Referance object is always considered up to date */
		if(obj != ref_obj)
		    continue;
	    }


	    /* Object is in range and up to date, add it to the list */
	    n = MAX(total_inrange_objects, 0);
	    total_inrange_objects = n + 1;
	    inrange_xsw_object = (xsw_object_struct **)realloc(
		inrange_xsw_object,
		total_inrange_objects * sizeof(xsw_object_struct *)
	    );
	    if(inrange_xsw_object == NULL)
	    {
		total_inrange_objects = 0;
		return;
	    }

	    /* Set pointer to object */
	    inrange_xsw_object[n] = obj;
	}


	/* Next, get all other types of objects */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    /* Skip these types of objects */
	    type = obj->type;
	    if((type == SW_OBJ_TYPE_HOME) ||
	       (type == SW_OBJ_TYPE_AREA) ||
	       (type == SW_OBJ_TYPE_WORMHOLE) ||
	       (type == SW_OBJ_TYPE_ELINK)
	    )
		continue;

	    /* Check if object is in the same sector as referance
	     * object and that it's recently updated
	     */
	    if(!SWMath3DInSameSectorPtr(obj, ref_obj))
		continue;

	    /* Object not updated recently? */
	    if((obj->last_updated + option.net_object_outdated_time)
		< cur_ms
	    )
	    {
		/* Referance object is always considered up to date */
		if(obj != ref_obj)
		    continue;
	    }


	    /* Object is in range and up to date, add it to the list */
	    n = MAX(total_inrange_objects, 0);
	    total_inrange_objects = n + 1;
	    inrange_xsw_object = (xsw_object_struct **)realloc(
		inrange_xsw_object,
		total_inrange_objects * sizeof(xsw_object_struct *)
	    );
	    if(inrange_xsw_object == NULL)
	    {
		total_inrange_objects = 0;
		return;
	    }

	    /* Set pointer to object */
	    inrange_xsw_object[n] = obj;
	}
}


/*
 *	Adds tar_obj_num to the in range objects list.
 *
 *	If check_range is false, then tar_obj is added to the list
 *	unconditionally.
 *
 *	If check_range is true then the tar_obj is checked to see if
 *	it is in range with ref_obj_num. If any only if tar_obj_num
 *	is in range with ref_obj_num will tar_obj_num be added to the
 *	list.
 */
int DBInRangeAdd(int ref_obj_num, int tar_obj_num, char check_range)
{
	int i;
	xsw_object_struct *ref_obj, *tar_obj;

	if(DBIsObjectGarbage(tar_obj_num))
	    return(-1);

	tar_obj = xsw_object[tar_obj_num];

	/* Remove all instances of the object in the list */
	for(i = 0; i < total_inrange_objects; i++)
	{
	    if(inrange_xsw_object[i] == tar_obj)
	    {
		int n;
		total_inrange_objects--;
		for(n = i; n < total_inrange_objects; n++)
		    inrange_xsw_object[n] = inrange_xsw_object[n + 1];
		i--;
		continue;
	    }
	}

	/* Check if target object is in range with referance object? */
	if(check_range)
	{
	    if(DBIsObjectGarbage(ref_obj_num))
		return(-1);

	    ref_obj = xsw_object[ref_obj_num];

	    /* Check if in range */
	    if(!SWMath3DInSameSectorPtr(tar_obj, ref_obj))
		return(0);
	}

	/* Add target object to in range list by type
	 *
	 * If the object is of type HOME or AREA then add it to the
	 * beginning of the list, otherwise append it to the end
	 */

	/* Sanitize total */
	if(total_inrange_objects < 0)
	    total_inrange_objects = 0;

	/* Increment total and allocate more pointers */
	total_inrange_objects++;
	inrange_xsw_object = (xsw_object_struct **)realloc(
	    inrange_xsw_object,
	    total_inrange_objects * sizeof(xsw_object_struct *)
	);
	if(inrange_xsw_object == NULL)
	{
	    total_inrange_objects = 0;
	    return(-3);
	}

	/* Add by type */
	if((tar_obj->type == SW_OBJ_TYPE_HOME) ||
	   (tar_obj->type == SW_OBJ_TYPE_AREA) ||
	   (tar_obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (tar_obj->type == SW_OBJ_TYPE_ELINK)
	)
	{
	    /* Home or area, add to beginning of list */

	    /* Shift pointers */
	    for(i = total_inrange_objects - 1; i > 0; i--)
		inrange_xsw_object[i] = inrange_xsw_object[i - 1];

	    inrange_xsw_object[0] = tar_obj;
	}
	else
	{
	    /* All other types, append to end */

	    i = total_inrange_objects - 1;
	    inrange_xsw_object[i] = tar_obj;
	}

	return(0);
}


/*
 *	Removes obj from the in range objects list.
 *
 *	Warning, the global inrange_xsw_object pointers will
 *	be reallocated!
 */
void DBInRangeDelete(xsw_object_struct *obj)
{
	int i, h;
	xsw_object_struct *tar_obj;

	if(obj == NULL)
	    return;

	if(total_inrange_objects < 0)
	    total_inrange_objects = 0;

	for(i = 0, h = -1;
	    i < total_inrange_objects;
	    i++
	)
	{
	    tar_obj = inrange_xsw_object[i];
	    if(tar_obj == NULL)
		continue;

	    if(tar_obj == obj)
	    {
		tar_obj = NULL;
		continue;
	    }

	    h = i;
	}

	/* Update inrange list size */
	total_inrange_objects = h + 1;
	if(total_inrange_objects > 0)
	{
	    inrange_xsw_object = (xsw_object_struct **)realloc(
		inrange_xsw_object,
		total_inrange_objects * sizeof(xsw_object_struct *)
	    );
	    if(inrange_xsw_object == NULL)
	    {
		total_inrange_objects = 0;
		return;
	    }
	}
	else
	{
	    free(inrange_xsw_object);
	    inrange_xsw_object = NULL;
	    total_inrange_objects = 0;
	}
}


/*
 *	Deletes the in range objects list (but not the objects
 *	themselves).
 */
void DBInRangeClear(void)
{
	free(inrange_xsw_object);
	inrange_xsw_object = NULL;
	total_inrange_objects = 0;
}
