#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"
#include "../include/unvfile.h"

#include "net.h"
#include "xsw.h"


int DBGetPlayerOwner(const int obj_num);
int DBGetTopObjectNumber(void);

int DBCreateObjectEconomyProduct(
	int obj_num,
	const char *product_name,
	int ocs_code,
	sw_credits_t sell_price,
	sw_credits_t buy_price,
	float product_amount,
	float product_max
);

void DBObjectTractor(int obj_num, int tar_obj_num);
void DBSortEconomyProducts(int obj_num);

int DBValidateObjectName(const char *name);
int DBValidateObjectPassword(const char *password);

float DBGetObjectVisibility(const int obj_num);
float DBGetObjectVisibilityPtr(xsw_object_struct *obj);

float DBGetObjectEffectiveScannerRange(int obj_num);
float DBGetObjectEffectiveScannerRangePtr(xsw_object_struct *obj);

char DBIsObjectInEffectiveScannerRange(
	int obj_num, int tar_obj_num,
	float scanner_range
);
char DBIsObjectInEffectiveScannerRangePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj,
	float scanner_range
);

int DBGetObjectNumByPtr(xsw_object_struct *obj);
const char *DBGetFormalNameStr(const int obj_num);
const char *DBGetObjectVectorName(const float theta);

void DBSetPlayerObject(const int obj_num);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *      Returns the player object that owns the specified object (not
 *      always just the immediate owner).
 */
int DBGetPlayerOwner(const int obj_num)
{
	const int max_recusions = 256;
	int level = 0;
        int i = obj_num;
        xsw_object_struct *obj;

        while(!DBIsObjectGarbage(i) && (level < max_recusions))
        {
            obj = xsw_object[i];

            if(obj->type == SW_OBJ_TYPE_PLAYER)
                return(i);

            /* If a non-player object owns itself then this is an
             * error, return -1
             */
            if(i == obj->owner)
                return(-1);

            i = obj->owner;
	    level++;
        }

        return(-1);
}

/*
 *	Returns the index of the non-garbage object with the highest
 *	index or -1 if there are no objects found.
 */
int DBGetTopObjectNumber(void)
{
        const int total = total_objects;
        int i, h = -1;
        xsw_object_struct *obj, **list = xsw_object;

        for(i = 0; i < total; i++)
        {
            obj = list[i];
            if(obj == NULL)
                continue;

            if(obj->type <= SW_OBJ_TYPE_GARBAGE)
                continue;

            h = i;
        }

        return(h);
}


/*
 *	Allocates a new economy product on the object.
 */
int DBCreateObjectEconomyProduct(
	int obj_num,
	const char *product_name,
	int ocs_code,
	sw_credits_t sell_price,
	sw_credits_t buy_price,
	float product_amount,
	float product_max
)
{
	int i, n;
	xsw_ecoproduct_struct *product;
	xsw_ecodata_struct *eco;
	xsw_object_struct *obj;

	if(product_name == NULL)
	    return(-2);

	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	obj = xsw_object[obj_num];
	eco = obj->eco;
	if(eco == NULL)
	{
	    if(UNVAllocEco(obj))
		return(-3);
	    eco = obj->eco;
	}

	/* Sanitize total */
	if(eco->total_products < 0)
	    eco->total_products = 0;

	/* Is product_name in list? */
	for(i = 0, n = -1; i < eco->total_products; i++)
	{
	    product = eco->product[i];
	    if(product == NULL)
	    {
		n = i;
		continue;
	    }

	    if(!strcasecmp(product->name, product_name))
		break;
	}
	if(i >= eco->total_products)
	{
	    /* No such product in list, allocate a new one */

	    if(n > -1)
	    { 
		/* Have allocated pointer n */
		i = n;
	    }
	    else
	    {
		/* No unallocated pointers, allocate more */
		i = eco->total_products;
		eco->total_products = i + 1;
		eco->product = (xsw_ecoproduct_struct **)realloc(
		    eco->product,   
		    eco->total_products * sizeof(xsw_ecoproduct_struct *)
		);
		if(eco->product == NULL)
		{
		    eco->total_products = 0;
		    return(-3);
		}
	    }

	    /* Allocate structure */
	    eco->product[i] = (xsw_ecoproduct_struct *)calloc(
		1, sizeof(xsw_ecoproduct_struct)
	    );
	    if(eco->product[i] == NULL)
		return(-3);
	}

	product = eco->product[i];

	/* New product index is now i */
	strncpy(
	    product->name, product_name,
	    sizeof(product->name)
	);
	product->name[sizeof(product->name) - 1] = '\0';

	product->ocs_code = ocs_code;
	product->sell_price = sell_price;
	product->buy_price = buy_price;
	product->amount = product_amount;
	product->amount_max = product_max;

	return(0);
}


/*
 *	Sets the source object to tractor lock the target object.
 *
 *	If tar_obj_num is -1 then the source object will remove
 *	tractor lock on all objects.
 */
void DBObjectTractor(int obj_num, int tar_obj_num)
{
	int n;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Remove all tractor locks? */
	if(tar_obj_num < 0)
	{
	    free(obj->tractored_object);
	    obj->tractored_object = NULL;
	    obj->total_tractored_objects = 0;
	    return;
	}

	/* Is the object already being tractored? */
	for(n = 0; n < obj->total_tractored_objects; n++)
	{
	    if(obj->tractored_object[n] == tar_obj_num)
		return;
	}

	/* Maximum tractorable objects reached? */
	if(obj->total_tractored_objects >= SW_TRACTOR_BEAM_MAX_OBJECTS)
	    return;

	/* Add to the tractored objects list */
	n = MAX(obj->total_tractored_objects, 0);
	obj->total_tractored_objects = n + 1;
	obj->tractored_object = (int *)realloc(
	    obj->tractored_object,
	    obj->total_tractored_objects * sizeof(int)
	);
	if(obj->tractored_object == NULL)
	{
	    obj->total_tractored_objects = 0;
	    return;
	}

	/* Set new tractored object */
	obj->tractored_object[n] = tar_obj_num;
}

/*
 *	Alphabitizes the economy products on the object by the name of
 *	the product.
 */
void DBSortEconomyProducts(int obj_num)
{
	int i, n, k, strc;
	char **strv;
	xsw_ecoproduct_struct *product, **products_list;
	xsw_ecodata_struct *eco;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];
	eco = obj->eco;
	if(eco == NULL)
	    return;
	     
	if(eco->total_products <= 0)
	    return;

	/* Build products name list */
	strc = 0;
	strv = NULL;
	for(i = 0; i < eco->total_products; i++)
	{
	    product = eco->product[i];
	    if(product == NULL)
		continue;

	    /* Copy name to string list */
	    strc++;
	    strv = (char **)realloc(strv, strc * sizeof(char *));
	    if(strv == NULL)
		return;

	    strv[strc - 1] = STRDUP(product->name);
	    if(strv[strc - 1] == NULL)
	    {
		strc--;
		continue;
	    }
	}

	/* Got no strings? */
	if(strc == 0)
	    return;

	/* Sort strings */
	strv = StringQSort(strv, strc); 
	if(strv == NULL)
	    return;


	/* Copy products pointers to tempory pointer array */
	products_list = (xsw_ecoproduct_struct **)calloc(
	    1,
	    eco->total_products * sizeof(xsw_ecoproduct_struct *)
	);
	memcpy(
	    products_list,		/* Destination */
	    eco->product,		/* Source */
	    eco->total_products * sizeof(xsw_ecoproduct_struct *)
	);


	/* Sort products */
	for(i = 0, k = 0; i < strc; i++)
	{
	    if(k >= eco->total_products)
		break;

	    if(strv[i] == NULL)
		continue;

	    /* Get pointer matching string in tempory products list */
	    for(n = 0; n < eco->total_products; n++)
	    {
		product = products_list[n];
		if(product == NULL)
		    continue;

		if(!strcasecmp(product->name, strv[i]))
		    break;
	    }
	    if(n < eco->total_products)
	    {
		/* Matched product name, now put it in order */
		eco->product[k] = products_list[n];
		k++;
	    }
	    else
	    {  
		fprintf(
		    stderr,
"DBSortEconomyProducts(): Warning: Missing name from tmp products array (#%i %i %i)\n",
		    obj_num, strc,
		    eco->total_products
		);
	    }

	} 
	/* Reallocate the rest of the product pointers */
	if(k > 0)
	{
	    eco->total_products = k;
	    eco->product = (xsw_ecoproduct_struct **)realloc(
		eco->product,
		eco->total_products * sizeof(xsw_ecoproduct_struct *)
	    );
	    if(eco->product == NULL)
	    {
		eco->total_products = 0;
		/* Don't return, need to free other stuff */
	    }
	}
	else
	{    
	    free(eco->product);
	    eco->product = NULL;
	    eco->total_products = 0;
	}

	/* Delete tempory products list */
	free(products_list);

	/* Delete tempory product names list */
	strlistfree(strv, strc);
}


/*      
 *      Validates name for an object.  Returns non-zero if it is
 *      NOT valid.
 *
 *      -1  general error.
 *      0   no error.
 *      1   name too long.
 *      2   name too short.
 */
int DBValidateObjectName(const char *name)
{
	int i, n;
	const char *s = SW_INVALID_NAME_CHARACTERS;

	/* Null pointer? */
	if(name == NULL)
	    return(-1);

	/* Empty string? */
	if(name[0] == '\0')
	    return(2);

	/* Invalid characters? */
	for(i = 0; name[i] != '\0'; i++)
	{
	    if(i >= SW_OBJ_NAME_MAX)
		return(1);

	    for(n = 0; s[n] != '\0'; n++)
	    {
		if(name[i] == s[n])
		    return(3);
	    }
	}

	return(0);
}

/*
 *      Checks if this is a valid password.
 */
int DBValidateObjectPassword(const char *password)
{
	return(0);
}


/*
 *	Returns object's current visibility.
 *
 *	Takes into account the object's location, area objects that
 *	produce limited visibility will be applied to the returned
 *	visibility.
 */
float DBGetObjectVisibility(const int obj_num)
{
	return(UNVGetObjectVisibility(
	    xsw_object, total_objects,
	    obj_num
	));
}

float DBGetObjectVisibilityPtr(xsw_object_struct *obj)
{       
	return(UNVGetObjectVisibilityPtr(obj));
}


/*
 *	Returns the object's current effective scanner range in real
 *	units.
 *
 *	Takes into account the object's location, area objects that
 *	produce limited visibility will be applied to the returned
 *	scanner range.
 */
float DBGetObjectEffectiveScannerRange(int obj_num)
{
	return(UNVGetObjectEffectiveScannerRange(
            xsw_object, total_objects,
            obj_num
        ));
}

float DBGetObjectEffectiveScannerRangePtr(xsw_object_struct *obj)
{
	return(UNVGetObjectEffectiveScannerRangePtr(obj));
}

/*
 *	Returns true if the target object is within the object's
 *	effective scanner range.
 *
 *	The scanner_range must be the effective scanner range obtained
 *	from DBGetObjectEffectiveScannerRange*().
 */
char DBIsObjectInEffectiveScannerRange(
	int obj_num, int tar_obj_num,
	float scanner_range
)
{
	return(UNVIsObjectInEffectiveScannerRange(
	    xsw_object, total_objects,
	    obj_num, tar_obj_num, scanner_range
	));
}

char DBIsObjectInEffectiveScannerRangePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj,
	float scanner_range
)
{
	return(UNVIsObjectInEffectiveScannerRangePtr(
	    obj, tar_obj, scanner_range
	));
}


/*
 *	Returns the object's number from the specified pointer.
 */
int DBGetObjectNumByPtr(xsw_object_struct *obj)
{
	int i;

	if(obj == NULL)
	    return(-1);

	for(i = 0; i < total_objects; i++)
	{
	    if(xsw_object[i] == obj)
		return(i);
	}

	return(-1);
}


/*
 *      Returns a statically allocated string containing the object's
 *      complete formal name in the format of name(#12345).
 *
 *      Or SW_OBJ_NAME_GARBAGE if the object is garbage.
 */
const char *DBGetFormalNameStr(const int obj_num)
{
	if(DBIsObjectGarbage(obj_num))
	    return(SW_OBJ_NAME_GARBAGE);
	else
	    return(UNVGetObjectFormalName(
		obj_num, xsw_object[obj_num]
	    ));
}


/*
 *      Returns the appropriate `shipsman direction' name
 *      based on theta in radians.
 */
const char *DBGetObjectVectorName(const float theta)
{
	if(theta < 0.39269908f) return("forward");
	else if(theta < 1.1780972f) return("starboard bow");
	else if(theta < 1.9634954f) return("starboard");
	else if(theta < 2.7488936f) return("aft starboard");
	else if(theta < 3.5342917f) return("aft");
	else if(theta < 4.3196899f) return("aft port");
	else if(theta < 5.1050881f) return("port");
	else if(theta < 5.8904862f) return("port bow");
	else return("forward");
}


/*
 *	Sets local player object number and all its referances
 *	to that of obj_num.
 *
 *	If obj_num is -1 or garbage, then player object referances
 *	will be reset.
 */
void DBSetPlayerObject(const int obj_num)
{
	xsw_net_parms_struct *net = &net_parms;
	xsw_bridge_win_struct *bw = xsw_bridge_win;

	if(DBIsObjectGarbage(obj_num))
	{
	    net->player_obj_num = -1;
	    net->player_obj_ptr = NULL;
	}
	else
	{
	    net->player_obj_num = obj_num;
	    net->player_obj_ptr = xsw_object[obj_num];
	}

	/* Recreate the selected weapon viewscreen label */
	if(bw != NULL)
	{
	    VSDrawUpdateWeaponLabel(
		&bw->vs_weapon_image, bw->vs_weapon_buf
	    );
	}
}
