#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "keymaps.h"
#ifdef JS_SUPPORT
# include "jsmaps.h"
#endif
#include "gctl.h"
#include "xsw.h"


xsw_gctl_struct *GCtlNew(const int controller);

int GCtlManageGUIKeyEvent(
        xsw_gctl_struct *gctl, event_t *event,
	void *bw_ptr
);
void GCtlResetKeys(xsw_gctl_struct *gctl);

void GCtlManageJoysticks(xsw_gctl_struct *gctl);

void GCtlManage(
        xsw_gctl_struct *gctl,
        const time_t cur_ms, const time_t cur_sec, const float tc
);

void GCtlDelete(xsw_gctl_struct *gctl);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Switches to and initializes controller.
 *	Previous controller will be automatically shut down.
 *
 *	Returns non zero on error.
 */
xsw_gctl_struct *GCtlNew(const int type)
{
	int i;
#ifdef JS_SUPPORT
	xsw_jsmap_struct *jsm;
#endif
	xsw_gctl_struct *gctl = XSW_GCTL(calloc(
	    1, sizeof(xsw_gctl_struct)
	));
	if(gctl == NULL)
	    return(NULL);

	gctl->type = type;
	switch(type)
	{
#ifdef JS_SUPPORT
	  case XSW_CONTROLLER_JOYSTICK:
	    for(i = 0; i < total_jsmaps; i++)
	    {
		jsm = xsw_jsmap[i];
		if(jsm == NULL)
		    continue;

		if(STRISEMPTY(jsm->device_name))
		    continue;

		if(JSMapInitializeJoystick(
		    jsm,
		    jsm->device_name,
		    fname.js_calib
		))
		{
		    char *s = (char *)malloc(
			STRLEN(jsm->device_name) + 80
		    );
		    sprintf(
			s,
"An error occured while initializing joystick `%s'.",
			jsm->device_name
		    );
		    XSWPrintWarning(s);
		    free(s);
		}
	    }
	    break;
#endif	/* JS_SUPPORT */

	  case XSW_CONTROLLER_POINTER:
	    /* Not supported yet */
	    GCtlDelete(gctl);
	    return(NULL);

	    break;

	  default: /* Default to XSW_CONTROLLER_KEYBOARD */
	    /* No initializing of keyboard needed */

	    break;
	}

	GCtlResetKeys(gctl);

	return(gctl);
}


/*
 *	Manages the key event.
 *
 *	Returns the number of events handled.
 */
int GCtlManageGUIKeyEvent(
        xsw_gctl_struct *gctl, event_t *event,
	void *bw_ptr
)
{
	int events_handled = 0;
	bool_t key_pressed, shift_key_state;
	keycode_t keycode;
	event_key_t *key;
	xsw_bridge_win_struct *bw = XSW_BRIDGE_WIN(bw_ptr);

	if((gctl == NULL) || (event == NULL) || (bw == NULL))
	    return(events_handled);

	key = &event->xkey;

	/* Skip if the bridge window is not in focus */
	if(!bw->has_focus || (prompt_mode != XSW_PROMPT_NONE))
	    return(events_handled);

	/* Get key state and key code */
	key_pressed = (event->type == KeyPress) ? True : False;
	shift_key_state = (key->state & OSW_KEY_STATE_SHIFT) ? True : False;
	keycode = key->keycode;

	/* Turn left */
	if(keycode == KeymapGetKeycode(XSW_KM_TURN_LEFT))
	{
	    if(key_pressed)
	    {
		OSWKBAutoRepeatOff();
	        gctl->pks_turn_left = True;
	    }
	    else
	    {
	        gctl->pks_turn_left = False;
	    }
	    events_handled++;
	}
	/* Turn right */
	else if(keycode == KeymapGetKeycode(XSW_KM_TURN_RIGHT))
	{
	    if(key_pressed)
	    {
		OSWKBAutoRepeatOff();
		gctl->pks_turn_right = True;
	    }
	    else
	    {
		gctl->pks_turn_right = False;
	    }
	    events_handled++;
	}
	/* Throttle up */
	else if(keycode == KeymapGetKeycode(XSW_KM_THROTTLE_INC))
	{
	    if(key_pressed)
	    {
		OSWKBAutoRepeatOff();
	        gctl->pks_throttle_inc = True;
	    }
	    else
	    {
	        gctl->pks_throttle_inc = False;
	    }
	    events_handled++;
	}
	/* Throttle down */
	else if(keycode == KeymapGetKeycode(XSW_KM_THROTTLE_DEC))
	{
	    if(key_pressed)
	    {
		OSWKBAutoRepeatOff();
		gctl->pks_throttle_dec = True;
	    }
	    else
	    {
		gctl->pks_throttle_dec = False;
	    }
	    events_handled++;
	}
	/* Throttle idle */
	else if(keycode == KeymapGetKeycode(XSW_KM_THROTTLE_IDLE))
	{
	    if(key_pressed)
	    {
		OSWKBAutoRepeatOff();
		gctl->pks_throttle_idle = True;
	    }
	    else
	    {
		gctl->pks_throttle_idle = False;
	    }
	    events_handled++;
	}
	/* Throttle Max */
	else if(keycode == KeymapGetKeycode(XSW_KM_THROTTLE_MAX))
	{
	    if(key_pressed)
	    {
		OSWKBAutoRepeatOff();
		gctl->pks_throttle_max = True;  
	    }
	    else
	    {   
		gctl->pks_throttle_max = False;
	    }
	    events_handled++;
	}
	/* Fire Weapon */
	else if(keycode == KeymapGetKeycode(XSW_KM_FIRE_WEAPON))
	{
	    if(key_pressed)
	    {
		gctl->fire_weapon = True;
		OSWKBAutoRepeatOff();
		gctl->pks_fire_weapon = True;
	    }
	    else
	    {
		gctl->fire_weapon = False;
		gctl->pks_fire_weapon = False;
	    }
	    events_handled++;
	}
	/* Omni Directrional Thrust Mode */
	else if(keycode == KeymapGetKeycode(XSW_KM_OMNI_DIR_THRUST))
	{
	    if(key_pressed)
	    {
		gctl->omni_dir_thrust = True;
		OSWKBAutoRepeatOff();
		gctl->pks_omni_dir_thrust = True;
	    }
	    else
	    {
		gctl->omni_dir_thrust = False;
		gctl->pks_omni_dir_thrust = False;
	    }
	    events_handled++;
	}
	/* External Dampers */
	else if(keycode == KeymapGetKeycode(XSW_KM_EXTERNAL_DAMPERS))
	{
	    if(key_pressed)
	    {
		gctl->external_dampers = True;
		OSWKBAutoRepeatOff();
		gctl->pks_external_dampers = True;
	    }
	    else
	    {
		gctl->external_dampers = False;
		gctl->pks_external_dampers = False;
	    }
	    events_handled++;
	}

	/* Turn auto repeat back on only when all states false */
	if(!gctl->pks_turn_left && !gctl->pks_turn_right &&
	   !gctl->pks_throttle_inc && !gctl->pks_throttle_dec &&
	   !gctl->pks_throttle_idle && !gctl->pks_throttle_max &&
	   !gctl->pks_fire_weapon &&
	   !gctl->pks_omni_dir_thrust &&
	   !gctl->pks_external_dampers
	)
	    OSWKBAutoRepeatOn();

	return(events_handled);
}

/*
 *	Resets all key states and turns keyboard repeating on.
 */
void GCtlResetKeys(xsw_gctl_struct *gctl)
{
	if(gctl == NULL)
	    return;

	gctl->fire_weapon = False;
	gctl->omni_dir_thrust = False;
	gctl->external_dampers = False;

	gctl->pks_turn_left = False;
	gctl->pks_turn_right = False;
	gctl->pks_throttle_inc = False;
	gctl->pks_throttle_dec = False;
	gctl->pks_throttle_idle = False;
	gctl->pks_throttle_max = False;
	gctl->pks_fire_weapon = False;
	gctl->pks_omni_dir_thrust = False;
	gctl->pks_external_dampers = False;

	if(IDC())
	{
	    OSWKBAutoRepeatOn();
	}
}


/*
 *	Manages the joystick event.
 *
 *	Any joystick events received will be processed and the
 *	gctl will be updated.
 *
 *	If a joystick button is pressed and the button is mapped to
 *	a keyboard key then a KeyPress or KeyRelease event will be
 *	queued.
 */
void GCtlManageJoysticks(xsw_gctl_struct *gctl)
{
#ifdef JS_SUPPORT
	int i, n;
	float d1, d2;
	event_t *event;
	js_data_struct *jsd;
	xsw_jsmap_struct *jsm;
	xsw_jsmap_axis_struct *axis;
	xsw_jsmap_button_struct *button;
	scale_bar_struct *scale;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_option_struct *opt = &option;

	if(gctl == NULL)
	    return;

	/* Iterate through joystick mappings list */
	for(i = 0; i < total_jsmaps; i++)
	{
	    jsm = xsw_jsmap[i];
	    if(jsm == NULL)
		continue;

	    jsd = &jsm->jsd;
	    if(JSUpdate(jsd) != JSGotEvent)
		continue;

	    /* Iterate through axis list */
	    for(n = 0; n < jsm->total_axises; n++)
	    {
		axis = jsm->axis[n];
		if(axis == NULL)
		    continue;

		switch(axis->op)
		{
		  case JSMAP_AXIS_OP_TURN:
#if defined(HAVE_LIBJSW)
		    gctl->turn = JSGetAxisCoeffNZ(jsd, n);
#endif
		    break;

		  case JSMAP_AXIS_OP_THROTTLE:
		    switch(opt->throttle_mode)
		    {
		      case XSW_THROTTLE_MODE_BIDIRECTIONAL:
			/* Bi directional mode */
#if defined(HAVE_LIBJSW)
			gctl->throttle = JSGetAxisCoeffNZ(jsd, n);
#endif
			break;
		      case XSW_THROTTLE_MODE_INCREMENTAL:
#if defined(HAVE_LIBJSW)
			gctl->throttle = JSGetAxisCoeffNZ(jsd, n);
#endif
			break;
		      default:	/* Normal mode */
#if defined(HAVE_LIBJSW)
			gctl->throttle = (JSGetAxisCoeff(jsd, n) + 1.0)
			    / 2.0;
#endif
			break;
		    }
		    break;

		  case JSMAP_AXIS_OP_THRUST_DIR:
#if defined(HAVE_LIBJSW)
		    gctl->thrust_dir = JSGetAxisCoeffNZ(jsd, n);
#endif
		    break;

		  case JSMAP_AXIS_OP_AIM_WEAPON_HEADING:
#if defined(HAVE_LIBJSW)
		    gctl->aim_weapon_heading = JSGetAxisCoeffNZ(jsd, n);
#endif
		    break;

		  case JSMAP_AXIS_OP_AIM_WEAPON_PITCH:
#if defined(HAVE_LIBJSW)
		    gctl->aim_weapon_pitch = JSGetAxisCoeffNZ(jsd, n);
#endif
		    break;

		  case JSMAP_AXIS_OP_SCANNER_ZOOM:
		    if(bw == NULL)
			break;
		    scale = &bw->scanner_sb;
		    d1 = scale->pos;
		    d2 = scale->pos_max - scale->pos_min;
#if defined(HAVE_LIBJSW)
		    scale->pos = (JSGetAxisCoeff(jsd, n) - 1.0)	/ -2.0 * d2;
#endif
		    /* Any change? */
		    if(d1 != scale->pos)
		    {
			/* Sanitize values */
			if(scale->pos < scale->pos_min)
			    scale->pos = scale->pos_min;
			else if(scale->pos > scale->pos_max)
			    scale->pos = scale->pos_max;
			BridgeWinScannerScaleValueChangedCB(scale, scale->pos, bw);
			ScaleBarQueueDraw(scale);
		    }
		    break;

		  case JSMAP_AXIS_OP_VS_ZOOM:
		    if(bw == NULL)
			break;
		    d1 = bw->viewscreen_zoom;
#if defined(HAVE_LIBJSW)
		    bw->viewscreen_zoom = (JSGetAxisCoeff(jsd, n) - 1.0)
			/ -2.0;
#endif
		    if(d1 != bw->viewscreen_zoom)
		    {
			if(bw->viewscreen_zoom < XSW_VS_ZOOM_MIN)
			    bw->viewscreen_zoom = XSW_VS_ZOOM_MIN;
			else if(bw->viewscreen_zoom > XSW_VS_ZOOM_MAX)
			    bw->viewscreen_zoom = XSW_VS_ZOOM_MAX;
		    }
		    break;
		}
	    }

	    /* Iterate through the buttons list */
	    for(n = 0; n < jsm->total_buttons; n++)
	    {
		button = jsm->button[n];
		if(button == NULL)
		    continue;

#if defined(HAVE_LIBJSW)
		if(n >= jsm->jsd.total_buttons)
		    break;
		if(jsm->jsd.button[n] == NULL)
		    continue;

		/* Change in button state? */
		if(jsm->jsd.button[n]->state == JSButtonStateOn)
		{
		    if(button->state)
			continue;
		    else
			button->state = True;
		}
		else
		{
		    if(button->state)
			button->state = False;
		    else
			continue;
		}
#endif
		/* Format a GUI KeyPress event to be sent to the GUI
		 * as if a keyboard key was pressed
		 */
		event = (event_t *)calloc(1, sizeof(event_t));
		if((event != NULL) && IDC())
		{
		    event_key_t *key = &event->xkey;
		    key->type = button->state ?
			KeyPress : KeyRelease;
		    key->serial = 0l;
		    key->send_event = True;
		    key->display = osw_gui[0].display;
		    key->window = (bw != NULL) ? bw->toplevel : None;
		    key->root = osw_gui[0].root_win;
		    key->subwindow = None;
		    key->time = CurrentTime;
		    key->x = 0;
		    key->y = 0;
		    key->x_root = 0;
		    key->y_root = 0;
		    key->state = 0;
		    key->keycode = button->keycode;
		    key->same_screen = True;
		    OSWSendEvent(
		        KeyPressMask | KeyReleaseMask,
		        event,
		        True
		    );
		}
		free(event);
	    }
	}
#endif	/* JS_SUPPORT */
}


/*
 *	Manages the game controller based on the specified
 *	controller type.
 */
void GCtlManage(
        xsw_gctl_struct *gctl,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	osw_gui_struct *gui = &osw_gui[0];
	xsw_option_struct *opt = &option;

	if(gctl == NULL)
	    return;

	/* If in keyboard mode then update the game controller
	 * states for axis oriented functions
	 */
	if((gctl->type == XSW_CONTROLLER_KEYBOARD) && IDC())
	{
	    const bool_t shift_key_state = (gui->key_state & OSW_KEY_STATE_SHIFT) ? True : False;
	    float pks_x_accel, pks_y_accel;

	    if(shift_key_state)
		pks_x_accel = gctl->aim_weapon_heading;
	    else
		pks_x_accel = gctl->turn;

	    /* Turn left (x) */
	    if(gctl->pks_turn_left)
	    {
		pks_x_accel -= 0.02 * tc;
	    }
	    /* Turn right (x) */
	    else if(gctl->pks_turn_right)
	    {
		pks_x_accel += 0.02 * tc;
	    }
	    /* Turn left (x) decelleration */
	    else if(pks_x_accel < 0.0)
	    {
		pks_x_accel += 0.02 * tc;
	    }
	    /* Turn right (x) decelleration */
	    else if(pks_x_accel > 0.0)
	    {
		pks_x_accel -= 0.02 * tc;
	    }

	    if(pks_x_accel > 1.0)
		pks_x_accel = 1.0;
	    else if(pks_x_accel < -1.0)
		pks_x_accel = -1.0;

	    /* Null zone */
	    if((pks_x_accel > (-0.01 * tc)) &&
	       (pks_x_accel < (0.01 * tc))
	    )
		pks_x_accel = 0.0;


	    if(shift_key_state)
	    {
		gctl->aim_weapon_heading = pks_x_accel;
	    }
	    else
	    {
		gctl->turn = pks_x_accel;

		/* Stop aiming */
		pks_x_accel = gctl->aim_weapon_heading;
		if(pks_x_accel < 0.0f)
		    pks_x_accel += 0.02f * tc;
		else if(pks_x_accel > 0.0f)
		    pks_x_accel -= 0.02f * tc;
		if((pks_x_accel > (-0.01f * tc)) &&
		   (pks_x_accel < (0.01f * tc))
		)
		    pks_x_accel = 0.0f;
		gctl->aim_weapon_heading = pks_x_accel;
	    }


	    pks_y_accel = gctl->throttle;

	    /* Turottle increase (y) */
	    if(gctl->pks_throttle_inc)
	    {
		const bool_t fast = (shift_key_state ||
		    (opt->throttle_mode == XSW_THROTTLE_MODE_BIDIRECTIONAL)) ? True : False;
		pks_y_accel += (fast ? 0.01f : 0.002f) * tc;
		if(pks_y_accel > 1.0f)
		    pks_y_accel = 1.0f;
	    }
	    /* Throttle decrease (y) */
	    if(gctl->pks_throttle_dec)
	    {
		const bool_t fast = (shift_key_state ||
		    (opt->throttle_mode == XSW_THROTTLE_MODE_BIDIRECTIONAL)) ? True : False;
		pks_y_accel -= (fast ? 0.01f : 0.002f) * tc;
		if(pks_y_accel < -1.0f)
		    pks_y_accel = -1.0f;
	    }
	    /* Throttle null zone (y), if was not 0 */
	    if((pks_y_accel < 0.005f) &&
	       (pks_y_accel > -0.005f) &&
	       (gctl->throttle != 0.0f)
	    )
		pks_y_accel = 0.0f;

	    /* Throttle set to idle? */
	    if(gctl->pks_throttle_idle && !shift_key_state)
		pks_y_accel = 0.0f;

	    /* Throttle set to max? */ 
	    if(gctl->pks_throttle_max && !shift_key_state)
		pks_y_accel = 1.0f;

	    gctl->throttle = pks_y_accel;

	    /* Thrust dir */
	    gctl->thrust_dir = 0.0f;
	}
}


/*
 *	Shuts down and deletes the game controller.
 */
void GCtlDelete(xsw_gctl_struct *gctl)
{
	if(gctl == NULL)
	    return;

#ifdef JS_SUPPORT
	/* Close all joysticks */
	if(total_jsmaps > 0)
	{
	    int i;
	    for(i = 0; i < total_jsmaps; i++)
		JSMapShutdownJoystick(xsw_jsmap[i]);
	}
#endif	/* JS_SUPPORT */

	/* Close pointer device */

	free(gctl);
}
