#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/objects.h"

#include "ignoredobj.h"


ignored_obj_struct *IgnoredObjNew(void);
void IgnoredObjDelete(ignored_obj_struct *iobj);

int IgnoredObjListMatch(
	ignored_obj_struct **list, const int total,
	const int obj_num, const char *name
);

int IgnoredObjListAdd(
	ignored_obj_struct ***list, int *total,
	const int obj_num, const char *name
);
void IgnoredObjListRemove(
	ignored_obj_struct ***list, int *total,
	const int obj_num, const char *name
);
void IgnoredObjListDelete(
	ignored_obj_struct ***list, int *total,
	const int iobj_num
);
void IgnoredObjListClear(ignored_obj_struct **list, const int total);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a new ignored object.
 */
ignored_obj_struct *IgnoredObjNew(void)
{
	return(IGNORED_OBJ(calloc(
	    1, sizeof(ignored_obj_struct)
	)));
}

/*
 *	Deletes the ignored object.
 */
void IgnoredObjDelete(ignored_obj_struct *iobj)
{
	if(iobj == NULL)
	    return;

	free(iobj);
}


/*
 *	Looks for an ignored object in the list that matches the
 *	specified object number and name.
 *
 *	If obj_num is -1 then only the name will be matched.
 *
 *	If name is NULL then only the obj_num will be matched.
 *
 *	Returns the index number of the matched ignored object or
 *	negative on failed match.
 */
int IgnoredObjListMatch(
	ignored_obj_struct **list, const int total,
	const int obj_num, const char *name
)
{
	if((list == NULL) || (total <= 0))
	    return(-1);

	/* Match both the number and the name? */
	if((obj_num > -1) && (name != NULL))
	{
	    int i;
	    ignored_obj_struct *iobj;

	    for(i = 0; i < total; i++)
	    {
		iobj = list[i];
		if(iobj == NULL)
		    continue;

		if(iobj->obj_num != obj_num)
		    continue;

		if(STRISEMPTY(iobj->name))
		    continue;

		/* Case sensitive name matching */
		if(strcmp(iobj->name, name))
		    continue;

		return(i);
	    }
	}
	/* Match only the name? */
	else if(name != NULL)
	{
	    int i;
	    ignored_obj_struct *iobj;

	    for(i = 0; i < total; i++)
	    {
		iobj = list[i];
		if(iobj == NULL)
		    continue;

		if(STRISEMPTY(iobj->name))
		    continue;

		/* Case sensitive name matching */
		if(strcmp(iobj->name, name))
		    continue;

		return(i);
	    }
	}
	/* Match only the number? */
	else if(obj_num > -1)
	{
	    int i;
	    ignored_obj_struct *iobj;

	    for(i = 0; i < total; i++)
	    {
		iobj = list[i];
		if(iobj == NULL)
		    continue;

		if(iobj->obj_num != obj_num)
		    continue;

		return(i);
	    }
	}

	return(-1);
}


/*
 *	Adds an ignored object to the list.
 *
 *	Returns the index number of the ignored object that was added
 *	to the list or negative on error.
 */
int IgnoredObjListAdd(
	ignored_obj_struct ***list, int *total,
	const int obj_num, const char *name
)
{
	int i;
	ignored_obj_struct *iobj;

	if((list == NULL) || (total == NULL))
	    return(-1);

	for(i = 0; i < *total; i++)
	{
	    iobj = (*list)[i];
	    if(iobj == NULL)
		break;
	}
	if(i < *total)
	{
	    IgnoredObjDelete((*list)[i]);
	    (*list)[i] = iobj = IgnoredObjNew();
	}
	else
	{
	    i = MAX((*total), 0);
	    *total = i + 1;
	    *list = (ignored_obj_struct **)realloc(
		*list,
		(*total) * sizeof(ignored_obj_struct *)
	    );
	    if(*list == NULL)
	    {
		*total = 0;
		return(-3);
	    }
	    (*list)[i] = iobj = IgnoredObjNew();
	}

	if(iobj == NULL)
	    return(-3);

	iobj->obj_num = obj_num;

	if(name != NULL)
	{
	    strncpy(iobj->name, name, sizeof(iobj->name));
	    iobj->name[sizeof(iobj->name) - 1] = '\0';
	}

	return(i);
}

/*
 *	Removes all ignored objects that match the specified object
 *	number and name.
 */
void IgnoredObjListRemove(
	ignored_obj_struct ***list, int *total,
	const int obj_num, const char *name
)
{
	int iobj_num;

	if((list == NULL) || (total == NULL))
	    return;

	while(1)
	{
	    iobj_num = IgnoredObjListMatch(
		*list, *total,
		obj_num, name
	    );
	    if(iobj_num < 0)
		break;

	    IgnoredObjListDelete(list, total, iobj_num);
	}
}

/*
 *	Deletes the ignored object from the list.
 */
void IgnoredObjListDelete(
	ignored_obj_struct ***list, int *total,
	const int iobj_num
)
{
	int i;

	if((list == NULL) || (total == NULL))
	    return;

	/* Is the ignored object number valid? */
	if((iobj_num < 0) || (iobj_num >= *total))
	    return;

	/* Delete the ignored object */
	IgnoredObjDelete((*list)[iobj_num]);

	/* Reduce the total */
	*total = (*total) - 1;

	/* Shift the pointers in the list */
	for(i = iobj_num; i < *total; i++)
	    (*list)[i] = (*list)[i + 1];

	/* Reallocate the list */
	if(*total > 0)
	{
	    *list = (ignored_obj_struct **)realloc(
		*list,
		(*total) * sizeof(ignored_obj_struct *)
	    );
	    if(*list == NULL)
	    {
		*total = 0;
		return;
	    }
	}
	else
	{
	    free(*list);
	    *list = NULL;
	    *total = 0;
	}
}

/*
 *	Deletes the list of ignored objects.
 */
void IgnoredObjListClear(ignored_obj_struct **list, const int total)
{
	int i;

	for(i = 0; i < total; i++)
	    IgnoredObjDelete(list[i]);
	free(list);
}

