#include <stdlib.h>
#include <string.h>

#include "../include/disk.h"

#include "xsw.h"
#include "splash.h"

#include "imagepaths.h"

#include "icons/info.h"
#include "icons/warning.h"
#include "icons/error.h"
#include "icons/question.h"
#include "images/esm.h"		/* Energy Saver Mode */


int ImgIsAllocated(int img_num);
int ImgIsLoaded(int img_num);

void ImgResize(
	int img_num, unsigned int width, unsigned int height
);
int ImgCreate(int img_num);
int ImgOpenFile(int img_num, const char *filename, bool_t load_pixmap);
int ImgOpenData(int img_num, const u_int8_t *data, bool_t load_pixmap);
void ImgUnload(int img_num);

int ImgLoadAll(void *sw_ptr);
void ImgUnloadAll(void);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the image is allocated.
 */
int ImgIsAllocated(int img_num)
{
	if((xsw_image == NULL) ||
	   (img_num < 0) ||
	   (img_num >= total_images)
	)
	    return(0);
	if(xsw_image[img_num] == NULL)
	    return(0);
	else
	    return(1);
}


/*
 *	Checks if the image is loaded.
 */
int ImgIsLoaded(int img_num)
{
	if(!ImgIsAllocated(img_num))
	    return(0);
	else if(xsw_image[img_num]->image == NULL)
	    return(0);
	else
	    return(1);
}


/*
 *	Resizes the image.
 */
void ImgResize(
	int img_num, unsigned int width, unsigned int height
)
{
	image_t	*old_image = NULL,
		*new_image = NULL;
	xsw_image_struct *img;

	if(!ImgIsLoaded(img_num))
	    return;

	img = xsw_image[img_num];

	old_image = img->image;
	if(old_image == NULL)
	    return;

	/* Create a new image of the new size */
	if(OSWCreateImage(&new_image, width, height))
	    return;

	/* Copy/resize the old image to the new image */
	WidgetResizeImageBuffer(
	    new_image->bits_per_pixel,
	    (u_int8_t *)new_image->data,	/* Target */
	    (const u_int8_t *)old_image->data,	/* Source */
	    new_image->width, new_image->height,
	    new_image->bytes_per_line,
	    old_image->width, old_image->height,
	    old_image->bytes_per_line
	);

	/* Delete the old image */
	OSWDestroyImage(&old_image);

	/* Set the new image */
	img->image = new_image;

	/* Resize the pixmap and mask as needed */
	if(img->pixmap != None)
	{
	    OSWDestroyPixmap(&img->pixmap);
	    OSWDestroyBitmap(&img->mask);
	    img->pixmap = OSWCreatePixmapFromImage(new_image);
	    img->mask = OSWCreateMaskFromImage(new_image);
	}
}


/*
 *	Allocates an image structure explicitly.
 */
int ImgCreate(int img_num)
{
	int i, prev_total_images;
	xsw_image_struct *img;

	if(img_num < 0)
	    return(-1);

	/* Already allocated? */
	if(ImgIsAllocated(img_num))
	{
	    ImgUnload(img_num);
	    return(0);
	}

	/* Record previous total */
	prev_total_images = total_images;

	/* Adjust global variable total_images */
	total_images = img_num + 1;


	/* Reallocate pointers */
	xsw_image = (xsw_image_struct **)realloc(
	    xsw_image,
	    total_images * sizeof(xsw_image_struct *)
	);
	if(xsw_image == NULL)
	{
	    total_images = 0;
	    return(-3);
	}

	/* Allocate each new entry */
	for(i = prev_total_images; i < total_images; i++)
	{
	    xsw_image[i] = img = (xsw_image_struct *)calloc(
		1, sizeof(xsw_image_struct)
	    );
	    if(img != NULL)
	    {
		img->load_state = 0;
		img->filename = NULL;
		img->image = NULL;
		img->pixmap = None;
		img->mask = None;
	    }
	}

	return(0);
}


/*
 *	Opens the image file to the specified image number.
 */
int ImgOpenFile(int img_num, const char *filename, bool_t load_pixmap)
{
	char *path;
	xsw_image_struct *img = NULL;

	if(STRISEMPTY(filename))
	    return(-1);

	/* Complete the path as needed */
	if(ISPATHABSOLUTE(filename))
	    path = STRDUP(filename);
	else
	    path = STRDUP(PrefixPaths(dname.images, filename));

	/* Image already allocated? */
	if(ImgIsAllocated(img_num))
	{
	    /* Unload it before loading the new one */
	    ImgUnload(img_num);
	    img = xsw_image[img_num];
	}
	else
	{
	    /* Allocate a new image at the specified number */
	    if(ImgCreate(img_num))
	    {
		free(path);
		return(-1);
	    }

	    img = xsw_image[img_num];
	}
	if(img == NULL)
	{
	    free(path);
	    return(-1);
	}

	/* Set file name */
	img->filename = path;

	/* Open the image */
	img->image = WidgetLoadImageFromTgaFile(path);
	if(img->image == NULL)
	{
	    fprintf(
		stderr,
"ImgOpenFile(): Warning: Unable to open image image: %s.\n",
		filename
	    );
	    ImgUnload(img_num);
	    return(-1);
	}

	/* Load pixmap too? */
	if(load_pixmap)
	{
	    img->pixmap = OSWCreatePixmapFromImage(img->image);
	    img->mask = OSWCreateMaskFromImage(img->image);
	}

	/* Mark the image as loaded */
	img->load_state = 1;

	return(0);
}

/*
 *	Opens the TGA image data to the specified image number.
 */
int ImgOpenData(int img_num, const u_int8_t *data, bool_t load_pixmap)
{
	xsw_image_struct *img = NULL;
	
	if(data == NULL)
	    return(-1);

	/* Image already allocated? */
	if(ImgIsAllocated(img_num))
	{
	    /* Unload it before loading the new one */
	    ImgUnload(img_num);
	    img = xsw_image[img_num];
	}
	else
	{   
	    /* Allocate a new image at the specified number */
	    if(ImgCreate(img_num))
		return(-1);
	     
	    img = xsw_image[img_num];
	}                              
	if(img == NULL)
	    return(-1);

	/* Open the image from the TGA data */
	img->image = WidgetLoadImageFromTgaData(data);
	if(img->image == NULL)
	{
	    fprintf(
		stderr,
"ImgOpenData(): Warning: Unable to open TGA image data: %p\n",
		data
	    );
	    ImgUnload(img_num);
	    return(-1);
	}

	/* Load pixmap too? */
	if(load_pixmap)
	{
	    img->pixmap = OSWCreatePixmapFromImage(img->image);
	    img->mask = OSWCreateMaskFromImage(img->image);
	}

	/* Mark as loaded */
	img->load_state = 1;

	return(0);
}


/*
 *	Unloads the image.
 */
void ImgUnload(int img_num)
{
	xsw_image_struct *img;

	if(!ImgIsAllocated(img_num))
	    return;

	img = xsw_image[img_num];

	free(img->filename);
	img->filename = NULL;

	OSWDestroyImage(&img->image);
	OSWDestroyPixmap(&img->pixmap);
	OSWDestroyBitmap(&img->mask);

	/* Mark as unloaded */
	img->load_state = 0;
}


/*
 *	Loads all the images.
 */
int ImgLoadAll(void *sw_ptr)
{
	int imgs_loaded = 0;
	const char *msg;
	const int total_imgs = 49;	/* Images to be loaded */
	splash_win_struct *sw = SPLASH_WIN(sw_ptr);

#define LOAD_FILE(_i_,_s_)	{	\
 SplashProgress(			\
  sw,					\
  (float)imgs_loaded / (float)total_imgs, \
  msg					\
 );					\
 ImgOpenFile((_i_),(_s_),False);	\
 imgs_loaded++;				\
}
#define LOAD_FILE_PIXMAP(_i_,_s_) {	\
 SplashProgress(			\
  sw,					\
  (float)imgs_loaded / (float)total_imgs, \
  msg					\
 );					\
 ImgOpenFile((_i_),(_s_),True);		\
 imgs_loaded++;				\
}
#define LOAD_DATA(_i_,_d_)	{	\
 SplashProgress(			\
  sw,					\
  (float)imgs_loaded / (float)total_imgs, \
  msg					\
 );					\
 ImgOpenData((_i_),(_d_),False);	\
 imgs_loaded++;				\
}

	/* Startup splash background is not loaded since it is already
	 * loaded by the splash window, the splash window at this
	 * point should already be initialized
	 */

	/* XSW_IMG_SPLASH_BG */


	/* Viewscreen Markings */
	msg = "Loading Images: Viewscreen Markings...";
	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_MARK_OBJECT,
	    IMGPATH_VIEWSCREEN_MARK_OBJECT
	);
	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_MARK_VESSEL,
	    IMGPATH_VIEWSCREEN_MARK_VESSEL
	);
	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_MARK_INCOMING_FIRE,
	    IMGPATH_VIEWSCREEN_MARK_INCOMING_FIRE
	);

	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_WEAPON_PROJECTILE,
	    IMGPATH_VIEWSCREEN_WEAPON_PROJECTILE
	);
	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_WEAPON_PULSE,
	    IMGPATH_VIEWSCREEN_WEAPON_PULSE
	);
	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_WEAPON_STREAM,
	    IMGPATH_VIEWSCREEN_WEAPON_STREAM
	);

	/* Scanner markingss */
	msg = "Loading Images: Scanner Markings...";
	LOAD_FILE(
	    XSW_IMG_SCANNER_MARK_UNKNOWN,
	    IMGPATH_SCANNER_MARK_UNKNOWN
	);
	LOAD_FILE(
	    XSW_IMG_SCANNER_MARK_LOCKED,
	    IMGPATH_SCANNER_MARK_LOCKED
	);
	LOAD_FILE(
	    XSW_IMG_SCANNER_MARK_WEAPON,
	    IMGPATH_SCANNER_MARK_WEAPON
	);
	LOAD_FILE(
	    XSW_IMG_SCANNER_MARK_HOME,
	    IMGPATH_SCANNER_MARK_HOME
	);
	LOAD_FILE(
	    XSW_IMG_SCANNER_MARK_AREA,
	    IMGPATH_SCANNER_MARK_AREA
	);

	/* Main window icons */
	msg = "Loading Images: Icons...";
	LOAD_FILE(
	    XSW_IMG_ICON_XSW,
	    IMGPATH_ICON_BRIDGE
	);
	LOAD_FILE(
	    XSW_IMG_ICON_ECONOMY,
	    IMGPATH_ICON_ECONOMY
	);
	LOAD_FILE(
	    XSW_IMG_ICON_OPTIONS,
	    IMGPATH_ICON_OPTIONS
	);
	LOAD_FILE(
	    XSW_IMG_ICON_UNIVERSE,
	    IMGPATH_ICON_UNIVERSE
	);
	LOAD_FILE(
	    XSW_IMG_ICON_STARCHART,
	    IMGPATH_ICON_STARCHART
	);

	/* Universe List icons */
	LOAD_FILE_PIXMAP(
	    XSW_IMG_ICON_UNIVERSE_SMALL,
	    IMGPATH_UNIV_ICON_SMALL
	);
	LOAD_FILE_PIXMAP(
	    XSW_IMG_ICON_UNIVERSE_LOGIN_SMALL,
	    IMGPATH_UNIV_ICON_LOGIN_SMALL
	);
	LOAD_FILE_PIXMAP(
	    XSW_IMG_ICON_UNIVERSE_OUTDATED_SMALL,
	    IMGPATH_UNIV_ICON_OUTDATED_SMALL
	);
	LOAD_FILE_PIXMAP(
	    XSW_IMG_ICON_UNIVERSE_UNKNOWN_SMALL,
	    IMGPATH_UNIV_ICON_UNKNOWN_SMALL
	);

        /* Info icon */
        LOAD_DATA(
            XSW_IMG_DIALOG_ICON_INFO,
            info_tga
        );
	/* Warning icon */
	LOAD_DATA(
	    XSW_IMG_DIALOG_ICON_WARNING,
	    warning_tga
	);
	/* Error icon */
	LOAD_DATA(
	    XSW_IMG_DIALOG_ICON_ERROR,
	    error_tga
	);
        /* Question icon */
        LOAD_DATA(
            XSW_IMG_DIALOG_ICON_QUESTION,
            question_tga
        );

	/* Energy saver mode banner */
	msg = "Loading Images: Bridge Panels...";
	LOAD_DATA(
	    XSW_IMG_VIEWSCREEN_ENERGY_SAVER_MODE,
	    esm_tga
	);

	/* Stats console */
	LOAD_FILE(
	    IMG_CODE_STATS_CON1,
	    IMGPATH_BPANEL_L1
	);
	LOAD_FILE(
	    IMG_CODE_STATS_CON2,
	    IMGPATH_BPANEL_L2
	);
	LOAD_FILE(
	    IMG_CODE_STATS_CON3,
	    IMGPATH_BPANEL_L3
	);
	LOAD_FILE(
	    IMG_CODE_STATS_CON4,
	    IMGPATH_BPANEL_L4
	);
	LOAD_FILE(
	    IMG_CODE_SCANNER,
	    IMGPATH_BPANEL_R1
	);
	LOAD_FILE(
	    IMG_CODE_SRO_CON1,
	    IMGPATH_BPANEL_R2
	);
	LOAD_FILE(
	    IMG_CODE_SRO_CON2,
	    IMGPATH_BPANEL_R3
	);
	LOAD_FILE(
	    IMG_CODE_SRO_CON3,
	    IMGPATH_BPANEL_R4
	);
	LOAD_FILE(
	    IMG_CODE_MESG_CON,
	    IMGPATH_BPANEL_MESG
	);

	/* Bridge panel outlines */
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_HULL,
	    IMGPATH_BPANEL_OL_HULL
	);
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_POWER,
	    IMGPATH_BPANEL_OL_POWER
	);
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_VIS,
	    IMGPATH_BPANEL_OL_VIS
	);
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_SHIELDS,
	    IMGPATH_BPANEL_OL_SHIELDS
	);
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_DMGCTL,
	    IMGPATH_BPANEL_OL_DMGCTL
	);
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_ANTIMATTER,
	    IMGPATH_BPANEL_OL_ANTIMATTER
	);
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_LTHROTTLE,
	    IMGPATH_BPANEL_OL_THROTTLE_L
	);
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_RTHROTTLE,
	    IMGPATH_BPANEL_OL_THROTTLE_R
	);
	LOAD_FILE(
	    IMG_CODE_BPANEL_OL_THRUSTVECTOR,
	    IMGPATH_BPANEL_OL_THRUSTVECTOR
	);

	/* StarChart window */
	msg = "Loading Images: Starchart Markings...";
	LOAD_FILE_PIXMAP(
	    XSW_IMG_STARCHART_ZOOM_IN,
	    IMGPATH_STARCHART_ZOOM_IN
	);
	LOAD_FILE_PIXMAP(
	    XSW_IMG_STARCHART_ZOOM_OUT,
	    IMGPATH_STARCHART_ZOOM_OUT
	);
	LOAD_FILE_PIXMAP(
	    XSW_IMG_STARCHART_JUMP_TO_PLAYER,
	    IMGPATH_STARCHART_JUMP_TO_PLAYER
	);

	/* Large message screen background is laoded as needed
	 * later on
	 */

	/* IMGPATH_MESGWIN_BG */


	/* Lens flares */
	msg = "Loading Images: Special Effects...";
	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_LENSFLARE01,
	    IMGPATH_VIEWSCREEN_LENSFLARE01
	);
	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_LENSFLARE02,
	    IMGPATH_VIEWSCREEN_LENSFLARE02
	);

	/* Strobe glow 1 */
	LOAD_FILE(
	    XSW_IMG_VIEWSCREEN_STROBEGLOW01,
	    IMGPATH_VIEWSCREEN_STROBEGLOW01
	);

	return(0);

#undef LOAD_FILE
#undef LOAD_FILE_PIXMAP
#undef LOAD_DATA
}


/*
 *	Deletes the entire images list.
 */
void ImgUnloadAll(void)
{
	int i;
	xsw_image_struct *img;

	for(i = 0; i < total_images; i++)
	{
	    img = xsw_image[i];
	    if(img == NULL)
		continue;

	    ImgUnload(i);
	    free(img->filename);
	    free(img);
	    xsw_image[i] = img = NULL;
	}

	free(xsw_image);
	xsw_image = NULL;
	total_images = 0;
}
