#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>

#include "../include/tga.h"
#include "../include/osw-x.h"
#include "../include/imgrefs.h"
#include "../include/objects.h"

#include "particle.h"
#include "xswimgrefs.h"
#include "imgrefsfio.h"
#include "xsw.h"

#include "images/unknown.h"


imgref_struct *ImgRefGetPtr(
	imgref_struct **imgrefs_list, const int nimgrefs,
	const int imgref_num
);
int ImgRefIsLoaded(imgref_struct *imgref);

void ImgRefReset(
	imgref_struct **imgrefs_list, const int nimgrefs,
	const int imgref_num
);
int ImgRefCreateExplicit(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const int imgref_num
);

int ImgRefLoadAsDefault(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const int imgref_num
);
int ImgRefLoad(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const int imgref_num
);
void ImgRefUnload(
	imgref_struct **imgrefs_list, const int nimgrefs,
	const int imgref_num
);

void ImgRefClear(imgref_struct ***imgrefs_list, int *nimgrefs);

void ImgRefReclaimMemory(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	xsw_object_struct **objs_list, const int nobjs,
	xsw_particle_struct **particles_list, const int nparticles
);

void ImgRefManage(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const time_t cur_ms, const time_t cur_sec, const float tc,
	void (*load_completed_cb)(const int, void *),
	void *load_completed_data
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Maximum pixels to load per iteration when using async image
 *	loading:
 */
#define IMGREF_ASYNC_PIXEL_LOAD_MAX_ITERATION	100000


/*
 *	Returns the pointer to the imgref in the specified list.
 */
imgref_struct *ImgRefGetPtr(
	imgref_struct **imgrefs_list, const int nimgrefs,
	const int imgref_num
)
{
	if((imgrefs_list == NULL) ||                       
	   (imgref_num < 0) ||
	   (imgref_num >= nimgrefs)
	)
	    return(NULL);
	else
	    return(imgrefs_list[imgref_num]);
}

/*
 *	Checks if the imgref is allocated and loaded.
 *
 *	Note that this does not check if the imgref has
 *	IMGREF_NO_IMAGE set.
 */
int ImgRefIsLoaded(imgref_struct *imgref)
{
	if(imgref == NULL)
	    return(0);
	else
	    return((imgref->image_data != NULL) ? 1 : 0);
}

/*
 *	Unloads the imgref's image and resets all of its values.
 */
void ImgRefReset(
	imgref_struct **imgrefs_list, const int nimgrefs,  
	const int imgref_num
)
{
	int i;
	imgref_struct *imgref = ((imgref_num >= 0) && (imgref_num < nimgrefs)) ?
	    imgrefs_list[imgref_num] : NULL;
	if(imgref == NULL)
	    return;

	/* Delete the image */
	ImgRefUnload(imgrefs_list, nimgrefs, imgref_num);

	/* Reset to default values */
	imgref->options = 0;

	free(imgref->path);
	imgref->path = NULL;

	imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;
	imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;

	imgref->merge_mode = IMGREF_MERGE_NORMAL;
	imgref->frame_determinant = IMGREF_FRAME_DETERMINE_BY_HEADING;
	imgref->layer_placement = IMGREF_LAYER_FG;
	imgref->effects = 0;
	imgref->magnification = 1;

	imgref->total_frames = 1;

	imgref->width = 0;
	imgref->height = 0;

	imgref->fwidth = 0;
	imgref->fheight = 0;

	free(imgref->name);
	imgref->name = NULL;
	free(imgref->description);
	imgref->description = NULL;

	for(i = 0; i < imgref->nlights; i++)
	    ImgRefLightDelete(imgref->light[i]);
	free(imgref->light);
	imgref->light = NULL;
	imgref->nlights = 0;
}

/*
 *	Creates the imgref explicitly as the specified imgref number.
 *
 *	If the imgref already exists then it will be deleted and then
 *	recreated.
 *
 *	Returns 0 on success or non-zero on error.
 */
int ImgRefCreateExplicit(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const int imgref_num
)
{
	int i, prev_total;
	imgref_struct *imgref;

	if((imgrefs_list == NULL) || (nimgrefs == NULL))
	    return(-2);

	if(*nimgrefs < 0)
	    *nimgrefs = 0;

	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	    return(-2);

	/* Already allocated? */
	if(imgref_num < *nimgrefs)
	{
	    /* Delete it first */
	    imgref = (*imgrefs_list)[imgref_num];
	    ImgRefReset(*imgrefs_list, *nimgrefs, imgref_num);
	    ImgRefDelete(imgref);
	    (*imgrefs_list)[imgref_num] = NULL;
	}
	/* Not allocated? */
	else if(imgref_num >= *nimgrefs)
	{
	    /* Allocate the pointer array to accomidate for it */
	    prev_total = *nimgrefs;
	    *nimgrefs = imgref_num + 1;
	    *imgrefs_list = (imgref_struct **)realloc(
		*imgrefs_list,
		(*nimgrefs) * sizeof(imgref_struct *)
	    );
	    if(*imgrefs_list == NULL)
	    {
		*nimgrefs = 0;
		return(-3);
	    }

	    /* Reset new pointers */
	    for(i = prev_total; i < *nimgrefs; i++)
		(*imgrefs_list)[i] = NULL;
	}

	/* Allocate a new imgref */
	(*imgrefs_list)[imgref_num] = imgref = ImgRefNew();
	if(imgref == NULL)
	    return(-3);

	/* Reset values */
	ImgRefReset(*imgrefs_list, *nimgrefs, imgref_num);

	return(0);
}

/*
 *	Loads the imgref as the default imgref (if possible).
 *
 *	If the imgref does not exist then it will be allocated if
 *	possible.
 *
 *	Returns 0 on success or non-zero on error.
 */
int ImgRefLoadAsDefault(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const int imgref_num
)
{
	int i, status;
	imgref_struct *imgref, *def_imgref;
	osw_gui_struct *gui = &osw_gui[0];

	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	    return(-2);

	/* Allocate this imgref as needed */
	imgref = (imgref_num < *nimgrefs) ?
	    (*imgrefs_list)[imgref_num] : NULL;
	if(imgref == NULL)
	{
	    if(ImgRefCreateExplicit(imgrefs_list, nimgrefs, imgref_num))
		return(-3);

	    imgref = (*imgrefs_list)[imgref_num];
	}

	/* Delete all resources associated with this imgref */
	ImgRefReset(*imgrefs_list, *nimgrefs, imgref_num);


	/* If this the default imgref itself that needs to be loaded
	 * then load this imgref without falling back to the default
	 * imgref (since there wouldn't be one)
	 */
	if(imgref_num == IMGREF_DEFAULT_IMGREF)
	{
	    tga_data_struct *td;

/* Explicitly set image format for now */
imgref->img_fmt = IMGREF_IMGFMT_TGA;
	    switch(imgref->img_fmt)
	    {
	      /* TGA format */
	      case IMGREF_IMGFMT_TGA:
		td = (tga_data_struct *)calloc(
		    1, sizeof(tga_data_struct)
		);
		if(td == NULL)
		{
		    imgref->lib_data = NULL;
		    imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;
		    return(-3);
		}
		imgref->lib_data = td;

		/* Open the TGA image from data */
		status = TgaReadFromData(
		    unknown_tga,
		    td,
		    TgaGetDepthForX(
			gui->bits_per_pixel,
			gui->depth
		    )
		);
		if(status != TgaSuccess)
		{
		    char *s;
		    const char *error_msg;
		    switch(status)
		    {
		      case TgaNoBuffers:
			error_msg = "Out of memory";
			break;
		      case TgaBadHeader:
			error_msg = "Bad header";
			break;
		      case TgaBadValue:
			error_msg = "Bad value";
			break;
		      case TgaNoFile:
			error_msg = "No file";
			break;
		      case TgaNoAccess:
			error_msg = "No access";
			break;
		      default:
			error_msg = "General error";
			break;
		    }
		    s = (char *)malloc(40 + STRLEN(error_msg) + 80);
		    sprintf(
			s,
"ImgRef %i: Unable to open image data: %s.\n",
			imgref_num, error_msg
		    );
		    XSWPrintWarning(s);
		    free(s);

		    TgaDestroyData(td);
		    free(td);
		    imgref->lib_data = NULL;
		    imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;

		    return(-1);
		}

		/* Set the image data to point to the image data on
		 * the library
		 */
		imgref->image_data = (u_int8_t *)td->data;

		imgref->width = td->width;
		imgref->height = td->height;

		/* Mark as loaded and done */
		imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
		break;

	      /* Unknown format */
	      default:
		if(imgref->img_fmt > -1)
		{
		    char *s = (char *)malloc(40 + 40 + 80);
		    sprintf(
			s,
"ImgRef %i: Unsupported image format `%i'.\n",
			imgref_num, imgref->img_fmt
		    );
		    XSWPrintWarning(s);
		    free(s);
		}
		break;
	    }

	    /* Sanitize total frames */
	    if(imgref->total_frames > IMGREF_MAX_FRAMES)
		imgref->total_frames = IMGREF_MAX_FRAMES;
	    if(imgref->total_frames < 1)
		imgref->total_frames = 1;

	    /* Calculate frame width and height */
	    imgref->fwidth = imgref->width;
	    imgref->fheight = (int)imgref->height /
		(int)imgref->total_frames;

	    return(0);
	}


	/* Is the default imgref loaded? */
	def_imgref = ImgRefGetPtr(*imgrefs_list, *nimgrefs, IMGREF_DEFAULT_IMGREF);
	if(!ImgRefIsLoaded(def_imgref))
	    return(-1);

	/* Copy all values from the default imgref to this imgref */
	memcpy(imgref, def_imgref, sizeof(imgref_struct));
	imgref->path = STRDUP(def_imgref->path);

	/* Set this imgref's library data to NULL so that its library
	 * data and image data are not deallocated. This basically marks
	 * this imgref as sharing the image_data from another imgref
	 */
	imgref->lib_data = NULL;

	/* Copy lights */
	if(def_imgref->nlights > 0)
	{
	    imgref_light_struct *tar_light, *src_light;

	    imgref->light = (imgref_light_struct **)malloc(
		def_imgref->nlights * sizeof(imgref_light_struct *)
	    );
	    if(imgref->light == NULL)
	    {
		imgref->nlights = 0;
		return(-3);
	    }
	    for(i = 0; i < def_imgref->nlights; i++)
	    {
		src_light = def_imgref->light[i];
	        if(src_light == NULL)
		{
		    imgref->light[i] = NULL;
		    continue;
		}

		imgref->light[i] = tar_light = (imgref_light_struct *)malloc(
		    sizeof(imgref_light_struct)
		);
		if(tar_light != NULL)
		    memcpy(tar_light, src_light, sizeof(imgref_light_struct));
	    }
	}

	return(0);
}

/*
 *	Opens the ImgRef's image.
 *
 *	If the imgref is not allocated then it will be allocated and
 *	its image will be loaded as default (if possible).
 *
 *	If the imgref is set to IMGREF_NO_IMAGE then -2 will be
 *	returned.
 *
 *	Returns 0 on success or non-zero on error.
 */
int ImgRefLoad(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const int imgref_num
)
{
	struct stat stat_buf;
	imgref_struct *imgref;
	osw_gui_struct *gui = &osw_gui[0];
	xsw_option_struct *opt = &option;

	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	    return(-2);

	/* If this imgref is not allocated then load it as the
	 * default if possible
	 */
	imgref = (imgref_num < *nimgrefs) ?
	    (*imgrefs_list)[imgref_num] : NULL;
	if(imgref == NULL)
	    return(ImgRefLoadAsDefault(imgrefs_list, nimgrefs, imgref_num));

	/* Check if this imgref is not suppose to have an image */
	if(imgref->options & IMGREF_NO_IMAGE)
	    return(-2);


	/* Unload the image if it is already loaded */
	ImgRefUnload(*imgrefs_list, *nimgrefs, imgref_num);


	/* Does not have a filename? */
	if(imgref->path == NULL)
	    return(ImgRefLoadAsDefault(imgrefs_list, nimgrefs, imgref_num));

	/* Is that filename pointing to an existing file? */
	if(stat(imgref->path, &stat_buf))
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		char *s = (char *)malloc(
		    STRLEN(imgref->path) + STRLEN(error_msg) +
		    80
		);
		*error_msg = toupper(*error_msg);
		sprintf(
		    s,
		    "%s: %s.",
		    imgref->path,
		    error_msg
		);
		XSWPrintWarning(s);
		free(s);
	    }
	    free(error_msg);

	    return(ImgRefLoadAsDefault(imgrefs_list, nimgrefs, imgref_num));
	}


	/* Open image async? */
	if(opt->async_image_loading)
	{
	    int status;
	    tga_data_struct *td;

/* Explicitly set format for now */
imgref->img_fmt = IMGREF_IMGFMT_TGA;

	    switch(imgref->img_fmt)
	    {
	      /* TGA format */
	      case IMGREF_IMGFMT_TGA:
		td = (tga_data_struct *)calloc(
		    1, sizeof(tga_data_struct)
		);
		if(td == NULL)
		{
		    imgref->lib_data = NULL;
		    imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;
		    return(-3);
		}
		imgref->lib_data = td;

		/* Set up for partial tga read */
		status = TgaStartReadPartialFromFile(
		    imgref->path,
		    td,
		    TgaGetDepthForX(
			gui->bits_per_pixel,
			gui->depth
		    )
		);
		if(status != TgaSuccess)
		{
		    char *s;
		    const char *error_msg;
		    switch(status)
		    {
		      case TgaNoBuffers:
			error_msg = "Out of memory";
			break;
		      case TgaBadHeader:
			error_msg = "Bad header";
			break;
		      case TgaBadValue:
			error_msg = "Bad value";
			break;
		      case TgaNoFile:
			error_msg = "No file";
			break;
		      case TgaNoAccess:
			error_msg = "No access";
			break;
		      default:
			error_msg = "General error";
			break;
		    }
		    s = (char *)malloc(
			40 + STRLEN(imgref->path) +
			STRLEN(error_msg) + 80
		    );
		    sprintf(
			s,
"ImgRef %i: %s: Unable to open the image file: %s.\n",
			imgref_num, imgref->path,
			error_msg
		    );
		    XSWPrintWarning(s);
		    free(s);

		    TgaDestroyData(td);
		    free(td);
		    imgref->lib_data = NULL;
		    imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;

		    return(-1);
		}

		/* Set the image data to point to the image data on
		 * the library
		 */
		imgref->image_data = (u_int8_t *)td->data;

		imgref->width = td->width;
		imgref->height = td->height;

		/* Mark load process as currently loading */
		imgref->load_progress = IMGREF_LOAD_PROGRESS_LOADING;
		break;

	      /* Unknown format */
	      default:
		if(imgref->img_fmt > -1)
		{
		    char *s = (char *)malloc(40 + 40 + 80);
		    sprintf(
			s,
"ImgRef %i: Unsupported image format `%i'.\n",
			imgref_num, imgref->img_fmt
		    );
		    XSWPrintWarning(s);
		    free(s);
		}
		break;
	    }

	    /* Sanitize total frames */
	    if(imgref->total_frames > IMGREF_MAX_FRAMES)
		imgref->total_frames = IMGREF_MAX_FRAMES;
	    if(imgref->total_frames < 1)
		imgref->total_frames = 1;

	    /* Calculate frame width and height */
	    imgref->fwidth = imgref->width;
	    imgref->fheight = (int)imgref->height /
		(int)imgref->total_frames;
	}
	else
	{
	    /* Open entire image at once */
	    int status;
	    tga_data_struct *td;

/* Explicitly set format for now */
imgref->img_fmt = IMGREF_IMGFMT_TGA;

	    switch(imgref->img_fmt)
	    {
	      /* TGA format */
	      case IMGREF_IMGFMT_TGA:
		td = (tga_data_struct *)calloc(
		    1, sizeof(tga_data_struct)
		);
		if(td == NULL)
		{
		    imgref->lib_data = NULL;
		    imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;
		    return(-3);
		}
		imgref->lib_data = td;

		/* Open the image */
		status = TgaReadFromFile(
		    imgref->path,
		    td,
		    TgaGetDepthForX(
			gui->bits_per_pixel,
			gui->depth
		    )
		);
		if(status != TgaSuccess)
		{
		    char *s;
		    const char *error_msg;
		    switch(status)
		    {
		      case TgaNoBuffers:
			error_msg = "Out of memory";
			break;
		      case TgaBadHeader:
			error_msg = "Bad header";
			break;
		      case TgaBadValue:
			error_msg = "Bad value";
			break;
		      case TgaNoFile:
			error_msg = "No file";
			break;
		      case TgaNoAccess:
			error_msg = "No access";
			break;
		      default:
			error_msg = "General error";
			break;
		    }
		    s = (char *)malloc(
			40 + STRLEN(imgref->path) +
			STRLEN(error_msg) + 80
		    );
		    sprintf(
			s,
"ImgRef %i: %s: Unable to open the image file: %s.\n",
			imgref_num, imgref->path,
			error_msg
		    );
		    XSWPrintWarning(s);
		    free(s);

		    TgaDestroyData(td);
		    free(td);
		    imgref->lib_data = NULL;
		    imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;

		    return(-1);
		}

		/* Set the image data to point to the image data on
		 * the library
		 */
		imgref->image_data = (u_int8_t *)td->data;

		imgref->width = td->width;
		imgref->height = td->height;

		/* Mark the load process as done */
		imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
		break;

	      /* Unknown format */
	      default:
		if(imgref->img_fmt > -1)
		{
		    char *s = (char *)malloc(40 + 40 + 80);
		    sprintf(
			s,
"ImgRef %i: Unsupported image format `%i'.\n",
			imgref_num, imgref->img_fmt
		    );
		    XSWPrintWarning(s);
		    free(s);
		}
		break;
	    }

	    /* Sanitize total frames */
	    if(imgref->total_frames > IMGREF_MAX_FRAMES)
		imgref->total_frames = IMGREF_MAX_FRAMES;
	    if(imgref->total_frames < 1)
		imgref->total_frames = 1;

	    /* Calculate frame width and height */
	    imgref->fwidth = imgref->width;
	    imgref->fheight = (int)imgref->height /
		(int)imgref->total_frames;
	}

	return(0);
}

/*
 *	Unloads the ImgRef's image and marks it as not loaded.
 */
void ImgRefUnload(
	imgref_struct **imgrefs_list, const int nimgrefs,
	const int imgref_num
)
{
	bool_t deleted_image_data = False;
	void *image_data;
	imgref_struct *imgref = ((imgref_num >= 0) && (imgref_num < nimgrefs)) ?
	    imgrefs_list[imgref_num] : NULL;
	if(imgref == NULL)
	    return;

	image_data = imgref->image_data;

	/* Unload by the image format type
	 *
	 * If the lib_data is NULL and image_data is not NULL then
	 * it indicates that this imgref's image data is shared and
	 * that the actual (master) image data is on another imgref
	 * who's lib_data and image_data are both not NULL
	 */
	switch(imgref->img_fmt)
	{
	  case IMGREF_IMGFMT_TGA:
	    if(imgref->lib_data != NULL)
	    {
	        TgaDestroyData((tga_data_struct *)imgref->lib_data);
		free(imgref->lib_data);
		imgref->lib_data = NULL;

		deleted_image_data = True;
	    }
	    break;

	  default:
	    if(imgref->lib_data != NULL)
	    {
		free(imgref->lib_data);
		imgref->lib_data = NULL;

		deleted_image_data = True;
	    }
	    break;
	}

	/* Was an image deleted? */
	if(deleted_image_data)
	{
	    int i;
	    imgref_struct *imgref2;

	    /* Reset all imgrefs who's image_data refers to this
	     * imgref's image_data (including this imgref itself)
	     */
	    for(i = 0; i < nimgrefs; i++)
	    {
		imgref2 = imgrefs_list[i];
		if(imgref2 == NULL)
		    continue;

		if(imgref2->image_data == image_data)
		    imgref2->image_data = NULL;
	    }
	}

	/* Mark this imgref's image as not loaded */
	imgref->image_data = NULL;
	imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;
	imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
}

/*
 *	Deletes all the imgrefs in the list and the list itself.
 */
void ImgRefClear(imgref_struct ***imgrefs_list, int *nimgrefs)
{
	int i, n;

	if((imgrefs_list == NULL) || (nimgrefs == NULL))
	    return;

	n = *nimgrefs;
	for(i = 0; i < n; i++)
	{
	    ImgRefReset(*imgrefs_list, n, i);
	    ImgRefDelete((*imgrefs_list)[i]);
	}
	free(*imgrefs_list);
	*imgrefs_list = NULL;
	*nimgrefs = 0;
}

/*
 *	Unloads all imgrefs that are not being used by any objects.
 */
void ImgRefReclaimMemory(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	xsw_object_struct **objs_list, const int nobjs,
	xsw_particle_struct **particles_list, const int nparticles
)
{
	char in_use;
	int i, n;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	xsw_particle_struct *particle;

	if((imgrefs_list == NULL) || (nimgrefs == NULL))
	    return;

	if(*nimgrefs < 0) 
	    *nimgrefs = 0;

	/* Iterate through the ImgRefs list */
	for(i = 0; i < *nimgrefs; i++)
	{
	    imgref = (*imgrefs_list)[i];
	    if(imgref == NULL)
		continue;

	    /* Skip the default imgref */
	    if(i == IMGREF_DEFAULT_IMGREF)
		continue;

	    /* This imgref's image must stay loaded? */
	    if(imgref->options & IMGREF_STAY_LOADED)
		continue;

	    /* Not loaded? */
	    if(!ImgRefIsLoaded(imgref))
		continue;

	    in_use = 0;

	    if(!in_use)
	    {
		/* Iterate through the objects list */
		for(n = 0; n < nobjs; n++)
		{
		    obj = objs_list[n];
		    if(obj == NULL)
			continue;

		    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
			continue;

		    /* Is this object using this ImgRef? */
		    if(obj->imgref == i)
		    {
			in_use = 1;
			break;
		    }
		}
	    }

	    if(!in_use)
	    {
		/* Iterate through the particles list */
		for(n = 0; n < nparticles; n++)
		{
		    particle = particles_list[n];
		    if(particle == NULL)
			continue;

		    if(particle->type <= XSW_PARTICLE_TYPE_UNKNOWN)
			continue;

		    if(particle->imgref_num == i)
		    {
			in_use = 1;
			break;
		    }
		}
	    }

	    /* Unload this ImgRef if it is not in use */
	    if(!in_use)
		ImgRefUnload(*imgrefs_list, *nimgrefs, i);
	}
}

/*
 *	Manages the list of ImgRefs.
 *
 *	Continues the loading of each imgref's image as needed when
 *	passive loading is specified.
 *
 *	Updates each imgref's lights.
 */
void ImgRefManage(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const time_t cur_ms, const time_t cur_sec, const float tc,
	void (*load_completed_cb)(const int, void *),
	void *load_completed_data
)
{
	int	i, p, status, load_completed,
		pixels_loaded = 0;
	imgref_struct *imgref;
	imgref_light_struct *light;
	osw_gui_struct *gui = &osw_gui[0];
	xsw_option_struct *opt = &option;

	if((imgrefs_list == NULL) || (nimgrefs == NULL))
	    return;

	/* Iterate through the ImgRefs list */
	for(i = 0; i < *nimgrefs; i++)
	{
	    imgref = (*imgrefs_list)[i];
	    if(imgref == NULL)
		continue;

	    /* Update lights */
	    for(p = 0; p < imgref->nlights; p++)
	    {
		light = imgref->light[p];
		if(light == NULL)
		    continue;

		/* Strobe? */
		if((light->strobe_off_int > 0l) &&
		   (light->strobe_on_int > 0l)
		)
		{
		    if(light->strobe_state)
		    {
			/* Strobe is in on interval */

			if(light->strobe_next <= cur_ms)
			{
			    /* Strobe just turned off */
			    light->strobe_state = 0l;

			    light->strobe_next = cur_ms +
				light->strobe_off_int;
			}
		    }
		    else
		    {
			/* Strobe is in off interval */
			if(light->strobe_next <= cur_ms)
			{
			    /* Strobe just turned on */
			    light->strobe_state = 1;

			    light->strobe_next = cur_ms +
				light->strobe_on_int;
			}
		    }
		}
	    }


	    /* ***************************************************** */
	    /* Passive image loading */

	    /* Can we still load more pixels and is this imgref's
	     * image in the loading process?
	     */
	    if((pixels_loaded < opt->async_image_pixels) &&
	       (imgref->load_progress == IMGREF_LOAD_PROGRESS_LOADING)
	    )
	    {
		tga_data_struct *td;

		/* Library data not allocated? */
		if(imgref->lib_data == NULL)
		    continue;

		/* Reset load_completed marker */
		load_completed = 0;

		/* Continue loading by format */
		switch(imgref->img_fmt)
		{
		  case IMGREF_IMGFMT_TGA:
		    td = (tga_data_struct *)imgref->lib_data;
		    status = TgaReadPartialFromFile(
			td,
			TgaGetDepthForX(
			    gui->bits_per_pixel,
			    gui->depth
			),
			IMGREF_ASYNC_PIXEL_LOAD_MAX_ITERATION
		    );
		    if(status != TgaSuccess)
		    {
			/* Load error, reset this imgref and load it
			 * as the default
			 */
			ImgRefReset(*imgrefs_list, *nimgrefs, i);
			ImgRefLoadAsDefault(imgrefs_list, nimgrefs, i);

			load_completed = 1;
			imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
		    }
		    /* Done loading? */
		    if(td->cur_load_pixel < 0)
		    {
			/* Image is done loading, mark it as such */
			load_completed = 1;
			imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
		    }
		    /* Increment pixels loaded */
		    pixels_loaded += IMGREF_ASYNC_PIXEL_LOAD_MAX_ITERATION;
		    break;

		  default:
		    imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
		    break;
		}

		/* Has this imgref's load progress just completed? */
		if(load_completed)
		{
		    /* Report load complete */
		    if(load_completed_cb != NULL)
			load_completed_cb(
			    i,
			    load_completed_data
			);
		}


	    }
	}
}
