#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/cfgfmt.h"
#include "../include/string.h"
#include "../include/fio.h"
#include "../include/unvmath.h"

#include "../include/imgrefs.h"
#include "imgrefsfio.h"


imgref_struct *ImgRefNew(void);
void ImgRefDelete(imgref_struct *imgref);

imgref_light_struct *ImgRefLightNew(void);
void ImgRefLightDelete(imgref_light_struct *light);

int ImgRefsFileOpen(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const char *path,
	void (*progress_cb)(void *, unsigned long, unsigned long),
	void *progress_data,
	void (*print_warning)(const char *s)
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a new ImgRef.
 */
imgref_struct *ImgRefNew(void)
{
	return(IMGREF(calloc(1, sizeof(imgref_struct))));
}

/*
 *	Deletes the ImgRef.
 *
 *	The loaded image must be deleted prior to this call.
 */
void ImgRefDelete(imgref_struct *imgref)
{
	int i;

	if(imgref == NULL)
	    return;

#if 0
	/* Warn if the image was not loaded */
	if(imgref->lib_data != NULL)
	{

	}
#endif

	free(imgref->path);
	free(imgref->name);
	free(imgref->description);

	for(i = 0; i < imgref->nlights; i++)
	    ImgRefLightDelete(imgref->light[i]);
	free(imgref->light);

	free(imgref);
}


/*
 *	Creates a new ImgRef light.
 */
imgref_light_struct *ImgRefLightNew(void)
{
	return(IMGREF_LIGHT(calloc(1, sizeof(imgref_light_struct))));
}

/*
 *	Deletes the ImgRef light.
 */
void ImgRefLightDelete(imgref_light_struct *light)
{
	if(light == NULL)
	    return;

	free(light);
}


/*
 *	Opens the ImgRefs from file.
 */
int ImgRefsFileOpen(
	imgref_struct ***imgrefs_list, int *nimgrefs,
	const char *path,
	void (*progress_cb)(void *, unsigned long, unsigned long),
	void *progress_data,
	void (*print_warning)(const char *s)
)
{
	int	lines_read = 0,
		imgrefs_loaded = 0,
		imgref_num = -1;
	unsigned long file_size = 0l;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	struct stat stat_buf;
	imgref_struct *imgref = NULL;

	/* Report initial progress */
	if(progress_cb != NULL)
	    progress_cb(progress_data, 0l, file_size);

	if((imgrefs_list == NULL) || (nimgrefs == NULL) || STRISEMPTY(path))
	    return(-2);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for reading");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    if(print_warning != NULL)
		print_warning(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;

	    if(progress_cb != NULL)
		progress_cb(progress_data, (unsigned long)ftell(fp), file_size);
	}

#define WARN_NO_SUCH_PARM(_parm_)               {       \
 char *s = (char *)malloc(                              \
  STRLEN(path) + 40 + STRLEN(_parm_) + 80               \
 );                                                     \
 sprintf(                                               \
  s,                                                    \
"%s: Line %i: Unsupported parameter `%s'.",             \
  path, lines_read, (_parm_)                            \
 );                                                     \
 if(print_warning != NULL)				\
  print_warning(s);					\
 free(s);                                               \
}
#define WARN_NO_SUCH_FILE(_v_,_error_code_)     {       \
 const int error_code = (_error_code_);                 \
 char *s, *error_msg = STRDUP(strerror(error_code));    \
 if(error_msg == NULL)                                  \
  error_msg = strdup("no such file");                   \
 *error_msg = toupper(*error_msg);                      \
							\
 s = (char *)malloc(                                    \
  STRLEN(path) + 40 + STRLEN(parm) + STRLEN(_v_) +      \
  STRLEN(error_msg) + 80                                \
 );                                                     \
 sprintf(                                               \
  s,                                                    \
"%s: Line %i: %s: Warning: %s: %s.",                    \
  path, lines_read, parm, (_v_), error_msg              \
 );                                                     \
 if(print_warning != NULL)				\
  print_warning(s);					\
 free(s);                                               \
							\
 free(error_msg);                                       \
}

	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';

	    /* Report progress */
	    if(progress_cb != NULL)
		progress_cb(progress_data, (unsigned long)ftell(fp), file_size);


	    /* BeginImgRef */
	    if(!strcasecmp(parm, "BeginImgRef") ||
	       !strcasecmp(parm, "BeginISRefEntry")
	    )
	    {
		/* Increment imgrefs_loaded */
		imgrefs_loaded++;

		/* Get the imgref's index */
		imgref_num = ATOI(val);

		/* Already exists? */
		if((imgref_num >= 0) && (imgref_num < *nimgrefs))
		{
		    imgref = (*imgrefs_list)[imgref_num];
		    if((imgref != NULL) ?
			!STRISEMPTY(imgref->path) : 0
		    )
		    {
			char *s = (char *)malloc(
			    STRLEN(path) + 40 + STRLEN(parm) +
			    40 + 80
			);
			sprintf(
			    s,
"%s: Line %i: %s: Warning: ImgRef %i redefined.",
			    path, lines_read, parm, imgref_num
			);
			if(print_warning != NULL)
			    print_warning(s);
			free(s);
		    }
		}

		/* Create a new */
		if((imgref_num >= 0) && (imgref_num < *nimgrefs))
		{
		    /* Pointer already exists, create the imgref as needed */
		    imgref = (*imgrefs_list)[imgref_num];
		    if(imgref == NULL)
			(*imgrefs_list)[imgref_num] = imgref = ImgRefNew();
		}
		else
		{
		    /* Need to allocate more pointers */
		    const int prev_total = *nimgrefs;
		    int i;

		    if(imgref_num < 0)
			imgref_num = MAX(*nimgrefs, 0);

		    if(imgref_num >= *nimgrefs)
			*nimgrefs = imgref_num + 1;

		    *imgrefs_list = (imgref_struct **)realloc(
			*imgrefs_list,
			(*nimgrefs) * sizeof(imgref_struct *)
		    );
		    for(i = prev_total; i < *nimgrefs; i++)
			(*imgrefs_list)[i] = NULL;

		    (*imgrefs_list)[imgref_num] = imgref = ImgRefNew();
		}
		if(imgref == NULL)
		{
		    char *s = (char *)malloc(
			STRLEN(path) + 40 + STRLEN(parm) +
			40 + 80
		    );
		    sprintf(
			s,
"%s: Line %i: %s: Warning: Unable to create ImgRef %i.",
			path, lines_read, parm, imgref_num
		    );
		    if(print_warning != NULL)
			print_warning(s);
		    free(s);
		    break;
		}

		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* Report progress */
		    if(progress_cb != NULL)
			progress_cb(progress_data, (unsigned long)ftell(fp), file_size);


		    /* Name */
		    if(!strcasecmp(parm, "Name"))
		    {
			free(imgref->name);
			imgref->name = STRDUP(val);
		    }
		    /* Description */
		    else if(!strcasecmp(parm, "Description") ||
			    !strcasecmp(parm, "Desc")
		    )
		    {
			free(imgref->description); 
			imgref->description = STRDUP(val);
		    }
		    /* Path */
		    else if(!strcasecmp(parm, "Path") ||
		            !strcasecmp(parm, "FileName")
		    )
		    {
			free(imgref->path);
			imgref->path = STRDUP(val);
		    }
		    /* LoadNow */
		    else if(!strcasecmp(parm, "LoadNow"))
		    {
			/* Same as IMGREF_STAY_LOADED */
			imgref->options |= IMGREF_STAY_LOADED;
		    }
		    /* NoImage */
		    else if(!strcasecmp(parm, "NoImage"))
		    {
			imgref->options |= IMGREF_NO_IMAGE;
		    }
		    /* StayLoaded */
		    else if(!strcasecmp(parm, "StayLoaded"))
		    {
			imgref->options |= IMGREF_STAY_LOADED;
		    }
		    /* MergeMode */
		    else if(!strcasecmp(parm, "MergeMode"))
		    {
			if(isdigit(*val))
			{
			    imgref->merge_mode = atoi(val);
			}
			else
			{
			    if(!strcasecmp(val, "Additive"))
				imgref->merge_mode =
				    IMGREF_MERGE_ADDITIVE;
			    else if(!strcasecmp(val, "Subtractive"))
				imgref->merge_mode =
				    IMGREF_MERGE_SUBTRACTIVE;
			    else	/* "Normal" */
				imgref->merge_mode =
				    IMGREF_MERGE_NORMAL;
			}
		    }
		    /* HasTransparency */
		    else if(!strcasecmp(parm, "HasTransparency"))
		    {
			imgref->options |= IMGREF_HAS_TRANSPARENCY;
		    }
		    /* FrameDeterminat */
		    else if(!strcasecmp(parm, "FrameDeterminat"))
		    {
			if(isdigit(*val))
			{
			    imgref->frame_determinant = atoi(val);
			}
			else
			{
			    if(!strcasecmp(val, "ByAnimation"))
				imgref->frame_determinant =
				    IMGREF_FRAME_DETERMINE_BY_ANIMATION;
			    else	/* Default to "ByHeading" */
				imgref->frame_determinant =
				    IMGREF_FRAME_DETERMINE_BY_HEADING;
			}
		    }
		    /* LayerPlacement */
		    else if(!strcasecmp(parm, "LayerPlacement"))
		    {
			if(isdigit(*val))
			{
			    imgref->layer_placement = atoi(val);
			}
			else
			{
			    if(!strcasecmp(val, "BackgroundTiled") ||
			       !strcasecmp(val, "BgTiled")
			    )
				imgref->layer_placement =
				    IMGREF_LAYER_BG_TILED;
			    else if(!strcasecmp(val, "BackgroundStatic") ||
				    !strcasecmp(val, "BgStatic")
			    )
				imgref->layer_placement =
				    IMGREF_LAYER_BG_STATIC;
			    else	/* Default to Foreground */
				imgref->layer_placement =
				    IMGREF_LAYER_FG;
			}
		    }
		    /* Effects */
		    else if(!strcasecmp(parm, "Effects"))
		    {
			const char *s3;

			imgref->effects = 0;

			strtoupper(val);
			s3 = strstr(val, "STARGLOW");
			if(s3 != NULL)
			    imgref->effects |= IMGREF_EFFECTS_STARGLOW;
			s3 = strstr(val, "FADEINGLOW");
			if(s3 != NULL)
			    imgref->effects |= IMGREF_EFFECTS_FADEINGLOW;
			s3 = strstr(val, "FADEOUTGLOW");
			if(s3 != NULL)
			    imgref->effects |= IMGREF_EFFECTS_FADEOUTGLOW;
		    }
		    /* Magnification */
		    else if(!strcasecmp(parm, "Magnification"))
		    {
			imgref->magnification = atof(val);
			if(imgref->magnification < 1)
			{
			    char *s = (char *)malloc(
				STRLEN(path) + 40 + STRLEN(parm) +
				STRLEN(val) + 80
			    );
			    sprintf(
				s,
"%s: Line %i: %s: Warning: Value `%s' is too small.",
				path, lines_read, parm, val
			    );
			    if(print_warning != NULL)
				print_warning(s);
			    free(s);

			    imgref->magnification = 1;
			}
		    }

		    /* TotalFrames */
		    else if(!strcasecmp(parm, "TotalFrames") ||
			    !strcasecmp(parm, "Frames")
		    )
		    {
			imgref->total_frames = atoi(val);
			if(imgref->total_frames < 1)
			{
			    char *s = (char *)malloc(
				STRLEN(path) + 40 + STRLEN(parm) +
				STRLEN(val) + 80
			    );
			    sprintf(
				s,
"%s: Line %i: %s: Warning: Value `%s' is too small.",
				path, lines_read, parm, val
			    );
			    if(print_warning != NULL)
				print_warning(s);
			    free(s);

			    imgref->total_frames = 1;
			}
			else if(imgref->total_frames > IMGREF_MAX_FRAMES)
			{
			    char *s = (char *)malloc(
				STRLEN(path) + 40 + STRLEN(parm) +
				STRLEN(val) + 40 + 80
			    );
			    sprintf(
				s,
"%s: Line %i: %s: Warning: Value `%s' is too big (IMGREF_MAX_FRAMES=%i).",
				path, lines_read, parm, val,
				IMGREF_MAX_FRAMES
			    );
			    if(print_warning != NULL)
				print_warning(s);
			    free(s);

			    imgref->total_frames = IMGREF_MAX_FRAMES;
			}
		    }
		    /* Width */
		    else if(!strcasecmp(parm, "Width"))
		    {
			/* Calculated by information on image file */
		    }
		    /* Height */
		    else if(!strcasecmp(parm, "Height"))
		    {
			/* Calculated by information on image file */
		    }
		    /* FrameWidth */
		    else if(!strcasecmp(parm, "FrameWidth"))
		    {
			/* Calculated by information on image file */
		    }
		    /* FrameHeight */
		    else if(!strcasecmp(parm, "FrameHeight"))
		    {
			/* Calculated by information on image file */
		    }

		    /* ************************************************* */
		    /* BeginPointLight */
		    else if(!strcasecmp(parm, "BeginPointLight"))
		    {
			imgref_light_struct *light;
			const int lignt_num = MAX(imgref->nlights, 0);
			imgref->nlights = lignt_num + 1;
			imgref->light = (imgref_light_struct **)realloc(
			    imgref->light,
			    imgref->nlights * sizeof(imgref_light_struct *)
			);
			if(imgref->light == NULL)
			{
			    imgref->nlights = 0;
			    break;
			}

			imgref->light[lignt_num] = light = ImgRefLightNew();
			if(light == NULL)
			    break;

			while(1)
			{
			    /* Get the next line */
			    free(s);
			    s = FReadNextLineAllocCount(
				fp, UNIXCFG_COMMENT_CHAR, &lines_read
			    );
			    if(s == NULL)
				break;

			    /* Get the parameter from the line */
			    s2 = StringCfgParseParm(s);
			    if(s2 == NULL)
				continue;
			    strncpy(parm, s2, sizeof(parm));
			    parm[sizeof(parm) - 1] = '\0';   

			    /* Get the value from the line */
			    s2 = StringCfgParseValue(s);
			    if(s2 == NULL)
				s2 = "0";
			    strncpy(val, s2, sizeof(val));
			    val[sizeof(val) - 1] = '\0';

			    /* Report progress */
			    if(progress_cb != NULL)
				progress_cb(progress_data, (unsigned long)ftell(fp), file_size);


			    /* Theta */
			    if(!strcasecmp(parm, "Theta"))
			    {
				light->theta = SANITIZERADIANS(ATOF(val));
			    }
			    /* Radius */
			    else if(!strcasecmp(parm, "Radius"))
			    {
			        light->radius = MAX(ATOF(val), 0.0f);
			    }

			    /* ColorAlpha */
			    else if(!strcasecmp(parm, "ColorAlpha"))
			    {
			        light->a = CLIP(ATOI(val), 0x00, 0xFF);
			    }
			    /* ColorRed */
			    else if(!strcasecmp(parm, "ColorRed"))
			    {
				light->r = CLIP(ATOI(val), 0x00, 0xFF);
			    }
			    /* ColorGreen */
			    else if(!strcasecmp(parm, "ColorGreen"))
			    {
				light->g = CLIP(ATOI(val), 0x00, 0xFF);
			    }
			    /* ColorBlue */
			    else if(!strcasecmp(parm, "ColorBlue"))
			    {
				light->b = CLIP(ATOI(val), 0x00, 0xFF);
			    }

			    /* StrobeOffInterval */
			    else if(!strcasecmp(parm, "StrobeOffInterval"))
			    {
				light->strobe_off_int = MAX(ATOL(val), 0l);
			    }
			    /* StrobeOnInterval */
			    else if(!strcasecmp(parm, "StrobeOnInterval"))
			    {
			        light->strobe_on_int = MAX(ATOL(val), 0l);
			    }

			    /* EndPointLight */
			    else if(!strcasecmp(parm, "EndPointLight"))
			    {
				break;
			    }
			    /* Unsupported parameter */
			    else
			    {
				WARN_NO_SUCH_PARM(parm);
			    }
			}
		    }

		    /* EndImgRef */
		    else if(!strcasecmp(parm, "EndImgRef") ||
			    !strcasecmp(parm, "EndISRefEntry")
		    )
		    {
			/* Check and warn about any errors */
			if(!(imgref->options & IMGREF_NO_IMAGE))
			{
			    if(imgref->path == NULL)
			    {
				char *s = (char *)malloc(
				    STRLEN(path) + 40 + 40 + 80
				);
				sprintf(
				    s,
"%s: Line %i: Warning: Imgref %i has no filename.",
				    path, lines_read, imgref_num
				);
				if(print_warning != NULL)
				    print_warning(s);
				free(s);
			    }
			}

			/* End of imgref, so break */
			break;
		    }

		    /* Unsupported parameter */
		    else
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }

	    /* Unsupported parameter */
	    else     
	    {           
		WARN_NO_SUCH_PARM(parm);
	    }
	}
	free(s);

#undef WARN_NO_SUCH_PARM
#undef WARN_NO_SUCH_FILE

	/* Close the file */
	fclose(fp);

	/* Report final progress */
	if(progress_cb != NULL)
	    progress_cb(progress_data, file_size, file_size);

	return(0);
}
