#ifdef JS_SUPPORT

#include <stdlib.h>
#include <string.h>
#if defined(HAVE_LIBJSW)
# include <jsw.h>
#endif

#include "jsmaps.h"


xsw_jsmap_struct *JSMapNew(void);

void JSMapInitializeJoystick(xsw_jsmap_struct *jsm);
void JSMapSyncWithData(xsw_jsmap_struct *jsm);
void JSMapShutdownJoystick(xsw_jsmap_struct *jsm);

void JSMapDelete(xsw_jsmap_struct *jsm);
void JSMapsListClear(
	xsw_jsmap_struct **list, const int total
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a new jsmap.
 */
xsw_jsmap_struct *JSMapNew(void)
{
	xsw_jsmap_struct *jsm = XSW_JSMAP(calloc(
	    1, sizeof(xsw_jsmap_struct)
	));
	if(jsm == NULL)
	    return(NULL);

	jsm->device_name = NULL;
#if defined(HAVE_LIBJSW)
	jsm->jsd.fd = -1;
#endif
	jsm->axis = NULL;
	jsm->total_axises = 0;
	jsm->button = NULL;
	jsm->total_buttons = 0;

	return(jsm);
}


/*
 *	Initializes the jsmap's joystick.
 *
 *	Returns 0 on success or non-zero on error.
 */
int JSMapInitializeJoystick(
	xsw_jsmap_struct *jsm,
	const char *device_name,
	const char *calibration_file
)
{
	if(jsm == NULL)
	    return(-2);

	/* Set the new joystick device name as needed */
	if(jsm->device_name != device_name)
	{
	    free(jsm->device_name);
	    jsm->device_name = STRDUP(device_name);
	}

	/* Shutdown the joystick first */
	JSMapShutdownJoystick(jsm);

#if defined(HAVE_LIBJSW)
	{
	    /* Initialize this joystick */
	    js_data_struct *jsd = &jsm->jsd;
	    const int status = JSInit(
		jsd, device_name, calibration_file,
		JSFlagNonBlocking
	    );
	    if(status != JSSuccess)
	    {
		/* An error occured while initializing the joystick */
		JSClose(jsd);
		return(-1);
	    }

	    JSMapSyncWithData(jsm);
	}

	return(0);
#else
# warning JSMapInitializeJoystick() No joystick library specified to initialize the joystick.
	return(-1);
#endif
}

/*
 *	Updates the number of axises and buttons on the jsmap to match
 *	the number of axises and buttons on the jsmap's data.
 */
void JSMapSyncWithData(xsw_jsmap_struct *jsm)
{
	int i;
	js_data_struct *jsd;

	if(jsm == NULL)
	    return;

	/* Get the joystick data */
	jsd = &jsm->jsd;

	if(jsm->total_axises < 0)
	    jsm->total_axises = 0;
	if(jsd->total_axises < 0)
	    jsd->total_axises = 0;

	/* Too many axises? */
	if(jsm->total_axises > jsd->total_axises)
	{
	    /* Delete excess axises */
	    for(i = jsm->total_axises - 1; i >= jsd->total_axises; i--)
		free(jsm->axis[i]);

	    jsm->total_axises = jsd->total_axises;
	    if(jsm->total_axises > 0)
	    {
		jsm->axis = (xsw_jsmap_axis_struct **)realloc(
		    jsm->axis,
		    jsm->total_axises * sizeof(xsw_jsmap_axis_struct *)
		);
		if(jsm->axis == NULL)
		{
		    jsm->total_axises = 0;
		    return;
		}
	    }
	    else
	    {
		free(jsm->axis);
		jsm->axis = NULL;  
		jsm->total_axises = 0;
	    }
	}
	/* Too few axises? */
	else if(jsm->total_axises < jsd->total_axises)
	{
	    /* Create additional axises */
	    xsw_jsmap_axis_struct *axis;

	    jsm->axis = (xsw_jsmap_axis_struct **)realloc(
		jsm->axis,
		jsd->total_axises * sizeof(xsw_jsmap_axis_struct *)
	    );
	    if(jsm->axis == NULL)
	    {
		jsm->total_axises = 0;
		return;
	    }
	    for(i = jsm->total_axises; i < jsd->total_axises; i++)
	    {
		jsm->axis[i] = axis = XSW_JSMAP_AXIS(calloc(
		    1, sizeof(xsw_jsmap_axis_struct)
		));
		if(axis == NULL)
		    continue;

		/* Reset values */
		axis->op = JSMAP_AXIS_OP_NONE;
	    }

	    jsm->total_axises = jsd->total_axises;
	}


	if(jsm->total_buttons < 0)
	    jsm->total_buttons = 0;
	if(jsd->total_buttons < 0)
	    jsd->total_buttons = 0;

	/* Too many buttons? */
	if(jsm->total_buttons > jsd->total_buttons)
	{
	    /* Delete excess buttons */
	    for(i = jsm->total_buttons - 1; i >= jsd->total_buttons; i--)
		free(jsm->button[i]);

	    jsm->total_buttons = jsd->total_buttons;
	    if(jsm->total_buttons > 0)
	    {
		jsm->button = (xsw_jsmap_button_struct **)realloc(
		    jsm->button,
		    jsm->total_buttons * sizeof(xsw_jsmap_button_struct *)
		);
		if(jsm->button == NULL) 
		{
		    jsm->total_buttons = 0;
		    return;
		}
	    }
	    else
	    {
		free(jsm->button);
		jsm->button = NULL;
		jsm->total_buttons = 0;
	    }
	}
	/* Too view buttons? */
	else if(jsm->total_buttons < jsd->total_buttons)
	{
	    /* Create additional buttons */
	    xsw_jsmap_button_struct *btn;

	    jsm->button = (xsw_jsmap_button_struct **)realloc(
		jsm->button,
		jsd->total_buttons * sizeof(xsw_jsmap_button_struct *)
	    );  
	    if(jsm->button == NULL)
	    {
		jsm->total_buttons = 0;
		return;
	    }         
	    for(i = jsm->total_buttons; i < jsd->total_buttons; i++)
	    {
		jsm->button[i] = btn = XSW_JSMAP_BUTTON(calloc(
		    1, sizeof(xsw_jsmap_button_struct)
		));
		if(btn == NULL)
		    continue;
	
		/* Reset values */
		btn->keycode = 0;
		btn->key_state = 0l;
		btn->state = False;
	    }

	    /* Set new total buttons on joystick map */
	    jsm->total_buttons = jsd->total_buttons;
	}
}

/*
 *	Shuts down the jsmap's joystick device.
 */
void JSMapShutdownJoystick(xsw_jsmap_struct *jsm)
{
	if(jsm == NULL)
	    return;

#if defined(HAVE_LIBJSW)
	JSClose(&jsm->jsd);
#endif
}


/*
 *	Shuts down the jsmap's joystick and deletes the jsmap.
 */
void JSMapDelete(xsw_jsmap_struct *jsm)
{
	int i;

	if(jsm == NULL)
	    return;

	JSMapShutdownJoystick(jsm);

	free(jsm->device_name);

	for(i = 0; i < jsm->total_axises; i++)
	    free(jsm->axis[i]);
	free(jsm->axis);

	for(i = 0; i < jsm->total_buttons; i++)
	    free(jsm->button[i]);
	free(jsm->button);

	free(jsm);
}

/*
 *	Deletes the list of jsmaps.
 */
void JSMapsListClear(
	xsw_jsmap_struct **list, const int total   
)
{
	int i;

	for(i = 0; i < total; i++)
	    JSMapDelete(list[i]);
	free(list);
}


#endif	/* JS_SUPPORT */
