#include <stdlib.h>
#include <string.h>

#include "../include/string.h"

#include "keymaps.h"
#include "keymapwin.h"
#include "xsw.h"

#include "../include/oswkeyvals.h"


static int KeymapWinAppendListRow(
	xsw_keymap_win_struct *kmw,
	clist_struct *clist,
	const char *name,
	keycode_t keycode, key_state_t key_state
);
static keycode_t KeymapWinGetKeycode(clist_struct *clist, int row_num);
static void KeymapWinSetKeycode(
	xsw_keymap_win_struct *kmw,
	clist_struct *clist, int row_num,
	keycode_t keycode, key_state_t key_state
);

static void KeymapWinUnfocusAll(xsw_keymap_win_struct *kmw);

static void KeymapWinFetch(xsw_keymap_win_struct *kmw);
static void KeymapWinApply(xsw_keymap_win_struct *kmw);
static void KeymapWinSetDefault(xsw_keymap_win_struct *kmw, int item_num);

static void KeymapWinPromptKey(xsw_keymap_win_struct *kmw);
static int KeymapWinCheckDups(xsw_keymap_win_struct *kmw, bool_t warn);

static void KeymapWinResizeCB(void *widget, void *data);
static void KeymapWinDrawCB(void *widget, void *data);

static void KeymapWinCListDataDestroyCB(void *data);
static void KeymapWinCListActivateCB(clist_struct *clist, void *data);
static void KeymapWinCListSelectCB(
	clist_struct *clist, int row_num, void *data
);
static void KeymapWinCListUnselectCB(
	clist_struct *clist, int row_num, void *data  
);
static void KeymapWinOkPBCB(push_button_struct *btn, void *data);
static void KeymapWinApplyPBCB(push_button_struct *btn, void *data);
static void KeymapWinCancelPBCB(push_button_struct *btn, void *data);
static void KeymapWinScanKeyPBCB(push_button_struct *btn, void *data);
static void KeymapWinDefaultPBCB(push_button_struct *btn, void *data);
static void KeymapWinDefaultAllPBCB(push_button_struct *btn, void *data);

xsw_keymap_win_struct *KeymapWinNew(void);
void KeymapWinResize(xsw_keymap_win_struct *kmw);
void KeymapWinQueueResize(xsw_keymap_win_struct *kmw);
void KeymapWinDraw(xsw_keymap_win_struct *kmw, const int amount);
void KeymapWinQueueDraw(xsw_keymap_win_struct *kmw, const int amount);
int KeymapWinManage(xsw_keymap_win_struct *kmw, event_t *event);
void KeymapWinMap(xsw_keymap_win_struct *kmw, win_t toplevel);
void KeymapWinUnmap(xsw_keymap_win_struct *kmw);
void KeymapWinDelete(xsw_keymap_win_struct *kmw);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sizes (in pixels):
 */
#define KEYMAP_WIN_DEF_WIDTH		400
#define KEYMAP_WIN_DEF_HEIGHT		480

#define KEYMAP_WIN_MIN_WIDTH		380
#define KEYMAP_WIN_MIN_HEIGHT		350

#define KEYMAP_WIN_MARGIN		10

#define KEYMAP_WIN_PROMPT_HEIGHT	80

#define KEYMAP_WIN_BTN_WIDTH		70
#define KEYMAP_WIN_BTN_HEIGHT		28

#define KMW_DEF_LIST_KEYCODE_POS	240

#define KEYMAP_WIN_LIST_TO_END_MARGIN	180


/*
 *	Keymap list draw positions (in pixels):
 */
#define KEYMAP_WIN_LIST_POS_NAME	KEYMAP_WIN_MARGIN
#define KEYMAP_WIN_LIST_POS_KEYCODE	240


/*
 *	Appends a new keymap to the list.
 *
 *	Called by KeymapWinFetch().
 */
static int KeymapWinAppendListRow(
	xsw_keymap_win_struct *kmw,
	clist_struct *clist,
	const char *name,
	keycode_t keycode, key_state_t key_state
)
{
	const char *keycode_name;
	char *s, *v[2];
	int row_num;
	keymap_struct *km;

	/* Append row */
	row_num = CListRowAppend(clist);
	if(row_num < 0)
	    return(-1);

	v[0] = STRDUP(name);

	/* Add keycode to row (colum 1) */
	keycode_name = OSWGetKeyCodeName(keycode);
	s = (char *)malloc(STRLEN(keycode_name) + 80 + 80 + 80);
#if 1
	sprintf(
	    s,  
	    "%s",
	    keycode_name
	);
#else
	sprintf(
	    s,
	    "%s%s%s%s",
	    (key_state & OSW_KEY_STATE_CTRL) ? "CTRL + " : "",
	    (key_state & OSW_KEY_STATE_ALT) ? "ALT + " : "",
	    (key_state & OSW_KEY_STATE_SHIFT) ? "SHIFT + " : "",
	    keycode_name
	);
#endif
	strtoupper(s);
	v[1] = s;

	CListSetRowTexts(
	    clist, row_num,
	    (const char **)v, sizeof(v) / sizeof(char *)
	);
	free(v[0]);
	free(v[1]);

	/* Set the keycode as the row data */
	km = KEYMAP(calloc(1, sizeof(keymap_struct)));
	if(km != NULL)
	{
	    km->keycode = keycode;
	    km->key_state = key_state;
	    CListRowSetData(
		clist, row_num,
		km, KeymapWinCListDataDestroyCB
	    );
	}

	return(row_num);
}


/*
 *	Gets the keycode from the specified row number.
 */
static keycode_t KeymapWinGetKeycode(clist_struct *clist, int row_num)
{
	keymap_struct *km;

	if(clist == NULL)
	    return(0);

	km = KEYMAP(CListRowGetData(clist, row_num));
	return((km != NULL) ? km->keycode : 0);
}

/*
 *	Sets keycode on row_num to the value of keycode.
 *	Changes the label and client data pointer on the item.
 */
static void KeymapWinSetKeycode(
	xsw_keymap_win_struct *kmw,
	clist_struct *clist, int row_num,
	keycode_t keycode, key_state_t key_state
)
{
	const char *keycode_name;
	char *s;
	keymap_struct *km;

	if(keycode < 0)
	    return;

	/* Set new label */
	keycode_name = OSWGetKeyCodeName(keycode);
	s = (char *)malloc(STRLEN(keycode_name) + 80 + 80 + 80);
#if 1
	sprintf(
	    s,
	    "%s",
	    keycode_name
	);
#else
	sprintf(
	    s,
	    "%s%s%s%s",
	    (key_state & OSW_KEY_STATE_CTRL) ? "CTRL + " : "",
	    (key_state & OSW_KEY_STATE_ALT) ? "ALT + " : "",
	    (key_state & OSW_KEY_STATE_SHIFT) ? "SHIFT + " : "",
	    keycode_name
	);
#endif
	strtoupper(s);
	CListSetItemText(
	    clist,
	    row_num, 1,			/* Row, column */
	    s
	);
	free(s);

	/* Set the row's data as the new keycode */
	km = KEYMAP(CListRowGetData(clist, row_num));
	if(km != NULL)
	{
	    km->keycode = keycode;
	    km->key_state = key_state;
	}
}


/*
 *	Unfocuses all widgets.
 */
static void KeymapWinUnfocusAll(xsw_keymap_win_struct *kmw)
{
	if(kmw == NULL)
	    return;

	WidgetSetFocus(&kmw->list, False);
}


/*
 *	Loads global keymap values into keymap window's colum list.
 *
 *	Any existing keymap entries will be deleted first.
 */
static void KeymapWinFetch(xsw_keymap_win_struct *kmw)
{
	const int m = KeymapsGetTotal();
	int i;
	clist_struct *clist;
	keymap_struct *km;

	if(kmw == NULL)
	    return;

	clist = &kmw->list;

	/* Delete all existing keymap rows */
	CListClear(clist);

	/* Add keymaps to list */
	for(i = 0; i < m; i++)
	{
	    km = KeymapGetPtr(i);
	    if(km == NULL)
		continue;

	    KeymapWinAppendListRow(
		kmw, clist,
	        km->name,
		km->keycode, km->key_state
	    );
	}

	CListAutosizeColumns(clist);
}

/*
 *	Sets global keymap values to that of the local
 *	keymap values of the keymap window.
 */
static void KeymapWinApply(xsw_keymap_win_struct *kmw)
{
	const int m = KeymapsGetTotal();
	int i;
	clist_struct *clist;
	keymap_struct *km;

	if(kmw == NULL) 
	    return; 

	clist = &kmw->list;

	for(i = 0; i < m; i++)
	{
	    km = KeymapGetPtr(i);
	    if(km == NULL)
		continue;

	    km->keycode = KeymapWinGetKeycode(clist, i);
	}
}


/*
 *	Sets the default keycode value for keymap item item_num.
 */
static void KeymapWinSetDefault(xsw_keymap_win_struct *kmw, int item_num)
{
	clist_struct *clist;

#define SET_KEY(_keyval_)	{	\
 KeymapWinSetKeycode(			\
  kmw, clist, item_num,			\
  OSWGetKeyCodeFromKeyVal(_keyval_),	\
  0					\
 );					\
}

	if(kmw == NULL) 
	    return; 

	clist = &kmw->list;

	/* Set default keycode for keymap item */
	switch(item_num)
	{
	  case XSW_KM_HELP:
	    SET_KEY(OSW_KEY_F1);
	    break;
	  case XSW_KM_EXIT:
	    SET_KEY(OSW_KEY_x);
	    break;
	  case XSW_KM_TURN_LEFT:
	    SET_KEY(OSW_KEY_Left);
	    break;
	  case XSW_KM_TURN_RIGHT:
	    SET_KEY(OSW_KEY_Right);
	    break;
	  case XSW_KM_THROTTLE_INC:
	    SET_KEY(OSW_KEY_Up);
	    break;
	  case XSW_KM_THROTTLE_DEC:
	    SET_KEY(OSW_KEY_Down);
	    break;
	  case XSW_KM_THROTTLE_IDLE:
	    SET_KEY(OSW_KEY_Delete);
	    break;
	  case XSW_KM_THROTTLE_MAX:
	    SET_KEY(OSW_KEY_Insert);
	    break;
	  case XSW_KM_FIRE_WEAPON:
	    SET_KEY(OSW_KEY_space);
	    break;

	  case XSW_KM_OMNI_DIR_THRUST:
	    SET_KEY(OSW_KEY_Alt_L);
	    break;
	  case XSW_KM_EXTERNAL_DAMPERS:
	    SET_KEY(OSW_KEY_b);
	    break;

	  case XSW_KM_LIGHTS_VECTOR:
	    SET_KEY(OSW_KEY_l);
	    break;
	  case XSW_KM_LIGHTS_STROBE:
	    SET_KEY(OSW_KEY_p);
	    break;
	  case XSW_KM_LIGHTS_LUMINATION:
	    SET_KEY(OSW_KEY_v);
	    break;

	  case XSW_KM_WEAPON_FREQ:
	    SET_KEY(OSW_KEY_y);
	    break;
	  case XSW_KM_SHIELD:
	    SET_KEY(OSW_KEY_s);
	    break;
	  case XSW_KM_SHIELD_FREQ:
	    SET_KEY(OSW_KEY_f);
	    break;
	  case XSW_KM_DMGCTL:
	    SET_KEY(OSW_KEY_c);
	    break;
	  case XSW_KM_CLOAK:
	    SET_KEY(OSW_KEY_z);
	    break;
	  case XSW_KM_INTERCEPT_SET:
	    SET_KEY(OSW_KEY_i);
	    break;
	  case XSW_KM_HAIL:
	    SET_KEY(OSW_KEY_h);
	    break;
	  case XSW_KM_COM_CHANNEL_SET:
	    SET_KEY(OSW_KEY_k);
	    break;
	  case XSW_KM_SEND_MESSAGE:
	    SET_KEY(OSW_KEY_apostrophe);
	    break;
	  case XSW_KM_IGNORE_OBJECT:
	    SET_KEY(OSW_KEY_r);
	    break;

	  case XSW_KM_VIEWSCREEN_ZOOM_IN:
	    SET_KEY(OSW_KEY_equal);
	    break;
	  case XSW_KM_VIEWSCREEN_ZOOM_OUT:
	    SET_KEY(OSW_KEY_minus);
	    break;
	  case XSW_KM_VIEWSCREEN_ZOOM_AUTO:
	    SET_KEY(OSW_KEY_backslash);
	    break;

	  case XSW_KM_SCANNER_ZOOM_IN:
	    SET_KEY(OSW_KEY_period);
	    break;
	  case XSW_KM_SCANNER_ZOOM_OUT:
	    SET_KEY(OSW_KEY_comma);
	    break;
	  case XSW_KM_SCANNER_ZOOM_MIN:
	    SET_KEY(OSW_KEY_n);
	    break;
	  case XSW_KM_SCANNER_ZOOM_MAX:
	    SET_KEY(OSW_KEY_o);
	    break;
	  case XSW_KM_SCANNER_ORIENT:
	    SET_KEY(OSW_KEY_g);
	    break;

	  case XSW_KM_ENGINE_STATE:
	    SET_KEY(OSW_KEY_asciitilde);
	    break;
	  case XSW_KM_THROTTLE_MODE:
	    SET_KEY(OSW_KEY_F8);
	    break;

	  case XSW_KM_LOCK_NEXT:
	    SET_KEY(OSW_KEY_j);
	    break;
	  case XSW_KM_LOCK_NEAREST:
	    SET_KEY(OSW_KEY_Tab);
	    break;
	  case XSW_KM_UNLOCK:
	    SET_KEY(OSW_KEY_u);
	    break;

	  case XSW_KM_WEAPONS_ONLINE:
	    SET_KEY(OSW_KEY_w);
	    break;
	  case XSW_KM_WEAPON_SELECT_NEXT:
	    SET_KEY(OSW_KEY_a);
	    break;
	  case XSW_KM_WEAPON_SELECT_PREV:
	    SET_KEY(OSW_KEY_q);
	    break;
	  case XSW_KM_WEAPON_SELECT_1:
	    SET_KEY(OSW_KEY_1);
	    break;
	  case XSW_KM_WEAPON_SELECT_2:
	    SET_KEY(OSW_KEY_2);
	    break;
	  case XSW_KM_WEAPON_SELECT_3:
	    SET_KEY(OSW_KEY_3);
	    break;
	  case XSW_KM_WEAPON_SELECT_4:
	    SET_KEY(OSW_KEY_4);
	    break;
	  case XSW_KM_WEAPON_SELECT_5:
	    SET_KEY(OSW_KEY_5);
	    break;
	  case XSW_KM_WEAPON_SELECT_6:
	    SET_KEY(OSW_KEY_6);
	    break;
	  case XSW_KM_WEAPON_SELECT_7:
	    SET_KEY(OSW_KEY_7);
	    break;
	  case XSW_KM_WEAPON_SELECT_8:
	    SET_KEY(OSW_KEY_8);
	    break;
	  case XSW_KM_WEAPON_SELECT_9:
	    SET_KEY(OSW_KEY_9);
	    break;
	  case XSW_KM_WEAPON_SELECT_10:
	    SET_KEY(OSW_KEY_0);
	    break;

	  case XSW_KM_TRACTOR_BEAM_LOCK:
	    SET_KEY(OSW_KEY_t);
	    break;

	  case XSW_KM_AUTO_DESTRUCT:
	    SET_KEY(OSW_KEY_d);
	    break;

	  case XSW_KM_VS_MARKINGS:
	    SET_KEY(OSW_KEY_F2);
	    break;

	  case XSW_KM_VS_LABELS:
	    SET_KEY(OSW_KEY_F3);
	    break;

	  case XSW_KM_ENERGY_SAVER_MODE:
	    SET_KEY(OSW_KEY_F4);
	    break;

	  case XSW_KM_NET_INTERVAL_DEC:
	    SET_KEY(OSW_KEY_braceleft);
	    break;
	  case XSW_KM_NET_INTERVAL_INC:
	    SET_KEY(OSW_KEY_braceright);
	    break;

	  case XSW_KM_MSG_SCROLL_UP:
	    SET_KEY(OSW_KEY_Page_Up);
	    break;
	  case XSW_KM_MSG_SCROLL_DOWN:
	    SET_KEY(OSW_KEY_Page_Down);
	    break;

	  case XSW_KM_ECONOMY:
	    SET_KEY(OSW_KEY_semicolon);
	    break;
	  case XSW_KM_STARCHART:
	    SET_KEY(OSW_KEY_m);
	    break;

	  case XSW_KM_CONNECT:
	    SET_KEY(OSW_KEY_F9);
	    break;
	  case XSW_KM_CONNECT_LAST:
	    SET_KEY(OSW_KEY_F12);
	    break;
	  case XSW_KM_DISCONNECT:
	    SET_KEY(OSW_KEY_F10);
	    break;
	  case XSW_KM_REFRESH:
	    SET_KEY(OSW_KEY_F5);
	    break;

	  case XSW_KM_CLIENT_CMD:
	    SET_KEY(OSW_KEY_slash);
	    break;
	  case XSW_KM_SERVER_CMD:
	    SET_KEY(OSW_KEY_e);
	    break;

	  case XSW_KM_SCREEN_SHOT:
	    SET_KEY(OSW_KEY_Sys_Req);
	    break;
	}

#undef SET_KEYCODE
}


/*
 *	Maps the key scan prompt window and sets everything up for
 *	scanning a key.
 */
static void KeymapWinPromptKey(xsw_keymap_win_struct *kmw)
{
	if(kmw == NULL)
	    return;

	/* Set up query */
	kmw->key_prompt_mode = True;

	/* Map keyprompt window */
	OSWMapRaised(kmw->key_prompt_win);
}


/*
 *	Checks key maps (set in keymap win items, not globals) for
 *	any duplicates, warns and returns number of duplicates.
 */
static int KeymapWinCheckDups(xsw_keymap_win_struct *kmw, bool_t warn)
{
	int i, n, dups_found;
	clist_struct *clist;
	keymap_struct *km, *km2;

	if(kmw == NULL)
	    return(0);

	clist = &kmw->list;

	/* Iterate through all keymap */
	for(i = 0, dups_found = 0;
	    i < clist->total_rows;
	    i++
	)
	{
	    /* Get keycode */
	    km = KEYMAP(CListRowGetData(clist, i));
	    if(km == NULL)
		continue;

	    /* Skip unset keycodes */
	    if(km->keycode == 0)
		continue;

	    /* Check for duplicates */
	    for(n = 0; n < clist->total_rows; n++)
	    {
		/* Skip current row i */
		if(n == i)
		    continue;

		/* Get keycode */
		km2 = KEYMAP(CListRowGetData(clist, n));
		if(km2 == NULL)
		    continue;

		/* Check if both items have the same keycode */
		if(km->keycode == km2->keycode)
		{
		    dups_found++;

		    /* Print warning */
		    if(warn)
		    {
			char *keycode_name = STRDUP(
			    OSWGetKeyCodeName(km->keycode)
			);
			const char 	*name = CListGetItemText(clist, i, 0),
					*name2 = CListGetItemText(clist, n, 0);
			char *s = (char *)malloc(
			    STRLEN(name) + STRLEN(name2) +
			    STRLEN(keycode_name) + (2 * 80)
			);
			strtoupper(keycode_name);
			sprintf(
			    s,
"Keymaps `%s' and `%s' are both\n\
mapped to the same key `%s'.",
			    name, name2, keycode_name
			);
			free(keycode_name);
			XSWDialog(
			    "Keymap Conflict",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    kmw->toplevel
			);
			free(s);
		    }
		}
	    }
	}

	return(dups_found);
}


/*
 *	Resize callback.
 */
static void KeymapWinResizeCB(void *widget, void *data)
{
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(widget);
	if(kmw == NULL)
	    return;

	KeymapWinResize(kmw);
}

/*
 *	Draw callback.
 */
static void KeymapWinDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(widget);
	if(kmw == NULL)
	    return;

	if(kmw->map_state)
	    KeymapWinDraw(kmw, amount);
}


/*
 *	Keymap clist data destroy callback.
 */
static void KeymapWinCListDataDestroyCB(void *data)
{
	keymap_struct *km = KEYMAP(data);
	if(km == NULL)
	    return;

	free(km);
}

/*
 *	Keymap clist activate callback.
 */
static void KeymapWinCListActivateCB(clist_struct *clist, void *data)
{
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	KeymapWinScanKeyPBCB(&kmw->scan_key_btn, kmw);
}

/*
 *	Keymap clist select callback.
 */
static void KeymapWinCListSelectCB(
	clist_struct *clist, int row_num, void *data  
)
{
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	if(kmw->map_state)
	    KeymapWinQueueDraw(kmw, KMW_DRAW_AMOUNT_DESCRIPTION);
}

/*
 *	Keymap clist unselect callback.
 */
static void KeymapWinCListUnselectCB(
	clist_struct *clist, int row_num, void *data  
)
{
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	if(kmw->map_state)
	    KeymapWinQueueDraw(kmw, KMW_DRAW_AMOUNT_DESCRIPTION);
}

/*
 *	OK callback.
 */
static void KeymapWinOkPBCB(push_button_struct *btn, void *data)
{
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	KeymapWinApply(kmw);
	KeymapWinUnmap(kmw);
}

/*
 *	Apply callback.
 */
static void KeymapWinApplyPBCB(push_button_struct *btn, void *data)
{
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	KeymapWinApply(kmw);
}

/*
 *	Cancel callback.
 */
static void KeymapWinCancelPBCB(push_button_struct *btn, void *data)
{
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	KeymapWinUnmap(kmw);
}

/*
 *	Scan Key callback.
 */
static void KeymapWinScanKeyPBCB(push_button_struct *btn, void *data)
{
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	KeymapWinPromptKey(kmw);
}

/*
 *	Default callback.
 */
static void KeymapWinDefaultPBCB(push_button_struct *btn, void *data)
{
	int i;
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	i = CListGetFirstSelectedRow(&kmw->list);
	if(i < 0)
	    return;

	KeymapWinSetDefault(kmw, i);
	CListAutosizeColumns(&kmw->list);
}

/*
 *	Default All callback.
 */
static void KeymapWinDefaultAllPBCB(push_button_struct *btn, void *data)
{
	const int m = KeymapsGetTotal();
	int i;
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(data);
	if(kmw == NULL)
	    return;

	for(i = 0; i < m; i++)
	    KeymapWinSetDefault(kmw, i);
	CListAutosizeColumns(&kmw->list);
}


/*
 *	Creates a new Keymap window.
 */
xsw_keymap_win_struct *KeymapWinNew(void)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	font_t *font;
	pixel_t pixel;
	pixmap_t pixmap;
	win_t w, parent = osw_gui[0].root_win;
	push_button_struct *btn;
	clist_struct *clist;
	widget_global_struct *wglobal = &widget_global;
	xsw_keymap_win_struct *kmw = XSW_KEYMAP_WIN(calloc(
	    1, sizeof(xsw_keymap_win_struct)
	));
	if(kmw == NULL)
	    return(NULL);

	kmw->map_state = False;
	kmw->has_focus = False;
	kmw->disabled = False;
	kmw->freeze_count = 0;
	kmw->busy_count = 0;
	kmw->x = 0;
	kmw->y = 0;
	kmw->width = KEYMAP_WIN_DEF_WIDTH;
	kmw->height = KEYMAP_WIN_DEF_HEIGHT;
	kmw->visibility_state = VisibilityFullyObscured;
	kmw->key_prompt_mode = False;

	/* Create toplevel */
	x = kmw->x;
	y = kmw->y;
	width = kmw->width;
	height = kmw->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	kmw->toplevel = w;
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);

	/* WM properties */
	if(ImgIsAllocated(XSW_IMG_ICON_OPTIONS))
	    pixmap = OSWCreatePixmapFromImage(
		xsw_image[XSW_IMG_ICON_OPTIONS]->image
	    );
	else
	    pixmap = wglobal->std_icon_pm;
	OSWSetWindowWMProperties(
	    w,
	    PROG_NAME_FULL ": Key Mappings",	/* Title */
	    "Key Mappings",		/* Icon title */
	    pixmap,			/* Icon */
	    False,			/* Let WM set coordinates? */
	    kmw->x, kmw->y,
	    kmw->width, kmw->height,
	    kmw->width, kmw->height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);   

	parent = w;


	/* Keymaps clist */
	clist = &kmw->list;
	x = KEYMAP_WIN_MARGIN;
	y = KEYMAP_WIN_MARGIN;
	width = kmw->width - (2 * KEYMAP_WIN_MARGIN);
	height = 300;
	CListInit(
	    clist, parent,
	    x, y,
	    width, height
	);
	clist->activate_cb = KeymapWinCListActivateCB;
	clist->activate_data = kmw;
	clist->select_cb = KeymapWinCListSelectCB;
	clist->select_data = kmw;
	clist->unselect_cb = KeymapWinCListUnselectCB;
	clist->unselect_data = kmw;

	font = wglobal->std_font;
	pixel = osw_gui[0].black_pix;
	CListColumnAppend(
	    &kmw->list,
	    "Name",
	    font, pixel,
	    CLIST_COLUMN_FLAG_ALIGN_LEFT,
	    0
	);
	CListColumnAppend(
	    &kmw->list,
	    "Key",
	    font, pixel,
	    CLIST_COLUMN_FLAG_ALIGN_LEFT,
	    KMW_DEF_LIST_KEYCODE_POS
	);


	/* Description window */
	y += (int)height + (KEYMAP_WIN_MARGIN / 2);
	width = kmw->width - (2 * KEYMAP_WIN_MARGIN);
	height = 65;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	kmw->desc = w;
	OSWSetWindowInput(w, ExposureMask);


	/* Scan Key button */
	btn = &kmw->scan_key_btn;
	x = KEYMAP_WIN_MARGIN;
	y += height + (KEYMAP_WIN_MARGIN / 2);
	width = KEYMAP_WIN_BTN_WIDTH;
	height = KEYMAP_WIN_BTN_HEIGHT;
	PBtnInit(  
	    btn, parent,
	    x, y, width, height,
	    "Scan Key",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    KeymapWinScanKeyPBCB, kmw
	);
	accel_keys[0] = OSW_KEY_Return;
	accel_mods[0] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Default button */
	btn = &kmw->default_btn;
	x = (int)kmw->width - (2 * KEYMAP_WIN_MARGIN) -
		KEYMAP_WIN_BTN_WIDTH - 90;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Default",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    KeymapWinDefaultPBCB, kmw
	);
	accel_keys[0] = OSW_KEY_d;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Default All button */
	btn = &kmw->default_all_btn;
	width = 90;
	height = KEYMAP_WIN_BTN_HEIGHT;
	x = (int)kmw->width - KEYMAP_WIN_MARGIN - 90;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Default All",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    KeymapWinDefaultAllPBCB, kmw
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);


	/* OK button */
	btn = &kmw->ok_btn;
	width = KEYMAP_WIN_BTN_WIDTH;
	height = KEYMAP_WIN_BTN_HEIGHT;
	x = KEYMAP_WIN_MARGIN;
	y = (int)kmw->height - (int)height - KEYMAP_WIN_MARGIN;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "OK",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    KeymapWinOkPBCB, kmw
	);
	accel_keys[0] = OSW_KEY_o;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Apply button */
	btn = &kmw->apply_btn;
	x = ((int)kmw->width - KEYMAP_WIN_BTN_WIDTH) / 2;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Apply",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    KeymapWinApplyPBCB, kmw
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Cancel button */
	btn = &kmw->cancel_btn;
	x = (int)kmw->width - KEYMAP_WIN_MARGIN -
	    KEYMAP_WIN_BTN_WIDTH;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    KeymapWinCancelPBCB, kmw
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL; 
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);


	/* Scan key window */
	x = KEYMAP_WIN_MARGIN;
	y = ((int)kmw->height - KEYMAP_WIN_PROMPT_HEIGHT) / 2;
	width = kmw->width - (2 * KEYMAP_WIN_MARGIN);
	height = KEYMAP_WIN_PROMPT_HEIGHT;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	kmw->key_prompt_win = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | ButtonPressMask
	);

	return(kmw);
}

/*
 *	Resizes the keymap window.
 */
void KeymapWinResize(xsw_keymap_win_struct *kmw)
{
	int x, y;
	unsigned int width, height;
	win_t w;

	if(kmw == NULL)
	    return;

	w = kmw->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	kmw->x = x;
	kmw->y = y;
	kmw->width = width;
	kmw->height = height;

}

/*
 *	Queues the Keymap Window to resize.
 */
void KeymapWinQueueResize(xsw_keymap_win_struct *kmw)
{
	if(kmw == NULL)
	    return;

	WidgetQueueResizeAction(
	    KeymapWinResizeCB, kmw, NULL
	);
}

/*
 *	Draws the Keymap Window.
 */
void KeymapWinDraw(xsw_keymap_win_struct *kmw, const int amount)
{
	widget_global_struct *wglobal = &widget_global;

	if(kmw == NULL)
	    return;

	/* Description buffer */
	if(kmw->desc_buf == None)
	{
	    unsigned int width, height;
	    OSWGetWindowGeometry(
		kmw->desc, NULL, NULL, &width, &height, NULL
	    );
	    if(OSWCreatePixmap(
		&kmw->desc_buf, width, height
	    ))
		return;
	}
	/* Key prompt buffer */
	if(kmw->key_prompt_buf == None)
	{
	    unsigned int width, height;
	    OSWGetWindowGeometry(
		kmw->key_prompt_win, NULL, NULL, &width, &height, NULL
	    );
	    if(OSWCreatePixmap(
		&kmw->key_prompt_buf, width, height
	    ))
		return;
	}


	/* Redraw background */
	if(amount == KMW_DRAW_AMOUNT_COMPLETE)
	{
	    unsigned int width, height;
	    win_t w = kmw->toplevel;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(osw_gui[0].black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile( 
		    w, wglobal->std_bkg_img,
		    width, height
		);

		OSWSetForeground(wglobal->surface_shadow_pix);
		OSWDrawLine(
		    w,
		    0,
		    (int)height - KEYMAP_WIN_BTN_HEIGHT - 25,
		    (int)width,
		    (int)height - KEYMAP_WIN_BTN_HEIGHT - 25
		);
 
		OSWSetForeground(wglobal->surface_highlight_pix);
		OSWDrawLine(
		    w,
		    0,
		    (int)height - KEYMAP_WIN_BTN_HEIGHT - 24,
		    (int)width,
		    (int)height - KEYMAP_WIN_BTN_HEIGHT - 24
		);
	    }
	}

	/* Description */
	if((amount == KMW_DRAW_AMOUNT_COMPLETE) ||
	   (amount == KMW_DRAW_AMOUNT_DESCRIPTION)
	)
	{
	    unsigned int width, height;
	    win_t w = kmw->desc;
	    pixmap_t pixmap = kmw->desc_buf;
	    clist_struct *clist = &kmw->list;
	    keymap_struct *km;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );
	    OSWClearPixmap(
		pixmap, width, height,
		wglobal->force_mono ?
		    osw_gui[0].black_pix :
		    wglobal->surface_editable_pix
	    );

	    /* Get selected item number */
	    km = KeymapGetPtr(CListGetLastSelectedRow(clist));
	    if(km != NULL)
	    {
		int	x, y,
			lbearing, ascent, descent,
			font_height;
		font_t *font = wglobal->std_font;
		const char	*s = km->description,
				*s_end;

		OSWSetFont(font);
		if(wglobal->force_mono)
		    OSWSetForeground(osw_gui[0].white_pix);
		else
		    OSWSetForeground(wglobal->editable_text_pix);

		x = 2 + 2;
		y = 2 + 2;
		while(s != NULL)
		{
		    if(*s == '\0')
			break;

		    s_end = strchr(s, '\n');
		    if(s_end == NULL)
			s_end = strchr(s, '\r');

		    if(s_end != NULL)
		    {
			const int len = (int)(s_end - s);
			OSWGetTextBounds(
			    font, s, len,
			    &lbearing, NULL, NULL, &ascent, &descent
			);
			font_height = ascent + descent;
			OSWDrawStringLimited(
			    pixmap,
			    x - lbearing,
			    y + ascent,
			    s, len
			);
			y += font_height;
		    }
		    else
		    {
			OSWGetTextBounds(          
			    font, s, -1,
			    &lbearing, NULL, NULL, &ascent, &descent
			);  
			font_height = ascent + descent;
			OSWDrawString(
			    pixmap,
			    x - lbearing,
			    y + ascent,
			    s
			);
			y += font_height;
		    }

		    s = (s_end != NULL) ? (s_end + 1) : NULL;
	        }
	    }

	    /* Draw frame */
	    WidgetFrameButtonPixmap(
		pixmap,
		True,
		width, height,
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix
	    );

	    OSWPutDrawableToWindow(w, pixmap);
	}


	/* Key prompt window */
	if(kmw->key_prompt_mode &&
	   ((amount == KMW_DRAW_AMOUNT_COMPLETE) ||
	    (amount == KMW_DRAW_AMOUNT_SCAN_KEY)
	   )
	)
	{
	    int	x, y,
		lbearing, rbearing, ascent, descent, font_height;
	    unsigned int width, height;
	    const char *s;
	    font_t *font = widget_global.std_font;   
	    win_t w = kmw->key_prompt_win;
	    pixmap_t pixmap = kmw->key_prompt_buf;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );
	    OSWClearPixmap(
		pixmap, width, height,
		wglobal->force_mono ?
		    osw_gui[0].white_pix :
		    xsw_color.keymap_query_bg
	    );

	    OSWSetForeground(
		wglobal->force_mono ?
		    osw_gui[0].black_pix :
		    xsw_color.keymap_query_fg
	    );

	    OSWSetFont(font);
	    x = KEYMAP_WIN_MARGIN;
	    y = KEYMAP_WIN_MARGIN;
	    s =
"Press a key on the keyboard to set the keycode.";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, NULL, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		s
	    );
	    y += font_height + (KEYMAP_WIN_MARGIN / 2);

	    s = "Press Button1 to cancel.";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, NULL, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		s
	    );
	    y += font_height;

	    s = "Press Button2 to set default.";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, NULL, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		s
	    );
	    y += font_height;

	    s = "Press Button3 to clear.";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, NULL, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		s
	    );

	    OSWPutDrawableToWindow(w, pixmap);
	}
}

/*
 *	Queues the Keymap Window to draw.
 */
void KeymapWinQueueDraw(xsw_keymap_win_struct *kmw, const int amount)
{
	if(kmw == NULL)
	    return;

	WidgetQueueDrawAction(
	    KeymapWinDrawCB, kmw, (void *)amount
	);
}

/*
 *	Manages the keymap window event.
 */
int KeymapWinManage(xsw_keymap_win_struct *kmw, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;
	event_configure_t *configure;
	event_key_t *key;
	event_button_t *button;
	event_visibility_t *visibility;

	if((kmw == NULL) || (event == NULL))
	    return(events_handled);

	if(!kmw->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    configure = &event->xconfigure;
	    if(w == kmw->toplevel)
	    {
		KeymapWinQueueResize(kmw);
		if(kmw->map_state)
		    KeymapWinQueueDraw(kmw, KMW_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if((w == kmw->toplevel) || (w == kmw->desc))
	    {
		KeymapWinQueueDraw(kmw, KMW_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    else if(w == kmw->key_prompt_win)
	    {
		KeymapWinQueueDraw(kmw, KMW_DRAW_AMOUNT_SCAN_KEY);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!kmw->has_focus)
	        return(events_handled);

	    /* Scan keycode mode? */
	    if(kmw->key_prompt_mode)
	    {
		OSWKBAutoRepeatOff();
		events_handled++;
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!kmw->has_focus)  
		return(events_handled);

	    /* Scan keycode mode? */
	    if(kmw->key_prompt_mode)
	    {
		clist_struct *clist = &kmw->list;
		const int i = CListGetLastSelectedRow(clist);
		if(i > -1)
		{
		    KeymapWinSetKeycode(
			kmw, clist, i,
			key->keycode, key->state
		    );
		    CListAutosizeColumns(clist);
		}

		OSWKBAutoRepeatOn();

		/* Check and warn about duplicate keycodes */
		KeymapWinCheckDups(kmw, True);

		/* Unmap the scan key prompt */
		kmw->key_prompt_mode = False;
		OSWUnmapWindow(kmw->key_prompt_win);

		events_handled++;
	    }
	    break;

	  case ButtonPress:
	    button = &event->xbutton;

	    /* Scan keycode mode? */
	    if(kmw->key_prompt_mode)
	    {
		clist_struct *clist = &kmw->list;
		const int i = CListGetFirstSelectedRow(clist);
		switch(button->button)
		{
		  case Button1:
		    /* Cancel */
		    kmw->key_prompt_mode = False;
		    OSWUnmapWindow(kmw->key_prompt_win);
		    break;
		  case Button2:
		    /* Set default */
		    KeymapWinSetDefault(kmw, i);
		    kmw->key_prompt_mode = False;
		    OSWUnmapWindow(kmw->key_prompt_win);
		    break;
		  case Button3:
		    /* Clear */
		    KeymapWinSetKeycode(
			kmw, clist, i,
			0, 0
		    );
		    CListAutosizeColumns(clist);
		    kmw->key_prompt_mode = False;
		    OSWUnmapWindow(kmw->key_prompt_win);
		    break;
		}
		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;

	    break;

	  case MapNotify:
	    if(w == kmw->toplevel)
	    {
		if(!kmw->map_state)
		    KeymapWinMap(kmw, None);
		events_handled++;
	    }
	    break;

	  case UnmapNotify:
	    if(w == kmw->toplevel)
	    {
		if(kmw->map_state)
		    KeymapWinUnmap(kmw);
		events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == kmw->toplevel)
	    {
		kmw->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == kmw->toplevel)
	    {
		KeymapWinUnfocusAll(kmw);
		kmw->has_focus = False;
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == kmw->toplevel)
	    {
		kmw->visibility_state = visibility->state;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(kmw->toplevel, event, &t))
	    {
		OSWSetInputFocus(kmw->toplevel, RevertToParent, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(kmw->toplevel, event))
	    {
		KeymapWinUnmap(kmw);
		events_handled++;
		return(events_handled);
	    }
	    break; 
	}

	/* Manage widgets */
	if(events_handled == 0)  
	    events_handled += WidgetManage(&kmw->ok_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&kmw->apply_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&kmw->cancel_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&kmw->scan_key_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&kmw->default_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&kmw->default_all_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&kmw->list, event);

	return(events_handled);
}

/*
 *	Maps the keymap window.
 */
void KeymapWinMap(xsw_keymap_win_struct *kmw, win_t toplevel)
{
	win_t w;

	if(kmw == NULL)
	    return;

	if(kmw->toplevel == None)
	    return;

	if(!kmw->map_state)
	{
	    KeymapWinFetch(kmw);
	    CListUnselectAllRows(&kmw->list);
	    CListSelectRow(&kmw->list, 0);
	}

	KeymapWinUnfocusAll(kmw);

	w = kmw->toplevel;
	OSWSetTransientFor(w, toplevel);
	if(toplevel != None)
	    WidgetCenterWindowToWindow(w, toplevel);
	else
	    WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_ROOT);

	/* Map */
	if(!kmw->map_state)
	{
	    OSWMapRaised(kmw->toplevel);
	    kmw->map_state = True;
	    kmw->visibility_state = VisibilityUnobscured;
	    OSWMapWindow(kmw->desc);
	    WidgetMap(&kmw->list);
	    WidgetMap(&kmw->ok_btn);
	    WidgetMap(&kmw->cancel_btn);
	    WidgetMap(&kmw->apply_btn);
	    WidgetMap(&kmw->scan_key_btn);
	    WidgetMap(&kmw->default_btn);
	    WidgetMap(&kmw->default_all_btn);
	}

	/* Focus */
	OSWGrabFocus(kmw->toplevel);
	kmw->list.has_focus = True;
}

/*
 *	Unmaps the keymap window.
 */
void KeymapWinUnmap(xsw_keymap_win_struct *kmw)
{
	if(kmw == NULL)
	    return;

	OSWUnmapWindow(kmw->toplevel);
	kmw->has_focus = False;
	kmw->map_state = False;

	WidgetUnmap(&kmw->ok_btn);
	WidgetUnmap(&kmw->cancel_btn);
	WidgetUnmap(&kmw->apply_btn);
	WidgetUnmap(&kmw->scan_key_btn);
	WidgetUnmap(&kmw->default_btn);
	WidgetUnmap(&kmw->default_all_btn);
	WidgetUnmap(&kmw->list);

	CListClear(&kmw->list);
	OSWDestroyPixmap(&kmw->key_prompt_buf);
	OSWDestroyPixmap(&kmw->desc_buf);
}

/*
 *	Deletes the Keymap window.
 */
void KeymapWinDelete(xsw_keymap_win_struct *kmw)
{
	if(kmw == NULL)
	    return;

	CListClear(&kmw->list);

	WidgetDestroy(&kmw->ok_btn);
	WidgetDestroy(&kmw->cancel_btn);
	WidgetDestroy(&kmw->apply_btn);
	WidgetDestroy(&kmw->scan_key_btn);
	WidgetDestroy(&kmw->default_btn);
	WidgetDestroy(&kmw->default_all_btn);
	OSWDestroyWindow(&kmw->key_prompt_win);
	OSWDestroyPixmap(&kmw->key_prompt_buf);
	WidgetDestroy(&kmw->list);
	OSWDestroyWindow(&kmw->desc);
	OSWDestroyPixmap(&kmw->desc_buf);
	OSWDestroyWindow(&kmw->toplevel);

	free(kmw);
}
