#include <stdlib.h>
#include <string.h>

#include "msglist.h"
#include "msgwin.h"

#include "xsw.h"


void MsgListSetSize(const int nmessages);
static void MsgListAppendIterate(
	const char *s, const int msg_len,
	pixel_t fg_pix, pixel_t bg_pix,
	const unsigned int flags,
	xsw_bridge_win_struct *bw,
	xsw_msg_win_struct *mw
);
void MsgListAppend(
	const char *s,
	pixel_t fg_pix, pixel_t bg_pix,
	const unsigned int flags
);

void MsgListUnmarkAll(void);

void MsgListCopyToClipboard(win_t w);

void MsgListQueueDrawAllWindows(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Reallocates the messages list to the number of messages
 *	specified by nmessages.
 *
 *	This function should not be used to initialize or delete the
 *	messages since the bridge window and messages window will be
 *	redrawn.
 */
void MsgListSetSize(const int nmessages)
{
	scroll_bar_struct *sb;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_msg_win_struct *mw = xsw_msg_win;

	/* Update messages list size */
	total_messages = MAX(nmessages, 0);
	if(total_messages > 0)
	{
	    mesg_list = (xsw_mesg_struct *)realloc(
		mesg_list,
		total_messages * sizeof(xsw_mesg_struct)
	    );
	    if(mesg_list == NULL)
		total_messages = 0;
	    if(total_messages > 0)
	    {
		/* Clear */
		memset(mesg_list, 0x00, total_messages * sizeof(xsw_mesg_struct));
	    }
	}
	else
	{
	    free(mesg_list);
	    mesg_list = NULL;
	    total_messages = 0;
	}

	/* Update the bridge window message box scroll bars */
	if(bw != NULL)
	{
	    sb = &bw->mesg_box_sb;
	    SBarQueueResize(
		sb,
		sb->width, sb->height,
#if 1
		sb->width,
#else
		(bw->mesg_box_longest_line * bw->mesg_box_char_width) +
		    SCROLLBAR_YBAR_WIDTH + (2 * 10),
#endif
		bw->mesg_box_line_spacing * (total_messages + 1)
	    );
	    /* Scroll to end */
	    SBarSetPositionX(sb, 0.0f);
	    SBarSetPositionY(sb, (float)sb->total_height);
	}

	/* Update the messages window scroll bars */
	if(mw != NULL)
	{
	    sb = &mw->sb;
	    SBarQueueResize(
		sb,
		sb->width, sb->height,
		(mw->longest_line * mw->char_width) +
		    SCROLLBAR_YBAR_WIDTH + (2 * MSG_WIN_MARGIN),
		mw->line_spacing * (total_messages + 5)
	    );
	    /* Scroll to end */ 
	    SBarSetPositionX(sb, 0.0f);
	    SBarSetPositionY(sb, (float)sb->total_height);
	}

	MsgListQueueDrawAllWindows();
}


/*
 *	Adds the message to the messages list.
 */
static void MsgListAppendIterate(
	const char *s, const int msg_len,
	pixel_t fg_pix, pixel_t bg_pix,
	const unsigned int flags,
	xsw_bridge_win_struct *bw,
	xsw_msg_win_struct *mw
)
{
	int i, m, len, msg_pos, longest_line;
	xsw_mesg_struct *msg, *msg_prev;

	/* Message lines are numbered from bottom to top, so message
	 * index 0 would be the last line
	 */

	longest_line = 0;

	/* Add the specified message to the list, adding multiple lines
	 * as needed
	 */
	msg_pos = 0;
	while(msg_pos < msg_len)
	{
	    /* Shift messages "up" */
	    m = total_messages - 1;
	    for(i = 0; i < m; i++)
	    {
		msg = &mesg_list[i];
		msg_prev = &mesg_list[i + 1];

		strncpy(
		    msg->message, msg_prev->message,
		    sizeof(msg->message)
		);
		msg->message[sizeof(msg->message) - 1] = '\0';

		msg->flags = msg_prev->flags;

		msg->fg_pix = msg_prev->fg_pix;
		msg->bg_pix = msg_prev->bg_pix;

		msg->sel_start = msg_prev->sel_start;
		msg->sel_end = msg_prev->sel_end;

		/* Update the longest line */
		len = STRLEN(msg->message);
		if(len > longest_line)
		    longest_line = len;
	    }

	    /* Get the first message position (highest index) in the
	     * message list, this where we are going to put the current
	     * segment of the specified message to
	     */
	    msg = &mesg_list[m];

	    len = MIN(
		(int)(sizeof(msg->message) - 1),
		MESG_WIN_COL_WRAP
	    );
	    strncpy(msg->message, s + msg_pos, len);
	    msg->message[len] = '\0';

	    msg->flags = flags;

	    msg->fg_pix = fg_pix;
	    msg->bg_pix = bg_pix;

	    msg->sel_start = -1;
	    msg->sel_end = -1;

	    /* Update the longest line */
	    len = STRLEN(msg->message);
	    if(len > longest_line)
		longest_line = len;

	    /* Increment the specified message's current position
	     * based on how much of it we added on this message
	     */
	    msg_pos += len;
	}

	/* Update the longest line on the message windows */
	if(bw != NULL)
	    bw->mesg_box_longest_line = longest_line;
	if(mw != NULL)
	    mw->longest_line = longest_line;
}

/*
 *	Adds the message to the messages list and draws all message
 *	windows.
 */
void MsgListAppend(
	const char *s,
	pixel_t fg_pix, pixel_t bg_pix,
	const unsigned int flags
)
{
	int len;
	const char *s_end;
	scroll_bar_struct *sb;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_msg_win_struct *mw = xsw_msg_win;

	/* This function may be called when not connected to the GUI,
	 * so check if we are connected first
	 */
	if(!IDC() || (s == NULL))
	    return;

	/* Add each line */
	do {

	    s_end = strpbrk(s, "\n\r");
	    len = (s_end != NULL) ? (int)(s_end - s) : STRLEN(s);

	    MsgListAppendIterate(
		s, len,
		fg_pix, bg_pix,
		flags,
		bw, mw
	    );

	    s = s_end;

	} while(s != NULL);

	/* Update the bridge window message box scroll bars */
	if(bw != NULL)
	{
	    sb = &bw->mesg_box_sb;
	    SBarQueueResize(
		sb,
		sb->width, sb->height,
#if 1
		sb->width,
#else
		(bw->mesg_box_longest_line * bw->mesg_box_char_width) +
		    SCROLLBAR_YBAR_WIDTH + (2 * 10),
#endif
		bw->mesg_box_line_spacing * (total_messages + 1)
	    );
	}

	/* Update the messages window's scroll bars */
	if(mw != NULL)
	{
	    sb = &mw->sb;
	    SBarQueueResize(
		sb,
		sb->width, sb->height,
		(mw->longest_line * mw->char_width) +
		    SCROLLBAR_YBAR_WIDTH + (2 * MSG_WIN_MARGIN),
		mw->line_spacing * (total_messages + 5)
	    );
	}

	/* Redraw all message windows */
	MsgListQueueDrawAllWindows();
}

/*
 *	Unmark all messages in the message list.
 */
void MsgListUnmarkAll(void)
{
	int i;
	xsw_mesg_struct *msg;

	for(i = 0; i < total_messages; i++)
	{
	     msg = &mesg_list[i];
	     msg->sel_start = -1;
	     msg->sel_end = -1;
	}
}


/*
 *	Puts the primary message buffer to the DDE setting the specified
 *	window as the owner.
 */
void MsgListCopyToClipboard(win_t w)
{
	bool_t last_line_full = False;
	int	i,
		sel_start, sel_end, sel_len, msg_len,
		buf_pos = 0, buf_len = 0;
	char *buf = NULL;
	xsw_mesg_struct *msg;

	/* Iterate through all message lines */
	for(i = 0; i < total_messages; i++)
	{
	    msg = &mesg_list[i];
	    sel_start = msg->sel_start;
	    sel_end = msg->sel_end;

	    /* This line has no selection? */
	    if((sel_start < 0) || (sel_end < 0))
		continue;

	    /* Get the content length of this line */
	    msg_len = (int)strlen(msg->message);
	    if(msg_len <= 0)
		continue;

	    /* Sanitize selected start and end positions */
	    if(sel_start >= msg_len)
		sel_start = msg_len - 1;
	    if(sel_end >= msg_len)
		sel_end = msg_len - 1;

	    /* Swap selection start/end positions as needed */
	    if(sel_start > sel_end)
	    {
		const int tmp_i = sel_end;
		sel_end = sel_start;
		sel_start = tmp_i;
	    }

	    /* Get the length of the selection of this line, make
	     * sure that it does not exceed this line's message content
	     * length
	     */
	    sel_len = sel_end - sel_start + 1;
	    if((sel_start + sel_len) > msg_len)
		sel_len = msg_len - sel_start;

	    if(sel_len <= 0)
		continue;

	    /* Need to append a carrage return to the buffer? */
	    if((buf_len > 0) && !last_line_full)
	    {
		buf_len += 1;
		buf = (char *)realloc(buf, buf_len * sizeof(char));
		if(buf == NULL)
		    return;
		buf_pos = buf_len - 1;

		buf[buf_pos] = '\n';
	    }

	    /* Add this message to the buffer */
	    buf_pos = buf_len;
	    buf_len += sel_len;
	    buf = (char *)realloc(buf, buf_len * sizeof(char));
	    if(buf == NULL)
		return;

	    memcpy(
		buf + buf_pos,			/* Target */
		msg->message + sel_start,	/* Source */
		sel_len
	    );

	    /* Check if this line is "full" in which case the next line
	     * should not have a carrage return
	     */
	    last_line_full = (msg_len >= (int)MIN(sizeof(msg->message), MESG_WIN_COL_WRAP)) ?
		True : False;
	}

	/* Null terminate and put to DDE */
	if(buf != NULL)
	{
	    buf = (char *)realloc(buf, (buf_len + 1) * sizeof(char));
	    buf[buf_len] = '\0';
	    OSWDDESetString(w, buf);
	    free(buf);
	}
}


/*
 *	Queues all windows that display messages to draw.
 */
void MsgListQueueDrawAllWindows(void)
{
	scroll_bar_struct *sb;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_msg_win_struct *mw = xsw_msg_win;

	if(mw == NULL)
	    return;

	/* Bridge window message box */
	if((bw != NULL) ? bw->map_state : False)
	{
	    sb = &bw->mesg_box_sb;
	    BridgeQueueDrawMessages(bw);
	    SBarQueueDraw(sb);
	}

	/* Message window */
	if((mw != NULL) ? mw->map_state : False)
	{
	    sb = &mw->sb;
	    MsgWinQueueDraw(mw);
	    SBarQueueDraw(sb);
	}
}
