#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <time.h>
#include <fcntl.h>
#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <unistd.h>

#include "../include/netio.h"

#include "net.h"
#include "xsw.h"


int NetConnect(const char *host, int port, time_t timeout);
void NetShutdownConnection(int how, char is_initiator);
void NetCloseConnection(time_t timeout);
void NetWaitClose(time_t timeout);
void NetDisconnect(time_t timeout);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Opens a connection.
 *
 *	If timeout is positive then it specifies the timeout in seconds.
 *
 *	Returns the socket number or -1 on error.
 */
int NetConnect(const char *host, int port, time_t timeout)
{
	int new_sock;			/* socket number */
	char *dhost;
	struct hostent *he;		/* Host pointer */
	struct sockaddr_in haddr;
	xsw_net_parms_struct *net = &net_parms;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_option_struct *opt = &option;

	if((host == NULL) || (port < 0))
	    return(-2);

	/* Copy host name */
	dhost = strdup(host);

	/* Close socket and reset net values as needed */
	NetDisconnect(timeout);


	/* Begin connecting */

	/* Get new hostname */
	he = gethostbyname(dhost);
	if(he == NULL)
	{
	    const int error_code = h_errno;
	    char *s;

#define PRINT_WARNING(_error_msg_)	{	\
 char *error_msg = STRDUP(_error_msg_);		\
 if(error_msg != NULL) {			\
  *error_msg = toupper(*error_msg);		\
  s = (char *)malloc(				\
   STRLEN(dhost) + STRLEN(error_msg) + 80	\
  );						\
  sprintf(					\
   s,						\
"Get host error: %s: %s.",			\
   dhost, error_msg				\
  );						\
  XSWPrintWarning(s);				\
  free(s);					\
 }						\
 free(error_msg);				\
}

	    switch(error_code)
	    {
	      case HOST_NOT_FOUND:
		PRINT_WARNING(
"host not found"
		);
		s = (char *)malloc(STRLEN(dhost) + (5 * 80));
	        sprintf(
		    s,
"Unable to find the host:\n\
\n\
    %s\n\
\n\
Please verify that the address is correct and try again.",
		    dhost
		);
		XSWDialog(
		    "Connect Failed",
                    XSW_IMG_DIALOG_ICON_WARNING,
		    s,
		    (bw != NULL) ? bw->toplevel : None
		);
		free(s);
		break;

	      case NO_ADDRESS:
#if (NO_DATA != NO_ADDRESS)
	      case NO_DATA:
#endif
                PRINT_WARNING(
"unable to obtain the host's IP address"
		);
		s = (char *)malloc(STRLEN(dhost) + (5 * 80));
		sprintf(
		    s,
"Unable to obtain the IP address for the host:\n\
\n\
    %s\n\
\n\
Please verify that the address is correct and try again.",
		    dhost
		);
		XSWDialog(
		    "Connect Failed",
                    XSW_IMG_DIALOG_ICON_WARNING,
		    s,
		    (bw != NULL) ? bw->toplevel : None
		);
		free(s);
		break;
 
	      case NO_RECOVERY:
                PRINT_WARNING(
"Non-recoverably name server error, please try again"
		);
		s = (char *)malloc(STRLEN(dhost) + (6 * 80));
		sprintf(
		    s,
"A non-recoverable name server error occurred while\n\
attempting to obtain the information for host:\n\
\n\
    %s\n\
\n\
Please try again later.",
		    dhost
		);
		XSWDialog(
		    "Connect Failed",
                    XSW_IMG_DIALOG_ICON_ERROR,
		    s,
		    (bw != NULL) ? bw->toplevel : None
		);
		free(s);
		break;

	      case TRY_AGAIN:
		PRINT_WARNING(
"temporary error occured, please try again"
		);
		s = (char *)malloc(STRLEN(dhost) + (6 * 80));
		sprintf(
		    s,
"A temporary error occurred on the name server while\n\
attempting to obtain the information for host:\n\
\n\
    %s\n\
\n\
Please try again later.",
		    dhost
		);
		XSWDialog(
		    "Connect Failed",
                    XSW_IMG_DIALOG_ICON_WARNING,
		    s,
		    (bw != NULL) ? bw->toplevel : None
		);
		free(s);
		break;

	      default:
		PRINT_WARNING(
"unknown error"
		);
		s = (char *)malloc(STRLEN(dhost) + (5 * 80));
		sprintf(
		    s,
"An unknown error occured while getting the host:\n\
\n\
    %s\n\
\n\
Please verify that the address is correct and try again.",
		    dhost
		);
		XSWDialog(
		    "Connect Failed",
                    XSW_IMG_DIALOG_ICON_WARNING,
		    s,
		    (bw != NULL) ? bw->toplevel : None
		);
		free(s);
		break;
	    }

#undef PRINT_WARNING

	    free(dhost);
	    return(-1);
	}

	/* Create a new socket */
	new_sock = socket(AF_INET, SOCK_STREAM, 0);
	if(new_sock < 0)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unknown error");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(STRLEN(error_msg) + 80);
	    sprintf(
		s,
"Create socket error: %s.",
		error_msg
	    );
	    XSWPrintError(s);
	    free(s);

	    s = (char *)malloc(STRLEN(error_msg) + (3 * 80));
	    sprintf(
		s,
"The following error occured while creating a new socket:\n\
\n\
    %s",
		error_msg
	    );
	    XSWDialog(
		"Connect Failed",
                XSW_IMG_DIALOG_ICON_ERROR,
		s,
		(bw != NULL) ? bw->toplevel : None
	    );
	    free(s);

	    free(error_msg);

	    free(dhost);

	    return(-1);
	}

	haddr.sin_family = AF_INET;		/* In hbo */
	haddr.sin_port = htons(port);		/* In nbo */
	haddr.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&(haddr.sin_zero), 0x00, 8);	/* Zero the rest of struct */

	/* Open connection */
	if(connect(
	    new_sock,
	    (struct sockaddr *)&haddr,
	    sizeof(struct sockaddr)
	   ) == -1
	)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unknown error");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(dhost) + 40 + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
"Connect error: %s: Port %i: %s.",
		dhost, port, error_msg
	    );
	    XSWPrintWarning(s);
	    free(s);

	    s = (char *)malloc(
		STRLEN(error_msg) + STRLEN(dhost) +
		40 + (7 * 80)
	    );
	    sprintf(
		s,
"%s:\n\
\n\
    %s  Port: %i\n\
\n\
Please verify that the address and port number are correct\n\
and that the site has a server running and accepting\n\
connections on the specified port number.",
		error_msg, dhost, port
	    );
	    XSWDialog(
		"Connect Failed",
                XSW_IMG_DIALOG_ICON_WARNING,
		s,
		(bw != NULL) ? bw->toplevel : None
	    );
	    free(s);
	    free(error_msg);

	    free(dhost);

	    return(-1);
	}

	/* Set socket to non-blocking */
	if(fcntl(new_sock, F_SETFL, O_NONBLOCK) == -1)
	{
	    const int error_code = errno;
	    if(opt->show_net_errors)
	    {
		char *s, *error_msg = STRDUP(strerror(error_code));
		if(error_msg == NULL)
		    error_msg = strdup("unknown error");
		*error_msg = toupper(*error_msg);

		s = (char *)malloc(40 + STRLEN(error_msg) + 80);
		sprintf(
		    s,
"Set socket O_NONBLOCK error: Socket %i: %s.",
		    new_sock,
		    error_msg
		);
		XSWPrintError(s);
		free(s);
		free(error_msg);
	    }
	}

	/* Connectioned successfully */

	/* Set up network parameters */
	net->socket = new_sock;

	/* Mark connection state as loggin in */
	net->connection_state = XSW_CON_STATE_NEGOTIATING;

	/* Mark connection start time */
	net->con_start = time(NULL);

	free(dhost);

	return(new_sock);
}


/*
 *	Shuts down write to the socket that is connected to the server
 *	and sets the connection state to XSW_CON_STATE_CLOSE_WAIT.
 *
 *	The how must be set to one of XSW_CON_SHUTDOWN_*.
 *
 *	If is_initiator is set to 1 then we are the initiator of the
 *	shutdown.
 */
void NetShutdownConnection(int how, char is_initiator)
{
	int lhow;
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	/* Set connection state to close wait */
	net->connection_state = XSW_CON_STATE_CLOSE_WAIT;

	if(is_initiator)
	    net->flags |= XSW_CON_FLAG_DISCON_INITIATOR;

	/* Shutdown the socket */
	switch(how)
	{
	  case XSW_CON_SHUTDOWN_WR:
	    lhow = SHUT_WR;
	    break;
	  case XSW_CON_SHUTDOWN_RDWR:
	    lhow = SHUT_RDWR;
	    break;
	  default:
	    lhow = SHUT_RD;
	    break;
	}
	if(shutdown(net->socket, lhow))
	{
	    const int error_code = errno;
	    if(opt->show_net_errors)
	    {
		char *s, *error_msg = STRDUP(strerror(error_code));
		if(error_msg == NULL)
		    error_msg = strdup("unknown error");
		*error_msg = toupper(*error_msg);
		s = (char *)malloc(STRLEN(error_msg) + 40 + 80);
		sprintf(
		    s,
"Shutdown socket error: Socket %i: %s.",
		    net->socket,
		    error_msg
		);
		XSWPrintWarning(s);
		free(s);
		free(error_msg);
	    }
	}
}

/*
 *	Closes the socket that is connected to the server and sets
 *	the connection state to XSW_CON_STATE_NOT_CONNECTED and the
 *	socket to -1.
 *
 *      If timeout is positive then it specifies the timeout in seconds.
 */
void NetCloseConnection(time_t timeout)
{
	int error_code;
	time_t start_time;
	xsw_net_parms_struct *net = &net_parms;
        xsw_option_struct *opt = &option;

	/* Set connection state to not connected */
	net->connection_state = XSW_CON_STATE_NOT_CONNECTED;

	/* Socket already closed? */
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	/* Close the socket */
	start_time = time(NULL);
	while(close(net->socket) != 0)
	{
	    error_code = errno;
	    if(error_code == EAGAIN)
	    {
		/* Timmed out? */
		if(timeout > 0l)
		{
		    if(time(NULL) > (start_time + timeout))
			break;
		}
		continue;
	    }
	    else
	    {
		if(opt->show_net_errors)
		{
		    char *s, *error_msg = STRDUP(strerror(error_code));
		    if(error_msg == NULL)
			error_msg = strdup("unknown error");
		    *error_msg = toupper(*error_msg);
		    s = (char *)malloc(STRLEN(error_msg) + 80);
		    sprintf(
			s,  
"Close socket error: Socket %i: %s.",
			net->socket,
			error_msg
		    );
		    XSWPrintWarning(s);
		    free(s);
		    free(error_msg);
		}
		break;
	    }
	}
	net->socket = -1;
}

/*
 *	Waits for the socket to close by calling recv() and discarding
 *	the data until it returns 0 or error. When 0 or error is
 *	returned the socket will be closed and the socket will be set
 *	to -1.
 *
 *	This function does not check if the connection state is set to
 *	CLOSE_WAIT or if the connection was shut down prior to this
 *	call.
 *
 *	If timeout is positive then it specifies the timeout in seconds.
 */
void NetWaitClose(time_t timeout)
{
	const time_t start_time = time(NULL);
	char buf[1024];
	xsw_net_parms_struct *net = &net_parms;

	while(XSW_CON_IS_CONNECTED(net))
	{
	    if(NetRecv(
		net->socket, buf, sizeof(buf)
	    ) <= 0)
		break;

	    /* Timmed out? */
	    if(timeout > 0l)
	    {
		if(time(NULL) > (start_time + timeout))
		    break;
	    }
	}
	NetCloseConnection(timeout);
}

/*
 *	Closes the socket that is connectioned to the server and sets
 *	all net values to indicate not connected.
 *
 *	If timeout is positive then it specifies the timeout in seconds.
 */
void NetDisconnect(time_t timeout)
{
	xsw_net_parms_struct *net = &net_parms;

	/* Shutdown and close the connection as needed */
	if(XSW_CON_IS_CONNECTED(net))
	{
	    if(XSW_CON_STATE_IS_CLOSE_WAIT(net))
	    {
		NetWaitClose(timeout);
	    }
	    else
	    {
		NetShutdownConnection(XSW_CON_SHUTDOWN_RDWR, 1);
		NetWaitClose(timeout);
	    }
	}

	net->connection_state = XSW_CON_STATE_NOT_CONNECTED;
	net->socket = -1;
	net->flags = 0;

	net->player_obj_num = -1;
	net->player_obj_ptr = NULL;

	net->con_start = 0l;
	net->login_delay_warning_last = 0l;

	/* Do not reset login name, login password, or address */

	free(net->title);
	net->title = NULL;

	net->net_int = XSW_SWSERV_DEF_UPDATE_INT;
	net->disconnect_send_count = 0;
	net->bad_send_count = 0;

	*net->recvbuf = '\0';
	net->recvbuf_len = 0;
}
