#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#include "../include/netio.h"


int NetIsSocketWritable(int s, int *error_code_rtn);
int NetIsSocketReadable(int s, int *error_code_rtn);

int NetSend(int socket, const void *buf, int nbytes);
int NetRecv(int socket, void *buf, int nbytes);


/*
 *	Checks if the socket is valid and writeable.
 *
 *	If error_code_rtn is not NULL then it will be set to the value
 *	of errno if the return is false.
 */
int NetIsSocketWritable(int s, int *error_code_rtn)
{
	struct timeval t;  
	fd_set writefds;

	if(s < 0)
	{
	    if(error_code_rtn != NULL)
		*error_code_rtn = EBADF;
	    return(0);
	}

	t.tv_sec = 0;
	t.tv_usec = 0;
	FD_ZERO(&writefds);
	FD_SET(s, &writefds);
	if(select(
	    s + 1, 
	    NULL, &writefds, NULL,
	    &t
	) == -1)
	{
	    if(error_code_rtn != NULL)
		*error_code_rtn = errno;
	    return(0);
	}

	if(FD_ISSET(s, &writefds))
	{
	    if(error_code_rtn != NULL)
		*error_code_rtn = 0;
	    return(1);
	}
	else
	{
	    if(error_code_rtn != NULL)
		*error_code_rtn = errno;
	    return(0);
	}
}

/*
 *	Checks if the socket is valid and contains data to be read.
 *
 *	If error_code_rtn is not NULL then it will be set to the value
 *	of errno if the return is false.
 */
int NetIsSocketReadable(int s, int *error_code_rtn)
{
	struct timeval t;
	fd_set readfds;

	if(s < 0)
	{
	    if(error_code_rtn != NULL)
		*error_code_rtn = EBADF;
	    return(0);
	}

	t.tv_sec = 0;
	t.tv_usec = 0;
	FD_ZERO(&readfds);
	FD_SET(s, &readfds);
	if(select(
	    s + 1,
	    &readfds, NULL, NULL,
	    &t
	) == -1)
	{
	    if(error_code_rtn != NULL)
		*error_code_rtn = errno;
	    return(0);
	}

	if(FD_ISSET(s, &readfds))
	{
	    if(error_code_rtn != NULL)
		*error_code_rtn = 0;
	    return(1);
	}
	else
	{
	    if(error_code_rtn != NULL)
		*error_code_rtn = errno;
	    return(0);
	}
}

/*
 *	Sends the data to the socket.
 */
int NetSend(int socket, const void *buf, int nbytes)
{
        if(socket < 0)
        {
            errno = EBADF;
            return(-1);
        }
        if(buf == NULL)
        {
            errno = EFAULT;
            return(-1);
        }
	return((int)send(socket, buf, nbytes, MSG_NOSIGNAL));
}

/*
 *	Receives data from the socket.
 */
int NetRecv(int socket, void *buf, int nbytes)
{
	if(socket < 0)
	{
	    errno = EBADF;
	    return(-1);
	}
	if(buf == NULL)
	{
	    errno = EFAULT;
	    return(-1);
	}
        return((int)recv(socket, buf, nbytes, MSG_NOSIGNAL));
}
