#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "../include/netio.h"
#include "../include/swnetcodes.h"
#include "../include/swsoundcodes.h"

#include "../include/reality.h"
#include "../include/unvmath.h"
#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "particle.h"
#include "xswimgrefs.h"
#include "sound.h"
#include "sndrefs.h"
#include "ignoredobj.h"
#include "net.h"
#include "starchartwin.h"
#include "xsw.h"
#include "ocsnames.h"		/* Must be after xsw.h for types to be defined */


static void NetHandleVersion(const char *arg);
static void NetHandleLogin(const char *arg);
static void NetHandleLogout(const char *arg);

static void NetHandleLiveMesg(const char *arg);
static void NetHandleSysMesg(const char *arg);

static void NetHandleWhoAmI(const char *arg);
static void NetHandleSetLegend(const char *arg);
static void NetHandleSetImageSet(const char *arg);
static void NetHandleSetSoundSet(const char *arg);
static void NetHandleSetForward(const char *arg);
static void NetHandleSetOCSN(const char *arg);
static void NetHandleCreateObject(const char *arg);
static void NetHandleRecycleObject(const char *arg);
static void NetHandleSetObjSect(const char *arg);
static void NetHandleSetFObjSect(const char *arg);
static void NetHandleSetPoseObj(const char *arg);
static void NetHandleSetFPoseObj(const char *arg);
static void NetHandleSetObjectValues(const char *arg);
static void NetHandleSetObjectMaximums(const char *arg);
static void NetHandleSetName(const char *arg);
static void NetHandleSetShieldVis(const char *arg);
static void NetHandleSetChannel(const char *arg);
static void NetHandleSetScore(const char *arg);
static void NetHandleSetEngine(const char *arg);
static void NetHandleSetObjParticleTrail(const char *arg);
static void NetHandleCreateEffect(const char *arg);
static void NetHandleCreateShockwave(const char *arg);
static void NetHandleNotifyHit(const char *arg);
static void NetHandleNotifyDestroy(const char *arg);
static void NetHandleTractorBeamLock(const char *arg);
static void NetHandleHail(const char *arg);
static void NetHandleComMessage(const char *arg);
static void NetHandleWeaponDisarm(const char *arg);
static void NetHandleSetAutoDestruct(const char *arg);
static void NetHandleSetDescription(const char *arg);
static void NetHandleEcoSetValues(const char *arg);
static void NetHandleEcoSetProductValues(const char *arg);
static void NetHandleStarChartAddObject(const char *arg);
static void NetHandleStarChartSetObjectName(const char *arg);
static void NetHandleStarChartSetObjectEmpire(const char *arg);
static void NetHandleStarChartRecycleObject(const char *arg);

static void NetHandlePlaySound(const char *arg);

static void NetHandleSetWeaponValues(const char *arg);

static void NetHandleExtCmd(const char *arg);
int NetHandleRecv(void);


#define ASSERT_NOT_NAN(x)	assert(!isnan(x))

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : True)


/*
 *	Version.
 */
static void NetHandleVersion(const char *arg)
{
	int status, major = 0, minor = 0, release = 0;

	/* CS_CODE_VERSION format:
	 *
	 *	major minor release
	 */
	status = sscanf(
	    arg,
"%i %i %i",
	    &major, &minor, &release
	);
	if(status >= 2)
	{
/* TODO */

	}

	/* Do not send version response */
}


/*
 *	Login.
 */
static void NetHandleLogin(const char *arg)
{
	int status;
	char *sndbuf;
	xsw_net_parms_struct *net = &net_parms;

	/* Send version */
	NetSendVersion();
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	/* Format login */
	sndbuf = (char *)malloc(
	    40 +
	    STRLEN(net->login_name) +
	    STRLEN(net->login_password) +
	    40 +
	    80
	);
	sprintf(
	    sndbuf,
"%i %s;%s;%i\n",
	    CS_CODE_LOGIN,
	    (net->login_name != NULL) ?
		net->login_name : "",
	    (net->login_password != NULL) ?
		net->login_password : "",
	    XSW_CLIENT_TYPE_CODE
	);
	status = NetSendData(sndbuf);
	free(sndbuf);
	if(status)
	{
	    XSWPrintWarning(
"Unable to send login data."
	    );
	    return;
	}

	/* We have sent our name and password, that does not mean that
	 * we are logged in yet, we need to wait for a NET_CMD_WHOAMI
	 * response from the server so that we know what is our player
	 * object
	 *
	 * Once we recieve a NET_CMD_WHOAMI and the player object then
	 * we are considered connected and logged in
	 *
	 * IMPORTANT: The connection_state will be set to
	 * XSW_CON_STATE_CONNECTED once we receive all of the following:
	 *
	 *	* Player object number
	 *	* Position of player object
	 *	* Sector of player object
	 */
}

/*
 *	Disconnect.
 */
static void NetHandleLogout(const char *arg)
{
	xsw_net_parms_struct *net = &net_parms;
	if(!XSW_CON_STATE_IS_CLOSE_WAIT(net))
	    NetShutdownConnection(XSW_CON_SHUTDOWN_RDWR, False);
}

/*
 *	Live message.
 */
static void NetHandleLiveMesg(const char *arg)
{
	if(STRISEMPTY(arg))
	    return;

	XSWPrint(arg);

	/* Log this message? */
	if(option.log_net)
	    XSWLog(arg);
}

/*
 *	System message.
 */
static void NetHandleSysMesg(const char *arg)
{
	xsw_net_parms_struct *net = &net_parms;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	win_t toplevel = (bw != NULL) ? bw->toplevel : None;

	/* Get system message command code */
	const int sys_cmd = StringGetNetCommand(arg);

	/* Get system message argument */
	char *sys_arg = STRDUP(StringGetNetArgument(arg));

	if(sys_arg == NULL)
	    return;

	/* Handle by system message command code */
	switch(sys_cmd)
	{
	  case CS_SYSMESG_CODE_LOGINFAIL:
	    if(!XSW_CON_STATE_IS_CONNECTED(net))
	    {
		XSWDialog(
		    "Login Failed",
		    XSW_IMG_DIALOG_ICON_WARNING,
		    sys_arg,
		    toplevel
		);
	    }
	    else
	    {
		if(option.show_server_errors)
		    XSWDialog(
			"Login Failed",
			XSW_IMG_DIALOG_ICON_WARNING,
			sys_arg,
			toplevel
		    );
	    }
	    break;

	  case CS_SYSMESG_CODE_LOGINSUCC:
	    /* Ignore this */
	    break;

	  case CS_SYSMESG_CODE_ABNDISCON:
	    if(!option.show_server_errors)
		break;

	    break;

	  case CS_SYSMESG_CODE_BADVALUE:
	    if(!option.show_server_errors)
		break;

	    break;

	  case CS_SYSMESG_CODE_BADARG:
	    if(!option.show_server_errors)
		break;

	    break;

	  case CS_SYSMESG_CODE_WARNING:
	    if(!option.show_server_errors)
		break;
	    XSWDialog(
		"Server Warning",
		XSW_IMG_DIALOG_ICON_WARNING,
		sys_arg,
		toplevel
	    );
	    break;

	  default:	/* CS_SYSMESG_CODE_ERROR */
	    if(!option.show_server_errors)
		break;
	    XSWDialog(
		"Server Error",
		XSW_IMG_DIALOG_ICON_ERROR,
		sys_arg,
		toplevel
	    );
	    break;
	}

	free(sys_arg);
}

/*
 *	Who am I.
 */
static void NetHandleWhoAmI(const char *arg)
{
	int len, obj_num = -1;
	const char *s, *s_end;
	char *name = NULL;
	xsw_net_parms_struct *net = &net_parms;

	/* Format CS_CODE_WHOAMI:
	 *
	 *	obj_num;name
	 */
	s = arg;

	/* Get the object number */
	if(*s == '#')
	    s++;
	obj_num = ATOI(s);

	/* Get the name */
	s = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	if(s != NULL)
	{
	    s++;
	    s_end = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	    len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    if(len > 0)
	    {
		name = (char *)malloc((len + 1) * sizeof(char));
		memcpy(name, s, len);
		name[len] = '\0';
	    }
	}

	/* Object number invalid? */
	if((obj_num < 0) || (obj_num >= SW_OBJECTS_MAX))
	{
	    free(name);
	    return;
	}

	/* Player object was not set? */
	if(net->player_obj_num < 0)
	{
	    /* Raise the bridge window when a whoami is received */
	    BridgeWinMap(xsw_bridge_win);

	    /* Update the looping sounds as needed */
	    if(option.sounds >= XSW_SOUNDS_ENGINE)
		XSWUpdateLoopSounds();

	    /* Update the background music as needed */
	    if(option.music)
		XSWUpdateBackgroundMusic();
	}

	/* Set the player (even if it has not yet been created) */
	DBSetPlayerObject(obj_num);

	/* Set global player object number explicitly incase the object
	 * was invalid and DBSetPlayerObject() would set it to -1
	 */
	net->player_obj_num = obj_num;

	/* Mark that we got the player object */
	net->flags |= XSW_CON_FLAG_SET_PLAYER_NUM;

	/* The player object may not have been created yet, we do
	 * not create it in this function.
	 *
	 * The reason is because we do not know the imgref_num
	 * thus we do not have enough information to pass to
	 * DBCreateExplicitObject().
	 *
	 * The player object will be created when the minimal
	 * information has been recieved in one of the recieve
	 * handling functions.
	 */
	if(!DBIsObjectGarbage(net->player_obj_num))
	{
	    /* Draw the bridge panel to reflect the setting of the
	     * player object
	     */
	    BridgeWinDrawPanel(obj_num, BW_PANEL_P2);
	}

	free(name);
}

/*
 *	Handles a set units.
 */
static void NetHandleSetLegend(const char *arg)
{
	float	ru_to_au = 0.0f,
		antimatter_low_warning = 0.0f,
		asteroid_field_damage_rate = 0.0f,
		auto_destruct_damage_range = 0.0f,
		cloak_antimatter_rate = 0.0f,
		com_range = 0.0f,
		damage_control_rate = 0.0f;
	long	asteroid_field_damage_int = 0l,
		disarmed_weapon_lifespan = 0l;
	float	engine_antimatter_rate = 0.0f,
		no_antimatter_thrust_coeff = 0.0f,
		on_fire_hp_coeff = 0.0f,
		tractor_beam_antimatter_rate = 0.0f,
		tractor_beam_range = 0.0f,
		transporter_range = 0.0f;
	sw_legend_struct *legend = &sw_legend;

	/* Format SWEXTCMD_SET_LEGEND:
	 *
	 *	ru_to_au
	 *	antimatter_low_warning 
	 *	asteroid_field_damage_int
	 *	asteroid_field_damage_rate
	 *	auto_destruct_damage_range
	 *	cloak_antimatter_rate
	 *	com_range
	 *	damage_control_rate
	 *	disarmed_weapon_lifespan
	 *	engine_antimatter_rate
	 *	no_antimatter_thrust_coeff
	 *	on_fire_hp_coeff
	 *	tractor_beam_antimatter_rate
	 *	tractor_beam_range
	 *	transporter_range
	 */
	sscanf(
	    arg,
"%f %f %ld %f %f %f %f %f %ld %f %f %f %f %f %f",
	    &ru_to_au,
	    &antimatter_low_warning,
	    &asteroid_field_damage_int,
	    &asteroid_field_damage_rate,
	    &auto_destruct_damage_range,
	    &cloak_antimatter_rate,
	    &com_range,
	    &damage_control_rate,
	    &disarmed_weapon_lifespan,
	    &engine_antimatter_rate,
	    &no_antimatter_thrust_coeff,
	    &on_fire_hp_coeff,
	    &tractor_beam_antimatter_rate,
	    &tractor_beam_range,
	    &transporter_range
	);
	ASSERT_NOT_NAN(ru_to_au);
	ASSERT_NOT_NAN(antimatter_low_warning);
	ASSERT_NOT_NAN(asteroid_field_damage_rate);
	ASSERT_NOT_NAN(auto_destruct_damage_range);
	ASSERT_NOT_NAN(cloak_antimatter_rate);
	ASSERT_NOT_NAN(com_range);
	ASSERT_NOT_NAN(damage_control_rate);
	ASSERT_NOT_NAN(engine_antimatter_rate);
	ASSERT_NOT_NAN(no_antimatter_thrust_coeff);
	ASSERT_NOT_NAN(on_fire_hp_coeff);
	ASSERT_NOT_NAN(tractor_beam_antimatter_rate);
	ASSERT_NOT_NAN(tractor_beam_range);
	ASSERT_NOT_NAN(transporter_range = 0.0f);

	if(ru_to_au > 0.0f)
	    legend->ru_to_au = ru_to_au;
	if(antimatter_low_warning > 0.0f)
	    legend->antimatter_low_warning = antimatter_low_warning;
	if(asteroid_field_damage_int > 0l)
	    legend->asteroid_field_damage_int = asteroid_field_damage_int;
	if(asteroid_field_damage_rate > 0.0f)
	    legend->asteroid_field_damage_rate = asteroid_field_damage_rate;
	if(auto_destruct_damage_range > 0.0f)
	    legend->auto_destruct_damage_range = auto_destruct_damage_range;
	if(cloak_antimatter_rate > 0.0f)
	    legend->cloak_antimatter_rate = cloak_antimatter_rate;
	if(com_range > 0.0f)
	    legend->com_range = com_range;
	if(damage_control_rate > 0.0f)
	    legend->damage_control_rate = damage_control_rate;
	legend->disarmed_weapon_lifespan = (time_t)disarmed_weapon_lifespan;
	if(engine_antimatter_rate > 0.0f)
	    legend->engine_antimatter_rate = engine_antimatter_rate;
	if(no_antimatter_thrust_coeff > 0.0f)
	    legend->no_antimatter_thrust_coeff = no_antimatter_thrust_coeff;
	if(on_fire_hp_coeff > 0.0f)
	    legend->on_fire_hp_coeff = on_fire_hp_coeff;
	if(tractor_beam_antimatter_rate > 0.0f)
	    legend->tractor_beam_antimatter_rate = tractor_beam_antimatter_rate;
	if(tractor_beam_range > 0.0f)
	    legend->tractor_beam_range = tractor_beam_range;
	if(transporter_range > 0.0f)
	    legend->transporter_range = transporter_range;

	/* Redraw the bridge window */
	BridgeWinDrawAll(xsw_bridge_win);

#ifdef STARCHARTWIN_H
	SChtSetLegend(xsw_scht, legend);
#endif
}

/*
 *	Handles a set title.
 */
static void NetHandleSetTitle(const char *arg)
{
	const char *title;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* Format SWEXTCMD_SETTITLE:
	 *
	 *	title
	 */
	title = arg;

	free(net->title);
	net->title = STRDUP(title);

	obj = net->player_obj_ptr;

	/* Update title */
	BridgeWinSetTitle(
	    xsw_bridge_win,
	    net->address, net->title,
	    (obj != NULL) ? obj->name : NULL
	);
}

/*
 *	Handles a set image set file.
 */
static void NetHandleSetImageSet(const char *arg)
{
	const char *s;
	char filename[PATH_MAX + NAME_MAX];
	struct stat stat_buf;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	win_t toplevel = (bw != NULL) ? bw->toplevel : None;

	/* Format CS_CODE_IMAGESET:
	 *
	 *	path
	 */
	s = arg;
	strncpy(filename, s, sizeof(filename));
	filename[sizeof(filename) - 1] = '\0';

	/* Check if the file exists */
	if(ISPATHABSOLUTE(filename))
	{
	    /* Paths different? */
	    if(strcmp(fname.imgrefs, filename))
	    {
		/* Does the file not exist? */
		if(stat(filename, &stat_buf))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find universe required ImgRefs file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing ImgRefs File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel         
			);
		    free(s);
		    return;
		}
		/* Is it a directory? */
		if(S_ISDIR(stat_buf.st_mode))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find universe required ImgRefs file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing ImgRefs File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}

		strncpy(fname.imgrefs, filename, sizeof(fname.imgrefs));
		fname.imgrefs[sizeof(fname.imgrefs) - 1] = '\0';
		XSWOpenImgRefs(fname.imgrefs);
	    }
	}
	else
	{
	    s = PrefixPaths(dname.images, filename);
	    if(s == NULL)
		return;
	    strncpy(filename, s, sizeof(filename));
	    filename[sizeof(filename) - 1] = '\0';

	    /* Paths different? */
	    if(strcmp(fname.imgrefs, filename))
	    {
		/* Does file exist? */
		if(stat(filename, &stat_buf))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find universe required ImgRefs file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing ImgRefs File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}
		/* Is it a directory? */
		if(S_ISDIR(stat_buf.st_mode))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find universe required ImgRefs file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing ImgRefs File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}

		strncpy(fname.imgrefs, filename, sizeof(fname.imgrefs));
		fname.imgrefs[sizeof(fname.imgrefs) - 1] = '\0';
		XSWOpenImgRefs(fname.imgrefs);
	    }
	}
}

/*
 *	Handles a set sound set file.
 */
static void NetHandleSetSoundSet(const char *arg)
{
	const char *s;
	char filename[PATH_MAX + NAME_MAX];
	struct stat stat_buf;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	win_t toplevel = (bw != NULL) ? bw->toplevel : None;

	/* Format CS_CODE_SOUNDSET:
	 *
	 *	path
	 */
	s = arg;
	strncpy(filename, s, sizeof(filename));
	filename[sizeof(filename) - 1] = '\0';

	/* Check if already loaded */
	if(ISPATHABSOLUTE(filename))
	{
	    /* Paths different? */
	    if(strcmp(fname.sndrefs, filename))
	    {
		/* Does file exist? */
		if(stat(filename, &stat_buf))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find universe required SndRefs file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing SndRefs File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}
		/* Is it a directory? */
		if(S_ISDIR(stat_buf.st_mode))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find universe required SndRefs file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing SndRefs File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}

		strncpy(fname.sndrefs, filename, sizeof(fname.sndrefs));
		fname.sndrefs[sizeof(fname.sndrefs) - 1] = '\0';
		XSWOpenSndRefs(fname.sndrefs);
	    }
	}
	else
	{
	    s = PrefixPaths(dname.sounds, filename);
	    if(s == NULL)
		return;

	    strncpy(filename, s, sizeof(filename));
	    filename[sizeof(filename) - 1] = '\0';
 
	    /* Paths different? */
	    if(strcmp(fname.sndrefs, filename))
	    {
		/* Does it exist? */
		if(stat(filename, &stat_buf))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find universe required SndRefs file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing SndRefs File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}
		/* Is it a directory? */
		if(S_ISDIR(stat_buf.st_mode))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find universe required SndRefs file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing SndRefs File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}

		strncpy(fname.sndrefs, filename, sizeof(fname.sndrefs));
		fname.sndrefs[sizeof(fname.sndrefs) - 1] = '\0';
		XSWOpenSndRefs(fname.sndrefs);
	    }
	}
}

/*
 *	Handles a forward.
 */
static void NetHandleSetForward(const char *arg)
{
	const char *s;
	char	*src_url,
		*url,
		*address,
		*login_name,
		*login_password;
	int port;
	xsw_net_parms_struct *net = &net_parms;

	/* CS_CODE_FORWARD forward:
	 *
	 *	url
	 */
	src_url = STRDUP(arg);

	/* Get the address (must be specified) */
	s = StringParseAddress(src_url);
	if(s == NULL)
	    return;

	address = strdup(s);

	/* Get the port */
	port = StringParsePort(src_url);
	if(port < 0)
	    port = net->port;

	/* Get the login name */
	s = StringParseName(src_url);
	login_name = (s != NULL) ?
	    strdup(s) : STRDUP(net->login_name);

	/* Get the login password */
	s = StringParsePassword(src_url);
	login_password = (s != NULL) ?
	    strdup(s) : STRDUP(net->login_password);

	free(src_url);

	/* Format the new url and connect */
	url = (char *)malloc(
	    STRLEN(login_name) +
	    STRLEN(login_password) +
	    STRLEN(address) +
	    40 +
	    80
	);
	sprintf(
	    url,
	    "swserv://%s:%s@%s:%i",
	    (login_name != NULL) ? login_name : "",
	    (login_password != NULL) ? login_password : "",
	    (address != NULL) ? address : "",
	    port
	);
	XSWConnect(url);
	free(url);
	free(login_name);
	free(login_password);
	free(address);
}

/*
 *	Handles a set object create script names file.
 */
static void NetHandleSetOCSN(const char *arg)
{
	char loaded_new = 0;
	const char *s;
	struct stat stat_buf;
	char filename[PATH_MAX + NAME_MAX];
	xsw_net_parms_struct *net = &net_parms;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	win_t toplevel = (bw != NULL) ? bw->toplevel : None;

	/* Format SWEXTCMD_SETOCSN:
	 *
	 *	path
	 */
	s = arg;
	strncpy(filename, s, sizeof(filename));
	filename[sizeof(filename) - 1] = '\0';


	/* Check if already loaded */
	if(ISPATHABSOLUTE(filename))
	{
	    /* Paths different? */
	    if(strcmp(fname.ocsn, filename))
	    {
		/* Does file exist? */
		if(stat(filename, &stat_buf))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find the universe required OCS Names file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing OCS Names File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}
		/* Is it a directory? */
		if(S_ISDIR(stat_buf.st_mode))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find the universe required OCS Names file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing OCS Names File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}

		strncpy(fname.ocsn, filename, sizeof(fname.ocsn));
		fname.ocsn[sizeof(fname.ocsn) - 1] = '\0';
		XSWOpenOCSNs(fname.ocsn);

		loaded_new = 1;
	    }
	}
	else
	{
	    s = PrefixPaths(dname.etc, filename);
	    if(s == NULL)
		return;

	    strncpy(filename, s, sizeof(filename));
	    filename[sizeof(filename) - 1] = '\0';

	    /* Paths different? */
	    if(strcmp(fname.ocsn, filename))
	    {
		/* Does file exist? */
		if(stat(filename, &stat_buf))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find the universe required OCS Names file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing OCS Names File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}
		/* Is it a directory? */
		if(S_ISDIR(stat_buf.st_mode))
		{
		    char *s = (char *)malloc(strlen(filename) + 256);
		    sprintf(
			s,
"Unable to find the universe required OCS Names file:\n\
\n\
    %s",
			filename
		    );
		    if(option.show_net_errors)
			XSWDialog(
			    "Missing OCS Names File",
			    XSW_IMG_DIALOG_ICON_WARNING,
			    s,
			    toplevel
			);
		    free(s);
		    return;
		}
  
		strncpy(fname.ocsn, filename, sizeof(fname.ocsn));
		fname.ocsn[sizeof(fname.ocsn) - 1] = '\0';
		XSWOpenOCSNs(fname.ocsn);

		loaded_new = 1;
	    }
	}

	/* Update resources needed when a new OCSN is loaded */
	if(loaded_new)
	{
	    /* Remove weapons from player object so they get recreated */
	    if(!DBIsObjectGarbage(net->player_obj_num))
		UNVAllocObjectWeapons(
		    net->player_obj_ptr,
		    0
		);

	    if(bw != NULL)
	    {
		/* Recreate the selected weapon viewscreen label */
		VSDrawUpdateWeaponLabel(
		    &bw->vs_weapon_image,
		    bw->vs_weapon_buf
		);
	    }
	}
}


/*
 *	Handles a create object request.
 */
static void NetHandleCreateObject(const char *arg)
{
/*	char set_back_player_obj = 0; */
	char loaded_imgref = 0;
	int obj_num = -1;
	imgref_struct *imgref;
	xsw_animation_struct *anim;
	xsw_object_struct *obj;

	int type = SW_OBJ_TYPE_GARBAGE;
	int imgref_num = 0;
	int owner = -1;
	long size = 0l;

	int locked_on = -1;
	int intercepting_object = -1;
	float scanner_range = 0.0f;

	long	sect_x = 0l, sect_y = 0l, sect_z = 0l;
	float	x = 0.0f, y = 0.0f, z = 0.0f;
	float	heading = 0.0f * PI,
		pitch = 0.0f * PI,
		bank = 0.0f * PI;
	float	velocity = 0.0f;
	float	velocity_heading = 0.0f * PI,
		velocity_pitch = 0.0f * PI,
		velocity_bank = 0.0f * PI;

	int current_frame = 0, total_frames = 0, cycle_times = 0;
	long anim_int = 0l, lifespan = 0l;

	time_t cur_ms;

	xsw_net_parms_struct *net = &net_parms;

	/* CS_CODE_CREATEOBJ format:
	 *
	 *	obj_num
	 *	type imgref_num owner size
	 *	locked_on intercepting_object scanner_range
	 *	sect_x sect_y sect_z
	 *	x y z
	 *	heading pitch bank
	 *	velocity
	 *	velocity_heading velocity_pitch velocity_bank
	 *	current_frame anim_int total_frames cycle_times
	 *	lifespan
	 */
	sscanf(
	    arg,
"%i\
 %i %i %i %ld\
 %i %i %f\
 %ld %ld %ld\
 %f %f %f\
 %f %f %f\
 %f\
 %f %f %f\
 %i %ld %i %i\
 %ld",
	    &obj_num,
	    &type, &imgref_num, &owner, &size,
	    &locked_on, &intercepting_object, &scanner_range,
	    &sect_x, &sect_y, &sect_z,
	    &x, &y, &z,
	    &heading, &pitch, &bank,
	    &velocity,
	    &velocity_heading, &velocity_pitch, &velocity_bank,
	    &current_frame, &anim_int, &total_frames, &cycle_times,
	    &lifespan
	);

	ASSERT_NOT_NAN(scanner_range);
	ASSERT_NOT_NAN(x);
	ASSERT_NOT_NAN(y);
	ASSERT_NOT_NAN(z);
	ASSERT_NOT_NAN(heading);
	ASSERT_NOT_NAN(pitch);
	ASSERT_NOT_NAN(bank);
	ASSERT_NOT_NAN(velocity);
	ASSERT_NOT_NAN(velocity_heading);
	ASSERT_NOT_NAN(velocity_pitch);
	ASSERT_NOT_NAN(velocity_bank);

	/* Object number must be valid */
	if((obj_num < 0) || (obj_num >= SW_OBJECTS_MAX))
	    return;

	/* Imgref number must be a valid number */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	    return;

#if 1
	/* Does the object already exist? */
	if(!DBIsObjectGarbage(obj_num))
	{

	}
	else
	{
	    /* The object does not exist, create it */
            if(DBCreateExplicitObject(
                obj_num,
                imgref_num,
                type,
                owner,
                sect_x, sect_y, sect_z,
                x, y, z,
                heading, pitch, bank
            ) < 0)
                return;
	}

        obj = xsw_object[obj_num];
#else
	/* If the object already exists, we must recycle it first */
	if(!DBIsObjectGarbage(obj_num))
	{
	    /* Need to make note to set back referance to player object
	     * after creation since we are about to recycle it
	     */
	    if(obj_num == net->player_obj_num)
	        set_back_player_obj = 1;

	    /* Recycle the object */
	    DBRecycleObject(obj_num);
	}

	/* Create the object */
	if(DBCreateExplicitObject(
	    obj_num,
	    imgref_num,
	    type,
	    owner,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    heading, pitch, bank
	) < 0)
	    return;

	obj = xsw_object[obj_num];

	/* Restore the player object as needed */
	if(set_back_player_obj)
	    DBSetPlayerObject(obj_num);
#endif

	/* Need to add a little antimatter so reality engine will
	 * allow this object to move if it is intended to move
	 */
	obj->antimatter_max = 1.0f;
	obj->antimatter = 1.0f;

	/* Is this the player object? */
	if(obj_num == net->player_obj_num)
	{
	    /* Player object is intercepting something, let server
	     * set its attitude
	     */
	    if(obj->intercepting_object > -1)
	    {
		obj->heading = heading;
		obj->pitch = pitch;
		obj->bank = bank;
		SWMathSetUnitVector2D(
		    &obj->attitude_vector_compoent,
		    obj->heading
		);
	    }

	    /* Already connected and logged in? */
	    if(XSW_CON_STATE_IS_CONNECTED(net))
	    {
		/* Redraw the bridge window and do not continue further */
		BridgeWinDrawAll(xsw_bridge_win);
		return;
	    }
	    else if(XSW_CON_STATE_IS_NEGOTIATING(net))
	    {
		/* Still logging in */

		/* Mark that we got the player's sector and position */
		net->flags |= XSW_CON_FLAG_SET_PLAYER_SECT;
		net->flags |= XSW_CON_FLAG_SET_PLAYER_POS;

		/* Did we get the player object number yet? */
		if(net->flags & XSW_CON_FLAG_SET_PLAYER_NUM)
		{
		    /* Mark that we are now logged in and set the
		     * player object references
		     */
		    net->connection_state = XSW_CON_STATE_CONNECTED;
		    DBSetPlayerObject(obj_num);
		}
	    }
	}

	/* Load the imgref as needed */
	imgref = ImgRefGetPtr(xsw_imgref, total_imgrefs, imgref_num);
	if(!ImgRefIsLoaded(imgref))
	{
	    ImgRefLoad(&xsw_imgref, &total_imgrefs, imgref_num);
	    loaded_imgref = 1;
	}

	/* Get the most recent time */
	cur_ms = MilliTime();

	/* Set the values of the new object */
	obj->type = type;

	obj->imgref = imgref_num;
	obj->owner = owner;
	obj->size = size;

	obj->locked_on = DBIsObjectGarbage(locked_on) ? -1 : locked_on;
	obj->intercepting_object = intercepting_object;
	obj->scanner_range = scanner_range;

	obj->velocity = velocity;
	obj->velocity_heading = velocity_heading;
	obj->velocity_pitch = velocity_pitch;
	obj->velocity_bank = velocity_bank;
	SWMathSetUnitVector2D(
	    &obj->momentum_vector_compoent,
	    obj->velocity_heading
	);

	obj->birth_time = cur_ms;
	obj->lifespan = lifespan;

	anim = &obj->animation;
	anim->current_frame = current_frame;
	anim->interval = anim_int;
	anim->total_frames = total_frames;
	anim->cycle_count = 0;
	anim->cycle_times = cycle_times;
	anim->last_interval = cur_ms;

	/* Refresh the last time it was updated */
	obj->last_updated = cur_ms;


	/* Is this the newly created player object? */
	if(obj_num == net->player_obj_num)
	{
	    /* Redraw the bridge window */
	    BridgeWinDrawAll(xsw_bridge_win);
	}
}

/*
 *	Handles a recycle object request.
 */
static void NetHandleRecycleObject(const char *arg)
{
	int obj_num = -1;

	/* CS_CODE_RECYCLEOBJ format:
	 *
	 *	obj_num
	 */
	sscanf(
	    arg,
	    "%i",
	    &obj_num
	);

	DBRecycleObject(obj_num);
}

/*
 *	Handles a set object sector request.
 */
static void NetHandleSetObjSect(const char *arg)
{
	char changed_sector = 0;
	int obj_num = -1;
	long sect_x = 0l, sect_y = 0l, sect_z = 0l;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_SETOBJSECT format:
	 *
	 *	obj_num
	 *	sect_x sect_y sect_z
	 */
	sscanf(
	    arg,
	    "%i %ld %ld %ld",
	    &obj_num,
	    &sect_x, &sect_y, &sect_z
	);

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Check of sector value changed */
	if((obj->sect_x != sect_x) ||
	   (obj->sect_y != sect_y) ||
	   (obj->sect_z != sect_z)
	)
	    changed_sector = 1;

	/* If not logged in but connected, then allow server to set
	 * sector position
	 */
	if(obj_num == net->player_obj_num)
	{
	    /* Do not set sector if already logged in */
	    if(XSW_CON_STATE_IS_CONNECTED(net))
		return;

	    /* Mark that got the player's sector */
	    net->flags |= XSW_CON_FLAG_SET_PLAYER_SECT;

	    /* If other required information about player object
	     * are recieved then set the connection state to
	     * XSW_CON_STATE_CONNECTED and set all player references
	     */
	    if((net->flags & XSW_CON_FLAG_SET_PLAYER_NUM) &&
	       (net->flags & XSW_CON_FLAG_SET_PLAYER_POS)
	    )
	    {
		net->connection_state = XSW_CON_STATE_CONNECTED;
		DBSetPlayerObject(obj_num);
	    }
	}

	if(changed_sector)
	{
	    obj->sect_x = sect_x;
	    obj->sect_y = sect_y;
	    obj->sect_z = sect_z;
	}

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;
}


/*
 *      Handles a forced set object sector request.
 */
static void NetHandleSetFObjSect(const char *arg)
{
	char changed_sector = 0;
	int obj_num = -1;
	long sect_x = 0l, sect_y = 0l, sect_z = 0l;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_SETFOBJSECT format:
	 *
	 *	obj_num
	 *	sect_x sect_y sect_z
	 */
	sscanf(
	    arg,
	    "%i %ld %ld %ld",
	    &obj_num,
	    &sect_x, &sect_y, &sect_z
	);

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Check of sector value changed */
	if((obj->sect_x != sect_x) ||
	   (obj->sect_y != sect_y) ||
	   (obj->sect_z != sect_z)
	)
	    changed_sector = 1;

	if(changed_sector)
	{
	    obj->sect_x = sect_x;
	    obj->sect_y = sect_y;
	    obj->sect_z = sect_z;

	    /* Is this the player object? */
	    if(obj_num == net->player_obj_num)
		XSWObjectSectorChangedNotify(
		    obj_num,
		    cur_millitime, cur_systime, time_compensation
		);
	}

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;

	/* Must respond with a sect that has the exact same values */
	NetSendObjSect(obj_num);
}


/*
 *	Handles a pose object request, velocity, pitch and coordinates
 *	(except sector) are set by this function.
 *
 *	Player object is not posed here unless it is intercepting
 *	something.
 */
static void NetHandleSetPoseObj(const char *arg)
{
	char imgref_changed = 0;
	char loaded_imgref = 0;
	int obj_num = -1;
	int type = SW_OBJ_TYPE_GARBAGE;
	int imgref_num = 0;
	long size = 0l;
	float	x = 0.0f, y = 0.0f, z = 0.0f;
	float	heading = 0.0f * PI,
		pitch = 0.0f * PI,
		bank = 0.0f * PI;
	float	velocity = 0.0f;
	float	velocity_heading = 0.0f * PI,
		velocity_pitch = 0.0f * PI,
		velocity_bank = 0.0f * PI;
	int current_frame = 0;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;


	/* CS_CODE_POSEOBJ format:
	 *
	 *	obj_num,
	 *	type, imgref_num, size,
	 *	x, y, z,
	 *	heading, pitch, bank,
	 *	velocity,
	 *	velocity_heading, velocity_pitch, velocity_bank,
	 *	current_frame
	 */
	sscanf(
	    arg,
"%i\
 %i %i %ld\
 %f %f %f\
 %f %f %f\
 %f\
 %f %f %f\
 %i",
	    &obj_num,
	    &type,
	    &imgref_num,
	    &size,
	    &x, &y, &z,
	    &heading, &pitch, &bank,
	    &velocity,
	    &velocity_heading, &velocity_pitch, &velocity_bank,
	    &current_frame
	);

	ASSERT_NOT_NAN(x);
	ASSERT_NOT_NAN(y);
	ASSERT_NOT_NAN(z);
	ASSERT_NOT_NAN(heading);
	ASSERT_NOT_NAN(pitch);
	ASSERT_NOT_NAN(bank);
	ASSERT_NOT_NAN(velocity);
	ASSERT_NOT_NAN(velocity_heading);
	ASSERT_NOT_NAN(velocity_pitch);
	ASSERT_NOT_NAN(velocity_bank);

	/* Object number must be valid */
	if((obj_num < 0) || (obj_num >= SW_OBJECTS_MAX))
	    return;

	/* ImgRef number must be valid */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	    return;

	/* Has the object not yet been created? */
	if(DBIsObjectGarbage(obj_num))
	{
	    /* Create it explicitly */
	    if(DBCreateExplicitObject(
		obj_num,
		imgref_num,
		type,
		-1,
		0l, 0l, 0l,
		x, y, z,
		heading, pitch, bank
	    ) < 0)
		return;

	    obj = xsw_object[obj_num];

	    /* Need to add initial antimatter to allow for initial
	     * movement
	     */
	    obj->antimatter_max = 1.0f;
	    obj->antimatter = 1.0f;

	    /* Redraw the bridge window if this is the newly created
	     * player object
	     */
	    if(obj_num == net->player_obj_num)
		BridgeWinDrawAll(xsw_bridge_win);

	    /* Request the name of this new object so that scanner
	     * contacts show the appropriate name for this object
	     */
	    NetSendReqName(obj_num);

	    if(!XSW_CON_IS_CONNECTED(net))
		return;
	}

	obj = xsw_object[obj_num];


	/* Check if the imgref has changed */
	if(obj->imgref != imgref_num)
	{
	    /* Record imgref number change and set new number */
	    imgref_changed = 1;
	    obj->imgref = imgref_num;
	}


	/* Do not update local player object once the connection state
	 * is XSW_CON_STATE_CONNECTED, with the exception that the player
	 * intercepting something
	 */
	if(obj_num == net->player_obj_num)
	{
	    if(obj->intercepting_object > -1)
	    {
		obj->heading = heading;
		obj->pitch = pitch;
		obj->bank = bank;
		SWMathSetUnitVector2D(
		    &obj->attitude_vector_compoent,
		    obj->heading
		);
	    }

	    /* Redraw bridge console panels as needed */
	    if(imgref_changed)
	    {
		/* Draw the bridge panel to reflect the change in the
		 * player object's imgref
		 */
		BridgeWinDrawPanel(obj_num, BW_PANEL_P2);
	    }

	    /* Do not continue if already connected */
	    if(XSW_CON_STATE_IS_CONNECTED(net))
		return;

	    /* Mark that we got the player's position */
	    net->flags |= XSW_CON_FLAG_SET_PLAYER_POS;

	    if((net->flags & XSW_CON_FLAG_SET_PLAYER_NUM) &&
	       (net->flags & XSW_CON_FLAG_SET_PLAYER_SECT)
	    )
	    {
		net->connection_state = XSW_CON_STATE_CONNECTED;

		/* Set player object referance */
		DBSetPlayerObject(obj_num);
	    }
	}

	/* Check if the player object is locked on to this object */
	if(net->player_obj_ptr != NULL)
	{
	    xsw_object_struct *player_obj = net->player_obj_ptr;
	    if(obj_num == player_obj->locked_on)
	    {
		if(imgref_changed)
		{
		    /* Draw the bridge panel to reflect the change in
		     * the locked on object's imgref
		     */
		    BridgeWinDrawPanel(obj_num, BW_PANEL_S1);
		}
	    }
	}


	/* Load the imgref as needed */
	imgref = ImgRefGetPtr(xsw_imgref, total_imgrefs, imgref_num);
	if(!ImgRefIsLoaded(imgref))
	{
	    ImgRefLoad(&xsw_imgref, &total_imgrefs, imgref_num);
	    loaded_imgref = 1;
	}

	/* Set new location and apperance values on object */
	obj->type = type;
	/* Already set imgref number */
	obj->size = size;

	obj->x = x;
	obj->y = y;
	obj->z = z;

	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	SWMathSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);

	obj->velocity = velocity;
	obj->velocity_heading = velocity_heading;
	obj->velocity_pitch = velocity_pitch;
	obj->velocity_bank = velocity_bank;
	SWMathSetUnitVector2D(
	    &obj->momentum_vector_compoent,
	    obj->velocity_heading
	);

	/* Increase total frames if current frame exceeds total */
	if(current_frame >= obj->animation.total_frames)
	    obj->animation.total_frames = current_frame + 1;

	/* Update animation frame only if not updating locally */
	if(!option.local_updates)
	    obj->animation.current_frame = current_frame;


	/* Refresh the last time it was updated */
	obj->last_updated = loaded_imgref ? MilliTime() : cur_millitime;
}

/*
 *	Handles a force pose object.
 */
static void NetHandleSetFPoseObj(const char *arg)
{
	char	loaded_imgref = 0,
		imgref_changed = 0;
	int obj_num = -1;
	int type = SW_OBJ_TYPE_GARBAGE;
	int imgref_num = 0;
	long size = 0l;
	float	x = 0.0f, y = 0.0f, z = 0.0f;
	float	heading = 0.0f * PI,
		pitch = 0.0f * PI,
		bank = 0.0f * PI;
	float	velocity = 0.0f;
	float	velocity_heading = 0.0f * PI,
		velocity_pitch = 0.0f * PI,
		velocity_bank = 0.0f * PI;
	int current_frame = 0;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* CS_CODE_FORCEPOSEOBJ format:
	 *
	 *	obj_num,
	 *	type, imgref_num, size,
	 *	x, y, z,
	 *	heading, pitch, bank,
	 *	velocity,
	 *	velocity_heading, velocity_pitch, velocity_bank,
	 *	current_frame
	 */
	sscanf(
	    arg,
"%i\
 %i %i %ld\
 %f %f %f\
 %f %f %f\
 %f\
 %f %f %f\
 %i",
	    &obj_num,
	    &type, &imgref_num, &size,
	    &x,	&y, &z,
	    &heading, &pitch, &bank,
	    &velocity,
	    &velocity_heading, &velocity_pitch, &velocity_bank,
	    &current_frame
	);

	ASSERT_NOT_NAN(x);
	ASSERT_NOT_NAN(y);
	ASSERT_NOT_NAN(z);
	ASSERT_NOT_NAN(heading);
	ASSERT_NOT_NAN(pitch);
	ASSERT_NOT_NAN(bank);
	ASSERT_NOT_NAN(velocity);
	ASSERT_NOT_NAN(velocity_heading);
	ASSERT_NOT_NAN(velocity_pitch);
	ASSERT_NOT_NAN(velocity_bank);

	/* Object number must be valid */
	if((obj_num < 0) || (obj_num >= SW_OBJECTS_MAX))
	    return;

	/* ImgRef number must be valid */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	    return;
	     
	/* Has the object not yet been created? */
	if(DBIsObjectGarbage(obj_num))
	{
	    if(DBCreateExplicitObject(
		obj_num,
		imgref_num,
		type,
		-1,
		0l, 0l, 0l,
		x, y, z,
		heading, pitch, bank
	    ) < 0)
		return;

	    obj = xsw_object[obj_num];

	    /* Antimatter should be set positive to ensure the allowment
	     * of movement in simulation
	     */
	    obj->antimatter_max = 1.0f;
	    obj->antimatter = 1.0f;

	    /* Redraw the bridge window if this is the newly created
	     * player object
	     */
	    if(obj_num == net->player_obj_num)
		BridgeWinDrawAll(xsw_bridge_win);

	    /* Request the name of this new object so that scanner
	     * contacts show the appropriate name for this object
	     */
	    NetSendReqName(obj_num);

	    if(!XSW_CON_IS_CONNECTED(net))
		return;
	}

	obj = xsw_object[obj_num];

	/* Check if the imgref has changed */
	if(obj->imgref != imgref_num)
	{
	    imgref_changed = 1;
	    obj->imgref = imgref_num;
	}


	/* Handle connection state change if any */
	if((obj_num == net->player_obj_num) &&
	   !XSW_CON_STATE_IS_CONNECTED(net)
	)
	{
	    /* This is force pose, player object may be changed */

	    /* Mark that we got the player's position */
	    net->flags |= XSW_CON_FLAG_SET_PLAYER_POS;

	    if((net->flags & XSW_CON_FLAG_SET_PLAYER_NUM) && 
	       (net->flags & XSW_CON_FLAG_SET_PLAYER_SECT)
	    )
	    {
		net->connection_state = XSW_CON_STATE_CONNECTED;

		/* Set player object referance */
		DBSetPlayerObject(obj_num);
	    }
	}


	/* Is this the player object? */
	if(obj_num == net->player_obj_num)
	{
	    if(imgref_changed)
	    {
		/* Draw the bridge panel to reflect the change in the
		 * player object's imgref
		 */
		BridgeWinDrawPanel(obj_num, BW_PANEL_P2);
	    }
	}

	/* Check if the player object is locked on to this object */
	if(net->player_obj_ptr != NULL)
	{
	    xsw_object_struct *player_obj = net->player_obj_ptr;
	    if(obj_num == player_obj->locked_on)
	    {
		if(imgref_changed)
		{
                    /* Draw the bridge panel to reflect the change in
		     * the locked on object's imgref
		     */
		    BridgeWinDrawPanel(obj_num, BW_PANEL_S1);
		}
	    }
	}


	/* Load the imgref as needed */
	imgref = ImgRefGetPtr(xsw_imgref, total_imgrefs, imgref_num);
	if(!ImgRefIsLoaded(imgref))
	{
	    ImgRefLoad(&xsw_imgref, &total_imgrefs, imgref_num);
	    loaded_imgref = 1;
	}


	/* Set force posed values on the object */
	obj->type = type;
	/* imgref already set */
	obj->size = size;

	obj->x = x;
	obj->y = y;
	obj->z = z;

	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	SWMathSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);

	obj->velocity = velocity;
	obj->velocity_heading = velocity_heading;
	obj->velocity_pitch = velocity_pitch;
	obj->velocity_bank = velocity_bank;
	SWMathSetUnitVector2D(
	    &obj->momentum_vector_compoent,
	    obj->velocity_heading
	);

	/* Increase total frames if current frame exceeds total */
	if(current_frame >= obj->animation.total_frames)
	    obj->animation.total_frames = current_frame + 1;

	/* Update animation frame only if not updating locally */
	if(!option.local_updates)
	    obj->animation.current_frame = current_frame;

	/* Refresh the last time it was updated */
	obj->last_updated = loaded_imgref ? MilliTime() : cur_millitime;

	/* Must respond with a pose that has the exact same values */
	NetSendObjPose(obj_num);
}

/*
 *	Handles set object values.
 */
static void NetHandleSetObjectValues(const char *arg)
{
	char	loc_changed = 0,
		changed_locked,
		changed_intercepting,
		changed_selected_weapon,
		lighting_changed;

	int obj_num = -1;
	int loc_type = SW_LOC_TYPE_SPACE;
	int loc_obj_num = -1;
	int locked_on = -1;
	int intercepting_object = -1;
	int thrust_rev_state = 0;
	float thrust_dir = 0.0f * PI;
	float thrust = 0.0f;
	float throttle = 0.0f;
	int lighting = 0;		/* sw_flags_t */
	float hp = 0.0f;
	float power = 0.0f;
	float antimatter = 0.0f, prev_antimatter;
	int shield_state = SW_SHIELD_STATE_NONE, prev_shield_state;
	int selected_weapon = -1;
	int cloak_state = SW_CLOAK_STATE_NONE, prev_cloak_state;
	float cloak_strength = 0.0f;
	float visibility = 0.0f;
	int damage_control = SW_DMGCTL_STATE_OFF;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;


	/* SWEXTCMD_STDOBJVALS format:
	 *
	 *	obj_num
	 *	loc_type loc_obj_num locked_on intercepting_object
	 *	thrust_rev_state thrust_dir thrust throttle
	 *	lighting hp power
	 *	antimatter shield_state selected_weapon cloak_state
	 *	cloak_strength visibility damage_control
	 */
	sscanf(
	    arg,
"%i\
 %i %i %i %i\
 %i %f %f %f\
 %i %f %f\
 %f %i %i %i\
 %f %f %i",
	    &obj_num,
	    &loc_type, &loc_obj_num, &locked_on, &intercepting_object,
	    &thrust_rev_state, &thrust_dir, &thrust, &throttle,
	    &lighting, &hp, &power,
	    &antimatter, &shield_state, &selected_weapon, &cloak_state,
	    &cloak_strength, &visibility, &damage_control
	);

	ASSERT_NOT_NAN(thrust_dir);
	ASSERT_NOT_NAN(thrust);
	ASSERT_NOT_NAN(throttle);
	ASSERT_NOT_NAN(hp);
	ASSERT_NOT_NAN(power);
	ASSERT_NOT_NAN(antimatter);
	ASSERT_NOT_NAN(cloak_strength);
	ASSERT_NOT_NAN(visibility);

	/* Object not created yet? */
	if(DBIsObjectGarbage(obj_num))
	{
	    /* Note we do not create the object since this function
	     * does not get enough information to create an object
	     */
	    return;
	}

	obj = xsw_object[obj_num];


	/* Set object values */

	/* Check if the locked on object has changed */
	if(DBIsObjectGarbage(locked_on))
	{
	    changed_locked = (obj->locked_on > -1) ? 1 : 0;
	    obj->locked_on = -1;
	}
	else
	{
	    changed_locked = (obj->locked_on == locked_on) ? 0 : 1;
	    obj->locked_on = locked_on;
	}

	/* Click if the intercepting object has changed */
	changed_intercepting =
	    (obj->intercepting_object != intercepting_object) ? 1 : 0;
	obj->intercepting_object = intercepting_object;

	/* Check if the selected weapon has changed */
	changed_selected_weapon =
	    (obj->selected_weapon != selected_weapon) ? 1 : 0;

	/* Set the velocity and thrust values only if this is not the
	 * player object
	 *
	 * The player object must not be affected by movement
	 * settings given by server
	 */
	if(obj_num != net->player_obj_num)
	{
	    obj->loc_type = loc_type;
	    if(obj->loc_obj_num != loc_obj_num)
		loc_changed = 1;
	    obj->loc_obj_num = loc_obj_num;
	    obj->thrust_dir = thrust_dir;
	    obj->thrust = thrust;
	    obj->throttle = throttle;
	}
	lighting_changed = (obj->lighting != (sw_flags_t)lighting) ? 1 : 0;
	obj->lighting = (sw_flags_t)lighting;
	obj->hp = hp;
	obj->power = power;
	prev_antimatter = obj->antimatter;
	obj->antimatter = antimatter;
	prev_shield_state = obj->shield_state;
	obj->shield_state = shield_state;
	obj->selected_weapon = selected_weapon;
	prev_cloak_state = obj->cloak_state;
	obj->cloak_state = cloak_state;
	obj->cloak_strength = cloak_strength;
	obj->visibility = visibility;
	obj->damage_control = damage_control;


	/* Mark the last updated time */
	obj->last_updated = cur_millitime;


	/* Is this the player? */
	if(obj_num == net->player_obj_num)
	{
	    sw_legend_struct *legend = &sw_legend;
	    xsw_bridge_win_struct *bw = xsw_bridge_win;

	    /* Draw the bridge panels to reflect the player object's
	     * new values
	     */
	    if(lighting_changed)
		BridgeWinDrawPanel(obj_num, BW_PANEL_P2);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PHULL);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PPOWER);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PSHIELDS);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PVIS);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PDMGCTL);
	    if(changed_intercepting)
		BridgeWinDrawPanel(obj_num, BW_PANEL_PINAME);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PWEAPONS);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PANTIMATTER);
#if 0
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PENGINESTATE);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PTHRUSTVECTOR);
#endif
	    if(changed_locked)
	    {
		BridgeWinDrawPanel(obj_num, BW_PANEL_PWLOCK);
		BridgeWinDrawPanel(locked_on, BW_PANEL_S1);
		BridgeWinDrawPanel(locked_on, BW_PANEL_S2);
		BridgeWinDrawPanel(locked_on, BW_PANEL_S3);

		if((obj->locked_on > -1) &&
		   (option.sounds >= XSW_SOUNDS_EVENTS)
		)
		    SoundPlay(
			&sound, SW_SNDREF_SCAN_BEEP,
			1.0f, 1.0f,
			0, 0
		    );

		/* Update the background music as needed */
		if(option.music)
		    XSWUpdateBackgroundMusic();
	    }

	    /* Change in shield state? */
	    if(obj->shield_state != prev_shield_state)
	    {
#if 0
		if((obj->shield_state == SW_SHIELD_STATE_UP) &&
		   (prev_shield_state == SW_SHIELD_STATE_DOWN)
		)
		    SoundPlay(
			&sound, SW_SNDREF_SHIELD_UP,
			1.0f, 1.0f,
			0, 0
		    );
		else if(((obj->shield_state == SW_SHIELD_STATE_DOWN) ||
			 (obj->shield_state == SW_SHIELD_STATE_NONE)) &&
			(prev_shield_state == SW_SHIELD_STATE_UP)
		)
		    SoundPlay(
			&sound, SW_SNDREF_SHIELD_DOWN,
			1.0f, 1.0f,
			0, 0
		    );
#endif
	    }

	    /* Change in cloak state? */
	    if(obj->cloak_state != prev_cloak_state)
	    {
#if 0
		if((obj->cloak_state == SW_CLOAK_STATE_UP) &&
		   (prev_cloak_state == SW_CLOAK_STATE_DOWN)
		)
		    SoundPlay(
			&sound, SW_SNDREF_CLOAK_UP,
			1.0f, 1.0f,
			0, 0
		    );
		else if(((obj->cloak_state == SW_CLOAK_STATE_DOWN) ||
			 (obj->cloak_state == SW_CLOAK_STATE_NONE)) &&
			(prev_cloak_state == SW_CLOAK_STATE_UP)
		)
		    SoundPlay(
			&sound, SW_SNDREF_CLOAK_DOWN,
			1.0f, 1.0f,
			0, 0
		    );
#endif
	    }

	    /* Antimatter low warning check */
	    if((legend->antimatter_low_warning > 0.0f) &&
	       (obj->antimatter_max > legend->antimatter_low_warning)
	    )
	    {
		const float	warning_threshold = legend->antimatter_low_warning,
				danger_threshold = warning_threshold / 2.0f;

		/* Reduction in antimatter? */
		if(obj->antimatter < prev_antimatter)
		{
		    if(((obj->antimatter <= warning_threshold) &&
		        (prev_antimatter > warning_threshold)) ||
		       ((obj->antimatter <= danger_threshold) &&
			(prev_antimatter > danger_threshold))
		    )
		    {
			XSWPrintWarning(
"Antimatter low."
			);
			if(option.sounds >= XSW_SOUNDS_EVENTS)
			    SoundPlay(
				&sound, SW_SNDREF_ANTIMATTER_LOW,
				1.0f, 1.0f,
				0, 0
			    );
		    }
		    /* Completely out of antimatter? */
		    else if((obj->antimatter <= 0.0f) &&
			    (prev_antimatter > 0.0f)
		    )
		    {
			XSWPrintWarning(
"Out of antimatter."
			);
			if(option.sounds >= XSW_SOUNDS_EVENTS)
			    SoundPlay(
				&sound, SW_SNDREF_ANTIMATTER_LOW,
				1.0f, 1.0f,
				0, 0
			    );
		    }
		}
		else if(obj->antimatter > prev_antimatter)
		{
		    /* Warn if antimatter is still low despite increase */
		    if(obj->antimatter <= warning_threshold)
		    {
			XSWPrintWarning(
"Antimatter low."
			);
			if(option.sounds >= XSW_SOUNDS_EVENTS)
			    SoundPlay(
				&sound, SW_SNDREF_ANTIMATTER_LOW,
				1.0f, 1.0f,     
				0, 0
			    );
		    }
		}
	    }

	    /* Change in selected weapon? */
	    if(changed_selected_weapon)
	    {
		if(bw != NULL)
		{
		    /* Recreate selected weapon stats label on viewscreen */
		    VSDrawUpdateWeaponLabel(
			&bw->vs_weapon_image,
			bw->vs_weapon_buf
		    );
		}
	    }
	}
	/* Is this the object that the player is locked on? */
	else if(net_parms.player_obj_ptr != NULL)
	{
	    xsw_object_struct *player_obj = net_parms.player_obj_ptr;
	    if(player_obj->locked_on == obj_num)
	    {
		if(lighting_changed || loc_changed)
		{
		    BridgeWinDrawPanel(obj_num, BW_PANEL_S1);
		}
		if(changed_selected_weapon)
		{
		     BridgeWinDrawPanel(obj_num, BW_PANEL_SWEAPONS);
		}
	    }
	}
}

/*
 *	Handles a set object maximums request.
 */
static void NetHandleSetObjectMaximums(const char *arg)
{
	char loaded_imgref = 0;
	char imgref_changed = 0;
	int obj_num = -1;
	int type = SW_OBJ_TYPE_GARBAGE;
	int imgref_num = 0;
	int owner = -1;
	long size = 0l;
	float scanner_range = 0.0f;
	float velocity_max = 0.0f;
	float thrust_power = 0.0f;
	float turnrate = 0.0f;
	float hp_max = 0.0f;
	float power_max = 0.0f;
	float power_purity = 0.0f;
	float core_efficency = 0.0f;
	float antimatter_max = 0.0f;
	int total_weapons = 0;
	float visibility = 0.0f;
	imgref_struct *imgref;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_STDOBJMAXS format:
	 *
	 *	obj_num
	 *	type imgref_num owner size scanner_range
	 *	velocity_max thrust_power turnrate hp_max power_max
	 *	power_purity core_efficency antimatter_max
	 *	total_weapons visibility
	 */
	sscanf(
	    arg,
"%i\
 %i %i %i %ld %f\
 %f %f %f %f %f\
 %f %f %f\
 %i %f",
	    &obj_num,
	    &type, &imgref_num, &owner, &size, &scanner_range,
	    &velocity_max, &thrust_power, &turnrate, &hp_max, &power_max,
	    &power_purity, &core_efficency, &antimatter_max,
	    &total_weapons, &visibility
	);

	ASSERT_NOT_NAN(scanner_range);
	ASSERT_NOT_NAN(velocity_max);
	ASSERT_NOT_NAN(thrust_power);
	ASSERT_NOT_NAN(turnrate);
	ASSERT_NOT_NAN(hp_max);
	ASSERT_NOT_NAN(power_max);
	ASSERT_NOT_NAN(power_purity);
	ASSERT_NOT_NAN(core_efficency);
	ASSERT_NOT_NAN(antimatter_max);
	ASSERT_NOT_NAN(visibility);

	/* Object number must be valid */
	if((obj_num < 0) || (obj_num >= SW_OBJECTS_MAX))
	    return;

	/* ImgRef number must be valid */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	    return;

	/* Has the object not yet been created? */
	if(DBIsObjectGarbage(obj_num))
	{
	    if(DBCreateExplicitObject(
		obj_num,
		imgref_num,
		type,
		owner,
		0l, 0l, 0l,
		0.0f, 0.0f, 0.0f,
		0.0f * PI, 0.0f * PI, 0.0f * PI
	    ) < 0)
		return;

	    /* Redraw the bridge window if this is the newly created
	     * player object 
	     */
	    if(obj_num == net->player_obj_num)
		BridgeWinDrawAll(xsw_bridge_win);

	    /* Request the name of this new object so that scanner
	     * contacts show the appropriate name for this object
	     */
	    NetSendReqName(obj_num);

	    if(!XSW_CON_IS_CONNECTED(net))
		return;
	}

	obj = xsw_object[obj_num];

	/* Check if the imgref has changed */
	if(obj->imgref != imgref_num)
	{
	    imgref_changed = 1;
	    obj->imgref = imgref_num;
	}

	/* Is this the player object? */
	if(obj_num == net->player_obj_num)
	{
	    if(imgref_changed)
	    {
		/* Draw the bridge panel to reflect the change in the
		 * player object's imgref
		 */
		BridgeWinDrawPanel(obj_num, BW_PANEL_P2);
	    }

	    /* Scanner range change? */
	    if(obj->scanner_range != scanner_range)
	    {
		obj->scanner_range = scanner_range;

		/* Recreate range label */
		ScannerUpdateLabels(xsw_bridge_win, obj_num);
	    }
	}


	/* Load the imgref as needed */
	imgref = ImgRefGetPtr(xsw_imgref, total_imgrefs, imgref_num);
	if(!ImgRefIsLoaded(imgref))
	{
	    ImgRefLoad(&xsw_imgref, &total_imgrefs, imgref_num);
	    loaded_imgref = 1;
	}


	/* Set new maximum values on object */
	obj->type = type;
	/* Already set imgref number */
	obj->owner = owner;
	obj->size = size;
	obj->scanner_range = scanner_range;

	obj->velocity_max = velocity_max; 
	obj->thrust_power = thrust_power;
	obj->turnrate = turnrate;
	obj->hp_max = hp_max;
	obj->power_max = power_max;

	obj->power_purity = power_purity;
	obj->core_efficency = core_efficency;
	obj->antimatter_max = antimatter_max;

	/* Need to reallocate weapons? */
	if(obj->total_weapons != total_weapons)
	    UNVAllocObjectWeapons(obj, total_weapons);

	obj->visibility = visibility;


	/* Refresh the last time it was updated */
	obj->last_updated = (loaded_imgref) ? MilliTime() : cur_millitime;


	if(obj_num == net->player_obj_num)
	{
	    /* Draw the bridge panel's to reflect the player object's
	     * new values
	     */
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PHULL);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PPOWER);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PSHIELDS);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PVIS);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PDMGCTL);

	    BridgeWinDrawPanel(obj_num, BW_PANEL_PINAME);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PWLOCK);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PWEAPONS);

	    BridgeWinDrawPanel(obj_num, BW_PANEL_PANTIMATTER);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PENGINESTATE);
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PTHRUSTVECTOR);
	}
}

/*
 *	Set name.
 */
static void NetHandleSetName(const char *arg)
{
	int obj_num = -1;
	const char *s;
	char	*name = NULL,
		*empire = NULL;
	font_t *font;
	const WColorStruct *c_fg, *c_empire;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	/* NET_CMD_SETNAME format:
	 *
	 *	"obj_num;name;other"
	 */
	s = arg;

	/* Get object number */
	if(s != NULL)
	{
	    if(*s == '#')
		s++;
	    obj_num = atoi(s);

	    s = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	    if(s != NULL)
		s++;
	}

	/* Get name */
	if(s != NULL)
	{
	    const char *s_end = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	    const int len = (s_end != NULL) ? (s_end - s) : strlen(s);
	    if(len > 0)
	    {
		name = (char *)malloc((len + 1) * sizeof(char));
		if(name != NULL)
		{
		    memcpy(name, s, len);
		    name[len] = '\0';
		}
	    }

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}

	/* Get empire */
	if(s != NULL)
	{
	    const char *s_end = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	    const int len = (s_end != NULL) ? (s_end - s) : strlen(s);
	    if(len > 0)
	    {
		empire = (char *)malloc((len + 1) * sizeof(char));
		if(empire != NULL)
		{
		    memcpy(empire, s, len);
		    empire[len] = '\0';
		}
	    }

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}

	/* Object not created yet? */
	if(DBIsObjectGarbage(obj_num))
	{
	    free(name);
	    free(empire);
	    return;
	}

	obj = xsw_object[obj_num];

	/* Set new name */
	if(!STRISEMPTY(name))
	{
	    strncpy(obj->name, name, sizeof(obj->name));
	    obj->name[sizeof(obj->name) - 1] = '\0';
	}

	/* Set new empire */
	if(!STRISEMPTY(empire))
	{
	    strncpy(obj->empire, empire, sizeof(obj->empire));
	    obj->empire[sizeof(obj->empire) - 1] = '\0';
	}


	/* Create viewscreen label image for that object */
	font = xsw_font.vs_standard;
	c_fg = &xsw_color.vs_label_fg_cv;
	switch(UNVMatchIFFPtr(net->player_obj_ptr, obj))
	{
	  case SW_IFF_HOSTILE:
	    c_empire = &xsw_color.bp_hostile_cv;
	    break;
	  case SW_IFF_FRIENDLY:
	    c_empire = &xsw_color.bp_friendly_cv;
	    break;
	  default:
	    c_empire = &xsw_color.bp_unknown_cv;
	    break;
	}
	switch(opt->show_formal_label)
	{
	  case XSW_SHOW_FORMAL_LABEL_ALWAYS:
	    VSLabelAdd(
		DBGetFormalNameStr(obj_num),
		c_fg, c_empire, font,
		obj_num, obj
	    );
	    break;
	  case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
	    if(obj->owner == net->player_obj_num)
		VSLabelAdd(
		    DBGetFormalNameStr(obj_num),
		    c_fg, c_empire, font,
		    obj_num, obj
		);
	    else
		VSLabelAdd(
		    obj->name,
		    c_fg, c_empire, font,
		    obj_num, obj
		);
	    break;
	  default:	/* XSW_SHOW_FORMAL_LABEL_NEVER */
	    VSLabelAdd(
		obj->name,
		c_fg, c_empire, font,
		obj_num, obj
	    );
	    break;
	}


	if(obj_num == net->player_obj_num)
	{
	    /* Update the name on the title and the bridge panel */
	    BridgeWinSetTitle(
		xsw_bridge_win,
		net->address, net->title, obj->name
	    );
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PNAME);
	}
	else if(net->player_obj_ptr != NULL)
	{
	    xsw_object_struct *player_obj = net->player_obj_ptr;
	    if(obj_num == player_obj->locked_on)
	    {
		/* Update the name on the bridge panel */
		BridgeWinDrawPanel(obj_num, BW_PANEL_SNAME);
		BridgeWinDrawPanel(obj_num, BW_PANEL_SEMPIRE);
	    }
	}

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;

	free(name);
	free(empire);
}

/*
 *	Handles a set shield visibility request on an object.
 */
static void NetHandleSetShieldVis(const char *arg)
{
	int obj_num = -1;
	float shield_visibility = 0.0f;
	xsw_object_struct *obj;

	/* SWEXTCMD_SETSHIELDVIS format:
	 *
	 *	obj_num shield_visibility
	 */
	sscanf(
	    arg,
	    "%i %f",
	    &obj_num,
	    &shield_visibility
	);

	ASSERT_NOT_NAN(shield_visibility);

	/* Object not created yet? */
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Set new shield visibility */
	obj->shield_visibility = CLIP(
	    shield_visibility, 0.0, 1.0
	);

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;
}

/*
 *	Set channel.
 */
static void NetHandleSetChannel(const char *arg)
{
	int obj_num = -1;
	int channel = 0;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_SETCHANNEL format:
	 *
	 *	obj_num channel
	 */
	sscanf(
	    arg,
	    "%i %i",
	    &obj_num,
	    &channel
	);

	/* Object not created yet? */
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Set new channel */
	obj->com_channel = channel;

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;

	if(obj_num == net->player_obj_num)
	{
	    /* Draw the bridge panel to reflect the new com channel */
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PCOMCHANNEL);
	}
}

/*
 *	Set score.
 */
static void NetHandleSetScore(const char *arg)
{
	int obj_num = -1;
	xsw_object_struct *obj;
	xsw_score_struct *score;
	float credits = 0.0f;
	float rmu = 0.0f, rmu_max = 0.0f;
	float damage_given = 0.0f, damage_received = 0.0f;
	int kills = 0;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_SETSCORE format:
	 *
	 *	Client sends:
	 *
	 *	obj_num
	 *
	 *	Server sends:
	 *
	 *	obj_num
	 *	credits rmu rmu_max
	 *	damage_given damage_received kills
	 */
	sscanf(
	    arg,
	    "%i %f %f %f %f %f %i",
	    &obj_num,
	    &credits, &rmu, &rmu_max,
	    &damage_given, &damage_received, &kills
	);

	ASSERT_NOT_NAN(credits);
	ASSERT_NOT_NAN(rmu);
	ASSERT_NOT_NAN(rmu_max);
	ASSERT_NOT_NAN(damage_given);
	ASSERT_NOT_NAN(damage_received);

	/* Object not created yet? */
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Allocate scores as needed */
	if(UNVAllocScores(obj))
	    return;

	score = obj->score;

	/* Set new scores */
	score->credits = (sw_credits_t)credits;
	score->rmu = (sw_rmu_t)rmu;
	score->rmu_max = (sw_rmu_t)rmu_max;
	score->damage_given = damage_given;
	score->damage_received = damage_received;
	score->kills = kills;

	if(obj_num == net->player_obj_num)
	{
	    /* Draw the bridge panel to reflect the player object's
	     * new score
	     */
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PSCORE);
	}

	/* Redraw the economy window if it is mapped */
	if((xsw_ecow != NULL) ? xsw_ecow->map_state : False)
	    EcoWinUpdateCredits(xsw_ecow, obj_num);

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;
}

/*
 *	Set engine.
 */
static void NetHandleSetEngine(const char *arg)
{
	int obj_num = -1;
	int engine_state = SW_ENGINE_STATE_NONE;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_SETENGINE format:
	 *
	 *	obj_num engine_state
	 */
	sscanf(
	    arg,
	    "%i %i",
  	    &obj_num, &engine_state
	);

	/* Object not created yet? */
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Engine state changed? */
	if(obj->engine_state != engine_state)
	{
	    obj->engine_state = engine_state;

	    if(obj_num == net->player_obj_num)
	    {
		/* Draw the bridge panel to reflect the player object's
		 * new engine state
		 */
		BridgeWinDrawPanel(obj_num, BW_PANEL_PENGINESTATE);
		switch(engine_state)
		{
		  case SW_ENGINE_STATE_ON:
		    XSWPrint("Engines: On");
		    break;
                  case SW_ENGINE_STATE_STARTING:
                    XSWPrint("Engines: Starting...");
                    break;
                  case SW_ENGINE_STATE_OFF:
                    XSWPrint("Engines: Off");
                    break;
		}
	    }
	}

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;
}

/*
 *	Set object particle trail.
 */
static void NetHandleSetObjParticleTrail(const char *arg)
{
	int obj_num = -1;
	int op = SW_PARTICLE_TRAIL_OP_FIRE_OFF;
	float	offset = 0.0f,
		heading = 0.0f * PI,
		pitch = 0.0f * PI,
		bank = 0.0f * PI;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_SET_OBJ_PARTICLE_TRAIL format:
	 *
	 *	obj_num op offset heading pitch bank
	 */
	sscanf(
	    arg,
	    "%i %i %f %f %f %f",
	    &obj_num, &op, &offset, &heading, &pitch, &bank
	);

	ASSERT_NOT_NAN(offset);
	ASSERT_NOT_NAN(heading);
	ASSERT_NOT_NAN(pitch);
	ASSERT_NOT_NAN(bank);

	/* Object not created yet? */
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	switch(op)
	{
	  case SW_PARTICLE_TRAIL_OP_FIRE_OFF:
	    obj->ext_flags &= ~XSW_OBJ_FLAG_ON_FIRE;
	    break;
	  case SW_PARTICLE_TRAIL_OP_FIRE_ON:
	    obj->ext_flags |= XSW_OBJ_FLAG_ON_FIRE;
	    break;
	  case SW_PARTICLE_TRAIL_OP_SMOKE_OFF:
	    obj->ext_flags &= ~XSW_OBJ_FLAG_SMOKE_TRAIL;
	    break;
	  case SW_PARTICLE_TRAIL_OP_SMOKE_ON:
	    obj->ext_flags |= XSW_OBJ_FLAG_SMOKE_TRAIL;
	    break;
	}

	/* Is this the player object? */
	if(obj_num == net->player_obj_num)
	{


	}

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;
}

/*
 *	Create effect.
 */
static void NetHandleCreateEffect(const char *arg)
{
	int     type,
		imgref_num = -1,
		ref_obj_num = -1;
	unsigned int flags = 0;
	float   x = 0.0f, y = 0.0f, z = 0.0f;
	long    sect_x = 0l,
		sect_y = 0l,
		sect_z = 0l,
		size_start = 0l,
		size_end = 0l,
		life_span = 0l;

	/* SWEXTCMD_CREATE_EFFECT format:
	 *
	 *	flags imgref_num sect_x sect_y sect_z x y z
	 *	size_start size_end life_span ref_obj_num
	 */
	sscanf(
	    arg,
"%ui %i\
 %ld %ld %ld\
 %f %f %f\
 %ld %ld %ld\
 %i",
	    &flags, &imgref_num,
	    &sect_x, &sect_y, &sect_z,
	    &x, &y, &z,
	    &size_start, &size_end, &life_span,
	    &ref_obj_num
	);
	  
	ASSERT_NOT_NAN(x);
	ASSERT_NOT_NAN(y);
	ASSERT_NOT_NAN(z);

	if(flags & SW_EFFECT_FLAG_DRIFT)
	    type = XSW_PARTICLE_TYPE_SMOKE;
	else
	    type = XSW_PARTICLE_TYPE_EFFECT;

	/* Create a particle as the effect */
	ParticleCreate(
	    &xsw_particle, &total_particles,
	    type,
	    imgref_num,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    0.0f * PI, 0.0f * PI, 0.0f * PI,
	    size_start, size_end,
	    cur_millitime, life_span
	);
}

/*
 *	Create shockwave.
 */
static void NetHandleCreateShockwave(const char *arg)
{
	int	type = 0,
		ref_obj_num = -1;
	float	x = 0.0f, y = 0.0f, z = 0.0f,
		damage_hull = 0.0f,
		damage_shield = 0.0f;
	long	sect_x = 0l,
		sect_y = 0l,
		sect_z = 0l,
		size_start = 0l,
		size_end = 0l,
		life_span = 0l;

	/* SWEXTCMD_CREATE_SHOCKWAVE format:
	 *
	 *	type sect_x sect_y sect_z x y z size_start size_end
	 *	life_span damage_hull damage_shield ref_obj_num
	 */
	sscanf(
	    arg,
"%i %ld %ld %ld %f %f %f\
 %ld %ld %ld %f %f %i",
	    &type, &sect_x, &sect_y, &sect_z, &x, &y, &z,
	    &size_start, &size_end, &life_span,
	    &damage_hull, &damage_shield, &ref_obj_num
	);

	ASSERT_NOT_NAN(x);
	ASSERT_NOT_NAN(y);
	ASSERT_NOT_NAN(z);

	/* Create a particle as the shockwave */
	ParticleCreate(
	    &xsw_particle, &total_particles,
	    XSW_PARTICLE_TYPE_SHOCKWAVE,
	    -1,					/* No ImgRef */
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    0.0f * PI, 0.0f * PI, 0.0f * PI,
	    size_start, size_end,
	    cur_millitime, life_span
	);
}


/*
 *	Notify hit.
 */
static void NetHandleNotifyHit(const char *arg)
{
	int	wep_obj_num = -1, wep_owner_obj_num, tar_obj_num = -1,
		shield_state = SW_SHIELD_STATE_DOWN;
	float	total_damage = 0.0f, bearing = 0.0f * PI,
		damage_hull = 0.0f, damage_shield = 0.0f,
		prev_hp = 0.0f, prev_power = 0.0f,
		hp = 0.0f, power = 0.0f;
	xsw_object_struct *wep_obj, *tar_obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_NOTIFYHIT format:
	 *
	 *	wep_obj_num tar_obj_num
	 *	total_damage bearing
	 *	damage_hull damage_shield
	 *	hp power
	 */
	sscanf(
	    arg,
"%i %i\
 %f %f\
 %f %f\
 %f %f",
	    &wep_obj_num, &tar_obj_num,
	    &total_damage, &bearing,
	    &damage_hull, &damage_shield,
	    &hp, &power
	);

	ASSERT_NOT_NAN(total_damage);
	ASSERT_NOT_NAN(bearing);
	ASSERT_NOT_NAN(damage_hull);
	ASSERT_NOT_NAN(damage_shield);
	ASSERT_NOT_NAN(hp);
	ASSERT_NOT_NAN(power);

	/* Get the weapon object if it is valid */
	wep_obj = DBIsObjectGarbage(wep_obj_num) ?
	    NULL : xsw_object[wep_obj_num];

	/* Get the weapon object's owner if available */
	wep_owner_obj_num = (wep_obj != NULL) ?
	    wep_obj->owner : -1;

	/* Get the target object if it is valid */
	tar_obj = DBIsObjectGarbage(tar_obj_num) ?
	    NULL : xsw_object[tar_obj_num];
	if(tar_obj != NULL)
	{
	    prev_hp = tar_obj->hp,
	    prev_power = tar_obj->power;
	    shield_state = tar_obj->shield_state;

	    /* Update the target object's values */
	    tar_obj->hp = hp;
	    tar_obj->power = power;

	    /* Player object exists? */
	    if(net->player_obj_ptr != NULL)
	    {
		xsw_object_struct *player_obj = net->player_obj_ptr;
	
		/* Is the player the target object? */
		if(net->player_obj_num == tar_obj_num)
		{
		    const int obj_num = tar_obj_num;
		    BridgeWinDrawPanel(obj_num, BW_PANEL_PHULL);
		    BridgeWinDrawPanel(obj_num, BW_PANEL_PPOWER);
		    BridgeWinDrawPanel(obj_num, BW_PANEL_PSHIELDS);
		}
		/* Is the player locked on the target object? */
		else if(player_obj->locked_on == tar_obj_num)
		{
		    const int obj_num = tar_obj_num;
		    BridgeWinDrawPanel(obj_num, BW_PANEL_SHULL);
		    BridgeWinDrawPanel(obj_num, BW_PANEL_SPOWER);
		    BridgeWinDrawPanel(obj_num, BW_PANEL_SSHIELDS);
		}
	    }
	}

	/* Notify hit */
	XSWHitNotify(
	    wep_obj_num,	/* Object that hit tar_obj_num */
	    tar_obj_num,	/* Object that was hit */
	    wep_owner_obj_num,	/* Owner of the object that hit
				 * tar_obj_num */
	    damage_hull,
	    damage_shield,
	    bearing
	);

	/* Notify damage */
	XSWDamageNotify(
	    tar_obj_num,
	    prev_hp, hp,
	    prev_power, power,
	    shield_state
	);
}

/*
 *	Notify destroy.
 */
static void NetHandleNotifyDestroy(const char *arg)
{
	int reason = SW_DESTROY_REASON_UNSPECIFIC;
	int	destroyed_obj_num = -1, destroyer_obj_num = -1,
		destroyer_obj_owner_num = -1;

	/* SWEXTCMD_NOTIFYDESTROY format:
	 *
	 *	reason
	 *	destroyed_obj_num destroyer_obj_num
	 *	destroyer_obj_owner_num
	 */
	sscanf(
	    arg,
	    "%i %i %i %i",
	    &reason,
	    &destroyed_obj_num, &destroyer_obj_num,
	    &destroyer_obj_owner_num
	);

	XSWDestroyNotify(
	    destroyer_obj_num,		/* Object that caused destruction */
	    destroyed_obj_num,		/* Object that was destroyed */
	    destroyer_obj_owner_num,	/* Owner of destroyer_obj */
	    reason			/* One of SW_DESTROY_REASON_* */
	);
}

/*
 *	Tractor beam lock.
 */
static void NetHandleTractorBeamLock(const char *arg)
{
	int obj_num = -1, tar_obj_num = -1;
	int prev_ntractored_objects;
	xsw_object_struct *obj;

	/* SWEXTCMD_TRACTORBEAMLOCK format:
	 *
	 *      obj_num tar_obj_num
	 */
	sscanf(
	    arg,
	    "%i %i",
	    &obj_num, &tar_obj_num
	);

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];
	prev_ntractored_objects = obj->total_tractored_objects;

	/* Lock/unlock tractor beam */
	DBObjectTractor(obj_num, tar_obj_num);

#if 0
	if(option.sounds >= XSW_SOUNDS_EVENTS)
	{
	    if(obj->total_tractored_objects > prev_ntractored_objects)
		SoundPlay(
		    &sound,
		    SW_SNDREF_TRACTOR_BEAM_ENGAGE,
		    1.0f, 1.0f,
		    0, 0
		);
	    else if(obj->total_tractored_objects < prev_ntractored_objects)
		SoundPlay(
		    &sound,
		    SW_SNDREF_TRACTOR_BEAM_DISENGAGE,
		    1.0f, 1.0f,
		    0, 0
		);
	}
#endif

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;
}

/*
 *	Hail.
 */
static void NetHandleHail(const char *arg)
{
	int obj_num = -1, tar_obj_num = -1, channel = 0;
	float bearing = 0.0f * PI;
	char *s;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_HAIL format:
	 *
	 *	obj_num tar_obj_num
	 *	bearing channel
	 */
	sscanf(
	    arg,
	    "%i %i %f %i",
	    &obj_num,		/* Who sent the hail */
	    &tar_obj_num,	/* Who the hail is for */
	    &bearing,
	    &channel
	);

	ASSERT_NOT_NAN(bearing);
 
        /* Get the object that sent this hail */
	obj = DBIsObjectGarbage(obj_num) ?
	    NULL : xsw_object[obj_num];

        /* Check if this object is on the ignored objects list */
        if(IgnoredObjListMatch(
            ignored_obj_list, total_ignored_objs,
            obj_num, (obj != NULL) ? obj->name : NULL   
        ) > -1)
        {
            return;
        }

	/* Is the object sending the hail known? */
	if(obj != NULL)
	{
	    /* We sent this hail? */
	    if(obj_num == net->player_obj_num)
	    {
		xsw_object_struct *tar_obj = DBIsObjectGarbage(tar_obj_num) ?
		    NULL : xsw_object[tar_obj_num];

		/* We sent that hail */
		if(option.sounds >= XSW_SOUNDS_EVENTS)
		    SoundPlay(
			&sound, SW_SNDREF_HAIL_OUTGOING,
			1.0f, 1.0f,
			0,
			0
		    );

		s = (char *)malloc(
		    40 +
		    ((tar_obj != NULL) ? STRLEN(tar_obj->name) : 0) +
		    80
		);
		if(tar_obj != NULL)
		    sprintf(
			s,
"Hail sent on channel %.2f to %s.",
		        (float)channel / 100.0f,
			tar_obj->name
		    );
		else if(tar_obj_num > -1)
		    sprintf(
			s,
"Hail sent on channel %.2f to unknown object.",
			(float)channel / 100.0f
		    );
		else
		    sprintf(     
			s,
"General hail sent on channel %.2f.",
			(float)channel / 100.0f
		    );
		XSWPrint(s);
		free(s);
	    }
	    else
	    {
		/* Someone else sent this hail */
		if(option.sounds >= XSW_SOUNDS_EVENTS)
		    SoundPlay(
			&sound, SW_SNDREF_HAIL_INCOMING,
			1.0f, 1.0f,
			0,
			0
		    );

		s = (char *)malloc(STRLEN(obj->name) + 40 + 40 + 80);
	        sprintf(
		    s,
"%s hailing on channel %.2f, bearing %.0f'.",
		    obj->name,
		    (float)channel / 100.0f,
		    (float)RADTODEG(bearing)
	        );
		XSWPrintBold(s);
		free(s);
	    }
	}
	else
	{
	    if(option.sounds >= XSW_SOUNDS_EVENTS)
		SoundPlay(
		    &sound, SW_SNDREF_HAIL_INCOMING,
		    1.0f, 1.0f,
		    0,
		    0
		);

	    s = (char *)malloc(40 + 40 + 80);
	    sprintf(
		s,
"Unknown hail received on channel %.2f, bearing %.0f'.",
		(float)channel / 100.0f,
		RADTODEG(bearing)
	    );
	    XSWPrintBold(s);
	    free(s);
	}

	/* Mark the last updated time */
	if(obj != NULL)
	    obj->last_updated = cur_millitime;
}

/*
 *	Com message.
 */
static void NetHandleComMessage(const char *arg)
{
	int obj_num = -1, tar_obj_num = -1;
	const char *s;
	char *larg, *msg, *msg2;
	float bearing = 0.0f * PI;
	int channel = 0;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_COMMESSAGE format:
	 *
	 *	obj_num tar_obj_num
	 *	bearing channel;message
	 */
	s = strchr(arg, CS_STRING_DELIMINATOR_CHAR);
	if(s != NULL)
	{
	    const int len = s - arg;

	    larg = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(larg, arg, len);
	    larg[len] = '\0';

	    s++;
	    while(ISBLANK(*s))
		s++;
	    msg = STRDUP(s);
	}
	else
	{
	    larg = STRDUP(arg);
	    msg = STRDUP("");
	}

	sscanf(
	    larg,
	    "%i %i %f %i",
	    &obj_num,		/* Object sending the message */
	    &tar_obj_num,	/* Object to recieve or -1 for any */
	    &bearing,
	    &channel
	);
	free(larg);

	ASSERT_NOT_NAN(bearing);

	/* Get the object that sent this message */
	obj = DBIsObjectGarbage(obj_num) ?
	    NULL : xsw_object[obj_num];

	/* Check if this object is on the ignored objects list */
	if(IgnoredObjListMatch(
	    ignored_obj_list, total_ignored_objs,
	    obj_num, (obj != NULL) ? obj->name : NULL
	) > -1)
	{
	    free(msg);
	    return;
	}

	/* Is the sending object known? */
	if(obj != NULL)
	{
	    /* Format the message based on if the player sent this
	     * message or it was sent by another object
	     */
	    msg2 = (char *)malloc(
		STRLEN(obj->name) + 40 + 40 + STRLEN(msg) + 80
	    );
	    if(obj == net->player_obj_ptr)
		sprintf(
		    msg2,
"<%s> -> %.2f: \"%s\"",
		    obj->name,
		    (float)((float)channel / 100.0f),
		    msg
		);
	    else
	        sprintf(
		    msg2,
"<%s> %.0f'@%.2f: \"%s\"",
		    obj->name,
		    RADTODEG(bearing),
		    (float)channel / 100.0f,
		    msg
	        );
	}
	else
	{
	    /* Sending object was not specified or is unknown, format
	     * the message accordingly
	     */
	    msg2 = (char *)malloc(
		40 + 40 + STRLEN(msg) + 80
	    );
	    sprintf(
		msg2,
"*Unknown* %.0f'@%.2f: \"%s\"",
		RADTODEG(bearing),
		(float)channel / 100.0f,
		msg
	    );
	}
	XSWPrint(msg2);

	/* Log this message? */
	if(option.log_net)
	    XSWLog(msg2);

	free(msg2);
	free(msg);
}

/*
 *	Weapon disarm.
 */
static void NetHandleWeaponDisarm(const char *arg)
{
	int	obj_num = -1,
		tar_obj_num = -1;
	xsw_option_struct *opt = &option;
	xsw_net_parms_struct *net = &net_parms;  

	/* SWEXTCMD_WEPDISARM format:
	 *
	 *	obj_num tar_obj_num
	 */
	sscanf(
	    arg,
	    "%i %i",
	    &obj_num, &tar_obj_num
	);

	/* Ignore if the object is not the player */
	if(obj_num != net->player_obj_num)
	    return;

	/* Weapon object valid? */
	if(!DBIsObjectGarbage(tar_obj_num))
	{
	    /* Print disarm notify */
	    char *s, *name;
	    xsw_object_struct *wep_obj_ptr = xsw_object[tar_obj_num];

	    /* Print disarm notify */
	    switch(opt->show_formal_label)
	    {
	      case XSW_SHOW_FORMAL_LABEL_ALWAYS:
		name = STRDUP(DBGetFormalNameStr(tar_obj_num));
		break;
	      case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
		name = (wep_obj_ptr->owner == net->player_obj_num) ?
		    STRDUP(DBGetFormalNameStr(tar_obj_num)) :
		    STRDUP(wep_obj_ptr->name);
		break;
	      default:	/* XSW_SHOW_FORMAL_LABEL_NEVER */
		name = STRDUP(wep_obj_ptr->name);
		break;
	    }
	    s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
		"Disarmed weapon `%s'.",
		name
	    );
	    XSWPrint(s);
	    free(s);
	    free(name);
	}
}

/*
 *	Auto destruct.
 */
static void NetHandleSetAutoDestruct(const char *arg)
{
	int obj_num = -1;
	unsigned long sec = 0l;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;  

	/* SWEXTCMD_AUTODESTRUCT format:
	 *
	 *	obj_num sec
	 */
	sscanf(
	    arg,
	    "%i %lu",
	    &obj_num, &sec
	);

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];
	if(sec > 0l)
	    obj->auto_destruct_next = (time_t)(cur_systime + sec);
	else
	    obj->auto_destruct_next = 0l;

	if(obj_num == net->player_obj_num)
	{
	    char *s = (char *)malloc(40 + 80);
	    if(sec > 0l)
	    {
		sprintf(
		    s,
"Setting auto-destruct for %lu seconds...",
		    sec
		);
		XSWPrintWarning(s);
	    }
	    else
	    {
		strcpy(s, "Auto-destruct aborted.");
		XSWPrintBold(s);
	    }
	    free(s);
	}
}

/*
 *	Set description.
 */
static void NetHandleSetDescription(const char *arg)
{
	int obj_num;
	const char *desc;
	xsw_object_struct *obj;
	xsw_option_struct *opt = &option;
	xsw_net_parms_struct *net = &net_parms;

	/* SWEXTCMD_SETDESCRIPTION format:
	 *
	 *	obj_num;description
	 */
	while(ISBLANK(*arg))
	    arg++;
	obj_num = ATOI(arg);

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	desc = strchr(arg, CS_STRING_DELIMINATOR_CHAR);
	if(desc != NULL)
	{
	    desc++;

	    free(obj->description);
	    obj->description = strdup(desc);

	    /* Just print it if it is received */
	    if(obj_num != net->player_obj_num)
	    {
		char *s, *name;

		switch(opt->show_formal_label)
		{
		  case XSW_SHOW_FORMAL_LABEL_ALWAYS:
		    name = STRDUP(DBGetFormalNameStr(obj_num));
		    break;
		  case XSW_SHOW_FORMAL_LABEL_AS_NEEDED:
		    name = (obj->owner == net->player_obj_num) ?
			STRDUP(DBGetFormalNameStr(obj_num)) :
			STRDUP(obj->name);
		    break;
		  default:	/* XSW_SHOW_FORMAL_LABEL_NEVER */
		    name = STRDUP(obj->name);
		    break;
		}
		s = (char *)malloc(
		    STRLEN(desc) + STRLEN(name) + 80
		);
		sprintf(
		    s,
		    "%s - %s", name,
		    desc
		);
		XSWPrint(s);
		free(s);
		free(name);
	    }
	}
	else
	{
	    free(obj->description);
	    obj->description = NULL;
	}

}

/*
 *	Economy set values.
 */
static void NetHandleEcoSetValues(const char *arg)
{
	int obj_num = -1;
	eco_flags_t flags = 0l;
	float tax_general = 0.0f, tax_friend = 0.0f, tax_hostile = 0.0f;
	xsw_ecodata_struct *eco;
	xsw_object_struct *obj; 

	/* SWEXTCMD_ECO_SETVALUES format:
	 *
	 *      obj_num flags
	 *	tax_general tax_friend tax_hostile
	 */
	sscanf(
	    arg,
	    "%i %ld %f %f %f",
	    &obj_num, &flags,
	    &tax_general, &tax_friend, &tax_hostile
	);

	ASSERT_NOT_NAN(tax_general);
	ASSERT_NOT_NAN(tax_friend);
	ASSERT_NOT_NAN(tax_hostile);

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Allocate scores and eco (as needed) on object */
	if(UNVAllocScores(obj))
	    return;
	if(UNVAllocEco(obj))
	    return;

	/* Get pointer to eco structure, make sure it's allocated */
	eco = obj->eco;
	if(eco == NULL)
	    return;

	/* Set values */
	eco->flags = flags;
	eco->tax_general = tax_general;
	eco->tax_friend = tax_friend;
	eco->tax_hostile = tax_hostile;
}

/*
 *	Economy set product values.
 */
static void NetHandleEcoSetProductValues(const char *arg)
{
	int obj_num = -1;
	char	*s,
		larg[CS_DATA_MAX_LEN];
	xsw_object_struct *obj;
	xsw_ecoproduct_struct product;
	xsw_net_parms_struct *net = &net_parms;

	strncpy(larg, arg, sizeof(larg));
	larg[sizeof(larg) - 1] = '\0';

	/* SWEXTCMD_ECO_SETPRODUCTVALUES format:
	 *
	 *	obj_num
	 *	ocs_code sell_price buy_price
	 *	product_amount product_max;product_name
	 */
	s = strchr(larg, CS_STRING_DELIMINATOR_CHAR);
	if(s == NULL)
	    return;

	memset(&product, 0x00, sizeof(xsw_ecoproduct_struct));
	strncpy(product.name, s + 1, sizeof(product.name));
	product.name[sizeof(product.name) - 1] = '\0';

	*s = '\0';

	sscanf(
	    larg,
	    "%i %i %f %f %f %f",
	    &obj_num,
	    &product.ocs_code,
	    &product.sell_price, &product.buy_price,
	    &product.amount, &product.amount_max
	);

	ASSERT_NOT_NAN(product.sell_price);
	ASSERT_NOT_NAN(product.buy_price);
	ASSERT_NOT_NAN(product.amount);
	ASSERT_NOT_NAN(product.amount_max);

	/* Check if obj_num is valid */
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Allocate scores and eco (as needed) on object */
	if(UNVAllocScores(obj))
	    return;
	if(UNVAllocEco(obj))
	    return;

	/* Allocate product on object (as needed) and update values */
	if(DBCreateObjectEconomyProduct(
	    obj_num,
	    product.name,
	    product.ocs_code,
	    product.sell_price,
	    product.buy_price,
	    product.amount,
	    product.amount_max
	))
	    return;

	EcoWinAddProduct(
	    xsw_ecow,
	    obj_num,
	    product
	);

	if(obj_num != net->player_obj_num)
	{
	    /* Map the economy window as needed */
	    if((xsw_ecow != NULL) ? !xsw_ecow->map_state : False)
	        EcoWinMap(xsw_ecow);
	}
}


/*
 *	Starchart add object.
 */
static void NetHandleStarChartAddObject(const char *arg)
{
	int obj_num = -1;
	int type = SW_OBJ_TYPE_GARBAGE;
	int imgref_num = 0;
	long size = 0l;
	long	sect_x = 0l, sect_y = 0l, sect_z = 0l;
	float	x = 0.0f, y = 0.0f, z = 0.0f;
	float	heading = 0.0f * PI,
		pitch = 0.0f * PI,
		bank = 0.0f * PI;

	/* SWEXTCMD_STARCHART_ADD_OBJECT format:
	 *
	 *      obj_num
	 *      type imgref_num size
	 *      sect_x sect_y sect_z
	 *      x y z
	 *      heading pitch bank
	 */
	sscanf(
	    arg,
"%i\
 %i %i %ld\
 %ld %ld %ld\
 %f %f %f\
 %f %f %f",
	    &obj_num,
	    &type, &imgref_num, &size,
	    &sect_x, &sect_y, &sect_z,
	    &x, &y, &z,
	    &heading, &pitch, &bank
	);

	ASSERT_NOT_NAN(x);
	ASSERT_NOT_NAN(y);
	ASSERT_NOT_NAN(z);
	ASSERT_NOT_NAN(heading);
	ASSERT_NOT_NAN(pitch);
	ASSERT_NOT_NAN(bank);

#ifdef STARCHARTWIN_H
	SChtAddObject(
	    xsw_scht, obj_num,
	    type, imgref_num, size,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    heading, pitch, bank
	);
#endif
}

/*
 *	Starchart set object name.
 */
static void NetHandleStarChartSetObjectName(const char *arg)
{
	const char *s;
	char name[SW_OBJ_NAME_MAX];

	/* Put the last object in the StarChart in context */

	s = strchr(arg, CS_STRING_DELIMINATOR_CHAR);
	if(s == NULL)
	    return;

	strncpy(name, s + 1, sizeof(name));
	name[sizeof(name) - 1] = '\0';

#ifdef STARCHARTWIN_H
	SChtSetObjectName(xsw_scht, name);
#endif
}

/*
 *	Starchart set object empire.
 */
static void NetHandleStarChartSetObjectEmpire(const char *arg)
{
	const char *s;
	char empire[SW_OBJ_EMPIRE_MAX];

	/* Put the last object in the StarChart in context */

	s = strchr(arg, CS_STRING_DELIMINATOR_CHAR);
	if(s == NULL)
	    return;

	strncpy(empire, s + 1, sizeof(empire));
	empire[sizeof(empire) - 1] = '\0';

#ifdef STARCHARTWIN_H
	SChtSetObjectEmpire(xsw_scht, empire);
#endif
}

/*
 *	Starchart recycle object.
 */
static void NetHandleStarChartRecycleObject(const char *arg)
{
	int obj_num = -1;

	/* SWEXTCMD_STARCHART_RECYCLE_OBJECT format:
	 *
	 *      obj_num
	 */
	sscanf(
	    arg,
	    "%i",
	    &obj_num
	);

#ifdef STARCHARTWIN_H
	SChtRecycleObject(xsw_scht, obj_num);
#endif
}


/*
 *	Play sound.
 */
static void NetHandlePlaySound(const char *arg)
{
	int sound_code = 0;
	float left_volume = 0.0f, right_volume = 0.0f;

	/* Check sound play level option */
	if(option.sounds <= XSW_SOUNDS_NONE)
	    return;

	/* CS_CODE_PLAYSOUND format:
	 *
	 *	sound_code left_volume right_volume
	 */
	sscanf(
	    arg,
	    "%i %f %f",
	    &sound_code,
	    &left_volume, &right_volume
	);

	ASSERT_NOT_NAN(left_volume);
	ASSERT_NOT_NAN(right_volume);

	if(option.sounds > XSW_SOUNDS_NONE)
	    SoundPlay(
		&sound, sound_code,
		left_volume,
		right_volume,
		0,		/* Effects */
		0		/* Priority */
	    );
}

/*
 *	Set weapon values.
 */
static void NetHandleSetWeaponValues(const char *arg)
{
	char changed_name = 0;
	char changed_amount = 0;

	int obj_num = -1;
	int wep_num = -1;
	int ocsn_num;

	int ocs_code = -1;
	int emission_type = SW_WEP_EMISSION_STREAM;
	int amount = 0;
	int max = 0;

	float damage_hull = 0.0f, damage_shields = 0.0f;
	long range = 0l;
	float create_power = 0.0f;

	long delay = 0l;
	long last_used = 0l;
	int fire_sndref = 0;
	unsigned long flags = 0;		/* sw_flags_t */

	xsw_object_struct *obj;
	xsw_weapon_struct *wep;

	xsw_ocsn_struct *ocsn;
	xsw_net_parms_struct *net = &net_parms;


	/* NET_CMD_SETWEAPONVAL format:
	 *
	 *	obj_num wep_num
	 *	ocs_code emission_type amount max
	 *	damage_hull damage_shields range create_power
	 *	delay last_used fire_sndref flags
	 */
	sscanf(
	    arg,
"%i %i\
 %i %i %i %i\
 %f %f %ld %f\
 %ld %ld %i %lu",
	    &obj_num, &wep_num,
	    &ocs_code, &emission_type, &amount, &max,
	    &damage_hull, &damage_shields, &range, &create_power,
	    &delay, &last_used, &fire_sndref, &flags
	);

	ASSERT_NOT_NAN(damage_hull);
	ASSERT_NOT_NAN(damage_shields);
	ASSERT_NOT_NAN(create_power);

	/* Object does not exist? */
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Make sure that the weapon number is valid */
	if((wep_num < 0) || (wep_num >= SW_OBJ_WEAPONS_MAX))
	    return;

	/* Allocate more weapons as needed */
	if(wep_num >= obj->total_weapons)
	{
	    if(UNVAllocObjectWeapons(obj, wep_num + 1))
		return;

	    changed_name = 1;
	    changed_amount = 1;
	}

	/* Set weapon values */
	wep = obj->weapons[wep_num];

	wep->flags = (sw_flags_t)flags;
	wep->ocs_code = ocs_code;
	wep->emission_type = emission_type;

	if(amount != wep->amount)
	{
	    changed_amount = 1;
	    wep->amount = amount;
	}
	wep->max = max;

	wep->damage_hull = damage_hull;
	wep->damage_shields = damage_shields;
	wep->range = range;
	wep->create_power = create_power;

	wep->delay = delay;

	wep->fire_sndref = fire_sndref;

	/* Get name for weapon type by matching type with OCS */
	ocsn_num = OCSNGetByCode(xsw_ocsn, total_ocsns, ocs_code);
	ocsn = OCSNGetPtr(xsw_ocsn, total_ocsns, ocsn_num);
	if(ocsn != NULL)
	{
	    const char *name = ocsn->name;
	    if((name != NULL) ? strcmp(wep->name, name) : 0)
	    {
	        strncpy(wep->name, name, sizeof(wep->name));
	        wep->name[sizeof(wep->name) - 1] = '\0';
		changed_name = 1;
	    }
	}
	else
	{
	    const char *name = "Unknown";
	    if(strcmp(wep->name, name))
	    {
		strncpy(wep->name, name, sizeof(wep->name));
		wep->name[sizeof(wep->name) - 1] = '\0';
		changed_name = 1;
	    }
	}

#if 0
	/* The wep->last_used is handled locally since cur_millitime
	 * on the client and server may differ, last_used will be set
	 * by the client's EventHandleWeaponFire() function
	 */
	wep->last_used = last_used;
#endif

	/* Mark the last updated time */
	obj->last_updated = cur_millitime;

	/* Redraw displays */
	if(obj_num == net->player_obj_num)
	{
	    xsw_bridge_win_struct *bw = xsw_bridge_win;

	    /* Draw the bridge panel to reflect the player object's
	     * weapon values
	     */
	    BridgeWinDrawPanel(obj_num, BW_PANEL_PWEAPONS);

	    if(changed_name || changed_amount)
	    {
		if(wep_num == obj->selected_weapon)
		{
		    if(bw != NULL)
		    {
			/* Recreate selected weapon stats label on viewscreen */
			VSDrawUpdateWeaponLabel(
			    &bw->vs_weapon_image,
			    bw->vs_weapon_buf
			);
		    }
		}
	    } 
	}
}

/*
 *	Handles extended net commands.
 *
 *	Called by NetHandleRecv().
 */
static void NetHandleExtCmd(const char *arg)
{
	int ext_cmd;
	const char *s;
	char ext_arg[CS_DATA_MAX_LEN];

	/* Get command */
	ext_cmd = StringGetNetCommand(arg);
	if(ext_cmd < 0)
	    return;

	/* Get argument */
	s = StringGetNetArgument(arg);
	if(s == NULL)
	    return;

	strncpy(ext_arg, s, sizeof(ext_arg));
	ext_arg[sizeof(ext_arg) - 1] = '\0';


	/* Handle external command */
	switch(ext_cmd)
	{
	  /* Systems and enviroment */
	  case SWEXTCMD_SETOCSN:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETOCSN: %s\n", ext_arg);
	    NetHandleSetOCSN(ext_arg);
	    break;

	  case SWEXTCMD_SET_LEGEND:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SET_LEGEND: %s\n", ext_arg);
	    NetHandleSetLegend(ext_arg);
	    break;

	  case SWEXTCMD_SETTITLE:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETTITLE: %s\n", ext_arg);
	    NetHandleSetTitle(ext_arg);
	    break;


	  /* Standards */
	  case SWEXTCMD_STDOBJVALS:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_STDOBJVALS: %s\n", ext_arg);
	    NetHandleSetObjectValues(ext_arg);
	    break;

	  case SWEXTCMD_STDOBJMAXS:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_STDOBJMAXS: %s\n", ext_arg);
	    NetHandleSetObjectMaximums(ext_arg);
	    break;

	  case SWEXTCMD_STDWEPVALS:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_STDWEPVALS: %s\n", ext_arg);
	    NetHandleSetWeaponValues(ext_arg);
	    break;


	  /* Setting */
	  case SWEXTCMD_SETOBJNAME:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETOBJNAME: %s\n", ext_arg);
	    NetHandleSetName(ext_arg);
	    break;

	  case SWEXTCMD_SETOBJSECT:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETOBJSECT: %s\n", ext_arg);
	    NetHandleSetObjSect(ext_arg);
	    break;

	  case SWEXTCMD_SETFOBJSECT:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETFOBJSECT: %s\n", ext_arg);
	    NetHandleSetFObjSect(ext_arg);
	    break;

	  case SWEXTCMD_SETWEAPON:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETWEAPON: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_SETINTERCEPT:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETINTERCEPT: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_SETWEPLOCK:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETWEPLOCK: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_SETSHIELDS:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETSHIELDS: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_SETDMGCTL:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETDMGCTL: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_SETCLOAK:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETCLOAK: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_SETSHIELDVIS:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETSHIELDVIS: %s\n", ext_arg);
	    NetHandleSetShieldVis(ext_arg);
	    break;

	  case SWEXTCMD_SETLIGHTING:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETLIGHTING: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_SETCHANNEL:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETCHANNEL: %s\n", ext_arg);
	    NetHandleSetChannel(ext_arg);
	    break;

	  case SWEXTCMD_SETSCORE:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETSCORE: %s\n", ext_arg);
	    NetHandleSetScore(ext_arg);
	    break;

	  case SWEXTCMD_SETENGINE:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETENGINE: %s\n", ext_arg);
	    NetHandleSetEngine(ext_arg);
	    break;

	  case SWEXTCMD_SET_OBJ_PARTICLE_TRAIL:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SET_OBJ_PARTICLE_TRAIL: %s\n", ext_arg);
	    NetHandleSetObjParticleTrail(ext_arg);
	    break;

	  case SWEXTCMD_CREATE_EFFECT:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_CREATE_EFFECT: %s\n", ext_arg);
	    NetHandleCreateEffect(ext_arg);
	    break;

	  case SWEXTCMD_CREATE_SHOCKWAVE:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_CREATE_SHOCKWAVE: %s\n", ext_arg);
	    NetHandleCreateShockwave(ext_arg);
	    break;


	  /* Request */
	  case SWEXTCMD_REQNAME:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_REQNAME: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_REQSECT:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_REQSECT: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;


	  /* Hit and destroy notifies */
	  case SWEXTCMD_NOTIFYHIT:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_NOTIFYHIT: %s\n", ext_arg);
	    NetHandleNotifyHit(ext_arg);
	    break;

	  case SWEXTCMD_NOTIFYDESTROY:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_NOTIFYDESTROY: %s\n", ext_arg);
	    NetHandleNotifyDestroy(ext_arg);
	    break;


	  /* Actions */
	  case SWEXTCMD_FIREWEAPON:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_FIREWEAPON: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_TRACTORBEAMLOCK:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_TRACTORBEAMLOCK: %s\n", ext_arg);
	    NetHandleTractorBeamLock(ext_arg);
	    break;

	  case SWEXTCMD_HAIL:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_HAIL: %s\n", ext_arg);
	    NetHandleHail(ext_arg);
	    break;

	  case SWEXTCMD_COMMESSAGE:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_COMMESSAGE: %s\n", ext_arg);
	    NetHandleComMessage(ext_arg);
	    break;

	  case SWEXTCMD_WORMHOLEENTER:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_WORMHOLEENTER: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_ELINKENTER:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_ELINKENTER: %s\n", ext_arg);
	    /* Client does not handle this command */
	    break;

	  case SWEXTCMD_WEPDISARM:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_WEPDISARM: %s\n", ext_arg);
	    NetHandleWeaponDisarm(ext_arg);
	    break;

	  case SWEXTCMD_AUTODESTRUCT:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_AUTODESTRUCT: %s\n", ext_arg);
	    NetHandleSetAutoDestruct(ext_arg);
	    break;

	  case SWEXTCMD_SETDESCRIPTION:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_SETDESCRIPTION: %s\n", ext_arg);
	    NetHandleSetDescription(ext_arg);
	    break;

	  case SWEXTCMD_REQDESCRIPTION:
	    /* Client does not handle this command */
	    break;



	  /* Economy */
	  case SWEXTCMD_ECO_REQVALUES:
	    /* Client does not handle this */
	    break;

	  case SWEXTCMD_ECO_SETVALUES:
	    NetHandleEcoSetValues(ext_arg);
	    break;

	  case SWEXTCMD_ECO_SETPRODUCTVALUES:
	    NetHandleEcoSetProductValues(ext_arg);
	    break;

	  case SWEXTCMD_ECO_BUY:
	    /* Client does not handle this */
	    break;

	  case SWEXTCMD_ECO_SELL:
	    /* Client does not handle this */
	    break;


	  /* StarChart */
	  case SWEXTCMD_STARCHART_ADD_OBJECT:
	    NetHandleStarChartAddObject(ext_arg);
	    break;

	  case SWEXTCMD_STARCHART_SET_OBJECT_NAME:
	    NetHandleStarChartSetObjectName(ext_arg);
	    break;

	  case SWEXTCMD_STARCHART_SET_OBJECT_EMPIRE:
	    NetHandleStarChartSetObjectEmpire(ext_arg);
	    break;

	  case SWEXTCMD_STARCHART_SET_OBJECT_RECYCLE:
	    NetHandleStarChartRecycleObject(ext_arg);
	    break;


	  default:
	    if(debug.level == DEBUG_LEVEL_NETWORK)
		fprintf(stderr, "Client: SWEXTCMD_???: %s\n", ext_arg);
	    break;
	}
}


/*
 *	Receives incoming data from the server and handles it.
 *
 *      Any segments that are received incomplete will be stored in
 *      the carryover buffer for use in the next call to this function.
 */
int NetHandleRecv(void)
{
	int bytes_read;

	char recvbuf[CS_DATA_MAX_BACKLOG];
	int recvbuf_cnt;
	char *recvbuf_ptr;

	char segbuf[CS_DATA_MAX_LEN];
	int segbuf_cnt;
	char *segbuf_ptr;

	char arg[CS_DATA_MAX_LEN];
	int command;

	xsw_net_parms_struct *net = &net_parms;
	xsw_net_stat_struct *nstat = &net_stat;
	xsw_option_struct *opt = &option;

	/* Check for collected data, this is in case a network
	 * statement was not received in whole and the collected
	 * data will be put into this receive buffer
	 */
	if(net->recvbuf_len < (int)sizeof(recvbuf))
	{
	    /* Sanitize the collected data length */
	    if(net->recvbuf_len > (int)sizeof(net->recvbuf))
		net->recvbuf_len = (int)sizeof(net->recvbuf);
	    else if(net->recvbuf_len < 0)
		net->recvbuf_len = 0;

	    /* Copy the collected data to the current recv buffer */
	    if(net->recvbuf_len > 0)
	        memcpy(
		    recvbuf,
		    net->recvbuf,
		    net->recvbuf_len
	        );
	}
	else
	{
	    /* Collected data has overflowed */
	    net->recvbuf_len = 0;
	}

	/* Get data from server */
	bytes_read = NetRecv(
	    net->socket,
	    recvbuf + net->recvbuf_len,
	    sizeof(recvbuf) - net->recvbuf_len
	);
	if(bytes_read == 0)
	{
	    /* Socket was polled and there were bytes to be read but
	     * recv() returns 0, this implies that the socket has been
	     * closed at the other end
	     */
	    XSWDisconnect();
	    return(0);
	}

	/* Receive error? */
	if(bytes_read < 0)
	{
	    const int error_code = errno;
	    switch(error_code)
	    {
	      case EAGAIN:	/* No data available */
#if (EWOULDBLOCK != EAGAIN)
	      case EWOULDBLOCK:
#endif
	      case EINTR:	/* Interrupted by signal */
		return(0);
		break;

	      case EPIPE:	/* Connection has been shut down */
		net->connection_state = XSW_CON_STATE_CLOSE_WAIT;
		XSWDisconnect();
		return(0);
		break;

	      case EBADF:	/* Invalid descriptor */
	      case ECONNRESET:	/* Connection terminated by peer */
	      case EFAULT:	/* Receive buffer out of address space */
	      case ENOTCONN:    /* Descriptor is not connected */
	      case ENOTSOCK:	/* Descriptor is not a socket */
		if(True)
		{
		    char *s, *error_msg = STRDUP(strerror(error_code));
		    if(error_msg == NULL)
			error_msg = strdup("unknown error");
		    *error_msg = toupper(*error_msg);
		    s = (char *)malloc(
			40 + STRLEN(error_msg) + 80
		    );
		    sprintf(
			s,
"Receive data fatal error: Socket %i: %s.",
			net->socket, error_msg
		    );
		    XSWPrintError(s);
		    free(s);
		    free(error_msg);
		}
		net->connection_state = XSW_CON_STATE_CLOSE_WAIT;
		XSWDisconnect();
		return(-1);
		break;

	      default:		/* Other error */
		if(opt->show_net_errors)
		{
		    char *s, *error_msg = STRDUP(strerror(error_code));
		    if(error_msg == NULL)
			error_msg = strdup("unknown error");
		    *error_msg = toupper(*error_msg);
		    s = (char *)malloc(
			40 + STRLEN(error_msg) + 80
		    );
		    sprintf(
			s,
"Receive data error: Socket %i: %s.",
			net->socket, error_msg
		    );
		    XSWPrintWarning(s);
		    free(s);
		    free(error_msg);
		}
		return(-1);
		break;
	    }
	}

	/* Increment recieved bytes per interval */
	nstat->rx_interval += (unsigned long)bytes_read;

	/* Add the collected data length to the bytesread to get
	 * correct length of the receive buffer
	 */
	bytes_read += net->recvbuf_len;

	/* Make sure bytes read does not overflow */
	if(bytes_read > (int)sizeof(recvbuf))
	    bytes_read = (int)sizeof(recvbuf);

	/* Set null character at the end of recvbuf just to be safe */
	recvbuf[sizeof(recvbuf) - 1] = '\0';

if(debug.level == DEBUG_LEVEL_NETWORK)
    fprintf(stderr, "Client recieved: %i bytes.\n", bytes_read);


	/* Begin parsing the recieved buffer */
	recvbuf_cnt = 0;
	recvbuf_ptr = recvbuf;
	while(recvbuf_cnt < bytes_read)
	{
	    /* Do not continue to process the received data if the
	     * connection is set to CLOSE_WAIT or not connected
	     */
	    if(XSW_CON_STATE_IS_CLOSE_WAIT(net) ||
	       !XSW_CON_IS_CONNECTED(net)
	    )
		break;

	    /* Get start of segment buffer */
	    segbuf_cnt = 0;
	    segbuf_ptr = segbuf;

	    /* Begin parsing data from recvbuf to segbuf */
	    while(1)
	    {
		/* Segment buffer overflowed? */
		if(segbuf_cnt >= (int)sizeof(segbuf))
		{
		    /* Increment recvbuf_cnt to the next deliminator
		     * (if any)
		     */
		    while(recvbuf_cnt < bytes_read)
		    {
			if((*recvbuf_ptr == '\n') ||
			   (*recvbuf_ptr == '\r') ||
			   (*recvbuf_ptr == '\0')
			)
			{
			    recvbuf_cnt++;
			    recvbuf_ptr++;
			    break;
			}
			recvbuf_cnt++;
			recvbuf_ptr++;
		    }
		    /* Null terminating character for segbuf will
		     * be added farther below
		     */
		    break;
		}
		/* End of recieve buffer reached without deliminator? */
		else if(recvbuf_cnt >= bytes_read)
		{
		    *segbuf_ptr = '\0';

		    /* Copy the contents of the segment buffer to
		     * net->recvbuf for processing in the next
		     * call to this function
		     */
		    net->recvbuf_len = MIN(
			segbuf_cnt, (int)sizeof(net->recvbuf)
		    );

		    if(net->recvbuf_len > 0)
		        memcpy(
			    net->recvbuf,
			    segbuf,
			    net->recvbuf_len
		        );

		    /* Return, do not continue, this is so that the
		     * collected data buffer is not reset and its
		     * contents will get carryed over to the next call
		     * to this function
		     */
		    return(0);

		    break;
		}
		/* Deliminator in recieve buffer encountered? */
		else if((*recvbuf_ptr == '\n') ||
		        (*recvbuf_ptr == '\r') ||
			(*recvbuf_ptr == '\0')
		)
		{
		    *segbuf_ptr = '\0';

		    recvbuf_cnt++;
		    recvbuf_ptr++;

		    break;
		}
		/* All else copy the data from the recieve buffer to
		 * the segment buffer
		 */
		else
		{
		    *segbuf_ptr = *recvbuf_ptr;

		    recvbuf_cnt++;
		    recvbuf_ptr++;

		    segbuf_cnt++;
		    segbuf_ptr++;
		}
	    }

	    /* Skip this segment if it is empty */
	    if(*segbuf == '\0')
		continue;

	    /* Null terminate the segment buffer just in case */
	    segbuf[sizeof(segbuf) - 1] = '\0';


	    /* Get the command */
	    command = StringGetNetCommand(segbuf);
	    if(command < 0)
		continue;

	    /* Get the argument */
	    strncpy(arg, StringGetNetArgument(segbuf), sizeof(arg));
	    arg[sizeof(arg) - 1] = '\0';


	    /* See which command to perform */
	    switch(command)
	    {
	      case CS_CODE_VERSION:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_VERSION: %s\n", arg);
		NetHandleVersion(arg);
		break;

	      case CS_CODE_LOGIN:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_LOGIN: %s\n", arg);
		NetHandleLogin(arg);
		break;

	      case CS_CODE_LOGOUT:
		if(debug.level == DEBUG_LEVEL_NETWORK)  
		    fprintf(stderr, "Client: CS_CODE_LOGOUT: %s\n", arg);
		NetHandleLogout(arg);
		break;

	      case CS_CODE_WHOAMI:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_WHOAMI: %s\n", arg);
		NetHandleWhoAmI(arg);
		    break;

	      case CS_CODE_REFRESH:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_REFRESH: %s\n", arg);
		/* Client dosen't handle this */
		break;

	      case CS_CODE_INTERVAL:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_INTERVAL: %s\n", arg);
		/* Client dosen't handle this */
		break;

	      case CS_CODE_IMAGESET:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_IMAGESET: %s\n", arg);
		NetHandleSetImageSet(arg);
		break;

	      case CS_CODE_SOUNDSET:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_SOUNDSET: %s\n", arg);
		NetHandleSetSoundSet(arg);
		break;

	      case CS_CODE_FORWARD:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_FORWARD: %s\n", arg);
		NetHandleSetForward(arg);
		break;

	      case CS_CODE_LIVEMESSAGE:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_LIVEMESSAGE: %s\n", arg);
		NetHandleLiveMesg(arg);
		break;

	      case CS_CODE_SYSMESSAGE:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_SYSMESSAGE: %s\n", arg);
		NetHandleSysMesg(arg);
		break;

	      case CS_CODE_LITERALCMD:
		/* Client dosen't handle this */
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_LITERALCMD: %s\n", arg);
		break;


	      case CS_CODE_PLAYSOUND:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_PLAYSOUND: %s\n", arg);
		NetHandlePlaySound(arg);
		break;


	      case CS_CODE_CREATEOBJ:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_CREATEOBJ: %s\n", arg);
		NetHandleCreateObject(arg);
		break;

	      case CS_CODE_RECYCLEOBJ:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_RECYCLEOBJ: %s\n", arg);
		NetHandleRecycleObject(arg);
		break;

	      case CS_CODE_POSEOBJ:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_POSEOBJ: %s\n", arg);
		NetHandleSetPoseObj(arg);
		break;

	      case CS_CODE_FORCEPOSEOBJ:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: CS_CODE_FORCEPOSEOBJ: %s\n", arg);
		NetHandleSetFPoseObj(arg);
		break;


	      case CS_CODE_EXT:
		NetHandleExtCmd(arg);
		break;

	      default:
		if(debug.level == DEBUG_LEVEL_NETWORK)
		    fprintf(stderr, "Client: ???: %s\n", arg);

		/* Print unsupported? */
		if(opt->show_net_errors)
		    XSWPrintWarning(arg);
		break;
	    }
	}

	/* Reset the collected data buffer since parsing was complete
	 * and successful
	 */
	net->recvbuf_len = 0;

	return(0);
}
