#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>

#include "../include/string.h"

#include "../include/netio.h"
#include "../include/swnetcodes.h"

#include "../include/reality.h"
#include "../include/unvmath.h"

#include "net.h"
#include "xsw.h"


int NetSendData(const char *data);
void NetSendGlobalMessage(const char *msg);
void NetSendVersion(void);
void NetSendLogout(void);
void NetSendRefresh(void);
void NetSendSetInterval(void);
void NetSendExec(const char *arg);
void NetSendSetImageSet(const char *arg);
void NetSendSetSoundSet(const char *arg);
void NetSendSetOCSN(const char *arg);
void NetSendObjSect(int obj_num);
void NetSendObjPose(int obj_num);
void NetSendObjThrottle(int obj_num);
void NetSendSetLighting(int obj_num, sw_flags_t lighting);
void NetSendSetChannel(int obj_num, int channel);
void NetSendWeaponsState(int obj_num, int state);
void NetSendSelectWeapon(int obj_num, int sel_wep_num);
void NetSendSetShields(
	int obj_num, int shield_state,
	float shield_frequency
);
void NetSendSetDmgCtl(int obj_num, char damage_control);
void NetSendSetCloak(int obj_num, char cloak_state);
void NetSendWeaponsLock(int obj_num, int tar_obj_num);
void NetSendWeaponsUnlock(int obj_num);
void NetSendReqName(int obj_num);
void NetSendIntercept(int obj_num, const char *arg);
void NetSendSetEngine(const int obj_num, const int engine_state);
void NetSendFireWeapon(int obj_num, float freq);
void NetSendTractorBeamLock(int obj_num, int tar_obj_num);
void NetSendHail(int obj_num, int tar_obj_num, int channel);
void NetSendComMessage(
	int obj_num, int tar_obj_num,
	int channel, const char *msg
);
void NetSendWormHoleEnter(int obj_num, int tar_obj_num);
void NetSendELinkEnter(int obj_num, int tar_obj_num);
void NetSendWeaponDisarm(int obj_num, int tar_obj_num);
void NetSendAutoDestruct(int obj_num, time_t sec);
void NetSendReqDescription(int obj_num);

void NetSendEcoReqValues(int obj_num);
void NetSendEcoBuy(
        const int customer_obj_num, const int proprietor_obj_num,
        xsw_ecoproduct_struct *product
);
void NetSendEcoSell(
        const int customer_obj_num, const int proprietor_obj_num,
        xsw_ecoproduct_struct *product
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sends the data and updates the net stats.
 *
 *	Returns:
 *
 *	0	Data sent successfully.
 *	-1	General error, not connected, or unable to send.
 *	-2	Bad value.
 *	-3	Systems error.
 *	-4	Interrupted by signal.
 *	-5	Send failed and the connection has been closed by this
 *		function.
 */
int NetSendData(const char *data)
{
	int bytes_sent, len;
	xsw_net_parms_struct *net = &net_parms;
	xsw_net_stat_struct *nstat = &net_stat;
	xsw_option_struct *opt = &option;

	if(data == NULL)
	    return(-2);

	/* Not connected or the connection is on CLOSE_WAIT? */
	if(!XSW_CON_IS_CONNECTED(net) ||
           XSW_CON_STATE_IS_CLOSE_WAIT(net)
	)
	    return(-1);

	/* Get minimized length of the data to send, not to exceed the
	 * maximum statement size
	 */
	len = MIN(strlen(data), CS_DATA_MAX_LEN);

	/* Send the data */
	bytes_sent = NetSend(net->socket, data, len);
	if(bytes_sent > 0)
	{
	    /* Send successful, record number of bytes sent */
	    nstat->sx_interval += (unsigned long)bytes_sent;
	}
	else
	{
	    const int error_code = errno;
	    switch(error_code)
	    {
	      case EAGAIN:
#if (EWOULDBLOCK != EAGAIN)
	      case EWOULDBLOCK:
#endif
	      case ENOBUFS:	/* Out of memory, try again */
		return(-1);
		break;

	      case EPIPE:	/* Connection has been shut down */
		net->connection_state = XSW_CON_STATE_CLOSE_WAIT;
		XSWDisconnect();
		return(-5);
		break;    

	      case EBADF:	/* Invalid descriptor */
	      case ENOTSOCK:	/* Descriptor is not a socket */
		/* Fatal error */
		if(True)
		{
		    char *s, *error_msg = STRDUP(strerror(error_code));
		    if(error_msg == NULL)
			error_msg = strdup("unknown error");
		    *error_msg = toupper(*error_msg);
		    s = (char *)malloc(
			40 + STRLEN(error_msg) + 80
		    );
		    sprintf(
			s,
"Send data fatal error: Socket %i: %s.",
			net->socket, error_msg
		    );
		    XSWPrintError(s);
		    free(s);
		    free(error_msg);
		}
		net->connection_state = XSW_CON_STATE_CLOSE_WAIT;
		XSWDisconnect();
		return(-5);
		break;

	      case EINTR:	/* Signal occured */
		return(-4);
		break;

	      case ENOMEM:	/* Out of memory */
		return(-3);
		break;

	      default:
		/* Print this non-fatal error? */
		if(opt->show_net_errors)
		{
		    char *s, *error_msg = STRDUP(strerror(error_code));
		    if(error_msg == NULL)
			error_msg = strdup("unknown error");
		    *error_msg = toupper(*error_msg);
		    s = (char *)malloc(
			40 + STRLEN(error_msg) + 80
		    );
		    sprintf(
			s,
"Send data error: Socket %i: %s.",
			net->socket, error_msg
		    );
		    XSWPrintWarning(s);
		    free(s);
		    free(error_msg);
		}
		return(-1);
		break;
	    }
	}

	return(0);
}

/*
 *	Sends a global message.
 */
void NetSendGlobalMessage(const char *msg)
{
	int len;
	char *sndbuf, lmesg[CS_MESG_MAX];

	if(STRISEMPTY(msg))
	    return;

	len = strlen(msg);

	/* Copy message */
	strncpy(lmesg, msg, sizeof(lmesg));
	lmesg[sizeof(lmesg) - 1] = '\0';
	strstrip(lmesg);

	sndbuf = (char *)malloc(
	    40 + STRLEN(lmesg) + 80
	);

	/* CS_CODE_LIVEMESSAGE format:
	 *
	 *	message
	 */
	sprintf(
	    sndbuf,
	    "%i %s\n",
	    CS_CODE_LIVEMESSAGE,
	    lmesg
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Sends the version of our client.
 */
void NetSendVersion(void)
{
	char *sndbuf = (char *)malloc(40 + 40 + 40 + 40 + 80);
	sprintf(
	    sndbuf,
"%i %i %i %i\n",
	    CS_CODE_VERSION,
	    PROG_VERSION_MAJOR,
	    PROG_VERSION_MINOR,
	    PROG_VERSION_RELEASE
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Sends a request to logout.
 */
void NetSendLogout(void)
{
	char *sndbuf = (char *)malloc(40 + 80);
	sprintf(
	    sndbuf,
	    "%i\n",
	    CS_CODE_LOGOUT
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Sends a request for refresh.
 */
void NetSendRefresh(void)
{
	char *sndbuf = (char *)malloc(40 + 80);
	sprintf(
	    sndbuf,
	    "%i\n",
	    CS_CODE_REFRESH
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Sends a set network interval.
 */
void NetSendSetInterval(void)
{
	char *sndbuf;
	xsw_net_parms_struct *net = &net_parms;

	sndbuf = (char *)malloc(40 + 40 + 80);

	/* Format:
	 *
	 *	interval
	 */
	sprintf(
	    sndbuf,
	    "%i %ld\n",
	    CS_CODE_INTERVAL,
	    net->net_int
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Sends a command.
 */
void NetSendExec(const char *arg)
{
	char *msg, *sndbuf;

	if(STRISEMPTY(arg))
	    return;

	msg = (char *)malloc(CS_MESG_MAX);
	strncpy(msg, arg, CS_MESG_MAX);
	msg[CS_MESG_MAX - 1] = '\0';

	sndbuf = (char *)malloc(40 + STRLEN(msg) + 80);

	/* Format:
	 *
	 *	arg
	 */
	sprintf(
	    sndbuf,
	    "%i %s\n",
	    CS_CODE_LITERALCMD,
	    msg
	);
	NetSendData(sndbuf);
	free(sndbuf);
	free(msg);
}

/*
 *	Sends a request for the current image set file name.
 */
void NetSendSetImageSet(const char *arg)
{
	char *sndbuf = (char *)malloc(
	    40 + 80
	);

	/* Format:
	 *
	 *
	 */
	sprintf(
	    sndbuf,
	    "%i\n",
	    CS_CODE_IMAGESET
	);
	NetSendData(sndbuf);
	free(sndbuf);
} 

/*
 *      Sends a request for the current sound set file name.
 */
void NetSendSetSoundSet(const char *arg)
{
	char *sndbuf = (char *)malloc(
	    40 + 80
	);

	/* Format:
	 *
	 *
	 */
	sprintf(
	    sndbuf,
	    "%i\n", 
	    CS_CODE_SOUNDSET
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *      Sends a request for the current object create script
 *	names file name.
 */
void NetSendSetOCSN(const char *arg)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 80
	);

	/* Format:
	 *
	 *
	 */
	sprintf(
	    sndbuf,
	    "%i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETOCSN
	);
	NetSendData(sndbuf);
	free(sndbuf); 
}

/*
 *	Object sector.
 */
void NetSendObjSect(int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Format:
	 *
	 *	obj_num
	 *	sect_x sect_y sect_z
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 40 + 40 + 80
	);
	sprintf(
	    sndbuf,
"%i %i\
 %i %ld %ld %ld\n",
	    CS_CODE_EXT, SWEXTCMD_SETOBJSECT,
	    obj_num,
	    obj->sect_x, obj->sect_y, obj->sect_z
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Object pose.
 */
void NetSendObjPose(int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Format:
	 *
	 *	obj_num
	 *	type imgref_num size
	 *	x y z
	 *	heading pitch bank
	 *	velocity
	 *	velocity_heading velocity_pitch velocity_bank
	 *	current_frame
	 */
	sndbuf = (char *)malloc(   
	    40 + 40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 +
	    40 + 40 + 40 +
	    40 + 80
	);
	sprintf(
	    sndbuf,
"%i %i\
 %i %i %ld\
 %.4f %.4f %.4f\
 %.4f 0 0\
 %.5f\
 %.4f 0 0\
 %i\n",
	    CS_CODE_POSEOBJ,
	    obj_num,
	    obj->type, obj->imgref, obj->size,
	    obj->x, obj->y, obj->z,
	    obj->heading,
	    obj->velocity,
	    obj->velocity_heading,
	    obj->animation.current_frame
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Throttle.
 */
void NetSendObjThrottle(int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Format:
	 *
	 *	obj_num throttle
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);
	sprintf(
	    sndbuf,
	    "%i %i %i %.4f\n",
	    CS_CODE_EXT, SWEXTCMD_SETTHROTTLE,
	    obj_num,
	    (float)((obj->engine_state == SW_ENGINE_STATE_ON) ?
		obj->throttle : 0.0f
	    )
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Set lighting.
 */
void NetSendSetLighting(int obj_num, sw_flags_t lighting)
{
	int status;
	char *sndbuf;
 
	/* SWEXTCMD_SETLIGHTING format:
	 *
	 *      obj_num lighting
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);
	sprintf(
	    sndbuf,
	    "%i %i %i %lu\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETLIGHTING,
	    obj_num, (unsigned long)lighting
	);
	status = NetSendData(sndbuf);
	free(sndbuf);
	if(status)
	    return;

	/* Update locally */
	if(!DBIsObjectGarbage(obj_num) && option.local_updates)
	{
	    xsw_object_struct *obj = xsw_object[obj_num];
	    obj->lighting = lighting;
	    if(obj_num == net_parms.player_obj_num)
                BridgeWinDrawPanel(obj_num, BW_PANEL_P2);
	}
}

/*
 *	Sends set channel.
 */
void NetSendSetChannel(int obj_num, int channel)
{
	int status;
	char *sndbuf;

	/* Sanitize channel */
	if(((float)channel / 100.0f) > SW_FREQ_MAX)
	    channel = (int)(SW_FREQ_MAX * 100.0f);
	if(((float)channel / 100.0f) < SW_FREQ_MIN)
	    channel = (int)(SW_FREQ_MIN * 100.0f);

	/* SWEXTCMD_SETCHANNEL format:
	 *
	 *	obj_num channel
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	); 
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETCHANNEL,
	    obj_num, channel
	);
	status = NetSendData(sndbuf);
	free(sndbuf);
	if(status)
	    return;

	/* Update locally as needed */
	if(!DBIsObjectGarbage(obj_num) && option.local_updates)
	{
	    xsw_object_struct *obj = xsw_object[obj_num];
	    obj->com_channel = channel;
	}
}

/*
 *	Weapon state (weapons online or offline).
 */
void NetSendWeaponsState(int obj_num, int state)
{
	/* No network protocol available for this yet */
	return;
}

/*
 *	Select weapon.
 */
void NetSendSelectWeapon(int obj_num, int sel_wep_num)
{
	int status;
	char *sndbuf;

	/* Selected weapon number valid? */
	if((sel_wep_num < 0) || (sel_wep_num >= SW_OBJ_WEAPONS_MAX))
	    return;

	/* SWEXTCMD_SETWEAPON format:
	 *
	 *	obj_num sel_wep_num
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	); 
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETWEAPON,
	    obj_num, sel_wep_num
	);
	status = NetSendData(sndbuf);
	free(sndbuf);
	if(status)
	    return;

#if 0
	/* Update locally as needed */
	if(!DBIsObjectGarbage(obj_num) && option.local_updates)
	{  
	    xsw_object_struct *obj = xsw_object[obj_num];
	    obj->selected_weapon = sel_wep_num;
	}
#endif
}

/*
 *	Shield state and frequency.
 */
void NetSendSetShields(
	int obj_num, int shield_state,
	float shield_frequency
)
{
	int status;
	char *sndbuf;

	if(shield_frequency > SW_FREQ_MAX)
	    shield_frequency = SW_FREQ_MAX;
	if(shield_frequency < SW_FREQ_MIN)
	    shield_frequency = SW_FREQ_MIN;

	/* Format:
	 *
	 *	obj_num shield_state shield_frequency
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 40 + 80
	); 
	sprintf(
	    sndbuf,
"%i %i\
 %i %i %.4f\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETSHIELDS,
	    obj_num, shield_state, shield_frequency
	);
	status = NetSendData(sndbuf);
	free(sndbuf);
	if(status)
	    return;

	/* Update locally */
	if(!DBIsObjectGarbage(obj_num) && option.local_updates)
	{
	    xsw_object_struct *obj = xsw_object[obj_num];
/*
	    obj->shield_state = shield_state;
 */
	    obj->shield_frequency = shield_frequency;
	}
}

/*
 *	Set damage control.
 */
void NetSendSetDmgCtl(int obj_num, char damage_control)
{
	int status;
	char *sndbuf;

	/* Format:
	 *
	 *	obj_num damage_control
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	); 
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETDMGCTL,
	    obj_num, damage_control
	);
	status = NetSendData(sndbuf);
	free(sndbuf);
	if(status)
	    return;

#if 0
	/* Update locally */
	if(!DBIsObjectGarbage(obj_num) && option.local_updates)
	{
	    xsw_object_struct *obj = xsw_object[obj_num];
	    obj->damage_control = damage_control;
	}
#endif
}


/*
 *	Sends set cloak.
 */
void NetSendSetCloak(int obj_num, char cloak_state)
{
	int status;
	char *sndbuf;

	/* Format:
	 *
	 *	obj_num cloak_state
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETCLOAK,
	    obj_num, cloak_state
	);
	status = NetSendData(sndbuf);
	free(sndbuf);
	if(status)
	    return;

#if 0
	/* Update locally */
	if(!DBIsObjectGarbage(obj_num) && option.local_updates)
	{
	    xsw_object_struct *obj = xsw_object[obj_num];
	    if(obj->cloak_state != CLOAK_STATE_NONE)
		obj->cloak_state = cloak_state;
	}
#endif
}


/*
 *	Send weapons lock for object_num.
 */
void NetSendWeaponsLock(int obj_num, int tar_obj_num)
{
	char *sndbuf;

	/* Format:
	 *
	 *	obj_num tar_obj_num
	 *
	 * tar_obj_num can be -1 for unlock or -2 for lock next
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	); 
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETWEPLOCK,
	    obj_num, tar_obj_num
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Send unlock weapons for obj_num.
 */
void NetSendWeaponsUnlock(int obj_num)
{
	NetSendWeaponsLock(obj_num, -1);
}


/*
 *	Sends a request name.
 */
void NetSendReqName(int obj_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 80
	); 

	/* Format:
	 *
	 *	obj_num
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i\n",
	    CS_CODE_EXT, SWEXTCMD_REQNAME,
	    obj_num
	);
	NetSendData(sndbuf);
	free(sndbuf);
}


/*
 *	Sents a intercept object or unintercept object.
 */
void NetSendIntercept(int obj_num, const char *arg)
{
	int status;
	char *sndbuf;
	xsw_net_parms_struct *net = &net_parms;

	/* If arg is NULL, set it to "#off" */
	if(arg == NULL)
	    arg = "#off";

	/* Format:
	 *
	 *	obj_num arg
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + STRLEN(arg) + 80
	);
	sprintf(
	    sndbuf,
"%i %i\
 %i %s\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETINTERCEPT,
	    obj_num, arg
	);
	status = NetSendData(sndbuf);
	free(sndbuf);
	if(status)
	    return;

	/* Update locally as needed */
	if(option.local_updates)
	{
	    if(obj_num == net->player_obj_num)
	    {
		if(!strcasecmp(arg, "#off"))
		{
		    if(!DBIsObjectGarbage(obj_num))
			xsw_object[obj_num]->intercepting_object = -1;
		}

		BridgeWinDrawPanel(
		    obj_num,
		    BW_PANEL_PINAME
		);
	    }
	}
}

/*
 *	Set engine state.
 */
void NetSendSetEngine(const int obj_num, const int engine_state)
{
	char *sndbuf;

	/* Format:
	 *
	 *	obj_num engine_state
	 */
	sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_SETENGINE,
	    obj_num, engine_state
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Fire weapon.
 */
void NetSendFireWeapon(int obj_num, float freq)
{
	float heading, pitch, bank;
	char *sndbuf;
	xsw_object_struct *obj;
	xsw_player_values_struct *pvals = &player_values;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Calculate firing direction */
	heading = SANITIZERADIANS(
	    obj->heading +
	    pvals->weapon_fire_heading
	);
	pitch = 0.0f * PI;
	bank = 0.0f * PI;

	/* Sanitize weapons frequency */
	if(freq > SW_FREQ_MAX)
	    freq = SW_FREQ_MAX;
	else if(freq < SW_FREQ_MIN)
	    freq = SW_FREQ_MIN;

	/* Format:
	 *
	 *	obj_num
	 *	sect_x sect_y sect_z
	 *	x y z
	 *	heading pitch bank
	 *	velocity
	 *	velocity_heading velocity_pitch velocity_bank
	 *	freq yield
	 */
	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 + 40 + 40 +
	    40 +
	    40 + 40 + 40 +
	    40 + 40 + 80
	);
	sprintf(
	    sndbuf,
"%i %i\
 %i\
 %ld %ld %ld\
 %.4f %.4f %.4f\
 %.4f %.4f %.4f\
 %.4f\
 %.4f %.4f %.4f\
 %.2f %.4f\n",
	    CS_CODE_EXT, SWEXTCMD_FIREWEAPON,
	    obj_num,
	    obj->sect_x, obj->sect_y, obj->sect_z,
	    obj->x, obj->y, obj->z,
	    heading, pitch, bank,	/* Firing direction */
	    obj->velocity,
	    obj->velocity_heading, 0.0 * PI, 0.0 * PI,
	    freq, pvals->weapon_yield
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Tractor beam lock.
 */
void NetSendTractorBeamLock(int obj_num, int tar_obj_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);

	/* Format:
	 *
	 *	obj_num tar_obj_num
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT, SWEXTCMD_TRACTORBEAMLOCK,
	    obj_num, tar_obj_num
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Hail.
 */
void NetSendHail(int obj_num, int tar_obj_num, int channel)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 40 + 40 + 80
	);

	/* SWEXTCMD_HAIL format:
	 *
	 *	obj_num tar_obj_num bearing channel
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i %.4f %i\n",
	    CS_CODE_EXT, SWEXTCMD_HAIL,
	    obj_num, tar_obj_num,
	    0.0f * PI, channel
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Com message.
 */
void NetSendComMessage(
	int obj_num, int tar_obj_num,
	int channel, const char *msg
)
{
	char *sndbuf;

	if(STRISEMPTY(msg))
	    return;

	/* Format:
	 *
	 *	obj_num tar_obj_num
	 *	bearing channel;message
	 */
	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 + 40 + 40 + STRLEN(msg) + 80
	);
	sprintf(
	    sndbuf,
"%i %i\
 %i %i %.4f %i;%s\n",
	    CS_CODE_EXT, SWEXTCMD_COMMESSAGE,
	    obj_num,
	    tar_obj_num,	/* Can be -1 */
	    0.0f * PI,		/* Bearing (ignored) */
	    channel,
	    msg
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Wormhole Enter.
 */
void NetSendWormHoleEnter(int obj_num, int tar_obj_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);

	/* Format:
	 *
	 *	obj_num tar_obj_num
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT, SWEXTCMD_WORMHOLEENTER,
	    obj_num, tar_obj_num
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Elink enter.
 */
void NetSendELinkEnter(int obj_num, int tar_obj_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);                        

	/* Format:
	 *
	 *	obj_num, tar_obj_num
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT, SWEXTCMD_ELINKENTER,
	    obj_num, tar_obj_num
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Disarm weapon.
 */
void NetSendWeaponDisarm(int obj_num, int tar_obj_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);                        

	/* Format:
	 *
	 *	obj_num tar_obj_num
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\n",
	    CS_CODE_EXT,
	    SWEXTCMD_WEPDISARM,
	    obj_num, tar_obj_num
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Auto destruct.
 */
void NetSendAutoDestruct(int obj_num, time_t sec)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 40 + 80
	);                        
				     
	/* Format:
	 *
	 *	obj_num sec
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %lu\n",
	    CS_CODE_EXT,
	    SWEXTCMD_AUTODESTRUCT,
	    obj_num, (unsigned long)sec
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Request description.
 */
void NetSendReqDescription(int obj_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 80
	);

	/* Format:
	 *
	 *	obj_num
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i\n",
	    CS_CODE_EXT, SWEXTCMD_REQDESCRIPTION,
	    obj_num
	);
	NetSendData(sndbuf);
	free(sndbuf);
}


/*
 *	Request economy values.
 */
void NetSendEcoReqValues(int obj_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 40 + 40 + 80
	);

	/* Format:
	 *
	 *	obj_num
	 */
	sprintf(
	    sndbuf, 
"%i %i\
 %i\n",
	    CS_CODE_EXT, SWEXTCMD_ECO_REQVALUES,
	    obj_num
	);
	NetSendData(sndbuf);
	free(sndbuf);
}

/*
 *	Economy buy.
 *
 *	Purchases the specified product from the proprietor to the
 *	customer.
 */
void NetSendEcoBuy(
	const int customer_obj_num, const int proprietor_obj_num,
	xsw_ecoproduct_struct *product
)
{
	char *sndbuf;

	if(product == NULL)
	    return;

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 +
	    80
	);

	/* Format:
	 *
	 *	customer_obj_num proprietor_obj_num
	 *	ocs_code amount
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\
 %i %.4f\n",
	    CS_CODE_EXT, SWEXTCMD_ECO_BUY,
	    customer_obj_num, proprietor_obj_num,
	    product->ocs_code, product->amount
	);
	NetSendData(sndbuf);
	free(sndbuf);
}


/*
 *	Economy sell.
 *
 *	Sells the specified product from the customer to the
 *	proprietor.
 */
void NetSendEcoSell(
	const int customer_obj_num, const int proprietor_obj_num,
	xsw_ecoproduct_struct *product
)
{
	char *sndbuf;

	if(product == NULL)
	    return;

	sndbuf = (char *)malloc( 
	    40 + 40 +
	    40 + 40 +
	    80
	);

	/* Format:
	 *
	 *	customer_obj_num proprietor_obj_num
	 *	ocs_code amount
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i\
 %i %.4f\n",
	    CS_CODE_EXT, SWEXTCMD_ECO_SELL,
	    customer_obj_num, proprietor_obj_num,
	    product->ocs_code, product->amount
	);
	NetSendData(sndbuf);
	free(sndbuf);
}
