#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "gctl.h"
#include "optwin.h"
#include "xsw.h"

#include "../include/oswkeyvals.h"


static void CREATE_PROMPT(
	xsw_options_win_struct *optwin, prompt_struct *prompt,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *name,
	unsigned int buf_len
);
static void CREATE_BROWSE_BUTTON(
	xsw_options_win_struct *optwin,
	prompt_struct *prompt,		/* Reference prompt */
	push_button_struct *btn,
	void (*activate_cb)(push_button_struct *, void *)
);
static void DRAW_OUTLINE(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *name,
	pixel_t fg_pix,
	pixel_t bg_pix,
	pixel_t text_pix
);
static void DRAW_LABEL(
	drawable_t d,
	int x, int y,
	const char *text,
	pixel_t pix
);

static void OptWinResizeCB(void *widget, void *data);
static void OptWinDrawCB(void *widget, void *data);

xsw_options_win_struct *OptWinNew(void);
void OptWinTabRemap(xsw_options_win_struct *optwin, int tab);
void OptWinResize(xsw_options_win_struct *optwin);
void OptWinQueueResize(xsw_options_win_struct *optwin);
void OptWinDraw(xsw_options_win_struct *optwin);
void OptWinQueueDraw(xsw_options_win_struct *optwin);
int OptWinManage(xsw_options_win_struct *optwin, event_t *event);
void OptWinUpdateWidgets(xsw_options_win_struct *optwin);
void OptWinMap(xsw_options_win_struct *optwin, win_t toplevel);
void OptWinUnmap(xsw_options_win_struct *optwin);
void OptWinDelete(xsw_options_win_struct *optwin);


#define OPTWIN_DEF_WIDTH		640
#define OPTWIN_DEF_HEIGHT		480

#define OPTWIN_MARGIN			10

#define OPTWIN_ICON_BTN_WIDTH		26
#define OPTWIN_ICON_BTN_HEIGHT		26

#define OPTWIN_BTN_WIDTH		70
#define OPTWIN_BTN_HEIGHT		28

#define OPTWIN_PULIST_HEIGHT		26
#define OPTWIN_PROMPT_HEIGHT		30

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Toggle button array names:
 */

/* Throttle mode */
static char *throttle_mode_names[] = {
	"Normal", 
	"Bi-Directional",
	"Incremental"
};
/* Show formal name labels */
static char *show_formal_label_names[] = {
	"Never",
	"As needed",
	"Always"
};
/* Sound amount */
static char *sound_amount_names[] = {
	"None",
	"Events",
	"Engines",
	"All"
};
/* Sound servers */
static char *sound_server_names[] = {
	"None",
	"YIFF",
	"EsounD",
	"MikMod"
};


/*
 *	Creates a prompt.
 */
static void CREATE_PROMPT(
	xsw_options_win_struct *optwin, prompt_struct *prompt,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *name,
	unsigned int buf_len
)
{
	PromptInit(
	    prompt, optwin->toplevel,
	    x, y, 
	    width, height,
	    PROMPT_STYLE_FLUSHED,
	    name,
	    buf_len,
	    0
	);
}

/*
 *	Creates a browse button.
 */
static void CREATE_BROWSE_BUTTON(
	xsw_options_win_struct *optwin,
	prompt_struct *prompt,		/* Reference prompt */
	push_button_struct *btn,
	void (*activate_cb)(push_button_struct *, void *)
)
{
	int x, y;
	const unsigned int	width = OPTWIN_ICON_BTN_WIDTH,
				height = OPTWIN_ICON_BTN_HEIGHT;
	widget_global_struct *wglobal = &widget_global;

	if(prompt != NULL)
	{
	    x = prompt->x + (int)prompt->width + 2;
	    y = prompt->y +
		(((int)prompt->height - (int)height) / 2);
	}
	else
	{
	    x = 0;
	    y = 0;
	}

	PBtnInit(
	    btn, optwin->toplevel,
	    x, y, width, height,
	    NULL,
	    PBTN_LABEL_ALIGN_CENTER,
	    wglobal->browse_files_pixmap,
	    wglobal->browse_files_mask,
	    activate_cb, optwin
	);
}


/*
 *	Draws an outline and label.
 */
static void DRAW_OUTLINE(
	drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *name,
	pixel_t fg_pix,
	pixel_t bg_pix,
	pixel_t text_pix
)
{
	const int margin = 2;
	int	lbearing, rbearing, swidth, ascent, descent,
		text_start_pos,
		text_len;		/* In pixels (including margin) */
	font_t *font = widget_global.std_font;

	if((d == None) || (name == NULL))
	    return;

	OSWGetTextBounds(
	    font, name, -1,
	    &lbearing, &rbearing, &swidth,
	    &ascent, &descent
	);
	text_start_pos = OPTWIN_MARGIN;
	text_len = swidth + (2 * margin);

	/* Draw shadows */
	OSWSetForeground(bg_pix);
	OSWDrawLine(d,			/* left */
	     x,
	     y + (int)height,
	     x,
	     y
	);
	OSWDrawLine(d,			/* top1 */
	     x,
	     y,
	     x + (int)text_start_pos,
	     y
	);
	OSWDrawLine(d,			/* Top2 */
	     x + (int)text_len + (int)text_start_pos,
	     y,
	     x + (int)width,
	     y
	);
	OSWDrawLine(d,			/* right */
	    x + (int)width - 1,
	    y + 1,
	    x + (int)width - 1,
	    y + (int)height - 1
	);
	OSWDrawLine(d,			/* bottom */
	    x + 1,
	    y + (int)height - 1,
	    x + (int)width - 1,
	    y + (int)height - 1
	);

	/* Draw highlights */
	OSWSetForeground(fg_pix);
	OSWDrawLine(d,			/* left */
	     x + 1,
	     y + (int)height - 2,
	     x + 1,
	     y + 1
	);
	OSWDrawLine(d,			/* top1 */
	     x + 1,
	     y + 1,
	     x + text_start_pos,
	     y + 1
	);
	OSWDrawLine(d,			/* Top2 */
	     x + text_len + text_start_pos,
	     y + 1,
	     x + (int)width - 2,
	     y + 1
	);
	OSWDrawLine(d,			/* right */
	    x + (int)width,
	    y + 1,
	    x + (int)width,
	    y + (int)height
	);
	OSWDrawLine(d,			/* bottom */
	    x + 1,
	    y + (int)height,
	    x + (int)width,
	    y + (int)height
	);

	/* Draw text */
	OSWSetForeground(text_pix);
	OSWDrawString(
	    d,
	    x + 2 + text_start_pos + margin - lbearing,
	    y - ((ascent + descent) / 2) + ascent,
	    name
	);
}

/*
 *	Draws a label, sets appropriate foreground pixel.
 */
static void DRAW_LABEL(
	drawable_t d,
	int x, int y,
	const char *text,
	pixel_t pix
)
{
	int lbearing, rbearing, swidth, ascent, descent;
	font_t *font = OSWQueryCurrentFont();

	if((d == None) || (text == NULL))
	    return;

	OSWGetTextBounds(
	    font, text, -1,
	    &lbearing, &rbearing, &swidth,
	    &ascent, &descent
	);
	OSWSetForeground(pix);
	OSWDrawString(
	    d,
	    x - lbearing,
	    y + ascent,
	    text
	);
}


/*
 *	Resize callback.
 */
static void OptWinResizeCB(void *widget, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(widget);
	if(optwin == NULL)
	    return;

	OptWinResize(optwin);
}

/*
 *	Draw callback.
 */
static void OptWinDrawCB(void *widget, void *data)
{
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(widget);
	if(optwin == NULL)
	    return;

	if(optwin->map_state)
	    OptWinDraw(optwin);
}


/*
 *	Creates a new options window.
 */
xsw_options_win_struct *OptWinNew(void)
{
	int x, y;
	unsigned int width, height;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	pixmap_t pixmap;
	image_t *img;
	push_button_struct *btn;
	check_button_struct *cb;
	radio_struct *radio;
	popup_list_struct *pulist;
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	widget_global_struct *wglobal = &widget_global;
	xsw_options_win_struct *optwin = XSW_OPTIONS_WIN(calloc(
	    1, sizeof(xsw_options_win_struct)
	));
	if(optwin == NULL)
	    return(NULL);

	optwin->map_state = False;
	optwin->has_focus = False;
	optwin->disabled = False;
	optwin->freeze_count = 0;
	optwin->busy_count = 0;
	optwin->x = 0;
	optwin->y = 0;
	optwin->width = OPTWIN_DEF_WIDTH;
	optwin->height = OPTWIN_DEF_HEIGHT;
	optwin->visibility_state = VisibilityFullyObscured;

	optwin->tab = OPTWIN_TAB_CODE_GENERAL;
	optwin->has_changes = False;

	/* Toplevel */
	x = optwin->x;
	y = optwin->y;
	width = optwin->width;
	height = optwin->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	optwin->toplevel = w;
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);

	/* WM properties */
	if(ImgIsAllocated(XSW_IMG_ICON_OPTIONS))
	    pixmap = OSWCreatePixmapFromImage(
		xsw_image[XSW_IMG_ICON_OPTIONS]->image
	    );
	else
	    pixmap = wglobal->std_icon_pm;
	OSWSetWindowWMProperties(
	    w,
	    PROG_NAME_FULL ": Options",
	    "Options",		/* Icon title */
	    pixmap,		/* Icon */
	    False,		/* Let WM set coordinates? */
	    optwin->x, optwin->y,
	    100, 100,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);

	parent = w;

	/* Create buttons */

	/* OK button */
	btn = &optwin->ok_btn;
	width = OPTWIN_BTN_WIDTH;
	height = OPTWIN_BTN_HEIGHT;
	x = ((int)optwin->width / 5 / 2) - ((int)width / 2);
	y = (int)optwin->height - OPTWIN_BTN_HEIGHT - OPTWIN_MARGIN;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "OK",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinOKPBCB, optwin
	);
	accel_keys[0] = OSW_KEY_o;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_Return;
	accel_mods[1] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Apply button */
	btn = &optwin->apply_btn;
	x += (int)optwin->width / 5;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Apply",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinApplyPBCB, optwin
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Save button */
	btn = &optwin->save_btn;
	x += (int)optwin->width / 5;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Save",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinSavePBCB, optwin
	);
	accel_keys[0] = OSW_KEY_s;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Defaults button */
	btn = &optwin->defaults_btn;
	x += (int)optwin->width / 5;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Defaults",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinDefaultsPBCB, optwin
	);
	accel_keys[0] = OSW_KEY_d;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	/* Cancel button */
	btn = &optwin->cancel_btn;
	x += (int)optwin->width / 5;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinCancelPBCB, optwin
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);


	/* Create tabs */
	img = wglobal->tab_unselected_img;

	/* General */
	x = 2 * OPTWIN_MARGIN;
	y = OPTWIN_MARGIN;
	OSWCreateWindow(
	    &optwin->general_tab, parent,
	    x, y,
	    img->width, img->height
	);
	OSWSetWindowInput(optwin->general_tab, OSW_EVENTMASK_BUTTON);

	/* Graphics */
	x += OPTWIN_MARGIN + (int)img->width;
	OSWCreateWindow(
	    &optwin->graphics_tab, parent,
	    x, y,
	    img->width, img->height
	);
	OSWSetWindowInput(optwin->graphics_tab, OSW_EVENTMASK_BUTTON);

	/* Sounds */
	x += OPTWIN_MARGIN + (int)img->width;
	OSWCreateWindow(
	    &optwin->sounds_tab, parent,
	    x, y,
	    img->width, img->height
	);
	OSWSetWindowInput(optwin->sounds_tab, OSW_EVENTMASK_BUTTON);

	/* Network */
	x += OPTWIN_MARGIN + (int)img->width;
	OSWCreateWindow(
	    &optwin->network_tab, parent,
	    x, y,
	    img->width, img->height
	);
	OSWSetWindowInput(optwin->network_tab, OSW_EVENTMASK_BUTTON);

	/* Misc */
	x += OPTWIN_MARGIN + (int)img->width;
	OSWCreateWindow(
	    &optwin->misc_tab, parent,
	    x, y,
	    img->width, img->height
	);
	OSWSetWindowInput(optwin->misc_tab, OSW_EVENTMASK_BUTTON);


	/* Create pages */
#define NUM_STR_LEN			40

	/* General Page */

	/* Show Universes List On Startup */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 67;
	cb = &optwin->show_univ_list_win_cb;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Universes List On Startup"
	);
	height = optwin->show_univ_list_win_cb.height;

	/* Show Economy Window On Orbit */
	cb = &optwin->show_eco_win_cb;
	y += (int)height + (OPTWIN_MARGIN / 2);
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Economy Window On Orbit"
	);

	/* CPU Priority */
	pulist = &optwin->cpu_priority_pulist;
	x = 420;
	y = 67;
	width = 190;
	height = 0;
	PUListInit(
	    pulist, parent,
	    x, y, width, height,
	    4,				/* Items visible */
	    PULIST_POPUP_BELOW
	);
	PUListAddItem(
	    pulist,
	    0,
	    "Highest",
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    "High",
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    "Medium",
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    "Lowest",
	    None, None
	);

	/* Controller Popup List */
	pulist = &optwin->controller_pulist;
	x = 70;
	y = 180;
	width = 150;
	height = 0;
	PUListInit(
	    pulist, parent,
	    x, y, width, height,
	    3,			/* Items visable */
	    PULIST_POPUP_BELOW
	);
	PUListAddItem(
	    pulist,
	    0,
	    "Keyboard",
	    None, None
	);
#ifdef JS_SUPPORT
	PUListAddItem(
	    pulist, 
	    0,
	    "Joystick",
	    None, None 
	); 
#else
	PUListAddItem(
	    pulist, 
	    PULIST_ITEM_FLAG_DISABLED,
	    "Joystick",
	    None, None 
	); 
#endif
	PUListAddItem(
	    pulist,
	    PULIST_ITEM_FLAG_DISABLED,
	    "Pointer",
	    None, None
	);

	/* Throttle Mode */
	radio = &optwin->throttle_mode_radio;
	x = 120;
	y = 215;
	RadioInit(
	    radio, parent,
	    x, y,			/* x, y */
	    sizeof(throttle_mode_names) / sizeof(char *),
	    0,				/* Default armed button */
	    throttle_mode_names,	/* Array of names */
	    sizeof(throttle_mode_names) / sizeof(char *),
	    RADIO_ALIGN_HORIZONTAL
	);

	/* Map Keyboard Button */
	x = 500;
	y = 180;
	width = 110;
	height = OPTWIN_BTN_HEIGHT;
	PBtnInit(
	    &optwin->keymap_btn, parent,
	    x, y, width, height,
	    "Map Keyboard",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinKeymapPBCB, optwin
	);
	PBtnSetTooltip(
	    &optwin->keymap_btn,
"Click this to configure keyboard operations"
	);
	y += (int)height + 5;

#ifdef JS_SUPPORT
	width = 110;
	height = OPTWIN_BTN_HEIGHT;
	PBtnInit(
	    &optwin->jsmap_btn, parent,
	    x, y, width, height,
	    "Map Joystick",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinJSMapPBCB, optwin
	);
	PBtnSetTooltip(
	    &optwin->jsmap_btn,
"Click this to configure joystick operations"
	);
#endif	/* JS_SUPPORT */

#if 0
	/* OCS Names */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 300;
	width = 320;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->ocsn_prompt,
	    x, y, width, height,
	    "OCS Names:",
	    PATH_MAX + NAME_MAX
	);
	PromptSetTooltip(
	    &optwin->ocsn_prompt,
"Enter the OCS names file. This file is usually automatically set\
 by the server."
	);
	CREATE_BROWSE_BUTTON(
	    optwin,
	    &optwin->ocsn_prompt,
	    &optwin->ocsn_browse_btn,
	    OptWinBrowseOCSNsPBCB
	);
	y += (int)height;

	/* ImgRefs */
	CREATE_PROMPT(
	    optwin, &optwin->imgrefs_prompt, 
	    x, y, width, height,
	    "ImgRefs:",
	    PATH_MAX + NAME_MAX
	);
	PromptSetTooltip(
	    &optwin->imgrefs_prompt,
"Enter the ImgRefs file. This file is usually automatically set\
 by the server."
	);
	CREATE_BROWSE_BUTTON(
	    optwin,
	    &optwin->imgrefs_prompt,
	    &optwin->imgrefs_browse_btn,
	    OptWinBrowseImgRefsPBCB
	);
	y += (int)height;

	/* SndRefs */
	CREATE_PROMPT(
	    optwin, &optwin->sndrefs_prompt,
	    x, y, width, height,
	    "SndRefs:",
	    PATH_MAX + NAME_MAX
	);
	PromptSetTooltip(
	    &optwin->sndrefs_prompt,
"Enter the SndRefs file. This file is usually automatically set\
 by the server."
	);
	CREATE_BROWSE_BUTTON(
	    optwin,
	    &optwin->sndrefs_prompt,
	    &optwin->sndrefs_browse_btn,
	    OptWinBrowseSndRefsPBCB
	);

	/* Link prompts togeather */ 
	optwin->ocsn_prompt.prev = &optwin->sndrefs_prompt;
	optwin->ocsn_prompt.next = &optwin->imgrefs_prompt;

	optwin->imgrefs_prompt.prev = &optwin->ocsn_prompt;
	optwin->imgrefs_prompt.next = &optwin->sndrefs_prompt;

	optwin->sndrefs_prompt.prev = &optwin->imgrefs_prompt;
	optwin->sndrefs_prompt.next = &optwin->ocsn_prompt;
#endif

	/* Units */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 300;
	width = 0;
	height = OPTWIN_PULIST_HEIGHT;
	LabelInit(
	    &optwin->units_label, parent,
	    x, y, width, height,
	    "Units:"
	);
	width = optwin->units_label.width;
	x += (int)width;
	width = 150;
	pulist = &optwin->units_pulist;
	PUListInit(
	    pulist, parent,
	    x, y, width, height,
	    4,			/* Items visible */
	    PULIST_POPUP_BELOW
	);
	/* Add the unit values, the index must match the XSW_UNTIS_*
	 * value
	 */
	PUListAddItem(
	    pulist,
	    0,
	    "Metric",
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    "English",
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    "Parsecs",
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    "Program Internal",
	    None, None
	);

	/* Automatic Viewscreen Zoom */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y += (int)height + (OPTWIN_MARGIN / 2);
	cb = &optwin->auto_zoom_cb;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Auto Zoom"
	);
	CBtnSetTooltip(
	    cb,
"Check this to automatically zoom the viewscreen based on the\
 locked-on object."
	); 
	y += 25;

	/* Local Simulation */
	cb = &optwin->local_updates_cb;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Local Simulation"
	);
	CBtnSetTooltip(
	    cb,
"Check this to simulate objects locally (improves simulation during\
 periods of lag)."
	);
	y += 25;

	/* Notify Scanner Contacts */
	cb = &optwin->notify_scanner_contacts_cb;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Contacts Notify"
	);
	CBtnSetTooltip(
	    cb,
"Check this to print a message and beep whenever an object enters\
 or leaves your vessel's scanner."
	);

	/* Configure VSMenu */
	x = 500;
	y = 300;
	width = 110;
	height = OPTWIN_BTN_HEIGHT;
	PBtnInit(
	    &optwin->configure_vsmenu_btn, parent,
	    x, y, width, height,
	    "Configure Menu",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinConfigureVSMenuPBCB, optwin
	);
	PBtnSetTooltip(
	    &optwin->configure_vsmenu_btn,
"Click this to configure the viewscreen menu"
	);


	/* Graphics Page */

	/* Show Viewscreen Selected Weapon */
	cb = &optwin->show_viewscreen_selected_weapon_cb;
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 67;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Selected Weapon"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display the selected weapon (when weapons are\
 online) on the viewscreen."
	);

	/* Show Viewscreen Markings */
	cb = &optwin->show_viewscreen_markings_cb;
	y += 25;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Markings"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display the heading arrow, weapon direction arrow\
 (when weapons are online), and the lock cursor (when the scanner\
 is locked on an object) on the viewscreen."
	);

	/* Show Viewscreen Object Labels */
	cb = &optwin->show_viewscreen_object_labels_cb;
	y += 25;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Object Labels"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display the object's name next to it on the viewscreen."
	);

	/* Show Viewscreen Netstats */
	cb = &optwin->show_viewscreen_netstats_cb;
	y += 25;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Netstats"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display the netstats label on the viewscreen."
	);

	/* Formal Object Names */
	radio = &optwin->show_formal_label_radio;
	x = 60;
	y = 190;
	RadioInit(
	    radio, parent,
	    x, y,
	    sizeof(show_formal_label_names) / sizeof(char *),
	    0,				/* Default armed button */
	    show_formal_label_names,	/* Names */
	    sizeof(show_formal_label_names) / sizeof(char *),
	    RADIO_ALIGN_HORIZONTAL
	);
 

	/* Lens Flares */ 
	cb = &optwin->show_lens_flares_cb;
	x = 80 + (int)(((int)optwin->width - 80) * 0.3f);
	y = 67;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Lens Flares"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display lens flares when bright light sources\
 (such as stars and explosions) are visible on the viewscreen."
	);

	/* Nebula Glow */
	cb = &optwin->show_nebula_glow_cb;
	y += 25;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Nebula Glow"
	);
	CBtnSetTooltip(
	    cb,
"Check this to perform gamma affects while in nebulas when bright\
 light sources (such as stars and explosions) are visible on the\
 viewscreen. Otherwise, gamma effects are abscent while in nebulas."
	);

	/* Strobe Glow */
	cb = &optwin->show_strobe_glow_cb;
	y += 25;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Strobe Glow"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display strobe lights with \"stary arms\"."
	);

	/* Fires */
	cb = &optwin->show_fires_cb;
	y += 25;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Fires"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display flame particles on objects that are\
 on fire."
	);

	/* Smoke Trails */
	cb = &optwin->show_smoke_trails_cb;
	x = 80 + (int)(((int)optwin->width - 80) * 0.6f);
	y = 67;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Smoke Trails"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display trails of smoke particles behind\
 objects that are to produce smoke."
	);

	/* Max Particles */
	y += 25;
	width = 150;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->max_particles_prompt,
	    x, y, width, height,
	    "Max Particles:",
	    40
	);
	PromptSetTooltip(
	    &optwin->max_particles_prompt,
"Enter the maximum number of particles that can exist at once."
	);


	/* Passive Image Loading */
	cb = &optwin->async_image_loading_cb;
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 295; 
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Passive Image Loading"
	);
	CBtnSetTooltip(
	    cb,
"Check this to open images asynchronously."
	);
	height = (int)cb->height;

	/* Pixels/Cycle Limit */
	y += (int)height;
	width = 230;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->async_image_pixels_prompt,
	    x, y, width, height,
	    "Limit (pixels/cycle):",
	    40
	);
	PromptSetTooltip(
	    &optwin->async_image_pixels_prompt,
"Enter the maximum number of pixels to read per cycle when opening\
 images asynchronously."
	);

	/* Link prompts togeather */


	/* Sounds Page */

	/* Sound Server Type */
	radio = &optwin->server_type_radio;
	x = 60;
	y = 90;
	RadioInit(
	    radio, parent,
	    x, y,
	    sizeof(sound_server_names) / sizeof(char *),
	    0,				/* Initially armed button */
	    sound_server_names,		/* Labels */
	    sizeof(sound_server_names) / sizeof(char *),
	    RADIO_ALIGN_HORIZONTAL
	);
	radio->check_button[0]->disabled = False;
#ifdef HAVE_Y2
	radio->check_button[1]->disabled = False;
#else
	radio->check_button[1]->disabled = True;
#endif	/* HAVE_Y2 */
#ifdef HAVE_ESD
	radio->check_button[2]->disabled = False;
#else
	radio->check_button[2]->disabled = True;
#endif	/* HAVE_ESD */
	radio->check_button[3]->disabled = True;

	/* Start Command Prompt */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y += (int)radio->height + (OPTWIN_MARGIN / 2);
	width = 360;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->sound_server_prompt,
	    x, y, width, height,
	    "Start Command:",
	    PATH_MAX + NAME_MAX
	);
	PromptSetTooltip(
	    &optwin->sound_server_prompt,
"Enter the command and any arguments that are needed to start the\
 specified sound server."
	);

	/* Address */
	y += (int)height;
	CREATE_PROMPT(
	    optwin, &optwin->sound_con_arg_prompt,
	    x, y, width, height,
	    "Address:",
	    PATH_MAX + NAME_MAX
	);
	PromptSetTooltip(
	    &optwin->sound_con_arg_prompt,
"Enter the address or url of the specified\
 sound server (for YIFF the format is \"<address>:<port>\",\
 such as \"localhost:9433\")."
	);


	/* Test Sound Button */
	x = 420;
	y = 67;
	width = 100;
	height = OPTWIN_BTN_HEIGHT;
	PBtnInit(
	    &optwin->sound_test_btn, parent,
	    x, y, width, height,
	    "Test Sound",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinTestSoundPBCB, optwin
	);

	/* Flip Stereo */
	cb = &optwin->flip_stereo_cb;
	y += (int)height + OPTWIN_MARGIN;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Flip Stereo"
	);


	/* Sounds Level */
	radio = &optwin->sounds_radio;
	x = 60;
	y = 285;
	RadioInit(
	    radio, parent,
	    x, y,
	    sizeof(sound_amount_names) / sizeof(char *),
	    0,				/* Initially armed button */
	    sound_amount_names,		/* Labels */
	    sizeof(sound_amount_names) / sizeof(char *),
	    RADIO_ALIGN_HORIZONTAL
	);

	/* Music */
	x = 420;
	y = 265;
	cb = &optwin->music_cb;
	CBtnInit(
	    cb, parent,
	    x, y,
	    False,
	    "Music"
	);

	/* Link prompts togeather */
	optwin->sound_server_prompt.prev = &optwin->sound_con_arg_prompt;
	optwin->sound_server_prompt.next = &optwin->sound_con_arg_prompt;

	optwin->sound_con_arg_prompt.prev = &optwin->sound_server_prompt;
	optwin->sound_con_arg_prompt.next = &optwin->sound_server_prompt;


	/* Net Page */

	/* Default Address */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 67;
	width = 310;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->def_address_prompt,
	    x, y, width, height,
	    "Default Address:",
	    MAX_URL_LEN
	);
	/* Default Port */
	y += (int)height;
	CREATE_PROMPT(
	    optwin, &optwin->def_port_prompt,
	    x, y, width, height,
	    "Default Port:", 
	    NUM_STR_LEN
	);

	/* Bandwidth */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 200;
	width = 310;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->net_bandwidth_prompt,
	    x, y, width, height,
	    "Bandwidth (bytes/sec):",
	    NUM_STR_LEN
	);
	PromptSetTooltip(
	    &optwin->net_bandwidth_prompt,
"Enter the bandwidth of your network connection in bytes per second.\
 It is recommended that you enter a value slightly smaller than\
 the maximum bandwidth in order to avoid overflows."
	);

	/* Show errors */
	cb = &optwin->show_net_errors_cb;
	x = 370;
	y = 67;
	CBtnInit(
	    cb, parent,
	    x, y,
	    option.show_net_errors ? True : False,
	    "Show Network Errors"
	);
	CBtnSetTooltip(
	    cb,
"Check this to print network warnings and errors, or uncheck this\
 to only print fatal network errors."
	);

	/* Show server errors */
	cb = &optwin->show_server_errors_cb;
	y += 25;
	CBtnInit(
	    cb, parent,
	    x, y,
	    option.show_server_errors ? True : False,
	    "Show Server Message Dialogs"
	);
	CBtnSetTooltip(
	    cb,
"Check this to display message dialogs from the server."
	);

	/* Interval */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 290;
	width = 310;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->net_int_prompt,
	    x, y, width, height,
	    "Update Interval (ms):",
	    NUM_STR_LEN
	);
	PromptSetTooltip(
	    &optwin->net_int_prompt,
"Enter the Network Update Interval in milliseconds. If Auto\
 Interval Tunning is enabled then this value is automatically\
 set and updated."
	);

	/* Auto Interval */
	cb = &optwin->auto_interval_cb;
	x = 370;
	y += 5;
	CBtnInit(
	    cb, parent,
	    x, y,
	    auto_interval_tune.state ? True : False,
	    "Auto Interval Tunning"
	);
	CBtnSetTooltip(
	    cb,
"Check this to automatically adjust the Network Update\
 Interval based on the specified bandwidth."
	);

	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y += 25;
	width = 310;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->net_timeout_prompt,
	    x, y,
	    width, height,
	    "Network Operations Timeout (sec):",
	    NUM_STR_LEN
	);
	PromptSetTooltip(
	    &optwin->net_timeout_prompt,
"Enter the Network Operations Timeout in seconds. This applies to\
 resolving host addresses, connecting sockets, shuting down sockets,\
 and closing sockets."
	);

	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y += 30;
	width = 310;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->net_object_outdated_prompt,
	    x, y,
	    width, height,
	    "Object Outdated Time (ms):",
	    NUM_STR_LEN
	);
	PromptSetTooltip(
	    &optwin->net_object_outdated_prompt,
"Enter the timeout in milliseconds in which objects not updated\
 by the server for this long of a period will not be displayed."
	);


	/* Link prompts togeather */
	optwin->def_address_prompt.prev = &optwin->net_object_outdated_prompt;
	optwin->def_address_prompt.next = &optwin->def_port_prompt;

	optwin->def_port_prompt.prev = &optwin->def_address_prompt;
	optwin->def_port_prompt.next = &optwin->net_bandwidth_prompt;

	optwin->net_bandwidth_prompt.prev = &optwin->def_port_prompt;
	optwin->net_bandwidth_prompt.next = &optwin->net_int_prompt;

	optwin->net_int_prompt.prev = &optwin->net_bandwidth_prompt;
	optwin->net_int_prompt.next = &optwin->net_timeout_prompt;

	optwin->net_timeout_prompt.prev = &optwin->net_int_prompt.prev;
	optwin->net_timeout_prompt.next = &optwin->net_object_outdated_prompt;

	optwin->net_object_outdated_prompt.prev = &optwin->net_timeout_prompt;
	optwin->net_object_outdated_prompt.next = &optwin->def_address_prompt;


	/* Misc Page */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 67;
	width = 330;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->xsw_local_toplevel_path_prompt,
	    x, y, width, height,
	    "Local Toplevel:",
	    PATH_MAX
	);
	y += (int)height;
	CREATE_PROMPT(
	    optwin, &optwin->xsw_toplevel_path_prompt,
	    x, y, width, height,
	    "Toplevel:",
	    PATH_MAX
	);
	y += (int)height;
	CREATE_PROMPT(
	    optwin, &optwin->xsw_etc_path_prompt,
	    x, y, width, height,
	    "Etc:",
	    PATH_MAX
	);
	y += (int)height;
	CREATE_PROMPT(
	    optwin, &optwin->xsw_images_path_prompt,
	    x, y, width, height,
	    "Images:",
	    PATH_MAX
	);
	y += (int)height;
	CREATE_PROMPT(
	    optwin, &optwin->xsw_sounds_path_prompt,
	    x, y, width, height,
	    "Sounds:",
	    PATH_MAX
	);
	y += (int)height;
	CREATE_PROMPT(
	    optwin, &optwin->xsw_downloads_path_prompt,
	    x, y, width, height,
	    "Downloads:",
	    PATH_MAX
	);

#ifdef JS_SUPPORT
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 290;
	width = 300;
	height = OPTWIN_PROMPT_HEIGHT;
	CREATE_PROMPT(
	    optwin, &optwin->js_calib_path_prompt,
	    x, y, width, height,
	    "Joystick Calibration:",
	    PATH_MAX + NAME_MAX
	);
	CREATE_BROWSE_BUTTON(
	    optwin,
	    &optwin->js_calib_path_prompt,
	    &optwin->js_calib_path_browse_btn,
	    OptWinBrowseJSCalPBCB
	);
#endif /* JS_SUPPORT */

	PBtnInit(
	    &optwin->refresh_memory_btn, parent,
	    440, 67,
	    140, OPTWIN_BTN_HEIGHT,
	    "Refresh Memory",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinRefreshMemoryPBCB, optwin
	);

	/* Link prompts togeather */
#ifdef JS_SUPPORT
	optwin->xsw_local_toplevel_path_prompt.prev =
	    &optwin->js_calib_path_prompt;
#else	/* JS_SUPPORT */
	optwin->xsw_local_toplevel_path_prompt.prev =
	    &optwin->xsw_downloads_path_prompt;
#endif	/* JS_SUPPORT */
	optwin->xsw_local_toplevel_path_prompt.next =
	    &optwin->xsw_toplevel_path_prompt;

	optwin->xsw_toplevel_path_prompt.prev =
	    &optwin->xsw_local_toplevel_path_prompt;
	optwin->xsw_toplevel_path_prompt.next =
	    &optwin->xsw_etc_path_prompt;

	optwin->xsw_etc_path_prompt.prev =
	    &optwin->xsw_toplevel_path_prompt;
	optwin->xsw_etc_path_prompt.next =
	    &optwin->xsw_images_path_prompt;

	optwin->xsw_images_path_prompt.prev =
	    &optwin->xsw_etc_path_prompt;
	optwin->xsw_images_path_prompt.next =
	    &optwin->xsw_sounds_path_prompt;

	optwin->xsw_sounds_path_prompt.prev =
	    &optwin->xsw_images_path_prompt;
	optwin->xsw_sounds_path_prompt.next =
	    &optwin->xsw_downloads_path_prompt;

	optwin->xsw_downloads_path_prompt.prev =
	    &optwin->xsw_sounds_path_prompt;
#ifdef JS_SUPPORT
	optwin->xsw_downloads_path_prompt.next =
	    &optwin->js_calib_path_prompt;
#else	/* JS_SUPPORT */
	optwin->xsw_downloads_path_prompt.next =
	    &optwin->xsw_local_toplevel_path_prompt;
#endif	/* JS_SUPPORT */

#ifdef JS_SUPPORT
	optwin->js_calib_path_prompt.prev =
	    &optwin->xsw_downloads_path_prompt;
	optwin->js_calib_path_prompt.next =
	    &optwin->xsw_local_toplevel_path_prompt;
#endif	/* JS_SUPPORT */

#undef NUM_STR_LEN

	return(optwin);
}



/*
 *	Remaps the widgets and windows on the options window
 *	depending on the givin tab and sets the new tab.
 *
 *	This function will also do all the widget redrawing.
 */
void OptWinTabRemap(xsw_options_win_struct *optwin, int tab)
{
	if(optwin == NULL)
	    return;

	/* Sanitize tab */
	if((tab != OPTWIN_TAB_CODE_GENERAL) &&
	   (tab != OPTWIN_TAB_CODE_GRAPHICS) &&
	   (tab != OPTWIN_TAB_CODE_SOUNDS) &&
	   (tab != OPTWIN_TAB_CODE_NET) &&
	   (tab != OPTWIN_TAB_CODE_MISC)
	)
	    tab = OPTWIN_TAB_CODE_GENERAL;

	/* Set new tab */
	optwin->tab = tab;


	/* General */
	if(optwin->tab == OPTWIN_TAB_CODE_GENERAL)
	{
	    WidgetMap(&optwin->show_univ_list_win_cb);
	    WidgetMap(&optwin->show_eco_win_cb);
	    WidgetMap(&optwin->cpu_priority_pulist);
	    WidgetMap(&optwin->controller_pulist);
	    WidgetMap(&optwin->throttle_mode_radio);
	    WidgetMap(&optwin->keymap_btn);
#ifdef JS_SUPPORT
	    WidgetMap(&optwin->jsmap_btn);
#endif	/* JS_SUPPORT */

	    WidgetMap(&optwin->units_pulist);
	    WidgetMap(&optwin->units_label);
	    WidgetMap(&optwin->auto_zoom_cb);
	    WidgetMap(&optwin->local_updates_cb);
	    WidgetMap(&optwin->notify_scanner_contacts_cb);
            WidgetMap(&optwin->configure_vsmenu_btn);
#if 0
	    WidgetMap(&optwin->imgrefs_prompt);
	    WidgetMap(&optwin->ocsn_prompt);
	    WidgetMap(&optwin->sndrefs_prompt);

	    WidgetMap(&optwin->imgrefs_browse_btn);
	    WidgetMap(&optwin->ocsn_browse_btn);
	    WidgetMap(&optwin->sndrefs_browse_btn);
#endif
	}
	else
	{
	    WidgetUnmap(&optwin->show_univ_list_win_cb);
	    WidgetUnmap(&optwin->show_eco_win_cb);
	    WidgetUnmap(&optwin->cpu_priority_pulist);
	    WidgetUnmap(&optwin->controller_pulist);
	    WidgetUnmap(&optwin->throttle_mode_radio);
	    WidgetUnmap(&optwin->keymap_btn);
#ifdef JS_SUPPORT
	    WidgetUnmap(&optwin->jsmap_btn);
#endif  /* JS_SUPPORT */

	    WidgetUnmap(&optwin->units_pulist);
	    WidgetUnmap(&optwin->units_label);
	    WidgetUnmap(&optwin->auto_zoom_cb);
	    WidgetUnmap(&optwin->local_updates_cb);
	    WidgetUnmap(&optwin->notify_scanner_contacts_cb);
            WidgetUnmap(&optwin->configure_vsmenu_btn);
#if 0
	    WidgetUnmap(&optwin->imgrefs_prompt);
	    WidgetUnmap(&optwin->ocsn_prompt);
	    WidgetUnmap(&optwin->sndrefs_prompt); 

	    WidgetUnmap(&optwin->imgrefs_browse_btn);
	    WidgetUnmap(&optwin->ocsn_browse_btn);
	    WidgetUnmap(&optwin->sndrefs_browse_btn);
#endif
	}

	/* Graphics */
	if(optwin->tab == OPTWIN_TAB_CODE_GRAPHICS)
	{
	    WidgetMap(&optwin->show_viewscreen_selected_weapon_cb);
	    WidgetMap(&optwin->show_viewscreen_markings_cb);
	    WidgetMap(&optwin->show_viewscreen_object_labels_cb);
	    WidgetMap(&optwin->show_viewscreen_netstats_cb);
	    WidgetMap(&optwin->show_formal_label_radio);
	    WidgetMap(&optwin->show_lens_flares_cb);
	    WidgetMap(&optwin->show_nebula_glow_cb);
	    WidgetMap(&optwin->show_strobe_glow_cb);
	    WidgetMap(&optwin->show_fires_cb);
	    WidgetMap(&optwin->show_smoke_trails_cb);
	    WidgetMap(&optwin->max_particles_prompt);
	    WidgetMap(&optwin->async_image_loading_cb);
	    WidgetMap(&optwin->async_image_pixels_prompt);
	}
	else
	{
	    WidgetUnmap(&optwin->show_viewscreen_selected_weapon_cb);
	    WidgetUnmap(&optwin->show_viewscreen_markings_cb);
	    WidgetUnmap(&optwin->show_viewscreen_object_labels_cb);
	    WidgetUnmap(&optwin->show_viewscreen_netstats_cb);
	    WidgetUnmap(&optwin->show_formal_label_radio); 
	    WidgetUnmap(&optwin->show_lens_flares_cb);
	    WidgetUnmap(&optwin->show_nebula_glow_cb);
	    WidgetUnmap(&optwin->show_strobe_glow_cb);
	    WidgetUnmap(&optwin->show_fires_cb);
	    WidgetUnmap(&optwin->show_smoke_trails_cb);
	    WidgetUnmap(&optwin->max_particles_prompt);
	    WidgetUnmap(&optwin->async_image_loading_cb);
	    WidgetUnmap(&optwin->async_image_pixels_prompt);
	}

	/* Sounds */
	if(optwin->tab == OPTWIN_TAB_CODE_SOUNDS)
	{
	    WidgetMap(&optwin->sounds_radio);
	    WidgetMap(&optwin->music_cb);
	    WidgetMap(&optwin->server_type_radio);
	    WidgetMap(&optwin->sound_server_prompt);
	    WidgetMap(&optwin->sound_con_arg_prompt);
	    WidgetMap(&optwin->sound_test_btn);
	    WidgetMap(&optwin->flip_stereo_cb);
	}
	else
	{
	    WidgetUnmap(&optwin->sounds_radio);
	    WidgetUnmap(&optwin->music_cb);
	    WidgetUnmap(&optwin->server_type_radio);
	    WidgetUnmap(&optwin->sound_server_prompt);
	    WidgetUnmap(&optwin->sound_con_arg_prompt);
	    WidgetUnmap(&optwin->sound_test_btn);
	    WidgetUnmap(&optwin->flip_stereo_cb);
	}

	/* Network */
	if(optwin->tab == OPTWIN_TAB_CODE_NET)
	{
	    WidgetMap(&optwin->def_address_prompt);
	    WidgetMap(&optwin->def_port_prompt);
	    WidgetMap(&optwin->show_net_errors_cb);
	    WidgetMap(&optwin->show_server_errors_cb);
	    WidgetMap(&optwin->net_bandwidth_prompt);
	    WidgetMap(&optwin->net_int_prompt);
	    WidgetMap(&optwin->auto_interval_cb);
	    WidgetMap(&optwin->net_timeout_prompt);
	    WidgetMap(&optwin->net_object_outdated_prompt);
	}
	else
	{
	    WidgetUnmap(&optwin->def_address_prompt); 
	    WidgetUnmap(&optwin->def_port_prompt);
	    WidgetUnmap(&optwin->show_net_errors_cb);
	    WidgetUnmap(&optwin->show_server_errors_cb);
	    WidgetUnmap(&optwin->net_bandwidth_prompt);
	    WidgetUnmap(&optwin->net_int_prompt);
	    WidgetUnmap(&optwin->auto_interval_cb);
	    WidgetUnmap(&optwin->net_timeout_prompt);
	    WidgetUnmap(&optwin->net_object_outdated_prompt);
	}

	/* Misc */
	if(optwin->tab == OPTWIN_TAB_CODE_MISC)
	{
	    WidgetMap(&optwin->xsw_local_toplevel_path_prompt);
	    WidgetMap(&optwin->xsw_toplevel_path_prompt);
	    WidgetMap(&optwin->xsw_etc_path_prompt);
	    WidgetMap(&optwin->xsw_images_path_prompt);
	    WidgetMap(&optwin->xsw_sounds_path_prompt);
	    WidgetMap(&optwin->xsw_downloads_path_prompt);
#ifdef JS_SUPPORT
	    WidgetMap(&optwin->js_calib_path_prompt);
	    WidgetMap(&optwin->js_calib_path_browse_btn);
#endif /* JS_SUPPORT */
	    WidgetMap(&optwin->refresh_memory_btn);
	}
	else
	{
	    WidgetUnmap(&optwin->xsw_local_toplevel_path_prompt);
	    WidgetUnmap(&optwin->xsw_toplevel_path_prompt);
	    WidgetUnmap(&optwin->xsw_etc_path_prompt);
	    WidgetUnmap(&optwin->xsw_images_path_prompt);
	    WidgetUnmap(&optwin->xsw_sounds_path_prompt);
	    WidgetUnmap(&optwin->xsw_downloads_path_prompt);
#ifdef JS_SUPPORT
	    WidgetUnmap(&optwin->js_calib_path_prompt);
	    WidgetUnmap(&optwin->js_calib_path_browse_btn);
#endif
	    WidgetUnmap(&optwin->refresh_memory_btn);
	}
}

/*
 *	Resizes the options window.
 */
void OptWinResize(xsw_options_win_struct *optwin)
{
	int x, y;
	unsigned int width, height;
	win_t w;

	if(optwin == NULL)
	    return;

	w = optwin->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);
	optwin->x = x;
	optwin->y = y;
	optwin->width = width;
	optwin->height = height;
}

/*
 *	Queues the options window to resize.
 */
void OptWinQueueResize(xsw_options_win_struct *optwin)
{
	if(optwin == NULL)
	    return;

	WidgetQueueResizeAction(
	    OptWinResizeCB, optwin, NULL
	);
}

/*
 *	Draws the options window.
 */
void OptWinDraw(xsw_options_win_struct *optwin)
{
	int	lbearing, rbearing, swidth, ascent, descent,
		font_height;
	unsigned int width, height;
	char *s;
	font_t *font;
	pixmap_t pixmap;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if(optwin == NULL)
	    return;

	/* Recreate buffers as needed */
	if(optwin->tab_win_buf == None)
	{
	    OSWGetWindowGeometry(
		optwin->general_tab, NULL, NULL, &width, &height, NULL
	    );
	    if(OSWCreatePixmap(
		&optwin->tab_win_buf, width, height
	    ))
		return;
	}

	font = wglobal->std_font;
	OSWSetFont(font);

	/* Clear toplevel buffer */
	WidgetPutImageTile(
	    optwin->toplevel,
	    wglobal->std_bkg_img,
	    optwin->width, optwin->height
	);

	/* Draw frame on toplevel */
	OSWGetWindowGeometry(
	    optwin->general_tab, NULL, NULL, &width, &height, NULL
	);
	OSWSetForeground(wglobal->surface_highlight_pix);
	OSWDrawLine(
	    optwin->toplevel,
	    OPTWIN_MARGIN,
	    OPTWIN_MARGIN + (int)height - 1,
	    (int)optwin->width - OPTWIN_MARGIN,
	    OPTWIN_MARGIN + (int)height - 1
	);
	OSWDrawLine(
	    optwin->toplevel,
	    OPTWIN_MARGIN,
	    OPTWIN_MARGIN + (int)height - 1,
	    OPTWIN_MARGIN,
	    (int)optwin->height - (OPTWIN_BTN_HEIGHT + OPTWIN_MARGIN +
		OPTWIN_MARGIN)
	);

	OSWSetForeground(wglobal->surface_shadow_pix);
	OSWDrawLine(
	    optwin->toplevel,
	    OPTWIN_MARGIN,
	    (int)optwin->height - (OPTWIN_BTN_HEIGHT + OPTWIN_MARGIN
		+ OPTWIN_MARGIN),
	    (int)optwin->width - OPTWIN_MARGIN - 1,
	    (int)optwin->height - (OPTWIN_BTN_HEIGHT + OPTWIN_MARGIN
		+ OPTWIN_MARGIN)
	);
	OSWDrawLine(
	    optwin->toplevel,
	    (int)optwin->width - OPTWIN_MARGIN,
	    OPTWIN_MARGIN + (int)height,
	    (int)optwin->width - OPTWIN_MARGIN,
	    (int)optwin->height - (OPTWIN_BTN_HEIGHT +
		OPTWIN_MARGIN + OPTWIN_MARGIN)
	);


	/* Redraw by selected tab */

	/* General */
	w = optwin->general_tab;
	pixmap = optwin->tab_win_buf;
	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);
	WidgetPutImageTile(
	    pixmap, wglobal->std_bkg_img,
	    width, height
	);
	if(optwin->tab == OPTWIN_TAB_CODE_GENERAL)
	{
	    int lbearing, rbearing, swidth, ascent, descent, font_height;
	    const char *s;
	    radio_struct *radio;
	    popup_list_struct *pulist;

	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_selected_img,
		0, 0,
		True
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		(2 * OPTWIN_MARGIN), 55,
		290, 90,
		"Show",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		330, 55,
		290, 90,
		"Performance",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		(2 * OPTWIN_MARGIN), 165,
		optwin->width - (4 * OPTWIN_MARGIN), 102,
		"Controller",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		(2 * OPTWIN_MARGIN), 285,
		optwin->width - (4 * OPTWIN_MARGIN), 135,
		"Viewscreen & Scanner",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    s = "CPU Priority:",
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    pulist = &optwin->cpu_priority_pulist;
	    OSWSetForeground(wglobal->normal_text_pix);
	    OSWDrawString(
		optwin->toplevel,
		pulist->x - swidth - 2 - lbearing,
		pulist->y + (((int)pulist->height - font_height) / 2) +
		    ascent,
		s
	    );

	    s = "Type:",
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    pulist = &optwin->controller_pulist;
	    OSWSetForeground(wglobal->normal_text_pix);
	    OSWDrawString(
		optwin->toplevel,
		pulist->x - swidth - 2 - lbearing,
		pulist->y + (((int)pulist->height - font_height) / 2) +
		    ascent,
		s
	    );

	    s = "Throttle Mode:",
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    radio = &optwin->throttle_mode_radio;
	    OSWSetForeground(wglobal->normal_text_pix);
	    OSWDrawString(
		optwin->toplevel,
		radio->x - swidth - 2 - lbearing,
		radio->y + (((int)radio->height - font_height) / 2) +
		    ascent,
		s
	    );

	}
	else
	{
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_unselected_img,
		0, 0,
		True 
	    );
	}
	OSWSetForeground(wglobal->normal_text_pix);
	s = "General";
	OSWGetTextBounds(
	    font, s, -1,
	    &lbearing, &rbearing, &swidth, &ascent, &descent
	);
	font_height = ascent + descent;
	OSWDrawString(
	    pixmap,
	    (((int)width - swidth) / 2) - lbearing,
	    (((int)height - font_height) / 2) + 3 + ascent,
	    s
	);
	OSWPutDrawableToWindow(w, pixmap);


	/* Graphics */
	w = optwin->graphics_tab;
	pixmap = optwin->tab_win_buf;
	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);
	WidgetPutImageTile(
	    pixmap, wglobal->std_bkg_img,
	    width, height
	);
	if(optwin->tab == OPTWIN_TAB_CODE_GRAPHICS)
	{ 
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_selected_img,
		0, 0,
		True
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		2 * OPTWIN_MARGIN, 55,
		optwin->width - (4 * OPTWIN_MARGIN), 210,
		"Details",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_LABEL(
		optwin->toplevel,
		40, 170,
		"Use Formal Object Names:",
		wglobal->normal_text_pix
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		2 * OPTWIN_MARGIN, 280,
		optwin->width - (4 * OPTWIN_MARGIN), 140,
		"Images",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );
	} 
	else
	{
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_unselected_img,
		0, 0,
		True
	    );
	}
	OSWSetForeground(wglobal->normal_text_pix);
	s = "Graphics";
	OSWGetTextBounds(
	    font, s, -1,
	    &lbearing, &rbearing, &swidth, &ascent, &descent
	);
	font_height = ascent + descent;
	OSWDrawString(
	    pixmap,
	    (((int)width - swidth) / 2) - lbearing,
	    (((int)height - font_height) / 2) + 3 + ascent,
	    s      
	);
	OSWPutDrawableToWindow(w, pixmap);


	/* Sounds */
	w = optwin->sounds_tab;
	pixmap = optwin->tab_win_buf;
	OSWGetWindowGeometry(
	    optwin->sounds_tab, NULL, NULL, &width, &height, NULL
	);
	WidgetPutImageTile(
	    pixmap, wglobal->std_bkg_img,
	    width, height
	);
	if(optwin->tab == OPTWIN_TAB_CODE_SOUNDS)
	{
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_selected_img,
		0, 0,
		True
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		2 * OPTWIN_MARGIN, 55,
		optwin->width - (4 * OPTWIN_MARGIN), 180,
		"Sound Server",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		2 * OPTWIN_MARGIN, 250,
		optwin->width - (4 * OPTWIN_MARGIN), 170,
		"Sound Options",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_LABEL(
		optwin->toplevel,
		40, 70,
		"Sound Server Type:",
		wglobal->normal_text_pix
	    );

	    DRAW_LABEL(
		optwin->toplevel,
		40, 265,
		"Amount Level:",
		wglobal->normal_text_pix
	    );
	}
	else
	{
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_unselected_img,
		0, 0,
		True
	    );
	}
	OSWSetForeground(wglobal->normal_text_pix);
	s = "Sounds";
	OSWGetTextBounds(
	    font, s, -1,
	    &lbearing, &rbearing, &swidth, &ascent, &descent
	);
	font_height = ascent + descent;
	OSWDrawString(
	    pixmap,
	    (((int)width - swidth) / 2) - lbearing,
	    (((int)height - font_height) / 2) + 3 + ascent,
	    s
	);
	OSWPutDrawableToWindow(w, pixmap);


	/* Network */
	w = optwin->network_tab;
	pixmap = optwin->tab_win_buf;
	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);
	WidgetPutImageTile(
	    pixmap, wglobal->std_bkg_img,
	    width, height
	);
	if(optwin->tab == OPTWIN_TAB_CODE_NET)
	{
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_selected_img,
		0, 0,
		True 
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		2 * OPTWIN_MARGIN, 55,
		optwin->width - (4 * OPTWIN_MARGIN), 120,
		"General",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		2 * OPTWIN_MARGIN, 190,
		optwin->width - (4 * OPTWIN_MARGIN), 70,
		"Bandwidth",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
	        2 * OPTWIN_MARGIN, 275,
		optwin->width - (4 * OPTWIN_MARGIN), 145,
		"Timmings",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	}
	else
	{
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_unselected_img,
		0, 0,
		True
	    );
	}
	OSWSetForeground(wglobal->normal_text_pix);
	s = "Network";
	OSWGetTextBounds(
	    font, s, -1,
	    &lbearing, &rbearing, &swidth, &ascent, &descent
	);
	font_height = ascent + descent;
	OSWDrawString(
	    pixmap,
	    (((int)width - swidth) / 2) - lbearing,
	    (((int)height - font_height) / 2) + 3 + ascent,
	    s
	);
	OSWPutDrawableToWindow(w, pixmap);


	/* Misc */
	w = optwin->misc_tab;
	pixmap = optwin->tab_win_buf;
	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);
	WidgetPutImageTile(
	    pixmap, wglobal->std_bkg_img,
	    width, height
	);
	if(optwin->tab == OPTWIN_TAB_CODE_MISC)
	{
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_selected_img,
		0, 0,
		True
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		2 * OPTWIN_MARGIN, 55,
		355, 205,
		"Directories",
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

	    DRAW_OUTLINE(
		optwin->toplevel,
		2 * OPTWIN_MARGIN, 275,
		355, 145,
		"Files", 
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );

#if 0
	    DRAW_OUTLINE(
		optwin->toplevel,
		390, 160,
		230, 260,
		"Dimensions", 
		wglobal->surface_highlight_pix,
		wglobal->surface_shadow_pix,
		wglobal->normal_text_pix
	    );
#endif
	}
	else
	{
	    WidgetPutImageNormal(
		pixmap,
		wglobal->tab_unselected_img,
		0, 0,
		True
	    );
	}
	OSWSetForeground(wglobal->normal_text_pix);
	s = "Misc";
	OSWGetTextBounds(
	    font, s, -1,
	    &lbearing, &rbearing, &swidth, &ascent, &descent
	);
	font_height = ascent + descent;
	OSWDrawString(
	    pixmap,
	    (((int)width - swidth) / 2) - lbearing,
	    (((int)height - font_height) / 2) + 3 + ascent,
	    s
	);
	OSWPutDrawableToWindow(w, pixmap);
}

/*
 *	Queues the options window to draw.
 */
void OptWinQueueDraw(xsw_options_win_struct *optwin)
{
	if(optwin == NULL)
	    return;

	WidgetQueueDrawAction(
	    OptWinDrawCB, optwin, NULL
	);
}

/*
 *	Manages the options window event.
 */
int OptWinManage(xsw_options_win_struct *optwin, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_visibility_t *visibility;

	if((optwin == NULL) || (event == NULL))
	    return(events_handled);

	if(!optwin->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == optwin->toplevel)
	    {
		OptWinQueueResize(optwin);
		OptWinQueueDraw(optwin);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == optwin->toplevel)
	    {
		OptWinQueueDraw(optwin);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!optwin->has_focus)
		return(events_handled);

	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!optwin->has_focus)
		return(events_handled);

	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    /* Tabs */
	    if(w == optwin->general_tab)
	    {
		optwin->tab = OPTWIN_TAB_CODE_GENERAL;
		events_handled++;
	    }
	    else if(w == optwin->graphics_tab)
	    {
		optwin->tab = OPTWIN_TAB_CODE_GRAPHICS;
		events_handled++;
	    }
	    else if(w == optwin->sounds_tab)
	    {
		optwin->tab = OPTWIN_TAB_CODE_SOUNDS;
		events_handled++;
	    }
	    else if(w == optwin->network_tab)
	    {
		optwin->tab = OPTWIN_TAB_CODE_NET;
		events_handled++;
	    }
	    else if(w == optwin->misc_tab)
	    {
		optwin->tab = OPTWIN_TAB_CODE_MISC;
		events_handled++;
	    }

	    /* Handle remapping of tabs if the event was handled */
	    if(events_handled > 0)
	    {
		OptWinTabRemap(optwin, optwin->tab);
		OptWinQueueDraw(optwin);
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    break;

	  case FocusIn:
	    if(w == optwin->toplevel)
	    {
		optwin->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == optwin->toplevel)
	    {
		optwin->has_focus = False;
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == optwin->toplevel)
	    {
		optwin->visibility_state = visibility->state;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(optwin->toplevel, event, &t))
	    {
		OSWSetInputFocus(optwin->toplevel, RevertToParent, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(optwin->toplevel, event))
	    {
		OptWinUnmap(optwin);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}


	/* Manage widgets */

#define WIDGET_MANAGE(_widget_)		{		\
 if(events_handled == 0)				\
  events_handled += WidgetManage((_widget_), event);	\
}
	/* Buttons */
	WIDGET_MANAGE(&optwin->ok_btn);
	WIDGET_MANAGE(&optwin->apply_btn);
	WIDGET_MANAGE(&optwin->save_btn);
	WIDGET_MANAGE(&optwin->defaults_btn);
	WIDGET_MANAGE(&optwin->cancel_btn);

	WIDGET_MANAGE(&optwin->show_univ_list_win_cb);
	WIDGET_MANAGE(&optwin->show_eco_win_cb);
	WIDGET_MANAGE(&optwin->cpu_priority_pulist);
	WIDGET_MANAGE(&optwin->controller_pulist);
	WIDGET_MANAGE(&optwin->throttle_mode_radio);
	WIDGET_MANAGE(&optwin->keymap_btn);
#ifdef JS_SUPPORT
	WIDGET_MANAGE(&optwin->jsmap_btn);
#endif
	WIDGET_MANAGE(&optwin->units_pulist);
	WIDGET_MANAGE(&optwin->units_label);
	WIDGET_MANAGE(&optwin->auto_zoom_cb);
	WIDGET_MANAGE(&optwin->local_updates_cb);
	WIDGET_MANAGE(&optwin->notify_scanner_contacts_cb);
        WIDGET_MANAGE(&optwin->configure_vsmenu_btn);
#if 0
	WIDGET_MANAGE(&optwin->imgrefs_prompt);
	WIDGET_MANAGE(&optwin->imgrefs_browse_btn);
	WIDGET_MANAGE(&optwin->ocsn_prompt);
	WIDGET_MANAGE(&optwin->ocsn_browse_btn);
	WIDGET_MANAGE(&optwin->sndrefs_prompt);
	WIDGET_MANAGE(&optwin->sndrefs_browse_btn);
#endif
	WIDGET_MANAGE(&optwin->show_viewscreen_selected_weapon_cb);
	WIDGET_MANAGE(&optwin->show_viewscreen_markings_cb);
	WIDGET_MANAGE(&optwin->show_viewscreen_object_labels_cb);
	WIDGET_MANAGE(&optwin->show_viewscreen_netstats_cb);
	WIDGET_MANAGE(&optwin->show_lens_flares_cb);
	WIDGET_MANAGE(&optwin->show_nebula_glow_cb);
	WIDGET_MANAGE(&optwin->show_strobe_glow_cb);
	WIDGET_MANAGE(&optwin->show_fires_cb);
	WIDGET_MANAGE(&optwin->show_smoke_trails_cb);
	WIDGET_MANAGE(&optwin->max_particles_prompt);
	WIDGET_MANAGE(&optwin->show_formal_label_radio);
	WIDGET_MANAGE(&optwin->async_image_loading_cb);
	WIDGET_MANAGE(&optwin->async_image_pixels_prompt);

	WIDGET_MANAGE(&optwin->server_type_radio);
	WIDGET_MANAGE(&optwin->sound_server_prompt);
	WIDGET_MANAGE(&optwin->sound_con_arg_prompt);
	WIDGET_MANAGE(&optwin->sound_test_btn);
	WIDGET_MANAGE(&optwin->sounds_radio);
	WIDGET_MANAGE(&optwin->music_cb);
	WIDGET_MANAGE(&optwin->flip_stereo_cb);

	WIDGET_MANAGE(&optwin->def_address_prompt);
	WIDGET_MANAGE(&optwin->def_port_prompt);
	WIDGET_MANAGE(&optwin->show_net_errors_cb);
	WIDGET_MANAGE(&optwin->show_server_errors_cb);
	WIDGET_MANAGE(&optwin->net_bandwidth_prompt);
	WIDGET_MANAGE(&optwin->net_int_prompt);
	WIDGET_MANAGE(&optwin->auto_interval_cb);
	WIDGET_MANAGE(&optwin->net_timeout_prompt);
	WIDGET_MANAGE(&optwin->net_object_outdated_prompt);

	WIDGET_MANAGE(&optwin->xsw_local_toplevel_path_prompt);
	WIDGET_MANAGE(&optwin->xsw_toplevel_path_prompt);
	WIDGET_MANAGE(&optwin->xsw_etc_path_prompt);
	WIDGET_MANAGE(&optwin->xsw_images_path_prompt);
	WIDGET_MANAGE(&optwin->xsw_sounds_path_prompt);
	WIDGET_MANAGE(&optwin->xsw_downloads_path_prompt);
#ifdef JS_SUPPORT
	WIDGET_MANAGE(&optwin->js_calib_path_prompt);
	WIDGET_MANAGE(&optwin->js_calib_path_browse_btn);
#endif
	WIDGET_MANAGE(&optwin->refresh_memory_btn);

#undef WIDGET_MANAGE

	return(events_handled);
}

/*
 *	Updates the options window's widgets to reflect current data
 *	values.
 */
void OptWinUpdateWidgets(xsw_options_win_struct *optwin)
{
	xsw_option_struct *opt = &option;

	if(optwin == NULL)
	    return;

#ifdef JS_SUPPORT
	WidgetSetDisabled(
	    &optwin->jsmap_btn,
	    (opt->controller == XSW_CONTROLLER_JOYSTICK) ? False : True
	);
#endif

}

/*
 *	Maps the options window.
 */
void OptWinMap(xsw_options_win_struct *optwin, win_t toplevel)
{
	win_t w;

	if(optwin == NULL)
	    return;

	if(optwin->toplevel == None)
	    return;

	/* Get values */
	OptWinFetchValues(optwin);

	w = optwin->toplevel;
	OSWSetTransientFor(w, toplevel);
	if(toplevel != None)
	    WidgetCenterWindowToWindow(w, toplevel);
	else
	    WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_ROOT);

	/* Map */
	if(!optwin->map_state)
	{
	    OSWMapRaised(optwin->toplevel);
	    optwin->map_state = True;
	    WidgetMap(&optwin->ok_btn);
	    WidgetMap(&optwin->cancel_btn);
	    WidgetMap(&optwin->apply_btn);
	    WidgetMap(&optwin->defaults_btn);
	    WidgetMap(&optwin->save_btn);
	    OSWMapWindow(optwin->general_tab);
	    OSWMapWindow(optwin->graphics_tab);
	    OSWMapWindow(optwin->sounds_tab);
	    OSWMapWindow(optwin->network_tab);
	    OSWMapWindow(optwin->misc_tab);
	}

	/* Set tab */
	OptWinTabRemap(optwin, optwin->tab);

	/* Reset has changes mark */
	optwin->has_changes = False;

	/* Focus */
	OSWGrabFocus(optwin->toplevel);
}


/*
 *	Unmaps the options window.
 */
void OptWinUnmap(xsw_options_win_struct *optwin)
{
	if(optwin == NULL)
	    return;

	OSWUnmapWindow(optwin->toplevel);
	optwin->map_state = False;
	optwin->has_focus = False;
	optwin->visibility_state = VisibilityFullyObscured;

	/* General */
	WidgetUnmap(&optwin->show_univ_list_win_cb);
	WidgetUnmap(&optwin->show_eco_win_cb);
	WidgetUnmap(&optwin->cpu_priority_pulist);
	WidgetUnmap(&optwin->controller_pulist);
	WidgetUnmap(&optwin->throttle_mode_radio);
	WidgetUnmap(&optwin->keymap_btn);
#ifdef JS_SUPPORT
	WidgetUnmap(&optwin->jsmap_btn);
#endif

	WidgetUnmap(&optwin->units_pulist);
	WidgetUnmap(&optwin->units_label);
	WidgetUnmap(&optwin->auto_zoom_cb);
	WidgetUnmap(&optwin->local_updates_cb);
	WidgetUnmap(&optwin->notify_scanner_contacts_cb);
        WidgetUnmap(&optwin->configure_vsmenu_btn);
#if 0
	WidgetUnmap(&optwin->imgrefs_prompt);
	WidgetUnmap(&optwin->ocsn_prompt);
	WidgetUnmap(&optwin->sndrefs_prompt);

	WidgetUnmap(&optwin->imgrefs_browse_btn);
	WidgetUnmap(&optwin->ocsn_browse_btn);
	WidgetUnmap(&optwin->sndrefs_browse_btn);
#endif

	/* Graphics */
	WidgetUnmap(&optwin->show_viewscreen_selected_weapon_cb);
	WidgetUnmap(&optwin->show_viewscreen_markings_cb);
	WidgetUnmap(&optwin->show_viewscreen_object_labels_cb);
	WidgetUnmap(&optwin->show_viewscreen_netstats_cb);
	WidgetUnmap(&optwin->show_formal_label_radio);
	WidgetUnmap(&optwin->show_lens_flares_cb);
	WidgetUnmap(&optwin->show_nebula_glow_cb);
	WidgetUnmap(&optwin->show_strobe_glow_cb);
	WidgetUnmap(&optwin->show_fires_cb);
	WidgetUnmap(&optwin->show_smoke_trails_cb);
	WidgetUnmap(&optwin->max_particles_prompt);
	WidgetUnmap(&optwin->async_image_loading_cb);
	WidgetUnmap(&optwin->async_image_pixels_prompt);

	/* Sounds */
	WidgetUnmap(&optwin->sounds_radio);
	WidgetUnmap(&optwin->music_cb);
	WidgetUnmap(&optwin->server_type_radio);
	WidgetUnmap(&optwin->sound_server_prompt);
	WidgetUnmap(&optwin->sound_con_arg_prompt);
	WidgetUnmap(&optwin->sound_test_btn);
	WidgetUnmap(&optwin->flip_stereo_cb);

	/* Network */
	WidgetUnmap(&optwin->def_address_prompt);
	WidgetUnmap(&optwin->def_port_prompt);
	WidgetUnmap(&optwin->show_net_errors_cb);
	WidgetUnmap(&optwin->show_server_errors_cb);
	WidgetUnmap(&optwin->net_bandwidth_prompt);
	WidgetUnmap(&optwin->net_int_prompt);
	WidgetUnmap(&optwin->auto_interval_cb);
	WidgetUnmap(&optwin->net_timeout_prompt);
	WidgetUnmap(&optwin->net_object_outdated_prompt);

	/* Misc */
	WidgetUnmap(&optwin->xsw_local_toplevel_path_prompt);
	WidgetUnmap(&optwin->xsw_toplevel_path_prompt);
	WidgetUnmap(&optwin->xsw_etc_path_prompt);
	WidgetUnmap(&optwin->xsw_images_path_prompt);
	WidgetUnmap(&optwin->xsw_sounds_path_prompt);
	WidgetUnmap(&optwin->xsw_downloads_path_prompt);
#ifdef JS_SUPPORT
	WidgetUnmap(&optwin->js_calib_path_prompt);
	WidgetUnmap(&optwin->js_calib_path_browse_btn);
#endif
	WidgetUnmap(&optwin->refresh_memory_btn);


	/* Buttons */
	WidgetUnmap(&optwin->ok_btn);
	WidgetUnmap(&optwin->cancel_btn);
	WidgetUnmap(&optwin->apply_btn);
	WidgetUnmap(&optwin->defaults_btn);
	WidgetUnmap(&optwin->save_btn);


	/* Set has_changes to False */
	optwin->has_changes = False;
}

/*
 *	Deletes the options window.
 */
void OptWinDelete(xsw_options_win_struct *optwin)
{
	if(optwin == NULL)
	    return;

	/* General */
	WidgetDestroy(&optwin->show_univ_list_win_cb);
	WidgetDestroy(&optwin->show_eco_win_cb);
	WidgetDestroy(&optwin->cpu_priority_pulist);
	WidgetDestroy(&optwin->controller_pulist);
	WidgetDestroy(&optwin->throttle_mode_radio);
	WidgetDestroy(&optwin->keymap_btn);
#ifdef JS_SUPPORT
	WidgetDestroy(&optwin->jsmap_btn);
#endif

	WidgetDestroy(&optwin->units_pulist);
	WidgetDestroy(&optwin->units_label);
	WidgetDestroy(&optwin->auto_zoom_cb);
	WidgetDestroy(&optwin->local_updates_cb);
	WidgetDestroy(&optwin->notify_scanner_contacts_cb);
        WidgetDestroy(&optwin->configure_vsmenu_btn);
#if 0
	WidgetDestroy(&optwin->imgrefs_prompt);
	WidgetDestroy(&optwin->ocsn_prompt);
	WidgetDestroy(&optwin->sndrefs_prompt);

	WidgetDestroy(&optwin->imgrefs_browse_btn);
	WidgetDestroy(&optwin->ocsn_browse_btn);
	WidgetDestroy(&optwin->sndrefs_browse_btn);
#endif

	/* Graphics */
	WidgetDestroy(&optwin->show_viewscreen_selected_weapon_cb);
	WidgetDestroy(&optwin->show_viewscreen_markings_cb);
	WidgetDestroy(&optwin->show_viewscreen_object_labels_cb);
	WidgetDestroy(&optwin->show_viewscreen_netstats_cb);
	WidgetDestroy(&optwin->show_formal_label_radio);
	WidgetDestroy(&optwin->show_lens_flares_cb);
	WidgetDestroy(&optwin->show_nebula_glow_cb);
	WidgetDestroy(&optwin->show_strobe_glow_cb);
	WidgetDestroy(&optwin->show_fires_cb);
	WidgetDestroy(&optwin->show_smoke_trails_cb);
	WidgetDestroy(&optwin->max_particles_prompt);
	WidgetDestroy(&optwin->async_image_loading_cb);
	WidgetDestroy(&optwin->async_image_pixels_prompt);

	/* Sounds */
	WidgetDestroy(&optwin->sounds_radio);
	WidgetDestroy(&optwin->music_cb);
	WidgetDestroy(&optwin->server_type_radio);
	WidgetDestroy(&optwin->sound_server_prompt);
	WidgetDestroy(&optwin->sound_con_arg_prompt);
	WidgetDestroy(&optwin->sound_test_btn);
	WidgetDestroy(&optwin->flip_stereo_cb);

	/* Network */
	WidgetDestroy(&optwin->def_address_prompt);
	WidgetDestroy(&optwin->def_port_prompt);
	WidgetDestroy(&optwin->show_net_errors_cb);
	WidgetDestroy(&optwin->show_server_errors_cb);
	WidgetDestroy(&optwin->net_bandwidth_prompt);
	WidgetDestroy(&optwin->net_int_prompt);
	WidgetDestroy(&optwin->auto_interval_cb);
	WidgetDestroy(&optwin->net_timeout_prompt);
	WidgetDestroy(&optwin->net_object_outdated_prompt);

	/* Misc */
	WidgetDestroy(&optwin->xsw_local_toplevel_path_prompt);
	WidgetDestroy(&optwin->xsw_toplevel_path_prompt);
	WidgetDestroy(&optwin->xsw_etc_path_prompt);
	WidgetDestroy(&optwin->xsw_images_path_prompt);
	WidgetDestroy(&optwin->xsw_sounds_path_prompt);
	WidgetDestroy(&optwin->xsw_downloads_path_prompt);
#ifdef JS_SUPPORT
	WidgetDestroy(&optwin->js_calib_path_prompt);
	WidgetDestroy(&optwin->js_calib_path_browse_btn);
#endif
	WidgetDestroy(&optwin->refresh_memory_btn);


	/* Buttons */
	WidgetDestroy(&optwin->ok_btn);
	WidgetDestroy(&optwin->cancel_btn);
	WidgetDestroy(&optwin->apply_btn);
	WidgetDestroy(&optwin->defaults_btn);
	WidgetDestroy(&optwin->save_btn);

	/* Tabs */
	OSWDestroyWindow(&optwin->general_tab);
	OSWDestroyWindow(&optwin->graphics_tab);
	OSWDestroyWindow(&optwin->sounds_tab);
	OSWDestroyWindow(&optwin->network_tab);
	OSWDestroyWindow(&optwin->misc_tab);
	OSWDestroyPixmap(&optwin->tab_win_buf);

	/* Toplevel */
	OSWDestroyWindow(&optwin->toplevel); 

	free(optwin);
}
