#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/disk.h"
#include "../include/cfgfmt.h"

#include "page.h"
#include "xsw.h"


static image_t *PageOpenImage(const char *path);

int PageFileOpen(
	page_struct *p,
	win_t w, shared_image_t *image,
	const char *path
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


static image_t *PageOpenImage(const char *path)
{
	return(WidgetLoadImageFromTgaFile(path));
}


/*
 *	Load main menu configuration from file.
 *
 *	Current configuration in mm will be deleted first.
 */
int PageFileOpen(
	page_struct *p,
	win_t w, shared_image_t *image,
	const char *path
)
{
	int	lines_read = 0,
		label_num = -1;
	char	*s, *s2, *s3,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	off_t filesize;
	struct stat stat_buf;
	page_label_struct *label_ptr;

	if((path == NULL) || (p == NULL))
	    return(-2);

	if(stat(path, &stat_buf))
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("no such file");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    XSWPrintWarning(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	/* Get file size */
	filesize = stat_buf.st_size;

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for reading");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,  
		"%s: %s.",
		path, error_msg
	    );
	    XSWPrintWarning(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	/* Delete resources (if previously allocated) */
	PageDestroy(p, w, image);

#define WARN_NO_SUCH_PARM(_parm_)               {       \
 char *s = (char *)malloc(                              \
  STRLEN(path) + 40 + STRLEN(_parm_) + 80               \
 );                                                     \
 sprintf(                                               \
  s,                                                    \
"%s: Line %i: Unsupported parameter `%s'.",             \
  path, lines_read, (_parm_)                            \
 );                                                     \
 XSWPrintWarning(s);                                    \
 free(s);                                               \
}
#define WARN_NO_SUCH_FILE(_v_,_error_code_)     {       \
 const int error_code = (_error_code_);                 \
 char *s, *error_msg = STRDUP(strerror(error_code));    \
 if(error_msg == NULL)                                  \
  error_msg = strdup("no such file");                   \
 *error_msg = toupper(*error_msg);                      \
							\
 s = (char *)malloc(                                    \
  STRLEN(path) + 40 + STRLEN(parm) + STRLEN(_v_) +      \
  STRLEN(error_msg) + 80                                \
 );                                                     \
 sprintf(                                               \
  s,                                                    \
"%s: Line %i: %s: Warning: %s: %s.",                    \
  path, lines_read, parm, (_v_), error_msg              \
 );                                                     \
 XSWPrintWarning(s);                                    \
 free(s);                                               \
							\
 free(error_msg);                                       \
}

	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Free previous line and allocate/read next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Fetch parameter */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Fetch value */
	    s2 = StringCfgParseValue(s);
	    /* Set value to "0" if NULL */
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';


	    /* BackgroundImage */
	    if(!strcasecmp(parm, "BackgroundImage") ||
	       !strcasecmp(parm, "ImageBackground")
	    )
	    {
		char *v = ISPATHABSOLUTE(val) ?
		    STRDUP(val) :
		    STRDUP(PrefixPaths(
			dname.images, val
		    ));

		free(p->bg_filename);
		p->bg_filename = STRDUP(v);

		/* Do not load background image now, it will be loaded
		 * when the page is mapped
		 */

		free(v);
	    }
	    /* BackgroundStyle */
	    else if(!strcasecmp(parm, "BackgroundStyle"))
	    {
		p->bg_image_draw_code = ATOI(val);
	    }
	    /* BeginLabel */
	    else if(!strcasecmp(parm, "BeginLabel"))
	    {
		/* Allocate new main menu label */
		label_num = PageCreateLabel(p);
		if(PageIsLabelAllocated(p, label_num))
		    label_ptr = p->label[label_num];
		else
		    continue;

		while(1)
		{
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Fetch parameter */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Fetch value */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';


		    /* ImageUnarmed */
		    if(!strcasecmp(parm, "ImageUnarmed") ||
		       !strcasecmp(parm, "UnarmedImage")
		    )
		    {
			image_t *img;
			char *full_path = ISPATHABSOLUTE(val) ?
			    STRDUP(val) :
			    STRDUP(PrefixPaths(
				dname.images, val
			    ));
			xsw_imglabel_struct *imglabel = &label_ptr->imglabel[
			    PAGE_LABEL_STATE_UNARMED
			];

			/* Unload the previous image as needed */
			ImgLabelReset(imglabel);

			/* Open the label's image */
			img = PageOpenImage(full_path);
			if(img == NULL)
			{
			    char *s = (char *)malloc(
				STRLEN(path) + 40 +
				STRLEN(full_path) + 80
			    );
			    sprintf(
				s,
"%s: Line %i: %s: Warning: Unable to open image.",
				path, lines_read, full_path
			    );
			    XSWPrintWarning(s);
			    free(s);
			}
			imglabel->image = img;

			free(full_path);
		    }
		    /* ImageArmed */
		    else if(!strcasecmp(parm, "ImageArmed") ||
			    !strcasecmp(parm, "ArmedImage")
		    )
		    {
			image_t *img;
			char *full_path = ISPATHABSOLUTE(val) ?
			    STRDUP(val) :
			    STRDUP(PrefixPaths(
				dname.images, val
			    ));
                        xsw_imglabel_struct *imglabel = &label_ptr->imglabel[
                            PAGE_LABEL_STATE_ARMED
                        ];

			/* Unload the previous image as needed */
			ImgLabelReset(imglabel);

			/* Open the label's image */
			img = PageOpenImage(full_path);
			if(img == NULL)
			{
			    char *s = (char *)malloc(
				STRLEN(path) + 40 +
				STRLEN(full_path) + 80
			    );
			    sprintf(
				s,
"%s: Line %i: %s: Warning: Unable to open image.",
				path, lines_read, full_path
			    );
			    XSWPrintWarning(s);
			    free(s);  
			}
			imglabel->image = img;

			free(full_path);
		    }
		    /* ImageHighlighted */
		    else if(!strcasecmp(parm, "ImageHighlighted") ||
			    !strcasecmp(parm, "HighlightedImage")
		    )
		    {
			image_t *img;
			char *full_path = ISPATHABSOLUTE(val) ?
			    STRDUP(val) :
			    STRDUP(PrefixPaths(
				dname.images, val
			    ));
                        xsw_imglabel_struct *imglabel = &label_ptr->imglabel[
                            PAGE_LABEL_STATE_HIGHLIGHTED
                        ];

			/* Unload the previous image as needed */
			ImgLabelReset(imglabel);

			/* Open the label's image */
			img = PageOpenImage(full_path);
			if(img == NULL)
			{
			    char *s = (char *)malloc(
				STRLEN(path) + 40 +  
				STRLEN(full_path) + 80
			    );
			    sprintf(
				s,
"%s: Line %i: %s: Warning: Unable to open image.",
				path, lines_read, full_path
			    );
			    XSWPrintWarning(s);
			    free(s);
			}
			imglabel->image = img;

			free(full_path);
		    }

		    /* ImageUnarmedPercentCoordinates */
		    else if(!strcasecmp(parm, "ImageUnarmedPosByPercent"))
		    {
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_UNARMED].pos_by_percent = 1;
		    }
		    /* ImageUnarmedSizeByPercent */
		    else if(!strcasecmp(parm, "ImageUnarmedSizeByPercent"))
		    {
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_UNARMED].size_by_percent = 1;
		    }
		    /* ImageUnarmedX */
		    else if(!strcasecmp(parm, "ImageUnarmedX") ||
			    !strcasecmp(parm, "UnarmedImageX")
		    )
		    {
			s3 = strchr(val, '%');
			if(s3 != NULL)
			{
			    label_ptr->imglabel[
				PAGE_LABEL_STATE_UNARMED].pos_by_percent = 1;
			    *s3 = '\0';
			}
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_UNARMED].x = atoi(val);
		    }
		    /* ImageUnarmedY */
		    else if(!strcasecmp(parm, "ImageUnarmedY") ||
			    !strcasecmp(parm, "UnarmedImageY")   
		    )
		    {
			s3 = strchr(val, '%');
			if(s3 != NULL)
			{
			    label_ptr->imglabel[
				PAGE_LABEL_STATE_UNARMED].pos_by_percent =
1;
			    *s3 = '\0';
			}
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_UNARMED].y = atoi(val);
		    }

		    /* ImageArmedPosByPercent */
		    else if(!strcasecmp(parm, "ImageArmedPosByPercent"))
		    {
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_ARMED].pos_by_percent = 1;
		    }
		    /* ImageArmedSizeByPercent */
		    else if(!strcasecmp(parm, "ImageArmedSizeByPercent"))
		    {
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_ARMED].size_by_percent = 1;
		    }
		    /* ImageArmedX */
		    else if(!strcasecmp(parm, "ImageArmedX") ||
			    !strcasecmp(parm, "ArmedImageX")
		    )
		    {
			s3 = strchr(val, '%');
			if(s3 != NULL)
			{
			    label_ptr->imglabel[
				PAGE_LABEL_STATE_ARMED].pos_by_percent = 1;
			    *s3 = '\0';
			}
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_ARMED].x = atoi(val);
		    }
		    /* ImageArmedY */
		    else if(!strcasecmp(parm, "ImageArmedY") ||
			    !strcasecmp(parm, "ArmedImageY")
		    )
		    {
			s3 = strchr(val, '%');
			if(s3 != NULL)
			{
			    label_ptr->imglabel[
				PAGE_LABEL_STATE_ARMED].pos_by_percent = 1;
			    *s3 = '\0';
			}
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_ARMED].y = atoi(val);
		    }

		    /* ImageHighlightedPosByPercent */
		    else if(!strcasecmp(parm, "ImageHighlightedPosByPercent"))
		    {
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_HIGHLIGHTED].pos_by_percent = 1;
		    }
		    /* ImageHighlightedSizeByPercent */
		    else if(!strcasecmp(parm, "ImageHighlightedSizeByPercent"))
		    {
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_HIGHLIGHTED].size_by_percent = 1;
		    }
		    /* ImageHighlightedX */
		    else if(!strcasecmp(parm, "ImageHighlightedX") ||
			    !strcasecmp(parm, "HighlightedImageX")
		    )
		    {
			s3 = strchr(val, '%');
			if(s3 != NULL)
			{
			    label_ptr->imglabel[
				PAGE_LABEL_STATE_HIGHLIGHTED].pos_by_percent = 1;
			    *s3 = '\0';
			}
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_HIGHLIGHTED].x = atoi(val);
		    }
		    /* ImageHighlightedY */
		    else if(!strcasecmp(parm, "ImageHighlightedY") ||
			    !strcasecmp(parm, "HighlightedImageY")
		    )
		    {
			s3 = strchr(val, '%');   
			if(s3 != NULL)
			{
			    label_ptr->imglabel[
				PAGE_LABEL_STATE_HIGHLIGHTED].pos_by_percent = 1;
			    *s3 = '\0';
			}
			label_ptr->imglabel[
			    PAGE_LABEL_STATE_HIGHLIGHTED].y = atoi(val);
		    }

		    /* ********************************************* */
		    /* AllowTransparency */
		    else if(!strcasecmp(parm, "AllowTransparency"))
		    {
			label_ptr->allow_transparency = 1;
		    }
		    /* Action */
		    else if(!strcasecmp(parm, "Action"))
		    {
			label_ptr->op_code = atoi(val);
		    }
		    /* HintMessage */
		    else if(!strcasecmp(parm, "HintMessage") ||
			    !strcasecmp(parm, "HintMesg")
		    )
		    {
			free(label_ptr->hint_mesg);
			label_ptr->hint_mesg = STRDUP(val);
		    }


		    /* EndLabel */
		    else if(!strcasecmp(parm, "EndLabel"))
		    {
			break;
		    }
		    /* Unsupported parameter */
		    else   
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }
	    /* Unsupported parameter */
	    else
	    {
		WARN_NO_SUCH_PARM(parm);
	    }
	}
	free(s);

#undef WARN_NO_SUCH_PARM
#undef WARN_NO_SUCH_FILE

	/* Close the file */
	fclose(fp);

	return(0);
}
