#include <stdio.h>
#include <stdlib.h>

#include "../include/reality.h"
#include "particle.h"


int ParticleCreate(
	xsw_particle_struct ***list, int *total,
	int type,
	int imgref_num,
	long sect_x, long sect_y, long sect_z,  
	float x, float y, float z,            
	float heading, float pitch, float bank,
	long size_start, long size_end,
	time_t birth_time,	/* In ms */    
	time_t life_span	/* In ms */
);
static void ParticleDelete(xsw_particle_struct *p);
void ParticleClear(xsw_particle_struct ***list, int *total);

void ParticleUpdateAll(
	xsw_particle_struct ***list, int *total,
	const time_t cur_ms, const time_t cur_sec, float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a new particle.
 */
int ParticleCreate(
	xsw_particle_struct ***list, int *total,
	int type,
	int imgref_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	long size_start, long size_end,
	time_t birth_time,	/* In ms */
	time_t life_span	/* In ms */
)
{
	int i, particle_num;
	xsw_particle_struct *p;

	if((list == NULL) || (total == NULL))
	    return(-2);

	for(i = 0; i < *total; i++)
	{                          
	    if((*list)[i] == NULL)
		break;
	}
	if(i < *total)
	{
	    particle_num = i;
	}
	else
	{
	    particle_num = MAX(*total, 0);
	    *total = particle_num + 1;
	    *list = (xsw_particle_struct **)realloc(
		*list, (*total) * sizeof(xsw_particle_struct *)
	    );
	    if(*list == NULL)
	    {
		*total = 0;
		return(-3);
	    }
	}

	(*list)[particle_num] = p = XSW_PARTICLE(calloc(
	    1, sizeof(xsw_particle_struct)
	));
	if(p == NULL)
	    return(-3);

	p->type = type;
	p->imgref_num = imgref_num;
	p->sect_x = sect_x;
	p->sect_y = sect_y;
	p->sect_z = sect_z;
	p->x = x;
	p->y = y;
	p->z = z;
	p->heading = heading;
	p->pitch = pitch;
	p->bank = bank;
	p->size_start = size_start;
	p->size_end = size_end;
	p->birth_time = birth_time;
	p->life_span = life_span;

	return(particle_num);
}

/*
 *	Deletes the particle.
 */
static void ParticleDelete(xsw_particle_struct *p)
{
	if(p == NULL)
	    return;

	free(p);
}

/*
 *	Deletes all particles in the list.
 */
void ParticleClear(xsw_particle_struct ***list, int *total)
{
	int i;

	if((list == NULL) || (total == NULL))
	    return;

	for(i = *total - 1; i >= 0; i--)
	    ParticleDelete((*list)[i]);
	free(*list);
	*list = NULL;
	*total = 0;
}


/*
 *	Updates all particles.
 */
void ParticleUpdateAll(
	xsw_particle_struct ***list, int *total,
	const time_t cur_ms, const time_t cur_sec, float tc
)
{
	int i;
	float d;
	xsw_particle_struct *p;

	if((list == NULL) || (total == NULL))
	    return;

	for(i = 0; i < *total; i++)
	{
	    p = (*list)[i];
	    if(p == NULL)
		continue;

	    /* Particle's life span has been exceeded? */
	    if(((p->birth_time + p->life_span) <= cur_ms) ||
	       (cur_ms < p->birth_time)
	    )
	    {
		ParticleDelete(p);
		(*list)[i] = NULL;
		continue;
	    }

	    /* Move by particle type */
	    switch(p->type)
	    {
	      case XSW_PARTICLE_TYPE_SMOKE:
		d = (float)((p->size_end + p->size_start) / 2l) * 2.0f *
		    ((float)SW_CYCLE_LAPSE_MS / 1000.0f) / 1000.0f;
		p->x -= d * tc;
		p->y += d * tc;
		break;

	      case XSW_PARTICLE_TYPE_EFFECT:
		break;

	      case XSW_PARTICLE_TYPE_SHOCKWAVE:
		break;
	    }
	}
}
