#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "../include/reality.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"

#include "../include/swsoundcodes.h"

#include "gctl.h"
#include "particle.h"
#include "sound.h"
#include "net.h"
#include "xsw.h"


static float REngTurnAngle(
	float theta_cur, float theta_to, float theta_limit,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

static void REngMove(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc,
	char allow_sect_change
);
static void REngHeading(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngThrottle(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void REngObjectStopped(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngThrust(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngAction(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngLocation(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngVisibility(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngShieldVisibility(
	xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngAnimation(
	xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngAutoDestruct(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void REngSpawnParticles(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

static void REngUpdateObject(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc,
	const int spawn_particles
);

int REngInit(void);
void REngManage(
	xsw_object_struct *player_obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void REngShutdown(void);


#define ASSERT_NOT_NAN(x)	(assert(!isnan(x)))

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)

#define INRANGE_INC(x,min,max)	\
(((x) > (max)) ? (max) : (((x) < (min)) ? (min) : (x)))


/*
 *	Turns the angle theta_cur towards the angle theta_to but
 *	no more than theta_limit (all units in radians).
 */
static float REngTurnAngle(
	float theta_cur, float theta_to, float theta_limit,
	const time_t cur_ms, const time_t cur_sec, const float tc
)  
{
	const float limit_tc = theta_limit * tc;
	float delta = SWMathAngleDelta(theta_to, theta_cur);

	/* Limit delta angle */
	if(delta > limit_tc)
	    delta = limit_tc;
	else if(delta < -limit_tc)
	    delta = -limit_tc;

	/* Turn current angle */
	theta_cur += delta;

	return(SANITIZERADIANS(theta_cur));
}


/*
 *	Moves the object.
 *
 *	If allow_sect_change is true then if the object's sector will be
 *	changed if the object moved out of the sector, otherwise it will
 *	warp to the other side of its current sector.
 */
static void REngMove(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc,
	char allow_sect_change  
)
{
	char ch_sect_px, ch_sect_nx, ch_sect_py, ch_sect_ny;
	float thrust_dir, thrust_mag, thrust_magc;
	float velocity_mag, velocity_magc;
	float x_vel, y_vel, z_vel;	/* Velocity delta */
	float x_velc, y_velc, z_velc;	/* Velocity delta /w time compensation */
	float x_thr, y_thr, z_thr;	/* Thrust */
	float x_thrc, y_thrc, z_thrc;	/* Thrust delta /w time compensation */
/*	float x_res, y_res, z_res;	*/	/* Resultant */
	xsw_net_parms_struct *net = &net_parms;

	/* Skip non-moving types of objects */
	if((obj->type == SW_OBJ_TYPE_STATIC) ||
	   (obj->type == SW_OBJ_TYPE_HOME) ||
	   (obj->type == SW_OBJ_TYPE_AREA) ||
	   (obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (obj->type == SW_OBJ_TYPE_ELINK)
	)
	    return;

	/* Stream weapons have location of their owner */
	if((obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)
	{
	    const int owner_obj_num = obj->owner;
	    if(!DBIsObjectGarbage(owner_obj_num))
	    {
		xsw_object_struct *owner_obj = xsw_object[owner_obj_num];
		obj->sect_x = owner_obj->sect_x;
		obj->sect_y = owner_obj->sect_y;
		obj->sect_z = owner_obj->sect_z;
		obj->x = owner_obj->x;
		obj->y = owner_obj->y;
		obj->z = owner_obj->z;
	    }
	    return;
	}

	/* Begin standard object movement */

	/* Sanitize velocity heading */
	obj->velocity_heading = SANITIZERADIANS(obj->velocity_heading);

	/* Calculate movement from velocity (momentum)
	 *
	 * Fetch velocity_mag, with compensation for lapsed time
	 */
	velocity_magc = obj->velocity * tc;
	velocity_mag = obj->velocity;

	/* Get momentum (velocity) vector compoents (with and
	 * without time compensation) from momentum unit vector
	 * compoents
	 */
	x_velc = obj->momentum_vector_compoent.i * velocity_magc;
	y_velc = obj->momentum_vector_compoent.j * velocity_magc;
	z_velc = obj->momentum_vector_compoent.k * velocity_magc;

	x_vel = obj->momentum_vector_compoent.i * velocity_mag;
	y_vel = obj->momentum_vector_compoent.j * velocity_mag;
	z_vel = obj->momentum_vector_compoent.k * velocity_mag;


	/* Calculate (not object relative) thrust direction */
	thrust_dir = SANITIZERADIANS(
	    obj->heading +
	    obj->thrust_dir +
	    (1.0f * PI)
	);

	ASSERT_NOT_NAN(thrust_dir);

	/* Calculate thrust magnitude, taking into account the
	 * object's thrust power, current power, and power
	 * purity
	 *
	 * The power coeff is in the range of 0.5 to 1.0, this is
	 * due to the fact that if there is no power the vessel cannot
	 * move and many players complain that this is unfair
	 */
	if(obj->power_max > 0.0f)
	{
	    const float power_coeff = (obj->power /
		obj->power_max / 2.0f) + 0.5f;
	    thrust_mag = obj->thrust * obj->power_purity *
		power_coeff;
	}
	else
	{
	    thrust_mag = obj->thrust * obj->power_purity;
	}

	ASSERT_NOT_NAN(thrust_mag);

	/* Calculate thrust magnitude with tc */
	thrust_magc = thrust_mag * tc;

	/* Get thrust vector compoents with and without time
	 * compensation
	 */
	x_thrc = thrust_magc * (float)sin((double)thrust_dir);
	y_thrc = thrust_magc * (float)cos((double)thrust_dir);
	z_thrc = 0.0f;

	x_thr = thrust_mag * (float)sin((double)thrust_dir);
	y_thr = thrust_mag * (float)cos((double)thrust_dir);
	z_thr = 0.0f;


	/* Move object by adding the calculated velocity and thrust
	 * vector compoents.
	 */
	ASSERT_NOT_NAN(obj->x);
	ASSERT_NOT_NAN(obj->y);
	ASSERT_NOT_NAN(obj->z);

	obj->x += x_velc + x_thrc;
	obj->y += y_velc + y_thrc;
/*      obj->z += z_velc + z_thrc; */
/* Force set z to 0 */
	obj->z = 0.0f;

	/* Do sector change as needed */
	if(allow_sect_change)
	{
	    sw_legend_struct *legend = &sw_legend;

	    ch_sect_px = 0;
	    ch_sect_nx = 0;
	    ch_sect_py = 0;
	    ch_sect_ny = 0;
/*
	    ch_sect_pz = 0;
	    ch_sect_nz = 0; 
 */

	    if(obj->x < legend->sect_x_min)
	    {
	        obj->x += legend->sect_x_len;
	        obj->sect_x -= 1;
		ch_sect_nx = 1;		/* Negative sect x change */
	    }
	    else if(obj->x > legend->sect_x_max)
	    {
		obj->x -= legend->sect_x_len;
		obj->sect_x += 1;
		ch_sect_px = 1;		/* Positive sect x change */
	    }

	    if(obj->y < legend->sect_y_min)
	    {
		obj->y += legend->sect_y_len;
		obj->sect_y -= 1;
		ch_sect_ny = 1;		/* Negative sect y change */
	    }
	    else if(obj->y > legend->sect_y_max)
	    {
		obj->y -= legend->sect_y_len;
		obj->sect_y += 1;
		ch_sect_py = 1;		/* Positive sect y change */
	    }

#if 0
	    if(obj->z < legend->sect_z_min)
	    {
		obj->z += legend->sect_z_len;
		obj->sect_z -= 1;
		ch_sect_nz = 1;		/* Negative sect z change */
	    }
	    else if(obj->z > legend->sect_z_max)
	    {
		obj->z -= legend->sect_z_len;
		obj->sect_z += 1;
		ch_sect_pz = 1;		/* Positive sect z change */
	    }
#endif

	    /* Check if there was a sector change */
	    if(ch_sect_nx || ch_sect_px ||
	       ch_sect_ny || ch_sect_py
/*
	       ch_sect_nz || ch_sect_pz
*/
	    )
	    {
		/* Update nessasary resources per sector change */
		int i, trac_obj_num;
		xsw_object_struct *trac_obj;

		/* Change sector for tractored objects */
		for(i = 0; i < obj->total_tractored_objects; i++)
		{
		    trac_obj_num = obj->tractored_object[i];
		    if(DBIsObjectGarbage(trac_obj_num))
			continue;

		    trac_obj = xsw_object[trac_obj_num];

		    trac_obj->sect_x = obj->sect_x;
		    trac_obj->sect_y = obj->sect_y;
/*                  trac_obj->sect_z = obj->sect_z; */

		    if(ch_sect_nx)
			trac_obj->x += legend->sect_x_len;
		    if(ch_sect_px)
			trac_obj->x -= legend->sect_x_len;

		    if(ch_sect_ny)
			trac_obj->y += legend->sect_y_len;
		    if(ch_sect_py)
			trac_obj->y -= legend->sect_y_len;
/*
		    if(ch_sect_nz)
			trac_obj->z += legend->sect_z_len;
		    if(ch_sect_pz)
			trac_obj->z -= legend->sect_z_len;
*/
		}

		/* Check if this is the player object */
		if(obj == net->player_obj_ptr)
		{
		    /* Send out sector position to server */
		    const int obj_num = net->player_obj_num;
		    NetSendObjSect(obj_num);
		    XSWObjectSectorChangedNotify(
			obj_num,
			cur_ms, cur_sec, tc
		    );
		    if(!XSW_CON_IS_CONNECTED(net))
			return;
		}
	    }
	}


	/*
	 *   Record new velocity and velocity_heading on object.
	 *   This value must not have the time compensation modifications
	 *   to it.
	 */     
	obj->velocity_heading = SWMathCoordinateDeltaVector(
	    x_vel + x_thr,
	    y_vel + y_thr
	);
	SWMathSetUnitVector2D(
	    &obj->momentum_vector_compoent,
	    obj->velocity_heading
	);

	ASSERT_NOT_NAN(obj->velocity_heading);

	obj->velocity = SWMath3DDistance(
	    x_vel + x_thr,
	    y_vel + y_thr,
	    0
	);

	ASSERT_NOT_NAN(obj->velocity);
}

/*
 *	Updates the object's heading.
 */
static void REngHeading(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_gctl_struct *gctl = xsw_gctl;
	xsw_net_parms_struct *net = &net_parms;

	/* Skip static objects */
	if(obj->type == SW_OBJ_TYPE_STATIC)
	   return;

	/* Update heading depending on object type */
	/* Stream weapons */
	if(obj->type == SW_OBJ_TYPE_STREAMWEAPON)
	{
	    /* Update heading for stream weapons so that they fire
	     * more realistically
	     */
	    const int tar_obj_num = obj->intercepting_object;
	    if(!DBIsObjectGarbage(tar_obj_num))
	    {
		/* Calculate the bearing to the object that this object
		 * is intercepting
		 */
		float bearing;
		xsw_object_struct *tar_obj = xsw_object[tar_obj_num];

		/* In same sector? */
		if(SWMath3DInSameSectorPtr(tar_obj, obj))
		    bearing = SWMathCoordinateDeltaVector(
			tar_obj->x - obj->x,
			tar_obj->y - obj->y
		    );
		else
		    bearing = SWMathCoordinateDeltaVector(
			tar_obj->sect_x - obj->sect_x,
			tar_obj->sect_y - obj->sect_y
		    );

		/* Turn heading */
		obj->heading = REngTurnAngle(
		    obj->heading, bearing,
		    100.0f * obj->turnrate,
		    cur_ms, cur_sec, tc
		);
	    }
	}
	/* Sphere weapons do not change heading */
	else if(obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	{


	}
	/* Player object */
	else if((obj == player_obj) && (gctl != NULL))
	{
	    /* Change heading only if not in omni directional thrust
	     * mode
	     */
	    if(!gctl->omni_dir_thrust)
	    {
		float delta = gctl->turn;

	        /* Adjust heading according to controller position */
	        obj->heading += delta * 100.0f * obj->turnrate * tc;

	        /* Switch off intercept if turn is too great */
	        if((delta > 0.3f) || (delta < -0.3f))
	        {
		    if(obj->intercepting_object > -1)
		    {
		        NetSendIntercept(
			    net->player_obj_num,
			    "#off"
			);
			if(!XSW_CON_IS_CONNECTED(net))
			    return;
		    }
		}

		obj->heading = SANITIZERADIANS(obj->heading);
	    }
	}

	/* Update attitude vector */
	SWMathSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);
}


/*
 *	Updates the throttle by checking the game controller
 *	positions.
 */
static void REngThrottle(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_gctl_struct *gctl = xsw_gctl;

	/* Skip static objects */
	if(obj->type == SW_OBJ_TYPE_STATIC)
	   return;

	/* Update heading for local player object */
	if((obj == player_obj) && (gctl != NULL))
	{
	    /* Check if omni directional thrust is on */
	    if(gctl->omni_dir_thrust)
	    {
		/* Omni directional thrust is on */

		switch(option.throttle_mode)
		{
		  case XSW_THROTTLE_MODE_INCREMENTAL:
		    /* Not sure of the best way to do this yet.. */
		    break;

		  /* XSW_THROTTLE_MODE_BIDIRECTIONAL or XSW_THROTTLE_MODE_NORMAL */
		  default:
		    obj->thrust_dir = SANITIZERADIANS(
		        (0.5 * PI) - atan2(gctl->throttle, gctl->turn) +
			(1.0 * PI)
		    );
		    obj->throttle = INRANGE_INC(
		        hypot(gctl->turn, gctl->throttle), 0.0, 1.0
		    );
		    break;
		}
	    }
	    else
	    {
		/* Omni directional thrust is off */

		switch(option.throttle_mode)
		{
		  case XSW_THROTTLE_MODE_INCREMENTAL:
		    if((obj->thrust_dir > (PI * 0.5)) &&
		       (obj->thrust_dir < (PI * 1.5))
		    )
		    {
			/* Thrusters going backwards (normal) */
			obj->throttle += (gctl->throttle
			    * 0.01f * tc
		        );

			obj->thrust_dir = SANITIZERADIANS(
			    ((0.5f * PI) * gctl->thrust_dir) + PI
			);

			/* Adjust throttle and set thrust dir */
			if(obj->throttle < 0.0f)
			{
			    obj->throttle *= -1.0f;
			    obj->thrust_dir = SANITIZERADIANS(   
				obj->thrust_dir + PI 
			    );
			}
		    }
		    else
		    {
			/* Thrusters going forwards (reverse) */
			obj->throttle += (gctl->throttle
			    * -0.01f * tc
			);

			obj->thrust_dir = SANITIZERADIANS(
			    ((0.5f * PI) * gctl->thrust_dir)
			);

			/* Adjust throttle and set thrust dir */
			if(obj->throttle < 0.0f)
			{
			    obj->throttle *= -1.0f;
			    obj->thrust_dir = SANITIZERADIANS(
				obj->thrust_dir + PI
			    );
			}
		    }
		    break;

		  /* XSW_THROTTLE_MODE_BIDIRECTIONAL or XSW_THROTTLE_MODE_NORMAL */
		  default:
		    obj->throttle = gctl->throttle;
		    if(obj->throttle < 0)
		    {
			obj->thrust_dir = SANITIZERADIANS(
			    ((0.5 * PI) * gctl->thrust_dir)
			);
			obj->throttle *= -1;
		    }
		    else
		    {
			obj->thrust_dir = SANITIZERADIANS(
			    ((0.5 * PI) * gctl->thrust_dir) + PI
			);
		    }
		    break;
		}
	    }
	}

	/* Sanitize throttle */
	if(obj->throttle < 0.0f)
	    obj->throttle = 0.0f;
	else if(obj->throttle > 1.0f)
	    obj->throttle = 1.0f;
}

/*
 *	Object stopped procedure.
 *
 *	Should be called whenever an object has just stopped from
 *	being in motion.
 */
void REngObjectStopped(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_option_struct *opt = &option;
	xsw_net_parms_struct *net = &net_parms;

	if(obj == player_obj)
	{
	    /* Player has just stopped moving */
	    int i, n;
	    xsw_object_struct *tar_obj;

	    /* Find nearest the HOME object in inrange objects
	     * list
	     */
	    for(i = 0; i < total_inrange_objects; i++)
	    {
		tar_obj = inrange_xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		if((tar_obj->type != SW_OBJ_TYPE_HOME) &&
		   (tar_obj->type != SW_OBJ_TYPE_WORMHOLE) &&
		   (tar_obj->type != SW_OBJ_TYPE_ELINK)
		)
		    continue;

		if(SWMath3DInContactPtr(obj, tar_obj))
		    break;
	    }
	    if(i < total_inrange_objects)
	    {
		/* Got nearest HOME, WORMHOLE, or ELINK object */
		tar_obj = inrange_xsw_object[i];
		n = DBGetObjectNumByPtr(tar_obj);
		if(n > -1)
		{
		    switch(tar_obj->type)
		    {
		      case SW_OBJ_TYPE_HOME:
			if(1)
			{
			    const char *name = (opt->show_formal_label == XSW_SHOW_FORMAL_LABEL_ALWAYS) ?
				DBGetFormalNameStr(n) : tar_obj->name;
			    char *s = (char *)malloc(
			        STRLEN(name) + 80
			    );
			    sprintf(
				s,
"Orbiting `%s'.",
				name
			    );
			    XSWPrint(s);
			    free(s);
			}
			/* Map economy window? */
			if(option.auto_map_eco_win)
			{
			    /* Map by requesting economy values */
			    NetSendReqName(n);
			    NetSendEcoReqValues(n);
			    if(!XSW_CON_IS_CONNECTED(net))
				return;
			}
			break;
		      case SW_OBJ_TYPE_WORMHOLE:
			if(next.wormhole_enter <= cur_ms)
			{
			    NetSendWormHoleEnter(
				net->player_obj_num,
				n
			    );
			    if(!XSW_CON_IS_CONNECTED(net))
				return;
			    next.wormhole_enter = cur_ms +
				SW_WORMHOLE_ENTER_DELAY;
			}
			break;
		      case SW_OBJ_TYPE_ELINK:
			NetSendELinkEnter(net->player_obj_num, n);
			if(!XSW_CON_IS_CONNECTED(net))
			    return;
			break;
		    }
		}
	    }
	}
}

/*
 *	Thrust, antimatter consumption, drag, and other calculations
 *	of velocity are performed here.
 */
static void REngThrust(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char just_stopped = 0;
	float throttle;
	sw_legend_struct *legend = &sw_legend;
	xsw_gctl_struct *gctl = xsw_gctl;

	/* Skip objects that do not move */
	if((obj->type == SW_OBJ_TYPE_STATIC) ||
	   (obj->type == SW_OBJ_TYPE_HOME) ||
	   (obj->type == SW_OBJ_TYPE_AREA) ||
	   (obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (obj->type == SW_OBJ_TYPE_ELINK)
	)
	   return;

	/* Get throttle */

	/* Check if object is the player object */	
	if(player_obj == obj)
	{
	    /* Are engines off? */
	    if(obj->engine_state == SW_ENGINE_STATE_ON)
		throttle = obj->throttle;
	    else
	        throttle = 0.0f;
	}
	else
	{
	    /* Not player object */

	    /* Fetch throttle based on object type */
	    switch(obj->type)
	    {
	      /* Projectile weapons always throttle at full */
	      case SW_OBJ_TYPE_WEAPON:
		throttle = 1.0f;
		break;

	      /* Stream weapon have no throttle value */
	      case SW_OBJ_TYPE_STREAMWEAPON:
		throttle = 0.0f;
		break;

	      /* Same as stream weapon */
	      case SW_OBJ_TYPE_SPHEREWEAPON:
		throttle = 0.0f;
		break;

	      /* All else get exact value */
	      default:
		throttle = obj->throttle;
		break;
	    }
	}

	/* Thrust */

	/* Thrust is calculated by taking the coefficent of the
	 * throttle and multiplying that with the thrust_power and
	 * including the time compensation (to compensate for velocity
	 * drag which has time compensation added to it)
	 */
	obj->thrust = MAX(
	    throttle * obj->thrust_power * tc,
	    0.0f
	);

	/* If there is no antimatter left then reduce the thrust
	 * output
	 */
	if(obj->antimatter <= 0.0f)
	    obj->thrust *= legend->no_antimatter_thrust_coeff;


	/* Drag */

	/* Is this the player object? */
	if((player_obj == obj) && (gctl != NULL))
	{
	    if(obj->velocity > 0.0f)
	    {
		if((obj->velocity_max > 0.0f) && (obj->thrust_power > 0.0f))
		{
		    /* Apply retro thrusters to the velocity? */
		    if(gctl->external_dampers)
			obj->velocity = MAX(
			    (obj->velocity - ((obj->thrust_power / 4) *
				obj->power_purity * tc)
			    ),
			    0.0f
			);

		    /* Apply normal drag to the velocity last */
		    obj->velocity = MAX(
			(obj->velocity - (obj->thrust_power *
			    (obj->velocity / obj->velocity_max) * tc)
			),
			0.0f
		    );
		}
		else
		{
		    obj->velocity = 0.0f;
		}
		if(obj->velocity <= 0.0f)
		{
		    obj->velocity = 0.0f;
		    just_stopped = 1;
		}
	    }
	}
	else
	{
	    /* Other object */
	    if(obj->velocity > 0.0f)
	    {
	        if((obj->velocity_max > 0.0f) && (obj->thrust_power > 0.0f))
 	            obj->velocity = MAX(
			(obj->velocity - (obj->thrust_power *
			    (obj->velocity / obj->velocity_max) * tc)
			),
			0.0f
		    );
	        else
	            obj->velocity = 0.0f;

		if(obj->velocity <= 0.0f)
		    obj->velocity = 0.0f;
	    }
	}

	/* Object just stopped? */
	if(just_stopped)
	    REngObjectStopped(player_obj, obj, cur_ms, cur_sec, tc);
}


/*
 *	Performs action as needed on object by it's parameter's
 *	values.
 *
 *	Weapons fire firing is handled here.
 */
static void REngAction(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int sel_wep_num;
	time_t dt;
	osw_gui_struct *gui = &osw_gui[0];
	xsw_weapon_struct *wep;
	xsw_gctl_struct *gctl = xsw_gctl;
	xsw_player_values_struct *pvals = &player_values;
	xsw_sound_struct *snd = &sound;
	xsw_net_parms_struct *net = &net_parms;

	/* Check if this is the player object */
	if((obj == player_obj) && (gctl != NULL))
	{
	    /* ***************************************************** */
	    /* Weapons Fire */
	    while(gctl->fire_weapon)
	    {
		/* Get pointer to selected weapon */
		sel_wep_num = obj->selected_weapon;
		if((sel_wep_num < 0) || (sel_wep_num >= obj->total_weapons))
		    break;

		wep = obj->weapons[sel_wep_num];
		if(wep == NULL)
		    break;

		/* Are the weapons offline? */
		if(!pvals->weapons_online)
		{
		    /* Weapons are offline, still count this as a fire
		     * by updating the time it was last used and play
		     * error sound
		     */
		    dt = wep->last_used + MIN(wep->delay, 1000l);
		    if(dt <= cur_ms)
		    {
			BridgeWarnWeaponsOffline(
			    xsw_bridge_win,
			    net->player_obj_num
			);
			wep->last_used = cur_ms;
		    }
		    break;
		}

		/* Check if we've waited the weapon's delay period.
		 * Note that since this is the client, we limit our
		 * delay check to be at most 1000 ms from the last
		 * time it was used.
		 */
		dt = wep->last_used + MIN(wep->delay, 1000l);
		if(dt >= cur_ms)
		    break;

		/* Out of stock? */
		if(wep->emission_type == SW_WEP_EMISSION_PROJECTILE)
		{
		    if((wep->max > 0) && (wep->amount == 0))
		    {
			char *s = (char *)malloc(STRLEN(wep->name) + 80);
			sprintf(
			    s,
"You do not have any `%ss' left.",
			    wep->name
			);
			XSWPrintWarning(s);
			free(s);
			if(option.sounds >= XSW_SOUNDS_EVENTS)
			    SoundPlay(
				snd, SW_SNDREF_STD_ERROR,
				1.0f, 1.0f,
				0, 0
			    );
		    }
		}

		/* Send the fire weapon command to server */
		NetSendFireWeapon(
		    net->player_obj_num,
		    pvals->weapon_freq
		);
		if(!XSW_CON_IS_CONNECTED(net))
		    return;

		/* Mark last millitime this weapon was used.
		 * The last used time sent by the server is ignored and
		 * the client keeps track of the last used time.  Since
		 * the firing times and weapon delay times are synced
		 * (or atleast reasonably synced) there shouldn't be any
		 * problem.
		 */
		wep->last_used = cur_ms;

		break;
	    }

	    /* ************************************************************ */
	    /* Aim weapon heading */

	    /* Get pointer to selected weapon */
	    sel_wep_num = obj->selected_weapon;
	    wep = NULL;
	    if((sel_wep_num >= 0) && (sel_wep_num < obj->total_weapons))
		wep = obj->weapons[sel_wep_num];

	    if(wep != NULL)
	    {
		if(wep->flags & SW_WEP_FLAG_FIXED)
		{
		    pvals->weapon_fire_heading = 0.0f * PI;
		}
		else if(gctl->pks_throttle_idle && (gui->key_state & OSW_KEY_STATE_SHIFT))
		{
		    pvals->weapon_fire_heading = 0.0f * PI;
/*		    pvals->weapon_fire_heading = wep->origin_heading; */
		}
		else if(gctl->pks_throttle_max && (gui->key_state & OSW_KEY_STATE_SHIFT))
		{
		    pvals->weapon_fire_heading = 1.0f * PI;
/*		    pvals->weapon_fire_heading = SANITIZERADIANS(
			wep->origin_heading + (1.0f * PI)
		    );
 */
		}
		else
		{
		    pvals->weapon_fire_heading = SANITIZERADIANS(
			pvals->weapon_fire_heading +
			(gctl->aim_weapon_heading *
			    (2 * PI / 180) *
			tc)
		    );
		}
	    }
	    else
	    {
		pvals->weapon_fire_heading = 0.0f * PI;
	    }
	}
}


/*
 *	Updates the object's location type.
 */
static void REngLocation(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, area_obj_num = -1;
	float d = 0.0f;			/* In screen units */
	xsw_object_struct	*tar_obj,
				*area_obj = NULL;
	xsw_option_struct *opt = &option;

	/* Skip area objects */
	if(obj->type == SW_OBJ_TYPE_AREA)
	    return;

	/* Iterate through the inrange objects list backwards and
	 * check all non-territory area objects first
	 */
	for(i = total_inrange_objects - 1; i >= 0; i--)
	{
	    tar_obj = inrange_xsw_object[i];
	    if(tar_obj == NULL)
		continue;

	    /* Not an area object? */
	    if(tar_obj->type != SW_OBJ_TYPE_AREA)
		continue;

	    /* Skip territory area objects */
	    if(tar_obj->loc_type == SW_LOC_TYPE_SPACE_TERRITORY)
		continue;

	    /* Skip itself or the player */
	    if((tar_obj == obj) || (tar_obj == player_obj))
		continue;

	    /* Not in same sector? */
	    if((obj->sect_x != tar_obj->sect_x) ||
	       (obj->sect_y != tar_obj->sect_y) ||
	       (obj->sect_z != tar_obj->sect_z)
	    )
		continue;

	    /* Calculate distance in screen units */
	    d = SWMath3DDistance(
		tar_obj->x - obj->x,
		tar_obj->y - obj->y,
		tar_obj->z - obj->z
	    ) * 1000.0f;
	    if(d > tar_obj->size)
		continue;

	    area_obj = tar_obj;
	    area_obj_num = DBGetObjectNumByPtr(area_obj);
	    break;
	}

	/* If the object was not in a non-territory area object then
	 * check if it is in a territory area object
	 */
	if(area_obj == NULL)
	{
	    /* Iterate through the inrange objects list backwards */
	    for(i = total_inrange_objects - 1; i >= 0; i--)
	    {
		tar_obj = inrange_xsw_object[i];
		if(tar_obj == NULL)
		    continue;

		/* Not an area object? */
		if(tar_obj->type != SW_OBJ_TYPE_AREA)
		    continue;

		/* Not a territory area object? */
		if(tar_obj->loc_type != SW_LOC_TYPE_SPACE_TERRITORY)
		    continue;

		/* Skip itself or the player */
		if((tar_obj == obj) || (tar_obj == player_obj))
		    continue;

		/* Not in same sector? */
		if((obj->sect_x != tar_obj->sect_x) ||
		   (obj->sect_y != tar_obj->sect_y) ||
		   (obj->sect_z != tar_obj->sect_z)
		)
		    continue;

		/* Calculate distance in screen units */
		d = SWMath3DDistance(
		    tar_obj->x - obj->x,
		    tar_obj->y - obj->y,
		    tar_obj->z - obj->z
		) * 1000.0f;
		if(d > tar_obj->size)
		    continue;

		area_obj = tar_obj;
		area_obj_num = DBGetObjectNumByPtr(area_obj);
		break;
	    }
	}

	/* Is the object now in normal space? */
	if(area_obj == NULL)
	{
	    /* Was the object already in an area object? */
	    if((obj->loc_type != SW_LOC_TYPE_SPACE) ||
	       (obj->loc_obj_num > -1)
	    )
	    {
		const int	obj_num = DBGetObjectNumByPtr(obj),
				prev_loc_type = obj->loc_type,
				prev_area_obj_num = obj->loc_obj_num;
		xsw_object_struct *prev_area_obj = DBIsObjectGarbage(prev_area_obj_num) ?
		    NULL : xsw_object[prev_area_obj_num];

		obj->loc_type = SW_LOC_TYPE_SPACE;
		obj->loc_obj_num = -1;

		/* Is this the player? */
		if(player_obj == obj)
	        {
		    char *s;
		    const char *name = (prev_area_obj != NULL) ?
			((opt->show_formal_label == XSW_SHOW_FORMAL_LABEL_ALWAYS) ?
			    DBGetFormalNameStr(prev_area_obj_num) : prev_area_obj->name) : NULL;

		    switch(prev_loc_type)
		    {
		      case SW_LOC_TYPE_SPACE_TERRITORY:
			s = (char *)malloc(STRLEN(name) + 80);
			if(!STRISEMPTY(name))
			    sprintf(
				s,
				"Leaving `%s'.",
				name
			    );
			else
			    strcpy(s, "Leaving territory.");
			XSWPrint(s);
			free(s);
			break;
		      case SW_LOC_TYPE_NEBULA:
			s = (char *)malloc(STRLEN(name) + 80);
			if(!STRISEMPTY(name))
			    sprintf(
				s,
				"Leaving `%s'.",                
				name
			    );
			else
			    strcpy(s, "Leaving nebula.");
			XSWPrint(s);
			free(s);
			break;
		      case SW_LOC_TYPE_ASTEROID_FIELD:
			s = (char *)malloc(STRLEN(name) + 80);
			if(!STRISEMPTY(name))
			    sprintf(
				s,
				"Leaving `%s'.",                
				name
			    );
			else
			    strcpy(s, "Leaving asteroid field.");
			XSWPrint(s);
			free(s);
			break;
		      case SW_LOC_TYPE_PULSAR:
			s = (char *)malloc(STRLEN(name) + 80);
			if(!STRISEMPTY(name))
			    sprintf(
				s,
				"Leaving `%s'.",                
				name
			    );
			else
			    strcpy(s, "Leaving pulsar.");
			XSWPrint(s);
			free(s);
			break;
 		    }

		    /* Redraw the bridge panels to reflect the object's
		     * change in location
		     */
		    BridgeWinDrawPanel(obj_num, BW_PANEL_P2);

		    /* Recreate scanner location type label */
		    ScannerUpdateLabels(xsw_bridge_win, obj_num);

		    /* Update the background music as needed */
		    if(option.music)
		        XSWUpdateBackgroundMusic();
	        }
		/* Is this the object that the player is locked on to? */
		else if((player_obj != NULL) ?
		    (player_obj->locked_on == obj_num) : 0
		)
		{
		    /* Redraw the bridge panels to reflect the object's
		     * change in location
		     */
		    BridgeWinDrawPanel(obj_num, BW_PANEL_S1);
		}
	    }
	}
	else
	{
	    /* Was the object in a different area object? */
	    if((obj->loc_type != area_obj->loc_type) ||
	       (obj->loc_obj_num != area_obj_num)
	    )
	    {
		const int obj_num = DBGetObjectNumByPtr(obj);

		obj->loc_type = area_obj->loc_type;
		obj->loc_obj_num = area_obj_num;

		/* Is this the player? */
		if(player_obj == obj)
	        {
		    sw_legend_struct *legend = &sw_legend;
		    sw_units_struct units;
		    const int n = DBGetObjectNumByPtr(area_obj);
		    const char *name = (opt->show_formal_label == XSW_SHOW_FORMAL_LABEL_ALWAYS) ?
			DBGetFormalNameStr(n) : area_obj->name;
		    const char *units_str;
		    char *s;
		    float area_velocity_max;

		    units.ru_to_au = legend->ru_to_au;
		    switch(option.units)
		    {
		      case XSW_UNITS_METRIC:
			area_velocity_max = SWConvertRUPerCycleToAUPerSecond(
			    &units, area_obj->velocity_max
			);
			units_str = "au/s";
			break;
		      case XSW_UNITS_ENGLISH:
			area_velocity_max = SWConvertRUPerCycleToAUPerSecond(
			    &units, area_obj->velocity_max
			);
			units_str = "au/s";
			break;
		      case XSW_UNITS_PARSECS:
			area_velocity_max = SWConvertRUPerCycleToMPPerSecond(
			    &units, area_obj->velocity_max
			);
			units_str = "mp/s";
			break;
		      default:
			area_velocity_max = SWConvertRUPerCycleToRUPerSecond(
			    &units, area_obj->velocity_max
			);
			units_str = "ru/s";
			break;
		    }

		    switch(area_obj->loc_type)
		    {
		      case SW_LOC_TYPE_SPACE_TERRITORY:
			s = (char *)malloc(STRLEN(name) + STRLEN(area_obj->empire) + 80);
			sprintf(
			    s,
"Entering `%s', occupied by the %s empire.",
			    name, area_obj->empire
			);
			XSWPrint(s);
			free(s);
			break;
		      case SW_LOC_TYPE_NEBULA:
			s = (char *)malloc(STRLEN(name) + 80 + 40);
			sprintf(
			    s,
"Entering `%s', visibility reduced to %i%%.",
			    name,
			    (int)(SW_VISIBILITY_NEBULA * 100.0f)
			);
		        XSWPrint(s);
			free(s);
			break;
		      case SW_LOC_TYPE_ASTEROID_FIELD:
			s = (char *)malloc(STRLEN(name) + 128 + 40 + 40 + STRLEN(units_str));
			sprintf(
			    s,
"Entering `%s', visibility reduced to %i%% and damage above speeds %.2f %s.",
			    name,
			    (int)(SW_VISIBILITY_ASTEROID_FIELD * 100.0),
			    area_velocity_max,
			    units_str
			);
			XSWPrintBold(s);
			free(s);
			break;
		      case SW_LOC_TYPE_PULSAR:
			s = (char *)malloc(STRLEN(name) + 128);
			sprintf(
			    s,
"Entering `%s', all cloaked and low visibility objects will be revealed.",
			    name
			);
			XSWPrintBold(s);
			free(s);
			break;
		    }

		    /* Redraw the bridge panels to reflect the object's
		     * change in location
		     */
		    BridgeWinDrawPanel(obj_num, BW_PANEL_P2);

		    /* Recreate scanner location type label */
		    ScannerUpdateLabels(xsw_bridge_win, obj_num);

		    /* Update the background music as needed */
		    if(option.music)
			XSWUpdateBackgroundMusic();
		}
		/* Is this the object that the player is locked on to? */
		else if((player_obj != NULL) ?
		    (player_obj->locked_on == obj_num) : 0
		)
		{
		    /* Redraw the bridge panels to reflect the object's
		     * change in location
		     */
		    BridgeWinDrawPanel(obj_num, BW_PANEL_S1);
		}
	    }
	}
}


/*
 *	Updates the object's visiblity.
 */
static void REngVisibility(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_net_parms_struct *net = &net_parms;

	/* Are the shields up? */
	if(obj->shield_state == SW_SHIELD_STATE_UP)
	{
	    if(obj->cur_visibility < 1.0f)
	    {
		obj->cur_visibility = MIN(
		    obj->cur_visibility + (0.01f * tc),
		    1.0f
		);
		if(obj == player_obj)
		    BridgeWinDrawPanel(
			net->player_obj_num,
			BW_PANEL_PVIS
		    );
	    }
	}
	/* Is the cloak up? */
	else if(obj->cloak_state == SW_CLOAK_STATE_UP)
	{
	    float v = (obj->power_max > 0.0f) ?
		obj->visibility - (obj->cloak_strength *
		    obj->power_purity * obj->power /
		    obj->power_max) :
		obj->visibility;

	    if(obj->cur_visibility > v)   
	    {
		obj->cur_visibility = MAX(
		    obj->cur_visibility - (0.01f * tc),
		    v
		);
		if(obj == player_obj)
		    BridgeWinDrawPanel(   
			net->player_obj_num,
			BW_PANEL_PVIS
		    );
	    }
	}
	/* All else adjust the visibility to normal */
	else
	{
	    if(obj->cur_visibility > obj->visibility)
	    {
		obj->cur_visibility = MAX(
		    obj->cur_visibility - (0.01f * tc),
		    obj->visibility
		);

		if(obj == player_obj)
		    BridgeWinDrawPanel(
			net->player_obj_num,
			BW_PANEL_PVIS
		    );
	    }
	    else if(obj->cur_visibility < obj->visibility)
	    {
		obj->cur_visibility = MIN(
		    obj->cur_visibility + (0.01f * tc),
		    obj->visibility
		);

		if(obj == player_obj)
		    BridgeWinDrawPanel(
			net->player_obj_num,
			BW_PANEL_PVIS
		    );
	    }
	}

	/* Clip */
	if(obj->cur_visibility < 0.0f)
	    obj->cur_visibility = 0.0f;
	else if(obj->cur_visibility > 1.0f)
	     obj->cur_visibility = 1.0f;
}


/*
 *	Updates the object's shield visibility.
 */
static void REngShieldVisibility(
	xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	/* Reduce shield visibility */
	obj->shield_visibility = MAX(
	    obj->shield_visibility - (
		((float)SW_CYCLE_LAPSE_MS /
		(float)SW_SHIELD_VISIBILITY_INTERVAL) * tc
	    ), 0.0f
	);
}


/*
 *	Increments animation frames on object.
 */
static void REngAnimation(
	xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_animation_struct *anim = &obj->animation;

	/* Handle infinate and finate repeating animations differently */
	if(anim->cycle_times <= 0)
	{
	    /* Repeats infinatly */
	    const time_t next_inc = anim->last_interval + anim->interval;

	    /* Time to increment frame? */
	    if(next_inc <= cur_ms)
	    {
		anim->current_frame++;

		/* Looped? */
		if(anim->current_frame >= anim->total_frames)
		    anim->current_frame = 0;

		anim->last_interval = cur_ms;
	    }
	}
	else
	{
	    /* Finite repeats */

	    /* Still has repeats to go? */
	    if(anim->cycle_count < anim->cycle_times)
	    {
		const time_t next_inc = anim->last_interval + anim->interval;

		/* Time to increment frame? */
		if(next_inc <= cur_ms)
		{
		    anim->current_frame++;

		    /* Looped? */
		    if(anim->current_frame >= anim->total_frames)
		    {
			anim->current_frame = 0;
			anim->cycle_count++;

			/* Cycled to end? */
			if(anim->cycle_count >= anim->cycle_times)
			{
			    /* Set current frame to the last frame */
			    anim->current_frame = anim->total_frames - 1;
			    if(anim->current_frame < 0)
				anim->current_frame = 0;

			    /* Frame will no longer be incremented */
			}
		    }

		    anim->last_interval = cur_ms;

		}
	    }
	}
}

/*
 *	Monitors auto destruct on the object.
 */
static void REngAutoDestruct(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	/* Skip if this is not the player */
	if(obj != player_obj)
	    return;

	/* Auto destruct schedualed? */
	if(obj->auto_destruct_next > 0l)
	{
	    xsw_player_values_struct *pvals = &player_values;

	    if(obj->auto_destruct_next > cur_sec)
	    {
		const time_t sec_left = obj->auto_destruct_next -
		    cur_sec;

		/* Time to warn? */
		if((sec_left < pvals->auto_destruct_last_sec_left) ||
		   (pvals->auto_destruct_last_sec_left == 0l)
		)
		{
		    if(sec_left <= 5)
		    {
			char s[80];
			sprintf(
			    s,
"Auto-destruct in %i seconds...",
			    (int)sec_left
			);
			XSWPrintWarning(s);
			if((sec_left == 5) &&
			   (option.sounds >= XSW_SOUNDS_EVENTS)
			)
			    SoundPlay(
				&sound, SW_SNDREF_EVENT_WARNING,
				1.0f, 1.0f,
				0, 0
			    );
		    }
		    else if((sec_left == 30) || (sec_left == 15) ||
			    (sec_left <= 10)
		    )
		    {
			char s[80];
			sprintf(
			    s,
"Auto-destruct in %i seconds...",
			    (int)sec_left
			);
			XSWPrint(s);
			if((sec_left == 10) &&
			   (option.sounds >= XSW_SOUNDS_EVENTS)
			)
			    SoundPlay(
				&sound, SW_SNDREF_EVENT_WARNING,
				1.0f, 1.0f,
				0, 0
			    );
		    }
		}

		pvals->auto_destruct_last_sec_left = sec_left;
	    }
	}
}


/*
 *	Spawns particles.
 *
 *	Calls to this function must be timmed properly.
 */
static void REngSpawnParticles(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_option_struct *opt = &option;

	/* Fires? */
	if((obj->ext_flags & XSW_OBJ_FLAG_ON_FIRE) &&
	   opt->show_fires
	)
	{
	    int imgref_num;
	    long particle_size;

	    if(obj->size > 90l)
	    {
		imgref_num = IMGREF_FIRE_90;
		particle_size = 45;
	    }
	    else if(obj->size > 60l)
	    {
		imgref_num = IMGREF_FIRE_60;
		particle_size = 30l;
	    }
	    else
	    {
		imgref_num = IMGREF_FIRE_30;
		particle_size = 15l;
	    }

	    ParticleCreate(
		&xsw_particle, &total_particles,
		XSW_PARTICLE_TYPE_SMOKE,
		imgref_num,
		obj->sect_x, obj->sect_y, obj->sect_z,
		obj->x, obj->y, obj->z,
		obj->heading, obj->pitch, obj->bank,
		particle_size, particle_size,
		cur_ms, XSW_PARTICLE_LIFE_SPAN
	    );
	}

	/* Smoke trails? */
	if((obj->ext_flags & XSW_OBJ_FLAG_SMOKE_TRAIL) &&
	   opt->show_smoke_trails
	)
	{
	    int imgref_num;
	    long particle_size;

	    if(obj->size > 90l)
	    {
		imgref_num = IMGREF_SMOKE_90;
		particle_size = 45;
	    }
	    else if(obj->size > 60l)
	    {
		imgref_num = IMGREF_SMOKE_60;
		particle_size = 30l;
	    }
	    else
	    {   
		imgref_num = IMGREF_SMOKE_30;
		particle_size = 15l;
	    }

	    ParticleCreate(
		&xsw_particle, &total_particles,
		XSW_PARTICLE_TYPE_SMOKE,
		imgref_num,
		obj->sect_x, obj->sect_y, obj->sect_z,
		obj->x, obj->y, obj->z,
		obj->heading, obj->pitch, obj->bank,
		particle_size, particle_size,
		cur_ms, XSW_PARTICLE_LIFE_SPAN
	    );
	}

}


/*
 *	Manages all reality engine responsibilities.
 */
static void REngUpdateObject(
	xsw_object_struct *player_obj, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc,
	const int spawn_particles
)
{
	xsw_net_parms_struct *net = &net_parms;

	REngVisibility(player_obj, obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngShieldVisibility(obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngHeading(player_obj, obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngThrottle(player_obj, obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngThrust(player_obj, obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngAction(player_obj, obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngLocation(player_obj, obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngMove(
	    player_obj, obj,
	    cur_ms, cur_sec, tc,
	    1
	);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngAnimation(obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	REngAutoDestruct(player_obj, obj, cur_ms, cur_sec, tc);
	if(!XSW_CON_IS_CONNECTED(net))
	    return;

	if(spawn_particles && (total_particles < option.max_particles))
	{
	    REngSpawnParticles(
		player_obj, obj,
		cur_ms, cur_sec, tc
	    );
	    if(!XSW_CON_IS_CONNECTED(net))
		return;
	}
}



/*
 *	Initializes the reality engine.
 */
int REngInit(void)
{
	/* No initialization needed */

	return(0);
}


/*
 *	Manages the reality engine.
 *
 *	Updates all objects in the inrange objects list.
 */
void REngManage(
	xsw_object_struct *player_obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char spawn_particles;
	int i;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;

	spawn_particles = (next.spawn_particles <= cur_ms) ? 1 : 0;

	/* Check if local updating is enabled */
	if(option.local_updates)
	{
	    /* Update all objects in range */
	    for(i = 0; i < total_inrange_objects; i++)
	    {
		obj = inrange_xsw_object[i];
		if(obj == NULL)
		    continue;

		if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		    continue;

		REngUpdateObject(
		    net->player_obj_ptr, obj,
		    cur_ms, cur_sec, tc,
		    spawn_particles
		);
		if(!XSW_CON_IS_CONNECTED(net))
		    break;
	    }
	}
	else if(net->player_obj_ptr != NULL)
	{
	    /* Update just the player object */
	    obj = net->player_obj_ptr;

	    REngUpdateObject(
		obj,		/* Player */
		obj,
		cur_ms, cur_sec, tc,
		spawn_particles
	    );
	}

	/* Update particles */
	ParticleUpdateAll(
	    &xsw_particle, &total_particles,
	    cur_ms, cur_sec, tc
	);

	if(spawn_particles)
	    next.spawn_particles = cur_ms + XSW_SPAWN_PARTICLES_INTERVAL;
}

/*
 *	Shuts down the reality engine.
 */
void REngShutdown(void)
{
	/* No shutdown procedure required */

	return;
}
