#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/swsoundcodes.h"
#include "../include/unvmath.h"

#include "sound.h"
#include "xsw.h"


static int SC_GET_OBJECT_NUM(xsw_object_struct *obj);

scanner_contact_struct *ScNew(void);
void ScDelete(scanner_contact_struct *sc);

int ScListFindByObject(xsw_object_struct *obj);
int ScListAddObject(
	scanner_contact_struct ***list, int *total,
	xsw_object_struct *obj
);
void ScListDeleteByObject(
	scanner_contact_struct ***list, int *total,
	xsw_object_struct *obj
);
void ScListClear(scanner_contact_struct **list, const int total);
void ScListReclaim(
	scanner_contact_struct ***list, int *total
);

int ScHandleContacts(
        scanner_contact_struct ***list, int *total,
	const int obj_num
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the object number of the specified object pointer.
 *
 *	Important: If the object is garbage, the valid object number
 *	will still be returned.
 */
static int SC_GET_OBJECT_NUM(xsw_object_struct *obj)
{
	const int total = total_objects;
	int i;
	xsw_object_struct **list = xsw_object;

	if(obj == NULL)
	    return(-1);

	for(i = 0; i < total; i++)
	{
	    if(list[i] == obj)
		return(i);
	}

	return(-1);
}


/*
 *	Creates a new scanner contact.
 */
scanner_contact_struct *ScNew(void)
{
	return(SCANNER_CONTACT(calloc(
	    1, sizeof(scanner_contact_struct)
	)));
}

/*
 *	Deletes the scanner contact.
 */
void ScDelete(scanner_contact_struct *sc)
{
	if(sc == NULL)
	    return;

	free(sc);
}


/*
 *	Returns the index of the scanner contact in the list that
 *	references the specified object or -1 on error.
 */
int ScListFindByObject(
	scanner_contact_struct **list, const int total,
	xsw_object_struct *obj
)
{
	int i, obj_num;
	scanner_contact_struct *sc;

	if(obj == NULL)
	    return(-1);

	obj_num = SC_GET_OBJECT_NUM(obj);
	if(obj_num < 0)
	    return(-1);

	for(i = 0; i < total; i++)
	{
	    sc = list[i];
	    if(sc == NULL)
		continue;

	    if(sc->obj_num == obj_num)
		return(i);
	}

	return(-1);
}

/*
 *	Adds the object to the scanner contact list.
 *
 *	Does not check if the object is already in the list.
 *
 *	Returns the index of the scanner contact added to the list or
 *	negative on error.
 */
int ScListAddObject(
	scanner_contact_struct ***list, int *total,
	xsw_object_struct *obj
)
{
	int i, obj_num;
	scanner_contact_struct *sc;

	if((list == NULL) || (total == NULL) || (obj == NULL))
	    return(-1);

	obj_num = SC_GET_OBJECT_NUM(obj);

	for(i = 0; i < *total; i++)
	{
	    sc = (*list)[i];
	    if(sc == NULL)
		break;

	    if(DBIsObjectGarbage(sc->obj_num))
		break;
	}
	if(i < *total)
	{
	    ScDelete((*list)[i]);
	    (*list)[i] = sc = ScNew();
	}
	else
	{
	    i = MAX((*total), 0);
	    *total = i + 1;
	    *list = (scanner_contact_struct **)realloc(
		*list,
		(*total) * sizeof(scanner_contact_struct *)
	    );
	    if(*list == NULL)
	    {
		*total = 0;
		return(-3);
	    }
	    (*list)[i] = sc = ScNew();
	}
	if(sc == NULL)
	    return(-3);

	sc->obj_num = obj_num;

	return(i);
}

/*
 *	Deletes all the scanner contacts in the list that refer to the
 *	specified object.
 */
void ScListDeleteByObject(
	scanner_contact_struct ***list, int *total,
	xsw_object_struct *obj
)
{
	int i, j;

	if((list == NULL) || (total == NULL))
	    return;

	/* Begin deleting all scanner contacts that refer to the
	 * specified object
	 */
	while(*total > 0)
	{
	    /* Find the next scanner contact that refers to the
	     * specified object
	     */
	    i = ScListFindByObject(*list, *total, obj);
	    if(i < 0)
		break;

	    /* Delete this scanner contact */
	    ScDelete((*list)[i]);

	    /* Reduce the total and shift the pointers in the list */
	    *total = (*total) - 1;
	    for(j = i; j < *total; j++)
		(*list)[j] = (*list)[j + 1];

	    /* Reallocate the list */
	    if(*total > 0)
	    {
		*list = (scanner_contact_struct **)realloc(
		    *list,
		    (*total) * sizeof(scanner_contact_struct *)
		);
		if(*list == NULL)
		{
		    *total = 0;
		    break;
		}
	    }
	    else
	    {
		free(*list);
		*list = NULL;
		*total = 0;
	    }
	}
}

/*
 *	Deletes the list of scanner contacts.
 */
void ScListClear(scanner_contact_struct **list, const int total)
{
	int i;

	for(i = 0; i < total; i++)
	    ScDelete(list[i]);

	free(list);
}

/*
 *	Deletes all the unused scanner contacts in the list.
 */
void ScListReclaim(
	scanner_contact_struct ***list, int *total
)
{
	int i, highest;
	scanner_contact_struct *sc;

	if((list == NULL) || (total == NULL))
	    return;

	/* Get the highest valid index */
	highest = -1;
	for(i = 0; i < *total; i++)
	{
	    sc = (*list)[i];
	    if(sc == NULL)
		continue;

	    if(DBIsObjectGarbage(sc->obj_num))
		continue;

	    highest = i;
	}
	if(highest > -1)
	{
	    /* Delete the unused scanner contacts */
	    for(i = highest + 1;
		i < *total;
	        i++
	    )
		ScDelete((*list)[i]);

	    *total = highest + 1;
	    if(*total > 0)
	    {
		(*list) = (scanner_contact_struct **)realloc(
		    *list,
		    (*total) * sizeof(scanner_contact_struct *)
		);
		if(*list == NULL)
		{
		    *total = 0;
		    return;
		}
	    }
	    else
	    {
		free(*list);
		*list = NULL;
		*total = 0;
	    }
	}
	else
	{
	    ScListClear(*list, *total);
	    *list = NULL;
	    *total = 0;
	}
}


/*
 *	Manages the scanner contacts.
 *
 *	The obj_num specifies the object that the scanner contacts
 *	relative to (this is usually the player object).
 *
 *	Calls to this function should be timmed properly, this
 *	function does NOT need to be called every time per loop.
 *
 *	This function is only used on the client to notify
 *	when a new object comes into scanner range of obj_num.
 */
int ScHandleContacts(
	scanner_contact_struct ***list, int *total,
	const int obj_num
)
{
	bool_t played_scanner_beep = False;
	const time_t cur_ms = cur_millitime;
	int i, status;
	float scanner_range;
	xsw_sound_struct *snd = &sound;
	xsw_object_struct *player_obj, *tar_obj, **obj_list;

	/* Skip if global option specifies not to do scanner notifying */
	if(!option.notify_scanner_contacts)
	    return(0);

	/* Check if object_num is valid */
	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	player_obj = xsw_object[obj_num];

	/* Get the current effective scanner range in real units */
	scanner_range = DBGetObjectEffectiveScannerRangePtr(player_obj);
	if(scanner_range <= 0.0f)
	    return(-1);

	/* Begin updating the contacts list */
	obj_list = inrange_xsw_object;
	for(i = 0; i < total_inrange_objects; i++)
	{
	    tar_obj = obj_list[i];
	    if(tar_obj == NULL)
		continue;
	    if(tar_obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Skip the following object types plus outdated ones and
	     * the player itself
	     */
	    if((tar_obj == player_obj) ||
	       (tar_obj->type == SW_OBJ_TYPE_WEAPON) ||
	       (tar_obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	       (tar_obj->type == SW_OBJ_TYPE_SPHEREWEAPON) ||
/* Animated objects are usually explosions, but sometimes not, so
 * skip these for now even though some may not be explosions
 */
	       (tar_obj->type == SW_OBJ_TYPE_ANIMATED)
	    )
		continue;

	    /* Is this object already in the scanner contacts list? */
	    if(ScListFindByObject(*list, *total, tar_obj) > -1)
	    {
		/* The object is in the scanner contact list, check if
		 * it is within scanner range
		 */

	        /* Outdated? */
		if((tar_obj->last_updated + option.net_object_outdated_time)
		    < cur_ms
		)
		{
		    status = 0;
		}
		else
		{
		    status = DBIsObjectInEffectiveScannerRangePtr(
			player_obj, tar_obj, scanner_range
		    );
		}
		/* Is this object not in scanner range JUST NOW? */
		if(!status)
		{
		    /* The object JUST LEFT scanner range */

		    /* Print message */
		    char *s = (char *)malloc(
			STRLEN(tar_obj->name) + 40 + 80
		    );
		    if((tar_obj->sect_x == player_obj->sect_x) &&
		       (tar_obj->sect_y == player_obj->sect_y) &&
		       (tar_obj->sect_z == player_obj->sect_z)
		    )
		    {
		        sprintf(
			    s,
"%s just left scanner range, bearing %.0f'.",
			    tar_obj->name,
			    RADTODEG(
				SWMathCoordinateDeltaVector(
				    tar_obj->x - player_obj->x,
				    tar_obj->y - player_obj->y
				)
			    )
			);
		    }
		    else
		    {
			sprintf(
			    s,
"%s just left scanner range, bearing %.0f'.",
			    tar_obj->name,
			    RADTODEG(
				SWMathCoordinateDeltaVector(
				    tar_obj->sect_x - player_obj->sect_x,
				    tar_obj->sect_y - player_obj->sect_y
				)
			    )
			);
		    }
		    XSWPrint(s);
		    free(s);

		    /* Remove all the scanner contacts that refer to
		     * this object
		     */
		    ScListDeleteByObject(list, total, tar_obj);
		}
	    }
	    else
	    {
		/* The object was not in the scanner contacts list,
		 * check if the object is within scanner range
		 */

		/* Outdated? */ 
		if((tar_obj->last_updated + option.net_object_outdated_time)
		    < cur_ms
		)
		{
		    status = 0;
		}
		else     
		{
		    status = DBIsObjectInEffectiveScannerRangePtr(
			player_obj, tar_obj, scanner_range
		    );
		}
		/* Is the object in scanner range now? */
		if(status)
		{
		    /* The object just entered scanner range */

		    /* Print message */
		    char *s = (char *)malloc(
			STRLEN(tar_obj->name) + 40 + 80
		    );
		    sprintf(
			s,
"%s just entered scanner range, bearing %.0f'.",
			tar_obj->name,
			RADTODEG(
			    SWMathCoordinateDeltaVector(
				tar_obj->x - player_obj->x,
				tar_obj->y - player_obj->y
			    )
			)
		    );
		    XSWPrintBold(s);
		    free(s);

		    /* Add this object to the scanner contacts list */
		    ScListAddObject(list, total, tar_obj);

		    /* Play scanner contacts sound */
		    if((option.sounds >= XSW_SOUNDS_EVENTS) &&
		       !played_scanner_beep
		    )
		    {
			SoundPlay(
			    snd,
			    SW_SNDREF_CONTACTS_BEEP,
			    1.0f, 1.0f,
			    0,
			    1
			);
			played_scanner_beep = True;
		    }
		}
	    }
	}

	/* Reclaim scanner contacts memory */
	ScListReclaim(list, total);

	return(0);
}
