#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "msgwin.h"
#include "net.h"
#include "xsw.h"


int ServScriptIsAllocated(int n);

int ServScriptStart(
	const char *filename,
	int argc, char **argv		/* Transfered */
);

void ServScriptDelete(int n);
void ServScriptDeleteAll(void);
void ServScriptReclaim(void);

void ServScriptDoSend(serv_script_struct *ss);
void ServScriptManage(void);

void ServScriptMapPrompt(const char *filename);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the server script n is allocated.
 */
int ServScriptIsAllocated(int n)
{
	if((serv_script == NULL) ||
	   (n < 0) ||
	   (n >= total_serv_scripts)
	)
	    return(0);
	else if(serv_script[n] == NULL)
	    return(0);
	else
	    return(1);
}


/*
 *	Allocates a new server script structure and sets it up.
 *
 *	The specified arguments in argv are transfered, the calling 
 *	function must not modify or delete them after this call.
 */
int ServScriptStart(
	const char *filename,
	int argc, char **argv
)
{
	int i, n;
	struct stat stat_buf;
	serv_script_struct *ss;
	xsw_net_parms_struct *net = &net_parms;

	/* Not connected? */
	if(net->connection_state != XSW_CON_STATE_CONNECTED)
	{
	    XSWPrintWarning("Not connected.");
	    return(-1);
	}

	/* Sanitize total */
	if(total_serv_scripts < 0)
	    total_serv_scripts = 0;

	/* Check for available pointer */
	for(i = 0; i < total_serv_scripts; i++)
	{
	    if(serv_script[i] == NULL)
		break;
	}
	if(i < total_serv_scripts)
	{
	    n = i;
	}
	else
	{
	    n = total_serv_scripts;
	    total_serv_scripts = n + 1;
	    serv_script = (serv_script_struct **)realloc(
		serv_script,
		total_serv_scripts * sizeof(serv_script_struct *)
	    );
	    if(serv_script == NULL)
	    {
		total_serv_scripts = 0;
		return(-3);
	    }
	}

	/* Allocate a new server script */
	serv_script[n] = ss = (serv_script_struct *)calloc(
	    1, sizeof(serv_script_struct)
	);
	if(ss == NULL)
	    return(-3);

	/* Open the server script file */
	if(filename != NULL)
	{
	    FILE *fp = fopen(filename, "rb");
	    ss->fp = fp;
	    if(fp == NULL)
	    {
		const int error_code = errno;
		char *s, *error_msg = STRDUP(strerror(error_code));
		if(error_msg == NULL)
		    error_msg = strdup("unknown error");
		*error_msg = toupper(*error_msg);
		s = (char *)malloc(
		    STRLEN(error_msg) + STRLEN(filename) + 80
		);
		sprintf(
		    s,
"%s: %s.",
		    filename, error_msg
		);
		XSWPrintWarning(s);
		free(s);
		free(error_msg);
		ServScriptDelete(n);
		return(-1);
	    }

	    /* Get file size */
	    if(!fstat(fileno(fp), &stat_buf))
		ss->filesize = stat_buf.st_size;

	    ss->filename = strdup(filename);
	}

	/* Transfer arguments (do not copy) */
	ss->argv = argv;
	ss->argc = argc;

	if(!STRISEMPTY(ss->filename))
	{
	    char *s = (char *)malloc(
		STRLEN(ss->filename) + 80
	    );
	    sprintf(
		s,
"Sending: %s...",
		ss->filename
	    );
	    XSWPrint(s);
	    free(s);
	}

	return(n);
}


/*
 *	Deletes the server script.
 */
void ServScriptDelete(int n)
{
	int i;
	serv_script_struct *ss;

	if(!ServScriptIsAllocated(n))
	    return;

	ss = serv_script[n];

	/* Close the file */
	if(ss->fp != NULL)
	{
	    fclose(ss->fp);
	    ss->fp = NULL;
	}

	free(ss->filename);

	for(i = 0; i < ss->argc; i++)
	    free(ss->argv[i]);
	free(ss->argv);

	free(ss);
	serv_script[n] = ss = NULL;
}

/*
 *	Deletes all server scripts.
 */
void ServScriptDeleteAll(void)
{
	int i;

	for(i = 0; i < total_serv_scripts; i++)
	    ServScriptDelete(i);
	free(serv_script);
	serv_script = NULL;
	total_serv_scripts = 0;
}

/*
 *	Reclaim memory used by server scripts.
 */
void ServScriptReclaim(void)
{
	int i, h;

	for(i = 0, h = -1; i < total_serv_scripts; i++)
	{
	    if(serv_script[i] != NULL)
		h = i;
	}

	total_serv_scripts = h + 1;

	if(total_serv_scripts > 0)
	{
	    serv_script = (serv_script_struct **)realloc(
		serv_script,
		total_serv_scripts * sizeof(serv_script_struct *)
	    );
	    if(serv_script == NULL)
	    {
		total_serv_scripts = 0;
		return;
	    }
	}
	else
	{
	    free(serv_script);
	    serv_script = NULL;
	    total_serv_scripts = 0;
	}
}


/*
 *	Sends the script data to the connection.
 */
void ServScriptDoSend(serv_script_struct *ss)
{
	int i, c, argn, status;
	int numstr_pos;
const int numstr_len = 48;
	char numstr[numstr_len];

	int sndbuf_pos;
	char sndbuf[CS_DATA_MAX_LEN];

	const char *arg;

const int sndbuf_len = CS_DATA_MAX_LEN - 10;
	xsw_net_parms_struct *net = &net_parms;


	if(ss == NULL)
	    return;

	/* Script file closed? */
	if(ss->fp == NULL)
	{
	    char *s = (char *)malloc(
		STRLEN(ss->filename) + 80
	    );
	    sprintf(
		s,
"%s: Done.",
		ss->filename
	    );
	    XSWPrint(s);    
	    free(s);

	    ss->filepos = ss->filesize;
	    return;
	}

	/* Not connected? */
	if(net->connection_state != XSW_CON_STATE_CONNECTED)
	{
	    char *s = (char *)malloc(
		STRLEN(ss->filename) + 80
	    );
	    sprintf(
		s,  
"%s: Lost connection, this script upload may not be complete.",
		ss->filename
	    );
	    XSWPrintWarning(s);
	    free(s);

	    ss->filepos = ss->filesize;
	    return;
	}

	/* Not time to send? */
/* TODO */

	/* Set literal command prefix to sndbuf */
	sprintf(sndbuf, "%i ", CS_CODE_LITERALCMD);

	/* Set sndbuf position */
	sndbuf_pos = strlen(sndbuf);

	/* Put data into sndbuf buffer */
	while(True)
	{
	    /* sndbuf position exceeded allowable sndbuf length? */
	    if(sndbuf_pos >= sndbuf_len)
	    {
		/* Set end characters */
		sndbuf[sndbuf_pos - 2] = '\n';
		sndbuf[sndbuf_pos - 1] = '\0';
		break;
	    }


	    /* Get next character in file */
	    c = fgetc(ss->fp);
	    ss->filepos++;

	    /* End of file? */
	    if((c == EOF) || (ss->filepos >= ss->filesize))
	    {
		char *s = (char *)malloc(
		    STRLEN(ss->filename) + 80
		);
		sprintf(
		    s,
"%s: Done.",
		    ss->filename
		);
		XSWPrint(s);
		free(s);

		/* Set file position to match size so that it gets
		 * deleted
		 */
		ss->filepos = ss->filesize;

		/* Set end characters */
		sndbuf[sndbuf_pos] = '\n';
		if((sndbuf_pos + 1) < sndbuf_len)
		    sndbuf[sndbuf_pos + 1] = '\0';

		break;
	    }

	    /* Backslash? */
	    if(c == '\\')
	    {
		/* Next character is to be accepted literally */

		/* Get next character in file */
		c = fgetc(ss->fp);
		ss->filepos++;

		/* End of file? */
		if((c == EOF) ||
		   (ss->filepos >= ss->filesize)
		)
		{
		    char *s = (char *)malloc(
			STRLEN(ss->filename) + 80
		    );
		    sprintf(
			s,
"%s: Done.",
			ss->filename
		    );
		    XSWPrint(s);
		    free(s);

		    /* Set file position to match size so it gets deleted */
		    ss->filepos = ss->filesize;

		    /* Set end characters */
		    sndbuf[sndbuf_pos] = '\n'; 
		    if((sndbuf_pos + 1) < sndbuf_len)  
			sndbuf[sndbuf_pos + 1] = '\0';
 
		    break;
		}

		/* Set character */
		sndbuf[sndbuf_pos] = (char)c;
		sndbuf_pos++;

		continue;
	    }

	    /* Substitute argument? */
	    if(c == '$')
	    {
		/* Get next character in file */
		c = fgetc(ss->fp);
		ss->filepos++;

		/* Get number */
		numstr_pos = 0;
		while(isdigit(c) && (numstr_pos < numstr_len))
 		{
		    numstr[numstr_pos] = (char)c;
		    numstr_pos++;

		    /* Get next character in file */
		    c = fgetc(ss->fp);
		    ss->filepos++;
		}
		if(numstr_pos < numstr_len)
		    numstr[numstr_pos] = '\0';
		else
		    numstr[numstr_len - 1] = '\0';

		/* Get argument number */
		argn = atoi(numstr);
		/* Put argument into sndbuf */
		if((argn >= 0) && (argn < ss->argc))
		{
		    arg = ss->argv[argn];
		    if(arg != NULL)
		    {
			for(i = 0; arg[i] != '\0'; i++)
			{
			    if(sndbuf_pos >= sndbuf_len)
				break;

			    sndbuf[sndbuf_pos] = arg[i];
			    sndbuf_pos++;
			}
		    }
		}

		/* Continue if argument filled sndbuf */
		if(sndbuf_pos >= sndbuf_len)
		    continue;
	    }

	    /* Newline? */
	    if((c == '\n') || (c == '\r'))
	    {
		/* Set end characters */
		sndbuf[sndbuf_pos] = '\n';
		if((sndbuf_pos + 1) < sndbuf_len)
		    sndbuf[sndbuf_pos + 1] = '\0';
		break;
	    }

	    /* Set character */
	    sndbuf[sndbuf_pos] = (char)c;
	    sndbuf_pos++;
	}
	sndbuf[sizeof(sndbuf) - 1] = '\0';

	/* Send to connection */
	status = NetSendData(sndbuf);
	if(status)
	{
	    char *s = (char *)malloc(
		STRLEN(ss->filename) + 80
	    );
	    sprintf(
		s,
"%s: Send data error, this script upload may not be complete.",
		ss->filename
	    );
	    XSWPrintWarning(s);
	    free(s);

	    ss->filepos = ss->filesize;
	}
}

/*
 *	Manages all server scripts.
 */
void ServScriptManage(void)
{
	int i;
	serv_script_struct *ss;

	for(i = 0; i < total_serv_scripts; i++)
	{
	    ss = serv_script[i];
	    if(ss == NULL)
		continue;

	    /* Send script */
	    ServScriptDoSend(ss);

	    /* Done sending or lost connection? */
	    if(ss->filepos >= ss->filesize)
	    {
		ServScriptDelete(i);
		ss = NULL;
		ServScriptReclaim();
		continue;
	    }
	}
}

/*
 *	Callback function to map the prompt and set up the initial
 *	value to "script <filename> "
 */
void ServScriptMapPrompt(const char *path)
{
	if(path != NULL)
	{
            xsw_msg_win_struct *mw = xsw_msg_win;

	    if((mw != NULL) ?
		mw->map_state : False
	    )
	    {
		prompt_struct *prompt = &mw->prompt;
		char *s = (char *)malloc(
		    STRLEN(path) + 80
		);
		sprintf(
		    s,
		    "script %s ",
		    path
		);
		PromptSetS(prompt, s);
		free(s);
		PromptUnmarkBuffer(prompt, PROMPT_POS_END);

		PromptSetLabel(
		    prompt,
		    "Client Command:"
		);
		prompt_mode = XSW_PROMPT_CLIENT_CMD;
		PromptMap(prompt);
		prompt->has_focus = True;
	    }
	    else if((xsw_bridge_win != NULL) ?
		xsw_bridge_win->map_state : False
	    )
	    {
		prompt_struct *prompt = &xsw_bridge_win->prompt;
		char *s = (char *)malloc(
		    STRLEN(path) + 80
		);
		sprintf(
		    s,
		    "script %s ",
		    path
		);
		PromptSetS(prompt, s);
		free(s);
		PromptUnmarkBuffer(prompt, PROMPT_POS_END);

		PromptSetLabel(
		    prompt,
		    "Client Command:"
		);
		prompt_mode = XSW_PROMPT_CLIENT_CMD;
		PromptMap(prompt);
		prompt->has_focus = True;
	    }
	}
}
