#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/cfgfmt.h"
#include "../include/string.h"

#include "../include/fio.h"
#include "../include/disk.h"

#include "sndrefs.h"


int SndRefsFileOpen(
	sndref_struct ***sndrefs_list, int *nsndrefs,
	const char *path,
	char **audio_mode_rtn,
	void (*progress_cb)(void *, unsigned long, unsigned long),
	void *progress_data,
	void (*print_warning)(const char *s)
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


int SndRefsFileOpen(
	sndref_struct ***sndrefs_list, int *nsndrefs,
	const char *path,
	char **audio_mode_rtn,
	void (*progress_cb)(void *, unsigned long, unsigned long),
	void *progress_data,
	void (*print_warning)(const char *s)
)
{
	int lines_read = 0;
	unsigned long file_size = 0l;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	struct stat stat_buf;

	/* Report initial progress */
	if(progress_cb != NULL)
	    progress_cb(progress_data, 0l, file_size);

	if((sndrefs_list == NULL) || (nsndrefs == NULL) || STRISEMPTY(path))
	    return(-2);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for reading");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    if(print_warning != NULL)
		print_warning(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;

	    if(progress_cb != NULL)
		progress_cb(progress_data, (unsigned long)ftell(fp), file_size);
	}

	/* Delete all the existing sndrefs */
	SndRefDeleteAll(sndrefs_list, nsndrefs);

	if(audio_mode_rtn != NULL)
	{
	    free(*audio_mode_rtn);
	    *audio_mode_rtn = NULL;
	}

#define WARN_NO_SUCH_PARM(_parm_)               {       \
 char *s = (char *)malloc(                              \
  STRLEN(path) + 40 + STRLEN(_parm_) + 80               \
 );                                                     \
 sprintf(                                               \
  s,                                                    \
"%s: Line %i: Unsupported parameter `%s'.",             \
  path, lines_read, (_parm_)                            \
 );                                                     \
 if(print_warning != NULL)				\
  print_warning(s);					\
 free(s);                                               \
}

#define WARN_NO_SUCH_FILE(_v_,_error_code_)     {       \
 const int error_code = (_error_code_);                 \
 char *s, *error_msg = STRDUP(strerror(error_code));    \
 if(error_msg == NULL)                                  \
  error_msg = strdup("no such file");                   \
 *error_msg = toupper(*error_msg);                      \
							\
 s = (char *)malloc(                                    \
  STRLEN(path) + 40 + STRLEN(parm) + STRLEN(_v_) +      \
  STRLEN(error_msg) + 80                                \
 );                                                     \
 sprintf(                                               \
  s,                                                    \
"%s: Line %i: %s: Warning: %s: %s.",                    \
  path, lines_read, parm, (_v_), error_msg              \
 );							\
 if(print_warning != NULL)				\
  print_warning(s);					\
 free(s);                                               \
							\
 free(error_msg);                                       \
}

	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';

	    /* Report progress */
	    if(progress_cb != NULL)
		progress_cb(progress_data, (unsigned long)ftell(fp), file_size);

	    /* Title */
	    if(!strcasecmp(parm, "Title"))
	    {
		/* Ignore this */
	    }
	    /* ToplevelDir */
	    else if(!strcasecmp(parm, "ToplevelDir"))
	    {
		/* Ignore this */
	    }

	    /* AudioModeName */
	    else if(!strcasecmp(parm, "AudioModeName") ||
		    !strcasecmp(parm, "ModeName") ||
		    !strcasecmp(parm, "AudioMode")
	    )
	    {
		if(audio_mode_rtn != NULL)
		{
		    free(*audio_mode_rtn);
		    *audio_mode_rtn = STRDUP(val);
		}
	    }


	    /* BeginSndRef */
	    else if(!strcasecmp(parm, "BeginSndRef") ||
		    !strcasecmp(parm, "BeginSoundItem") ||
		    !strcasecmp(parm, "BeginSoundSchemeItem")
	    )
	    {
		const int sndref_num = ATOI(val);
		sndref_struct *sndref;

		/* Create a new SndRef */
		if(SndRefNewExplicit(
		    sndrefs_list, nsndrefs, sndref_num
		) < 0)
		    continue;

		sndref = SndRefGetPtr(*sndrefs_list, *nsndrefs, sndref_num);
		if(sndref == NULL)
		    continue;

		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';


		    /* Name */
		    if(!strcasecmp(parm, "Name"))
		    {
			free(sndref->name);
			sndref->name = STRDUP(val);
		    }
		    /* Description */
		    else if(!strcasecmp(parm, "Description") ||
			    !strcasecmp(parm, "Desc")
		    )
		    {
			free(sndref->description);
			sndref->description = STRDUP(val);
		    }
		    /* Path */
		    else if(!strcasecmp(parm, "Path") ||
		            !strcasecmp(parm, "Filename")
		    )
		    {
			free(sndref->path);
			sndref->path = STRDUP(val);
		    }
		    /* EndSndRef */
		    else if(!strcasecmp(parm, "EndSndRef") ||
			    !strcasecmp(parm, "EndSoundItem") ||
			    !strcasecmp(parm, "EndSoundSchemeItem")
		    )
		    {
			break;
		    }
		    /* Unsupported parameter */
		    else
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }
	    /* Unsupported parameter */
	    else
	    {
		WARN_NO_SUCH_PARM(parm);
	    }
	}
	free(s);

#undef WARN_NO_SUCH_PARM
#undef WARN_NO_SUCH_FILE

	/* Close the file */
	fclose(fp);

	/* Report final progress */
	if(progress_cb != NULL)
	    progress_cb(progress_data, file_size, file_size);

	return(0);
}
