#include <stdlib.h>
#include <string.h>
#ifdef HAVE_ESD  
# include <esd.h>
#endif
#ifdef HAVE_Y2
# include <Y2/Y.h>
# include <Y2/Ylib.h>
#endif
#include "sound.h"
#include "sndrefs.h"
#include "xsw.h"


#ifdef HAVE_Y2
# ifndef DEF_Y_AUDIO_MODE_NAME
#  define DEF_Y_AUDIO_MODE_NAME		"Default"
# endif
#endif	/* HAVE_Y2 */


int SoundInit(xsw_sound_struct *snd, const int server_type);
int SoundChangeMode(xsw_sound_struct *snd, const char *mode_name);
void *SoundPlay(
	xsw_sound_struct *snd, const int sndref_num,
	const float vol_left,		/* 0.0 to 1.0 */
	const float vol_right,		/* 0.0 to 1.0 */
	const unsigned int effects,	/* Any of SOUND_EFFECT_* */
	const int priority		/* 0 or 1 */
);
void SoundStop(xsw_sound_struct *snd, void *playid);
int SoundChangeBackgroundMusic(
	xsw_sound_struct *snd, const int sndref_num,
	const unsigned int effects,	/* Any of SOUND_EFFECT_* */
	const int priority		/* 0 or 1 */
);
int SoundStopBackgroundMusic(xsw_sound_struct *snd);
int SoundManage(xsw_sound_struct *snd);
void SoundShutdown(xsw_sound_struct *snd);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Connects to the sound server and opens the sndrefs from file.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	General error.
 *	-2	Bad value.
 */
int SoundInit(xsw_sound_struct *snd, const int server_type)
{
	if(snd == NULL)
	    return(-2);

	/* Initialize by the sound server type */
	switch(server_type)
	{
	  case SOUND_SERVER_TYPE_YIFF:
#ifdef HAVE_Y2
	    if(STRISEMPTY(snd->con_arg))
	    {
		free(snd->con_arg);
		snd->con_arg = strdup("127.0.0.1:9433");
	    }

	    /* Connect to the sound server */
	    snd->con_data = (void *)YOpenConnection(
		snd->start_cmd,
		snd->con_arg
	    );
	    if(snd->con_data == NULL)
		return(-1);
#endif	/* HAVE_Y2 */
	    break;

	  case SOUND_SERVER_TYPE_ESOUND:
#ifdef ESD_H
	    /* Connect to the sound server */
	    snd->con_data = (void *)esd_open_sound(NULL);
	    if((int)snd->con_data < 0)
	    {
		snd->con_data = NULL;
		return(-1);
	    }
#endif /* ESD_H */
	    break;

	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}

	/* Reset values */
	snd->server_type = server_type;
	snd->playid_list = NULL;
	snd->nplayids = 0;
	snd->bg_music_playid = NULL;
	snd->bg_music_sndref_num = -1;
	snd->loop_tractor_beam_playid = NULL;

	return(0);
}


/*
 *	Sets the Audio mode.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	General error.
 *	-2	Bad value.
 *	-5	Sound not initialized.
 */
int SoundChangeMode(xsw_sound_struct *snd, const char *mode_name)
{
	int status = 0;
#ifdef HAVE_Y2
	int i;
	YAudioModeValuesStruct **yaudio_mode;
	int total_yaudio_modes;
#endif	/* HAVE_Y2 */
	if(snd == NULL)
	    return(-2);

	if(snd->con_data == NULL)
	    return(-5);

	/* Change mode by which sound server type: */
	switch(snd->server_type)
	{
	  case SOUND_SERVER_TYPE_YIFF:
#ifdef HAVE_Y2
	    if(mode_name == NULL)
		mode_name = DEF_Y_AUDIO_MODE_NAME; 

	    /* Get listing of available Audio modes */
	    yaudio_mode = YGetAudioModes(
		(YConnection *)snd->con_data,
		&total_yaudio_modes
	    );

	    /* Check if the specified Audio mode is in list */
	    for(i = 0; i < total_yaudio_modes; i++)
	    {
		if(yaudio_mode[i] == NULL)
		    continue;

		/* Audio mode names match? */
		if(!strcasecmp(yaudio_mode[i]->name, mode_name))
		    break;
	    }
	    /* Delete the Audio modes listing */
	    YFreeAudioModesList(yaudio_mode, total_yaudio_modes);

	    /* Audio mode name not in list? */
	    if(i >= total_yaudio_modes)
	    {
		status = -2;
		break;
	    }

	    /* Change to the specified Audio mode */
	    status = YChangeAudioModePreset(
		(YConnection *)snd->con_data, mode_name
	    );
	    if(status < 0)
		status = -1;
	    else
		status = 0;
#endif	/* HAVE_Y2 */
	    break;
	
	  case SOUND_SERVER_TYPE_ESOUND:
	    break;

	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}

	/* Set new audio mode */
	if(status == 0)
	{
	    free(snd->audio_mode_name);
	    snd->audio_mode_name = STRDUP(mode_name);
	}

	return(status);
}


/*
 *	Plays the sndref specified by sndref_num.
 *
 *	Returns the play id on success or NULL on error.
 */
void *SoundPlay(
	xsw_sound_struct *snd, const int sndref_num,
	const float vol_left,		/* 0.0 to 1.0 */
	const float vol_right,		/* 0.0 to 1.0 */
	const unsigned int effects,	/* Any of SOUND_EFFECT_* */
	const int priority		/* 0 or 1 */
)
{
	const char *path;
#ifdef ESD_H
	int id;
	char *esd_sndobj_path;
#endif
#ifdef HAVE_Y2
	YID yid;
#endif	/* HAVE_Y2 */
	sndref_struct *sndref;
	void *playid = NULL;

	if(snd == NULL)
	    return(playid);

	if(snd->con_data == NULL)
	    return(playid);

	/* SndRef not defined? */
	sndref = SndRefGetPtr(xsw_sndref, total_sndrefs, sndref_num);
	if(sndref == NULL)
	    return(playid);

	/* Get the path to the sound file */
	path = xsw_sndref[sndref_num]->path;
	if(path == NULL)
	    return(playid);

	/* Play by sound server type */
	switch(snd->server_type)
	{
	  case SOUND_SERVER_TYPE_YIFF:
#ifdef HAVE_Y2
	    if(snd->con_data != NULL)
	    {
		YConnection *con = (YConnection *)snd->con_data;
	        YEventSoundPlay value;

	        value.flags = YPlayValuesFlagPosition |
			      YPlayValuesFlagTotalRepeats |
			      YPlayValuesFlagVolume |
			      YPlayValuesFlagSampleRate;
	        value.position = 0l;
	        value.total_repeats = (effects & SOUND_EFFECT_REPEATING) ?
		    -1 : 1;
	        value.left_volume = vol_left;
	        value.right_volume = vol_right;
		value.back_left_volume = vol_left;
		value.back_right_volume = vol_right;
	        value.sample_rate = 0;
		yid = YStartPlaySoundObject(
		    con,
		    path,
		    &value
	        );
		if(yid != YIDNULL)
		    playid = (void *)yid;
	    }
#endif	/* HAVE_Y2 */
	    break;
  
	  case SOUND_SERVER_TYPE_ESOUND:
#ifdef ESD_H
	    fprintf(stderr, "Playing '%s'...", path);

	    /* We should be able to map sndrefs to sample id numbers */
	    esd_sndobj_path = (char *)malloc(
		STRLEN("xsw:") + STRLEN(path) + 80
	    );
	    sprintf(
		esd_sndobj_path,
		"xsw:%s",
		path
	    );
	    id = esd_sample_getid((int)snd->con_data, esd_sndobj_path);
	    free(esd_sndobj_path);

	    if(id < 0)
	    {
		fprintf(stderr,"(CACHING NOW)\n");
		id = esd_file_cache(
		    (int)snd->con_data,
		    "xsw",
		    path
		);
		if(id < 0)
		    fprintf(stderr, "esd_file_cache failed\n");
	    }
	    else
	    {
		fprintf(stderr, "(cached)\n");
	    }

	    if(id > -1)
	    {
		/* Stop the existing play (if any) */
		esd_sample_stop((int)snd->con_data, id);

		/* Set the volume */
		if(esd_set_default_sample_pan(
		    (int)snd->con_data,
		    id,
		    (int)(vol_left * ESD_VOLUME_BASE),
		    (int)(vol_right * ESD_VOLUME_BASE)
		   ) < 0
		)
		    fprintf(stderr, "esd_set_default_sample_pan() failed\n");

		/* Start playing */
		if(esd_sample_play((int)snd->con_data, id) < 0)
		{
		    fprintf(stderr, "esd_sample_play() failed\n");
		}
		else
		{
		    playid = (void *)id;
#if 0
/* Since we use the cached copies now.... do not free... */
		    if(esd_sample_free((int)snd->con_data, id) < 0)
		    {
			fprintf(stderr, "esd_sample_free() failed\n");
		    }
#endif
		}
	    }
	    else
	    {
		fprintf(stderr, "sample caching failure!\n");
	    }
#endif	/* ESD_H */
	    break;
   
	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}

	return(playid);
}

/*
 *	Stops the sound from playing.
 */
void SoundStop(xsw_sound_struct *snd, void *playid)
{
	if(snd == NULL)
	    return;

	if(snd->con_data == NULL)
	    return;

	switch(snd->server_type)
	{
	  case SOUND_SERVER_TYPE_YIFF:
#ifdef HAVE_Y2
	    if(snd->con_data != NULL)
	    {
		YConnection *con = (YConnection *)snd->con_data;
		YID yid = (YID)playid;
		YDestroyPlaySoundObject(con, yid);
	    }
#endif	/* HAVE_Y2 */
	    break;

	  case SOUND_SERVER_TYPE_ESOUND:
#ifdef ESD_H
	    esd_sample_stop((int)snd->con_data, (int)playid);
#endif	/* ESD_H */
	    break;

	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}
}

/*
 *	Sets the current background music to the one specified by
 *	sndref_num and stops the previous background music.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	General error.
 *	-2	Bad value or no such SndRef.  
 *	-5	Sound not initialized.
 */
int SoundChangeBackgroundMusic(
	xsw_sound_struct *snd, const int sndref_num,
	const unsigned int effects,	/* Any of SOUND_EFFECT_* */
	const int priority		/* 0 or 1 */
)
{
	char *path = NULL;
#ifdef HAVE_ESD
	int id;
#endif /* HAVE_ESD */
	sndref_struct *sndref;

	if(snd == NULL)
	    return(-2);

	if(snd->con_data == NULL)
	    return(-5);

	/* Get the path to the sound file from the SndRef */
	sndref = SndRefGetPtr(xsw_sndref, total_sndrefs, sndref_num);
	if(sndref != NULL)
	    path = sndref->path;

	/* Play by which sound server type: */
	switch(snd->server_type)
	{
	  case SOUND_SERVER_TYPE_YIFF:
#ifdef HAVE_Y2
	    /* Kill previous background music play */
	    if(snd->bg_music_playid != NULL)
	    {
		YDestroyPlaySoundObject(
		    (YConnection *)snd->con_data,
		    (YID)snd->bg_music_playid
		);
		snd->bg_music_playid = NULL;
		snd->bg_music_sndref_num = -1;
	    }

	    /* Begin new background music play */
	    if(!STRISEMPTY(path))
	    {
		YEventSoundPlay value;

		value.flags = YPlayValuesFlagPosition |
			      YPlayValuesFlagTotalRepeats |
			      YPlayValuesFlagVolume |
			      YPlayValuesFlagSampleRate;
		value.position = 0;
		value.total_repeats = -1;	/* Repeat infinatly */
		value.left_volume = 1.0;
		value.right_volume = 1.0;
		value.sample_rate = 0;

		snd->bg_music_playid = (void *)YStartPlaySoundObject(
		    (YConnection *)snd->con_data,
		    path,
		    &value
		);
		if(snd->bg_music_playid != NULL)
		{
		    snd->bg_music_sndref_num = sndref_num;
		}
	    }
#endif  /* HAVE_Y2 */
	    break;

	  case SOUND_SERVER_TYPE_ESOUND:
#ifdef ESD_H
	    id = esd_file_cache((int)snd->con_data, "xsw", path);
	    if(id < 0)
	    {
		fprintf(stderr, "esd_file_cache() failed\n");
		snd->bg_music_playid = NULL;
		snd->bg_music_sndref_num = -1;
	    }
	    else
	    {
		/* now that we have the new music cached, stop & restart */
		SoundStopBackgroundMusic(snd);
		// (int)snd->bg_music_playid = id;  // Dan S: forbidden in ansi c++. Cast rvalue not lvalue.
		snd->bg_music_playid = (void *)(&id);
		if(esd_sample_loop(
		    (int)snd->con_data,
		    (int)snd->bg_music_playid)
		   < 0
		)
		{
		    fprintf(stderr, "esd_sample_loop() failed\n");
		}
		else
		{
		    snd->bg_music_sndref_num = sndref_num;
		}
	    }
#endif	/* ESD_H */
	    break;

	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}

	return((snd->bg_music_sndref_num > -1) ? 0 : -2);
} 

/*
 *	Stops playing the background music.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	General error.
 *	-2	Bad avlue or no such SndRef.
 *	-5	Sound not initialized.
 */
int SoundStopBackgroundMusic(xsw_sound_struct *snd)
{
	if(snd == NULL)
	    return(-2);

	if(snd->con_data == NULL)
	    return(-5);

	switch(snd->server_type)
	{
	  case SOUND_SERVER_TYPE_YIFF:
#ifdef HAVE_Y2
	    /* Kill previous background music play */
	    if(snd->bg_music_playid != NULL)
		YDestroyPlaySoundObject(
		    (YConnection *)snd->con_data,
		    (YID)snd->bg_music_playid
		);
	
#endif  /* HAVE_Y2 */
	    break;

	  case SOUND_SERVER_TYPE_ESOUND:
#ifdef ESD_H
	    if(snd->bg_music_playid != NULL)
	    {
		if(esd_sample_stop(
		    (int)snd->con_data,
		    (int)snd->bg_music_playid)
		    < 0
		)
		    fprintf(stderr, "esd_sample_stop() failed\n");
		if(esd_sample_free(
		    (int)snd->con_data,
		    (int)snd->bg_music_playid)
		    < 0
		)
		    fprintf(stderr, "esd_sample_stop() failed\n");
	    }
#endif  /* ESD_H */
	    break;

	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}

	snd->bg_music_sndref_num = -1;
	snd->bg_music_playid = NULL;

	return(0);
}

/*
 *	Manages sound events.
 *
 *	Returns number of events handled.
 */
int SoundManage(xsw_sound_struct *snd)
{
	int events_handled = 0;
#ifdef HAVE_Y2
	YEvent yevent;
#endif	/* HAVE_Y2 */

	if(snd == NULL)
	    return(events_handled);

	switch(snd->server_type)
	{
	  case SOUND_SERVER_TYPE_YIFF:
#ifdef HAVE_Y2
	    if(snd->con_data != NULL)   
	    {
		YID yid;

		while(
		    YGetNextEvent(
			(YConnection *)snd->con_data,
			&yevent,
			False			/* Do not block */
		    ) > 0
		)
		{
		    events_handled++;

		    switch(yevent.type)
		    {
		      /* Disconnect or shutdown */
		      case YDisconnect: case YShutdown:
			YCloseConnection(
			    (YConnection *)snd->con_data,
			    False
			);
			snd->con_data = NULL;

			/* If sound was set to in the options, then notify
			 * about this lost connection
			 */
			if(snd->server_type > SOUND_SERVER_TYPE_NONE)
			{
			    snd->server_type = SOUND_SERVER_TYPE_NONE;
			}
			break;

		      /* Sound object play has stopped */
		      case YSoundObjectKill:
			yid = yevent.kill.yid;
			if(yid == (YID)snd->bg_music_playid)
			    snd->bg_music_playid = (void *)YIDNULL;
			else if(yid == (YID)snd->loop_tractor_beam_playid)
			    snd->loop_tractor_beam_playid = (void *)YIDNULL;
			break;

		      case YMixerChannel:
			break;
		    }
		}
	    }
#endif	/* HAVE_Y2 */
	    break;
  
	  case SOUND_SERVER_TYPE_ESOUND:
	    break;
 
	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}

	return(events_handled);
}

/*
 *	Deletes all sndrefs and disconnects from the sound server.
 */
void SoundShutdown(xsw_sound_struct *snd)
{
	if(snd == NULL)
	    return;

	/* Disconnect from sound server as needed */
	switch(snd->server_type)
	{
	  case SOUND_SERVER_TYPE_YIFF:
#ifdef HAVE_Y2
	    if(snd->con_data != NULL)
	    {
		YConnection *ycon = (YConnection *)snd->con_data;

#define STOP_SOUND(p)	{			\
 if(*(p) != NULL) {				\
  YDestroyPlaySoundObject(ycon, (YID)(*(p)));	\
  *(p) = NULL;					\
} }

		/* Stop the background music */
		STOP_SOUND(&snd->bg_music_playid);
		STOP_SOUND(&snd->loop_tractor_beam_playid);

#undef STOP_SOUND

		/* Close the connection to the Y sound server */
		YCloseConnection(ycon, False);
		ycon = NULL;
		snd->con_data = NULL;
	    }
#endif	/* HAVE_Y2 */
	    break;

	  case SOUND_SERVER_TYPE_ESOUND:
#ifdef ESD_H
	    snd->bg_music_playid = NULL;
	    snd->loop_tractor_beam_playid = NULL;
	    if(snd->con_data != NULL)
	    {
		SoundStopBackgroundMusic(snd);
		esd_close((int)snd->con_data);
		snd->con_data = NULL;
	    }
#endif  /* ESD_H */
	    break;

	  case SOUND_SERVER_TYPE_MIKMOD:
	    break;
	}

	free(snd->start_cmd);
	snd->start_cmd = NULL;
	free(snd->con_arg);
	snd->con_arg = NULL;
	free(snd->audio_mode_name);
	snd->audio_mode_name = NULL;
	free(snd->playid_list);
	snd->playid_list = NULL;
	snd->nplayids = 0;
}
