#include <stdlib.h>
#include <string.h>

#include "unvlist.h"


unvitem_struct *UnvItemNew(void);
void UnvItemDelete(unvitem_struct *item);

int UnvListItemIsAllocated(
	unvitem_struct **list, const int total,
	const int item_num
);

u_int8_t *UnvItemToData(unvitem_struct *item, int *len_rtn);
unvitem_struct *UnvDataToItem(const u_int8_t *data, int len);

int UnvListInsert(
	unvitem_struct ***list, int *total_rtn, 
	const int item_num,
	const char *alias,
	const char *url,
	const time_t last_connected,
	const char *comments
);
int UnvListAppend(
	unvitem_struct ***list, int *total_rtn,
	const char *alias,
	const char *url,
	const time_t last_connected,
	const char *comments
);
void UnvListClear(unvitem_struct **list, const int total);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a new universe item.
 */
unvitem_struct *UnvItemNew(void)
{
	return(UNVITEM(calloc(
	    1, sizeof(unvitem_struct)
	)));
}

/*
 *	Deletes the universe item.
 */
void UnvItemDelete(unvitem_struct *item)
{
	if(item == NULL)
	    return;

	free(item->alias);
	free(item->url);
	free(item->comments);
	free(item);
}


/*
 *	Checks if the universe item is allocated in the list.
 */
int UnvListItemIsAllocated(
	unvitem_struct **list, const int total, 
	const int item_num
)
{
	if((item_num < 0) || (item_num >= total))
	    return(0);
	else
	    return((list[item_num] != NULL) ? 1 : 0);
}

/*
 *	Converts the universe item to data.
 */
u_int8_t *UnvItemToData(unvitem_struct *item, int *len_rtn)
{
	int i, seg_len, len = 0;
	u_int8_t *data = NULL;

	if((item == NULL) || (len_rtn == NULL))
	    return(NULL);

	*len_rtn = len;

	/* Format is:
	 *
	 *	(char *)alias 0x00
	 *	(char *)url 0x00
	 *	(time_t)last_connected
	 *	(char *)comments 0x00
	 *	(unsigned int)times_connected
	 */

	/* Alias */
	i = len;
	seg_len = STRLEN(item->alias) + 1;
	len += seg_len;
	data = (u_int8_t *)realloc(
	    data, len * sizeof(u_int8_t)
	);
	if(data == NULL)
	    return(NULL);
	if(item->alias != NULL)
	    memcpy(data + i, item->alias, seg_len);

	/* URL */
	i = len;
	seg_len = STRLEN(item->url) + 1;
	len += seg_len;
	data = (u_int8_t *)realloc(
	    data, len * sizeof(u_int8_t)
	);
	if(data == NULL)
	    return(NULL);
	if(item->url != NULL)
	    memcpy(data + i, item->url, seg_len);

	/* Last Connected */
	i = len;
	seg_len = sizeof(time_t);
	len += seg_len;
	data = (u_int8_t *)realloc(
	    data, len * sizeof(u_int8_t)
	);
	if(data == NULL)
	    return(NULL);
	memcpy(data + i, &item->last_connected, seg_len);

	/* Comments */
	i = len;
	seg_len = STRLEN(item->comments) + 1;
	len += seg_len;                  
	data = (u_int8_t *)realloc(
	    data, len * sizeof(u_int8_t)
	);
	if(data == NULL)
	    return(NULL);
	if(item->comments != NULL)
	    memcpy(data + i, item->comments, seg_len);

	/* Times Connected */
	i = len;
	seg_len = sizeof(unsigned int);
	len += seg_len;
	data = (u_int8_t *)realloc(
	    data, len * sizeof(u_int8_t)
	);
	if(data == NULL)
	    return(NULL);
	memcpy(data + i, &item->times_connected, seg_len);

	*len_rtn = len;

	return(data);
}

/*
 *	Converts the data to a universe item.
 */
unvitem_struct *UnvDataToItem(const u_int8_t *data, int len)
{
	int i, seg_len;
	unvitem_struct *item;

	if((data == NULL) || (len <= 0))
	    return(NULL);

	item = UnvItemNew();
	if(item == NULL)
	    return(NULL);

	/* Format is:
	 *
	 *	(char *)alias 0x00
	 *	(char *)url 0x00  
	 *	(time_t)last_connected  
	 *	(char *)comments 0x00
	 *	(unsigned int)times_connected
	 */

	i = 0;

	/* Alias */
	seg_len = strlen((char *)(data + i)) + 1;
	if((i + seg_len) > len)
	    return(item);
	item->alias = (char *)malloc(seg_len * sizeof(char));
	if(item->alias != NULL)
	    memcpy(item->alias, data + i, seg_len);
	i += seg_len;

	/* URL */
	seg_len = strlen((char *)(data + i)) + 1;
	if((i + seg_len) > len)
	    return(item);
	item->url = (char *)malloc(seg_len * sizeof(char));
	if(item->url != NULL)
	    memcpy(item->url, data + i, seg_len);
	i += seg_len;

	/* Last connected */
	seg_len = sizeof(time_t);
	if((i + seg_len) > len)
	    return(item);
	memcpy(&item->last_connected, data + i, seg_len);
	i += seg_len;

	/* Comment */
	seg_len = strlen((char *)(data + i)) + 1;
	if((i + seg_len) > len)
	    return(item);
	item->comments = (char *)malloc(seg_len * sizeof(char));
	if(item->comments != NULL)
	    memcpy(item->comments, data + i, seg_len);
	i += seg_len;

	/* Times connected */
	seg_len = sizeof(unsigned int);
	if((i + seg_len) > len)
	    return(item);
	memcpy(&item->times_connected, data + i, seg_len);
/*	i += seg_len; */

	return(item);
}


/*
 *	Inserts a new universe item into the universes list.
 */
int UnvListInsert(
	unvitem_struct ***list, int *total_rtn,
	const int item_num,
	const char *alias,
	const char *url,
	const time_t last_connected,
	const char *comments
)
{
	int new_item_num;
	unvitem_struct *item;

	if((list == NULL) || (total_rtn == NULL))
	    return(-2);

	/* Append? */
	if((item_num < 0) || (item_num >= *total_rtn))
	{
	    const int i = MAX(*total_rtn, 0);
	    *total_rtn = i + 1;

	    *list = (unvitem_struct **)realloc(
		*list,
		(*total_rtn) * sizeof(unvitem_struct *)
	    );
	    if(*list == NULL)
	    {
		*total_rtn = 0;
		return(-3);
	    }

	    (*list)[i] = item = UnvItemNew();
	    if(item == NULL)
		return(-3);

	    new_item_num = i;
	}
	else
	{
	    /* Insert */
	    int i = MAX(*total_rtn, 0);
	    *total_rtn = i + 1;

	    item = UnvItemNew();
	    if(item == NULL)
		return(-3);

	    /* Allocate new item */
	    *list = (unvitem_struct **)realloc(
		*list,
		(*total_rtn) * sizeof(unvitem_struct *)
	    );
	    if(*list == NULL)
	    {
		*total_rtn = 0;
		UnvItemDelete(item);
		return(-3);
	    }

	    /* Shift entries */
	    for(i = (*total_rtn) - 1; i > item_num; i--)
		(*list)[i] = (*list)[i - 1];

	    /* Set new index to point to newly allocated pointer */
	    (*list)[item_num] = item;

	    new_item_num = item_num;
	}

	/* Set values */
	item->alias = STRDUP(alias);
	item->url = STRDUP(url);
	item->last_connected = last_connected;
	item->comments = STRDUP(comments);

	return(new_item_num);
}

/*
 *	Appends a new universe item to the universes list.
 */
int UnvListAppend(
	unvitem_struct ***list, int *total_rtn,
	const char *alias,
	const char *url,
	const time_t last_connected,
	const char *comments
)
{
	return(UnvListInsert(
	    list, total_rtn,
	    -1,
	    alias, url, last_connected, comments
	));
}

/*
 *	Deletes the list of universe items.
 */
void UnvListClear(unvitem_struct **list, const int total)
{
	int i;

	for(i = 0; i < total; i++)
	    UnvItemDelete(list[i]);

	free(list);
}        
