#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/cfgfmt.h"

#include "unvlist.h"
#include "unvlistwin.h"
#include "unvlistfile.h"


int UnvListFileOpen(
	const char *path,
	unvitem_struct ***list, int *total_rtn,
        void (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data,
        void (*print_warning)(const char *s)
);
int UnvListFileSave(
	const char *path,
	unvitem_struct **list, const int total,
	xsw_unvlist_win_struct *ulw,
        void (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data,
        void (*print_warning)(const char *s)
);


#define WARN_NO_SUCH_PARM(_parm_)		{	\
 char *s = (char *)malloc(				\
  STRLEN(path) + 40 + STRLEN(_parm_) + 80		\
 );							\
 sprintf(						\
  s,							\
"%s: Line %i: Unsupported parameter `%s'.",		\
  path, lines_read, (_parm_)				\
 );							\
 if(print_warning != NULL)				\
  print_warning(s);					\
 free(s);						\
}

#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Opens the universe items from the specified universe list file.
 */
int UnvListFileOpen(
	const char *path,
	unvitem_struct ***list, int *total_rtn,
        void (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data,
        void (*print_warning)(const char *s)
)
{
	int lines_read = 0;
	unsigned long file_size = 0l;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	struct stat stat_buf;

	if(progress_cb != NULL)
	    progress_cb(0l, file_size, progress_data);

	if(STRISEMPTY(path) || (list == NULL) || (total_rtn == NULL))
	    return(-2);

	/* Open the file for reading */
	fp = fopen(path, "rb");     
	if(fp == NULL)      
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		*error_msg = toupper(*error_msg);
		s = (char *)malloc(
		    STRLEN(path) + STRLEN(error_msg) + 80
		);
		sprintf(
		    s,
		    "%s: %s.",
		    path, error_msg
		);
		if(print_warning != NULL)
		    print_warning(s);
		free(s);
	    }
	    free(error_msg);
	    return(-1);
	}

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;
            if(progress_cb != NULL)
                progress_cb((unsigned long)ftell(fp), file_size, progress_data);
	}


	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';

            if(progress_cb != NULL)
                progress_cb(
		    (unsigned long)ftell(fp),
		    file_size,
		    progress_data
		);

	    /* BeginUniverseItem */
	    if(!strcasecmp(parm, "BeginUniverseItem") ||
	       !strcasecmp(parm, "BeginUniverseEntry")
	    )
	    {
		unvitem_struct *item;
		int i = UnvListAppend(
		    list, total_rtn,
		    NULL, NULL, 0l, NULL
		);
		if(i < 0)
		{
		    free(s);
		    fclose(fp);
		    return(-3);
		}

		item = (*list)[i];

		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);  
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

                    if(progress_cb != NULL)
                        progress_cb(
			    (unsigned long)ftell(fp),
			    file_size,
			    progress_data
			);

		    /* Alias */
		    if(!strcasecmp(parm, "Alias")) 
		    {
			item->alias = STRDUP(val);
		    }
		    /* URL */
		    else if(!strcasecmp(parm, "URL"))
		    {
			item->url = STRDUP(val);
		    }
		    /* LastConnected */
		    else if(!strcasecmp(parm, "LastConnected"))
		    {
			item->last_connected = atol(val);
		    }
		    /* Comments */
		    else if(!strcasecmp(parm, "Comments"))
		    {
			item->comments = STRDUP(val);
		    }
		    /* TimesConnected */
		    else if(!strcasecmp(parm, "TimesConnected"))
		    {
			item->times_connected = atoi(val);
		    }
		    /* EndUniverseItem */
		    else if(!strcasecmp(parm, "EndUniverseItem") ||
			    !strcasecmp(parm, "EndUniverseEntry")
		    )
		    {
			break;
		    }
		    /* Unknown parameter */
		    else
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }
	    /* Unknown parameter */
	    else
	    {
		WARN_NO_SUCH_PARM(parm);
	    }
	}
	free(s);

	/* Close the file */
	fclose(fp);

        if(progress_cb != NULL)
            progress_cb(file_size, file_size, progress_data);

	return(0);
}

/*
 *	Saves the universe list to the specified file.
 */
int UnvListFileSave(
	const char *path, 
	unvitem_struct **list, const int total,
	xsw_unvlist_win_struct *ulw,
        void (*progress_cb)(unsigned long, unsigned long, void *),
        void *progress_data,
        void (*print_warning)(const char *s)
)
{
	int i, n = 0;
	FILE *fp;
	list_struct *wlist;
	list_item_struct *item;
	unvitem_struct *uitem;

        if(progress_cb != NULL)
            progress_cb(0l, (unsigned long)n, progress_data);

	if(STRISEMPTY(path) || (ulw == NULL))
	    return(-2);

	/* Open the file for writing */
	fp = fopen(path, "wb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		*error_msg = toupper(*error_msg);
		s = (char *)malloc(
		    STRLEN(path) + STRLEN(error_msg) + 80
		);
		sprintf(            
		    s, 
		    "%s: %s.", 
		    path, error_msg 
		);
		if(print_warning != NULL)
		    print_warning(s);
		free(s); 
	    } 
	    free(error_msg); 
	    return(-1);
	}

	/* Header */
	fprintf(
	    fp,
"# ShipWars Universes List\n\
#\n"
	);

	/* Save by order on the universe list window */
	wlist = &ulw->list;
	n = wlist->total_items;
	for(i = 0; i < n; i++)
	{
	    item = wlist->item[i];
	    if(item == NULL)
		continue;

	    uitem = UNVITEM(ListGetItemData(wlist, i));
	    if(uitem == NULL)
		continue;  

            if(progress_cb != NULL)
                progress_cb(
		    (unsigned long)i, (unsigned long)n,
		    progress_data
		);

	    fprintf(fp, "BeginUniverseItem\n");

	    /* Alias */
	    if(!STRISEMPTY(uitem->alias))
		fprintf(
		    fp,
		    "\tAlias = %s\n",
		    item->text	/* Use name on list item */
		);
	    /* URL */
	    if(!STRISEMPTY(uitem->url))
		fprintf(
		    fp,
		    "\tURL = %s\n",
		    uitem->url
		);
	    /* LastConnected */
	    fprintf(
		fp,
		"\tLastConnected = %lu\n",
		uitem->last_connected
	    );
	    /* Comments */
	    if(!STRISEMPTY(uitem->comments))
		fprintf(
		    fp,
		    "\tComments = %s\n",
		    uitem->comments
		);
	    /* TimesConnected */
	    fprintf(
		fp,
		"\tTimesConnected = %i\n",
		uitem->times_connected
	    );

	    fprintf(fp, "EndUniverseItem\n\n");
	}

	fprintf(fp, "\n");

	/* Close the file */
	fclose(fp);

        if(progress_cb != NULL) 
            progress_cb((unsigned long)n, (unsigned long)n, progress_data);

	return(0);
}
