#include <stdlib.h>
#include <string.h>

#include "xsw.h"


int VSLabelIsAllocated(int n);
int VSLabelIsLoaded(int n);
int VSLabelGetByPointer(xsw_object_struct *obj);
int VSLabelGetHighest(void);

int VSLabelAdd(
	const char *text,
	const WColorStruct *c_fg,
	const WColorStruct *c_empire,
	font_t *font,
	int obj_num, xsw_object_struct *obj
);

void VSLabelDelete(int n);
void VSLabelDeleteByObjectPtr(xsw_object_struct *obj);
void VSLabelDeleteAll(void);
void VSLabelReclaimMemory(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)
#define IS_NUM_ODD(n)	((long)(n) & 1)


/*
 *	Checks if label n is allocated.
 */
int VSLabelIsAllocated(int n)
{
	if((vs_object_label == NULL) ||
	   (n < 0) ||
	   (n >= total_vs_object_labels)
	)
	    return(0);
	else if(vs_object_label[n] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Checks if label n is allocated and loaded.
 */
int VSLabelIsLoaded(int n)
{
	if(!VSLabelIsAllocated(n))
	    return(0);
	else if(vs_object_label[n]->image == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Returns the label number matching the object pointer.
 *	If no label can be found, -1 is returned.
 *	If a index number of 0 or greater is returned, that label
 *	can be assumed allocated and loaded.
 */
int VSLabelGetByPointer(xsw_object_struct *obj)
{
	int i;
	vs_object_label_struct *label;

	if(obj == NULL)
	    return(-1);

	for(i = 0; i < total_vs_object_labels; i++)
	{
	    label = vs_object_label[i];
	    if(label == NULL)
		continue;

	    if(label->obj == obj)
	    {
		if(VSLabelIsLoaded(i))
		    return(i);
		else
		    continue;
	    }
	}

	return(-1);
}

/*
 *	Returns the highest allocated label number that has an image
 *	loaded or -1 for none.
 */
int VSLabelGetHighest(void)
{
	int i, n;
	vs_object_label_struct *label;

	for(i = 0, n = -1; i < total_vs_object_labels; i++)
	{
	    label = vs_object_label[i];
	    if(label == NULL)
		continue;

	    if(label->image == NULL)
		continue;

	    n = i;
	}

	return(n);
}

/*
 *	Adds a new or replaces an existing label.
 */
int VSLabelAdd(
	const char *text,
        const WColorStruct *c_fg,
        const WColorStruct *c_empire,
	font_t *font,
	int obj_num, xsw_object_struct *obj
)
{
	int	i, k, n,
		lbearing, rbearing, swidth, ascent, descent,
		font_height;
	unsigned int width, height;
	char *new_text;
	pixel_t fg_pix, bg_pix;
	font_t *prev_font;
	pixmap_t pixmap;
	image_t *img;
	vs_object_label_struct *label;
	osw_gui_struct *gui = &osw_gui[0];

	if(obj == NULL)
	    return(-1);

	if(total_vs_object_labels < 0)
	    total_vs_object_labels = 0;

	/* Look for an available pointer */
	k = -1;
	for(i = 0; i < total_vs_object_labels; i++)
	{
	    label = vs_object_label[i];
	    if(label == NULL)
	    {
		k = i;
		break;
	    }

	    /* Check if the object already has a label */
	    if(label->obj == obj)
	    {
		/* Delete this label */
		VSLabelDelete(i);
		k = i;
		break;
	    }
	}
	/* Got available pointer? */
	if(k > -1)
	{
	    n = k;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_vs_object_labels;
	    total_vs_object_labels = n + 1;
	    vs_object_label = (vs_object_label_struct **)realloc(
		vs_object_label,
		total_vs_object_labels * sizeof(vs_object_label_struct *)
	    );
	    if(vs_object_label == NULL)
	    {
		total_vs_object_labels = 0;
		return(-3);
	    }
	    vs_object_label[n] = NULL;
	}

	/* Allocate new label */
	vs_object_label[n] = label = (vs_object_label_struct *)calloc(
	    1, sizeof(vs_object_label_struct)
	);
	if(label == NULL)
	    return(-3);

	/* Set object number and pointer */
	label->obj_num = obj_num;
	label->obj = obj;

	/* Format text */
	if(text == NULL)
	    text = "(null)";
	new_text = (char *)malloc(
	    STRLEN(text) + STRLEN(obj->empire) + 80
	);
	sprintf(
	    new_text,
	    "(%s) %s",
	    obj->empire,
	    text
	);

	/* Calculate size */
	prev_font = OSWQueryCurrentFont();
	OSWSetFont(font);
	OSWGetTextBounds(
	    font, new_text, -1,
	    &lbearing, &rbearing, &swidth,
	    &ascent, &descent
	);
	font_height = ascent + descent;
	width = (unsigned int)swidth + 4;
	height = (unsigned int)font_height + 4;

	/* Width and height must be even numbers */
	if(IS_NUM_ODD(width))
	    width += 1;
	if(IS_NUM_ODD(height))
	    height += 1;

	if(OSWCreatePixmap(&pixmap, width, height))
	{
	    VSLabelDelete(n);
	    free(new_text);
	    OSWSetFont(prev_font);
	    return(-1);
	}

	/* Allocate pixels */
	OSWLoadPixelRGB(&fg_pix, c_fg->r, c_fg->g, c_fg->b);
	OSWLoadPixelRGB(&bg_pix, 0x10, 0x10, 0x10);	/* Overriding bg */

	/* Clear background with black color so that it will be
	 * detected as transparent when converted to an image
	 */
	OSWClearPixmap(pixmap, width, height, gui->black_pix);

	/* Draw text outline */
	OSWSetForeground(bg_pix);
#define DRAW_STRING(_x_,_y_)				\
 OSWDrawString(						\
  pixmap, (_x_) - lbearing, (_y_) + ascent, new_text	\
 );
	DRAW_STRING(1, 0);
	DRAW_STRING(2, 1);
	DRAW_STRING(1, 2);
	DRAW_STRING(0, 1);
#undef DRAW_STRING

	/* Draw text foreground */
	OSWSetForeground(fg_pix);
	OSWDrawString(
	    pixmap,
	    1 - lbearing,
	    1 + ascent,
	    new_text
	);

	free(new_text);		/* No longer needed */

	/* Set back previous font */
	OSWSetFont(prev_font);

	/* Unload pixels */
	OSWDestroyPixel(&fg_pix);
	OSWDestroyPixel(&bg_pix);

	/* Copy graphics from the pixmap to the image */
	img = OSWGetImage(pixmap, 0, 0, width, height);
	OSWDestroyPixmap(&pixmap);

	/* Set image */
	label->image = img;

	return(0);
}

/*
 *	Deletes the label.
 */
void VSLabelDelete(int n)
{
	vs_object_label_struct *label;

	if(!VSLabelIsAllocated(n))
	    return;

	label = vs_object_label[n];
	label->obj_num = -1;
	label->obj = NULL;
	OSWDestroyImage(&label->image);
	free(label);
	vs_object_label[n] = NULL;
}

/*
 *	Deletes all labels matching the specified object.
 */
void VSLabelDeleteByObjectPtr(xsw_object_struct *obj)
{
	int i;
	vs_object_label_struct *label;

	if(obj == NULL)
	    return;

	for(i = 0; i < total_vs_object_labels; i++)
	{
	    label = vs_object_label[i];
	    if(label == NULL)
		continue;

	    if(label->obj == obj)
		VSLabelDelete(i);
	}
}

/*
 *	Deletes all labels.
 */
void VSLabelDeleteAll() 
{
	int i;

	for(i = 0; i < total_vs_object_labels; i++)
	    VSLabelDelete(i);
	free(vs_object_label);
	vs_object_label = NULL;
	total_vs_object_labels = 0;
}

/*
 *	Deletes all unallocated pointers in the list.
 */
void VSLabelReclaimMemory(void)
{
	int i, highest;

	/* Get highest allocated label */
	highest = VSLabelGetHighest();

	/* If there are no allocated labels then delete the list */
	if(highest < 0)
	{
	    VSLabelDeleteAll();
	    return;
	}

	/* Delete excess pointers */
	for(i = highest + 1; i < total_vs_object_labels; i++)
	    VSLabelDelete(i);

	/* Reallocate pointers */
	total_vs_object_labels = highest + 1;

	if(total_vs_object_labels > 0)
	{
	    vs_object_label = (vs_object_label_struct **)realloc(
	        vs_object_label,
	        total_vs_object_labels * sizeof(vs_object_label_struct *)
	    );
	    if(vs_object_label == NULL)
	    {
	        total_vs_object_labels = 0;
		return;
	    }
	}
	else
	{
	    free(vs_object_label);
	    vs_object_label = NULL;
	    total_vs_object_labels = 0;
	}
}
