#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "../include/widget.h"


static void RadioResizeCB(void *widget, void *data);
static void RadioDrawCB(void *widget, void *data);

int RadioGetActive(radio_struct *radio);
void RadioSetActive(radio_struct *radio, int n);

int RadioInit(
	radio_struct *radio, win_t parent,
	int x, int y,
	int ncheck_buttons,
	int start_sel_btn,
	char **names, int nnames,
	int alignment
);
void RadioResize(radio_struct *radio);
void RadioQueueResize(radio_struct *radio);
void RadioDraw(radio_struct *radio);
void RadioQueueDraw(radio_struct *radio);
int RadioManage(
	radio_struct *radio, event_t *xevent
);
void RadioMap(radio_struct *radio);
void RadioUnmap(radio_struct *radio);
void RadioDestroy(radio_struct *radio);


#ifndef MAX
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#endif


/*
 *	Size constants (in pixels):
 */
#define TB_MARGIN	2

#define TB_MIN_WIDTH	((2 * TB_MARGIN) + 2)
#define TB_MIN_HEIGHT	((2 * TB_MARGIN) + 2)

#define TB_CHAR_WIDTH   7
#define TB_CHAR_HEIGHT  14


#define TBA_MARGIN	2

#define TBA_MIN_WIDTH	((2 * TB_MARGIN) + 2)
#define TBA_MIN_HEIGHT	((2 * TB_MARGIN) + 2)


/*
 *	Resize callback.
 */
static void RadioResizeCB(void *widget, void *data)
{
	radio_struct *radio = (radio_struct *)widget;
	if(radio == NULL)
	    return;

	RadioResize(radio);
}

/*
 *	Draw callback.
 */
static void RadioDrawCB(void *widget, void *data)
{
	radio_struct *radio = (radio_struct *)widget;
	if(radio == NULL)
	    return;

	if(radio->map_state)
	    RadioDraw(radio);
}


/*
 *	Gets the selected button on the radio.
 */
int RadioGetActive(radio_struct *radio)
{
	return((radio != NULL) ? radio->armed_check_button : 0);
}

/*
 *	Sets the selected button on the radio.
 */
void RadioSetActive(radio_struct *radio, int n)
{
	if(radio == NULL)
	    return;

	if(radio->armed_check_button != n)
	{
	    int i, m = radio->ncheck_buttons;
	    check_button_struct *cb;

	    radio->armed_check_button = n;

	    for(i = 0; i < m; i++)
	    {
		cb = radio->check_button[i];
		if(cb == NULL)
		    continue;

		CBtnSetActive(cb, (i == n) ? True : False);
	    }
	}
}


/*
 *	Initializes a radio.
 */
int RadioInit(
	radio_struct *radio, win_t parent,
	int x, int y,
	int ncheck_buttons,
	int start_sel_btn,
	char **names, int nnames,
	int alignment
)
{
	int i, swidth, ascent, descent, font_height, longest;
	unsigned int width, height;
	image_t *img = NULL;
	font_t *font;
	win_t w;
	check_button_struct *cb;
	widget_global_struct *wglobal = &widget_global;

	if((radio == NULL) || (parent == None) ||
	   (names == NULL) || (ncheck_buttons == 0)
	)
	    return(-2);

	img = wglobal->check_button_unarmed_img;

	radio->type = WIDGET_TYPE_RADIO;
	radio->map_state = False;
	radio->has_focus = False;
	radio->disabled = False;
	radio->visibility_state = VisibilityFullyObscured;
	radio->x = x;
	radio->y = y;
	radio->width = 1;
	radio->height = 1;
	radio->font = font = wglobal->std_font;
	radio->next = NULL;
	radio->prev = NULL;
	radio->data = NULL;

	radio->align = alignment;
	radio->ncheck_buttons = ncheck_buttons;
	radio->armed_check_button = start_sel_btn;


	/* Calculate width and height */
	switch(alignment)
	{
	  /* Vertical alignment */
	  case RADIO_ALIGN_VERTICAL:
	    longest = 0;
	    font_height = 0;
	    for(i = 0; i < (int)nnames; i++)
	    {
		if(names[i] == NULL)
		    break;

		OSWGetTextBounds(
		    font, names[i], -1,
		    NULL, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;

		if(swidth > longest)
		    longest = swidth;
	    }
	    width = ((img != NULL) ? img->width : 16) +
		longest + (3 * TB_MARGIN);
	    height = (MAX(
		font_height,
		((img != NULL) ? (int)img->height : 16)
	    ) +	(int)(TB_MARGIN * 2)) * ncheck_buttons;
	    break;

	  /* Horizontal alignment */
	  case RADIO_ALIGN_HORIZONTAL:
	    width = 0;
	    font_height = 0;
	    for(i = 0; i < ncheck_buttons; i++)
	    {
		if(i >= nnames)
		    break;

		OSWGetTextBounds(
		    font, names[i], -1,
		    NULL, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;

		width += swidth +
		    (int)(TB_MARGIN * 5) +
		    ((img != NULL) ? img->width : 16);
	    }
	    height = MAX(
		    font_height,
		    ((img != NULL) ? (int)img->height : 16)
		) + (2 * TB_MARGIN);
	    break;

	  /* Cascade alignment (not supported yet) */
	  case RADIO_ALIGN_CASCADE:
	    fprintf(
		stderr,
"RadioInit(): Error: Alignment %i not supported.\n",
		alignment
	    );
	    return(-1);
	    break;

	  default:
	    fprintf(
		stderr,
"RadioInit(): Error: Alignment %i not supported.\n",
		alignment
	    );
	    return(-1);
	    break;
	}

	radio->width = MAX(width, TBA_MIN_WIDTH);
	radio->height = MAX(height, TBA_MIN_HEIGHT);

	/* Create toplevel */
	x = radio->x;
	y = radio->y;
	width = radio->width;
	height = radio->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	radio->toplevel = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask
	);
	parent = w;

	/* Create the toggle buttons */
	radio->check_button = (check_button_struct **)calloc(
	    radio->ncheck_buttons, sizeof(check_button_struct *)
	);
	if(radio->check_button == NULL)
	{
	    radio->ncheck_buttons = 0;
	    return(-3);
	}

	switch(alignment)
	{
	  /* Vertical alignment */
	  case RADIO_ALIGN_VERTICAL:
	    x = 0;
	    y = 0;
	    for(i = 0; i < radio->ncheck_buttons; i++)
	    {
		if(i >= (int)nnames)
		    break;

		radio->check_button[i] = cb = (check_button_struct *)calloc(
		    1, sizeof(check_button_struct)
		);
		if(cb == NULL)
		    return(-3);

		OSWGetTextBounds(
		    font, names[i], -1,
		    NULL, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;

		CBtnInit(
		    cb, parent,
		    x, y,
		    False,
		    names[i]
		);

		y += MAX(
		    font_height,
		    ((img != NULL) ? (int)img->height : 16)
		) + (2 * TB_MARGIN);
	    }
	    break;

	  /* Horizontal alignment */
	  case RADIO_ALIGN_HORIZONTAL:
	    x = 0;
	    y = 0;
	    for(i = 0; i < radio->ncheck_buttons; i++)
	    {
		if(i >= (int)nnames)
		    break;

		radio->check_button[i] = cb = (check_button_struct *)calloc(
		    1, sizeof(check_button_struct)
		);
		if(cb == NULL)
		    return(-3);

		OSWGetTextBounds(
		    font, names[i], -1,
		    NULL, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;

		CBtnInit(
		    cb, parent,
		    x, y,
		    False,
		    names[i]
		);

		x += (5 * TB_MARGIN) +
		    ((img != NULL) ? (int)img->width : 16) +
		    swidth;
	    }
	    break;

	  case RADIO_ALIGN_CASCADE:
	    break;
	}

	/* Set selected armed button */
	for(i = 0; i < radio->ncheck_buttons; i++)
	{
	    cb = radio->check_button[i];
	    if(cb == NULL)
	        continue;

	    if(i == radio->armed_check_button)
		cb->state = True;
	    else
		cb->state = False;
	}

	/* Add the widget to the regeristry */
	WidgetRegister(radio, radio->type);

	return(0);
}

/*
 *	Resizes the radio.
 */
void RadioResize(radio_struct *radio)
{
	int i, x, y, swidth, ascent, descent, font_height;
	unsigned int width, height;
	win_t w;
	image_t *img;
	check_button_struct *cb;
	widget_global_struct *wglobal = &widget_global;

	if(radio == NULL)
	    return;

	w = radio->toplevel;
	OSWGetWindowGeometry(w,  &x, &y, &width, &height, NULL);
	radio->x = x;
	radio->y = y;
	radio->width = width;
	radio->height = height;

	img = wglobal->check_button_unarmed_img;

	switch(radio->align)
	{
	  /* Vertical alignment */
	  case RADIO_ALIGN_VERTICAL:
	    x = 0;
	    y = 0;
	    for(i = 0; i < radio->ncheck_buttons; i++)
	    {
		cb = radio->check_button[i];
		if(cb == NULL)
		    continue;

		w = cb->toplevel;
		if(w == None)
		    continue;

		OSWGetTextBounds(
		    cb->font, cb->label, -1,
		    NULL, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;

		OSWMoveWindow(w, x, y);
		WidgetQueueResize(cb);

		y += MAX(
		    font_height,
		    ((img != NULL) ? (int)img->height : 16)
		) + (2 * TB_MARGIN);
	    }
	    break;

	  /* Horizontal alignment */
	  case RADIO_ALIGN_HORIZONTAL:
	    x = 0;
	    y = 0;
	    for(i = 0; i < radio->ncheck_buttons; i++)
	    {
		cb = radio->check_button[i];
		if(cb == NULL)
		    continue;

		w = cb->toplevel;
		if(w == None)
		    continue;

		OSWGetTextBounds(
		    cb->font, cb->label, -1,
		    NULL, NULL, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;

		OSWMoveWindow(w, x, y);
		WidgetQueueResize(cb);
		  
		x += (5 * TB_MARGIN) +
		    ((img != NULL) ? (int)img->width : 16) +
		    swidth;
	    }
	    break;
	}
}

/*
 *	Queues the radio to resize.
 */
void RadioQueueResize(radio_struct *radio)
{
	if(radio == NULL)
	    return;

	WidgetQueueResizeAction(RadioResizeCB, radio, NULL);
}

/*
 *	Draws the radio.
 */
void RadioDraw(radio_struct *radio)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(radio == NULL)
	    return;

	if(radio->toplevel != None)
	{
	    unsigned int width, height;
	    win_t w = radio->toplevel;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height 
		);
	    }
	}

}

/*
 *	Queues the radio to draw.
 */
void RadioQueueDraw(radio_struct *radio)
{
	if(radio == NULL)
	    return;

	WidgetQueueDrawAction(RadioDrawCB, radio, NULL);
}

/*
 *	Manages the radio event.
 */
int RadioManage(
	radio_struct *radio, event_t *event
)
{
	int i, events_handled = 0;
	win_t w;
	check_button_struct *cb;

	if((radio == NULL) || (event == NULL))
	    return(events_handled);

	if(!radio->map_state &&
	   (event->type != MapNotify)
	)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == radio->toplevel)
	    {
		if(radio->map_state)
		    RadioQueueDraw(radio);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == radio->toplevel)
	    {
		RadioQueueDraw(radio);
		events_handled++;
	    }
	    break;

	  case ButtonPress:
	    /* Skip if the entire radio is disabled */
	    if(radio->disabled)
		return(events_handled);

	    for(i = 0; i < radio->ncheck_buttons; i++)
	    {
		cb = radio->check_button[i];
		if(cb == NULL)
		    continue;

		if(cb->disabled)
		    continue;

		if(w == cb->toplevel)
		{
		    radio->armed_check_button = i;
		    events_handled++;
		}
	    }
	    if(events_handled > 0)
	    {
		for(i = 0; i < radio->ncheck_buttons; i++)
		{
		    cb = radio->check_button[i];
		    if(cb == NULL)
			continue;

		    if(cb->disabled)
			continue;

		    if(radio->armed_check_button == i)
		    {
			cb->has_focus = True;
			cb->state = True;
			WidgetQueueDraw(cb);
		    }
		    else
		    {
			if(cb->has_focus)
			{
			    cb->has_focus = False;
			    WidgetQueueDraw(cb);
			}
			if(cb->state)
			{
			    cb->state = False;
			    WidgetQueueDraw(cb);
			}
		    }
		}
	    }
	    break;

	  case VisibilityNotify:
	    if(w == radio->toplevel)
	    {
		radio->visibility_state = event->xvisibility.state;
		events_handled++;
	    }
	    break;
	}

	/* Manage events on each toggle button */
	if(events_handled == 0)
	{
	    for(i = 0; i < radio->ncheck_buttons; i++)
	    {
		events_handled += WidgetManage(radio->check_button[i], event);
		if(events_handled > 0)
		    break;
	    }
	}

	return(events_handled);
}


/*
 *	Maps the radio.
 */
void RadioMap(radio_struct *radio)
{
	if(radio == NULL)
	    return;

	if(!radio->map_state)
	{
	    int i, m = radio->ncheck_buttons;
	    check_button_struct *cb;

	    OSWMapRaised(radio->toplevel);
	    radio->map_state = True;

	    for(i = 0; i < m; i++)
	    {
		cb = radio->check_button[i];
		if(cb == NULL)
		    continue;

		CBtnSetActive(
		    cb, (radio->armed_check_button == i) ? True : False
		);
		WidgetMap(cb);
	    }
	}
}

/*
 *	Unmaps the radio.
 */
void RadioUnmap(radio_struct *radio)
{
	if(radio == NULL)
	    return;

	OSWUnmapWindow(radio->toplevel);
	radio->map_state = False;
}

/*
 *	Destroys the radio.
 */
void RadioDestroy(radio_struct *radio)
{
	int i;

	if(radio == NULL)
	    return;

	/* Destroy each toggle button */
	for(i = 0; i < radio->ncheck_buttons; i++)
	{
	    if(radio->check_button[i] == NULL)
		continue;

	    /* Destroy toggle button */
	    WidgetDestroy(radio->check_button[i]);

	    /* Free toggle button structure */
	    free(radio->check_button[i]);
	}
	free(radio->check_button);
	radio->check_button = NULL;

	radio->ncheck_buttons = 0;
	radio->armed_check_button = -1;

	OSWDestroyWindow(&radio->toplevel);

	radio->type = WIDGET_TYPE_UNKNOWN;
	radio->map_state = False;
	radio->has_focus = False;
	radio->disabled = False;
	radio->x = 0;
	radio->y = 0;
	radio->width = 0;
	radio->height = 0;
	radio->visibility_state = VisibilityFullyObscured;
	radio->font = NULL;
	radio->next = NULL;
	radio->prev = NULL;
}
