#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/swnetcodes.h"
#include "../include/swsoundcodes.h"

#include "../include/unvmain.h"
#include "../include/unvmath.h"
#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "particle.h"
#include "msgwin.h"
#include "unvlist.h"
#include "unvlistwin.h"
#include "ecowin.h"
#include "starchartwin.h"
#include "sound.h"
#include "net.h"
#include "xsw.h"


int XSWConnect(const char *url);
int XSWConnectLast(void);
void XSWSendDisconnect(void);
void XSWDisconnect(void);
void XSWRefresh(
	const time_t cur_ms, const time_t cur_sec, const float tc
);

void XSWObjectSectorChangedNotify(
	const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

void XSWDamageNotify(
	const int obj_num,
	const float prev_hp, const float hp,
	const float prev_power, const float power,
	const int shield_state
);
void XSWHitNotify(
	const int src_obj_num,		/* Object that hit tar_obj_num */
	const int tar_obj_num,		/* Object that got hit */
	const int owner_obj_num,	/* Owner of src_obj_num */
	const float dmg_structure,
	const float dmg_shield,
	const float bearing		/* Target to source, in radians */
);
void XSWDestroyNotify(
	const int src_obj_num,		/* Object that caused destruction */
	const int tar_obj_num,		/* Object that was destroyed */
	const int owner_obj_num,	/* Owner of src_obj_num */
	const int reason		/* One of SW_DESTROY_REASON_* */
);

static bool_t XSWUpdateLoopSoundsTractorBeamState(
	const int obj_num, xsw_object_struct *obj,
	float *vol_left_rtn, float *vol_right_rtn
);
void XSWUpdateLoopSounds(void);
void XSWUpdateBackgroundMusic(void);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Connects to the universe specified by url.
 *
 *	The url can be either a url in the format of:
 *
 *	"sw://[<name>:<password>@]<address>[:port]"
 *
 *	Or the name of a universe item.
 *
 */
int XSWConnect(const char *url)
{
	int port, status;
	char *s, *durl, *prev_address;
	xsw_net_parms_struct *net = &net_parms;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_option_struct *opt = &option;

	if((bw == NULL) || STRISEMPTY(url))
	    return(-2);

	durl = strdup(url);
	if(durl == NULL)
	    return(-3);

	strstrip(durl);

	/* Is this a URL? */
	if(strcasepfx(durl, "sw://") ||
	   strcasepfx(durl, "shipwars://") ||
	   strcasepfx(durl, "swserv://")
	)
	{


	}
	else
	{
	    int i;
	    unvitem_struct *item;

	    /* Not a URL, so assume it is the name of a universe
	     * item in the universes list
	     */
	    for(i = 0; i < total_univ_items; i++)
	    {
		item = univ_item[i];
		if(item == NULL)
		    continue;

		if(STRISEMPTY(item->alias))
		    continue;

		if(!strcasecmp(durl, item->alias))
		{
		    free(durl);
		    durl = STRDUP(item->url);
		    strstrip(durl);
		    break;
		}
	    }

	    /* No such universe item? */
	    if(i >= total_univ_items)
	    {
		s = (char *)malloc(STRLEN(durl) + 80);
		sprintf(
		    s,
"%s: No such universe.",
		    durl
		);
		XSWPrintWarning(s);
		free(s);

		s = (char *)malloc(STRLEN(durl) + 80);
		sprintf(
		    s,
"No such universe:\n\
\n\
    %s",
		    durl
		);
		XSWDialog(
		    "Connect Failed",
		    XSW_IMG_DIALOG_ICON_WARNING,
		    s,
		    bw->toplevel
		);
		free(s);

		free(durl);
		return(-1);
	    }
	}

	/* Record the current address */
	prev_address = STRDUP(net->address);

	/* Get the address from the URL (must be valid) */
	s = StringParseAddress(durl);
	if(s != NULL)
	{
	    free(net->address);
	    net->address = strdup(s);
	}
	else
	{
	    s = (char *)malloc(STRLEN(durl) + 80);
	    sprintf(
		s,
"%s: Unable to parse the URL.",
		durl
	    );
	    XSWPrintWarning(s);
	    free(s);

	    s = (char *)malloc(STRLEN(durl) + 80);
	    sprintf(
		s,
"Unable to parse the URL:\n\
\n\
    %s",
		durl
	    );
	    XSWDialog(
		"Connect Failed",
		XSW_IMG_DIALOG_ICON_WARNING,
		s,
		bw->toplevel
	    );
	    free(s);

	    free(durl);
	    free(prev_address);

	    return(-2);
	}

	/* Get the login name from the URL */
	s = StringParseName(durl);
	if(s != NULL)
	{
	    free(net->login_name);
	    net->login_name = strdup(s);
	}

	/* Get the password from the URL */
	s = StringParsePassword(durl);
	if(s != NULL)
	{
	    free(net->login_password);
	    net->login_password = strdup(s);
	}

	/* Get the port number from the URL */
	port = StringParsePort(durl);
	net->port = (port > -1) ? port : XSW_DEF_SWSERV_PORT;

	/* Currently connected? */
	if(XSW_CON_IS_CONNECTED(net))
	{
	    /* Disconect first
	     *
	     * Shutdown the socket that is connected to the server and
	     * wait for the closing of the socket
	     */
	    if(!XSW_CON_STATE_IS_CLOSE_WAIT(net))
	    {
		if(net->disconnect_send_count == 0)
		{
		    char *s;
		    s = (char *)malloc(STRLEN(prev_address) + 80);
		    sprintf(
			s,
"Disconnecting from %s...",
			prev_address
		    );
		    XSWPrint(s);
		    free(s);
		}

		NetSendLogout();
	    }
/* If the logout was never received by the server then there is a risk
 * of infiniate loop here because the socket will never be closed
 */
	    OSWBlockInput(True);
	    while(XSW_CON_IS_CONNECTED(net))
		XSWMainIteration(NULL);
	    OSWBlockInput(False);

	    /* Delete the ignored objects list */
	    IgnoredObjListClear(ignored_obj_list, total_ignored_objs);
	    ignored_obj_list = NULL;
	    total_ignored_objs = 0;

	    /* Delete all the objects */
	    DBDeleteAllObjects();
	    ParticleClear(&xsw_particle, &total_particles);

	    /* Reset the universe legend */
	    UNVLegendReset(&sw_legend);

	    /* Clear economy window */
	    EcoWinClearProducts(xsw_ecow);
	}

	/* Connect */
	status = NetConnect(net->address, net->port, opt->net_timeout);
	if(status < 0)
	{
	    /* Failed to connect */

	    /* Update the bridge window's title */
	    BridgeWinSetTitle(bw, NULL, NULL, NULL);
	}
	else
	{
	    /* Successful connect */

	    /* Update the bridge window's title */
	    BridgeWinSetTitle(bw, net->address, NULL, NULL);

	    s = (char *)malloc(STRLEN(net->address) + 80);
	    sprintf(
		s,
		"Connecting to %s...",
		(net->address != NULL) ? net->address : ""
	    );
	    XSWPrint(s);
	    free(s);

	    /* Clear starchart on successful connect? */
	    if(option.clear_chart_on_connect)
		SChtClear(xsw_scht);
	}

	/* Redraw the bridge window */
	BridgeWinDrawAll(bw);

	free(durl);
	free(prev_address);

	return(status);
}

/*
 *	Connects to the last universe.
 *
 *	Uses the last specified login name and password.
 */
int XSWConnectLast(void)
{
	int status;
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	/* Already connected? */
	if(XSW_CON_IS_CONNECTED(net))
	    return(0);

	/* Connect using last address and port */
	status = NetConnect(net->address, net->port, opt->net_timeout);
	if(status < 0)
	{
	    /* Could not connect */

	    /* Update the bridge window's title */
	    BridgeWinSetTitle(xsw_bridge_win, NULL, NULL, NULL);
	}
	else
	{
	    /* Successful connect */
	    char *s;

	    /* Update the bridge window's title */
	    BridgeWinSetTitle(xsw_bridge_win, net->address, NULL, NULL);

	    s = (char *)malloc(strlen(net->address) + 80);
	    sprintf(
		s, 
		"Connecting to %s...",
		(net->address != NULL) ?
		    net->address : ""
	    );
	    XSWPrint(s);
	    free(s);

	    /* Clear starchart on successful connect? */
	    if(option.clear_chart_on_connect)
		SChtClear(xsw_scht);
	}

	/* Redraw the bridge window */
	BridgeWinDrawAll(xsw_bridge_win);

	return(status);
}

/*
 *	Sends a disconnect to the universe and increments the disconnect
 *	send count. If the disconnect send count has exceeded the limit
 *	then the connection's socket will be shut down.
 *
 *	This does not actually close the socket to the connection, see
 *	XSWDisconnect().
 */
void XSWSendDisconnect(void)
{
	const int disconnect_send_try_limit = 3;
	xsw_net_parms_struct *net = &net_parms;

	/* Not connected? */
	if(!XSW_CON_IS_CONNECTED(net) ||
	   XSW_CON_STATE_IS_NOT_CONNECTED(net)
	)
	    return;

	if(net->disconnect_send_count >= disconnect_send_try_limit)
	{
	    char *s, *address = STRDUP(net->address);

	    s = (char *)malloc(STRLEN(address) + 80);
	    sprintf(
		s,
"Forcing disconnect from %s...",
		address
	    );
	    XSWPrintWarning(s);
	    free(s);

	    XSWDisconnect();

	    s = (char *)malloc(STRLEN(address) + 80);
	    sprintf(
		s,
"Forceably disconnected from %s.",
		address
	    );
	    XSWPrintWarning(s);
	    free(s);
	    free(address);
	}
	else
	{
	    if(!XSW_CON_STATE_IS_CLOSE_WAIT(net))
	    {
		if(net->disconnect_send_count == 0)
		{
		    char *s, *address = STRDUP(net->address);
		    s = (char *)malloc(STRLEN(address) + 80);
		    sprintf(
			s,
"Disconnecting from %s...",
			address
		    );
		    XSWPrint(s);
		    free(s);
		    free(address);
		}

		NetSendLogout();
		if(!XSW_CON_IS_CONNECTED(net))
		    return;
	    }
	    net->disconnect_send_count++;
	}
}

/*
 *	Disconnects from the universe.
 *
 *	Forceably shuts down and closes the connection's socket and sets
 *	everything up to the non-connected state and returns the bridge
 *	window to the menus.
 */
void XSWDisconnect(void)
{
	scale_bar_struct *scale;
	xsw_gctl_struct *gctl;
	xsw_player_values_struct *pvals = &player_values;
	xsw_bridge_win_struct *bw = xsw_bridge_win;
	xsw_option_struct *opt = &option;

	/* Delete the ignored objects list */
	IgnoredObjListClear(ignored_obj_list, total_ignored_objs);
	ignored_obj_list = NULL;
	total_ignored_objs = 0;

	/* Delete all the objects */
	DBDeleteAllObjects();
	ParticleClear(&xsw_particle, &total_particles);

	/* Reset the universe legend */
	UNVLegendReset(&sw_legend);

	/* Close the connection to the server and reset net values */
	NetDisconnect(opt->net_timeout);

	/* Update the bridge window's title */
	BridgeWinSetTitle(bw, NULL, NULL, NULL);

	/* Clear the economy window */
	EcoWinClearProducts(xsw_ecow);

	/* Reset game controller positions */
	gctl = xsw_gctl;
	if(gctl != NULL)
	{
	    gctl->turn = 0.0f;
	    gctl->throttle = 0.0f;
	    gctl->thrust_dir = 0.0f;
	    GCtlResetKeys(gctl);
	}

	/* Reset player values */
	pvals->weapon_fire_heading = (0.0f * PI);
	pvals->weapon_fire_pitch = (0.0f * PI);
	pvals->auto_destruct_last_sec_left = 0l;
	pvals->warning_weapons_lock = 0;
	pvals->warning_incoming_fire = 0;

	/* Reset bridge window viewscreen & scanner zoom */
	if(bw != NULL)
	{
	    bw->viewscreen_zoom = XSW_VS_ZOOM_MIN;
	}

	scale = &bw->scanner_sb;
	scale->pos = 80.0f;
	BridgeWinScannerScaleValueChangedCB(scale, scale->pos, bw);
	ScaleBarQueueDraw(scale);

	/* Redraw the bridge window */
	BridgeWinDrawAll(bw);
}

/*
 *	Refresh.
 */
void XSWRefresh(
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_net_parms_struct *net = &net_parms;

	/* Clear the starchart */
	SChtClear(xsw_scht);

	/* Delete all the object labels */
	VSLabelDeleteAll();

	/* Delete the scanner contacts list */
	ScListClear(scanner_contact_list, total_scanner_contacts);
	scanner_contact_list = NULL;
	total_scanner_contacts = 0;

	/* Update the in range objects list */
	DBInRangeUpdate(
	    net->player_obj_num,
	    cur_ms, cur_sec, tc
	);

	/* Connected and logged in? */
	while(XSW_CON_IS_CONNECTED(net) &&
	      XSW_CON_STATE_IS_CONNECTED(net)
	)
	{
	    /* Send refresh to server */
	    NetSendRefresh();
	    if(!XSW_CON_IS_CONNECTED(net))
		break;

	    break;
	}

	/* Redraw windows */
	BridgeWinDrawAll(xsw_bridge_win);
	if((xsw_msg_win != NULL) ?
	    xsw_msg_win->map_state : False
	)
	    MsgWinQueueDraw(xsw_msg_win);
}


/*
 *	Updates the object when it has changed sectors.
 *
 *	This function should be called whenever any object changes
 *	sectors.
 */
void XSWObjectSectorChangedNotify(
	const int obj_num,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Is this the player object? */
	if(obj == net_parms.player_obj_ptr)
	{
	    ScHandleContacts(
		&scanner_contact_list, &total_scanner_contacts,
		obj_num
	    );
	    DBReclaim(cur_ms, cur_sec, tc);
	    DBInRangeUpdate(
		obj_num,
		cur_ms, cur_sec, tc
	    );
	}
}

/*
 *	Damage notify.
 *
 *	This function should be called whenever an object has received
 *	damage.
 */
void XSWDamageNotify(
	const int obj_num,
	const float prev_hp, const float hp,
	const float prev_power, const float power,
	const int shield_state
)
{
	xsw_net_parms_struct *net = &net_parms;
	xsw_object_struct *obj;
	xsw_sound_struct *snd = &sound;
	xsw_option_struct *opt = &option;

	obj = DBIsObjectGarbage(obj_num) ?
	    NULL : xsw_object[obj_num];
	if(obj == NULL)
	    return;

	/* Is this the player object? */
	if(obj_num == net->player_obj_num)
	{
	    bool_t	hp_warning = False,
			hp_danger = False,
			shield_warning = False,
			shield_danger = False,
			shield_down = False,
			played_sound = False;
	    if(obj->hp_max > 0.0f)
	    {
		const float m = obj->hp_max;
		if(((prev_hp / m) > XSW_HP_WARNING_COEFF) &&
		   ((hp / m) <= XSW_HP_WARNING_COEFF)
		)
		    hp_warning = True;
		if(((prev_hp / m) > XSW_HP_DANGER_COEFF) &&
		   ((hp / m) <= XSW_HP_DANGER_COEFF)
		)
		    hp_danger = True;
	    }
	    if((obj->power_max > 0.0f) &&
	       (shield_state == SW_SHIELD_STATE_UP)
	    )
	    {
		const float m = obj->power_max;
		if(((prev_power / m) > XSW_POWER_WARNING_COEFF) &&
		   ((power / m) <= XSW_POWER_WARNING_COEFF)
		)
		    shield_warning = True;
		if(((prev_power / m) > XSW_POWER_DANGER_COEFF) &&
		   ((power / m) <= XSW_POWER_DANGER_COEFF)
		)
		    shield_danger = True;
		if((prev_power > 0.0f) && (power <= 0.0f) &&
		   (prev_hp >= obj->hp_max)
		)
		    shield_down = True;
	    }

	    if(shield_warning)
	    {
		XSWPrintWarning(
"Shields compromised."
		);
		if((opt->sounds >= XSW_SOUNDS_EVENTS) && !played_sound)
		{
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
		    played_sound = True;
		}
	    }
	    if(shield_danger)
	    {
		XSWPrintWarning(
"Shields failing."
		);
		if((opt->sounds >= XSW_SOUNDS_EVENTS) && !played_sound)
		{
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
		    played_sound = True;
		}
	    }
	    if(shield_down)
	    {
		XSWPrintWarning(
"Shields down."
		);
		if((opt->sounds >= XSW_SOUNDS_EVENTS) && !played_sound)
		{
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
		    played_sound = True;
		}
	    }
	    if(hp_warning)
	    {
		XSWPrintWarning(
"Hull structure compromised."
		);
		if((opt->sounds >= XSW_SOUNDS_EVENTS) && !played_sound)
		{
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
		    played_sound = True;
		}
	    }
	    if(hp_danger)
	    {
		XSWPrintError(
"Hull structure failing."
		);
		if((opt->sounds >= XSW_SOUNDS_EVENTS) && !played_sound)
		{
		    SoundPlay(
			snd, SW_SNDREF_EVENT_WARNING,
			1.0f, 1.0f,
			0, 0
		    );
		    played_sound = True;
		}
	    }
	}
}

/*
 *	Hit notify.
 *
 *	Plays the hit sound if the camera is in range of the object.
 *
 *	If the player object was hit then notification of the hit
 *	will be printed.
 *
 *	If the player object's weapon hit another object then 
 *	notification of the hit will be printed.
 *
 *	This function should be called whenever an object has been
 *	hit.
 */
void XSWHitNotify(
	const int src_obj_num,		/* Object that hit tar_obj_num */
	const int tar_obj_num,		/* Object that got hit */
	const int owner_obj_num,	/* Owner of src_obj_num */
	const float dmg_structure,
	const float dmg_shield,
	const float bearing		/* Target to source in radians */
)
{
	float vol_left, vol_right;
	int sndref_num = SW_SNDREF_DEFAULT;
	xsw_object_struct	*src_obj, *tar_obj,
				*owner_obj, *camera_obj;
	xsw_sound_struct *snd = &sound;

	/* Get pointers to source and target objects */
	if(DBIsObjectGarbage(src_obj_num))
	    src_obj = NULL;
	else
	    src_obj = xsw_object[src_obj_num];

	if(DBIsObjectGarbage(tar_obj_num))
	    tar_obj = NULL;
	else
	    tar_obj = xsw_object[tar_obj_num];

	if(DBIsObjectGarbage(owner_obj_num))
	    owner_obj = NULL;
	else
	    owner_obj = xsw_object[owner_obj_num];


	/* Get the camera object as the player object */
	camera_obj = net_parms.player_obj_ptr;
	if(camera_obj == NULL)
	    return;

	/* Get the distance from the camera object to target object */
	if(SWMath3DInSameSectorPtr(camera_obj, tar_obj))
	{
	    vol_left = SWMath3DDistance(
		tar_obj->x - camera_obj->x,
		tar_obj->y - camera_obj->y,
		tar_obj->z - camera_obj->z
	    );
	    /* Max distance is 3.0 real units */
	    vol_left = (3.0f - vol_left) / 3.0f;
	    if(vol_left < 0.0f)
		vol_left = 0.0f;

	    vol_right = vol_left;
	}
	else if(SWMath3DInSameSectorPtr(camera_obj, src_obj))
	{
	    vol_left = SWMath3DDistance(
		src_obj->x - camera_obj->x,
		src_obj->y - camera_obj->y,
		src_obj->z - camera_obj->z
	    );
	    /* Max distance is 3.0 real units */
	    vol_left = (3.0f - vol_left) / 3.0f;
	    if(vol_left < 0.0f)
		vol_left = 0.0f;

	    vol_right = vol_left;
	}
	else
	{
	    vol_left = 0.0f;
	    vol_right = 0.0f;
	}


	/* Is the target object the player? */
	if(tar_obj_num == net_parms.player_obj_num)
	{
	    /* Notify that we got hit */
	    char *s;
	    const char	*vname = DBGetObjectVectorName(bearing),
			*src_name = (owner_obj != NULL) ?
			    owner_obj->name :
			    ((src_obj != NULL) ? src_obj->name : "unknown");

	    if(dmg_shield > 0.0f)
	    {
		s = (char *)malloc(
		    STRLEN(vname) + STRLEN(src_name) +
		    40 + 40 + (2 * 80)
		);
		sprintf(
		    s,
"Recieved hit to %s from %s: %.2f damage to structure,\
 %.2f damage to shields.",
		    vname, src_name,
		    dmg_structure, dmg_shield
		);
	    }
	    else
	    {
		s = (char *)malloc(
		    STRLEN(vname) + STRLEN(src_name) +
		    40 + (2 * 80)
		);
		sprintf(
		    s,
"Recieved hit to %s from %s: %.2f damage to structure.",
		    vname, src_obj->name,
		    dmg_structure
		);
	    }
	    XSWPrintWarning(s);
	    free(s);

	    if(dmg_structure > 0.0f)
		sndref_num = SW_SNDREF_HIT_STRUCTURE;
	    else if(dmg_shield > 0.0f)
		sndref_num = SW_SNDREF_HIT_SHIELDS;
	    else
		sndref_num = SW_SNDREF_HIT_STRUCTURE;

	    if((option.sounds > XSW_SOUNDS_NONE) &&
	       ((vol_left > 0.0f) || (vol_right > 0.0f))
	    )
		SoundPlay(
		    snd, sndref_num,
		    vol_left, vol_right,
		    0,
		    0
		);
	}
	/* Is the weapon object's owner the player? */
	else if(owner_obj_num == net_parms.player_obj_num)
	{
	    /* Notify that we hit an object */
	    char *s;
	    const char	*vname = DBGetObjectVectorName(bearing),
			*src_name = (tar_obj != NULL) ?
			    tar_obj->name : "Unknown";

	    if(dmg_shield > 0.0f)
	    {
		s = (char *)malloc(
		    STRLEN(vname) + STRLEN(src_name) +
		    40 + 40 + (2 * 80)
		);
		sprintf(
		    s,
"%s hit %s: %.2f damage to structure,\
 %.2f damage to shields.",
		    src_name, vname,
		    dmg_structure, dmg_shield
		);
	    }
	    else
	    {
		s = (char *)malloc(
		    STRLEN(vname) + STRLEN(src_name) +
		    40 + 40 + (2 * 80)
		);
		sprintf(
		    s,
"%s hit %s: %.2f damage to structure.",
		    src_name, vname,
		    dmg_structure
		);
	    }
	    if((dmg_structure > 0.0f) || (dmg_shield > 0.0f))
		XSWPrint(s);
	    free(s);

	    if(dmg_structure > 0.0f)
		sndref_num = SW_SNDREF_OHIT_STRUCTURE;
	    else if(dmg_shield > 0.0f)
		sndref_num = SW_SNDREF_OHIT_SHIELDS;
	    else
		sndref_num = SW_SNDREF_OHIT_STRUCTURE;

	    if((option.sounds > XSW_SOUNDS_NONE) &&
	       ((vol_left > 0.0f) || (vol_right > 0.0f))
	    )
		SoundPlay(
		    snd, sndref_num,
		    vol_left, vol_right,
		    0,
		    0
		);
	}
	/* Someone else's weapon hit some object */
	else
	{
	    if(dmg_structure > 0.0f)
		sndref_num = SW_SNDREF_OHIT_STRUCTURE;
	    else if(dmg_shield > 0.0f)
		sndref_num = SW_SNDREF_OHIT_SHIELDS;
	    else
		sndref_num = SW_SNDREF_OHIT_STRUCTURE;

	    if((option.sounds > XSW_SOUNDS_NONE) &&
	       ((vol_left > 0.0f) || (vol_right > 0.0f))
	    )
		SoundPlay(
		    snd, sndref_num,
		    vol_left, vol_right,
		    0,
		    0
		);
	}
}

/*
 *	Object destroyed procedure.
 *
 *	This function should be called whenever an object has been
 *	destroyed.
 */
void XSWDestroyNotify(
	const int src_obj_num,		/* Object that caused destruction */
	const int tar_obj_num,		/* Object that was destroyed */
	const int owner_obj_num,	/* Owner of src_obj_num */
	const int reason		/* One of SW_DESTROY_REASON_* */
)
{
	xsw_object_struct *src_obj, *tar_obj;
	xsw_bridge_win_struct *bw = xsw_bridge_win;

	if(bw == NULL)
	    return;

	/* Get pointers to source and target objects (may be NULL) */
	if(DBIsObjectGarbage(src_obj_num))
	    src_obj = NULL;
	else
	    src_obj = xsw_object[src_obj_num];

	if(DBIsObjectGarbage(tar_obj_num))
	    tar_obj = NULL;
	else
	    tar_obj = xsw_object[tar_obj_num];

	/* Check if the destroyed object is the player */
	if(tar_obj_num == net_parms.player_obj_num)
	{
	    /* Player got destroyed */
	    char *s;

	    /* Change viewscreen page to display the "destroyed"
	     * page, this will not change it immediatly but
	     * when we are disconnected (which is very soon)
	     * the "destroyed" page is the page that will be shown
	     */
	    VSChangePage(bw, &bw->destroyed_page);

	    switch(reason)
	    {
	      case SW_DESTROY_REASON_SELF_DESTRUCT:
		s = strdup(
"Your vessel's power-core has over-loaded and detonated!"
		);
		break;
	      default:	/* SW_DESTROY_REASON_UNSPECIFIC */
		s = strdup(
"Your vessel has been destroyed!"
		);
		break;
	    }
	    XSWPrintError(s);
	    free(s);
	}
	else
	{
	    /* Some other object got destroyed */
	    char *s, *name;

	    if(tar_obj != NULL)
	    {
		name = STRDUP(tar_obj->name);
	    }
	    else
	    {
		name = (char *)malloc(80);
		sprintf(name, "Object #%i", tar_obj_num);
	    }
	    switch(reason)
	    {
	      case SW_DESTROY_REASON_SELF_DESTRUCT:
		s = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    s,
"%s's power-core has over-loaded and detonated.",
		    name
		);
		break;
	      default:	/* SW_DESTROY_REASON_UNSPECIFIC */
		s = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    s,
"%s has been destroyed.",
		    name
		);
		break;
	    }
	    if(owner_obj_num == net_parms.player_obj_num)
		XSWPrintBold(s);
	    else
		XSWPrint(s);
	    free(s);
	    free(name);
	}
}


/*
 *	Returns True if the object's tractor beam is engaged or
 *	if another object has its tractor beam engaged on this object.
 */
static bool_t XSWUpdateLoopSoundsTractorBeamState(
	const int obj_num, xsw_object_struct *obj,
	float *vol_left_rtn, float *vol_right_rtn
)
{
	if(obj == NULL)
	    return(False);

	/* Is the specified object tractoring an object? */
	if(obj->total_tractored_objects > 0)
	{
	    const int m = obj->total_tractored_objects;
	    int i;
	    for(i = 0; i < m; i++)
	    {
		if(obj->tractored_object[i] > -1)
		{
		    if(vol_left_rtn != NULL)
			*vol_left_rtn = 1.0f;
		    if(vol_right_rtn != NULL)
			*vol_right_rtn = 1.0f;
		    return(True);
		}
	    }
	}

	/* Is another object in range tractoring this object? */
	if(total_objects > 0)
	{
	    int i;
	    xsw_object_struct *obj2;

	    for(i = 0; i < total_inrange_objects; i++)
	    {
		obj2 = inrange_xsw_object[i];
		if(obj2 == NULL)
		    continue;

		if(obj2->type <= SW_OBJ_TYPE_GARBAGE)
		    continue;

		if(obj2 == obj)
		    continue;

		if(obj2->total_tractored_objects > 0)
		{
		    const int m = obj2->total_tractored_objects;
		    int i;
		    for(i = 0; i < m; i++)
		    {
			if(obj2->tractored_object[i] == obj_num)
			{
			    if(vol_left_rtn != NULL)
				*vol_left_rtn = 1.0f;
			    if(vol_right_rtn != NULL)
				*vol_right_rtn = 1.0f;
			    return(True);
			}
		    }
		}
	    }
	}

	return(False);
}

/*
 *	Updates the looping sounds.
 */
void XSWUpdateLoopSounds(void)
{
	int obj_num;
	sndref_struct *sndref;
	xsw_object_struct *obj;
	xsw_net_parms_struct *net = &net_parms;
	xsw_sound_struct *snd = &sound;

	if(snd == NULL)
	    return;

	/* Get the player object, which may not exist, so subsequent
	 * functions should check that
	 */
	obj_num = net->player_obj_num;
	obj = net->player_obj_ptr;

	/* Tractor beam */
	sndref = SndRefGetPtr(
	    xsw_sndref, total_sndrefs,
	    SW_SNDREF_LOOP_TRACTOR_BEAM
	);
	if(sndref != NULL)
	{
	    float vol_left, vol_right;

	    /* Check if the tractor beam looping sound should be on */
	    if(XSWUpdateLoopSoundsTractorBeamState(
		obj_num, obj,
		&vol_left, &vol_right
	    ))
	    {
		if(snd->loop_tractor_beam_playid == NULL)
		{
		    snd->loop_tractor_beam_playid = SoundPlay(
			snd, SW_SNDREF_LOOP_TRACTOR_BEAM,
			vol_left, vol_right,
			SOUND_EFFECT_REPEATING,
			0
		    );
		}
	    }
	    else
	    {
		if(snd->loop_tractor_beam_playid != NULL)
		{
		    SoundStop(snd, snd->loop_tractor_beam_playid);
		    snd->loop_tractor_beam_playid = NULL;
		}
	    }
	}
	else
	{
	    if(snd->loop_tractor_beam_playid != NULL)
	    {
		SoundStop(snd, snd->loop_tractor_beam_playid);
		snd->loop_tractor_beam_playid = NULL;
	    }
	}



}

/*
 *	Updates the background music.
 */
void XSWUpdateBackgroundMusic(void)
{
	int sndref_num = SW_SNDREF_BG_MUSIC_STANDARD;
	page_struct *page;
	xsw_sound_struct *snd = &sound;
	xsw_bridge_win_struct *bw = xsw_bridge_win;

	if((snd == NULL) || (bw == NULL))
	    return;

	/* Do not update the background music if the bridge window has
	 * not been created yet
	 */
	if(bw->toplevel == None)
	    return;

	/* Check if a page is shown instead of a viewscreen */
	page = bw->cur_page;
	if((page != NULL) ? page->map_state : False)
	{
/* TODO specific pages, always using main menu music for now */
	    sndref_num = SW_SNDREF_BG_MUSIC_MENU_MAIN;
	}
	else
	{
	    xsw_net_parms_struct *net = &net_parms;

	    /* View screen is mapped */
	    sndref_num = SW_SNDREF_BG_MUSIC_STANDARD;

	    if(net->player_obj_ptr != NULL)
	    {
		xsw_object_struct *obj = net->player_obj_ptr;

		/* Check if in nebula */
		if(obj->loc_type == SW_LOC_TYPE_NEBULA)
		    sndref_num = SW_SNDREF_BG_MUSIC_NEBULA;
		else if(obj->loc_type == SW_LOC_TYPE_ASTEROID_FIELD)
		    sndref_num = SW_SNDREF_BG_MUSIC_ASTEROID_FIELD;
#if 0
		else if(obj->loc_type == SW_LOC_TYPE_PULSAR) 
		    sndref_num = SW_SNDREF_BG_MUSIC_NEBULA;
#endif

		/* Check if locked on something */
		if(DBIsObjectGarbage(obj->locked_on))
		{
		    /* Player object is not locked on to anything */

		}
		else
		{
		    /* Player object is locked on something */
		    const int tar_obj_num = obj->locked_on;
		    xsw_object_struct *tar_obj = xsw_object[tar_obj_num];
		    const int tar_type = tar_obj->type;

		    /* Check if locked on a weapon, controlled, or
		     * player
		     */
		    if((tar_type == SW_OBJ_TYPE_CONTROLLED) ||
		       (tar_type == SW_OBJ_TYPE_PLAYER) ||
		       (tar_type == SW_OBJ_TYPE_WEAPON) ||
		       (tar_type == SW_OBJ_TYPE_STREAMWEAPON) ||
		       (tar_type == SW_OBJ_TYPE_SPHEREWEAPON)
		    )
		    {
			/* Not friendly? */
			const int iff = UNVMatchIFFPtr(obj, tar_obj);
			if((iff == SW_IFF_UNKNOWN) ||
			   (iff == SW_IFF_HOSTILE)
			)
			    sndref_num = SW_SNDREF_BG_MUSIC_ENGAGING;
		    }
		}
	    }
	    else
	    {
		/* Player object not valid, could be still logging in */

	    }
	}

	/* Need to change the background music? */
	if(sndref_num != snd->bg_music_sndref_num)
	{
	    /* Change the background music */
	    SoundChangeBackgroundMusic(snd, sndref_num, 0, 0);
	}
}
