#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>

#include "gctl.h"
#include "net.h"
#include "keymaps.h"
#include "xsw.h"


void XSWDialog(
	const char *title, int icon_img_code, const char *msg,
	win_t toplevel
);
int XSWGetResponse(
	const char *title, int icon_img_code, const char *msg,
	unsigned int btn_flags, win_t toplevel
);
void XSWPrint(const char *msg);
void XSWPrintWarning(const char *msg);
void XSWPrintError(const char *msg);
void XSWPrintBold(const char *msg);
void XSWPrintDebugStderr(int error_level);
void XSWPrintHelpStdout(void);
void XSWPrintVersionStdout(void);
void XSWPrintHelp(void);

dialog_struct *XSWDialogNew(void);
void XSWDialogDelete(dialog_struct *d);

cdlg_struct *XSWConfirmationDialogNew(void);
void XSWConfirmationDialogDelete(cdlg_struct *d);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Maps the dialog with the specified message.
 */
void XSWDialog(
	const char *title, int icon_img_code, const char *msg,
	win_t toplevel
)
{
	dialog_struct *d = xsw_dlg;
	image_t *icon = ImgIsAllocated(icon_img_code) ?
	    xsw_image[icon_img_code]->image : NULL;

	/* Not connected to GUI? */
	if(!IDC())
	{
	    printf("%s\n", msg);
	    return;
	}

	/* Initialize the dialog as needed */
	if(d == NULL)
	    xsw_dlg = d = XSWDialogNew();
	if(d == NULL)
	    return;

	/* Set the dialog's title and icon */
	DialogSetTransientFor(d, toplevel);
	DialogSetMapRelative(d, WIDGET_WINDOW_RELATIVE_POINTER);
	DialogSetTitle(d, title);
	DialogSetIcon(d, icon);
	DialogSetMessage(d, msg);
	DialogMap(d);
}

/*
 *	Maps the confirmation dialog and waits for user response.
 *
 *	Returns one of CDlg*.
 */
int XSWGetResponse(
	const char *title, int icon_img_code, const char *msg,
	unsigned int btn_flags, win_t toplevel
)
{
	int response;
	cdlg_struct *d;
	image_t *icon = ImgIsAllocated(icon_img_code) ?
	    xsw_image[icon_img_code]->image : NULL;

	/* Not connected to GUI? */
	if(!IDC())
	    return(CDlgError);

	/* Initialize the confirmation dialog as needed */
	d = xsw_cdlg;
	if(d == NULL)
	    xsw_cdlg = d = XSWConfirmationDialogNew();
	if(d == NULL)
	    return(CDlgError);

	/* Get response */
	response = CDlgGetResponse(
	    d,
	    title, msg, icon,
	    btn_flags,
	    WIDGET_WINDOW_RELATIVE_POINTER,
	    toplevel,
	    XSWMain, NULL,
	    XSWMainPop, NULL
	);

	/* Destroy the confirmation dialog */
	XSWConfirmationDialogDelete(d);
	xsw_cdlg = d = NULL;

	return(response);
}

/*
 *	Prints the message.
 *
 *	If connected to the GUI then prints the message to the message
 *	box, otherwise it prints it to stdout.
 *
 *	The message should not have a tailing newline character.
 */
void XSWPrint(const char *msg)
{
	if(msg == NULL)
	    return;

	if(IDC() && (xsw_bridge_win != NULL))
	    MsgListAppend(msg, xsw_color.bp_standard_text, 0, 0);
	else
	    fprintf(stdout, "%s\n", msg);
}

void XSWPrintWarning(const char *msg)
{
	if(msg == NULL)
	    return;

	if(IDC() && (xsw_bridge_win != NULL))
	    MsgListAppend(msg, xsw_color.bp_warning, 0, 0);
	else
	    fprintf(stderr, "%s\n", msg);
}

void XSWPrintError(const char *msg)
{
	if(msg == NULL)
	    return;

	if(IDC() && (xsw_bridge_win != NULL))
	    MsgListAppend(msg, xsw_color.bp_danger, 0, 0);
	else
	    fprintf(stderr, "%s\n", msg);
}

void XSWPrintBold(const char *msg)
{
	if(msg == NULL)
	    return;

	if(IDC() && (xsw_bridge_win != NULL))
	    MsgListAppend(msg, xsw_color.bold_text, 0, 0);
	else
	    fprintf(stdout, "%s\n", msg);
}


/*
 *	Prints debugging information to stderr.
 *
 *	This function should be called when a segfault is detected
 *	to print a generate state of the program for informational
 *	and minimal debugging purposes.
 */
void XSWPrintDebugStderr(int error_level)
{
	const char *con_state_str, *controller_str;
	xsw_net_parms_struct *net = &net_parms;
	xsw_option_struct *opt = &option;

	switch(net->connection_state)
	{
	  case XSW_CON_STATE_NOT_CONNECTED:
	    con_state_str = "NOT_CONNECTED";
	    break;
	  case XSW_CON_STATE_NEGOTIATING:
	    con_state_str = "NEGOTIATING";
	    break;
	  case XSW_CON_STATE_CONNECTED:
	    con_state_str = "CONNECTED";
	    break;
	  case XSW_CON_STATE_CLOSE_WAIT:
	    con_state_str = "CLOSE_WAIT";
	    break;
	  default:
	    con_state_str = "*badvalue*";
	    break;
	}

	switch(opt->controller)
	{
	  case XSW_CONTROLLER_KEYBOARD:
	    controller_str = "KEYBOARD";
	    break;
	  case XSW_CONTROLLER_JOYSTICK:
	    controller_str = "JOYSTICK";
	    break;
	  case XSW_CONTROLLER_POINTER:
	    controller_str = "POINTER";
	    break;
	  default:
	    controller_str = "*badvalue*";
	    break;
	}

	/* Print debug information */
	fprintf(
	    stderr,
"%s %s debug report:\n",
	    PROG_NAME_FULL, PROG_VERSION
	);

	fprintf(
	    stderr,
"runlevel: %i  prompt_mode: %i  connection_state: %s\n",
	    runlevel, prompt_mode, con_state_str
	);

	fprintf(
	    stderr,
"option.controller: %s  net interval: %i ms\n",
	    controller_str, (int)net->net_int
	);

	fprintf(
	    stderr,
"player object: #%i  total_objects: %i\n",
	    net->player_obj_num, total_objects
	);

}

/*
 *	Print usage of this program to stdout.
 */
void XSWPrintHelpStdout(void)
{
	fputs(PROG_USAGE_MESSAGE, stdout);
}

/*
 *	Print version of this program to stdout.
 */
void XSWPrintVersionStdout(void)
{
	fputs(
	    PROG_NAME_FULL " Version " PROG_VERSION "\n"
	    PROG_COPYRIGHT "\n",
	    stdout
	);
}

/*
 *	Prints a basic list of commands to the message box.
 */
void XSWPrintHelp(void)
{
	char *s;

	s = (char *)malloc(512);
	if(s == NULL)
	    return;

	sprintf(
	    s,
"%s = Help  %s = Connect  %s = Disconnect  %s = Refresh\
  %s = Client Command  %s = Server Command",
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_HELP)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_CONNECT)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_DISCONNECT)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_REFRESH)),

	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_CLIENT_CMD)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_SERVER_CMD))
	);
	XSWPrint(s);

	sprintf(
	    s,
"%s = Set Channel  %s = Send Hail  %s = Send Message  %s = Intercept\
  %s = Shield Frequency",
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_COM_CHANNEL_SET)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_HAIL)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_SEND_MESSAGE)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_INTERCEPT_SET)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_SHIELD_FREQ))
	);
	XSWPrint(s);

	sprintf(
	    s,
"%s = Lock Next  %s = Unlock  %s = Tractor Beam  %s = Shields  %s = Fire Weapon\
  %s = Auto Destruct",
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_LOCK_NEXT)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_UNLOCK)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_TRACTOR_BEAM_LOCK)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_SHIELD)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_FIRE_WEAPON)),
	    OSWGetKeyCodeName(KeymapGetKeycode(XSW_KM_AUTO_DESTRUCT))
	);
	XSWPrint(s);

	free(s);
}


/*
 *	Creates a new dialog.
 */
dialog_struct *XSWDialogNew(void)
{
	dialog_struct *d = (dialog_struct *)calloc(
	    1, sizeof(dialog_struct)
	);
	if(d == NULL)
	    return(NULL);

	if(DialogInit(d))
	{
	    free(d);
	    d = NULL;
	}

	return(d);
}

/*
 *	Deletes the dialog.
 */
void XSWDialogDelete(dialog_struct *d)
{
	if(d == NULL)
	    return;

	WidgetDestroy(d);
	free(d);
}


/*
 *	Creates a new confirmation dialog.
 */
cdlg_struct *XSWConfirmationDialogNew(void)
{
	cdlg_struct *d = (cdlg_struct *)calloc(
	    1, sizeof(cdlg_struct)
	);
	if(d == NULL)
	    return(NULL);

	if(CDlgInit(d))
	{
	    free(d);
	    d = NULL;
	}

	return(d);
}

/*
 *	Deletes the confirmation dialog.
 */
void XSWConfirmationDialogDelete(cdlg_struct *d)
{
	if(d == NULL)
	    return;

	CDlgDestroy(d);
	free(d);
}
