#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/strexp.h"
#include "../include/prochandle.h"
#include "../include/osw-x.h"		/* For image_t */

#include "../include/printerw.h"


static float CM_TO_PIXELS(float x);
static float INCH_TO_PIXELS(float x);

#define WRITE_PIXEL_INPUTS		\
	FILE *fp,			\
	const u_int8_t *p8,		\
	const unsigned long rm,		\
	const unsigned long gm,		\
	const unsigned long bm		\

static int WRITE_PIXEL_GREYSCALE_8(WRITE_PIXEL_INPUTS);
static int WRITE_PIXEL_COLOR_8(WRITE_PIXEL_INPUTS);
static int WRITE_PIXEL_GREYSCALE_15(WRITE_PIXEL_INPUTS);
static int WRITE_PIXEL_COLOR_15(WRITE_PIXEL_INPUTS);
static int WRITE_PIXEL_GREYSCALE_16(WRITE_PIXEL_INPUTS);
static int WRITE_PIXEL_COLOR_16(WRITE_PIXEL_INPUTS);
static int WRITE_PIXEL_GREYSCALE_24(WRITE_PIXEL_INPUTS);
static int WRITE_PIXEL_COLOR_24(WRITE_PIXEL_INPUTS);

int PrinterWritePSImage(
	FILE *fp, image_t *img,
	const int page,		/* Page number (starts from 0) */
	printer_parm_struct *parm,
	int (*progress_cb)(void *, const unsigned long, const unsigned long),
	void *progress_data
);

int PrinterRunPrint(
	const char *cmd,		/* Print command */
	const char *path,		/* Output postscript file */
	int (*progress_cb)(void *, const unsigned long, const unsigned long),
	void *progress_data
);
int PrinterPrintImage(
	image_t *image,
	const char *path,               /* Output postscript file */
	const char *cmd,                /* Print command */     
	printer_parm_struct *parm,
	int (*progress_cb)(void *, const unsigned long, const unsigned long),
	void *progress_data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static float CM_TO_PIXELS(float x)
{
	return(x * 72 / 2.56);
}

static float INCH_TO_PIXELS(float x)
{
	return(x * 72);
}


/*
 *	Pixel writers
 *
 *	Returns the number of characters written.
 */
static int WRITE_PIXEL_GREYSCALE_8(WRITE_PIXEL_INPUTS)
{
	const unsigned long v = *((const u_int8_t *)p8); 
	fprintf(
	    fp,
	    "%.2x",
	    (u_int8_t)((
		((v & rm) * 0xff / rm) +
		((v & gm) * 0xff / gm) +
		((v & bm) * 0xff / bm)) / 3
	    )
	);
	return(2);
}

static int WRITE_PIXEL_COLOR_8(WRITE_PIXEL_INPUTS)
{
	const unsigned long v = *((const u_int8_t *)p8);
	fprintf(
	    fp,
	    "%.2x%.2x%.2x",
	    (u_int8_t)((v & rm) * 0xff / rm),
	    (u_int8_t)((v & gm) * 0xff / gm),
	    (u_int8_t)((v & bm) * 0xff / bm)
	);
	return(6);
}

static int WRITE_PIXEL_GREYSCALE_15(WRITE_PIXEL_INPUTS)
{
	const unsigned long v = *((const u_int16_t *)p8);
	fprintf(
	    fp,
	    "%.2x",
	    (u_int8_t)((
		((v & rm) * 0xFF / rm) +
		((v & gm) * 0xFF / gm) +
		((v & bm) * 0xFF / bm)) / 3
	    )
	);
	return(2);
}

static int WRITE_PIXEL_COLOR_15(WRITE_PIXEL_INPUTS)
{
	const unsigned long v = *((const u_int16_t *)p8);
	fprintf(
	    fp,
	    "%.2x%.2x%.2x",
	    (u_int8_t)((v & rm) * 0xFF / rm),
	    (u_int8_t)((v & gm) * 0xFF / gm),
	    (u_int8_t)((v & bm) * 0xFF / bm)
	);
	return(6);
}

static int WRITE_PIXEL_GREYSCALE_16(WRITE_PIXEL_INPUTS)
{
	const unsigned long v = *((const u_int16_t *)p8);
	fprintf(
	    fp,
	    "%.2x",
	    (u_int8_t)((
		((v & rm) * 0xFF / rm) +
		((v & gm) * 0xFF / gm) +
		((v & bm) * 0xFF / bm)) / 3
	    )
	);
	return(2);
}

static int WRITE_PIXEL_COLOR_16(WRITE_PIXEL_INPUTS)
{
	const unsigned long v = *((const u_int16_t *)p8);
	fprintf(
	    fp,
	    "%.2x%.2x%.2x",
	    (u_int8_t)((v & rm) * 0xFF / rm),
	    (u_int8_t)((v & gm) * 0xFF / gm),
	    (u_int8_t)((v & bm) * 0xFF / bm)
	);
	return(6);
}

static int WRITE_PIXEL_GREYSCALE_24(WRITE_PIXEL_INPUTS)
{
	fprintf(
	    fp,
	    "%.2x",
	    (u_int8_t)(
		((int)p8[2] + (int)p8[1] + (int)p8[0]) / 3
	    )
	);
	return(2);
}

static int WRITE_PIXEL_COLOR_24(WRITE_PIXEL_INPUTS)
{
	fprintf(
	    fp,
	    "%.2x%.2x%.2x",
	    p8[2], p8[1], p8[0]
	);
	return(6);

}



/*
 *	Writes a PS file based onthe given image and parm data.
 *	The file fp must already be fopen()ed prior to calling
 *	this function and assumed to be positioned at the beginning
 *	of the file.  The file must be opened in write only mode.
 */
int PrinterWritePSImage(
	FILE *fp, image_t *img,
	const int page,		/* Page number (starts from 0) */
	printer_parm_struct *parm,
	int (*progress_cb)(void *, const unsigned long, const unsigned long),
	void *progress_data
)
{
	int bpp, bpl, depth;
	int (*write_func)(WRITE_PIXEL_INPUTS);

	if((fp == NULL) || (img == NULL) || (parm == NULL) ||
	   (page < 0)
	)
	    return(PrinterBadValue);

	/* Report initial progress */
	if(progress_cb != NULL)
	{
	    if(!progress_cb(progress_data, 0l, 0l))
		return(PrinterUserAborted);
	}

	bpp = img->bytes_per_pixel;
	bpl = img->bytes_per_line;
	depth = img->depth;

	/* Headings */
	/* Format identifier */
	fprintf(
	    fp,
"%%!PS-Adobe-2.0 EPSF-2.0\n"
	);
	/* Title */
	fprintf(
	    fp, 
"%%%%Title: %s\n",
	    parm->title
	);
	/* Creator */
	fprintf(
	    fp,
"%%%%Creator: %s\n",
	    "Generic Printer Wrapper Code"
	);
	/* Bounding box */
	switch(parm->units)
	{
	  case PrinterUnitCentimeters:
	    fprintf(
	       fp,
"%%%%BoundingBox: %i %i %i %i\n",
		(int)CM_TO_PIXELS(parm->x),
		(int)CM_TO_PIXELS(parm->y),
		(int)CM_TO_PIXELS(parm->x) + (int)img->width,
		(int)CM_TO_PIXELS(parm->y) + (int)img->height
	    );
	    break;

	  case PrinterUnitInches:
	    fprintf(
	       fp,
"%%%%BoundingBox: %i %i %i %i\n",
		(int)INCH_TO_PIXELS(parm->x),
		(int)INCH_TO_PIXELS(parm->y),
		(int)INCH_TO_PIXELS(parm->x) + (int)img->width,
		(int)INCH_TO_PIXELS(parm->y) + (int)img->height
	    );
	    break;

	  default:	/* PrinterUnitPixels */
	    fprintf(
		fp,
"%%%%BoundingBox: %i %i %i %i\n",
		(int)parm->x,
		(int)parm->y,
		(int)parm->x + (int)img->width,
		(int)parm->y + (int)img->height
	    );
	    break;
	}
	/* Pages */
	fprintf(
	    fp,
"%%%%Pages: %i\n",
	    parm->pages        
	);
	/* Fonts? */
	fprintf(
	    fp,   
"%%%%DocumentFonts:\n"
	);
	/* End comments? */
	fprintf(
	    fp,   
"%%%%EndComments\n"
	);
	/* End prolog? */
	fprintf(
	    fp,
"%%%%EndProlog\n\n"
	);


	/* Page number */
	fprintf(
	    fp,
"%%%%Page: %i %i\n\n",
	    page + 1,
	    parm->pages
	);

	/* Remember original state? */
	fprintf(   
	    fp,
"%% remember original state\n"
	);
	fprintf(   
	    fp,
"/origstate save def\n\n"
	);

	/* Temp dictionary? */
	fprintf(
	    fp,
"%% build a temporary dictionary\n"
	);
	fprintf(
	    fp,
"20 dict begin\n\n"
	);

	/* Define string to hold a scanline's worth of data */
	fprintf(
	    fp,
"%% define string to hold a scanline's worth of data\n"
	);
	if(parm->color_mode == PrinterColorModeColor)
	    fprintf(
		fp,
"/pix %i string def\n\n",
	        (int)img->width * 3
	    );
	else
	    fprintf(   
		fp,
"/pix %i string def\n\n",
		(int)img->width
	    );


	/* Define space for color conversions */
	fprintf(
	    fp,
"%% define space for color conversions\n"
	);
	fprintf(
	    fp,
"/grays %i string def  %% space for gray scale line\n",
	    (int)img->width
	);
	fprintf(
	    fp,
"/npixls 0 def\n"
	);
	fprintf(
	    fp,
"/rgbindx 0 def\n\n"
	);

	/* Lower left corner */
	fprintf(
	    fp,
"%% lower left corner\n"
	);
	switch(parm->units)
	{
	  case PrinterUnitCentimeters:
	    fprintf(
	       fp,
"%i %i translate\n\n",
		(int)CM_TO_PIXELS(parm->x),
		(int)CM_TO_PIXELS(parm->height) -
		    (int)CM_TO_PIXELS(parm->y) - (int)img->height
	    );
	    break;

	  case PrinterUnitInches:
	    fprintf(
	       fp,
"%i %i translate\n\n",
		(int)INCH_TO_PIXELS(parm->x),
		(int)INCH_TO_PIXELS(parm->height) -
		    (int)INCH_TO_PIXELS(parm->y) - (int)img->height
	    );
	    break;

	  default:	/* PrinterUnitPixels */
	    fprintf(
	       fp,
"%i %i translate\n\n",
		(int)parm->x,
		(int)parm->height - (int)parm->y - (int)img->height
	    );
	    break;
	};


	/* Size of image (on paper, in 1/72inch coords) */
	fprintf(
	    fp,
"%% size of image (on paper, in 1/72inch coords)\n"
	);
	fprintf(
	    fp,
"%.5f %.5f scale\n\n",
	    (float)img->width * 0.9993,
	    (float)img->height * 0.9993
	);

	/* Define `colorimage' if printing in color mode */
	if(parm->color_mode == PrinterColorModeColor)
	{
	    fprintf(
		fp,   
"%% define 'colorimage' if it isn't defined\n\
%%   ('colortogray' and 'mergeprocs' come from xwd2ps\n\
%%     via xgrab)\n"
	    );
	    fprintf(
		fp,
"/colorimage where   %% do we know about 'colorimage'?\n\
  { pop }           %% yes: pop off the 'dict' returned\n\
  {                 %% no:  define one\n\
    /colortogray {  %% define an RGB->I function\n\
      /rgbdata exch store    %% call input 'rgbdata'\n\
      rgbdata length 3 idiv\n\
      /npixls exch store\n\
      /rgbindx 0 store\n\
      0 1 npixls 1 sub {\n\
	grays exch\n\
	rgbdata rgbindx       get 20 mul    %% Red\n\
	rgbdata rgbindx 1 add get 32 mul    %% Green\n\
	rgbdata rgbindx 2 add get 12 mul    %% Blue\n\
	add add 64 idiv      %% I = .5G + .31R + .18B\n\
	put\n\
	/rgbindx rgbindx 3 add store\n\
      } for\n\
      grays 0 npixls getinterval\n\
    } bind def\n\n"
	    );

	    fprintf(
		fp,
"%% Utility procedure for colorimage operator.\n\
%% This procedure takes two procedures off the\n\
%% stack and merges them into a single procedure.\n\n"
	    );

	    fprintf(
		fp,
"/mergeprocs { %% def\n\
      dup length\n\
      3 -1 roll\n\
      dup\n\
      length\n\
      dup\n\
      5 1 roll\n\
      3 -1 roll\n\
      add\n\
      array cvx\n\
      dup\n\
      3 -1 roll\n\
      0 exch\n\
      putinterval\n\
      dup\n\
      4 2 roll\n\
      putinterval\n\
    } bind def\n\n"
	    );

	    fprintf(
		fp,
"/colorimage { %% def\n\
      pop pop     %% remove 'false 3' operands\n\
      {colortogray} mergeprocs\n\
      image\n\
    } bind def\n\
  } ifelse          %% end of 'false' case\n\n"
	    );
	}

	/* Dimensions of data */
	fprintf(
	    fp,
"%% dimensions of data\n"
	);
	switch(parm->color_mode)
	{
	  case PrinterColorModeColor:
	    fprintf(
		fp,
"%i %i %i\n",
	        (int)img->width, (int)img->height, 8
	    );
	    break;

	  case PrinterColorModeGreyScale:
	    fprintf(
		fp,
"%i %i %i\n",
		(int)img->width, (int)img->height, 8
	    );
	    break;

	  default:	/* PrinterColorModeBlackAndWhite */
	    fprintf(
		fp,
"%i %i %i\n",
		(int)img->width, (int)img->height, 1
	    );
	    break;
	}

	/* Mapping matrix */
	fprintf(
	    fp,
"%% mapping matrix\n"
	);
	fprintf(
	    fp,
"[%i %i %i %i %i %i]\n",
	    (int)img->width,
	    0,
	    0,
	    (int)img->height * -1,
	    0,
	    (int)img->height
	);

	fprintf(
	    fp,
"{currentfile pix readhexstring pop}\n"
	);

	/* Use color image if in color mode */
	if(parm->color_mode == PrinterColorModeColor)
	    fprintf(
		fp,
"false 3 colorimage\n\n"
	    );
	else
	    fprintf(
		fp,
"image\n\n"
	    );


	/* Write image data */
	write_func = NULL;
	switch(bpp)
	{
	  case 1:
	    switch(parm->color_mode)
	    {
	      case PrinterColorModeColor:
		write_func = WRITE_PIXEL_COLOR_8;
		break;
	      case PrinterColorModeGreyScale:
		write_func = WRITE_PIXEL_GREYSCALE_8;
		break;
	    }
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		switch(parm->color_mode)
		{
		  case PrinterColorModeColor:
		    write_func = WRITE_PIXEL_COLOR_15;
		     break;
		 case PrinterColorModeGreyScale:
		    write_func = WRITE_PIXEL_GREYSCALE_15;
		    break;
		}
		break;
	      case 16:
		switch(parm->color_mode)
		{   
		  case PrinterColorModeColor:
		    write_func = WRITE_PIXEL_COLOR_16;
		     break;
		 case PrinterColorModeGreyScale:
		    write_func = WRITE_PIXEL_GREYSCALE_16;
		    break;
		}
		break;
	    }
	    break;
	  case 3:
	  case 4:
	    switch(parm->color_mode)
	    {
	      case PrinterColorModeColor:
		write_func = WRITE_PIXEL_COLOR_24;
		break;
	      case PrinterColorModeGreyScale:
		write_func = WRITE_PIXEL_GREYSCALE_24;
		break;
	    }
	    break;
	}
	if(write_func != NULL)
	{
	    const int line_max = 70;
	    const unsigned long	rm = img->red_mask,
				gm = img->green_mask,
				bm = img->blue_mask;
	    int x, y, col = 0;

	    for(y = 0; y < (int)img->height; y++)
	    {
		if(progress_cb != NULL)      
		{
		    if(!progress_cb(
			progress_data,
			(unsigned long)y,
			(unsigned long)img->height
		    ))
			return(PrinterUserAborted);
		}

		for(x = 0; x < (int)img->width; x++)
		{
		    col += write_func(
			fp,
			img->data + (y * bpl) + (x * bpp),
			rm, gm, bm
		    );
		    if(col >= line_max)
		    {
			col = 0;
			fputc('\n', fp);
		    }
		}

		col = 0;
		fputc('\n', fp);
	    }
	}

	fputc('\n', fp);


	/* Show page */
	fprintf(
	    fp,
"showpage\n\n"
	);           

	/* Stop using temporary dictionary */
	fprintf(
	    fp,
"%% stop using temporary dictionary\n"
	);
	fprintf(
	    fp, 
"end\n\n"
	);
 
	/* Restore originate state */
	fprintf(
	    fp,
"%% restore original state\n"
	);
	fprintf(
	    fp,
"origstate restore\n\n"
	);

	/* Trailer */
	fprintf(
	    fp,
"%%%%Trailer\n"
	);

	if(progress_cb != NULL)      
	{
	    if(!progress_cb(
		progress_data, 
		(unsigned long)img->height,
		(unsigned long)img->height
	    ))
		return(PrinterUserAborted);
	}

	return(PrinterSuccess);
}


/*
 *	Executes the print command.
 */
int PrinterRunPrint(
	const char *cmd,		/* Print command */
	const char *path,		/* Output postscript file */
	int (*progress_cb)(void *, const unsigned long, const unsigned long),
	void *progress_data
)
{
#define PrinterFileToken	"%file"
	const char *s;
	char *fcmd, **strv;
	int i, n, len, strc;

	if(STRISEMPTY(path))
	    return(PrinterBadValue);

	if(STRISEMPTY(cmd))
	    cmd = PrinterDefPrintCmd;

	/* Count the number of occurances of the path substitution
	 * in the command
	 */
	n = 0;
	s = strstr(cmd, PrinterFileToken);
	while(s != NULL)
	{
	    n++;
	    s = strstr(s + 1, PrinterFileToken);
	}

	/* Allocate a new command with space for the path
	 * substitutions
	 */
	len = strlen(cmd) + (n * strlen(path)) + 10;
	fcmd = (char *)malloc((len + 1) * sizeof(char));
	if(fcmd == NULL)
	    return(PrinterNoBuffers);
	strcpy(fcmd, cmd);

	/* Substitute the paths in the new command*/
	substr(fcmd, PrinterFileToken, path);

	/* Explode the command */
	strv = strchrexp(fcmd, ';', &strc);
	for(i = 0; i < strc; i++)
	{
	    s = strv[i];
	    if(s == NULL)
		continue;

	    ExecB(s);
	    free(strv[i]);
	}
	free(strv);

	free(fcmd);

	return(PrinterSuccess);
}


/*
 *	Prints the image.
 */
int PrinterPrintImage(
	image_t *img,
	const char *path,               /* Output postscript file */
	const char *cmd,                /* Print command */
	printer_parm_struct *parm,
	int (*progress_cb)(void *, const unsigned long, const unsigned long),
	void *progress_data
)
{
	int status;
	struct stat stat_buf;
	FILE *fp;

	if((img == NULL) || (parm == NULL) || STRISEMPTY(path))
	    return(PrinterBadValue);

	if((img->width == 0) ||
	   (img->height == 0) ||
	   (img->data == NULL)
	)
	    return(PrinterBadValue);

	if((parm->width <= 0) ||
	   (parm->height <= 0)
	)
	    return(PrinterBadValue);

	if(parm->pages <= 0)
	    return(PrinterSuccess);

	/* Make sure that the specified output file does not exist */
	if(!stat(path, &stat_buf))
	    return(PrinterError);

	/* Create the output file */
	fp = fopen(path, "wb");
	if(fp == NULL)
	    return(PrinterError);

	status = PrinterWritePSImage(
	    fp, img,
	    0,
	    parm,
	    progress_cb, progress_data
	);

	fclose(fp);

	if(status != PrinterSuccess)
	    return(status);


	/* Print to the printer? */
	if(parm->dest == PrinterDestinationPrinter)
	{
	    /* Print to the printer */
	    status = PrinterRunPrint(cmd, path, progress_cb, progress_data);

	    /* The output file is no longer needed */
	    unlink(path);

	    if(status != PrinterSuccess)
		return(status);
	}
	else
	{
	    /* Print to file (nothing to do) */
	}

	return(PrinterSuccess);
}
