#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/xsw_ctype.h"
#include "../include/string.h"
#include "../include/reality.h"

#include "../include/unvmath.h"
#include "../include/unvmatch.h"


char UNVIsObjectGarbage(
	xsw_object_struct **xsw_object, int total_objects,
	int n
);

char UNVIsObjectNameIndex(const char *name);

/* Visibility */
float UNVGetObjectVisibility(
	xsw_object_struct **xsw_object, int total_objects,
	int n
);
float UNVGetObjectVisibilityPtr(xsw_object_struct *obj);

/* Effective Scanner Range */
float UNVGetObjectEffectiveScannerRange(
	xsw_object_struct **xsw_object, int total_objects,
	int n
);
float UNVGetObjectEffectiveScannerRangePtr(xsw_object_struct *obj);

char UNVIsObjectInEffectiveScannerRange(
	xsw_object_struct **xsw_object, int total_objects,
	int obj_num, int tar_obj_num,
	float scanner_range
);
char UNVIsObjectInEffectiveScannerRangePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj,
	float scanner_range
);

/* Intercept Matching */
int UNVMatchIntercept(
	xsw_object_struct **xsw_object, int total_objects,
	int obj_num, const char *arg
);
int UNVMatchInterceptByNumber(
	xsw_object_struct **xsw_object, int total_objects,
	int ref_obj, int tar_obj
);

/* Match By Name */
int UNVMatchObjectByName(
	xsw_object_struct **xsw_object, int total_objects,
	const char *name, int type
);

/* Match Nearest */
int UNVMatchObjectNearest(
	xsw_object_struct **xsw_object, int total_objects,
	int obj_num, int type
);

/* IFF Matching */
int UNVMatchIFF(
	xsw_object_struct **xsw_object, int total_objects,
	int ref_obj_num, int tar_obj_num
);
int UNVMatchIFFPtr(
	xsw_object_struct *ref_obj, xsw_object_struct *tar_obj
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns true if object n is invalid or is of
 *	type SW_OBJ_TYPE_GARBAGE.
 *
 *	Returns false if the object is valid and non-garbage.
 */
char UNVIsObjectGarbage(
	xsw_object_struct **xsw_object,
	int total_objects,
	int n
)
{
	if((xsw_object == NULL) ||
	   (n < 0) ||
	   (n >= total_objects)
	)
	    return(1);
	else if(xsw_object[n] == NULL)
	    return(1);
	else if(xsw_object[n]->type <= SW_OBJ_TYPE_GARBAGE)
	    return(1);
	else
	    return(0);
}

/*
 *	Checks if the object name is in index notation (returns true)
 *	or a regular name (returns false).
 */
char UNVIsObjectNameIndex(const char *name)
{
	if(name == NULL)
	    return(0);
	else if(*name == '#')
	    return(1);
	else
	    return(0);
}


/*
 *	Returns object's current visibility.
 *
 *	Takes into account the object's location (area objects that
 *	produce limited visibility) and its shield and cloak states.
 *
 *	Note that obj->cur_visibility reflects its visibility modified
 *	by its shield and cloak states.
 */
float UNVGetObjectVisibility(
	xsw_object_struct **xsw_object, int total_objects,
	int n
)
{
	if(UNVIsObjectGarbage(xsw_object, total_objects, n))
	    return(0.0f);
	else 
	    return(UNVGetObjectVisibilityPtr(xsw_object[n]));
}

float UNVGetObjectVisibilityPtr(xsw_object_struct *obj)
{
	if(obj == NULL)
	    return(0.0f);

	if(obj->type == SW_OBJ_TYPE_AREA)
	    return(obj->cur_visibility);
	/* Object in a nebula? */
	else if(obj->loc_type == SW_LOC_TYPE_NEBULA)
	    return(obj->cur_visibility * SW_VISIBILITY_NEBULA);
	/* Object in an asteroid field? */
	else if(obj->loc_type == SW_LOC_TYPE_ASTEROID_FIELD)
	    return(obj->cur_visibility * SW_VISIBILITY_ASTEROID_FIELD);
	/* Object in an pulsar? */
	else if(obj->loc_type == SW_LOC_TYPE_PULSAR)
	    return(1.0f);
	/* Object in normal space */
	else
	    return(obj->cur_visibility);
}


/*
 *	Returns the object's current effective scanner range in real
 *	units.
 *
 *	Takes into account the object's location, area objects that
 *	produce limited visibility will be applied to the returned
 *	scanner range.
 */
float UNVGetObjectEffectiveScannerRange(  
	xsw_object_struct **xsw_object, int total_objects,
	int n
)
{
	if(UNVIsObjectGarbage(xsw_object, total_objects, n))
	    return(0.0f);
	else
	    return(UNVGetObjectVisibilityPtr(xsw_object[n]));
}

float UNVGetObjectEffectiveScannerRangePtr(xsw_object_struct *obj)
{
	if(obj == NULL)
	    return(0.0f);

	switch(obj->loc_type)
	{
	  case SW_LOC_TYPE_NEBULA:
	    return(
		MAX(obj->scanner_range, 0.0f) * SW_VISIBILITY_NEBULA
	    );
	    break;
	  case SW_LOC_TYPE_ASTEROID_FIELD:
	    return(
		MAX(obj->scanner_range, 0.0f) * SW_VISIBILITY_ASTEROID_FIELD
	    );
	    break;
	  default:
	    return(MAX(obj->scanner_range, 0.0f));
	    break;
	}
}


/*
 *	Returns true if the target object is within the object's
 *	effective scanner range.
 *
 *	The scanner_range must be the effective scanner range obtained
 *	from UNVGetObjectEffectiveScannerRange*().
 */
char UNVIsObjectInEffectiveScannerRange(
	xsw_object_struct **xsw_object, int total_objects,
	int obj_num, int tar_obj_num,
	float scanner_range
)
{
	if(UNVIsObjectGarbage(xsw_object, total_objects, obj_num) ||
	   UNVIsObjectGarbage(xsw_object, total_objects, tar_obj_num)
	)
	    return(0);
	else
	    return(UNVIsObjectInEffectiveScannerRangePtr(
		xsw_object[obj_num], xsw_object[tar_obj_num],
		scanner_range
	    ));
}

char UNVIsObjectInEffectiveScannerRangePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj,
	float scanner_range
)
{
	float d, size;

	/* Not in the same sector or no scanning range? */
	if(!SWMath3DInSameSectorPtr(obj, tar_obj) || (scanner_range <= 0.0f))
	    return(0);

	/* Calculate the distance between the two objects in real units */
	size = MAX(((float)tar_obj->size / 1000.0f), 0.0f);
	d = MAX((SWMath3DDistance(
	    tar_obj->x - obj->x,
	    tar_obj->y - obj->y,
	    tar_obj->z - obj->z
	) - size), 0.0f);

	switch(tar_obj->type)
	{
	  case SW_OBJ_TYPE_AREA:
	    /* Always check area object ranges assuming the area
	     * object is at full visibility
	     */
	    return((d > scanner_range) ? 0 : 1);
	    break;

	  default:
	    /* Is the specified object is in a visibility obscured
	     * area?
	     */
	    if((obj->loc_type == SW_LOC_TYPE_NEBULA) ||
	       (obj->loc_type == SW_LOC_TYPE_ASTEROID_FIELD)
	    )
	    {
		/* Are both objects in the same area? */
		if(obj->loc_obj_num == tar_obj->loc_obj_num)
		{
		    /* Check objects in the same area with the target
		     * object's cur_visibility because the specified
		     * effective range has already been reduced
		     * by the visibility obscured area object and that
		     * the cur_visibility does not take into account
		     * the location that might affect it
		     */
		    if(tar_obj->cur_visibility > 0.0f)
			return(
	(d > (scanner_range * tar_obj->cur_visibility)) ? 0 : 1
			);
		    else
			return(0);
		}
		else
		{
		    /* Target object is either outside of the area
		     * object or within another area object, in either
		     * case take the target object's current visibility
		     * into account
		     */
		    const float tar_vis = UNVGetObjectVisibilityPtr(tar_obj);
		    if(tar_vis <= 0.0f)
			return(0);

		    return((d > (scanner_range * tar_vis)) ? 0 : 1);
		}
	    }    
	    else
	    {
		/* Not in a visibility obscured area, so check the
		 * range with the target object's current visibility
		 */
		const float tar_vis = UNVGetObjectVisibilityPtr(tar_obj);
		if(tar_vis <= 0.0f)
		    return(0);

		return((d > (scanner_range * tar_vis)) ? 0 : 1);
	    }
	    break;
	}
}


/*
 *	Intercepts the object specified by arg.
 *
 *	If arg is "" or "#off" then -1 is returned.
 *
 *	If the object specified in arg is found and is of type
 *	SW_OBJ_TYPE_STATIC then that object's number is returned
 *	without checking obj_num's scanner range.
 */
int UNVMatchIntercept(
	xsw_object_struct **xsw_object, int total_objects,
	int obj_num,
	const char *arg
)
{
	int tar_obj_num;
	char name[SW_OBJ_NAME_MAX];     
	xsw_object_struct *obj, *tar_obj;

	if(arg == NULL)
	    return(-1);

	if(UNVIsObjectGarbage(xsw_object, total_objects, obj_num))
	    return(-1);

	obj = xsw_object[obj_num];

	strncpy(name, arg, sizeof(name));
	name[sizeof(name) - 1] = '\0';
	strstrip(name);

	/* Turn intercept off? */
	if(!strcasecmp(name, "#off") || (*name == '\0'))
	    return(-1);

	/* Match the target object by name */
	tar_obj_num = UNVMatchObjectByName(
	    xsw_object,
	    total_objects,
	    name,
	    -1
	);
	if(UNVIsObjectGarbage(xsw_object, total_objects, tar_obj_num))
	    return(-1);

	tar_obj = xsw_object[tar_obj_num];

	/* Make sure we do not intercept itself */
	if(tar_obj_num == obj_num)
	    return(-1);

	/* Following types of objects may not be intercepted */
	if((tar_obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (tar_obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)
	    return(-1);

	/* Following objects don't need range check */
	if((tar_obj->type == SW_OBJ_TYPE_STATIC) ||
	   (tar_obj->type == SW_OBJ_TYPE_HOME) ||
	   (tar_obj->type == SW_OBJ_TYPE_AREA) ||
	   (tar_obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (tar_obj->type == SW_OBJ_TYPE_ELINK)
	)
	    return(tar_obj_num);

	/* Is tar_obj within scanner range of obj_num? */
	if(SWMath3DInRangePtr(
	    obj,
	    tar_obj,
	    obj->scanner_range * UNVGetObjectVisibilityPtr(tar_obj)
	))
	    return(tar_obj_num);
	else
	    return(-1);
}


/*
 *	Checks if tar_obj is within intercept range of ref_obj,
 *	and that both are valid.  Returns tar_obj if everything
 *	checks out okay or -1 if not.
 */
int UNVMatchInterceptByNumber(
	xsw_object_struct **xsw_object, int total_objects,
	int ref_obj_num, int tar_obj_num
)
{
	xsw_object_struct *obj;
	xsw_object_struct *tar_obj;


	if(UNVIsObjectGarbage(xsw_object, total_objects, ref_obj_num))
	    return(-1);
	else
	    obj = xsw_object[ref_obj_num];

	if(UNVIsObjectGarbage(xsw_object, total_objects, tar_obj_num))
	    return(-1);
	else
	    tar_obj = xsw_object[tar_obj_num];


	/* ref_obj_num and tar_obj_num cannot be the same */
	if(ref_obj_num == tar_obj_num)
	    return(-1);


	/* Skip these types of objects */
	if((tar_obj->type == SW_OBJ_TYPE_WEAPON) ||
	   (tar_obj->type == SW_OBJ_TYPE_STREAMWEAPON) ||
	   (tar_obj->type == SW_OBJ_TYPE_SPHEREWEAPON)
	)
	    return(-1);


	/* Following objects always matchable regardless of distance */
	if((tar_obj->type == SW_OBJ_TYPE_STATIC) ||
	   (tar_obj->type == SW_OBJ_TYPE_HOME) ||
	   (tar_obj->type == SW_OBJ_TYPE_AREA) ||
	   (tar_obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (tar_obj->type == SW_OBJ_TYPE_ELINK)
	)
	    return(tar_obj_num);

	/* If none of the above then do scanner range check */
	if(SWMath3DInRangePtr(
	    obj, tar_obj,
	    obj->scanner_range * UNVGetObjectVisibilityPtr(tar_obj)
	))
	    return(tar_obj_num);
	else
	    return(-1);
}


/*
 *	Matches an object by the specified name.
 *
 *	The type specifies the type of object to match or -1 for any
 *	type.
 */
int UNVMatchObjectByName(
	xsw_object_struct **xsw_object, int total_objects,
	const char *name, int type
)
{
	int obj_num;
	xsw_object_struct *obj;

	if(name == NULL)
	    return(-1);

	/* Skip leading spaces in name */
	while(ISBLANK(*name))
	    name++;

	/* Empty string? */
	if(*name == '\0')
	    return(-1);

	/* Number matching? */
	if(UNVIsObjectNameIndex(name))
	{
	    /* Skip leading '#' characters */
	    while(*name == '#')
		name++;

	    /* Get object number directly from name string */
	    obj_num = atoi(name);

	    /* Check if object is valid and allocated */
	    if((obj_num < 0) || (obj_num >= total_objects))
		return(-1);

	    obj = xsw_object[obj_num];
	    if(obj == NULL)
		return(-1);

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		return(-1);

	    /* Match all types of objects? */
	    if(type == -1)
		return(obj_num);

	    /* Match specific type */
	    if(obj->type == type)
		return(obj_num);
	    else
		return(-1);
	}


	/* Iterate through the list of objects using exact match
	 * first
	 */
	for(obj_num = 0; obj_num < total_objects; obj_num++)
	{
	    obj = xsw_object[obj_num];
	    if(obj == NULL)
		continue;

	    if((obj->type <= SW_OBJ_TYPE_GARBAGE) ||
	       (*obj->name == '\0')
	    )
		continue;

	    /* Match all types? */
	    if(type == -1)
	    {
		if(!strcasecmp(obj->name, name))
		    return(obj_num);
	    }
	    /* Match specific type */
	    else if(obj->type == type)
	    {
		if(!strcasecmp(obj->name, name))
		    return(obj_num);
	    }
	}

	/* Iterate through the list of objects a second time
	 * using partial match
	 */
	for(obj_num = 0; obj_num < total_objects; obj_num++)
	{
	    obj = xsw_object[obj_num];
	    if(obj == NULL)
		continue;

	    if((obj->type <= SW_OBJ_TYPE_GARBAGE) ||
	       (*obj->name == '\0')
	    )
		continue;

	    /* Match all types? */
	    if(type == -1)
	    {
		if(strcasestr(obj->name, name) != NULL)
		    return(obj_num);
	    }
	    /* Match specific type */
	    else if(obj->type == type)
	    {
		if(strcasestr(obj->name, name) != NULL)
		    return(obj_num);
	    }
	}

	return(-1);
}


/*
 *	Matches the object nearest to the specified object.
 *
 *	The type specifies the type of object to match or -1 for any
 *	type.
 */
int UNVMatchObjectNearest(
	xsw_object_struct **xsw_object, int total_objects,
	int obj_num, int type
)
{
	int i, closest_obj_num;
	float d, closest_distance, scanner_range;
	xsw_object_struct *obj, *tar_obj;

	if((xsw_object == NULL) || (total_objects <= 0))
	    return(-1);

	if(UNVIsObjectGarbage(xsw_object, total_objects, obj_num))
	    return(-1);

	obj = xsw_object[obj_num];

	/* Get the current effective scanner range in real units */
	scanner_range = UNVGetObjectEffectiveScannerRangePtr(obj);
	if(scanner_range <= 0.0f)
	    return(-1);

	closest_obj_num = -1;
	closest_distance = 0.0f;

	/* Iterate through the objects list */
	for(i = 0; i < total_objects; i++)
	{
	    tar_obj = xsw_object[i];
	    if(tar_obj == NULL)
		continue;

	    if(i == obj_num)
		continue;

	    /* Object type specified? */
	    if(type > SW_OBJ_TYPE_GARBAGE)
	    {
		/* Wrong object type? */
		if(tar_obj->type != type)
		    continue;
	    }
	    else
	    {
		if(tar_obj->type <= SW_OBJ_TYPE_GARBAGE)
		    continue;
	    }

	    /* Not within effective scanner range? */
	    if(!UNVIsObjectInEffectiveScannerRangePtr(
		obj, tar_obj, scanner_range
	    ))
		continue;

	    /* Calculate the distance between the two objects in real
	     * units
	     */
	    d = SWMath3DDistance(
		tar_obj->x - obj->x,
		tar_obj->y - obj->y,
		tar_obj->z - obj->z
	    ) - MAX((tar_obj->size / 1000.0f), 0.0f);

	    if(closest_obj_num > -1)
	    {
		if(d < closest_distance)
		{
		    closest_obj_num = i;
		    closest_distance = d;
		}
	    }
	    else
	    {
		closest_obj_num = i;
		closest_distance = d;
	    }
	}

	return(closest_obj_num);
}


/*
 *	Returns the IFF of ref_obj_num to tar_obj_num.
 */
int UNVMatchIFF(
	xsw_object_struct **xsw_object, int total_objects,
	int ref_obj_num, int tar_obj_num
)
{
	if(UNVIsObjectGarbage(xsw_object, total_objects, ref_obj_num))
	    return(SW_IFF_UNKNOWN);
	else if(UNVIsObjectGarbage(xsw_object, total_objects, tar_obj_num))
	    return(SW_IFF_UNKNOWN);
	else
	    return(UNVMatchIFFPtr(
		xsw_object[ref_obj_num],
		xsw_object[tar_obj_num]
	    ));
}

/*
 *	Returns the IFF of ref_obj to tar_obj
 */
int UNVMatchIFFPtr(
	xsw_object_struct *ref_obj, xsw_object_struct *tar_obj
)
{
	const char *s_ref, *s_tar;

	if((ref_obj == NULL) || (tar_obj == NULL))
	    return(SW_IFF_UNKNOWN);

	s_ref = ref_obj->empire;
	s_tar = tar_obj->empire;
	if((*s_ref == '\0') || (*s_tar == '\0'))
	    return(SW_IFF_UNKNOWN);

	if(strcasecmp(s_ref, s_tar))
	    return(SW_IFF_UNKNOWN);
	else
	    return(SW_IFF_FRIENDLY);
}




