#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/os.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/reality.h"

#include "mon.h"
#include "config.h"


#include "../include/cursors/std_arrow.xpm"
#include "../include/cursors/text_cur.xpm"
#include "../include/icons/info.h"
#include "../include/icons/warning.h"
#include "../include/icons/error.h"
#include "../include/icons/question.h"


static void MonSignalCB(int s);

int MonCreateMonitor(const int argc, char **argv);

void MonManageGUI(void *data);
void MonManageNet(void *data);
void MonManageNetReconnect(void *data);

void MonResetTimmers(const time_t cur_ms, const time_t cur_sec);

int MonInit(int argc, char **argv);
void MonMainIteration(void *data);
void MonMain(void *data);
void MonMainPop(void *data);
void MonMainPopAll(void *data);
void MonShutdown(void);


int		runlevel;
time_t		cur_millitime,
		cur_systime,
		lapsed_millitime;
float		time_compensation;

mon_option_struct	option;
mon_dname_struct	dname;
mon_next_struct		next;
mon_interval_struct	interval;

mon_color_struct	mon_color;
mon_font_struct		mon_font;
mon_image_struct	mon_image;
mon_cursor_struct	mon_cursor;

monitor_struct	**monitor;
int		total_monitors;

dialog_struct	*mon_dlg;
cdlg_struct	*mon_cdlg;


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Signal handler.
 */
static void MonSignalCB(int s)
{
	switch(s)
	{
	  case SIGHUP:
	    /* Reconnect all monitors */
	    if(total_monitors > 0)
	    {
		int i;
		monitor_struct *mon;
		for(i = 0; i < total_monitors; i++)
		{
		    mon = monitor[i];
		    if(mon == NULL)
			continue;

		    /* Not connected? */
		    if((mon->socket < 0) || mon->close_wait)
			continue;

		    /* No address set? */
		    if(STRISEMPTY(mon->address))
			continue;

		    /* Redraw */
		    if(mon->map_state)
			MonDraw(mon, MON_DRAW_AMOUNT_READOUT);

		    /* Reconnect */
		    MonNetConnect(mon, mon->address, mon->port, 45l);
		}
	    }
	    break;
	  case SIGINT:
	  case SIGTERM:
	  case SIGQUIT:
	  case SIGABRT:
	  case SIGKILL:
	    MonMainPopAll(NULL);
	    break;
	  case SIGPIPE:
	    break;
	  case SIGSEGV:
	    exit(1);
	    break;
	  case SIGCONT:
	    break;
	  case SIGSTOP:
	    break;
	}
}


/*
 *	Creates a new monitor.
 */
int MonCreateMonitor(const int argc, char **argv)
{
        int i, n;
        monitor_struct *m = MonNew(argc, argv);
	if(m == NULL)
	    return(-1);

        for(i = 0; i < total_monitors; i++)
        {
            if(monitor[i] == NULL)
                 break;
        }
        if(i < total_monitors)
        {
	    monitor[i] = m;
	    n = i;
        }
        else
        {
            n = MAX(total_monitors, 0);
            total_monitors = n + 1;
            monitor = (monitor_struct **)realloc(
                monitor,
                total_monitors * sizeof(monitor_struct *)
            );
            if(monitor == NULL)
            {
                total_monitors = 0;
		MonDelete(m);
                return(-3);
            }

	    monitor[n] = m;
        }

	MonMap(m);

	return(n);
}


/*
 *	Manages the GUI.
 */
void MonManageGUI(void *data)
{
	int events_handled, total_events_handled = 0;
	event_t event;

	/* Manage GUI events */
	while(OSWEventsPending() > 0)
	{
	    /* Reset events counter */
	    events_handled = 0;

	    /* Get event */
	    OSWWaitNextEvent(&event);

	    /* Manage widget event */ 
	    if(events_handled == 0)
		events_handled += WidgetManageGUI(&event);

	    /* Monitors */
	    if(events_handled == 0)
		events_handled += MonManageAll(&event);

	    /* Dialog */
	    if(events_handled == 0)
		events_handled += WidgetManage(mon_dlg, &event);

	    /* Confirmation Dialog */
	    if(events_handled == 0)
		events_handled += CDlgManage(mon_cdlg, &event);


	    /* Add up total events handled so far */
	    total_events_handled += events_handled;
	}

	/* Process queued widget events */
	WidgetMainIteration(NULL);
}

/*
 *	Manages the network connections.
 */
void MonManageNet(void *data)
{
	int i;
	monitor_struct *mon;

	for(i = 0; i < total_monitors; i++)
	{
	    mon = monitor[i];
	    if(mon == NULL)
		continue;

	    MonNetManage(mon);
	}
}

/*
 *	Checks if any monitor has lost connection and attempts to
 *	reconnect.
 */
void MonManageNetReconnect(void *data)
{
	int i;
	monitor_struct *mon;

	for(i = 0; i < total_monitors; i++)
	{
	    mon = monitor[i];
	    if(mon == NULL)
		continue;

	    /* Still connected? */
	    if((mon->socket > -1) || mon->close_wait)
		continue;

	    /* No address set? */
	    if(STRISEMPTY(mon->address))
		continue;

	    /* Redraw */
	    if(mon->map_state)
		MonDraw(mon, MON_DRAW_AMOUNT_READOUT);

	    /* Reconnect */
	    MonNetConnect(mon, mon->address, mon->port, 45l);
	}
}


/*
 *	Reset timmers.
 */
void MonResetTimmers(const time_t cur_ms, const time_t cur_sec)
{
	next.pipe_recv = cur_ms;
	next.systems_check = cur_ms;
	next.connect_check = cur_ms;
}


/*
 *	Initializes monitor.
 */
int MonInit(int argc, char **argv)
{
	int i, mon_num;
	const char *s, *arg;
	image_t *img;
	WColorStruct *c, color;
	mon_option_struct *opt;

	/* Set initial values */
	cur_millitime = OSWTimeMS();
	cur_systime = time(NULL);
	lapsed_millitime = 0l;
	time_compensation = 0.0f;

	opt = &option;
	opt->quiet_mode = False;

	interval.pipe_recv = PIPE_RECV_INT;
	interval.systems_check = SYSTEMS_CHECK_INT;
	interval.connect_check = CONNECT_CHECK_INT;

	MonResetTimmers(0l, 0l);


	/* Reset globals */
	strncpy(
	    dname.images, DEF_IMAGES_DIR,
	    sizeof(dname.images)
	);
	dname.images[sizeof(dname.images) - 1] = '\0';

	strncpy(
	    dname.server, DEF_SERVER_DIR,
	    sizeof(dname.server)
	);
	dname.server[sizeof(dname.server) - 1] = '\0';

	monitor = NULL;
	total_monitors = 0;


	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Quiet mode */
	    if(!strcasecmp(arg, "--quiet") ||
	       !strcasecmp(arg, "-quiet") ||
	       !strcasecmp(arg, "--q") ||
	       !strcasecmp(arg, "-q")
	    )
	    {
		opt->quiet_mode = True;
	    }
	    /* Images dir */
	    else if(!strcasecmp(arg, "--images_dir") ||
		    !strcasecmp(arg, "-images_dir") ||
		    !strcasecmp(arg, "--images-dir") ||
		    !strcasecmp(arg, "-images-dir") ||
		    !strcasecmp(arg, "--images") ||
		    !strcasecmp(arg, "-images") ||
		    !strcasecmp(arg, "--i") ||
		    !strcasecmp(arg, "-i")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    strncpy(
			dname.images, arg,
			sizeof(dname.images)
		    );
		    dname.images[sizeof(dname.images) - 1] = '\0';
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	    /* Server dir */
	    else if(!strcasecmp(arg, "--server_dir") ||
		    !strcasecmp(arg, "-server_dir") ||
		    !strcasecmp(arg, "--server-dir") ||
		    !strcasecmp(arg, "-server-dir") ||
		    !strcasecmp(arg, "--server") ||
		    !strcasecmp(arg, "-server") ||
		    !strcasecmp(arg, "--s") ||
		    !strcasecmp(arg, "-s")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    strncpy(
			dname.server, arg,
			sizeof(dname.server)
		    );
		    dname.server[sizeof(dname.server) - 1] = '\0';
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			argv[i - 1]
		    );
		}
	    }
	}


	/* Connect to GUI */
	if(OSWGUIConnect(argc, argv))
	    return(-1);

	/* Initialize widgets */  
	if(WidgetInitGlobals(argc, argv))
	    return(-1);


	/* Colors */
	OSWLoadPixelRGB(
	    &mon_color.readout_text,
	    0xff,
	    0x70,
	    0xd0
	);
	OSWLoadPixelRGB(
	    &mon_color.button_text,
	    0xf0,
	    0xa0,
	    0xff
	);
	OSWLoadPixelRGB(
	    &mon_color.messages_text,
	    0xff,
	    0xc0,
	    0xff
	);


	/* Fonts */
	OSWLoadFont(
	    &mon_font.readout,
"-adobe-helvetica-medium-r-normal-*-10-*-*-*-p-*-iso8859-1"
	);
	if(mon_font.readout == NULL)
	    OSWLoadFont(
		&mon_font.readout,
		"6x10"
	    );

	OSWLoadFont(
	    &mon_font.button,
"-adobe-helvetica-medium-r-normal-*-10-*-*-*-p-*-iso8859-1"
	);
	if(mon_font.button == NULL)
	    OSWLoadFont(
		&mon_font.button,
		"6x10"
	    );

	OSWLoadFont(
	    &mon_font.messages,
	    "7x14"
	);


	/* Cursors */
	c = &color;
	c->a = 0xff;
	c->r = 0xff;
	c->g = 0xff;
	c->b = 0xff;
	mon_cursor.standard = WidgetCreateCursorFromData(
	    std_arrow_xpm,
	    0, 0,
	    c, NULL
	);

	c = &color;
	c->a = 0xff;
	c->r = 0xff;
	c->g = 0xb3;
	c->b = 0xfa;
	mon_cursor.text = WidgetCreateCursorFromData(
	    text_cur_xpm,
	    8, 8,
	    c, NULL
	);



	/* Images */
	s = PrefixPaths(dname.images, IMG_NAME_READOUT_BKG);
	mon_image.readout_bkg = WidgetLoadImageFromTgaFile(s);

	s = PrefixPaths(dname.images, IMG_NAME_MESG_BKG);
	mon_image.mesg_bkg = WidgetLoadImageFromTgaFile(s);

	s = PrefixPaths(dname.images, IMG_NAME_BTN0);
	mon_image.btn_unarmed = WidgetLoadImageFromTgaFile(s);

	s = PrefixPaths(dname.images, IMG_NAME_BTN1);
	mon_image.btn_armed = WidgetLoadImageFromTgaFile(s);

	s = PrefixPaths(dname.images, IMG_NAME_BTN2);
	mon_image.btn_highlight = WidgetLoadImageFromTgaFile(s);

	mon_image.info = WidgetLoadImageFromTgaData(info_tga);
	mon_image.warning = WidgetLoadImageFromTgaData(warning_tga);
	mon_image.error = WidgetLoadImageFromTgaData(error_tga);
	mon_image.question = WidgetLoadImageFromTgaData(question_tga);

	s = PrefixPaths(dname.images, IMG_NAME_MONITOR_ICON);
	img = WidgetLoadImageFromTgaFile(s);
	mon_image.monitor_icon_pm = OSWCreatePixmapFromImage(img);
	OSWDestroyImage(&img);


	/* Dialogs */
	mon_dlg = MonDialogNew();
	mon_cdlg = NULL;


	/* Create the first Monitor */
	mon_num = MonCreateMonitor(argc, argv);


	return(0);
}

/*
 *	Manages one iteration of monitor.
 *
 *	Updates timmers and manages all resources.
 */
void MonMainIteration(void *data)
{
	time_t	cur_ms, cur_sec,
		new_systime = time(NULL),
		new_millitime = OSWTimeMS();
	float tc;

	/* Check if timmers have cycled */
	if((new_systime < cur_systime) ||
	   (new_millitime < cur_millitime)
	)
	{
	    /* Reset timmers and time lapses */
	    MonResetTimmers(new_millitime, new_systime);
	    lapsed_millitime = 0l;
	    time_compensation = 1.0f;
	}
	else
	{
	    /* Calculate time lapse ms and compensation coeff */
	    lapsed_millitime = new_millitime - cur_millitime;
	    time_compensation = (float)lapsed_millitime /
		(float)SW_CYCLE_LAPSE_MS;
	}
	 
	/* Set current time */
	cur_millitime = cur_ms = new_millitime;
	cur_systime = cur_sec = new_systime;
	tc = time_compensation;


	/* Manage the GUI */
	MonManageGUI(data);

	/* Manage the network connections */
	if(next.pipe_recv <= cur_ms)
	{
	    MonManageNet(data);
	    next.pipe_recv = cur_ms + interval.pipe_recv;
	}

	/* Reconnect check */
	if(next.connect_check <= cur_ms)
	{
	    MonManageNetReconnect(data);
	    next.connect_check = cur_ms + interval.connect_check;
	}

	/* Systems check */
	if(next.systems_check <= cur_ms)
	{

	    next.systems_check = cur_ms + interval.systems_check;
	}


	/* Switch to runlevel 1 if all windows are gone */
	if(runlevel > 0)
	{
	    bool_t need_pop = False;
	    int i;

	    for(i = 0; i < total_monitors; i++)
	    {
		if(monitor[i] != NULL)
		    break;
	    }
	    if(i >= total_monitors)
		need_pop = True;

	    if(need_pop)
		MonMainPopAll(NULL);
	}
}

/*
 *	Manages Monitor.
 *
 *      Pushes a monitor main loop and manages and blocks until
 *	MonitorMainPop() is called.
 */
void MonMain(void *data)
{
	const int entering_runlevel = runlevel;
	runlevel++;
	while(runlevel > entering_runlevel)
	{
	    usleep(8000);
	    MonMainIteration(data);
	}
}

/*
 *	Pops off the highest Monitor main loop.
 */
void MonMainPop(void *data)
{
	if(runlevel > 0)
	    runlevel--;
}
 
/*
 *      Pops off all Monitor main loops.
 */
void MonMainPopAll(void *data)
{
	runlevel = 0;
}

/*
 *	Shuts down Monitor.
 */
void MonShutdown(void)
{
	int i;

	/* Dialogs */
	MonDialogDelete(mon_dlg);
	mon_dlg = NULL;
	MonConfirmationDialogDelete(mon_cdlg);
	mon_cdlg = NULL;

	/* Monitors */
	for(i = 0; i < total_monitors; i++)
	    MonDelete(monitor[i]);
	free(monitor);
	monitor = NULL;
	total_monitors = 0;

	/* Cursors */
	WidgetDestroyCursor(&mon_cursor.text);        
	WidgetDestroyCursor(&mon_cursor.standard);

	/* Images */
	OSWDestroyPixmap(&mon_image.monitor_icon_pm);
	OSWDestroyImage(&mon_image.info);
	OSWDestroyImage(&mon_image.warning);
	OSWDestroyImage(&mon_image.error);
	OSWDestroyImage(&mon_image.question);
	OSWDestroyImage(&mon_image.btn_highlight);
	OSWDestroyImage(&mon_image.btn_armed);
	OSWDestroyImage(&mon_image.btn_unarmed);
	OSWDestroyImage(&mon_image.mesg_bkg);
	OSWDestroyImage(&mon_image.readout_bkg);

	/* Fonts */
	OSWUnloadFont(&mon_font.messages);
	OSWUnloadFont(&mon_font.button);
	OSWUnloadFont(&mon_font.readout);

	/* Colors */
	OSWDestroyPixel(&mon_color.readout_text);
	OSWDestroyPixel(&mon_color.button_text);
	OSWDestroyPixel(&mon_color.messages_text);

	/* Widgets */
	WidgetDestroyGlobals();

	/* GUI */
	OSWGUIDisconnect();
}


int main(int argc, char *argv[])
{
	int i, status;
	const char *arg;

	runlevel = 0;

	/* Set the signals to watch for */
	signal(SIGINT, MonSignalCB);
	signal(SIGTERM, MonSignalCB);
	signal(SIGQUIT, MonSignalCB);
	signal(SIGABRT, MonSignalCB);
	signal(SIGKILL, MonSignalCB);
	signal(SIGPIPE, MonSignalCB);
	signal(SIGSEGV, MonSignalCB);

	/* Handle help and version arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help */
	    if(!strcasecmp(arg, "--help") ||
	       !strcasecmp(arg, "-help") ||
	       !strcasecmp(arg, "--h") ||
	       !strcasecmp(arg, "-h") ||
	       !strcasecmp(arg, "-?") ||
	       !strcasecmp(arg, "/?")
	    )
	    {
		MonPrintHelpStdout();
		return(0);
	    }
	    /* Version */
	    else if(!strcasecmp(arg, "--version") ||
		    !strcasecmp(arg, "-version")
	    )
	    {
		MonPrintVersionStdout();
		return(0);
	    }
	}    

	/* Initialize */
	status = MonInit(argc, argv);
	if(status)
	{
	    MonShutdown();
	    return(-status);
	}

	/* Manage */
	MonMain(NULL);

	/* Shutdown */
	MonShutdown();

	return(0);
}
