#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../include/string.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/objects.h"

#include "mon.h"
#include "mesgwin.h"
#include "config.h"


int MonBtnInit(   
	mon_push_button_struct *btn,
	win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	char *label,
	void *client_data,
	int (*func_cb)(void *)
);
void MonBtnDraw(mon_push_button_struct *btn);
int MonBtnManage(mon_push_button_struct *btn, event_t *event);
void MonBtnDestroy(mon_push_button_struct *btn);

int MonIsAllocated(const int n);
monitor_struct *MonGetPtr(const int n);

static void MonDrawCB(void *widget, void *data);

monitor_struct *MonNew(const int argc, char **argv);
void MonDraw(monitor_struct *m, const int amount);
void MonQueueDraw(monitor_struct *m, const int amount);
int MonManage(monitor_struct *m, event_t *event);
int MonManageAll(event_t *event);
void MonMap(monitor_struct *m);
void MonUnmap(monitor_struct *m);
void MonDelete(monitor_struct *m);


#define MON_DEF_WIDTH			200
#define MON_DEF_HEIGHT			60

#define MON_BTN_WIDTH			80
#define MON_BTN_HEIGHT			12

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int MonBtnInit(
	mon_push_button_struct *btn,
	win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	char *label,
	void *client_data,
	int (*func_cb)(void *)
)
{
	if(btn == NULL)
	    return(-1);

	btn->x = x;
	btn->y = y;
	btn->width = width;
	btn->height = height;
	btn->parent = parent;
	btn->label = (label != NULL) ? strdup(label) : NULL;
	btn->client_data = client_data;
	btn->func_cb = func_cb;

	btn->state = MON_BTN_STATE_UNARMED;

	return(0);
}


void MonBtnDraw(mon_push_button_struct *btn)
{
	font_t *font;
	image_t *img;

	if(btn == NULL)
	    return;

	if((btn->width == 0) || (btn->height == 0))
	    return;

	font = mon_font.button;

	switch(btn->state)
	{
	  case MON_BTN_STATE_HIGHLIGHTED:
	    img = mon_image.btn_highlight;
	    break;
	  case MON_BTN_STATE_ARMED:
	    img = mon_image.btn_armed;
	    break;
	  default:
	    img = mon_image.btn_unarmed;
	    break;
	}

	if(widget_global.force_mono)
	{
	    OSWSetForeground(osw_gui[0].white_pix);
	    OSWDrawRectangle(
		btn->parent,
		btn->x, btn->y,
		btn->width, btn->height
	    );
	}
	else
	{
	    WidgetPutImageNormal(
	        btn->parent,
	        img,
	        btn->x, btn->y,
	        True
	    );
	    OSWSetForeground(mon_color.button_text);
	}

	if(btn->label != NULL)
	{
	    int	x, y,
		lbearing, swidth, ascent, descent, font_height;
	    const char *s = btn->label;

	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    x = btn->x + (((int)btn->width - swidth) / 2);
	    y = btn->y + (((int)btn->height - font_height) / 2);
	    OSWDrawString(
	        btn->parent,
		x - lbearing, y + ascent,
		s
	    );
	}
}

int MonBtnManage(mon_push_button_struct *btn, event_t *event)
{
	int x, y;
	int events_handled = 0;

	if((btn == NULL) || (event == NULL))
	    return(events_handled);

	if(event->xany.window != btn->parent)
	    return(events_handled);

	/* We now know that the event occured on the button's parent */
	switch(event->type)
	{
	  case ButtonPress:
	    x = event->xbutton.x;
	    y = event->xbutton.y;

	    if((x >= btn->x) &&
	       (y >= btn->y) &&
	       (x < (btn->x + (int)btn->width)) &&
	       (y < (btn->y + (int)btn->height))
	    )
	    {
		btn->state = MON_BTN_STATE_ARMED;
		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    x = event->xbutton.x;
	    y = event->xbutton.y;

	    if((x >= btn->x) &&
	       (y >= btn->y) &&
	       (x < (btn->x + (int)btn->width)) &&
	       (y < (btn->y + (int)btn->height))
	    )
	    {
		if(btn->state == MON_BTN_STATE_ARMED)
		{
		    if(btn->func_cb != NULL)
			btn->func_cb(btn->client_data);
		}

		btn->state = MON_BTN_STATE_HIGHLIGHTED;
		events_handled++;
	    }
	    break;

	  case MotionNotify:
	    x = event->xmotion.x;
	    y = event->xmotion.y;
	    if((x >= btn->x) &&
	       (y >= btn->y) &&
	       (x < (btn->x + (int)btn->width)) &&
	       (y < (btn->y + (int)btn->height))
	    )
	    {
		if(btn->state == MON_BTN_STATE_UNARMED)
		{
		    btn->state = MON_BTN_STATE_HIGHLIGHTED;
		    events_handled++;
		}
	    }
	    else if(btn->state != MON_BTN_STATE_UNARMED)
	    {
		btn->state = MON_BTN_STATE_UNARMED;
		events_handled++;
	    }
	    break;

	  case LeaveNotify:
	    btn->state = MON_BTN_STATE_UNARMED;
	    events_handled++;
	    break;
	}


	return(events_handled);
}


void MonBtnDestroy(mon_push_button_struct *btn)
{
	if(btn == NULL)
	    return;

	btn->x = 0;
	btn->y = 0;
	btn->width = 0;
	btn->height = 0;
	btn->parent = 0;

	free(btn->label);
	btn->label = NULL;

	btn->client_data = NULL;
	btn->func_cb = NULL;

	btn->state = MON_BTN_STATE_UNARMED;
}


int MonIsAllocated(const int n)
{
	if((monitor == NULL) ||
	   (n < 0) || (n >= total_monitors)
	)
	    return(0);
	else if(monitor[n] == NULL)
	    return(0);
	else
	    return(1);
}

monitor_struct *MonGetPtr(const int n)
{
	if((monitor == NULL) ||
	   (n < 0) || (n >= total_monitors)
	)
	    return(NULL);
	else
	    return(monitor[n]);
}


static void MonDrawCB(void *widget, void *data)
{
	int amount = (int)data;
	monitor_struct *m = MONITOR(widget);
	if(m == NULL)
	    return;

	if(m->map_state)
	    MonDraw(m, amount);
}


monitor_struct *MonNew(const int argc, char **argv)
{
	int i, x, y;
	unsigned int width, height;
	const char *arg;
	char cwd[PATH_MAX];
	win_t w, parent;
	image_t *img;
	menu_struct *menu;
	osw_gui_struct *gui = &osw_gui[0];
	monitor_struct *m = MONITOR(calloc(
	    1, sizeof(monitor_struct)
	));
	if(m == NULL)
	    return(NULL);

	if(getcwd(cwd, sizeof(cwd)) != NULL)
	    cwd[sizeof(cwd) - 1] = '\0';
	else
	    *cwd = '\0';

	/* Get position and sizes */
	if(gui->def_geometry_set)
	{
	    m->x = gui->def_toplevel_x;
	    m->y = gui->def_toplevel_y;
	    m->width = gui->def_toplevel_width;
	    m->height = gui->def_toplevel_height;
	}
	else
	{
	    m->x = 0;
	    m->y = 0;
	    m->width = MON_DEF_WIDTH;
	    m->height = MON_DEF_HEIGHT;
	}

	img = mon_image.readout_bkg;
	if(img != NULL)
	{
	    m->width = img->width;
	    m->height = img->height;
	}

	free(m->address);
	m->address = STRDUP(DEF_ADDRESS);

	m->port = DEF_PORT;

	free(m->name);
	m->name = STRDUP(DEF_LOGIN_NAME);
	free(m->password);
	m->password = STRDUP(DEF_LOGIN_PASSWORD);

	m->close_wait = False;
	m->socket = -1;

	m->last_error_mesg = NULL;

	m->frame = 0;

	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* User login name and password */
	    if(!strcasecmp(arg, "--user") ||
	       !strcasecmp(arg, "-user") ||
	       !strcasecmp(arg, "--u") ||
	       !strcasecmp(arg, "-u") ||
	       !strcasecmp(arg, "--login") ||
	       !strcasecmp(arg, "-login") ||
	       !strcasecmp(arg, "--l") ||
	       !strcasecmp(arg, "-l")
	    )
	    {
		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    free(m->name);
		    m->name = STRDUP(arg);
		}

		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		{
		    free(m->password);
		    m->password = STRDUP(arg);
		}
	    }
	    /* All else assume argument is address and port */
	    else if((*arg != '-') &&
		    (*arg != '+') &&
		    (m->socket < 0)
	    )
	    {
		free(m->address);
		m->address = STRDUP(arg);

		i++;
		arg = (i < argc) ? argv[i] : NULL;
		if(arg != NULL)
		    m->port = ATOI(arg);

		MonNetConnect(
		    m,
		    m->address,
		    m->port,
		    45l
		);
	    }
	}


	/* Create toplevel */
	parent = gui->root_win;
	OSWCreateWindow(
	    &w, parent,
	    m->x, m->y,
	    m->width, m->height
	);
	m->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    NOT_REACHABLE_TITLE_STRING,	/* Title */
	    PROG_NAME,			/* Icon title */
	    ((mon_image.monitor_icon_pm == 0) ?
		widget_global.std_icon_pm : mon_image.monitor_icon_pm),
	    !gui->def_geometry_set,	/* Let WM set coordinates? */
	    m->x, m->y,
	    m->width, m->height,
	    m->width, m->height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(
	    w,
	    OSW_EVENTMASK_TOPLEVEL | ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask | ExposureMask |
	    EnterWindowMask | LeaveWindowMask
	);
 
	parent = w;

	/* Initialize buttons */
	img = mon_image.btn_unarmed;
	x = ((img == NULL) ?
	    ((int)m->width - MON_BTN_WIDTH - 5) :
	    ((int)m->width - (int)img->width - 5)
	);
	y = 5;
	width = ((img == NULL) ? MON_BTN_WIDTH : img->width);
	height = ((img == NULL) ? MON_BTN_HEIGHT : img->height);
	MonBtnInit(
	    &m->shutdown_btn, parent,
	    x, y,
	    width, height,
	    "Shutdown",
	    m,
	    MonShutdownPBCB
	);

	y += (5 + (int)height);
	MonBtnInit(
	    &m->messages_btn, parent,
	    x, y,
	    width, height,
	    "Messages",
	    m, MonMessagesPBCB
	);

	/* Menu */
	menu = &m->menu;
	MenuInit(
	    menu, gui->root_win,
	    MonMenuActivateCB, m
	);
	MenuItemAppend(
	    menu,
	    MENU_ITEM_LABEL,
	    "Next Display",
	    None, None,
	    MON_ACTION_DISPLAY_NEXT
	);
	MenuItemAppend(
	    menu,
	    MENU_ITEM_LABEL,
	    "Previous Display",
	    None, None,
	    MON_ACTION_DISPLAY_PREV
	);
	MenuItemAppend(
	    menu,
	    MENU_ITEM_SEPARATOR,
	    NULL,
	    None, None,
	    MON_ACTION_NONE
	);
	MenuItemAppend(
	    menu,
	    MENU_ITEM_LABEL,
	    "Messages...",
	    None, None,
	    MON_ACTION_MAP_MESSAGES
	);
	MenuItemAppend(
	    menu,
	    MENU_ITEM_SEPARATOR,
	    NULL,
	    None, None,
	    MON_ACTION_NONE
	);
	MenuItemAppend(
	    menu,
	    MENU_ITEM_LABEL,
	    "Connect...",
	    None, None,
	    MON_ACTION_NEW_CONNECTION
	);
	MenuItemAppend(
	    menu,
	    MENU_ITEM_LABEL,
	    "New Monitor...",
	    None, None,
	    MON_ACTION_NEW_MONITOR
	);

	/* Connect window */
	ConWinInit(
	    &m->cw,
	    MonConnectCB, m
	);

	/* Messages window */
	MesgWinInit(&m->mesgwin, argc, argv);

	return(m);
}


void MonDraw(monitor_struct *m, const int amount)
{
	int x, y, yp, len;
	win_t w;
	pixmap_t pixmap;
	image_t *img;

	if(m == NULL)
	    return;

	/* Recreate buffers as needed */
	if(m->toplevel_buf == None)
	{
	    if(OSWCreatePixmap(
		&m->toplevel_buf, m->width, m->height
	    ))
		return;
	}

	/* Redraw toplevel */
	if(amount == MON_DRAW_AMOUNT_COMPLETE)
	{
	    unsigned int width, height;

	    w = m->toplevel;
	    pixmap = m->toplevel_buf;
	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );
	    img = mon_image.readout_bkg;

	    if(!widget_global.force_mono &&
	       (img != NULL)
	    )
		OSWPutImageToDrawable(img, pixmap);
	    else
		OSWClearPixmap(
		    pixmap, width, height,
		    osw_gui[0].black_pix
		);

	    OSWPutDrawableToWindow(w, pixmap);
	}

	/* Redraw readout */
	if((amount == MON_DRAW_AMOUNT_COMPLETE) ||
	   (amount == MON_DRAW_AMOUNT_READOUT)
	)
	{
	    unsigned int width, height;

	    w = m->toplevel;
	    pixmap = m->toplevel_buf;
	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    img = mon_image.readout_bkg;

	    x = 5;
	    y = 5;
	    width = 200;
	    height = (int)height - 10;

	    OSWSetFont(mon_font.readout);

	    if(!widget_global.force_mono &&
	       (img != NULL)
	    )
	    {
		OSWPutImageToDrawableSect(
		    img, pixmap,
		    x, y, 		/* Target */
		    x, y,		/* Source */
		    width, height
		);

		OSWSetForeground(mon_color.readout_text);
	    }
	    else
	    {
		OSWSetForeground(osw_gui[0].black_pix);
		OSWDrawSolidRectangle(
		    pixmap,
		    x, y,
		    width, height
		);       
	 
		OSWSetForeground(osw_gui[0].white_pix);
	    }

	    /* Connected or not? */
	    if(m->socket < 0)
	    {
		/* Not connected, draw error message */
		const char *s = m->last_error_mesg;
		if(s == NULL)
		    s = "Error:\n    Universe is unreachable.";

		yp = y + 9 + (12 * 0);
		while(s != NULL)
		{
		    len = strlinelen(s);
		    OSWDrawStringLimited(
			pixmap,
			x + 4,
			yp,
			s,
			len
		    );
		    s = strchr(s, '\n');
		    if(s != NULL)
			s += 1;

		    yp += 12;
		}
	    }
	    else
	    {
		char *s, *time_str1, *time_str2;
		mon_stats_struct *stats = &m->stats;

		/* Draw stats for curent frame */
		switch(m->frame)
		{
		  case 2:
		    /* Address, port, and PID */
		    s = (char *)malloc(
			STRLEN(m->address) + 80
		    );
		    sprintf(
			s,
			"Address: %s",
			m->address
		    );
		    OSWDrawString(
			pixmap,
			x + 4,
			y + 9 + (12 * 0),
			s
		    );
		    free(s);

		    s = (char *)malloc(40 + 80);
		    sprintf(
			s,
			"Port: %i",
			m->port
		    );
		    OSWDrawString(pixmap,
			x + 4,
			y + 9 + (12 * 1),
			s
		    );
		    free(s);

		    s = (char *)malloc(40 + 80);
		    sprintf(
			s,
			"PID: %i",
			stats->pid
		    );
		    OSWDrawString(pixmap,
			x + 4,
			y + 9 + (12 * 2),
			s
		    );
		    free(s);
		    break;

		  case 1:
		    /* Full memory info and break down */
		    s = (char *)malloc(40 + 80);
		    sprintf(
			s,
			"Total Memory Used: %ld bytes",
			stats->mem_total
		    );
		    OSWDrawString(
			pixmap,
			x + 4,
			y + 9 + (12 * 0), 
			s
		    );
		    free(s);

		    s = (char *)malloc(40 + 80);
		    sprintf(
			s,
			"Objects Memory: %ld bytes",
			stats->mem_obj
		    );
		    OSWDrawString(
			pixmap,
			x + 4,
			y + 9 + (12 * 1),
			s
		    );
		    free(s);

		    s = (char *)malloc(40 + 80);
		    sprintf(
			s,
			"Connections Memory: %ld bytes",
			stats->mem_con
		    );
		    OSWDrawString(
			pixmap,
			x + 4,
			y + 9 + (12 * 2),
			s
		    );
		    free(s);
		    break;

		  default:
		    /* Number of connections */
		    s = (char *)malloc(40 + 40 + 80);
		    sprintf(
			s,
"Connections: %i  Guests: %i",
			stats->total_connections,
			stats->guest_connections
		    );
	            OSWDrawString(
			pixmap,
		        x + 4,
		        y + 9 + (12 * 0),
		        s
	            );
		    free(s);

	            /* Objects */
		    s = (char *)malloc(40 + 40 + 80);
		    sprintf(
			s,
		        "Objects: %i  %ld bytes",
		        stats->total_objects,
		        stats->mem_obj
	            );
		    OSWDrawString(
			pixmap,
			x + 4,
			y + 9 + (12 * 1),
			s
		    );
		    free(s);

		    /* Uptime */
		    time_str1 = STRDUP(
			StringFormatTimePeriod(stats->uptime)
		    );
		    time_str2 = STRDUP(
			StringFormatTimePeriod(stats->next_save)
		    );
		    s = (char *)malloc(
			STRLEN(time_str1) + STRLEN(time_str2) + 80
		    );
		    sprintf(
			s,
"Uptime: %s  Next Save: %s",
			time_str1, time_str2
		    );
		    OSWDrawString(
			pixmap,
			x + 4,
			y + 9 + (12 * 2),
			s
		    );
		    free(s);
		    free(time_str1);
		    free(time_str2);
		    break;
		}
	    }

	    /* Put to window */
	    OSWCopyDrawablesCoord(
		w, pixmap,
		x, y,
		width, height,
		x, y
	    );
	}


	/* Redraw buttons */
	if((amount == MON_DRAW_AMOUNT_COMPLETE) ||
	   (amount == MON_DRAW_AMOUNT_BUTTONS)
	)
	{
	    MonBtnDraw(&m->shutdown_btn);
	    MonBtnDraw(&m->messages_btn);
	}
}

void MonQueueDraw(monitor_struct *m, const int amount)
{
	if(m == NULL)
	    return;

	WidgetQueueDrawAction(MonDrawCB, m, (void *)amount);
}

int MonManage(monitor_struct *m, event_t *event)
{
	int x, y, root_x, root_y;
	time_t t;
	win_t w;
	int events_handled = 0;

	if((m == NULL) || (event == NULL))
	    return(events_handled);

	if(!m->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	/* Buttons */
	events_handled += MonBtnManage(&m->shutdown_btn, event);
	events_handled += MonBtnManage(&m->messages_btn, event);

	/* Return of one of the buttons managed the event */
	if(events_handled > 0)
	{
	    MonQueueDraw(m, MON_DRAW_AMOUNT_BUTTONS);
	    return(events_handled);
	}

	/* Manage events for monitor normally */
	switch(event->type)
	{
	  case ButtonPress:
	    if(w == m->toplevel)
	    {
		switch(event->xbutton.button)
		{
		  case Button3:
		    /* Map menu */
		    OSWGetPointerCoords(
			m->toplevel,
			&root_x, &root_y,
			&x, &y
		    );

		    if(m->menu.map_state)
			MenuUnmap(&m->menu);
		    else
			MenuMapCoordinates(&m->menu, root_x, root_y);
		    break;

		  default:
		    /* Change frame */
		    m->frame++;
		    if(m->frame >= MON_MAX_FRAMES)
			m->frame = 0;
		    MonQueueDraw(m, MON_DRAW_AMOUNT_READOUT);
		    break;
		}
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == m->toplevel)
	    {
		MonQueueDraw(m, MON_DRAW_AMOUNT_COMPLETE);
	        events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == m->toplevel)
	    {
		m->has_focus = True;
		MonQueueDraw(m, MON_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == m->toplevel)
	    {
		m->has_focus = False;
		MonQueueDraw(m, MON_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case MapNotify:
	    if(w == m->toplevel)
	    {
		if(!m->map_state)
		    MonMap(m);
		events_handled++;
	    }
	    break;

	  case UnmapNotify:
	    if(w == m->toplevel)
	    {
		if(m->map_state)  
		    MonUnmap(m);
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus */  
	    if(OSWEventIsTakeFocus(m->toplevel, event, &t))
	    {
		OSWGrabFocusTime(m->toplevel, t);
		events_handled++;
	    }
	    /* WM Delete Window */
	    else if(OSWEventIsDeleteWindow(m->toplevel, event))
	    {
		MonUnmap(m);
		events_handled++;
	    }
	    break;
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(&m->menu, event);

	return(events_handled);
}


int MonManageAll(event_t *event)
{
	int i, m_deleted, events_handled = 0;
	monitor_struct *m;

	if(event == NULL)
	    return(events_handled);

	/* Manage each monitor window */
	for(i = 0, m_deleted = 0; i < total_monitors; i++)
	{
	    m = monitor[i];
	    if(m == NULL) 
		continue;

	    if(events_handled == 0)
		events_handled += MonManage(m, event);
	    if(events_handled == 0)
		events_handled += MesgWinManage(&m->mesgwin, event);
	    if(events_handled == 0)
		events_handled += ConWinManage(&m->cw, event);

	    /* Is this monitor unmapped? */
	    if(!m->map_state)
	    {
		/* Delete this monitor */
		MonDelete(m);
		monitor[i] = m = NULL;
		m_deleted++;
		continue;
	    }
	}

	/* Update window menus as needed */
/*
	if(m_deleted > 0)
	    ???
 */

	return(events_handled);
}


void MonMap(monitor_struct *m)
{
	if(m == NULL)
	    return;

	OSWMapRaised(m->toplevel);
	m->map_state = True;
	OSWGrabFocus(m->toplevel);
}


void MonUnmap(monitor_struct *m)
{
	if(m == NULL)
	    return;

	OSWUnmapWindow(m->toplevel);
	m->map_state = False;

	OSWDestroyPixmap(&m->toplevel_buf);
}


void MonDelete(monitor_struct *m)
{
	mon_stats_struct *stats;

	if(m == NULL)
	    return;

	/* Close connection properly */
	if(m->socket > -1)
	{
	    if(m->close_wait)
	    {
		MonNetWaitClose(m, 45l);
	    }
	    else
	    {
		MonNetSendLogout(m);
		MonNetShutdownConnection(
		    m, MON_NET_SHUTDOWN_RDWR, True
		);
		MonNetWaitClose(m, 45l);
	    }
	}
	MonNetDisconnect(m, 45l);

	free(m->address);
	m->address = NULL;
	m->port = 0;
	free(m->name);
	m->name = NULL;
	free(m->password);
	m->password = NULL;

	free(m->last_error_mesg);
	m->last_error_mesg = NULL;

	m->frame = 0;

	stats = &m->stats;
	free(stats->title);
	stats->title = NULL;

	MesgWinDestroy(&m->mesgwin);
	ConWinDestroy(&m->cw);

	WidgetDestroy(&m->menu);

	MonBtnDestroy(&m->shutdown_btn);
	MonBtnDestroy(&m->messages_btn);

	OSWDestroyWindow(&m->toplevel);
	OSWDestroyPixmap(&m->toplevel_buf);

	free(m);
}
