#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <signal.h>
#include <unistd.h>

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/objects.h"

#include "mon.h"
#include "mesgwin.h"


static bool_t MonAddressIsLocal(const char *address);

int MonShutdownPBCB(void *data);
int MonMessagesPBCB(void *data);
void MonMenuActivateCB(menu_struct *menu, int id, void *data);
void MonConnectCB(
	const char *address,
	int port,
	const char *name,
	const char *password,
	void *data
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the address is local.
 */
static bool_t MonAddressIsLocal(const char *address)
{
	if(address == NULL)
	    return(False);

	if(!strcasecmp(address, "127.0.0.1"))
	    return(True);
	if(!strcasecmp(address, "localhost"))
	    return(True);

	return(False);
}

/*
 *	Shutdown procedure callback.
 *
 *	Checks if monitor is connected to a universe on the
 *	same computer, if so then it sends a SIGTERM to
 *	the server process.
 */
int MonShutdownPBCB(void *data)
{
	int status;
	mon_stats_struct *stats;
	monitor_struct *m = MONITOR(data);
	if(m == NULL)
	    return(-1);

	/* Check if the universe that we are connected to
	 * is *not* on the same computer as this process
	 */
	if(!MonAddressIsLocal(m->address))
	{
	    char *s = (char *)malloc(
		STRLEN(m->address) + 40 + (80 * 8)
	    );
	    sprintf(
		s,
"Unable to shutdown the universe at:\n\
\n\
    %s  Port: %i\n\
\n\
The universe that you are connected to does not appear\n\
to be running on the same physical machine as this process.\n\
You must be running this program on the same computer as\n\
the universe's server to shut it down.",
		m->address, m->port
	    );
	    MonDialog(
		"Shutdown Failed",
		mon_image.warning,
		s,
		m->toplevel
	    );
	    free(s);
	    return(0);
	}


	/* Send SIGTERM to server process. */
	stats = &m->stats;
	if(stats->pid > 0)
	{
	    status = kill(stats->pid, SIGTERM);
	    if(status)
	    {
		switch(errno)
		{
		  case EINVAL:
		    fprintf(   
			stderr,
			"%i: Internal error, invalid signal `%i'.\n",
			stats->pid, SIGTERM
		    );
		    break;

		  case ESRCH:
		    fprintf(
			stderr, 
			"%i: No such process.\n",
			stats->pid
		    );
		    break;

		  case EPERM:
		    fprintf(
			stderr,
			"%i: Permission denied.\n",
			stats->pid
		    );
		    break;
		}
	    }
	}

	MonUnmap(m);

	return(0);
}

/*
 *	Message button callback.
 *
 *	Maps messages window.
 */
int MonMessagesPBCB(void *data)
{
	monitor_struct *m = MONITOR(data);
	if(m == NULL)
	    return(-1);

	MesgWinMap(&m->mesgwin);

	return(0);
}


/*
 *	Menu callback.
 */
void MonMenuActivateCB(menu_struct *menu, int id, void *data)
{
	monitor_struct *m = MONITOR(data);
	if(m == NULL)
	    return;

	switch(id)
	{
	  case MON_ACTION_NONE:
	    break;

	  case MON_ACTION_DISPLAY_NEXT:
	    if(m == NULL)
		break;

	    m->frame++;
	    if(m->frame > MON_MAX_FRAMES)
		m->frame = 0;

	    if(m->map_state)
		MonDraw(m, MON_DRAW_AMOUNT_READOUT);
	    break;

	  case MON_ACTION_DISPLAY_PREV:
	    if(m == NULL)
		break;

	    m->frame--;
	    if(m->frame < 0)
		m->frame = MON_MAX_FRAMES - 1;

	    if(m->map_state)
	        MonDraw(m, MON_DRAW_AMOUNT_READOUT);
	    break;

	  case MON_ACTION_MAP_MESSAGES:
	    if(m == NULL)
		break;
	    MesgWinMap(&m->mesgwin);
	    break;

	  case MON_ACTION_NEW_CONNECTION:
	    if(m != NULL)
	    {
		int x, y;
		unsigned int width, height;
		con_win_struct *cw = &m->cw;

		PromptSetS(&cw->address_prompt, m->address);
		PromptSetI(&cw->port_prompt, m->port);
		PromptSetS(&cw->name_prompt, m->name);
		PromptSetS(&cw->password_prompt, m->password);

		WidgetCenterWindow(
		    cw->toplevel,
		    WIDGET_WINDOW_RELATIVE_POINTER
		);
		OSWGetWindowGeometry(
		    cw->toplevel, &x, &y, &width, &height, NULL
		);
		cw->x = x;
		cw->y = y;
		cw->width = width;
		cw->height = height;

		ConWinMap(cw);
	    }
	    break;

	  case MON_ACTION_NEW_MONITOR:
	    MonCreateMonitor(0, NULL);
	    break;
	}
}

/*
 *	Connect window callback.
 *
 *	Connects the monitor to the specified address.
 */
void MonConnectCB(
	const char *address,
	int port,
	const char *name,
	const char *password,
	void *data
)
{
	monitor_struct *m = MONITOR(data);
	if(m == NULL)
	    return;

	free(m->name);
	m->name = STRDUP(name);

	free(m->password);
	m->password = STRDUP(password);

	MonNetConnect(
	    m,
	    address,
	    port,
	    45l
	);
}
