#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>

#include <netdb.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <fcntl.h>

#include "../include/string.h"
#include "../include/netio.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/auxstatcodes.h"

#include "mon.h"
#include "config.h"


static void MonNetSetErrorMesg(monitor_struct *m, const char *msg);

int MonNetConnect(
	monitor_struct *m, const char *address, int port,
	time_t timeout
);

int MonNetSend(monitor_struct *m, const char *buf);
void MonNetSendLogout(monitor_struct *m);

static int MonNetManageIterate(monitor_struct *m, const char *buf);
int MonNetManage(monitor_struct *m);

void MonNetShutdownConnection(
	monitor_struct *m, int how, bool_t is_initiator
);
void MonNetCloseConnection(monitor_struct *m, time_t timeout);
void MonNetWaitClose(monitor_struct *m, time_t timeout);
void MonNetDisconnect(monitor_struct *m, time_t timeout);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sets new error message on monitor m.
 */
static void MonNetSetErrorMesg(monitor_struct *m, const char *msg)
{
	if(m == NULL)
	    return;

	free(m->last_error_mesg);
	m->last_error_mesg = STRDUP(msg);
}


/*
 *	Connects to the specified universe.
 */        
int MonNetConnect(
	monitor_struct *m, const char *address, int port,
	time_t timeout
)
{
	char *daddress;
	struct hostent *he;
	struct sockaddr_in haddr;

	if((m == NULL) || (address == NULL) || (port < 0))
	    return(-1);

	daddress = strdup(address);
	if(daddress == NULL)
	    return(-3);

	/* Close socket and reset net values as needed */
	MonNetDisconnect(m, timeout);

	/* Reset window title */
	OSWSetWindowTitle(m->toplevel, NOT_REACHABLE_TITLE_STRING);
	OSWSetWindowTitle(m->mesgwin.toplevel, NOT_REACHABLE_TITLE_STRING);

	/* Set new address and port on the monitor */
	free(m->address);
	m->address = strdup(daddress);
	m->port = port;


	/* Get new host address info */
	he = gethostbyname(daddress);
	if(he == NULL)
	{
	    const int error_code = h_errno;
	    const char *error_msg;
	    char *s;
	    switch(error_code)
	    {
	      case HOST_NOT_FOUND:
		error_msg = "Unable to find host";
		break;
	      case NO_ADDRESS:
		error_msg = "Unable to obtain the IP address for the host";
		break;
	      case NO_RECOVERY:
		error_msg = "A non-recoverable name server error occurred\n\
while attempting to obtain the information for the host";
		break;
	      case TRY_AGAIN:
		error_msg = "A tempory error occurred on the name server\n\
while attempting to obtain the information for the host";
		break;
	      default:
		error_msg = "Unable to find host";
		break;
	    }
	    s = (char *)malloc(
		STRLEN(daddress) + STRLEN(error_msg) + (3 * 80)
	    );
	    sprintf(
		s,
"%s:\n\
\n\
    %s",
		daddress, error_msg
	    );
	    MonNetSetErrorMesg(m, s);
	    free(s);
	    free(daddress);
	    return(-1);
	}

	m->socket = socket(AF_INET, SOCK_STREAM, 0);
	if(m->socket < 0)
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		char *s = (char *)malloc(
		    STRLEN(daddress) + 40 +
		    STRLEN(error_msg) + (5 * 80)
		);
		*error_msg = toupper(*error_msg);
		sprintf(
		    s,
"Unable to connect to:\n\
\n\
    %s   Port: %i\n\
\n\
    %s",
		    daddress, port,
		    error_msg
		);
		MonNetSetErrorMesg(m, s);
		free(s);
	    }
	    free(error_msg);
	    free(daddress);
	    return(-1);
	}

	haddr.sin_family = AF_INET;		/* In hbo */ 
	haddr.sin_port = htons(m->port);	/* In nbo */
	haddr.sin_addr = *((struct in_addr *)he->h_addr);
	memset(&(haddr.sin_zero), 0x00, 8);	/* Zero the rest of struct */

	/* Connect the socket */
	if(connect(
	    m->socket, (struct sockaddr *)&haddr,
	    sizeof(struct sockaddr)
	))
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		char *s = (char *)malloc(
		    STRLEN(daddress) + 40 +
		    STRLEN(error_msg) + (5 * 80)
		);
		*error_msg = toupper(*error_msg);    
		sprintf(
		    s,
"Unable to connect to:\n\
\n\
    %s   Port: %i\n\
\n\
    %s.",
		    m->address, m->port,
		    error_msg
		);
		MonNetSetErrorMesg(m, s);
		free(s);
		free(error_msg);
	    }
	    MonNetCloseConnection(m, timeout);
	    free(daddress);
	    return(-1);
	}

	/* Set the socket to non-blocking */
	if(fcntl(m->socket, F_SETFL, O_NONBLOCK) == -1)
	{
            const int error_code = errno;
            char *error_msg = STRDUP(strerror(error_code));
            if(error_msg != NULL)
            {
                char *s = (char *)malloc(
                    STRLEN(error_msg) + (3 * 80)
                );
                *error_msg = toupper(*error_msg);
                sprintf(
                    s,
"Unable to set socket to O_NONBLOCK:\n\
\n\
    %s.",
                    error_msg
                );
                MonNetSetErrorMesg(m, s);
                free(s);
	    }
	    free(error_msg);
	}

	free(daddress);

	return(0);
}


/*
 *	Sends buf to where monitor m is connected.
 */
int MonNetSend(monitor_struct *m, const char *buf)
{
	if((m == NULL) || (buf == NULL))
	{
	    errno = EFAULT;
	    return(-1);
	}

	return(NetSend(m->socket, buf, strlen(buf)));
}

/*
 *	Sends a logout to the universe.
 */
void MonNetSendLogout(monitor_struct *m)
{
	char *sndbuf = (char *)malloc(
	    STRLEN(STAT_PFX_LOGOUT) + 80
	);
	sprintf( 
	    sndbuf,
	    "%s:\n",
	    STAT_PFX_LOGOUT
	);
	MonNetSend(m, sndbuf);
	free(sndbuf);
}

/*
 *	Processes the specified data.
 */
static int MonNetManageIterate(monitor_struct *m, const char *buf)
{
	int segments_handled = 0;
	const char *s, *s_end;
	char *parm = NULL, *val = NULL, *sndbuf;
	mon_stats_struct *stats;
	int i1, i2;
	long l1, l2, l3;

	if((m == NULL) || (buf == NULL))
	    return(segments_handled);

	stats = &m->stats;
	s = buf;

	/* Get the parameter */
	s_end = strchr(s, ':');
	if(s_end != NULL)
	{
	    const int len = s_end - s;
	    if(len > 0)
	    {
		parm = (char *)malloc((len + 1) * sizeof(char));
		memcpy(parm, s, len);
		parm[len] = '\0';
	    }
	    s = s_end + 1;
	}
	else
	{
	    parm = strdup(s);
	}

	/* Get the value */
	val = STRDUP(s);

	strstrip(val);
	strstrip(parm);

	if(parm == NULL)
	    parm = strdup("");
	if(val == NULL)
	    val = strdup("");

	/* Connections */
	if(!strcasecmp(parm, STAT_PFX_CONNECTIONS))
	{
	    /* Format:
	     *
	     *	total_connections guest_connections
	     */
	    sscanf(
		val,
		"%i %i",
		&i1,
		&i2
	    );
	    stats->total_connections = i1;
	    stats->guest_connections = i2;
	    segments_handled++;
	}
	/* Memory */
	else if(!strcasecmp(parm, STAT_PFX_MEMORY))
	{
	    /* Format:
	     *
	     *	mem_total mem_con mem_obj
	     */
	    sscanf(
		val,
		"%ld %ld %ld",
		&l1,
		&l2,
		&l3
	    );
	    stats->mem_total = l1;
	    stats->mem_con = l2;
	    stats->mem_obj = l3;
	    segments_handled++;
	}
	/* Message */
	else if(!strcasecmp(parm, STAT_PFX_MESSAGE))
	{
	    /* Format:
	     *
	     *	message
	     */
	    MesgWinAppend(
		&m->mesgwin, val,
		mon_color.messages_text
	    );
	    segments_handled++;
	}
	/* Next */
	else if(!strcasecmp(parm, STAT_PFX_NEXT))
	{
	    /* Format:
	     *
	     *	next_save next_export
	     */
	    sscanf(
		val,
		"%ld %ld",
		&l1,
		&l2
	    );
	    stats->next_save = l1;
	    stats->next_export = l2;
	    segments_handled++;
 	}
	/* Title */
	else if(!strcasecmp(parm, STAT_PFX_TITLE))
	{
	    /* Format:
	     *
	     *	stitle
	     */
	    free(stats->title);
	    stats->title = strdup(val);
	    OSWSetWindowTitle(m->toplevel, stats->title);
	    OSWSetWindowTitle(m->mesgwin.toplevel, stats->title);
	    segments_handled++;
	}
	/* Objects */
	else if(!strcasecmp(parm, STAT_PFX_OBJECTS))
	{
	    /* Format:
	     *
	     *	itotal_objects
	     */
	    sscanf(
		val,
		"%i",
		&i1
	    );
 	    stats->total_objects = i1;
	    segments_handled++;
	}
	/* PID */
	else if(!strcasecmp(parm, STAT_PFX_PID))
	{
	   /* Format:
	    *
	    *	ipid
	    */
	    sscanf(
		val,
		"%i",
		&i1
	    );
	    stats->pid = i1;
	    segments_handled++;
	}
	/* Request login */
	else if(!strcasecmp(parm, STAT_PFX_REQUESTLOGIN))
	{
	    /* Format:
	     *
	     */

	    /* Respond with login info */
	    sndbuf = (char *)malloc(
		STRLEN(STAT_PFX_LOGIN) + STRLEN(m->name) +
		STRLEN(m->password) + 80
	    );
	    sprintf(
		sndbuf,
		"%s: %s;%s\n",
		STAT_PFX_LOGIN,
		m->name,
		m->password
	    );
	    MonNetSend(m, sndbuf);
	    free(sndbuf);
	    segments_handled++;
	}
	/* Logout */
	else if(!strcasecmp(parm, STAT_PFX_LOGOUT))
	{
	    if(!m->close_wait)
		MonNetShutdownConnection(
		    m, MON_NET_SHUTDOWN_RDWR, False
		);
	    segments_handled++;
	}
	/* Uptime */
	else if(!strcasecmp(parm, STAT_PFX_UPTIME))
	{
	    /* Format is:
	     *
	     *	luptime
	     */
	    sscanf(
		val,
		"%ld",
		&l1
	    );
	    stats->uptime = l1;
	    segments_handled++;
	}

	free(parm);
	free(val);

	return(segments_handled);
}


int MonNetManage(monitor_struct *m)
{
	int	segments_handled = 0,
		bytes_read;

	char buf[AUXSTAT_MAX_LEN];
	char *buf_ptr;
	int buf_cnt;

	char workbuf[AUXSTAT_SEGMENT_MAX];
	char *workbuf_ptr;
	int workbuf_cnt;

	if(m == NULL)
	    return(segments_handled);

	if(m->socket < 0)
	    return(segments_handled);

	/* Get data */
	bytes_read = NetRecv(m->socket, buf, sizeof(buf));
	if(bytes_read == 0)
	{
	    MonNetDisconnect(m, 45l);
	    return(segments_handled);
	}

	/* Recieve error? */
	if(bytes_read < 0)
	{
	    const int error_code = errno;
	    switch(error_code)
	    {
#if (EAGAIN != EWOULDBLOCK)
	      case EWOULDBLOCK:
#endif
	      case EAGAIN:
		return(segments_handled);
		break;
	      case EBADF:
	      case ENOTSOCK:
	      case EFAULT:
		MonNetCloseConnection(m, 45l);
		MonNetDisconnect(m, 45l);
		return(segments_handled);
		break;
	      default:
		MonNetCloseConnection(m, 45l);
		MonNetDisconnect(m, 45l);
		return(segments_handled);
		break;
	    }
	}


	/* Null terminate */
	if(bytes_read > AUXSTAT_MAX_LEN)
	    bytes_read = AUXSTAT_MAX_LEN;

	buf[bytes_read - 1] = '\0';


	/* Begin parsing and handling */
	buf_cnt = 0;
	buf_ptr = buf;

	while(buf_cnt < bytes_read)
	{
	    /* Reset work buffer count and pointer */
	    workbuf_cnt = 0;
	    workbuf_ptr = workbuf;

	    /* Begin fetching data from recvbuf to workbuf */
	    while(1)
	    {
		/* Work buffer overflowed? */
		if(workbuf_cnt >= AUXSTAT_SEGMENT_MAX)
		{
		    /* Increment buf_cnt to next delimiter char */
		    while(buf_cnt < bytes_read)
		    {
			if((*buf_ptr == '\n') ||
			   (*buf_ptr == '\r') ||
			   (*buf_ptr == '\0')
			)
			{
			    buf_cnt++;
			    buf_ptr++;
			    break;
			}
			buf_cnt++;
			buf_ptr++;
		    }
		    /*   Null terminating character for workbuf will
		     *   be added farther below.
		     */
		    break;
		}

		/* End of buffer reached? */
		else if(buf_cnt >= bytes_read)
		{
		    *workbuf_ptr = '\0';

		    /*   Copy workbuf_cnt to net_parms.co_data for
		     *   processing in the next call to this function.
		     */
/*
		    net_parms.co_data_len = MIN(
			workbuf_cnt,
			AUXSTAT_SEGMENT_MAX
		    );

		    if(net_parms.co_data_len > 0)  
			memcpy(
			    net_parms.co_data,
			    workbuf,
			    net_parms.co_data_len
			);
 */                         
		    /*   Return, don't continue. This is so that the
		     *   the carry over buffer is not reset and its
		     *   values just set above get carryed over to
		     *   the next call to this function.
		     */
		    return(0);

		    break;
		}

		/* Deliminator in recieve buffer encountered? */
		else if((*buf_ptr == '\n') ||
			(*buf_ptr == '\r') ||
			(*buf_ptr == '\0')
		)
		{
		    *workbuf_ptr = '\0';

		    buf_cnt++;
		    buf_ptr++;

		    break;
		}

		/* Copy data from buffer to work buffer normally */
		else
		{
		    *workbuf_ptr = *buf_ptr;

		    buf_cnt++;  
		    buf_ptr++;

		    workbuf_cnt++;
		    workbuf_ptr++;
		}
	    }

	    /* Skip if workbuf is empty */
	    if(*workbuf == '\0')
		continue;

	    /* Set null terminating character for workbuf */
	    workbuf[sizeof(workbuf) - 1] = '\0';

	    /* Handle this segment */
	    segments_handled += MonNetManageIterate(m, workbuf);
	}

	if(segments_handled > 0)
	{
	    /* Redraw monitor readout */
	    if(m->map_state)
		MonQueueDraw(m, MON_DRAW_AMOUNT_READOUT);
	}

	return(segments_handled);
}

/*
 *	Shuts down the socket and sets m->close_wait to True.
 */
void MonNetShutdownConnection(
	monitor_struct *m, int how, bool_t is_initiator  
)
{
	int lhow;

	if(m == NULL)
	    return;

	if(m->socket < 0)
	    return;

	/* Set connection state to close wait */
	m->close_wait = True;

        /* Shutdown the socket */
	switch(how)
	{
	  case MON_NET_SHUTDOWN_WR:
	    lhow = SHUT_WR;
	    break;
	  case MON_NET_SHUTDOWN_RDWR:
	    lhow = SHUT_RDWR;
	    break;
	  default:
	    lhow = SHUT_RD;
	    break;
	}
	if(shutdown(m->socket, lhow))
	{
#if 0
            int error_code = errno;
            char *errmsg = STRDUP(strerror(error_code));
            if(errmsg != NULL)
            {
                char *s = (char *)malloc(STRLEN(errmsg) + 40 + 80);
                *errmsg = toupper(*errmsg);
                sprintf(
                    s,
"Socket %i shutdown error: %s.",
                    net->socket,
                    errmsg
                );
                XSWPrintError(s);
                free(s);
            }
            free(errmsg);
#endif
	}
}

/*
 *	Closes the connection's socket.
 *
 *	If timeout is positive then it specifies the timeout in seconds.
 */
void MonNetCloseConnection(monitor_struct *m, time_t timeout)
{
	int error_code;
	time_t start_time;

	if(m == NULL)
	    return;

	/* Socket already closed? */
	if(m->socket < 0)
	    return;

	/* Close the socket */
	start_time = time(NULL);
	while(close(m->socket) != 0)
	{
	    error_code = errno;
	    if(error_code == EAGAIN)
	    {
		/* Timmed out? */
		if(timeout > 0l)
		{
		    if(time(NULL) > (start_time + timeout))
			break;
		}
		continue;
	    }
	    else
	    {
#if 0
		char *errmsg = STRDUP(strerror(error_code));
		if(errmsg != NULL)
		{
		    char *s = (char *)malloc(STRLEN(errmsg) + 80);
		    *errmsg = toupper(*errmsg);
		    sprintf(
			s,
"Socket %i close error: %s.",
			m->socket,
			errmsg
		    );
		    XSWPrintError(s);
		    free(s);
		}
		free(errmsg);
#endif
		break;
	    }
	}
	m->socket = -1;
}

/*
 *	Waits for the socket to close by calling recv() and discarding
 *	the data until it returns 0. When 0 is returned the socket
 *	will be closed and the socket will be set to -1.
 *
 *	This function does not check if the connection state is set to
 *	CLOSE_WAIT or if the connection was shut down prior to this
 *	call.
 *
 *	If timeout is positive then it specifies the timeout in seconds.
 */
void MonNetWaitClose(monitor_struct *m, time_t timeout)
{
	const time_t start_time = time(NULL);
	char buf[1024];

	if(m == NULL)
	    return;

	/* Wait for recv() to return 0 or timeout */
	while(m->socket > -1)
	{
	    if(NetRecv(
		m->socket, buf, sizeof(buf)
	    ) <= 0)
		break;

	    /* Timmed out? */
	    if(timeout > 0l)
	    {
		if(time(NULL) > (start_time + timeout))
		    break;
	    }
	}

	/* Close the connection's socket */
	MonNetCloseConnection(m, timeout);
}

/*
 *	Shuts down the socket and then closes it as needed.
 */
void MonNetDisconnect(monitor_struct *m, time_t timeout)
{
	if(m == NULL)
	    return;

	/* Shutdown and close the connection as needed */
	if(m->socket > -1)
	{
	    if(m->close_wait)
	    {
		MonNetWaitClose(m, timeout);
	    }
	    else
	    {
		MonNetShutdownConnection(
		    m, MON_NET_SHUTDOWN_RDWR, True
		);
		MonNetWaitClose(m, timeout);
	    }
	}

	m->close_wait = False;
	m->socket = -1;
}
