#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>

#include "mon.h"


void MonDialog(
	const char *title, image_t *icon, const char *msg,
	win_t toplevel
);
int MonGetResponse(
	const char *title, image_t *icon, const char *msg,
	unsigned int btn_flags, win_t toplevel
);
void MonPrint(const char *msg);
void MonPrintWarning(const char *msg);
void MonPrintError(const char *msg);
void MonPrintHelpStdout(void);
void MonPrintVersionStdout(void);

dialog_struct *MonDialogNew(void);
void MonDialogDelete(dialog_struct *d);

cdlg_struct *MonConfirmationDialogNew(void);
void MonConfirmationDialogDelete(cdlg_struct *d);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Maps the dialog with the specified message.
 */
void MonDialog(
	const char *title, image_t *icon, const char *msg,
	win_t toplevel
)
{
	dialog_struct *d = mon_dlg;

	/* Not connected to GUI? */
	if(!IDC())
	{
	    printf("%s\n", msg);
	    return;
	}

        /* Initialize the dialog as needed */
        if(d == NULL)
            mon_dlg = d = MonDialogNew();
        if(d == NULL)
            return;

	/* Set the dialog's title and icon */
	DialogSetTitle(d, title);
	DialogSetIcon(d, icon);
	DialogSetMessage(d, msg);
	DialogSetTransientFor(d, toplevel);
	DialogMap(d);
        WidgetCenterWindow(d->toplevel, WIDGET_WINDOW_RELATIVE_POINTER);
}

/*
 *	Maps the confirmation dialog and waits for user response.
 *
 *	Returns one of CDlg*.
 */
int MonGetResponse(
	const char *title, image_t *icon, const char *msg,
	unsigned int btn_flags, win_t toplevel
)
{
	int response;
        cdlg_struct *d;

	/* Not connected to GUI? */
	if(!IDC())
	    return(CDlgError);

        /* Initialize the confirmation dialog as needed */
        d = mon_cdlg;
        if(d == NULL)
            mon_cdlg = d = MonConfirmationDialogNew();
        if(d == NULL)
            return(CDlgError);

	/* Get response */
	response = CDlgGetResponse(
	    d,
	    title, msg, icon,
	    btn_flags,
	    WIDGET_WINDOW_RELATIVE_POINTER,
	    toplevel,
	    MonMain, NULL,
	    MonMainPop, NULL
	);

        /* Destroy the confirmation dialog */
        MonConfirmationDialogDelete(d);
        mon_cdlg = d = NULL;

	return(response);
}

/*
 *	Prints the message.
 *
 *	If connected to the GUI then prints the message to the message
 *	box, otherwise it prints it to stdout.
 *
 *	The message should not have a tailing newline character.
 */
void MonPrint(const char *msg)
{
	if(msg == NULL)
	    return;
#if 0
	if(IDC() && (bridge_win.mesg_box != None))
	    MesgAdd(msg, xsw_color.bp_standard_text);
	else
#endif
	    printf("%s\n", msg);
}

void MonPrintWarning(const char *msg)
{
	if(msg == NULL)
	    return;

#if 0
	if(IDC() && (bridge_win.mesg_box != None))
	    MesgAdd(msg, xsw_color.bp_warning);
	else
#endif
	    fprintf(stderr, "%s\n", msg);
}

void MonPrintError(const char *msg)
{
	if(msg == NULL)
	    return;

#if 0
	if(IDC() && (bridge_win.mesg_box != None))
	    MesgAdd(msg, xsw_color.bp_danger);
	else
#endif
	    fprintf(stderr, "%s\n", msg);
}

/*
 *	Print usage of this program to stdout.
 */
void MonPrintHelpStdout(void)
{
	fputs(PROG_USAGE_MESSAGE, stdout);
}

/*
 *	Print version of this program to stdout.
 */
void MonPrintVersionStdout(void)
{
	fputs(
	    PROG_NAME " Version " PROG_VERSION "\n"
	    PROG_COPYRIGHT "\n",
	    stdout
	);
}


/*
 *	Creates a new dialog.
 */
dialog_struct *MonDialogNew(void)
{
        dialog_struct *d = (dialog_struct *)calloc(
            1, sizeof(dialog_struct)
        );
        if(d == NULL)
            return(NULL);

        if(DialogInit(d))
        {
            free(d);
            d = NULL;
        }

        return(d);
}

/*
 *	Deletes the dialog.
 */
void MonDialogDelete(dialog_struct *d)
{
        if(d == NULL)
            return;

        WidgetDestroy(d);
        free(d);
}


/*
 *	Creates a new confirmation dialog.
 */
cdlg_struct *MonConfirmationDialogNew(void)
{
        cdlg_struct *d = (cdlg_struct *)calloc(
            1, sizeof(cdlg_struct)
        );
        if(d == NULL)  
            return(NULL);

        if(CDlgInit(d))
        {
            free(d);
            d = NULL;
        }

        return(d);
}

/*
 *      Deletes the confirmation dialog.
 */
void MonConfirmationDialogDelete(cdlg_struct *d)
{
        if(d == NULL)
            return;

        CDlgDestroy(d);
        free(d);
}
