#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/widget.h"


static void LabelResizeCB(void *widget, void *data);
static void LabelDrawCB(void *widget, void *data);

static bool_t LabelGetTextExtents(
	font_t *font, const char *text,
	unsigned int *width_rtn, unsigned int *height_rtn
);

int LabelInit(
	label_struct *label, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *text
);
void LabelSetFont(label_struct *label, font_t *font);
void LabelSetJustify(label_struct *label, int justify);
void LabelSetBorder(
	label_struct *label, unsigned int x_border, unsigned int y_border
);
void LabelSetText(label_struct *label, const char *text);
void LabelSetTextColor(label_struct *label, pixel_t text_c);
void LabelMinimizeSize(label_struct *label);
void LabelResize(label_struct *label);
void LabelQueueResize(label_struct *label);
void LabelDraw(label_struct *label);
void LabelQueueDraw(label_struct *label);
int LabelManage(label_struct *label, event_t *event);
void LabelMap(label_struct *label);
void LabelUnmap(label_struct *label);
void LabelDestroy(label_struct *label);  


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Resize callback.
 */
static void LabelResizeCB(void *widget, void *data)
{
	label_struct *label = LABEL(widget);
	if(label == NULL)
	    return;

	LabelResize(label);
}

/*
 *	Draw callback.
 */
static void LabelDrawCB(void *widget, void *data)
{
	label_struct *label = LABEL(widget);
	if(label == NULL)
	    return;     

	if(label->map_state)
	    LabelDraw(label);
}


/*
 *	Calculates the extents of the text.
 */
static bool_t LabelGetTextExtents(
	font_t *font, const char *text,
	unsigned int *width_rtn, unsigned int *height_rtn
)
{
	int len, swidth, ascent, descent;
	const char *s, *s_end;
	unsigned int width = 0, height = 0;

	if(width_rtn != NULL)
	    *width_rtn = width;
	if(height_rtn != NULL)
	    *height_rtn = height;

	if((font == NULL) || (text == NULL))
	    return(False);

	s = text;
	while(s != NULL)
	{
	    s_end = strchr(s, '\n');
	    len = (s_end != NULL) ? (s_end - s) : strlen(s);
	    OSWGetTextBounds(
		font, s, len,
		NULL, NULL, &swidth, &ascent, &descent
	    );
	    height += (unsigned int)MAX((ascent + descent), 0);
	    if((unsigned int)swidth > width)
		width = (unsigned int)swidth;

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}

	if(width_rtn != NULL)
	    *width_rtn = width;
	if(height_rtn != NULL)
	    *height_rtn = height;

	return(True);
}


/*
 *	Initializes the label.
 */
int LabelInit(
	label_struct *label, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *text
)
{
	const unsigned int x_border = 0, y_border = 0;
	unsigned int	text_width, text_height,
			min_width, min_height;
	win_t w;
	font_t *font;
/*	osw_gui_struct *gui = &osw_gui[0]; */
	widget_global_struct *wglobal = &widget_global;

	if((label == NULL) || (parent == None))
	    return(-2);

	font = wglobal->std_font;

	LabelGetTextExtents(
	    font, text, &text_width, &text_height
	);
	min_width = text_width;
	min_height = text_height;

	/* Need to autocalculate the size? */
	if(width == 0)
	    width = text_width + (2 * x_border);
	if(height == 0)
	    height = text_height + (2 * x_border);

	if(width < min_width)
	    width = min_width;
	if(height < min_height)
	    height = min_height;

	label->type = WIDGET_TYPE_LABEL;
	label->map_state = False;
	label->has_focus = False;
	label->disabled = False;
	label->visibility_state = VisibilityFullyObscured;
	label->x = x;
	label->y = y;
	label->width = width;
	label->height = height;
	label->font = font;
	label->next = NULL;
	label->prev = NULL;
	label->data = NULL;

	label->text = STRDUP(text);
	label->justify = LABEL_JUSTIFY_LEFT;
	label->x_border = x_border;
	label->y_border = y_border;
	label->text_c = wglobal->normal_text_pix;

	/* Create toplevel */
	x = label->x;
	y = label->y;   
	width = label->width;
	height = label->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	label->toplevel = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    VisibilityChangeMask
	);

	/* Add button to widget regeristry */
	WidgetRegister(label, label->type);  

	return(0);
}

/*
 *	Sets the label's font.
 */
void LabelSetFont(label_struct *label, font_t *font)
{
	if(label == NULL)
	    return;

	if(font != label->font)
	{
	    label->font = font;
	    if(label->map_state)
		LabelQueueDraw(label);
	}
}  

/*
 *	Sets the label's justification.
 */
void LabelSetJustify(label_struct *label, int justify)
{
	if(label == NULL)
	    return;

	if(justify != label->justify)
	{
	    label->justify = justify;
	    if(label->map_state)
		LabelQueueDraw(label);
	}
}

/*
 *	Sets the label's border.
 */
void LabelSetBorder(
	label_struct *label, unsigned int x_border, unsigned int y_border
)
{
	if(label == NULL)
	    return;

	if((x_border != label->x_border) ||
	   (y_border != label->y_border)
	)
	{
	    label->x_border = x_border;
	    label->y_border = y_border;
	    if(label->map_state)
		LabelQueueDraw(label);
	}
}

/*
 *	Sets the label's text.
 */
void LabelSetText(label_struct *label, const char *text)
{
	if(label == NULL)                       
	    return;

	if(text != label->text)
	{
	    free(label->text);
	    label->text = STRDUP(text);
	    if(label->map_state)
		LabelQueueDraw(label);
	}
}

/*
 *	Sets the label's colors.
 */
void LabelSetTextColor(label_struct *label, pixel_t text_c)
{
	if(label == NULL)
	    return;

	if(text_c != label->text_c)
	{
	    label->text_c = text_c;
	    if(label->map_state)
		LabelQueueDraw(label);
	}
}

/*
 *	Resizes the label to fit its text and border.
 */
void LabelMinimizeSize(label_struct *label)
{
	unsigned int width, height, text_width, text_height;

        if(label == NULL)
            return;

	LabelGetTextExtents(
	    label->font, label->text,
	    &text_width, &text_height
	);

	width = text_width + (2 * label->x_border);
	height = text_height + (2 * label->y_border);

	OSWResizeWindow(label->toplevel, width, height);
	LabelQueueResize(label);
	if(label->map_state)
	    LabelQueueDraw(label);
}

/*
 *	Resizes the label.
 */
void LabelResize(label_struct *label)
{
	int x, y;
	unsigned int width, height;
	win_t w;
/*	widget_global_struct *wglobal = &widget_global; */

	if(label == NULL)
	    return;

	w = label->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);

	label->x = x;
	label->y = y;
	label->width = width;
	label->height = height;
}

/*
 *	Queues the label to resize.
 */
void LabelQueueResize(label_struct *label)
{
	if(label == NULL)                       
	    return;

	WidgetQueueResizeAction(LabelResizeCB, label, NULL);
}

/*
 *	Draws the label.
 */
void LabelDraw(label_struct *label)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(label == NULL)
	    return;

	if(label->toplevel != None)
	{
	    int x, y;
	    unsigned int	width = label->width,
				height = label->height,
				x_border = label->x_border;
	    font_t *font = label->font;
	    win_t w = label->toplevel;

	    /* Draw the background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);
	    }

	    /* Draw the text */
	    if(!STRISEMPTY(label->text))
	    {
		int	len,
			lbearing, swidth, ascent, descent,
			font_height;
		unsigned int text_width, text_height;
		const char *s, *s_end;

		OSWSetForeground(
		    label->disabled ?
		    wglobal->disabled_text_pix :
		    label->text_c
		);
		OSWSetFont(font);

		LabelGetTextExtents(
		    font, label->text, &text_width, &text_height
		);

		y = ((int)height - (int)text_height) / 2;
		s = label->text;
		while((s != NULL) && (y < (int)height))
		{
		    s_end = strchr(s, '\n');
		    len = (s_end != NULL) ? (int)(s_end - s) : strlen(s);
		    OSWGetTextBounds(
			font, s, len,
			&lbearing, NULL, &swidth, &ascent, &descent
		    );
		    font_height = ascent + descent;

		    if(len > 0)
		    {
			switch(label->justify)
			{
			  case LABEL_JUSTIFY_RIGHT:
			    x = (int)width - (int)x_border - swidth;
			    break;
			  case LABEL_JUSTIFY_CENTER:
			    x = ((int)width - swidth) / 2;
			    break;
			  default:	/* LABEL_JUSTIFY_LEFT */
			    x = (int)x_border;
			    break;
			}
			OSWDrawStringLimited(
			    w,
			    x - lbearing,
			    y + ascent,
			    s, len
			);
		    }

		    y += font_height;
		    s = (s_end != NULL) ? (s_end + 1) : NULL;
		}

	    }
	}
}

/*
 *	Queues the label to draw.
 */
void LabelQueueDraw(label_struct *label)
{
	if(label == NULL)
	    return;

	WidgetQueueDrawAction(LabelDrawCB, label, NULL);
}

/*
 *	Manages the label event.
 */
int LabelManage(label_struct *label, event_t *event)
{
	int events_handled = 0;
	win_t w;
	event_visibility_t *visibility;

	if((label == NULL) || (event == NULL))
	    return(events_handled);

	if(!label->map_state)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
          case ConfigureNotify:
            if(w == label->toplevel)
            {
		if(label->map_state)
		    LabelQueueDraw(label);
                events_handled++;
            }
            break;

	  case Expose:
	    if(w == label->toplevel)
	    {
		LabelQueueDraw(label);
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == label->toplevel)
	    {
		label->visibility_state = visibility->state;
		events_handled++;
	    }
	    break;
	}

	return(events_handled);
}

/*
 *	Maps the label.
 */
void LabelMap(label_struct *label)
{
	if(label == NULL)
	    return;

	if(!label->map_state)
	{
	    OSWMapRaised(label->toplevel);
	    label->map_state = True;
	}
}

/*
 *	Unmaps the label.
 */
void LabelUnmap(label_struct *label)
{
	if(label == NULL)
	    return;

	OSWUnmapWindow(label->toplevel);
	label->map_state = False;
}

/*
 *	Destroys the label.
 */
void LabelDestroy(label_struct *label)
{
	if(label == NULL)
	    return;

	OSWDestroyWindow(&label->toplevel);

	label->map_state = False;
	label->has_focus = False;
	label->disabled = False;
	label->x = 0;
	label->y = 0;
	label->width = 0;
	label->height = 0;
	label->visibility_state = VisibilityFullyObscured;
	label->font = NULL;
	label->next = NULL;
	label->prev = NULL;
	free(label->text);
	label->text = NULL;
	label->justify = LABEL_JUSTIFY_LEFT;
	label->x_border = 0;
	label->y_border = 0;
	label->text_c = None;
}
