#include <stdlib.h>
#include <string.h>
#include "../include/string.h"
#include "../include/graphics.h"
#include "../include/widget.h"


static void PBarResizeCB(void *widget, void *data);
static void PBarDrawCB(void *widget, void *data);

int PBarInit(
	progress_bar_struct *pb,
	win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	float start_val,
	float min, float max,
	const char *label,
	int completion_action
);
void PBarResize(progress_bar_struct *pb);
void PBarDraw(progress_bar_struct *pb);
void PBarQueueDraw(progress_bar_struct *pb);
int PBarManage(progress_bar_struct *pb, event_t *xevent);
void PBarMap(progress_bar_struct *pb);
void PBarUnmap(progress_bar_struct *pb);
void PBarDestroy(progress_bar_struct *pb);

void PBarUpdate(
	progress_bar_struct *pb, float value, const char *label
);


#ifndef MAX
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#endif

#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sizes (in pixels):
 */
#define PBAR_MARGIN			1


/*
 *	Resize callback.
 */
static void PBarResizeCB(void *widget, void *data)
{
	progress_bar_struct *pb = PROGRESS_BAR(widget);
	if(pb == NULL)
	    return;

	PBarResize(pb);
}

/*
 *	Draw callback.
 */
static void PBarDrawCB(void *widget, void *data)
{
	progress_bar_struct *pb = PROGRESS_BAR(widget);
	if(pb == NULL)  
	    return;

	if(pb->map_state)
	    PBarDraw(pb);
}


/*
 *	Initializes the progress bar.
 */
int PBarInit(
	progress_bar_struct *pb, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	float start_val,
	float min, float max,
	const char *label,
	int completion_action
)
{
	win_t w;
	widget_global_struct *wglobal = &widget_global;
	
	if((pb == NULL) || (parent == None))
	    return(-2);

	if(width == 0)
	    width = 1;
	if(height == 0)
	    height = 1;

	if(start_val < min)
	    start_val = min;
	if(start_val > max)
	    start_val = max;

	pb->type = WIDGET_TYPE_PROGRESS_BAR;
	pb->map_state = False;
	pb->has_focus = False;
	pb->disabled = False;
	pb->visibility_state = VisibilityFullyObscured;
	pb->x = x;
	pb->y = y;
	pb->width = width;
	pb->height = height;
	pb->font = wglobal->std_font;
	pb->next = NULL;
	pb->prev = NULL;
	pb->data = NULL;

	/* Create toplevel */
	x = pb->x;
	y = pb->y;
	width = pb->width;
	height = pb->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	pb->toplevel = w;
	pb->toplevel_buf = 0;
	pb->image_buf = NULL;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    VisibilityChangeMask
	);

	pb->label = STRDUP(label);
	pb->current = start_val;
	pb->min = min;
	pb->max = max;
	pb->completion_action = completion_action;

	/* Add widget to regeristry */
	WidgetRegister(pb, WIDGET_TYPE_PROGRESS_BAR);

	return(0);
}

/*
 *	Resizes the progress bar.
 */
void PBarResize(progress_bar_struct *pb)
{
	int x, y;
	unsigned int width, height;
	win_t w;

	if(pb == NULL)
	    return;

	w = pb->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);

	pb->x = x;
	pb->y = y;

	if((pb->width == width) && (pb->height == height))
	    return;

	pb->width = width;
	pb->height = height;

	OSWDestroyImage(&pb->image_buf);
	OSWDestroyPixmap(&pb->toplevel_buf);
}

/*
 *	Queues the progress bar to resize.
 */
void PBarQueueResize(progress_bar_struct *pb)
{
	if(pb == NULL)
	    return;

	WidgetQueueResizeAction(PBarResizeCB, pb, NULL);
}

/*
 *	Draws the progress bar.
 */
void PBarDraw(progress_bar_struct *pb)
{
	float gamma, progress_delta_dist, total_delta_dist;
	u_int8_t *data;
	u_int8_t *ptr8;
	u_int16_t *ptr16;
	u_int32_t *ptr32;
	int x_pos, x_pos_check;		/* In pixels */
	int x_col_pos, y_row_pos;	/* In pixels */
	unsigned int width, height;
	int bpp, bpl;
	image_t *img;
	win_t w;
	pixmap_t pixmap;
	WColorStruct color;
	win_attr_t wattr;
	osw_gui_struct *gui = &osw_gui[0];

	if(pb == NULL)
	    return;

	w = pb->toplevel;

	OSWGetWindowAttributes(w, &wattr);

	total_delta_dist = (float)(pb->max - pb->min);
	if(total_delta_dist < 0.0f)
	    total_delta_dist = 0.0f;

	progress_delta_dist = (float)(pb->current - pb->min);
	if(progress_delta_dist < 0.0f)
	    progress_delta_dist = 0.0f;

	/* Map as needed */
	if(!pb->map_state)
	{
	    OSWMapRaised(w);
	    pb->map_state = True;
	}

	OSWSetFont(pb->font);

	/* Recreate image buffer as needed */
	if(pb->image_buf == NULL)
	{
	    if(OSWCreateImage(
		&pb->image_buf,
		wattr.width, wattr.height
	    ))
		return;
	}
	img = pb->image_buf;
	data = (u_int8_t *)img->data;  

	/* Recreate graphics buffer as needed */
	if(pb->toplevel_buf == None)
	{
	    if(OSWCreatePixmap(
		&pb->toplevel_buf,
		wattr.width, wattr.height
	    ))
		return;
	}
	pixmap = pb->toplevel_buf;


	/* Get size of the image */
	width = img->width;
	height = img->height;


	/* Calculate x_pos, the x position that the progress bar ends */
	if(total_delta_dist != 0.0)
	{
	    x_pos = (int)((progress_delta_dist / total_delta_dist) *
		(float)MAX((int)wattr.width - (2 * PBAR_MARGIN), 0));
	    if(x_pos < 0)
		x_pos = 0;
	}
	else
	{
	    x_pos = 0;
	}


	/* Draw progress bar */
	if((data != NULL) && (width > 0) && (height > 0))
	{
	    /* Set colors */
	    color.a = 0xff;
	    color.r = 0xff;
	    color.g = 0x02;
	    color.b = 0x80;

	    bpp = img->bits_per_pixel / 8;
	    bpl = img->bytes_per_line;

	    /* Blit lines */
	    switch(img->depth)
	    {
	      case 8:
		x_col_pos = 0;
		y_row_pos = 0;
		x_pos_check = MIN((x_pos + 1), (int)width);

		gamma = (float)y_row_pos / (float)height;
		if(gamma > 0.5)
		    gamma = (2 * (1 - gamma)) + 0.2;
		else
		    gamma = (2 * gamma) + 0.2;
		while(y_row_pos < (int)height)
		{
		    ptr8 = (u_int8_t *)(data +
			(bpl * y_row_pos) +
			(bpp * x_col_pos)
		    );

		    if(x_col_pos < x_pos_check)
			*ptr8 = PACK8TO8(
			    (u_int8_t)(MIN(color.r * gamma, 0xff)),
			    (u_int8_t)(MIN(color.g * gamma, 0xff)),
			    (u_int8_t)(MIN(color.b * gamma, 0xff))
			);
		    else
			*ptr8 = 0x00;

		    /* Increment values */
		    x_col_pos++;
		    if(x_col_pos >= (int)width)
		    {
			x_col_pos = 0;
			y_row_pos++;

			gamma = (float)y_row_pos / (float)img->height;
			if(gamma > 0.5)
			    gamma = (2 * (1 - gamma)) + 0.2;
			else
			    gamma = (2 * gamma) + 0.2;
		    }
		}
	        break;

	      case 15:
		x_col_pos = 0;
		y_row_pos = 0;
		x_pos_check = MIN((x_pos + 1), (int)width);

		gamma = (float)y_row_pos / (float)height;
		if(gamma > 0.5)
		    gamma = (2 * (1 - gamma)) + 0.2;
		else
		    gamma = (2 * gamma) + 0.2;
		while(y_row_pos < (int)height)
		{
		    ptr16 = (u_int16_t *)(data +
			(bpl * y_row_pos) +
			(bpp * x_col_pos)
		    );

		    if(x_col_pos < x_pos_check)
			*ptr16 = PACK8TO15(
			    (u_int8_t)(MIN(color.r * gamma, 0xff)),
			    (u_int8_t)(MIN(color.g * gamma, 0xff)),
			    (u_int8_t)(MIN(color.b * gamma, 0xff))
			);
		    else
			*ptr16 = 0x00;

		    /* Increment values */
		    x_col_pos++;
		    if(x_col_pos >= (int)width)
		    {
			x_col_pos = 0;
			y_row_pos++;

			gamma = (float)y_row_pos / (float)img->height;
			if(gamma > 0.5)
			    gamma = (2 * (1 - gamma)) + 0.2;
			else  
			    gamma = (2 * gamma) + 0.2;
		    }
		}
		break;

	      case 16:
		x_col_pos = 0;
		y_row_pos = 0;
		x_pos_check = MIN((x_pos + 1), (int)width);

		gamma = (float)y_row_pos / (float)height;
		if(gamma > 0.5)
		    gamma = (2 * (1 - gamma)) + 0.2;
		else
		    gamma = (2 * gamma) + 0.2;
		while(y_row_pos < (int)height)
		{
		    ptr16 = (u_int16_t *)(data +
			(bpl * y_row_pos) +
			(bpp * x_col_pos)
		    );

		    if(x_col_pos < x_pos_check)
			*ptr16 = PACK8TO16(
			    (u_int8_t)(MIN(color.r * gamma, 0xff)),
			    (u_int8_t)(MIN(color.g * gamma, 0xff)),
			    (u_int8_t)(MIN(color.b * gamma, 0xff))
			);
		    else
			*ptr16 = 0x00;

		    /* Increment values */
		    x_col_pos++;
		    if(x_col_pos >= (int)width)
		    {
			x_col_pos = 0;
			y_row_pos++;

			gamma = (float)y_row_pos / (float)img->height;
			if(gamma > 0.5)
			    gamma = (2 * (1 - gamma)) + 0.2;
			else
			    gamma = (2 * gamma) + 0.2;
		    }   
		}
		break;

	      case 24:
	      case 32:
		x_col_pos = 0;
		y_row_pos = 0;
		x_pos_check = MIN((x_pos + 1), (int)width);

		gamma = (float)y_row_pos / (float)height;
		if(gamma > 0.5)
		    gamma = (2 * (1 - gamma)) + 0.2;
		else
		    gamma = (2 * gamma) + 0.2;
		while(y_row_pos < (int)height)
		{
		    ptr32 = (u_int32_t *)(data +
			(bpl * y_row_pos) +
			(bpp * x_col_pos)
		    );

		    if(x_col_pos < x_pos_check)
			*ptr32 = PACK8TO32(
			    (u_int8_t)color.a,
			    (u_int8_t)(MIN(color.r * gamma, 0xff)),
			    (u_int8_t)(MIN(color.g * gamma, 0xff)),
			    (u_int8_t)(MIN(color.b * gamma, 0xff))
			);
		    else
			*ptr32 = 0x00000000;

		    /* Increment values */
		    x_col_pos++;
		    if(x_col_pos >= (int)width)
		    {
			x_col_pos = 0;
			y_row_pos++;

			gamma = (float)y_row_pos / (float)img->height;
			if(gamma > 0.5)
			    gamma = (2 * (1 - gamma)) + 0.2;
			else
			    gamma = (2 * gamma) + 0.2;
		    }
		}
		break;
	    }

	    OSWPutImageToDrawable(img, pixmap);
	}


	/* Draw text */
	if(pb->label != NULL)
	{
	    if(widget_global.force_mono)
		OSWSetForeground(gui->white_pix);
	    else
	        OSWSetForeground(widget_global.pbar_text_pix);

	    OSWDrawString(
		w,
		12, ((int)wattr.height / 2) + 5,
	        pb->label
	    );
	}

	/* Draw depressed frame */
	WidgetFrameButtonPixmap(
	    pixmap,
	    True,
	    wattr.width, wattr.height,
	    widget_global.surface_highlight_pix,
	    widget_global.surface_shadow_pix
	);

	OSWPutDrawableToWindow(w, pixmap);
}

/*
 *	Queues the progress bar to draw.
 */
void PBarQueueDraw(progress_bar_struct *pb)
{
	if(pb == NULL)
	    return;

	WidgetQueueDrawAction(PBarDrawCB, pb, NULL);
}

/*
 *	Manages the progress bar event.
 */
int PBarManage(progress_bar_struct *pb, event_t *event)
{
	int events_handled = 0;
	win_t w;

	if((pb == NULL) || (event == NULL))
	    return(events_handled);

	if(!pb->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == pb->toplevel)
	    {
		if(pb->map_state)
		    PBarQueueDraw(pb);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == pb->toplevel)
	    {
		PBarQueueDraw(pb);
	        events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    if(w == pb->toplevel)
	    {
		pb->visibility_state = event->xvisibility.state;
		events_handled++;
	    }
	    break;
	}

	return(events_handled);
}

/*
 *	Maps the progress bar.
 */
void PBarMap(progress_bar_struct *pb)
{
	if(pb == NULL)
	    return;

	pb->map_state = False;
	PBarDraw(pb);
}

/*
 *	Unmaps the progress bar.
 */
void PBarUnmap(progress_bar_struct *pb)
{
	if(pb == NULL)
	    return;   

	OSWUnmapWindow(pb->toplevel);
	pb->map_state = False;

	OSWDestroyPixmap(&pb->toplevel_buf);
}

/*
 *	Destroys the progress bar.
 */
void PBarDestroy(progress_bar_struct *pb)
{
	if(pb == NULL)
	    return;

	OSWDestroyWindow(&pb->toplevel);
	OSWDestroyPixmap(&pb->toplevel_buf);
	OSWDestroyImage(&pb->image_buf);

	pb->type = WIDGET_TYPE_UNKNOWN;
	pb->map_state = False;
	pb->has_focus = False;
	pb->visibility_state = VisibilityFullyObscured;
	pb->x = 0;
	pb->y = 0;
	pb->width = 0;
	pb->height = 0;
	pb->font = NULL;
	pb->next = NULL;
	pb->prev = NULL;
	pb->data = NULL;

	pb->current = 0.0f;
	pb->min = 0.0f;
	pb->max = 0.0f;

	free(pb->label);
	pb->label = NULL;
}


/*
 *	Updates the progress bar.
 */
void PBarUpdate(
	progress_bar_struct *pb, float value, const char *label
)
{
	if(pb == NULL)
	    return;

	if(value > pb->max)
	    value = pb->max;
	if(value < pb->min)
	    value = pb->min;

	pb->current = value;

	/* Change name as needed */
	if(label != NULL)
	{
	    free(pb->label);
	    pb->label = strdup(label);
	}

	/* Redraw */
	if(pb->map_state)
	    PBarQueueDraw(pb);
}
