#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"

#include "../include/widget.h"


static bool_t PBtnMatchAccelKey(
	push_button_struct *btn,
	keycode_t keycode, key_state_t state
);
static void PBtnResizeCB(void *widget, void *data);
static void PBtnDrawCB(void *widget, void *data);

int PBtnInit(
	push_button_struct *btn, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *label, int label_align,
	pixmap_t pixmap, bitmap_t mask,
	void (*activate_cb)(push_button_struct *, void *),
	void *activate_data
);
void PBtnSetLabel(
	push_button_struct *btn,
	unsigned int width, unsigned int height,
	const char *label, int label_align,
	image_t *image
);
void PBtnSetTooltip(
	push_button_struct *btn, const char *mesg
);
void PBtnSetAccelKeys(
	push_button_struct *btn,
	const keyval_t *accel_keys, const key_state_t *accel_mods,
	int naccel_keys
);
void PBtnSetState(push_button_struct *btn, const int state);
void PBtnResize(push_button_struct *btn);
void PBtnQueueResize(push_button_struct *btn);
void PBtnDraw(push_button_struct *btn);
void PBtnQueueDraw(push_button_struct *btn);
int PBtnManage(push_button_struct *btn, event_t *event);
void PBtnMap(push_button_struct *btn);
void PBtnUnmap(push_button_struct *btn);
void PBtnDestroy(push_button_struct *btn);


#define PBTN_DEF_WIDTH			70
#define	PBTN_DEF_HEIGHT			28

#define PBTN_MARGIN			2

#define PBTN_MIN_WIDTH			(2 * PBTN_MARGIN)
#define PBTN_MIN_HEIGHT			(2 * PBTN_MARGIN)


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)
#define IS_NUM_ODD(n)	((long)(n) & 1)


/*
 *	Checks if keycode matches one of the keys specified in btn's
 *	hotkeys listing.
 */
static bool_t PBtnMatchAccelKey(
	push_button_struct *btn,
	keycode_t keycode, key_state_t state
)
{
	int i, m;
	keyval_t keyval;

	if((btn == NULL) || (keycode == 0))
	    return(False);

	if((btn->naccel_keys <= 0) || (btn->accel_keys == NULL))
	    return(False);

	keyval = OSWGetKeyValFromKeyCode(keycode, state);

	m = btn->naccel_keys;
	for(i = 0; i < m; i++)
	{
	    if(btn->accel_keys[i] == keyval)
	    {
		/* If accelerator modifier keys are set on the button
		 * then check them, otherwise return True
		 */
		if(btn->accel_mods != NULL)
		{
		    const key_state_t accel_mods = btn->accel_mods[i];
		    if(accel_mods & OSW_KEY_STATE_ALT)
		    {
			if(!(state & OSW_KEY_STATE_ALT))
			    continue;
		    }
		    else
		    {
			if(state & OSW_KEY_STATE_ALT)
			    continue;
		    }

		    if(accel_mods & OSW_KEY_STATE_CTRL)
		    {
			if(!(state & OSW_KEY_STATE_CTRL))
			    continue;
		    }
		    else
		    {
			if(state & OSW_KEY_STATE_CTRL)
			    continue;
		    }

		    if(accel_mods & OSW_KEY_STATE_SHIFT)
		    {
			if(!(state & OSW_KEY_STATE_SHIFT))
			    continue;
		    }
		    else
		    {
			if(state & OSW_KEY_STATE_SHIFT)
			    continue;
		    }

		    return(True);
		}
		else
		{
		    return(True);
		}
	    }
	}

	return(False);
}

/*
 *	Resize callback.
 */
static void PBtnResizeCB(void *widget, void *data)
{
	push_button_struct *btn = (push_button_struct *)widget;
	if(btn == NULL)
	    return;

	PBtnResize(btn);
}

/*
 *	Draw callback.
 */
static void PBtnDrawCB(void *widget, void *data)
{
	push_button_struct *btn = (push_button_struct *)widget;
	if(btn == NULL)
	    return;

	if(btn->map_state)
	    PBtnDraw(btn);
}


/*
 *	Initializes the button.
 */
int PBtnInit(
	push_button_struct *btn, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *label, int label_align,
	pixmap_t pixmap, bitmap_t mask,
	void (*activate_cb)(push_button_struct *, void *),
	void *activate_data
)
{
	font_t *font;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if((btn == NULL) || (parent == None))
	    return(-2);

	font = wglobal->pbtn_font;

	/* Autocalculate width? */
	if(width == 0)
	{
	    if(pixmap != None)
	    {
		unsigned int icon_width, icon_height;
		OSWGetPixmapGeometry(
		    pixmap, &icon_width, &icon_height, NULL
		);
		width = (int)icon_width + (2 * PBTN_MARGIN);
	    }

	    if(label != NULL)
	    {
		int swidth;
		if(OSWGetTextBounds(
		    font, label, -1,
		    NULL, NULL,
		    &swidth,
		    NULL, NULL
		))
		    width = swidth + (2 * PBTN_MARGIN);
	    }

	    if(width == 0)
		width = PBTN_DEF_WIDTH;
	}
	if((int)width > (int)gui->display_width)
	    width = gui->display_width;
	if((int)width < PBTN_MIN_WIDTH)
	    width = PBTN_MIN_WIDTH;

	/* Autocalculate height? */
	if(height == 0)
	{
	    if(pixmap != None)
	    {
		unsigned int icon_width, icon_height;
		OSWGetPixmapGeometry(
		    pixmap, &icon_width, &icon_height, NULL
		);
		height = (int)icon_height + (2 * PBTN_MARGIN);
	    }

	    if(label != NULL)
	    {
		int ascent, descent;
		if(OSWGetTextBounds(
		    font, label, -1,
		    NULL, NULL,
		    NULL,
		    &ascent, &descent
		))
		    height = (ascent + descent) + (2 * PBTN_MARGIN);
	    }

	    if(height == 0)
		height = PBTN_DEF_HEIGHT;
	}
	if((int)height > (int)gui->display_height)
	    height = gui->display_height;
	if((int)height < PBTN_MIN_HEIGHT)
	    height = PBTN_MIN_WIDTH;

	/* Width or height cannot be odd */
	if(IS_NUM_ODD(width))
	    width += 1;
	if(IS_NUM_ODD(height))
	    height += 1;

	btn->type = WIDGET_TYPE_PUSH_BUTTON;
	btn->map_state = False;
	btn->has_focus = False;
	btn->disabled = False;
	btn->visibility_state = VisibilityFullyObscured;
	btn->x = x;
	btn->y = y;
	btn->width = width;
	btn->height = height;
	btn->font = font;
	btn->next = NULL;
	btn->prev = NULL;
	btn->data = NULL;

	btn->is_default = False;
	btn->state = PBTN_UNARMED;

	btn->unarmed_img = NULL;
	btn->armed_img = NULL;
	btn->highlighted_img = NULL;

	btn->accel_keys = NULL;
	btn->accel_mods = NULL;
	btn->naccel_keys = 0;

	btn->label = STRDUP(label);
	btn->label_align = label_align;

	btn->pixmap = pixmap;
	btn->mask = mask;

	btn->pressed_cb = NULL;
	btn->pressed_data = NULL;
	btn->released_cb = NULL;
	btn->released_data = NULL;
	btn->activate_cb = activate_cb;
	btn->activate_data = activate_data;

	/* Create toplevel */
	x = btn->x;
	y = btn->y;
	width = btn->width;
	height = btn->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	btn->toplevel = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    KeyPressMask | KeyReleaseMask |
	    ButtonPressMask | ButtonReleaseMask |
	    EnterWindowMask | LeaveWindowMask |
	    VisibilityChangeMask
	);

	/* Add button to widget regeristry */
	WidgetRegister(btn, btn->type);

	return(0);
}


/*
 *	Sets new label or image on btn.
 *
 *	width and height specify the new size of the button,
 *	if either dimension is 0, that dimension will not be
 *	that of the image or label's `reasonable' size.
 */
void PBtnSetLabel(
	push_button_struct *btn,
	unsigned int width, unsigned int height,
	const char *label, int label_align,
	pixmap_t pixmap, bitmap_t mask
)
{
	osw_gui_struct *gui = &osw_gui[0];

	if(btn == NULL)
	    return;

	/* Set new alignment code */
	btn->label_align = label_align;

	/* Set new label text */
	free(btn->label);
	btn->label = STRDUP(label);

	btn->pixmap = pixmap;
	btn->mask = mask;

	/* Calculate needed width and height */
	if(width == 0)
	{
	    if(pixmap != None)
	    {
		unsigned int icon_width, icon_height;
		OSWGetPixmapGeometry(
		    pixmap, &icon_width, &icon_height, NULL
		);
		width = (int)icon_width + (2 * PBTN_MARGIN);
	    }

	    if(label != NULL)
	    {
		int swidth;
		if(OSWGetTextBounds(
		    btn->font, label, -1,
		    NULL, NULL,
		    &swidth,
		    NULL, NULL
		))
		    width = swidth + (2 * PBTN_MARGIN);
	    }

	    if(width == 0)
		width = PBTN_DEF_WIDTH;
	}
	if((int)width > (int)gui->display_width)
	    width = gui->display_width;
	if((int)width < PBTN_MIN_WIDTH)
	    width = PBTN_MIN_WIDTH;

	if(height == 0)
	{
	    if(pixmap != None)
	    {
		unsigned int icon_width, icon_height;
		OSWGetPixmapGeometry(
		    pixmap, &icon_width, &icon_height, NULL
		);
		height = (int)icon_height + (2 * PBTN_MARGIN);
	    }

	    if(label != NULL)
	    {
		int ascent, descent;
		if(OSWGetTextBounds(
		    btn->font, label, -1,
		    NULL, NULL,
		    NULL,
		    &ascent, &descent
		))
		    height = (ascent + descent) + (2 * PBTN_MARGIN);
	    }

	    if(height == 0)
		height = PBTN_DEF_HEIGHT;
	}
	if((int)height > (int)gui->display_height)
	    height = gui->display_height;
	if((int)height < PBTN_MIN_HEIGHT)
	    height = PBTN_MIN_HEIGHT;

	/* Width or height cannot be odd */
	if(IS_NUM_ODD(width))
	    width += 1;
	if(IS_NUM_ODD(height))
	    height += 1;

	PBtnQueueResize(btn);
}

/*
 *	Sets the button's hint message.
 */
void PBtnSetTooltip(
	push_button_struct *btn, const char *msg
)
{
	if(btn == NULL)
	    return;

	TooltipsSetMessage(NULL, btn->toplevel, msg);
}

/*
 *	Sets the button's accelerator keys.
 */
void PBtnSetAccelKeys(
	push_button_struct *btn, 
	const keyval_t *accel_keys, const key_state_t *accel_mods,
	int naccel_keys
)
{
	if(btn == NULL)
	    return;

	free(btn->accel_keys);
	btn->accel_keys = NULL;
	free(btn->accel_mods);
	btn->accel_mods = NULL;
	btn->naccel_keys = 0;

	if((accel_keys != NULL) && (naccel_keys > 0))
	{
	    int len = naccel_keys * sizeof(keyval_t);
	    btn->accel_keys = (keyval_t *)malloc(len);
	    memcpy(btn->accel_keys, accel_keys, len);
	    if(accel_mods != NULL)
	    {
		len = naccel_keys * sizeof(key_state_t);
		btn->accel_mods = (key_state_t *)malloc(len);
		memcpy(btn->accel_mods, accel_mods, len);
	    }

	    btn->naccel_keys = naccel_keys;
	}
	if(btn->map_state)
	    PBtnQueueDraw(btn);
}

/*
 *	Sets the button's state.
 */
void PBtnSetState(push_button_struct *btn, const int state)
{
	if(btn == NULL)
	    return;

	if(btn->state != state)
	{
	    btn->state = state;
	    if(btn->map_state)
		PBtnQueueDraw(btn);
	}
}

/*
 *	Resizes the button.
 */
void PBtnResize(push_button_struct *btn)
{
	int x, y;
	unsigned int width, height;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if(btn == NULL)
	    return;

	w = btn->toplevel;
	if(w == None)
	    return;

	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);

	btn->x = x;
	btn->y = y;
	btn->width = width;
	btn->height = height;

	/* Recreate button background images */
	OSWDestroyImage(&btn->armed_img);
	OSWDestroyImage(&btn->unarmed_img);
	OSWDestroyImage(&btn->highlighted_img);
	if(wglobal->button_unarmed_img != NULL)
	{
	    image_t	*img,
			*src_img = wglobal->button_unarmed_img;
	    OSWCreateImage(&img, width, height);
	    WidgetResizeImageBuffer(
		img->bits_per_pixel, 
		(u_int8_t *)img->data,
		(const u_int8_t *)src_img->data,
		img->width, img->height,
		img->bytes_per_line, 
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    btn->unarmed_img = img;
	}
	if(wglobal->button_armed_img != NULL)
	{
	    image_t	*img,
			*src_img = wglobal->button_armed_img;
	    OSWCreateImage(&img, width, height);
	    WidgetResizeImageBuffer(
		img->bits_per_pixel,
		(u_int8_t *)img->data,
		(const u_int8_t *)src_img->data,
		img->width, img->height,
		img->bytes_per_line,
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    btn->armed_img = img;
	}
	if(wglobal->button_highlighted_img != NULL)
	{
	    image_t	*img,
			*src_img = wglobal->button_highlighted_img;
	    OSWCreateImage(&img, width, height);
	    WidgetResizeImageBuffer(
		img->bits_per_pixel,
		(u_int8_t *)img->data,
		(const u_int8_t *)src_img->data,
		img->width, img->height, 
		img->bytes_per_line,
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    btn->highlighted_img = img;
	}
}

/*
 *	Queues the button to resize.
 */
void PBtnQueueResize(push_button_struct *btn)
{
	if(btn == NULL)
	    return;

	WidgetQueueResizeAction(PBtnResizeCB, btn, NULL);
}

/*
 *	Draws the button.
 */
void PBtnDraw(push_button_struct *btn)
{
	unsigned int width, height;
	image_t *tar_img, *src_img;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(btn == NULL)
	    return;

	w = btn->toplevel;
	if(w == None)
	    return;

	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);

	/* Create the button background images as needed */

	/* Unarmed background image */
	src_img = wglobal->button_unarmed_img;
	tar_img = btn->unarmed_img;
	if((src_img != NULL) && (tar_img == NULL))
	{
	    if(OSWCreateImage(&tar_img, width, height))
		return;
	    WidgetResizeImageBuffer(
		tar_img->bits_per_pixel,
		(u_int8_t *)tar_img->data,
		(const u_int8_t *)src_img->data,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    btn->unarmed_img = tar_img;
	}

	/* Armed background image */
	src_img = wglobal->button_armed_img;
	tar_img = btn->armed_img;
	if((src_img != NULL) && (tar_img == NULL))
	{
	    if(OSWCreateImage(&tar_img, width, height))
		return;
	    WidgetResizeImageBuffer(
		tar_img->bits_per_pixel,
		(u_int8_t *)tar_img->data,
		(const u_int8_t *)src_img->data,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    btn->armed_img = tar_img;
	}

	/* Highlighted background image */
	src_img = wglobal->button_highlighted_img;
	tar_img = btn->highlighted_img;
	if((src_img != NULL) && (tar_img == NULL))
	{
	    if(OSWCreateImage(&tar_img, width, height))
		return;
	    WidgetResizeImageBuffer(
		tar_img->bits_per_pixel,
		(u_int8_t *)tar_img->data,
		(const u_int8_t *)src_img->data,
		tar_img->width, tar_img->height,
		tar_img->bytes_per_line,
		src_img->width, src_img->height,
		src_img->bytes_per_line
	    );
	    btn->highlighted_img = tar_img;
	}


	/* Draw button background */
	if(wglobal->force_mono)
	{
	    OSWSetForeground(gui->black_pix);
	    OSWDrawSolidRectangle(w, 0, 0, width, height);
	    OSWSetForeground(gui->white_pix);
	    OSWDrawRectangle(
		w,
		0, 0,
		(int)width - 1,
		(int)height - 1
	    );
	}
	else
	{
	    if(btn->state == PBTN_ARMED)
		src_img = btn->armed_img;
	    else if(btn->state == PBTN_HIGHLIGHTED)
		src_img = btn->highlighted_img;
	    else
		src_img = btn->unarmed_img;
	    OSWPutImageToDrawable(src_img, w);
	}


	/* Draw pixmap */
	if((btn->pixmap != None) && !wglobal->force_mono)
	{
	    int x, y;
	    unsigned int icon_width, icon_height;
	    pixmap_t icon_pixmap = btn->pixmap;
	    bitmap_t icon_mask = btn->mask;

	    OSWGetPixmapGeometry(
		icon_pixmap, &icon_width, &icon_height, NULL
	    );

	    switch(btn->label_align)
	    {
	      case PBTN_LABEL_ALIGN_LEFT:
		x = PBTN_MARGIN;
		break;
	      case PBTN_LABEL_ALIGN_RIGHT:
		x = (int)width - (int)icon_width - PBTN_MARGIN;
		break;
 	      default:	/* Centered */
		x = ((int)width - (int)icon_width) / 2;
		break;
	    }
	    y = ((int)height - (int)icon_height) / 2;
	    OSWSetClipMask(icon_mask);
	    OSWSetClipOrigin(x, y);
	    OSWCopyDrawablesCoord(
		w, icon_pixmap,
		x, y,
		icon_width, icon_height,
		0, 0
	    );
	    OSWSetClipMask(None);
	}


	/* Draw text label */
	if((btn->label != NULL) && (btn->font != NULL))
	{
	    int	x, y,
		lbearing, rbearing, swidth, ascent, descent,
		font_height;
	    const char *s = btn->label;
	    font_t *font = btn->font;

	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing,
		&swidth,
		&ascent, &descent
	    );
	    font_height = ascent + descent;

	    /* Calculate the label position based on the label
	     * alignment
	     */
	    switch(btn->label_align)
	    {
	      case PBTN_LABEL_ALIGN_LEFT:
		x = PBTN_MARGIN;
		break;
	      case PBTN_LABEL_ALIGN_RIGHT:
		x = (int)width - swidth - PBTN_MARGIN;
		break;
	      default:	/* Centered */
		x = ((int)width - swidth) / 2;
		break;
	    }
	    y = ((int)height - font_height) / 2;

	    /* Set font and color */
	    OSWSetFont(font);
	    if(wglobal->force_mono)
		OSWSetForeground(gui->white_pix);
	    else if(btn->disabled)
		OSWSetForeground(wglobal->disabled_text_pix);
	    else
		OSWSetForeground(wglobal->normal_text_pix);
	    OSWDrawString(
		w,
		x - lbearing,
		y + ascent,
		s
	    );

	    /* Draw accelerator key underline? */
	    if((btn->accel_keys != NULL) && (btn->naccel_keys > 0))
	    {
		const int m = btn->naccel_keys;
		char c;
		int i;
		const char *s2 = s, *ul_pos = NULL;

		/* Seek ul_pos to accelerator key character in the
		 * label string
		 */
		while(*s2 != '\0')
		{
		    for(i = 0; i < m; i++)
		    {
			c = (char)(btn->accel_keys[i] & 0x000000FF);
			if(toupper(*s2) == toupper(c))
			{
			    ul_pos = s2;
			    break;
			}
		    }
		    if(ul_pos != NULL)
			break;
		    s2++;
		}
		/* Matched accelerator key character in the label? */
		if((ul_pos != NULL) ? (*ul_pos != '\0') : False)
		{
		    int x2, y2, ul_x, ul_width;

		    i = ul_pos - s;
		    OSWGetTextBounds(
			font, s, i,
			&lbearing, &rbearing, &ul_x,
			NULL, NULL
		    );
		    ul_x -= lbearing;
		    OSWGetTextBounds(
			font, s + i, 1,
			&lbearing, &rbearing, &ul_width,
			&ascent, &descent        
		    );
		    ul_width -= lbearing;
		    font_height = ascent + descent;

		    x2 = x + ul_x;
		    y2 = y + font_height;
		    OSWDrawLine(
			w,
			x2, y2,
			x2 + ul_width, y2
		    );
		}
	    }
	}
}

/*
 *	Queues the button to draw.
 */
void PBtnQueueDraw(push_button_struct *btn)
{
	if(btn == NULL)
	    return;

	WidgetQueueDrawAction(PBtnDrawCB, btn, NULL);
}

/*
 *	Manages the button event.
 */
int PBtnManage(push_button_struct *btn, event_t *event)
{
	int events_handled = 0;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_visibility_t *visibility;

	if((btn == NULL) || (event == NULL))
	    return(events_handled);

	if(!btn->map_state)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == btn->toplevel)
	    {
		if(btn->map_state)
		    PBtnQueueDraw(btn);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == btn->toplevel)
	    {
		PBtnQueueDraw(btn);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if((btn->map_state && !btn->disabled &&
		(btn->state != PBTN_ARMED)) ?
		PBtnMatchAccelKey(btn, key->keycode, key->state) :
		False
	    )
	    {
		TooltipsUnmap(NULL);
		OSWKBAutoRepeatOff();
		btn->state = PBTN_ARMED;
		if(btn->pressed_cb != NULL)
		{
		    btn->pressed_cb(
			btn,
			btn->pressed_data
		    );
		    if(!WidgetIsRegistered(btn))
		    {
			events_handled++;
			return(events_handled);
		    }
		}

		PBtnQueueDraw(btn);
		events_handled++;
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if((btn->map_state && !btn->disabled &&
		(btn->state == PBTN_ARMED)) ?
		PBtnMatchAccelKey(btn, key->keycode, key->state) :
		False
	    )
	    {
		TooltipsUnmap(NULL);
		OSWKBAutoRepeatOn();
		btn->state = PBTN_UNARMED;
		if(btn->released_cb != NULL)
		{
		    btn->released_cb(
			btn,
			btn->released_data
		    );
		    if(!WidgetIsRegistered(btn))
		    {
			events_handled++;
			return(events_handled);
		    }
		}
		if(btn->activate_cb != NULL)
		{
		    btn->activate_cb(
			btn,
			btn->activate_data
		    );
		    if(!WidgetIsRegistered(btn))
		    {
			events_handled++;
			return(events_handled);
		    }
		}

		PBtnQueueDraw(btn);
		events_handled++;
	    }
	    else if(btn->state == PBTN_ARMED)
	    {
		/* A key was released while this button was armed, it
		 * may mean that this button accelerator key with
		 * modifier key armed it but then the modified key was
		 * released before the key was released
		 *
		 * In any case, we should return this to its unarmed
		 * state
		 */
		btn->state = PBTN_UNARMED;
		PBtnQueueDraw(btn);
	    }
	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    if(w == btn->toplevel)
	    {
		if(!btn->disabled)
		{
		    btn->has_focus = True;

		    TooltipsUnmap(NULL);

		    btn->state = PBTN_ARMED;
		    if(btn->pressed_cb != NULL)
		    {
			btn->pressed_cb(
			    btn,
			    btn->pressed_data
			);
		        if(!WidgetIsRegistered(btn))
			{
			    events_handled++;
			    return(events_handled);
			}
		    }

		    PBtnQueueDraw(btn);
		}
		events_handled++;
	    }
	    else
	    {
		if(btn->has_focus)
		{
		    btn->has_focus = False;
		    PBtnQueueDraw(btn);
		}
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(w == btn->toplevel)
	    {
		if(!btn->disabled)
		{
		    /* Button released while the button was armed? */
		    if(btn->state == PBTN_ARMED)
		    {
			const int	x = (int)button->x,
					y = (int)button->y;
			const unsigned int	width = btn->width,
						height = btn->height;

			btn->state = PBTN_HIGHLIGHTED;

			/* Unmap the hint window */
			TooltipsUnmap(NULL);

			/* Button released within the button? */
			if((x >= 0) && (y >= 0) &&
			   (x < (int)width) && (y < (int)height)
			)
			{
			    /* Report the button being released and
			     * then report it being acivated
			     */
			    if(btn->released_cb != NULL)
			    {
				btn->released_cb(
				    btn,
				    btn->released_data
				);
				if(!WidgetIsRegistered(btn))
				{
				    events_handled++;
				    return(events_handled);
				}
			    }
			    if(btn->activate_cb != NULL)
			    {
				btn->activate_cb(
				    btn,
				    btn->activate_data
				);
				if(!WidgetIsRegistered(btn))
				{
				    events_handled++;
				    return(events_handled);
				}
			    }
			}

			PBtnQueueDraw(btn);
		    }
		}

		events_handled++;
	    }
	    break;

	  case EnterNotify:
	    if(w == btn->toplevel)
	    {
		switch(btn->state)
		{
		  case PBTN_UNARMED:
		    if(!btn->disabled)
		    {
			/* Was previously unarmed when entered, so set
			 * its state to highlighted
			 */
		        btn->state = PBTN_HIGHLIGHTED;
			/* Schedual hints window map */
			TooltipsSchedualMap(
			    NULL, (time_t)-1, btn->toplevel
			);
			PBtnQueueDraw(btn);
		    }
		    break;
		}
		events_handled++;
	    }
	    break;

	  case LeaveNotify:
	    if(w == btn->toplevel)
	    {
		/* Unmap the hint window */
		TooltipsUnmap(NULL);

		switch(btn->state)
		{
		  case PBTN_ARMED:
		    if(!btn->disabled)
		    {
			/* Leaving an armed button returns it to its
			 * released state without activating it
			 */
			btn->state = PBTN_UNARMED;

			/* Must also report this button being
			 * released (but not activated)
			 */
			if(btn->released_cb != NULL)
			{
			    btn->released_cb(
				btn,
				btn->released_data
			    );
			    if(!WidgetIsRegistered(btn))
			    {
				events_handled++;
				return(events_handled);
			    }
			}

			PBtnQueueDraw(btn);
 		    }
		    break;
		  case PBTN_HIGHLIGHTED:
		    if(!btn->disabled)
		    {
			/* Leaving a highlighted button returns it
			 * to its released state
			 */
			btn->state = PBTN_UNARMED;
			PBtnQueueDraw(btn);
 		    }
		    break;
	 	}
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if(w == btn->toplevel)
	    {
		btn->visibility_state = visibility->state;
		events_handled++;
	    }
	    break;
	}

	return(events_handled);
}

/*
 *	Maps the button.
 */
void PBtnMap(push_button_struct *btn)
{
	if(btn == NULL)
	    return;

	if(!btn->map_state)
	{
	    btn->state = PBTN_UNARMED;
	    OSWMapRaised(btn->toplevel);
	    btn->map_state = True;
	}
}

/*
 *	Unmaps the button.
 */
void PBtnUnmap(push_button_struct *btn)
{
	if(btn == NULL)
	    return;

	OSWUnmapWindow(btn->toplevel);
	btn->map_state = False;

	OSWDestroyImage(&btn->unarmed_img);
	OSWDestroyImage(&btn->armed_img);
	OSWDestroyImage(&btn->highlighted_img);
}

/*
 *	Destroys the button.
 */
void PBtnDestroy(push_button_struct *btn)
{
	if(btn == NULL)
	    return;

	/* Delete the associated hint message */
	TooltipsDeleteMessage(NULL, btn->toplevel);

	OSWDestroyWindow(&btn->toplevel);

	OSWDestroyImage(&btn->unarmed_img);
	OSWDestroyImage(&btn->armed_img);
	OSWDestroyImage(&btn->highlighted_img);

	free(btn->label); 
	btn->label = NULL;

	btn->type = WIDGET_TYPE_UNKNOWN;
	btn->map_state = False;
	btn->has_focus = False;
	btn->disabled = False;
	btn->x = 0;
	btn->y = 0;
	btn->width = 0;
	btn->height = 0;
	btn->visibility_state = VisibilityFullyObscured;
	btn->font = NULL;
	btn->next = NULL;
	btn->prev = NULL;

	btn->is_default = False;
	btn->state = PBTN_UNARMED;

	btn->pixmap = None;		/* Shared */
	btn->mask = None;		/* Shared */

	btn->label_align = PBTN_LABEL_ALIGN_CENTER;

	free(btn->accel_keys);
	btn->accel_keys = NULL;
	free(btn->accel_mods);
	btn->accel_mods = NULL;
	btn->naccel_keys = 0;

	btn->pressed_cb = NULL;
	btn->pressed_data = NULL;
	btn->released_cb = NULL;
	btn->released_data = NULL;
	btn->activate_cb = NULL;
	btn->activate_data = NULL;
}
