#include <stdlib.h>
#include <cmath>

#include "../include/widget.h"

#include "../include/oswkeyvals.h"


static void SBarResizeCB(void *widget, void *data);
static void SBarDrawCB(void *widget, void *data);
static void SBarScrollTimeoutCB(void *widget, void *data);

static void SBarAddScrollTimeout(
	scroll_bar_struct *sb,
	int op,
	time_t interval
);
static void SBarScrollRepeatStop(scroll_bar_struct *sb);

float SBarGetPositionX(scroll_bar_struct *sb);
float SBarGetPositionY(scroll_bar_struct *sb);

void SBarSetPositionX(scroll_bar_struct *sb, float x);
void SBarSetPositionY(scroll_bar_struct *sb, float y);

int SBarInit(
	scroll_bar_struct *sb, win_t parent,
	unsigned int width, unsigned int height,
	unsigned int total_width, unsigned int total_height
);
void SBarResize(
	scroll_bar_struct *sb,
	unsigned int width, unsigned int height,
	unsigned int total_width, unsigned int total_height
);
void SBarQueueResize(
	scroll_bar_struct *sb,
	unsigned int width, unsigned int height,
	unsigned int total_width, unsigned int total_height
);
void SBarDraw(scroll_bar_struct *sb);
void SBarQueueDraw(scroll_bar_struct *sb); 
int SBarManage(scroll_bar_struct *sb, event_t *event);
void SBarDestroy(scroll_bar_struct *sb);


#define FTOI_RNDOFF(x)	((int)((x) + 0.5f))

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Event Masks:
 */
#define SB_BTN_EVENTMASK	(ButtonPressMask | ButtonReleaseMask)

#define SB_SLIDER_EVENTMASK	(ExposureMask | \
				ButtonPressMask | ButtonReleaseMask | \
				PointerMotionMask)


/*
 *	Minimum length of handle (in pixels):
 */
#define SB_DEF_HANDLE_MIN_LEN		5

/*
 *	Default Scroll Increments:
 */
#define SB_DEF_X_STEP_INC		20
#define SB_DEF_Y_STEP_INC		20


/*
 *	Scroll Bar Operations:
 */
#define SB_OP_CODE_NONE			0
#define SB_OP_CODE_SCROLL_UP		1	/* One step */
#define SB_OP_CODE_SCROLL_DOWN		2	/* One step */
#define SB_OP_CODE_SCROLL_LEFT		3	/* One step */
#define SB_OP_CODE_SCROLL_RIGHT		4	/* One step */
#define SB_OP_CODE_PAGE_UP		5	/* One page */
#define SB_OP_CODE_PAGE_DOWN		6	/* One page */
#define SB_OP_CODE_PAGE_LEFT		7	/* One page */
#define SB_OP_CODE_PAGE_RIGHT		8	/* One page */


/*
 *	Resize callback.
 */
static void SBarResizeCB(void *widget, void *data)
{
	scroll_bar_struct *sb = (scroll_bar_struct *)widget;
	if(sb == NULL)
	    return;

	/* Note that SBarQueueResize() has already set the geometry
	 * so all we need to do is just pass them to SBarResize()
	 */
	SBarResize(
	    sb,
	    sb->width, sb->height, sb->total_width, sb->total_height
	);
}

/*
 *	Draw callback.
 */
static void SBarDrawCB(void *widget, void *data)
{
	scroll_bar_struct *sb = (scroll_bar_struct *)widget;
	if(sb == NULL)
	    return;

	SBarDraw(sb);
}

/*
 *	Scroll timeout callback.
 */
static void SBarScrollTimeoutCB(void *widget, void *data)
{
	int prev_x_pos, prev_y_pos;
	int op = (int)data;
	scroll_bar_struct *sb = (scroll_bar_struct *)widget;
	if(sb == NULL)
	    return;

	prev_x_pos = sb->x_win_pos;
	prev_y_pos = sb->y_win_pos;

	switch(op)
	{
	  case SB_OP_CODE_SCROLL_UP:
	    sb->y_win_pos -= sb->y_step_inc;
	    break;
	  case SB_OP_CODE_SCROLL_DOWN:
	    sb->y_win_pos += sb->y_step_inc;
	    break;
	  case SB_OP_CODE_SCROLL_LEFT:
	    sb->x_win_pos -= sb->x_step_inc;
	    break;
	  case SB_OP_CODE_SCROLL_RIGHT:
	    sb->x_win_pos += sb->x_step_inc;
	    break;

	  case SB_OP_CODE_PAGE_UP:
	    sb->y_win_pos -= sb->y_page_inc;
	    break;
	  case SB_OP_CODE_PAGE_DOWN:
	    sb->y_win_pos += sb->y_page_inc;
	    break;
	  case SB_OP_CODE_PAGE_LEFT:
	    sb->x_win_pos -= sb->x_page_inc;
	    break;
	  case SB_OP_CODE_PAGE_RIGHT:
	    sb->x_win_pos += sb->x_page_inc;
	    break;
	}

	/* Clip scroll position */
	if((sb->x_win_pos + (int)sb->width) > (int)sb->total_width)
	    sb->x_win_pos = (int)sb->total_width - (int)sb->width;
	if(sb->x_win_pos < 0)
	    sb->x_win_pos = 0;

	if((sb->y_win_pos + (int)sb->height) > (int)sb->total_height)
	    sb->y_win_pos = (int)sb->total_height - (int)sb->height;
	if(sb->y_win_pos < 0)
	    sb->y_win_pos = 0;

	if((prev_x_pos != sb->x_win_pos) || (prev_y_pos != sb->y_win_pos))
	{
	    SBarQueueDraw(sb);

	    /* Report value changed */
	    if(sb->value_changed_cb != NULL)
		sb->value_changed_cb(
		sb,
		(float)sb->x_win_pos,
		(float)sb->y_win_pos,
		sb->value_changed_data
	    );

	    /* Repeat this scroll if a button is still pressed
	     * anywhere on this scroll bar
	     */
	    if(sb->button_pressed > 0)
		SBarAddScrollTimeout(
		    sb, op,
		    widget_global.sbar_scroll_repeat_int
		);
	}
}

/*
 *	Scheduals a scroll timeout.
 */
static void SBarAddScrollTimeout(
	scroll_bar_struct *sb,
	int op,
	time_t interval
)
{
	if(sb == NULL)
	    return;

	WidgetTimeoutAdd(
	    SBarScrollTimeoutCB,
	    sb, (void *)op,
	    interval
	);
}

/*
 *	Removes all timeouts associated with the specified scroll bar.
 */
static void SBarScrollRepeatStop(scroll_bar_struct *sb)
{
	if(sb == NULL)
	    return;

	WidgetTimeoutRemoveByFuncWidget(
	    SBarScrollTimeoutCB, sb
	);
}

/*
 *	Returns the x scroll bar's current position.
 */
float SBarGetPositionX(scroll_bar_struct *sb)
{
	return((sb != NULL) ? (float)sb->x_win_pos : 0.0f);
}

/*
 *	Returns the y scroll bar's current position.
 */
float SBarGetPositionY(scroll_bar_struct *sb)
{
	return((sb != NULL) ? (float)sb->y_win_pos : 0.0f);
}


/*
 *	Sets the x scroll bar's position.
 *
 *	If the specified value has changed then the value_changed
 *	callback will be called.
 */
void SBarSetPositionX(scroll_bar_struct *sb, float x)
{
	int i;

	if(sb == NULL)
	    return;

	i = (int)x;
	if((i + (int)sb->width) > (int)sb->total_width)
	    i = (int)sb->total_width - (int)sb->width;
	if(i < 0)
	    i = 0;

	if(sb->x_win_pos != i)
	{
	    sb->x_win_pos = i;
	    SBarQueueDraw(sb);
	    if(sb->value_changed_cb != NULL)
		sb->value_changed_cb(
		    sb,
		    (float)sb->x_win_pos,
		    (float)sb->y_win_pos,
		    sb->value_changed_data
		);
	}
}

/*
 *	Sets the y scroll bar's position.
 *
 *	If the specified value has changed then the value_changed
 *	callback will be called.
 */
void SBarSetPositionY(scroll_bar_struct *sb, float y)
{
	int i;

	if(sb == NULL)
	    return;

	i = (int)y;
	if((i + (int)sb->height) > (int)sb->total_height)
	    i = (int)sb->total_height - (int)sb->height;
	if(i < 0)
	    i = 0;

	if(sb->y_win_pos != i)
	{
	    sb->y_win_pos = i;
	    SBarQueueDraw(sb);
	    if(sb->value_changed_cb != NULL)
		sb->value_changed_cb(
		    sb,
		    (float)sb->x_win_pos,
		    (float)sb->y_win_pos,
		    sb->value_changed_data
		);
	}
}


/*
 *	Initializes the scroll bar.
 */
int SBarInit(
	scroll_bar_struct *sb, win_t parent,
	unsigned int width, unsigned int height,
	unsigned int total_width, unsigned int total_height
)
{
	win_t parent2;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if((sb == NULL) || (parent == None))
	    return(-2);

	sb->type = WIDGET_TYPE_SCROLL_BAR;
	sb->map_state = False;
	sb->has_focus = False;
	sb->disabled = False;
	sb->visibility_state = VisibilityFullyObscured;
	sb->x = 0;
	sb->y = 0;
	sb->width = MAX(width, 1);
	sb->height = MAX(height, 1);
	sb->font = wglobal->std_font;
	sb->prev = NULL;
	sb->next = NULL;
	sb->data = NULL;

	sb->flags = 0;

	sb->button_pressed = 0;
	sb->button_last_x = 0;
	sb->button_last_y = 0;
	sb->button_pressed_time = 0l;

	sb->total_width = MAX(total_width, 1);
	sb->total_height = MAX(total_height, 1);

	sb->x_win_pos = 0;
	sb->y_win_pos = 0;

	sb->x_step_inc = SB_DEF_X_STEP_INC;
	sb->x_page_inc = (int)sb->width;
	sb->y_step_inc = SB_DEF_Y_STEP_INC;
	sb->y_page_inc = (int)sb->height;

	sb->x_origin_delta = 0;
	sb->y_origin_delta = 0;

	sb->value_changed_cb = NULL;
	sb->value_changed_data = NULL;

	/* Create the x scroll bar */
	OSWCreateWindow(
	    &sb->x_toplevel, parent,
	    0, 0, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT
	);
	OSWSetWindowCursor(sb->x_toplevel, gui->std_cursor);
	sb->x_toplevel_buf = None;
	parent2 = sb->x_toplevel;

	/* X left button */
	OSWCreateWindow(
	    &sb->x_left_btn, parent2,
	    0, 0, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT
	);
	OSWSetWindowInput(sb->x_left_btn, SB_BTN_EVENTMASK);

	/* X slider */
	OSWCreateWindow(
	    &sb->x_slider, parent2,
	    0, 0, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT
	);
	OSWSetWindowInput(sb->x_slider, SB_SLIDER_EVENTMASK);

	/* X right button */
	OSWCreateWindow(
	    &sb->x_right_btn, parent2,
	    0, 0, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT
	);
	OSWSetWindowInput(sb->x_right_btn, SB_BTN_EVENTMASK);


	/* Create the y scroll bar */
	OSWCreateWindow(
	    &sb->y_toplevel, parent,
	    0, 0, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT
	);
	OSWSetWindowCursor(sb->y_toplevel, gui->std_cursor);
	sb->y_toplevel_buf = None;
	parent2 = sb->y_toplevel;

	/* Y up button */
	OSWCreateWindow(
	    &sb->y_up_btn, parent2,
	    0, 0, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT
	);
	OSWSetWindowInput(sb->y_up_btn, SB_BTN_EVENTMASK);

	/* Y slider */
	OSWCreateWindow(
	    &sb->y_slider, parent2,
	    0, 0, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT
	);
	OSWSetWindowInput(sb->y_slider, SB_SLIDER_EVENTMASK);

	/* Y down button */
	OSWCreateWindow(
	    &sb->y_down_btn, parent2,
	    0, 0, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT
	);
	OSWSetWindowInput(sb->y_down_btn, SB_BTN_EVENTMASK);


	/* Set initial position, size, and map/unmap */
	SBarResize(sb, sb->width, sb->height, sb->total_width, sb->total_height); 

	sb->map_state = True;	/* Always considered mapped */

	/* Add widget to regeristry */
	WidgetRegister(sb, WIDGET_TYPE_SCROLL_BAR);

	return(0);
}

/*
 *	Resizes the scroll bar.
 */
void SBarResize(
	scroll_bar_struct *sb,
	unsigned int width, unsigned int height,
	unsigned int total_width, unsigned int total_height
)
{
	bool_t x_bar_map_state, y_bar_map_state;
	int x, y, prev_x_pos, prev_y_pos;

	if((sb == NULL) || (width == 0) || (height == 0))
	    return;

	prev_x_pos = sb->x_win_pos;
	prev_y_pos = sb->y_win_pos;

	/* Update geometry */
	sb->width = MAX(width, 1);
	sb->height = MAX(height, 1);
	sb->total_width = MAX(total_width, 1);
	sb->total_height = MAX(total_height, 1);

	/* Clip scroll positions */
	if((sb->x_win_pos + (int)width) > (int)total_width)
	    sb->x_win_pos = MAX(((int)total_width - (int)width), 0);
	if((sb->y_win_pos + (int)height) > (int)total_height)
	    sb->y_win_pos = MAX(((int)total_height - (int)height), 0);

	/* Update page increments */
	sb->x_page_inc = (int)width;
	sb->y_page_inc = (int)height;

	/* Get scroll bar map states */
	x_bar_map_state = (sb->width < sb->total_width) ? True : False;
	y_bar_map_state = (sb->height < sb->total_height) ? True : False;


	/* Destroy buffers */
	OSWDestroyPixmap(&sb->x_toplevel_buf);
	OSWDestroyPixmap(&sb->y_toplevel_buf);

	/* Resize the x scroll bar */

	/* X toplevel */
	x = 0;
	y = (int)sb->height - SCROLLBAR_XBAR_HEIGHT;
	width = (unsigned int)MAX(
	    (int)sb->width -
	    (y_bar_map_state ? SCROLLBAR_YBAR_WIDTH : 0),
	    1
	);
	height = SCROLLBAR_XBAR_HEIGHT;
	OSWMoveResizeWindow(sb->x_toplevel, x, y, width, height);

	/* X left button */
	x = 0;
	y = FTOI_RNDOFF(((float)height - SCROLLBAR_BTN_HEIGHT) / 2.0f);
	width = SCROLLBAR_BTN_WIDTH;
	height = SCROLLBAR_BTN_HEIGHT;
	OSWMoveResizeWindow(sb->x_left_btn, x, y, width, height);

	/* X slider */
	x += SCROLLBAR_BTN_WIDTH;
	width = (unsigned int)MAX(
	    (int)sb->width - (2 * SCROLLBAR_BTN_WIDTH) -
	    (y_bar_map_state ? SCROLLBAR_YBAR_WIDTH : 0),
	    1
	);
	OSWMoveResizeWindow(sb->x_slider, x, y, width, height);

	/* X right button */
	x += (int)width;
	width = SCROLLBAR_BTN_WIDTH;
	OSWMoveResizeWindow(sb->x_right_btn, x, y, width, height);


	/* Resize the y scroll bar */

	/* Y toplevel */
	x = (int)sb->width - SCROLLBAR_YBAR_WIDTH;
	y = 0;
	width = SCROLLBAR_YBAR_WIDTH;
	height = (unsigned int)MAX(
	    (int)sb->height -
	    (x_bar_map_state ? SCROLLBAR_XBAR_HEIGHT : 0),
	    1
	);
	OSWMoveResizeWindow(sb->y_toplevel, x, y, width, height);

	/* Y up button */
	x = FTOI_RNDOFF(((float)width - SCROLLBAR_BTN_WIDTH) / 2.0f);
	y = 0;
	width = SCROLLBAR_BTN_WIDTH;
	height = SCROLLBAR_BTN_HEIGHT;
	OSWMoveResizeWindow(sb->y_up_btn, x, y, width, height);

	/* Y slider */
	y += SCROLLBAR_BTN_HEIGHT;
	height = (unsigned int)MAX(
	    (int)sb->height - (2 * SCROLLBAR_BTN_HEIGHT) -
	    (x_bar_map_state ? SCROLLBAR_XBAR_HEIGHT : 0),
	    1
	);
	OSWMoveResizeWindow(sb->y_slider, x, y, width, height);

	/* Y down button */
	y += (int)height;
	height = SCROLLBAR_BTN_HEIGHT;
	OSWMoveResizeWindow(sb->y_down_btn, x, y, width, height);


	/* Map/unmap the scroll bars */
	if(x_bar_map_state)
	{
	    OSWMapRaised(sb->x_toplevel);
	    OSWMapWindow(sb->x_left_btn);
	    OSWMapWindow(sb->x_slider);
	    OSWMapWindow(sb->x_right_btn);
	}
	else
	{
	    OSWUnmapWindow(sb->x_toplevel);
	    OSWUnmapWindow(sb->x_left_btn);
	    OSWUnmapWindow(sb->x_slider);
	    OSWUnmapWindow(sb->x_right_btn);
	}

	if(y_bar_map_state)
	{
	    OSWMapRaised(sb->y_toplevel);
	    OSWMapWindow(sb->y_up_btn);
	    OSWMapWindow(sb->y_slider);
	    OSWMapWindow(sb->y_down_btn);
	}
	else
	{
	    OSWUnmapWindow(sb->y_toplevel);
	    OSWUnmapWindow(sb->y_up_btn);
	    OSWUnmapWindow(sb->y_slider);
	    OSWUnmapWindow(sb->y_down_btn);
	}

	/* Notify of value changed as needed */
	if((sb->x_win_pos != prev_x_pos) || (sb->y_win_pos != prev_y_pos))
	{
	    if(sb->value_changed_cb != NULL)
		sb->value_changed_cb(
		    sb,
		    (float)sb->x_win_pos,
		    (float)sb->y_win_pos,
		    sb->value_changed_data
		);
	}
}

/*
 *	Queues the scroll bar to resize.
 */
void SBarQueueResize(
	scroll_bar_struct *sb,
	unsigned int width, unsigned int height,
	unsigned int total_width, unsigned int total_height
)
{
	if((sb == NULL) || (width == 0) || (height == 0))
	    return;

	/* We need to record the geometry here because SBarResizeCB()
	 * will pass them on to SBarResize()
	 */
	sb->width = width;
	sb->height = height;
	sb->total_width = total_width;
	sb->total_height = total_height;

	/* Queue resize */
	WidgetQueueResizeAction(SBarResizeCB, sb, NULL);
}

/*
 *	Draws the scroll bar.
 */
void SBarDraw(scroll_bar_struct *sb)
{
	unsigned int	width, height,
			total_width, total_height,
			wwidth, wheight;
	float		x_pos, y_pos,
			handle_pos, handle_len;
	int		handle_posi, handle_leni;
	pixel_t		bg_pix, frame_pix, btn_pix, handle_pix;
	pixmap_t pixmap;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(sb == NULL)
	    return;

	width = MAX(sb->width, 1);
	height = MAX(sb->height, 1);
	total_width = MAX(sb->total_width, 1);
	total_height = MAX(sb->total_height, 1);

	/* Recreate buffers as needed */
	if(sb->x_toplevel_buf == None)
	{
	    OSWGetWindowGeometry(
		sb->x_toplevel, NULL, NULL, &wwidth, &wheight, NULL
	    );
	    OSWCreatePixmap(&sb->x_toplevel_buf, wwidth, wheight);
	}
	if(sb->y_toplevel_buf == None)
	{
	    OSWGetWindowGeometry(
		sb->y_toplevel, NULL, NULL, &wwidth, &wheight, NULL
	    );
	    OSWCreatePixmap(&sb->y_toplevel_buf, wwidth, wheight);
	}

	/* Get colors */
	if(wglobal->force_mono)
	{
	    bg_pix = gui->black_pix;
	    frame_pix = gui->white_pix;
	    btn_pix = gui->white_pix;
	    handle_pix = gui->white_pix;
	}
	else
	{
	    bg_pix = wglobal->scroll_bar_bg_pix;
	    frame_pix = wglobal->scroll_bar_frame_pix;
	    btn_pix = wglobal->scroll_bar_btn_pix;
	    handle_pix = wglobal->scroll_bar_handle_pix;
	}


	/* Begin drawing */

	/* X scroll bar left button */
	pixmap = sb->x_toplevel_buf;
	OSWClearPixmap(
	    pixmap, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT,
	    bg_pix
	);
	OSWSetForeground(btn_pix);
	OSWDrawSolidArc(
	    pixmap,
	    0,
	    0,
	    (2 * SCROLLBAR_BTN_WIDTH) - 2,
	    SCROLLBAR_BTN_HEIGHT - 1,
	    0.0f * PI, 2.0f * PI
	);
        OSWSetForeground(frame_pix);
        OSWDrawArc(
            pixmap,
            0,
            0,
            (2 * SCROLLBAR_BTN_WIDTH) - 2,
            SCROLLBAR_BTN_HEIGHT - 1,
            0.0f * PI, 2.0f * PI
        );
	OSWPutDrawableToWindow(sb->x_left_btn, pixmap);

	/* X scroll bar right button */
	OSWClearPixmap(
	    pixmap, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT,
	    bg_pix
	);
	OSWSetForeground(btn_pix);
	OSWDrawSolidArc(
	    pixmap,
	    0 - SCROLLBAR_BTN_WIDTH + 1,
	    0,
	    (2 * SCROLLBAR_BTN_WIDTH) - 2,
	    SCROLLBAR_BTN_HEIGHT - 1,
	    0.0f * PI, 2.0f * PI
	);
	OSWSetForeground(frame_pix);
        OSWDrawArc(
            pixmap,
            0 - SCROLLBAR_BTN_WIDTH + 1,
            0,
            (2 * SCROLLBAR_BTN_WIDTH) - 2,
            SCROLLBAR_BTN_HEIGHT - 1,
            0.0f * PI, 2.0f * PI
        );
	OSWPutDrawableToWindow(sb->x_right_btn, pixmap);


	/* Y scroll bar up button */
	pixmap = sb->y_toplevel_buf;
	OSWClearPixmap(
	    pixmap, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT,
	    bg_pix
	);
	OSWSetForeground(btn_pix);
	OSWDrawSolidArc(
	    pixmap,
	    0,
	    0,
	    SCROLLBAR_BTN_WIDTH - 1,
	    (2 * SCROLLBAR_BTN_HEIGHT) - 2,
	    0.0f * PI, 2.0f * PI
	);
	OSWSetForeground(frame_pix);
        OSWDrawArc(
            pixmap,
            0,
            0,
            SCROLLBAR_BTN_WIDTH - 1,
            (2 * SCROLLBAR_BTN_HEIGHT) - 2,
            0.0f * PI, 2.0f * PI
        );
	OSWPutDrawableToWindow(sb->y_up_btn, pixmap);

	/* Y scroll bar down button */
	OSWClearPixmap(
	    pixmap, SCROLLBAR_BTN_WIDTH, SCROLLBAR_BTN_HEIGHT,
	    bg_pix
	);
	OSWSetForeground(btn_pix);
	OSWDrawSolidArc(
	    pixmap,
	    0,
	    0 - SCROLLBAR_BTN_HEIGHT + 1,
	    SCROLLBAR_BTN_WIDTH - 1,
	    (2 * SCROLLBAR_BTN_HEIGHT) - 2,
	    0.0f * PI, 2.0f * PI
	);
	OSWSetForeground(frame_pix);
        OSWDrawArc(
            pixmap,
            0,
            0 - SCROLLBAR_BTN_HEIGHT + 1,
            SCROLLBAR_BTN_WIDTH - 1,
            (2 * SCROLLBAR_BTN_HEIGHT) - 2,
            0.0f * PI, 2.0f * PI
        );
	OSWPutDrawableToWindow(sb->y_down_btn, pixmap);


	/* X slider */
	pixmap = sb->x_toplevel_buf;
	OSWGetWindowGeometry(
	    sb->x_slider,
	    NULL, NULL, &wwidth, &wheight, NULL
	);
	OSWClearPixmap(
	    pixmap, wwidth, wheight, bg_pix
	);

	/* Calculate the handle's position & length */
	x_pos = (float)sb->x_win_pos;
	handle_pos = (float)wwidth * (x_pos / (float)total_width);
	handle_len = (float)wwidth * ((float)width / (float)total_width);

	/* Make sure that the handle keeps a minimum size */
	if(handle_len < SB_DEF_HANDLE_MIN_LEN)
	{
	    handle_pos -= (SB_DEF_HANDLE_MIN_LEN - handle_len) / 2.0f;
	    handle_len = SB_DEF_HANDLE_MIN_LEN;
	}

	/* Convert handle position to whole number */
	handle_posi = FTOI_RNDOFF(handle_pos);
	handle_leni = FTOI_RNDOFF(handle_len);

	/* Clip the handle's position */
	if((handle_posi + handle_leni) > (int)wwidth)
	    handle_posi = (int)wwidth - handle_leni;
	if(handle_posi < 0)
	    handle_posi = 0;

	/* Draw the handle */
	OSWSetForeground(handle_pix);
	OSWDrawSolidRectangle(
	    pixmap,
	    handle_posi,
	    0,
	    handle_leni,
	    wheight
	);
	OSWSetForeground(frame_pix);
	OSWDrawLine(
	    pixmap,
	    handle_posi,
	    0,
	    handle_posi,
	    wheight
	);
	OSWDrawLine(
	    pixmap,
	    handle_posi + handle_leni - 1,
	    0,
	    handle_posi + handle_leni - 1,
	    wheight
	);

	/* Draw the frame */
	OSWDrawRectangle(
	    pixmap,
	    0, 0,
	    wwidth - 1, wheight - 1
	);

	OSWPutDrawableToWindow(sb->x_slider, pixmap);


	/* Y slider */
	pixmap = sb->y_toplevel_buf;
	OSWGetWindowGeometry(
	    sb->y_slider,
	    NULL, NULL, &wwidth, &wheight, NULL
	);
	OSWClearPixmap(
	    pixmap, wwidth, wheight, bg_pix
	);

	/* Calculate the handle's position & length */
	y_pos = (float)sb->y_win_pos;
	handle_pos = (float)wheight * (y_pos / (float)total_height);
	handle_len = (float)wheight * ((float)height / (float)total_height);

	/* Make sure that the handle keeps a minimum length */
	if(handle_len < SB_DEF_HANDLE_MIN_LEN)
	{
	    handle_pos -= (SB_DEF_HANDLE_MIN_LEN - handle_len) / 2.0f;
	    handle_len = SB_DEF_HANDLE_MIN_LEN;
	}

	/* Convert the handle position & length to whole numbers */
	handle_posi = FTOI_RNDOFF(handle_pos);
	handle_leni = FTOI_RNDOFF(handle_len);

	/* Clip the handle's position */
	if((handle_posi + handle_leni) > (int)wheight)
	    handle_posi = (int)wheight - handle_leni;
	if(handle_posi < 0)
	    handle_posi = 0;

	/* Draw the handle */
	OSWSetForeground(handle_pix);
	OSWDrawSolidRectangle(
	    pixmap,
	    0,
	    handle_posi,
	    wwidth,
	    handle_leni
	);
	OSWSetForeground(frame_pix);
	OSWDrawLine(
	    pixmap,
	    0,
	    handle_posi,
	    wwidth,
	    handle_posi
	);
	OSWDrawLine(
	    pixmap,
	    0,
	    handle_posi + handle_leni - 1,
	    wwidth,
	    handle_posi + handle_leni - 1
	);

	/* Draw the frame */
	OSWDrawRectangle(
	    pixmap,
	    0, 0,
	    wwidth - 1, wheight - 1
	);

	OSWPutDrawableToWindow(sb->y_slider, pixmap);
}

/*
 *	Queues the scroll bar to draw.
 */
void SBarQueueDraw(scroll_bar_struct *sb)
{
	if(sb == NULL)
	    return;

	WidgetQueueDrawAction(SBarDrawCB, sb, NULL);
}

/*
 *	Manages the scroll bar event.
 */
int SBarManage(scroll_bar_struct *sb, event_t *event)
{
	int events_handled = 0;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	win_t w;
	event_button_t *button;
	event_key_t *key;
	event_motion_t *motion;
	event_visibility_t *visibility;

	if((sb == NULL) || (event == NULL))
	    return(events_handled);

	if(!sb->map_state && (event->type != MapNotify))
	    return(events_handled);

#define REPORT_VALUE_CHANGED(_sb_)	{	\
 if((_sb_)->value_changed_cb != NULL)		\
  (_sb_)->value_changed_cb(			\
  (_sb_),					\
  (float)(_sb_)->x_win_pos,			\
  (float)(_sb_)->y_win_pos,			\
  (_sb_)->value_changed_data			\
 );						\
}
	w = event->xany.window;

	switch(event->type)
	{
	  case Expose:
	    if((w == sb->x_toplevel) ||
	       (w == sb->y_toplevel) ||
	       (w == sb->x_slider) ||
	       (w == sb->y_slider) ||
	       (w == sb->x_left_btn) ||
	       (w == sb->x_right_btn) ||
	       (w == sb->y_up_btn) ||
	       (w == sb->y_down_btn)
	    )
	    {
		SBarQueueDraw(sb);
	        events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;

	    /* Skip if not in focus */
	    if(!sb->has_focus)
		return(events_handled);

	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Up */
	    if((keyval == OSW_KEY_Up) ||
               (keyval == OSW_KEY_KP_Up)
	    )
	    {
		const int y_min = 0;
		sb->y_win_pos = MAX(
		    (sb->y_win_pos - sb->y_step_inc),
		    y_min
		);
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Down */
	    else if((keyval == OSW_KEY_Down) ||
                    (keyval == OSW_KEY_KP_Down)
	    )
	    {
		const int y_max = MAX(
		    ((int)sb->total_height - (int)sb->height),
		    0
		);
		sb->y_win_pos = MIN(
		    (sb->y_win_pos + sb->y_step_inc),
		    y_max
		);
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Page Up */
            else if((keyval == OSW_KEY_Page_Up) ||
                    (keyval == OSW_KEY_KP_Page_Up)   
            )
	    {
		const int y_min = 0;
		sb->y_win_pos = MAX(
		    (sb->y_win_pos - (int)sb->height),
		    y_min
		);
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Page Down */
            else if((keyval == OSW_KEY_Page_Down) ||
                    (keyval == OSW_KEY_KP_Page_Down)   
            )
	    {
		const int y_max = MAX(
		    ((int)sb->total_height - (int)sb->height),
		    0
		);
		sb->y_win_pos = MIN(
		    (sb->y_win_pos + (int)sb->height),
		    y_max
		);
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Home */
            else if((keyval == OSW_KEY_Home) ||
                    (keyval == OSW_KEY_KP_Home)
            )
	    {
		const int y_min = 0;
		sb->y_win_pos = y_min;
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* End */
            else if((keyval == OSW_KEY_End) ||
                    (keyval == OSW_KEY_KP_End)
            )
	    {
		const int y_max = MAX(
		    ((int)sb->total_height - (int)sb->height),
		    0
		);
		sb->y_win_pos = y_max;
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Left */
            else if((keyval == OSW_KEY_Left) ||
                    (keyval == OSW_KEY_KP_Left)   
            )
	    {
		const int x_min = 0;
		sb->x_win_pos = MAX(
		    (sb->x_win_pos - sb->x_step_inc),
		    x_min
		);
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Right */
            else if((keyval == OSW_KEY_Right) ||
                    (keyval == OSW_KEY_KP_Right)
            )
	    {
		const int x_max = MAX(
		    ((int)sb->total_width - (int)sb->width),
		    0
		);
		sb->x_win_pos = MIN(
		    (sb->x_win_pos + sb->x_step_inc),
		    x_max
		);
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;

	    if(!sb->has_focus)
		return(events_handled);

	    break;

	  case ButtonPress:
	    button = &event->xbutton;

	    /* Change focus by which window buttonpress is on */
	    if((w == sb->x_slider) ||
	       (w == sb->y_slider) ||
	       (w == sb->x_left_btn) ||
	       (w == sb->x_right_btn) ||
	       (w == sb->y_up_btn) ||
	       (w == sb->y_down_btn)
	    )
	        sb->has_focus = True;
	    else
		sb->has_focus = False;

	    /* X slider */
	    if(w == sb->x_slider)
	    {
		const int button_num = button->button;
		const int	x = FTOI_RNDOFF((float)button->x),
				x_min = 0,
				x_max = MAX(
				    ((int)sb->total_width - (int)sb->width),
				    0
				);
		unsigned int wwidth, wheight;
		int	handle_x_min = 0,
			handle_x_max = 0,
			handle_len = 0;

		sb->button_pressed = button_num;

		OSWGetWindowGeometry(w, NULL, NULL, &wwidth, &wheight, NULL);

		/* Calculate the handle's position & length */
		if(sb->total_width > 0)
		{
		    handle_x_min = FTOI_RNDOFF(
			(float)sb->x_win_pos * (float)wwidth /
			    (float)sb->total_width
		    );
		    handle_len = FTOI_RNDOFF(
			(float)sb->width * (float)wwidth /
			    (float)sb->total_width
		    );
		}
		/* Make sure that handle stays at least a minimum
		 * length
		 */
		if(handle_len < SB_DEF_HANDLE_MIN_LEN)
		{
		    handle_x_min = MAX(
			FTOI_RNDOFF(
			    (float)handle_x_min -
			    ((SB_DEF_HANDLE_MIN_LEN - (float)handle_len) / 2.0f)
			), 0
		    );
		    handle_len = SB_DEF_HANDLE_MIN_LEN;
		}
		handle_x_max = handle_x_min + handle_len;

		/* Handle by button number */
		switch(button_num)
		{
		  case Button2:
		    if(wwidth > 0)
		    {
			sb->flags |= SCROLLBAR_FLAG_X_HANDLE_CLICKED;
			sb->x_win_pos = FTOI_RNDOFF(
			    (((float)x / (float)wwidth) *
			     (float)sb->total_width) -
			    ((float)sb->width / 2)
			);
			sb->x_origin_delta = 0;
		    }
		    break;
		  case Button4:
		  case Button5:
/* TODO */
		    break;
		  default:
		    /* Clicked on the trough? */
		    if((x < handle_x_min) || (x > handle_x_max))
		    {
			if(x < handle_x_min)
			{
			    sb->x_win_pos -= sb->x_page_inc;
			    SBarScrollRepeatStop(sb);
			    SBarAddScrollTimeout(
				sb, SB_OP_CODE_PAGE_LEFT,
				widget_global.sbar_scroll_repeat_delay
			    );
			}
			else
			{
			    sb->x_win_pos += sb->x_page_inc;
			    SBarScrollRepeatStop(sb);
			    SBarAddScrollTimeout(
				sb, SB_OP_CODE_PAGE_RIGHT,
				widget_global.sbar_scroll_repeat_delay
			    );
			}
		    }
		    else
		    {
			/* Clicked on the handle */
			sb->flags |= SCROLLBAR_FLAG_X_HANDLE_CLICKED;
			if(wwidth > 0)
			    sb->x_origin_delta = FTOI_RNDOFF(
				(float)sb->x_win_pos - (
				    ((float)sb->total_width / (float)wwidth) *
					(float)x) + ((float)sb->width / 2)
			    );
			else
			    sb->x_origin_delta = 0;
		    }
		    break;
		}

		/* Clip scroll position */
		if(sb->x_win_pos > x_max)
		    sb->x_win_pos = x_max;
		else if(sb->x_win_pos < x_min)
		    sb->x_win_pos = x_min;
			    
		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Y slider */
	    else if(w == sb->y_slider)
	    {
		const int button_num = button->button;
		const int	y = FTOI_RNDOFF((float)button->y),
				y_min = 0,
				y_max = MAX(
				    ((int)sb->total_height - (int)sb->height),
				    0
				);
		unsigned int wwidth, wheight;
		int	handle_y_min = 0,
			handle_y_max = 0,
			handle_len = 0;

		sb->button_pressed = button_num;

		OSWGetWindowGeometry(w, NULL, NULL, &wwidth, &wheight, NULL);

		/* Calculate the handle's bounds */
		if(sb->total_height > 0)
		{
		    handle_y_min = FTOI_RNDOFF(
			(float)sb->y_win_pos * (float)wheight /
			    (float)sb->total_height
		    );
		    handle_len = FTOI_RNDOFF(
			(float)sb->height * (float)wheight /
			    (float)sb->total_height
		    );
		}
		/* Make sure that handle stays at least a minimum 
		 * length   
		 */
		if(handle_len < SB_DEF_HANDLE_MIN_LEN)
		{
		    handle_y_min = MAX(
			FTOI_RNDOFF(
			    (float)handle_y_min -
			    ((SB_DEF_HANDLE_MIN_LEN - (float)handle_len) / 2.0f)
			), 0
		    );
		    handle_len = SB_DEF_HANDLE_MIN_LEN;
		}
		handle_y_max = handle_y_min + handle_len;

		/* Handle by button number */
		switch(button_num)
		{
		  case Button2:
		    if(wheight > 0)
		    {
			sb->flags |= SCROLLBAR_FLAG_Y_HANDLE_CLICKED;
			sb->y_win_pos = FTOI_RNDOFF(
			    (((float)y / (float)wheight) *
			     (float)sb->total_height) -
			    ((float)sb->height / 2)
			);
			sb->y_origin_delta = 0;
		    }
		    break;
		  case Button4:
		  case Button5:
/* TODO */
		    break;
		  default:
		    /* Clicked on the trough? */
		    if((y < handle_y_min) || (y > handle_y_max))
		    {
			if(y < handle_y_min)
			{
			    sb->y_win_pos -= sb->y_page_inc;
			    SBarScrollRepeatStop(sb);
			    SBarAddScrollTimeout(
			        sb, SB_OP_CODE_PAGE_UP,
			        widget_global.sbar_scroll_repeat_delay
			    );
		        }
		        else
		        {
			    sb->y_win_pos += sb->y_page_inc;
			    SBarScrollRepeatStop(sb);
			    SBarAddScrollTimeout(
				sb, SB_OP_CODE_PAGE_DOWN,
				widget_global.sbar_scroll_repeat_delay
			    );
		        }
		    }
		    else
		    {
		        /* Clicked on the handle */
			sb->flags |= SCROLLBAR_FLAG_Y_HANDLE_CLICKED;
			if(wheight > 0)
			    sb->y_origin_delta = FTOI_RNDOFF(
				(float)sb->y_win_pos - (
				    ((float)sb->total_height / (float)wheight) *
					(float)y) + ((float)sb->height / 2)
			    );
			else
			    sb->y_origin_delta = 0;
		    }
		    break;
		}

		/* Clip scroll position */
		if(sb->y_win_pos > y_max)
		    sb->y_win_pos = y_max;
		else if(sb->y_win_pos < y_min)
		    sb->y_win_pos = y_min;

		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Y up button */
	    else if(w == sb->y_up_btn)
	    {
		const int	button_num = button->button,
				y_min = 0;

		sb->flags |= SCROLLBAR_FLAG_Y_UP_BTN_CLICKED;
		sb->button_pressed = button_num;

		switch(button_num)
		{
		  case Button3:
		    sb->y_win_pos = MAX(
			(sb->y_win_pos - sb->y_page_inc),
			y_min
		    );
		    SBarScrollRepeatStop(sb);
		    SBarAddScrollTimeout(
			sb, SB_OP_CODE_PAGE_UP,
			widget_global.sbar_scroll_repeat_delay
		    );
		    break;
		  case Button2:
		    sb->y_win_pos = y_min;
		    break;
		  default:
		    sb->y_win_pos = MAX(
			(sb->y_win_pos - sb->y_step_inc),
			y_min
		    );
		    SBarScrollRepeatStop(sb);
		    SBarAddScrollTimeout(
			sb, SB_OP_CODE_SCROLL_UP,
			widget_global.sbar_scroll_repeat_delay
		    );
		    break;
		}

		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* Y down button */
	    else if(w == sb->y_down_btn)
	    {
		const int	button_num = button->button,
				y_max = MAX(
		    ((int)sb->total_height - (int)sb->height),
		    0
		);

		sb->flags |= SCROLLBAR_FLAG_Y_DOWN_BTN_CLICKED;
		sb->button_pressed = button_num;

		switch(button_num)
		{
		  case Button3:
		    sb->y_win_pos = MIN(
			(sb->y_win_pos + sb->y_page_inc),
			y_max
		    );
		    SBarScrollRepeatStop(sb);
		    SBarAddScrollTimeout(
			sb, SB_OP_CODE_PAGE_DOWN,
			widget_global.sbar_scroll_repeat_delay
		    );
		    break;
		  case Button2:
		    sb->y_win_pos = y_max;
		    break;
		  default:
		    sb->y_win_pos = MIN(
			(sb->y_win_pos + sb->y_step_inc),
			y_max
		    );
		    SBarScrollRepeatStop(sb);
		    SBarAddScrollTimeout(
			sb, SB_OP_CODE_SCROLL_DOWN,
			widget_global.sbar_scroll_repeat_delay
		    );
		    break;
		}

		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* X left button */
	    else if(w == sb->x_left_btn)
	    {
		const int	button_num = button->button,
				x_min = 0;

		sb->flags |= SCROLLBAR_FLAG_X_LEFT_BTN_CLICKED;
		sb->button_pressed = button_num;

		switch(button_num)
		{
		  case Button3:
		    sb->x_win_pos = MAX(
			(sb->x_win_pos - sb->x_page_inc),
			x_min
		    );
		    SBarScrollRepeatStop(sb);
		    SBarAddScrollTimeout(
			sb, SB_OP_CODE_PAGE_LEFT,
			widget_global.sbar_scroll_repeat_delay
		    );
		    break;
		  case Button2:
		    sb->x_win_pos = x_min;
		    break;
		  default:
		    sb->x_win_pos = MAX(
			(sb->x_win_pos - sb->x_step_inc),
			x_min
		    );
		    SBarScrollRepeatStop(sb);
		    SBarAddScrollTimeout(
			sb, SB_OP_CODE_SCROLL_LEFT,
			widget_global.sbar_scroll_repeat_delay
		    );
		    break;
		}

		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* X right button */
	    else if(w == sb->x_right_btn)
	    {
		const int	button_num = button->button,
				x_max = MAX(
		    ((int)sb->total_width - (int)sb->width),
		    0
		);

		sb->flags |= SCROLLBAR_FLAG_X_RIGHT_BTN_CLICKED;
		sb->button_pressed = button_num;

		switch(button_num)
		{
		  case Button3:
		    sb->x_win_pos = MIN(
			(sb->x_win_pos + sb->x_page_inc),
			x_max
		    );
		    SBarScrollRepeatStop(sb);
		    SBarAddScrollTimeout(
			sb, SB_OP_CODE_PAGE_RIGHT,
			widget_global.sbar_scroll_repeat_delay
		    );
		    break;
		  case Button2:
		    sb->x_win_pos = x_max;
		    break;
		  default:
		    sb->x_win_pos = MIN(
			(sb->x_win_pos + sb->x_step_inc),
			x_max
		    );
		    SBarScrollRepeatStop(sb);
		    SBarAddScrollTimeout(
			sb, SB_OP_CODE_SCROLL_RIGHT,
			widget_global.sbar_scroll_repeat_delay
		    );
		    break;
 		}

		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;

	    if(sb->button_pressed > 0)
	    {
		sb->flags &= ~SCROLLBAR_FLAG_X_LEFT_BTN_CLICKED;
		sb->flags &= ~SCROLLBAR_FLAG_X_HANDLE_CLICKED;
		sb->flags &= ~SCROLLBAR_FLAG_X_RIGHT_BTN_CLICKED;
		sb->flags &= ~SCROLLBAR_FLAG_Y_UP_BTN_CLICKED;
		sb->flags &= ~SCROLLBAR_FLAG_Y_HANDLE_CLICKED;
		sb->flags &= ~SCROLLBAR_FLAG_Y_DOWN_BTN_CLICKED;
		sb->button_pressed = 0;
		SBarScrollRepeatStop(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    /* x bar */
	    if((w == sb->x_slider) && (sb->button_pressed > 0) &&
	       (sb->flags & SCROLLBAR_FLAG_X_HANDLE_CLICKED)
	    )
	    {
		const int	x = (int)motion->x,
				x_min = 0,
				x_max = MAX(
				    ((int)sb->total_width - (int)sb->width),
				    0
				);
		unsigned int wwidth, wheight;

		OSWGetWindowGeometry(
		    sb->x_slider,
		    NULL, NULL, &wwidth, &wheight, NULL
		);

		if((wwidth > 0) && (sb->total_width > 0))
		    sb->x_win_pos = FTOI_RNDOFF(
			(float)x / (float)wwidth * (float)sb->total_width
		    );
		else
		    sb->x_win_pos = 0;

		sb->x_win_pos -= FTOI_RNDOFF(
		    ((float)sb->width / 2) - (float)sb->x_origin_delta
		);

		if(sb->x_win_pos > x_max)
		    sb->x_win_pos = x_max;
		if(sb->x_win_pos < x_min)
		    sb->x_win_pos = x_min;

		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    /* y bar */
	    else if((w == sb->y_slider) && (sb->button_pressed > 0) &&
	            (sb->flags & SCROLLBAR_FLAG_Y_HANDLE_CLICKED)
	    )
	    {
		const int	y = (int)motion->y,
				y_min = 0,
				y_max = MAX(
				    ((int)sb->total_height - (int)sb->height),
				    0
				);
		unsigned int wwidth, wheight;

		OSWGetWindowGeometry(
		    sb->y_slider,
		    NULL, NULL, &wwidth, &wheight, NULL
		);

		if((wheight > 0) && (sb->total_height > 0))
		    sb->y_win_pos = FTOI_RNDOFF(
			(float)y / (float)wheight * (float)sb->total_height
		    );
		else
		    sb->y_win_pos = 0;

		sb->y_win_pos -= FTOI_RNDOFF(
		    ((float)sb->height / 2) - (float)sb->y_origin_delta
		);

		if(sb->y_win_pos > y_max)
		    sb->y_win_pos = y_max;
		if(sb->y_win_pos < y_min)
		    sb->y_win_pos = y_min;

		REPORT_VALUE_CHANGED(sb);
		SBarQueueDraw(sb);
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    visibility = &event->xvisibility;
	    if((w == sb->x_toplevel) || (w == sb->y_toplevel))
	    {
		sb->visibility_state = visibility->state;
		events_handled++;
	    }
	    break;
	}

#undef REPORT_VALUE_CHANGED

	return(events_handled);
}

/*
 *	Destroys the scroll bar.
 */
void SBarDestroy(scroll_bar_struct *sb)
{
	if(sb == NULL)
	    return;

	/* Stop any scroll repeats */
	SBarScrollRepeatStop(sb);

	OSWDestroyWindow(&sb->x_left_btn);
	OSWDestroyWindow(&sb->x_slider);
	OSWDestroyWindow(&sb->x_right_btn);
	OSWDestroyWindow(&sb->x_toplevel);

	OSWDestroyWindow(&sb->y_up_btn);
	OSWDestroyWindow(&sb->y_slider);
	OSWDestroyWindow(&sb->y_down_btn);
	OSWDestroyWindow(&sb->y_toplevel);

	OSWDestroyPixmap(&sb->x_toplevel_buf);
	OSWDestroyPixmap(&sb->y_toplevel_buf);

	sb->type = WIDGET_TYPE_UNKNOWN;
	sb->map_state = False;
	sb->has_focus = False;
	sb->disabled = False;
	sb->visibility_state = VisibilityFullyObscured;
	sb->x = 0;
	sb->y = 0;
	sb->width = 0;
	sb->height = 0;

	sb->flags = 0;

	sb->button_pressed = 0;
	sb->button_last_x = 0;
	sb->button_last_y = 0;
	sb->button_pressed_time = 0l;

	sb->total_width = 0;
	sb->total_height = 0;

	sb->x_origin_delta = 0;
	sb->y_origin_delta = 0;

	sb->x_win_pos = 0;
	sb->y_win_pos = 0;

	sb->x_step_inc = 0;
	sb->x_page_inc = 0;
	sb->y_step_inc = 0;
	sb->y_page_inc = 0;

	sb->value_changed_cb = NULL;
	sb->value_changed_data = NULL;
}
