#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/widget.h"


static void TooltipsResizeCB(void *widget, void *data);
static void TooltipsDrawCB(void *widget, void *data);
static void TooltipsMapTimeoutCB(void *widget, void *data);

static tooltips_data_struct *TooltipsDataNew(void);
static void TooltipsDataDelete(tooltips_data_struct *d);

int TooltipsInit(tooltips_struct *tt);
void TooltipsResize(tooltips_struct *tt);
void TooltipsQueueResize(tooltips_struct *tt);
void TooltipsDraw(tooltips_struct *tt);
void TooltipsQueueDraw(tooltips_struct *tt);
int TooltipsManage(tooltips_struct *tt, event_t *event);
void TooltipsMap(tooltips_struct *tt);
void TooltipsUnmap(tooltips_struct *tt);
void TooltipsDestroy(tooltips_struct *tt);

static void TooltipsSetMessageLines(
	tooltips_data_struct *d, const char *msg
);
void TooltipsSetMessage(
	tooltips_struct *tt, win_t w, const char *msg
);
void TooltipsDeleteMessage(tooltips_struct *tt, win_t w);
void TooltipsDeleteAllMessages(tooltips_struct *tt);

void TooltipsSchedualMap(
	tooltips_struct *tt, time_t interval_ms, win_t w
);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Resize callback.
 */
static void TooltipsResizeCB(void *widget, void *data)
{
	tooltips_struct *tt = TOOLTIPS(data);
	if(tt == NULL)
	    return;

	TooltipsResize(tt);
}

/*
 *	Draw callback.
 */
static void TooltipsDrawCB(void *widget, void *data)
{
	tooltips_struct *tt = TOOLTIPS(data);
	if(tt == NULL)
	    return;

	if(tt->map_state)
	    TooltipsDraw(tt);
}

/*
 *	Map timeout callback.
 *
 *	Called when the Tooltips has been schedualed to be mapped
 *	and it is time to be mapped.
 */
static void TooltipsMapTimeoutCB(void *widget, void *data)
{
	tooltips_struct *tt = TOOLTIPS(widget);
	if(tt == NULL)
	    return;

	/* Double check to make sure that the Tooltips is not
	 * mapped, its window has been created, and that it has
	 * actually been schedualed to be mapped
	 */
	if(!tt->map_state && (tt->toplevel != None) &&
	   (tt->next_map > 0l)
	)
	{
	    TooltipsMap(tt);
	    tt->next_map = 0l;
	}
}


/*
 *	Creates a new Tooltips data.
 */
static tooltips_data_struct *TooltipsDataNew(void)
{
	tooltips_data_struct *d = TOOLTIPS_DATA(calloc(
	    1, sizeof(tooltips_data_struct)
	));
	if(d == NULL)
	    return(NULL);

	d->ref_win = None;
	d->message = NULL;
	d->nlines = 0;

	return(d);
}

/*
 *	Deletes the Tooltips data.
 */
static void TooltipsDataDelete(tooltips_data_struct *d)
{
	if(d == NULL)
	    return;

	strlistfree(d->message, d->nlines);
	free(d);
}


/*
 *	Initializes the Tooltips.
 */
int TooltipsInit(tooltips_struct *tt)
{
	osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL) 
	    tt = wglobal->tooltips; 
	if(tt == NULL)
	    return(-2);

	tt->type = WIDGET_TYPE_TOOLTIPS;
	tt->map_state = False;
	tt->has_focus = False;
	tt->disabled = False;
	tt->visibility_state = VisibilityFullyObscured;
	tt->x = 0;
	tt->y = 0;
	tt->width = 80;
	tt->height = 30;
	tt->font = wglobal->std_font;
	tt->prev = NULL;
        tt->next = NULL;
	tt->data = NULL;

	tt->next_map = 0l;
	tt->ref_win = None;

	tt->hint_data = NULL;
	tt->ndatas = 0;
	tt->ref_data_num = -1;

	/* Create toplevel */
	OSWCreateWindow(
	    &w, parent,
	    tt->x, tt->y,
	    tt->width, tt->height
	);
	tt->toplevel = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask
	);

	/* WM Properties */
	OSWSetWindowWMProperties(
	    w,
	    "Tooltips",
	    "Tooltips",
	    wglobal->std_icon_pm,
	    False,
	    tt->x, tt->y,
	    tt->width, tt->height,
	    tt->width, tt->height,
	    OSW_WINDOW_FRAME_NONE,
	    0, NULL
	);

        /* Add widget to regeristry */
        WidgetRegister(tt, WIDGET_TYPE_TOOLTIPS);

	return(0);
}

/*
 *	Resizes the Tooltips.
 */
void TooltipsResize(tooltips_struct *tt)
{
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

}

/*
 *	Queues the Tooltips to resize.
 */
void TooltipsQueueResize(tooltips_struct *tt)
{
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

	WidgetQueueResizeAction(TooltipsResizeCB, tt, NULL);
}

/*
 *	Redraws the Tooltips.
 */
void TooltipsDraw(tooltips_struct *tt)
{
	int	i, x, y,
		lbearing, rbearing, swidth, ascent, descent,
		font_height,
		nlines,
		data_num, x_margin = 5, y_margin = 3;
	unsigned int width, height;
	char *s, **msg;
	font_t *font;
	pixmap_t pixmap;
	win_t w;
	tooltips_data_struct *tt_data;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global; 

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

	/* Get the hint data referenced by the Tooltips  */
	data_num = tt->ref_data_num;
	tt_data = ((data_num >= 0) && (data_num < tt->ndatas)) ?
	    tt->hint_data[data_num] : NULL;
	if(tt_data == NULL)
	    return;

	msg = tt_data->message;
	nlines = tt_data->nlines;
	if((msg == NULL) || (nlines <= 0))
	    return;

	/* Calculate size */
	font = tt->font;
	width = 0;
	height = 0;
	for(i = 0; i < nlines; i++)
	{
	    s = msg[i];
	    if(s == NULL)
		continue;

	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;

	    if(swidth > (int)width)
		width = (unsigned int)swidth;

	    height += font_height;
	}
	width += (unsigned int)(2 * x_margin);
	height += (unsigned int)(2 * y_margin);

	/* Map as needed */
	if(!tt->map_state)
	{
	    int rx, ry;
	    unsigned int rwidth, rheight;

	    OSWGetWindowGeometry(
		tt_data->ref_win, NULL, NULL, &rwidth, &rheight, NULL
	    );
	    OSWGetPointerCoords(
		tt_data->ref_win, &rx, &ry, NULL, NULL
	    );
	    x = rx - ((int)width / 2);
	    if((x + (int)width) >= (int)gui->display_width)
		x = (int)gui->display_width - (int)width;
	    if(x < 0)
		x = 0;
	    y = ry + 20;
	    if((y + (int)height) >= (int)gui->display_height)
		y = (int)gui->display_height - (int)height;
	    if(y < 0)
		y = 0;

	    w = tt->toplevel;
	    OSWMoveResizeWindow(w, x, y, width, height);
	    tt->x = x;
	    tt->y = y;
	    tt->width = width;
	    tt->height = height;

	    /* Destroy buffer */
	    OSWDestroyPixmap(&tt->toplevel_buf);

	    /* Map window */
	    OSWMapRaised(w);

	    tt->map_state = True;
	    tt->visibility_state = VisibilityUnobscured;
	}

	OSWSetFont(tt->font);

	w = tt->toplevel;
	
	/* Recreate buffer as needed */
	if(tt->toplevel_buf == None)
	    OSWCreatePixmap(&tt->toplevel_buf, width, height);
	pixmap = tt->toplevel_buf;

	/* Redraw background */
	OSWClearPixmap(
	    pixmap, width, height,
	    wglobal->force_mono ?
		gui->white_pix : wglobal->hint_bkg_pix
	);

	if(wglobal->force_mono)
	    OSWSetForeground(gui->black_pix);
	else
	    OSWSetForeground(wglobal->hint_text_pix);

	/* Draw message */
	x = x_margin;
	y = y_margin;
	for(i = 0; i < nlines; i++)
	{
	    s = msg[i];
	    if(s == NULL)
		continue;

	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, NULL, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWDrawString(
		pixmap,
		x - lbearing,
		y + ascent,
		s
	    );
	    y += font_height;
	}
	OSWDrawRectangle(
	    pixmap, 0, 0, (int)width - 1, (int)height - 1
	);
	OSWPutDrawableToWindow(w, pixmap);
}

/*
 *	Queues the Tooltips to draw.
 */
void TooltipsQueueDraw(tooltips_struct *tt)
{
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

	WidgetQueueDrawAction(TooltipsDrawCB, tt, NULL);
}

/*
 *	Manages the Tooltips event.
 */
int TooltipsManage(tooltips_struct *tt, event_t *event)
{
	int events_handled = 0;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if((tt == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == tt->toplevel)
	    {
		TooltipsQueueResize(tt);
		if(tt->map_state)
		    TooltipsQueueDraw(tt);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == tt->toplevel)
	    {
		TooltipsQueueDraw(tt);
		events_handled++;
	    }
	    break;
	}

	return(events_handled);
}

/*
 *	Maps the Tooltips.
 */
void TooltipsMap(tooltips_struct *tt)
{
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

	tt->map_state = False;
	TooltipsDraw(tt);
}

/*
 *	Unmaps the Tooltips.
 */
void TooltipsUnmap(tooltips_struct *tt)
{
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

	WidgetTimeoutRemoveByFuncWidget(
	    TooltipsMapTimeoutCB, tt
	);
	tt->next_map = 0l;
	tt->ref_win = None;
	tt->ref_data_num = -1;

	OSWUnmapWindow(tt->toplevel);

	tt->map_state = False;
	tt->has_focus = False;
	tt->visibility_state = VisibilityFullyObscured;

	/* Destroy buffers */
	OSWDestroyPixmap(&tt->toplevel_buf);
}

/*
 *	Destroys the Tooltips.
 */
void TooltipsDestroy(tooltips_struct *tt)
{
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

	WidgetTimeoutRemoveByFuncWidget(
	    TooltipsMapTimeoutCB, tt
	);
	tt->next_map = 0l;
	tt->ref_win = None;
	tt->ref_data_num = -1;

	TooltipsDeleteAllMessages(tt);

	OSWDestroyWindow(&tt->toplevel);
	OSWDestroyPixmap(&tt->toplevel_buf);

	tt->type = WIDGET_TYPE_UNKNOWN;
	tt->map_state = False;
	tt->has_focus = False;
	tt->disabled = False;
	tt->visibility_state = VisibilityFullyObscured;
	tt->x = 0;
	tt->y = 0;
	tt->width = 0;
	tt->height = 0;
}


/*
 *	Sets the message as parsed lines on the Tooltips data.
 */
static void TooltipsSetMessageLines(
	tooltips_data_struct *d, const char *msg
)
{
	const int cutoff_len = 60;
	int len, line_len;
	const char *s;

	if((d == NULL) || (msg == NULL))
	    return;

	/* Clear existing message */
	strlistfree(d->message, d->nlines);
	d->message = NULL;
	d->nlines = 0;

	/* Set message */
	s = msg;
	while(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    len = STRLEN(s);
	    if(len > cutoff_len)
	    {
		char *buf;

		for(line_len = cutoff_len - 1; line_len > 0; line_len--)
		{
		    if(ISBLANK(s[line_len]))
		    {
			for(line_len--; line_len > 0; line_len--)
			{
			    if(!ISBLANK(s[line_len]))
			    {
				line_len++;
				break;
			    }
			}
			break;
		    }
		}
		if(line_len <= 0)
		    line_len = cutoff_len;

		buf = (char *)malloc(line_len + 1);
		memcpy(buf, s, line_len);
		buf[line_len] = '\0';
		d->message = strlistappend(
		    d->message, &d->nlines, buf
		);
		free(buf);
		s += line_len;
	    }
	    else
	    {
		d->message = strlistappend(
		    d->message, &d->nlines, s
		);
		s = NULL;
	    }
	}
}

/*
 *	Sets the hint message for the specified window.
 *
 *	If msg is NULL then the specified window's current hint
 *	message will be unset.
 */
void TooltipsSetMessage(
	tooltips_struct *tt, win_t w, const char *msg
)
{
	int i;
	tooltips_data_struct *d = static_cast<tooltips_data_struct*>(NULL);
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if((tt == NULL) || (w == None))
	    return;

	/* Unset message? */
	if(STRISEMPTY(msg))
	{
	    TooltipsDeleteMessage(tt, w);
	    return;
	}

	/* Check if the window's hint message has already been set */
	for(i = 0; i < tt->ndatas; i++)
	{
	    d = tt->hint_data[i];
	    if(d == NULL)
		continue;

	    if(d->ref_win == w)
		break;
	}
	if((i < tt->ndatas) && (d != NULL))
	{
	    d->ref_win = w;
	    TooltipsSetMessageLines(d, msg);
	}
	else
	{
	    i = MAX(tt->ndatas, 0);
	    tt->ndatas = i + 1;

	    tt->hint_data = (tooltips_data_struct **)realloc(
		tt->hint_data,
		tt->ndatas * sizeof(tooltips_data_struct *)
	    );
	    if(tt->hint_data == NULL)
	    {
		tt->ndatas = 0;
		return;
	    }

	    tt->hint_data[i] = d = TooltipsDataNew();
	    if(d != NULL)
	    {
		d->ref_win = w;
		TooltipsSetMessageLines(d, msg);
	    }
	}
}

/*
 *	Deletes the hint message for the specified window.
 */
void TooltipsDeleteMessage(tooltips_struct *tt, win_t w)
{
	int i, n, ndeleted = 0;
	tooltips_data_struct *d;
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if((tt == NULL) || (w == None))
	    return;

	for(i = 0; i < tt->ndatas; i++)
	{
	    d = tt->hint_data[i];
	    if(d == NULL)
		continue;

	    if(d->ref_win == w)
	    {
		TooltipsDataDelete(d);
		tt->ndatas--;
		for(n = i; n < tt->ndatas; n++)
		    tt->hint_data[n] = tt->hint_data[n + 1];
		i--;
		ndeleted++;
	    }
	}

	if(ndeleted > 0)
	{
	    if(tt->ndatas > 0)
	    {
		tt->hint_data = (tooltips_data_struct **)realloc(
		    tt->hint_data,
		    tt->ndatas * sizeof(tooltips_data_struct *)
		);
		if(tt->hint_data == NULL)
		    tt->ndatas = 0;
	    }
	    else
	    {
		free(tt->hint_data);
		tt->hint_data = NULL;
		tt->ndatas = 0;
	    }
	}
}

/*
 *	Deletes all Tooltips messages.
 */
void TooltipsDeleteAllMessages(tooltips_struct *tt)
{
	int i;
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

	for(i = 0; i < tt->ndatas; i++)
	    TooltipsDataDelete(tt->hint_data[i]);
	free(tt->hint_data);
	tt->hint_data = NULL;
	tt->ndatas = 0;
	tt->ref_data_num = -1;
}


/*
 *	Scheduals the Tooltips to map.
 *
 *	If hw is NULL then the global Tooltips will be used.
 *
 *	The interval_ms is in milliseconds, if it is (time_t)-1 then
 *	the default interval will be used.
 *
 *	Any previously schedualed mappings will be canceled by this
 *	call.
 */
void TooltipsSchedualMap(
	tooltips_struct *tt, time_t interval_ms, win_t w
)
{
	int i;
	tooltips_data_struct *d;
	widget_global_struct *wglobal = &widget_global;

	if(tt == NULL)
	    tt = wglobal->tooltips;
	if(tt == NULL)
	    return;

	if((tt->toplevel == None) || (w == None))
	{
	    TooltipsUnmap(tt);
	    return;
	}

	if(interval_ms == (time_t)-1)
	    interval_ms = wglobal->tooltips_map_int;
	if(interval_ms < 0l)
	    interval_ms = 0l;

	tt->next_map = OSWTimeMS() + interval_ms;
	tt->ref_win = w;
	for(i = 0; i < tt->ndatas; i++)
	{
	    d = tt->hint_data[i];
	    if(d == NULL)
		continue;

	    if(d->ref_win == w)
	    {
		tt->ref_data_num = i;
		break;
	    }
	}

	WidgetTimeoutRemoveByFuncWidget(    
	    TooltipsMapTimeoutCB, tt  
	);
	WidgetTimeoutAdd(
	    TooltipsMapTimeoutCB,
	    tt,
	    NULL,
	    interval_ms
	);
}
