#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <fcntl.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/netio.h"

#include "siteban.h"
#include "sysparm.h"
#include "swserv.h"
#include "../include/auxstatcodes.h"


static void AUXCloseSocket(const int n, const int s);

int AUXConIsAllocated(const int n);

int AUXConNew(
        const time_t cur_ms, const time_t cur_sec, const float tc
);

int AUXConSend(
	int n, sws_aux_connection_struct *aux_con,
	const char *sndbuf
);
void AUXConSendMessage(
	int n, sws_aux_connection_struct *aux_con,
	const char *msg
);
static void AUXConSendStats(
        const int n, sws_aux_connection_struct *aux_con,
        const time_t cur_ms, const time_t cur_sec, const float tc
);

static void AUXHandleLogin(
	int n, sws_aux_connection_struct *aux_con,
	const char *arg
);
static void AUXHandleLogout(
	int n, sws_aux_connection_struct *aux_con,
	const char *arg                           
);
static int AUXConHandleNetworkStatement(
	int n, sws_aux_connection_struct *aux_con,
	const char *buf
);
static void AUXConHandleRecv(
	const int n, sws_aux_connection_struct *aux_con,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void AUXConManageAll(
        const time_t cur_ms, const time_t cur_sec, const float tc
);

static void AUXConLogoutConnection(
	int n, sws_aux_connection_struct *aux_con 
);
void AUXConShutdownConnection(
	int n, int how, char is_initiator
);
void AUXConCloseConnection(int n);
void AUXConDelete(
	const int n,
        const time_t cur_ms, const time_t cur_sec, const float tc  
);
void AUXConDeleteAll(
        const time_t cur_ms, const time_t cur_sec, const float tc  
);

void AUXConManageNewConnections(
        const int socket,
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *      Closes the socket.
 *
 *      Will wait to ensure that the socket is safely closed and report
 *      any errors involving the closing of the socket.
 */
static void AUXCloseSocket(const int n, const int s)
{
	int error_code;
	const time_t start_time = time(NULL);

	if(s < 0)
	    return;

	while(close(s) != 0)
	{
	    error_code = errno;

	    /* Give up if more than 30 seconds elapsed */
	    if(time(NULL) > (start_time + 30l))
	    {
		if(sysparm.log_errors)
		{
		    char *msg = (char *)malloc(40 + 80 + 80 + 80);
		    if(n > -1)
			sprintf(
			    msg,
"AUX connection %i:\
 Error closing socket %i:\
 Timmed out (waited more than %ld seconds).",
			    n,
			    s,
			    30l
			);
		    else
			sprintf(
			    msg,
"New AUX connection:\
 Error closing socket %i:\
 Timmed out (waited more than %ld seconds).",
			    s,
			    30l 
			);
		    LogAppend(fname.log, msg);
		    free(msg);
		}
		break;
	    }

	    if(error_code == EAGAIN)
	    {
		sleep(1);
		continue; 
	    }
	    else
	    {
		char *msg = STRDUP(strerror(error_code));
		if(msg != NULL)
		{
		    *msg = toupper(*msg);
		    if(n > -1)
			fprintf(
			    stderr,
"AUXCloseSocket(): AUX connection %i: Error closing socket %i: %s.",
			    n,
			    s,
			    msg
			);
		    else
			fprintf(
			    stderr,
"AUXCloseSocket(): New AUX connection: Error closing socket %i: %s.",
			    s,
			    msg 
			);

		    if(sysparm.log_errors)
		    {
			char *buf = (char *)malloc(STRLEN(msg) + 40 + 80);
			if(n > -1)
			    sprintf(
				buf,
"AUX connection %i: Error closing socket %i: %s.",
				n,
				s,
				msg
			    );
			else
			    sprintf(
				buf,
"New AUX connection: Error closing socket %i: %s.",
				s,
				msg
			    );
			LogAppend(fname.log, buf);
			free(buf);
		    }

		    free(msg);
		}
		break;
	    }
	}
}


/*
 *	Checks if the AUX connection is allocated.
 */
int AUXConIsAllocated(const int n)
{
	if((aux_connection == NULL) ||
	   (n < 0) ||
	   (n >= total_aux_connections)
	)
	    return(0);
	else if(aux_connection[n] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Creates a new AUX connection from the specified socket.
 *
 *	Returns the new AUX connection's index or -1 on error.
 */
int AUXConNew(
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, n;
	sws_aux_connection_struct *aux_con;

	/* Sanitize total */
	if(total_aux_connections < 0)
	    total_aux_connections = 0;

	/* Search for an available pointer */
	for(i = 0; i < total_aux_connections; i++)
	{
	    aux_con = aux_connection[i];
	    if(aux_con == NULL)
		break;

	    if(aux_con->socket < 0)
		break;
	}
	if(i < total_aux_connections)
	{
	    /* Got available pointer */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_aux_connections;
	    total_aux_connections = n + 1;
	    aux_connection = (sws_aux_connection_struct **)realloc(
		aux_connection,
		total_aux_connections * sizeof(sws_aux_connection_struct *)
	    );
	    if(aux_connection == NULL)
	    {
		total_aux_connections = 0;
		return(-3);
	    }

	    /* Reset the new pointer to NULL so that it gets allocated
	     * below
	     */
	    aux_connection[n] = NULL;
	}

	/* Allocate as needed */
	if(aux_connection[n] == NULL)
	{
	    aux_connection[n] = aux_con = (sws_aux_connection_struct *)calloc(
		1, sizeof(sws_aux_connection_struct)
	    );
	    if(aux_con == NULL)
		return(-3);
	}

	aux_con = aux_connection[n];

	aux_con->socket = -1;
	aux_con->flags = 0;
	aux_con->format = SWS_AUX_CON_FORMAT_STANDARD;

	aux_con->conhost = NULL;
	aux_con->contime = cur_sec;
	aux_con->object_num = -1;

	aux_con->next = cur_ms;

	aux_con->bytes_received = 0l;
	aux_con->bytes_sent = 0l;
	aux_con->errors_received = 0l;
	aux_con->errors_sent = 0l;


	return(n);
}

/*
 *	Sends the data to the AUX connection.
 *
 *	Returns:
 *
 *	0	Data sent successfully.
 *	-1	General error.
 *	-2	Bad value.
 *	-3	Systems error.
 *	-4	Interrupted by signal.
 *	-5	Send failed and the connection has been closed by this
 *		function.
 */
int AUXConSend(
	int n, sws_aux_connection_struct *aux_con,
	const char *sndbuf
)
{
	int len, socket, bytes_written;

	if((aux_con == NULL) || (sndbuf == NULL))
	    return(-2);

	socket = aux_con->socket;
	if((socket < 0) || (aux_con->flags & SWS_AUX_CON_FLAG_CLOSE_WAIT))
	    return(-1);

	len = strlen(sndbuf);
	bytes_written = NetSend(socket, sndbuf, len);
	if(bytes_written == 0)
	{
	    return(0);
	}
	else if(bytes_written < 0)
	{
	    const int error_code = errno;
	    switch(error_code)
	    {
	      case EAGAIN:
#if (EWOULDBLOCK != EAGAIN)
	      case EWOULDBLOCK:
#endif
              case ENOBUFS:	/* Out of memory, try again */
		return(0);
		break;

              case EPIPE:	/* Connection has been shut down */
                aux_con->flags |= SWS_AUX_CON_FLAG_CLOSE_WAIT;
                AUXConCloseConnection(n);
                return(-5);
                break;

	      case EBADF:	/* Invalid descriptor */
	      case ENOTSOCK:  	/* Descriptor is not a socket */
		aux_con->errors_sent++;
		AUXConCloseConnection(n);
		return(-5);
		break;

              case EINTR:	/* Signal occured */
		aux_con->errors_sent++;
		return(-4);
		break;      

	      case ENOMEM:	/* Out of memory */
		aux_con->errors_sent++;
		return(-3);
		break;

	      default:
		aux_con->errors_sent++;
		return(-1);
		break;
	    }
	}
	else
	{
	    aux_con->bytes_sent += (unsigned long)bytes_written;
	}

	return(0);
}

/*
 *	Sends a message.
 */
void AUXConSendMessage(
	int n, sws_aux_connection_struct *aux_con,
	const char *msg
)
{
	char *sndbuf;

	if((aux_con == NULL) || STRISEMPTY(msg))
	    return;

	sndbuf = (char *)malloc(
	    STRLEN(STAT_PFX_MESSAGE) +
	    STRLEN(msg) +
	    80
	);
	sprintf(
	    sndbuf,
"%s: %s\n",
	    STAT_PFX_MESSAGE,
	    msg
	);
	AUXConSend(n, aux_con, sndbuf);
	free(sndbuf);
}

/*
 *	Ssend all server stats to all AUX pipes.
 */
static void AUXConSendStats(
	const int n, sws_aux_connection_struct *aux_con,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	const time_t start_time = sws_start_time;
	int i, total_con, total_guest_con;
	char *sndbuf;
	sws_memory_stats_struct m;
	time_t uptime;
	sw_legend_struct *legend = &sw_legend;
	sws_connection_struct *con;

	/* Calculate connections */
	total_con = 0;
	total_guest_con = 0;
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if((con->socket < 0) ||
	       (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
	    )
		continue;

	    if(con->flags & SWS_CON_FLAG_IS_GUEST)
	        total_guest_con++;

	    total_con++;
	}

	/* Get memory stats */
	SWSGetMemoryStats(&m);

	/* Calculate the uptime */
	uptime = ((start_time > 0l) && (start_time <= cur_sec)) ?
	    (cur_sec - start_time) : 0l;

	/* Format buffer */
	sndbuf = (char *)malloc(
	    STRLEN(STAT_PFX_CONNECTIONS) + 40 + 40 +
	    STRLEN(STAT_PFX_MEMORY) + 40 + 40 + 40 +
	    STRLEN(STAT_PFX_NEXT) + 40 + 40 +
	    STRLEN(STAT_PFX_TITLE) + STRLEN(legend->title) +
	    STRLEN(STAT_PFX_OBJECTS) + 40 +
	    STRLEN(STAT_PFX_PID) + 40 +
	    STRLEN(STAT_PFX_UPTIME) + 40 +
	    80
	);
	sprintf(
	    sndbuf,
"%s: %i %i\n\
%s: %ld %ld %ld\n\
%s: %ld %ld\n\
%s: %s\n\
%s: %i\n\
%s: %i\n\
%s: %ld\n",
	    STAT_PFX_CONNECTIONS, total_con, total_guest_con,
	    STAT_PFX_MEMORY, m.total, m.con, m.obj,
	    STAT_PFX_NEXT, (next.unv_save - cur_sec),
		(next.stats_export - cur_sec),
	    STAT_PFX_TITLE, legend->title,
	    STAT_PFX_OBJECTS, total_objects,
	    STAT_PFX_PID, getpid(),
	    STAT_PFX_UPTIME, uptime
	);
	AUXConSend(n, aux_con, sndbuf);
	free(sndbuf);
}


/*
 *	Handles a login.
 */
static void AUXHandleLogin(
	int n, sws_aux_connection_struct *aux_con,
	const char *arg
)
{
	int len, obj_num;  
	char	*name = NULL,
		*password = NULL;
	const char *s, *s_end;
	sw_legend_struct *legend = &sw_legend;
	xsw_object_struct *obj;

	/* Format is:
	 *
	 *	"name;password"
	 */
	s = arg;

	/* Get name */
	s_end = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	if(s_end != NULL)
	{
	    len = s_end - s;
	    if(len > 0)
	    {
		name = (char *)malloc((len + 1) * sizeof(char));
		memcpy(name, s, len);
		name[len] = '\0';
	    }
	}
	else
	{
	    name = strdup(s);
	}
	if(s_end != NULL)
	    s = s_end + 1;

	/* Get password */
	s_end = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	if(s_end != NULL) 
	{
	    len = s_end - s; 
	    if(len > 0)
	    {
		password = (char *)malloc((len + 1) * sizeof(char));
		memcpy(password, s, len);
		password[len] = '\0';
	    }
	}
	else
	{
	    password = strdup(s);
	}

	strstrip(name);
	strstrip(password);

	if(name == NULL)
	    name = strdup("");
	if(password == NULL)
	    password = strdup("");

	/* Does name specify a guest login */
	if(sysparm.allow_guest &&
	   !strcasecmp(name, sysparm.guest_login_name)
	)
	{
	    /* Guest connection */
	    char *s, *s2;

	    /* Mark this as a guest connection */
	    aux_con->flags |= SWS_AUX_CON_FLAG_IS_GUEST;

            /* Print welcome message */
            s = STRDUP(sysparm.mesg_welcome);

            s2 = strsub(s, "%title", legend->title);
            free(s);   
            s = s2;

            s2 = strsub(s, "%name", sysparm.guest_login_name);
            free(s);
            s = s2; 

            AUXConSendMessage(n, aux_con, s);
            free(s);

            AUXConSendMessage(
                n, aux_con,
"Event and log messages will not be shown to non-player logins."
	    );
	}
	else
	{
	    /* Find the player object by the name specified by the
	     * login name
	     */
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects, name, SW_OBJ_TYPE_PLAYER
	    );
	    if(DBIsObjectGarbage(obj_num))
	    {
		/* No such player object */
		AUXConSendMessage(
		    n, aux_con, 
		    sysparm.mesg_wrong_login
		);
		free(name);
		free(password);
		return;
	    }
	    else
	    {
		/* Found player object (user login) */

		obj = xsw_object[obj_num];

		/* Check password */
		if(CryptHandleVerify(password, obj->password) < 1)
		{
		    /* Wrong password */
		    AUXConSendMessage(
			n, aux_con,
			sysparm.mesg_wrong_login
		    );
		    free(name);
		    free(password);
		    return;
		}
		else
		{
		    /* Password matched */
		    char *s, *s2;

		    /* Log in connection */
		    aux_con->object_num = obj_num;

		    /* Print welcome message */
		    s = STRDUP(sysparm.mesg_welcome);

		    s2 = strsub(s, "%title", legend->title);
		    free(s);
		    s = s2;

		    s2 = strsub(s, "%name", obj->name);
		    free(s);
		    s = s2;

		    AUXConSendMessage(n, aux_con, s);
		    free(s);

		    if(!sysparm.console_quiet)
		    {
			const char *name = DBGetFormalNameStr(obj_num);
			fprintf(
			    stdout,
"AUX connection %i: Successfully logged in as `%s'.\n",
			    n,
			    name
			);
		    }

		    if(sysparm.log_net)
		    {
			const char *name = DBGetFormalNameStr(obj_num);
			char *s = (char *)malloc(
			    40 + STRLEN(name) + 80
			);
			sprintf(
			    s,
"AUX connection %i: Successfully logged in as `%s'.",
			    n,
			    name
			);
			LogAppend(fname.log, s);
			free(s);
		    }
		}
	    }
	}

	free(name);
	free(password);
}

/*
 *	Handles a logout.
 */
static void AUXHandleLogout(
	int n, sws_aux_connection_struct *aux_con,
	const char *arg
)
{
	sw_legend_struct *legend = &sw_legend;
	int obj_num = aux_con->object_num;
	if(!DBIsObjectGarbage(obj_num))
	{
	    /* Send leave message to connection */
	    xsw_object_struct *obj = xsw_object[obj_num];
	    char	*s = STRDUP(sysparm.mesg_leave),
			*s2 = strsub(s, "%title", legend->title),
			*mesg;
	    free(s);
	    mesg = strsub(s2, "%name", obj->name);
	    free(s2);
	    AUXConSendMessage(n, aux_con, mesg);
	    free(mesg);
	}

	if(!(aux_con->flags & SWS_AUX_CON_FLAG_CLOSE_WAIT))
	    AUXConShutdownConnection(
		n,
		SWS_CON_SHUTDOWN_RDWR,	/* Shutdown both read and write */
		0			/* We are not the disconnect initiator */
	    );
}

/*
 *	Handles the network statement.
 *
 *	Returns the number of statements handled.
 */
static int AUXConHandleNetworkStatement(
	int n, sws_aux_connection_struct *aux_con,
	const char *buf
)
{
	int len, statements_handled = 0;
	const char *s, *s_end;
	char	*parm = NULL,
		*val = NULL;

	if(buf == NULL)
	    return(statements_handled);

	/* Format is:
	 *
	 *      "parm:val"
	 */
	s = buf;

	/* Get parameter */
	s_end = strchr(s, ':');
	if(s_end != NULL)
	{
	    len = s_end - s;
	    if(len > 0)
	    {
		parm = (char *)malloc((len + 1) * sizeof(char));
		memcpy(parm, s, len);
		parm[len] = '\0';
	    }
	}
	else
	{
	    parm = strdup(s);
	}
	if(s_end != NULL)
	    s = s_end + 1;

	/* Get value */
	s_end = strchr(s, ':');
	if(s_end != NULL)
	{
	    len = s_end - s;
	    if(len > 0)
	    {
		val = (char *)malloc((len + 1) * sizeof(char));
		memcpy(val, s, len);
		val[len] = '\0';
	    }
	}    
	else
	{   
	    val = strdup(s);
	}

	strstrip(parm);
	strstrip(val);

	if(parm == NULL)
	    parm = strdup("");
	if(val == NULL)   
	    val = strdup("");

	/* Handle network statement */
	/* Login */
	if(!strcasecmp(parm, STAT_PFX_LOGIN))
	{
	    AUXHandleLogin(n, aux_con, val);
	    statements_handled++;
	}
	/* Logout */
	else if(!strcasecmp(parm, STAT_PFX_LOGOUT))
	{
	    AUXHandleLogout(n, aux_con, val);
	    statements_handled++;
	}

	free(parm);
	free(val);

	return(statements_handled);
}

/*
 *	Handles incoming data on AUX connection aux_con.
 *
 *	Returns the number of statements received.
 */
static void AUXConHandleRecv(
        const int n, sws_aux_connection_struct *aux_con,  
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int bytes_read, statements_handled = 0;
	char buf[AUXSTAT_MAX_LEN];
	char *buf_ptr;
	int buf_cnt;  

	char workbuf[AUXSTAT_SEGMENT_MAX];
	char *workbuf_ptr;
	int workbuf_cnt;  

	/* Recieve incoming data */
	bytes_read = NetRecv(aux_con->socket, buf, sizeof(buf));
	if(bytes_read == 0)
	{
	    /* recv() == 0 means that the socket has been shutdown()
	     * and all data flushed, in which case the socket should
	     * now be closed
	     */
	    AUXConCloseConnection(n);
	    return;
	}

	/* Recieve error? */
	if(bytes_read < 0)
	{
	    const int error_code = errno;
	    char *error_msg;

	    /* Handle error */
	    switch(error_code)
	    {
	      case EAGAIN:	/* No data available */
#if (EWOULDBLOCK != EAGAIN)
	      case EWOULDBLOCK:
#endif
	      case EINTR:	/* Interrupted by signal */
		return;
		break;

	      case EPIPE:	/* Connection has been shut down */
		aux_con->flags |= SWS_AUX_CON_FLAG_CLOSE_WAIT;
		AUXConCloseConnection(n);
		return;
		break;

	      case EBADF:	/* Invalid descriptor */
	      case ECONNRESET:	/* Connection terminated by peer */
	      case EFAULT:	/* Receive buffer out of address space */
	      case ENOTCONN:	/* Descriptor is not connected */
	      case ENOTSOCK:	/* Descriptor is not a socket */
		error_msg = STRDUP(strerror(error_code));
		if((error_msg != NULL) && sysparm.log_errors)  
		{
		    char *s = (char *)malloc(
			40 + STRLEN(error_msg) + 128
		    );
		    *error_msg = toupper(*error_msg);
		    sprintf(
			s,  
"AUX connection %i: Receive error and broken connection: %s.",
			n,
			error_msg
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
		AUXConCloseConnection(n);
		free(error_msg);
		return;
		break;

	      default:
		error_msg = STRDUP(strerror(error_code));
		if((error_msg != NULL) && sysparm.log_errors)
		{
		    char *s = (char *)malloc(
			40 + STRLEN(error_msg) + 128
		    );
		    *error_msg = toupper(*error_msg);
		    sprintf(
			s,
"Connection %i: Receive error: %s.",
			n,
			error_msg
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
		free(error_msg);
		return;
		break;
	    }
	}

	/* Increment bytes received */
	aux_con->bytes_received += (unsigned long)bytes_read;

	/* Null terminate */
	if(bytes_read > AUXSTAT_MAX_LEN)
	    bytes_read = AUXSTAT_MAX_LEN;

	buf[bytes_read - 1] = '\0';


	/* Begin parsing and handling */
	buf_cnt = 0;
	buf_ptr = buf;

	while(buf_cnt < bytes_read)
	{
	    /* Reset work buffer count and pointer */
	    workbuf_cnt = 0;
	    workbuf_ptr = workbuf;

	    /* Begin fetching data from recvbuf to workbuf */
	    while(1)
	    {
		/* Work buffer overflowed? */
		if(workbuf_cnt >= AUXSTAT_SEGMENT_MAX)
		{
		    /* Increment buf_cnt to next delimiter char */
		    while(buf_cnt < bytes_read)
		    {      
			if((*buf_ptr == '\n') ||
			   (*buf_ptr == '\r') ||
			   (*buf_ptr == '\0')
			)
			{
			    buf_cnt++;
			    buf_ptr++;
			    break;
			}
			buf_cnt++;
			buf_ptr++;
		    }
		    /*   Null terminating character for workbuf will
		     *   be added farther below.
		     */
		    break;
		}

		/* End of buffer reached? */
		else if(buf_cnt >= bytes_read)
		{
		    *workbuf_ptr = '\0';

#if 0
		    /* Copy workbuf_cnt to net_parms.co_data for
		     * processing in the next call to this function
		     */
		    net_parms.co_data_len = MIN(
			workbuf_cnt,
			AUXSTAT_SEGMENT_MAX
		    );
		    
		    if(net_parms.co_data_len > 0)
			memcpy(
			    net_parms.co_data,  
			    workbuf,
			    net_parms.co_data_len
			);
#endif
		    /* Return, don't continue. This is so that the
		     * the carry over buffer is not reset and its
		     * values just set above get carryed over to
		     * the next call to this function.
		     */
		    return;

		    break;
		}

		/* Deliminator in recieve buffer encountered? */
		else if((*buf_ptr == '\n') ||
			(*buf_ptr == '\r') ||
			(*buf_ptr == '\0')
		)
		{
		    *workbuf_ptr = '\0';

		    buf_cnt++;
		    buf_ptr++;

		    break;
		}

		/* Copy data from buffer to work buffer normally */
		else
		{
		    *workbuf_ptr = *buf_ptr;

		    buf_cnt++;
		    buf_ptr++;

		    workbuf_cnt++;
		    workbuf_ptr++;
		}
	    }

	    /* Skip if workbuf is empty */
	    if(*workbuf == '\0')
		continue;

	    /* Set null terminating character for workbuf */
	    workbuf[AUXSTAT_SEGMENT_MAX - 1] = '\0';

	    /* Handle this segment */
	    statements_handled += AUXConHandleNetworkStatement(
		n, aux_con,
		workbuf
	    );
	}

	return;
}

/*
 *	Checks all AUX connections for incoming data, processes it,
 *	and sends out any pending data.
 */
void AUXConManageAll(
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i;
	sws_aux_connection_struct *aux_con;

	/* Receive */
	for(i = 0; i < total_aux_connections; i++)
	{
	    aux_con = aux_connection[i];
	    if(aux_con == NULL)
		continue;

	    if(aux_con->socket < 0)
		continue;

	    /* Handle by format */
	    switch(aux_con->format)
	    {
	      default:	/* AUX_CON_FORMAT_STANDARD */
		AUXConHandleRecv(
		    i, aux_con,
		    cur_ms, cur_sec, tc
		);
		break;
	    }
	}

	/* Send */
	for(i = 0; i < total_aux_connections; i++)
	{
	    aux_con = aux_connection[i];
	    if(aux_con == NULL)   
		continue;

	    if((aux_con->socket < 0) ||
	       (aux_con->flags & SWS_AUX_CON_FLAG_CLOSE_WAIT)
	    )
		continue;

	    /* Handle by format */
	    switch(aux_con->format)
	    {
	      default:	/* AUX_CON_FORMAT_STANDARD */
		if(aux_con->next <= cur_ms)
		{
		    AUXConSendStats(
			i, aux_con,
			cur_ms, cur_sec, tc
		    );
		    aux_con->next = cur_ms + sysparm.int_aux_con_stats;
		}
		break;
	    }
	}
}


/*
 *	Notifies and logs the logout of this AUX connection.
 */
static void AUXConLogoutConnection(
	int n, sws_aux_connection_struct *aux_con
)
{
	int obj_num = aux_con->object_num;

	/* Notify console of disconnect as needed */
	if(!sysparm.console_quiet && (aux_con->socket > -1))
	{
	    if(DBIsObjectGarbage(obj_num))
		fprintf(
		    stdout,
"AUX connection %i: Shutting down socket...\n",
		    n
		);
	    else
		fprintf(
		    stdout,
"AUX connection %i: %s logged out and shutting down socket...\n",
		    n,
		    DBGetFormalNameStr(obj_num)
		);
	}

	/* Log logout and shutting down of connection? */
	if(sysparm.log_net && (aux_con->socket > -1))
	{                                                   
	    char *s;
	    if(DBIsObjectGarbage(obj_num))
	    {
		s = (char *)malloc(40 + 80);
		sprintf(
		    s,  
"AUX connection %i: Shutting down socket...",
		    n
		);
	    }
	    else
	    {
		char *name = STRDUP(DBGetFormalNameStr(obj_num));
		s = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    s,
"AUX connection %i: %s logged out and shutting down socket...",
		    n,
		    name
		);
		free(name);
	    }
	    LogAppend(fname.log, s);
	    free(s);
	}

/* Send logout? */
}

/*
 *	Logs out the AUX connection and shuts down its socket.
 */
void AUXConShutdownConnection(
	int n, int how, char is_initiator
)
{
	int lhow;
	sws_aux_connection_struct *aux_con;

	if(!AUXConIsAllocated(n))
	    return;

	aux_con = aux_connection[n];

	if((aux_con->socket < 0) ||
	   (aux_con->flags & SWS_AUX_CON_FLAG_CLOSE_WAIT)
	)
	    return;

	/* Mark the connection's socket as CLOSE_WAIT, it is important
	 * to do this now before any subsequent calls or else they
	 * will attempt to call this function to shut down this
	 * connection and cause infinate recursion
	 */
	aux_con->flags |= SWS_AUX_CON_FLAG_CLOSE_WAIT;

	/* Shutdown the socket */
	switch(how)
	{
	  case SWS_CON_SHUTDOWN_WR:
	    lhow = SHUT_WR;
	    break;
	  case SWS_CON_SHUTDOWN_RDWR:
	    lhow = SHUT_RDWR;
	    break;
	  default:
	    lhow = SHUT_RD;
	    break;
	}
	if(shutdown(aux_con->socket, lhow))
	{
	    const int error_code = errno;
	    if(sysparm.log_errors)
	    {
		char *errmsg = STRDUP(strerror(error_code));
		if(errmsg != NULL)
		{
		    char *s = (char *)malloc(STRLEN(errmsg) + 40 + 80);
		    *errmsg = toupper(*errmsg);
		    sprintf(
			s,
"AUX connection %i: shutdown() error: %s.",
			n,
			errmsg
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
		free(errmsg);
	    }
	}

	/* Logout connection after marking it as CLOSE_WAIT so that
	 * data will not be sent to it
	 */
	AUXConLogoutConnection(n, aux_con);

	/* Schedual stats export to now in order to export the
	 * connections list now
	 */
	next.stats_export = cur_systime;
}

/*
 *	Closes the AUX connection.
 */
void AUXConCloseConnection(int n)
{
	sws_aux_connection_struct *aux_con;

	if(!AUXConIsAllocated(n))
	    return;

	aux_con = aux_connection[n];

	/* If the socket is still connected then forceably close it
	 * (without shutting it down first)   
	 */
	if(aux_con->socket > -1)
	{
	    if(!(aux_con->flags & SWS_AUX_CON_FLAG_CLOSE_WAIT))
		AUXConShutdownConnection(
		    n, SWS_CON_SHUTDOWN_RDWR, 1
		);
	    AUXCloseSocket(n, aux_con->socket);
	    aux_con->socket = -1;

	    /* Notify close of AUX connection */
	    if(!sysparm.console_quiet)
		fprintf(
		    stdout,
"AUX connection %i: Disconnected.\n",
		    n
		);

	    /* Log disconnect */
	    if(sysparm.log_net)
	    {
		char *s = (char *)malloc(40 + 80);
		sprintf(
		    s,  
"AUX connection %i: Disconnected.", 
		    n
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	}

	aux_con->flags = 0;
	free(aux_con->conhost);
	aux_con->conhost = NULL;
	aux_con->object_num = -1;
}

/*
 *	Closes and deletes the AUX connection.
 */
void AUXConDelete(
	const int n,
        const time_t cur_ms, const time_t cur_sec, const float tc  
)
{
	sws_aux_connection_struct *aux_con;

	if(!AUXConIsAllocated(n))
	    return;

	aux_con = aux_connection[n];

	if(aux_con->socket > -1)
	{
	    if(!(aux_con->flags & SWS_AUX_CON_FLAG_CLOSE_WAIT))
		AUXConShutdownConnection(
		    n, SWS_CON_SHUTDOWN_RDWR, 1
		);
	    AUXConCloseConnection(n);
	}

	free(aux_con->conhost);
	free(aux_con);
	aux_connection[n] = NULL;
}

/*
 *	Closes and deletes all AUX connections.
 */
void AUXConDeleteAll(
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i;

	for(i = 0; i < total_aux_connections; i++)
	    AUXConDelete(i, cur_ms, cur_sec, tc);
	free(aux_connection);
	aux_connection = NULL;
	total_aux_connections = 0;
}


/*
 *	Checks and handles any new AUX connections from the specified
 *	socket.
 */
void AUXConManageNewConnections(
	const int socket,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, n, new_con;
	char *sndbuf;
	sws_aux_connection_struct *aux_con;

	int sin_size;
	int new_socket;
	struct sockaddr_in foreign_addr;

	siteban_ip_union ip;


	/* Answer the new connection */
	sin_size = sizeof(struct sockaddr_in);
	new_socket = accept(
	    socket,
	    (struct sockaddr *)&foreign_addr,
	    reinterpret_cast<socklen_t *>(&sin_size)
	);
	if(new_socket == -1)
	    return;

	/* Set socket nonblocking */
	fcntl(new_socket, F_SETFL, O_NONBLOCK);


	/* Get IP (in nbo) */
	ip.whole = foreign_addr.sin_addr.s_addr;

	/* Is this IP banned? */
	if(SiteBanIsBanned(&ip, 0))
	{
	    sndbuf = (char *)malloc(
		STRLEN(STAT_PFX_MESSAGE) + 40 + 40 + 40 + 40 + 80
	    );
	    sprintf(
		sndbuf,
		"%s: Your address %i.%i.%i.%i has been banned.\n",
		STAT_PFX_MESSAGE,
		ip.part_u8[0],
		ip.part_u8[1],
		ip.part_u8[2],
		ip.part_u8[3]
	    );
	    NetSend(new_socket, sndbuf, strlen(sndbuf));
	    free(sndbuf);

	    /* Close the new socket */
	    AUXCloseSocket(-1, new_socket);

	    /* Pretend that we did not do anything */
	    return;
	} 


	/* Too many connections? */
	for(i = 0, n = 0; i < total_aux_connections; i++)
	{
	    if(aux_connection[i] == NULL)
		continue;

	    if(aux_connection[i]->socket < 0)
		continue;

	    n++;
	}
	if(n >= sysparm.max_aux_connections)
	{
	    sndbuf = (char *)malloc(
		STRLEN(STAT_PFX_MESSAGE) + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%s: Maximum of %i AUX connections reached.\n",
		STAT_PFX_MESSAGE,
		sysparm.max_aux_connections
	    );
	    NetSend(new_socket,	sndbuf,	strlen(sndbuf));
	    free(sndbuf);

	    /* Close the new socket */
	    AUXCloseSocket(-1, new_socket);

	    return;
	}


	/* Get a new connection descriptor for new connection */
	new_con = AUXConNew(cur_ms, cur_sec, tc);
	if(AUXConIsAllocated(new_con))
	{
	    aux_con = aux_connection[new_con];
	}
	else
	{
	    if(sysparm.log_errors)
		LogAppend(
		    fname.log,
"AUXConManageNewConnections(): Unable to create a new AUX connection."
		);

	    /* Close the new socket */
	    AUXCloseSocket(-1, new_socket);

	    return;
	}

	aux_con->socket = new_socket;
	free(aux_con->conhost);
	aux_con->conhost = STRDUP(inet_ntoa(foreign_addr.sin_addr));

	/* Notify of new AUX connection */
	if(!sysparm.console_quiet)
	    fprintf(
		stdout,
"AUX connection %i: Got connection from `%s'.\n",
		new_con,
		(aux_con->conhost != NULL) ?
		    aux_con->conhost : ""
	    );

	/* Log connect */
	if(sysparm.log_net)
	{
	    char *s = (char *)malloc(
		40 + STRLEN(aux_con->conhost) + 80
	    );
	    sprintf(
		s,
"AUX connection %i: Got connection from `%s'.",
		new_con,
		(aux_con->conhost != NULL) ?
		    aux_con->conhost : ""
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	sndbuf = (char *)malloc(
	    STRLEN(STAT_PFX_REQUESTLOGIN) + 80
	);
	sprintf(
	    sndbuf,
	    "%s:\n",
	    STAT_PFX_REQUESTLOGIN
	);
	AUXConSend(new_con, aux_con, sndbuf);
	free(sndbuf);
}
