#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/string.h"

#include "cmd.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


#ifdef PLUGIN_SUPPORT
#include "plugins.h"
#endif 


int CmdInput(
	const int con_num, const char *cmd,
	const time_t cur_ms, const time_t cur_sec,
	const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Command input front end.
 *
 *	This is the front end for handling the server command received
 *	from a connection.
 */
int CmdInput(
	const int con_num, const char *cmd,
	const time_t cur_ms, const time_t cur_sec,
	const float tc
)
{
	int status, con_obj_num, command_handled;
	const char *s;
	char *cmd_name, *arg;
	xsw_object_struct *con_obj;
	sws_connection_struct *con;

	if(STRISEMPTY(cmd))
	    return(-2);

	/* The connection must be valid and connected */
	if(!ConIsConnected(con_num))
	    return(-1);

	con = connection[con_num];

	/* The connection must be logged in (have a valid object) */
	con_obj_num = con->object_num;
	if(DBIsObjectGarbage(con_obj_num))
	{
	    NetSendMessage(
		con_num,
"You must be logged in in order to execute a server command."
	    );
	    return(-1);
	}

	con_obj = xsw_object[con_obj_num];

	/* At this point the connection has been checked to be valid,
	 * connected, and logged in (have an object)
	 */

	/* Parse the command's name (first argument) */
	s = cmd;
	while(ISBLANK(*s))
	    s++;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, ' ');
	    const int len = (s_end != NULL) ? (s_end - s) : strlen(s);

	    cmd_name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(cmd_name, s, len);
	    cmd_name[len] = '\0';
	    strstrip(cmd_name);

	    if(s_end != NULL)
	    {
		s = s_end;
		while(ISBLANK(*s))
		    s++;
	    }
	    else
	    {
		s = NULL;
	    }
	}
	else
	{
	    cmd_name = strdup("");
	}

	/* Get the argument */
	if(s != NULL)
	{
	    arg = strdup(s);
	}
	else
	{
	    arg = strdup("");
	}

	/* Log this command */
	if(sysparm.log_general)
	{
	    char *name = STRDUP(DBGetFormalNameStr(con_obj_num));
	    char *s = (char *)malloc(
		STRLEN(name) + STRLEN(cmd_name) +
		STRLEN(arg) + 80
	    );
	    sprintf(
		s,
"%s: Executed: Command: `%s'  Argument: `%s'",
		name, cmd_name, arg
	    );
	    free(name);
	    LogAppend(fname.log, s);
	    free(s);
	}

	/* The command and argument are now properly parsed and stored
	 * as strings command and arg (respectivly) with the cmd
	 * variable holding the original entire command.
	 *
	 * Now begin handling of the command, first check if the command
	 * matches any of the critical surface level commands, which
	 * are debug, help, plugin, and sysparm.
	 *
	 * Afterwards, if the command is not handled it will be passed
	 * to any plugins to allow them to handle it.  If no plugins
	 * handle the command then pass it to the other internal server
	 * commands.
	 */
	status = 0;
	command_handled = 0;

	/* Debug */
	if(!strcasecmp(cmd_name, "debug"))
	{
	    status = CmdDebug(con_num, arg);
	    command_handled = 1;
	}
	/* Help */
	else if(!strcasecmp(cmd_name, "help") ||
		!strcasecmp(cmd_name, "?")
	)
	{
	    status = CmdHelp(con_num, arg);
	    command_handled = 1;
	}
	/* Version */ 
	else if(!strcasecmp(cmd_name, "version"))
	{
	    status = CmdVersion(con_num, arg);
	    command_handled = 1;
	} 
	/* Sysparm */
	else if(!strcasecmp(cmd_name, "sysparam") ||
		!strcasecmp(cmd_name, "sysparm") ||
		!strcasecmp(cmd_name, "tune")
	)
	{
	    status = CmdSysparm(con_num, arg, cur_ms, cur_sec, tc);
	    command_handled = 1;
	}
#ifdef PLUGIN_SUPPORT
	/* Plugin */
	else if(!strcasecmp(cmd_name, "plugin") ||
		!strcasecmp(cmd_name, "plugins")
	)
	{
	    status = CmdPlugin(con_num, arg);
	    command_handled = 1;
	}
#endif  /* PLUGIN_SUPPORT */
	if(command_handled)
	{
	    free(cmd_name);
	    free(arg);
	    return(status);
	}

#ifdef PLUGIN_SUPPORT
	/* If the command has not been handled above then pass it
	 * to the plugins
	 */
	if(!command_handled)
	{
	    int i;
	    for(i = 0; i < total_plugins; i++)
	    {
		if(PluginCommand(
		    i,
		    con_num, con_obj_num,
		    cmd_name, arg,
		    con_obj->permission.uid,
		    con_obj->permission.gid
		))
		{
		    status = 0;
		    command_handled = 1;
		    break;
		}
	    }

	    /* If the command was handled by a plugin then return
	     * instead of passing it to the other server commands
	     * below
	     */
	    if(command_handled)
	    {
		free(cmd_name);
		free(arg);
		return(status);
	    }
	}
#endif	/* PLUGIN_SUPPORT */

	/* Laston */
	if(!strcasecmp(cmd_name, "laston"))
	{
	    status = CmdLaston(con_num, arg);
	}
	/* Link */
	else if(strcasepfx(cmd_name, "li"))
	{
	    status = CmdLink(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Uptime */
	else if(strcasepfx(cmd_name, "upt"))
	{
	    status = CmdUptime(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Memory */
	else if(strcasepfx(cmd_name, "mem"))
	{
	    status = CmdMemory(con_num, arg);
	}
	/* Disk */
	else if(strcasepfx(cmd_name, "disk"))
	{
	    status = CmdDisk(con_num, arg);
	}
	/* Password */
	else if(strcasepfx(cmd_name, "passw"))
	{
	    status = CmdPassword(con_num, arg);
	}
	/* Set */
	else if(strcasepfx(cmd_name, "set"))
	{
	    status = CmdSet(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Wall */
	else if(strcasepfx(cmd_name, "wal"))
	{
	    status = CmdWall(con_num, arg);
	}
	/* Who */
	else if(strcasepfx(cmd_name, "who"))
	{
	    status = CmdWho(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Netstat */
	else if(strcasepfx(cmd_name, "nets"))
	{
	    status = CmdNetstat(con_num, arg);
	}
	/* Score */
	else if(strcasepfx(cmd_name, "score"))
	{
	    status = CmdScore(con_num, arg);
	}
	/* ETA */
	else if(strcasepfx(cmd_name, "eta"))
	{
	    status = CmdETA(con_num, arg);
	}
	/* Save */
	else if(strcasepfx(cmd_name, "save")) 
	{
	    status = CmdSaveUniverse(con_num, arg);
	}
	/* Shutdown */
	else if(strcasepfx(cmd_name, "shutdown"))
	{
	    status = CmdShutdown(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Shutdown */
	else if(strcasepfx(cmd_name, "siteb"))
	{
	    status = CmdSiteBan(con_num, arg);
	}
	/* Boot */
	else if(strcasepfx(cmd_name, "boot"))
	{
	    status = CmdBoot(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Sync time */
	else if(strcasepfx(cmd_name, "sync"))
	{
	    status = CmdSyncTime(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* ID */
	else if(strcasepfx(cmd_name, "id"))
	{
	    status = CmdID(con_num, arg);
	}
	/* Create Player */
	else if(strcasepfx(cmd_name, "createp"))
	{
	    status = CmdCreatePlayer(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Create */
	else if(strcasepfx(cmd_name, "create"))
	{
	    status = CmdCreate(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Chown */
	else if(strcasepfx(cmd_name, "cho"))
	{
	    status = CmdChown(con_num, arg);
	}
	/* Recover */
	else if(strcasepfx(cmd_name, "unrec") ||
		strcasepfx(cmd_name, "reco")
	)
	{
	   status = CmdRecover(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Recycle Player */
	else if(strcasepfx(cmd_name, "recyclep"))
	{
	    status = CmdRecyclePlayer(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Recycle */
	else if(strcasepfx(cmd_name, "rec"))
	{
	    status = CmdRecycle(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Examine */
	else if(strcasepfx(cmd_name, "ex"))
	{
	    status = CmdExamine(con_num, arg, cur_ms, cur_sec, tc);
	}
	/* Find */
	else if(strcasepfx(cmd_name, "find"))
	{
	    status = CmdFind(con_num, arg);
	}
	/* Economy Product Create */
	else if(strcasepfx(cmd_name, "ecoproductc") ||
		strcasepfx(cmd_name, "ecoprodc")
	)
	{
	    status = CmdEcoProductCreate(con_num, arg);
	}
	/* Economy Product Set */
	else if(strcasepfx(cmd_name, "ecoproducts") ||
		strcasepfx(cmd_name, "ecoprods")
	)
	{
	    status = CmdEcoProductSet(con_num, arg);
	}
	/* Economy Product Delete */
	else if(strcasepfx(cmd_name, "ecoproductd") ||
		strcasepfx(cmd_name, "ecoprodd")
	)
	{
	    status = CmdEcoProductDelete(con_num, arg);
	}
	/* Signup */
	else if(strcasepfx(cmd_name, "signup"))
	{
	    status = CmdSignup(con_num, arg, cur_ms, cur_sec, tc);
	}


	/* Test */
	else if(!strcasecmp(cmd_name, "test")) 
	{
	    status = CmdTest(con_num, arg);
	}

	/* Unknown command */
	else
	{
	    char *sndbuf = (char *)malloc(
		STRLEN(cmd_name) + 80
	    );
	    sprintf(
		sndbuf,
"%s: No such command or macro.",
		cmd_name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    status = -2;
	}

	free(cmd_name);
	free(arg);

	return(status);
}
