#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static int CMDECO_PERMISSION_CHECK(const int con_num, const int permission);
int CmdEcoProductCreate(const int con_num, const char *arg);
int CmdEcoProductSet(const int con_num, const char *arg);
int CmdEcoProductDelete(const int con_num, const char *arg);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the connection matches or surpasses the specified
 *	permission level.
 */
int CMDECO_PERMISSION_CHECK(const int con_num, const int permission)
{
	int obj_num;
	xsw_object_struct *obj;

	/* Connection must be logged in */
	if(!ConIsLoggedIn(con_num))
	    return(0);

	/* Get connection's object */
	obj_num = connection[con_num]->object_num;
	if(DBIsObjectGarbage(obj_num))
	    return(0);

	obj = xsw_object[obj_num];

	/* Check object's permission */
	if(obj->permission.uid <= permission)
	    return(1);
	else
	    return(0);
}


/*
 *	Creates an eco product on an object.
 *
 *	Syntax: "<object>=<product_name>"
 */
int CmdEcoProductCreate(const int con_num, const char *arg)
{
	int i, obj_num;
	const char *s;
	char *name, *product_name;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *prod;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Parse the object name from argument */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(name, s, len);
	    name[len] = '\0';
	    strstrip(name);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    name = NULL;
	}

	/* Parse the product name from the argument */
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    product_name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(product_name, s, len);
	    product_name[len] = '\0';
	    strstrip(product_name);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    product_name = NULL;
	}

	/* Missing argument? */
	if((name == NULL) || (product_name == NULL))
	{
	    NetSendMessage(
		con_num,
"Usage: `ecoprodcreate <object|me|this>=<product_name>'"
	    );
	    free(name);
	    free(product_name);
	    return(-2);
	}

	/* Match the object */
	if(STRISEMPTY(name))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con_obj_num;
	}
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}

	/* Make sure that the specified object is valid */
	if(DBIsObjectGarbage(obj_num))
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: No such object.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(product_name);
	    return(-2);
	}

	obj = xsw_object[obj_num];


	/* Permission check */
	if(obj->owner != con_obj_num)
	{
	    if(!CMDECO_PERMISSION_CHECK(con_num, SWS_ACCESS_UID_SETO))
	    {
		NetSendMessage(
		    con_num,
"ecoprodcreate: You do not own that object."
		);
		free(name);
		free(product_name);
		return(-1);
	    }
	}
	else
	{
	    if(!CMDECO_PERMISSION_CHECK(con_num, SWS_ACCESS_UID_SET))
	    {
		char *s = (char *)malloc(40 + 80);
		sprintf(
		    s,
"ecoprodcreate: Permission denied: Requires access level %i.",
		    SWS_ACCESS_UID_SET
		);
		NetSendMessage(con_num, s);
		free(s);
		free(name);
		free(product_name);
		return(-1);
	    }
	}

	/* Allocate economy data on the specified object as needed */
	if(UNVAllocEco(obj))
	{
	    NetSendMessage(
		con_num,
"ecoprodcreate: Memory allocation error."
	    );
	    free(name);
	    free(product_name);
	    return(-1);
	}

	eco = obj->eco;

	/* Check if the product already exists */
	for(i = 0; i < eco->total_products; i++)
	{
	    prod = eco->product[i];
	    if(prod == NULL)
		continue;

	    if(!strcasecmp(prod->name, product_name))
	    {
		char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
		char *s = (char *)malloc(
		    STRLEN(name1) + STRLEN(product_name) + 80
		);
		sprintf(
		    s,
"ecoprodcreate: %s: Economy product `%s' already exists.",
		    name1, product_name
		);
		NetSendMessage(con_num, s);
		free(s);
		free(name1);
		free(name);
		free(product_name);
		return(0);
	    }
	}


	/* Create new product */
	if(DBCreateObjectEconomyProduct(
	    obj_num,
	    product_name,
	    -1,			/* OCS Code */
	    1.0f,		/* Sell price */
	    1.0f,		/* Buy price */
	    1.0f,		/* Amount */
	    1.0f		/* Amount max */
	))
	{
	    char *s = (char *)malloc(
		STRLEN(name) + STRLEN(product_name) + 80
	    );
	    sprintf(
		s,
"ecoprodcreate: %s: Unable to create economy product `%s'.",
		name, product_name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(product_name);
	    return(-1);
	}
	else
	{
	    /* Print response and log */
	    char	*name1 = STRDUP(DBGetFormalNameStr(obj_num)),
			*name2 = STRDUP(DBGetFormalNameStr(con_obj_num));

	    char *s = (char *)malloc(
		STRLEN(name1) + STRLEN(product_name) + 80
	    );
	    sprintf(
		s,
"Created economy product `%s' on `%s'.",
		product_name, name1
	    );
	    NetSendMessage(con_num, s);
	    free(s);

	    if(sysparm.log_general)
	    {
		char *s = (char *)malloc(
		    STRLEN(name2) + STRLEN(product_name) +
		    STRLEN(name1) + 80
		);
		sprintf(
		    s,
"%s: Created economy product `%s' on `%s'.",
		    name2, product_name, name1
		);
		LogAppend(fname.log, s);
		free(s);
	    }

	    free(name1);
	    free(name2);
	}

	/* Sort the products by name */
	DBSortEconomyProducts(obj_num);

	free(name);
	free(product_name);

	return(0);
}

/*
 *	Sets parameter of product on object.
 *
 *	Syntax: "<object>=<product_name>=<parm>:<value>"
 */
int CmdEcoProductSet(const int con_num, const char *arg)
{
	int i, obj_num;
	const char *s;
	char *name, *product_name, *parm, *val;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *prod;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Parse the object name from argument */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(name, s, len);
	    name[len] = '\0';
	    strstrip(name);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    name = NULL;
	}

	/* Parse the product name from the argument */
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    product_name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(product_name, s, len);
	    product_name[len] = '\0';
	    strstrip(product_name);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    product_name = NULL;
	}

	/* Parse the parameter from the argument */
	if(s != NULL)
	{
	    const char *s_end = strchr(s, ':');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    parm = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(parm, s, len);
	    parm[len] = '\0';
	    strstrip(parm);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    parm = NULL;
	}

	/* Parse the value from the argument */
	if(s != NULL)
	{
	    val = STRDUP(s);
	    strstrip(val);
	}
	else
	{
	    val = NULL;
	}

	/* Missing argument? */
	if((name == NULL) || (product_name == NULL) ||
	   (parm == NULL) || (val == NULL)
	)
	{
	    NetSendMessage(
		con_num,
"Usage: `ecoprodset <object|me|this>=<product_name>=<parm>:<value>'"
	    );
	    free(name);
	    free(product_name);
	    free(parm);
	    free(val);
	    return(-1);
	}

	/* Match the object */
	if(STRISEMPTY(name))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con_obj_num;
	}
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}

	/* Make sure that the specified object is valid */
	if(DBIsObjectGarbage(obj_num))
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: No such object.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(product_name);
	    free(parm);
	    free(val);
	    return(-2);
	}

	obj = xsw_object[obj_num];

	/* Permission check */
	if(obj->owner != con_obj_num)
	{
	    if(!CMDECO_PERMISSION_CHECK(con_num, SWS_ACCESS_UID_SETO))
	    {
		NetSendMessage(
		    con_num,
"ecoprodset: You do not own that object."
		);
		free(name);
		free(product_name);
		free(parm);
		free(val);
		return(-1);
	    }
	}
	else
	{
	    if(!CMDECO_PERMISSION_CHECK(con_num, SWS_ACCESS_UID_SET))
	    {
		char *s = (char *)malloc(40 + 80);
		sprintf(
		    s,
"ecoprodset: Permission denied: Requires access level %i.",
		    SWS_ACCESS_UID_SET
		);
		NetSendMessage(con_num, s);
		free(s);
		free(name);
		free(product_name);
		free(parm);
		free(val);
		return(-1);
	    }
	}

	/* Allocate economy data on the specified object as needed */
	if(UNVAllocEco(obj))
	{
	    NetSendMessage(
		con_num,
"ecoprodset: Memory allocation error."
	    );
	    free(name);
	    free(product_name);
	    free(parm);
	    free(val);
	    return(-1);
	}

	eco = obj->eco;

	/* Match the product */
	prod = NULL;
	if(isdigit(*product_name))
	{
	    /* Match product by number */
	    i = atoi(product_name);
	    if((i >= 0) && (i < eco->total_products))
		prod = eco->product[i];
	}
	else
	{
	    /* Match product by name */
	    for(i = 0; i < eco->total_products; i++)
	    {
	        if(eco->product[i] == NULL)
		    continue;

	        if(!strcasecmp(eco->product[i]->name, product_name))
	        {
		    prod = eco->product[i];
		    break;
	        }
	    }
	}
	if(prod == NULL)
	{
	    char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
	    char *s = (char *)malloc(
		STRLEN(name1) + STRLEN(product_name) + 80
	    );
	    sprintf(
		s,
"ecoprodset: %s: No such economy product `%s'.",
		name1, product_name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name1);
	    free(name); 
	    free(product_name);
	    free(parm);
	    free(val);
	    return(-2);
	}


	/* Name */
	if(!strcmp(parm, "name"))
	{
	    strncpy(prod->name, val, sizeof(prod->name));
	    prod->name[sizeof(prod->name) - 1] = '\0';
	}
	/* OCS code */
	else if(!strcmp(parm, "ocs_code"))
	{
	    prod->ocs_code = atoi(val);
	}
	/* Buy price */
	else if(!strcmp(parm, "buy_price"))
	{
	    prod->buy_price = atof(val);
	}
	/* Sell price */
	else if(!strcmp(parm, "sell_price"))
	{
	    prod->sell_price = atof(val);
	}
	/* Amount */
	else if(!strcmp(parm, "amount"))
	{
	    prod->amount = atof(val);
	}
	/* Amount max */
	else if(!strcmp(parm, "amount_max"))
	{
	    prod->amount_max = atof(val);
	}
	else
	{
	    char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
	    char *s = (char *)malloc(
		STRLEN(name1) + STRLEN(parm) + 80
	    );
	    sprintf(
		s,
"ecoprodset: %s: No such economy product parameter `%s'.",
		name1, parm
	    );
	    NetSendMessage(con_num, s); 
	    free(s);
	    free(name1);
	    free(name);
	    free(product_name);
	    free(parm);
	    free(val);
	    return(-2);
	}

 	/* Print response and log */
	if(1)
	{
	    char	*name1 = STRDUP(DBGetFormalNameStr(obj_num)),
			*name2 = STRDUP(DBGetFormalNameStr(con_obj_num));
	    char *s = (char *)malloc(
		STRLEN(name1) + STRLEN(product_name) +
		STRLEN(parm) + STRLEN(val) + 80
	    );
	    sprintf(
		s,
"Set `%s' economy product `%s' parameter `%s' to `%s'.",
		name1, product_name, parm, val
	    );
	    NetSendMessage(con_num, s);
	    free(s);

	    if(sysparm.log_general)
	    {
		char *s = (char *)malloc(
		    STRLEN(name2) + STRLEN(product_name) +
		    STRLEN(parm) + STRLEN(val) + STRLEN(name1) + 80
		);
		sprintf(
		    s,
"%s: Set `%s' economy product `%s' parameter `%s' to `%s'.",
		    name2, name1, product_name, parm, val
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	    free(name1);
	    free(name2);
	}

	/* Sort the products by name */
	DBSortEconomyProducts(obj_num);

	free(name);
	free(product_name);
	free(parm);
	free(val);

	return(0);
}


/*
 *	Deletes a product entry on an object.
 *
 *	Syntax: "<object>=<product_name>"
 */
int CmdEcoProductDelete(const int con_num, const char *arg)
{
	int i, obj_num;
	const char *s;
	char *name, *product_name;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *prod;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Parse the object name from argument */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(name, s, len);
	    name[len] = '\0';
	    strstrip(name);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    name = NULL;
	}

	/* Parse the product name from the argument */
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    product_name = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(product_name, s, len);
	    product_name[len] = '\0';
	    strstrip(product_name);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    product_name = NULL;
	}

	/* Missing argument? */
	if((name == NULL) || (product_name == NULL))
	{
	    NetSendMessage(
		con_num,
"Usage: `ecoproddelete <object|me|this>=<product_name>'"
	    );
	    free(name);
	    free(product_name);
	    return(-2);
	}

	/* Match the object */
	if(STRISEMPTY(name))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "this"))
	{
	    obj_num = con_obj->locked_on;
	}
	else if(!strcasecmp(name, "me"))
	{
	    obj_num = con_obj_num;
	}
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name, -1
	    );
	}

	/* Make sure that the specified object is valid */
	if(DBIsObjectGarbage(obj_num))
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: No such object.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name);
	    free(product_name);
	    return(-2);
	}

	obj = xsw_object[obj_num];


	/* Permission check */
	if(obj->owner != con_obj_num)
	{
	    if(!CMDECO_PERMISSION_CHECK(con_num, SWS_ACCESS_UID_SETO))
	    {
		NetSendMessage(
		    con_num,
"ecoproddelete: You do not own that object."
		);
		free(name);
		free(product_name);
		return(-1);
	    }
	}
	else
	{
	    if(!CMDECO_PERMISSION_CHECK(con_num, SWS_ACCESS_UID_SET))
	    {
		char *s = (char *)malloc(40 + 80);
		sprintf(
		    s,
"ecoproddelete: Permission denied: Requires access level %i.",
		    SWS_ACCESS_UID_SET
		);
		NetSendMessage(con_num, s);
		free(s);
		free(name);
		free(product_name);
		return(-1);
	    }
	}

	/* Get pointer to the economy data on the object */
	eco = obj->eco;
	if(eco == NULL)
	    return(-1);


	/* Get pointer to product */
	prod = NULL;
	if(isdigit(*product_name))
	{
	    /* Match product by number */
	    i = atoi(product_name);
	    if((i >= 0) && (i < eco->total_products))
		prod = eco->product[i];
	}
	else
	{
	    /* Match product by name */
	    for(i = 0; i < eco->total_products; i++)
	    {
		if(eco->product[i] == NULL)
		    continue;

		if(!strcasecmp(eco->product[i]->name, product_name))
		{
		    prod = eco->product[i];
		    break;
		}
	    }
	}
	if(prod == NULL)
	{
	    char *name1 = STRDUP(DBGetFormalNameStr(obj_num));
	    char *s = (char *)malloc(
		STRLEN(name1) + STRLEN(product_name) + 80
	    );
	    sprintf(
		s,
"ecoproddelete: %s: No such economy product `%s'.",
		name1, product_name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    free(name1);
	    free(name);
	    free(product_name);
	    return(-2);
	}

	/* Delete the product */
	free(prod);
	eco->product[i] = prod = NULL;

	/* Print response and log */
	if(1)
	{
	    char	*name1 = STRDUP(DBGetFormalNameStr(obj_num)),
			*name2 = STRDUP(DBGetFormalNameStr(con_obj_num));
	    char *s = (char *)malloc(
		STRLEN(product_name) + STRLEN(name1) + 80
	    );
	    sprintf(
		s,
"Deleted economy product `%s' on `%s'.",
		product_name, name1
	    );
	    NetSendMessage(con_num, s);
	    free(s);

	    if(sysparm.log_general)
	    {
		char *s = (char *)malloc(
		    STRLEN(name2) + STRLEN(product_name) +
		    STRLEN(name1) + 80
		);
		sprintf(
		    s,
"%s: Deleted economy product `%s' on `%s'.",
		    name2, product_name, name1
		);
		LogAppend(fname.log, s);
		free(s);
	    }

	    free(name1);
	    free(name2);
	}

	/* If all the product pointers are NULL then delete the list */
	for(i = 0; i < eco->total_products; i++)
	{
	    if(eco->product[i] != NULL)
		break;
	}
	if(i >= eco->total_products)
	{
	    free(eco->product);
	    eco->product = NULL;
	    eco->total_products = 0;
	}

	/* Sort the products by name */
	DBSortEconomyProducts(obj_num);

	free(name);
	free(product_name);

	return(0);
}
