#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fnmatch.h>

#include "../include/string.h"
#include "../include/strexp.h"

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static int CmdFindMatch(const char *pattern, const char *string);
int CmdFind(const int con_num, const char *arg);


#define THIS_CMD_NAME	"find"

#define SPACES_40	"                                        "

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Matches string with pattern, using fnmatch().
 */
static int CmdFindMatch(const char *pattern, const char *string)
{
	int status;
	char	*tmp_pattern = strdup(pattern),
		*tmp_string = strdup(string);

	strtoupper(tmp_pattern);
	strtoupper(tmp_string);

	status = fnmatch(tmp_pattern, tmp_string, 0);

	free(tmp_pattern);
	free(tmp_string);

	return(status);
}


int CmdFind(const int con_num, const char *arg)
{
	int obj_num, objects_found;
	const char *s;
	char	*sndbuf, *pattern, *type_str,
		*obj_name, *owner_name, *delta_str, position_str[256];
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    NetSendMessage(
		con_num,
"Usage: `find <name_pattern>[=<type>]'"
	    );
	    return(-2);
	}

	/* Allowed to find? */
	if(con_obj->permission.uid > SWS_ACCESS_UID_FIND)
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_FIND
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* Parse the pattern from the argument */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    pattern = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(pattern, s, len * sizeof(char));
	    pattern[len] = '\0';
	    strstrip(pattern);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    pattern = NULL;
	}

	/* Parse the type string from the argument */
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    type_str = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(type_str, s, len * sizeof(char));
	    type_str[len] = '\0';
	    strstrip(type_str);

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    type_str = NULL;
	}

	if(STRISEMPTY(pattern))
	{
	    NetSendMessage(
		con_num,
"Usage: `find <name_pattern>[=<type>]'"
	    );
	    free(pattern);
	    free(type_str);
	    return(-2);
	}


	/* Begin search, iterate through the objects list */
	objects_found = 0;
	for(obj_num = 0; obj_num < total_objects; obj_num++)
	{
	    obj = xsw_object[obj_num];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Allowed to match others? */
	    if((obj->owner != con_obj_num) &&
	       (con_obj->permission.uid > SWS_ACCESS_UID_FINDO)
	    )
		continue;

	    /* Type match? */
	    if(!STRISEMPTY(type_str))
	    {
		int status;

		switch(obj->type)
		{
		  case SW_OBJ_TYPE_STATIC:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_STATIC))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_DYNAMIC:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_DYNAMIC))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_CONTROLLED:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_CONTROLLED))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_PLAYER:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_PLAYER))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_WEAPON:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_WEAPON))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_STREAMWEAPON:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_STREAMWEAPON))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_SPHEREWEAPON:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_SPHEREWEAPON))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_HOME:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_HOME))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_AREA:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_AREA))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_ANIMATED:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_ANIMATED))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_WORMHOLE:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_WORMHOLE))
			status = -1;
		    else
			status = 0;
		    break;

		  case SW_OBJ_TYPE_ELINK:
		    if(CmdFindMatch(type_str, SW_OBJ_TYPE_NAME_ELINK))
			status = -1;
		    else
			status = 0;
		    break;

		  default:
		    status = 0;
		    break;
		}
		if(status)
		    continue;
	    }

	    /* Match pattern in the object's name */
	    if(CmdFindMatch(pattern, obj->name))
		continue;

	    /* Consider this object found */
	    objects_found++;

	    /* First object found? */
	    if(objects_found == 1)
	    {
		/* Print heading for the first time */
		NetSendMessage(
		    con_num,
"Name                     Owner                    Location"
		);
	    }

	    /* Format delta string */
	    if(SWMath3DInSectorPtr(
		obj,
		con_obj->sect_x,
		con_obj->sect_y,
		con_obj->sect_z
	    ))
	    {
		float	dx = obj->x - con_obj->x,
			dy = obj->y - con_obj->y;
		const char *units_str = "RU";

		delta_str = (char *)malloc(
		    40 + 40 + STRLEN(units_str) + 80
		);
		sprintf(
		    delta_str,
		    "Brg: %.2f' D: %.2f %s",
		    RADTODEG(SWMathCoordinateDeltaVector(dx, dy)),
		    SWMath3DDistance(dx, dy, 0),
		    units_str
		);
	    }
	    else
	    {
		float dx = obj->sect_x - con_obj->sect_x;
		float dy = obj->sect_y - con_obj->sect_y;
		const char *units_str = "S";

		delta_str = (char *)malloc(
		    40 + 40 + STRLEN(units_str) + 80
		);
		sprintf(
		    delta_str,
		    "Brg: %.2f' D: %.2f %s",
		    RADTODEG(SWMathCoordinateDeltaVector(dx, dy)),
		    SWMath3DDistance(dx, dy, 0),
		    units_str
		);
	    }

	    /* Format position string */
	    UNVLocationFormatString(
		position_str,
		&obj->sect_x, &obj->sect_y, &obj->sect_z,
		&obj->x, &obj->y, &obj->z,
		sizeof(position_str)
	    );

	    /* Format names */
	    obj_name = STRDUP(DBGetFormalNameStr(obj_num));
	    owner_name = STRDUP(DBGetFormalNameStr(obj->owner));

	    sndbuf = (char *)malloc(
		STRLEN(obj_name) + STRLEN(owner_name) +
		STRLEN(delta_str) + STRLEN(position_str) +
		STRLEN(SPACES_40) + (2 * 80)
	    );

	    strcpy(sndbuf, obj_name);
	    strcat(sndbuf, SPACES_40);
	    strcpy(sndbuf + 25, owner_name);
	    strcat(sndbuf, SPACES_40);
	    strcpy(sndbuf + 50, delta_str);
	    strcat(sndbuf, SPACES_40);
	    strcpy(sndbuf + 78, position_str);

	    NetSendMessage(con_num, sndbuf);

	    free(delta_str);
	    free(owner_name);
	    free(obj_name);
	    free(sndbuf);
	}

	/* Print number of objects found */
	if(objects_found > 0)
	{
	    sndbuf = (char *)malloc(40 + 80);
	    sprintf(
		sndbuf,
"*** %i %s Found ***",
		objects_found,
		(objects_found == 1) ? "Object" : "Objects"
	    );
	}
	else
	{
	    sndbuf = (char *)malloc(80);
	    strcpy(sndbuf, "*** No Objects Found ***");
	}
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);

	free(pattern);
	free(type_str);

	return(0);
}
