#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fnmatch.h>

#include "../include/string.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static int CmdLastonSortLastConnectedCB(const void *a, const void *b);
static int CmdLastonMatch(
	const int con_obj_num, xsw_object_struct *con_obj,
	const char *s, const char *name
);
static xsw_object_struct **CmdLastonGetPlayersList(
	const int con_obj_num, xsw_object_struct *con_obj,
	const char *s, int *total_rtn
);

int CmdLaston(const int con_num, const char *arg);


#define THIS_CMD_NAME	"laston"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sort by last connected time callback.
 *
 *	The more recently connected will be listed last.
 */
static int CmdLastonSortLastConnectedCB(const void *a, const void *b)
{
	xsw_object_struct	*o1 = *((xsw_object_struct **)a),
				*o2 = *((xsw_object_struct **)b);
	return((int)(o1->last_connected - o2->last_connected));
}

/*
 *	Checks if the specified name matches the specified search string
 *	s relative to the specified object con_obj.
 */
static int CmdLastonMatch(
	const int con_obj_num, xsw_object_struct *con_obj,
	const char *s, const char *name
)
{
	int status;
	char	*s2 = STRDUP(s),
		*name2 = STRDUP(name);

	if((s2 == NULL) || (name2 == NULL))
	{
	    free(s2);
	    free(name2);
	    return(0);
	}

	if(!strcasecmp(s2, "me"))
	{
	    if(!strcmp(con_obj->name, name2))
	    {
		free(s2);
		free(name2);
		return(1);
	    }
	}

	if(!strcasecmp(s2, "this"))
	{
	    const int obj_num = con_obj->locked_on;
	    if(!DBIsObjectGarbage(obj_num))
	    {
		xsw_object_struct *obj = xsw_object[obj_num];
		if(!strcmp(obj->name, name2))
		{
		    free(s2);
		    free(name2);
		    return(1);
		}
	    }
	}

	strtoupper(s2);
	strtoupper(name2);

	status = fnmatch(s2, name2, 0);

	free(s2);
	free(name2);

	return(!status);
}

/*
 *	Returns a list of player objects that matches the specified
 *	search string s.
 *
 *	The returned list must be deleted by the calling function, but
 *	not each object.
 */
static xsw_object_struct **CmdLastonGetPlayersList(
	const int con_obj_num, xsw_object_struct *con_obj,
	const char *s, int *total_rtn
)
{
	int i, n, total = 0;
	xsw_object_struct *obj, **list = NULL;

	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type != SW_OBJ_TYPE_PLAYER)
		continue;

	    if(!CmdLastonMatch(con_obj_num, con_obj, s, obj->name))
		continue;

	    n = total;
	    total++;
	    list = DBReallocList(list, total);
	    if(list == NULL)
	    {
		total = 0;
		break;
	    }

	    list[n] = obj;
	}

	if(total_rtn != NULL)
	    *total_rtn = total;

	return(list);
}


int CmdLaston(const int con_num, const char *arg)
{
	int i, obj_num, con_obj_num, nobj_matched, total;
	const char *name;
	xsw_object_struct *obj, *con_obj, **list;
	sws_connection_struct *con = connection[con_num];

	con_obj_num = con->object_num;
	con_obj = xsw_object[con_obj_num];

	/* Print usage? */ 
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 80);
	    sprintf(
		s,
"Usage: `%s <player|me|this>'",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Get name */
	name = arg;

	/* No permission check needed */

	/* Get the list of matched player objects */
	list = CmdLastonGetPlayersList(
	    con_obj_num, con_obj,
	    name, &total
	);

	/* Sort the list */
	if(list != NULL)
	    qsort(
		list, total,
		sizeof(xsw_object_struct *),
		CmdLastonSortLastConnectedCB
	    );

	/* Iterate through the matched objects list */
	nobj_matched = 0;
	for(i = 0; i < total; i++)
	{
	    obj = list[i];
	    if(obj == NULL)
		continue;

	    obj_num = DBGetObjectNumByPtr(obj);

	    /* Already connected? */
	    if(ConGetByObject(obj_num) > -1)
	    {
		char *s = (char *)malloc(STRLEN(obj->name) + 1024);
		*s = '\0';
		strcat(s, obj->name);
		strcat(s, "                              ");
		s[24] = '\0';
		strcat(s, "Online");
		NetSendMessage(con_num, s);
		free(s);
	    }
	    else
	    {
		const char *ts = (obj->last_connected > 0l) ?
		    StringFormatTimePeriod(
			cur_systime - obj->last_connected
		) : NULL;
		char *s = (char *)malloc(STRLEN(obj->name) + STRLEN(ts) + 1024);
		*s = '\0';
		strcat(s, obj->name);
		strcat(s, "                              ");
		s[24] = '\0';
		if(ts != NULL)
		{
		    strcat(s, ts);
		    strcat(s, " ago");
		}
		else
		{
		    strcat(s, "Never connected");
		}
		NetSendMessage(con_num, s);
		free(s);
	    }
	    nobj_matched++;
	}

	free(list);

	/* No objects matched? */
	if(nobj_matched == 0)
	{
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"%s: No such player.",
		name
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	return(0);
}
