#ifdef PLUGIN_SUPPORT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/prochandle.h"

#include "plugins.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static void CmdPluginPrintUsage(const int con_num);
int CmdPlugin(const int con_num, const char *arg);


#define THIS_CMD_NAME	"plugin"


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static void CmdPluginPrintUsage(const int con_num)
{
	char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + (1 * 80));
	sprintf(
	    s,
 "Usage: `%s [load <plugin> [args...]] [unload <id>]'",
	    THIS_CMD_NAME
	);
	NetSendMessage(con_num, s);
	free(s);
}


int CmdPlugin(const int con_num, const char *arg)
{
	int i, n, argc;
	char **argv;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];
	plugin_id_t id;
	plugin_struct *pi;

	/* Print listing of plugins if no argument is given */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    int j;
	    const char *arg;
	    char *s, *all_args;

	    /* Check permissions to view */
	    if(con_obj->permission.uid > SWS_ACCESS_UID_PLUGINVIEW)
	    {
		char *s = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 40 + 80
		);
		sprintf(
		    s,
"%s: Permission denied: Requires access level %i.",
		    THIS_CMD_NAME,
		    SWS_ACCESS_UID_PLUGINVIEW
		);
		NetSendMessage(con_num, s);
		free(s);
		return(0);
	    }

	    NetSendMessage(
		con_num,
		"-     ID CMD"
	    );

	    /* Print list of loaded plugins and their arguments */
	    for(i = 0, n = 0; i < total_plugins; i++)
	    {
		pi = plugin[i];
		if(pi == NULL)
		    continue;

		all_args = NULL;
		for(j = 0; j < pi->argc; j++)
		{
		    arg = pi->argv[j];
		    if(arg == NULL)
			continue;

		    if(strchr(arg, ' ') != NULL)
		    {
			all_args = strinsstr(
			    all_args, -1, " \""
			);
			all_args = strinsstr(
			    all_args, -1, arg
			);
			all_args = strinsstr(
			    all_args, -1, "\""
			);
		    }
		    else
		    {
			all_args = strinsstr(
			    all_args, -1, " "
			);
			all_args = strinsstr(
			    all_args, -1, arg
			);
		    }
		}
		if(all_args == NULL)
		    strdup("");

		s = (char *)malloc(
		    40 + STRLEN(pi->name) + STRLEN(all_args) + 80
		);
		sprintf(
		    s,
		    "- %6d %s%s",
		    i + 1,		/* Printed ID is 1 + the actual ID */
		    pi->name,
		    (all_args != NULL) ? all_args : ""
		);
		NetSendMessage(con_num, s);
		free(s);
		free(all_args);

		n++;
	    }

	    /* Print footer */
	    if(n > 0)
	    {
		s = (char *)malloc(40 + 80);
		sprintf(
		    s,
"%i %s loaded.",
		    n,
		    (n > 1) ? "plugins" : "plugin"
		);
		NetSendMessage(con_num, s);
		free(s);
	    }
	    else
	    {
		NetSendMessage(
		    con_num,
"No plugins loaded."
		);
	    }

	    /* Print usage */
	    CmdPluginPrintUsage(con_num);
	}
	else
	{
	    /* Check permissions to load or unload */
	    if(con_obj->permission.uid > SWS_ACCESS_UID_PLUGINOP)
	    {
		char *s = (char *)malloc(
		    STRLEN(THIS_CMD_NAME) + 40 + 80
		);
		sprintf(
		    s,
"%s: Permission denied: Requires access level %i.",
		    THIS_CMD_NAME,
		    SWS_ACCESS_UID_PLUGINOP
		);
		NetSendMessage(con_num, s);
		free(s);
		return(0);
	    }

	    /* Parse arguments */
	    argv = ExecExplodeCommand(arg, &argc);

	    /* Got at least 2 arguments? */
	    if(argc < 2)
	    {
		/* Nope, print usage */
		CmdPluginPrintUsage(con_num);
	    }
	    else
	    {
		/* Got atleast 2 arguments */

		/* Load plugin? */
		if(strcasepfx(argv[0], "l"))
		{
		    /* Load plugin */
		    int pl_argc = argc - 2;
		    char	*s,
				**pl_argv = (pl_argc > 0) ? &argv[2] : NULL;
		    const char *path = argv[1];

		    /* Complete path? */
		    if(!ISPATHABSOLUTE(path))
		    {
			const char *full_path = PrefixPaths(dname.plugins, argv[1]);
			free(argv[1]);
			argv[1] = STRDUP(full_path);
			path = argv[1];
		    }

		    /* Load plugin */
		    id = PluginLoad(
			path,			/* Path */
			0,			/* Flags */
			pl_argc, pl_argv,	/* Arguments */
			con_num		/* Connection Number */
		    );
		    switch(id)
		    {
		      case -1:	/* Plugin not found */
			s = (char *)malloc(STRLEN(path) + 80);
			sprintf(
			    s,
"%s: No such plugin by that name found.",
			    path
			);
			break;
		      case -2:
			s = (char *)malloc(
			    STRLEN(path) +
			    STRLEN(SWPLUGIN_INIT_FUNCNAME) +
			    STRLEN(SWPLUGIN_MANAGE_FUNCNAME) +
			    STRLEN(SWPLUGIN_SHUTDOWN_FUNCNAME)
			);
			sprintf(
			    s,
"%s: Missing %s, %s, or %s function(s).",
			    path,
			    SWPLUGIN_INIT_FUNCNAME,
			    SWPLUGIN_MANAGE_FUNCNAME,
			    SWPLUGIN_SHUTDOWN_FUNCNAME
			);
			break;
		      case -3:
			s = (char *)malloc(STRLEN(path) + 80);
			sprintf(
			    s,
"%s: Unknown error occured during loading of plugin.",
			    path
			);
			break;
		      case -4:
			s = (char *)malloc(STRLEN(path) + 80);
			sprintf(
			    s,
"%s: Exited abruptly but normally upon initialization.",
			    path
			);
			break;
		      default:
			s = (char *)malloc(STRLEN(path) + 40 + 80);
			sprintf(
			    s,
"Loaded plugin: ID %i: %s",
			    id + 1,
			    path
			);
			break;
		    }
		    NetSendMessage(con_num, s);
		    free(s);
		}
		/* Unload plugin? */
		else if(strcasepfx(argv[0], "u"))
		{
		    char *s;

		    /* Unload plugin */
		    id = atoi(argv[1]) - 1;
		    if(PluginIsLoaded(id))
		    {
			pi = plugin[id];
			s = (char *)malloc(STRLEN(pi->name) + 40 + 80);
			sprintf(
			    s,
"Unloaded plugin: ID %i: %s",
			    id + 1,
			    pi->name
			);
			NetSendMessage(con_num, s);
			free(s);
			PluginUnload(id, con_num);
		    }
		    else
		    {
			s = (char *)malloc(40 + 80);
			sprintf(
			    s,
"%i: No such plugin id.",
			    id + 1
			);
			NetSendMessage(con_num, s);
			free(s);
		    }
		}
		else
		{
		    /* Unsupported operation, print usage */
		    CmdPluginPrintUsage(con_num);
		}
	    }

	    /* Delete parsed arguments */
	    strlistfree(argv, argc);
	}

	return(0);
}

#endif	/* PLUGIN_SUPPORT */
