#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>

#include "../include/string.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static void PRINT_USAGE(const int con_num);
int CmdShutdown(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define THIS_CMD_NAME	"shutdown"

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static void PRINT_USAGE(const int con_num)
{
	char *sndbuf = (char *)malloc(
	    STRLEN(THIS_CMD_NAME) + 80
	);
	sprintf(
	    sndbuf,
"Usage: `%s <seconds|now|abort>'",
	    THIS_CMD_NAME
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);
}

/*
 *	Scheduals or aborts a server shutdown.
 */
int CmdShutdown(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int con_obj_num;
	time_t t = 0l;
	char *sndbuf;
	xsw_object_struct *con_obj;
	sws_connection_struct *con = connection[con_num];

	/* Get connection's object number (assumed valid) */
	con_obj_num = con->object_num;
	con_obj = xsw_object[con_obj_num];

	/* Check the connection's permission */
	if(con_obj->permission.uid > SWS_ACCESS_UID_SHUTDOWN)
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 40 + 80
	    );
	    sprintf(
		sndbuf,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_SHUTDOWN
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	    return(-1);
	}

	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    if(next.shutdown > 0l)
	    {
		const time_t t = MAX((next.shutdown - cur_sec), 0l);
		char *time_str = STRDUP(StringFormatTimePeriod(t));
		sndbuf = (char *)malloc(
		    STRLEN(time_str) + STRLEN(THIS_CMD_NAME) + (2 * 80)
		);
		sprintf(
		    sndbuf,
"The server has been schedualed to shutdown in %s,\
 type `%s abort' to abort it.",
		    time_str,
		    THIS_CMD_NAME
		);
		NetSendMessage(con_num, sndbuf);
		free(sndbuf);
		free(time_str);
	    }
	    PRINT_USAGE(con_num);
	    return(0);
	}

	/* Specified argument is in seconds? */
	if(isdigit(*arg))
	{
	    t = atol(arg);
	}
	else
	{
	    while(*arg == '-')
		arg++;

	    /* Shutdown now? */
	    if(toupper(*arg) == 'N')
	    {
		t = 0l;
	    }
	    /* Abort shutdown? */
	    else if((toupper(*arg) == 'A') || (toupper(*arg) == 'C'))
	    {
		/* Shutdown in process? */
		if(next.shutdown > 0l)
		{
		    /* Cancel shutdown */
		    next.shutdown = 0l;

		    sndbuf = (char *)malloc(80);
		    sprintf( 
			sndbuf,
"*** Server shutdown aborted. ***"
		    );
		    NetSendMessage(-1, sndbuf);
		    free(sndbuf);

		    if(!sysparm.console_quiet)
			fprintf(
			    stdout,
"%s shutdown aborted.\n",
			    PROG_NAME
			);

		    if(sysparm.log_general)  
		    {
			char	*name = STRDUP(DBGetFormalNameStr(con_obj_num)),
				*s = (char *)malloc(STRLEN(name) + 80);
			sprintf(
			    s,
"%s: Aborted server shutdown.",
			    name
			);
			LogAppend(fname.log, s);
			free(s);
			free(name);
		    }
		}
		else
		{
		    sndbuf = (char *)malloc(80);
		    sprintf(
			sndbuf,
"There was no server shutdown schedualed to be aborted."
		    );
		    NetSendMessage(con_num, sndbuf);
		    free(sndbuf);
		}
		return(0);
	    }
	    else
	    {
		/* Unsupported argument */
		PRINT_USAGE(con_num);
		return(-2);
	    }
	}

	/* Shutdown already schedualed? */
	if(next.shutdown > 0l)
	{
	    sndbuf = (char *)malloc(
		STRLEN(THIS_CMD_NAME) + 128
	    );
	    sprintf(
		sndbuf,
"There is already a server shutdown schedualed, type `%s abort' to abort it.",
		THIS_CMD_NAME
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	/* Log who requested shutdown */
	if(sysparm.log_general)
	{
	    char *name = STRDUP(DBGetFormalNameStr(con_obj_num));
	    if(t > 0l)
	    {
                char	*time_str = STRDUP(StringFormatTimePeriod(t)),
			*s = (char *)malloc(
		    STRLEN(name) + STRLEN(time_str) + 80
		);
		sprintf(
		    s,
"%s: Requesting server shutdown in %s.",
		    name, time_str
		);
		LogAppend(fname.log, s);
		free(s);
		free(time_str);
	    }
	    else
	    {
		char *s = (char *)malloc(
                    STRLEN(name) + 80
                );
		sprintf(
		    s,
"%s: Requesting immediate server shutdown.",
		    name
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	    free(name);
	}

	/* Schedual shutdown */
	SWSShutdownStart(t);

	return(0);
}
