#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fnmatch.h>
#include <sys/types.h>

#include "../include/string.h"
#include "../include/unvutil.h"
#include "../include/unvmatch.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


#define THIS_CMD_NAME	"sysparm"


static int CmdSysparmMatchParm(const char *str, const char *parm);
static char *CmdSysparmFmtDeltaTime(time_t dt);
int CmdSysparm(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns true if str matches parm, using fnmatch() style
 *	wildcard matching.
 */
static int CmdSysparmMatchParm(const char *str, const char *parm)
{
	return(!fnmatch(str, parm, 0));
}


/*
 *	Returns a statically allocated string containing a verbose
 *	explaination of the delta time in seconds.
 */
static char *CmdSysparmFmtDeltaTime(time_t dt)
{
	static char s[256];

	if(dt < 0l)
	    dt = 0l;

	if(dt < 60)
	    sprintf(s, "%ld s", dt);
	else if(dt < 3600)
	    sprintf(s, "%ld m", dt / 60);
	else if(dt < 86400)
	    sprintf(s, "%ld h", dt / 3600);
	else
	    sprintf(s, "%ld d", dt / 86400);
	s[sizeof(s) - 1] = '\0';

	return(s);
}


/*
 *	Sysparm.
 */
int CmdSysparm(
	const int con_num, const char *arg,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int matches = 0;
	const char *s, *parm_name;
	char *sndbuf, *parm, *val;
	sw_legend_struct *legend = &sw_legend;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;
	xsw_object_struct *con_obj = xsw_object[con_obj_num];

	/* Check if allowed to view or set system parameters */
	if(con_obj->permission.uid > SWS_ACCESS_UID_SYSPARM)
	{
	    char *s = (char *)malloc(STRLEN(THIS_CMD_NAME) + 40 + 80);
	    sprintf(
		s,
"%s: Permission denied: Requires access level %i.",
		THIS_CMD_NAME, SWS_ACCESS_UID_SYSPARM
	    );
	    NetSendMessage(con_num, s);
	    free(s);
	    return(-1);
	}

	/* Parse the parameter */
	s = arg;
	if(s != NULL)
	{
	    const char *s_end = strchr(s, '=');
	    const int len = (s_end != NULL) ? (s_end - s) : STRLEN(s);
	    parm = (char *)malloc((len + 1) * sizeof(char));
	    if(len > 0)
		memcpy(parm, s, len);
	    parm[len] = '\0';
	    strstrip(parm);

	    if(STRISEMPTY(parm))
	    {
		free(parm);
		parm = STRDUP("*");
	    }

	    s = (s_end != NULL) ? (s_end + 1) : NULL;
	}
	else
	{
	    parm = STRDUP("*");
	}

	/* Parse the value */
	if(s != NULL)
	{
	    val = STRDUP(s);
	    strstrip(val);
	    s = NULL;
	}
	else
	{
	    val = STRDUP("");
	}

	if((parm == NULL) || (val == NULL))
	{
	    free(parm);
	    free(val);
	    return(-3);
	}


	/* Check input parameter through each parameter, set the
	 * new value if new value is specified
	 */
	parm_name = "allow_guest";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.allow_guest = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.allow_guest ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "allow_guest_signup";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.allow_guest_signup = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf, 
		"%s = %s",
		parm_name,
		sysparm.allow_guest_signup ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "antimatter_low_warning";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->antimatter_low_warning = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    if(legend->antimatter_low_warning < 1.0f)
	        sprintf(
		    sndbuf,
		    "%s = %.3f",
		    parm_name,
		    legend->antimatter_low_warning
		);
	    else
		sprintf(
		    sndbuf,
		    "%s = %.2f",
		    parm_name,
		    legend->antimatter_low_warning
		);
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "asteroid_field_damage_int";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->asteroid_field_damage_int = MAX(atol(val), 1000l);
		NetSendLegend(-1);
 	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld",
		parm_name,
		legend->asteroid_field_damage_int
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "asteroid_field_damage_rate";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->asteroid_field_damage_rate = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
 	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %f",
		parm_name,
		legend->asteroid_field_damage_rate
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "auto_destruct_damage_range";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->auto_destruct_damage_range = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %.4f",
		parm_name,
		legend->auto_destruct_damage_range
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "cease_fire";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.cease_fire = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.cease_fire ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "cloak_antimatter_rate";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->cloak_antimatter_rate = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %.4f",
		parm_name,
		legend->cloak_antimatter_rate
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "com_range";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->com_range = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %.4f ru",
		parm_name,
		legend->com_range
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "con_notify"; 
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.con_notify = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.con_notify ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "cpu_priority";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		if(isdigit(*val))
		{
		    sysparm.cpu_priority = atoi(val);
		}
		else
		{
		    if(!strcasecmp(val, "highest"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_HIGHEST;
		    else if(strcasepfx(val, "h"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_HIGH;
		    else if(strcasepfx(val, "m"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_MEDIUM;
		    else if(strcasepfx(val, "l"))
			sysparm.cpu_priority = SWS_CPU_PRIORITY_LOWEST;
		}
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    switch(sysparm.cpu_priority)
	    {
	      case SWS_CPU_PRIORITY_HIGHEST:
		sprintf(
		    sndbuf,
		    "%s = highest",
		    parm_name
		);
		break;
	      case SWS_CPU_PRIORITY_HIGH:
		sprintf(
		    sndbuf,
		    "%s = high",
		    parm_name
		);
		break;
	      case SWS_CPU_PRIORITY_MEDIUM:
		sprintf(
		    sndbuf,
		    "%s = medium",
		    parm_name
		);
		break;
	      default:	/* SWS_CPU_PRIORITY_LOWEST */
		sprintf(
		    sndbuf,
		    "%s = lowest",
		    parm_name
		);
		break;
	    }
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "damage_control_rate";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->damage_control_rate = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %f",
		parm_name,
		legend->damage_control_rate
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "disarmed_weapon_lifespan";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t nval;

		if(isdigit(*val))
		{
		    nval = (time_t)atol(val);
		}
		else
		{
		    if(!strcasecmp(val, "infinate") ||
		       strcasepfx(val, "i")
		    )
			nval = 0l;
		    else if(!strcasecmp(val, "no_change") ||
			    !strcasecmp(val, "no change") || 
			    strcasepfx(val, "n")
		    )
			nval = 1l;
		    else
			nval = 1l;
		}
		legend->disarmed_weapon_lifespan = nval;
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    if(legend->disarmed_weapon_lifespan == 0l)
		sprintf(
		    sndbuf,
		    "%s = infinate",
		    parm_name
		);
	    else if(legend->disarmed_weapon_lifespan == 1l)
		sprintf(
		    sndbuf,
		    "%s = no_change",
		    parm_name
		);
	    else
		sprintf(
		    sndbuf,
		    "%s = %ld ms",
		    parm_name,
		    legend->disarmed_weapon_lifespan
		);
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "engine_antimatter_rate";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->engine_antimatter_rate = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %f",
		parm_name,
		legend->engine_antimatter_rate
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "file_imgrefs";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    legend->imgrefs, val,
		    sizeof(legend->imgrefs)
		);
		legend->imgrefs[sizeof(legend->imgrefs) - 1] = '\0';
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(
		STRLEN(parm_name) + STRLEN(legend->imgrefs) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		legend->imgrefs
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "file_ocsn"; 
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    legend->ocsn, val, sizeof(legend->ocsn)
		);
		legend->ocsn[sizeof(legend->ocsn) - 1] = '\0';
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(
		STRLEN(parm_name) + STRLEN(legend->ocsn) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		legend->ocsn   
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "file_sndrefs";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    legend->sndrefs, val, sizeof(legend->sndrefs)
		);
		legend->sndrefs[sizeof(legend->sndrefs) - 1] = '\0';
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(
		STRLEN(parm_name) + STRLEN(legend->sndrefs) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		legend->sndrefs
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "guest_login_name";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    sysparm.guest_login_name, val,
		    sizeof(sysparm.guest_login_name)
		);
		sysparm.guest_login_name[
		    sizeof(sysparm.guest_login_name) - 1] = '\0';
	    }
	    sndbuf = (char *)malloc(                  
		STRLEN(parm_name) + STRLEN(sysparm.guest_login_name) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.guest_login_name
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "guest_start_direction";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];

	    matches++;
	    if(!STRISEMPTY(val))
	    {
		UNVParseDirection(
		    val,
		    &legend->guest_start_heading,
		    &legend->guest_start_pitch,
		    &legend->guest_start_bank
		);
		NetSendLegend(-1);
	    }
	    UNVDirectionFormatString(
		text,
		&legend->guest_start_heading,
		&legend->guest_start_pitch,
		&legend->guest_start_bank,
		sizeof(text)
	    );
	    sndbuf = (char *)malloc(                  
		STRLEN(parm_name) + STRLEN(text) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		text
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "guest_start_location";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];

	    matches++;
	    if(!STRISEMPTY(val))
	    { 
		UNVParseLocation(
		    val,
		    &legend->guest_start_sect_x,
		    &legend->guest_start_sect_y,
		    &legend->guest_start_sect_z,
		    &legend->guest_start_x,
		    &legend->guest_start_y,
		    &legend->guest_start_z
		);
		NetSendLegend(-1);
	    }
	    UNVLocationFormatString(
		text,
		&legend->guest_start_sect_x,
		&legend->guest_start_sect_y,
		&legend->guest_start_sect_z,
		&legend->guest_start_x,
		&legend->guest_start_y,
		&legend->guest_start_z,
		sizeof(text)
	    );
	    sndbuf = (char *)malloc(                  
		STRLEN(parm_name) + STRLEN(text) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		text
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "hide_players_home";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.hide_players_home = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.hide_players_home ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);

	    /* Need to update all connected an non-connected player
	     * objects's hidden state when this is updated
	     */
	    DBReclaim(cur_ms, cur_sec, tc);
	}

	parm_name = "hit_player_bonus";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.hit_player_bonus = MAX(atof(val), 0.0f);
	    }

	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    if(sysparm.hit_player_bonus == 0.0f)
		sprintf(
		    sndbuf,
		    "%s = %s",
		    parm_name,
		    "none"
		);
	    else
		sprintf(
		    sndbuf,
		    "%s = %f",
		    parm_name,
		    sysparm.hit_player_bonus
		);
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "homes_destroyable";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.homes_destroyable = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.homes_destroyable ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_aux_con_stats";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_aux_con_stats = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_aux_con_stats
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_memory_clean";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_memory_clean = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_memory_clean
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_new_connection_poll"; 
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_new_connection_poll = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_new_connection_poll
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_object_values";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_object_values = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_object_values
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_os_stats";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In seconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_os_stats = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld s  (next: %s)",
		parm_name,
		sysparm.int_os_stats,
		CmdSysparmFmtDeltaTime(next.os_stats - cur_sec)
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_stats_export";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In seconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_stats_export = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld s  (next: %s)",
		parm_name,
		sysparm.int_stats_export,
		CmdSysparmFmtDeltaTime(next.stats_export - cur_sec)
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_system_check";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_system_check = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_system_check
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_unv_save";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In seconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_unv_save = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld s  (next: %s)",
		parm_name,
		sysparm.int_unv_save,
		CmdSysparmFmtDeltaTime(next.unv_save - cur_sec)
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "int_weapon_values";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In milliseconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.int_weapon_values = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld ms",
		parm_name,
		sysparm.int_weapon_values
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "killer_gets_credits";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.killer_gets_credits = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		(sysparm.killer_gets_credits) ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "log_errors"; 
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.log_errors = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.log_errors ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "log_events";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.log_events = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.log_events ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "log_general";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.log_general = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.log_general ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "log_net";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.log_net = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.log_net ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "login_timeout";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		time_t dt = atol(val);		/* In seconds */
		if(dt < 1l)
		    dt = 1l;
		sysparm.login_timeout = dt;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %ld s",
		parm_name,
		sysparm.login_timeout
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "lost_found_owner";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		const int obj_num = UNVMatchObjectByName(
		    xsw_object, total_objects,
		    val, SW_OBJ_TYPE_PLAYER
		);
		if(DBIsObjectGarbage(obj_num))
		{
		    sndbuf = (char *)malloc(STRLEN(val) + 80);
		    sprintf(
			sndbuf,
			"%s: No such player object.",
			val
		    );
		}
		else
		{
		    const char *name = DBGetFormalNameStr(obj_num);
		    legend->lost_found_owner = obj_num;
		    sndbuf = (char *)malloc(STRLEN(parm_name) + STRLEN(name) + 80);
		    sprintf(
			sndbuf,
			"%s = %s",
			parm_name, name
		    );
		}
		NetSendLegend(-1);
	    }
	    else
	    {
		const char *name = DBGetFormalNameStr(legend->lost_found_owner);
		sndbuf = (char *)malloc(STRLEN(parm_name) + STRLEN(name) + 80);
		sprintf(
		    sndbuf,
		    "%s = %s",
		    parm_name, name
		);
	    }
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "max_aux_connections";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.max_aux_connections = MAX(ATOI(val), 0);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.max_aux_connections
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "max_connections";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.max_connections = MAX(ATOI(val), 1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.max_connections
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "max_failed_logins";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		int i = atoi(val);
		if(i < 1)
		    i = 1;
		sysparm.max_failed_logins = i;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.max_failed_logins
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "max_guests";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		int i = atoi(val);
		if(i < 0)
		    i = 0;
		sysparm.max_guests = i;
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.max_guests
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "mesg_leave";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    sysparm.mesg_leave, val, sizeof(sysparm.mesg_leave)
		);
		sysparm.mesg_leave[sizeof(sysparm.mesg_leave) - 1] = '\0';
	    }
	    sndbuf = (char *)malloc(
		STRLEN(parm_name) + STRLEN(sysparm.mesg_leave) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.mesg_leave
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "mesg_no_guests";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    sysparm.mesg_no_guests, val,
		    sizeof(sysparm.mesg_no_guests)
		);
		sysparm.mesg_no_guests[
		    sizeof(sysparm.mesg_no_guests) - 1] = '\0';
	    }
	    sndbuf = (char *)malloc(
		STRLEN(parm_name) + STRLEN(sysparm.mesg_no_guests) + 80  
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.mesg_no_guests
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "mesg_welcome";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    sysparm.mesg_welcome, val,
		    sizeof(sysparm.mesg_welcome)
		);
		sysparm.mesg_welcome[
		    sizeof(sysparm.mesg_welcome) - 1] = '\0';
	    }
	    sndbuf = (char *)malloc(
		STRLEN(parm_name) + STRLEN(sysparm.mesg_welcome) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.mesg_welcome
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "mesg_wrong_login";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    sysparm.mesg_wrong_login, val,
		    sizeof(sysparm.mesg_wrong_login)
		);
		sysparm.mesg_wrong_login[
		    sizeof(sysparm.mesg_wrong_login) - 1] = '\0';
	    }
	    sndbuf = (char *)malloc(
		STRLEN(parm_name) + STRLEN(sysparm.mesg_wrong_login) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.mesg_wrong_login
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "movement_limit_tolorance";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.movement_limit_tolorance = MAX(atof(val), 1.0f);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %f",
		parm_name,
		sysparm.movement_limit_tolorance
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "no_antimatter_thrust_coeff";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->no_antimatter_thrust_coeff = CLIP(atof(val), 0.0f, 1.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %.3f",
		parm_name,
		legend->no_antimatter_thrust_coeff
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "on_fire_hp_coeff";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->on_fire_hp_coeff = CLIP(atof(val), 0.0f, 1.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %.3f",
		parm_name,
		legend->on_fire_hp_coeff
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "player_start_direction";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];
	 
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		UNVParseDirection(
		    val,
		    &legend->player_start_heading,
		    &legend->player_start_pitch,
		    &legend->player_start_bank
		);
		NetSendLegend(-1);
	    }
	    UNVDirectionFormatString(
		text,
		&legend->player_start_heading,
		&legend->player_start_pitch,
		&legend->player_start_bank,
		sizeof(text)
	    );
	    sndbuf = (char *)malloc(STRLEN(parm_name) + STRLEN(text) + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		text
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "player_start_location";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    char text[256];
		
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		UNVParseLocation(
		    val,
		    &legend->player_start_sect_x,
		    &legend->player_start_sect_y,
		    &legend->player_start_sect_z,
		    &legend->player_start_x,
		    &legend->player_start_y,
		    &legend->player_start_z
		);
		NetSendLegend(-1);
	    }
	    UNVLocationFormatString(
		text,
		&legend->player_start_sect_x,
		&legend->player_start_sect_y,
		&legend->player_start_sect_z,
		&legend->player_start_x,
		&legend->player_start_y,
		&legend->player_start_z,
		sizeof(text)
	    );
	    sndbuf = (char *)malloc(STRLEN(parm_name) + STRLEN(text) + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		text
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "recycled_objects_max";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.recycled_objects_max = MAX(ATOI(val), 0);
/* Need to reallocate recycled objects list */
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %i",
		parm_name,
		sysparm.recycled_objects_max
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "report_destroyed_weapons";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.report_destroyed_weapons = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.report_destroyed_weapons ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "respawn_players";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.respawn_players = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.respawn_players ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "ru_to_au";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->ru_to_au = MAX(atof(val), 0.0f);
		if(legend->ru_to_au <= 0.0f)
		    legend->ru_to_au = SW_DEF_RU_TO_AU;
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    if(legend->ru_to_au < 1.0f)
		sprintf(
		    sndbuf,
		    "%s = %.5f",
		    parm_name,
		    legend->ru_to_au
		);
	    else
		sprintf(
		    sndbuf,
		    "%s = %.3f",
		    parm_name,
		    legend->ru_to_au
		);
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "send_starchart";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.send_starchart = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.send_starchart ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "unv_title";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		strncpy(
		    legend->title, val, sizeof(legend->title)
		);
		legend->title[
		    sizeof(legend->title) - 1] = '\0';
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(
		STRLEN(parm_name) + STRLEN(legend->title) + 80
	    );
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		legend->title
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "tractor_beam_antimatter_rate";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->tractor_beam_antimatter_rate = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %.4f",
		parm_name,
		legend->tractor_beam_antimatter_rate
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "tractor_beam_range";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->tractor_beam_range = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %.4f ru",
		parm_name,
		legend->tractor_beam_range
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "transporter_range";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		legend->transporter_range = MAX(atof(val), 0.0f);
		NetSendLegend(-1);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %.4f ru",
		parm_name,
		legend->transporter_range
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	parm_name = "use_server_ai";
	if(CmdSysparmMatchParm(parm, parm_name))
	{
	    matches++;
	    if(!STRISEMPTY(val))
	    {
		sysparm.use_server_ai = StringIsYes(val);
	    }
	    sndbuf = (char *)malloc(STRLEN(parm_name) + 40 + 80);
	    sprintf(
		sndbuf,
		"%s = %s",
		parm_name,
		sysparm.use_server_ai ? "yes" : "no"
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}



	/* No matches? */
	if(matches == 0)
	{
	    sndbuf = (char *)malloc(STRLEN(parm) + 80);
	    sprintf(
		sndbuf,
"%s: Unsupported parameter.",
		parm
	    );
	    NetSendMessage(con_num, sndbuf);
	    free(sndbuf);
	}

	free(parm);
	free(val);

	return(0);
}
