#include <stdlib.h>
#include <string.h>

#include "net.h"
#include "swserv.h"


int ConIsAllocated(int condescriptor);
int ConIsConnected(int condescriptor);
int ConIsLoggedIn(int condescriptor);
int ConGetByObject(int obj_num);
int ConGetTop(void);

int ConNew(void);
void ConReset(int condescriptor);
void ConRecycle(int condescriptor);
void ConReclaim(void);
void ConDeleteAll(
        const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if condescriptor is allocated.
 */
int ConIsAllocated(int condescriptor)
{
	if((condescriptor < 0) ||
	   (condescriptor >= total_connections) ||
	   (connection == NULL)
	)
	    return(0);
	else if(connection[condescriptor] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Checks if the connection is connected, regardless of if the
 *	connection is logged in or if the socket is on close wait.
 */
int ConIsConnected(int condescriptor)
{
	if(!ConIsAllocated(condescriptor))
	    return(0);
	else if(connection[condescriptor]->socket < 0)
	    return(0);
	else
	    return(1);
}

/*
 *	Checks if the connection and logged in, if it's object is set.
 *
 *	Does not check if connection's object is valid.
 */
int ConIsLoggedIn(int condescriptor)
{
	sws_connection_struct *con;

	if(!ConIsConnected(condescriptor))
	    return(0);

	con = connection[condescriptor];
	if(con->flags & SWS_CON_FLAG_CLOSE_WAIT)
	    return(0);

	return((con->object_num > -1) ? 1 : 0);
}

/*
 *	Returns the first connection that is logged in and who's object
 *	matches the specified object.
 */
int ConGetByObject(int obj_num)
{
	int i;
	sws_connection_struct *con;

	if(obj_num < 0)
	    return(-1);

	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if((con->socket < 0) ||
	       (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
	    )
		continue;

	    if(con->object_num == obj_num)
		return(i);
	}

	return(-1);
}

/*
 *	Returns the highest connection number that is connected or
 *	-1 if there are no connections.
 */
int ConGetTop(void)
{
	int i, h;
	sws_connection_struct *con;

	for(i = 0, h = -1; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    h = i;
	}

	return(h);
}


/*
 *	Creates a new connection in the connections list.
 *
 *	Returns the connection number of -1 on error.
 *
 *	Global highest_connection will be updated as needed.
 */
int ConNew(void)
{
	int i, n;
	sws_connection_struct *con;

	/* Sanitize total */
	if(total_connections < 0)
	    total_connections = 0;

	/* Search for available structure already allocated */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		break;

	    if(con->socket < 0)
	        break;
	}
	if(i < total_connections)
	{
	    /* Got available pointer */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_connections;
	    total_connections = n + 1;
	    connection = (sws_connection_struct **)realloc(
	        connection,
	        total_connections * sizeof(sws_connection_struct *)
	    );
	    if(connection == NULL)
	    {
	        total_connections = 0;
		return(-1);
	    }

	    /* Reset new pointer to NULL so it gets allocated below */
	    connection[n] = NULL;
	}

	/* Allocate connection as needed */
	con = connection[n];
	if(con == NULL)
	{
	    connection[n] = con = (sws_connection_struct *)calloc(
		1, sizeof(sws_connection_struct)
	    );
	    if(con == NULL)
		return(-1);
	}

	/* Reset new connection */
	ConReset(n);

	/* Increment global highest connection as needed */
	if(n >= highest_connection)
	    highest_connection = n + 1;

	return(n);
}

/*
 *	Resets the connection's values to defaults.
 *
 *	If the socket is connected it will be reset but not closed.
 */
void ConReset(int condescriptor)
{
	sws_connection_struct *con;

	if(!ConIsAllocated(condescriptor))
	    return;

	con = connection[condescriptor];

	con->socket = -1;
	con->flags = 0;
	con->client_type = 0;

	free(con->conhost);
	con->conhost = NULL;
	con->object_num = -1;
	con->contime = 0l;
	con->bytes_received = 0;
	con->bytes_sent = 0;
	con->errors_received = 0;
	con->errors_sent = 0;

	con->badlogins = 0;

	con->obj_update_int = SWS_DEF_OBJECT_UPDATE_INT;
	con->obj_update_next = 0l;

	con->conbuf_items = 0;
}

/*
 *	Recycles the connection.
 *
 *	If the socket is connected it will be reset but not closed.
 */
void ConRecycle(int condescriptor)
{
	if(!ConIsAllocated(condescriptor))
	    return;

	ConReset(condescriptor);
}

/*
 *	Reclaims connections memory.
 */
void ConReclaim(void)
{
	int i, h;
	sws_connection_struct *con;

	if(total_connections < 0)
	    total_connections = 0;

	for(i = 0, h = -1; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket < 0)
		continue;

	    h = i;
	}

	for(i = total_connections - 1; i > h; i--)
	{
	    free(connection[i]);
	    connection[i] = NULL;
	}

	total_connections = h + 1;


	/* Reallocate or free pointer array */
	if(total_connections > 0)
	{
	    connection = (sws_connection_struct **)realloc(
	        connection,
	        total_connections * sizeof(sws_connection_struct *)
	    );
	    if(connection == NULL)
	    {
	        total_connections = 0;
		return;
	    }
	}
	else
	{
	    free(connection);
	    connection = NULL;
	    total_connections = 0;
	}
}

/*
 *	Deletes all connections.
 *
 *	Any connection who's socket is still connected will be
 *	forcably closed.
 */
void ConDeleteAll(
        const time_t cur_ms, const time_t cur_sec, const float tc  
)
{
	int i;
	sws_connection_struct *con;

	/* Logout, forceably close, and reset all connections first */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if(con->socket > -1)
	    {
		if(!(con->flags & SWS_CON_FLAG_CLOSE_WAIT))
		    NetShutdownConnection(
			i, SWS_CON_SHUTDOWN_RDWR, 1,
			cur_ms, cur_sec, tc
		    );
		NetCloseConnection(i, cur_ms, cur_sec, tc);
	    }

	    ConReset(i);
	}

	/* Delete all connections */
	for(i = 0; i < total_connections; i++)
	    free(connection[i]);
	free(connection);
	connection = NULL;
	total_connections = 0;
}
