#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/imgrefs.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"
#include "../include/swnetcodes.h"

#include "effects.h"
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


char DBIsObjectGarbage(int obj_num);

xsw_object_struct **DBReallocList(
	xsw_object_struct **list, const int total
);
xsw_object_struct *DBNewObject(void);
int DBCreateObject(
	int imgref_num,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int DBCreateExplicitObject(
	int obj_num,
	int imgref_num,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int DBCreateObjectByOPM(
	const char *opm_name,
	const char *name,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int DBCreateObjectByOPMPtr(
	xsw_object_struct *opm_ptr,
	const char *name,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

void DBDeleteObject(xsw_object_struct *obj);
void DBDeleteAllObjects(void);

void DBRecycleObject(
	const int obj_num,
        const time_t cur_ms, const time_t cur_sec, const float tc  
);

static void DBReclaimIteration(
	sw_legend_struct *legend,
	const int obj_num, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
void DBReclaim(
	const time_t cur_ms, const time_t cur_sec, const float tc
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if object number is valid, allocated and garbage.
 *
 *	Returns false if object is valid, allocated and non-garbage.
 */
char DBIsObjectGarbage(int obj_num)
{
	return(
	    UNVIsObjectGarbage(
		xsw_object,
		total_objects,
		obj_num
	    )
	);
}

/*
 *	Reallocates given object pointer array.
 *
 *	If num_objects <= 0 then cur_obj_ptrs is free'ed if it is not
 *	NULL.
 */
xsw_object_struct **DBReallocList(
	xsw_object_struct **list, const int total
)
{
	if(total > 0)
	{
	    return((xsw_object_struct **)realloc(
		list,
		total * sizeof(xsw_object_struct *)
	    ));
	}
	else
	{
	    free(list);
	    return(NULL);
	}
}

/*
 *	Creates a new object.
 */
xsw_object_struct *DBNewObject(void)
{
	xsw_object_struct *obj = SW_OBJECT(calloc(
	    1, sizeof(xsw_object_struct)
	));
	if(obj == NULL)
	    return(NULL);

	UNVResetObject(obj);

	return(obj);
}

/*
 *	Creates a new object in the database.
 *
 *	Returns the new object's index number or -1 on error.
 */
int DBCreateObject(
	int imgref_num,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, object_rtn, prev_total_objects;
	xsw_object_struct *obj;

	/* Type may not be garbage */
	if(type <= SW_OBJ_TYPE_GARBAGE)
	{
	    fprintf(
		stderr,
"DBCreateObject: Error: Useless request to create garbage.\n"
	    ); 
	    return(-1);
	}

	/* Is the imgref an invalid value? */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	{
	    fprintf(
		stderr,
"DBCreateObject: Error: ImgRef value %i out of range.\n",
		imgref_num
	    );
	    return(-1);
	}

	/* Look for an available object already in the list */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type == SW_OBJ_TYPE_GARBAGE)
		break;
	}
	if(i < total_objects)
	{
	    obj = xsw_object[i];

	    UNVResetObject(obj);
	    obj->type = type;
	    obj->owner = owner_obj_num;
	    obj->imgref = imgref_num;
	    obj->sect_x = sect_x;
	    obj->sect_y = sect_y;
	    obj->sect_z = sect_z;
	    obj->x = x;
	    obj->y = y;
	    obj->z = z;
	    obj->heading = heading;
	    obj->pitch = pitch;
	    obj->bank = bank;
	    SWMathSetUnitVector2D(
		&obj->attitude_vector_compoent,
		obj->heading
	    );
	    obj->birth_time = cur_ms;
	    obj->last_updated = cur_ms;
	    obj->animation.last_interval = cur_ms;

	    return(i);
	}

	/* No garbage objects, can we allocate more? */
	if((total_objects + SWS_OBJECT_ALLOCATE_AHEAD) > SW_OBJECTS_MAX)
	    return(-1);

	/* Allocate more objects in memory */
	prev_total_objects = total_objects;
	total_objects += SWS_OBJECT_ALLOCATE_AHEAD;

	/* Allocate pointer array */
	xsw_object = DBReallocList(xsw_object, total_objects);
	if(xsw_object == NULL)
	{
	    total_objects = 0;
	    return(-3);
	}

	/* Allocate each object */
	for(i = prev_total_objects; i < total_objects; i++)
	{
	    xsw_object[i] = DBNewObject();
	    if(xsw_object[i] == NULL)
	    {
		total_objects = i;
		return(-3);
	    }
	}

	/* New object will be the first object allocated above */
	object_rtn = prev_total_objects;

	/* Set values on the new object */
	obj = xsw_object[object_rtn];
	obj->type = type;
	obj->owner = owner_obj_num;
	obj->imgref = imgref_num;
	obj->sect_x = sect_x;
	obj->sect_y = sect_y;
	obj->sect_z = sect_z;
	obj->x = x;
	obj->y = y;
	obj->z = z;
	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	SWMathSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);
	obj->birth_time = cur_ms;
	obj->last_updated = cur_ms;
	obj->animation.last_interval = cur_ms;

	return(object_rtn);
}

/*
 *	Creates a new object in the database at the index specified by
 *	object_num.
 *
 *	If the object already exists then it will be recycled and
 *	recreated.
 *
 *	Returns the new object's index number or -1 on error.
 */
int DBCreateExplicitObject(
	int obj_num,
	int imgref_num,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, obj_num_rtn;
	xsw_object_struct *obj;

	/* Type may not be garbage or error */
	if(type <= SW_OBJ_TYPE_GARBAGE)
	{
	    fprintf(
		stderr,
"DBCreateExplicitObject(): Useless request to create garbage.\n"
	    );
	    return(-1);
	}

	/* Is the object number an invalid value? */
	if((obj_num < 0) || (obj_num >= SW_OBJECTS_MAX))
	{
	    fprintf(
		stderr,
"DBCreateExplicitObject(): Object number %i out of range.\n",
		obj_num
	    );
	    return(-1);
	}

	/* Is the imgref an invalid value? */
	if((imgref_num < 0) || (imgref_num >= IMGREF_MAX_IMGREFS))
	{
	    fprintf(
		stderr,
"DBCreateExplicitObject(): Error: ImgRef value %i out of range.\n",
		imgref_num
	    );
	    return(-1);
	}

	/* If obj_num is not garbage then recycle it first */
	if(!DBIsObjectGarbage(obj_num))
	    DBRecycleObject(obj_num, cur_ms, cur_sec, tc);

	/* Allocate memory if obj_num is greater than total_objects */
	if(obj_num >= total_objects)
	{
	    /* Can we allocate more? */
	    if((obj_num + SWS_OBJECT_ALLOCATE_AHEAD) >= SW_OBJECTS_MAX)
	    {
		fprintf(
		    stderr,
"DBCreateExplicitObject(): Maximum objects %i reached.\n",
		    SW_OBJECTS_MAX
		);
		return(-3);
	    }

	    /* Reallocate objects pointer array */ 
	    xsw_object = DBReallocList(
		xsw_object,
		obj_num + SWS_OBJECT_ALLOCATE_AHEAD + 1
	    );
	    if(xsw_object == NULL)
	    {
		total_objects = 0;
		return(-3);
	    }

	    /* Allocate each new object */
	    for(i = total_objects;
		i <= (obj_num + SWS_OBJECT_ALLOCATE_AHEAD);
		i++
	    )
	    {
		xsw_object[i] = (xsw_object_struct *)DBNewObject();
		if(xsw_object[i] == NULL)
		{
		    total_objects = i;
		    return(-3);
		}
	    }

	    /* Adjust global total */
	    i = total_objects;
	    total_objects = obj_num + SWS_OBJECT_ALLOCATE_AHEAD + 1;

	    /* Reset newly created objects */
	    while(i < total_objects)
	    {
		UNVResetObject(xsw_object[i]);
		i++;
	    } 
	}

	/* Create explicit object */
	obj_num_rtn = obj_num;

	/* Reset values on the new object */
	obj = xsw_object[obj_num_rtn];
	UNVResetObject(obj);
	obj->type = type; 
	obj->owner = owner_obj_num;
	obj->imgref = imgref_num;
	obj->sect_x = sect_x;
	obj->sect_y = sect_y;
	obj->sect_z = sect_z;
	obj->x = x;
	obj->y = y;
	obj->z = z;
	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	SWMathSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);
	obj->birth_time = cur_ms;
	obj->last_updated = cur_ms;
	obj->animation.last_interval = cur_ms;

	return(obj_num_rtn);
}


/*
 *	Creates an object and models it to the specified OPM.
 */
int DBCreateObjectByOPM(
	const char *opm_name,
	const char *name,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int opm_num;

	if((opm_name == NULL) || (type <= SW_OBJ_TYPE_GARBAGE))
	    return(-1);

	/* Get the OPM */
	opm_num = OPMGetByName(opm_name, type);
	if(OPMIsGarbage(opm_num))
	{
	    fprintf(
		stderr,
		"DBCreateObjectByOPM(): %s: No such OPM.\n",
		opm_name
	    );
	    return(-2);
	}

	return(DBCreateObjectByOPMPtr(
	    opm[opm_num], name, type,
	    owner_obj_num,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    heading, pitch, bank,
	    cur_ms, cur_sec, tc
	));
}

/*
 *	Creates an object and models it to the specified OPM.
 */
int DBCreateObjectByOPMPtr(
	xsw_object_struct *opm_ptr,
	const char *name,
	int type,
	int owner_obj_num,
	long sect_x, long sect_y, long sect_z,
	float x, float y, float z,
	float heading, float pitch, float bank,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int obj_num;
	xsw_object_struct *obj;

	if((opm_ptr == NULL) || (type <= SW_OBJ_TYPE_GARBAGE))
	    return(-1);

	/* Create the new object */
	obj_num = DBCreateObject(
	    IMGREF_DEFAULT_IMGREF,
	    type,
	    owner_obj_num,
	    sect_x, sect_y, sect_z,
	    x, y, z,
	    heading, pitch, bank,
	    cur_ms, cur_sec, tc
	);
	if(DBIsObjectGarbage(obj_num))
	    return(-3);

	obj = xsw_object[obj_num];

	/* Set model the new object */
	OPMModelObjectPtr(obj, opm_ptr);

	/* Set values */
	strncpy(
	    obj->name, (name != NULL) ? name : "",
	    sizeof(obj->name)
	);
	obj->name[sizeof(obj->name) - 1] = '\0';

	obj->owner = owner_obj_num;

	obj->sect_x = sect_x;
	obj->sect_y = sect_y;
	obj->sect_z = sect_z;

	obj->x = x; 
	obj->y = y;
	obj->z = z;

	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;

	SWMathSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);

	obj->birth_time = cur_ms;
	obj->last_updated = cur_ms;
	obj->animation.last_interval = cur_ms;

	return(obj_num);
}

/*
 *	Deletes the object.
 */
void DBDeleteObject(xsw_object_struct *obj)
{
	UNVDeleteObject(obj);
}

/*
 *	Deletes all objects.
 */     
void DBDeleteAllObjects(void)
{
	UNVDeleteAllObjects(xsw_object, total_objects);
	xsw_object = NULL;
	total_objects = 0;
}

/*
 *	Recycles the object.
 *
 *	If the object is a player then all of its posessions will be
 *	be chowned to the global lost and found owner.
 *
 *	If the object is a player and is still connected then the
 *	connection will be disconnected first.
 *
 *	All members on the object will be deleted and reset.
 *
 *	The objects type will be set to garbage.
 */
void DBRecycleObject(
	const int obj_num,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	xsw_object_struct *obj;
	sw_legend_struct *legend = &sw_legend;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* If this is a player then perform additional checks */
	if(obj->type == SW_OBJ_TYPE_PLAYER)
	{
	    int i;
	    sws_connection_struct *con;

	    /* Change ownership of all objects owned by this player
	     * object to the global lost_found_owner object
	     */
	    if(!DBIsObjectGarbage(legend->lost_found_owner))
	    {
		char *name1, *name2;
		xsw_object_struct *tobj;

		name2 = STRDUP(DBGetFormalNameStr(
		    legend->lost_found_owner
		));

		for(i = 0; i < total_objects; i++)
		{
		    tobj = xsw_object[i];
		    if(tobj == NULL)
			continue;

		    if(tobj->type <= SW_OBJ_TYPE_GARBAGE)
			continue;

		    if(tobj->owner != obj_num)
			continue;

		    /* Skip object itself */
		    if(i == obj_num)
			continue;

		    tobj->owner = legend->lost_found_owner;

		    name1 = STRDUP(DBGetFormalNameStr(i));

		    if(sysparm.log_general)
		    {
			char *s = (char *)malloc(
			    STRLEN(name1) + STRLEN(name2) + 80
			);
			sprintf(
			    s,
"Chowned %s to %s due to recycling of previous owner.",
			    name1, name2
			);
			LogAppend(fname.log, s);
			free(s);
		    }

		    free(name1);
		}

		free(name2);
	    }
	    /* If no valid lost and found owner then set to owner 0 */
	    else
	    {
		xsw_object_struct *tobj;

		for(i = 0; i < total_objects; i++)
		{
		    tobj = xsw_object[i];
		    if(tobj == NULL)
			continue;
		    if(tobj->type <= SW_OBJ_TYPE_GARBAGE)
			continue;
		    if(tobj->owner != obj_num)
			continue;

		    /* Skip object itself */
		    if(i == obj_num)
			continue;

		    tobj->owner = 0;

		    if(sysparm.log_general)
		    {
			char	*name = STRDUP(DBGetFormalNameStr(i)),
				*s = (char *)malloc(STRLEN(name) + 80);
			sprintf(
			    s,
"Chowned %s to #0 due to recycling of previous owner.",
			    name
			);
			free(name);
			LogAppend(fname.log, s);
			free(s);
		    }
		}
	    }


	    /* Logout all connections logged in as this object */
	    for(i = 0; i < total_connections; i++)
	    {
		con = connection[i];
		if(con == NULL)
		    continue;

		if((con->socket < 0) ||
		   (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
		)
		    continue;

		if(con->object_num != obj_num)
		    continue;

		/* Mark the connection as having no object (not logged
		 * in) and then shut it down, this is to prevent
		 * NetShutdownConnection() from calling this function
		 * again to recycle the object if it was a guest
		 */
		con->object_num = -1;
		NetShutdownConnection(
		    i, SWS_CON_SHUTDOWN_WR, 1,
		    cur_ms, cur_sec, tc
		);
	    }
	}

	/* Reset and delete all allocated substructures on this
	 * object
	 */
	UNVResetObject(obj);

	/* Set this object to type garbage to mark it as recycled */  
	obj->type = SW_OBJ_TYPE_GARBAGE;
}


/*
 *	Recycles non-connected Guest objects.
 *
 *	Fixes the object's connected and hidden flags (if it is a
 *	player).
 *
 *	Fixes its on fire state (if it is a player or controlled
 *	object).
 *
 *	Fixes hp and power values.
 */
static void DBReclaimIteration(
	sw_legend_struct *legend,
	const int obj_num, xsw_object_struct *obj,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int type;

	if(obj == NULL)
	    return;

	type = obj->type;
	if(type <= SW_OBJ_TYPE_GARBAGE)
	    return;

	/* Invalid owner or owner not set? */
	if(DBGetPlayerOwner(obj_num) < 0)
	{
	    const int prev_owner = obj->owner;
	    int new_owner;

	    if(type == SW_OBJ_TYPE_PLAYER)
		new_owner = obj_num;
	    else
		new_owner = (legend->lost_found_owner > -1) ?
		    legend->lost_found_owner : 0;

	    obj->owner = new_owner;

	    if(sysparm.log_errors)
	    {
		char	*name1 = STRDUP(DBGetFormalNameStr(obj_num)),
			*name2 = STRDUP(DBGetFormalNameStr(new_owner));
		char *s = (char *)malloc(
		    STRLEN(name1) + STRLEN(name2) + 40 + 80
		);
		sprintf(
		    s,
"Chowned `%s' to `%s' due to its previous owner #%i being invalid.",
		    name1, name2, prev_owner
		);
		LogAppend(fname.log, s);
		free(s);
		free(name1);
		free(name2);
	    }
	}

	/* Player? */
	if(type == SW_OBJ_TYPE_PLAYER)
	{
	    /* Is this a stray guest object that is not connected? */
	    if(strpfx(obj->name, sysparm.guest_login_name))
	    {
		/* Not connected? */
		if(ConGetByObject(obj_num) < 0)
		{
		    if(sysparm.log_general || sysparm.log_errors)
		    {
			const char *name = DBGetFormalNameStr(obj_num);
			char *s = (char *)malloc(STRLEN(name) + 80);
			sprintf(
			    s,
"Recycled unconnected guest object `%s'.",
			    name
			);
			LogAppend(fname.log, s);
			free(s);
		    }

		    /* Recycle this object */
		    DBRecycleObject(obj_num, cur_ms, cur_sec, tc);

		    /* Notify all connections about this recycle */
		    NetSendRecycleObject(-1, obj_num);

		    return;
		}
	    }

	    /* Hide players at home objects? */
	    if(sysparm.hide_players_home)
	    {
		/* Not connected? */
		if(ConGetByObject(obj_num) < 0)
		{
		    /* Not marked hidden yet? */
		    if(!(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON))
		    {
			int i;
			xsw_object_struct *tar_obj;

			/* Search for a home object that this
			 * player is at
			 */
			for(i = 0; i < total_objects; i++)
			{
			    tar_obj = xsw_object[i];
			    if(tar_obj == NULL)
				continue;

			    if(tar_obj->type != SW_OBJ_TYPE_HOME)
				continue;

			    if(i == obj_num)
				continue;

			    /* At this home object? */
			    if(SWMath3DInContactPtr(obj, tar_obj))
			    {
				/* Mark this player as hidden */
				obj->ext_flags |= SWS_OBJ_FLAG_HIDEFROMCON;
				break;
			    }
			}
		    }
		}
		else
		{
		    /* Object is connected */

		    /* Is hidden? */
		    if(obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    {
			/* Mark as not hidden */
			obj->ext_flags &= ~SWS_OBJ_FLAG_HIDEFROMCON;
		    }

		    /* Not marked as connected? */
		    if(!(obj->ext_flags & SWS_OBJ_FLAG_CONNECTED))
		    {
			/* Mark as connected */
			obj->ext_flags |= SWS_OBJ_FLAG_CONNECTED;
		    }
		}
	    }
	}

	/* Player or controlled? */
	if((type == SW_OBJ_TYPE_PLAYER) ||
	   (type == SW_OBJ_TYPE_CONTROLLED)
	)
	{
	    /* Update the object's on fire state */
	    if(obj->hp_max > 0.0f)
		EffectsSetObjectOnFire(
		    obj_num, obj,
		    ((obj->hp / obj->hp_max) <= legend->on_fire_hp_coeff) ? 1 : 0
		);
	    else
		EffectsSetObjectOnFire(obj_num, obj, 0);
	}

	/* General checks for all object types */

	/* Clip the hit points */
	if(obj->hp > obj->hp_max)
	    obj->hp = obj->hp_max;

	/* Clip the power */
	if(obj->power_max < 0.0f)
	    obj->power_max = 0.0f;
	if(obj->power > obj->power_max)
	    obj->power = obj->power_max;
	if(obj->power < 0.0f)
	    obj->power = 0.0f;

}

/*
 *	Performs maintainance on the universe database.
 *
 *	Reallocates unused object pointers.
 *
 *	Updates player object connected and hidden flags.
 *
 *	Sanitizes hp and power values on all objects.
 */        
void DBReclaim(
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, highest_obj_num;
	sw_legend_struct *legend = &sw_legend;

	if(total_objects < 0)
	    total_objects = 0;

	if((total_objects == 0) || (xsw_object == NULL))
	    return;

	/* Get the highest non-garbage object */
	highest_obj_num = DBGetTopObjectNumber();
	if((highest_obj_num >= 0) && (highest_obj_num < total_objects))
	{
	    /* Delete all allocated objects from highest_obj_num + 1
	     * to total_objects - 1
	     */
	    for(i = total_objects - 1; i > highest_obj_num; i--)
	    {
		DBDeleteObject(xsw_object[i]);
		xsw_object[i] = NULL;
	    }

	    /* Reallocate object pointers */
	    total_objects = highest_obj_num + 1;
	    if(total_objects > 0)
	    {
	        xsw_object = DBReallocList(
		    xsw_object, total_objects
	        );
		if(xsw_object == NULL)
		{
		    total_objects = 0;
		    return;
	        }
	    }
	    else
	    {
		free(xsw_object);
		xsw_object = NULL;
		total_objects = 0;
	    }
	}


	/* Perform maintainance on each object */
	for(i = 0; i < total_objects; i++)
	    DBReclaimIteration(
		legend,
		i, xsw_object[i],
		cur_ms, cur_sec, tc
	    );
}
