#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include "../include/disk.h"
#include "../include/unvfile.h"

#include "sysparm.h"
#include "swserv.h"


int DBOpenFromFile(
	const char *path,
	const time_t cur_ms, const time_t cur_sec, const float tc
);
int DBSaveToFile(const char *path);
int DBEmergencySave(xsw_object_struct **list, const int total);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


int DBOpenFromFile(
	const char *path,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int total = 0;
	sw_legend_struct *legend = &sw_legend;

	/* Delete all objects currently in memory */
	DBDeleteAllObjects();

	/* Load objects from the universe file into the universe
	 * database
	 */
	xsw_object = UNVFileOpen(
	    path,
	    &total,
	    legend,
	    NULL,	/* Client data */
	    NULL	/* Notify callback function */
	);
	total_objects = total;

	/* Delete unused object pointers, update object hidden states
	 * and sanitize values
	 */
	DBReclaim(cur_ms, cur_sec, tc);

	return(0);
}


int DBSaveToFile(const char *path)
{
	const char *parent;
	mode_t mode;
	int status;
	sw_legend_struct *legend = &sw_legend;

	/* Update universe legend information */
	strcpy(legend->version, PROG_VERSION);

	parent = GetParentDir(path);
	if(parent != NULL)
	{
	    struct stat stat_buf;
	    if(stat(parent, &stat_buf))
		mode = SWSCurrentFileMode();
	    else
		mode = stat_buf.st_mode &
		    (S_IRUSR | S_IWUSR |
		     S_IRGRP | S_IWGRP |
		     S_IROTH | S_IWOTH
		    );
	}
	else
	{
	    mode = SWSCurrentFileMode();
	}

	/* Save to file */
	status = UNVFileSave(
	    path, mode,
	    xsw_object,
	    total_objects,
	    legend,
	    NULL,	/* Client data */
	    NULL	/* Notify callback function */
	);

	return(status);
}


static int DBEmergencySavePlanA(
	xsw_object_struct **list, const int total
)
{
	int status;
	char *full_path;

	fprintf(
	    stderr,
	    "%s\n",
"*** Server: Emergency database save in progress using \"Plan A\"..."
	);
	if(sysparm.log_errors)
	    LogAppend(
		fname.log,
"*** Server: Emergency database save in progress using \"Plan A\"..."
	    );

	if(STRISEMPTY(fname.unv_out))
	{
	    char *path = STRDUP(PrefixPaths(
		dname.db, "untitled.unv"
	    ));
	    full_path = (char *)malloc(STRLEN(path) + 40 + 80);
	    sprintf(
		full_path,
		"%s.EMERGENCY.%i",
		path, getpid()
	    );
	    free(path);
	}
	else
	{
	    const char *path = fname.unv_out;
	    full_path = (char *)malloc(STRLEN(path) + 40 + 80);
	    sprintf(
		full_path,
		"%s.EMERGENCY.%i",
		path, getpid()
	    );
	}

	/* Attempt to save */
	status = DBSaveToFile(full_path);
	if(status)
	{
	    fprintf(
		stderr,
		"%s\n",
"*** Server: Emergency database save using \"Plan A\" has failed."
	    );
	    if(sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Server: Emergency database save using \"Plan A\" has failed."
		);
	}
	else
	{
	    fprintf(
		stderr,
"*** Server: Database sucessfully saved to `%s'.\n",
		full_path
	    );
	    if(sysparm.log_errors)
	    {
		char *s = (char *)malloc(
		    STRLEN(full_path) + 80
		);
		sprintf(
		    s,
"*** Server: Database sucessfully saved to `%s'.",
		    full_path
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	}

	free(full_path);

	return(status);
}

static int DBEmergencySavePlanB(
	xsw_object_struct **list, const int total
)
{
	int status, fd;
	char *full_path;
	const char *parent = getenv("TMPDIR");

	fprintf(
	    stderr,
	    "%s\n",
"*** Server: Emergency database save in progress using \"Plan B\"..."
	);
	if(sysparm.log_errors)
	    LogAppend(
		fname.log,
"*** Server: Emergency database save in progress using \"Plan B\"..."
	    );

	if(parent == NULL)
#if defined(P_tmpdir)
	    parent = P_tmpdir;
#elif defined(WIN32)
	    parent = "C:\\TEMP";
#else
	    parent = "/tmp";
#endif
	full_path = STRDUP(PrefixPaths(
	    parent, "emergency.unv.XXXXXX"
	));
	fd = mkstemp(full_path);
	if(fd > -1)
	{
	    close(fd);
	}
	else
	{
	    free(full_path);
	    return(-1);
	}

	/* Attempt to save */
	status = DBSaveToFile(full_path);
	if(status)
	{
	    fprintf(
		stderr,
		"%s\n",
"*** Server: Emergency database save using \"Plan B\" has failed."
	    );
	    if(sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Server: Emergency database save using \"Plan B\" has failed."
		);
	}
	else
	{
	    fprintf(
		stderr,
"*** Server: Database sucessfully saved to `%s'.\n",
		full_path
	    );
	    if(sysparm.log_errors)
	    {
		char *s = (char *)malloc(
		    STRLEN(full_path) + 80
		);
		sprintf(
		    s,
"*** Server: Database sucessfully saved to `%s'.",
		    full_path
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	}

	free(full_path);

	return(status);
}


int DBEmergencySave(xsw_object_struct **list, const int total)
{
	/* Plan A */
	if(!DBEmergencySavePlanA(list, total))
	    return(0);

	/* Plan B */
	if(!DBEmergencySavePlanB(list, total))
	    return(0);

	return(-1);
}
