#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvutil.h"

#include "swserv.h"
#include "sysparm.h"


void DBPurgeRecycledObject(const int rec_obj_num);
void DBPurgeAllRecycledObjects(void);

int DBSaveRecycledObject(const int obj_num);

int DBRecoverRecycledObject(
	const int rec_obj_num, const int owner_obj_num
);
int DBRecoverRecycledObjectByName(
	const char *name, const int owner_obj_num
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Deletes the specified object from the recycled objects list.
 */
void DBPurgeRecycledObject(const int rec_obj_num)
{
	int i;

	if((rec_obj_num < 0) || (rec_obj_num >= total_recycled_objects))
	    return;

	DBDeleteObject(recycled_xsw_object[rec_obj_num]);

	total_recycled_objects--;

	for(i = rec_obj_num; i < total_recycled_objects; i++)
	    recycled_xsw_object[i] = recycled_xsw_object[i + 1];

	if(total_recycled_objects > 0)
	{
	    recycled_xsw_object = DBReallocList(
		recycled_xsw_object, total_recycled_objects
	    );
	    if(recycled_xsw_object == NULL)
		total_recycled_objects = 0;
	}
	else
	{
	    free(recycled_xsw_object);
	    recycled_xsw_object = NULL;
	    total_recycled_objects = 0;
	}
}

/*
 *	Deletes the recycled objects list.
 */
void DBPurgeAllRecycledObjects(void)
{
	int i;

	for(i = 0; i < total_recycled_objects; i++)
	    DBDeleteObject(recycled_xsw_object[i]);
	free(recycled_xsw_object);
	recycled_xsw_object = NULL;
	total_recycled_objects = 0;
}


/*
 *	Coppies the object specified by obj_num from the database to
 *	the recycled objects list.
 *
 *	The specified object in the database will not actually be
 *	recycled or modified in any way.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	General error.
 *	-2	Bad value.
 *	-3	Systems error or memory allocation error.
 *	-4	Operation aborted.
 *	-5	sysparm.recycled_objects_max is set to 0.
 */
int DBSaveRecycledObject(const int obj_num)
{
	xsw_object_struct *obj;

	if(sysparm.recycled_objects_max <= 0)
	    return(-5);

	if(DBIsObjectGarbage(obj_num))
	    return(-2);

	obj = xsw_object[obj_num];

	/* Need to purge the last recycled object to make room for
	 * this object?
	 */
	if(total_recycled_objects >= sysparm.recycled_objects_max)
	{
	    const int i_last = total_recycled_objects - 1;
	    int i;

	    /* Delete the last recycled object */
	    DBDeleteObject(recycled_xsw_object[i_last]);

	    /* Shift pointers */
	    for(i = i_last; i > 0; i--)
		recycled_xsw_object[i] = recycled_xsw_object[i - 1];

	    /* Copy the specified object to the first index of the
	     * ercycled objects list
	     */
	    recycled_xsw_object[0] = UNVDupObject(obj);
	}
	else
	{
	    /* Increase allocation */
	    const int i_last = MAX(total_recycled_objects, 0);
	    int i;

	    total_recycled_objects = i_last + 1;

	    recycled_xsw_object = DBReallocList(
		recycled_xsw_object, total_recycled_objects
	    );
	    if(recycled_xsw_object == NULL)
		return(-3);

	    /* Shift pointers */
	    for(i = i_last; i > 0; i--)
		recycled_xsw_object[i] = recycled_xsw_object[i - 1];

	    /* Copy the specified object to the first index of the
	     * ercycled objects list
	     */
	    recycled_xsw_object[0] = UNVDupObject(obj);
	}

	return(0);
}


/*
 *	Recovers the specified recycled object to the database.
 *
 *	Does not purge (deletes) the specified recycled object.
 *
 *	Returns the index of the recovered database object or -1 on
 *	error.
 */
int DBRecoverRecycledObject(
	const int rec_obj_num, const int owner_obj_num
)
{
	int i, obj_num;
	xsw_object_struct *obj, *rec_obj;

	rec_obj = ((rec_obj_num >= 0) && (rec_obj_num < total_recycled_objects)) ?
	    recycled_xsw_object[rec_obj_num] : NULL;
	if(rec_obj == NULL)
	    return(-2);

	/* Look for an available pointer in the database, either an
	 * allocated object that is garbage or a NULL pointer
	 */
	for(i = 0; i < total_objects; i++)
	{
	    if(DBIsObjectGarbage(i))
		break;
	}
	if(i < total_objects)
	{
	    /* Found a garbage object or a NULL pointer, delete the
	     * object as needed and copy the recycled object to the
	     * database object
	     */
	    obj_num = i;
	    DBDeleteObject(xsw_object[obj_num]);
	    xsw_object[obj_num] = obj = UNVDupObject(rec_obj);
	}
	else
	{
	    /* No garbage objects or NULL pointers, append an object to
	     * the database and copy the recycled object to the
	     * appended object
	     */
	    obj_num = i = MAX(total_objects, 0);
	    total_objects = i + 1;
	    xsw_object = DBReallocList(xsw_object, total_objects);
	    if(xsw_object == NULL)
		return(-3);
	    xsw_object[obj_num] = obj = UNVDupObject(rec_obj);
	}

	if(obj == NULL)
	    return(-3);

	/* Set owner */
	if(obj->type == SW_OBJ_TYPE_PLAYER)
	    obj->owner = obj_num;
	else
	    obj->owner = owner_obj_num;

	return(obj_num);
}

/*
 *	Recovers a recycled object that matches the specified name.
 *
 *	The owner_obj_num specifies the object performing this
 *	recovery and will be the owner of the recovered object.
 *
 *	The recycled object (if successfully recovered) will be purged.
 *
 *	Returns the index of the recovered database object or one of
 *	the following on error:
 *
 *	-1	General error.
 *	-2	Bad value or no such recycled object found.
 *	-3	Systems error or memory allocation error.
 *	-7	Insufficient permission to recover.
 */
int DBRecoverRecycledObjectByName(
	const char *name, const int owner_obj_num
)
{
	int i, owner_uid, rec_obj_num, obj_num;
	xsw_object_struct *obj, *owner_obj;

	if(STRISEMPTY(name) || DBIsObjectGarbage(owner_obj_num))
	    return(-2);

	owner_obj = xsw_object[owner_obj_num];

	/* Get the owner's uid level */
	owner_uid = owner_obj->permission.uid;

	/* Is the owner not allowed to recover? */
	if(owner_uid > SWS_ACCESS_UID_RECOVER)
	    return(-7);

	/* Search for recycled objects starting from the oldest */
	for(i = total_recycled_objects - 1; i >= 0; i--)
	{
	    obj = recycled_xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Name does not match? (case sensitive) */
	    if(strstr(obj->name, name) == NULL)
		continue;

	    /* Matched this recycled object
	     *
	     * Is it a player?
	     */
	    if(obj->type == SW_OBJ_TYPE_PLAYER)
	    {
		/* This is a player, check if the owner has no create
		 * player permission or is not allowed to recover
		 * other object
		 */
		if((owner_uid > SWS_ACCESS_UID_CREATEPLAYER) ||
		   (owner_uid > SWS_ACCESS_UID_RECOVERO)
		)
		    return(-7);
	    }
	    else
	    {
		/* Not the owner of this object? */
		if(obj->owner != owner_obj_num)
		{
		    /* Check if the owner is not allowed to recover
		     * other's objects
		     */
		    if(owner_uid > SWS_ACCESS_UID_RECOVERO)
			return(-7);
		}
	    }

	    break;
	}
	rec_obj_num = i;

	/* No recycled object found? */
	if((rec_obj_num < 0) ||
	   (rec_obj_num >= total_recycled_objects)
	)
	    return(-2);

	/* Attempt to recover the object */
	obj_num = DBRecoverRecycledObject(rec_obj_num, owner_obj_num);

	/* Recovery successful? */
	if(!DBIsObjectGarbage(obj_num))
	{
	    /* Purge the recycled object */
	    DBPurgeRecycledObject(rec_obj_num);
	}

	return(obj_num);
}
