#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "sysparm.h"
#include "swserv.h"


int DBGetPlayerOwner(const int obj_num);
int DBGetTopObjectNumber(void);

int DBObjectTractor(const int obj_num, const int tar_obj_num);
int DBIsObjectTractorablePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj
);
void DBObjectUntractor(const int obj_num, const int tar_obj_num);

int DBObjectDoSetSector(
	const int obj_num,
	const long sect_x, const long sect_y, const long sect_z,
	const char allow_wrapping
);

int DBCreateObjectEconomyProduct(
	const int obj_num,
	const char *product_name,
	const int ocs_code,
	const sw_credits_t sell_price,
	const sw_credits_t buy_price,
	const float amount, const float amount_max
);
void DBSortEconomyProducts(const int obj_num);

int DBGetObjectNumByPtr(xsw_object_struct *obj);
const char *DBGetTypeName(const int type);
const char *DBGetFormalNameStr(const int obj_num);
const char *DBGetOCSOPMName(const int ocs_num);

int DBValidateObjectName(const char *name);
int DBValidateObjectPassword(const char *password);

float DBGetObjectVisibility(const int obj_num);
float DBGetObjectVisibilityPtr(xsw_object_struct *obj);

float DBGetObjectEffectiveScannerRange(const int obj_num);
float DBGetObjectEffectiveScannerRangePtr(xsw_object_struct *obj);

char DBIsObjectInEffectiveScannerRange(
	const int obj_num, const int tar_obj_num,
	const float scanner_range
);
char DBIsObjectInEffectiveScannerRangePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj,
	const float scanner_range
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the player object that owns the specified object (not
 *      always just the immediate owner).
 */
int DBGetPlayerOwner(const int obj_num)
{
	const int max_recusions = 256;
	int level = 0;
	int i = obj_num;
	xsw_object_struct *obj;

	while(!DBIsObjectGarbage(i) && (level < max_recusions))
	{
	    obj = xsw_object[i];

	    if(obj->type == SW_OBJ_TYPE_PLAYER)
		return(i);

	    /* If a non-player object owns itself then this is an
	     * error, return -1
	     */
	    if(i == obj->owner)
		return(-1);

	    i = obj->owner;
	    level++;
	}

	return(-1);
}

/*
 *	Returns the index of the non-garbage object with the highest
 *	index or -1 if there are no objects found.
 */
int DBGetTopObjectNumber(void)
{
	const int total = total_objects;
	int i, h = -1;
	xsw_object_struct *obj, **list = xsw_object;

	for(i = 0; i < total; i++)
	{
	    obj = list[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    h = i;
	}

	return(h);
}


/*
 *      Sets the object to lock tractor beam on the target object.
 *
 *	If tar_obj_num is -1 then the object's tractor beam will be
 *	unlocked.
 */
int DBObjectTractor(const int obj_num, const int tar_obj_num)
{
	int n;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	obj = xsw_object[obj_num];

	/* Unlock tractor beam? */
	if(tar_obj_num < 0)
	{
	    free(obj->tractored_object);
	    obj->tractored_object = NULL;
	    obj->total_tractored_objects = 0;
	    return(0);
	}


	/* tar_obj is assumed valid */

	/* Is the object already in the list? */
	for(n = 0; n < obj->total_tractored_objects; n++)
	{
	    if(obj->tractored_object[n] == tar_obj_num)
		return(0);
	}

	/* Sanitize total */
	if(obj->total_tractored_objects < 0)
	    obj->total_tractored_objects = 0;

	/* Maximum tractorable objects reached? */
	if(obj->total_tractored_objects >= SW_TRACTOR_BEAM_MAX_OBJECTS)
	    return(-3);

	/* Add to the tractored objects list */
	n = obj->total_tractored_objects;
	obj->total_tractored_objects = n + 1;
	obj->tractored_object = (int *)realloc(
	    obj->tractored_object,
	    obj->total_tractored_objects * sizeof(int)
	);
	if(obj->tractored_object == NULL)
	{
	    obj->total_tractored_objects = 0;
	    return(-3);
	}

	/* Set new tractored object */
	obj->tractored_object[n] = tar_obj_num;

	return(0);
}


/*
 *	Checks if the object can tractor the target object.
 */
int DBIsObjectTractorablePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj
)
{
	sw_legend_struct *legend = &sw_legend;

	if((obj == NULL) || (tar_obj == NULL))
	    return(0);

	/* These object types may not be tractored */
	if((tar_obj->type == SW_OBJ_TYPE_HOME) ||
	   (tar_obj->type == SW_OBJ_TYPE_AREA) ||
	   (tar_obj->type == SW_OBJ_TYPE_WORMHOLE) ||
	   (tar_obj->type == SW_OBJ_TYPE_ELINK)
	)
	    return(0);

	/* Check if objects are in range */
	if(!SWMath3DInRangePtr(obj, tar_obj, legend->tractor_beam_range))
	    return(0);

	/* Object may not tractor while cloaked */
	if(obj->cloak_state == SW_CLOAK_STATE_UP)
	    return(0);

	/* Check if target object has shields up and if so, check
	 * if target object's power is greater than the source
	 * object's
	 */
	if((tar_obj->shield_state == SW_SHIELD_STATE_UP) &&
	   ((tar_obj->power * tar_obj->power_purity) >=
	    (obj->power * obj->power_purity)
	   )
	)
	    return(0);


	return(1);
} 

/*
 *	Untractors the target object.
 */
void DBObjectUntractor(int obj_num, int tar_obj_num)
{
	int i, n;
	xsw_object_struct *obj;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Sanitize total */
	if(obj->total_tractored_objects < 0)
	    obj->total_tractored_objects = 0;

	/* If tar_obj_num is -1 then unlock all tractor beam objects
	 * for now
	 */
	if(tar_obj_num < 0)
	{
	    free(obj->tractored_object);
	    obj->tractored_object = NULL;
	    obj->total_tractored_objects = 0;
	    return;
	}

	/* tar_obj_num is assumed valid */

	/* Untractor object */
	for(i = 0; i < obj->total_tractored_objects; i++)
	{
	    if(obj->tractored_object[i] == tar_obj_num)
		obj->tractored_object[i] = -1;
	}

	/* Reclaim tractored objects memory */
	n = -1;
	for(i = 0; i < obj->total_tractored_objects; i++)
	{
	    if(DBIsObjectGarbage(obj->tractored_object[i]))
		continue;
	    else
		n = i;
	}

	/* Reclaim memory as needed */
	obj->total_tractored_objects = n + 1;
	if(obj->total_tractored_objects > 0)
	{
	    obj->tractored_object = (int *)realloc(
		obj->tractored_object,
		obj->total_tractored_objects * sizeof(int)
	    );
	    if(obj->tractored_object == NULL)
	    {
		obj->total_tractored_objects = 0;
		return;
	    }
	}
	else
	{
	    free(obj->tractored_object);
	    obj->tractored_object = NULL;
	    obj->total_tractored_objects = 0;
	}
}


/*
 *	Procedure to change sector of an object in accordance with
 *	sector changing rules. Returns 0 on success, -1 on error,
 *	or -3 if the sector change would violate sector changing
 *	rules.
 *
 *	Tractored objects (if any) will have their sectors changed as
 *	well.
 *
 *	If allow_wrapping is set to 1, then the position of
 *	the object (and its tractored objects if any) will have
 *	their positions wrapped.
 */
int DBObjectDoSetSector(
	const int obj_num,
	const long sect_x, const long sect_y, const long sect_z,
	const char allow_wrapping
)
{
	int i, tract_obj_num;
	long sect_dx, sect_dy, sect_dz;
	xsw_object_struct *obj, *tract_obj;

	if(DBIsObjectGarbage(obj_num))
	    return(-1);

	obj = xsw_object[obj_num];

	/* Is there a change in sectors? */
	if((obj->sect_x != sect_x) ||
	   (obj->sect_y != sect_y) ||
	   (obj->sect_z != sect_z)
	)
	{
	    sw_legend_struct *legend = &sw_legend;

	    /* Get sector change deltas */
	    sect_dx = sect_x - obj->sect_x;
	    sect_dy = sect_y - obj->sect_y; 
	    sect_dz = sect_z - obj->sect_z;

	    if(allow_wrapping)
	    {
		/* Wrap object position within sector */
		if(sect_dx < 0)
		    obj->x += legend->sect_x_len;
		else if(sect_dx > 0)
		    obj->x -= legend->sect_x_len;

		if(sect_dy < 0)
		    obj->y += legend->sect_y_len;
		else if(sect_dy > 0)
		    obj->y -= legend->sect_y_len;
#if 0
		if(sect_dz < 0)
		    obj->z += legend->sect_z_len;
		else if(sect_dz > 0)
		    obj->z -= legend->sect_z_len;
#endif
	    }

	    /* Sanitize coordinates */
	    if(obj->x < legend->sect_x_min)
		obj->x = legend->sect_x_min + 1;
	    if(obj->x > legend->sect_x_max)
		obj->x = legend->sect_x_max - 1;

	    if(obj->y < legend->sect_y_min)
		obj->y = legend->sect_y_min + 1;
	    if(obj->y > legend->sect_y_max)
		obj->y = legend->sect_y_max - 1;

/*
	    if(obj->z < legend->sect_z_min)
		obj->z = legend->sect_z_min + 1;
	    if(obj->z > legend->sect_z_max)
		obj->z = legend->sect_z_max - 1;
 */

	    /* Set new sector of object */
	    obj->sect_x = sect_x;
	    obj->sect_y = sect_y;
	    obj->sect_z = sect_z;

	    /* Go through list of tractored objects */
	    for(i = 0; i < obj->total_tractored_objects; i++)
	    {
		tract_obj_num = obj->tractored_object[i];
		if(DBIsObjectGarbage(tract_obj_num))
		    continue;

		tract_obj = xsw_object[tract_obj_num];

		/* Set tractored object's sector to match object's */
		tract_obj->sect_x = sect_x; 
		tract_obj->sect_y = sect_y;
		tract_obj->sect_z = sect_z; 

		/* Move tractored object's coordinate position
		 * within the sector to the object's position
		 */
		tract_obj->x = obj->x;
		tract_obj->y = obj->y;
/*
		tract_obj->z = obj->z;
 */
	    }
	}

	return(0);
}


/*
 *	Creates an economy product on the specified object.
 *
 *	If a product already exists by the specified product_name
 *	then that product will be updated instead of a new one being
 *	created on the object.
 *
 *	Returns the 0 on success or non-zero on error.
 */
int DBCreateObjectEconomyProduct(
	const int obj_num,
	const char *product_name,
	const int ocs_code,
	const sw_credits_t sell_price,
	const sw_credits_t buy_price,
	const float amount,
	const float amount_max
)
{
	int i;
	xsw_object_struct *obj;
	xsw_ecodata_struct *eco;
	xsw_ecoproduct_struct *prod;

	if(STRISEMPTY(product_name))
	    return(-2);

	if(DBIsObjectGarbage(obj_num))
	    return(-2);

	obj = xsw_object[obj_num];

	/* Allocate the object's economy data as needed */
	if(obj->eco == NULL)
	{
	    if(UNVAllocEco(obj))
		return(-3);
	}

	eco = obj->eco;

	/* Does the product already exist on this object? */
	for(i = 0; i < eco->total_products; i++)
	{
	    prod = eco->product[i];
	    if(prod == NULL)
		continue;

	    if(!strcasecmp(prod->name, product_name))
		break;
	}
	if(i >= eco->total_products)
	{
	    /* No such product, create a new one */
	    i = MAX(eco->total_products, 0);
	    eco->total_products = i + 1;
	    eco->product = (xsw_ecoproduct_struct **)realloc(
		eco->product,
		eco->total_products * sizeof(xsw_ecoproduct_struct *)
	    );
	    if(eco->product == NULL)
	    {
		eco->total_products = 0;
		return(-3);
	    }
	    eco->product[i] = prod = SW_ECO_PRODUCT(calloc(
		1, sizeof(xsw_ecoproduct_struct)
	    ));
	    if(prod == NULL)
		return(-3);
	}
	else
	{
	    /* Get the existing product */
	    prod = eco->product[i];
	}

	/* Set the product's values */
	strncpy(
	    prod->name, product_name, sizeof(prod->name)
	);
	prod->name[sizeof(prod->name) - 1] = '\0';

	prod->ocs_code = ocs_code;
	prod->sell_price = sell_price;
	prod->buy_price = buy_price;
	prod->amount = amount;
	prod->amount_max = amount_max;

	return(0);
}


/*
 *	Sorts the economy products on the specified object by their
 *	product names.
 */
void DBSortEconomyProducts(const int obj_num)
{
	int i, n, nproducts;
	const char *prod_name;
	char **names_list;
	xsw_ecodata_struct *eco;
	xsw_object_struct *obj;
	xsw_ecoproduct_struct *prod, **tmp_prod_list;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	eco = obj->eco;
	if(eco == NULL)
	    return;

	/* Delete any NULL economy products on the object */
	for(i = 0; i < eco->total_products; i++)
	{
	    if(eco->product[i] == NULL)
	    {
		eco->total_products--;
		for(n = i; n < eco->total_products; n++)
		    eco->product[n] = eco->product[n + 1];
		i--;
		continue;
	    }
	}

	nproducts = eco->total_products;
	if(nproducts <= 0)
	    return;

	/* Build the list of product names to be sorted */
	names_list = (char **)malloc(nproducts * sizeof(char *));
	for(i = 0; i < nproducts; i++)
	{
	    prod = eco->product[i];
	    if(prod == NULL)
	    {
		names_list[i] = STRDUP("");
		continue;
	    }

	    names_list[i] = STRDUP(prod->name);
	}

	/* Sort the product names list */
	names_list = StringQSort(names_list, nproducts);
	if(names_list == NULL)
	    return;

	/* Copy products pointers to tempory pointer array */
	tmp_prod_list = (xsw_ecoproduct_struct **)malloc(
	    nproducts * sizeof(xsw_ecoproduct_struct *)
	);
	memcpy(
	    tmp_prod_list,	/* Destination */
	    eco->product,	/* Source */
	    nproducts * sizeof(xsw_ecoproduct_struct *)
	);

	/* Sort the products with respect to the product names list */
	for(i = 0; i < nproducts; i++)
	{
	    prod_name = names_list[i];
	    if(prod_name == NULL)
		continue;

	    /* Match the product name with one in the tmp products
	     * list
	     */
	    eco->product[i] = NULL;
	    for(n = 0; n < nproducts; n++)
	    {
		prod = tmp_prod_list[n];
		if(prod == NULL)
		    continue;

		if(!strcasecmp(prod->name, prod_name))
		{
		    eco->product[i] = prod;
		    tmp_prod_list[n] = NULL;
		    break;
		}
	    }
	    if(eco->product[i] == NULL)
	    {
		const char *name = DBGetFormalNameStr(obj_num);
		fprintf(
		    stderr,
"DBSortEconomyProducts(): %s: Warning: Missing product `%s' from the tmp products list\n",
		    name, prod_name
		);
	    }
	}

	/* Delete the tmp products list */
	free(tmp_prod_list);

	/* Delete the product names list */
	strlistfree(names_list, nproducts);
}


/*
 *	Returns the object's number from the specified pointer.
 */
int DBGetObjectNumByPtr(xsw_object_struct *obj) 
{
        int i;

        if(obj == NULL)
            return(-1);

        for(i = 0; i < total_objects; i++)
        {
            if(xsw_object[i] == obj)
                return(i);
        }

        return(-1);
}


/*
 *      Return a statically allocated string containing the official
 *      name for the given object type code.
 */     
const char *DBGetTypeName(const int type)
{
	switch(type)
	{
	  case SW_OBJ_TYPE_ERROR:
	    return(SW_OBJ_TYPE_NAME_ERROR);
	    break;

	  case SW_OBJ_TYPE_GARBAGE:
	    return(SW_OBJ_TYPE_NAME_GARBAGE);
	    break;

	  case SW_OBJ_TYPE_STATIC:
	    return(SW_OBJ_TYPE_NAME_STATIC);
	    break; 

	  case SW_OBJ_TYPE_DYNAMIC:
	    return(SW_OBJ_TYPE_NAME_DYNAMIC);
	    break;

	  case SW_OBJ_TYPE_CONTROLLED:
	    return(SW_OBJ_TYPE_NAME_CONTROLLED);
	    break;

	  case SW_OBJ_TYPE_PLAYER:
	    return(SW_OBJ_TYPE_NAME_PLAYER);
	    break;
  
	  case SW_OBJ_TYPE_WEAPON:
	    return(SW_OBJ_TYPE_NAME_WEAPON);
	    break;

	  case SW_OBJ_TYPE_STREAMWEAPON:
	    return(SW_OBJ_TYPE_NAME_STREAMWEAPON);
	    break;

	  case SW_OBJ_TYPE_SPHEREWEAPON:
	    return(SW_OBJ_TYPE_NAME_SPHEREWEAPON);
	    break;

	  case SW_OBJ_TYPE_HOME:
	    return(SW_OBJ_TYPE_NAME_HOME);
	    break;

	  case SW_OBJ_TYPE_AREA:
	    return(SW_OBJ_TYPE_NAME_AREA);
	    break;

	  case SW_OBJ_TYPE_ANIMATED:
	    return(SW_OBJ_TYPE_NAME_ANIMATED);
	    break;

	  case SW_OBJ_TYPE_WORMHOLE:
	    return(SW_OBJ_TYPE_NAME_WORMHOLE);
	    break;

	  case SW_OBJ_TYPE_ELINK:
	    return(SW_OBJ_TYPE_NAME_ELINK);
	    break;

	  default:
	    return("*Unknown*");
	    break;
	}

	return("*Unknown*");
}


/*
 *      Returns a statically allocated format name string in
 *      the format of Name(#12345)
 */
const char *DBGetFormalNameStr(const int obj_num)
{
	if(DBIsObjectGarbage(obj_num))
	    return(SW_OBJ_NAME_GARBAGE);
	else
	    return(UNVGetObjectFormalName(
		obj_num, xsw_object[obj_num]
	    ));
}

/*
 *	Returns a pointer to the name of the OPM referanced
 *	by the given OCS code. Can return NULL on no match or
 *	error.
 *
 *	Do not free the returned pointer.
 */
const char *DBGetOCSOPMName(const int ocs_code)
{
	const int ocs_num = OCSGetByCode(ocs_code);
	if(ocs_num < 0)
	    return(NULL);

	return(ocs[ocs_num]->opm_name);
}

/*      
 *	Checks if the specified name is valid for an object to have.
 *      
 *	-1	General error
 *	0	Success
 *	1	Name too long
 *	2	Name too short.
 */        
int DBValidateObjectName(const char *name)
{
	int i, n;
	char *s = SW_INVALID_NAME_CHARACTERS;

	/* Null pointer? */
	if(name == NULL)
	    return(-1);

	/* Empty string? */
	if(*name == '\0')
	    return(2);

	/* Invalid characters? */
	for(i = 0; name[i] != '\0'; i++)
	{
	    if(i >= SW_OBJ_NAME_MAX)
		return(1);

	    for(n = 0; s[n] != '\0'; n++)
	    {
		if(name[i] == s[n])
		    return(3);
	    }
	}

	return(0);
}


/*         
 *      Checks if this is a valid password.
 */        
int DBValidateObjectPassword(const char *password)
{
	return(0);
}


/*
 *	Returns object's current visibility.
 *
 *	Takes into account the object's location, area objects that
 *	produce limited visibility will be applied to the returned
 *	visibility.
 */
float DBGetObjectVisibility(const int obj_num)
{
	return(UNVGetObjectVisibility(
	    xsw_object, total_objects, obj_num
	));
}

float DBGetObjectVisibilityPtr(xsw_object_struct *obj)
{
	return(UNVGetObjectVisibilityPtr(obj));
}


/*
 *	Returns the object's current effective scanner range in real
 *	units.
 *
 *	Takes into account the object's location, area objects that
 *	produce limited visibility will be applied to the returned
 *	scanner range.
 */
float DBGetObjectEffectiveScannerRange(const int obj_num)
{
	if(DBIsObjectGarbage(obj_num))
	    return(0.0f);
	else
	    return(DBGetObjectEffectiveScannerRangePtr(
		xsw_object[obj_num]
	    ));
}

float DBGetObjectEffectiveScannerRangePtr(xsw_object_struct *obj)
{
	if(obj == NULL)
	    return(0.0f);

	switch(obj->loc_type)
	{
	  case SW_LOC_TYPE_SPACE_TERRITORY:
	    return(MAX(obj->scanner_range, 0.0f));
	    break;
	  case SW_LOC_TYPE_NEBULA:
	    return(
		MAX(obj->scanner_range, 0.0f) * SW_VISIBILITY_NEBULA
	    );
	    break;
	  case SW_LOC_TYPE_ASTEROID_FIELD:
	    return(
		MAX(obj->scanner_range, 0.0f) * SW_VISIBILITY_ASTEROID_FIELD
	    );
	    break;
	  case SW_LOC_TYPE_PULSAR:
	    return(MAX(obj->scanner_range, 0.0f));
	    break;
	  default:	/* SW_LOC_TYPE_SPACE */
	    return(MAX(obj->scanner_range, 0.0f));
	    break;
	}
}


/*
 *	Returns true if the target object is within the object's
 *	effective scanner range.
 *
 *	The scanner_range must be the effective scanner range obtained
 *	from DBGetObjectEffectiveScannerRange*().
 */
char DBIsObjectInEffectiveScannerRange(
	const int obj_num, const int tar_obj_num,
	const float scanner_range
)
{
	return(UNVIsObjectInEffectiveScannerRange(
	    xsw_object, total_objects,
	    obj_num, tar_obj_num, scanner_range
	));
}

char DBIsObjectInEffectiveScannerRangePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj,
	const float scanner_range
)
{
	return(UNVIsObjectInEffectiveScannerRangePtr(
	    obj, tar_obj, scanner_range
	));
}
