#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>

#include "swserv.h"


int IncomingSocketIsAllocated(const int n);

int IncomingSocketNew(const int type, const int port);
void IncomingSocketDelete(const int n);
void IncomingSocketDeleteAll(void);

void IncomingSocketReclaim(void);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the incoming socket is allocated.
 */
int IncomingSocketIsAllocated(const int n)
{
	if((n < 0) || (n >= total_incoming_sockets))
	    return(0);
	else if(incoming_socket[n] == NULL)
	    return(0);
	else
	    return(1);
}


/*                  
 *	Creates a new incoming socket to listen for incoming
 *	connections on the specified port number.
 *
 *	Returns the new incoming socket number or -1 on error.
 */
int IncomingSocketNew(const int type, const int port)
{
	int i, n, status;
	struct sockaddr_in addr;
	incoming_socket_struct *insock;

	if(port < 0)
	    return(-2);

	if(total_incoming_sockets < 0)
	    total_incoming_sockets = 0;

	/* Check if there is already an incoming socket listening to
	 * the specified port number
	 */
	for(i = 0; i < total_incoming_sockets; i++)
	{
	    insock = incoming_socket[i];
	    if(insock == NULL)
		continue;

	    if(insock->port == port)
		return(i);
	}

	/* Look for available pointer */
	for(i = 0; i < total_incoming_sockets; i++)
	{
	    if(incoming_socket[i] == NULL)
		break;
	}
	if(i < total_incoming_sockets)
	{
	    /* Found available pointer */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_incoming_sockets;
	    total_incoming_sockets = n + 1;
	    incoming_socket = (incoming_socket_struct **)realloc(
		incoming_socket,
		total_incoming_sockets * sizeof(incoming_socket_struct *)
	    );
	    if(incoming_socket == NULL)
	    {
		total_incoming_sockets = 0;
		return(-3);
	    }
	}

	/* Allocate a new incoming socket */
	incoming_socket[n] = insock = SWS_INCOMING_SOCKET(calloc(
	    1, sizeof(incoming_socket_struct)
	));
	if(insock == NULL)   
	    return(-3);

	/* Set up values */
	insock->type = type;
	insock->port = port;
	insock->start_time = cur_systime;

	/* Create a socket */
	insock->socket = socket(AF_INET, SOCK_STREAM, 0);
	if(insock->socket < 0)
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unknown error");
	    *error_msg = toupper(*error_msg);
	    fprintf(
		stderr,
"Unable to create a socket for the incoming socket on port %i: %s.\n",
		port,
		error_msg
	    );
	    free(error_msg);

	    IncomingSocketDelete(n);

	    return(-1);
	} 
 
	/* Set up the the address structure */
	addr.sin_family = AF_INET;
	addr.sin_port = htons(insock->port);
	addr.sin_addr.s_addr = INADDR_ANY;
	memset(&addr.sin_zero, 0x00, 8);

	/* Bind the socket */
	status = bind(  
	    insock->socket,
	    (struct sockaddr *)&addr,
	    sizeof(struct sockaddr) 
	);
	if(status == -1)
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unknown error");
	    *error_msg = toupper(*error_msg);
	    fprintf(
		stderr,
"Unable to bind socket %i for the incoming socket on port %i: %s.\n",
		insock->socket,
		insock->port,
		error_msg
	    );
	    free(error_msg);

            IncomingSocketDelete(n);

            return(-1);
	}

	/* Set up the listening and backlogged connection limit */
	if(listen(insock->socket, SWS_LISTEN_SOCKET_BACKLOG) == -1)
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unknown error");
	    *error_msg = toupper(*error_msg);
	    fprintf(
		stderr,
"Unable to set up socket %i for listening for the incoming socket on\
 port %i: %s.\n",
		insock->socket,
		insock->port,
		error_msg
	    );
	    free(error_msg);

            IncomingSocketDelete(n);

            return(-1);
	}

	/* Set listening port socket to nonblocking */
	if(fcntl(insock->socket, F_SETFL, O_NONBLOCK) == -1)
	{
	    const int error_code = errno;
            char *error_msg = STRDUP(strerror(error_code));
            if(error_msg == NULL)
                error_msg = strdup("unknown error");
            *error_msg = toupper(*error_msg);
            fprintf(
                stderr,
"Unable to set up socket %i to non-blocking for the incoming socket on\
 port %i: %s.\n",
                insock->socket,
                insock->port,
                error_msg
            );
            free(error_msg);

            IncomingSocketDelete(n);

	    return(-1);
	}

	return(n);
}

/*
 *	Closes and deletes the incoming socket.
 */
void IncomingSocketDelete(const int n)
{ 
	incoming_socket_struct *insock;

	if(!IncomingSocketIsAllocated(n))
	    return;

	insock = incoming_socket[n];

	/* Close the socket as needed */
	if(insock->socket > -1)
	{
	    const time_t start_time = time(NULL);
	    int error_code;

	    shutdown(insock->socket, SHUT_RDWR);
            while(close(insock->socket) != 0)
            {
		error_code = errno;
                if(error_code == EAGAIN)
		{
                    /* Give up if more than 30 seconds elapsed */
                    if(time(NULL) > (start_time + 30l))
			break;

                    sleep(1);
                    continue;
		}
                else
		{
		    char *s = STRDUP(strerror(error_code));
		    if(s == NULL)
			s = strdup("unknown error");
		    *s = toupper(*s);
		    fprintf(
			stderr,
"Unable to close incoming socket %i (socket = %i): %s.\n",
			n, insock->socket, s
		    );
		    free(s);
                    break;
		}
            }
	    insock->socket = -1;
	}

	/* Delete this incoming socket */
	free(insock);
	incoming_socket[n] = insock = NULL;
}

/*
 *	Closes and deletes all incomming sockets.
 */
void IncomingSocketDeleteAll(void)
{
	int i;

	for(i = 0; i < total_incoming_sockets; i++)
	    IncomingSocketDelete(i);
	free(incoming_socket);
	incoming_socket = NULL;
	total_incoming_sockets = 0;
}


/*
 *	Reclaims incoming socket memory.
 */
void IncomingSocketReclaim(void)
{
	int i, n;
	incoming_socket_struct *insock;

	n = -1;
	for(i = 0; i < total_incoming_sockets; i++)
	{
	    insock = incoming_socket[i];
	    if(insock == NULL)
		continue;

	    n = i;
	}

	total_incoming_sockets = n + 1;
	if(total_incoming_sockets > 0)
	{
	    incoming_socket = (incoming_socket_struct **)realloc(
		incoming_socket,
		total_incoming_sockets * sizeof(incoming_socket_struct *)
	    );
	    if(incoming_socket == NULL)
	    {
		total_incoming_sockets = 0;
		return;
	    }
	}
	else
	{
	    free(incoming_socket);
	    incoming_socket = NULL;
	    total_incoming_sockets = 0;
	}
}
