#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/stat.h>

#if defined(__WIN32__)
# include <windows.h>
#else
# include <unistd.h>
#endif

#ifndef _USE_BSD
# define _USE_BSD
#endif
#include <sys/resource.h>
#include <sys/wait.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/prochandle.h"
#include "../include/df.h"
#include "../include/netio.h"
#include "../include/unvmain.h"
#include "../include/unvmatch.h"

#ifdef PLUGIN_SUPPORT
# include "plugins.h"
#endif
#include "siteban.h"
#include "cfgfile.h"
#include "net.h"
#include "swserv.h"
#include "sysparm.h"


static void SWSSignalCB(int s);

mode_t SWSCurrentFileMode(void);
const char *SWSCurrentDirectory(void);

static void SWSPrintDebugStderr(int error_level);
static void SWSPrintHelpStdout(void);
static void SWSPrintVersionStdout(void);
void SWSGetOSStats(void);
void SWSGetMemoryStats(sws_memory_stats_struct *buf);
static int SWSCheckBadSockets(
	const time_t cur_ms, const time_t cur_sec, const float tc
);
static void SWSReset(
	const int reset_level,
	const time_t cur_ms, const time_t cur_sec, const float tc
);

pid_t SWSReadPIDFile(const char *path);
void SWSWritePIDFile(const char *path, const pid_t p);
int SWSSaveUnv(void);
static void SWSExportStats(void);
void SWSResetTimmers(
	const time_t cur_ms, const time_t cur_sec, const float tc
);

void SWSShutdownStart(const time_t t);

int SWSInit(int argc, char *argv[]);
void SWSMainIteration(void *data);
void SWSMain(void *data);
void SWSMainPop(void *data);
void SWSMainPopAll(void *data);
void SWSShutdown(const char save_unv);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/* Signal makeup definations for some platforms */
#ifndef SIGPOLL
# define SIGPOLL	SIGUSR1
#endif
#ifndef SIGPWR
# define SIGPWR		SIGUSR2
#endif


int		sws_runlevel;
int		sws_reset;

time_t		sws_start_time;
time_t		cur_millitime;
time_t		cur_systime;
time_t		lapsed_millitime;
float		time_compensation;

sws_debug_struct	sws_debug;
sws_next_struct		next;
sws_sysparm_struct	sysparm;
swserv_dname_struct	dname;
swserv_fname_struct	fname;
sws_os_stat_struct	os_stat;

sws_connection_struct	**connection;
int			total_connections;
int			highest_connection;

sws_aux_connection_struct	**aux_connection;
int			total_aux_connections;

incoming_socket_struct	**incoming_socket;
int			total_incoming_sockets;

sw_legend_struct	sw_legend;

xsw_object_struct	**xsw_object;
int			total_objects;

xsw_object_struct	**recycled_xsw_object;
int			total_recycled_objects;

sws_ocs_struct	**ocs;
int		total_ocss;

xsw_object_struct	**opm;
int			total_opms;

sws_std_opm_struct	sws_std_opm;


/*
 *	Segfault counter.
 */
static int	segfault_count;


/*
 *	Signal callback.
 */
static void SWSSignalCB(int s)
{
	const time_t	cur_ms = cur_millitime,
			cur_sec = cur_systime;
	const float tc = time_compensation;
	int status;

	switch(s)
	{
	  case SIGHUP:
	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"%s received signal: SIGHUP: Reloading configuration...\n",
		    PROG_NAME
		);
	    if(sysparm.log_general)
		LogAppend(
		    fname.log,
"*** Received signal: SIGHUP: Reloading configuration..."
		);
	    sws_reset = SWS_RESET_CFG;
	    break;

	  case SIGINT:
	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"%s received signal: SIGINT: Server shutting down...\n",
		    PROG_NAME
		);
	    if(sysparm.log_general || sysparm.log_net || sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Received signal: SIGINT: Server shutting down..."
		);
	    SWSMainPopAll(NULL);
	    break;

	  case SIGTERM:
	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"%s received signal: SIGTERM: Server shutting down...\n",
		    PROG_NAME
		);
	    if(sysparm.log_general || sysparm.log_net || sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Received signal: SIGTERM: Server shutting down..."
		);
	    SWSMainPopAll(NULL);
	    break;

	  case SIGPIPE:
	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"%s received signal: SIGPIPE: Broken pipe.\n",
		    PROG_NAME
		);
	    if(sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Received signal: SIGPIPE: Broken pipe."
		);
	    /* Check if any sockets died */
	    SWSCheckBadSockets(cur_ms, cur_sec, tc);
	    break;

	  case SIGSEGV:
	    SWSPrintDebugStderr(1);

	    /* Print and log */
	    if(!sysparm.console_quiet)
		printf(
"%s received signal: SIGSEGV: Server performing emergency save and shutting down...\n",
		    PROG_NAME
		);
	    if(sysparm.log_general || sysparm.log_errors)
		LogAppend(
		    fname.log,
"*** Received signal: SIGSEGV: Server performing emergency save and shutting down..."
		);
	    status = DBEmergencySave(xsw_object, total_objects);
	    if(!sysparm.console_quiet)
	    {
		if(status)
		    printf(
"Emergency universe save failed, error code %i.\n",
			status
		    );
		else
		    printf(
"Emergency universe save complete and successful.\n"
		    );
	    }
	    if(sysparm.log_general || sysparm.log_errors)
	    {
		if(status)
		{
		    char *buf = (char *)malloc(40 + 128);
		    sprintf(
			buf,
"Server: Emergency universe save failed: Error code %i.",
			status
		    );
		    LogAppend(fname.log, buf);
		    free(buf);
		}
		else
		{
		    LogAppend(
			fname.log,
"Server: Emergency universe save complete and successful."
		    );
		}
	    }
	    exit(1);
	    break;

	  default:
	    /* Log the receiving of an unsupported signal */
	    if(!sysparm.console_quiet)
		printf(
"%s received signal: %i: Unsupported signal.\n",
		    PROG_NAME,
		    s
		);
	    if(sysparm.log_errors)
	    {
		char *msg = (char *)malloc(40 + 80);
		sprintf(
		    msg,
"*** Received unsupported signal %i.",
		    s
		);
		LogAppend(fname.log, msg);
		free(msg);
	    }
	    /* Perform an emergency save just in case */
	    DBEmergencySave(xsw_object, total_objects);
	    break;
	}
}


/*
 *	Returns the current permissions for files.
 */
mode_t SWSCurrentFileMode(void)
{
	const mode_t m = umask(0);
	umask(m);
	return(
	    (~m) &
	    (S_IRUSR | S_IWUSR |
	     S_IRGRP | S_IWGRP |
	     S_IROTH | S_IWOTH)
	);
}

/*
 *	Returns the current working directory.
 *
 *	The returned string must not be modified or deleted.
 */
const char *SWSCurrentDirectory(void)
{
	static char path[PATH_MAX];
	return(getcwd(path, sizeof(path)));
}


/*
 *	Procedure to print debug information.
 */
static void SWSPrintDebugStderr(int error_level)
{
	/* Heading */
	fprintf(
	    stderr,
"%s %s debug report:\n",
	    PROG_NAME_FULL, PROG_VERSION
	);

	fprintf(
	    stderr,
"runlevel: %i  total_connections (allocated): %i  highest_connection: %i\n",
	    sws_runlevel, total_connections, highest_connection
	);

	fprintf(
	    stderr,
"total_objects (allocated): %i\n",
	    total_objects
	);

	/* Footer */
	fprintf(stderr, "%s", SWS_DEBUG_FOOTER);
}


/*
 *	Prints the help message to stdout.
 */
static void SWSPrintHelpStdout(void)
{
	fprintf(
	    stdout,
	    "%s",
	    PROG_USAGE_MESSAGE
	);
}


/*
 *	Prints the version information to stdout.
 */
static void SWSPrintVersionStdout(void)
{
	fprintf(
	    stdout,
	    "%s",
	    PROG_NAME_FULL " Version " PROG_VERSION "\n"
	    PROG_COPYRIGHT "\n"
	);
}


/*
 *	Procedure to update OS stats.
 */
void SWSGetOSStats(void)
{
#ifdef __WIN32__
	char CurrDir[257];
	int SectorsPerCluster, BytesPerSector, FreeClusters, TotalClusters;

	if(GetCurrentDirectory(sizeof(CurrDir),CurrDir))
	{
	    CurrDir[3] = '\0'; /* only need the drive */
	    if(GetDiskFreeSpace(CurrDir,
				(PDWORD)&SectorsPerCluster,
				(PDWORD)&BytesPerSector,
				(PDWORD)&FreeClusters,
				(PDWORD)&TotalClusters
	       )
	    )
	    {
		os_stat.disk_used = (BytesPerSector *
		    SectorsPerCluster) * FreeClusters;

		os_stat.disk_total = (BytesPerSector *
		    SectorsPerCluster) * TotalClusters;

		return;
	    }
	}

	os_stat.disk_used = 0;
	os_stat.disk_total = 0;

#else
	int i;
	int total_df_stats;
	df_stat_struct **df_stat;


	/* Get disk space stats */
	df_stat = DiskFreeGetListing(&total_df_stats);
	if((df_stat != NULL) && (total_df_stats > 0))
	{
	    /*   Look for device stat index for the home
	     *   directory of this program.  Must do search in
	     *   reverse order to catch higher mounted paths first.
	     */
	    for(i = total_df_stats - 1; i >= 0; i--)
	    {
		if(df_stat[i] == NULL)
		    continue;

		if(strpfx(
		    dname.toplevel,
		    df_stat[i]->mounted_on
		))
		    break;
	    }
	    if(i < total_df_stats)
	    {
		os_stat.disk_used = df_stat[i]->used;
		os_stat.disk_total = df_stat[i]->total;
	    }
	    else
	    {
		os_stat.disk_used = 0;
		os_stat.disk_total = 0;
	    }
	}

	/* Free listing */
	DiskFreeDeleteListing(df_stat, total_df_stats);


	return;
#endif /* __WIN32__ */
}


/*
 *	Procedure to get statistics of all memory used by the
 *	program.
 */
void SWSGetMemoryStats(sws_memory_stats_struct *buf)
{
	int i, n;
	xsw_object_struct *obj;
	xsw_weapon_struct *wep;
	xsw_ecodata_struct *eco;
#ifdef PLUGIN_SUPPORT
	plugin_struct *pi;
#endif	/* PLUGIN_SUPPORT */

	if(buf == NULL)
	    return;

	/* Reset total */
	buf->total = 0;

	/* AUX pipes */
	buf->aux = (total_aux_connections * sizeof(sws_aux_connection_struct *)) +
	    (total_aux_connections * sizeof(sws_aux_connection_struct));
	buf->total += buf->aux;

	/* Connections */
	buf->con = (total_connections * sizeof(sws_connection_struct *)) +
	    (total_connections * sizeof(sws_connection_struct));
	buf->total += buf->con;

	/* Objects */
	buf->obj = (total_objects * sizeof(xsw_object_struct *)) +
	    (total_objects * sizeof(xsw_object_struct));

	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    /* Password */
	    if(obj->password != NULL)
		buf->obj += (strlen(obj->password) + 1) * sizeof(char);

	    /* Description */
	    if(obj->description != NULL)
		buf->obj += (strlen(obj->description) + 1) * sizeof(char);

	    /* Elink */
	    if(obj->elink != NULL)
		buf->obj += (strlen(obj->elink) + 1) * sizeof(char);

	    /* Tractored objects */
	    buf->obj += obj->total_tractored_objects * sizeof(int);

	    /* Weapons */
	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;

		buf->obj += sizeof(xsw_weapon_struct);
	    }
	    buf->obj += obj->total_weapons *
		sizeof(xsw_weapon_struct *);

	    /* Score */
	    if(obj->score != NULL)
		buf->obj += sizeof(xsw_score_struct);

	    /* Economy */
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		buf->obj += sizeof(xsw_ecodata_struct);

		for(n = 0; n < eco->total_products; n++)
		{
		    if(eco->product[n] == NULL)
			continue;

		    buf->obj += sizeof(xsw_ecoproduct_struct);
		}
		buf->obj += eco->total_products *
		    sizeof(xsw_ecoproduct_struct *);
	    }
	}

	buf->total += buf->obj;


	/* Recycled objects */
	buf->rec_obj = (total_recycled_objects * sizeof(xsw_object_struct *))
	    + (total_recycled_objects * sizeof(xsw_object_struct));

	for(i = 0; i < total_recycled_objects; i++)
	{
	    obj = recycled_xsw_object[i];
	    if(obj == NULL)
		continue;

	    /* Password */
	    if(obj->password != NULL)
		buf->rec_obj += (strlen(obj->password) + 1) * sizeof(char);

	    /* Description */
	    if(obj->description != NULL)
		buf->rec_obj += (strlen(obj->description) + 1) * sizeof(char);

	    /* Elink */
	    if(obj->elink != NULL)
		buf->rec_obj += (strlen(obj->elink) + 1) * sizeof(char);

	    /* Tractored objects */
	    buf->rec_obj += obj->total_tractored_objects * sizeof(int);

	    /* Weapons */
	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;
	
		buf->rec_obj += sizeof(xsw_weapon_struct);
	    }
	    buf->rec_obj += obj->total_weapons *
		sizeof(xsw_weapon_struct *);

	    /* Score */
	    if(obj->score != NULL)
		buf->rec_obj += sizeof(xsw_score_struct);

	    /* Economy */
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		buf->rec_obj += sizeof(xsw_ecodata_struct);

		for(n = 0; n < eco->total_products; n++) 
		{
		    if(eco->product[n] == NULL)
			continue;

		    buf->rec_obj += sizeof(xsw_ecoproduct_struct);
		}
		buf->rec_obj += eco->total_products *
		    sizeof(xsw_ecoproduct_struct *);
	    }
	}
	buf->total += buf->rec_obj;


	/* Object parameter models */
	buf->opm = (total_opms * sizeof(xsw_object_struct *)) +
	    (total_opms * sizeof(xsw_object_struct));

	for(i = 0; i < total_opms; i++)
	{
	    obj = opm[i];

	    /* Password */
	    if(obj->password != NULL)
		buf->opm += (strlen(obj->password) + 1) * sizeof(char);

	    /* Description */
	    if(obj->description != NULL)
		buf->opm += (strlen(obj->description) + 1) * sizeof(char);

	    /* Elink */
	    if(obj->elink != NULL)
		buf->opm += (strlen(obj->elink) + 1) * sizeof(char);

	    /* Tractored objects */
	    buf->opm += obj->total_tractored_objects * sizeof(int);

	    /* Weapons */
	    for(n = 0; n < obj->total_weapons; n++)
	    {
		wep = obj->weapons[n];
		if(wep == NULL)
		    continue;

		buf->opm += sizeof(xsw_weapon_struct);
	    }
	    buf->opm += obj->total_weapons *
		sizeof(xsw_weapon_struct *);

	    /* Score */
	    if(obj->score != NULL)
		buf->opm += sizeof(xsw_score_struct);

	    /* Economy */
	    eco = obj->eco;
	    if(eco != NULL)
	    {
		buf->opm += sizeof(xsw_ecodata_struct);

		for(n = 0; n < eco->total_products; n++)
		{
		    if(eco->product[n] == NULL)
			continue;

		    buf->opm += sizeof(xsw_ecoproduct_struct);
		}
		buf->opm += eco->total_products *
		    sizeof(xsw_ecoproduct_struct *);
	    }
	}
	buf->total += buf->opm;


	/* Object create scripts */
	buf->ocs = (total_ocss * sizeof(sws_ocs_struct *)) +
	    (total_ocss * sizeof(sws_ocs_struct));

	buf->total += buf->ocs;


	/* Plugins */
#ifdef PLUGIN_SUPPORT
	buf->plugins = (total_plugins * sizeof(SWServContext *)) +
	    (total_plugins * sizeof(SWServContext));

	for(i = 0; i < total_plugins; i++)
	{
	    pi = plugin[i];
	    if(pi == NULL)
		continue;

	    if(pi->name != NULL)
		buf->plugins += (strlen(pi->name) + 1) * sizeof(char);

	    for(n = 0; n < pi->argc; n++)
	    {
		if(pi->argv[n] != NULL)
		    buf->plugins += (strlen(pi->argv[n]) + 1) * sizeof(char);
	    }
	    buf->plugins += pi->argc * sizeof(char *);
	}
#else	/* PLUGIN_SUPPORT */
	buf->plugins = 0;
#endif	/* PLUGIN_SUPPORT */

	buf->total += buf->plugins;
}


/*
 *	Checks all connections for any bad or dead sockets and closes
 *	them as needed.
 *
 *	Returns the number of connections closed.
 */
static int SWSCheckBadSockets(
	const time_t cur_ms, const time_t cur_sec, const float tc  
)
{
	int i, bad_sockets = 0;
	sws_connection_struct *con;

	/* Iterate through all connections */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    if((con->socket < 0) ||
	       (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
	    )
		continue;

#if 0
/* This this is not the right way to check for bad connections,
 * skip it for now
 */
	    if(!NetIsSocketWritable(con->socket))
	    {
		shutdown(con->socket, SHUT_RDWR);
		close(con->socket);
		con->socket = -1;

		NetCloseConnection(i, cur_ms, cur_sec, tc);
	    }
#endif
	}

	return(bad_sockets);
}


/*
 *	Reloads the configuration and/or the universe.
 *
 *	If reset_level is SWS_RESET_CFG then the configuration will
 *	be reloaded.
 *
 *	If reset_level is SWS_RESET_CFG_UNV then the configuration
 *	and the universe will be reloaded.
 */
static void SWSReset(
	const int reset_level,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	char log_general;
	char *log_file;

	if(reset_level == SWS_RESET_NONE)
	    return;

	log_general = sysparm.log_general;
	log_file = STRDUP(fname.log);

	/* Log reset */
	if(log_general)
	{
	    char *s = (char *)malloc(40 + 80);
	    sprintf(
		s,
"Server: Level %i reset in progress...",
		reset_level
	    );
	    LogAppend(log_file, s);
	    free(s);
	}

	/* Reload the universe? */
	if(reset_level >= SWS_RESET_CFG_UNV)
	{
	    int i;

	    /* Notify all the connections that they need to relogin */
	    NetSendMessage(
		-1,
"*** Universe reloading from original, you will need to login again..."
	    );

	    /* Disconnect all connections */
	    for(i = 0; i < total_connections; i++)
		NetCloseConnection(i, cur_ms, cur_sec, tc);

	    /* Delete the current universe and open the original
	     * universe
	     */
	    DBOpenFromFile(
		fname.unv_in,
		cur_ms, cur_sec, tc
	    );

	    /* Logout, close, and delete all connections */
	    ConDeleteAll(cur_ms, cur_sec, tc);
	}

	/* Reload the configuration? */
	if(reset_level >= SWS_RESET_CFG)
	{
	    NetSendMessage(
		-1,
"Server: Reloading the configuration..."
	    );

	    /* Configuration */
	    CfgFileOpen(
		fname.cfg,
		cur_ms, cur_sec
	    );

	    /* Object Parameter Models */
	    OPMFileOpen(fname.opm);

	    /* Object Create Scripts */
	    OCSFileOpen(fname.ocs);

	    NetSendMessage(
		-1,
"Server: Configuration reload complete."
	    );
	}

	/* Refresh memory? */
	if(reset_level == SWS_RESET_REFRESH)
	{
	    NetSendMessage(
		-1,
"Server: Memory refresh in progress..."
	    );

	    /* Universe database */
	    DBReclaim(cur_ms, cur_sec, tc);

	    /* Object parameter models */
	    OPMReclaim();

	    /* Object create scripts */
	    OCSReclaim();

	    /* Connections */
	    ConReclaim();

	    NetSendMessage(
		-1,
"Server: Memory refresh complete."
	    );
	}

	/* Log reset */
	if(log_general)
	{
	    LogAppend(
		log_file,
"Server: Reset complete."
	    );
	}

	free(log_file);
}


/*
 *	Gets the PID from the file specified by path.
 */
pid_t SWSReadPIDFile(const char *path)
{
	pid_t p = 0;
	FILE *fp;

        if(STRISEMPTY(path))
            return(p);

        fp = fopen(path, "rb");
        if(fp == NULL)
            return(p);

	fread(&p, sizeof(p), 1, fp);
	fclose(fp);

	return(p);
}

/*
 *	Writes the PID to the file specified by path.
 */
void SWSWritePIDFile(const char *path, const pid_t p)
{
	FILE *fp;

	if(STRISEMPTY(path))
	    return;

	fp = fopen(path, "wb");
	if(fp == NULL)
	    return;

	fprintf(fp, "%i\n", p);
	fclose(fp);
}

/*
 *	Saves the universe to file.
 *
 *	Returns 0 on success or non-zero on error.
 */
int SWSSaveUnv(void)
{
	int status;

	if(!sysparm.console_quiet)
	    printf(
"%s: Universe save in progress...\n",
		PROG_NAME
	    );

	/* Notify all connections */
	NetSendMessage(
	    -1,
"Server: Universe save in progress..."
	);

	/* Log */
	if(sysparm.log_general)
	    LogAppend(
		fname.log,
"Server: Universe save in progress..."
	    );

	/* Save the universe */
	status = DBSaveToFile(fname.unv_out);

	/* Error? */
	if(status)
	{
	    char *s = (char *)malloc(40 + (2 * 80));
	    sprintf(
		s,
"Server: Universe save failed: Error code %i: Attempting emergency universe save...",
		status
	    );
	    NetSendMessage(-1, s);
	    free(s);
	}
	else
	{
	    NetSendMessage(
		-1,
"Server: Universe save complete and successful."
	    );
	}

	if(sysparm.log_general)
	{
	    if(status)
	    {
		char *s = (char *)malloc(40 + (2 * 80));
		sprintf(
		    s,
"Server: Universe save failed: Error code %i: Attempting emergency universe save...",
		    status
		);
		LogAppend(fname.log, s);
		free(s);
	    }
	    else
	    {
		LogAppend(
		    fname.log,
"Server: Universe save complete and successful."
		);
	    }
	}

	if(!sysparm.console_quiet)
	{
	    if(status)
		printf(
"%s: Universe save failed, attempting emergency universe save...\n",
		    PROG_NAME
		);
	    else
		printf(
"%s: Universe save complete and successful.\n",
		    PROG_NAME
		);
	}

	/* Attempt emergancy save incase of error (ignore emergency
	 * save status result)
	 */
	if(status)
	{
	    int status2 = DBEmergencySave(xsw_object, total_objects);
	    if(status2)
	    {
		if(sysparm.log_general || sysparm.log_errors)
		{
		    char *s = (char *)malloc(40 + 80);
		    sprintf(
			s,
"Server: Emergency universe save failed: Error code %i.",
			status2
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
		printf(
"%s: Emergency universe save failed, error code %i.\n",
		    PROG_NAME, status2
		);
	    }
	    else
	    {
		if(sysparm.log_general || sysparm.log_errors) 
		    LogAppend(
			fname.log,
"Server: Emergency universe save complete and successful."
		    );
		printf(
"%s: Emergency universe save complete and successful.\n",
		    PROG_NAME
		);
	    }
	}

	return(status);
}

/*
 *	Exports statistics.
 */
static void SWSExportStats(void)
{
	ExportScores(fname.scores_export);
	ExportConList(fname.conlist_export);
}

/*
 *	Resets all timmers.
 */
void SWSResetTimmers(
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	const char shutdown_schedualed = (next.shutdown > 0l) ? 1 : 0;
	int i, n;
	sws_aux_connection_struct *aux_con;
	sws_connection_struct *con;
	xsw_object_struct *obj;

	/* Next instances */
	next.system_check = cur_ms;
	next.object_values = cur_ms;
	next.memory_clean = cur_ms;
	next.new_connection_poll = cur_ms;
	next.os_stats = cur_sec;
	next.unv_save = cur_sec;
	next.stats_export = cur_sec;
	next.shutdown = shutdown_schedualed ? cur_sec : 0l;
	next.asteroid_field_damage = cur_ms;
	next.need_weapon_values = 0;

	/* Connections */
	for(i = 0; i < total_connections; i++)
	{
	    con = connection[i];
	    if(con == NULL)
		continue;

	    con->obj_update_next = cur_ms;
	    con->obj_last_sect = (con->obj_last_sect > 0l) ?
		cur_ms : 0l;
	    con->obj_last_pose = (con->obj_last_pose > 0l) ?
		cur_ms : 0l;
	}

	/* AUX Connections */
	for(i = 0; i < total_aux_connections; i++)
	{
	    aux_con = aux_connection[i];
	    if(aux_con == NULL)
		continue;

	    aux_con->next = cur_ms;
	}

	/* Objects */
	for(i = 0; i < total_objects; i++)
	{
	    obj = xsw_object[i];
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    obj->last_updated = cur_ms;
	    if(obj->last_cloak > 0l)
		obj->last_cloak = cur_ms;
	    if(obj->birth_time > 0l)
		obj->birth_time = cur_ms;
	    obj->animation.last_interval = cur_ms;

	    /* Auto destruct needs to be unset since there is no way   
	     * to calculate it after the timmers being reset
	     */
	    obj->auto_destruct_next = 0l;

	    for(n = 0; n < obj->total_weapons; n++)
		obj->weapons[n]->last_used = 0l;
	}
}


/*
 *	Starts the SWServ shutdown process.
 *
 *	The t specifies the delay (in seconds) before SWServ shuts down,
 *	to shutdown as soon as possible set t to 0.
 *
 *	All connections will be notified about the shutdown process.
 *
 *	The global next.shutdown will be set to when the shutdown
 *	should occure and SWSMainIteration() will check for that.
 */
void SWSShutdownStart(const time_t t)
{
	const time_t cur_sec = (cur_systime > 0l) ?
	    cur_systime : time(NULL);

	if(!sysparm.console_quiet)
	{
	    if(t > 0l)
	    {
		char *time_str = STRDUP(StringFormatTimePeriod(t));
	        fprintf(
		    stdout,
"%s shutting down in %s...\n",
		    PROG_NAME,
		    time_str
		);
		free(time_str);
	    }
	    else
	    {
		fprintf(
		    stdout,
"%s shutting down...\n",
		    PROG_NAME
		);
	    }
	}


	/* Notify all connections that the server is shutting down */
	if(t > 0l)
	{
	    char	*time_str = STRDUP(StringFormatTimePeriod(t)),
			*s = (char *)malloc(STRLEN(time_str) + 80);
	    sprintf(
		s,
"*** Server shutting down in %s! ***",
		time_str
	    );
	    NetSendMessage(-1, s);
	    free(s);
	    free(time_str);
	}
	else
	{
	    NetSendMessage(
		-1,
"*** Server shutting down! ***"
	    );
	}

	/* Schedual shutdown */
	next.shutdown = cur_sec + t;
}


/*
 *	Initializes SWServ.
 */
int SWSInit(int argc, char *argv[])
{
	char run_in_foreground = 0;
	int i, n, status;
	time_t	cur_ms = MilliTime(),
		cur_sec = time(NULL);
	float tc;
	const char *s, *arg;
	char cwd[PATH_MAX];
	struct stat stat_buf;
	sw_legend_struct *legend = &sw_legend;

	/* Check if command line arguments specify to run in
	 * SWServ in the foreground
	 */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    if(!strcasecmp(arg, "--foreground") ||
	       !strcasecmp(arg, "--fg") ||
	       !strcasecmp(arg, "-foreground") ||
	       !strcasecmp(arg, "-fg")
	    )
	    {
		run_in_foreground = 1;
		break;
	    }
	}
	/* Do not run in foreground? */
	if(!run_in_foreground)
	{
	    /* Run in the background */
	    switch(fork())
	    {
	      case -1:
		perror("fork");
		return(-3);
		break;
	      case 0:	/* Child */
		break;
	      default:	/* Parent */
		exit(0);
		break;
	    }
	}


	/* Initialize time zone */
	tzset();

	/* Reset globals */
	sws_reset = SWS_RESET_NONE;

	sws_start_time = cur_sec;
	cur_millitime = cur_ms;
	cur_systime = cur_sec;
	lapsed_millitime = 0l;
	time_compensation = tc = 1.0f;

	sws_debug.level = SWS_DEBUG_LEVEL_NONE;
	sws_debug.value = 0.0f;


	/* Reset sysparm */
	sysparm.allow_guest = 1;
	sysparm.allow_guest_signup = 1;
	sysparm.cease_fire = 0;
	sysparm.con_notify = 1;
	sysparm.console_quiet = 0;
	sysparm.cpu_priority = SWS_CPU_PRIORITY_HIGHEST;
	strncpy(
	    sysparm.guest_login_name, SWS_DEF_GUEST_NAME,
	    sizeof(sysparm.guest_login_name)
	);
	sysparm.guest_login_name[
	    sizeof(sysparm.guest_login_name) - 1] = '\0';
	sysparm.hide_players_home = 1;
	sysparm.hit_player_bonus = SWS_DEF_HIT_PLAYER_BONUS;
	sysparm.homes_destroyable = 0;
	sysparm.int_system_check = SWS_DEF_INT_SYSTEM_CHECK;
	sysparm.int_new_connection_poll = SWS_DEF_INT_NEW_CONNECTION_POLL;
	sysparm.int_object_values = SWS_DEF_INT_OBJECT_VALUES;
	sysparm.int_weapon_values = SWS_DEF_INT_WEAPON_VALUES;
	sysparm.int_memory_clean = SWS_DEF_INT_MEMORY_CLEAN;
	sysparm.int_aux_con_stats = SWS_DEF_INT_AUX_CON_STATS;
	sysparm.int_os_stats = SWS_DEF_INT_OS_STATS;	/* In seconds */
	sysparm.int_stats_export = SWS_DEF_INT_STATS_EXPORT;	/* In seconds */
	sysparm.int_unv_save = SWS_DEF_INT_UNV_SAVE;	/* In seconds */
	sysparm.killer_gets_credits = 1;
	sysparm.log_general = 1;
	sysparm.log_events = 1;
	sysparm.log_net = 1;
	sysparm.log_errors = 1;
	sysparm.login_timeout = SWS_DEF_LOGIN_TIMEOUT;
	sysparm.max_aux_connections = SWS_DEF_MAX_AUX_STATS_CONNECTIONS;
	sysparm.max_connections = SWS_DEF_MAX_CONNECTIONS;
	sysparm.max_failed_logins = SWS_DEF_MAX_FAILED_LOGINS;
	sysparm.max_guests = SWS_DEF_MAX_GUESTS;
	strncpy(
	    sysparm.mesg_wrong_login,
	    SWS_DEF_WRONG_LOGIN_MESG,
	    sizeof(sysparm.mesg_wrong_login)
	);
	sysparm.mesg_wrong_login[
	    sizeof(sysparm.mesg_wrong_login) - 1] = '\0';
	strncpy(
	    sysparm.mesg_no_guests,
	    SWS_DEF_NO_GUESTS_MESG,
	    sizeof(sysparm.mesg_no_guests)
	);
	sysparm.mesg_no_guests[
	    sizeof(sysparm.mesg_no_guests) - 1] = '\0';
	strncpy(
	    sysparm.mesg_welcome,
	    SWS_DEF_WELCOME_MESG,
	    sizeof(sysparm.mesg_welcome)
	);
	sysparm.mesg_welcome[
	    sizeof(sysparm.mesg_welcome) - 1] = '\0';
	strncpy(
	    sysparm.mesg_leave,
	    SWS_DEF_LEAVE_MESG,
	    sizeof(sysparm.mesg_leave)
	);
	sysparm.mesg_leave[
	    sizeof(sysparm.mesg_leave) - 1] = '\0';
	sysparm.movement_limit_tolorance = SWS_DEF_MOVEMENT_LIMIT_TOLORANCE;
	sysparm.recycled_objects_max = SWS_DEF_RECYCLED_OBJECTS_MAX;
	sysparm.report_destroyed_weapons = 0;
	sysparm.respawn_players = 1;
	sysparm.send_starchart = 1;
	sysparm.single_connection = 1;
	sysparm.use_server_ai = 1;

	/* Reset next scheduals */
	next.system_check = cur_ms;
	next.object_values = cur_ms;
	next.memory_clean = cur_ms;
	next.new_connection_poll = cur_ms;
	next.os_stats = cur_sec;
	next.unv_save = cur_sec + sysparm.int_unv_save;
	next.stats_export = cur_sec;
	next.shutdown = 0l;			/* Do not shutdown */
	next.asteroid_field_damage = cur_ms;
	next.need_weapon_values = 0;

	/* Reset objects database */
	xsw_object = NULL;
	total_objects = 0;

	/* Reset the recycled objects list */
	recycled_xsw_object = NULL;
	total_recycled_objects = 0;

	/* Reset object create scripts */
	ocs = NULL;
	total_ocss = 0;

	/* Reset object parameter models */
	opm = NULL;
	total_opms = 0;

	/* Reset standard opms indices */
	sws_std_opm.debris_num = -1;
	sws_std_opm.debris_50_num = -1;
	sws_std_opm.debris_100_num = -1;
	sws_std_opm.debris_150_num = -1;
	sws_std_opm.debris_200_num = -1;
	sws_std_opm.debris_250_num = -1;
	sws_std_opm.debris_300_num = -1;
	sws_std_opm.debris_starbase_num = -1;
	sws_std_opm.guest_num = -1;
	sws_std_opm.new_player_num = -1;

	/* Reset connections */
	connection = NULL;
	total_connections = 0;
	highest_connection = 0;

	/* Reset incoming sockets */
	incoming_socket = NULL;
	total_incoming_sockets = 0;

	/* Reset sitebans */
	siteban = NULL;
	total_sitebans = 0;

	/* Reset directories */
	strncpy(
	    dname.toplevel, SWS_TOPLEVEL_DIR,
	    sizeof(dname.toplevel)
	);
	strncpy(
	    dname.bin, SWS_BIN_DIR,
	    sizeof(dname.bin)
	);
	strncpy(
	    dname.etc, SWS_ETC_DIR,
	    sizeof(dname.etc)
	);
	strncpy(
	    dname.db, SWS_DB_DIR,
	    sizeof(dname.db)
	);
	strncpy(
	    dname.logs, SWS_LOGS_DIR,
	    sizeof(dname.logs)
	);
	strncpy(
	    dname.plugins, SWS_PLUGINS_DIR,
	    sizeof(dname.plugins)
	);
	strncpy(
	    dname.public_html, SWS_PUBLIC_HTML_DIR,
	    sizeof(dname.public_html)
	);
	strncpy(
	    dname.tmp, SWS_TMP_DIR,
	    sizeof(dname.tmp)
	);

	strncpy(
	    fname.monitor, "monitor",
	    sizeof(fname.monitor)
	);
	fname.monitor_set = 0;
	strncpy(
	    fname.unv_in, SWS_DEF_UNV_IN_FILE,
	    sizeof(fname.unv_in)
	);
	fname.unv_in_set = 0;
	strncpy(
	    fname.unv_out, SWS_DEF_UNV_OUT_FILE,
	    sizeof(fname.unv_out)
	);
	fname.unv_out_set = 0;
	strncpy(
	    fname.opm, SWS_DEF_OPM_FILE,
	    sizeof(fname.opm)
	);
	fname.opm_set = 0;
	strncpy(
	    fname.ocs, SWS_DEF_OCS_FILE,
	    sizeof(fname.ocs)
	);
	fname.ocs_set = 0;

	s = PrefixPaths(dname.logs, SWS_DEF_LOG_FILE);
	strncpy(
	    fname.log,
	    (s != NULL) ? s : SWS_DEF_LOG_FILE,
	    sizeof(fname.log)
	);

	s = PrefixPaths(dname.tmp, SWS_DEF_PID_FILE);
	strncpy(
	    fname.pid,
	    (s != NULL) ? s : SWS_DEF_PID_FILE,
	    sizeof(fname.pid)
	);

	*fname.conlist_export = '\0';
	*fname.scores_export = '\0';
	*fname.events_export = '\0';

#ifdef __WIN32__
	strcpy(cwd, ".");
#else
	s = SWSCurrentDirectory();
	strncpy(cwd, (s != NULL) ? s : "/", sizeof(cwd));
	cwd[sizeof(cwd) - 1] = '\0';
#endif /* __WIN32_ */
	s = PrefixPaths(cwd, SWS_CFG_FILE);
	strncpy(
	    fname.cfg,
	    (s != NULL) ? s : SWS_CFG_FILE,
	    sizeof(fname.cfg)
	);


	total_aux_connections = 0;
	aux_connection = NULL;

	UNVLegendReset(legend);

	/* Initialize the universe */
	UNVInit();


	/* Parse arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    if(strcasepfx(arg, "--mon") ||
	       strcasepfx(arg, "-mon")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.monitor, argv[i], sizeof(fname.monitor));
		    fname.monitor[sizeof(fname.monitor) - 1] = '\0';
		    fname.monitor_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    else if(strcasepfx(arg, "--unvin") ||
		    strcasepfx(arg, "-unvin") ||
		    strcasepfx(arg, "--dbin") ||
		    strcasepfx(arg, "-dbin")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.unv_in, argv[i], sizeof(fname.unv_in));
		    fname.unv_in[sizeof(fname.unv_in) - 1] = '\0';
		    fname.unv_in_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    else if(strcasepfx(arg, "--unvout") ||
		    strcasepfx(arg, "-unvout") ||
		    strcasepfx(arg, "--dbout") ||
		    strcasepfx(arg, "-dbout")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.unv_out, argv[i], sizeof(fname.unv_out));
		    fname.unv_out[sizeof(fname.unv_out) - 1] = '\0';
		    fname.unv_out_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Object parameter values */
	    else if(strpfx(arg, "--opm") ||
		    strpfx(arg, "-opm")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.opm, argv[i], sizeof(fname.opm));
		    fname.opm[sizeof(fname.opm) - 1] = '\0';
		    fname.opm_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Object create scripts */
	    else if(strpfx(arg, "--ocs") ||
		    strpfx(arg, "-ocs")
	    )
	    {
		i++;
		if(i < argc)
		{
		    strncpy(fname.ocs, argv[i], sizeof(fname.ocs));
		    fname.ocs[sizeof(fname.ocs) - 1] = '\0';
		    fname.ocs_set = 1;
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Port */
	    else if(strpfx(arg, "--p") ||
		    strpfx(arg, "-p")
	    )
	    {
		i++;
		if(i < argc)
		{
		    const int port = atoi(argv[i]);
		    const int insock_num = IncomingSocketNew(
			SWS_INCOMING_SOCKET_TYPE_STANDARD,
			port
		    );
		    if(insock_num < 0)
			return(-1);
		}
		else
		{
		    fprintf(
			stderr,
			"%s: Requires argument.\n",
			arg
		    );
		    return(-2);
		}
	    }
	    /* Quiet */
	    else if(strpfx(arg, "--q") ||
		    strpfx(arg, "-q")
	    )
	    {
		sysparm.console_quiet = 1;
	    }
	    /* Foreground (just checked, don't handle since it was
	     * handled above)
	     */
	    else if(!strcasecmp(arg, "--foreground") ||
		    !strcasecmp(arg, "--fg") ||
		    !strcasecmp(arg, "-foreground") ||
		    !strcasecmp(arg, "-fg")
	    )
	    {

	    }
	    /* Check if argument is prefixed with '-' or '+' */
	    else if((*arg == '-') ||
		    (*arg == '+')
	    )
	    {
		fprintf(
		    stderr,
		    "%s: Unsupported argument.\n",
		    arg
		);
	    }
	    /* Sole argument is configuration file to use */
	    else
	    {
		strncpy(fname.cfg, arg, sizeof(fname.cfg));
		fname.cfg[sizeof(fname.cfg) - 1] = '\0';
	    }
	}

/* Check for any errors in arguments */


	/* Print start up messages */
	if(!sysparm.console_quiet)
	{
	    printf(
		"%s\n",
		PROG_NAME_FULL " " PROG_VERSION
	    );
	}


	/* Open the configuration from file */
	if(!stat(fname.cfg, &stat_buf))
	{
	    status = CfgFileOpen(
		fname.cfg,
		cur_ms, cur_sec
	    );
	}
	else
	{
#if defined(__WIN32__)
	    status = -1;
#else
	    /* Is in the etc dir? */
	    s = PrefixPaths(dname.etc, SWS_CFG_FILE);
	    strncpy(
		fname.cfg, (s != NULL) ? s : SWS_CFG_FILE,
		sizeof(fname.cfg)
	    );
	    fname.cfg[sizeof(fname.cfg) - 1] = '\0';

	    if(!stat(fname.cfg, &stat_buf))
	    {
		status = CfgFileOpen(
		    fname.cfg,
		    cur_ms, cur_sec
		);
	    }
	    else
	    {
		/* Is in the global etc directory? */
		s = PrefixPaths(ETC_DIR, SWS_CFG_FILE);
		strncpy(
		    fname.cfg,
		    (s != NULL) ? s : SWS_CFG_FILE,
		    sizeof(fname.cfg)
		);
		fname.cfg[sizeof(fname.cfg) - 1] = '\0';
		status = CfgFileOpen(
		    fname.cfg,
		    cur_ms, cur_sec
		);
	    }
#endif
	}
	/* Error loading the configuration file? */
	if(status < 0)
	{
	    fprintf(
		stderr,
"%s: An error occured while loading the configuration file.\n",
		fname.cfg
	    );
	    return(-1);
	}

	/* Log startup, this is the earliest that we can log this
	 * since we just obtained the log option and the log file from
	 * the configuration file
	 */
	if(sysparm.log_general)
	    LogAppend(
		fname.log,
"*** " PROG_NAME_FULL " " PROG_VERSION " starting..."
	    );

	/* Write the PID file after getting the configuration */
	SWSWritePIDFile(fname.pid, getpid());

	/* Open a standard incoming connections listening socket to
	 * the default port as needed
	 */
	for(i = 0, n = 0; i < total_incoming_sockets; i++)
	{
	    if(incoming_socket[i] == NULL)
		continue;

	    /* Is this a standard incoming socket? */
	    if(incoming_socket[i]->type == SWS_INCOMING_SOCKET_TYPE_STANDARD)
		n++;
	}
	if(n <= 0)
	{
	    /* No standard incoming listening sockets exist */
	    fprintf(
		stderr,
"Error: No standard incoming sockets established.\n\
There were either no incoming sockets specified in the configuration\n\
file or they were unable to be created.\n"
	    );
	    return(-1);
	}


	/* ********************************************************* */
	/* Check if directories are valid */

	/* Server toplevel/home directory */
	if(!ISPATHDIR(dname.toplevel))
	{
	    fprintf(
		stderr,
"%s: Error: Invalid directory.\n",
		dname.toplevel
	    );
	    return(-1);
	}

	/* Server etc directory */
	if(!ISPATHDIR(dname.etc))
	{
	    fprintf(
		stderr,
		"%s: Warning: Invalid directory.\n",
		dname.etc
	    );
	    /* Non-fatal */
	}

	/* Server logs directory */
	if(!ISPATHDIR(dname.logs))
	{
	    fprintf(
		stderr,
		"%s: Warning: Invalid directory.\n",
		dname.logs
	    );
	    /* Non-fatal */
	}

	/* Server bin directory */
	if(!ISPATHDIR(dname.bin))
	{
	    fprintf(
		stderr,
		"%s: Warning: Invalid directory.\n",
		dname.bin
	    );
	    /* Non-fatal */
	}

	/* Server database directory */
	if(!ISPATHDIR(dname.db))
	{
	    fprintf(
		stderr,
		"%s: Error: Invalid directory.\n",
		dname.db
	    );
	    return(-1);
	}


	/* Run the monitor as needed, this needs to be done after
	 * the rcfile has been parsed so that the AUX Stat listening
	 * socket has been created
	 */
	if(fname.monitor_set &&
	   ISPATHEXECUTABLE(fname.monitor)
	)
	{
	    incoming_socket_struct *insock;

	    /* Use first AUX Stats listening socket port number */
	    for(i = 0, n = 0; i < total_incoming_sockets; i++)
	    {
		insock = incoming_socket[i];
		if(insock == NULL)
		    continue;

		if(insock->type == SWS_INCOMING_SOCKET_TYPE_AUXSTATS)
		    n = insock->port;
	    }

	    if(n > 0)
	    {
		char cmd[1024 + PATH_MAX + NAME_MAX];
	        sprintf(
		    cmd,
		    "%s %s %i",
		    fname.monitor,
		    "localhost",
		    n
	        );
	        Exec(cmd);
	    }
	    else
	    {
		fprintf(
		    stderr,
"Warning: Requested use of monitor but no AUX Stats port available.\n"
		);
		fprintf(
		    stderr,
"Warning: Monitor not executed.\n"
		);
	    }
	}


	/* Open the input universe */
	if(DBOpenFromFile(
	    fname.unv_in,
	    cur_ms, cur_sec, tc
	))
	{
	    fprintf(
		stderr,
"%s: Error: An error occured while loading the universe file.\n",
		fname.unv_in
	    );
	    return(-1);
	}


	/* Load the Object Parameters Models */
	status = OPMFileOpen(fname.opm);
	if(status < 0)
	{
	    fprintf(
		stderr,
"%s: Warning: An error occured while loading the object parameter models file.\n",
		fname.opm
	    );
	    /* Non-fatal error */
	}

	/* Load Object Create Scripts */
	status = OCSFileOpen(fname.ocs);
	if(status < 0)
	{
	    fprintf(
		stderr,
"%s: Warning: An error occured while loading the object create scripts file.\n",
		fname.ocs
	    );
	    /* Non-fatal error */
	}


	/* Print startup messages? */
	if(!sysparm.console_quiet)
	{
	    incoming_socket_struct *insock;

	    /* Universe title */
	    fprintf(
		stdout,
		"Universe: `%s'\n",
		legend->title
	    );

	    /* Configuration file */
	    fprintf(
		stdout,
		"Configuration file: %s\n",
		fname.cfg   
	    );

	    /* Listening port(s), both standard and aux */
	    for(i = 0; i < total_incoming_sockets; i++)
	    {
		insock = incoming_socket[i];
		if(insock == NULL)
		    continue;

		if(insock->type == SWS_INCOMING_SOCKET_TYPE_STANDARD)
		    fprintf(
			stdout,
			"Standard connection port: %i\n",
			insock->port
		    );
		if(insock->type == SWS_INCOMING_SOCKET_TYPE_AUXSTATS)
		    fprintf(
			stdout,
			"AUX monitoring port: %i\n",
			insock->port
		    );

	    }

	    /* Processs ID */
	    fprintf(stdout, "Process ID: %i\n", getpid());
	}

	/* Initialize the reality engine */
	REngInit();

	/* Log the end of startup */
	if(sysparm.log_general)
	    LogAppend(
		fname.log,
"*** " PROG_NAME_FULL " " PROG_VERSION " started."
	    );

	return(0);
}

/*
 *	Manages one iteration of SWServ.
 */
void SWSMainIteration(void *data)
{
	time_t	cur_ms, cur_sec,
		new_systime = time(NULL),
		new_millitime = MilliTime();
	float tc;
	sw_legend_struct *legend = &sw_legend;

	/* Check if timmers have cycled */
	if((new_systime < cur_systime) ||
	   (new_millitime < cur_millitime)
	)
	{
	    /* Reset timmers and time lapses */
	    lapsed_millitime = 0;
	    time_compensation = 1.0f;
	    SWSResetTimmers(new_millitime, new_systime, time_compensation);
	}
	else
	{
	    /* Calculate time lapse ms and compensation coeff */
	    lapsed_millitime = new_millitime - cur_millitime;
	    time_compensation = (float)lapsed_millitime /
		(float)SW_CYCLE_LAPSE_MS;
#if 0
	    if(time_compensation < 1.0f)
	        time_compensation = 1.0f;
#endif
	}

	/* Set current time */
	cur_systime = cur_sec = new_systime;
	cur_millitime = cur_ms = new_millitime;
	tc = time_compensation;


	/* Reset schedualed? */
	if(sws_reset != SWS_RESET_NONE)
	{
	    SWSReset(sws_reset, cur_ms, cur_sec, tc);
	    sws_reset = SWS_RESET_NONE;
	}


	/* Time for system updates? */
	if(next.system_check <= cur_ms)
	{
	    /* Shutdown schedualed? */
	    if(next.shutdown > 0l)
	    {
		/* Time to shutdown? */
		if(next.shutdown <= cur_sec)
		{
		    SWSMainPopAll(NULL);
		    return;
		}
	    }

	    /* Poll incoming sockets? */
	    if(next.new_connection_poll <= cur_ms)
	    {
		int i;
		incoming_socket_struct *insock;
		sws_connection_struct *con;

		/* Iterate through each incoming socket */
		for(i = 0; i < total_incoming_sockets; i++)
		{
		    insock = incoming_socket[i];
		    if(insock == NULL)
			continue;

		    /* Handle by socket type */
		    switch(insock->type)
		    {
		      case SWS_INCOMING_SOCKET_TYPE_AUXSTATS:
			AUXConManageNewConnections(
			    insock->socket,
			    cur_ms, cur_sec, tc
			);
			break;
		      default:	/* SWS_INCOMING_SOCKET_TYPE_STANDARD */
			NetManageNewConnections(
			    insock->socket,
			    cur_ms, cur_sec, tc
			);
			break;
		    }
		}

		/* Disconnect any sockets that are not logged in and
		 * have timmed out
		 */
		for(i = 0; i < total_connections; i++)
		{
		    con = connection[i];
		    if(con == NULL)
			continue;

		    /* Skip if the connection's socket is not
		     * connected or is on CLOSE_WAIT
		     */
		    if((con->socket < 0) ||
		       (con->flags & SWS_CON_FLAG_CLOSE_WAIT)
		    )
			continue;

		    /* Not logged in? */
		    if(!DBIsObjectGarbage(con->object_num))
			continue;

		    /* Timmed out? */
		    if((con->contime + sysparm.login_timeout) <
			cur_sec
		    )
		    {
			if(sysparm.log_net)
			{
			    char *s = (char *)malloc(80 + 40);
			    sprintf(
				s,
"Connection %i: Timmed out.",
				i
			    );
			    LogAppend(fname.log, s);
			    free(s);
			}

			/* Close connection */
			NetCloseConnection(i, cur_ms, cur_sec, tc);

			continue;
		    }
		}

		/* Schedual next poll for new connections */
		next.new_connection_poll = cur_ms +
		    sysparm.int_new_connection_poll;
	    }

	    /* Time to fetch OS stats? */
	    if(next.os_stats <= cur_sec)
	    {
		SWSGetOSStats();

		/* Schedual next fetch OS stats */
		next.os_stats = cur_sec + sysparm.int_os_stats;
	    }

	    /* Time to clean memory? */
	    if(next.memory_clean <= cur_ms)
	    {
		/* Universe objects */
		DBReclaim(cur_ms, cur_sec, tc);

		/* Object parameter macros */
		OPMReclaim();

		/* Object create scripts */
		OCSReclaim();

		/* Connections */
		ConReclaim();

		/* Schedual next memory clean */
		next.memory_clean = cur_ms + sysparm.int_memory_clean;
	    }

	    /* Time to save the universe? */
	    if(next.unv_save <= cur_sec)
	    {
		SWSSaveUnv();

		/* Schedual next universe save */
		next.unv_save = cur_sec + sysparm.int_unv_save;
	    }

	    /* Time to export stats? */
	    if(next.stats_export <= cur_sec)
	    {
		SWSExportStats();

		/* Schedual next time to export stats */
		next.stats_export = cur_sec + sysparm.int_stats_export;
	    }


	    /* Schedual next time for system check */
	    next.system_check = cur_ms + sysparm.int_system_check;
	}


	/* Manage objects */
	REngManage(
	    legend,
	    cur_ms, cur_sec, tc
	);
	if(next.asteroid_field_damage <= cur_ms)
	{
	    int i;

	    for(i = 0; i < total_objects; i++)
		REngAsteroidFieldDamage(
		    legend, i,
		    cur_ms, cur_sec, tc
		);

	    next.asteroid_field_damage = cur_ms + legend->asteroid_field_damage_int;
	}


	/* Network */

	/* AUX connections */
	AUXConManageAll(cur_ms, cur_sec, tc);

	/* Send out pending outgoing network data */
	NetManageSend(cur_ms, cur_sec, tc);

	/* Recieve and process incoming network data */
	NetManageRecv(cur_ms, cur_sec, tc);


#ifdef PLUGIN_SUPPORT
	/* Plugins management */
	PluginManage();
#endif	/* PLUGIN_SUPPORT */
}

/*
 *	Manages SWServ.
 *
 *	Pushes an SWServ main loop and manages and blocks until
 *	SWSMainPop() is called.
 */
void SWSMain(void *data)
{
	const int entering_runlevel = sws_runlevel;
	sws_runlevel++;
	while(sws_runlevel > entering_runlevel)
	{
	    switch(sysparm.cpu_priority)
	    {
	      case SWS_CPU_PRIORITY_HIGHEST:
		/* No sleeping */
		break;
	      case SWS_CPU_PRIORITY_HIGH:
		usleep(0);
		break;
	      case SWS_CPU_PRIORITY_MEDIUM:
		usleep(SW_CYCLE_LAPSE_US / 2);
		break;
	      default:	/* SWS_CPU_PRIORITY_LOWEST */
		usleep(SW_CYCLE_LAPSE_US);
		break;
	    }

	    SWSMainIteration(data);
	}
}

/*
 *	Pops off the highest SWServ main loop.
 */
void SWSMainPop(void *data)
{
	if(sws_runlevel > 0)
	    sws_runlevel--;
}

/*
 *	Pops off all SWServ main loops.
 */
void SWSMainPopAll(void *data)
{
	sws_runlevel = 0;
}

/*
 *	Shuts down SWServ.
 *
 *	If save_unv is true then the universe will be saved.
 */
void SWSShutdown(const char save_unv)
{
	const time_t	cur_ms = cur_millitime,
			cur_sec = cur_systime;
	const float tc = time_compensation;

#ifdef PLUGIN_SUPPORT
	/* Unload plugins first so that they may modify things before
	 * the final universe save and shutdown
	 */
	PluginUnloadAll();
#endif

	/* Logout, close, and delete all connections
	 *
	 * This is needed so that Guest objects are not saved
	 */
	ConDeleteAll(cur_ms, cur_sec, tc);

	/* Save the universe? */
	if(save_unv)
	    SWSSaveUnv();

	/* Begin shutdown */

	/* Reality engine */
	REngShutdown();

	/* Recycled objects buffers */
	DBPurgeAllRecycledObjects();

	/* XSW Objects */
	DBDeleteAllObjects();

	/* Object Create Scripts */
	OCSDeleteAll();

	/* Object Parameter Models */
	OPMDeleteAll();

	/* Universe management resources */
	UNVShutdown();

	/* AUX Monitoring Connections */
	AUXConDeleteAll(cur_ms, cur_sec, tc);

	/* Site Bans */
	SiteBanDeleteAll();

	/* Incoming Sockets */
	IncomingSocketDeleteAll();

	/* Remove the PID file */
	if(!STRISEMPTY(fname.pid))
	    unlink(fname.pid);
}


int main(int argc, char *argv[])
{
	int i, status;
	const char *arg;

	segfault_count = 0;
	sws_runlevel = 0;

	/* Select signals to watch for */
	signal(SIGHUP, SWSSignalCB);
	signal(SIGINT, SWSSignalCB);
	signal(SIGQUIT, SWSSignalCB);
	signal(SIGILL, SWSSignalCB);
	signal(SIGABRT, SWSSignalCB);
	signal(SIGBUS, SWSSignalCB);
	signal(SIGKILL, SWSSignalCB);
	signal(SIGUSR1, SWSSignalCB);
	signal(SIGSEGV, SWSSignalCB);
	signal(SIGUSR2, SWSSignalCB);
	signal(SIGPIPE, SWSSignalCB);
	signal(SIGTERM, SWSSignalCB);
	signal(SIGCONT, SWSSignalCB);
	signal(SIGSTOP, SWSSignalCB);

	/* Handle help and version arguments */
	for(i = 1; i < argc; i++)
	{
	    arg = argv[i];
	    if(arg == NULL)
		continue;

	    /* Help */
	    if(!strcasecmp(arg, "--help") ||
	       !strcasecmp(arg, "-help") ||
	       !strcasecmp(arg, "--h") ||
	       !strcasecmp(arg, "-h") ||
	       !strcasecmp(arg, "-?") ||
	       !strcasecmp(arg, "/?")
	    )
	    {
		SWSPrintHelpStdout();
		return(0);
	    }
	    /* Version */
	    else if(!strcasecmp(arg, "--version") ||
		    !strcasecmp(arg, "-version")
	    )
	    {
		SWSPrintVersionStdout();
		return(0);
	    }
	}

	/* Initialize */
	status = SWSInit(argc, argv);
	if(status)
	{
	    /* Shutdown and do not save the universe */
	    SWSShutdown(0);
	    fprintf(
		stderr,
		"%s\n",
PROG_NAME_FULL " " PROG_VERSION " shutting down due to error during initialization."
	    );
	    return(-status);
	}

	/* Manage */
	SWSMain(NULL);

	/* Save the universe and shutdown */
	SWSShutdown(1);

	/* Notify and log shutdown */
	if(!sysparm.console_quiet)
	    fprintf(
		stdout,
		"%s shut down.\n",
		PROG_NAME_FULL
	    );
	if(sysparm.log_general)
	    LogAppend(
		fname.log,
"*** " PROG_NAME_FULL " " PROG_VERSION " shut down."
	    );

	return(0);
}
