#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/swnetcodes.h"

#ifdef PLUGIN_SUPPORT
# include "plugins.h"
#endif
#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static void NetComMesgAddStatic(char *msg);
static int NetComMesgLocTypeStatic(const int loc_type);

void NetHandleComMesg(const int con_num, const char *arg);
void NetSendComMesg(
	const int con_num,
	const int src_obj_num,		/* Sender */
	const int tar_obj_num,		/* Receiver or -1 for general */
	const float bearing,		/* Target to source in radians */
	const int channel,
	const char *msg,
	const unsigned int options	/* Any of NET_COM_MSG_* */
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Garbles the message with "static" by adding randing '#'
 *	characters to the specified string.
 *
 *	The specified string will be modified.
 */
static void NetComMesgAddStatic(char *msg)
{
	const time_t cur_ms = cur_millitime;
	char *s;

	if(msg == NULL)
	    return;

	srand((unsigned int)cur_ms);

	s = msg;
	while(*s != '\0')
	{
#ifndef RAND_MAX
# define RAND_MAX	((unsigned int)-1)
#endif
	    if(((float)rand() / (float)RAND_MAX) > 0.72f)
		*s = '#';
	    s++;
	}
}

/*
 *	Checks if the specified location type is one that should have
 *	com messages garbled with static.
 */
static int NetComMesgLocTypeStatic(const int loc_type)
{
	return(
	    ((loc_type == SW_LOC_TYPE_NEBULA) ||
             (loc_type == SW_LOC_TYPE_ASTEROID_FIELD)
	    ) ? 1 : 0
	);
}


void NetHandleComMesg(const int con_num, const char *arg)
{
	char *s, *darg, *msg;
	int	obj_num = -1,
		tar_obj_num = -1;
	float	bearing = 0.0f * PI;
	int	channel = 0;
	sw_legend_struct *legend = &sw_legend;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;

	/* Copy the argument for parsing and editing purposes */
	darg = STRDUP(arg);

	/* SWEXTCMD_COMMESSAGE format:
	 *
	 *	src_obj tar_obj bearing channel;message
	 */

	/* Parse the message from the argument */
	s = strchr(darg, CS_STRING_DELIMINATOR_CHAR);
	if(s == NULL)
	{
	    /* No message, do not process this */
	    free(darg);
	    return;
	}
	msg = STRDUP(s + 1);

	/* Remove the message from darg */
	*s = '\0';

	/* No message? */
	if(msg == NULL)
	{
	    /* Do not process this */
	    free(darg);
	    return;
	}

	/* Parse arguments */
	sscanf(
	    darg,
	    "%i %i %f %i",
	    &obj_num,
	    &tar_obj_num,	/* Can be -1 */
	    &bearing,
	    &channel
	);

	free(darg);

	/* Connection must own the object */
	if(con_obj_num != obj_num)
	{
	    free(msg);
	    return;
	}

	if(DBIsObjectGarbage(obj_num))
	{
	    free(msg);
	    return;
	}

	obj = xsw_object[obj_num];

	/* Send message to a specific object? */
	if(tar_obj_num > -1)
	{
	    /* Are the objects within com range? */
	    if(SWMath3DInRange(
		xsw_object, total_objects,
		obj_num, tar_obj_num, legend->com_range
	    ))
	    {
		const int tar_con_num = ConGetByObject(tar_obj_num);
		xsw_object_struct *tar_obj = xsw_object[tar_obj_num];
		unsigned int options = 0;

		/* Is the target object listening to the specified
		 * com channel and is the target object connected?
		 */
		if((tar_obj->com_channel == channel) &&
		   (tar_con_num > -1)
		)
		{
		    /* Check if either the source or target object
		     * is in a location where the message should be
		     * garbled with static
		     */
		    if(NetComMesgLocTypeStatic(tar_obj->loc_type) ||
		       NetComMesgLocTypeStatic(obj->loc_type)
		    )
			options |= NET_COM_MSG_STATIC;

		    /* Calculate the bearing from the target to the
		     * source object
		     */
		    bearing = SWMathCoordinateDeltaVector(
			obj->x - tar_obj->x,
			obj->y - tar_obj->y
		    );

		    /* Send the message to this connection */
		    NetSendComMesg(
			tar_con_num,
			obj_num,	/* Sender */
			tar_obj_num,	/* Receiver */
			bearing,	/* Target to source */
			channel,
			msg,
			options
		    );
		}
	    }
	}
	else
	{
	    /* Send message to all objects in com range */
	    int i, tar_con_num;
	    unsigned int options;
	    xsw_object_struct *tar_obj;

	    /* Iterate through the objects list */
	    for(i = 0; i < total_objects; i++)
	    {
		/* Is this object not valid or not in com range? */
		if(!SWMath3DInRange(
		    xsw_object, total_objects,
		    obj_num, i, legend->com_range
		))
		    continue;

		tar_obj = xsw_object[i];

		/* Is this object not listening to the specified com
		 * channel?
		 */
		if(tar_obj->com_channel != channel)
		    continue;

		/* Get the connection of this object and skip if it is
		 * not connected
		 */
		tar_con_num = ConGetByObject(i);
		if(tar_con_num < 0)
		    continue;

		options = 0;

		/* Check if either the source or target object is in a
		 * location where the message should be garbled with
		 * static
		 */
		if(NetComMesgLocTypeStatic(tar_obj->loc_type) ||
		   NetComMesgLocTypeStatic(obj->loc_type)
		)
		    options |= NET_COM_MSG_STATIC;

		/* Calculate the bearing from the target to the source
		 * object
		 */
		bearing = SWMathCoordinateDeltaVector(
		    obj->x - tar_obj->x,
		    obj->y - tar_obj->y
		);

		/* Send the message to this connection */
		NetSendComMesg(
		    tar_con_num,
		    obj_num,		/* Sender */
		    -1,			/* Receiver (-1 means general message) */
		    bearing,		/* Target to source */
		    channel,
		    msg,
		    options		/* Options */
		);
	    }
	}

	/* Log this message? */
	if(sysparm.log_general || sysparm.log_events || sysparm.log_net)
	{
	    const char *name = DBGetFormalNameStr(obj_num);
	    s = (char *)malloc(
		STRLEN(name) + 40 + STRLEN(msg) + 80
	    );
	    sprintf(
		s,
"%s: Channel %.2f: Message: \"%s\"",
		name,
		(float)channel / 100.0f,
		msg
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

#ifdef PLUGIN_SUPPORT
	/* Forward this com message to the plugins */
	if(total_plugins > 0)
	{
	    int i;
	    for(i = 0; i < total_plugins; i++)
		PluginHailNotify(
		    i,
		    con_num,
		    obj_num,		/* Sender */
		    tar_obj_num,	/* Target object or -1 */
		    channel,		/* Com channel */
		    msg			/* Message */
		);
	}
#endif

	free(msg);
}


void NetSendComMesg(
	const int con_num,
	const int src_obj_num,		/* Sender */
	const int tar_obj_num,		/* Receiver or -1 for general */
	const float bearing,		/* Target to source in radians */
	const int channel,
	const char *msg,
	const unsigned int options	/* Any of NET_COM_MSG_* */
)
{
	char *dmsg, *sndbuf;

	if(msg == NULL)
	    return;

	/* Copy the message as dmsg and make sure that it is not
	 * longer than CS_MESG_MAX characters
	 */
	dmsg = STRDUP(msg);
	if(STRLEN(dmsg) > CS_MESG_MAX)
	    dmsg[CS_MESG_MAX] = '\0';

	/* Add static to the message? */
	if(options & NET_COM_MSG_STATIC)
	    NetComMesgAddStatic(dmsg);

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 + 40 + 40 +
	    STRLEN(dmsg) +
	    80
	);

	/* SWEXTCMD_COMMESSAGE format:
	 *
	 *	src_obj_num tar_obj_num bearing channel;message
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i %.4f %i;%s\n",
	    CS_CODE_EXT,
	    SWEXTCMD_COMMESSAGE,
	    src_obj_num,
	    tar_obj_num,
	    bearing,
	    channel,
	    dmsg
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);

	free(dmsg);
}
