#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/imgrefs.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static void NetBadLoginShutdownConnection(
	const int con_num,
        const time_t cur_ms, const time_t cur_sec, const float tc
);

void NetHandleLogin(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetSendLogin(const int con_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Shuts down the connection and records it as being disconnected
 *	because it sent too many bad logins.
 */
static void NetBadLoginShutdownConnection(
	const int con_num,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	if(sysparm.log_errors)
	{
	    char *s = (char *)malloc(40 + 40 + 80);
	    sprintf(
		s,
"Connection %i: Sent %i or more bad logins.",
		con_num,
		sysparm.max_failed_logins
	    );
	    LogAppend(fname.log, s);
	    free(s);
	}

	NetShutdownConnection(
	    con_num, SWS_CON_SHUTDOWN_WR, 1,
	    cur_ms, cur_sec, tc
	);
}

void NetHandleLogin(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int i, n, obj_num, client_type;
	const char *s;
	char	*sndbuf,
		name[SW_OBJ_NAME_MAX],
		password[SW_OBJ_PASSWORD_MAX],
		guest_name[SW_OBJ_NAME_MAX + 256];
	sw_legend_struct *legend = &sw_legend;
	sws_connection_struct *con = connection[con_num];
	xsw_object_struct *obj = NULL;

	/* Already logged in? */
	if(!DBIsObjectGarbage(con->object_num))
	{
#if 0
	    NetSendMessage(
		con_num,
		"You are already logged in."
	    );
#endif
	    return;
	}

	/* No argument specified? */
	if((arg != NULL) ? (*arg == '\0') : 1)
	{
	    /* Respond with correct login format instruction */
	    NetSendMessage(
		con_num,
"Login syntax incorrect, the correct format is `<name>;<password>'."
	    );
	    return;
	}

	/* Close this connection if it gave too many bad logins */
	if(con->badlogins >= sysparm.max_failed_logins)
	{
	    NetBadLoginShutdownConnection(con_num, cur_ms, cur_sec, tc);
	    return;
	}


	/* ********************************************************** */
	/* Check if at least one deliminator exists */
	s = strchr(arg, CS_STRING_DELIMINATOR_CHAR);
	if(s == NULL)
	{
	    /* Missing deliminator character, print usage response */
	    NetSendMessage(
		con_num,
"Login syntax incorrect, the correct format is `<name>;<password>'."
	    );

	    /* Send a rerequest for login */
	    sndbuf = (char *)malloc(40 + 80);
	    sprintf(sndbuf, "%i\n", CS_CODE_LOGIN);
	    NetSendDataToConnection(con_num, sndbuf);
	    free(sndbuf);

	    /* Record a bad login */
	    con->badlogins++;
	    if(con->badlogins >= sysparm.max_failed_logins)
		NetBadLoginShutdownConnection(con_num, cur_ms, cur_sec, tc);

	    return;
	}


	/* Begin parsing */
	s = arg;

	/* Parse name */
	if(s != NULL)
	{
	    char *s2;

	    strncpy(name, s, sizeof(name));
	    name[sizeof(name) - 1] = '\0';

	    /* Remove deliminator */
	    s2 = strchr(name, CS_STRING_DELIMINATOR_CHAR);
	    if(s2 != NULL)
		*s2 = '\0';

	    strstrip(name);

	    /* Seek to next deliminator */
	    s = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	    if(s != NULL)
		s++;
	}
	else
	{
	    *name = '\0';
	}

	/* Parse password */
	if(s != NULL)
	{
	    char *s2;

	    strncpy(password, s, sizeof(password));
	    password[sizeof(password) - 1] = '\0';

	    /* Remove deliminator */
	    s2 = strchr(password, CS_STRING_DELIMINATOR_CHAR);
	    if(s2 != NULL) 
		*s2 = '\0';

	    strstrip(password);

	    /* Seek to next deliminator */
	    s = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	    if(s != NULL)
		s++;
	}
	else
	{
	    *password = '\0';
	}

	/* Parse client type code */
	if(s != NULL)
	{
	    client_type = ATOI(s);

	    /* Seek to next deliminator */
	    s = strchr(s, CS_STRING_DELIMINATOR_CHAR);
	    if(s != NULL)
		s++;
	}
	else
	{
	    client_type = 0;
	}


	/* Check if this is a guest login (case insensitive) */
	if(sysparm.allow_guest &&
	   !strcasecmp(name, sysparm.guest_login_name)
	)
	{
	    int guest_opm_num;

	    /* Check if there are too many guest connections */
	    if(NetTotalGuestsConnected() >= sysparm.max_guests)
	    {
		/* Warn bad or missing guest OPM message */
		sndbuf = (char *)malloc(
		    40 + STRLEN(sysparm.guest_login_name) + 80
		);
		sprintf(
		    sndbuf,
"The maximum capacity of %i %s connections has been reached.",
		    sysparm.max_guests,
		    sysparm.guest_login_name
		);
		NetSendSysMessage(
		    con_num,
		    CS_SYSMESG_CODE_LOGINFAIL,
		    sndbuf
		);
		free(sndbuf);

		if(sysparm.log_errors)
		{
		    const char *guest_login_name = sysparm.guest_login_name;
		    char *s = (char *)malloc(
			40 + 40 + STRLEN(guest_login_name) +
			(2 * 80)
		    );
		    sprintf(
			s,
"Connection %i: Denied login due to it exceeding the maximum capacity\
 of %i %s connections.",
			con_num, sysparm.max_guests, guest_login_name
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}

		/* Shutdown the connection */
		NetShutdownConnection(
		    con_num, SWS_CON_SHUTDOWN_WR, 1,
		    cur_ms, cur_sec, tc
		);

		return;
	    }

	    /* Create a new object for the guest */
	    obj_num = DBCreateObject(
		IMGREF_DEFAULT_IMGREF,	/* ImgRef */
		SW_OBJ_TYPE_PLAYER,	/* Type */
		-1,			/* Owner */
		0l, 0l, 0l,		/* Sector */
		0.0f, 0.0f, 0.0f,	/* x, y, z */
		0.0f * PI, 0.0f * PI, 0.0f * PI,
		cur_ms, cur_sec, tc
	    );
	    if(DBIsObjectGarbage(obj_num))
	    {
		char *s = (char *)malloc(
		    STRLEN(sysparm.guest_login_name) + 80
		);
		sprintf(
		    s,
"Unable to create a %s object.",
		    sysparm.guest_login_name
		);
		NetSendSysMessage(
		    con_num,
		    CS_SYSMESG_CODE_LOGINFAIL,
		    s
		);
		free(s);

		/* Log */
		if(sysparm.log_errors)
		{
		    const char *guest_login_name = sysparm.guest_login_name;
		    char *s = (char *)malloc(
			40 + STRLEN(guest_login_name) +
			STRLEN(guest_login_name) + 80
		    );
		    sprintf(
			s,
"Connection %i: Unable to create a %s object for this %s login.",
			con_num, guest_login_name, guest_login_name
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}

		/* Shutdown the connection */
		NetShutdownConnection(
		    con_num, SWS_CON_SHUTDOWN_WR, 1,
		    cur_ms, cur_sec, tc
		);

		return;
	    }

	    obj = xsw_object[obj_num];


	    /* Set up guest object using OPM */
	    guest_opm_num = OPMGetByName(
		sysparm.guest_login_name,
		SW_OBJ_TYPE_PLAYER
	    );
	    if(OPMIsGarbage(guest_opm_num))
	    {
		/* Warn about missing OPM */
		char *s = (char *)malloc(
		    STRLEN(sysparm.guest_login_name) + 80
		);
		sprintf(
		    s,
"Warning: OPM `%s' not defined in the OPM file.",
		    sysparm.guest_login_name
		);
		NetSendMessage(con_num, s);
		free(s);

		/* Log */
		if(sysparm.log_errors)
		{
		    const char *guest_login_name = sysparm.guest_login_name;
		    char *s = (char *)malloc(
			STRLEN(guest_login_name) + (2 * 80)
		    );
		    sprintf(
			s,
"Warning: OPM `%s' used for new guest objects is not defined in the OPM file.",
			guest_login_name
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}
	    }
	    else
	    {
		/* Model new guest object according to OPM */
		OPMModelObject(obj_num, guest_opm_num);

		/* Move guest object to guest start location */
		obj->sect_x = legend->guest_start_sect_x;
		obj->sect_y = legend->guest_start_sect_y;
		obj->sect_z = legend->guest_start_sect_z;

		obj->x = legend->guest_start_x;
		obj->y = legend->guest_start_y;
		obj->z = legend->guest_start_z;

		/* Set guest object to guest start direction */
		obj->heading = legend->guest_start_heading;
		obj->pitch = legend->guest_start_pitch;
		obj->bank = legend->guest_start_bank;

		/* Must set guest to own itself */
		obj->owner = obj_num;
	    }

	    /* Set guest name (post fix a number to it) */
	    for(i = 0; i < sysparm.max_guests; i++)
	    {
		sprintf(
		    guest_name,
		    "%s %i",
		    sysparm.guest_login_name,
		    i + 1
		);
		/* See if another object has this exact same name */
	        for(n = 0; n < total_objects; n++)
	        {
		    if(xsw_object[n] == NULL)
		        continue;
		    if(xsw_object[n]->type <= SW_OBJ_TYPE_GARBAGE)
			continue;

		    if(!strcasecmp(xsw_object[n]->name, guest_name))
			break;
		}
		if(n >= total_objects)
		    break;	/* No other objects have this name */
	    }
	    strncpy(obj->name, guest_name, sizeof(obj->name));
	    obj->name[sizeof(obj->name) - 1] = '\0';

	    /* Remove password for guest objects */
	    free(obj->password);
	    obj->password = NULL;


	    /* Set client type and IS guest */
	    con->client_type = client_type;
	    con->flags |= SWS_CON_FLAG_IS_GUEST;
	}
	/* Check if guest login, but guest login not allowed */
	else if(!sysparm.allow_guest &&
		!strcasecmp(name, sysparm.guest_login_name)
	)
	{
	    /* Notify that guest connections are not allowed */
	    char	*s = STRDUP(sysparm.mesg_no_guests),
			*s2 = strsub(s, "%title", legend->title),
	    		*mesg;
	    free(s);
	    mesg = strsub(s2, "%name", sysparm.guest_login_name);
	    free(s2);
	    NetSendSysMessage(
		con_num,
		CS_SYSMESG_CODE_LOGINFAIL,
		mesg
	    );
	    free(mesg);

	    /* Shutdown the connection */
	    NetShutdownConnection(
		con_num, SWS_CON_SHUTDOWN_WR, 1,
		cur_ms, cur_sec, tc
	    );

	    return;
	}
	/* Get object number for regular login */
	else
	{
	    obj_num = UNVMatchObjectByName(
		xsw_object, total_objects,
		name, SW_OBJ_TYPE_PLAYER
	    );
	    if(DBIsObjectGarbage(obj_num))
	    {
		con->badlogins++;	/* Count this as a bad login */

		/* Log unmatched object name login as an error */
		if(sysparm.log_errors)
		{
		    char *s = (char *)malloc(40 + STRLEN(name) + 80);
		    sprintf(
			s,
"Connection %i: No player object found for login name `%s'.",
			con_num,
			name
		    );
		    LogAppend(fname.log, s);
		    free(s);
		}

		/* Send bad login message */
		NetSendSysMessage(
		    con_num,
		    CS_SYSMESG_CODE_LOGINFAIL,
		    sysparm.mesg_wrong_login
		);

		/* Give the connection another try? */
		if(con->badlogins < sysparm.max_failed_logins)
		{
		    /* Resend request for login */
		    sndbuf = (char *)malloc(40 + 80);
		    sprintf(sndbuf, "%i\n", CS_CODE_LOGIN);
		    NetSendDataToConnection(con_num, sndbuf);
		    free(sndbuf);
		}
		else
		{
		    /* No, shut it down for sending too many bad
		     * logins
		     */
		    NetBadLoginShutdownConnection(con_num, cur_ms, cur_sec, tc);
		}

		return;
	    }

	    obj = xsw_object[obj_num];


	    /* Set client type and NOT guest */
	    con->client_type = client_type;
	    con->flags &= ~SWS_CON_FLAG_IS_GUEST;
	}


	/* Get and check password (for non guest logins) */
	if(!(con->flags & SWS_CON_FLAG_IS_GUEST))
	{
	    if(CryptHandleVerify(password, obj->password) < 1)
	    {
		/* Send bad login message */
		NetSendSysMessage(
		    con_num,
		    CS_SYSMESG_CODE_LOGINFAIL,
		    sysparm.mesg_wrong_login
		);

		/* Resend a request for login */
		sndbuf = (char *)malloc(40 + 80);
		sprintf(sndbuf, "%i\n", CS_CODE_LOGIN);
		NetSendDataToConnection(con_num, sndbuf);
		free(sndbuf);

		/* Log */
		if(sysparm.log_errors)
		{
		    char	*name = STRDUP(DBGetFormalNameStr(obj_num)),
				*s = (char *)malloc(
			40 + STRLEN(name) + STRLEN(password) + 80
		    );
		    sprintf(
			s,
"Connection %i: Failed to log in as `%s' with bad password `%s'.",
			con_num, name, password
		    );
		    LogAppend(fname.log, s);
		    free(s);
		    free(name);
		}

		/* Record a bad login */
		con->badlogins++;
		if(con->badlogins >= sysparm.max_failed_logins)
		    NetBadLoginShutdownConnection(con_num, cur_ms, cur_sec, tc);

		return;
	    }
	}


	/* All login checks passed! */

	/* Set the connection's object number */
	con->object_num = obj_num;


	/* Allow only one connection per object? */
	if(sysparm.single_connection)
	{
	    /* Disconnect extra connections of obj_num */
	    sws_connection_struct *con2;

	    for(i = 0; i < total_connections; i++)
	    {
		con2 = connection[i];
		if(con2 == NULL)
		    continue;

		if((con2->object_num < 0) ||
		   (con2->flags & SWS_CON_FLAG_CLOSE_WAIT)
		)
		    continue;

		if(i == con_num)
		    continue;

		/* Is this connection logged in as the same object? */
		if(con2->object_num == obj_num)
		{
		    /* Notify and disconnect it */
		    NetSendMessage(
			i,
"Disconnecting due to new login of the same object as this connection."
		    );

		    if(sysparm.log_errors)
		    {
			char	*name = STRDUP(DBGetFormalNameStr(obj_num)),
				*s = (char *)malloc(
					40 + STRLEN(name) + (2 * 80)
				);
			sprintf(
			    s,
"Connection %i: Disconnected due to new login of the same object `%s'.",
			    i, name
			);
			LogAppend(fname.log, s);
			free(s);
			free(name);
		    }

		    NetShutdownConnection(
			i, SWS_CON_SHUTDOWN_WR, 1,
			cur_ms, cur_sec, tc
		    );
		}
	    }
	}

	/* Mark the object as connected and unhide it
	 *
	 * This needs to be done after the above force single
	 * connections check to ensure that the object is marked
	 * properly
	 */
	if(obj != NULL)
	{
	    obj->ext_flags |= SWS_OBJ_FLAG_CONNECTED;
	    obj->ext_flags &= ~SWS_OBJ_FLAG_HIDEFROMCON;
	    obj->last_connected = cur_sec;
	    obj->last_updated = cur_ms;
	}

	/* Set initial last sect & pose times */
	con->flags &= ~SWS_CON_FLAG_FSECT_WAIT;
	con->flags &= ~SWS_CON_FLAG_FPOSE_WAIT;
	con->obj_last_sect = cur_ms;
	con->obj_last_pose = cur_ms;

	/* If objects can be hidden then this object must be refreshed
	 * to all connections with objects within scanner range
	 */
	if(sysparm.hide_players_home)
	{
	    xsw_object_struct *obj2;
            sws_connection_struct *con2;

	    for(i = 0; i < total_connections; i++)
	    {
                con2 = connection[i];
                if(con2 == NULL)
                    continue;

                if((con2->object_num < 0) ||
                   (con2->flags & SWS_CON_FLAG_CLOSE_WAIT)
                )
                    continue;

                if(i == con_num)
                    continue;

		if(DBIsObjectGarbage(con2->object_num))
		    continue;

		obj2 = xsw_object[con2->object_num];

                if(obj->owner != con2->object_num)
                {
                    if(DBGetObjectVisibilityPtr(obj) <= 0.0f)
                        continue;
                }

                /* Not in scanner range? */
		if(!SWMath3DInRangePtr(
		    obj, obj2,
		    obj2->scanner_range
                ))
                    continue;

		NetSendRefreshObject(i, obj_num, 1);
	    }
	}

	/* Notify all connections about the new connection? */
	if(sysparm.con_notify && (obj != NULL))
	{
	    char	*s = STRDUP("%name has connected."),
			*mesg = strsub(
			    s, "%name",
			    DBIsObjectGarbage(obj_num) ?
				"*unknown*" : obj->name
			);
	    free(s);

	    for(i = 0; i < total_connections; i++)
	    {
		if(connection[i] == NULL)
		    continue;
		if(connection[i]->object_num < 0)
		    continue;
	 
		/* Skip current connection */
		if(i == con_num)
		    continue;

		NetSendMessage(i, mesg);
	    }

	    free(mesg);
	}


	if(!sysparm.console_quiet)
	{
	    char *name = STRDUP(DBGetFormalNameStr(obj_num));
	    fprintf(
		stdout,
"Connection %i: Successfully logged in as `%s'.\n",
		con_num,
		name
	    );
	    free(name);
	}

	/* Log successful connection */
	if(sysparm.log_net)
	{
	    char	*name = STRDUP(DBGetFormalNameStr(obj_num)),
			*s = (char *)malloc(40 + STRLEN(name) + 80);
	    sprintf(
		s,
"Connection %i: Successfully logged in as `%s'.",
		con_num,
		name
	    );
	    LogAppend(fname.log, s);
	    free(s);
	    free(name);
	}


	/* Print welcome message */
	if(obj != NULL)
	{
	    char	*s = STRDUP(sysparm.mesg_welcome),
			*s2 = strsub(s, "%title", legend->title),
			*mesg;
	    free(s);
	    mesg = strsub(
		s2, "%name",
		DBIsObjectGarbage(obj_num) ? "*unknown*" : obj->name
	    );
	    free(s2);
	    NetSendMessage(con_num, mesg);
	    free(mesg);
	}

#if 0
	sndbuf = (char *)malloc(STRLEN() + 80);
	sprintf(
	    sndbuf,
	    "You are now connected as %s.",
	    DBGetFormalNameStr(obj_num)
	);
	NetSendMessage(con_num, sndbuf);
	free(sndbuf);
#endif

	/* Refresh connection for the first time */
	NetSendRefresh(con_num);

	/* Schedual stats export to now in order to export the
	 * connections list now
	 */
	next.stats_export = cur_sec;
}


void NetSendLogin(const int con_num)
{
	char *sndbuf = (char *)malloc(
	    40 + 80
	);

	/* CS_CODE_LOGIN format:
	 *
	 */
	sprintf(
	    sndbuf,
	    "%i\n",
	    CS_CODE_LOGIN
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
