#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "sysparm.h"
#include "swserv.h"


static int PoseDistanceLimiting(
	sws_connection_struct *con,
	xsw_object_struct *obj,
	float x, float y, float z,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetHandleObjectPose(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetSendObjectPose(int con_num, int obj_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sets the pose position but limits it to a distance based on
 *	the object's maximum velocity.
 *
 *	Returns true if the movement was limited.
 */
static int PoseDistanceLimiting(
	sws_connection_struct *con,
	xsw_object_struct *obj,
	float x, float y, float z,
	const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	const float	dx = x - obj->x,
			dy = y - obj->y,
			dz = z - obj->z,
			d = SWMath3DDistance(dx, dy, dz);
	const time_t dt = CLIP(
	    (cur_ms - con->obj_last_pose),
	    0l, 10000l
	);
	const float dlimit = MAX(
	    (obj->velocity_max * (float)dt / (float)SW_CYCLE_LAPSE_MS),
	    0.0f
	);
	sws_sysparm_struct *parm = &sysparm;

#if 0
printf("\rd=%.3f dlimit=%.3f dt=%ldms res=%.3f  ",
 d, dlimit, dt,
 dlimit * parm->movement_limit_tolorance
); fflush(stdout);
#endif

	/* Distance moved exceeded the distance limit (include
	 * tolorance)?
	 */
	if((d > (dlimit * parm->movement_limit_tolorance)) &&
	   (d > 0.0f) && (dt > 0l)
	)
	{
	    const float dmaxed = dlimit * parm->movement_limit_tolorance;
	    const float dcoeff = dmaxed / d;
	    obj->x += dx * dcoeff;
	    obj->y += dy * dcoeff;
	    obj->z += dz * dcoeff;
#if 0
if(dmaxed > 0.0f)
 printf("\nMovement exceeded %i%%\n",
  (int)(d / dmaxed * 100.0f)
 );
#endif

	    /* Update odometer */
	    obj->odometer += dmaxed;

	    return(1);
	}
	else
	{
	    obj->x = x;
	    obj->y = y;
	    obj->z = z;

            /* Update odometer */
            obj->odometer += d;

	    return(0);
	}
} 
 
void NetHandleObjectPose(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int distance_limited = 0;
	int obj_num = -1;
	int type = SW_OBJ_TYPE_GARBAGE;
	int imgref_num = 0;
	long size = 0l;
	float x = 0.0f, y = 0.0f, z = 0.0f;
	float heading = 0.0f * PI, pitch = 0.0f * PI, bank = 0.0f * PI;
	float velocity = 0.0f;
	float velocity_heading = 0.0f * PI;
	float velocity_pitch = 0.0f * PI;
	float velocity_bank = 0.0f * PI;
	int current_frame = 0;
	xsw_object_struct *obj;
	sws_connection_struct *con = connection[con_num];

	/* CS_CODE_POSEOBJ format:
	 *
	 *      obj_num
	 *      type imgref_num size
	 *      x y z
	 *      heading pitch bank
	 *      velocity
	 *	velocity_heading velocity_pitch velocity_bank
	 *      frame
	 */
	sscanf(
	    arg,
"%i\
 %i %i %ld\
 %f %f %f\
 %f %f %f\
 %f\
 %f %f %f\
 %i",
	    &obj_num,
	    &type, &imgref_num, &size,
	    &x, &y, &z,
	    &heading, &pitch, &bank,
	    &velocity,
	    &velocity_heading, &velocity_pitch, &velocity_bank,
	    &current_frame
	);

	/* Connection must own object */         
	if(con->object_num != obj_num)
	    return;
 
	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Waiting for a force pose response? */
	if(con->flags & SWS_CON_FLAG_FPOSE_WAIT)
	{
	    /* Is this a valid response?
	     *
	     * Since there may be errors comparing float values parsed
	     * from the client we need to add a margin of error when
	     * checking this
	     */
	    if(SWMath3DDistance(
		x - obj->x,
		y - obj->y,
		z - obj->z
	    ) <= 1.0f)
	    {
		/* Switch off force pose wait */
		con->flags &= ~SWS_CON_FLAG_FPOSE_WAIT;
		con->obj_last_pose = cur_ms;
	    }
	    else
	    {
		/* Invalid response to a force pose, check if timmed
		 * out
		 */
		const time_t fpose_timeout = 3000l;
		if((con->obj_last_pose + fpose_timeout) < cur_ms)
		{
		    /* Timmed out, switch off force pose wait */
		    con->flags &= ~SWS_CON_FLAG_FPOSE_WAIT;
		}
		else
		{
		    /* Resend force pose */
		    NetSendObjectForcePose(
			con_num, obj_num,
			cur_ms, cur_sec, tc
		    );
		}
	    }
	    return;
	}

	/* Last pose time older (not from the same iteration)?
	 *
	 * This is to prevent the handling of subsequent pose
	 * statements that were received bunched togeather until the
	 * next main iteration
	 */
	if(con->obj_last_pose >= cur_ms)
	    return;

#if 0
/* Do not set type, imgref, or size */
	obj->type = type;
	obj->imgref = imgref_num;
	obj->size = size;
#endif

	/* Set these only if object has enough antimatter */
/*	if(obj->antimatter > 0.0f) */
	if(1)
	{
	    distance_limited = PoseDistanceLimiting(
		con, obj,
		x, y, z,
		cur_ms, cur_sec, tc
	    );

	    obj->velocity = velocity;

	    obj->velocity_heading = velocity_heading;
	    obj->velocity_pitch = velocity_pitch;
	    obj->velocity_bank = velocity_bank;
	    SWMathSetUnitVector2D(
		&obj->momentum_vector_compoent,
		obj->velocity_heading
	    );
	}

	obj->heading = heading;
	obj->pitch = pitch;
	obj->bank = bank;
	SWMathSetUnitVector2D(
	    &obj->attitude_vector_compoent,
	    obj->heading
	);

#if 0
	obj->animation.current_frame = current_frame;
#endif

	/* If the distance was limited then we need to send a force
	 * pose in order to correct the client
	 */
	if(distance_limited)
	    NetSendObjectForcePose(
		con_num, obj_num,
		cur_ms, cur_sec, tc
	    );

	con->obj_last_pose = cur_ms;
}


void NetSendObjectPose(int con_num, int obj_num)
{
	char *sndbuf;
	xsw_object_struct *obj;  

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	sndbuf = (char *)malloc(
	    40 + 40 +
	    40 + 40 + 40 +
            40 + 40 + 40 +
            40 + 40 + 40 +
            40 +
            40 + 40 + 40 +
            40 + 80
	);

	/* CS_CODE_POSEOBJ format:
	 *
	 *      obj_num
	 *      type imgref_num size
	 *      x y z
	 *      heading pitch bank
	 *      velocity
	 *      velocity_heading velocity_pitch velocity_bank
	 *      current_frame
	 */
	sprintf(
	    sndbuf,
"%i %i\
 %i %i %ld\
 %.4f %.4f %.4f\
 %.4f 0 0\
 %.5f\
 %.4f 0 0\
 %i\n",
	    CS_CODE_POSEOBJ,
	    obj_num,
	    obj->type, obj->imgref, obj->size,
	    obj->x, obj->y, obj->z,
	    obj->heading,
	    obj->velocity,
	    obj->velocity_heading,
	    obj->animation.current_frame
	);
	NetSendDataToConnection(con_num, sndbuf);
	free(sndbuf);
}
