#include <stdio.h>

#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/swnetcodes.h"

#include "net.h"
#include "swserv.h"


static int WEPLOCK_GET_NEXT_OBJECT(
	const int obj_num, xsw_object_struct *obj,
	const int start_obj_num
);
static int WEPLOCK_GET_NEAREST_OBJECT(
	const int obj_num, xsw_object_struct *obj
);
static char WEPLOCK_IS_IN_SCANNER_RANGE(
	const int obj_num, xsw_object_struct *obj,
	const int tar_obj_num, xsw_object_struct *tar_obj
);

void NetHandleSetWeaponsLock(
	const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
);
void NetSendSetWeaponsLock(const int con_num, const int obj_num);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static int WEPLOCK_GET_NEXT_OBJECT(
	const int obj_num, xsw_object_struct *obj,
	const int start_obj_num
)
{
	int i;
	float scanner_range;
	xsw_object_struct *tar_obj;

	/* Get the current effective scanner range in real units */
	scanner_range = DBGetObjectEffectiveScannerRangePtr(obj);
	if(scanner_range <= 0.0f)
	    return(-1);

	/* Begin search, note that start_obj_num could be -1 */
	for(i = MAX(start_obj_num + 1, 0);
	    i < total_objects;
	    i++
	)
	{
	    tar_obj = xsw_object[i];
	    if(tar_obj == NULL)
		continue;
	    if(tar_obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue; 
	    if(tar_obj == obj)
		continue;

	    if(!DBIsObjectInEffectiveScannerRangePtr(
		obj, tar_obj, scanner_range
	    ))
		continue;

	    /* Got match */
	    return(i);
	}

	return(-1);
}

static int WEPLOCK_GET_NEAREST_OBJECT(
	const int obj_num, xsw_object_struct *obj
)
{
        return(UNVMatchObjectNearest(
            xsw_object, total_objects,
            obj_num, -1
        ));
}

static char WEPLOCK_IS_IN_SCANNER_RANGE(
	const int obj_num, xsw_object_struct *obj,
	const int tar_obj_num, xsw_object_struct *tar_obj
)
{
	float scanner_range;

	if(tar_obj == NULL)
	    return(0);

	/* Get the current effective scanner range in real units */
	scanner_range = DBGetObjectEffectiveScannerRangePtr(obj);
	if(scanner_range <= 0.0f)
	    return(0);

	return(DBIsObjectInEffectiveScannerRangePtr(
	    obj, tar_obj, scanner_range
	));
}


void NetHandleSetWeaponsLock(
        const int con_num, const char *arg,
        const time_t cur_ms, const time_t cur_sec, const float tc
)
{
	int	obj_num = -1,
		tar_obj_num = -1, prev_obj_num;
	xsw_object_struct	*obj,
				*new_locked_obj = NULL;
	sws_connection_struct *con = connection[con_num];
	const int con_obj_num = con->object_num;

	/* Format NET_CMD_WEAPONSLOCK:
	 *
	 *	obj_num tar_obj_num
	 */
	sscanf(
	    arg,
	    "%i %i",
	    &obj_num, &tar_obj_num
	);

	if(obj_num != con_obj_num)
	    return;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Record previously locked on object */
	prev_obj_num = obj->locked_on;
	if(DBIsObjectGarbage(prev_obj_num))
	{
	    /* Was not locked on to anything, reset values */
	    obj->locked_on = -1;
	    prev_obj_num = -1;
	}

	/* Handle parsed target object value as an argument for
	 * operation
	 */
	switch(tar_obj_num)
	{
	  case -1:	/* Unlock */
	    obj->locked_on = -1;
	    break;

	  case -2:	/* Lock on the next object */
	    while(1)
	    {
		/* Match next object */
		obj->locked_on = WEPLOCK_GET_NEXT_OBJECT(
		    obj_num, obj, prev_obj_num
		);
		/* Unable to match next object? */
		if(DBIsObjectGarbage(obj->locked_on))
		    break;

		/* Get the next object */
		new_locked_obj = xsw_object[obj->locked_on];

		/* Check if the matched object is marked hidden */
		if(new_locked_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		{
		    /* Set previous object to current and try again */
		    prev_obj_num = obj->locked_on;
		    new_locked_obj = NULL;
		    continue;
		}

		/* Match successful, break */
		break;
	    }
	    break;

	  case -3:	/* Lock on the nearest object */
	    obj->locked_on = WEPLOCK_GET_NEAREST_OBJECT(
		obj_num, obj
	    );
	    if(!DBIsObjectGarbage(obj->locked_on))
	    {
		new_locked_obj = xsw_object[obj->locked_on];
		if(new_locked_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		    new_locked_obj = NULL;
	    }
	    break;

	  default:	/* Lock explicitly on the specified object */
	    if(!DBIsObjectGarbage(tar_obj_num))
	    {
		new_locked_obj = xsw_object[tar_obj_num];

		if(WEPLOCK_IS_IN_SCANNER_RANGE(
		    obj_num, obj,
		    tar_obj_num, new_locked_obj) &&
		   !(new_locked_obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON)
		)
		{
		    obj->locked_on = tar_obj_num;
		}
		else
		{
		    new_locked_obj = NULL;
		    obj->locked_on = -1;
		}
	    }
	    break;
	}


	/* Check if locked on a new valid object */
	if(new_locked_obj != NULL)
	{
	    const int locked_obj_num = obj->locked_on;
	    int wep_num;

	    /* Need to send set values of target object first */
	    NetSendObjectMaximums(con_num, locked_obj_num);

	    /* Do not send object standard values just yet, the
	     * schedualing for the next object values update will be
	     * set to now just below
	     */

	    NetSendObjectName(con_num, locked_obj_num);
	    NetSendObjectDescription(con_num, locked_obj_num);
	    NetSendSetEngine(con_num, locked_obj_num);
	    NetSendSetChannel(
		con_num,
		locked_obj_num,
		new_locked_obj->com_channel
	    );

	    /* Send weapon values */
	    for(wep_num = 0;
		wep_num < new_locked_obj->total_weapons;
		wep_num++
	    )
		NetSendWeaponValues(
		    con_num, locked_obj_num, wep_num
		);
	}

	/* Set next object values update to now so it gets updated now */
	next.object_values = cur_ms;
}


void NetSendSetWeaponsLock(const int con_num, const int obj_num)
{
	return;
}
