#ifdef PLUGIN_SUPPORT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/disk.h"

extern "C" {
#ifdef __linux__
# include <dlfcn.h>
#endif
#ifdef __SOLARIS__
# include <dlfcn.h>
#endif
}

#include "plugins.h"
#include "sysparm.h"
#include "swserv.h"


/* Utilities */
int PluginIsLoaded(plugin_id_t id);
plugin_struct *PluginGetPointerFromID(plugin_id_t id);
plugin_id_t PluginMatchByName(char *name);

/* Load, Manage, and Unload */
plugin_id_t PluginLoad(
	const char *path,
	const unsigned int flags,
	const int argc, char **argv,
	const int con_num
);
void PluginManage(void);
void PluginUnload(plugin_id_t id, const int con_num);
void PluginUnloadAll(void);

/* Commands List */
char **PluginHelpCommandList(
	plugin_id_t id, int *ncommands
);

/* Command */
int PluginCommand(
	plugin_id_t id,
	int con_num, int obj_num,
	const char *cmd, const char *arg,
	int uid, int gid
);

/* Hail Notify */
void PluginHailNotify(
	plugin_id_t id,
	int con_num, int obj_num, int tar_obj_num,
	int channel, const char *message
);

/* Hit Notify */
void PluginHitNotify(
	plugin_id_t id,
	const int wobj_num,
	const int tobj_num,
	const float bearing,	/* Target to weapon object */
	const float dmg_structure,
	const float dmg_shield
);

/* Destroy Notify */
void PluginDestroyNotify(
	plugin_id_t id,
	int reason,		/* One of SW_DESTROY_REASON_* */
	int destroyed_obj_num,
	int destroyer_obj_num,
	int destroyer_obj_owner_num
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


plugin_struct	**plugin = NULL;
int		total_plugins = 0;


/*
 *	Checks if the plugin specified by id is allocated and loaded.
 */
int PluginIsLoaded(plugin_id_t id)
{
	if((plugin == NULL) || (id < 0) || (id >= total_plugins))
	    return(0);
	else if(plugin[id] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Returns the pointer to the plugin by the specified id.
 *
 *	Can return NULL on error or if the if refers to an unloaded
 *	plugin.
 */
plugin_struct *PluginGetPointerFromID(plugin_id_t id)
{
	if(PluginIsLoaded(id))
	    return(plugin[id]);
	else
	    return(NULL);
}

/*
 *	Matches a plugin by the specified name.
 *
 *	Returns the plugin id or -1 on no match.
 */
plugin_id_t PluginMatchByName(char *name)
{
	int i;
	plugin_struct *pi;

	if(name == NULL)
	    return(-1);

	for(i = 0; i < total_plugins; i++)
	{
	    pi = plugin[i];
	    if(pi == NULL)
		continue;

	    /* Skip if not loaded */
	    if(pi->handle == NULL)
		continue;

	    /* Skip if has no name */
	    if(pi->name == NULL)
		continue;

	    /* Skip if names do not match */
	    if(strcasecmp(pi->name, name))
		continue;

	    return(i);
	}

	return(-1);
} 


/*
 *	Loads the plugin and allocates a new plugin in the list.
 *
 *	The specified path must be an absolute path.
 *
 *	Return values are:
 *
 *	0 or greater	The loaded plugin's id.
 *	-1		Invalid path or plugin not found.
 *	-2		Missing init, manage, or shutdown function.
 *	-3		Systems error.
 *	-4		Init function called returned -1.
 */
plugin_id_t PluginLoad(
	const char *path,
	const unsigned int flags,
	const int argc, char **argv,
	const int con_num
)
{
	int i, status, largc;
	char **largv;
	const char *name;
	void *handle = NULL;
	struct stat stat_buf;
	plugin_id_t n;
	plugin_struct *pi;
	SWPLUGIN_INIT_FUNCRTN (*init_func)(
		SWPLUGIN_INIT_FUNCPROTO
	) = NULL;
	SWPLUGIN_MANAGE_FUNCRTN (*manage_func)(
		SWPLUGIN_MANAGE_FUNCPROTO
	) = NULL;
	SWPLUGIN_SHUTDOWN_FUNCRTN (*shutdown_func)(
		SWPLUGIN_SHUTDOWN_FUNCPROTO
	) = NULL;
	SWServContext *ctx;

	if(path == NULL)
	    return(-1);

	if(stat(path, &stat_buf))
	    return(-1);

	name = strrchr(path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = path;

	/* Load the plugin */
#if defined(__linux__) || defined(__SOLARIS__)
	handle = dlopen(
	    path,
	    RTLD_LAZY |
	    RTLD_GLOBAL
	);
	if(handle == NULL)
	{
	    char *buf, *s = STRDUP(dlerror());
	    if(s == NULL)
		s = strdup("unknown error");
	    *s = toupper(*s);

	    buf = (char *)malloc(STRLEN(name) + STRLEN(s) + 80);
	    sprintf(
		buf,
"%s: Unable to load plugin: %s.",
		name, s
	    );

	    if(sysparm.log_errors)
		LogAppend(fname.log, buf);

	    fprintf(stderr, "%s\n", buf);

	    free(buf);
	    free(s);

	    return(-3);
	}
#endif

	/* Get and check pointers to the initialize, manage, and
	 * shutdown functions on the plugin
	 */
#if defined(__linux__) || defined(__SOLARIS__)
	init_func = (SWPLUGIN_INIT_FUNCRTN(*)(SWPLUGIN_INIT_FUNCPROTO))
	    dlsym(handle, SWPLUGIN_INIT_FUNCNAME);
	if(init_func == NULL)
	{
	    char *buf = (char *)malloc(
		STRLEN(name) + STRLEN(SWPLUGIN_INIT_FUNCNAME) + 80
	    );
	    sprintf(
		buf,
"%s: Missing initialize function `%s()'.",
		name,
		SWPLUGIN_INIT_FUNCNAME
	    );
	    if(sysparm.log_errors)
		LogAppend(fname.log, buf);
	    fprintf(stderr, "%s\n", buf);
	    free(buf);
	    dlclose(handle);
	    return(-2);
	}
	manage_func = (SWPLUGIN_MANAGE_FUNCRTN(*)(SWPLUGIN_MANAGE_FUNCPROTO))
	    dlsym(handle, SWPLUGIN_MANAGE_FUNCNAME);
	if(manage_func == NULL)
	{
	    char *buf = (char *)malloc(
		STRLEN(name) + STRLEN(SWPLUGIN_MANAGE_FUNCNAME) + 80
	    );
	    sprintf(
		buf,
"%s: Missing management function `%s()'.",
		name,
		SWPLUGIN_MANAGE_FUNCNAME
	    );
	    if(sysparm.log_errors)
		LogAppend(fname.log, buf);
	    fprintf(stderr, "%s\n", buf);
	    free(buf);
	    dlclose(handle);
	    return(-2);
	}
	shutdown_func = (SWPLUGIN_SHUTDOWN_FUNCRTN(*)(SWPLUGIN_SHUTDOWN_FUNCPROTO))
	    dlsym(handle, SWPLUGIN_SHUTDOWN_FUNCNAME);
	if(shutdown_func == NULL)
	{
	    char *buf = (char *)malloc(
		STRLEN(name) + STRLEN(SWPLUGIN_SHUTDOWN_FUNCNAME) + 80
	    );
	    sprintf(
		buf,
"%s: Missing shutdown function `%s()'.",
		name,
		SWPLUGIN_SHUTDOWN_FUNCNAME
	    );
	    if(sysparm.log_errors)
		LogAppend(fname.log, buf);
	    fprintf(stderr, "%s\n", buf);
	    free(buf);
	    dlclose(handle);
	    return(-2);
	}
#endif

	/* Create the context for this plugin */
	ctx = (SWServContext *)calloc(1, sizeof(SWServContext));
	PluginSetupContext(ctx);

	/* Copy arguments */
	if(argc > 0)
	{
	    largc = argc;
	    largv = (char **)calloc(largc, sizeof(char *));
	    if(largv != NULL)
	    {
		for(i = 0; i < largc; i++)
		    largv[i] = strdup(argv[i]);
	    }
	    else
	    {
		largc = 0;
	    }
	}
	else
	{
	    largc = 0;
	    largv = NULL;
	}

	/* Call initialize function */
	status = (*init_func)(
	    largc, largv,
	    con_num, ctx
	);
	if(status != 0)
	{
	    /* Init function reported error so unload this plugin */
	    char *buf = (char *)malloc(
		STRLEN(name) + STRLEN(SWPLUGIN_INIT_FUNCNAME) + 40 + 80
	    );
	    sprintf(
		buf,
"%s: Initialize function `%s()' returned error code `%i', plugin not loaded.",
		name,
		SWPLUGIN_INIT_FUNCNAME,
		status
	    );
	    if(sysparm.log_errors)
		LogAppend(fname.log, buf);
	    fprintf(stderr, "%s\n", buf);
	    free(buf);

#if defined(__linux__) || defined(__SOLARIS__)
	    dlclose(handle);
#endif
	    strlistfree(largv, largc);
	    free(ctx);
	    return(-4);
	}


	/* Sanitize total */
	if(total_plugins < 0)
	    total_plugins = 0;

	/* Search for available pointer in array */
	for(i = 0; i < total_plugins; i++)
	{
	    if(plugin[i] == NULL)
		break;
	}
	if(i < total_plugins)
	{
	    /* Got available pointer */
	    n = i;
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = total_plugins;
	    total_plugins = n + 1;
	    plugin = (plugin_struct **)realloc(
		plugin,
		total_plugins * sizeof(plugin_struct *)
	    );
	    if(plugin == NULL)
	    {
		total_plugins = 0;
#if defined(__linux__) || defined(__SOLARIS__)
		dlclose(handle);
#endif
		strlistfree(largv, largc);
		free(ctx);
		return(-3);
	    }

	    /* Reset new pointer to NULL so it gets allocated below */
	    plugin[n] = NULL;
	}

	/* Allocate structure as needed */
	if(plugin[n] == NULL)
	{
	    plugin[n] = (plugin_struct *)calloc(
		1, sizeof(plugin_struct)
	    );
	    if(plugin[n] == NULL)
	    {
#if defined(__linux__) || defined(__SOLARIS__)
		dlclose(handle);
#endif  /* __linux__ */
		strlistfree(largv, largc);
		free(ctx);
		return(-3);
	    }
	}

	/* Set values to plugin record structure */
	pi = plugin[n];
	pi->name = strdup(path);

	pi->argc = largc;
	pi->argv = largv;

	pi->handle = handle;
	pi->flags = flags;
	pi->ctx = ctx;

	pi->init_func = init_func;
	pi->manage_func = manage_func;
	pi->shutdown_func = shutdown_func;

	/* Get pointers to optional functions on plugin */
	pi->help_commands_list_func = (
	    SWPLUGIN_HELP_COMMANDS_LIST_FUNCRTN(*)(SWPLUGIN_HELP_COMMANDS_LIST_FUNCPROTO)
	)dlsym(handle, SWPLUGIN_HELP_COMMANDS_LIST_FUNCNAME);

	pi->command_func = (
	    SWPLUGIN_COMMAND_FUNCRTN(*)(SWPLUGIN_COMMAND_FUNCPROTO)
	)dlsym(handle, SWPLUGIN_COMMAND_FUNCNAME);

	pi->hail_notify_func = (
	    SWPLUGIN_HAIL_NOTIFY_FUNCRTN(*)(SWPLUGIN_HAIL_NOTIFY_FUNCPROTO)
	)dlsym(handle, SWPLUGIN_HAIL_NOTIFY_FUNCNAME);

	pi->hit_notify_func = (
	    SWPLUGIN_HIT_NOTIFY_FUNCRTN(*)(SWPLUGIN_HIT_NOTIFY_FUNCPROTO)
	)dlsym(handle, SWPLUGIN_HIT_NOTIFY_FUNCNAME);

	pi->destroy_notify_func = (
	    SWPLUGIN_DESTROY_NOTIFY_FUNCRTN(*)(SWPLUGIN_DESTROY_NOTIFY_FUNCPROTO)
	)dlsym(handle, SWPLUGIN_DESTROY_NOTIFY_FUNCNAME);

	/* Print plugin loaded */
	if(!sysparm.console_quiet)
	{
	    fprintf(
		stdout,
"Loaded plugin: ID %i: %s\n",
		n + 1, name
	    );
#if 0
	    fprintf(
		stdout,
"\tID: %i\n",
		n + 1
	    );
	    if(pi->argc > 0)
	    {
		fprintf(
		    stdout,
"\tArguments: "
		);
		for(i = 0; i < pi->argc; i++)
		{
		    fprintf(stdout, "\"%s\"", pi->argv[i]);
		    if(i < (pi->argc - 1))
			fprintf(stdout, " ");
		}
		fprintf(stdout, "\n");
	    }
#endif
	}         

	if(sysparm.log_general)
	{
	    char *buf, *all_args_str = NULL;

	    for(i = 0; i < pi->argc; i++)
	    {
		all_args_str = strcatalloc(
		    all_args_str, " "
		);
		all_args_str = strcatalloc(
		    all_args_str, pi->argv[i]
		);
	    }
	    if(all_args_str == NULL)
		all_args_str = strdup("");

	    if(con_num > -1)
	    {
		if(ConIsLoggedIn(con_num))
		{
		    sws_connection_struct *con = connection[con_num];
		    char *name = STRDUP(DBGetFormalNameStr(con->object_num));
		    buf = (char *)malloc(
			STRLEN(name) + 40 + STRLEN(path) +
			STRLEN(all_args_str) + 80
		    );
		    sprintf(
			buf,
"%s: Loaded plugin: ID %i: %s%s",
			name, n + 1, path, all_args_str
		    );
		    free(name);
		}
		else
		{
		    buf = (char *)malloc(
			40 + 40 + STRLEN(path) + STRLEN(all_args_str) + 80
		    );
		    sprintf(
			buf,
"Connection %i: Loaded plugin: ID %i: %s%s",
			con_num, n + 1, path, all_args_str
		    );
		}
	    }
	    else
	    {
		buf = (char *)malloc(
		    40 + STRLEN(path) + STRLEN(all_args_str) + 80
		);
		sprintf(
		    buf,
"Server: Loaded plugin: ID %i: %s%s",
		    n + 1, path, all_args_str
		);
	    }
	    LogAppend(fname.log, buf);
	    free(buf);
	    free(all_args_str);
	}

	return(n);
}


/*
 *	Manages all plugins, this function needs to be called
 *	once per loop.
 */
void PluginManage(void)
{
	int i;
	plugin_struct *pi;

	for(i = 0; i < total_plugins; i++)
	{
	    pi = plugin[i];
	    if(pi == NULL)
		continue;

	    /* Management function valid? */
	    if(pi->manage_func != NULL)
	    {
		/* Call management function */
		if(pi->manage_func(pi->ctx) != 0)
		{
		    /* Management function returned a value indicating
		     * that it needs to be unloaded now
		     */
		    PluginUnload(i, -1);
		    continue;
		}
	    }
	}
}

/*
 *	Unloads and deletes the plugin.
 */
void PluginUnload(plugin_id_t id, const int con_num)
{
	const char *name, *path;
	plugin_struct *pi = PluginGetPointerFromID(id);
	if(pi == NULL)
	    return;

	path = pi->name;
	if(path != NULL)
	{
	    name = strrchr(path, DIR_DELIMINATOR);
	    if(name != NULL)
		name++;
	    else
		name = path;
	}
	else
	{
	    name = "*noname*";
	}

	/* Was this plugin loaded? */
	if(pi->handle != NULL)
	{
	    /* Call the plugin's shutdown function */
	    if(pi->shutdown_func != NULL)
		pi->shutdown_func(pi->ctx);

	    if(!sysparm.console_quiet)
		fprintf(  
		    stdout,
"Unloaded plugin: ID %i: %s\n",
		    id + 1, name
		);

	    if(sysparm.log_general)
	    {
		char *buf;

		if(con_num > -1)
		{
		    if(ConIsLoggedIn(con_num))
		    {
			sws_connection_struct *con = connection[con_num];
			char *name = STRDUP(DBGetFormalNameStr(con->object_num));
			buf = (char *)malloc(
			    STRLEN(name) + 40 + STRLEN(path) + 80
			);
			sprintf(
			    buf,
"%s: Unloaded plugin: ID %i: %s",
			    name, id + 1, path
			);
			free(name);
		    }
		    else
		    {
			buf = (char *)malloc(
			    40 + 40 + STRLEN(path) + 80
			);
			sprintf(
			    buf,
"Connection %i: Unloaded plugin: ID %i: %s",
			    con_num, id + 1, path
			);
		    }
		}
		else
		{
		    buf = (char *)malloc(
			40 + STRLEN(path) + 80
		    );
		    sprintf(
			buf,
"Server: Unloaded plugin: ID %i: %s",
			id + 1, path
		    );
		}
		LogAppend(fname.log, buf);
		free(buf);
	    }

#if defined(__linux__) || defined(__SOLARIS__)
	    /* Unload */
	    dlclose(pi->handle);
#endif
	    pi->handle = NULL;		/* Mark unloaded */
	}

	/* Delete resources */
	free(pi->name);
	strlistfree(pi->argv, pi->argc);
	free(pi->ctx);
	free(pi);
	plugin[id] = NULL;
}

/*
 *	Unloads and deletes all plugins.
 */
void PluginUnloadAll(void)
{
	int i;

	for(i = total_plugins - 1; i >= 0; i--)
	    PluginUnload(i, -1);

	free(plugin);
	plugin = NULL;
	total_plugins = 0;
}


/*
 *	Calls the help commands list function on the plugin (if any)
 *	and returns a pointer to a list of strings.
 *
 *	The returned strings should not be modified or deleted.
 */
char **PluginHelpCommandList(
	plugin_id_t id,
	int *ncommands
)
{
	plugin_struct *pi = PluginGetPointerFromID(id);

	if(ncommands != NULL)
	    *ncommands = 0;

	if(pi == NULL)
	    return(NULL);

	if(pi->help_commands_list_func == NULL)
	    return(NULL);

	return(pi->help_commands_list_func(
	    ncommands, pi->ctx
	));
}

/*
 *	Sends the command specified by cmd with the argument specified
 *	by arg to the plugin's command callback (if any) and returns
 *	1 if the command was handled or 0 if it was not.
 */
int PluginCommand(
	plugin_id_t id,
	int con_num, int obj_num,
	const char *cmd, const char *arg,
	int uid, int gid 
)
{
	plugin_struct *pi = PluginGetPointerFromID(id);

	/* Command and argument must be given (both may not be NULL) */
	if((cmd == NULL) || (arg == NULL))
	    return(0);

	if(pi == NULL)
	    return(0);

	if(pi->command_func == NULL)
	    return(0);

	return(pi->command_func(
	    cmd,
	    arg,
	    con_num,
	    obj_num,
	    uid, gid,
	    pi->ctx
	));
}

/*
 *	Sends the hail or message to the specified plugin's hail notify
 *	callback (if any).
 *
 *	If the message is NULL then this will be treated as a hail,
 *	otherwise this will be treated as a com message.
 */
void PluginHailNotify(
	plugin_id_t id,
	int con_num,
	int obj_num,                     
	int tar_obj_num,
	int channel,    
	const char *message
)
{
	plugin_struct *pi = PluginGetPointerFromID(id);
	if(pi == NULL)
	    return;

	if(pi->hail_notify_func == NULL)
	    return;

	pi->hail_notify_func(
	    con_num,
	    obj_num,
	    tar_obj_num,
	    channel,
	    message,
	    pi->ctx
	);
}

/*
 *	Notifies the plugin about an object being hit.
 */
void PluginHitNotify(
	plugin_id_t id,
	const int wobj_num,
	const int tobj_num,
	const float bearing,	/* Target to weapon */
	const float dmg_structure,
	const float dmg_shield
)
{
	const float dmg_total = dmg_structure + dmg_shield;
	plugin_struct *pi = PluginGetPointerFromID(id);
	if(pi == NULL)
	    return;

	if(pi->hit_notify_func == NULL)
	    return;

	pi->hit_notify_func(
	    wobj_num,
	    tobj_num,
	    dmg_total,
	    bearing,
	    dmg_structure,
	    dmg_shield,
	    pi->ctx
	);
}         

/*
 *	Notifies the plugin about an object being destroyed.
 */
void PluginDestroyNotify(
	plugin_id_t id,
	int reason,		/* One of SW_DESTROY_REASON_* */
	int destroyed_obj_num,
	int destroyer_obj_num,
	int destroyer_obj_owner_num
)
{
	plugin_struct *pi = PluginGetPointerFromID(id);
	if(pi == NULL)
	    return;

	if(pi->destroy_notify_func == NULL)
	    return;

	pi->destroy_notify_func(
	    reason,
	    destroyed_obj_num,
	    destroyer_obj_num,
	    destroyer_obj_owner_num,
	    pi->ctx
	);
}


#endif	/* PLUGIN_SUPPORT */
