#ifdef PLUGIN_SUPPORT

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/objects.h"
#include "../include/unvmatch.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"
#include "../include/swsoundcodes.h"
#include "../include/swexporteventscodes.h"

#include "plugins.h"
#include "net.h"
#include "swserv.h"
#include "weapons.h"
#include "sysparm.h"


static void get_version(int *major, int *minor, int *release);

static void set_data(SWServContext *ctx, void *data);
static void *get_data(SWServContext *ctx);           

static void print(const char *s);
static void print_err(const char *s);

static int con_get_obj_num(int con_num);
static void con_print(int con_num, const char *msg);
static void con_print_dialog(int con_num, int code, const char *msg);

static int sysparm_get_i(const char *parm);
static float sysparm_get_f(const char *parm);
static const char *sysparm_get_s(const char *parm);

static xsw_object_struct *obj_get_pointer(
	xsw_object_struct **list, int total, int obj_num
);
static int obj_create(xsw_object_struct ***list, int *total, int t);
static void obj_recycle(xsw_object_struct ***list, int *total, int n);
static int obj_get_player_owner(xsw_object_struct **list, int total, int n);
static int obj_is_connected(xsw_object_struct **list, int total, int n);
static int obj_is_hidden(xsw_object_struct **list, int total, int n);
static unsigned int wep_get_flags(xsw_object_struct **list, int total, int n);
static void wep_set_flags(
	xsw_object_struct **list, int total,
	int n, unsigned int flags
);
static void wep_unset_flags(
	xsw_object_struct **list, int total,
	int n, unsigned int flags
);
static int obj_in_vector_range(
	xsw_object_struct **list, int total,
	int src_obj_num, int tar_obj_num,
	float range,		/* In real units */
	float bearing,		/* In radians */
	float variance		/* In radians */
);
static void obj_fire_weapon(
	xsw_object_struct **list, int total, int obj_num,
	float freq, float yield
);
static void obj_send_message(
	xsw_object_struct **list, int total, int obj_num,
	int tar_obj_num,	/* Can be -1 for all */
	int channel,
	const char *msg
);
static void obj_sync_connections(
	xsw_object_struct **list, int total, int obj_num,
	char is_new
);

static xsw_object_struct *opm_get_pointer(
	xsw_object_struct **list, int total, int opm_num
);
static int opm_get_by_name(
	xsw_object_struct **list, int total, const char *name
);
static int opm_model_object(
	xsw_object_struct **list, int total, int opm_num,
	int obj_num
);


void PluginSetupContext(SWServContext *ctx);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the version number.
 */
static void get_version(int *major, int *minor, int *release)
{
	if(major != NULL)
	    *major = PROG_VERSION_MAJOR;
	if(minor != NULL)
	    *minor = PROG_VERSION_MINOR;
	if(release != NULL)
	    *release = PROG_VERSION_RELEASE;
}

/*
 *	Sets the user data.
 */
static void set_data(SWServContext *ctx, void *data)
{
	if(ctx == NULL)
	    return;

	ctx->data = data;
}

/*
 *	Gets the user data.
 */
static void *get_data(SWServContext *ctx)
{
	return((ctx != NULL) ? ctx->data : NULL);
}


/*
 *	Prints the specified message to stdout.
 */
static void print(const char *s)
{
	if(s == NULL)
	    return;

	if(sysparm.console_quiet)
	    return;

	fprintf(
	    stdout,
	    "%s",
	    s
	);
	fflush(stdout);
}

/*
 *	Prints the specified message to stderr.
 */
static void print_err(const char *s)
{
	if(s == NULL)
	    return;

	if(sysparm.console_quiet)
	    return;

	fprintf(
	    stderr,
	    "%s",
	    s
	);
	fflush(stdout);
}


/*
 *	Gets the object number associated with the specified connection.
 */
static int con_get_obj_num(int con_num)
{
	if(ConIsLoggedIn(con_num))
	    return(connection[con_num]->object_num);
	else
	    return(-1);
}

/*
 *	Prints the specified message to the specified connection.
 */
static void con_print(int con_num, const char *msg)
{
	NetSendMessage(con_num, msg);
}

/*
 *	Prints the specified message to the specified connection.
 *
 *	The code can be one of CS_SYSMESG_CODE_*.
 */
static void con_print_dialog(int con_num, int code, const char *msg)
{
	NetSendSysMessage(con_num, code, msg);
}


/*
 *	Gets the specified sysparm as an int.
 */
static int sysparm_get_i(const char *parm)
{
	sw_legend_struct *legend = &sw_legend;
	sws_sysparm_struct *list = &sysparm;

	if(!strcasecmp(parm, "allow_guest"))
	    return((int)list->allow_guest);
	else if(!strcasecmp(parm, "allow_guest_signup"))
	    return((int)list->allow_guest_signup);
	else if(!strcasecmp(parm, "asteroid_field_damage_int"))
	    return((int)legend->asteroid_field_damage_int);      
	else if(!strcasecmp(parm, "asteroid_field_damage_rate"))
	    return((int)legend->asteroid_field_damage_rate);      
	else if(!strcasecmp(parm, "cease_fire"))
	    return((int)list->cease_fire);
	else if(!strcasecmp(parm, "cloak_antimatter_rate"))
	    return((int)legend->cloak_antimatter_rate);
	else if(!strcasecmp(parm, "com_range"))
	    return((int)legend->com_range);
	else if(!strcasecmp(parm, "con_notify"))
	    return((int)list->con_notify);
	else if(!strcasecmp(parm, "console_quiet"))
	    return((int)list->console_quiet);

	else if(!strcasecmp(parm, "damage_control_rate"))
	    return((int)legend->damage_control_rate);
	else if(!strcasecmp(parm, "disarmed_weapon_lifespan"))
	    return((int)legend->disarmed_weapon_lifespan);
	else if(!strcasecmp(parm, "engine_antimatter_rate"))
	    return((int)legend->engine_antimatter_rate);

	else if(!strcasecmp(parm, "hide_players_home"))
	    return((int)list->hide_players_home);

	else if(!strcasecmp(parm, "hit_player_bonus"))
	    return((int)list->hit_player_bonus);

	else if(!strcasecmp(parm, "homes_destroyable"))
	    return((int)list->homes_destroyable);

	else if(!strcasecmp(parm, "int_system_check"))
	    return((int)list->int_system_check);
	else if(!strcasecmp(parm, "int_new_connection_poll"))
	    return((int)list->int_new_connection_poll);
	else if(!strcasecmp(parm, "int_object_values"))
	    return((int)list->int_object_values);
	else if(!strcasecmp(parm, "int_weapon_values"))
	    return((int)list->int_weapon_values);
	else if(!strcasecmp(parm, "int_memory_clean"))
	    return((int)list->int_memory_clean);
	else if(!strcasecmp(parm, "int_aux_con_stats"))
	    return((int)list->int_aux_con_stats);
	else if(!strcasecmp(parm, "int_os_stats"))
	    return((int)list->int_os_stats);
	else if(!strcasecmp(parm, "int_stats_export"))
	    return((int)list->int_stats_export);
	else if(!strcasecmp(parm, "int_unv_save"))
	    return((int)list->int_unv_save);

	else if(!strcasecmp(parm, "killer_gets_credits"))
	    return((int)list->killer_gets_credits);
	else if(!strcasecmp(parm, "log_general"))
	    return((int)list->log_general);
	else if(!strcasecmp(parm, "log_events"))
	    return((int)list->log_events);
	else if(!strcasecmp(parm, "log_net"))
	    return((int)list->log_net);
	else if(!strcasecmp(parm, "log_errors"))
	    return((int)list->log_errors);

	else if(!strcasecmp(parm, "login_timeout"))
	    return((int)list->login_timeout);

	else if(!strcasecmp(parm, "max_aux_connections"))
	    return((int)list->max_aux_connections);
	else if(!strcasecmp(parm, "max_connections"))
	    return((int)list->max_connections);
	else if(!strcasecmp(parm, "max_failed_logins"))
	    return((int)list->max_failed_logins);
	else if(!strcasecmp(parm, "max_guests"))
	    return((int)list->max_guests);

	else if(!strcasecmp(parm, "movement_limit_tolorance"))
	    return((int)list->movement_limit_tolorance);
        else if(!strcasecmp(parm, "on_fire_hp_coeff"))
            return((int)legend->on_fire_hp_coeff);

	else if(!strcasecmp(parm, "recycled_objects_max"))
	    return((int)list->recycled_objects_max);

	else if(!strcasecmp(parm, "report_destroyed_weapons"))
	    return((int)list->report_destroyed_weapons);
	else if(!strcasecmp(parm, "respawn_players"))
	    return((int)list->respawn_players);
	else if(!strcasecmp(parm, "send_starchart"))
	    return((int)list->send_starchart);
	else if(!strcasecmp(parm, "single_connection"))
	    return((int)list->single_connection);

	else if(!strcasecmp(parm, "tractor_beam_antimatter_rate"))
	    return((int)legend->tractor_beam_antimatter_rate);
	else if(!strcasecmp(parm, "tractor_beam_range"))
	    return((int)legend->tractor_beam_range);

	else if(!strcasecmp(parm, "transporter_range"))
	    return((int)legend->transporter_range);

	else if(!strcasecmp(parm, "use_server_ai"))
	    return((int)list->use_server_ai);

	return(0);
}

/*
 *	sysparm_get_f() wrapper.
 */
static float sysparm_get_f(const char *parm)
{
	sw_legend_struct *legend = &sw_legend;
	sws_sysparm_struct *list = &sysparm;

	if(!strcasecmp(parm, "allow_guest")) 
	    return((float)list->allow_guest);
	else if(!strcasecmp(parm, "allow_guest_signup"))
	    return((float)list->allow_guest_signup);
	else if(!strcasecmp(parm, "asteroid_field_damage_int"))
	    return((float)legend->asteroid_field_damage_int);
	else if(!strcasecmp(parm, "asteroid_field_damage_rate"))
	    return((float)legend->asteroid_field_damage_rate);
	else if(!strcasecmp(parm, "cease_fire")) 
	    return((float)list->cease_fire);
	else if(!strcasecmp(parm, "cloak_antimatter_rate"))
	    return((float)legend->cloak_antimatter_rate);
	else if(!strcasecmp(parm, "com_range"))
	    return((float)legend->com_range);
	else if(!strcasecmp(parm, "con_notify"))
	    return((float)list->con_notify);
	else if(!strcasecmp(parm, "console_quiet"))
	    return((float)list->console_quiet);

	else if(!strcasecmp(parm, "damage_control_rate"))
	    return((float)legend->damage_control_rate);
	else if(!strcasecmp(parm, "disarmed_weapon_lifespan"))
	    return((float)legend->disarmed_weapon_lifespan);
	else if(!strcasecmp(parm, "engine_antimatter_rate"))
	    return((float)legend->engine_antimatter_rate);

	else if(!strcasecmp(parm, "hide_players_home"))
	    return((float)list->hide_players_home);

	else if(!strcasecmp(parm, "hit_player_bonus"))
	    return((float)list->hit_player_bonus);

	else if(!strcasecmp(parm, "homes_destroyable"))
	    return((float)list->homes_destroyable);

	else if(!strcasecmp(parm, "int_system_check"))
	    return((float)list->int_system_check);
	else if(!strcasecmp(parm, "int_new_connection_poll"))
	    return((float)list->int_new_connection_poll);
	else if(!strcasecmp(parm, "int_object_values"))
	    return((float)list->int_object_values);
	else if(!strcasecmp(parm, "int_weapon_values"))
	    return((float)list->int_weapon_values);
	else if(!strcasecmp(parm, "int_memory_clean"))
	    return((float)list->int_memory_clean);
	else if(!strcasecmp(parm, "int_aux_con_stats")) 
	    return((float)list->int_aux_con_stats);
	else if(!strcasecmp(parm, "int_os_stats"))
	    return((float)list->int_os_stats);
	else if(!strcasecmp(parm, "int_stats_export"))
	    return((float)list->int_stats_export);
	else if(!strcasecmp(parm, "int_unv_save"))
	    return((float)list->int_unv_save);

	else if(!strcasecmp(parm, "killer_gets_credits"))
	    return((float)list->killer_gets_credits);
	else if(!strcasecmp(parm, "log_general"))
	    return((float)list->log_general);
	else if(!strcasecmp(parm, "log_events"))
	    return((float)list->log_events);
	else if(!strcasecmp(parm, "log_net"))
	    return((float)list->log_net);
	else if(!strcasecmp(parm, "log_errors"))
	    return((float)list->log_errors);
				       
	else if(!strcasecmp(parm, "login_timeout"))
	    return((float)list->login_timeout);

	else if(!strcasecmp(parm, "max_aux_connections"))
	    return((float)list->max_aux_connections);
	else if(!strcasecmp(parm, "max_connections"))
	    return((float)list->max_connections);
	else if(!strcasecmp(parm, "max_failed_logins"))
	    return((float)list->max_failed_logins);
	else if(!strcasecmp(parm, "max_guests"))
	    return((float)list->max_guests);

	else if(!strcasecmp(parm, "movement_limit_tolorance"))
	    return((float)list->movement_limit_tolorance);
        else if(!strcasecmp(parm, "on_fire_hp_coeff"))
            return((float)legend->on_fire_hp_coeff);

	else if(!strcasecmp(parm, "recycled_objects_max"))
	    return((float)list->recycled_objects_max);

	else if(!strcasecmp(parm, "report_destroyed_weapons"))
	    return((float)list->report_destroyed_weapons);
	else if(!strcasecmp(parm, "respawn_players"))
	    return((float)list->respawn_players);
	else if(!strcasecmp(parm, "send_starchart"))
	    return((float)list->send_starchart);
	else if(!strcasecmp(parm, "single_connection"))
	    return((float)list->single_connection);

	else if(!strcasecmp(parm, "tractor_beam_antimatter_rate"))
	    return((float)legend->tractor_beam_antimatter_rate);  
	else if(!strcasecmp(parm, "tractor_beam_range"))
	    return((float)legend->tractor_beam_range);

	else if(!strcasecmp(parm, "transporter_range"))
	    return((float)legend->transporter_range);

	else if(!strcasecmp(parm, "use_server_ai"))
	    return((float)list->use_server_ai);

	return(0.0f);
}

/*
 *	sysparm_get_s() wrapper.
 */
static const char *sysparm_get_s(const char *parm)
{
	sws_sysparm_struct *list = &sysparm;

	if(!strcasecmp(parm, "guest_login_name"))
	    return(list->guest_login_name);

	else if(!strcasecmp(parm, "mesg_wrong_login"))
	    return(list->mesg_wrong_login);
	else if(!strcasecmp(parm, "mesg_no_guests"))
	    return(list->mesg_no_guests);
	else if(!strcasecmp(parm, "mesg_welcome"))
	    return(list->mesg_welcome);
	else if(!strcasecmp(parm, "mesg_leave"))
	    return(list->mesg_leave);

	return(NULL);
}


/*
 *	obj_get_pointer() wrapper.
 */
static xsw_object_struct *obj_get_pointer(
	xsw_object_struct **list, int total, int obj_num
)
{
	if((list == NULL) || (obj_num < 0) || (obj_num >= total))
 	    return(NULL);
	else
	    return(list[obj_num]);
}

/*
 *	obj_create() wrapper.
 */
static int obj_create(xsw_object_struct ***list, int *total, int t)
{
	int i, n;
	xsw_object_struct *obj;

	/* Search for available pointer */
	for(i = 0; i < *total; i++)
	{
	    obj = (*list)[i];
	    if(obj == NULL)
		break;
	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		break;
	}
	if(i < *total)
	{
	    n = i;      /* Got available pointer */
	}
	else
	{
	    /* Need to allocate more pointers */
	    n = *total;
	    *total = *total + 1;

	    *list = (xsw_object_struct **)realloc(
		*list,
		(*total) * sizeof(xsw_object_struct *)
	    );
	    if(*list == NULL)
	    {
		*total = 0;
		return(-3);
	    }

	    (*list)[n] = NULL;
	}

	/* Allocate new object structure as needed */
	if((*list)[n] == NULL)
	{
	    (*list)[n] = (xsw_object_struct *)calloc(
		1, sizeof(xsw_object_struct)
	    );
	    if((*list)[n] == NULL)
		return(-1);
	}

	obj = (*list)[n];
	UNVResetObject(obj);
	obj->type = t;

	return(n);
}

/*
 *	obj_recycle() wrapper.
 */
static void obj_recycle(xsw_object_struct ***list, int *total, int n)
{
        const time_t	cur_ms = cur_millitime,
                        cur_sec = cur_systime;
        const float tc = time_compensation;

	if(*list == xsw_object)
	{
	    NetSendRecycleObject(-1, n);
	    DBRecycleObject(n, cur_ms, cur_sec, tc);
	}
	else 
	{
	    if(!UNVIsObjectGarbage(*list, *total, n))
		UNVResetObject((*list)[n]);
	}
}

/*
 *	Returns the player object that owns the specified object (not
 *	always just the immediate owner).
 */
static int obj_get_player_owner(xsw_object_struct **list, int total, int n)
{
	if(list == xsw_object)
	    return(DBGetPlayerOwner(n));
	else
	    return(-1);
}

/*
 *	Checks if the object is a player that is currently connected.
 */
static int obj_is_connected(xsw_object_struct **list, int total, int n)
{
	xsw_object_struct *obj = obj_get_pointer(
	    list, total, n
	);
	if(obj == NULL)
	    return(0);

	return((obj->ext_flags & SWS_OBJ_FLAG_CONNECTED) ? 1 : 0);
}

/*
 *	Checks if the object is marked hidden.
 */
static int obj_is_hidden(xsw_object_struct **list, int total, int n)
{
	xsw_object_struct *obj = obj_get_pointer(
	    list, total, n
	);
	if(obj == NULL)
	    return(0);

	return((obj->ext_flags & SWS_OBJ_FLAG_HIDEFROMCON) ? 1 : 0);
}

/*
 *	Gets the weapon object's flags.
 */
static unsigned int wep_get_flags(
	xsw_object_struct **list, int total, int n
)
{
	unsigned int flags, rtn_flags;
	xsw_object_struct *obj = obj_get_pointer(
	    list, total, n
	);
	if(obj == NULL)   
	    return(0);

	flags = obj->ext_flags;
	rtn_flags = 0;
	if(flags & SWS_OBJ_FLAG_WEP_PENETRATE_SHIELD)
	    rtn_flags |= SW_WEP_FLAG_PENETRATE_SHIELD;
	if(flags & SWS_OBJ_FLAG_WEP_DESTROYABLE)
	    rtn_flags |= SW_WEP_FLAG_DESTROYABLE;
	return(rtn_flags);
}

/*
 *	Sets the weapon object's flags.
 */
static void wep_set_flags(
	xsw_object_struct **list, int total,
	int n, unsigned int flags
)
{
	xsw_object_struct *obj = obj_get_pointer(
	    list, total, n
	);
	if(obj == NULL)   
	    return;

	if(flags & SW_WEP_FLAG_PENETRATE_SHIELD)
	    obj->ext_flags |= SWS_OBJ_FLAG_WEP_PENETRATE_SHIELD;
	if(flags & SW_WEP_FLAG_DESTROYABLE)
	    obj->ext_flags |= SWS_OBJ_FLAG_WEP_DESTROYABLE;
        if(flags & SW_WEP_FLAG_SMOKE_TRAIL)
            obj->ext_flags |= SWS_OBJ_FLAG_SMOKE_TRAIL;
        if(flags & SW_WEP_FLAG_SHOCKWAVE_HIT)
            obj->ext_flags |= SWS_OBJ_FLAG_WEP_SHOCKWAVE_HIT;
}

/*
 *	Unsets the weapon object's flags.
 */
static void wep_unset_flags(
	xsw_object_struct **list, int total,
	int n, unsigned int flags
)
{
	xsw_object_struct *obj = obj_get_pointer(
	    list, total, n
	);
	if(obj == NULL)
	    return;

	if(flags & SW_WEP_FLAG_PENETRATE_SHIELD)
	    obj->ext_flags &= ~SWS_OBJ_FLAG_WEP_PENETRATE_SHIELD;
	if(flags & SW_WEP_FLAG_DESTROYABLE)
	    obj->ext_flags &= ~SWS_OBJ_FLAG_WEP_DESTROYABLE;
        if(flags & SW_WEP_FLAG_SMOKE_TRAIL)
            obj->ext_flags &= ~SWS_OBJ_FLAG_SMOKE_TRAIL;
        if(flags & SW_WEP_FLAG_SHOCKWAVE_HIT)
            obj->ext_flags &= ~SWS_OBJ_FLAG_WEP_SHOCKWAVE_HIT;
}

/*
 *	obj_in_vector_range() wrapper.
 */
static int obj_in_vector_range(
	xsw_object_struct **list, int total,
	int src_obj_num, int tar_obj_num,
	float range,		/* In real units */
	float bearing,		/* In radians */
	float variance		/* In radians */
)
{
	return(SWMath3DInVectorContact(
	    list, total,
	    src_obj_num, tar_obj_num,
	    bearing, variance,
	    range * 1000.0f	/* Convert to screen units */
	));
}

/*
 *	Instructs the object to fire its selected weapon.
 */
static void obj_fire_weapon(
	xsw_object_struct **list, int total, int obj_num,
	float freq, float yield
)
{
	const time_t	cur_ms = cur_millitime,
			cur_sec = cur_systime;
	const float tc = time_compensation;
	int con_num, sel_wep_num, objects_created;
	char loc_str[256];
	xsw_object_struct *obj;
	xsw_weapon_struct *sel_wep;

	if(UNVIsObjectGarbage(list, total, obj_num))
	    return;

	obj = list[obj_num];

	/* Sanitize weapon power yield */
	if(yield < 0.0f)
	    yield = 0.0f;
	else if(yield > 1.0f)
	    yield = 1.0f;

	/* Get the selected weapon */
	sel_wep_num = obj->selected_weapon;
	if((sel_wep_num < 0) || (sel_wep_num >= obj->total_weapons))
	    return;

	sel_wep = obj->weapons[sel_wep_num];
	if(sel_wep == NULL)
	    return;

	/* Did we wait the minimum delay before we can fire again? */
	if((sel_wep->delay + sel_wep->last_used) >=                  
	    cur_ms
	)
	    return;

	/* Is the object cloaked and the weapon may not be used
	 * while cloaked?
	 */
	if((obj->cloak_state == SW_CLOAK_STATE_UP) &&
	   !(sel_wep->flags & SW_WEP_FLAG_FIRE_CLOAKED)
	)                   
	    return;

	/* If weapon is expendable then there is any left */
	if(sel_wep->emission_type == SW_WEP_EMISSION_PROJECTILE)
	{
	    if(sel_wep->amount <= 0)
		return;
	}

	/* Does the object have enough power to fire weapon? */
	if(obj->power < (sel_wep->create_power * yield))
	    return;

	/* Search for a connetion assocated with this object */
	con_num = ConGetByObject(obj_num);

	/* Fire weapon by emission type */
	/* Stream weapons */
	if(sel_wep->emission_type == SW_WEP_EMISSION_STREAM)
	{
	    objects_created = WeaponsFire(
		sel_wep->ocs_code,              /* OCS */
		obj_num,			/* Owner */
		sel_wep->emission_type,         /* Emission type */
		obj->x,
		obj->y,
		obj->heading,
		obj->pitch,
		sel_wep->damage_hull * yield,
		sel_wep->damage_shields * yield,
		sel_wep->range,
		freq,
		sel_wep->flags,
		cur_ms, cur_sec, tc
	    );
	}     
	/* Projectile */
	else if(sel_wep->emission_type == SW_WEP_EMISSION_PROJECTILE)
	{                   
	    objects_created = WeaponsFire(
		sel_wep->ocs_code,              /* OCS */
		obj_num,			/* Owner */
		sel_wep->emission_type,         /* Emission type */
		obj->x,
		obj->y,
		obj->heading,  
		obj->pitch,
		sel_wep->damage_hull * yield,
		sel_wep->damage_shields * yield,
		sel_wep->range,
		freq,
		sel_wep->flags,
		cur_ms, cur_sec, tc
	    );
	}     
	/* All else assume pulse */
	else
	{
	    objects_created = WeaponsFire(
		sel_wep->ocs_code,              /* OCS */
		obj_num,			/* Owner */
		sel_wep->emission_type,         /* Emission type */
		obj->x,
		obj->y,
		obj->heading,
		obj->pitch,
		sel_wep->damage_hull * yield,
		sel_wep->damage_shields * yield,
		sel_wep->range,
		freq,
		sel_wep->flags,
		cur_ms, cur_sec, tc
	    );
	}     

	/* No weapons fired? */
	if(objects_created <= 0)
	    return;


	/* Format location string */
	UNVLocationFormatString(
	    loc_str,
	    &obj->sect_x, &obj->sect_y, &obj->sect_z,
	    &obj->x, &obj->y, &obj->z,  
	    sizeof(loc_str)
	);


	/* Log weapons fire */
	if(sysparm.log_events)
	{     
	    char text[(2 * SW_OBJ_NAME_MAX) + 80 + 80 + sizeof(loc_str)];
	    sprintf(
		text,
"%s: Fired weapon `%s' (ocs %i) at %s",   
		DBGetFormalNameStr(obj_num),
		DBGetOCSOPMName(sel_wep->ocs_code),
		sel_wep->ocs_code,
		loc_str
	    );
	    LogAppend(fname.log, text);
	}

	/* Mark the last time it was fired */
	sel_wep->last_used = cur_ms;

	/* Consume power that was needed to create this weapon */
	switch(sel_wep->emission_type)
	{
	  case SW_WEP_EMISSION_PROJECTILE:
	    obj->power -= sel_wep->create_power * objects_created * yield;

	    /* Decrease stock for projectiles */
	    sel_wep->amount -= objects_created;
	    if(sel_wep->amount < 0) 
		sel_wep->amount = 0;

	    /* Play fire sound? */
	    if(!(sel_wep->flags & SW_WEP_FLAG_NO_FIRE_SOUND) &&
		(con_num > -1)
	    )
		NetSendPlaySound(
		    con_num,
		    (sel_wep->fire_sndref > 0) ?
			sel_wep->fire_sndref :
			SW_SNDREF_FIRE_PROJECTILE,
		    1.0f, 1.0f
		);

	    /* Export projectile weapons fire */
	    if(1)
	    {
		char text[(3 * SW_OBJ_NAME_MAX) + 512];
		sprintf(
		    text,
		    "%s fired %s at %s.",
		    DBGetFormalNameStr(obj_num),
		    DBGetOCSOPMName(sel_wep->ocs_code),
		    loc_str
		);
		ExportEvents(
		    fname.events_export,
		    EVST_TYPE_PROJECTILEWEAPON_FIRED,
		    text
		);
	    }
	    break;

	  case SW_WEP_EMISSION_PULSE:
	    obj->power -= sel_wep->create_power * objects_created * yield;

	    /* Play fire sound? */
	    if(!(sel_wep->flags & SW_WEP_FLAG_NO_FIRE_SOUND) &&
	       (con_num > -1)
	    )
		NetSendPlaySound(
		    con_num,
		    (sel_wep->fire_sndref > 0) ?
			sel_wep->fire_sndref :
			SW_SNDREF_FIRE_PULSE,
		    1.0f, 1.0f
		);
	    break;

	  default:      /* SW_WEP_EMISSION_STREAM */
	    obj->power -= sel_wep->create_power * objects_created * yield;

	    /* Play fire sound */
	    if(!(sel_wep->flags & SW_WEP_FLAG_NO_FIRE_SOUND) &&
	       (con_num > -1)
	    )
		NetSendPlaySound(
		    con_num,
		    (sel_wep->fire_sndref > 0) ?
			sel_wep->fire_sndref :
			SW_SNDREF_FIRE_STREAM,
		    1.0f, 1.0f
		);
	    break;
	}

	/* Sanitize power on the firing object */
	if(obj->power < 0.0f)
	    obj->power = 0.0f;

	/* Report new object and weapons values to objects nearby */
	NetFireWeaponSyncObjectValues(
	    obj, obj_num, sel_wep_num,
	    cur_ms, cur_sec, tc
	);
}

/*
 *	Sends a message to the object specified by tar_obj_num
 *	on the specified com channel.
 *
 *	If tar_obj_num is -1 then the message is sent to all objects
 *	in com range of obj_num.
 *
 *	If msg is NULL then a hail will be sent.
 */
static void obj_send_message(
	xsw_object_struct **list, int total, int obj_num,
	int tar_obj_num,        /* Can be -1 for all */
	int channel,
	const char *msg
)
{
	sw_legend_struct *legend = &sw_legend;
	xsw_object_struct *obj;

	if(list != xsw_object)
	    return;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Send to specific object? */
	if(tar_obj_num > -1)
	{
	    /* Object in com range? */
	    if(SWMath3DInRange(
		xsw_object, total_objects,
		obj_num, tar_obj_num, legend->com_range
	    ))
	    { 
		const int tar_con_num = ConGetByObject(tar_obj_num);
		xsw_object_struct *tar_obj = xsw_object[tar_obj_num];
		if((tar_obj->com_channel == channel) && (tar_con_num > -1))
		{
		    /* Get bearing from target to source object */
		    const float bearing = SWMathCoordinateDeltaVector(
			obj->x - tar_obj->x,
			obj->y - tar_obj->y
		    );

		    /* Send message to that connection */
		    if(msg != NULL)
			NetSendComMesg(
			    tar_con_num,
			    obj_num,		/* Sender */
			    tar_obj_num,	/* Receiver */
			    bearing,		/* Target to source */
			    channel,
			    msg,
			    0			/* Options */
			);
		    else
			NetSendHail(
			    tar_con_num,
			    obj_num,		/* Sender */
			    tar_obj_num,	/* Receiver */
			    bearing,		/* Target to source */
			    channel
			);
		}     
	    }     
	}
	else
	{
	    int i, tar_con_num;
	    float bearing;
	    xsw_object_struct *tar_obj;

	    /* Send message to all objects in com range */  
	    for(i = 0; i < total_objects; i++)
	    { 
		/* Object not valid or not in com range? */
		if(!SWMath3DInRange(
		    xsw_object, total_objects,
		    obj_num, i, legend->com_range
		))
		    continue;

		tar_obj = xsw_object[i];

		/* Skip of object_count is not listening on channel */
		if(tar_obj->com_channel != channel)
		    continue;

		/* Get the connection of this object (if any) */
		tar_con_num = ConGetByObject(i);
		if(tar_con_num < 0) 
		    continue;

		/* Get bearing from target to source object */
		bearing = SWMathCoordinateDeltaVector(
		    obj->x - tar_obj->x,
		    obj->y - tar_obj->y
		);    

		/* Send message to that connection */
		if(msg != NULL)
		    NetSendComMesg(
			tar_con_num,
			obj_num,		/* Sender */
			-1,			/* General message */
			bearing,		/* Target to source */
			channel,
			msg,
			0			/* Options */
		    );
		else
		    NetSendHail(
			tar_con_num,
			obj_num,		/* Sender */
			-1,			/* Receiver */
			bearing,		/* Target to source */
			channel
		    );
	    }
	}

#ifdef PLUGIN_SUPPORT
	/* Forward this message to plugins */
	if(total_plugins > 0)
	{
	    int i;
	    for(i = 0; i < total_plugins; i++)
		PluginHailNotify(
		    i,
		    -1,			/* No connection */
		    obj_num,		/* Sender */
		    tar_obj_num,	/* Receiver or -1 for general */
		    channel,            /* Com channel */
		    msg                 /* Message */
		);
	}
#endif
}

/*
 *	Sends the specified object's current values to all connections
 *	who's object is in range with this object.
 *
 *	If is_new is true then a create object will be sent as well.
 */
static void obj_sync_connections(
	xsw_object_struct **list, int total, int obj_num,
	char is_new
)
{
	int con_num, wep_num, con_obj_num;
	xsw_object_struct *obj, *con_obj;
	sws_connection_struct *con;

	if(list != xsw_object)
	    return;

	if(DBIsObjectGarbage(obj_num))
	    return;

	obj = xsw_object[obj_num];

	/* Iterate through the connections list */
	for(con_num = 0; con_num < total_connections; con_num++)
	{
	    con = connection[con_num];
	    if(con == NULL)
		continue;

	    con_obj_num = con->object_num;
	    if(DBIsObjectGarbage(con_obj_num))
		continue;

	    con_obj = xsw_object[con_obj_num];

	    /* Object in range with connection object? */
	    if(!SWMath3DInRangePtr(
		con_obj, obj,
		con_obj->scanner_range
	    ))
		continue;

	    if(is_new)
		NetSendCreateObject(con_num, obj_num);
	    NetSendObjectMaximums(con_num, obj_num);
	    NetSendObjectValues(con_num, obj_num);
	    NetSendObjectName(con_num, obj_num);

	    for(wep_num = 0; wep_num < obj->total_weapons; wep_num++)
		NetSendWeaponValues(con_num, obj_num, wep_num);
	}   
}


/*
 *	opm_get_pointer() wrapper.
 */
static xsw_object_struct *opm_get_pointer(
	xsw_object_struct **list, int total, int opm_num
)
{
	if((list == NULL) || (opm_num < 0) || (opm_num >= total))
	    return(NULL);
	else
	    return(list[opm_num]);
}

/*
 *	opm_get_by_name() wrapper.
 */
static int opm_get_by_name(
	xsw_object_struct **list, int total, const char *name
)
{
	return(OPMGetByName(name, -1));
}

/*
 *	opm_model_object() wrapper.
 */
static int opm_model_object(
	xsw_object_struct **list, int total, int opm_num,
	int obj_num
)
{
	return(OPMModelObject(obj_num, opm_num));
}


/*
 *	Sets up the Plugin Context.
 */
void PluginSetupContext(SWServContext *ctx)
{
	ctx->data = NULL;

	ctx->cur_millitime = &cur_millitime;
	ctx->cur_systime = &cur_systime;
	ctx->lapsed_millitime = &lapsed_millitime;

	ctx->time_compensation = &time_compensation;

	ctx->xsw_object = &xsw_object;
	ctx->total_objects = &total_objects;

	ctx->legend = &sw_legend;

	ctx->opm = &opm;
	ctx->total_opms = &total_opms;

	ctx->ocs = (void ***)&ocs;
	ctx->total_ocss = &total_ocss;

	ctx->connection = (void ***)&connection;
	ctx->total_connections = &total_connections;

	ctx->set_data = set_data;
	ctx->get_data = get_data;

	ctx->con_is_connected = ConIsConnected;
	ctx->con_is_logged_in = ConIsLoggedIn;
	ctx->con_get_obj_num = con_get_obj_num;
	ctx->con_print = con_print;
	ctx->con_print_dialog = con_print_dialog;

	ctx->sysparm_get_i = sysparm_get_i;
	ctx->sysparm_get_f = sysparm_get_f;
	ctx->sysparm_get_s = sysparm_get_s;

	ctx->obj_is_garbage = UNVIsObjectGarbage;
	ctx->obj_get_pointer = obj_get_pointer;
	ctx->obj_create = obj_create;
	ctx->obj_recycle = obj_recycle;
	ctx->obj_get_player_owner = obj_get_player_owner;
	ctx->obj_is_connected = obj_is_connected;
	ctx->obj_is_hidden = obj_is_hidden;
	ctx->wep_get_flags = wep_get_flags;
	ctx->wep_set_flags = wep_set_flags;
	ctx->wep_unset_flags = wep_unset_flags;
	ctx->obj_in_same_sector = SWMath3DInSameSector;
	ctx->obj_in_contact = SWMath3DInContact;
	ctx->obj_in_range = SWMath3DInRange;
	ctx->obj_in_vector_range = obj_in_vector_range;
	ctx->obj_fire_weapon = obj_fire_weapon;
	ctx->obj_send_message = obj_send_message;
	ctx->obj_sync_connections = obj_sync_connections;

	ctx->opm_get_pointer = opm_get_pointer;
	ctx->opm_get_by_name = opm_get_by_name;
	ctx->opm_model_object = opm_model_object;

	ctx->get_version = get_version;

	ctx->print = print;
	ctx->print_err = print_err;

}


#endif	/* PLUGIN_SUPPORT */
