#include "blitting.h"


static void BlitBufCirclePixel8(void *p, const u_int32_t v);
static void BlitBufCirclePixel15(void *p, const u_int32_t v);
static void BlitBufCirclePixel16(void *p, const u_int32_t v);
static void BlitBufCirclePixel24(void *p, const u_int32_t v);
static void BlitBufCirclePixel32(void *p, const u_int32_t v);

static void BlitBufCircleCircle(
	unsigned int bpp,
	u_int8_t *tar_buf,     
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float thickness,     
	float radius,   
	const u_int32_t v,
	void (*pixel_op_func)(void *, const u_int32_t v)
);

void BlitBufCircle(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width,  unsigned int tar_height,
	unsigned int tar_bpl,
	float thickness,		/* In pixels */
	float radius,			/* In pixels */
	const WColorStruct *c
);


/* Maximum radius (in pixels) */
#define MAX_RADIUS		2000

#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static void BlitBufCirclePixel8(void *p, const u_int32_t v)
{
	*(u_int8_t *)p = (u_int8_t)v;
}

static void BlitBufCirclePixel15(void *p, const u_int32_t v)
{
	*(u_int16_t *)p = (u_int16_t)v;
}

static void BlitBufCirclePixel16(void *p, const u_int32_t v)
{
	*(u_int16_t *)p = (u_int16_t)v;
}

static void BlitBufCirclePixel24(void *p, const u_int32_t v)
{
	u_int8_t *p8 = (u_int8_t *)p;
	p8[0] = (u_int8_t)((v & 0x000000FF) >> 0);
	p8[1] = (u_int8_t)((v & 0x0000FF00) >> 8);
	p8[2] = (u_int8_t)((v & 0x00FF0000) >> 16);
}

static void BlitBufCirclePixel32(void *p, const u_int32_t v)
{
	*(u_int32_t *)p = v;
}


static void BlitBufCircleCircle(
	unsigned int bpp,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	float thickness,
	float radius,
	const u_int32_t v,
	void (*pixel_op_func)(void *, const u_int32_t v)
)
{
	int	x, y,
		q, q0, r,
		cx, cy;

	if((tar_buf == NULL) || (tar_width == 0) || (tar_height == 0))
	    return;

	/* Sanitize radius */
	r = (int)((radius < 0) ? (radius * -1) : (radius));
	if(r > MAX_RADIUS)
	    r = MAX_RADIUS;

	/* Circle entirely not visible on target? */
	if(((tar_x + r) < 0) ||
	   ((tar_y + r) < 0) ||
	   ((tar_x - r) >= (int)tar_width) ||
	   ((tar_y - r) >= (int)tar_height)
	)
	    return;

	q = q0 = r * r;
	x = 0;
	y = r;

	while(x <= y)
	{
	    if(q > q0)
	    {
		/* Outside circle */
		y--;
		q -= 2 * y;
	    }
	    else
	    {
		/* Inside circle */
		x++;
		q += 2 * x;
	    }


	    /* Draw for primary arc */

	    /* Translate */
	    cx = tar_x + x;
	    cy = tar_y + y;

	    /* Check if 'on screen' */
	    if((cx >= 0) && (cx < (int)tar_width) &&
	       (cy >= 0) && (cy < (int)tar_height)
	    )
	    {
		pixel_op_func(
		    tar_buf +
			(cy * (int)tar_bpl) +
			(cx * (int)bpp),
		    v
		);
	    }


	    /* Draw for vertical mirror */

	    /* Translate */
	    cx = tar_x + x;
	    cy = tar_y - y;

	    /* Check if 'on screen' */
	    if((cx >= 0) && (cx < (int)tar_width) &&
	       (cy >= 0) && (cy < (int)tar_height)
	    )
	    {
		pixel_op_func(
		    tar_buf + 
			(cy * (int)tar_bpl) + 
			(cx * (int)bpp),     
		    v
		);
	    }


	    /* Draw for horizontal mirror */

	    /* Translate */
	    cx = tar_x - x;
	    cy = tar_y + y;

	    /* Check if 'on screen' */
	    if((cx >= 0) && (cx < (int)tar_width) &&
	       (cy >= 0) && (cy < (int)tar_height)
	    )
	    {
		pixel_op_func(
		    tar_buf + 
			(cy * (int)tar_bpl) + 
			(cx * (int)bpp),     
		    v
		);
	    }


	    /* Draw for 180 degree rotation */

	    /* Translate */
	    cx = tar_x - x;
	    cy = tar_y - y;

	    /* Check if 'on screen' */
	    if((cx >= 0) && (cx < (int)tar_width) &&
	       (cy >= 0) && (cy < (int)tar_height)
	    )
	    {
		pixel_op_func(
		    tar_buf +
			(cy * (int)tar_bpl) +
			(cx * (int)bpp),
		    v
		);
	    }


	    /* Draw for secondary arc */

	    /* Translate */
	    cx = tar_x + y;
	    cy = tar_y + x;

	    /* Check if 'on screen' */
	    if((cx >= 0) && (cx < (int)tar_width) &&
	       (cy >= 0) && (cy < (int)tar_height)
	    )
	    {
		pixel_op_func(
		    tar_buf +
			(cy * (int)tar_bpl) +
			(cx * (int)bpp),
		    v
		);
	    }


	    /* Draw for vertical mirror */

	    /* Translate */
	    cx = tar_x + y;
	    cy = tar_y - x;

	    /* Check if 'on screen' */
	    if((cx >= 0) && (cx < (int)tar_width) &&
	       (cy >= 0) && (cy < (int)tar_height)
	    )
	    {
		pixel_op_func(
		    tar_buf +
			(cy * (int)tar_bpl) +
			(cx * (int)bpp),
		    v
		);
	    }


	    /* Draw for horizontal mirror */

	    /* Translate */
	    cx = tar_x - y;
	    cy = tar_y + x;

	    /* Check if 'on screen' */
	    if((cx >= 0) && (cx < (int)tar_width) &&
	       (cy >= 0) && (cy < (int)tar_height)
	    )
	    {
		pixel_op_func(
		    tar_buf +
			(cy * (int)tar_bpl) +
			(cx * (int)bpp),
		    v
		);
	    }


	    /* Draw for 180 degree rotation */

	    /* Translate */
	    cx = tar_x - y;
	    cy = tar_y - x;

	    /* Check if 'on screen' */
	    if((cx >= 0) && (cx < (int)tar_width) &&
	       (cy >= 0) && (cy < (int)tar_height)
	    )
	    {
		pixel_op_func(
		    tar_buf +
			(cy * (int)tar_bpl) +
			(cx * (int)bpp),
		    v
		);
	    }
	}
}


void BlitBufCircle(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf,
	int tar_x, int tar_y,
	unsigned int tar_width,  unsigned int tar_height,
	unsigned int tar_bpl,
	float thickness,		/* In pixels */
	float radius,			/* In pixels */
	const WColorStruct *c
)
{
	u_int32_t v;
	void (*pixel_op_func)(void *, u_int32_t v);

	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;

	switch(depth)
	{
	  case 8:
	    v = PACK8TO8(c->r, c->g, c->b);
	    break;
	  case 15:
	    v = PACK8TO15(c->r, c->g, c->b);
	    break;
	  case 16: 
	    v = PACK8TO16(c->r, c->g, c->b);
	    break;
	  case 24:
	  case 32:
	    v = PACK8TO32(c->a, c->r, c->g, c->b);
	    break;
	  default:
	    v = 0x00000000;
	    break;
	}

	switch(bpp)
	{
	  case 1:
	    pixel_op_func = BlitBufCirclePixel8;
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		pixel_op_func = BlitBufCirclePixel15;
		break;
	      case 16:
		pixel_op_func = BlitBufCirclePixel16;
		break;
	      default:  
		return;
		break;
	    }
	    break;
	  case 3: 
	    pixel_op_func = BlitBufCirclePixel24;
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24: 
		pixel_op_func = BlitBufCirclePixel24;
		break;
	      case 32:
		pixel_op_func = BlitBufCirclePixel32;
		break;
	      default:
		return;
		break;
	    }
	    break;
	  default:
	    return;
	    break;
	}

	BlitBufCircleCircle(
	    bpp,
	    tar_buf,
	    tar_x, tar_y,
	    tar_width, tar_height, tar_bpl,
	    thickness,
	    radius,
	    v, pixel_op_func
	);
}
