// client/blitstd.cpp
/*
				  Buffer Blitting: Standard Blitting Routines

	Functions:

	BlitBufNormal8()
	BlitBufNormal15()
	BlitBufNormal16()
	BlitBufNormal24()
	BlitBufNormal32()
	BlitBufNormal()

	BlitBufAbsolute8()
	BlitBufAbsolute15()
	BlitBufAbsolute16()
	BlitBufAbsolute24()
	BlitBufAbsolute32()
	BlitBufAbsolute()

	BlitBufAdditive8()
	BlitBufAdditive15()
	BlitBufAdditive16()
	BlitBufAdditive32()
	BlitBufAdditive()

	---

	These functions perform blitting of two buffers.
	A porition of the source buffer to a position on the destination
	buffer with `zooming' support.

	Zooming is currently limited to 0.01 to 1.00 ranges.

	1-16-00(KB):    Macro'ized inner pixel loops.
					Added XSWRect and support to clean up data management.
	1-17-00(KB):    Reorganized setup code for scaled blits.
	1-18-00(KB):    Macroized blit preparation code.
	1-19-00(KB):    Debugging changes.
 */

#include <assert.h>
#include <string.h>
#include "blitting.h"

/*
 *   Checks if tx, ty, sy, and sy are 'on the screen':
 *
 *      The variables beginning with a t means target, and respectively
 *  x, y, width, and height.
 *
 *      The variables beginning with a s means source, and respectively
 *      x, y, width, and height.
 */

#ifndef MAX
	#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
	#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#endif


/* XSWRect */
typedef struct tag_Rect
{

	int     left,
	top,
	right,
	bottom;

} XSWRect;


/* XSWRect utility macro's */
#define RECT_WIDTH(a)	(((a).right) - ((a).left))
#define RECT_HEIGHT(a)	(((a).bottom) - ((a).top))
#define RECT_VALID(a)	((RECT_WIDTH(a) > 0) && (RECT_HEIGHT(a) > 0))


/*
 *	Blit preparation macro:
 */
#define BLIT_PREP                                                               \
    float           Source_Step_X;                                              \
    float           Source_Step_Y;                                              \
    XSWRect         Source;                                                     \
    XSWRect         Dest;                                                       \
    float           Scale = (float)zoom * (float)magnification;                 \
    assert(tar_buf && src_buf);	                                               \
    assert((tar_width != 0) && (src_width != 0));                                 \
    assert((tar_height != 0) && (src_height != 0));                               \
    assert((src_copy_width != 0) && (src_copy_height != 0));                      \
    assert((zoom >= 0.0) && (magnification >= 0.0));                              \
    if((tar_x >= (int)tar_width) || (tar_y >= (int)tar_height))                   \
	return;                                                                 \
    if((src_x >= (int)src_width) || (src_y >= (int)src_height))                   \
	return;                                                                 \
    if(Scale > BLIT_MAGNIFICATION_MAX)						\
	Scale = BLIT_MAGNIFICATION_MAX;    					 \
    Source.left     =   src_x;                                                  \
    Source.top      =   src_y;                                                  \
    Source.right    =   src_x + src_copy_width;                                 \
    Source.bottom   =   src_y + src_copy_height;                                \
    Dest.left       =   tar_x;                                                  \
    Dest.top        =   tar_y;                                                  \
    Dest.right      =   tar_x + static_cast<int>(Scale*src_copy_width);         \
    Dest.bottom     =   tar_y + static_cast<int>(Scale*src_copy_height);        \
    Source_Step_X   = (float)RECT_WIDTH(Source) / (float)RECT_WIDTH(Dest);      \
    Source_Step_Y   = (float)RECT_HEIGHT(Source) / (float)RECT_HEIGHT(Dest);    \
    if(Dest.left < 0)                                                         \
    {                                                                           \
    	Source.left     -=  static_cast<int>(Dest.left * Source_Step_X);        \
	Dest.left       =   0;                                                  \
	if( Dest.left >= Dest.right )						\
	    return;								\
    }                                                                           \
    if( Dest.top < 0 )                                                          \
    {                                                                           \
	Source.top      -=  static_cast<int>(Dest.top * Source_Step_Y);         \
	Dest.top        =   0;                                                  \
	if( Dest.top >= Dest.bottom )						\
	    return;								\
    }                                                                           \
    if( Dest.right > (int)tar_width )                                           \
    {                                                                           \
	Source.right    +=  static_cast<int>((Dest.right - tar_width) * Source_Step_X);   \
	Dest.right      =   tar_width;                                          \
	if( Dest.right <= Dest.left )						\
	    return;                                   				\
    }                                                                           \
    if( Dest.bottom > (int)tar_height )                                              \
    {                                                                           \
	Source.bottom   +=  static_cast<int>((Dest.bottom-tar_height) * Source_Step_Y); \
	Dest.bottom     =   tar_height;                                         \
	if( Dest.bottom <= Dest.top )						\
	    return;       				                            \
    }                                                                           \
    if( !RECT_VALID(Source) || !RECT_VALID(Dest) )                              \
	return;



/*
	Pixel loop definition macro.

	Used to define the pixel loop in a consistent manner.
	Requires that PIX_PLOT is set according to requirements.
 */
#define PIX_LOOP(size)	{			\
    u_int32_t row, col;				\
    float source_y_frac = 0;			\
    for(row = 0; row < Dest_Height; row++) {	\
	float source_x_frac;			\
	const size *source_ptr_row = pSource + Source_Stride * \
	    (int)source_y_frac;			\
	source_x_frac = 0;			\
	for(col = 0; col < Dest_Width; col++) {	\
	    const size source = *(source_ptr_row + (int)source_x_frac); \
	    PIX_PLOT				\
	    source_x_frac += Source_Step_X;	\
	}					\
	pDest           += Dest_Stride;		\
	source_y_frac   += Source_Step_Y;	\
    }						\
}



/*
 *                          NORMAL BLITTING
 */

static void BlitBufNormal8(
						  u_int8_t        *pDest,
						  u_int32_t       Dest_Stride,
						  u_int32_t       Dest_Width,
						  u_int32_t       Dest_Height,
						  const u_int8_t  *pSource,
						  u_int32_t       Source_Stride,
						  float           Source_Step_X,
						  float           Source_Step_Y,
						  float           Alpha   )
{
	if ( Alpha >= 1 )
	{
#define PIX_PLOT                \
	{                               \
	    if( source != 0x0000 )      \
		pDest[col] = source;    \
	}
		PIX_LOOP(u_int8_t)
#undef PIX_PLOT
		
	}

	else
	{
#define PIX_PLOT                                                                                        \
	{                                                                                               \
	    if(source != 0x00)                                                                          \
	    {                                                                                           \
		int tr, tg, tb;                                                                         \
		tr = (pDest[col]) & 0xE0;                                                               \
		tg = (pDest[col]) & 0x1C;                                                               \
		tb = (pDest[col]) & 0x03;                                                               \
		pDest[col] =    (( (u_int8_t)((Alpha * ((source & 0xE0) - tr)) + tr) ) & 0xE0) |        \
				(( (u_int8_t)((Alpha * ((source & 0x1C) - tg)) + tg) ) & 0x1C) |        \
				(( (u_int8_t)((Alpha * ((source & 0x03) - tb)) + tb) ) & 0x03) ;        \
	    }                                                                                           \
	}
		PIX_LOOP(u_int8_t)
#undef PIX_PLOT
		
	}
}


static void BlitBufNormal15(
						   u_int16_t       *pDest,
						   u_int32_t       Dest_Stride,
						   u_int32_t       Dest_Width,
						   u_int32_t       Dest_Height,
						   const u_int16_t *pSource,
						   u_int32_t       Source_Stride,
						   float           Source_Step_X,
						   float           Source_Step_Y,
						   float           Alpha   )
{
	if ( Alpha >= 1 )
	{
#define PIX_PLOT                \
	{                               \
	    if( source != 0x0000 )      \
		pDest[col] = source;    \
	}

		PIX_LOOP(u_int16_t)
#undef PIX_PLOT
		
	}

	else
	{
#define PIX_PLOT                                                                                \
	{                                                                                               \
	    if(source != 0x0000)                                                                        \
	    {                                                                                           \
		int tr, tg, tb;                                                                         \
		tr = (pDest[col]) & 0x7C00;                                                             \
		tg = (pDest[col]) & 0x03E0;                                                             \
		tb = (pDest[col]) & 0x001F;                                                             \
		pDest[col] =    (( (u_int16_t)((Alpha * ((source & 0x7C00) - tr)) + tr) ) & 0x7C00) |   \
				(( (u_int16_t)((Alpha * ((source & 0x03E0) - tg)) + tg) ) & 0x03E0) |   \
				(( (u_int16_t)((Alpha * ((source & 0x001F) - tb)) + tb) ) & 0x001F) ;   \
	    }                                                                                           \
	}
		PIX_LOOP(u_int16_t)
#undef PIX_PLOT
		
	}
}


static void BlitBufNormal16(
						   u_int16_t       *pDest,
						   u_int32_t       Dest_Stride,
						   u_int32_t       Dest_Width,
						   u_int32_t       Dest_Height,
						   const u_int16_t *pSource,
						   u_int32_t       Source_Stride,
						   float           Source_Step_X,
						   float           Source_Step_Y,
						   float           Alpha   )
{
	if ( Alpha >= 1 )
	{
#define PIX_PLOT	{		\
	    if(source != 0x0000)	\
		pDest[col] = source;	\
}
		PIX_LOOP(u_int16_t)
#undef PIX_PLOT
		
	}

	else
	{
#define PIX_PLOT			{		\
	    if(source != 0x0000) {			\
		const u_int16_t dv = pDest[col],	\
				tr = dv & 0xF800,	\
				tg = dv & 0x07E0,	\
				tb = dv & 0x001F;	\
		pDest[col] =    (( (u_int16_t)((Alpha * ((source & 0xF800) - tr)) + tr) ) & 0xF800) |   \
				(( (u_int16_t)((Alpha * ((source & 0x07E0) - tg)) + tg) ) & 0x07E0) |   \
				(( (u_int16_t)((Alpha * ((source & 0x001F) - tb)) + tb) ) & 0x001F) ;   \
	    }                                                                                           \
}
		PIX_LOOP(u_int16_t)
#undef PIX_PLOT
		
	}
}

static void BlitBufNormal24(
	unsigned int bpp,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	int src_x, int src_y,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	unsigned int src_copy_width, unsigned int src_copy_height,
	float zoom,
	float vis,
	float magnification
)
{
	int	tar_x_col, tar_y_row,
		src_x_col, src_y_row,
		src_width_check, src_height_check;
	unsigned int x_pixels_inc, y_pixels_inc;	/* For zooming */
	u_int8_t *tar_ptr;
	const u_int8_t *src_ptr;

	if((tar_buf == NULL) || (src_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (src_width == 0) || (src_height == 0) ||
	   (src_copy_width == 0) || (src_copy_height == 0) ||
	   (zoom <= 0.0f) || (vis <= 0.0f)
	)
	    return;

	if(zoom > 1.0f)
	    zoom = 1.0f;

	/* Sanitize magnification */
	if(magnification < BLIT_MAGNIFICATION_MIN)
	    magnification = BLIT_MAGNIFICATION_MIN;
	else if(magnification > BLIT_MAGNIFICATION_MAX)
	    magnification = BLIT_MAGNIFICATION_MAX;

	/* Source totally not visable on target? */
	if(((tar_x + (int)((int)src_copy_width *
		magnification * zoom)) < 0) ||
	   ((tar_y + (int)((int)src_copy_height *
		magnification * zoom)) < 0) ||
	   (tar_x >= (int)tar_width) ||
	   (tar_y >= (int)tar_height) ||
	   ((src_x + (int)src_copy_width) < 0) ||
	   ((src_y + (int)src_copy_height) < 0) ||
	   (src_x >= (int)src_width) ||
	   (src_y >= (int)src_height)
	)
	    return;

	/* Convert source units to 256 (for zooming) */
	src_x <<= 8;
	src_y <<= 8;
	src_copy_width <<= 8;
	src_copy_height <<= 8;
	src_width <<= 8;
	src_height <<= 8;

	/* Precalculate source checks */
	src_width_check = (int)(src_x + (int)src_copy_width);
	src_height_check = (int)(src_y + (int)src_copy_height);

	/* Calculate source increments for zooming affect. */
	x_pixels_inc = (unsigned int)(256 / magnification / zoom);
	y_pixels_inc = (unsigned int)(256 / magnification / zoom);

	/* Sanitize target values. */
	if(tar_x < 0)
	{
	    src_x += (int)(tar_x * -256 / magnification / zoom);
	    tar_x = 0;
	}
	if(tar_y < 0)
	{
	    src_y += (int)(tar_y * -256 / magnification / zoom);
	    tar_y = 0;
	}

	if(src_x < 0)
	    src_x = 0;
	if(src_y < 0)
	    src_y = 0;

	if(src_width_check > (int)src_width)
	    src_width_check = src_width;
	if(src_height_check > (int)src_height)
	    src_height_check = src_height;

	/* Set up starting numbers */
	tar_x_col = tar_x;
	tar_y_row = tar_y;
	src_x_col = src_x;
	src_y_row = src_y;

	if(vis >= 1.0f)
	{
	    while(src_y_row < src_height_check)
	    {
		/* Skip if target is off the screen. */
		if((tar_x_col >= (int)tar_width) ||
		   (tar_y_row >= (int)tar_height)
		)
		{
		    /* Increment the x positions */
		    src_x_col += (int)x_pixels_inc;
		    tar_x_col++;

		    /* Go to the next line? */
		    if(src_x_col >= src_width_check)
		    {
			src_x_col = src_x;
			src_y_row += (int)y_pixels_inc;

			tar_x_col = tar_x;
			tar_y_row++;
		    }

		    continue;
		}


		/* Get the pointer to the current source pixel */
		src_ptr = src_buf +
		    ((src_y_row >> 8) * (int)src_bpl) +
		    ((src_x_col >> 8) * (int)bpp);

		/* Is this source pixel transparent? */
		if((src_ptr[0] == 0x00) &&
		   (src_ptr[1] == 0x00) &&
		   (src_ptr[2] == 0x00)
		)
		{
		    /* Increment the x positions */
		    src_x_col += (int)x_pixels_inc;
		    tar_x_col++;

		    /* Go to the next line. */
		    if(src_x_col >= src_width_check)
		    {
			src_x_col = src_x;
			src_y_row += (int)y_pixels_inc;

			tar_x_col = tar_x;
			tar_y_row++;
		    }

		    continue;
		}

		/* Copy this pixel */
		memcpy(
		    tar_buf +
			(tar_y_row * (int)tar_bpl) +
			(tar_x_col * (int)bpp),
		    src_ptr,
		    bpp
		);

		/* Increment the x positions */
		src_x_col += (int)x_pixels_inc;
		tar_x_col++;

		/* Go to the next line? */
		if(src_x_col >= src_width_check)
		{
		    src_x_col = src_x;
		    src_y_row += (int)y_pixels_inc;

		    tar_x_col = tar_x;   
		    tar_y_row++;
		}
	    }
	}
	else
	{
	    /* Decreased visibility */

	    while(src_y_row < src_height_check)
	    {
		/* Skip if target is off the screen */
		if((tar_x_col >= (int)tar_width) ||
		   (tar_y_row >= (int)tar_height)
		)
		{
		    /* Increment the x positions */
		    src_x_col += x_pixels_inc;
		    tar_x_col++;

		    /* Go to the next line? */
		    if(src_x_col >= src_width_check)
		    {
			src_x_col = src_x;
			src_y_row += y_pixels_inc;

			tar_x_col = tar_x;
			tar_y_row++;
		    }

		    continue;
		}

		/* Get the pointer to the current source pixel */
		src_ptr = src_buf +
		    ((src_y_row >> 8) * (int)src_bpl) +
		    ((src_x_col >> 8) * (int)bpp);

		/* Is this source pixel transparent? */
		if((src_ptr[0] == 0x00) &&
		   (src_ptr[1] == 0x00) &&
		   (src_ptr[2] == 0x00)
		)
		{
		    /* Increment the x positions */
		    src_x_col += x_pixels_inc;
		    tar_x_col++;

		    /* Go to the next line? */
		    if(src_x_col >= src_width_check)
		    {
			src_x_col = src_x;
			src_y_row += y_pixels_inc;

			tar_x_col = tar_x;
			tar_y_row++;
		    }

		    continue;
		}   

		/* Get the pointer to the current target pixel */
		tar_ptr = tar_buf +
		    (tar_y_row * (int)tar_bpl) +
		    (tar_x_col * (int)bpp);

		/* Copy this pixel with the visibility coefficient
		 * applied to it
		 */
		tar_ptr[0] = (u_int8_t)(
		    (vis * (src_ptr[0] - tar_ptr[0])) + tar_ptr[0]
		);
		tar_ptr[1] = (u_int8_t)(
		    (vis * (src_ptr[1] - tar_ptr[1])) + tar_ptr[1]
		);
		tar_ptr[2] = (u_int8_t)(
		    (vis * (src_ptr[2] - tar_ptr[2])) + tar_ptr[2]
		);

		/* Increment the x positions */
		src_x_col += x_pixels_inc;
		tar_x_col++;

		/* Go to the next line? */
		if(src_x_col >= src_width_check)
		{
		    src_x_col = src_x;
		    src_y_row += y_pixels_inc;

		    tar_x_col = tar_x;
		    tar_y_row++;
		}
	    }
	}
}

static void BlitBufNormal32(
	u_int32_t *pDest,
	u_int32_t Dest_Stride,
	u_int32_t Dest_Width, u_int32_t Dest_Height,
	const u_int32_t *pSource,
	u_int32_t Source_Stride,
	float Source_Step_X, float Source_Step_Y,
	float Alpha
)
{
	if(Alpha >= 1.0f)
	{
#define PIX_PLOT	{			\
 if(source & 0x00FFFFFF) /* Ignore alpha */	\
  pDest[col] = source;				\
}
		PIX_LOOP(u_int32_t);
#undef PIX_PLOT
	}
	else
	{
#define PIX_PLOT	{			\
 if(source & 0x00FFFFFF) { /* Ignore alpha */	\
  const u_int8_t *pTar8 = (const u_int8_t *)&(pDest[col]),	\
		 *pSrc8 = (const u_int8_t *)&source;		\
  pDest[col] = (u_int32_t)(source & 0xFF000000) | \
   ((u_int32_t)((Alpha * (pSrc8[2] - pTar8[2])) + pTar8[2]) << 16) | \
   ((u_int32_t)((Alpha * (pSrc8[1] - pTar8[1])) + pTar8[1]) << 8) | \
    (u_int32_t)((Alpha * (pSrc8[0] - pTar8[0])) + pTar8[0]); \
 }						\
}
		PIX_LOOP(u_int32_t);
#undef PIX_PLOT
	}
}


void BlitBufNormal(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	int src_x, int src_y,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	unsigned int src_copy_width, unsigned int src_copy_height,
	float zoom,
	float vis,
	float magnification
)
{
	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;
	if(src_bpl == 0)
	    src_bpl = src_width * bpp;

	BLIT_PREP

	switch(bpp)
	{
	  case 1:
	    {
		u_int8_t *pDest = ((u_int8_t*)tar_buf) + Dest.left + Dest.top*tar_width;
		const u_int8_t *pSource = ((const u_int8_t*)src_buf) + Source.left + Source.top*src_width;
		BlitBufNormal8(
		    pDest,
		    tar_width,
		    RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
		    pSource,
		    src_width,
		    Source_Step_X, Source_Step_Y,
		    vis
		);
	    }
	    break;

	  case 2:
	    switch(depth)
	    {
	      case 15:
		{
		    u_int16_t *pDest = ((u_int16_t *)tar_buf) + Dest.left + Dest.top*tar_width;
		    const u_int16_t *pSource = ((const u_int16_t *)src_buf) + Source.left + Source.top*src_width;
		    BlitBufNormal15(
			pDest,
			tar_width,
			RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
			pSource,
			src_width,
			Source_Step_X, Source_Step_Y,
			vis
		    );
		}
		break;
	      case 16:
		{
		    u_int16_t *pDest = ((u_int16_t*)tar_buf) + Dest.left + Dest.top*tar_width;
		    const u_int16_t   *pSource    = ((const u_int16_t*)src_buf) + Source.left + Source.top*src_width;
		    BlitBufNormal16(
			pDest,
			tar_width,
			RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
			pSource,
			src_width,
			Source_Step_X, Source_Step_Y,
			vis
		    );
		}
		break;
	    }
	    break;

	  case 3:
	    {
		BlitBufNormal24(
		    bpp,
		    tar_buf, src_buf,
		    tar_x, tar_y,
		    tar_width, tar_height, tar_bpl,
		    src_x, src_y,
		    src_width, src_height, src_bpl,
		    src_copy_width, src_copy_height,
		    zoom, vis, magnification
		);
	    }
	    break;

	  case 4:
	    switch(depth)
	    {
	      case 24:
	      case 32:
		{
		    u_int32_t *pDest = ((u_int32_t *)tar_buf) + Dest.left +
			(Dest.top * tar_width);
		    const u_int32_t *pSource = ((const u_int32_t *)src_buf) +
			Source.left + (Source.top * src_width);
		    BlitBufNormal32(
			pDest,
			tar_width,
			RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
			pSource,
			src_width,
			Source_Step_X, Source_Step_Y,
			vis
		    );
		}
		break;
	    }
	    break;
	}
}



/*
 *	Absolute:
 *
 *	Same as normal blitting except that it does not check for
 *	transparency or perform alpha blending.
 */
static void BlitBufAbsolute8(
							u_int8_t        *pDest,
							u_int32_t       Dest_Stride,
							u_int32_t       Dest_Width,
							u_int32_t       Dest_Height,
							const u_int8_t  *pSource,
							u_int32_t       Source_Stride,
							float           Source_Step_X,
							float           Source_Step_Y   )
{
#define PIX_PLOT	{	\
 pDest[col] = source;		\
}
	PIX_LOOP(u_int8_t)
#undef PIX_PLOT
	
}



static void BlitBufAbsolute16(
							 u_int16_t       *pDest,
							 u_int32_t       Dest_Stride,
							 u_int32_t       Dest_Width,
							 u_int32_t       Dest_Height,
							 const u_int16_t *pSource,
							 u_int32_t       Source_Stride,
							 float           Source_Step_X,
							 float           Source_Step_Y   )
{
#define PIX_PLOT	{	\
 pDest[col] = source;		\
}
	PIX_LOOP(u_int16_t)
#undef PIX_PLOT
	
}


static void BlitBufAbsolute24(
	unsigned int bpp,
	u_int8_t *tar_buf,
	const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	int src_x, int src_y,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	unsigned int src_copy_width, unsigned int src_copy_height,
	float zoom,		/* 0.1 to 1.0 */
	float magnification	/* Must be >= 1.0 */
)
{
	int tar_x_col, tar_y_row;
	int src_x_col, src_y_row;

	unsigned int x_pixels_inc, y_pixels_inc;    /* For zooming. */

	int src_width_check;   /* src_x + src_copy_width */
	int src_height_check;  /* src_y + src_copy_height */

	if((tar_buf == NULL) || (src_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (src_width == 0) || (src_height == 0) ||
	   (src_copy_width == 0) || (src_copy_height == 0) ||
	   (zoom <= 0.0f)
	)
	    return;

	/* Sanitize zoom */
	if(zoom > 1.0f)
	    zoom = 1.0f;

	/* Sanitize magnification. */
	if(magnification < BLIT_MAGNIFICATION_MIN)
	    magnification = BLIT_MAGNIFICATION_MIN;
	else if(magnification > BLIT_MAGNIFICATION_MAX)
	    magnification = BLIT_MAGNIFICATION_MAX;

	/* Source totally not visable on target? */
	if(((tar_x + (int)((int)src_copy_width *
		magnification * zoom)) < 0) ||
	   ((tar_y + (int)((int)src_copy_height *
		magnification * zoom)) < 0) ||
	   (tar_x >= (int)tar_width) ||
	   (tar_y >= (int)tar_height) ||
	   ((src_x + (int)src_copy_width) < 0) ||
	   ((src_y + (int)src_copy_height) < 0) ||
	   (src_x >= (int)src_width) ||
	   (src_y >= (int)src_height)
	)
	    return;


	/* Convert source units to 256 (for zooming) */
	src_x <<= 8;
	src_y <<= 8;
	src_copy_width <<= 8;
	src_copy_height <<= 8;
	src_width <<= 8;
	src_height <<= 8;

	/* Precalculate source checks */
	src_width_check = (int)(src_x + (int)src_copy_width);
	src_height_check = (int)(src_y + (int)src_copy_height);
	
	/* Calculate source increments for zooming affect. */
	x_pixels_inc = (unsigned int)(256 / magnification / zoom);
	y_pixels_inc = (unsigned int)(256 / magnification / zoom);

	/* Sanitize target values. */
	if(tar_x < 0)
	{
	    src_x += (int)(tar_x * -256 / magnification / zoom);
	    tar_x = 0;
	}
	if(tar_y < 0)
	{
	    src_y += (int)(tar_y * -256 / magnification / zoom);
	    tar_y = 0;
	}

	if(src_x < 0)
	    src_x = 0;
	if(src_y < 0)
	    src_y = 0;

	if(src_width_check > (int)src_width)
	    src_width_check = src_width;
	if(src_height_check > (int)src_height)
	    src_height_check = src_height;
	
	/* Set up starting numbers */
	tar_x_col = tar_x;
	tar_y_row = tar_y;
	src_x_col = src_x;
	src_y_row = src_y;
	
 
	/* Begin blitting src_buf to tar_buf */
	while(src_y_row < src_height_check)
	{
	    /* Skip if off the screen. */
	    if((tar_x_col >= (int)tar_width) ||
	       (tar_y_row >= (int)tar_height)
	    )
	    {
		/* Increment the x positions */
		src_x_col += (int)x_pixels_inc;
		tar_x_col++;

		/* Go to the next line? */
		if(src_x_col >= src_width_check)
		{
		    src_x_col = src_x;
		    src_y_row += (int)y_pixels_inc;
	
		    tar_x_col = tar_x;  
		    tar_y_row++;
		}

		continue;
	    }

	    memcpy(
		tar_buf +
		    (tar_y_row * (int)tar_bpl) +
		    (tar_x_col * (int)bpp),
		src_buf +
		    ((src_y_row >> 8) * (int)src_bpl) +
		    ((src_x_col >> 8) * (int)bpp),
		bpp
	    );

	    /* Increment the x positions */
	    src_x_col += (int)x_pixels_inc;
	    tar_x_col++;

	    /* Go to the next line? */
	    if(src_x_col >= src_width_check)
	    {
		src_x_col = src_x;
		src_y_row += (int)y_pixels_inc;
	     
		tar_x_col = tar_x;   
		tar_y_row++;
	    }
	}
}

static void BlitBufAbsolute32(
							 u_int32_t       *pDest,
							 u_int32_t       Dest_Stride,
							 u_int32_t       Dest_Width,
							 u_int32_t       Dest_Height,
							 const u_int32_t *pSource,
							 u_int32_t       Source_Stride,
							 float           Source_Step_X,
							 float           Source_Step_Y   )
{
#define PIX_PLOT	{	\
 pDest[col] = source;		\
}
	PIX_LOOP(u_int32_t)
#undef PIX_PLOT
	
}


void BlitBufAbsolute(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,                      
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	int src_x, int src_y,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	unsigned int src_copy_width, unsigned int src_copy_height,
	float zoom,		/* 0.1 to 1.0 supported */
	float magnification
)
{
	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;
	if(src_bpl == 0)
	    src_bpl = src_width * bpp;

	BLIT_PREP

	switch(bpp)
	{
	  case 1:
	    {
		u_int8_t    *pDest      = ((u_int8_t*)tar_buf) + Dest.left + Dest.top*tar_width;
		const u_int8_t *pSource = ((u_int8_t*)src_buf) + Source.left + Source.top*src_width;
		BlitBufAbsolute8(
		    pDest,
		    tar_width,
		    RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
		    pSource,
		    src_width,
		    Source_Step_X, Source_Step_Y
		);
	    }
	    break;
	  case 2:
	    /* 15 and 16 bit depth would be the same here */
	    {
		u_int16_t   *pDest      = ((u_int16_t*)tar_buf) + Dest.left + Dest.top*tar_width;
		const u_int16_t *pSource = ((u_int16_t*)src_buf) + Source.left + Source.top*src_width;
		BlitBufAbsolute16(
		    pDest,
		    tar_width,
		    RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
		    pSource,
		    src_width,
		    Source_Step_X, Source_Step_Y
		);
	    }
	    break;
	  case 3:
	    {
		BlitBufAbsolute24(
		    bpp,
		    tar_buf, src_buf,
		    tar_x, tar_y,
		    tar_width, tar_height, tar_bpl,
		    src_x, src_y,
		    src_width, src_height, src_bpl, 
		    src_copy_width, src_copy_height,
		    zoom, magnification
		);
	    }
	    break;
	  case 4:
	    switch(depth)
	    {
	      case 24:
	      case 32:
		{
		    u_int32_t *pDest = ((u_int32_t *)tar_buf) + Dest.left +
			(Dest.top * tar_width);
		    const u_int32_t *pSource = ((u_int32_t *)src_buf) +
			Source.left + (Source.top * src_width);
		    BlitBufAbsolute32(
			pDest,
			tar_width,
			RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
			pSource,
			src_width,
			Source_Step_X, Source_Step_Y
		    );
		}
		break;
	    }
	    break;
	}
}


/*
 *	Additive
 */

void BlitBufAdditive8(
					 u_int8_t        *pDest,
					 u_int32_t       Dest_Stride,
					 u_int32_t       Dest_Width,
					 u_int32_t       Dest_Height,
					 const u_int8_t  *pSource,
					 u_int32_t       Source_Stride,
					 float           Source_Step_X,
					 float           Source_Step_Y   )
{
#define PIX_PLOT                                \
    {                                               \
	int r, g, b;                                \
	int dest = pDest[col];                      \
	r = MIN((source&0xE0) + (dest&0xE0), 0xE0); \
	g = MIN((source&0x1C) + (dest&0x1C), 0x1C); \
	b = MIN((source&0x03) + (dest&0x03), 0x03); \
	pDest[col] = r | g | b;                     \
    }
	PIX_LOOP(u_int8_t)
#undef PIX_PLOT
	
}




void BlitBufAdditive15(
					  u_int16_t       *pDest,
					  u_int32_t       Dest_Stride,
					  u_int32_t       Dest_Width,
					  u_int32_t       Dest_Height,
					  const u_int16_t *pSource,
					  u_int32_t       Source_Stride,
					  float           Source_Step_X,
					  float           Source_Step_Y   )
{
#define PIX_PLOT                                        \
    {                                                       \
	int r, g, b;                                        \
	u_int32_t dest  = (u_int32_t)pDest[col];            \
	r = (source & 0x00007C00) + (dest & 0x00007C00);    \
	g = (source & 0x000003E0) + (dest & 0x000003E0);    \
	b = (source & 0x0000001F) + (dest & 0x0000001F);    \
	pDest[col] =    (u_int16_t)(MIN(r, 0x00007C00) |    \
				    MIN(g, 0x000003E0) |    \
				    MIN(b, 0x0000001F) );   \
    }
	PIX_LOOP(u_int16_t)
#undef PIX_PLOT
	
}



void BlitBufAdditive16(
					  u_int16_t       *pDest,
					  u_int32_t       Dest_Stride,
					  u_int32_t       Dest_Width,
					  u_int32_t       Dest_Height,
					  const u_int16_t *pSource,
					  u_int32_t       Source_Stride,
					  float           Source_Step_X,
					  float           Source_Step_Y   )
{
#define PIX_PLOT                                        \
    {                                                       \
	int r, g, b;                                        \
	u_int32_t dest  = (u_int32_t)pDest[col];            \
	r = (source & 0x0000F800) + (dest & 0x0000F800);    \
	g = (source & 0x000007E0) + (dest & 0x000007E0);    \
	b = (source & 0x0000001F) + (dest & 0x0000001F);    \
	pDest[col] =    (u_int16_t)(MIN(r, 0x0000F800) |    \
				    MIN(g, 0x000007E0) |    \
				    MIN(b, 0x0000001F) );   \
    }
	PIX_LOOP(u_int16_t)
#undef PIX_PLOT
	
}

static void BlitBufAdditive24(
	unsigned int bpp,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	int src_x, int src_y,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	unsigned int src_copy_width, unsigned int src_copy_height,
	float zoom,
	float magnification
)
{
	int	tar_x_col, tar_y_row,
		src_x_col, src_y_row,
		src_width_check, src_height_check;
	unsigned int x_pixels_inc, y_pixels_inc;	/* For zooming */
	u_int8_t *tar_ptr;
	const u_int8_t *src_ptr;

	if((tar_buf == NULL) || (src_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (src_width == 0) || (src_height == 0) ||
	   (src_copy_width == 0) || (src_copy_height == 0) ||
	   (zoom <= 0.0f)
	)
	    return;

	if(zoom > 1.0f)
	    zoom = 1.0f;

	/* Sanitize magnification */
	if(magnification < BLIT_MAGNIFICATION_MIN)
	    magnification = BLIT_MAGNIFICATION_MIN;
	else if(magnification > BLIT_MAGNIFICATION_MAX)
	    magnification = BLIT_MAGNIFICATION_MAX;

	/* Source totally not visable on target? */
	if(((tar_x + (int)((int)src_copy_width *
		magnification * zoom)) < 0) ||
	   ((tar_y + (int)((int)src_copy_height *
		magnification * zoom)) < 0) ||
	   (tar_x >= (int)tar_width) ||
	   (tar_y >= (int)tar_height) ||
	   ((src_x + (int)src_copy_width) < 0) ||
	   ((src_y + (int)src_copy_height) < 0) ||
	   (src_x >= (int)src_width) ||
	   (src_y >= (int)src_height)
	)
	    return;

	/* Convert source units to 256 (for zooming) */
	src_x <<= 8;
	src_y <<= 8;
	src_copy_width <<= 8;
	src_copy_height <<= 8;
	src_width <<= 8;
	src_height <<= 8;

	/* Precalculate source checks */
	src_width_check = (int)(src_x + (int)src_copy_width);
	src_height_check = (int)(src_y + (int)src_copy_height);

	/* Calculate source increments for zooming affect. */
	x_pixels_inc = (unsigned int)(256 / magnification / zoom);
	y_pixels_inc = (unsigned int)(256 / magnification / zoom);

	/* Sanitize target values. */
	if(tar_x < 0)
	{
	    src_x += (int)(tar_x * -256 / magnification / zoom);
	    tar_x = 0;
	}
	if(tar_y < 0)
	{
	    src_y += (int)(tar_y * -256 / magnification / zoom);
	    tar_y = 0;
	}

	if(src_x < 0)
	    src_x = 0;
	if(src_y < 0)
	    src_y = 0;

	if(src_width_check > (int)src_width)
	    src_width_check = src_width;
	if(src_height_check > (int)src_height)
	    src_height_check = src_height;

	/* Set up starting numbers */
	tar_x_col = tar_x;
	tar_y_row = tar_y;
	src_x_col = src_x;
	src_y_row = src_y;

	if(1)
	{
	    while(src_y_row < src_height_check)
	    {
		/* Skip if target is off the screen. */
		if((tar_x_col >= (int)tar_width) ||
		   (tar_y_row >= (int)tar_height)
		)
		{
		    /* Increment the x positions */
		    src_x_col += (int)x_pixels_inc;
		    tar_x_col++;

		    /* Go to the next line? */
		    if(src_x_col >= src_width_check)
		    {
			src_x_col = src_x;
			src_y_row += (int)y_pixels_inc;

			tar_x_col = tar_x;
			tar_y_row++;
		    }

		    continue;
		}


		/* Get the pointer to the current source pixel */
		src_ptr = src_buf +
		    ((src_y_row >> 8) * (int)src_bpl) +
		    ((src_x_col >> 8) * (int)bpp);

		/* Get the pointer to the current target pixel */
		tar_ptr = tar_buf +
		    (tar_y_row * (int)tar_bpl) +
		    (tar_x_col * (int)bpp);

		/* Copy this pixel additively */
		tar_ptr[0] = (u_int8_t)MIN(
		    (u_int16_t)tar_ptr[0] + (u_int16_t)src_ptr[0],
		    0x00FF
		);
		tar_ptr[1] = (u_int8_t)MIN(
		    (u_int16_t)tar_ptr[1] + (u_int16_t)src_ptr[1],
		    0x00FF
		);
		tar_ptr[2] = (u_int8_t)MIN(
		    (u_int16_t)tar_ptr[2] + (u_int16_t)src_ptr[2],
		    0x00FF
		);

		/* Increment the x positions */
		src_x_col += (int)x_pixels_inc;
		tar_x_col++;

		/* Go to the next line? */
		if(src_x_col >= src_width_check)
		{
		    src_x_col = src_x;
		    src_y_row += (int)y_pixels_inc;

		    tar_x_col = tar_x;   
		    tar_y_row++;
		}
	    }
	}
}

void BlitBufAdditive32(
					  u_int32_t       *pDest,
					  u_int32_t       Dest_Stride,
					  u_int32_t       Dest_Width,
					  u_int32_t       Dest_Height,
					  const u_int32_t *pSource,
					  u_int32_t       Source_Stride,
					  float           Source_Step_X,
					  float           Source_Step_Y   )
{
#define PIX_PLOT	{				\
 const u_int32_t cDest = pDest[col];			\
 const u_int32_t r = (source & 0x00FF0000) + (cDest & 0x00FF0000), \
		 g = (source & 0x0000FF00) + (cDest & 0x0000FF00), \
		 b = (source & 0x000000FF) + (cDest & 0x000000FF); \
 pDest[col] = (u_int32_t)(MIN(r, 0x00FF0000) |		\
			  MIN(g, 0x0000FF00) |		\
			  MIN(b, 0x000000FF)		\
 );							\
}
	PIX_LOOP(u_int32_t);
#undef PIX_PLOT
}


void BlitBufAdditive(
	unsigned int bpp, unsigned int depth,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	int tar_x, int tar_y,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tar_bpl,
	int src_x, int src_y,
	unsigned int src_width, unsigned int src_height,
	unsigned int src_bpl,
	unsigned int src_copy_width, unsigned int src_copy_height,
	float zoom,			/* 0.1 to 1.0 supported */
	float magnification		/* 1.0 to 255.0 supported */
)
{
	if(tar_bpl == 0)
	    tar_bpl = tar_width * bpp;
	if(src_bpl == 0)
	    src_bpl = src_width * bpp;

	BLIT_PREP

	switch(bpp)
	{
	  case 1:
	    {
		u_int8_t *pDest = ((u_int8_t*)tar_buf) + Dest.left + Dest.top*tar_width;
		const u_int8_t *pSource = ((const u_int8_t*)src_buf) + Source.left + Source.top*src_width;
		BlitBufAdditive8(
		    pDest,
		    tar_width,
		    RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
		    pSource,
		    src_width,
		    Source_Step_X, Source_Step_Y
		);
	    }
	    break;
	  case 2:
	    switch(depth)
	    {
	      case 15:
		{
		    u_int16_t *pDest = ((u_int16_t*)tar_buf) + Dest.left + Dest.top*tar_width;
		    const u_int16_t *pSource = ((const u_int16_t*)src_buf) + Source.left + Source.top*src_width;
		    BlitBufAdditive15(
			pDest,
			tar_width,
			RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
			pSource,
			src_width,
			Source_Step_X, Source_Step_Y
		    );
		}
		break;
	      case 16:
		{
		    u_int16_t *pDest = ((u_int16_t*)tar_buf) + Dest.left + Dest.top*tar_width;
		    const u_int16_t *pSource = ((const u_int16_t*)src_buf) + Source.left + Source.top*src_width;
		    BlitBufAdditive16(
			pDest,
			tar_width,
			RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
			pSource,
			src_width,
			Source_Step_X, Source_Step_Y
		    );
		}
		break;
	    }
	    break;
	  case 3:
	    {
		BlitBufAdditive24(
		    bpp,
		    tar_buf, src_buf,
		    tar_x, tar_y, tar_width, tar_height, tar_bpl,
		    src_x, src_y, src_width, src_height, src_bpl,
		    src_copy_width, src_copy_height,
		    zoom, magnification
		);
	    }
	    break;
	  case 4:
	    switch(depth)
	    {
#if 0
	      case 24:
		{
		    u_int8_t *pDest = (u_int8_t *)tar_buf +
			((Dest.left + (Dest.top * tar_width)) * bpp);
		    const u_int8_t *pSource = (const u_int8_t *)src_buf +
			((Source.left + (Source.top * src_width)) * bpp);
		    BlitBufAdditive24(                                     
			bpp,
			pDest,
			tar_bpl,
			RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
			pSource,
			src_bpl,
			Source_Step_X, Source_Step_Y
		    );
		}
		break;
#endif
	      case 24:
	      case 32:
		{
		    u_int32_t *pDest = ((u_int32_t*)tar_buf) + Dest.left + Dest.top*tar_width;
		    const u_int32_t *pSource = ((const u_int32_t*)src_buf) + Source.left + Source.top*src_width;
		    BlitBufAdditive32(
			pDest,
			tar_width,
			RECT_WIDTH(Dest), RECT_HEIGHT(Dest),
			pSource,
			src_width,
			Source_Step_X, Source_Step_Y
		    );
		}
		break;
	    }
	    break;
	}
}


#if 0
/*
	Original blitting preparation code prior to macroization.
	!DO NOT DELETE AS THE COMPRESSED MACRO IS NOT EASILLY EDITED!
 */
u_int32_t       Dest_Width;
u_int32_t       Dest_Height;
float           Source_Step_X;
float           Source_Step_Y;

XSWRect         Source;
XSWRect         Dest;

float       Scale = (float)zoom * (float)magnification;

/* Prophylactics. */
assert( tar_buf && src_buf );										 // Bad buffer involved.
assert( tar_width != 0 && src_width != 0 );							 // Bad width involved.
assert( tar_height != 0 && src_height != 0 );						 // Bad height involved.
assert( src_copy_width != 0 && src_copy_height != 0 );				 // Bad copy parameters involved.
assert( zoom >= 0.0 && magnification >= 0.0 );						 // Bad scaling information.

if ( tar_x >= tar_width || tar_y >= tar_height )
	return;
if ( src_x >= src_width || src_y >= src_height )
	return;

/* Sanitize Scale. */
if ( Scale > BLIT_MAGNIFICATION_MAX )	 Scale = BLIT_MAGNIFICATION_MAX;
/* if( Scale < BLIT_MAGNIFICATION_MIN )    Scale = BLIT_MAGNIFICATION_MIN; */

assert( Scale==1.0f );

/* Precalculations. */

/* Compute rectangles. */
Source.left     =   src_x;
Source.top      =   src_y;
Source.right    =   src_x + src_copy_width;
Source.bottom   =   src_y + src_copy_height;

Dest.left       =   tar_x;
Dest.top        =   tar_y;
Dest.right      =   tar_x + Scale*src_copy_width;
Dest.bottom     =   tar_y + Scale*src_copy_height;

/*  Source and Dest in standard StretchBlit format.
	This is where we would insert an OpenGL call to perform the blit. */

/* Compute stepping information. */
Source_Step_X   = (float)RECT_WIDTH(Source) / (float)RECT_WIDTH(Dest);
Source_Step_Y   = (float)RECT_HEIGHT(Source) / (float)RECT_HEIGHT(Dest);

/* Clip destination rectangle against buffer size. */
if ( Dest.left < 0 )
{
	Source.left     -=  Dest.left * Source_Step_X;
	Dest.left       =   0;
}

if ( Dest.top < 0 )
{
	Source.top      -=  Dest.top * Source_Step_Y;
	Dest.top        =   0;
}

if ( Dest.right > tar_width )
{
	Source.right    +=  (Dest.right-tar_width) * Source_Step_X;
	Dest.right      =   tar_width;
}

if ( Dest.bottom > tar_height )
{
	Source.bottom   +=  (Dest.bottom-tar_height) * Source_Step_Y;
	Dest.bottom     =   tar_height;
}

/* Revalidate rects. */
if ( !RECT_VALID(Source) || !RECT_VALID(Dest) )
	return;

#endif
