#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/stat.h>

#include "../include/cfgfmt.h"
#include "../include/string.h"
#include "../include/fio.h"
#include "../include/disk.h"

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "keymap.h"
#include "ue.h"
#include "cfgfile.h"
#include "config.h"


int UECfgFileOpen(const char *path);
int UECfgFileSave(const char *path);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Opens the configuration from the specified file.
 */
int UECfgFileOpen(const char *path)
{
	int i, lines_read = 0;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	unsigned long file_size = 0l;
	FILE *fp;
	struct stat stat_buf;
	ue_option_struct *opt = &option;

	if(STRISEMPTY(path))
	    return(-2);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	    return(-1);

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;
	}

	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';


	    if(!strcasecmp(parm, "VersionMajor"))
	    {
		opt->rc_version_major = atoi(val);
	    }
	    else if(!strcasecmp(parm, "VersionMinor"))
	    {
		opt->rc_version_minor = atoi(val);
	    }
	    else if(!strcasecmp(parm, "VersionRelease"))
	    {
		opt->rc_version_release = atoi(val);
	    }

	    else if(!strcasecmp(parm, "RecordWindowGeometry"))
	    {
		opt->record_window_geometry = StringIsYes(val);
	    }
	    else if(!strcasecmp(parm, "ShowObjectLabels"))
	    {
		opt->show_object_labels = StringIsYes(val);
	    }
	    else if(!strcasecmp(parm, "ShowGeometryLabels"))
	    {
		opt->show_geometry_labels = StringIsYes(val);
	    }
	    else if(!strcasecmp(parm, "ShowGrid"))
	    {
		opt->show_grid = StringIsYes(val);
	    }
	    else if(!strcasecmp(parm, "ShowContactBounds"))
	    {
		opt->show_contact_bounds = StringIsYes(val);
	    }
	    else if(!strcasecmp(parm, "AnimateObjectDisplay"))
	    {
		opt->animate_object_display = StringIsYes(val);
	    }
	    else if(!strcasecmp(parm, "GridSpacing"))
	    {
		/* In real units */
		opt->grid_spacing = MAX(atof(val), 0.0f);
	    }
	    else if(!strcasecmp(parm, "ObjectMarkerRadius"))
	    {
		/* In pixels */
		opt->object_marker_radius = MAX(atoi(val), 0);
	    }
	    else if(!strcasecmp(parm, "UEWGeometry"))
	    {
		int vi[4];
		rectangle_t *rect = &opt->uew_geometry;
		sscanf(
		    val,
		    "%i %i %i %i",
		    &vi[0], &vi[1], &vi[2], &vi[3]
		);
		rect->x = vi[0];
		rect->y = vi[1];
		rect->width = MAX(vi[2], 0);
		rect->height = MAX(vi[3], 0);
	    }
	    else if(!strcasecmp(parm, "ObjectLabelFontName"))
	    {
		free(opt->object_label_font_name);
		opt->object_label_font_name = STRDUP(val);
	    }
	    else if(!strcasecmp(parm, "GeometryLabelFontName"))
	    {
		free(opt->geometry_label_font_name);
		opt->geometry_label_font_name = STRDUP(val);
	    }

	    else if(!strcasecmp(parm, "ToplevelDir"))
	    {
		if(ISPATHABSOLUTE(val))
		{
		    strncpy(
			dname.toplevel, val,
			sizeof(dname.toplevel)
		    );
		    dname.toplevel[sizeof(dname.toplevel) - 1] = '\0';
		}
		else
		{
		    fprintf(
			stderr,
"%s: Line %i: %s: Warning: `%s' is not an absolute path.\n",
			path, lines_read, parm, val
		    );
		}
	    }
	    else if(!strcasecmp(parm, "EtcDir"))
	    {
		char *full_path = ISPATHABSOLUTE(val) ?
		    STRDUP(val) :
		    STRDUP(PrefixPaths(dname.toplevel, val));
		if(full_path != NULL)
		{
		    strncpy(
			dname.etc, full_path,
			sizeof(dname.etc)
		    );
		    dname.etc[sizeof(dname.etc) - 1] = '\0';
		}
		free(full_path);
	    }
	    else if(!strcasecmp(parm, "ImagesDir"))
	    {
		char *full_path = ISPATHABSOLUTE(val) ?
		    STRDUP(val) :
		    STRDUP(PrefixPaths(dname.toplevel, val));
		if(full_path != NULL)
		{
		    strncpy(
			dname.images, full_path,
			sizeof(dname.images)
		    );
		    dname.images[sizeof(dname.images) - 1] = '\0';
		}
		free(full_path);
	    }
	    else if(!strcasecmp(parm, "SoundsDir"))
	    {
		char *full_path = ISPATHABSOLUTE(val) ?
		    STRDUP(val) :
		    STRDUP(PrefixPaths(dname.toplevel, val));
		if(full_path != NULL)
		{
		    strncpy(
			dname.sounds, full_path,
			sizeof(dname.sounds)
		    );
		    dname.sounds[sizeof(dname.sounds) - 1] = '\0';
		}
		free(full_path);
	    }
	    else if(!strcasecmp(parm, "ServerDir"))
	    {
		char *full_path = ISPATHABSOLUTE(val) ?
		    STRDUP(val) :
		    STRDUP(PrefixPaths("/home", val));
		if(full_path != NULL)
		{
		    strncpy(
			dname.server, full_path,
			sizeof(dname.server)
		    );
		    dname.server[sizeof(dname.server) - 1] = '\0';
		}
		free(full_path);
	    }
	    else if(!strcasecmp(parm, "OPMFile"))
	    {
		strncpy(fname.opm, val, sizeof(fname.opm));
		fname.opm[sizeof(fname.opm) - 1] = '\0';
	    }

	    else if(!strcasecmp(parm, "BeginKeyMap"))
	    {
		/* Begin reading this block */
		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL) continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* Get all keymaps */
		    for(i = 0; i < UE_TOTAL_KEYMAPS; i++)
		    {
			if(!strcasecmp(keymap_name[i], parm))
			    keymap[i].keycode = atoi(val);
		    }

		    if(!strcasecmp(parm, "EndKeyMap"))
			break;
		}
	    }
	    else
	    {
		fprintf(
		    stderr,
"%s: Line %i: Unknown parameter: %s\n",
		    path, lines_read, parm
		);
	    }
	}

	/* Close the file */
	fclose(fp);

	return(0);
}


int UECfgFileSave(const char *path)
{
	int i;
	FILE *fp;
	struct stat stat_buf;
	const rectangle_t *rect;
	ue_option_struct *opt = &option;

	/* Check if filename does not already exist, if it does
	 * not then attempt to create parent directories.
	 */
	if(stat(path, &stat_buf))
	{
	    char *parent = GetParentDir(path);

	    if(parent != NULL)
	    {
		if(rmkdir(
		    parent,
		    S_IRUSR | S_IWUSR | S_IXUSR |
		    S_IRGRP |           S_IXGRP |
		    S_IROTH |           S_IXOTH
		))
		{
		    fprintf(
			stderr,
"%s: Unable to create directory.\n",
			parent
		    );
		    return(-1);
		}
	    }
	}

	/* Open the file for writing */
	fp = fopen(path, "wb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *error_msg = STRDUP(strerror(error_code));
	    if(error_msg != NULL)
	    {
		*error_msg = toupper(*error_msg);
		fprintf(
		    stderr,
"%s: %s.\n",
		    path, error_msg
		);
	    }
	    free(error_msg);
	    return(-1);
	}


	/* Heading comment */
	fprintf(
	    fp,
"# %s Version %s Configuration File\n\
#\n",
	    PROG_NAME_FULL,
	    PROG_VERSION
	);

	/* Version */
	fprintf(
	    fp,
"# This configuration file version:\n\
#\n"
	);
	fprintf(
	    fp, 
	    "VersionMajor = %i\n",
	    PROG_VERSION_MAJOR
	);
	fprintf(
	    fp, 
	    "VersionMinor = %i\n",
	    PROG_VERSION_MINOR
	);
	fprintf(
	    fp, 
	    "VersionRelease = %i\n",
	    PROG_VERSION_RELEASE
	);
	fprintf(
	    fp, 
	    "\n\n"
	);

	/* Directories & Files */
	fprintf(
	    fp,
"# Directories and files:\n\
#\n"
	);
	if(!STRISEMPTY(dname.toplevel))
	    fprintf(
		fp,
		"ToplevelDir = %s\n",
		dname.toplevel
	    );
	if(!STRISEMPTY(dname.etc))
	    fprintf(
		fp,
		"EtcDir = %s\n",
		dname.etc
	    );
	if(!STRISEMPTY(dname.images))
	    fprintf(
		fp,
		"ImagesDir = %s\n",
		dname.images
	    );
	if(!STRISEMPTY(dname.sounds))
	    fprintf(
		fp,
		"SoundsDir = %s\n",
		dname.sounds
	    );
	if(!STRISEMPTY(dname.server))
	    fprintf(
		fp,
		"ServerDir = %s\n",
		dname.server
	    );
	if(!STRISEMPTY(fname.opm))
	    fprintf(
		fp,
		"OPMFile = %s\n",
		fname.opm
	    );
	fprintf(
	    fp, 
	    "\n\n"
	);

	/* Appearance */
	fprintf(
	    fp,
"# Appearance:\n\
#\n"
	);
	fprintf(
	    fp,
	    "RecordWindowGeometry = %s\n",
	    opt->record_window_geometry ? "yes" : "no"
	);
	rect = &opt->uew_geometry;
	fprintf(
	    fp, 
	    "UEWGeometry = %i %i %i %i\n",
	    (int)rect->x, (int)rect->y,
	    (int)rect->width, (int)rect->height
	);
	fprintf(
	    fp,
	    "\n\n"  
	);


	/* Fonts */
	fprintf(
	    fp, 
"# Fonts:\n\
#\n"
	);
	if(!STRISEMPTY(opt->object_label_font_name))
	    fprintf(
		fp,
		"ObjectLabelFontName = %s\n",
		opt->object_label_font_name
	    );
	if(!STRISEMPTY(opt->geometry_label_font_name))
	    fprintf(
		fp, 
		"GeometryLabelFontName = %s\n",
		opt->geometry_label_font_name
	    );
	fprintf(
	    fp, 
	    "\n\n"
	);


	/* Graphics */
	fprintf(
	    fp,
"# Graphics:\n\
#\n"
	);
	fprintf(
	    fp,
	    "ShowObjectLabels = %s\n",
	    opt->show_object_labels ? "yes" : "no"
	);
	fprintf(
	    fp,
	    "ShowGeometryLabels = %s\n",
	    opt->show_geometry_labels ? "yes" : "no"
	);
	fprintf(
	    fp,
	    "ShowGrid = %s\n",
	    opt->show_grid ? "yes" : "no"
	);
	fprintf(
	    fp,
	    "ShowContactBounds = %s\n",
	    opt->show_contact_bounds ? "yes" : "no"
	);
	fprintf(
	    fp,
	    "AnimateObjectDisplay = %s\n",
	    opt->animate_object_display ? "yes" : "no"
	);
	fprintf(
	    fp,
	    "GridSpacing = %.6f\n",
	    opt->grid_spacing
	);
	fprintf(
	    fp,
	    "ObjectMarkerRadius = %i\n",
	    opt->object_marker_radius
	);
	fprintf(
	    fp,
	    "\n\n"
	);


	/* Key Mappings */
	fprintf(
	    fp, 
"# Key Mappings\n\
#\n"
	);
	fprintf(
	    fp, 
	    "BeginKeyMap\n"
	);
	for(i = 0; i < UE_TOTAL_KEYMAPS; i++)
	{
	    fprintf(
		fp,
		"\t%s = %i\n",
		keymap_name[i],
		keymap[i].keycode
	    );
	}
	fprintf(
	    fp, 
	    "EndKeyMap\n"
	);
	fprintf(
	    fp, 
	    "\n\n"
	);


	/* Close the file */
	fclose(fp);

	return(0);
}



