#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>

#include "../include/tga.h"
#include "../include/osw-x.h"
#include "../include/imgrefs.h"

#include "imgrefsfio.h"

int ImgRefIsLoaded(imgref_struct *imgref);
int ImgRefLoad(imgref_struct *imgref);
void ImgRefUnload(imgref_struct *imgref);

void ImgRefClear(imgref_struct **imgref, int total);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Checks if the ImgRef's image is loaded.
 */
int ImgRefIsLoaded(imgref_struct *imgref)
{
	if(imgref == NULL)
	    return(0);
	else if(imgref->image_data == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Loads the ImgRef's image.
 */
int ImgRefLoad(imgref_struct *imgref)
{
	int status;
	struct stat stat_buf;
	tga_data_struct *td;
	osw_gui_struct *gui = &osw_gui[0];

	if(imgref == NULL)
	    return(-1);

	/* This imgref not to have image? */
	if(imgref->options & IMGREF_NO_IMAGE)
	    return(-2);

	/* Already loaded? */
	if(ImgRefIsLoaded(imgref))
	    return(0);

	/* Missing filename? */
	if(STRISEMPTY(imgref->path))
	    return(-2);

	/* Is that filename pointing to an existing file? */
	if(stat(imgref->path, &stat_buf))
	    return(-2);

	/* Set up for passive async loading */
	if(1)
	{
/* Explicitly set file format for now */
imgref->img_fmt = IMGREF_IMGFMT_TGA;

	    switch(imgref->img_fmt)
	    {
	      /* TGA format */
	      case IMGREF_IMGFMT_TGA:
		imgref->lib_data = (void *)calloc(
		    1, sizeof(tga_data_struct)
		);
		if(imgref->lib_data == NULL)
		{
		    imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;
		    return(-1);
		}
		td = (tga_data_struct *)imgref->lib_data;

		/* Set up for partial tga read */
		status = TgaStartReadPartialFromFile(
		    imgref->path,
		    td,
                    TgaGetDepthForX(
                        gui->bits_per_pixel,
                        gui->depth
                    )
		);
		if(status != TgaSuccess)
		{
		    fprintf(
			stderr,
"ImgRefLoad(): %s: TgaStartReadPartialFromFile() error code %i.\n",
			imgref->path, status
		    );

		    /* Destroy the tga_data */
		    TgaDestroyData(td);

		    imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;

		    free(imgref->lib_data);
		    imgref->lib_data = NULL;

		    return(-1);
		}

		/* Link image_data pointer */
		imgref->image_data = (u_int8_t *)td->data;

		/* Set width and height from tga_data */
		imgref->width = td->width;
		imgref->height = td->height;

		/* Mark load process as currently loading */
		imgref->load_progress = IMGREF_LOAD_PROGRESS_LOADING;
		break;

	      /* Unknown format */
	      default:
		break;
	    }

	    /* Sanitize number of frames specified on imgref entry */
	    if(imgref->total_frames > IMGREF_MAX_FRAMES)
		imgref->total_frames = IMGREF_MAX_FRAMES;
	    if(imgref->total_frames < 1)
		imgref->total_frames = 1;

	    /* Calculate frame width and height */
	    imgref->fwidth = imgref->width;
	    imgref->fheight = (int)imgref->height /
		(int)imgref->total_frames;
	}

	return(0);
}

/*
 *	Unloads the ImgRef's image.
 */
void ImgRefUnload(imgref_struct *imgref)
{
	if(imgref == NULL)
	    return;

	if(!ImgRefIsLoaded(imgref))
	    return;

	/* Unload by format */
	switch(imgref->img_fmt)
	{
	  case IMGREF_IMGFMT_TGA:
	    if(imgref->lib_data != NULL)
	    {
		/* Delete the image and library data */
		TgaDestroyData((tga_data_struct *)imgref->lib_data);
		free(imgref->lib_data);
		imgref->lib_data = NULL;
		imgref->image_data = NULL;
	    }
	    break;

	  default:
	    if(imgref->lib_data != NULL)
	    {
		free(imgref->lib_data);
		imgref->lib_data = NULL;
		free(imgref->image_data);
		imgref->image_data = NULL;
	    }
	    break; 
	}

	imgref->img_fmt = IMGREF_IMGFMT_UNKNOWN;
	imgref->load_progress = IMGREF_LOAD_PROGRESS_DONE;
}


/*
 *	Deletes all the imgrefs in the list.
 */
void ImgRefClear(imgref_struct ***list, int *total)
{
	int i;
	imgref_struct *imgref;

	for(i = 0; i < *total; i++)
	{
	    imgref = (*list)[i];
	    ImgRefUnload(imgref);
	    ImgRefDelete(imgref);
	}
	free(*list);
	*list = NULL;
	*total = 0;
}
