#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/fio.h"
#include "../include/disk.h"
#include "../include/cfgfmt.h"

#include "ocsinfofio.h"


ocs_info_struct *OCSInfoNew(void);
void OCSInfoDelete(ocs_info_struct *ocs_info);

int OCSInfoFileOpen(
	ocs_info_struct ***list, int *total,
	const char *path,
	void (*progress_cb)(unsigned long, unsigned long, void *),
	void *progress_data,
	void (*print_warning)(const char *s)
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Creates a new OCS Info.
 */
ocs_info_struct *OCSInfoNew(void)
{
	return(OCS_INFO(calloc(1, sizeof(ocs_info_struct))));
}

/*
 *	Deletes the OCS info.
 */
void OCSInfoDelete(ocs_info_struct *ocs_info)
{
	if(ocs_info == NULL)
	    return;

	free(ocs_info->name);
	free(ocs_info->icon_file);
	free(ocs_info->icon_eco_file);
	free(ocs_info);
}


/*
 *	Opens the OCS names from the specified file.
 */
int OCSInfoFileOpen(
	ocs_info_struct ***list, int *total,
	const char *path,
	void (*progress_cb)(unsigned long, unsigned long, void *),
	void *progress_data,
	void (*print_warning)(const char *s)
)
{
	int	lines_read = 0,
		ocs_info_num = -1;
	unsigned long file_size = 0l;
	char	*s, *s2,
		parm[CFG_PARAMETER_MAX],
		val[CFG_VALUE_MAX];
	FILE *fp;
	struct stat stat_buf;
	ocs_info_struct *ocs_info;

        /* Report initial progress */
        if(progress_cb != NULL)
            progress_cb(0l, file_size, progress_data);

	if((list == NULL) || (total == NULL) || STRISEMPTY(path))
	    return(-2);

	/* Open the file for reading */
	fp = fopen(path, "rb");
	if(fp == NULL)
	{
	    const int error_code = errno;
	    char *s, *error_msg = STRDUP(strerror(error_code));
	    if(error_msg == NULL)
		error_msg = strdup("unable to open file for reading");
	    *error_msg = toupper(*error_msg);

	    s = (char *)malloc(
		STRLEN(path) + STRLEN(error_msg) + 80
	    );
	    sprintf(
		s,
		"%s: %s.",
		path, error_msg
	    );
	    if(print_warning != NULL)
		print_warning(s);
	    free(s);

	    free(error_msg);

	    return(-1);
	}

	if(!fstat(fileno(fp), &stat_buf))
	{
	    file_size = (unsigned long)stat_buf.st_size;

	    if(progress_cb != NULL)
		progress_cb((unsigned long)ftell(fp), file_size, progress_data);
	}

#define WARN_NO_SUCH_PARM(_parm_)               {       \
 char *s = (char *)malloc(                              \
  STRLEN(path) + 40 + STRLEN(_parm_) + 80               \
 );                                                     \
 sprintf(                                               \
  s,                                                    \
"%s: Line %i: Unsupported parameter `%s'.",             \
  path, lines_read, (_parm_)                            \
 );                                                     \
 if(print_warning != NULL)				\
  print_warning(s);					\
 free(s);                                               \
}
 
#define WARN_NO_SUCH_FILE(_v_,_error_code_)     {       \
 const int error_code = (_error_code_);                 \
 char *s, *error_msg = STRDUP(strerror(error_code));    \
 if(error_msg == NULL)                                  \
  error_msg = strdup("no such file");                   \
 *error_msg = toupper(*error_msg);                      \
							\
 s = (char *)malloc(                                    \
  STRLEN(path) + 40 + STRLEN(parm) + STRLEN(_v_) +      \
  STRLEN(error_msg) + 80                                \
 );                                                     \
 sprintf(                                               \
  s,                                                    \
"%s: Line %i: %s: Warning: %s: %s.",                    \
  path, lines_read, parm, (_v_), error_msg              \
 );                                                     \
 if(print_warning != NULL)				\
  print_warning(s);					\
 free(s);                                               \
							\
 free(error_msg);                                       \
}

	/* Begin reading the file */
	s = NULL;
	while(1)
	{
	    /* Get the next line */
	    free(s);
	    s = FReadNextLineAllocCount(
		fp, UNIXCFG_COMMENT_CHAR, &lines_read
	    );
	    if(s == NULL)
		break;

	    /* Get the parameter from the line */
	    s2 = StringCfgParseParm(s);
	    if(s2 == NULL)
		continue;
	    strncpy(parm, s2, sizeof(parm));
	    parm[sizeof(parm) - 1] = '\0';

	    /* Get the value from the line */
	    s2 = StringCfgParseValue(s);
	    if(s2 == NULL)
		s2 = "0";
	    strncpy(val, s2, sizeof(val));
	    val[sizeof(val) - 1] = '\0';

	    /* Report progress */
	    if(progress_cb != NULL)
		progress_cb((unsigned long)ftell(fp), file_size, progress_data);


	    /* BeginHeader */
	    if(!strcasecmp(parm, "BeginHeader"))
	    {
		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* Title */
		    if(!strcasecmp(parm, "Title"))
		    {


		    }
		    /* EndHeader */
		    else if(!strcasecmp(parm, "EndHeader"))
		    {
			break;
		    }
		}
		continue;
	    }
	    /* BeginOCSN */
	    else if(!strcasecmp(parm, "BeginOCSN") ||
		    !strcasecmp(parm, "BeginOCSNEntry") ||
		    !strcasecmp(parm, "BeginOCSEntry")
	    )
	    {
		/* Create a new OCS Info */
		ocs_info_num = MAX((*total), 0);
		*total = ocs_info_num + 1;
		*list = (ocs_info_struct **)realloc(
		    *list,
		    (*total) * sizeof(ocs_info_struct *)
		);
		(*list)[ocs_info_num] = ocs_info = OCSInfoNew();
		if(ocs_info == NULL)
		    break;

		while(1)
		{
		    /* Get the next line */
		    free(s);
		    s = FReadNextLineAllocCount(
			fp, UNIXCFG_COMMENT_CHAR, &lines_read
		    );
		    if(s == NULL)
			break;

		    /* Get the parameter from the line */
		    s2 = StringCfgParseParm(s);
		    if(s2 == NULL)
			continue;
		    strncpy(parm, s2, sizeof(parm));
		    parm[sizeof(parm) - 1] = '\0';

		    /* Get the value from the line */
		    s2 = StringCfgParseValue(s);
		    if(s2 == NULL)
			s2 = "0";
		    strncpy(val, s2, sizeof(val));
		    val[sizeof(val) - 1] = '\0';

		    /* Report progress */
		    if(progress_cb != NULL)
			progress_cb((unsigned long)ftell(fp), file_size, progress_data);


		    /* Code */
		    if(!strcasecmp(parm, "Code") ||
		       !strcasecmp(parm, "Type")
		    )
		    {
			ocs_info->code = ATOI(val);
		    }
		    /* Name */
		    else if(!strcasecmp(parm, "Name"))
		    {
			free(ocs_info->name);
			ocs_info->name = STRDUP(val);
		    }
		    /* IconFile */
		    else if(!strcasecmp(parm, "IconFile"))
		    {
			free(ocs_info->icon_file);
			ocs_info->icon_file = STRDUP(val);
		    }
		    /* EcoMiniIconFile */
		    else if(!strcasecmp(parm, "EcoMiniIconFile") ||
			    !strcasecmp(parm, "EcoIconFile")
		    )
		    {
			free(ocs_info->icon_eco_file);
			ocs_info->icon_eco_file = STRDUP(val);
		    }
		    /* EndOCSNEntry */
		    else if(!strcasecmp(parm, "EndOCSNEntry") ||
			    !strcasecmp(parm, "EndOCSEntry")
		    )
		    {
			break;
		    }
		    /* Unsupported property */
		    else
		    {
			WARN_NO_SUCH_PARM(parm);
		    }
		}
	    }
	    /* Unsupported property */
	    else
	    {
		WARN_NO_SUCH_PARM(parm);
	    }
	}
	free(s);

#undef WARN_NO_SUCH_PARM
#undef WARN_NO_SUCH_FILE

	/* Close the file */
	fclose(fp);

	/* Report final progress */
	if(progress_cb != NULL)
	    progress_cb(file_size, file_size, progress_data);

	return(0);
}
