#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/osw-x.h"
#include "../include/widget.h"

#include "keymap.h"
#include "uew.h"
#include "optwin.h"
#include "ue.h"

#include "../include/oswkeyvals.h"


static void OptWinDrawOutline(
	optwin_struct *ow, drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *name,
	pixel_t fg_pix, pixel_t bg_pix, pixel_t text_pix
);
static void OptWinCreateBrowseButton(
	optwin_struct *ow,
	prompt_struct *prompt,		/* Reference prompt */
	win_t parent,
	push_button_struct *btn,
	void (*activate_cb)(push_button_struct *, void *)
);

/* Set Tab */
void OptWinSetTab(optwin_struct *ow, const int tab_num);

/* Get/Set Values */
static void OptWinGetValues(optwin_struct *ow);
static void OptWinSetValues(optwin_struct *ow);

/* Keymaps */
static void OptWinKeymapListItemSetKeyCode(
	optwin_struct *ow, keycode_t keycode
);

/* Callbacks */
static void OptWinDrawCB(void *widget, void *data);
static void OptWinBrowseDirectoryCB(push_button_struct *btn, void *data);
static void OptWinBrowseFileCB(push_button_struct *btn, void *data);
static void OptWinKeymapCListActivateCB(
	clist_struct *clist, void *data
);
static void OptWinScanKeyPBCB(push_button_struct *btn, void *data);
static void OptWinOKPBCB(push_button_struct *btn, void *data);
static void OptWinApplyPBCB(push_button_struct *btn, void *data);
static void OptWinCancelPBCB(push_button_struct *btn, void *data);

/* Options Window */
optwin_struct *OptWinNew(void);
void OptWinDraw(optwin_struct *ow, const int amount);
void OptWinQueueDraw(optwin_struct *ow, const int amount);
int OptWinManage(optwin_struct *ow, event_t *event);
void OptWinMap(optwin_struct *ow);
void OptWinMapValues(optwin_struct *ow, win_t toplevel);
void OptWinUnmap(optwin_struct *ow);
void OptWinDelete(optwin_struct *ow);


#define OPTWIN_DEF_TITLE		PROG_NAME_FULL ": Options"
#define OPTWIN_DEF_ICON_TITLE		"Options"

#define OPTWIN_WIDTH			640
#define OPTWIN_HEIGHT			480

#define OPTWIN_MARGIN			10

#define OPTWIN_PROMPT_HEIGHT		30

#define OPTWIN_ICON_BTN_WIDTH		26
#define OPTWIN_ICON_BTN_HEIGHT		26

#define OPTWIN_BTN_WIDTH		70
#define OPTWIN_BTN_HEIGHT		28


#define OPTWIN_OUTLINE_NAME_POS	0.10	/* Coefficient */

#define OPTWIN_CHAR_WIDTH		7
#define OPTWIN_CHAR_HEIGHT		14


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Draws an outline box with label.
 */
static void OptWinDrawOutline(
	optwin_struct *ow, drawable_t d,
	int x, int y,
	unsigned int width, unsigned int height,
	const char *text,
	pixel_t fg_pix, pixel_t bg_pix, pixel_t text_pix
)
{
	int	lbearing, rbearing, swidth, ascent, descent,
		font_height, text_start_pos, text_len;
	font_t *font = widget_global.std_font;

	if((ow == NULL) || (d == None) || (text == NULL))
	    return;

	text_start_pos = OPTWIN_MARGIN;
	OSWGetTextBounds(
	    font, text, -1,
	    &lbearing, &rbearing, &swidth, &ascent, &descent
	);
	font_height = ascent + descent;
	text_len = swidth + (2 * 2);

	/* Draw shadows */
	OSWSetForeground(bg_pix);
	OSWDrawLine(
	    d,				/* left */
	    x + 2,
	    y + (int)height - 2,
	    x + 2,
	    y + 2
	);
	OSWDrawLine(
	    d,				/* top1 */
	    x + 4,
	    y + 3,
	    x + text_start_pos,
	    y + 3
	);
	OSWDrawLine(
	    d,				/* Top2 */
	    x + text_len + text_start_pos,
	    y + 3,
	    x + (int)width - 2,
	    y + 3
	);
	OSWDrawLine(
	    d,				/* right */
	    x + (int)width - 3,
	    y + 3,
	    x + (int)width - 3,
	    y + (int)height - 3
	);
	OSWDrawLine(
	    d,				/* bottom */
	    x + 3,
	    y + (int)height - 3,
	    x + (int)width - 3,
	    y + (int)height - 3
	);

	/* Draw highlights */
	OSWSetForeground(fg_pix);
	OSWDrawLine(
	    d,				/* left */
	    x + 3,
	    y + (int)height - 3,
	    x + 3,
	    y + 3
	);
	OSWDrawLine(
	    d,				/* top1 */
	    x + 3,
	    y + 3,
	    x + text_start_pos,
	    y + 3
	);
	OSWDrawLine(
	    d,				/* Top2 */
	    x + text_len + text_start_pos,
	    y + 3,
	    x + (int)width - 3,
	    y + 3
	);
	OSWDrawLine(
	    d,				/* right */
	    x + (int)width - 2,
	    y + 3,
	    x + (int)width - 2,
	    y + (int)height - 2
	);
	OSWDrawLine(
	    d,				/* bottom */
	    x + 3,
	    y + (int)height - 2,
	    x + (int)width - 2,
	    y + (int)height - 2
	);

	/* Draw text */
	OSWSetForeground(text_pix);
	OSWSetFont(font);
	OSWDrawString(
	    d,
	    x + 2 + text_start_pos + 2 - lbearing,
	    y + 2 - (font_height / 2) + ascent,
	    text
	);
}

/*
 *	Creates a browse button.
 *
 *	The activate_cb's data will be passed as the specified
 *	reference prompt.
 */
static void OptWinCreateBrowseButton(
	optwin_struct *ow,
	prompt_struct *prompt,		/* Reference prompt */
	win_t parent,
	push_button_struct *btn,
	void (*activate_cb)(push_button_struct *, void *)
)
{
	int x, y;
	const unsigned int	width = OPTWIN_ICON_BTN_WIDTH,
				height = OPTWIN_ICON_BTN_HEIGHT;
	widget_global_struct *wglobal = &widget_global;

	if(prompt != NULL)
	{
	    x = prompt->x + (int)prompt->width + 2;
	    y = prompt->y +
		(((int)prompt->height - (int)height) / 2);
	}
	else
	{
	    x = 0;
	    y = 0;
	}

	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    NULL,
	    PBTN_LABEL_ALIGN_CENTER,
	    wglobal->browse_files_pixmap,
	    wglobal->browse_files_mask,
	    activate_cb, prompt
	);   
}


/*
 *	Draw callback.
 */
static void OptWinDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	optwin_struct *ow = OPTWIN(widget);
	if(ow == NULL)
	    return;

	if(ow->map_state)
	    OptWinDraw(ow, amount);
}


/*
 *	Maps/unmaps widgets according to the specified tab.
 */
void OptWinSetTab(optwin_struct *ow, const int tab_num)
{
	int prev_tab;
	bool_t just_map = False;

	if(ow == NULL)
	    return;

	/* No change? */
	if(ow->tab == tab_num)
	    just_map = True;

	prev_tab = ow->tab;
	ow->tab = tab_num;

	/* Unmap previous tab */
	if(!just_map)
	{
	    switch(prev_tab)
	    {
	      case OPTWIN_TAB_GENERAL:
		WidgetUnmap(&ow->record_window_geometry_cb);
	        break;

	      case OPTWIN_TAB_FONTS:
		WidgetUnmap(&ow->geometry_label_font_prompt);
	        WidgetUnmap(&ow->object_label_font_prompt);
		break;

	      case OPTWIN_TAB_DISPLAY:
	        WidgetUnmap(&ow->show_object_labels_cb);
		WidgetUnmap(&ow->show_geometry_labels_cb);
		WidgetUnmap(&ow->show_grid_cb);
		WidgetUnmap(&ow->show_contact_bounds_cb); 
		WidgetUnmap(&ow->animate_object_display_cb);
		WidgetUnmap(&ow->grid_spacing_prompt);
		WidgetUnmap(&ow->object_marker_radius_prompt);
		break;

	      case OPTWIN_TAB_LOCATIONS:
		WidgetUnmap(&ow->toplevel_path_prompt);
		WidgetUnmap(&ow->toplevel_path_btn);
		WidgetUnmap(&ow->etc_path_prompt);
		WidgetUnmap(&ow->etc_path_btn);
		WidgetUnmap(&ow->images_path_prompt);
		WidgetUnmap(&ow->images_path_btn);
		WidgetUnmap(&ow->sounds_path_prompt);
		WidgetUnmap(&ow->sounds_path_btn);
		WidgetUnmap(&ow->server_path_prompt);
		WidgetUnmap(&ow->server_path_btn);
		WidgetUnmap(&ow->opm_file_prompt);
		WidgetUnmap(&ow->opm_file_btn);
		break;

	      case OPTWIN_TAB_KEYMAP:
		WidgetUnmap(&ow->keymaps_list);
		WidgetUnmap(&ow->scan_key_btn);
		OSWUnmapWindow(ow->scan_key_win);
		ow->scanning_key = False;
		break;
	    }
	}

	/* Map newly selected tab */
	switch(ow->tab)
	{
	  case OPTWIN_TAB_GENERAL:
	    WidgetMap(&ow->record_window_geometry_cb);
	    break;

	  case OPTWIN_TAB_FONTS:
	    WidgetMap(&ow->geometry_label_font_prompt);
	    WidgetMap(&ow->object_label_font_prompt);
	    break;

	  case OPTWIN_TAB_DISPLAY:
	    WidgetMap(&ow->show_object_labels_cb);
	    WidgetMap(&ow->show_geometry_labels_cb);
	    WidgetMap(&ow->show_grid_cb);
	    WidgetMap(&ow->show_contact_bounds_cb); 
	    WidgetMap(&ow->animate_object_display_cb);
	    WidgetMap(&ow->grid_spacing_prompt);
	    WidgetMap(&ow->object_marker_radius_prompt);
	    break;

	  case OPTWIN_TAB_LOCATIONS:
	    WidgetMap(&ow->toplevel_path_prompt);
	    WidgetMap(&ow->toplevel_path_btn);
	    WidgetMap(&ow->etc_path_prompt);
	    WidgetMap(&ow->etc_path_btn); 
	    WidgetMap(&ow->images_path_prompt);
	    WidgetMap(&ow->images_path_btn);
	    WidgetMap(&ow->sounds_path_prompt);
	    WidgetMap(&ow->sounds_path_btn);
	    WidgetMap(&ow->server_path_prompt);
	    WidgetMap(&ow->server_path_btn);
	    WidgetMap(&ow->opm_file_prompt);
	    WidgetMap(&ow->opm_file_btn);
	    break;

	  case OPTWIN_TAB_KEYMAP:
	    WidgetMap(&ow->keymaps_list);
	    WidgetMap(&ow->scan_key_btn);
	    if(ow->scanning_key)
		OSWMapRaised(ow->scan_key_win);
	    else
	        OSWUnmapWindow(ow->scan_key_win);
	    break;
	}
}


/*
 *	Gets the options values and sets them to the Option Window's
 *	widgets.
 */
static void OptWinGetValues(optwin_struct *ow)
{
	int i, row_num;
	char *v[2];
	keycode_t keycode;
	clist_struct *clist;
	ue_option_struct *opt = &option;

	if(ow == NULL)
	    return;

	/* General */
	CBtnSetActive(
	    &ow->record_window_geometry_cb,
	    opt->record_window_geometry
	);


	/* Fonts */
	PromptSetS(
	    &ow->object_label_font_prompt,
	    opt->object_label_font_name
	);
	PromptSetS(
	    &ow->geometry_label_font_prompt,
	    opt->geometry_label_font_name
	);


	/* Display */
	CBtnSetActive(
	    &ow->show_object_labels_cb,
	    opt->show_object_labels
	);
	CBtnSetActive(
	    &ow->show_geometry_labels_cb,
	    opt->show_geometry_labels
	);
	CBtnSetActive(
	    &ow->show_grid_cb,
	    opt->show_grid
	);
	CBtnSetActive(
	    &ow->show_contact_bounds_cb,
	    opt->show_contact_bounds
	);
	CBtnSetActive(
	    &ow->animate_object_display_cb,
	    opt->animate_object_display
	);
	PromptSetF2(&ow->grid_spacing_prompt, opt->grid_spacing, 2);
	PromptSetI(&ow->object_marker_radius_prompt, opt->object_marker_radius);


	/* Paths */
	PromptSetS(&ow->toplevel_path_prompt, dname.toplevel);
	PromptSetS(&ow->etc_path_prompt, dname.etc);
	PromptSetS(&ow->images_path_prompt, dname.images);
	PromptSetS(&ow->sounds_path_prompt, dname.sounds);
	PromptSetS(&ow->server_path_prompt, dname.server);
	PromptSetS(&ow->opm_file_prompt, fname.opm);


	/* Keymaps List */
	clist = &ow->keymaps_list;
	CListClear(clist);
	for(i = 0; i < UE_TOTAL_KEYMAPS; i++)
	{
	    row_num = CListRowAppend(&ow->keymaps_list);
	    if(row_num < 0)
		break;

	    keycode = keymap[i].keycode;

	    v[0] = STRDUP(keymap_name[i]);
	    v[1] = STRDUP(OSWGetKeyCodeName(keycode));

	    CListSetRowTexts(
		clist, row_num,
		(const char **)v, sizeof(v) / sizeof(char *)
	    );

	    free(v[0]);
	    free(v[1]);

	    CListRowSetData(clist, row_num, (void *)keycode, NULL);
	}

	ow->has_changes = False;
}

/*
 *	Sets the values from the Options Window's widgets to the
 *	options.
 */
static void OptWinSetValues(optwin_struct *ow)
{
	int i, n;
	const char *s;
	clist_struct *clist;
	ue_option_struct *opt = &option;

	if(ow == NULL)
	    return;


	/* General */
	opt->record_window_geometry = CBtnGetActive(
	    &ow->record_window_geometry_cb
	);


	/* Fonts */
	free(opt->object_label_font_name);
	opt->object_label_font_name = STRDUP(PromptGetS(
	    &ow->object_label_font_prompt
	));
	free(opt->geometry_label_font_name);
	opt->geometry_label_font_name = STRDUP(PromptGetS(
	    &ow->geometry_label_font_prompt
	));


	/* Display */
	opt->show_object_labels = CBtnGetActive(
	    &ow->show_object_labels_cb
	);
	opt->show_geometry_labels = CBtnGetActive(
	    &ow->show_geometry_labels_cb
	);
	opt->show_grid = CBtnGetActive(
	    &ow->show_grid_cb
	);
	opt->show_contact_bounds = CBtnGetActive(
	    &ow->show_contact_bounds_cb
	);
	opt->animate_object_display = CBtnGetActive(
	    &ow->animate_object_display_cb
	);
 	opt->grid_spacing = MAX(
	    PromptGetF(&ow->grid_spacing_prompt),
	    0.01f
	);
	opt->object_marker_radius = MAX(
	    PromptGetI(&ow->object_marker_radius_prompt),
	    1
	);


	/* Paths */
	s = PromptGetS(&ow->toplevel_path_prompt);
	if(s != NULL)
	{
	    strncpy(
		dname.toplevel, s,
		sizeof(dname.toplevel)
	    );
	    dname.toplevel[sizeof(dname.toplevel) - 1] = '\0';
	}

	s = PromptGetS(&ow->etc_path_prompt);
	if(s != NULL)
	{
	    strncpy(dname.etc, s, sizeof(dname.etc));
	    dname.etc[sizeof(dname.etc) - 1] = '\0';
	}

	s = PromptGetS(&ow->images_path_prompt);
	if(s != NULL)
	{
	    strncpy(dname.images, s, sizeof(dname.images));
	    dname.images[sizeof(dname.images) - 1] = '\0';
	}

	s = PromptGetS(&ow->sounds_path_prompt);
	if(s != NULL)
	{
	    strncpy(dname.sounds, s, sizeof(dname.sounds));
	    dname.sounds[sizeof(dname.sounds) - 1] = '\0';
	}

	s = PromptGetS(&ow->server_path_prompt);
	if(s != NULL) 
	{
	    strncpy(dname.server, s, sizeof(dname.server));
	    dname.server[sizeof(dname.server) - 1] = '\0';
	}

	s = PromptGetS(&ow->opm_file_prompt);
	if(s != NULL)
	{
	    strncpy(fname.opm, s, sizeof(fname.opm));
	    fname.opm[sizeof(fname.opm) - 1] = '\0';
	}


	/* Keymaps */
	clist = &ow->keymaps_list;
	n = MIN(clist->total_rows, UE_TOTAL_KEYMAPS);
	for(i = 0; i < n; i++)
	    keymap[i].keycode = (keycode_t)CListRowGetData(clist, i);

	/* Reset has changes mark */
	ow->has_changes = False;


	/* Update all universe editor windows due to changes */
	if(total_uews > 0)
	{
	    uew_struct *uew;
	    for(i = 0; i < total_uews; i++)
	    {
		uew = ue_uew[i];
		if(uew == NULL)
		    continue;

		UEWUpdateWidgets(uew);
		if(uew->map_state)
		    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);
	    }
	}
}


/*
 *	Sets the keycode for the currently selected item on the
 *	keymaps list.
 */
static void OptWinKeymapListItemSetKeyCode(optwin_struct *ow, keycode_t keycode)
{
	int row_num, column_num;
	clist_struct *clist;

	if(ow == NULL)
	    return;

	clist = &ow->keymaps_list;
	row_num = CListGetLastSelectedRow(clist);
	if(row_num < 0)
	    return;

	column_num = 1;

	CListSetItemText(
	    clist, row_num, column_num,
	    OSWGetKeyCodeName(keycode)
	);
	CListRowSetData(clist, row_num, (void *)keycode, NULL);
}

/*
 *	Browse directory callback.
 */
static void OptWinBrowseDirectoryCB(push_button_struct *btn, void *data)
{
	int npaths;
	char *path, **paths_list;

	prompt_struct *prompt = PROMPT(data);
	if((btn == NULL) || (prompt == NULL))
	    return;

	path = STRDUP(PrefixPaths(
	    GetParentDir(PromptGetS(prompt)),
	    "*"
	));

	if(!UESelectFile(
	    "Select Directory",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    WidgetGetParent(prompt)
	))
	{
	    free(path);
	    return;
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);                            

	PromptSetS(prompt, path);

	free(path);
}

/*
 *	Browse file callback.
 */
static void OptWinBrowseFileCB(push_button_struct *btn, void *data)
{
	int npaths;
	char *path, **paths_list;

	prompt_struct *prompt = PROMPT(data);
	if((btn == NULL) || (prompt == NULL))
	    return;

	path = STRDUP(PrefixPaths(
	    GetParentDir(PromptGetS(prompt)),
	    "*"
	));

	if(!UESelectFile(
	    "Select File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    WidgetGetParent(prompt)
	))
	{
	    free(path);
	    return;
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);                            

	PromptSetS(prompt, path);

	free(path);
}

/*
 *	Keymap list callback.
 */
static void OptWinKeymapCListActivateCB(
	clist_struct *clist, void *data
)
{
	optwin_struct *ow = (optwin_struct *)data;
	if(ow == NULL)
	    return;

	OptWinScanKeyPBCB(&ow->scan_key_btn, ow);
}

/*
 *	Scan key button callback.
 */
static void OptWinScanKeyPBCB(push_button_struct *btn, void *data)
{
	optwin_struct *ow = (optwin_struct *)data;
	if(ow == NULL)
	    return;

	OSWMapRaised(ow->scan_key_win);
	ow->scanning_key = True;
}

/*
 *	OK button callback.
 */
static void OptWinOKPBCB(push_button_struct *btn, void *data)
{
	optwin_struct *ow = (optwin_struct *)data;
	if(ow == NULL)
	    return;

	OptWinSetValues(ow);
	OptWinUnmap(ow);
}

/*      
 *	Apply button callback.
 */
static void OptWinApplyPBCB(push_button_struct *btn, void *data)
{
	optwin_struct *ow = (optwin_struct *)data;
	if(ow == NULL)
	    return;

	OptWinSetValues(ow);
}

/*
 *	Cancel button callback.
 */
static void OptWinCancelPBCB(push_button_struct *btn, void *data)
{
	optwin_struct *ow = (optwin_struct *)data;
	if(ow == NULL)
	    return;

	OptWinUnmap(ow);
}


/*
 *	Creates a new Options Window.
 */
optwin_struct *OptWinNew(void)
{
	int x, y, x2, y2;
	unsigned int width, height, width2, height2;
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
	font_t *font;
	win_t w, parent;
	push_button_struct *btn;
	clist_struct *clist;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;
	optwin_struct *ow = OPTWIN(calloc(1, sizeof(optwin_struct)));
	if(ow == NULL)
	    return(NULL);

	ow->map_state = False;
	ow->has_focus = False;
	ow->x = 0;
	ow->y = 0;
	ow->width = OPTWIN_WIDTH;
	ow->height = OPTWIN_HEIGHT;

	ow->has_changes = False;
	ow->tab = OPTWIN_TAB_GENERAL;
	ow->scanning_key = False;


	/* Create toplevel */
	parent = gui->root_win;
	x = ow->x;
	y = ow->y;
	width = ow->width;
	height = ow->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	ow->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    OPTWIN_DEF_TITLE,
	    OPTWIN_DEF_ICON_TITLE,
	    wglobal->std_icon_pm,	/* Icon */
	    False,			/* Let WM set coordinates? */
	    ow->x, ow->y,
	    100, 72,
	    gui->display_width, gui->display_height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	WidgetCenterWindow(w, WIDGET_WINDOW_RELATIVE_ROOT);

	parent = w;

	/* Create tab windows */
	width = ((wglobal->tab_unselected_img == NULL) ?
	    10 : wglobal->tab_unselected_img->width
	);
	height = ((wglobal->tab_unselected_img == NULL) ?
	    10 : wglobal->tab_unselected_img->height
	);

	/* General Tab */
	x = (2 * OPTWIN_MARGIN);
	y = OPTWIN_MARGIN;
	OSWCreateWindow(
	    &ow->general_tab_win, parent,
	    x, y, width, height
	);
	OSWSetWindowInput(ow->general_tab_win, OSW_EVENTMASK_BUTTON);

	/* Fonts Tab */
	x += (int)width + OPTWIN_MARGIN;
	OSWCreateWindow(
	    &ow->fonts_tab_win, parent,
	    x, y, width, height
	);
	OSWSetWindowInput(ow->fonts_tab_win, OSW_EVENTMASK_BUTTON);

	/* Display Tab */
	x += (int)width + OPTWIN_MARGIN;
	OSWCreateWindow(
	    &ow->display_tab_win, parent,
	    x, y, width, height
	);
	OSWSetWindowInput(ow->display_tab_win, OSW_EVENTMASK_BUTTON);

	/* Locations Tab */
	x += (int)width + OPTWIN_MARGIN;
	OSWCreateWindow(
	    &ow->paths_tab_win, parent,
	    x, y, width, height
	);
	OSWSetWindowInput(ow->paths_tab_win, OSW_EVENTMASK_BUTTON);

	/* Keymaps Tab */
	x += (int)width + OPTWIN_MARGIN;
	OSWCreateWindow(
	    &ow->keymap_tab_win, parent,
	    x, y, width, height
	);
	OSWSetWindowInput(ow->keymap_tab_win, OSW_EVENTMASK_BUTTON);


	/* Buttons */
	btn = &ow->ok_btn;
	width = OPTWIN_BTN_WIDTH;
	height = OPTWIN_BTN_HEIGHT;
	x = ((int)ow->width / 3 / 2) - ((int)width / 2);
	y = (int)ow->height - OPTWIN_MARGIN - OPTWIN_BTN_HEIGHT;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "OK",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinOKPBCB, ow
	);
	accel_keys[0] = OSW_KEY_o;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_Return;
	accel_mods[1] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 2);

	btn = &ow->apply_btn;
	x += ((int)ow->width / 3);
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Apply", 
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinApplyPBCB, ow
	);
	accel_keys[0] = OSW_KEY_a;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 1);

	btn = &ow->cancel_btn;
	x += ((int)ow->width / 3);
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinCancelPBCB, ow
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);


	/* General */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 65;
	CBtnInit(
	    &ow->record_window_geometry_cb, parent,
	    x, y,
	    False,
	    "Record Window Geometry"
	);
	height = ow->record_window_geometry_cb.height;


	/* Fonts */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 65;
	width = (unsigned int)MAX(
	    (int)ow->width - (6 * OPTWIN_MARGIN) - (4 * 2),
	    1
	);
	height = OPTWIN_PROMPT_HEIGHT;
	PromptInit(
	    &ow->object_label_font_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Object Labels:",
	    256, 0
	);

	y += OPTWIN_PROMPT_HEIGHT;
	PromptInit(
	    &ow->geometry_label_font_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Geometry Labels:",
	    256, 0
	);

	/* Link prompts togeather */
	ow->object_label_font_prompt.prev = &ow->geometry_label_font_prompt;
	ow->object_label_font_prompt.next = &ow->geometry_label_font_prompt;

	ow->geometry_label_font_prompt.prev = &ow->object_label_font_prompt;
	ow->geometry_label_font_prompt.next = &ow->object_label_font_prompt;


	/* Display */
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 65;
	CBtnInit(
	    &ow->show_object_labels_cb, parent,
	    x, y,
	    False,
	    "Show Object Labels"
	);
	height = ow->show_object_labels_cb.height;

	y += (int)height + (OPTWIN_MARGIN / 2);
	CBtnInit(
	    &ow->show_geometry_labels_cb, parent,
	    x, y,
	    False,
	    "Show Geometry Labels"
	);

	y += (int)height + OPTWIN_MARGIN;
	CBtnInit(
	    &ow->show_grid_cb, parent,
	    x, y,
	    False,
	    "Show Grid"
	);

	width2 = 130;
	height2 = OPTWIN_PROMPT_HEIGHT;
	x2 = x + (int)ow->show_grid_cb.width + OPTWIN_MARGIN;
	y2 = y + ((int)height - (int)height2) / 2;
	PromptInit(
	    &ow->grid_spacing_prompt, parent,
	    x2, y2, width2, height2,
	    PROMPT_STYLE_FLUSHED,
	    "Spacing:",
	    40, 0
	);
	PromptSetTooltip(
	    &ow->grid_spacing_prompt,
"Enter the grid spacing in real units"
	);

	y += (int)height + OPTWIN_MARGIN;
	width = 180;
	height = OPTWIN_PROMPT_HEIGHT;
	PromptInit(
	    &ow->object_marker_radius_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Object Marker Radius:",
	    40, 0
	);
	PromptSetTooltip(
	    &ow->object_marker_radius_prompt,
"Enter the radius of the object markers in pixels"
	);

	y += (int)height + OPTWIN_MARGIN;
	CBtnInit(
	    &ow->show_contact_bounds_cb, parent,
	    x, y,
	    False,
	    "Show Contact Bounds"
	);


	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y += (int)height + 45;
	CBtnInit(
	    &ow->animate_object_display_cb, parent,
	    x, y,
	    False,
	    "Animate"
	);


	/* Locations */
	width = (unsigned int)MAX(
	    (int)ow->width - (6 * OPTWIN_MARGIN) - (4 * 2) - 2 -
		OPTWIN_ICON_BTN_WIDTH,
	    1
	);
	height = OPTWIN_PROMPT_HEIGHT;
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 65;
	PromptInit(
	    &ow->toplevel_path_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Toplevel:",
	    PATH_MAX + NAME_MAX, 0
	);
	OptWinCreateBrowseButton(
	    ow,
	    &ow->toplevel_path_prompt,
	    parent,
	    &ow->toplevel_path_btn,
	    OptWinBrowseDirectoryCB
	);

	y += (int)height + (OPTWIN_MARGIN / 2);
	PromptInit(
	    &ow->etc_path_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Etc:",
	    PATH_MAX + NAME_MAX, 0
	);
	OptWinCreateBrowseButton(
	    ow,
	    &ow->etc_path_prompt,
	    parent,
	    &ow->etc_path_btn,
	    OptWinBrowseDirectoryCB
	);

	y += (int)height;
	PromptInit(
	    &ow->images_path_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Images:",
	    PATH_MAX + NAME_MAX, 0
	);
	OptWinCreateBrowseButton(
	    ow,
	    &ow->images_path_prompt,
	    parent,
	    &ow->images_path_btn,
	    OptWinBrowseDirectoryCB
	);

	y += (int)height;
	PromptInit(
	    &ow->sounds_path_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Sounds:",
	    PATH_MAX + NAME_MAX, 0
	);
	OptWinCreateBrowseButton(
	    ow,
	    &ow->sounds_path_prompt,
	    parent,
	    &ow->sounds_path_btn,
	    OptWinBrowseDirectoryCB
	);

	y += (int)height + (OPTWIN_MARGIN / 2);
	PromptInit(
	    &ow->server_path_prompt, parent,
	    x, y, width, height, 
	    PROMPT_STYLE_FLUSHED,
	    "Server:",
	    PATH_MAX + NAME_MAX, 0
	);
	OptWinCreateBrowseButton(
	    ow,
	    &ow->server_path_prompt,
	    parent,
	    &ow->server_path_btn,
	    OptWinBrowseDirectoryCB
	);

	y = 275;
	PromptInit(
	    &ow->opm_file_prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "OPM:",
	    PATH_MAX + NAME_MAX, 0
	);
	OptWinCreateBrowseButton( 
	    ow,
	    &ow->opm_file_prompt,
	    parent,
	    &ow->opm_file_btn,
	    OptWinBrowseFileCB
	);


	/* Link prompts togeather */
	ow->toplevel_path_prompt.prev = &ow->opm_file_prompt;
	ow->toplevel_path_prompt.next = &ow->etc_path_prompt;

	ow->etc_path_prompt.prev = &ow->toplevel_path_prompt;
	ow->etc_path_prompt.next = &ow->images_path_prompt;

	ow->images_path_prompt.prev = &ow->etc_path_prompt;
	ow->images_path_prompt.next = &ow->sounds_path_prompt;

	ow->sounds_path_prompt.prev = &ow->images_path_prompt;
	ow->sounds_path_prompt.next = &ow->server_path_prompt;

	ow->server_path_prompt.prev = &ow->sounds_path_prompt;
	ow->server_path_prompt.next = &ow->opm_file_prompt;

	ow->opm_file_prompt.prev = &ow->server_path_prompt;
	ow->opm_file_prompt.next = &ow->toplevel_path_prompt;


	/* Keymaps */
	clist = &ow->keymaps_list;
	x = (3 * OPTWIN_MARGIN) + (2 * 2);
	y = 65;
	font = wglobal->std_font;
	CListInit(
	    clist, parent,
	    x, y,
	    300, 300
	);
	clist->activate_cb = OptWinKeymapCListActivateCB;
	clist->activate_data = ow;
	CListColumnAppend(
	    clist,
	    "Keymap",
	    font,
	    wglobal->editable_text_pix,
	    0,		/* Attributes */
	    0		/* Start pos */
	);
	CListColumnAppend(
	    clist,
	    "Keycode",
	    font,
	    wglobal->editable_text_pix,
	    0,		/* Attributes */
	    200		/* Start pos */
	);

	btn = &ow->scan_key_btn;
	PBtnInit(
	    btn, parent,
	    x + clist->width + OPTWIN_MARGIN, y,
	    OPTWIN_BTN_WIDTH, OPTWIN_BTN_HEIGHT,
	    "Scan Key",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    OptWinScanKeyPBCB, ow
	);

	width = (unsigned int)MAX(
	    (int)ow->width - (2 * OPTWIN_MARGIN),
	    1
	);
	height = 70;
	OSWCreateWindow(
	    &ow->scan_key_win, parent,
	    OPTWIN_MARGIN,
	    ((int)ow->height / 2) - ((int)height / 2),
	    width, height
	);
	OSWSetWindowInput(
	    ow->scan_key_win,
	    ButtonPressMask | ButtonReleaseMask | ExposureMask
	);
	OSWLoadPixelRGB(&ow->scan_key_fg_pix,
	    0xff,
	    0xff,
	    0xff
	);
	OSWLoadPixelRGB(&ow->scan_key_bg_pix,
	    0x18,
	    0x30,
	    0xff
	);

	return(ow);
}

/*
 *	Draws the Options Window.
 */
void OptWinDraw(optwin_struct *ow, const int amount)
{
	int x, y;
	unsigned int width, height;
	const char *s;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(ow == NULL)
	    return;

	/* Toplevel */
	if(amount == OPTWIN_DRAW_AMOUNT_COMPLETE)
	{
	    image_t *img;
	    win_t w = ow->toplevel;
	    width = ow->width;
	    height = ow->height;

	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, width, height);
		OSWSetForeground(gui->white_pix);
	    }
	    else
	    {
		WidgetPutImageTile(
		    w, wglobal->std_bkg_img,
		    width, height
		);
		OSWSetForeground(wglobal->normal_text_pix);
	    }

	    /* Draw main outline */
	    img = wglobal->tab_unselected_img;
	    x = OPTWIN_MARGIN;
	    y = OPTWIN_MARGIN + ((img != NULL) ? (int)img->height : 0) - 1;

	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->white_pix);
		OSWDrawLine(
		    w,
		    x,
		    (int)ow->height - OPTWIN_BTN_HEIGHT - (2 * OPTWIN_MARGIN),
		    x,
		    y
		);
		OSWDrawLine(
		    w,
		    x,
		    y,
		    (int)ow->width - OPTWIN_MARGIN,
		    y
		);
		OSWDrawLine(
		    w,
		    (int)ow->width - OPTWIN_MARGIN,
		    y,
		    (int)ow->width - OPTWIN_MARGIN,
		    (int)ow->height - OPTWIN_BTN_HEIGHT - (2 * OPTWIN_MARGIN)
		);
		OSWDrawLine(
		    w,
		    x,
		    (int)ow->height - OPTWIN_BTN_HEIGHT - (2 * OPTWIN_MARGIN),
		    (int)ow->width - OPTWIN_MARGIN,
		    (int)ow->height - OPTWIN_BTN_HEIGHT - (2 * OPTWIN_MARGIN)
		);
	    }
	    else
	    {
		OSWSetForeground(wglobal->surface_highlight_pix);
		OSWDrawLine(
		    w,
		    x, 
		    (int)ow->height - OPTWIN_BTN_HEIGHT - (2 * OPTWIN_MARGIN),
		    x,
		    y
		);
		OSWDrawLine(
		    w,
		    x,
		    y,
		    (int)ow->width - OPTWIN_MARGIN,
		    y
		);

		OSWSetForeground(wglobal->surface_shadow_pix);
		OSWDrawLine(
		    w,
		    (int)ow->width - OPTWIN_MARGIN,
		    y,
		    (int)ow->width - OPTWIN_MARGIN,
		    (int)ow->height - OPTWIN_BTN_HEIGHT - (2 * OPTWIN_MARGIN)
		);
		OSWDrawLine(
		    w, 
		    x, 
		    (int)ow->height - OPTWIN_BTN_HEIGHT - (2 * OPTWIN_MARGIN),
		    (int)ow->width - OPTWIN_MARGIN,
		    (int)ow->height - OPTWIN_BTN_HEIGHT - (2 * OPTWIN_MARGIN)
		);
	    }


	    /* Draw stuff specific to each tab */

	    if(ow->tab == OPTWIN_TAB_GENERAL)
	    {
		OptWinDrawOutline(
		    ow, w,
		    x + OPTWIN_MARGIN, y + OPTWIN_MARGIN,
		    (int)ow->width - (4 * OPTWIN_MARGIN), 55,
		    "General",
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix,
		    wglobal->normal_text_pix
		);
	    }
	    else if(ow->tab == OPTWIN_TAB_FONTS)
	    {
		OptWinDrawOutline(
		    ow, w,
		    x + OPTWIN_MARGIN, y + OPTWIN_MARGIN,
		    (int)ow->width - (4 * OPTWIN_MARGIN), 90,
		    "Fonts",
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix,
		    wglobal->normal_text_pix
		);
	    }
	    else if(ow->tab == OPTWIN_TAB_DISPLAY)
	    {
		OptWinDrawOutline(
		    ow, w,
		    x + OPTWIN_MARGIN, y + OPTWIN_MARGIN,
		    (int)ow->width - (4 * OPTWIN_MARGIN), 170,
		    "View",
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix,
		    wglobal->normal_text_pix
		);
		OptWinDrawOutline(
		    ow, w,
		    x + OPTWIN_MARGIN, y + OPTWIN_MARGIN + 185,
		    (int)ow->width - (4 * OPTWIN_MARGIN), 60,
		    "Object Display",
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix,
		    wglobal->normal_text_pix
		);
	    }
	    else if(ow->tab == OPTWIN_TAB_LOCATIONS)
	    {
		OptWinDrawOutline(
		    ow, w,
		    x + OPTWIN_MARGIN, y + OPTWIN_MARGIN,
		    (int)ow->width - (4 * OPTWIN_MARGIN), 195,
		    "Directories",
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix,
		    wglobal->normal_text_pix
		);
		OptWinDrawOutline(
		    ow, w,
		    x + OPTWIN_MARGIN, y + OPTWIN_MARGIN + 210,
		    (int)ow->width - (4 * OPTWIN_MARGIN), 60,
		    "Files",
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix,
		    wglobal->normal_text_pix
		);  
	    }
	    else if(ow->tab == OPTWIN_TAB_KEYMAP)
	    {
		OptWinDrawOutline(
		    ow, w,
		    x + OPTWIN_MARGIN, y + OPTWIN_MARGIN,
		    (int)ow->width - (4 * OPTWIN_MARGIN), 330,
		    "Keymaps",
		    wglobal->surface_highlight_pix,
		    wglobal->surface_shadow_pix,
		    wglobal->normal_text_pix
		);
	    }
	}


	/* Tabs */
	if((amount == OPTWIN_DRAW_AMOUNT_COMPLETE) ||
	   (amount == OPTWIN_DRAW_AMOUNT_TABS)
	)
	{
	    const int tab = ow->tab;
	    int	lbearing, rbearing, swidth, ascent, descent,
		font_height;
	    font_t *font = wglobal->std_font;

	    OSWSetFont(font);

	    /* General Tab */
	    win_t w = ow->general_tab_win;
	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	    WidgetPutImageTile(
		w, wglobal->std_bkg_img,
		width, height
	    );
	    WidgetPutImageNormal(
		w,
		(tab == OPTWIN_TAB_GENERAL) ?
		    wglobal->tab_selected_img :
		    wglobal->tab_unselected_img,
		0, 0, True
	    );
	    s = "General";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWSetForeground(
		wglobal->force_mono ?
		    gui->white_pix :
		    wglobal->normal_text_pix
	    );
	    OSWDrawString(
		w,
		(((int)width - swidth) / 2) - lbearing,
		(((int)height - font_height) / 2) + 3 + ascent,
		s
	    );


	    /* Fonts Tab */
	    w = ow->fonts_tab_win;
	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	    WidgetPutImageTile( 
		w, wglobal->std_bkg_img,
		width, height
	    );
	    WidgetPutImageNormal(
		w,
		(tab == OPTWIN_TAB_FONTS) ?
		    wglobal->tab_selected_img :
		    wglobal->tab_unselected_img,
		0, 0, True
	    );
	    s = "Fonts";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWSetForeground(
		wglobal->force_mono ?
		    gui->white_pix :
		    wglobal->normal_text_pix
	    );
	    OSWDrawString(
		w,
		(((int)width - swidth) / 2) - lbearing,
		(((int)height - font_height) / 2) + 3 + ascent,
		s
	    );


	    /* Display Tab */
	    w = ow->display_tab_win;
	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	    WidgetPutImageTile(
		w, wglobal->std_bkg_img,
		width, height
	    );
	    WidgetPutImageNormal(
		w,
		(tab == OPTWIN_TAB_DISPLAY) ?
		    wglobal->tab_selected_img :
		    wglobal->tab_unselected_img,
		0, 0, True
	    );
	    s = "Display";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWSetForeground(
		wglobal->force_mono ?
		    gui->white_pix :
		    wglobal->normal_text_pix
	    );
	    OSWDrawString(
		w,
		(((int)width - swidth) / 2) - lbearing,
		(((int)height - font_height) / 2) + 3 + ascent,
		s
	    );


	    /* Locations Tab */
	    w = ow->paths_tab_win;
	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	    WidgetPutImageTile(
		w, wglobal->std_bkg_img,
		width, height
	    );
	    WidgetPutImageNormal(
		w,
		(tab == OPTWIN_TAB_LOCATIONS) ?
		    wglobal->tab_selected_img :
		    wglobal->tab_unselected_img,
		0, 0, True
	    );
	    s = "Locations";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWSetForeground(
		wglobal->force_mono ?
		    gui->white_pix :
		    wglobal->normal_text_pix
	    );
	    OSWDrawString(
		w,
		(((int)width - swidth) / 2) - lbearing,
		(((int)height - font_height) / 2) + 3 + ascent,
		s
	    );


	    /* Keymaps */
	    w = ow->keymap_tab_win;
	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	    WidgetPutImageTile(
		w, wglobal->std_bkg_img,
		width, height
	    );
	    WidgetPutImageNormal(
		w,
		(tab == OPTWIN_TAB_KEYMAP) ?
		    wglobal->tab_selected_img :
		    wglobal->tab_unselected_img,
		0, 0, True
	    );
	    s = "Keymaps";
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    OSWSetForeground(
		wglobal->force_mono ?
		    gui->white_pix :
		    wglobal->normal_text_pix
	    );
	    OSWDrawString(
		w,
		(((int)width - swidth) / 2) - lbearing,
		(((int)height - font_height) / 2) + 3 + ascent,
		s
	    );
	}


	/* Draw scan key prompt window */
	if((amount == OPTWIN_DRAW_AMOUNT_COMPLETE) || 
	   (amount == OPTWIN_DRAW_AMOUNT_SCANKEY)
	)
	{
	    if(ow->scanning_key)
	    {
		int lbearing, ascent, descent, font_height;
		const char *s;
		font_t *font = wglobal->std_font;
		win_t w = ow->scan_key_win;

		OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);

		OSWClearPixmap(
		    w,
		    width, height,
		    wglobal->force_mono ?
			gui->white_pix :
			ow->scan_key_bg_pix
		);
		OSWSetForeground(
		    wglobal->force_mono ?
			gui->black_pix :
			ow->scan_key_fg_pix
		);

		OSWSetFont(font);
		y = 0;
		s = "Press a key on the keyboard to set the keycode.";
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    w,
		    OPTWIN_MARGIN - lbearing,
		    y + OPTWIN_MARGIN + ascent,
		    s
		);

		y += font_height + (OPTWIN_MARGIN / 2);
		s = "Press Button1 to cancel.";
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    w,
		    OPTWIN_MARGIN - lbearing,
		    y + OPTWIN_MARGIN + ascent,
		    s
		);

		y += font_height;
		s = "Press Button3 to clear.";
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, NULL, NULL, &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    w,
		    OPTWIN_MARGIN - lbearing,
		    y + OPTWIN_MARGIN + ascent,
		    s
		);
	    }
	}
}

/*
 *	Queues the Options Window to draw.
 */
void OptWinQueueDraw(optwin_struct *ow, const int amount)
{
	if(ow == NULL)
	    return;

	WidgetQueueDrawAction(
	    OptWinDrawCB, ow, (void *)amount
	);
}

/*
 *	Manages the Options Window event.
 */
int OptWinManage(optwin_struct *ow, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;

	if((ow == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case Expose:
	    if(w == ow->toplevel)
	    {
		OptWinQueueDraw(ow, OPTWIN_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    else if((w == ow->general_tab_win) ||
		    (w == ow->fonts_tab_win) ||
		    (w == ow->display_tab_win) ||
		    (w == ow->paths_tab_win) ||
		    (w == ow->keymap_tab_win)
	    )
	    {
		OptWinQueueDraw(ow, OPTWIN_DRAW_AMOUNT_TABS);
		events_handled++;
	    }
	    else if(w == ow->scan_key_win)
	    {
		OptWinQueueDraw(ow, OPTWIN_DRAW_AMOUNT_SCANKEY);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!ow->has_focus)
		return(events_handled);

	    /* Scan key */
	    if(ow->scanning_key)
	    {
		OptWinKeymapListItemSetKeyCode(ow, event->xkey.keycode);

		OSWUnmapWindow(ow->scan_key_win);
		ow->scanning_key = False;

		events_handled++;
	    }
	    break;

	  case KeyRelease:
	    if(!ow->has_focus)
		return(events_handled);

	    break;

	  case ButtonPress:
	    /* Scanning key? */
	    if(ow->scanning_key)
	    {
		switch(event->xbutton.button)
		{
		  case Button1:
		    /* Cancel */
		    break;
		  case Button3:
		    OptWinKeymapListItemSetKeyCode(ow, 0);
		    break;
		}

		OSWUnmapWindow(ow->scan_key_win);
		ow->scanning_key = False;

		events_handled++;
	    }
	    else if(w == ow->general_tab_win)
	    {
		OptWinSetTab(ow, OPTWIN_TAB_GENERAL);
		events_handled++;
	    }
	    else if(w == ow->fonts_tab_win)
	    {
		OptWinSetTab(ow, OPTWIN_TAB_FONTS);
		events_handled++;
	    }
	    else if(w == ow->display_tab_win)
	    {
		OptWinSetTab(ow, OPTWIN_TAB_DISPLAY);
		events_handled++;
	    }
	    else if(w == ow->paths_tab_win)
	    {
		OptWinSetTab(ow, OPTWIN_TAB_LOCATIONS);
		events_handled++;
	    }
	    else if(w == ow->keymap_tab_win)
	    {
		OptWinSetTab(ow, OPTWIN_TAB_KEYMAP);
		events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == ow->toplevel)
	    {
		ow->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == ow->toplevel)
	    {
		ow->has_focus = False;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(ow->toplevel, event, &t))
	    {
		OSWGrabFocusTime(ow->toplevel, t);
		events_handled++;
		return(events_handled);
	    }
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(ow->toplevel, event))
	    {
		OptWinCancelPBCB(&ow->cancel_btn, ow);
		events_handled++;
		return(events_handled);
	    }
	    break;
	}

#define MANAGE_WIDGET(_w_)	{		\
 if(events_handled == 0)			\
  events_handled += WidgetManage((_w_), event);	\
}

	/* Manage widgets */
	MANAGE_WIDGET(&ow->ok_btn);
	MANAGE_WIDGET(&ow->apply_btn);
	MANAGE_WIDGET(&ow->cancel_btn);

	MANAGE_WIDGET(&ow->record_window_geometry_cb);

	MANAGE_WIDGET(&ow->geometry_label_font_prompt);
	MANAGE_WIDGET(&ow->object_label_font_prompt);

	MANAGE_WIDGET(&ow->show_object_labels_cb);
	MANAGE_WIDGET(&ow->show_geometry_labels_cb);
	MANAGE_WIDGET(&ow->show_grid_cb);
	MANAGE_WIDGET(&ow->show_contact_bounds_cb);
	MANAGE_WIDGET(&ow->animate_object_display_cb);
	MANAGE_WIDGET(&ow->grid_spacing_prompt);
	MANAGE_WIDGET(&ow->object_marker_radius_prompt);

	MANAGE_WIDGET(&ow->toplevel_path_prompt);
	MANAGE_WIDGET(&ow->toplevel_path_btn);
	MANAGE_WIDGET(&ow->etc_path_prompt);
	MANAGE_WIDGET(&ow->etc_path_btn);
	MANAGE_WIDGET(&ow->images_path_prompt);
	MANAGE_WIDGET(&ow->images_path_btn);
	MANAGE_WIDGET(&ow->sounds_path_prompt);
	MANAGE_WIDGET(&ow->sounds_path_btn);
	MANAGE_WIDGET(&ow->server_path_prompt);
	MANAGE_WIDGET(&ow->server_path_btn);
	MANAGE_WIDGET(&ow->opm_file_prompt);
	MANAGE_WIDGET(&ow->opm_file_btn);

	MANAGE_WIDGET(&ow->keymaps_list);
	MANAGE_WIDGET(&ow->scan_key_btn);

#undef MANAGE_WIDGET

	return(events_handled);
}

/*
 *	Maps the Options Window.
 */
void OptWinMap(optwin_struct *ow)
{
	if(ow == NULL)
	    return;

	if(!ow->map_state)
	{
	    OSWMapRaised(ow->toplevel);
	    ow->map_state = True;

	    OSWMapWindow(ow->general_tab_win);
	    OSWMapWindow(ow->fonts_tab_win);
	    OSWMapWindow(ow->display_tab_win);
	    OSWMapWindow(ow->paths_tab_win);
	    OSWMapWindow(ow->keymap_tab_win);

	    OptWinSetTab(ow, ow->tab);

	    WidgetMap(&ow->ok_btn);
	    WidgetMap(&ow->apply_btn);
	    WidgetMap(&ow->cancel_btn);

	    /* Map/unmap scan key prompt window */
	    if(ow->scanning_key)
		OSWMapRaised(ow->scan_key_win);
	    else
		OSWUnmapWindow(ow->scan_key_win);
	}

	OSWGrabFocus(ow->toplevel);   
}

/*
 *	Gets the option values and maps the Options Window.
 */
void OptWinMapValues(optwin_struct *ow, win_t toplevel)
{
	if(ow == NULL)
	    return;

	OptWinGetValues(ow);
	WidgetCenterWindowToWindow(ow->toplevel, toplevel);
	OptWinMap(ow);
}

/*
 *	Unmaps the Options Window.
 */
void OptWinUnmap(optwin_struct *ow)
{
	if(ow == NULL)
	    return;

	/* General */
	WidgetUnmap(&ow->record_window_geometry_cb);

	/* Fonts */
	WidgetUnmap(&ow->geometry_label_font_prompt); 
	WidgetUnmap(&ow->object_label_font_prompt);

	/* Display */
	WidgetUnmap(&ow->show_object_labels_cb);
	WidgetUnmap(&ow->show_geometry_labels_cb);
	WidgetUnmap(&ow->show_grid_cb);
	WidgetUnmap(&ow->show_contact_bounds_cb);
	WidgetUnmap(&ow->animate_object_display_cb);
	WidgetUnmap(&ow->grid_spacing_prompt);
	WidgetUnmap(&ow->object_marker_radius_prompt);

	/* Paths */
	WidgetUnmap(&ow->toplevel_path_prompt);
	WidgetUnmap(&ow->toplevel_path_btn);
	WidgetUnmap(&ow->etc_path_prompt);
	WidgetUnmap(&ow->etc_path_btn);
	WidgetUnmap(&ow->images_path_prompt);
	WidgetUnmap(&ow->images_path_btn);
	WidgetUnmap(&ow->sounds_path_prompt);
	WidgetUnmap(&ow->sounds_path_btn);
	WidgetUnmap(&ow->server_path_prompt);
	WidgetUnmap(&ow->server_path_btn);
	WidgetUnmap(&ow->opm_file_prompt);
	WidgetUnmap(&ow->opm_file_btn);

	/* Keymaps */
	WidgetUnmap(&ow->keymaps_list);
	WidgetUnmap(&ow->scan_key_btn);
	OSWUnmapWindow(ow->scan_key_win);
	ow->scanning_key = False;


	OSWUnmapWindow(ow->general_tab_win);
	OSWUnmapWindow(ow->fonts_tab_win);
	OSWUnmapWindow(ow->display_tab_win);
	OSWUnmapWindow(ow->paths_tab_win);
	OSWUnmapWindow(ow->keymap_tab_win);

	WidgetUnmap(&ow->ok_btn);
	WidgetUnmap(&ow->apply_btn);
	WidgetUnmap(&ow->cancel_btn);

	OSWUnmapWindow(ow->toplevel);
	ow->map_state = False;
}

/*
 *	Deletes the Options Window.
 */
void OptWinDelete(optwin_struct *ow)
{
	if(ow == NULL)
	    return;

	WidgetDestroy(&ow->record_window_geometry_cb);

	WidgetDestroy(&ow->geometry_label_font_prompt);
	WidgetDestroy(&ow->object_label_font_prompt);

	WidgetDestroy(&ow->show_object_labels_cb);
	WidgetDestroy(&ow->show_geometry_labels_cb);
	WidgetDestroy(&ow->show_grid_cb);
	WidgetDestroy(&ow->show_contact_bounds_cb);
	WidgetDestroy(&ow->animate_object_display_cb);
	WidgetDestroy(&ow->grid_spacing_prompt);
	WidgetDestroy(&ow->object_marker_radius_prompt);

	WidgetDestroy(&ow->toplevel_path_prompt);
	WidgetDestroy(&ow->toplevel_path_btn);
	WidgetDestroy(&ow->etc_path_prompt);
	WidgetDestroy(&ow->etc_path_btn);
	WidgetDestroy(&ow->images_path_prompt);
	WidgetDestroy(&ow->images_path_btn);
	WidgetDestroy(&ow->sounds_path_prompt);
	WidgetDestroy(&ow->sounds_path_btn);
	WidgetDestroy(&ow->server_path_prompt);
	WidgetDestroy(&ow->server_path_btn);
	WidgetDestroy(&ow->opm_file_prompt);
	WidgetDestroy(&ow->opm_file_btn);

	WidgetDestroy(&ow->keymaps_list);
	WidgetDestroy(&ow->scan_key_btn);
	OSWDestroyWindow(&ow->scan_key_win);
	ow->scanning_key = False;
	OSWDestroyPixel(&ow->scan_key_fg_pix);
	OSWDestroyPixel(&ow->scan_key_bg_pix);

	WidgetDestroy(&ow->ok_btn);
	WidgetDestroy(&ow->apply_btn);
	WidgetDestroy(&ow->cancel_btn);

	OSWDestroyWindow(&ow->general_tab_win);
	OSWDestroyWindow(&ow->fonts_tab_win);
	OSWDestroyWindow(&ow->display_tab_win);
	OSWDestroyWindow(&ow->paths_tab_win);
	OSWDestroyWindow(&ow->keymap_tab_win);

	OSWDestroyWindow(&ow->toplevel);

	free(ow);
}
