#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "../include/string.h"
#include "../include/disk.h"
#include "../include/osw-x.h"
#include "../include/widget.h"
#include "../include/printerw.h"
#include "../include/objects.h"
#include "../include/unvutil.h"

#include "uew.h"
#include "printdlg.h"
#include "ue.h"
#include "config.h"

#include "../include/oswkeyvals.h"


static float PrintDlgCmToPixels(const float x);
static float PrintDlgInchToPixels(const float x);
static float PrintDlgPixelsToCm(const float x);
static float PrintDlgPixelsToInch(const float x);

void PrintDlgGetPaperSizePixels(
	print_dlg_struct *pw,
	unsigned int *paper_width,
	unsigned int *paper_height
);

static image_t *PrintDlgRotateImage(image_t *img);
static image_t *PrintDlgCreateSpool(print_dlg_struct *pw);
static int PrintDlgPrint(print_dlg_struct *pw);

static void PrintDlgDrawCB(void *widget, void *data);
static void PrintDlgPrintCB(push_button_struct *btn, void *data);
static void PrintDlgCancelCB(push_button_struct *btn, void *data);
static int PrintProgressCB(
        void *data, const unsigned long i, const unsigned long n
);

print_dlg_struct *PrintDlgNew(void *src_uew_ptr);
void PrintDlgDraw(print_dlg_struct *pw, const int amount);
void PrintDlgQueueDraw(print_dlg_struct *pw, const int amount);
int PrintDlgManage(print_dlg_struct *pw, event_t *event);
void PrintDlgMap(print_dlg_struct *pw);
void PrintDlgMapValues(print_dlg_struct *pw);
void PrintDlgUnmap(print_dlg_struct *pw);
void PrintDlgDelete(print_dlg_struct *pw);


#define PRINT_DLG_DEF_PRINT_CMD		PREFIX "/bin/lpr %file"
#define PRINT_DLG_DEF_FILE		"universe.ps"

#define PRINT_DLG_TITLE			"Print"
#define PRINT_DLG_ICON_TITLE		"Print"

#define PRINT_DLG_MARGIN		10

#define PRINT_DLG_WIDTH			520
#define PRINT_DLG_HEIGHT		400

#define PRINT_DLG_BUTTON_WIDTH		70
#define PRINT_DLG_BUTTON_HEIGHT		28

#define PRINT_DLG_PROMPT_HEIGHT		30


/*
 *	Drag Operations:
 */
#define PRINT_DLG_DRAG_OP_NONE		0
#define PRINT_DLG_DRAG_OP_TRANSLATE	1


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)
#define IS_NUM_ODD(n)	((long)(n) & 1)


/*
 *	Unit conversions:
 */
static float PrintDlgCmToPixels(const float x)
{
	return(x * 72 / 2.56);
}

static float PrintDlgInchToPixels(const float x)
{
	return(x * 72);
}

static float PrintDlgPixelsToCm(const float x)
{
	return(x / 72 * 2.54);
}

static float PrintDlgPixelsToInch(const float x)
{
	return(x / 72);
}


/*
 *	Fetches paper size specified on printer window pw.
 */
void PrintDlgGetPaperSizePixels(
	print_dlg_struct *pw,
	unsigned int *paper_width,
	unsigned int *paper_height
)
{
	if((pw == NULL) ||
	   (paper_width == NULL) ||
	   (paper_height == NULL)
	)
	    return;


	switch(pw->paper_size_radio.armed_check_button)
	{
	  /* A4 (21.0 x 29.7 cm), 8.27 x 11.69 inch */
	  case 3:
	    *paper_width = (unsigned int)(PrintDlgInchToPixels(8.26772));
	    *paper_height = (unsigned int)(PrintDlgInchToPixels(11.69291));
	    break;

	  /* Executive, 7.5 x 10 inch */
	  case 2:
	    *paper_width = (unsigned int)(PrintDlgInchToPixels(7.5));
	    *paper_height = (unsigned int)(PrintDlgInchToPixels(10));
	    break;

	  /* Legal, 8.5 x 14 inch */
	  case 1:
	    *paper_width = (unsigned int)(PrintDlgInchToPixels(8.5));
	    *paper_height = (unsigned int)(PrintDlgInchToPixels(14));
	    break;

	  /* Legal, 8.5 by 11 inch */
	  default:
	    *paper_width = (unsigned int)(PrintDlgInchToPixels(8.5));
	    *paper_height = (unsigned int)(PrintDlgInchToPixels(11));
	    break;
	}
}


/*
 *	Creates a duplicate of image rotated 90 degrees
 *	counter-clockwise.
 */
static image_t *PrintDlgRotateImage(image_t *img)
{
	int	bpp,
		tar_bpl, tar_row, tar_col,
		src_bpl, src_row, src_col;
	unsigned int tar_width, tar_height, src_width, src_height;
	u_int8_t *tar, *src;
	image_t *rimg = NULL;

	if(img == NULL)
	    return(NULL);

	bpp = img->bytes_per_pixel;
	src_width = img->width;
	src_height = img->height;
	src_bpl = img->bytes_per_line;

	/* Create new image with swapped width and height */
	if(OSWCreateImage(&rimg, src_height, src_width))
	    return(NULL);

	tar_width = rimg->width;
	tar_height = rimg->height;
	tar_bpl = rimg->bytes_per_line;

	tar = rimg->data;
	src = img->data;

	tar_row = (int)tar_height - 1;
	tar_col = 0;
	src_row = 0;
	src_col = 0;

	/* 8 bits */
	if(bpp == 1)
	{
	    while((src_row < (int)src_height) &&
		  (tar_col < (int)tar_width)
	    )
	    {
		tar[(tar_row * tar_bpl) +
		    (tar_col * bpp)] = src[(src_row * src_bpl) +
					(src_col * bpp)];
		src_col++;
		tar_row--;
		if(src_col >= (int)src_width)
		{
		    src_row++;
		    src_col = 0;
		}
		if(tar_row < 0)
		{
		    tar_col++;
		    tar_row = (int)tar_height - 1;
		}
	    }
	}
	else if(bpp == 2)
	{
	    while((src_row < (int)src_height) &&
		  (tar_col < (int)tar_width)
	    )
	    {
		*(u_int16_t *)(tar +
		    (tar_row * tar_bpl) +
		    (tar_col * bpp)
		) = *(u_int16_t *)(src +
		    (src_row * src_bpl) +
		    (src_col * bpp)
		);
		src_col++;
		tar_row--;
		if(src_col >= (int)src_width)
		{
		    src_row++;
		    src_col = 0;
		}
		if(tar_row < 0)
		{
		    tar_col++;
		    tar_row = (int)tar_height - 1;
		}  
	    }
	}
	else if(bpp == 3)
	{
	    u_int8_t *tar_ptr8, *src_ptr8;

	    while((src_row < (int)src_height) &&
		  (tar_col < (int)tar_width)
	    )
	    {
		tar_ptr8 = tar +
		    (tar_row * tar_bpl) +
		    (tar_col * bpp);
		src_ptr8 = src +
		    (src_row * src_bpl) +
		    (src_col * bpp);
		*tar_ptr8++ = *src_ptr8++;
                *tar_ptr8++ = *src_ptr8++;
                *tar_ptr8 = *src_ptr8;
		src_col++;
		tar_row--;
		if(src_col >= (int)src_width)
		{
		    src_row++;
		    src_col = 0;
		}
		if(tar_row < 0)
		{
		    tar_col++;
		    tar_row = (int)tar_height - 1;
		}
	    }
	}
	else if(bpp == 4)
	{
	    while((src_row < (int)src_height) &&
		  (tar_col < (int)tar_width)
	    )
	    {
	        *(u_int32_t *)(tar +
		    (tar_row * tar_bpl) +
		    (tar_col * bpp)
		) = *(u_int32_t *)(src +
		    (src_row * src_bpl) +
		    (src_col * bpp)
		);
		src_col++;
		tar_row--;
		if(src_col >= (int)src_width)
		{
		    src_row++;
		    src_col = 0;
		}
		if(tar_row < 0)
		{
		    tar_col++;
		    tar_row = (int)tar_height - 1;
		}  
	    }
	}

	return(rimg);
}


/*
 *	Creates a printable image based on the specified print
 *	window's referenced universe editor's universe.
 *
 *	Note that this function's drawing procedure should be (nearly)
 *	identical to UEWViewDraw().
 */
static image_t *PrintDlgCreateSpool(print_dlg_struct *pw)
{
	bool_t draw_greyscale;
	unsigned int width, height;
	win_t w;
	pixmap_t pixmap;
	image_t *new_img;
	uew_struct *uew;
	ue_option_struct *opt = &option;

	if(pw == NULL)
	    return(NULL);

	draw_greyscale = (RadioGetActive(&pw->color_mode_radio) == 1) ? False : True;

	uew = UEW(pw->src);
	if(uew == NULL)
	    return(NULL);

        w = uew->view;
        if(w == None)
        {
            UEWStatus(
                uew,  
"Unable to obtain the view from the universe editor."
            );
            return(NULL);
        }

	OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	if((width == 0) || (height == 0))
	{
            UEWStatus(
                uew,
"Universe editor's view is too small to render an image from."
            );
            return(NULL);
	}

        if(IS_NUM_ODD(width))
            width += 1;
        if(IS_NUM_ODD(height))
            height += 1;

	/* Create a pixmap the same size as the image */
	if(OSWCreatePixmap(&pixmap, width, height))
	{
	    UEWStatus(
		uew,
"Unable to create pixmap."
	    );
	    return(NULL);
	}

	/* Update options */
	opt->show_object_labels = CBtnGetActive(&pw->obj_labels_cb);
	opt->show_geometry_labels = CBtnGetActive(&pw->label_geometry_cb);

	/* Draw the view on to the pixmap */
	UEWViewDrawPixmap(
	    uew, pixmap,
	    None,			/* No view window */
	    True,			/* Draw printer friendly */
	    draw_greyscale
	);

	/* Get new image from pixmap */
	new_img = OSWGetImage(
	    pixmap,
	    0, 0, width, height
	);
	if(new_img == NULL)
	    UEWStatus(
		uew,
"Unable to obtain the rendered view contents."
	    );

	OSWDestroyPixmap(&pixmap);

	return(new_img);
}


/*
 *	Print.
 */
static int PrintDlgPrint(print_dlg_struct *pw)
{
	unsigned int	width, height,
			paper_width, paper_height;
	win_t toplevel;
	image_t *img, *rs_img;
	printer_parm_struct parm;
	uew_struct *uew;

	if(pw == NULL)
	    return(-2);

	toplevel = pw->toplevel;
	uew = UEW(pw->src);
	if(uew == NULL)
	{
	    UEDialog(
		"Print Failed",
		UE_ICON_WARNING,
"There is no universe associated with this\n\
print dialog.",
		toplevel
	    );
	    return(-2);
	}

	/* Print size may not be zero */
	if((PromptGetF(&pw->width_prompt) <= 0) ||
	   (PromptGetF(&pw->height_prompt) <= 0)
	)
	{
            UEDialog(
                "Print Failed",
                UE_ICON_WARNING,
"The paper size is too small.",
                toplevel
            );
	    return(-2);
	}

	/* Get the paper size in pixels */
	PrintDlgGetPaperSizePixels(pw, &paper_width, &paper_height);


	/* Set up the print parameters */

	parm.options = 0;
	parm.title[0] = '\0';
	parm.header[0] = '\0';
	parm.footer[0] = '\0';
	parm.pages = 1;

	/* Color Mode */
	switch(RadioGetActive(&pw->color_mode_radio))
	{
	  case 1:
	    parm.color_mode = PrinterColorModeColor;
	    break;
	  default:
	    parm.color_mode = PrinterColorModeGreyScale;
	    break;
	}

	/* Units */
	switch(pw->units)
	{
	  case PRINT_DLG_UNITS_CM:
	    parm.units = PrinterUnitCentimeters;
	    break;
	  case PRINT_DLG_UNITS_INCH:
	    parm.units = PrinterUnitInches;
	    break;
	  default:	/* PRINT_DLG_UNITS_PIXELS */
	    parm.units = PrinterUnitPixels;
	    break;
	}

	/* Get x and y offsets from upper left corner (no unit conv) */
	parm.x = PromptGetF(&pw->x_prompt);
	parm.y = PromptGetF(&pw->y_prompt);

	/* Set paper size */
	switch(pw->units)
	{
	  case PRINT_DLG_UNITS_CM:
	    parm.width = PrintDlgPixelsToCm(paper_width);
	    parm.height = PrintDlgPixelsToCm(paper_height);
	    break;
	  case PRINT_DLG_UNITS_INCH:
	    parm.width = PrintDlgPixelsToInch(paper_width);
	    parm.height = PrintDlgPixelsToInch(paper_height);
	    break;
	  default:	/* PRINT_DLG_UNITS_PIXELS */
	    parm.width = (float)paper_width;
	    parm.height = (float)paper_height;
	    break;
	}

	/* Destination */
	switch(RadioGetActive(&pw->print_to_radio))
	{
	  case 1:
	    parm.dest = PrinterDestinationFile;
	    break;
	  default:
	    parm.dest = PrinterDestinationPrinter;
	    break;
	}

	/* Create the spool image */
	img = PrintDlgCreateSpool(pw);
	if(img == NULL)
	{
            UEDialog(
                "Print Failed",
                UE_ICON_WARNING,
"Unable to render an image of the universe for\n\
printing.",
                toplevel
            );
	    return(-1);
	}

	/* Resize the spool image */
	switch(pw->units)
	{
	  case PRINT_DLG_UNITS_CM:
	    width = (unsigned int)(PrintDlgCmToPixels(
		PromptGetF(&pw->width_prompt)
	    ));
	    height = (unsigned int)(PrintDlgCmToPixels(
		PromptGetF(&pw->height_prompt)
	    ));
	    break;
	  case PRINT_DLG_UNITS_INCH:
	    width = (unsigned int)(PrintDlgInchToPixels(
		PromptGetF(&pw->width_prompt)
	    ));
	    height = (unsigned int)(PrintDlgInchToPixels(
		PromptGetF(&pw->height_prompt)
	    ));
	    break;
	  default:
	    width = (unsigned int)(PromptGetI(&pw->width_prompt));
	    height = (unsigned int)(PromptGetF(&pw->height_prompt));
	    break;
	}
	if(OSWCreateImage(&rs_img, width, height))
	{
            UEDialog(
                "Print Failed",
		UE_ICON_WARNING,
"Unable to create a resized image of the\n\
universe.",
                toplevel
            );      
	    OSWDestroyImage(&img);
	    return(-3);
	}
	WidgetResizeImageBuffer(
	    img->bits_per_pixel,
	    rs_img->data,			/* Target */
	    img->data,				/* Source */
	    rs_img->width, rs_img->height, rs_img->bytes_per_line,
	    img->width, img->height, img->bytes_per_line
	);
	OSWDestroyImage(&img);
	img = rs_img;

	/* Rotate spool image if orientation is set to landscape */
	if(RadioGetActive(&pw->orientation_radio) == 1)
	{
	    rs_img = PrintDlgRotateImage(img);
	    if(rs_img == NULL)
	    {
                UEDialog(
                    "Print Failed",
                    UE_ICON_WARNING,
"Unable to rotate the rendered image of the\n\
universe.",
		    toplevel
		);      
		OSWDestroyImage(&img);
		return(-1);
	    }
	    OSWDestroyImage(&img);
	    img = rs_img;
	}

	/* Check if the image is out of bounds of the size of the paper */
	switch(pw->units)
	{
	  case PRINT_DLG_UNITS_CM:
	    width = (int)(PrintDlgPixelsToCm((int)img->width));
	    height = (int)(PrintDlgPixelsToCm((int)img->height));
	    break;
	  case PRINT_DLG_UNITS_INCH:
	    width = (int)(PrintDlgPixelsToInch((int)img->width));
	    height = (int)(PrintDlgPixelsToInch((int)img->height));
	    break;
	  default:
	    width = img->width;
	    height = img->height;
	    break;
	}
	if(((int)parm.x < 0) ||
	   ((int)parm.y < 0) ||   
	   (((int)parm.x + (int)width) > (int)parm.width) ||
	   (((int)parm.y + (int)height) > (int)parm.height)
	)
	{
	    int response = UEGetResponse(
		"Confirm Print",
		UE_ICON_WARNING,
"The image geometry may be partially or entirly\n\
off the bounds of the paper.\n\
\n\
Do you want to continue printing?",
		CDlgBtnYes | CDlgBtnNo,
		pw->toplevel
	    );
	    if(response == CDlgNo)
	    {
		OSWDestroyImage(&img);
		return(-4);
	    }
	}
	 

	/* Print to file? */
	if(RadioGetActive(&pw->print_to_radio) == 1)
	{
	    /* Print to file */
	    int status;
	    char *path = STRDUP(PromptGetS(&pw->print_file_prompt));

	    if(STRISEMPTY(path))
	    {
                UEDialog(
                    "Print Failed",
                    UE_ICON_WARNING,
"There is no output file specified.",
                    toplevel
                );
		free(path);
		OSWDestroyImage(&img);
		return(-2);
	    }

	    /* Complete the path as needed */
	    if(!ISPATHABSOLUTE(path))
	    {
		char *full_path = STRDUP(PrefixPaths(
		    UECurrentDirectory(), path
		));
		free(path);
		path = full_path;
	    }
	    if(STRISEMPTY(path))
	    {
                UEDialog(
                    "Print Failed",
                    UE_ICON_WARNING,
"Unable to complete the path of the output file.",
                    toplevel
                );
		free(path);
		OSWDestroyImage(&img);
		return(-2);
	    }


	    /* Check if the file already exists */
	    if(UECheckFileExists(path, NULL))
	    {
		/* Confirm overwrite */
		int response;
		char *s = (char *)malloc(
		    STRLEN(path) + (2 * 80)
		);
		sprintf(
		    s,
"Overwrite existing file:\n\
\n\
    %s",
		    path
		);
		response = UEGetResponse(
		    "Confirm Overwrite",
		    UE_ICON_WARNING,
		    s,
		    CDlgBtnYes | CDlgBtnNo,
		    pw->toplevel
		);
		free(s);
		if(response != CDlgYes)
		{
		    free(path);
		    OSWDestroyImage(&img);
		    return(-4);
		}
		unlink(path);
	    }

	    /* Print to file */
	    status = PrinterPrintImage(
		img, path,
		PromptGetS(&pw->print_cmd_prompt),
		&parm,
		PrintProgressCB, pw
	    );
	    UEWStatusProgress(uew, 0.0f);
	    if(status == PrinterSuccess)
	    {
		char *s = (char *)malloc(STRLEN(path) + 80);
		sprintf(
		    s,
"Printed to file `%s'.",
		    path
		);
		UEWStatus(uew, s);
		free(s);
	    }
	    else
	    {
		char *s = (char *)malloc(STRLEN(path) + (2 * 80));
		sprintf(
		    s,
"An error occured while printing the universe to:\n\
\n\
    %s",
		    path
		);
		UEWStatus(uew, "Print failed.");
		UEDialog(
		    "Print Failed",
		    UE_ICON_ERROR,
		    s,
		    pw->toplevel
		);
		free(s);
		free(path);
		OSWDestroyImage(&img);
		return(-1);
	    }

	    free(path);
	}
	else
	{
	    /* Print to printer */
	    int fd, status;
	    char	*path,
			*tmp_dir = STRDUP(getenv("TMPDIR"));
	    if(tmp_dir == NULL)
#ifdef P_tmpdir
		tmp_dir = STRDUP(P_tmpdir);
#else
		tmp_dur = STRDUP("/tmp");
#endif

	    /* Generate a tempory file name */
	    path = STRDUP(PrefixPaths(tmp_dir, PROG_NAME "_XXXXXX"));
	    free(tmp_dir);
	    fd = mkstemp(path);
	    if(fd < 0)
	    {
		char *s = (char *)malloc(
		    STRLEN(path) + (2 * 80)
		);
		sprintf(
		    s,
"Unable to create tempory file:\n\
\n\
    %s",
		    path
		);
		UEWStatus(uew, "Print failed.");
		UEDialog(
		    "Print Failed",
		    UE_ICON_ERROR,
		    s,
		    pw->toplevel
		);
		free(s);
		OSWDestroyImage(&img);
		free(path);
		return(-1);
	    }
	    close(fd);

	    /* Print */
	    status = PrinterPrintImage(
		img, path,
		PromptGetS(&pw->print_cmd_prompt),
		&parm,
		PrintProgressCB, pw
	    );
            UEWStatusProgress(uew, 0.0f);
	    if(status == PrinterSuccess)
	    {
		UEWStatus(uew, "Printed universe.");
	    }
	    else
	    {
		UEWStatus(uew, "Print failed.");
		UEDialog(
		    "Print Failed",
		    UE_ICON_ERROR,
"An error occured while printing the universe.",
		    pw->toplevel
		);
		OSWDestroyImage(&img);
		free(path);
		return(-1);
	    }

	    free(path);
	}

	/* Delete the spool image */
	OSWDestroyImage(&img);

	return(0);
}


/*
 *	Draw callback.
 */
static void PrintDlgDrawCB(void *widget, void *data)
{
	const int amount = (int)data;
	print_dlg_struct *pw = PRINT_DLG(widget);
	if(pw == NULL)
	    return;

	if(pw->map_state)
	    PrintDlgDraw(pw, amount);
}

/*
 *	Print button callback.
 */
static void PrintDlgPrintCB(push_button_struct *btn, void *data)
{
	print_dlg_struct *pw = PRINT_DLG(data);
	if(pw == NULL)
	    return;

	if(!PrintDlgPrint(pw))
	    PrintDlgUnmap(pw);
}

/*
 *	Cancel button callback.
 */
static void PrintDlgCancelCB(push_button_struct *btn, void *data)
{
	print_dlg_struct *pw = PRINT_DLG(data);
	if(pw == NULL)
	    return;

	PrintDlgUnmap(pw);
}

/*
 *	Print progress callback.
 */
static int PrintProgressCB(
	void *data, const unsigned long i, const unsigned long n
)
{
        print_dlg_struct *pw = PRINT_DLG(data);
        if(pw == NULL)
            return(0);

	UEWStatusProgress(
	    UEW(pw->src),
	    (n > 0) ? ((float)i / (float)n) : 0.0f
	);

	return(1);	/* Keep processing */
}


/*
 *	Creates a print dialog.
 */
print_dlg_struct *PrintDlgNew(void *src_uew_ptr)
{
	int x, y;
	unsigned int width, height;
	char *strarray[10];
	keyval_t accel_keys[4];
	key_state_t accel_mods[4];
        osw_gui_struct *gui = &osw_gui[0];
	win_t w, parent = gui->root_win;
	push_button_struct *btn;
	prompt_struct *prompt;
	popup_list_struct *pulist;
	radio_struct *radio;
        widget_global_struct *wglobal = &widget_global;
	uew_struct *uew = UEW(src_uew_ptr);
	print_dlg_struct *pw = PRINT_DLG(calloc(1, sizeof(print_dlg_struct)));

	pw->map_state = False;
	pw->has_focus = False;
	pw->x = 0;
	pw->y = 0;
	pw->width = PRINT_DLG_WIDTH;
	pw->height = PRINT_DLG_HEIGHT;

	pw->preview_buf = None;
	pw->preview_img = NULL;

	pw->units = PRINT_DLG_UNITS_INCH;

	pw->src = src_uew_ptr;

	pw->drag_op = PRINT_DLG_DRAG_OP_NONE;
	pw->button_pressed = 0;
	pw->button_last_x = 0;
	pw->button_last_y = 0;
	pw->button_pressed_time = 0l;


	/* Create toplevel */
	x = pw->x;
	y = pw->y;
	width = pw->width;
	height = pw->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	pw->toplevel = w;
	OSWSetWindowWMProperties(
	    w,
	    PRINT_DLG_TITLE ": Untitled",
	    PRINT_DLG_ICON_TITLE,
	    wglobal->std_icon_pm,
	    False,			/* Let WM set coordinates? */
	    x, y,
	    width, height,
	    width, height,
	    OSW_WINDOW_FRAME_DIALOG,
	    0, NULL
	);
	OSWSetWindowInput(w, OSW_EVENTMASK_TOPLEVEL);
	if(uew != NULL)
	    OSWSetTransientFor(w, uew->toplevel);

	parent = w;

	/* Create preview window */
	OSWCreateWindow(
	    &w, parent,
	    10, 10, 10, 10
	);
	pw->preview = w;
	OSWSetWindowInput(
	    w,
	    ExposureMask | ButtonPressMask | ButtonReleaseMask |
		PointerMotionMask
	);

	/* Load paper pixel colors */
	OSWLoadPixelRGB(
	    &pw->paper_bg_pix,
	    0xff,
	    0xff,
	    0xff
	);
	OSWLoadPixelRGB(
	    &pw->paper_fg_pix,
	    0x00,
	    0x00,
	    0x00
	);



	/* Print To */
	radio = &pw->print_to_radio;
	strarray[0] = "Printer";
	strarray[1] = "File";
	strarray[2] = NULL;
	x = 105;
	y = PRINT_DLG_MARGIN;
	RadioInit(
	    radio, parent,
	    x, y,
	    2,
	    0,		/* Initially selected button */
	    strarray,
	    2,
	    RADIO_ALIGN_HORIZONTAL
	);

	/* Print command prompt */
	prompt = &pw->print_cmd_prompt;
	x = PRINT_DLG_MARGIN;
	y = 35;
	width = (unsigned int)MAX(
	    ((int)pw->width - (2 * PRINT_DLG_MARGIN)),
	    1
	);
	height = PRINT_DLG_PROMPT_HEIGHT;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Print Command:",
	    PATH_MAX + NAME_MAX, 0
	);
	PromptSetS(prompt, PRINT_DLG_DEF_PRINT_CMD);

	/* Print File */
	prompt = &pw->print_file_prompt;
	y += (int)height;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "File:",
	    PATH_MAX + NAME_MAX, 0
	);
	PromptSetS(prompt, PRINT_DLG_DEF_FILE);


	/* Amount of objects */
	pulist = &pw->objects_pulist;
	x = 2 * PRINT_DLG_MARGIN;
	y = 130;
	width = (unsigned int)MAX(
	    (((int)pw->width / 3) - (5 * PRINT_DLG_MARGIN)),
	    1
	);
	height = 0;
	PUListInit(
	    pulist, parent,
	    x, y, width, height,
	    3,			/* Total items visible */
	    PULIST_POPUP_BELOW
	);
	PUListAddItem(
	    pulist,
	    0,
	    "All Objects",
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    "Selected Objects",
	    None, None
	);
	PUListAddItem(
	    pulist,
	    0,
	    "Astronomical Objects",
	    None, None
	);
	height = pulist->height;

	/* Paper position prompts */
	prompt = &pw->x_prompt;
	x = PRINT_DLG_MARGIN;
	y += (int)height + (PRINT_DLG_MARGIN / 2);
	width = 160;
	height = PRINT_DLG_PROMPT_HEIGHT;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "        X:",
	    40, 0
	);
	PromptSetI(prompt, 0);

	prompt = &pw->y_prompt;
	y += (int)height;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "        Y:",
	    40, 0
	);
	PromptSetI(prompt, 0);

	prompt = &pw->width_prompt;
	y += (int)height + (PRINT_DLG_MARGIN / 2);
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    " Width:",
	    40, 0
	);
	PromptSetI(prompt, 1);

	prompt = &pw->height_prompt;
	y += (int)height;
	PromptInit(
	    prompt, parent,
	    x, y, width, height,
	    PROMPT_STYLE_FLUSHED,
	    "Height:",
	    40, 0
	);
	PromptSetI(&pw->height_prompt, 1);

	/* Label Objects */
	y += (int)height + (PRINT_DLG_MARGIN / 2);
	CBtnInit(
	    &pw->obj_labels_cb, parent,
	    x, y,
	    option.show_object_labels,
	    "Label Objects"
	);
	height = pw->obj_labels_cb.height;

	/* Label Geometry */
	y += (int)height + (PRINT_DLG_MARGIN / 2);
	CBtnInit(
	    &pw->label_geometry_cb, parent,
	    x, y,
	    option.show_geometry_labels,
	    "Label Geometry"
	);


	/* Orientation */
	radio = &pw->orientation_radio;
	strarray[0] = "Portrait";
	strarray[1] = "Landscape";
	strarray[2] = NULL;
	x = ((int)pw->width / 3 * 2) + 15;
	y = 130;
	RadioInit(
	    radio, parent,
	    x, y,
	    2,
	    0,			/* Initially selected button */
	    strarray,
	    2,
	    RADIO_ALIGN_VERTICAL
	);

	/* Color mode */
	radio = &pw->color_mode_radio;
	strarray[0] = "GreyScale";
	strarray[1] = "Color"; 
	strarray[2] = NULL;
	y = 195;
	RadioInit(
	    radio, parent,
	    x, y,
	    2,
	    0,			/* Initially selected button */
	    strarray,
	    2,
	    RADIO_ALIGN_VERTICAL
	);

	/* Paper size */
	radio = &pw->paper_size_radio;
	strarray[0] = "Letter (8.5 x 11 in)";
	strarray[1] = "Legal (8.5 x 14 in)";
	strarray[2] = "Executive (7.5 x 10 in)";
	strarray[3] = "A4 (21.0 x 29.7 cm)";
	strarray[4] = NULL;
	y = 260;
	RadioInit(
	    radio, parent,
	    x, y,
	    4,
	    0,			/* Initially selected button */
	    strarray,
	    4,
	    RADIO_ALIGN_VERTICAL
	);
	height = radio->height;


	/* Print button */
	btn = &pw->print_btn;
	width = PRINT_DLG_BUTTON_WIDTH;
	height = PRINT_DLG_BUTTON_HEIGHT;
	x = PRINT_DLG_MARGIN;
	y = (int)pw->height - PRINT_DLG_MARGIN - (int)height;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Print",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    PrintDlgPrintCB, pw
	);
	accel_keys[0] = OSW_KEY_p;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_Return;
	accel_mods[1] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 2);

	/* Cancel button */
	btn = &pw->cancel_btn;
	x = (int)pw->width - PRINT_DLG_MARGIN - (int)width;
	PBtnInit(
	    btn, parent,
	    x, y, width, height,
	    "Cancel",
	    PBTN_LABEL_ALIGN_CENTER,
	    None, None,
	    PrintDlgCancelCB, pw
	);
	accel_keys[0] = OSW_KEY_c;
	accel_mods[0] = OSW_KEY_STATE_CTRL;
	accel_keys[1] = OSW_KEY_w;
	accel_mods[1] = OSW_KEY_STATE_CTRL;
	accel_keys[2] = OSW_KEY_Escape;
	accel_mods[2] = 0;
	PBtnSetAccelKeys(btn, accel_keys, accel_mods, 3);

	return(pw);
}

/*
 *	Draws the print dialog.
 */
void PrintDlgDraw(print_dlg_struct *pw, const int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(pw == NULL)
	    return;

	/* Toplevel */
	if(amount == PRINT_DLG_DRAW_AMOUNT_COMPLETE)
	{
	    int y;
	    unsigned int wwidth, wheight;
	    font_t *font = wglobal->std_font;
	    win_t w = pw->toplevel;

	    OSWGetWindowGeometry(w, NULL, NULL, &wwidth, &wheight, NULL);
	     
	    /* Redraw background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(w, 0, 0, wwidth, wheight);
		OSWSetForeground(gui->white_pix);
	    }
	    else
	    {
		WidgetPutImageTile( 
		    w, wglobal->std_bkg_img,
		    wwidth, wheight
		);


		/* Draw upper HR */
		OSWSetForeground(wglobal->surface_shadow_pix);
		y = 100;
		OSWDrawLine(w, 0, y, (int)wwidth, y);

		OSWSetForeground(wglobal->surface_highlight_pix);
		y += 1;
		OSWDrawLine(w, 0, y, (int)wwidth, y);


		/* Draw lower HR */
		OSWSetForeground(wglobal->surface_shadow_pix); 
		y = (int)wheight - PRINT_DLG_MARGIN -
		    PRINT_DLG_BUTTON_HEIGHT - PRINT_DLG_MARGIN;
		OSWDrawLine(w, 0, y, (int)wwidth, y);

		OSWSetForeground(wglobal->surface_highlight_pix);
		y += 1;
		OSWDrawLine(w, 0, y, (int)wwidth, y);

		OSWSetForeground(wglobal->normal_text_pix);
	    }

	    OSWSetFont(font);
	    OSWDrawString(
		w,
		55, 23,
		"Print To:"
	    );
	    OSWDrawString(
		w,
		14, 123,
		"Print:"
	    );
	    OSWDrawString(
		w, 
		((int)pw->width / 3 * 2) + 5, 123,
		"Orientation:"
	    );
	    OSWDrawString(
		w,
		((int)pw->width / 3 * 2) + 5, 190,
		"Color Mode:"
	    );
	    OSWDrawString(
		w,
		((int)pw->width / 3 * 2) + 5, 256,
		"Paper Size:"
	    );
	}

	/* Preview */
	if((amount == PRINT_DLG_DRAW_AMOUNT_COMPLETE) ||
	   (amount == PRINT_DLG_DRAW_AMOUNT_PREVIEW)
	)
	{
	    int i, x, y;
	    unsigned int	wwidth, wheight, pv_width, pv_height,
				paper_width, paper_height;
	    float coeff, width_coeff, height_coeff;
	    win_t w = pw->preview;
	    pixmap_t pixmap;

	    /* Get the size of the paper */
	    PrintDlgGetPaperSizePixels(pw, &paper_width, &paper_height);

	    /* Calculate the size of the preview */
	    coeff = (paper_width > 0) ?
		((float)paper_height / (float)paper_width) : 0.0f;

	    wwidth = (unsigned int)MAX(
		(((int)pw->width / 3) - (2 * PRINT_DLG_MARGIN)),
		1
	    );
	    wheight = (unsigned int)MAX(
		((int)wwidth * coeff),
		1
	    );

	    i = (int)pw->height - 120 - PRINT_DLG_BUTTON_HEIGHT - 40;
	    if((int)wheight > i)
	    {
		wheight = i;
		if(coeff > 0.0f)
		    wwidth = (unsigned int)(MAX(
			((int)wheight / coeff), 1
		    ));
	    }

	    /* Calculate the position of the preview */
	    x = ((int)pw->width / 2) - ((int)wwidth / 2);
	    y = 120 +
		(((int)pw->height - 120 - PRINT_DLG_BUTTON_HEIGHT -
		    (4 * PRINT_DLG_MARGIN)) / 2) -
 		((int)wheight / 2);

	    /* Move and resize the preview window */
	    OSWMoveResizeWindow(w, x, y, wwidth, wheight);

            /* Draw a shadow on the parent window */
	    if(pw->toplevel != None)
	    {
		win_t w = pw->toplevel;
		OSWSetForeground(
		    wglobal->force_mono ?
                        gui->black_pix :
                        pw->paper_fg_pix
                );
                OSWDrawSolidRectangle(
                    w,
                    x + (int)wwidth, y + 5,
		    5, wheight
                );
                OSWDrawSolidRectangle(
                    w,
                    x + 5, y + (int)wheight,
                    wwidth, 5
                );
	    }

	    /* Resize or create the preview pixmap as needed */
	    if(pw->preview_buf != None)
	    {
		unsigned int prev_width, prev_height;
		OSWGetPixmapGeometry(
		    pw->preview_buf,
		    &prev_width, &prev_height, NULL
		);
		if((prev_width != wwidth) || (prev_height != wheight))
		{
		    OSWDestroyPixmap(&pw->preview_buf);
		    OSWCreatePixmap(&pw->preview_buf, wwidth, wheight);
		}
	    }
	    else
	    {
		OSWCreatePixmap(&pw->preview_buf, wwidth, wheight);
	    }
	    pixmap = pw->preview_buf;

	    /* Draw the background */
	    OSWClearPixmap(
		pixmap, wwidth, wheight,
		wglobal->force_mono ?
		    gui->white_pix :
		    pw->paper_bg_pix
	    );

	    /* Draw the outline of the printed area */
	    if(RadioGetActive(&pw->orientation_radio) == 1)
	    {
		/* Landscape */

		switch(pw->units)
		{
		  case PRINT_DLG_UNITS_CM:
		    x = (int)(PrintDlgCmToPixels(
			PromptGetF(&pw->x_prompt)
		    ) * (float)wwidth / (float)paper_width);
		    y = (int)(PrintDlgCmToPixels(
			PromptGetF(&pw->y_prompt)
		    ) * (float)wheight / (float)paper_height);

		    width_coeff = PrintDlgCmToPixels(
			PromptGetF(&pw->height_prompt)) / (float)paper_width;
		    height_coeff = PrintDlgCmToPixels(
			PromptGetF(&pw->width_prompt)) / (float)paper_height;

		    pv_width = (unsigned int)(width_coeff * (float)wwidth);
		    pv_height = (unsigned int)(height_coeff * (float)wheight);
		    break;

		  case PRINT_DLG_UNITS_INCH:
		    x = (int)(PrintDlgInchToPixels(
			PromptGetF(&pw->x_prompt)
			) * (float)wwidth / (float)paper_width);
		    y = (int)(PrintDlgInchToPixels(
			PromptGetF(&pw->y_prompt)
			) * (float)wheight / (float)paper_height);
 
		    width_coeff = PrintDlgInchToPixels(
			PromptGetF(&pw->height_prompt)) / (float)paper_width;
		    height_coeff = PrintDlgInchToPixels(
			PromptGetF(&pw->width_prompt)) / (float)paper_height;

		    pv_width = (unsigned int)(width_coeff * (float)wwidth);
		    pv_height = (unsigned int)(height_coeff * (float)wheight);
		    break;

		  default:	/* PRINT_DLG_UNITS_PIXELS */
		    x = (int)(PromptGetF(&pw->x_prompt) * (float)wwidth /
			(float)paper_width);
		    y = (int)(PromptGetF(&pw->y_prompt) * (float)wheight /
			(float)paper_height);

		    width_coeff = PromptGetF(&pw->height_prompt) / (float)paper_width;
		    height_coeff = PromptGetF(&pw->width_prompt) / (float)paper_height;

		    pv_width = (unsigned int)(width_coeff * (float)wwidth);
		    pv_height = (unsigned int)(height_coeff * (float)wheight);
		    break;
		}
	    }
	    else
	    {
		/* Portrait */

		switch(pw->units)
		{
		  case PRINT_DLG_UNITS_CM:
		    x = (int)(PrintDlgCmToPixels(
			PromptGetF(&pw->x_prompt)
			) * (float)wwidth / (float)paper_width);
		    y = (int)(PrintDlgCmToPixels(
			PromptGetF(&pw->y_prompt)
			) * (float)wheight / (float)paper_height);

		    width_coeff = PrintDlgCmToPixels(
			PromptGetF(&pw->width_prompt)) / (float)paper_width;
		    height_coeff = PrintDlgCmToPixels(
			PromptGetF(&pw->height_prompt)) / (float)paper_height;

		    pv_width = (unsigned int)(width_coeff * (float)wwidth); 
		    pv_height = (unsigned int)(height_coeff * (float)wheight);
		    break;

		  case PRINT_DLG_UNITS_INCH:
		    x = (int)(PrintDlgInchToPixels(
			PromptGetF(&pw->x_prompt)
			) * (float)wwidth / (float)paper_width);
		    y = (int)(PrintDlgInchToPixels(
			PromptGetF(&pw->y_prompt)
			) * (float)wheight / (float)paper_height);

		    width_coeff = PrintDlgInchToPixels(
			PromptGetF(&pw->width_prompt)) / (float)paper_width;
		    height_coeff = PrintDlgInchToPixels(
			PromptGetF(&pw->height_prompt)) / (float)paper_height;

		    pv_width = (unsigned int)(width_coeff * (float)wwidth);
		    pv_height = (unsigned int)(height_coeff * (float)wheight);
		    break;

		  default:      /* PRINT_DLG_UNITS_PIXELS */
		    x = (int)(PromptGetF(&pw->x_prompt) * (float)wwidth /
			(float)paper_width);
		    y = (int)(PromptGetF(&pw->y_prompt) * (float)wheight /
			(float)paper_height);

		    width_coeff = PromptGetF(&pw->width_prompt) / (float)paper_width;
		    height_coeff = PromptGetF(&pw->height_prompt) / (float)paper_height;

		    pv_width = (unsigned int)(width_coeff * (float)wwidth);
		    pv_height = (unsigned int)(height_coeff * (float)wheight);
		    break;
		}
	    }

	    /* Draw the preview contents */
	    if((pv_width > 0) && (pv_height > 0))
	    {
		/* Preview image has not been created yet? */
		if(pw->preview_img == NULL)
		{
		    /* Create the image */
		    image_t *img = PrintDlgCreateSpool(pw);

		    /* Rotate the image as needed */
		    if((img != NULL) &&
		       (RadioGetActive(&pw->orientation_radio) == 1)
		    )
		    {
			image_t *rimg = PrintDlgRotateImage(img);
			if(rimg != NULL)
			{
			    OSWDestroyImage(&img);
			    img = rimg;
			}
		    }

		    /* Resize the image */
		    if(img != NULL)
		    {
			image_t *rsimg;
			OSWCreateImage(&rsimg, pv_width, pv_height);
			if(rsimg != NULL)
			{
			    WidgetResizeImageBuffer(
				img->bits_per_pixel,
				rsimg->data, img->data,
				rsimg->width, rsimg->height,
				rsimg->bytes_per_line,
				img->width, img->height,
				img->bytes_per_line
			    );
			    OSWDestroyImage(&img);
			    img = rsimg;
			}
		    }

		    pw->preview_img = img;
		}

		/* Got a rendered preview image? */
		if(pw->preview_img != NULL)
		{
		    OSWPutImageToDrawablePos(
			pw->preview_img, pixmap, x, y
		    );
		}
		else
		{
		    /* Failed to obtain a rendered preview image */
		    int x2, y2,
			lbearing, rbearing, swidth, ascent, descent,
			font_height;
		    const char *s = "NO PREVIEW";
		    font_t *font = wglobal->std_font;

                    OSWSetForeground(
                        wglobal->force_mono ?
                            gui->black_pix :
                            pw->paper_fg_pix
                    );
		    OSWSetFont(font);
		    OSWGetTextBounds(
			font, s, -1,
			&lbearing, &rbearing, &swidth, &ascent, &descent
		    );
		    font_height = ascent + descent;
		    x2 = x + (((int)pv_width - swidth) / 2);
		    y2 = y + (((int)pv_height - font_height) / 2);
		    OSWDrawString(
			pixmap,
			x2 - lbearing,
			y2 + ascent,
			s
		    );
		}

		/* Draw the content outline */
	        OSWSetForeground(
		    wglobal->force_mono ?
			gui->black_pix :
			pw->paper_fg_pix
	        );
	        OSWDrawRectangle(
		    pixmap,
		    x, y, pv_width - 1, pv_height - 1
	        );
	    }

	    /* Draw a frame around the paper */
            OSWSetForeground(
                wglobal->force_mono ?
                    gui->black_pix :
                    pw->paper_fg_pix
            );
            OSWDrawRectangle(
                pixmap,
                0, 0, wwidth - 1, wheight - 1
	    );

	    OSWPutDrawableToWindow(w, pixmap);
	}
}

/*
 *	Queues the print dialog to draw.
 */
void PrintDlgQueueDraw(print_dlg_struct *pw, const int amount)
{
	if(pw == NULL)
	    return;

	WidgetQueueDrawAction(   
	    PrintDlgDrawCB, pw, (void *)amount
	);
}

/*
 *	Manages the print dialog.
 */
int PrintDlgManage(print_dlg_struct *pw, event_t *event)
{
	int events_handled = 0;
	time_t t;
	win_t w;
	event_button_t *button;

	if((pw == NULL) || (event == NULL))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case Expose:
	    if(w == pw->toplevel)
	    {
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_COMPLETE);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    if(!pw->has_focus)
		return(events_handled);

	    break;

	  case KeyRelease:
	    if(!pw->has_focus)
		return(events_handled);

	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    if(w == pw->preview)
	    {
		WCursor *wcur = ue_cursor.translate_wcur;
		pw->drag_op = PRINT_DLG_DRAG_OP_TRANSLATE;
                pw->button_pressed = button->button;
		pw->button_last_x = button->x;
		pw->button_last_y = button->y;
                pw->button_pressed_time = button->time;
                OSWGrabPointerTime(
                    pw->preview,
                    True,
                    ButtonReleaseMask | PointerMotionMask,    
                    GrabModeAsync, GrabModeAsync,
                    None,
                    (wcur != NULL) ? wcur->cursor : None,
                    button->time 
                );
		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(w == pw->preview)
	    {
		pw->drag_op = PRINT_DLG_DRAG_OP_NONE;
                pw->button_pressed = 0;
		if(OSWIsPointerGrabbed())
		    OSWUngrabPointerTime(button->time);
		events_handled++;
	    }
	    break;

	  case MotionNotify:
	    if(w == pw->preview)
	    {
		if(pw->drag_op == PRINT_DLG_DRAG_OP_TRANSLATE)
		{
		    const int	x = event->xmotion.x,
				y = event->xmotion.y;
		    int px, py;
		    unsigned int	width, height,
					paper_width, paper_height;

		    OSWGetWindowGeometry(
			pw->preview, NULL, NULL, &width, &height, NULL
		    );

		    /* Get paper width and height (in pixels) */
		    PrintDlgGetPaperSizePixels(pw, &paper_width, &paper_height);

		    switch(pw->units)
		    {
		      case PRINT_DLG_UNITS_CM:
			px = (int)(PrintDlgCmToPixels(
			    PromptGetF(&pw->x_prompt)
			));
			py = (int)(PrintDlgCmToPixels(
			    PromptGetF(&pw->y_prompt)
			));
 
			PromptSetF2(
			    &pw->x_prompt,
			    PrintDlgPixelsToCm(
			        px + (((float)paper_width / (float)width) *
 				(x - pw->button_last_x))
			    ), 2
			);
			PromptSetF2(
			    &pw->y_prompt,
			    PrintDlgPixelsToCm(
				py + (((float)paper_height / (float)height) *
				(y - pw->button_last_y))
			    ), 2
			);
			break;

		      case PRINT_DLG_UNITS_INCH:
			px = (int)(PrintDlgInchToPixels(
			    PromptGetF(&pw->x_prompt)
			));
			py = (int)(PrintDlgInchToPixels(
			    PromptGetF(&pw->y_prompt)
			));

			PromptSetF2(
			    &pw->x_prompt,
			    PrintDlgPixelsToInch(
				px + (((float)paper_width / (float)width) *
				(x - pw->button_last_x))
			    ),
			    2
			);
			PromptSetF2(
			    &pw->y_prompt,
			    PrintDlgPixelsToInch(
				py + (((float)paper_height / (float)height) *
				(y - pw->button_last_y))
			    ),
			    2
			);
			break;

		      default:      /* PRINT_DLG_UNITS_PIXELS */
			px = PromptGetI(&pw->x_prompt);
			py = PromptGetI(&pw->y_prompt);

			PromptSetI(
			    &pw->x_prompt,
			    (int)(px + (((float)paper_width / (float)width) *
				(float)(x - pw->button_last_x)))
			);
			PromptSetI(
			    &pw->y_prompt,
			    (int)(py + (((float)paper_height / (float)height) *
				(float)(y - pw->button_last_y)))
			);
			break;
		    }

		    PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);

		    pw->button_last_x = x;
		    pw->button_last_y = y;
		}

		events_handled++;
	    }
	    break;

	  case FocusIn:
	    if(w == pw->toplevel)
	    {
		pw->has_focus = True;
		events_handled++;
	    }
	    break;

	  case FocusOut:
	    if(w == pw->toplevel)
	    {
		pw->has_focus = False;
		events_handled++;
	    }
	    break;

	  case ClientMessage:
	    /* WM Take Focus? */
	    if(OSWEventIsTakeFocus(pw->toplevel, event, &t))
	    {
		OSWGrabFocusTime(pw->toplevel, t);
		events_handled++;
		return(events_handled);
	    }   
	    /* WM Delete Window? */
	    else if(OSWEventIsDeleteWindow(pw->toplevel, event))
	    {
		PrintDlgCancelCB(&pw->cancel_btn, pw);
	        events_handled++;
		return(events_handled);
	    }
	    break;
	}


	if(events_handled == 0)
	    events_handled += WidgetManage(&pw->cancel_btn, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&pw->print_btn, event);

	if(events_handled == 0)
	{
	    radio_struct *radio = &pw->orientation_radio;
	    const int i = radio->armed_check_button;
	    events_handled += WidgetManage(radio, event);
	    if((events_handled > 0) &&
	       (i != radio->armed_check_button)
	    )
	    {
	        OSWDestroyImage(&pw->preview_img);
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	    }
	}

	if(events_handled == 0)
	{
	    radio_struct *radio = &pw->color_mode_radio;
	    const int i = radio->armed_check_button;
	    events_handled += WidgetManage(radio, event);
	    if((events_handled > 0) &&
	       (i != radio->armed_check_button)
	    )
	    {
		OSWDestroyImage(&pw->preview_img);
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	    }
	}

	if(events_handled == 0)
	{
	    radio_struct *radio = &pw->paper_size_radio;
	    const int i = radio->armed_check_button;
	    events_handled += WidgetManage(radio, event);
	    if((events_handled > 0) &&
	       (i != radio->armed_check_button)
	    )
	    {
		OSWDestroyImage(&pw->preview_img);
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	    }
	} 

	if(events_handled == 0)
	{
	    check_button_struct *cb = &pw->obj_labels_cb;
	    const bool_t b = CBtnGetActive(cb);
	    events_handled += CBtnManage(cb, event);
	    if((events_handled > 0) &&
	       (b != CBtnGetActive(cb))
	    )
	    {
		OSWDestroyImage(&pw->preview_img);
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	    }
	}
	if(events_handled == 0)
	{
	    check_button_struct *cb = &pw->label_geometry_cb;
	    const bool_t b = CBtnGetActive(cb);
	    events_handled += CBtnManage(cb, event);
	    if((events_handled > 0) &&
	       (b != CBtnGetActive(cb))
	    )
	    {
		OSWDestroyImage(&pw->preview_img);
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	    }
	}

	if(events_handled == 0)
	{
	    events_handled += WidgetManage(&pw->x_prompt, event);
	    if((events_handled > 0) &&
	       (event->type == KeyPress)
	    )
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	}
	if(events_handled == 0)
	{
	    events_handled += WidgetManage(&pw->y_prompt, event);
	    if((events_handled > 0) &&
	       (event->type == KeyPress)
	    )   
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	}
	if(events_handled == 0)
	{
	    events_handled += WidgetManage(&pw->width_prompt, event);
	    if((events_handled > 0) &&
	       (event->type == KeyPress)
	    )
	    {
		OSWDestroyImage(&pw->preview_img);
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	    }
	}
	if(events_handled == 0)
	{
	    events_handled += WidgetManage(&pw->height_prompt, event);
	    if((events_handled > 0) &&
	       (event->type == KeyPress)
	    )
	    {
		OSWDestroyImage(&pw->preview_img);
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	    }
	}

	if(events_handled == 0)
	{
	    popup_list_struct *pulist = &pw->objects_pulist;
	    const int i = PUListGetSelItem(pulist);
	    events_handled += PUListManage(pulist, event);
	    if((events_handled > 0) &&
	       (i != PUListGetSelItem(pulist))
	    )
	    {
		OSWDestroyImage(&pw->preview_img);
		PrintDlgQueueDraw(pw, PRINT_DLG_DRAW_AMOUNT_PREVIEW);
	    }
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(&pw->print_cmd_prompt, event);
	if(events_handled == 0)
	    events_handled += WidgetManage(&pw->print_file_prompt, event);

	if(events_handled == 0)
	    events_handled += WidgetManage(&pw->print_to_radio, event);

	return(events_handled);
}

/*
 *	Maps the print dialog.
 */
void PrintDlgMap(print_dlg_struct *pw)
{
	if(pw == NULL)
	    return;

	OSWMapRaised(pw->toplevel);
	pw->map_state = True;

	OSWMapWindow(pw->preview);

	WidgetMap(&pw->cancel_btn);
	WidgetMap(&pw->print_btn);

	WidgetMap(&pw->orientation_radio);
	WidgetMap(&pw->color_mode_radio);
	WidgetMap(&pw->paper_size_radio);

	WidgetMap(&pw->obj_labels_cb);
	WidgetMap(&pw->label_geometry_cb);

	WidgetMap(&pw->x_prompt);
	WidgetMap(&pw->y_prompt);
	WidgetMap(&pw->width_prompt);
	WidgetMap(&pw->height_prompt);

	WidgetMap(&pw->objects_pulist);

	WidgetMap(&pw->print_cmd_prompt);
	WidgetMap(&pw->print_file_prompt);

	WidgetMap(&pw->print_to_radio);

	/* Destroy preview document image to force recreation */
	OSWDestroyImage(&pw->preview_img);
}


/*
 *	Gets the values from the specified universe editor and maps
 *	the print dialog.
 */
void PrintDlgMapValues(print_dlg_struct *pw)
{
	unsigned int paper_width, paper_height;
	sw_legend_struct *legend;
	uew_struct *uew;

	if(pw == NULL)
	    return;

	uew = UEW(pw->src);
	if(uew == NULL)
	    return;

	legend = &uew->legend;
	if(!STRISEMPTY(legend->title))
	{
	    char *s = (char *)malloc(
		STRLEN(PRINT_DLG_TITLE) +
		STRLEN(legend->title) + 80
	    );
	    sprintf(
		s,
		"%s: %s",
		PRINT_DLG_TITLE, legend->title
	    );
	    OSWSetWindowTitle(pw->toplevel, s);
	    free(s);
	}

	WidgetCenterWindowToWindow(pw->toplevel, uew->toplevel);

	/* Get paper width and height (in pixels) */
	PrintDlgGetPaperSizePixels(pw, &paper_width, &paper_height);

	/* Get information from the universe editor */
	if(uew->view != None)
	{
	    unsigned int wwidth, wheight;
	    win_t w = uew->view;

	    OSWGetWindowGeometry(w, NULL, NULL, &wwidth, &wheight, NULL);

	    switch(pw->units)
	    {
	      case PRINT_DLG_UNITS_CM:
		PromptSetF2(
		    &pw->width_prompt,
		    PrintDlgPixelsToCm(wwidth),
		    2
		);
		PromptSetF2(
		    &pw->height_prompt,
		    PrintDlgPixelsToCm(wheight),
		    2
		);

		if(pw->orientation_radio.armed_check_button == 1)
		{
		    PromptSetF2(
			&pw->x_prompt,
		        PrintDlgPixelsToCm(
			    ((int)paper_width / 2) -
			    ((int)wheight / 2)
			),
			2
		    );
		    PromptSetF2(
			&pw->y_prompt,
			PrintDlgPixelsToCm(
			    ((int)paper_height / 2) -
			    ((int)wwidth / 2)
			),
			2
		    );
		}
		else
		{
		    PromptSetF2(
			&pw->x_prompt,
			PrintDlgPixelsToCm(
			    ((int)paper_width / 2) - 
			    ((int)wwidth / 2)  
			),
			2
		    );
		    PromptSetF2(
			&pw->y_prompt,
			PrintDlgPixelsToCm(                
			    ((int)paper_height / 2) -
			    ((int)wheight / 2)
			),
			2
		    );
		}
		break;

	      case PRINT_DLG_UNITS_INCH:
		PromptSetF2(
		    &pw->width_prompt,
		    PrintDlgPixelsToInch(wwidth),
		    2
		);
		PromptSetF2(
		    &pw->height_prompt,
		    PrintDlgPixelsToInch(wheight),
		    2
		);

		if(pw->orientation_radio.armed_check_button == 1)
		{
		    PromptSetF2(
			&pw->x_prompt,
			PrintDlgPixelsToInch(
			    ((int)paper_width / 2) -
			    ((int)wheight / 2)
			),
			2
		    );
		    PromptSetF2(
			&pw->y_prompt,
			PrintDlgPixelsToInch(
			    ((int)paper_height / 2) -
			    ((int)wwidth / 2)
			),
			2
		    );
		}
		else
		{
		    PromptSetF2(
			&pw->x_prompt,
			PrintDlgPixelsToInch(
			    ((int)paper_width / 2) -
			    ((int)wwidth / 2)
			),
			2
		    );
		    PromptSetF2(
			&pw->y_prompt,
			PrintDlgPixelsToInch(
			    ((int)paper_height / 2) -
			    ((int)wheight / 2)
			),
			2
		    );
		}
		break;

	      default:		/* PRINT_DLG_UNITS_PIXELS */
		PromptSetI(&pw->width_prompt, wwidth);
		PromptSetI(&pw->height_prompt, wheight);

		if(pw->orientation_radio.armed_check_button == 1)
		{
		    PromptSetI(
			&pw->x_prompt,
			((int)paper_width - (int)wheight) / 2
		    );
		    PromptSetI(
			&pw->y_prompt,
			((int)paper_height - (int)wwidth) / 2
		    );
		}
		else
		{ 
		    PromptSetI(
			&pw->x_prompt,
			((int)paper_width - (int)wwidth) / 2
		    );
		    PromptSetI(
			&pw->y_prompt,
			((int)paper_height - (int)wheight) / 2
		    );
		}
		break;
	    }
	}

	/* Map */
	PrintDlgMap(pw);
}

/*
 *	Unmaps the print dialog.
 */
void PrintDlgUnmap(print_dlg_struct *pw)
{
	if(pw == NULL)
	    return;

	WidgetUnmap(&pw->cancel_btn);
	WidgetUnmap(&pw->print_btn);

	WidgetUnmap(&pw->orientation_radio);
	WidgetUnmap(&pw->color_mode_radio);
	WidgetUnmap(&pw->paper_size_radio);

	WidgetUnmap(&pw->label_geometry_cb);
	WidgetUnmap(&pw->obj_labels_cb);

	WidgetUnmap(&pw->x_prompt);
	WidgetUnmap(&pw->y_prompt);
	WidgetUnmap(&pw->width_prompt);
	WidgetUnmap(&pw->height_prompt);

	WidgetUnmap(&pw->objects_pulist);
	
	WidgetUnmap(&pw->print_cmd_prompt);
	WidgetUnmap(&pw->print_file_prompt);

	WidgetUnmap(&pw->print_to_radio);

	OSWUnmapWindow(pw->toplevel);
	pw->map_state = False;

	OSWDestroyImage(&pw->preview_img);
	OSWDestroyPixmap(&pw->preview_buf);
}

/*
 *	Deletes the print dialog.
 */
void PrintDlgDelete(print_dlg_struct *pw)
{
	if(pw == NULL)
	    return;

	OSWDestroyPixel(&pw->paper_bg_pix);
	OSWDestroyPixel(&pw->paper_fg_pix);


	WidgetDestroy(&pw->cancel_btn);
	WidgetDestroy(&pw->print_btn);

	WidgetDestroy(&pw->orientation_radio);
	WidgetDestroy(&pw->color_mode_radio);
	WidgetDestroy(&pw->paper_size_radio);

	WidgetDestroy(&pw->label_geometry_cb);
	WidgetDestroy(&pw->obj_labels_cb);

	WidgetDestroy(&pw->x_prompt);
	WidgetDestroy(&pw->y_prompt);
	WidgetDestroy(&pw->width_prompt);
	WidgetDestroy(&pw->height_prompt);

	WidgetDestroy(&pw->objects_pulist);

	WidgetDestroy(&pw->print_cmd_prompt);
	WidgetDestroy(&pw->print_file_prompt);

	WidgetDestroy(&pw->print_to_radio);

	OSWDestroyWindow(&pw->preview);
	OSWDestroyWindow(&pw->toplevel);

	OSWDestroyImage(&pw->preview_img);
	OSWDestroyPixmap(&pw->preview_buf);

	free(pw);
}
