#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>

#include "ue.h"


void UEDialog(
	const char *title,
	const int icon_code,		/* One of UE_ICON_* */
	const char *msg,
	win_t toplevel
);
int UEGetResponse(
	const char *title,
	const int icon_code,		/* One of UE_ICON_* */
	const char *msg,
	unsigned int btn_flags,		/* Any of CDlgBtn* */
	win_t toplevel
);
void UEPrint(const char *msg);
void UEPrintWarning(const char *msg);
void UEPrintError(const char *msg);
void UEPrintBold(const char *msg);
void UEPrintDebugStderr(int error_level);
void UEPrintHelpStdout(void);
void UEPrintVersionStdout(void);

dialog_struct *UEDialogNew(void);
void UEDialogDelete(dialog_struct *d);

cdlg_struct *UEConfirmationDialogNew(void);
void UEConfirmationDialogDelete(cdlg_struct *d);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Maps the dialog with the specified message.
 *
 *	The dialog will be automatically created as needed.
 *
 *	The icon_code must be one of UE_ICON_*.
 *
 *	The msg specifies the message.
 *
 *	The toplevel specifies the relative toplevel window or you may
 *	specify None.
 */
void UEDialog(
	const char *title,
	const int icon_code,		/* One of UE_ICON_* */
	const char *msg,
	win_t toplevel
)
{
	dialog_struct *d = ue_dlg;
	image_t *icon = NULL;

	/* Not connected to GUI? */
	if(!IDC())
	{
	    printf("%s\n", msg);
	    return;
	}

	/* Initialize the dialog as needed */
	if(d == NULL)
	    ue_dlg = d = UEDialogNew();
	if(d == NULL)
	    return;

	switch(icon_code)
	{
          case UE_ICON_DEFAULT:
            icon = ue_image.icon_unvedit;
            break;
	  case UE_ICON_INFO:
	    icon = ue_image.icon_info;
	    break;
          case UE_ICON_WARNING:
            icon = ue_image.icon_warning;
            break;
          case UE_ICON_ERROR:
            icon = ue_image.icon_error;
            break;
          case UE_ICON_QUESTION:
            icon = ue_image.icon_question;
            break;
	}

	/* Set the dialog's title and icon */
        DialogSetTransientFor(d, toplevel);
        DialogSetMapRelative(d, WIDGET_WINDOW_RELATIVE_POINTER);
	DialogSetTitle(d, title);
	DialogSetIcon(d, icon);
	DialogSetMessage(d, msg);
	DialogMap(d);
}

/*
 *	Maps the confirmation dialog and waits for user response.
 *
 *	Returns one of CDlg*.
 */
int UEGetResponse(
	const char *title,
	const int icon_code,		/* One of UE_ICON_* */
	const char *msg,
	unsigned int btn_flags,		/* Any of CDlgBtn* */
	win_t toplevel
)
{
	int response;
	cdlg_struct *d;
	image_t *icon = NULL;

	/* Not connected to GUI? */
	if(!IDC())
	    return(CDlgError);

	/* Initialize the confirmation dialog as needed */
	d = ue_cdlg;
	if(d == NULL)
	    ue_cdlg = d = UEConfirmationDialogNew();
	if(d == NULL)
	    return(CDlgError);

        switch(icon_code)
        {
          case UE_ICON_DEFAULT:
            icon = ue_image.icon_unvedit;
            break;
          case UE_ICON_INFO:
            icon = ue_image.icon_info;
            break;
          case UE_ICON_WARNING:
            icon = ue_image.icon_warning;
            break;
          case UE_ICON_ERROR:
            icon = ue_image.icon_error;
            break;
          case UE_ICON_QUESTION:
            icon = ue_image.icon_question;
            break;
        }

	/* Get response */
	response = CDlgGetResponse(
	    d,
	    title, msg, icon,
	    btn_flags,
	    WIDGET_WINDOW_RELATIVE_POINTER,
	    toplevel,
	    UEMain, NULL,
	    UEMainPop, NULL
	);

	/* Destroy the confirmation dialog */
	UEConfirmationDialogDelete(d);
	ue_cdlg = d = NULL;

	return(response);
}

/*
 *	Prints the message.
 *
 *	If connected to the GUI then prints the message to the message
 *	box, otherwise it prints it to stdout.
 *
 *	The message should not have a tailing newline character.
 */
void UEPrint(const char *msg)
{
	if(msg == NULL)
	    return;

#if 0
	if(IDC() && (bridge_win.mesg_box != None))
	    MesgAdd(msg, xsw_color.bp_standard_text, 0, 0);
	else
	    fprintf(stdout, "%s\n", msg);
#endif
}

void UEPrintWarning(const char *msg)
{
	if(msg == NULL)
	    return;

#if 0
	if(IDC() && (bridge_win.mesg_box != None))
	    MesgAdd(msg, xsw_color.bp_warning, 0, 0);
	else
	    fprintf(stderr, "%s\n", msg);
#endif
}

void UEPrintError(const char *msg)
{
	if(msg == NULL)
	    return;

#if 0
	if(IDC() && (bridge_win.mesg_box != None))
	    MesgAdd(msg, xsw_color.bp_danger, 0, 0);
	else                                         
	    fprintf(stderr, "%s\n", msg);
#endif
}

void UEPrintBold(const char *msg)
{
        if(msg == NULL)
            return;

#if 0
        if(IDC() && (bridge_win.mesg_box != None))
            MesgAdd(msg, xsw_color.bold_text, 0, 0);
        else
            fprintf(stdout, "%s\n", msg);
#endif
}


/*
 *	Print usage of this program to stdout.
 */
void UEPrintHelpStdout(void)
{
	fputs(PROG_USAGE_MESSAGE, stdout);
}

/*
 *	Print version of this program to stdout.
 */
void UEPrintVersionStdout(void)
{
	fputs(
	    PROG_NAME_FULL " Version " PROG_VERSION "\n"
	    PROG_COPYRIGHT "\n",
	    stdout
	);
}

/*
 *	Creates a new dialog.
 */
dialog_struct *UEDialogNew(void)
{
	dialog_struct *d = (dialog_struct *)calloc(
	    1, sizeof(dialog_struct)
	);
	if(d == NULL)
	    return(NULL);

	if(DialogInit(d))
	{
	    free(d);
	    d = NULL;
	}

	return(d);
}

/*
 *	Deletes the dialog.
 */
void UEDialogDelete(dialog_struct *d)
{
	if(d == NULL)
	    return;

	WidgetDestroy(d);
	free(d);
}


/*
 *	Creates a new confirmation dialog.
 */
cdlg_struct *UEConfirmationDialogNew(void)
{
	cdlg_struct *d = (cdlg_struct *)calloc(
            1, sizeof(cdlg_struct)
        );
        if(d == NULL)
            return(NULL);

        if(CDlgInit(d))
        {
            free(d);
            d = NULL;
        }

        return(d);
}

/*
 *	Deletes the confirmation dialog.
 */
void UEConfirmationDialogDelete(cdlg_struct *d)
{
        if(d == NULL)
            return;

	CDlgDestroy(d);
        free(d);
}
