#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "../include/osw-x.h"
#include "../include/widget.h"

#include "../include/objects.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"

#include "blitting.h"
#include "uew.h"
#include "ue.h"


void UEWObjectDisplayDrawPixmap(
	pixmap_t pixmap, shared_image_t *simg,
	xsw_object_struct *obj,
	imgref_struct *imgref, const int imgref_frame
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define FTOI_RNDOFF(f)	((int)((f) + 0.5f))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Draws the object's imgref to the specified pixmap.
 *
 *	The simg specifies the shared image that will be used to
 *	render the compoents of the image.
 *
 *	The object specifies the object who's properties will be used
 *	to draw the contact bounds. You may pass this value as NULL for
 *	no drawing of the contact bounds.
 *
 *	In addition to the imgref, the background, the frame, and the
 *	geometry labels will all be drawn to the pixmap according to
 *	the global options.
 */
void UEWObjectDisplayDrawPixmap(
	pixmap_t pixmap, shared_image_t *simg,
	xsw_object_struct *obj,
	imgref_struct *imgref, const int imgref_frame
)
{
	unsigned int width, height;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;
	ue_option_struct *opt = &option;

	if(pixmap == None)
	    return;

	OSWGetPixmapGeometry(pixmap, &width, &height, NULL);

	/* Is there an imgref to be displayed? */
	if(ImgRefIsLoaded(imgref) && (simg != NULL))
	{
	    int x, y, xc, yc;
	    font_t *font = ue_font.geometry_label;
	    WColorStruct c;

	    const float	zoom_x = MIN(
		(float)simg->width /
		(float)MAX(imgref->fwidth * imgref->magnification, 1),
		1.0f
	    ),
			zoom_y = MIN(
		(float)simg->height /
		(float)MAX(imgref->fheight * imgref->magnification, 1),
		1.0f
	    ),
			zoom = MIN(zoom_x, zoom_y);

	    /* Draw the background for the imgref's image */
	    c.a = 0xff;
	    c.r = 0x00;
	    c.g = 0x00;
	    c.b = 0x00;
	    BlitBufClear(
		simg->bytes_per_pixel, simg->depth,
		simg->data,
		simg->width, simg->height, simg->bytes_per_line,
		&c
	    );

	    /* Draw the imgref's image */
	    xc = (int)simg->width / 2;
	    yc = (int)simg->height / 2;

	    x = xc - FTOI_RNDOFF(
		(((int)imgref->fwidth / 2 * imgref->magnification) * zoom)
	    );
	    y = yc - FTOI_RNDOFF(
		(((int)imgref->fheight / 2 * imgref->magnification) * zoom)
	    );
	    BlitBufAbsolute(
		simg->bytes_per_pixel, simg->depth,
		simg->data,
		imgref->image_data,
		x, y,
		simg->width, simg->height,
		simg->bytes_per_line,
		0, imgref->fheight * imgref_frame,
		imgref->width, imgref->height, 0,
		imgref->fwidth, imgref->fheight,
		zoom,
		imgref->magnification
	    );

	    /* Draw point lights */
	    if(imgref->nlights > 0)
	    {
		const int n = imgref->nlights;
		const float theta = imgref_frame * 0.39269908;
		int i;
		float theta2;
		imgref_light_struct *light;

		for(i = 0; i < n; i++)
		{
		    light = imgref->light[i];
		    if(light == NULL)
			continue;

		    /* Is this light suppose to strobe? */
		    if((light->strobe_off_int > 0l) &&
		       (light->strobe_on_int > 0l) &&
		       opt->animate_object_display
		    )
		    {
			/* Is the strobe currently off? */
			if(!light->strobe_state)
			    continue;
		    }

		    /* Calculate the color of light or strobe */
		    c.a = light->a;
		    c.r = light->r;
		    c.g = light->g;
		    c.b = light->b;

		    theta2 = SANITIZERADIANS(light->theta + theta);
		    x = xc + FTOI_RNDOFF(
			(float)sin((double)theta2) * light->radius *
			zoom
		    );
		    y = yc - FTOI_RNDOFF(
			(float)cos((double)theta2) * light->radius *
			zoom
		    );

		    BlitBufPointLight(
			simg->bytes_per_pixel, simg->depth,
			simg->data,
			x, y,
			simg->width, simg->height,
			simg->bytes_per_line,
			1,			/* Radius */
			&c			/* Color */
		    );
		}
	    }

	    /* Draw contact bounds */
	    if((obj != NULL) && opt->show_contact_bounds)
	    {
		WColorStruct c;
		c.a = 0xFF;
		c.r = 0xFF;
		c.g = 0xB3;
		c.b = 0xFA;
                BlitBufCircle(
                    simg->bytes_per_pixel, simg->depth,
                    (u_int8_t *)simg->data,
		    (int)(simg->width / 2),
		    (int)(simg->height / 2),
                    simg->width, simg->height, simg->bytes_per_line,
                    1,
                    (int)(obj->size * zoom),
                    &c
                );
	    }

	    OSWPutSharedImageToDrawable(simg, pixmap);

	    /* Draw the geometry label */
	    if(opt->show_geometry_labels)
	    {
		int lbearing, rbearing, swidth, ascent, descent,
		    font_height;
		char *s;

		if(imgref->magnification == 1.0f)
		{
		    s = (char *)malloc(40 + 40 + 80);
		    sprintf(
			s,
			"%ix%i%s",
			imgref->fwidth,
			imgref->fheight,
			(imgref->layer_placement == IMGREF_LAYER_BG_TILED) ?
			   " Tiled" : ""
		    );
		}
		else
		{
		    s = (char *)malloc(40 + 40 + 40 + 40 + 80);
		    sprintf(
			s,
			"%ix%i Mag:%ix%i%s",
			imgref->fwidth,
			imgref->fheight,
			(int)(imgref->magnification * imgref->fwidth),
			(int)(imgref->magnification * imgref->fheight),
			(imgref->layer_placement == IMGREF_LAYER_BG_TILED) ?
			    " Tiled" : ""
		    );
		}     
		OSWSetForeground(ue_color.view_label);
		OSWSetFont(font);
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, &rbearing, &swidth, &ascent, &descent
		);
		font_height = ascent + descent;
		x = (int)width - swidth - 4;
		y = (int)height - font_height - 4;
		OSWDrawString(
		    pixmap, 
		    x - lbearing,
		    y + ascent,
		    s
		);
		free(s);
	    }
	}
	else
	{
	    /* Redraw the background */
	    if(wglobal->force_mono)
	    {
		OSWSetForeground(gui->black_pix);
		OSWDrawSolidRectangle(pixmap, 0, 0, width, height);
	    }
	    else
	    {
		OSWSetForeground(ue_color.view_bg);
		OSWDrawSolidRectangle(pixmap, 0, 0, width, height);
	    }
	}

	WidgetFrameButtonPixmap(
	    pixmap, True,
	    width, height,
	    (wglobal->force_mono) ?
		gui->white_pix :
		wglobal->surface_highlight_pix,
	    (wglobal->force_mono) ?
		gui->white_pix :
		wglobal->surface_shadow_pix
	);
}
