#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "../include/string.h"
#include "../include/disk.h"

#include "../include/objects.h"
#include "../include/unvmain.h"
#include "../include/unvutil.h"
#include "../include/unvfile.h"

#include "clipboard.h"
#include "imgrefsfio.h"
#include "uhw.h"
#include "propdlg.h"
#include "wepw.h"
#include "ecow.h"
#include "uew.h"
#include "ue.h"
#include "uescrshot.h"
#include "config.h"


xsw_object_struct *UEWGetObjectPtr(uew_struct *uew, const int obj_num);
bool_t UEWIsObjectSelected(uew_struct *uew, const int obj_num);

void UEWDoChownAll(
	uew_struct *uew,
	const int old_owner, const int new_owner
);

void UEWSetOPMFile(uew_struct *uew, const char *path);
void UEWOpenImgRefs(uew_struct *uew, const char *path);

static void UEWUpdateWindowMenuIterate(
	uew_struct *uew, menu_struct *menu
);
static void UEWUpdateWindowMenu(uew_struct *uew);
void UEWDoUpdateWindowMenus(void);

void UEWNewUniverse(uew_struct *uew);
void UEWOpenFile(uew_struct *uew, const char *path);
void UEWOpen(uew_struct *uew);
void UEWOpenNew(uew_struct *uew);
void UEWInsert(uew_struct *uew);
void UEWOpenDNDURL(uew_struct *uew, const char *url);
int UEWSave(uew_struct *uew);
int UEWSaveAs(uew_struct *uew);
int UEWSaveSelectedAs(uew_struct *uew);
int UEWSaveAsImage(uew_struct *uew);
void UEWInsertObject(
	uew_struct *uew, xsw_object_struct *obj,
	bool_t keep_obj_position
);
void UEWCut(uew_struct *uew);
void UEWCopy(uew_struct *uew);
void UEWPaste(uew_struct *uew);
void UEWRemodel(uew_struct *uew);
void UEWDeleteObject(uew_struct *uew);
void UEWCheckUnv(
	uew_struct *uew, const bool_t verbose, const bool_t automatic
);
void UEWSelectOPMFile(uew_struct *uew);
int UEWExportOPMFile(uew_struct *uew);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? (int)strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Returns the pointer to the specified object number.
 */
xsw_object_struct *UEWGetObjectPtr(uew_struct *uew, const int obj_num)
{
	if(uew == NULL)
	    return(NULL);
	else
	    return(SW_OBJECT(CListRowGetData(
		&uew->objects_list, obj_num
	    )));
}

/*
 *	Checks if obj_num is selected on uew.
 */
bool_t UEWIsObjectSelected(uew_struct *uew, const int obj_num)
{
	int i, n, row_num;
	clist_struct *clist;

	if((uew == NULL) || (obj_num < 0))
	    return(False);

	clist = &uew->objects_list;
	n = clist->total_sel_rows;
	row_num = obj_num;
	for(i = 0; i < n; i++)
	{
	    if(clist->sel_row[i] == row_num)
		return(True);
	}

	return(False);
}


/*
 *	Chowns all objects owned by old_owner to the new_owner.
 *
 *	Note that old_owner itself will be chowned to new_owner as well.
 *
 *	The old_owner may not be -1.
 *
 *	The new_owner can be -1 to specify to unchown.
 */
void UEWDoChownAll(
	uew_struct *uew,
	const int old_owner, const int new_owner
)
{
	int i, n, last_sel_obj;
	clist_struct *clist;
	xsw_object_struct *obj;

	if((uew == NULL) || (old_owner < 0))
	    return;

	clist = &uew->objects_list;
	last_sel_obj = uew->last_selected_object;

	n = clist->total_rows;
	for(i = 0; i < n; i++)
	{
	    obj = SW_OBJECT(CListRowGetData(clist, i));
	    if(obj == NULL)
		continue;

	    if(obj->owner != old_owner)
		continue;

	    /* Need to apply property values before setting owner as
	     * needed
	     */
	    if(i == last_sel_obj)
		UEWPropsDoSetValues(uew, i);

	    obj->owner = new_owner;

	    if(i == last_sel_obj)
	        UEWPropsDoGetValues(uew, i);
	}
}


/*
 *	Set the OPM file.
 *
 *	Updates the OPM popup list.
 *
 *	Updates the global OPM file location.
 */
void UEWSetOPMFile(uew_struct *uew, const char *path)
{
	const char *name;
	char *s, *error_msg;
	int total, nopms_opened = 0;
	popup_list_struct *pulist;
	xsw_object_struct **list;

	if((uew == NULL) || STRISEMPTY(path))
	    return;

	UEWSetBusy(uew, True);

	/* Check if the file does not exist */
	if(!UECheckFileExists(path, &error_msg))
	{
	    char *s = (char *)malloc(
		STRLEN(error_msg) + STRLEN(path) + (3 * 80)
	    );
	    sprintf(
		s,
"%s:\n\
\n\
    %s",
		error_msg, path
	    );
	    UEDialog(
		"Missing OPM File",
		UE_ICON_WARNING,
		s,
		uew->toplevel
	    );
	    free(s);
	    free(error_msg);
	    UEWSetBusy(uew, False);
	    return;
	}

	name = strrchr(path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = path;

	pulist = &uew->opm_pulist;

	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Opening the OPM file `%s'...",
	    name
	);
	UEWStatus(uew, s);
	free(s);

	/* Delete all existing OPMs */
	PUListClear(pulist);

	/* Open the OPM file */
	list = UNVFileOpen(
	    path, &total,
	    NULL,
	    UEWUnvFIOProgressCB, uew
	);
	if((list != NULL) && (total > 0))
	{
	    int i, n;
	    xsw_object_struct *obj;

	    for(i = 0; i < total; i++)
	    {
		obj = list[i];
		if(obj == NULL)
		    continue;

		/* Skip garbage and no name OPMs */
		if((obj->type == SW_OBJ_TYPE_GARBAGE) ||
		   STRISEMPTY(obj->name)
		)
		{
		    UNVDeleteObject(obj);
		    continue;
		}

		n = PUListAddItem(
		    pulist,
		    0,
		    obj->name,
		    None, None
		);
		PUListItemSetData(pulist, n, obj, UEWOPMListDestroyCB);

		nopms_opened++;
	    }
	}
	free(list);

	PUListSetSelItem(pulist, 0);

	/* Report status */
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Opened %i %s from the OPM file `%s'.",
	    nopms_opened,
	    (nopms_opened == 1) ? "OPM" : "OPMs",
	    name
	);
	UEWStatus(uew, s);
	free(s);
	UEWStatusProgress(uew, 0.0f);

	if(fname.opm != path)
	{
	    strncpy(fname.opm, path, sizeof(fname.opm));
	    fname.opm[sizeof(fname.opm) - 1] = '\0';
	}

	UEWSetBusy(uew, False);
}

/*
 *	Opens the imgrefs file.
 */
void UEWOpenImgRefs(uew_struct *uew, const char *path)
{
	int nimgrefs_opened = 0;
	const char *name;
	char *s, *full_path, *error_msg;
	popup_list_struct *pulist;

	if((uew == NULL) || STRISEMPTY(path))
	    return;

	UEWSetBusy(uew, True);

	full_path = ISPATHABSOLUTE(path) ?
	    STRDUP(path) :
	    STRDUP(PrefixPaths(dname.images, path));

	/* Check if the file does not exist */
	if(!UECheckFileExists(full_path, &error_msg))
	{
	    char *s = (char *)malloc(
		STRLEN(error_msg) + STRLEN(full_path) + (3 * 80)
	    );
	    sprintf(
		s,
"%s:\n\
\n\
    %s",
		error_msg, full_path
	    );
	    UEDialog(
		"Missing ImgRefs File",
		UE_ICON_WARNING,
		s,
		uew->toplevel
	    );
	    free(s);
	    free(error_msg);
	    UEWSetBusy(uew, False);
	    return;
	}

	name = strrchr(full_path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = full_path;

	pulist = &uew->imgrefs_pulist;

	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Opening the ImgRefs file `%s'...",
	    name
	);
	UEWStatus(uew, s);
	free(s);

	/* Delete all the existing ImgRefs */
	ImgRefClear(&uew->imgref, &uew->total_imgrefs);
	PUListClear(pulist);

	/* Open the ImgRefs file */
	if(!ImgRefsFileOpen(
	    &uew->imgref, &uew->total_imgrefs,
	    full_path,
	    UEWImgRefsFIOProgressCB, uew,
	    UEPrintWarning
	))
	{
	    int i, n;
	    char *s;
	    imgref_struct *imgref;

	    /* Iterate through the loaded imgrefs */
	    for(i = 0; i < uew->total_imgrefs; i++)
	    {
		imgref = uew->imgref[i];
		if(imgref == NULL)
		    continue;

		/* Complete the filename as needed */
		if(!STRISEMPTY(imgref->path))
		{
		    char *error_msg;

		    if(!ISPATHABSOLUTE(imgref->path))
		    {
			char *name = STRDUP(imgref->path);
			free(imgref->path);
			imgref->path = STRDUP(PrefixPaths(
			    dname.images, name
			));
			free(name);
		    } 

		    /* Check if the file does not exist? */
		    if(!UECheckFileExists(imgref->path, &error_msg))
		    {
			s = (char *)malloc(
			    STRLEN(imgref->path) +
			    STRLEN(error_msg) + 80
			);
			sprintf(
			    "%s:\n\n    %s",
			    error_msg, imgref->path
			);
			UEDialog(
			    "Missing ImgRef File",
			    UE_ICON_WARNING,
			    s,
			    uew->toplevel
			);
			free(s);
			free(error_msg);
		    }
		}

		/* Add this imgref to the list */
		if(!STRISEMPTY(imgref->name))
		{
		    s = STRDUP(imgref->name);
		}
		else
		{
		    s = (char *)malloc(40 + 80);
		    sprintf(
			s,
			"ImgRef #%i",
			i
		    );
		}
		n = PUListAddItem(
		    pulist,
		    0,
		    s,
		    None, None
		);
		free(s);

		/* Set the ImgRef's number as the data */
		PUListItemSetData(
		    pulist, n,
		    (void *)i, NULL
		);

		nimgrefs_opened++;
	    }
	}

	/* Report status */
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Opened %i %s from the ImgRefs file `%s'.",
	    nimgrefs_opened,
	    (nimgrefs_opened == 1) ? "ImgRef" : "ImgRefs",
	    name
	);
	UEWStatus(uew, s);
	free(s);
	UEWStatusProgress(uew, 0.0f);

	UEWSetBusy(uew, False);

	free(full_path);
}


/*
 *	Updates the menu with a listing of all windows.
 *
 *	Called by UEWUpdateWindowMenu().
 */
static void UEWUpdateWindowMenuIterate(
	uew_struct *uew, menu_struct *menu
)
{
	int	i,
		win_num = 0,
		op_code = UEW_OP_WINDOW_START;
	char *s;
	const char *name;
	uew_struct *uew2;
	uhw_struct *uhw;
	prop_dlg_struct *prop_dlg;
	wepw_struct *wepw;
	ecow_struct *ecow;

	if((uew == NULL) || (menu == NULL))
	    return;

	/* Delete all existing items in the menu */
	MenuDeleteAllItems(menu);

	/* Add initial items */
	MenuItemAppend(
	    menu,
	    MENU_ITEM_LABEL,
	    "New Universe Editor",
	    None, None,
	    UEW_OP_NEW_UEW
	);
	MenuItemAppend(
	    menu,
	    MENU_ITEM_SEPARATOR,
	    NULL,
	    None, None,
	    UEW_OP_NONE 
	);

	/* Universe Editor Windows */
	for(i = 0; i < total_uews; i++)
	{
	    if(op_code >= UEW_OP_WINDOW_END)
		break;

	    uew2 = ue_uew[i];
	    if(uew2 == NULL)
		continue;

	    name = uew2->legend.title;
	    if(STRISEMPTY(name))
		name = "Untitled";
	    s = (char *)malloc(40 + STRLEN(name) + 80);
	    sprintf(
		s,
		"%i Universe: %s",
		win_num + 1, name
	    );
	    MenuItemAppend(
		menu,
		MENU_ITEM_LABEL,
		s,
		None, None,
		op_code   
	    );
	    free(s);

	    win_num++;
	    op_code++;
	}

	/* Universe Header Windows */
	uhw = uew->uhw;
	if((uhw != NULL) ? uhw->map_state : False)
	{
	    s = (char *)malloc(40 + 80);
	    sprintf(
		s,
		"%i Header",
		win_num + 1
	    );
	    MenuItemAppend(
		menu,
		MENU_ITEM_LABEL,
		s,
		None, None,
		op_code
	    );
	    free(s);

	    win_num++;
	    op_code++;
	}

	/* Properties Dialogs */
	for(i = 0; i < uew->total_prop_dlgs; i++)
	{
	    if(op_code >= UEW_OP_WINDOW_END)
		break;

	    prop_dlg = uew->prop_dlg[i];
	    if(prop_dlg == NULL)
		continue;

	    if(!prop_dlg->map_state)
		continue;

	    name = UNVGetObjectFormalName(
		prop_dlg->obj_num,
		UEWGetObjectPtr(uew, prop_dlg->obj_num)
	    );
	    s = (char *)malloc(40 + STRLEN(name) + 80);
	    sprintf(
		s,
		"%i Properties: %s",
		win_num + 1, name
	    );
	    MenuItemAppend(
		menu, 
		MENU_ITEM_LABEL,
		s,
		None, None, 
		op_code
	    );
	    free(s);

	    win_num++;
	    op_code++;
	}

	/* Weapon Windows */
	for(i = 0; i < uew->total_wepws; i++)
	{
	    if(op_code >= UEW_OP_WINDOW_END)
		break;

	    wepw = uew->wepw[i];
	    if(wepw == NULL)
		continue;

	    if(!wepw->map_state)
		continue;

	    name = UNVGetObjectFormalName(
		wepw->obj_num,
		UEWGetObjectPtr(uew, wepw->obj_num)
	    );
	    s = (char *)malloc(40 + STRLEN(name) + 80);
	    sprintf(
		s,
		"%i Weapons: %s",
		win_num + 1, name
	    );
	    MenuItemAppend(
		menu, 
		MENU_ITEM_LABEL,
		s,
		None, None, 
		op_code
	    );
	    free(s);

	    win_num++;
	    op_code++;
	}

	/* Economy windows */
	for(i = 0; i < uew->total_ecows; i++)
	{
	    if(op_code >= UEW_OP_WINDOW_END)
		break;

	    ecow = uew->ecow[i];
	    if(ecow == NULL)
		continue;

	    if(!ecow->map_state)
		continue;

	    name = UNVGetObjectFormalName(
		ecow->obj_num,
		UEWGetObjectPtr(uew, ecow->obj_num)
	    );
	    s = (char *)malloc(40 + STRLEN(name) + 80);
	    sprintf(
		s,
		"%i Economy: %s",
		win_num + 1, name
	    );
	    MenuItemAppend(
		menu,
		MENU_ITEM_LABEL,
		s,
		None, None,
		op_code
	    );
	    free(s);

	    win_num++;
	    op_code++;
	}
}

/*
 *	Updates the Universe Editor WIndow's window menu.
 */
static void UEWUpdateWindowMenu(uew_struct *uew)
{
	menu_struct *menu;
	menu_bar_struct *mb;

	if(uew == NULL)
	    return;

	mb = &uew->mb;
	menu = MenuBarGetMenuFromItem(
	    mb,
	    uew->window_menu_num
	);
	UEWUpdateWindowMenuIterate(uew, menu);
}

/*
 *	Updates all the windows menus on all the Universe Editors.
 */
void UEWDoUpdateWindowMenus(void)
{
	int i;
	for(i = 0; i < total_uews; i++)
	    UEWUpdateWindowMenu(ue_uew[i]);
}


/*
 *	New universe.
 */
void UEWNewUniverse(uew_struct *uew)
{
	clist_struct *clist;

	if(uew == NULL)
	    return;

	UEWSetBusy(uew, True);

	clist = &uew->objects_list;

	/* Unsaved changes? */
	if(uew->has_changes)
	{
	    /* Query user to save changes */
	    const int response = UEGetResponse(
		"Confirm Save Universe",
		UE_ICON_WARNING,
"The current universe has unsaved changes,\n\
do you want to save those changes?",
		CDlgBtnNo | CDlgBtnYes | CDlgBtnCancel,
		uew->toplevel
	    );
	    if(response == CDlgYes)
	    {
		if(STRISEMPTY(uew->unv_file))
		{
		    UEDialog(
"Untitled Universe",
			UE_ICON_INFO,
"The current universe has not been saved to a specified\n\
file name. To save it under a new file name, go to\n\
File->Save As...",
			uew->toplevel
		    );
		    UEWSetBusy(uew, False);
		    return;
		}

		/* Save */
		if(UEWSave(uew))
		{
		    UEWSetBusy(uew, False);
		    return;
		}
	    }
	    else if(response == CDlgNo)
	    {
		/* No save */
	    }
	    else
	    {
		/* Cancel */
		UEWSetBusy(uew, False);
		return;
	    }
	}

	/* Delete all the objects */
	CListClear(clist);

	/* Delete all the imgrefs */
	ImgRefClear(&uew->imgref, &uew->total_imgrefs);

	/* Reset the universe legend */
	UNVLegendReset(&uew->legend);

	/* Move back to center */
	uew->sect_x = 0l;
	uew->sect_y = 0l;
	uew->sect_z = 0l;
	uew->cx = 0.0f;
	uew->cy = 0.0f;
	uew->cz = 0.0f;
	uew->zoom = UEW_VIEW_DEF_ZOOM;

	/* Unset the current file path */
	free(uew->unv_file);
	uew->unv_file = NULL;

	/* Reset the has changes marker */
	uew->has_changes = False;

	UEWUpdateTitle(uew);

	/* Clear the view status prompts */
	UEWViewStatusGetValues(uew, NULL);

	UEWStatus(uew, "New universe created.");
	UEWStatusProgress(uew, 0.0f);

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	UEWDoUpdateWindowMenus();

	UEWSetBusy(uew, False);
}

/*
 *	Opens the specified universe file.
 */
void UEWOpenFile(uew_struct *uew, const char *path)
{
	int total, nobjs_opened = 0;
	char *s, *error_msg;
	const char *name;
	clist_struct *clist;
	sw_legend_struct *legend;
	xsw_object_struct **list;

	if((uew == NULL) || STRISEMPTY(path))
	    return;

	UEWSetBusy(uew, True);

	/* Check if the file does not exist */
	if(!UECheckFileExists(path, &error_msg))
	{
	    char *s = (char *)malloc(
		STRLEN(error_msg) + STRLEN(path) + (3 * 80)
	    );
	    sprintf(
		s,
"%s:\n\
\n\
    %s",
		error_msg, path
	    );
	    UEDialog(
		"Missing Universe File",
		UE_ICON_WARNING,
		s,
		uew->toplevel
	    );
	    free(s);
	    free(error_msg);
	    UEWSetBusy(uew, False);
	    return;
	}

	clist = &uew->objects_list;
	legend = &uew->legend;

	name = strrchr(path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = path;

	/* Delete all the objects */
	CListClear(clist);

	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Opening the universe file `%s'...",
	    name
	);
	UEWStatus(uew, s);
	free(s);

	/* Open the universe file */
	list = UNVFileOpen(
	    path, &total,
	    legend,
	    UEWUnvFIOProgressCB, uew
	);
	if((list != NULL) && (total > 0))
	{
	    int i, row_num;
	    xsw_object_struct *obj;

	    for(i = 0; i < total; i++)
	    {
		obj = list[i];
		if(obj == NULL)
		    continue;

		/* Skip garbage objects */
		if(obj->type == SW_OBJ_TYPE_GARBAGE)
		{
		    UNVDeleteObject(obj);
		    continue;
		}

		row_num = CListRowAppend(clist);
		if(row_num > -1)
		{
		    const int obj_num = row_num;
		    char *v[2], *s;

		    v[0] = s = (char *)malloc(40);
		    sprintf(s, "#%i", obj_num);

		    v[1] = STRDUP(obj->name);

		    CListSetRowTexts(
			clist, row_num,
			(const char **)v, sizeof(v) / sizeof(char *)
		    );

		    free(v[0]);
		    free(v[1]);

		    /* Set the object as the row's data */
		    CListRowSetData(
			clist, row_num,
			obj, UEWObjectsCListDestroyCB
		    );

		    nobjs_opened++;
		}
		else
		{
		    UNVDeleteObject(obj);
		    continue;
		}
	    }
	}
	free(list);

	/* Set the newly opened universe file name */
	free(uew->unv_file);
	uew->unv_file = STRDUP(path);

	/* Open new imgrefs */
	UEWOpenImgRefs(uew, legend->imgrefs);

	/* Reset the has changes marker */
	uew->has_changes = False;

	UEWUpdateTitle(uew);

	/* Move to the player start location */
	uew->sect_x = legend->player_start_sect_x;
	uew->sect_y = legend->player_start_sect_y;
	uew->sect_z = legend->player_start_sect_z;
	uew->cx = legend->player_start_x;
	uew->cy = legend->player_start_y;
	uew->cz = legend->player_start_z;

	uew->zoom = UEW_VIEW_DEF_ZOOM;

	/* Clear the view's status prompts */
	UEWViewStatusGetValues(uew, NULL);

	/* Report status */
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Opened %i %s from the universe file `%s'.",
	    nobjs_opened,
	    (nobjs_opened == 1) ? "object" : "objects",
	    name
	);
	UEWStatus(uew, s);
	free(s);
	UEWStatusProgress(uew, 0.0f);

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	UEWDoUpdateWindowMenus();

	UEWSetBusy(uew, False);
}

/*
 *	Open universe.
 */
void UEWOpen(uew_struct *uew)
{
	int i, npaths;
	char *path, **paths_list;
	clist_struct *clist;
	uew_struct *uew2;

	if(uew == NULL)
	    return;

	clist = &uew->objects_list;

	/* Unsaved changes? */
	if(uew->has_changes)
	{
	    /* Query user to save changes */
	    const int response = UEGetResponse(
		"Confirm Save Universe",
		UE_ICON_WARNING,
"The current universe has unsaved changes,\n\
do you want to save those changes?",
		CDlgBtnNo | CDlgBtnYes | CDlgBtnCancel,
		uew->toplevel
	    );
	    if(response == CDlgYes)
	    {
		if(STRISEMPTY(uew->unv_file))
		{
		    UEDialog(
"Untitled Universe",
			UE_ICON_INFO,
"The current universe has not been saved to a specified\n\
file name. To save it under a new file name, go to\n\
File->Save As...",
			uew->toplevel
		    );
		    return;
		}

		/* Save */
		if(UEWSave(uew))
		    return;
	    }
	    else if(response == CDlgNo)
	    {
		/* No save */
	    }
	    else
	    {
		/* Cancel */
		return;
	    }
	}

	/* Query user for universe file to open */
	if(uew->unv_file != NULL)
	    path = STRDUP(PrefixPaths(
		GetParentDir(uew->unv_file),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	else 
	    path = STRDUP(PrefixPaths(
		UECurrentDirectory(),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	if(!UESelectFile(
	    "Open Universe",
	    "Open", "Cancel",
	    path, &paths_list, &npaths,
	    uew->toplevel
	))
	{
	    free(path);
	    return;
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);

	/* Already opened? */
	for(i = 0; i < total_uews; i++)
	{
	    uew2 = ue_uew[i];
	    if(uew2 == NULL)
		continue;

	    /* Skip current */
	    if(uew2 == uew)
		continue;

	    if(STRISEMPTY(uew2->unv_file))
		continue;

	    if(!strcmp(uew2->unv_file, path))
	    {
		if(UEGetResponse(
		    "Open Universe",
		    UE_ICON_QUESTION,
"This file appears to be already opened in another\n\
window. Are you sure you want to open this file?",
		    CDlgBtnNo | CDlgBtnYes,
		    uew->toplevel
		) != CDlgYes)
		{
		    free(path);
		    return;
		}
	    }
	}

	/* Open the universe file */
	UEWOpenFile(uew, path);

	free(path);
}

/*
 *	Opens a universe file to a new universe editor.
 */
void UEWOpenNew(uew_struct *uew)
{
	int i, npaths;
	char *path, **paths_list;
	uew_struct *uew2;

	/* Query user for universe file to open */
	if(uew->unv_file != NULL)
	    path = STRDUP(PrefixPaths(
		GetParentDir(uew->unv_file),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	else 
	    path = STRDUP(PrefixPaths(
		UECurrentDirectory(),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	if(!UESelectFile(
	    "Open Universe",
	    "Open", "Cancel",
	    path, &paths_list, &npaths,
	    uew->toplevel
	))
	{
	    free(path);
	    return;
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);

	/* Already opened? */
	for(i = 0; i < total_uews; i++)
	{
	    uew2 = ue_uew[i];
	    if(uew2 == NULL)
		continue;

	    /* Skip current */
	    if(uew2 == uew)
		continue;

	    if(STRISEMPTY(uew2->unv_file))
		continue;

	    if(!strcmp(uew2->unv_file, path))
	    {
		if(UEGetResponse(
		    "Open Universe",
		    UE_ICON_QUESTION,
"This file appears to be already opened in another\n\
window. Are you sure you want to open this file?",
		    CDlgBtnNo | CDlgBtnYes,
		    uew->toplevel
		) != CDlgYes)
		{
		    free(path);
		    return;
		}
	    }
	}

	/* Create a new universe editor and open the selected
	 * universe file
	 */
	if(UECreateUEW(path) < 0)
	{
	    UEDialog(
		"Create Universe Editor Failed",
		UE_ICON_ERROR,
"Unable to create a new Universe Editor.",
		uew->toplevel
	    );
	}

	free(path);
}

/*
 *	Inserts the objects from the specified universe file.
 */
void UEWInsert(uew_struct *uew)
{
	int npaths, total, ins_row_num, nobjs_inserted = 0;
	const char *name;
	char *s, *path, **paths_list;
	clist_struct *clist;
	xsw_object_struct **list;

	if(uew == NULL)
	    return;

	UEWSetBusy(uew, True);

	clist = &uew->objects_list;
	ins_row_num = CListGetLastSelectedRow(clist);

	/* Query user for universe file to insert */
	if(uew->unv_file != NULL)
	    path = STRDUP(PrefixPaths(
		GetParentDir(uew->unv_file),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	else
	    path = STRDUP(PrefixPaths(
		UECurrentDirectory(),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	if(!UESelectFile(
	    "Insert Universe",
	    "Insert", "Cancel",
	    path, &paths_list, &npaths,
	    uew->toplevel
	))
	{
	    UEWSetBusy(uew, False);
	    free(path);
	    return;
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);

	name = strrchr(path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = path;

	/* Unselect all objects */
	CListUnselectAllRows(clist);

	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Inserting the universe file `%s'...",
	    name  
	);
	UEWStatus(uew, s);
	free(s);

	/* Open the universe file */
	list = UNVFileOpen(
	    path, &total,
	    NULL,		/* Ignore the legend */
	    UEWUnvFIOProgressCB, uew
	);
	if((list != NULL) && (total > 0))
	{
	    int i, row_num;
	    xsw_object_struct *obj;

	    for(i = 0; i < total; i++)
	    {
		obj = list[i];
		if(obj == NULL)
		    continue;

		/* Skip garbage objects */
		if(obj->type == SW_OBJ_TYPE_GARBAGE)
		{
		    UNVDeleteObject(obj);
		    continue;
		}

		/* Insert or append this object */
		if(ins_row_num > -1)
		{
		    row_num = CListRowInsert(clist, ins_row_num);
		    ins_row_num++;
		}
		else
		{
		    row_num = CListRowAppend(clist);
		}
		if(row_num > -1)
		{           
		    const int obj_num = row_num;
		    char *v[2], *s;

		    v[0] = s = (char *)malloc(40);
		    sprintf(s, "#%i", obj_num);

		    v[1] = STRDUP(obj->name);

		    CListSetRowTexts(
			clist, row_num,
			(const char **)v, sizeof(v) / sizeof(char *)
		    );

		    free(v[0]);
		    free(v[1]);

		    /* Set the object as the row's data */
		    CListRowSetData(  
			clist, row_num,
			obj, UEWObjectsCListDestroyCB
		    );
		}
		else
		{
		    UNVDeleteObject(obj);
		    continue;
		}

		if(!uew->has_changes)
		{
		    uew->has_changes = True;
		    UEWUpdateTitle(uew);
		}

		nobjs_inserted++;
	    }    
	}    
	free(list);

	/* Report status */
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Inserted %i %s from the universe file `%s'.",
	    nobjs_inserted,
	    (nobjs_inserted == 1) ? "object" : "objects",
	    name
	);
	UEWStatus(uew, s);
	free(s);
	UEWStatusProgress(uew, 0.0f);

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	UEWDoUpdateWindowMenus();

	UEWSetBusy(uew, False);

	free(path);
}

/*
 *	Opens the universe file from the specified DND URL.
 */
void UEWOpenDNDURL(uew_struct *uew, const char *url)
{
	int i;
	const char *s;
	char *path;
	uew_struct *uew2;

	if((uew == NULL) || STRISEMPTY(url))
	    return;

	s = strstr(url, "://");
	if(s != NULL)
	{
	    s += STRLEN("://");
	    path = STRDUP(strchr(s, '/'));
	}
	else
	{
	    path = STRDUP(url);
	}

	if(STRISEMPTY(path))
	{
	    free(path);
	    return;
	}

	/* Unsaved changes? */
	if(uew->has_changes)
	{
	    /* Query user to save changes */
	    const int response = UEGetResponse(
		"Confirm Save Universe",
		UE_ICON_WARNING,
"The current universe has unsaved changes,\n\
do you want to save those changes?",
		CDlgBtnNo | CDlgBtnYes | CDlgBtnCancel,
		uew->toplevel
	    );
	    if(response == CDlgYes)
	    {
		if(STRISEMPTY(uew->unv_file))
		{
		    UEDialog(
"Untitled Universe",
			UE_ICON_INFO,
"The current universe has not been saved to a specified\n\
file name. To save it under a new file name, go to\n\
File->Save As...",
			uew->toplevel
		    );
		    free(path);
		    return;
		}

		/* Save */
		if(UEWSave(uew))
		{
		    free(path);
		    return;
		}
	    }
	    else if(response == CDlgNo)
	    {
		/* No save */
	    }
	    else
	    {
		/* Cancel */
		free(path);
		return;
	    }
	}

	/* Already opened? */
	for(i = 0; i < total_uews; i++)
	{
	    uew2 = ue_uew[i];
	    if(uew2 == NULL)
		continue;

	    /* Skip current */
	    if(uew2 == uew)
		continue;

	    if(STRISEMPTY(uew2->unv_file))
		continue;

	    if(!strcmp(uew2->unv_file, path))
	    {
		if(UEGetResponse(
		    "Open Universe",
		    UE_ICON_QUESTION,
"This file appears to be already opened in another\n\
window. Are you sure you want to open this file?",
		    CDlgBtnNo | CDlgBtnYes,
		    uew->toplevel
		) != CDlgYes)
		{
		    free(path);
		    return;
		}
	    }
	}

	/* Open the universe file */
	UEWOpenFile(uew, path);

	free(path);
}


/*
 *	Saves the universe to the specified file.
 *
 *	The current object's properties will be applied before the
 *	universe is saved.
 *
 *	The has_changes will be reset to False on successful save.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	Error.
 *	-2	Bad value.
 *	-3	Systems error or memory allocation error.
 *	-4	User aborted or canceled.
 */
int UEWSaveFile(uew_struct *uew, const char *path)
{
	int n, obj_num, status, nobjs_saved = 0;
	mode_t mode;
	char *s;
	const char *name;
	struct stat stat_buf;
	clist_struct *clist;
	sw_legend_struct *legend;
	xsw_object_struct **list;

	if((uew == NULL) || STRISEMPTY(path))
	    return(-2);

	UEWSetBusy(uew, True);

	clist = &uew->objects_list;
	legend = &uew->legend;

	name = strrchr(path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = path;

	if(stat(path, &stat_buf))
	{
	    mode = UECurrentFileMode();
	}
	else
	{
	    mode = stat_buf.st_mode;
	}

	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Saving the universe file `%s'...",
	    name
	);
	UEWStatus(uew, s);
	free(s);

	/* Apply values to the last selected object */
	obj_num = uew->last_selected_object;
	UEWPropsDoSetValues(uew, obj_num);

	/* Update universe header */
	strncpy(
	    legend->version, PROG_VERSION,
	    sizeof(legend->version)
	);
	legend->version[sizeof(legend->version) - 1] = '\0';

	/* Save to file */
	n = clist->total_rows;
	if(n > 0)
	{
	    int i;
	    list = (xsw_object_struct **)malloc(
		n * sizeof(xsw_object_struct *)
	    );
	    for(i = 0; i < n; i++)
		list[i] = SW_OBJECT(CListRowGetData(clist, i));
	}
	else
	{
	    list = NULL;
	}
	status = UNVFileSave(
	    path, mode,
	    list, n,
	    legend,
	    UEWUnvFIOProgressCB, uew
	);
	free(list);
	UEWStatusProgress(uew, 0.0f);
	nobjs_saved = n;

	if(status < 0)
	{
	    s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"Error saving the universe to the file `%s'.",
		name
	    );
	    UEWStatus(uew, s);
	    free(s);

	    s = (char *)malloc(
		STRLEN(path) + (2 * 80)
	    );
	    sprintf(
		s,
"An error occured while saving the universe to the file:\n\n\
    %s",
		path
	    );
	    UEDialog(
		"Save Universe Error",
		UE_ICON_ERROR,
		s,
		uew->toplevel
	    );
	    free(s);

	    UEWSetBusy(uew, False);

	    return(-1);
	}

	/* Reset the has changes marker */
	uew->has_changes = False;

	UEWUpdateTitle(uew);

	/* Report status */
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Saved %i %s to the universe to the file `%s'.",
	    nobjs_saved,
	    (nobjs_saved == 1) ? "object" : "objects",
	    name
	);
	UEWStatus(uew, s);
	free(s);

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	UEWDoUpdateWindowMenus();

	UEWSetBusy(uew, False);

	return(0);
}

/*
 *	Saves the current universe.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	Error.
 *	-2	Bad value.
 *	-3	Systems error or memory allocation error.
 *	-4	User aborted or canceled.
 */
int UEWSave(uew_struct *uew)
{
	if(uew == NULL)
	    return(-2);

	if(STRISEMPTY(uew->unv_file))
	    return(UEWSaveAs(uew));
	else
	    return(UEWSaveFile(uew, uew->unv_file));
}

/*
 *	Queries the user for a file name to save the universe to
 *	and then saves the universe to the specified file name.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	Error.
 *	-2	Bad value.
 *	-3	Systems error or memory allocation error.
 *	-4	User aborted or canceled.
 */
int UEWSaveAs(uew_struct *uew)
{
	int status, npaths;
	char *path, **paths_list;

	if(uew == NULL)
	    return(-2);

	/* Query user for universe file to save to */
	if(uew->unv_file != NULL)
	    path = STRDUP(PrefixPaths(
		GetParentDir(uew->unv_file),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	else
	    path = STRDUP(PrefixPaths(
		UECurrentDirectory(),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	if(!UESelectFile(
	    "Save Universe",
	    "Save", "Cancel",
	    path, &paths_list, &npaths,
	    uew->toplevel
	))
	{
	    free(path);
	    return(-4);
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);

	if(STRISEMPTY(path))
	{
	    free(path);
	    return(-4);
	}

	/* Check if the file already exists */
	if(UECheckFileExists(path, NULL))
	{
	    /* Confirm overwrite */
	    int response;
	    char *s = (char *)malloc(STRLEN(path) + 80);
	    sprintf(
		s,
"Overwrite existing file:\n\n    %s",
		path
	    );
	    response = UEGetResponse(
		"Confirm Overwrite",
		UE_ICON_WARNING,
		s,
		CDlgBtnNo | CDlgBtnYes,
		uew->toplevel
	    );
	    free(s);
	    if(response != CDlgYes)
	    {
		free(path);
		return(-4);
	    }
	}

	status = UEWSaveFile(uew, path);
	if(!status)
	{
	    /* Save successful, record the new file name for this
	     * universe
	     */
	    free(uew->unv_file);
	    uew->unv_file = STRDUP(path);

	    UEWDoUpdateWindowMenus();
	}

	free(path);

	return(status);
}

/*
 *	Queries the user for a universe file name to save the selected
 *	object(s) to and then saves the objects to the specified
 *	universe file name.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	Error.
 *	-2	Bad value.
 *	-3	Systems error or memory allocation error.
 *	-4	User aborted or canceled.
 */
int UEWSaveSelectedAs(uew_struct *uew)
{
	int i, row_num, status, npaths, total, nobjs_saved = 0;
	mode_t mode;
	const char *name;
	char *s, *path, **paths_list;
	struct stat stat_buf;
	clist_struct *clist;
	sw_legend_struct *legend;
	xsw_object_struct **list;

	if(uew == NULL)
	    return(-2);

	UEWSetBusy(uew, True);

	legend = &uew->legend;
	clist = &uew->objects_list;

	/* Create a list of selected objects */
	total = clist->total_sel_rows;
	list = (total > 0) ?
	    (xsw_object_struct **)malloc(
		total * sizeof(xsw_object_struct *)
	    ) : NULL;
	for(i = 0; i < total; i++)
	{
	    row_num = clist->sel_row[i];
	    list[i] = SW_OBJECT(CListRowGetData(clist, row_num));
	}

	/* No objects selected? */
	if((list == NULL) || (total <= 0))
	{
	    UEDialog(
		"Save Selected Objects",
		UE_ICON_WARNING,
		"No objects selected.",
		uew->toplevel
	    );
	    UEWSetBusy(uew, False);
	    free(list);
	    return(-2);
	}

	/* Query user for the universe file to save to */
	if(uew->unv_file != NULL)
	    path = STRDUP(PrefixPaths(
		GetParentDir(uew->unv_file),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	else
	    path = STRDUP(PrefixPaths(
		UECurrentDirectory(),
		UE_FILE_FORMAT_EXTENSION_MASK_UNV
	    ));
	if(!UESelectFile(
	    "Save Selected Objects",
	    "Save", "Cancel",
	    path, &paths_list, &npaths,
	    uew->toplevel
	))
	{
	    UEWSetBusy(uew, False);
	    free(list);
	    free(path);
	    return(-4);
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);

	if(STRISEMPTY(path))
	{
	    UEWSetBusy(uew, False);
	    free(list);
	    free(path);
	    return(-4);
	}

	name = strrchr(path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = path;

	/* Check if the file already exists */
	if(UECheckFileExists(path, NULL))
	{
	    /* Confirm overwrite */
	    int response;
	    s = (char *)malloc(STRLEN(path) + 80);
	    sprintf(
		s,
"Overwrite existing file:\n\n    %s",
		path
	    );
	    response = UEGetResponse(
		"Confirm Overwrite",
		UE_ICON_WARNING,
		s,
		CDlgBtnNo | CDlgBtnYes,
		uew->toplevel
	    );
	    free(s);
	    if(response != CDlgYes)
	    {
		UEWSetBusy(uew, False);
		free(list);
		free(path);
		return(-4);
	    }
	}

	if(stat(path, &stat_buf))
	{
	    mode = UECurrentFileMode();
	}
	else
	{
	    mode = stat_buf.st_mode;
	}

	/* Save the selected objects */
	status = UNVFileSave(
	    path, mode,
	    list, total,
	    legend,
	    UEWUnvFIOProgressCB, uew
	);
	UEWStatusProgress(uew, 0.0f);
	nobjs_saved = total;
	free(list);
	list = NULL;
	total = 0;

	if(status < 0)
	{
	    s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"Error saving the %s to the file `%s'.",
		(nobjs_saved == 1) ? "object" : "objects",
		name
	    );
	    UEWStatus(uew, s);
	    free(s);

	    s = (char *)malloc(
		STRLEN(path) + (3 * 80)
	    );
	    sprintf(
		s,
"An error occured while saving the %s\n\
to the file:\n\n\
    %s",
		(nobjs_saved == 1) ? "object" : "objects",
		path
	    );
	    UEDialog(
		"Save Universe Error",
		UE_ICON_ERROR,
		s,
		uew->toplevel
	    );
	    free(s);

	    UEWSetBusy(uew, False);
	    free(path);

	    return(-1);
	}

	/* Report status */
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Saved %i %s to the file `%s'.",
	    nobjs_saved,
	    (nobjs_saved == 1) ? "object" : "objects",
	    name
	);
	UEWStatus(uew, s);
	free(s);

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	UEWSetBusy(uew, False);

	free(path);

	return(status);
}

/*
 *	Saves a screenshot image of the current view contents.
 *
 *	Returns:
 *
 *	0	Success.
 *	-1	Error.
 *	-2	Bad value.
 *	-3	Systems error or memory allocation error.
 *	-4	User aborted or canceled.
 */
int UEWSaveAsImage(uew_struct *uew)
{
	int status, npaths;
	const char *name;
	char *s, *path, **paths_list;
	drawable_t d;

	if(uew == NULL)
	    return(-2);

	UEWSetBusy(uew, True);

	/* Query user for the image file to save to */
	if(uew->unv_file != NULL)
	    path = STRDUP(PrefixPaths(
		GetParentDir(uew->unv_file),
		UE_FILE_FORMAT_EXTENSION_MASK_TGA
	    ));
	else
	    path = STRDUP(PrefixPaths(
		UECurrentDirectory(),
		UE_FILE_FORMAT_EXTENSION_MASK_TGA
	    ));
	if(!UESelectFile(
	    "Save As Image",
	    "Save", "Cancel",
	    path, &paths_list, &npaths,
	    uew->toplevel
	))
	{
	    UEWSetBusy(uew, False);
	    free(path);
	    return(-4);
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);

	if(STRISEMPTY(path))
	{
	    UEWSetBusy(uew, False);
	    free(path);
	    return(-4);
	}

	name = strrchr(path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = path;

	/* Check if the file already exists */
	if(UECheckFileExists(path, NULL))
	{
	    /* Confirm overwrite */
	    int response;
	    s = (char *)malloc(STRLEN(path) + 80);
	    sprintf(
		s,
"Overwrite existing file:\n\n    %s",
		path
	    );
	    response = UEGetResponse(
		"Confirm Overwrite",
		UE_ICON_WARNING,
		s,
		CDlgBtnNo | CDlgBtnYes,
		uew->toplevel
	    );
	    free(s);
	    if(response != CDlgYes)
	    {
		UEWSetBusy(uew, False);
		free(path);
		return(-4);
	    }
	}

	/* Get the view buffer or the window */
	d = uew->view_buf;
	if(d == None)
	    d = uew->view;

	/* Save the view buffer or window to the image file */
	status = UEScrShotWindow(path, d);
	if(status < 0)
	{
	    s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"Error saving the screenshot of the view to the file `%s'.",
		name
	    );
	    UEWStatus(uew, s);
	    free(s);

	    s = (char *)malloc(
		STRLEN(path) + (3 * 80)
	    );
	    sprintf(
		s,
"An error occured while saving the screenshot of the view\n\
to the file:\n\n\
    %s",
		path
	    );
	    UEDialog(
		"Save Image",
		UE_ICON_ERROR,
		s,
		uew->toplevel
	    );
	    free(s);

	    UEWSetBusy(uew, False);
	    free(path);

	    return(-1);
	}

	/* Report status */
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Saved the screenshot of the view to the file `%s'.",
	    name
	);
	UEWStatus(uew, s);
	free(s);

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	UEWSetBusy(uew, False);

	free(path);

	return(status);
}

/*
 *	Inserts an object using the specified object as a reference.
 *
 *	If obj is NULL then the new object will have default values.
 */
void UEWInsertObject(
	uew_struct *uew, xsw_object_struct *obj,
	bool_t keep_obj_position
)
{
	int i, row_num, obj_num;
	char *v[2], *s;
	clist_struct *clist;
	sw_legend_struct *legend;
	xsw_object_struct *new_obj;

	if(uew == NULL)
	    return;

	legend = &uew->legend;
	clist = &uew->objects_list;

	/* Record the last selected object (before unselecting) */
	obj_num = uew->last_selected_object;
	row_num = obj_num;

	/* Unselect all the objects */
	CListUnselectAllRows(clist);

	/* Insert a new row in the objects list */
	row_num = CListRowInsert(clist, row_num);
	if(row_num < 0)
	    return;

	/* Create the new object */
	if(obj == NULL)
	{
	    new_obj = UNVDupObject(&unv_garbage_object);
	    if(new_obj == NULL)
		return;

	    strncpy(new_obj->name, "New Object", sizeof(new_obj->name));
	    new_obj->name[sizeof(new_obj->name) - 1] = '\0';

	    new_obj->type = SW_OBJ_TYPE_STATIC;
	}
	else
	{
	    new_obj = UNVDupObject(obj);
	    if(new_obj == NULL)
		return;
	}


	v[0] = s = (char *)malloc(80);
	sprintf(s, "#%i", row_num);

	v[1] = STRDUP(new_obj->name);

	CListSetRowTexts(
	    clist, row_num,
	    (const char **)v, sizeof(v) / sizeof(char *)
	);

	free(v[0]);
	free(v[1]);

	/* Set the newly inserted object as the row's data */
	CListRowSetData(
	    clist, row_num,
	    new_obj, UEWObjectsCListDestroyCB
	);

	/* Update all the object numbers due to a new object being
	 * inserted in the list
	 */
	for(i = 0; i < clist->total_rows; i++)
	    UEWObjectsListItemUpdateObjNum(uew, i);

	/* Move the new object to center of view position? */
	if(!keep_obj_position)
	{
	    new_obj->sect_x = uew->sect_x;
	    new_obj->sect_y = uew->sect_y;
	    new_obj->sect_z = uew->sect_z;
	    new_obj->x = uew->cx;
	    new_obj->y = uew->cy;
	    new_obj->z = uew->cz;

	    /* Also set default owner */
	    new_obj->owner = legend->lost_found_owner;
	}

	/* Update the has changes mark */
	if(!uew->has_changes)
	{
	    uew->has_changes = True;
	    UEWUpdateTitle(uew);
	}

	/* Select the new object */
	CListSelectRow(clist, row_num);

	if(obj != NULL)
	{
	    UEWStatus(uew, "Inserted object.");
	}
	else
	{
	    UEWStatus(uew, "Inserted new object.");
	}
}

/*
 *	Coppies the selected object(s) to the clipboard.
 */
void UEWCut(uew_struct *uew)
{
	int i, n, row_num, obj_num, objects_cut = 0;
	clist_struct *clist;
	xsw_object_struct *obj, **list;

	if(uew == NULL)
	    return;

	clist = &uew->objects_list;
	if(clist->total_sel_rows <= 0)
	{
	    UEDialog(
		"Cut Objects",
		UE_ICON_WARNING,
		"No objects selected.",
		uew->toplevel
	    );
	    return;
	}

	/* Copy the objects first */
	UEWCopy(uew);

	/* Create a list of objects to be deleted */
	n = clist->total_sel_rows;
	list = (xsw_object_struct **)malloc(
	    n * sizeof(xsw_object_struct *)
	);
	for(i = 0; i < n; i++)
	    list[i] = SW_OBJECT(CListRowGetData(
		clist, clist->sel_row[i]  
	    ));

	/* Iterate through the list of objects to be deleted */
	for(i = 0; i < n; i++)
	{
	    obj = list[i];
	    if(obj == NULL)
		continue;

	    row_num = CListRowGetByData(clist, obj);
	    if(row_num < 0)
		continue;

	    obj_num = row_num;

	    /* Need to make sure no other objects own this object */
	    UEWDoChownAll(
		uew,
		obj_num,	/* Previous owner (this object) */
		((uew->legend.lost_found_owner < 0) ?
		    0 : uew->legend.lost_found_owner)
	    );
	      
	    /* Delete this object */
	    CListRowRemove(clist, row_num);
	    objects_cut++;

	    /* Update the has changes mark */
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}

	/* Delete the list of objects to be deleted */
	free(list);

	/* Update all the object numbers due to an object being removed
	 * from the list
	 */
	for(i = 0; i < clist->total_rows; i++)
	    UEWObjectsListItemUpdateObjNum(uew, i);

	if(objects_cut > 0)
	{
	    char *s = (char *)malloc(40 + 80);
	    sprintf(
		s,
"Cut %i %s.",
		objects_cut,
		(objects_cut == 1) ? "object" : "objects"
	    );
	    UEWStatus(uew, s);
	    free(s);
	}
	else
	{
	    UEWStatus(uew, "No objects cut.");
	}
}

/*
 *	Coppies the selected object(s) to the clipboard.
 */
void UEWCopy(uew_struct *uew)
{
	int i, n, row_num, obj_num, objects_coppied;
	clist_struct *clist;
	xsw_object_struct *obj;

	if(uew == NULL)
	    return;

	clist = &uew->objects_list;
	if(clist->total_sel_rows <= 0)
	{
	    UEDialog(
		"Copy Objects",
		UE_ICON_WARNING,
		"No objects selected.",
		uew->toplevel
	    );
	    return;
	}

	obj_num = uew->last_selected_object;
	UEWPropsDoSetValues(uew, obj_num);

	UEClipboardClear(ue_clipboard);
	objects_coppied = 0;
	obj = NULL;
	n = clist->total_sel_rows;
	for(i = 0; i < n; i++)
	{
	    row_num = clist->sel_row[i];
	    obj_num = row_num;
	    obj = SW_OBJECT(CListRowGetData(clist, row_num));
	    if(obj == NULL)
		continue;

	    UEClipboardAppendObject(ue_clipboard, obj);
	    objects_coppied++;
	}

	if((objects_coppied == 1) && (obj != NULL))
	{
	    const char *name = UNVGetObjectFormalName(obj_num, obj);
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"Coppied object `%s' to the clipboard.",
		name
	    );
	    UEWStatus(uew, s);
	    free(s);
	}
	else
	{
	    char *s = (char *)malloc(40 + 80);
	    sprintf(
		s,
"Coppied %i objects.",
		objects_coppied
	    );
	    UEWStatus(uew, s);
	    free(s);
	}
}

/*
 *	Pastes the object(s) from the clipboard.
 */
void UEWPaste(uew_struct *uew)
{
	int i, total, objects_pasted;
	char *s;
	xsw_object_struct **list, *obj;

	if(uew == NULL)
	    return;

	/* Get the objects from the clipboard */
	list = UEClipboardPasteObjects(ue_clipboard, &total);
	if((list == NULL) || (total <= 0))
	{
	    UEDialog(
		"Paste Objects",
		UE_ICON_INFO,
		"The clipboard is empty.",
		uew->toplevel
	    );
	    free(list);
	    return;
	}

	/* Paste the objects */
	objects_pasted = 0;
	for(i = total - 1; i >= 0; i--)
	{
	    obj = list[i];
	    if(obj == NULL)
		continue;

	    UEWInsertObject(uew, obj, True);
	    UNVDeleteObject(obj);
	    objects_pasted++;
	}
	free(list);

	s = (char *)malloc(40 + 80);
	sprintf(
	    s,
"Pasted %i %s.",
	    objects_pasted,
	    (objects_pasted == 1) ? "object" : "objects"
	);
	UEWStatus(uew, s);
	free(s);
}


/*
 *	Remodel the selected object(s) to the selected OPM.
 */
void UEWRemodel(uew_struct *uew)
{
	int i, n, row_num, obj_num, objects_remodeled = 0;
	char *s;
	clist_struct *clist;
	popup_list_struct *pulist;
	xsw_object_struct *obj, *robj, *opm;

	if(uew == NULL)
	    return;

	pulist = &uew->opm_pulist;
	opm = SW_OBJECT(PUListItemGetData(
	    pulist,
	    PUListGetSelItem(pulist)
	));
	clist = &uew->objects_list;
	n = clist->total_sel_rows;

	if(opm == NULL)
	{
	    UEDialog(
		"Remodel Object",
		UE_ICON_WARNING,
		"No OPM selected.",
		uew->toplevel
	    );
	    return;
	}
	if(n <= 0)
	{
	    UEDialog(
		"Remodel Object",
		UE_ICON_WARNING,
		"No objects selected.",
		uew->toplevel
	    );
	    return;
	}

	/* Confirm remodel */
	row_num = (n == 1) ? clist->sel_row[0] : -1;
	obj_num = row_num;
	obj = SW_OBJECT(CListRowGetData(clist, row_num));
	if(obj != NULL)
	{
	    int response;
	    const char *name = UNVGetObjectFormalName(obj_num, obj);
	    char *s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"Remodel object `%s'?",
		name
	    );
	    response = UEGetResponse(           
		"Confirm Remodel",
		UE_ICON_QUESTION,
		s,
		CDlgBtnNo | CDlgBtnYes,
		uew->toplevel
	    );
	    free(s);
	    if(response != CDlgYes)
		return;
	}
	else
	{
	    int response;
	    char *s = (char *)malloc(40 + 80);
	    sprintf(
		s,
"Remodel %i %s?",
		n,
		(n == 1) ? "object" : "objects"
	    );
	    response = UEGetResponse(           
		"Confirm Remodel",
		UE_ICON_QUESTION,
		s,
		CDlgBtnNo | CDlgBtnYes,
		uew->toplevel
	    );
	    free(s);
	    if(response != CDlgYes)
		return;
	}

	/* Remodel the selected objects */
	for(i = 0; i < n; i++)
	{
	    row_num = clist->sel_row[i];
	    obj_num = row_num;
	    obj = SW_OBJECT(CListRowGetData(clist, row_num));
	    if(obj == NULL)
		continue;

	    /* Create a new object from the OPM */
	    robj = UNVDupObject(opm);
	    if(robj == NULL)
		continue;

	    /* Preserve the name, password, position, and direction */
	    strncpy(robj->name, obj->name, sizeof(robj->name));
	    robj->name[sizeof(robj->name) - 1] = '\0';
	    free(robj->password);
	    robj->password = STRDUP(obj->password);
	    robj->sect_x = obj->sect_x;
	    robj->sect_y = obj->sect_y;
	    robj->sect_z = obj->sect_z;
	    robj->x = obj->x;
	    robj->y = obj->y;
	    robj->z = obj->z;
	    robj->heading = obj->heading;
	    robj->pitch = obj->pitch;
	    robj->bank = obj->bank;

	    /* Delete the existing object and set the new object to
	     * replace it
	     */
	    CListRowSetData(
		clist, row_num,
		robj, UEWObjectsCListDestroyCB
	    );

	    obj = NULL;	/* Original object has been deleted */

	    UEWObjectsListItemUpdateObjName(uew, obj_num);

	    if(obj_num == uew->last_selected_object)
		UEWDisplayObject(uew, obj_num, False);

	    /* Update the has changes mark */
	    if(!uew->has_changes)
	    {                              
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }

	    objects_remodeled++;
	}

	s = (char *)malloc(40 + 80);
	sprintf(
	    s,
"Remodeled %i %s.",
	    objects_remodeled,
	    (objects_remodeled == 1) ? "object" : "objects"
	);
	UEWStatus(uew, s);
	free(s);
}

/*
 *	Deletes all the selected objects.
 */
void UEWDeleteObject(uew_struct *uew)
{
	bool_t yes_to_all = False;
	int i, n, row_num, obj_num, objects_deleted = 0;
	clist_struct *clist;
	xsw_object_struct *obj, **list;

	if(uew == NULL)
	    return;

	clist = &uew->objects_list;

	if(clist->total_sel_rows <= 0)
	{
	    UEDialog(
		"Delete Object",
		UE_ICON_WARNING,
		"No objects selected.",
		uew->toplevel
	    );
	    return;
	}

	/* Create a list of objects to be deleted */
	n = clist->total_sel_rows;
	list = (xsw_object_struct **)malloc(
	    n * sizeof(xsw_object_struct *)
	);
	for(i = 0; i < n; i++)
	    list[i] = SW_OBJECT(CListRowGetData(
		clist, clist->sel_row[i]
	    ));

	/* Iterate through the list of objects to be deleted */
	for(i = 0; i < n; i++)
	{
	    obj = list[i];
	    if(obj == NULL)
		continue;

	    row_num = CListRowGetByData(clist, obj);
	    if(row_num < 0)
		continue;

	    obj_num = row_num;

	    /* Confirm delete? */
	    if(!yes_to_all)
	    {
		int response;
		const char *name = UNVGetObjectFormalName(obj_num, obj);
		char *s = (char *)malloc(STRLEN(name) + 80);
		sprintf(
		    s,
"Delete object `%s'?",
		    name
		);
		response = UEGetResponse(
		    "Confirm Delete",
		    UE_ICON_QUESTION,
		    s,
		    CDlgBtnNo | CDlgBtnYes |
		    CDlgBtnYesToAll | CDlgBtnCancel,
		    uew->toplevel
		);
		free(s);
		if(response == CDlgYesToAll)
		{
		    yes_to_all = True;
		}
		else if(response == CDlgYes)
		{

		}
		else if(response == CDlgCancel)
		{
		    break;
		}
		else
		{
		    continue;
		}
	    }

	    /* Need to make sure no other objects own this object */
	    UEWDoChownAll(
		uew,
		obj_num,	/* Previous owner (this object) */
		((uew->legend.lost_found_owner < 0) ?
		    0 : uew->legend.lost_found_owner)
	    );

	    /* Delete this object */
	    CListRowRemove(clist, row_num);
	    objects_deleted++;

	    /* Update the has changes mark */
	    if(!uew->has_changes)
	    {
		uew->has_changes = True;
		UEWUpdateTitle(uew);
	    }
	}

	/* Delete the list of objects to be deleted */
	free(list);

	/* Update all the object numbers due to an object being removed
	 * from the list
	 */
	for(i = 0; i < clist->total_rows; i++)
	    UEWObjectsListItemUpdateObjNum(uew, i);

	if(objects_deleted > 0)
	{
	    char *s = (char *)malloc(40 + 80);
	    sprintf(
		s,
"Deleted %i %s.",
		objects_deleted,
		(objects_deleted == 1) ? "object" : "objects"
	    );
	    UEWStatus(uew, s);
	    free(s);
	}
	else
	{
	    UEWStatus(uew, "No objects deleted.");
	}
}

/*
 *	Check Universe.
 */
void UEWCheckUnv(
	uew_struct *uew, const bool_t verbose, const bool_t automatic
)
{




}

/*
 *	Select OPM File.
 */
void UEWSelectOPMFile(uew_struct *uew)
{
	int npaths;
	char *path, **paths_list;

	if(uew == NULL)
	    return;

	/* Query user for the OPM file to select */
	if(!STRISEMPTY(fname.opm))
	    path = STRDUP(PrefixPaths(
		GetParentDir(fname.opm),
		UE_FILE_FORMAT_EXTENSION_MASK_OPM
	    ));
	else
	    path = STRDUP(PrefixPaths(
		UECurrentDirectory(),
		UE_FILE_FORMAT_EXTENSION_MASK_OPM
	    ));
	if(!UESelectFile(   
	    "Select OPM File",
	    "Select", "Cancel",
	    path, &paths_list, &npaths,
	    uew->toplevel  
	))
	{ 
	    free(path);
	    return;
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);

	UEWSetOPMFile(uew, path);

	free(path);
}

/*
 *	Exports the selected objects to an OPM file.
 */
int UEWExportOPMFile(uew_struct *uew)
{
	int i, row_num, status, npaths, total, nobjs_exported = 0;
	mode_t mode;
	const char *name;
	char *s, *path, **paths_list;
	struct stat stat_buf;
	clist_struct *clist;
	sw_legend_struct *legend;
	xsw_object_struct **list;

	if(uew == NULL)
	    return(-2);

	UEWSetBusy(uew, True);

	legend = &uew->legend;
	clist = &uew->objects_list;

	/* Create a list of selected objects */
	total = clist->total_sel_rows;
	list = (total > 0) ?
	    (xsw_object_struct **)malloc(
		total * sizeof(xsw_object_struct *)
	    ) : NULL;
	for(i = 0; i < total; i++)
	{
	    row_num = clist->sel_row[i];
	    list[i] = SW_OBJECT(CListRowGetData(clist, row_num));
	}

	/* No objects selected? */
	if((list == NULL) || (total <= 0))
	{
	    UEDialog(
		"Export OPM File",
		UE_ICON_WARNING,
		"No objects selected.",
		uew->toplevel
	    );
	    UEWSetBusy(uew, False);
	    free(list);
	    return(-2);
	}

	/* Query user for the OPM file to export to */
	if(!STRISEMPTY(fname.opm))
	    path = STRDUP(PrefixPaths(
		GetParentDir(fname.opm),
		UE_FILE_FORMAT_EXTENSION_MASK_OPM
	    ));
	else
	    path = STRDUP(PrefixPaths(
		UECurrentDirectory(),
		UE_FILE_FORMAT_EXTENSION_MASK_OPM
	    ));
	if(!UESelectFile(
	    "Export OPM File",
	    "Export", "Cancel",
	    path, &paths_list, &npaths,
	    uew->toplevel
	))
	{
	    UEWSetBusy(uew, False);
	    free(list);
	    free(path);
	    return(-4);
	}
	free(path);
	path = STRDUP((npaths > 0) ? paths_list[npaths - 1] : NULL);
	strlistfree(paths_list, npaths);

	if(STRISEMPTY(path))
	{
	    UEWSetBusy(uew, False);
	    free(list);
	    free(path);
	    return(-4);
	}

	name = strrchr(path, DIR_DELIMINATOR);
	if(name != NULL)
	    name++;
	else
	    name = path;

	/* Check if the file already exists */
	if(UECheckFileExists(path, NULL))
	{
	    /* Confirm overwrite */
	    int response;
	    s = (char *)malloc(STRLEN(path) + 80);
	    sprintf(
		s,
"Overwrite existing file:\n\n    %s",
		path
	    );
	    response = UEGetResponse(
		"Confirm Overwrite",
		UE_ICON_WARNING,
		s,
		CDlgBtnNo | CDlgBtnYes,
		uew->toplevel
	    );
	    free(s);
	    if(response != CDlgYes)
	    {
		UEWSetBusy(uew, False);
		free(list);
		free(path);
		return(-4);
	    }
	}

	if(stat(path, &stat_buf))
	{
	    mode = UECurrentFileMode();
	}
	else
	{
	    mode = stat_buf.st_mode;
	}

	/* Export the selected objects to the OPM file which is the
	 * same format as a universe file
	 */
	status = UNVFileSave(
	    path, mode,
	    list, total,
	    legend,
	    UEWUnvFIOProgressCB, uew
	);
	UEWStatusProgress(uew, 0.0f);
	nobjs_exported = total;
	free(list);
	list = NULL;
	total = 0;

	if(status < 0)
	{
	    s = (char *)malloc(STRLEN(name) + 80);
	    sprintf(
		s,
"Error exporting the %s to the file `%s'.",
		(nobjs_exported == 1) ? "object" : "objects",
		name
	    );
	    UEWStatus(uew, s);
	    free(s);

	    s = (char *)malloc(
		STRLEN(path) + (3 * 80)
	    );
	    sprintf(
		s,
"An error occured while exporting the %s\n\
to the OPM file:\n\n\
    %s",
		(nobjs_exported == 1) ? "object" : "objects",
		path
	    );
	    UEDialog(
		"Export OPM File Error",
		UE_ICON_ERROR,
		s,
		uew->toplevel
	    );
	    free(s);

	    UEWSetBusy(uew, False);
	    free(path);

	    return(-1);
	}

	/* Report status */
	s = (char *)malloc(STRLEN(name) + 80);
	sprintf(
	    s,
"Exported %i %s to the file `%s'.",
	    nobjs_exported,
	    (nobjs_exported == 1) ? "object" : "objects",
	    name
	);
	UEWStatus(uew, s);
	free(s);

	if(uew->map_state)
	    UEWQueueDraw(uew, UEW_DRAW_COMPLETE);

	UEWSetBusy(uew, False);

	free(path);

	return(status);
}
