#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <cmath>

#include "../include/osw-x.h"
#include "../include/widget.h"
#include "../include/objects.h"
#include "../include/unvutil.h"

#include "uew.h"
#include "ue.h"


bool_t UEWViewConvertPositionToWindowCoordinates(
	const unsigned long wwidth, const unsigned long wheight,
	const float sect_width, const float sect_height,
	const long cur_sect_x, const long cur_sect_y,
	const float cur_x, const float cur_y,
	const long sect_x, const long sect_y,
	const float x, const float y,
	const float zoom,
	int *wx_rtn, int *wy_rtn
);

int UEWViewMatchObject(
	uew_struct *uew,
	const int x, const int y,
	const int start_obj_num
);
void UEWViewStatusGetValues(
	uew_struct *uew, xsw_object_struct *obj
);

static void UEWViewDrawObject(
	uew_struct *uew, drawable_t d,
	const unsigned int width, const unsigned int height,
	sw_legend_struct *legend,
	const long sect_x_min, const long sect_x_max,
	const long sect_y_min, const long sect_y_max,
	const long cur_sect_x, const long cur_sect_y,
	const float cur_x, const float cur_y,
	const int obj_num, xsw_object_struct *obj,
	const float zoom,
	const bool_t draw_printer_friendly,
	const bool_t draw_greyscale
);
void UEWViewDrawPixmap(
	uew_struct *uew, pixmap_t pixmap,
	win_t w,				/* View window */
	const bool_t draw_printer_friendly,	/* Draw printer friendly version */
	const bool_t draw_greyscale		/* Draw greyscale */
);
void UEWViewDraw(uew_struct *uew);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define FTOI_RNDOFF(f)	((int)((f) + 0.5f))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Converts the object position in real units to window
 *	coordinates.
 *
 *	The wwidth and wheight specify the size of the window in pixels.
 *
 *	The sect_width and sect_height specify the size of a sector in
 *	real units.
 *
 *	The cur_sect_x and cur_sect_y specify the current sector
 *	supposedly displayed on the window.
 *
 *	The cur_x and cur_y specify the current center position in
 *	the current sector in real units supposedly displayed on the
 *	window.
 *
 *	The sect_x and sect_y specify the sector.
 *
 *	The x and y specify the position in real units to convert.
 *
 *	The zoom specifies the zoom of the window.
 */
bool_t UEWViewConvertPositionToWindowCoordinates(
	const unsigned long wwidth, const unsigned long wheight,
	const float sect_width, const float sect_height,
	const long cur_sect_x, const long cur_sect_y,
	const float cur_x, const float cur_y,
	const long sect_x, const long sect_y,
	const float x, const float y,
	const float zoom,
	int *wx_rtn, int *wy_rtn
)
{
	/* Calculate the sector deltas */
	const long	d_sect_x = sect_x - cur_sect_x,
			d_sect_y = sect_y - cur_sect_y;

	int	wx = 0,
		wy = 0;

	/* Calculate X coordinate position in units of pixels */
	if(d_sect_x == 0)
	{
	    wx = FTOI_RNDOFF((x - cur_x) * zoom);
	}
	else if(d_sect_x < 0)
	{
	    wx = FTOI_RNDOFF(
		(((sect_width / 2) - x) +
		((d_sect_x + 1) * sect_width * -1) +
		(cur_x + (sect_width / 2))) * zoom * -1
	    );
	}
	else if(d_sect_x > 0)
	{
	    wx = FTOI_RNDOFF(
		(((sect_width / 2) - cur_x) +
		((d_sect_x - 1) * sect_width) +
		(x + (sect_width / 2))) * zoom
	    );
	}
	wx += (int)wwidth / 2;

	/* Calculate Y coordinate position in units of pixels */
	if(d_sect_y == 0)
	{
	    wy = FTOI_RNDOFF((y - cur_y) * zoom);
	}
	else if(d_sect_y < 0)
	{
	    wy = FTOI_RNDOFF(
		(((sect_height / 2) - y) +
		((d_sect_y + 1) * sect_height * -1) +
		(cur_y + (sect_height / 2))) * zoom * -1
	    );
	}
	else if(d_sect_y > 0)
	{
	    wy = FTOI_RNDOFF(
		(((sect_height / 2) - cur_y) +
		((d_sect_y - 1) * sect_height) +
		(y + (sect_height / 2))) * zoom
	    );
	}
	wy = ((int)wheight / 2) - wy;

	if(wx_rtn != NULL)
	    *wx_rtn = wx;
	if(wy_rtn != NULL)
	    *wy_rtn = wy;

	return(True);
}

/*
 *	Matches object on uew, only returns match.
 *
 *	Does not actually select the object.
 */
int UEWViewMatchObject(
	uew_struct *uew, 
	const int x, const int y,
	const int start_obj_num
)
{
	ue_option_struct *opt = &option;
	const int marker_radius = opt->object_marker_radius;
	int i, total, obj_x = 0, obj_y = 0, sw, sh, swv, shv;
	unsigned int width, height;
	long sect_x_min, sect_x_max, sect_y_min, sect_y_max;
	float zoom;
	win_t w;
	clist_struct *clist;
	sw_legend_struct *legend;
	xsw_object_struct *obj;

	if(uew == NULL)
	    return(-2);

	legend = &uew->legend;
	zoom = CLIP(uew->zoom, UEW_VIEW_ZOOM_MIN, UEW_VIEW_ZOOM_MAX);
	w = uew->view;
	clist = &uew->objects_list;
	if(w == None)
	    return(-2);

	OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);


	/* Calculate the zoomed width of a sector */
	sw = MAX(FTOI_RNDOFF(legend->sect_x_len * zoom), 1);

	/* Calculate the number of sectors visible on the width */
	swv = (sw > 0) ? ((int)width / sw) : 0;

	/* Calculate the zoomed height of a sector */
	sh = MAX(FTOI_RNDOFF(legend->sect_y_len * zoom), 1);

	/* Calculate the number of sectors visible on the height */
	shv = (sh > 0) ? ((int)height / sh) : 0;

	/* Calculate sector bounds and add "good measure" */
	sect_x_min = uew->sect_x - (long)swv - 1;
	sect_x_max = uew->sect_x + (long)swv + 1;
	sect_y_min = uew->sect_y - (long)shv - 1;
	sect_y_max = uew->sect_y + (long)shv + 1;

	/* Iterate through the objects list */
	total = clist->total_rows;
	for(i = MAX(start_obj_num, 0); i < total; i++)
	{
	    obj = SW_OBJECT(CListRowGetData(clist, i));
	    if(obj == NULL)
		continue;

	    if(obj->type <= SW_OBJ_TYPE_GARBAGE)
		continue;

	    /* Not in the same sector as the clicked on sector? */
	    if((obj->sect_x < sect_x_min) ||
	       (obj->sect_x > sect_x_max) ||
	       (obj->sect_y < sect_y_min) ||
	       (obj->sect_y > sect_y_max)
	    )
		continue;

	    /* Calculate the center of the object in window coordinates */
	    UEWViewConvertPositionToWindowCoordinates(
		width, height,
		legend->sect_x_len, legend->sect_y_len,
		uew->sect_x, uew->sect_y,
		uew->cx, uew->cy,
		obj->sect_x, obj->sect_y,
		obj->x, obj->y,
		zoom,
		&obj_x, &obj_y
	    );

	    /* Check if x and y are within object's coordinates
	     * in pixels
	     */
	    if((x >= (obj_x - marker_radius)) &&
	       (x < (obj_x + marker_radius)) &&
	       (y >= (obj_y - marker_radius)) &&
	       (y < (obj_y + marker_radius))
	    )
		return(i);
	}

	return(-1);
}


/*
 *	Sets the prompt values on view status to the values of the
 *	specified object.
 *
 *	If the object is NULL then the prompts will be cleared.
 */
void UEWViewStatusGetValues(
	uew_struct *uew, xsw_object_struct *obj
)
{
	bool_t obj_is_valid;

	if(uew == NULL)
	    return;

	if(!uew->map_state)
	    return;

	/* Check if the object is valid */
	if(obj == NULL)
	    obj_is_valid = False;
	else if(obj->type <= SW_OBJ_TYPE_GARBAGE)
	    obj_is_valid = False;
	else
	    obj_is_valid = True;

	if(obj_is_valid)
	{
	    PromptSetF2(&uew->obj_cx_prompt, obj->x, 2);
	    PromptSetF2(&uew->obj_cy_prompt, obj->y, 2);
	    PromptSetF2(&uew->obj_cz_prompt, obj->z, 2);

	    PromptSetL(&uew->obj_sect_x_prompt, obj->sect_x);
	    PromptSetL(&uew->obj_sect_y_prompt, obj->sect_y);
	    PromptSetL(&uew->obj_sect_z_prompt, obj->sect_z);

	    PromptSetS(&uew->obj_name_prompt, obj->name);

	    PromptSetF2(
		&uew->obj_heading_prompt,
		obj->heading * (180 / PI),
		2
	    );

	    PromptSetL(&uew->obj_size_prompt, obj->size); 
	}
	else
	{
	    PromptSetS(&uew->obj_cx_prompt, "");
	    PromptSetS(&uew->obj_cy_prompt, "");
	    PromptSetS(&uew->obj_cz_prompt, "");

	    PromptSetS(&uew->obj_sect_x_prompt, "");
	    PromptSetS(&uew->obj_sect_y_prompt, "");
	    PromptSetS(&uew->obj_sect_z_prompt, "");

	    PromptSetS(&uew->obj_name_prompt, "");

	    PromptSetS(&uew->obj_heading_prompt, "");
	    PromptSetS(&uew->obj_size_prompt, "");
	}
}


/*
 *	Draws the object.
 */
static void UEWViewDrawObject(
	uew_struct *uew, drawable_t d,
	const unsigned int width, const unsigned int height,
	sw_legend_struct *legend,
	const long sect_x_min, const long sect_x_max,
	const long sect_y_min, const long sect_y_max,
	const long cur_sect_x, const long cur_sect_y,
	const float cur_x, const float cur_y,
	const int obj_num, xsw_object_struct *obj,
	const float zoom,
	const bool_t draw_printer_friendly,
	const bool_t draw_greyscale
)
{
	ue_option_struct *opt = &option;
	const int marker_radius = opt->object_marker_radius;
	int type, x, y;
	osw_gui_struct *gui = &osw_gui[0];

	if(obj == NULL)
	    return;

	type = obj->type;
	if(type <= SW_OBJ_TYPE_GARBAGE)
	    return;

	/* Is this object out of the sector bounds so that we do not
	 * need to draw it?
	 */
	if((obj->sect_x < sect_x_min) ||
	   (obj->sect_x > sect_x_max) ||
	   (obj->sect_y < sect_y_min) ||
	   (obj->sect_y > sect_y_max)
	)
	    return;

	/* Calculate the center of the object in window coordinates */
	UEWViewConvertPositionToWindowCoordinates(
	    width, height,
	    legend->sect_x_len, legend->sect_y_len,
	    cur_sect_x, cur_sect_y,
	    cur_x, cur_y,
	    obj->sect_x, obj->sect_y,
	    obj->x, obj->y,
	    zoom,
	    &x, &y
	);

	/* Draw the object marker */
	if(draw_greyscale)
	{
	    OSWSetForeground(gui->black_pix);
	}
	else
	{
	    if(UEWIsObjectSelected(uew, obj_num))
		OSWSetForeground(ue_color.view_obj_selected);
	    else if(type == SW_OBJ_TYPE_DYNAMIC)
		OSWSetForeground(ue_color.view_obj_dynamic);
	    else if(type == SW_OBJ_TYPE_CONTROLLED)
		OSWSetForeground(ue_color.view_obj_controlled);
	    else if(type == SW_OBJ_TYPE_PLAYER)
		OSWSetForeground(ue_color.view_obj_player);
	    else if(type == SW_OBJ_TYPE_HOME)
		OSWSetForeground(ue_color.view_obj_home);
	    else if(type == SW_OBJ_TYPE_AREA)
		OSWSetForeground(ue_color.view_obj_area);
	    else if(type == SW_OBJ_TYPE_WORMHOLE)
		OSWSetForeground(ue_color.view_obj_wormhole);
	    else if(type == SW_OBJ_TYPE_ELINK)
		OSWSetForeground(ue_color.view_obj_wormhole);
	    else
		OSWSetForeground(ue_color.view_obj_static);
	}

	/* Draw contact bounds? */
	if(opt->show_contact_bounds)
	{
	    const int radius = (int)((obj->size / 1000l) * zoom);
	    if(radius > marker_radius)
	    {
		const int diameter = 2 * radius;
		OSWDrawArc(
		    d,
		    x - radius, y - radius,
		    diameter, diameter,
		    0.0f * PI, 2.0f * PI
		);
	    }
	}

	/* Draw the object's marker */
	OSWDrawArc(
	    d,
	    x - marker_radius, y - marker_radius,
	    2 * marker_radius, 2 * marker_radius,
	    0.0f * PI, 2.0f * PI
	);

	/* Draw the object's heading */
	switch(type)
	{
	  case SW_OBJ_TYPE_DYNAMIC:
	  case SW_OBJ_TYPE_CONTROLLED:
	  case SW_OBJ_TYPE_WEAPON:
	  case SW_OBJ_TYPE_PLAYER:
	  case SW_OBJ_TYPE_WORMHOLE:
	    {
		const int marker_len = 2 * marker_radius;
		int	x2 = x + FTOI_RNDOFF(sin(obj->heading) * marker_len),
			y2 = y - FTOI_RNDOFF(cos(obj->heading) * marker_len);
		OSWDrawLine(
		    d,
		    x, y,
		    x2, y2
		);
	    }
	    break;
	}

	/* Draw the object's label */
	if(opt->show_object_labels)
	{
	    int     x2, y2,
		    lbearing, rbearing, swidth, ascent, descent,
		    font_height;
	    font_t *font = ue_font.object_label;
	    const char *name = UNVGetObjectFormalName(obj_num, obj);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, name, -1,
		&lbearing, &rbearing, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    x2 = x - (swidth / 2);
	    y2 = y - marker_radius - 2 - font_height;
	    OSWDrawString(
		d,
		x2 - lbearing,
		y2 + ascent,
		name
	    );
	}     

}

/*
 *	Draws the view on the specified pixmap.
 *
 *	The w specifies the view window, which should be None if
 *	printer_friendly is True in order to draw the printer friendly
 *	version.
 *
 *	The printer_friendly specifies that the view is to be drawn
 *	with a white background with greyscale sector grids and object
 *	markers.
 *
 *	The draw_greyscale specifies that the view is to be drawn
 *	in greyscale (no color).
 */
void UEWViewDrawPixmap(
	uew_struct *uew, pixmap_t pixmap,
	win_t w,				/* View window */
	const bool_t draw_printer_friendly,	/* Draw printer friendly version */
	const bool_t draw_greyscale		/* Draw greyscale */
)
{
	int	i, obj_num, total,
		x, y,
		sw, sh, swv, shv,
		pos, inc;
	unsigned int width, height;
	long	sect_cur,
		sect_x_min, sect_x_max, sect_y_min, sect_y_max;
	float zoom;
	drawable_t d;
	clist_struct *clist;
	sw_legend_struct *legend;
	xsw_object_struct *obj;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;
	ue_option_struct *opt = &option;

	if(uew == NULL)
	    return;

	if(pixmap != None)
	{
	    OSWGetPixmapGeometry(pixmap, &width, &height, NULL);
	    d = pixmap;
	}
	else if(w != None)
	{
	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	    d = w;
	}
	else
	{
	    return;
	}

	legend = &uew->legend;
	zoom = CLIP(uew->zoom, UEW_VIEW_ZOOM_MIN, UEW_VIEW_ZOOM_MAX);
	clist = &uew->objects_list;

	/* Draw the background */
	if(draw_printer_friendly)
	    OSWSetForeground(gui->white_pix);
	else
	    OSWSetForeground(
		draw_greyscale ?
		    gui->black_pix :
		    ue_color.view_bg
	    );
	OSWDrawSolidRectangle(d, 0, 0, width, height);


	/* Draw the minor (not sector) grids */
	if(draw_printer_friendly)
	{
	    if(ue_color.nview_grids > 0)
		OSWSetForeground(
		    draw_greyscale ?
			gui->black_pix :
			ue_color.view_grid[ue_color.nview_grids - 1]
		);
	    else
		OSWSetForeground(gui->black_pix);
	}
	else
	{
	    if(ue_color.nview_grids > 0)
		OSWSetForeground(
		    draw_greyscale ?
			gui->white_pix :
			ue_color.view_grid[ue_color.nview_grids / 2]
		);
	    else
		OSWSetForeground(gui->white_pix);
	}

	inc = FTOI_RNDOFF(opt->grid_spacing * zoom);
	if(opt->show_grid && (inc >= 20))
	{
	    /* Vertical grids */
	    swv = (int)width / inc;
	    x = FTOI_RNDOFF(
		(((opt->grid_spacing / 2) - uew->cx) * zoom) +
		((int)width / 2)
	    );
	    for(pos = x % inc; pos < (int)width; pos += inc)
	    {
		if(pos < 0)
		    continue;

		OSWDrawLine(
		    d,
		    pos, 0,
		    pos, (int)height
		);
	    }

	    /* Horizontal grids */
	    shv = (int)height / inc;
	    y = FTOI_RNDOFF(
		(((opt->grid_spacing / 2) - uew->cy) * zoom) +
		((int)height / 2)
	    );
	    for(pos = ((int)height - y) % inc;
		pos < (int)height;
		pos += inc
	    )
	    {
		if(pos < 0)
		    continue;

		OSWDrawLine(
		    d, 
		    0, pos,
		    (int)width, pos
		);
	    }
	}

	/* Draw sector grids? */

	/* Calculate the zoomed width of a sector */
	sw = MAX(FTOI_RNDOFF(legend->sect_x_len * zoom), 1);

	/* Calculate the number of sectors visible along the width */
	swv = MAX(((sw > 0) ? ((int)width / sw) : 1), 1);

	/* Calculate the zoomed height of a sector */
	sh = MAX(FTOI_RNDOFF(legend->sect_y_len * zoom), 1);

	/* Calculate the number of sectors visible along the height */
	shv = MAX(((sh > 0) ? ((int)height / sh) : 1), 1);

	if(opt->show_grid)
	{
	    /* Draw vertical sector grids */
	    if(draw_printer_friendly)
	    {
		if(ue_color.nview_grids > 0)
		    OSWSetForeground(
			draw_greyscale ?
			    gui->black_pix :
			    ue_color.view_grid[0]
		    );
		else
		    OSWSetForeground(gui->black_pix);
	    }
	    else
	    {
		if(ue_color.nview_grids > 0)
		    OSWSetForeground(
			draw_greyscale ?
			    gui->white_pix :
			    ue_color.view_grid[ue_color.nview_grids - 1]
		    );
		else
		    OSWSetForeground(gui->white_pix);
	    }

	    /* Calculate the position of the starting grid */
	    x = FTOI_RNDOFF(
	        (((legend->sect_x_len / 2) - uew->cx) * zoom) +
	        ((int)width / 2) - (swv * sw)
	    );
	    sect_cur = uew->sect_x - swv + 1;

	    for(pos = x, inc = sw; pos < (int)width; pos += inc, sect_cur++)
	    {
	        if(pos < 0)
		    continue;

	        OSWDrawLine(
		    d,
		    pos, 0,
		    pos, (int)height
		);
	    }


	    /* Draw horizontal sector grids */

	    /* Calculate the position of the starting grid */
	    y = FTOI_RNDOFF(
	        ((int)height -
		    ((((legend->sect_y_len / 2) - uew->cy) * zoom) +
		    ((int)height / 2))
	        ) - (shv * sh)
	    );
	    sect_cur = uew->sect_y + shv;

	    for(pos = y, inc = sh; pos < (int)height; pos += inc, sect_cur--)
	    {
		if(pos < 0)
		    continue;

		OSWDrawLine(
		    d,
		    0, pos,
		    (int)width, pos
		);
	    }
	}


	/* Draw objects */

	/* Calculate sector bounds and add `good measure' */
	sect_x_min = uew->sect_x - (long)swv - 1;
	sect_x_max = uew->sect_x + (long)swv + 1;
	sect_y_min = uew->sect_y - (long)shv - 1;
	sect_y_max = uew->sect_y + (long)shv + 1;

	/* Draw unselected objects */
	total = clist->total_rows;
	for(i = 0; i < total; i++)
	{
	    obj = SW_OBJECT(CListRowGetData(clist, i));
	    obj_num = i;

	    if(UEWIsObjectSelected(uew, obj_num))
		continue;

	    UEWViewDrawObject(
		uew, d,
		width, height,
		legend,
		sect_x_min, sect_x_max,
		sect_y_min, sect_y_max,
		uew->sect_x, uew->sect_y,
		uew->cx, uew->cy,
		obj_num, obj,
		zoom, draw_printer_friendly, draw_greyscale
	    );
	}
	/* Draw selected objects */
	for(i = 0; i < clist->total_sel_rows; i++)
	{
	    obj_num = clist->sel_row[i];
	    obj = SW_OBJECT(CListRowGetData(clist, obj_num));
	    UEWViewDrawObject(
		uew, d,  
		width, height,
		legend,
		sect_x_min, sect_x_max, 
		sect_y_min, sect_y_max,
		uew->sect_x, uew->sect_y,
		uew->cx, uew->cy,
		obj_num, obj,
		zoom, draw_printer_friendly, draw_greyscale
	    );
	}

	/* Draw the cursor (cross hairs) */
	if(draw_printer_friendly)
	    OSWSetForeground(
		draw_greyscale ?
		    gui->black_pix :
		    ue_color.view_cursor
	    );
	else
	    OSWSetForeground(
		draw_greyscale ?
		    gui->white_pix :
		    ue_color.view_cursor  
	    );
	x = ((int)width / 2) - 7;
	y = (int)height / 2;
	OSWDrawLine(
	    d,
	    x, y,
	    x + 14, y
	);
	x = (int)width / 2;
	y = ((int)height / 2) - 7;
	OSWDrawLine(
	    d,
	    x, y,
	    x, y + 14
	);

	/* Draw coordinate text */
	if(opt->show_geometry_labels)
	{
	    int lbearing, swidth, ascent, descent, font_height;
	    const float displayed_zoom = (zoom - UEW_VIEW_ZOOM_MIN) /
		(UEW_VIEW_ZOOM_MAX - UEW_VIEW_ZOOM_MIN);
	    font_t *font = ue_font.geometry_label;
	    char s[40 + 40 + 40 + 40 + 40 + 80];
	    sprintf(
		s,
	        "Sect: %ld, %ld  XY: %.2f, %.2f  Zoom: %i%%",
		uew->sect_x, uew->sect_y,
		uew->cx, uew->cy,
		FTOI_RNDOFF(displayed_zoom * 100.0f)
	    );
	    if(draw_printer_friendly)
		OSWSetForeground(
		    draw_greyscale ?
			gui->black_pix :
			ue_color.view_label
		);
	    else
		OSWSetForeground(
		    draw_greyscale ?
			gui->white_pix :
			ue_color.view_label  
		);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    x = 10;
	    y = (int)height - font_height - 10;
	    OSWDrawString(
		d,
		x - lbearing,
		y + ascent,
		s
	    );
	}

	/* Draw the select rectangular line? */
	if((uew->drag_op == UEW_DRAG_OP_VIEW_SELECT_RECTANGULAR) &&
	   (w != None)
	)
	{
	    int	start_x = uew->button_last_x,
		start_y = uew->button_last_y;

	    OSWGetPointerCoords(w, NULL, NULL, &x, &y);
	    if(start_x > x)
	    {
		const int t = x;
		x = start_x;
		start_x = t;
	    }
	    if(start_y > y)
	    {
		const int t = y;
		y = start_y;
		start_y = t;
	    }

	    OSWSetForeground(ue_color.view_cursor);
	    OSWDrawRectangle(
		d,
		start_x,
		start_y,
		(unsigned int)MAX((x - start_x - 1), 1),
		(unsigned int)MAX((y - start_y - 1), 1)
	    );
	}

	/* Draw the measure line? */
	if((uew->drag_op == UEW_DRAG_OP_VIEW_MEASURE) &&
	   (w != None)
	)
	{
	    const int	start_x = uew->button_last_x,
			start_y = uew->button_last_y;
	    int lbearing, swidth, ascent, descent, font_height;
	    float d_ru;
	    char s[40 + 40 + 40 + 40 + 40 + 80];
	    font_t *font = ue_font.geometry_label;

	    OSWGetPointerCoords(w, NULL, NULL, &x, &y);
	    OSWSetForeground(ue_color.view_cursor);
	    OSWDrawLine(d, start_x, start_y, x, y);

	    d_ru = hypot(
		(float)(x - start_x) / zoom,
		(float)(y - start_y) / zoom
	    );
	    sprintf(
		s,
		"Dist: %.2f au (%.2f ru)",
		d_ru * legend->ru_to_au, d_ru
	    );
	    OSWSetForeground(ue_color.view_label);
	    OSWSetFont(font);
	    OSWGetTextBounds(
		font, s, -1,
		&lbearing, NULL, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    x = x - (swidth * x / (int)width);
	    if(y >= start_y)
		y = y + 2;
	    else
		y = y - 2 - font_height;
	    OSWDrawString(
		d,
		x - lbearing,
		y + ascent,
		s
	    );
	}

	/* Draw the DND focus frame?? */
	if((w == uew->dnd_tar_w) && (w != None))
	{
	    OSWSetForeground(wglobal->surface_selected_pix);
	    OSWDrawRectangle(d, 0, 0, width - 1, height - 1);
	}
}

/*
 *	Draws view window on uew n.
 */
void UEWViewDraw(uew_struct *uew)
{
	pixmap_t pixmap;
	win_t w;

	if(uew == NULL)
	    return;

	w = uew->view;
	if(w == None)
	    return;

	/* Recreate the view pixmap as needed */
	if(uew->view_buf == None)
	{
	    unsigned int width, height;
	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );
	    OSWCreatePixmap(&uew->view_buf, width, height);
	}

	pixmap = uew->view_buf;

	/* Draw the view pixmap */
	UEWViewDrawPixmap(uew, pixmap, w, False, False);

	/* Put the view pixmap to the view window */
	OSWPutDrawableToWindow(w, pixmap);
}
