#include <stdio.h>
#include <sys/types.h>
#include <cmath>

#include "../include/reality.h"
#include "../include/objects.h"

#include "../include/unvmatch.h"
#include "../include/unvmath.h"


/* Conversion */
float RADTODEG(const float r);
float DEGTORAD(const float d);
float SANITIZERADIANS(const float r);
float SANITIZEDEGREES(const float d);

float SWConvertRUToAU(sw_units_struct *units, const float x);
float SWConvertRUToKM(sw_units_struct *units, const float x);
float SWConvertRUToMiles(sw_units_struct *units, const float x);
float SWConvertRUToKiloMiles(sw_units_struct *units, const float x);
float SWConvertRUToLY(sw_units_struct *units, const float x);
float SWConvertRUToParsecs(sw_units_struct *units, const float x);
float SWConvertRUToMP(sw_units_struct *units, const float x);
float SWConvertRUToSU(sw_units_struct *units, const float x);

float SWConvertRUPerCycleToRUPerSecond(
	sw_units_struct *units, const float x
);
float SWConvertRUPerCycleToAUPerSecond(
	sw_units_struct *units, const float x
);
float SWConvertRUPerCycleToKMPerSecond(
	sw_units_struct *units, const float x
);
float SWConvertRUPerCycleToMilesPerSecond(
	sw_units_struct *units, const float x
);
float SWConvertRUPerCycleToKiloMilesPerSecond(
	sw_units_struct *units, const float x
);
float SWConvertRUPerCycleToMPPerSecond(
	sw_units_struct *units, const float x
);

/* Polar To Coordinate Conversion */
float SWMathRectangularToPolarX(const float input_angle, const float input_radius);
float SWMathRectangularToPolarY(const float input_angle, const float input_radius);

/* Vector Unitlization */
void SWMathSetUnitVector2D(
	sw_vector_compoent_struct *v,
	const float bearing
);
void SWMathSetUnitVector3D(
	sw_vector_compoent_struct *v,
	const float bearing, const float pitch
);

/* Angle & Vectors Calculations */
float SWMathAngleDelta(const float theta_to, const float theta_cur);
void SWMathVectorAdd(
	const float v1_dir, const float v1_mag,
	const float v2_dir, const float v2_mag,
	float *v_dir_rtn, float *v_mag_rtn
);
float SWMathCoordinateDeltaVector(const float dx, const float dy);

/* 3D Distance */
float SWMath3DDistance(
	const float delta_x, const float delta_y, const float delta_z
);

/* Sector Checking */
int SWMath3DInSectorPtr(
	xsw_object_struct *obj_ptr,
	const long sect_x, const long sect_y, const long sect_z
);
int SWMath3DInSameSector(
	xsw_object_struct **xsw_object, const int total_objects,
	const int obj_num1, const int obj_num2
);
int SWMath3DInSameSectorPtr(
	xsw_object_struct *obj1, xsw_object_struct *obj2
);

/* Spherical Contact */
int SWMath3DInContact(
	xsw_object_struct **xsw_object, const int total_objects,
	const int obj_num1, const int obj_num2
);
int SWMath3DInContactPtr(
	xsw_object_struct *obj1, xsw_object_struct *obj2
);

/* Vector Contact */
int SWMath3DInVectorContact(
	xsw_object_struct **xsw_object, const int total_objects,
	const int obj_num1, const int obj_num2,
	const float heading,
	const float heading_variance,
	const float range
);

/* Range Contact */
int SWMath3DInRange(
	xsw_object_struct **xsw_object, const int total_objects,
	const int obj_num, const int tar_obj_num,
	const float distance		/* In real units */
);
int SWMath3DInRangePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj,
	const float distance		/* In real units */
);

/* Maximum range by velocity and time */
float SWMathMaxRangeByVelocity(const float velocity_max, const time_t t_ms);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define ABSOLUTE(x)	(((x) < 0.0f) ? -(x) : (x))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Converts radians to degrees.
 */
float RADTODEG(const float r)
{
	return(r * 180.0f / PI);
}

/*
 *	Converts degrees to radians.
 */        
float DEGTORAD(const float d)
{
	return(d * PI / 180.0f);
}

/*
 *	Sanitizes radian value to [(0.0  * PI), (2.0 * PI)].
 */
float SANITIZERADIANS(const float r)
{
	float _r = r;
	while(_r >= (2.0f * PI))
	    _r -= (2.0f * PI);
	while(_r < (0.0f * PI))
	    _r += (2.0f * PI);
	return(_r);
}

/*
 *      Sanitizes degree value to [0', 360'].
 */
float SANITIZEDEGREES(const float d)
{
	float _d = d;
	while(_d > 360.0f)
	    _d -= 360.0f;
	while(_d < 0.0f)
	    _d += 360.0f;
	return(_d);
}


/*
 *	Convert SW Real units to Astronomical units.
 */
float SWConvertRUToAU(sw_units_struct *units, const float x)
{
	if(units != NULL)
	    return(x * units->ru_to_au);
	else
	    return(0.0f);
}

/*
 *	Convert SW Real units to kilometers.
 */
float SWConvertRUToKM(sw_units_struct *units, const float x)
{
	if(units != NULL)
	    return(x * units->ru_to_au * 149597870.0f);
	else
	    return(0.0f);
}

/*
 *	Convert SW Real units to miles.
 */
float SWConvertRUToMiles(sw_units_struct *units, const float x)
{
        if(units != NULL)
            return(x * units->ru_to_au * 240724344.0f);
        else
            return(0.0f);
}

/*
 *	Convert SW Real units to kilomiles.
 */
float SWConvertRUToKiloMiles(sw_units_struct *units, const float x)
{
        if(units != NULL)
            return(x * units->ru_to_au * 240724.344f);
        else
            return(0.0f);
}

/*
 *	Convert SW Real units to light years.
 */
float SWConvertRUToLY(sw_units_struct *units, const float x)
{
	if(units != NULL)
	    return(x * units->ru_to_au / 63271.47239f);
	else
	    return(0.0f);
}

/*
 *	Convert SW Real units to Parsecs.
 */
float SWConvertRUToParsecs(sw_units_struct *units, const float x)
{
	if(units != NULL)
	    return(x * units->ru_to_au / 206265.0f);
	else
	    return(0.0f);
}

/*
 *	Convert SW Real units to MilliParsecs.
 */
float SWConvertRUToMP(sw_units_struct *units, const float x)
{
	if(units != NULL)
	    return(x * units->ru_to_au / 206.265f);
	else
	    return(0.0f);
}


/*
 *	Convert SW Real units to screen units.
 */
float SWConvertRUToSU(sw_units_struct *units, const float x)
{
	return(x * 1000.0f);
}


/*
 *	Convert SW Real units per cycle to SW Real units per second.
 */
float SWConvertRUPerCycleToRUPerSecond(
	sw_units_struct *units, const float x
)
{
	return(
	    x * (1000.0f / (float)SW_CYCLE_LAPSE_MS)
	);
}

/*
 *	Convert SW Real units per cycle to Astronomical units per
 *	second.
 */
float SWConvertRUPerCycleToAUPerSecond(
	sw_units_struct *units, const float x
)
{
	return(
	    SWConvertRUToAU(units, x) *
	    (1000.0f / (float)SW_CYCLE_LAPSE_MS)
	);
}

/*
 *	Convert SW Real units per cycle to kilometers per second.
 */
float SWConvertRUPerCycleToKMPerSecond(
	sw_units_struct *units, const float x
)
{
	return(
	    SWConvertRUToKM(units, x) *
	    (1000.0f / (float)SW_CYCLE_LAPSE_MS)
	);
}

/*
 *	Convert SW Real units per cycle to miles per second.
 */
float SWConvertRUPerCycleToMilesPerSecond(
	sw_units_struct *units, const float x
)
{
	return(
	    SWConvertRUToMiles(units, x) *
	    (1000.0f / (float)SW_CYCLE_LAPSE_MS)
	);
}

/*
 *	Convert SW Real units per cycle to kilomiles per second.
 */
float SWConvertRUPerCycleToKiloMilesPerSecond(
	sw_units_struct *units, const float x
)
{
	return(
	    SWConvertRUToKiloMiles(units, x) *
	    (1000.0f / (float)SW_CYCLE_LAPSE_MS)
	);
}

/*
 *	Convert SW Real units per cycle to MilliParsecs per second.
 */
float SWConvertRUPerCycleToMPPerSecond(
	sw_units_struct *units, const float x
)
{
	return(
	    SWConvertRUToMP(units, x) *
	    (1000.0f / (float)SW_CYCLE_LAPSE_MS)
	);
}


/*
 *	Returns the X axis coefficient of the given polar bearing
 *	coordinates.
 */
float SWMathRectangularToPolarX(const float theta, const float r)
{
	return(
	    r *
	    (float)cos((double)(SANITIZERADIANS((0.5f * PI) - theta)))
	);
}

/*
 *	Returns the Y axis coefficient of the given polar bearing
 *	coordinates.
 */
float SWMathRectangularToPolarY(const float theta, const float r)
{
	return(
	    r *
	    (float)sin((double)(SANITIZERADIANS((0.5f * PI) - theta)))
	);
}


/* Not sure, but it's probable that the following two functions
 * can be replaced with:
 *
 * Given a vector, normalize it to a unit vector.
 * It works the same for 2D and 3D vectors as member k
 * will remain 0 if that is the current value.
 *
void MuNormalizeVector3D(xsw_vector_component_struct *v)
{
	float	Magnitude = v->i*v->i + v->j*v->j + v->k*v->k;
	if(Mag == 0.0f)
		return;		// Vector has no length so no normalization possible.

	Magnitude = 1.0f / (float)sqrt((double)MAX(Magnitude, 0.0f));

	v->i *= Magnitude;
	v->j *= Magnitude;
	v->k *= Magnitude;
}

 */

/*
 *	Sets vector compoent structure members i and j to
 *	2D unit vector compoents.  Member k is set to 0.
 */
void SWMathSetUnitVector2D(
	sw_vector_compoent_struct *v,
	const float bearing
)
{
	if(v == NULL)
	    return;

	v->i = (float)sin((double)bearing);
	v->j = (float)cos((double)bearing);
	v->k = 0.0f;
}

/*
 *	Sets vector compoent structure members i, j, and k to
 *	3D unit vector compoents.
 */
void SWMathSetUnitVector3D(
	sw_vector_compoent_struct *v,
	const float bearing, const float pitch
)
{
	float cos_pitch;

	if(v == NULL)
	    return;

	cos_pitch = (float)cos((double)pitch);

	v->i = (float)sin((double)bearing) * cos_pitch;
	v->j = (float)cos((double)bearing) * cos_pitch;
	v->k = (float)sin((double)pitch) * -1.0f;
}


/* Not sure what the next function is used for but a proper
 * vector addition is simple:

void MuAddVector(sw_vector_compoent_struct* res, sw_vector_compoent_struct* v1, sw_vector_compoent_struct* v2)
{
	res->i = v1->i + v2->i;
	res->j = v1->j + v2->j;
	res->k = v1->k + v2->k;
}
 */


/* 
 *	Returns the shorter delta of theta_to minus theta_cur in
 *	radians.
 */
float SWMathAngleDelta(const float theta_to, const float theta_cur)
{
	const float delta = theta_to - theta_cur;

	/* If delta was calculated to be going the longer opposite
	 * way then change it to go in the shorter correct way
	 */
	if(delta > (1.0f * PI))
	    return(delta - (2.0f * PI));
	else if(delta < (-1.0f * PI))
	    return(delta + (2.0f * PI));
	else
	    return(delta);
}

/*
 *	Adds two vectors togeather, setting the new vector's
 *	direction and magnitude values in v_dir_rtn and v_mag_rtn
 *	(respectivly).
 */
void SWMathVectorAdd(
	const float v1_dir, const float v1_mag,
	const float v2_dir, const float v2_mag,
	float *v_dir_rtn, float *v_mag_rtn
)
{
	float x1, y1, x2, y2, xr, yr;

	if((v_dir_rtn == NULL) || (v_mag_rtn == NULL))
	    return;

	/* Calculate right angle displacements */
	x1 = (float)sin((double)v1_dir) * ABSOLUTE(v1_mag);
	y1 = (float)cos((double)v1_dir) * ABSOLUTE(v1_mag);
 
	x2 = (float)sin((double)v2_dir) * ABSOLUTE(v2_mag);
	y2 = (float)cos((double)v2_dir) * ABSOLUTE(v2_mag);

	/* Add vectors togeather */
	xr = x1 + x2;
	yr = y1 + y2;

	*v_mag_rtn = (float)hypot((double)xr, (double)yr);
	*v_dir_rtn = SWMathCoordinateDeltaVector(xr, yr);
}

/*
 *	Calculates vector of the unitless delta coordinates.
 */
float SWMathCoordinateDeltaVector(const float dx, const float dy)
{
	return(SANITIZERADIANS(
	    (0.5f * PI) - (float)atan2((double)dy, (double)dx)
	));
}


/*
 *	Calculates the vector length (the vector magnitude).
 */
float SWMath3DDistance(const float dx, const float dy, const float dz)
{
	return((float)sqrt(
	    (double)MAX((dx * dx) + (dy * dy) + (dz * dz), 0.0f)
	));
}


/* Often useful optimization for comparisons.
 *
 * Calculate the squared length or magnitude of a vector.
 *
 */
float MuMagnitude2Vector(sw_vector_compoent_struct *v)
{
	return((v->i * v->i) + (v->j * v->j) + (v->k * v->k));
}

/* The above can be used to optimize distance measurement
 * and comparison code.  Example:

sw_vector_compoent_struct v1;
sw_vector_compoent_struct v2;

	// Initialize the vectors here..

	// Perform distance comparison.
	if( SWMath3DDistance(v1) > SWMath3DDistance(v2) )

	// Is the same as:
	if( MuMagnitude2Vector(v1) > MuMagnitude2Vector(v2) )

 * The difference between the two versions is that the
 * second version will save you 2 square roots which are expensive.
 */	


/*
 *	Checks if the object is in the given sector.
 */
int SWMath3DInSectorPtr(
	xsw_object_struct *obj,
	const long sect_x, const long sect_y, const long sect_z
)
{
	if(obj == NULL)
	    return(0);

	if((obj->sect_x == sect_x) &&
	   (obj->sect_y == sect_y) &&
	   (obj->sect_z == sect_z)
	)
	    return(1);
	else
	    return(0);
}

/*
 *	Checks if both objects are valid and in the same sector.
 */
int SWMath3DInSameSector(
	xsw_object_struct **xsw_object, const int total_objects,
	const int obj_num1, const int obj_num2
)
{
	xsw_object_struct *o1_ptr, *o2_ptr;

	/* obj_num1 and obj_num2 must be valid */
	if(UNVIsObjectGarbage(xsw_object, total_objects, obj_num1) ||
	   UNVIsObjectGarbage(xsw_object, total_objects, obj_num2)
	)
	    return(0);

	o1_ptr = xsw_object[obj_num1];
	o2_ptr = xsw_object[obj_num2];

	return(
	    (o1_ptr->sect_x == o2_ptr->sect_x) &&
	    (o1_ptr->sect_y == o2_ptr->sect_y) &&
	    (o1_ptr->sect_z == o2_ptr->sect_z)
	);
}

/*
 *	Checks if the two objects are in the same sector.
 */
int SWMath3DInSameSectorPtr(
	xsw_object_struct *obj1, xsw_object_struct *obj2
)
{
	if((obj1 == NULL) || (obj2 == NULL))
	    return(0);

	if(obj1 == obj2)
	    return(1);

	return(
	    (obj1->sect_x == obj2->sect_x) &&
	    (obj1->sect_y == obj2->sect_y) &&
	    (obj1->sect_z == obj2->sect_z)
	);
}        


/*
 *	Checks if the two objects are in contact, takes size
 *	of each object into account.
 */
int SWMath3DInContact(
	xsw_object_struct **xsw_object, const int total_objects,
	const int obj_num1, const int obj_num2  
)
{
	float d, s1, s2;	/* In screen units */
	xsw_object_struct *o1_ptr, *o2_ptr;


	/* Check if objects are valid and in the same sector */
	if(!SWMath3DInSameSector(
	    xsw_object, total_objects, obj_num1, obj_num2
	))
	    return(0);


	/* Get object pointers */
	o1_ptr = xsw_object[obj_num1];
	o2_ptr = xsw_object[obj_num2];


	/* Both objects must have a size greater than 0 */
	s1 = (float)((o1_ptr->size < 0) ? 0 : o1_ptr->size);
	s2 = (float)((o2_ptr->size < 0) ? 0 : o2_ptr->size);


	/* Get the distance in Screen units */
	d = SWMath3DDistance(
	    o2_ptr->x - o1_ptr->x,
	    o2_ptr->y - o1_ptr->y,
	    o2_ptr->z - o1_ptr->z
	) * 1000.0f;			/* Convert to Screen units */


	/* Is distance less than the two sizes? */
	if(d <= (s1 + s2))
	    return(1);
	else
	    return(0);
}

/*
 *	Checks if the two objects are in the same sector and in contact.
 */
int SWMath3DInContactPtr(
	xsw_object_struct *obj1, xsw_object_struct *obj2
)
{
	float d, s1, s2;	/* In screen units */

	if((obj1 == NULL) || (obj2 == NULL))
	    return(0);

	if((obj1->sect_x != obj2->sect_x) ||
	   (obj1->sect_y != obj2->sect_y) ||
	   (obj1->sect_z != obj2->sect_z)
	)
	    return(0);


	/* Both objects must have a size greater than 0 */
	s1 = ((obj1->size < 0) ? 0 : obj1->size);
	s2 = ((obj2->size < 0) ? 0 : obj2->size);

	/* Get the distance in screen units */
	d = SWMath3DDistance(
	    obj2->x - obj1->x,
	    obj2->y - obj1->y,
	    obj2->z - obj1->z
	) * 1000.0f;

	/* Is distance less than the two sizes? */
	if(d <= (s1 + s2))
	    return(1);
	else
	    return(0);
}


/*
 *	Checks if object 1 is in contact with object 2 in the specified
 *	bearing from object 1 to object 2 within the specified
 *	variance and range in screen units.
 *
 *	The heading is the bearing from object 1 to object 2.
 *
 *	The heading_variance is the absolute variance in radians.
 *
 *	The size of object 1 is not taken into account, however the
 *	size of object 2 is.
 */
int SWMath3DInVectorContact(
	xsw_object_struct **xsw_object, const int total_objects,
	const int obj_num1, const int obj_num2,
	const float heading,		/* Object 1 to object 2 */
	const float heading_variance,
	const float range		/* In screen units */
)
{
	float _heading, _heading_variance;
	float d, s2;			/* In screen units */
	float hdg_res_norm, hdg_res_lower, hdg_res_upper;
	float hdgb_min, hdgb_max;	/* Heading bounds */
	xsw_object_struct *o1_ptr, *o2_ptr;

	/* Check if objects are valid and in the same sector */
	if(!SWMath3DInSameSector(
	    xsw_object, total_objects, obj_num1, obj_num2
	))
	    return(0);

	/* Get object pointers */
	o1_ptr = xsw_object[obj_num1];
	o2_ptr = xsw_object[obj_num2];

	/* Get size of object 2 in screen units */
	s2 = MAX(o2_ptr->size, 0l);

	/* Sanitize heading */
	_heading = SANITIZERADIANS(heading);
	_heading_variance = ABSOLUTE(heading_variance);

	/* Calculate min and max heading bounds */
	hdgb_min = _heading - _heading_variance;
	hdgb_max = _heading + _heading_variance;

	/* Calculate the distance in screen units */
	d = SWMath3DDistance(
	    o2_ptr->x - o1_ptr->x,
	    o2_ptr->y - o1_ptr->y,
	    o2_ptr->z - o1_ptr->z
	) * 1000.0f;

	/* If out of range, then definatly not in contact (do not take
	 * object 1's size into account)
	 */
	if(d > (s2 + ABSOLUTE(range)))
	    return(0);

	/* At this point, target object is within range */

	/* Calculate bearing from object 1 to object 2 */
	hdg_res_norm = SWMathCoordinateDeltaVector(
	    o2_ptr->x - o1_ptr->x,
	    o2_ptr->y - o1_ptr->y
	);
	hdg_res_lower = hdg_res_norm - (2.0f * PI);
	hdg_res_upper = hdg_res_norm + (2.0f * PI);

	/* Check heading result order; normal, upper, lower */
	if((hdg_res_norm >= hdgb_min) &&
	   (hdg_res_norm <= hdgb_max)
	)
	    return(1);
	else if((hdg_res_upper >= hdgb_min) &&
		(hdg_res_upper <= hdgb_max)
	)
	    return(1);
	else if((hdg_res_lower >= hdgb_min) &&
		(hdg_res_lower <= hdgb_max)
	)
	    return(1);
	else
	    return(0);
}

/*
 *	Checks if the two objects are in the same sector and in range,
 *	both objects must be valid.
 *
 *	The distance is in real units and is usually the object's
 *	scanner range.
 */
int SWMath3DInRange(
	xsw_object_struct **xsw_object, const int total_objects,
	const int obj_num, const int tar_obj_num,
	const float distance		/* In real units */
)
{
	float	distance_su,
		distance_apart,
		size1, size2;
	xsw_object_struct *obj, *tar_obj;

	/* Objects are not in the same sector? */
	if(!SWMath3DInSameSector(
	    xsw_object, total_objects,
	    obj_num, tar_obj_num
	))
	    return(0);

	obj = xsw_object[obj_num];
	tar_obj = xsw_object[tar_obj_num];

	/* Convert distance to absolute screen units */
	distance_su = (distance < 0.0f) ?
	    (distance * -1000.0f) : (distance * 1000.0f);

	/* Get the sizes of both objects in screen units */
	size1 = (float)MAX(obj->size, 0l);
	size2 = (float)MAX(tar_obj->size, 0l);

	/* Calculate distance apart in screen units */
	distance_apart = SWMath3DDistance(
	    tar_obj->x - obj->x,
	    tar_obj->y - obj->y,
	    tar_obj->z - obj->z
	) * 1000.0f;

	/* See if distance apart is closer than distance */
	if((distance_apart - (size1 + size2)) <= distance_su)
	    return(1);
	else
	    return(0);
}

/*
 *	Checks if the specified target object is within range of
 *	the specified object.
 *
 *	The distance is in real units and is usually the object's 
 *	scanner range.
 */
int SWMath3DInRangePtr(
	xsw_object_struct *obj, xsw_object_struct *tar_obj,
	const float distance
)
{
	float	distance_su,
		distance_apart,		/* In screen units */
		size1, size2;

	if((obj == NULL) || (tar_obj == NULL))
	    return(0);

	if(obj == tar_obj)
	    return(1);

	/* Objects are not in the same sector? */
	if((obj->sect_x != tar_obj->sect_x) ||
	   (obj->sect_y != tar_obj->sect_y) ||
	   (obj->sect_z != tar_obj->sect_z)
	)
	    return(0);

	/* Convert distance to absolute screen units */
	distance_su = (distance < 0.0f) ?
	    (distance * -1000.0f) : (distance * 1000.0f);

	/* Get the sizes of both objects in screen units */
	size1 = (float)MAX(obj->size, 0l);
	size2 = (float)MAX(tar_obj->size, 0l);

	/* Calculate distance apart in screen units */
	distance_apart = SWMath3DDistance(
	    tar_obj->x - obj->x,
	    tar_obj->y - obj->y,
	    tar_obj->z - obj->z
	) * 1000.0f;

	/* See if distance apart is closer than the specified
	 * distance
	 */
	if((distance_apart - (size1 + size2)) <= distance_su)
	    return(1);
	else
	    return(0);
}

/*
 *	Calculates the `theoretical' maximum range based on
 *	given maximum velocity v_max and time t.
 *
 *	If v_max is in Real units, then the returned value will be
 *	in Real units.  Same goes for XSW Screen units.
 *
 *	The t_ms specifies the length of time in in milliseconds.
 */
float SWMathMaxRangeByVelocity(const float v_max, const time_t t_ms)
{
	return((v_max / (float)SW_CYCLE_LAPSE_MS) * (float)MAX(t_ms, 0l));
}
