#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "../include/xsw_ctype.h"
#include "../include/string.h"
#include "../include/reality.h"
#include "../include/objects.h"

#include "../include/unvmain.h"
#include "../include/unvmath.h"
#include "../include/unvutil.h"
#include "../include/unvmatch.h"
#include "../include/unvfile.h"


int UNVAllocScores(xsw_object_struct *obj);
int UNVAllocObjectWeapons(xsw_object_struct *obj, int total_weapons);
int UNVAllocEco(xsw_object_struct *obj);

const char *UNVGetObjectFormalName(
	const int obj_num, xsw_object_struct *obj
);

xsw_object_struct *UNVDupObject(xsw_object_struct *obj);
void UNVResetObject(xsw_object_struct *obj);
void UNVDeleteObject(xsw_object_struct *obj);
void UNVDeleteAllObjects(xsw_object_struct **list, int total);

void UNVParseLocation(
	const char *s,
	long *sect_x, long *sect_y, long *sect_z,
	float *x, float *y, float *z
);
void UNVLocationFormatString(
	char *s,
	const long *sect_x, const long *sect_y, const long *sect_z,
	const float *x, const float *y, const float *z,
	int len
);

void UNVParseDirection(
	const char *s,
	float *heading, float *pitch, float *bank
);
void UNVDirectionFormatString(
	char *s,
	const float *heading, const float *pitch, const float *bank,
	int len
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : TRUE)


/*
 *	Allocates a score structure (as needed) on the object.
 */
int UNVAllocScores(xsw_object_struct *obj)
{
	if(obj == NULL)
	    return(-1);

	if(obj->score == NULL)
	{
	    obj->score = (xsw_score_struct *)calloc(
		1, sizeof(xsw_score_struct)
	    );
	    if(obj->score == NULL)
		return(-3);
	}

	return(0);
}

/*
 *	Allocates or deletes weapons on object as needed.
 */
int UNVAllocObjectWeapons(xsw_object_struct *obj, int total_weapons)
{
	int i, prev;
	xsw_weapon_struct *wep;

	if(obj == NULL)
	    return(-1);

	/* Delete all weapons? */
	if(total_weapons <= 0)
	{
	    /* Delete all weapons on object */
	    for(i = 0; i < obj->total_weapons; i++)
		free(obj->weapons[i]);
	    free(obj->weapons);
	    obj->weapons = NULL;
	    obj->total_weapons = 0;
	    obj->selected_weapon = -1;
	}
	else
	{
	    /* Total weapons is 1 or greater */

	    /* Sanitize total */
	    if(total_weapons > SW_OBJ_WEAPONS_MAX)
		total_weapons = SW_OBJ_WEAPONS_MAX;

	    /* Sanitize total on object */
	    if(obj->total_weapons < 0)
		obj->total_weapons = 0;

	    /* Record previous total on object */
	    prev = obj->total_weapons;


	    /* Delete weapons? */
	    if(total_weapons < prev)
	    {
		/* Delete weapons on object */
		for(i = total_weapons; i < prev; i++)
		{
		    /* Unselect weapon if it will be deleted */
		    if(obj->selected_weapon == i)
			obj->selected_weapon = -1;

		    free(obj->weapons[i]);
		    obj->weapons[i] = NULL;
		}

		obj->weapons = (xsw_weapon_struct **)realloc(
		    obj->weapons,
		    total_weapons * sizeof(xsw_weapon_struct *)
		);
		if(obj->weapons == NULL)
		{
		    obj->total_weapons = 0;
		    obj->selected_weapon = -1;
		    return(-1);
		}

		/* Set new total on object */
		obj->total_weapons = total_weapons;
	    }
	    /* Allocate more weapons? */
	    else if(total_weapons > prev)
	    {
		/* Allocate more weapons on object */
		obj->weapons = (xsw_weapon_struct **)realloc(
		    obj->weapons,
		    total_weapons * sizeof(xsw_weapon_struct *)
		);
		if(obj->weapons == NULL)
		{
		    obj->total_weapons = 0;
		    obj->selected_weapon = -1;
		    return(-1);
		}

		for(i = prev; i < total_weapons; i++)
		{
		    obj->weapons[i] = (xsw_weapon_struct *)malloc(
			sizeof(xsw_weapon_struct)
		    );
		    if(obj->weapons[i] == NULL)
		    {
			obj->total_weapons = i;
			obj->selected_weapon = -1;
			return(-1);
		    }
		    else
		    {
			/* Reset values */
			wep = obj->weapons[i];

			wep->flags = 0;
			wep->ocs_code = 0;
			wep->emission_type = SW_WEP_EMISSION_STREAM;
			wep->amount = 0;
			wep->max = 0;
			wep->damage_hull = 0.0f;
			wep->damage_shields = 0.0f;
			wep->range = 0l;
			wep->create_power = 0.0f;
			wep->delay = 0l;
			wep->last_used = 0l;

			wep->use_sndref = 0;
			wep->fire_sndref = 0;
			wep->hit_sndref = 0;
			wep->reclaim_sndref = 0;
			*wep->name = '\0';
		    }
		}

		/* Set new total on object */
		obj->total_weapons = total_weapons;
	    }
	}

	/* Sanitize selected weapon */
	if(obj->selected_weapon >= obj->total_weapons)
	    obj->selected_weapon = obj->total_weapons - 1;
	if(obj->selected_weapon < 0)
	    obj->selected_weapon = -1;

	return(0);
}

/*
 *	Allocates an economy structure on the object as needed.
 *
 *	Returns 0 on success or non-zero on error.
 */
int UNVAllocEco(xsw_object_struct *obj)
{
	if(obj == NULL)
	    return(-2);

	if(obj->eco == NULL)
	{
	    obj->eco = (xsw_ecodata_struct *)calloc(
		1, sizeof(xsw_ecodata_struct)
	    );
	    if(obj->eco == NULL)
		return(-3);
	}
	
	return(0);
}

/*
 *	Returns a statically allocated string containing the object's
 *	formal name.
 */
const char *UNVGetObjectFormalName(
	const int obj_num, xsw_object_struct *obj
)
{
	char *s;
	static char name[SW_OBJ_NAME_MAX + 80];

	if(obj == NULL)
	    return(SW_OBJ_NAME_GARBAGE);

	if(obj->type <= SW_OBJ_TYPE_GARBAGE)
	    return(SW_OBJ_NAME_GARBAGE);

	if(obj_num < 0)
	{
	    /* Object index number is invalid, implies garbage */
	    strncpy(name, SW_OBJ_NAME_GARBAGE, sizeof(name));
	    name[sizeof(name) - 1] = '\0';
	    return(name);
	}
	else if(obj->type <= SW_OBJ_TYPE_GARBAGE)
	{
	    /* Object type is garbage */
	    strncpy(name, SW_OBJ_NAME_GARBAGE, sizeof(name));
	    name[sizeof(name) - 1] = '\0';
	    return(name);
	}

	sprintf(
	    name,
	    "%s(#%i",
	    obj->name,
	    obj_num
	);
	s = name;
	while(*s != '\0')
	    s++;

	/* Type */
	switch(obj->type)
	{
	  case SW_OBJ_TYPE_DYNAMIC:
	    *s++ = 'D';
	    break;

	  case SW_OBJ_TYPE_CONTROLLED:
	    *s++ = 'C';
	    break;

	  case SW_OBJ_TYPE_PLAYER:
	    *s++ = 'P';
	    break;

	  case SW_OBJ_TYPE_WEAPON:
	    *s++ = 'W';
	    break;

	  case SW_OBJ_TYPE_STREAMWEAPON:
	    *s++ = 'W';
	    break;

	  case SW_OBJ_TYPE_SPHEREWEAPON:
	    *s++ = 'W';
	    break;

	  case SW_OBJ_TYPE_HOME:
	    *s++ = 'H';
	    break;

	  case SW_OBJ_TYPE_AREA:
	    *s++ = 'A';
	    break;

	  case SW_OBJ_TYPE_ANIMATED:
	    *s++ = 'V';
	    break;

	  case SW_OBJ_TYPE_WORMHOLE:
	    *s++ = 'G';
	    break;

	  case SW_OBJ_TYPE_ELINK:
	    *s++ = 'L';
	    break;

	  default:
	    *s++ = '-';
	    break;
	}

	if(obj->eco != NULL)
	    *s++ = 'E';

	*s++ = ')';

	/* Null terminate */
	*s = '\0';


	return(name);
}


/*
 *	Coppies the object.
 */
xsw_object_struct *UNVDupObject(xsw_object_struct *obj)
{
	xsw_object_struct *new_obj;

	if(obj == NULL)
	    return(NULL);

	new_obj = SW_OBJECT(malloc(sizeof(xsw_object_struct)));
	if(new_obj == NULL)
	    return(NULL);

	/* Copy all values
	 *
	 * Note that this coppies pointers to allocated memory as well,
	 * however we will copy the allocated memory and update the
	 * pointers below as well (make sure that all allocated memory
	 * are coppied and that the pointers are updated)
	 */
	memcpy(new_obj, obj, sizeof(xsw_object_struct));


	/* Copying all allocated memory */

	new_obj->password = STRDUP(obj->password);
	new_obj->description = STRDUP(obj->description);
	new_obj->elink = STRDUP(obj->elink);

	/* Tractored objects */
	if(obj->total_tractored_objects > 0)
	{
	    const int n = obj->total_tractored_objects;
	    new_obj->tractored_object = (int *)malloc(
		n * sizeof(int)
	    );
	    if((new_obj->tractored_object != NULL) &&
	       (obj->tractored_object != NULL)
	    )
	    {
		memcpy(
		    new_obj->tractored_object,
		    obj->tractored_object,
		    n * sizeof(int)  
		);
	    }
	    else
	    {
		new_obj->total_tractored_objects = 0;
	    }
	}

	/* Weapons */
	if((obj->weapons != NULL) && (obj->total_weapons > 0))
	{
	    const int nweapons = obj->total_weapons;

	    new_obj->weapons = (xsw_weapon_struct **)malloc(
		nweapons * sizeof(xsw_weapon_struct *)
	    );
	    if(new_obj->weapons != NULL)
	    {
		int i;
		xsw_weapon_struct *src_wep, *tar_wep;

		for(i = 0; i < nweapons; i++)
		{
		    src_wep = obj->weapons[i];
		    if(src_wep != NULL)
		    {
			tar_wep = SW_WEAPON(malloc(
			    sizeof(xsw_weapon_struct)
			));
			if(tar_wep != NULL)
			    memcpy(
				tar_wep, src_wep,
				sizeof(xsw_weapon_struct)
			    );
		    }
		    else
		    {
			tar_wep = NULL;
		    }
		    new_obj->weapons[i] = tar_wep;
		}
	    }
	    else
	    {
		new_obj->total_weapons = 0;
	    }
	}
	else
	{
	    new_obj->total_weapons = 0;
	    new_obj->weapons = NULL;
	}

	/* Score */
	if(obj->score != NULL)
	{
	    xsw_score_struct *tar_score;
	    new_obj->score = tar_score = SW_SCORE(malloc(
		sizeof(xsw_score_struct)
	    ));
	    if(tar_score != NULL)
		memcpy(tar_score, obj->score, sizeof(xsw_score_struct));
	}

	/* Economy */
	if(obj->eco != NULL)
	{
	    xsw_ecodata_struct *src_eco = obj->eco, *tar_eco;

	    new_obj->eco = tar_eco = SW_ECO_DATA(malloc(
		sizeof(xsw_ecodata_struct)
	    ));
	    if(tar_eco != NULL)
	    {
		const int nproducts = src_eco->total_products;

		memcpy(tar_eco, src_eco, sizeof(xsw_ecodata_struct));

		/* Copy each product */
		if((nproducts > 0) && (src_eco->product != NULL))
		{
		    tar_eco->product = (xsw_ecoproduct_struct **)malloc(
			nproducts * sizeof(xsw_ecoproduct_struct *)
		    );
		    if(tar_eco->product != NULL)
		    {
			int i;
			xsw_ecoproduct_struct *src_prod, *tar_prod;

			for(i = 0; i < nproducts; i++)
			{
			    src_prod = src_eco->product[i];
			    if(src_prod != NULL)
			    {
				tar_prod = SW_ECO_PRODUCT(malloc(
				    sizeof(xsw_ecoproduct_struct)
				));
				if(tar_prod != NULL)
				    memcpy(
					tar_prod, src_prod,
					sizeof(xsw_ecoproduct_struct)
				    );
			    }
			    else
			    {
				tar_prod = NULL;
			    }
			    tar_eco->product[i] = tar_prod;
			}
		    }
		    else
		    {
			tar_eco->total_products = 0;
		    }
		}
		else
		{
		    tar_eco->total_products = 0;
		    tar_eco->product = NULL;
		}
	    }
	}

	return(new_obj);
}

/*
 *	Resets the object's values and deletes all of its resources.
 */
void UNVResetObject(xsw_object_struct *obj)
{
	int i;

	if(obj == NULL)
	    return;

	/* Delete weapons */
	for(i = 0; i < obj->total_weapons; i++)
	    free(obj->weapons[i]);
	free(obj->weapons);
	obj->weapons = NULL;
	obj->total_weapons = 0;
	obj->selected_weapon = -1;
 
	/* Delete tractored objects */
	free(obj->tractored_object);
	obj->tractored_object = NULL;
	obj->total_tractored_objects = 0;

	/* Delete scores */
	free(obj->score);
	obj->score = NULL;

	/* Delete economy */
	if(obj->eco != NULL)
	{
	    xsw_ecodata_struct *eco = obj->eco;

	    for(i = 0; i < eco->total_products; i++)
		free(eco->product[i]);
	    free(eco->product);
	    eco->product = NULL;
	    eco->total_products = 0;

	    free(obj->eco);
	    obj->eco = NULL;
	}

	/* Delete password */
	free(obj->password);
	obj->password = NULL;

	/* Delete description */
	free(obj->description);
	obj->description = NULL;

	/* Delete elink */
	free(obj->elink);
	obj->elink = NULL;

	/* Reset values to that of the unv_garbage_object */
	memcpy(
	    obj, &unv_garbage_object,
	    sizeof(xsw_object_struct)
	);
}

/*
 *	Deletes the object.
 */
void UNVDeleteObject(xsw_object_struct *obj)
{
	if(obj == NULL)
	    return;

	/* Delete the object's resources */
	UNVResetObject(obj);

	free(obj);
}


/*
 *	Deletes the list of objects.
 */
void UNVDeleteAllObjects(xsw_object_struct **list, int total)
{
	int i;

	for(i = 0; i < total; i++)
	    UNVDeleteObject(list[i]);

	free(list);
}


/*
 *	Parses the location string s, changing the pointed to
 *	position variables only if their values are specified in
 *	the string.  Any number of the inputs may be NULL, if s is
 *	NULL, then no operation will be performed.
 *
 *	Format of string s is:
 *
 *		"<sect_x>,<sect_y>,<sect_z>:<x>,<y>,<z>"
 *
 *	Example:
 *
 *		"4,-5,0:34.0,-542.81,-12.2"
 */
void UNVParseLocation(
	const char *s,
	long *sect_x, long *sect_y, long *sect_z,
	float *x, float *y, float *z
)
{
	const char value_deliminator = ',';
	const char dimension_deliminator = ':';

	/* Sector x */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(sect_x != NULL)
		*sect_x = atol(s);

	    /* Seek value deliminator */
	    s = strchr(s, value_deliminator);
	    if(s != NULL)
		s++;
	}

	/* Sector y */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(sect_y != NULL)
		*sect_y = atol(s);

	    /* Seek value deliminator */
	    s = strchr(s, value_deliminator);
	    if(s != NULL)
		s++;
	}

	/* Sector z */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(sect_z != NULL)
		*sect_z = atol(s);  

	    /* Seek dimension deliminator */
	    s = strchr(s, dimension_deliminator);
	    if(s != NULL)
		s++;
	}

	/* X */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(x != NULL)
		*x = atof(s);

	    /* Seek next value deliminator */
	    s = strchr(s, value_deliminator);
	    if(s != NULL)
		s++;
	}

	/* Y */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(y != NULL)
		*y = atof(s);

	    /* Seek next value deliminator */
	    s = strchr(s, value_deliminator);
	    if(s != NULL)
		s++;
	}

	/* Z */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(z != NULL)
		*z = atof(s);

	    /* Seek next value deliminator */
	    s = strchr(s, value_deliminator);
	    if(s != NULL)
		s++;
	}
}

/*
 *	Formats string s in the standard universe location notation
 *	from the given the inputs. If any of the inputs are NULL then 0
 *	will be replaced for that value.
 *
 *	The len specifies the number of bytes that have been allocated
 *	for string s.
 *
 *	See UNVParseLocation() for the string format.
 */
void UNVLocationFormatString(
	char *s,
	const long *sect_x, const long *sect_y, const long *sect_z,
	const float *x, const float *y, const float *z,
	int len
)
{
	const char value_deliminator[] = ", ";
	const char dimension_deliminator[] = " : ";
	const char *sp;
	int cur_len = 0;
	char num_str[80];

	if((s == NULL) || (len < 1))
	    return;

	*s = '\0';      /* Reset string */


	/* Sect X */
	sprintf(
	    num_str, "%ld",
	    (sect_x != NULL) ? *sect_x : 0l
	);
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);

	sp = value_deliminator;
	cur_len += strlen(sp);
	if(cur_len < len)
	    strcat(s, sp);

	/* Sect Y */
	sprintf(
	    num_str, "%ld",
	    (sect_y != NULL) ? *sect_y : 0l
	);
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);

	sp = value_deliminator;
	cur_len += strlen(sp);
	if(cur_len < len)
	    strcat(s, sp);

	/* Sect Z */
	sprintf(
	    num_str, "%ld",
	    (sect_z != NULL) ? *sect_z : 0l
	);
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);

	sp = dimension_deliminator;
	cur_len += strlen(sp);
	if(cur_len < len)
	    strcat(s, sp);


	/* X */
	sprintf(num_str, "%.4f",
	    (x != NULL) ? *x : 0.0f
	);
#if 0
	s2 = num_str;	/* Seek to end and strip 0's */
	while(*s2 != '\0')
	    s2++;
	s2--;
	while(s2 > num_str)
	{
	    if((*s2 != '0') &&
	       (*s2 != '.')
	    )
		break;

	    *s2-- = '\0';
	}
#endif
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);

	sp = value_deliminator;
	cur_len += strlen(sp);
	if(cur_len < len)
	    strcat(s, sp);

	/* Y */
	sprintf(num_str, "%.4f",
	    (y != NULL) ? *y : 0.0f
	);
#if 0
	s2 = num_str;      /* Seek to end and strip 0's */
	while(*s2 != '\0')
	    s2++;
	s2--;
	while(s2 > num_str)
	{
	    if((*s2 != '0') &&
	       (*s2 != '.')
	    )
		break;

	    *s2-- = '\0';
	}
#endif
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);

	sp = value_deliminator;
	cur_len += strlen(sp);
	if(cur_len < len)
	    strcat(s, sp);

	/* Z */
	sprintf(num_str, "%.4f",
	    (z != NULL) ? *z : 0.0f
	);
#if 0
	s2 = num_str;      /* Seek to end and strip 0's */
	while(*s2 != '\0')
	    s2++;
	s2--;
	while(s2 > num_str)
	{
	    if((*s2 != '0') &&
	       (*s2 != '.')   
	    )
		break;

	    *s2-- = '\0';
	}
#endif
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);
}


/*
 *      Parses the direction string s, changing the pointed to
 *      direction variables only if their values are specified in
 *      the string.  Any number of the inputs may be NULL, if s is
 *      NULL, then no operation will be performed.
 *
 *      Format of string s is:
 *
 *              "<heading>,<pitch>,<bank>"
 *
 *      Example:
 *
 *              "2.32,6.22,5.14"
 */     
void UNVParseDirection(
	const char *s,
	float *heading, float *pitch, float *bank
)
{
	const char value_deliminator = ',';

	/* Heading */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(heading != NULL)
		*heading = SANITIZERADIANS(atof(s));

	    /* Seek value deliminator */
	    s = strchr(s, value_deliminator);
	    if(s != NULL)
		s++;
	}

	/* Pitch */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(pitch != NULL)
		*pitch = SANITIZERADIANS(atol(s));

	    /* Seek value deliminator */
	    s = strchr(s, value_deliminator);
	    if(s != NULL)
		s++;
	}

	/* Bank */
	if(s != NULL)
	{
	    while(ISBLANK(*s))
		s++;

	    if(bank != NULL)
		*bank = SANITIZERADIANS(atof(s));

	    /* Seek next value deliminator */
	    s = strchr(s, value_deliminator);
	    if(s != NULL)
		s++;
	}
}

/*
 *	Formats string s in the standard universe direction notation
 *	from the given the inputs. If any of the inputs are NULL then 0
 *	will be replaced for that value. All units are in radians.
 *
 *	The len specifies the number of bytes that have been allocated
 *	for string s.
 *
 *	See UNVParseDirection() for the string format.
 */
void UNVDirectionFormatString(
	char *s,
	const float *heading, const float *pitch, const float *bank,
	int len
)
{
	const char value_deliminator[] = ", ";
	const char *sp;
	int cur_len = 0;
	char num_str[80];

	if((s == NULL) || (len < 1))
	    return; 

	*s = '\0';      /* Reset string */

	/* Heading */
	sprintf(num_str, "%f",
	    (heading != NULL) ? *heading : 0.0f
	);
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);

	sp = value_deliminator;
	cur_len += strlen(sp);
	if(cur_len < len)
	    strcat(s, sp);

	/* Pitch */
	sprintf(num_str, "%f",
	    (pitch != NULL) ? *pitch : 0.0f
	);
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);

	sp = value_deliminator;
	cur_len += strlen(sp);
	if(cur_len < len)
	    strcat(s, sp);

	/* Bank */
	sprintf(num_str, "%f",
	    (bank != NULL) ? *bank : 0.0f
	);
	cur_len += strlen(num_str);
	if(cur_len < len)
	    strcat(s, num_str);
}





