#include <stdio.h>
#include <stdlib.h>
#include "../include/widget.h"


static void ScaleBarDrawCB(void *widget, void *data);

int ScaleBarInit(
	scale_bar_struct *sb, win_t parent,
	int x, int y,
	unsigned int length,
	int style,
	int ticks,
	int orientation,
	float pos_min, float pos_max, float pos,
	bool_t flip_pos
);
void ScaleBarDraw(scale_bar_struct *sb);
void ScaleBarQueueDraw(scale_bar_struct *sb);
int ScaleBarManage(scale_bar_struct *sb, event_t *xevent);
void ScaleBarMap(scale_bar_struct *sb);
void ScaleBarUnmap(scale_bar_struct *sb);
void ScaleBarDestroy(scale_bar_struct *sb);


/*
 *	Sizes (in pixels):
 */
#define SCALEBAR_BAR_MIN_LEN			10
#define SCALEBAR_FLUSHED_HANDLE_LENGTH		36


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Draw callback.
 */
static void ScaleBarDrawCB(void *widget, void *data)
{
	scale_bar_struct *sb = (scale_bar_struct *)widget;
	if(sb == NULL)
	    return;

	if(sb->map_state)
	    ScaleBarDraw(sb);
}


/*
 *	Initializes the scale bar.
 */
int ScaleBarInit(
	scale_bar_struct *sb, win_t parent,
	int x, int y,
	unsigned int length,
	int style,
	int ticks,
	int orientation,
	float pos_min, float pos_max, float pos,
	bool_t flip_pos
)
{
	unsigned int width, height;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if((sb == NULL) || (parent == None))
	    return(-2);

	sb->type = WIDGET_TYPE_SCALE_BAR;
	sb->map_state = False;
	sb->has_focus = False;
	sb->disabled = False;
	sb->visibility_state = VisibilityFullyObscured;
	sb->font = wglobal->scale_bar_font;
	sb->next = NULL;
	sb->prev = NULL;
	sb->data = NULL;

	sb->options = ((flip_pos) ? SCALEBAR_FLAG_FLIP_POS : 0);
	sb->orientation = orientation;
	sb->style = style;
	sb->ticks = ticks;

	sb->button_pressed = 0;
	sb->button_last_x = 0;
	sb->button_last_y = 0;
	sb->button_pressed_time = 0l;

	/* Sanitize and set positions */
	if(pos_min > pos_max)
	    pos_min = pos_max;
	sb->pos_min = pos_min;
	sb->pos_max = pos_max;
	sb->pos = CLIP(pos, pos_min, pos_max);

	sb->bkg_buf = None;

	if((int)length < SCALEBAR_BAR_MIN_LEN)
	    length = SCALEBAR_BAR_MIN_LEN;
	sb->length = length;

	sb->value_changed_cb = NULL;
	sb->value_changed_data = NULL;

	/* Calculate the geometry */
	if(style == SCALEBAR_STYLE_STANDARD_VALUE)
	{
	    switch(orientation)
	    {
	      case SCALEBAR_ORIENT_VERTICAL:
		width = SCALEBAR_BAR_STANDARD_WIDTH;
		height = length;
		break;
	      default:		/* SCALEBAR_ORIENT_HORIZONTAL */
		width = length;
		height = SCALEBAR_BAR_STANDARD_WIDTH;
		break;
	    }
	}
	else if(style == SCALEBAR_STYLE_FLUSHED)
	{
	    switch(orientation)
	    {
	      case SCALEBAR_ORIENT_VERTICAL:
		width = SCALEBAR_BAR_FLUSHED_WIDTH;
		height = length;
		break;
	      default:		/* SCALEBAR_ORIENT_HORIZONTAL */
		width = length;
		height = SCALEBAR_BAR_FLUSHED_WIDTH;
		break;
	    }
	}
	else
	{
	    switch(orientation)
	    {
	      case SCALEBAR_ORIENT_VERTICAL:
		width = SCALEBAR_BAR_STANDARD_WIDTH;
		height = length;
		break;
	      default:		/* SCALEBAR_ORIENT_HORIZONTAL */
		width = length;
		height = SCALEBAR_BAR_STANDARD_WIDTH;
		break;
	    }
	}

	/* Create toplevel */
	OSWCreateWindow(
	    &w, parent,
	    x, y,
	    width, height
	);
	sb->toplevel = w;
	sb->toplevel_buf = None;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    KeyPressMask | KeyReleaseMask |
	    ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask |
	    VisibilityChangeMask
	);
	OSWSetWindowCursor(w, gui->std_cursor);
	sb->x = x;
	sb->y = y;
	sb->width = width;
	sb->height = height;

	/* Add this widget to the regeristry */
	WidgetRegister(sb, sb->type);

	return(0);
}

/*
 *	Draw scale bar.
 */
void ScaleBarDraw(scale_bar_struct *sb)
{
	int x, y, m, tick_spacing;
	unsigned int width, height;
	float pos, pos_len, handle_pos;
	win_t w;
	pixmap_t pixmap;
	image_t *img;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(sb == NULL)
	    return;

	w = sb->toplevel;
	OSWGetWindowGeometry(
	    w, NULL, NULL, &width, &height, NULL
	);

	/* Recreate buffers as needed */
	if(sb->toplevel_buf == None)
	    OSWCreatePixmap(
		&sb->toplevel_buf,
		width, height
	    );
	if(sb->bkg_buf == None)
	    OSWCreatePixmap(&sb->bkg_buf, width, height);

	/* Map as needed */
	if(!sb->map_state)
	{
	    /* Redraw background */
	    if(sb->style == SCALEBAR_STYLE_FLUSHED)
	    {
		/* Draw background */
		pixmap = sb->bkg_buf;
		OSWClearPixmap(
		    pixmap,
		    width, height,
		    (wglobal->force_mono) ? gui->black_pix :
			wglobal->scroll_bar_bg_pix
		);
		if(wglobal->force_mono)
		    OSWSetForeground(gui->white_pix);
		else
		    OSWSetForeground(wglobal->scroll_bar_frame_pix);
		/* Frame */
		OSWDrawRectangle(
		    pixmap,
		    0, 0,
		    (int)width - 1, (int)height - 1
		);
	    }
	    else	/* SCALEBAR_STYLE_STANDARD or SCALEBAR_STYLE_STANDARD_VALUE */
	    {
		/* Draw background */
		pixmap = sb->bkg_buf;
		if(wglobal->force_mono)
		    OSWClearPixmap(
			pixmap, width, height,  
			gui->black_pix
		    );
		else
		    WidgetPutImageTile(
		        pixmap, wglobal->std_bkg_img,
		        width, height
		    );

		/* Slide channel and ticks */
		switch(sb->orientation)
		{
		  case SCALEBAR_ORIENT_VERTICAL:
		    img = wglobal->scalebar_v_img;
		    if(wglobal->force_mono)
		    {
			OSWSetForeground(gui->white_pix);
			OSWDrawRectangle(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    6,
			    10,
			    (int)height - 12
			);
		    }
		    else if(img != NULL)
		    {
			OSWSetForeground(wglobal->surface_editable_pix);
			OSWDrawSolidRectangle(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    (int)img->height / 2,
			    10,
			    (int)height - (int)img->height
			);
			OSWSetForeground(wglobal->surface_shadow_pix);
			/* Top horizontal */
			OSWDrawLine(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    ((int)img->height / 2),
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 5,
			    ((int)img->height / 2)
			);
			OSWDrawLine(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 4,
			    ((int)img->height / 2) + 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 4,
			    ((int)img->height / 2) + 1
			);
			/* Left vertical */
			OSWDrawLine(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    ((int)img->height / 2),
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    ((int)height - ((int)img->height / 2))
			);
			OSWDrawLine(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 4,
			    ((int)img->height / 2) + 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 4,
			    ((int)height - ((int)img->height / 2)) - 1
			);
			OSWSetForeground(wglobal->surface_highlight_pix);
			/* Bottom horizontal */
			OSWDrawLine(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 4,
			    ((int)height - ((int)img->height / 2)),
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 5,
			    ((int)height - ((int)img->height / 2))
			);
			OSWDrawLine(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 3,
			    ((int)height - ((int)img->height / 2)) + 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 4,
			    ((int)height - ((int)img->height / 2)) + 1
			);
			/* Right vertical */
			OSWDrawLine(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 5,
			    ((int)img->height / 2) + 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 5,
			    ((int)height - ((int)img->height / 2))
			);
			OSWDrawLine(
			    pixmap,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 4,
			    ((int)img->height / 2) + 2,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 4,
			    ((int)height - ((int)img->height / 2)) - 1 
			);

			/* Change color for drawing ticks */
			OSWSetForeground(wglobal->normal_text_pix);
		    }

		    /* Ticks */
		    if((sb->ticks >= 2) && (img != NULL))
		    {
		        for(y = (int)img->height / 2,
			    tick_spacing =
				((int)height - (int)img->height) /
				(sb->ticks - 1),
			    m = (int)height -
				((int)img->height / 2);
			    y <= m;
			    y += tick_spacing
			)
		        {
			    OSWDrawLine(
				pixmap,
				0, y,
				4, y
			    );
			    OSWDrawLine(sb->bkg_buf,
				(int)width - 5, y,
				(int)width - 1, y
			    );
			}
		    }
		    break;
	     
		  default:      /* SCALEBAR_ORIENT_HORIZONTAL */
		    img = wglobal->scalebar_h_img;
		    if(wglobal->force_mono)
		    {
			OSWSetForeground(gui->white_pix);
			OSWDrawRectangle(
			    pixmap,
			    6,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    (int)width - 12,
			    10
			);
		    }
		    else if(img != NULL)
		    {
			OSWSetForeground(wglobal->surface_editable_pix);
			OSWDrawSolidRectangle(
			    pixmap,
			    (int)img->width / 2,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    (int)width - (int)img->width,
			    10
			);
			OSWSetForeground(wglobal->surface_shadow_pix);
			/* Top horizontal */
			OSWDrawLine(
			    pixmap,
			    ((int)img->width / 2),
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    ((int)width - ((int)img->width / 2)) - 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5
			);
			OSWDrawLine(
			    pixmap,
			    ((int)img->width / 2) + 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 4,
			    ((int)width - ((int)img->width / 2)) - 2,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 4
			);
			/* Left vertical */
			OSWDrawLine(
			    pixmap,
			    ((int)img->width / 2),
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 5,
			    ((int)img->width / 2),
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 5
			);
			OSWDrawLine(
			    pixmap,
			    ((int)img->width / 2) + 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 4,
			    ((int)img->width / 2) + 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 4
			);

			OSWSetForeground(wglobal->surface_highlight_pix);
			/* Bottom horizontal */
			OSWDrawLine(
			    pixmap,
			    ((int)img->width / 2) + 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 5,
			    ((int)width - ((int)img->width / 2)) - 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 5
			);
			OSWDrawLine(
			    pixmap,
			    ((int)img->width / 2) + 2,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 4,
			    ((int)width - ((int)img->width / 2)) - 2,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 4
			);
			/* Right vertical */
			OSWDrawLine(
			    pixmap,
			    ((int)width - ((int)img->width / 2)) - 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 4,
			    ((int)width - ((int)img->width / 2)) - 1,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 5
			);
			OSWDrawLine(
			    pixmap,
			    ((int)width - ((int)img->width / 2)) - 2,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) - 3,
			    ((int)width - ((int)img->width / 2)) - 2,
			    (SCALEBAR_BAR_STANDARD_WIDTH / 2) + 4
			);

			/* Change color for drawing ticks */
			OSWSetForeground(wglobal->normal_text_pix);
		    }
		    /* Ticks */
		    if((sb->ticks >= 2) && (img != NULL))
		    {
			for(x = (int)img->width / 2,
			    tick_spacing =
				((int)width - (int)img->width) /
				(sb->ticks - 1),
			    m = (int)width -
				((int)img->width / 2);
			    x <= m;
			    x += tick_spacing
			)
			{
			    OSWDrawLine(
				pixmap,
				x, 0,
				x, 4
			    );
			    OSWDrawLine(
				pixmap,
				x, (int)height - 5,
				x, (int)height - 1
			    );
			}
		    }
		    break;
		}
	    }


	    OSWMapRaised(sb->toplevel);

	    sb->map_state = True;
	    sb->visibility_state = VisibilityUnobscured;
	}


	/* Clear toplevel buffer */
	OSWCopyDrawables(
	    sb->toplevel_buf,
	    sb->bkg_buf,
	    width, height
	);

	pixmap = sb->toplevel_buf;

	/* Flushed style */
	if(sb->style == SCALEBAR_STYLE_FLUSHED)
	{
	    /* Calculate positions */
	    pos = sb->pos;
	    pos_len = sb->pos_max - sb->pos_min;
	    if(pos_len == 0) pos_len = 1;

	    if(sb->options & SCALEBAR_FLAG_FLIP_POS)
		pos = pos_len - pos;

	    switch(sb->orientation)
	    {
	      case SCALEBAR_ORIENT_VERTICAL:
		handle_pos = (pos / pos_len) *
		    ((int)height - SCALEBAR_FLUSHED_HANDLE_LENGTH);

		if(wglobal->force_mono)
		    OSWSetForeground(gui->white_pix);
		else
		    OSWSetForeground(wglobal->scroll_bar_handle_pix);
		OSWDrawSolidRectangle(
		    pixmap,
		    1, (int)handle_pos,
		    (int)width - 2, SCALEBAR_FLUSHED_HANDLE_LENGTH
		);

		if(wglobal->force_mono)
		    OSWSetForeground(gui->black_pix);
		else
		    OSWSetForeground(wglobal->scroll_bar_frame_pix);
		OSWDrawLine(
		    pixmap,
		    1, (int)handle_pos,
		    (int)width - 2, (int)handle_pos
		);
		OSWDrawLine(
		    pixmap,
		    1, static_cast<int>(handle_pos + SCALEBAR_FLUSHED_HANDLE_LENGTH),
		    (int)width - 2, static_cast<int>(handle_pos + SCALEBAR_FLUSHED_HANDLE_LENGTH)
		);

		if(wglobal->force_mono)
		    OSWSetForeground(gui->black_pix);
		else
		    OSWSetForeground(wglobal->scroll_bar_bg_pix);
		OSWDrawLine(
		    pixmap,
		    1, static_cast<int>(handle_pos + (SCALEBAR_FLUSHED_HANDLE_LENGTH / 2)),
		    (int)width - 2,
		    static_cast<int>(handle_pos + (SCALEBAR_FLUSHED_HANDLE_LENGTH / 2))
		);
		break;

	      default:      /* SCALEBAR_ORIENT_HORIZONTAL */
		handle_pos = (pos / pos_len) *
		    ((int)width - SCALEBAR_FLUSHED_HANDLE_LENGTH);

		if(wglobal->force_mono)
		    OSWSetForeground(gui->white_pix);
		else
		    OSWSetForeground(wglobal->scroll_bar_handle_pix);
		OSWDrawSolidRectangle(
		    pixmap,
		    static_cast<int>(handle_pos), 1,
		    SCALEBAR_FLUSHED_HANDLE_LENGTH, (int)height - 2
		);

		if(wglobal->force_mono)
		    OSWSetForeground(gui->black_pix);
		else
		    OSWSetForeground(wglobal->scroll_bar_frame_pix);
		OSWDrawLine(
		    pixmap,
		    static_cast<int>(handle_pos), 1,
		    static_cast<int>(handle_pos), (int)height - 2
		);
		OSWDrawLine(
		    pixmap,
		    static_cast<int>(handle_pos + SCALEBAR_FLUSHED_HANDLE_LENGTH), 1,
		    static_cast<int>(handle_pos + SCALEBAR_FLUSHED_HANDLE_LENGTH), (int)height - 2
		);

		if(wglobal->force_mono)
		    OSWSetForeground(gui->black_pix);
		else
		    OSWSetForeground(wglobal->scroll_bar_bg_pix);
		OSWDrawLine(
		    pixmap,
		    static_cast<int>(handle_pos + (SCALEBAR_FLUSHED_HANDLE_LENGTH / 2)), 1,
		    static_cast<int>(handle_pos + (SCALEBAR_FLUSHED_HANDLE_LENGTH / 2)), (int)height - 2
		);
		break;  
	    }
	}
	/* Standard style */
	else	/* SCALEBAR_STYLE_STANDARD or SCALEBAR_STYLE_STANDARD_VALUE */
	{
	    /* Calculate positions */
	    pos = sb->pos - sb->pos_min;
	    pos_len = sb->pos_max - sb->pos_min;
	    if(pos_len == 0) pos_len = 1;

	    if(sb->options & SCALEBAR_FLAG_FLIP_POS)
		pos = pos_len - pos;

	    switch(sb->orientation)
	    {
	      case SCALEBAR_ORIENT_VERTICAL:
		img = wglobal->scalebar_v_img;
		if(img != NULL)
		{
		    handle_pos = (pos / pos_len) *
			((int)height - (int)img->height);
		    WidgetPutImageNormal(
			pixmap,
			img,
			((int)width / 2) - ((int)img->width / 2),
			static_cast<int>(handle_pos),
			True
		    );
		}
		break;

	      default:      /* SCALEBAR_ORIENT_HORIZONTAL */
		img = wglobal->scalebar_h_img;
		if(img != NULL)
		{
		    if(sb->style == SCALEBAR_STYLE_STANDARD_VALUE)
		    {
			char *s = (char *)malloc(40);
		        sprintf(s, "%.2f", pos);
		        if(wglobal->force_mono)
			    OSWSetForeground(gui->white_pix);
		        else
		            OSWSetForeground(wglobal->normal_text_pix);
		        OSWDrawString(
			    pixmap,
			    ((pos / pos_len) > 0.5) ? 10 :
				((int)width / 2) + 6,
			    7,
			    s
		        );
			free(s);
		    }

		    handle_pos = (pos / pos_len) *
			((int)width - (int)img->width);
		    WidgetPutImageNormal(
			pixmap,
			img,
			(int)handle_pos,
			((int)height - (int)img->height) / 2,
			True
		    );
		}
		break;
	    }
	}

	OSWPutDrawableToWindow(w, pixmap);
}

/*
 *	Queues the scale bar to draw.
 */
void ScaleBarQueueDraw(scale_bar_struct *sb)
{
	if(sb == NULL)
	    return;

	WidgetQueueDrawAction(ScaleBarDrawCB, sb, NULL);
}

/*
 *	Manages the scale bar event.
 */
int ScaleBarManage(scale_bar_struct *sb, event_t *event)
{
	int events_handled = 0;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	widget_global_struct *wglobal = &widget_global;

	if((sb == NULL) || (event == NULL))
	    return(events_handled);

	if(!sb->map_state && (event->type != MapNotify))
	    return(0);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == sb->toplevel)
	    {
		if(sb->map_state)
		    ScaleBarQueueDraw(sb);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == sb->toplevel)
	    {
		ScaleBarQueueDraw(sb);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!sb->has_focus)
		return(events_handled); 

	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!sb->has_focus)
		return(events_handled); 

	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    if(w == sb->toplevel)
	    {
		int	x = (int)button->x,
			y = (int)button->y,
			orientation = sb->orientation,
			style = sb->style;
		unsigned int width, height;
		float	pos,
			pos_min = sb->pos_min,
			pos_max = sb->pos_max;

		sb->button_pressed = button->button;

		if(pos_min > pos_max)
		    pos_min = pos_max;

		pos = CLIP(sb->pos, pos_min, pos_max);

		OSWGetWindowGeometry(
		    w, NULL, NULL, &width, &height, NULL
		);

		/* Handle by style */
		if(style == SCALEBAR_STYLE_FLUSHED)
		{
		    /* Flushed style */
		    switch(orientation)
		    {
		      case SCALEBAR_ORIENT_VERTICAL:
			y -= ((SCALEBAR_FLUSHED_HANDLE_LENGTH / 2) + 1);
			if(y < 0) y = 0;
			if(y > ((int)height - SCALEBAR_FLUSHED_HANDLE_LENGTH - 3))
			    y = (int)height - SCALEBAR_FLUSHED_HANDLE_LENGTH - 3;
			sb->pos = (pos_max - pos_min) *
			    ((float)y / (float)((int)height -
			    SCALEBAR_FLUSHED_HANDLE_LENGTH - 3));
			break;

		      default:	/* SCALEBAR_ORIENT_HORIZONTAL */
			x -= ((SCALEBAR_FLUSHED_HANDLE_LENGTH / 2) + 1);
			if(x < 0)
			    x = 0;
			if(x > ((int)width - SCALEBAR_FLUSHED_HANDLE_LENGTH - 2))
			    x = (int)width - SCALEBAR_FLUSHED_HANDLE_LENGTH - 2;
			sb->pos = (pos_max - pos_min) *
			    ((float)x / (float)((int)width -
			    SCALEBAR_FLUSHED_HANDLE_LENGTH - 2));
			break;
		    }
		}
		/* Standard style */
		else
		{
		    image_t *img;
		    switch(orientation)
		    {
		      case SCALEBAR_ORIENT_VERTICAL:
			img = wglobal->scalebar_v_img;
			if(img != NULL)
			{
			    y -= (((int)img->height / 2) + 1);
			    if(y < 0)
				y = 0;
			    if(y > ((int)height - (int)img->height - 3))
				y = (int)height - (int)img->height - 3;
			    sb->pos = (pos_max - pos_min) *
				((float)y / (float)((int)height -
				(int)img->height - 3));
			}
			break;

		      default:	/* SCALEBAR_ORIENT_HORIZONTAL */
			img = wglobal->scalebar_h_img;
			if(img != NULL)
			{
			    x -= (((int)img->width / 2) + 1);
			    if(x < 0)
				x = 0;
			    if(x > ((int)width - (int)img->width - 2))
			    x = (int)width - (int)img->width - 2;
			    sb->pos = (pos_max - pos_min) *
				((float)x / (float)((int)width -
				(int)img->width - 2));
			}
			break;
		    }
		}

		/* Flip position? */
		if(sb->options & SCALEBAR_FLAG_FLIP_POS)
		    sb->pos = (pos_max - pos_min) - sb->pos;

		sb->pos += pos_min;

		/* Call changed function */
		if(sb->value_changed_cb != NULL)
		    sb->value_changed_cb(
			sb,
			sb->pos,
			sb->value_changed_data
		    );

		ScaleBarQueueDraw(sb);
		events_handled++;
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(w == sb->toplevel)
	    {
		if(sb->button_pressed > 0)
		{
		    sb->button_pressed = 0;
		    ScaleBarQueueDraw(sb);
		    events_handled++;
		}
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if((w == sb->toplevel) &&
	       (sb->button_pressed > 0)
	    )
	    {
		int	x = (int)motion->x,
			y = (int)motion->y,
			orientation = sb->orientation,
			style = sb->style;
		unsigned int width, height;
		float	pos,
			pos_min = sb->pos_min,
			pos_max = sb->pos_max;

		if(pos_min > pos_max)
		    pos_min = pos_max;

		pos = CLIP(sb->pos, pos_min, pos_max);

		OSWGetWindowGeometry(
		    w, NULL, NULL, &width, &height, NULL
		);

		/* Handle by style */
		if(style == SCALEBAR_STYLE_FLUSHED)
		{
		    /* Flushed */
		    switch(orientation)
		    {
		      case SCALEBAR_ORIENT_VERTICAL:
			y -= ((SCALEBAR_FLUSHED_HANDLE_LENGTH / 2) + 1);
			if(y < 0)
			    y = 0;
			if(y > ((int)height - SCALEBAR_FLUSHED_HANDLE_LENGTH - 3))
			    y = (int)height - SCALEBAR_FLUSHED_HANDLE_LENGTH - 3;
			sb->pos = (pos_max - pos_min) *
			    ((float)y / (float)((int)height -
			    SCALEBAR_FLUSHED_HANDLE_LENGTH - 3));
			break;

		      default:	/* SCALEBAR_ORIENT_HORIZONTAL */
			x -= ((SCALEBAR_FLUSHED_HANDLE_LENGTH / 2) + 1);
			if(x < 0)
			    x = 0;
			if(x > ((int)width - SCALEBAR_FLUSHED_HANDLE_LENGTH - 2))
			    x = (int)width - SCALEBAR_FLUSHED_HANDLE_LENGTH - 2;
			sb->pos = (pos_max - pos_min) *
			    ((float)x / (float)((int)width -
			    SCALEBAR_FLUSHED_HANDLE_LENGTH - 2));
			break;
		    }
		}
		/* Standard style */
		else
		{
		    image_t *img;
		    switch(orientation)
		    {
		      case SCALEBAR_ORIENT_VERTICAL:
			img = wglobal->scalebar_v_img;
			if(img != NULL)
			{
			    y -= (((int)img->height / 2) + 1);
			    if(y < 0)
				y = 0;
			    if(y > ((int)height - (int)img->height - 3))
				y = (int)height - (int)img->height - 3;
			    sb->pos = (pos_max - pos_min) *
				((float)y / (float)((int)height -
				(int)img->height - 3));
			}
			break;

		      default:	/* SCALEBAR_ORIENT_HORIZONTAL */
			img = wglobal->scalebar_h_img;
			if(img != NULL)
			{
			    x -= (((int)img->width / 2) + 1);
			    if(x < 0) x = 0;
			    if(x > ((int)width - (int)img->width - 2))
			    x = (int)width - (int)img->width - 2;
			    sb->pos = (pos_max - pos_min) *
				((float)x / (float)((int)width -
				(int)img->width - 2));
			}
			break;
		    }
		}

		/* Flip position? */
		if(sb->options & SCALEBAR_FLAG_FLIP_POS)
		    sb->pos = (pos_max - pos_min) - sb->pos;

	        sb->pos += pos_min;

		/* Call value changed function */
		if(sb->value_changed_cb != NULL)
		    sb->value_changed_cb(
			sb,
			sb->pos,
			sb->value_changed_data	/* Data */
		    );

		ScaleBarQueueDraw(sb);
		events_handled++;
	    }
	    break;

	  case VisibilityNotify:
	    if(w == sb->toplevel)
	    {
		sb->visibility_state = event->xvisibility.state;
		events_handled++;
	    }
	    break;
	}

	return(events_handled);
}

/*
 *	Maps the scale bar.
 */
void ScaleBarMap(scale_bar_struct *sb)
{
	if(sb == NULL)
	    return;

	sb->map_state = False;
	ScaleBarDraw(sb);
}

/*
 *	Unmaps the scale bar.
 */
void ScaleBarUnmap(scale_bar_struct *sb)
{
	if(sb == NULL)
	    return;

	OSWUnmapWindow(sb->toplevel);
	sb->map_state = False;
	sb->has_focus = False;
	sb->visibility_state = VisibilityFullyObscured;

	/* Destroy buffers */
	OSWDestroyPixmap(&sb->toplevel_buf);
	OSWDestroyPixmap(&sb->bkg_buf);
}

/*
 *	Destroys the scale bar.
 */
void ScaleBarDestroy(scale_bar_struct *sb)
{
	if(sb == NULL)
	    return;

	OSWDestroyWindow(&sb->toplevel);
	OSWDestroyPixmap(&sb->toplevel_buf);
	OSWDestroyPixmap(&sb->bkg_buf);

	sb->type = WIDGET_TYPE_UNKNOWN;
	sb->map_state = False;
	sb->has_focus = False;
	sb->disabled = False;
	sb->visibility_state = VisibilityFullyObscured;
	sb->x = 0;
	sb->y = 0;
	sb->width = 0;
	sb->height = 0;
	sb->font = NULL;
	sb->next = NULL;
	sb->prev = NULL;

	sb->options = 0;
	sb->orientation = SCALEBAR_ORIENT_HORIZONTAL;
	sb->style = SCALEBAR_STYLE_STANDARD;
	sb->ticks = 0;

	sb->button_pressed = 0;
	sb->button_last_x = 0;
	sb->button_last_y = 0;
	sb->button_pressed_time = 0l;

	sb->pos = 0.0f;
	sb->pos_min = 0.0f;
	sb->pos_max = 0.0f;

	sb->length = 0;

	sb->value_changed_cb = NULL;
	sb->value_changed_data = NULL;
}
