#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef HAVE_LIBXPM
#include <X11/xpm.h>
#endif

#include "../include/graphics.h"
#include "../include/widget.h"


pixel_t WidgetGetPixel(char *clsp);

WCursor *WidgetCreateCursorFromID(cursor_id_t id);
WCursor *WidgetCreateCursorFromFile(
	const char *path,
	int hot_x, int hot_y,
	const WColorStruct *c_fg, 
	const WColorStruct *c_bg
);
WCursor *WidgetCreateCursorFromData(
	char **xpmdata,
	int hot_x, int hot_y,
	const WColorStruct *c_fg,
	const WColorStruct *c_bg
);
void WidgetDestroyCursor(WCursor **wcursor);

void WidgetSetWindowCursor(win_t w, WCursor *wcursor);

void WidgetCenterWindow(win_t w, int relation);
void WidgetCenterWindowToWindow(win_t w, win_t rel_w);

void WidgetResizeImageBuffer(
	unsigned int bits_per_pixel,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tbpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int sbpl
);

void WidgetPutImageTile(
	drawable_t tar_d, image_t *src_img,
	unsigned int tar_width, unsigned int tar_height
);
void WidgetPutPixmapTile(
	drawable_t tar_d, pixmap_t src_pm,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int src_width, unsigned int src_height
);

void WidgetFrameButton(
	win_t w, bool_t state,
	pixel_t fg_pix, pixel_t bg_pix
);
void WidgetFrameButtonPixmap(
	pixmap_t pixmap,
	bool_t state,
	unsigned int width, unsigned int height,
	pixel_t fg_pix, pixel_t bg_pix
);

image_t *WidgetCreateImageText(
	const char *s, font_t *font,
	pixel_t fg_pix, pixel_t bg_pix, bool_t outline
);

void WidgetCopyImage(
	image_t *tar_img, image_t *src_img
);
void WidgetCopyImageToShared(
	shared_image_t *tar_img, image_t *src_img
);

void WidgetPutImageNormal(
	drawable_t d, image_t *img,
	int tar_x, int tar_y,
	bool_t allow_transparency
);
void WidgetPutImageRaised(
	drawable_t d, image_t *img,
	int tar_x, int tar_y,
	unsigned int altitude
);

void WidgetAdjustImageGamma(
	image_t *img,
	float r, float g, float b
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)
#define IS_NUM_ODD(n)	((long)(n) & 1)
#define SUBMIN(a,b)	((a) < (b) ? (a) : 0)


/*
 *      Allocates a pixel given a color specification string.
 *	The pixel is the ID for the allocated color.
 */
pixel_t WidgetGetPixel(char *clsp)
{
	XColor xc;
	osw_gui_struct *gui = &osw_gui[0];

	/* Make sure color spec clsp is valid */
	if(clsp == NULL)
	    return(gui->white_pix);

	/* Parse color */
	memset(&xc, 0x00, sizeof(XColor));
	switch(XParseColor(
	    gui->display,
	    gui->colormap,
	    clsp, &xc
	))
	{
	  case BadColor:
	    fprintf(
		stderr,
		"%s: BadColor.\n",
		clsp
	    );
	    break;
	  case BadValue:
	    fprintf(
		stderr,
		"%s: BadValue.\n",
		clsp
	    );
	    break;
	}

	/* Allocate color */
	if(XAllocColor(gui->display, gui->colormap, &xc) == 0)
	{
	    fprintf(
		stderr,
		"%s: Unable to allocate color.\n",
		clsp
	    );
	    return(gui->white_pix);
	}
  
	/* Return the pixel */
	return(xc.pixel);
}


/*
 *	Creates a new widget cursor from the specified ID.
 *
 *	The ID is a default cursor defined on the system.
 */
WCursor *WidgetCreateCursorFromID(cursor_id_t id)
{
	cursor_t cursor = OSWCursorOpenID(id);
	WCursor *wcursor;

	if(cursor == None)
	    return(NULL);

	/* Allocate the WCursor */
	wcursor = (WCursor *)calloc(1, sizeof(WCursor));
	if(wcursor == NULL)
	{
	    OSWDestroyCursor(&cursor);
	    return(NULL); 
	}

	wcursor->cursor = cursor;

	return(wcursor);
}

/*
 *	Creates a new widget cursor from an XPM file.
 */
WCursor *WidgetCreateCursorFromFile(
	const char *path,
	int hot_x, int hot_y,
	const WColorStruct *c_fg, 
	const WColorStruct *c_bg
)
{
#if defined(HAVE_X) && defined(HAVE_LIBXPM)
	osw_gui_struct *gui = &osw_gui[0];
	int status;
	unsigned int	width, height,
			req_width, req_height;
	char *dpath;
	display_t *dpy = gui->display;
	bitmap_t mask;
	pixmap_t pixmap;
	win_t w = gui->root_win;
	WCursor *wcursor;
	XColor fg_xc, bg_xc;
	XpmAttributes xpmattr;
	struct stat stat_buf;

	if((w == 0) || (path == NULL))
	    return(NULL);

	/* Get the file's stats */
	if(stat(path, &stat_buf))
	{
	    char *s = STRDUP(strerror(errno));
	    if(s == NULL)
		s = strdup("no such file");
	    *s = toupper(*s);
	    fprintf(
		stderr,
		"%s: %s.\n",
		path, s
	    );
	    free(s);
	    return(NULL);
	}

	/* Set up XPM attributes */
	memset(&xpmattr, 0x00, sizeof(XpmAttributes));
	xpmattr.closeness = WIDGET_XPM_DEF_COLOR_CLOSENESS;
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;

	/* We need to load the XPM data to a bitmap in order to
	 * create a cursor, so the depth must be 1 bit
	 */
	xpmattr.depth = 1;

	/* Load the XPM file to a pixmap and mask */
	dpath = strdup(path);
	status = XpmReadFileToPixmap(
	    dpy, w,
	    dpath,
	    &pixmap,
	    &mask,
	    &xpmattr
	);
	free(dpath);
	if(status != XpmSuccess)
	{
	    fprintf(
		stderr,
		"%s: Unable to open XPM image.\n",
		path
	    );
	    return(NULL);
	}

	/* Get size of loaded pixmap and mask */
	width = xpmattr.width;
	height = xpmattr.height;

	/* Get the recommend cursor size */
	if(!XQueryBestCursor(
	    dpy, w,
	    width, height,
	    &req_width, &req_height
	))
	{
	    req_width = width;
	    req_height = height;
	}

	/* Need to reduce size of pixmap and mask? */
	if((width > req_width) || (height > req_height))
	{
	    bitmap_t tmp_mask;
	    pixmap_t tmp_pixmap;
	    gc_val_t gcv;
	    gc_t gc;

	    /* Create resized pixmap and mask */
	    OSWCreateBitmap(&tmp_mask, req_width, req_height);
	    OSWCreatePixmap(&tmp_pixmap, req_width, req_height);

	    /* Copy to the resized pixmap */
	    if(tmp_pixmap != None)
	    {
		XCopyArea(
		    dpy,
		    pixmap,		/* Src */
		    tmp_pixmap,		/* Tar */
		    gui->gc,
		    0, 0,
		    req_width, req_height,
		    0, 0
		);
		OSWDestroyPixmap(&pixmap);
		pixmap = tmp_pixmap;
	    }

	    /* Create gc for the mask and copy to the resized mask */
	    gcv.function = GXcopy;
	    gcv.plane_mask = 1;			/* Monochrome */
	    gcv.foreground = gui->white_pix;
	    gcv.background = gui->black_pix;;
	    gc = XCreateGC(
		dpy,
		mask,
		GCFunction | GCPlaneMask |
		GCForeground | GCBackground,
		&gcv
	    );
	    if((tmp_mask != None) && (gc != None))
	    {
		XCopyArea(
		    dpy,
		    mask,		/* Source */
		    tmp_mask,		/* Target */
		    gc,
		    0, 0,
		    req_width, req_height,
		    0, 0
		);
		OSWDestroyBitmap(&mask);
		mask = tmp_mask;
	    }

	    XFreeGC(dpy, gc);
	}

	/* Set colors */
	if(c_fg != NULL)
	{
	    fg_xc.red	= ((u_int16_t)c_fg->r << 8) | (u_int16_t)c_fg->r;
	    fg_xc.green	= ((u_int16_t)c_fg->g << 8) | (u_int16_t)c_fg->g;
	    fg_xc.blue	= ((u_int16_t)c_fg->b << 8) | (u_int16_t)c_fg->b;
	}
	else
	{
	    fg_xc.red	= 0xffff;
	    fg_xc.green	= 0xffff;
	    fg_xc.blue	= 0xffff;
	}
	if(c_bg != NULL)
	{
	    bg_xc.red	= ((u_int16_t)c_bg->r << 8) | (u_int16_t)c_bg->r;
	    bg_xc.green	= ((u_int16_t)c_bg->g << 8) | (u_int16_t)c_bg->g;
	    bg_xc.blue	= ((u_int16_t)c_bg->b << 8) | (u_int16_t)c_bg->b;
	}
	else
	{
	    bg_xc.red	= 0x0000;
	    bg_xc.green = 0x0000;
	    bg_xc.blue	= 0x0000;
	}

	/* Allocate the WCursor */
	wcursor = (WCursor *)calloc(1, sizeof(WCursor));
	if(wcursor == NULL)
	{
	    OSWDestroyPixmap(&pixmap);
	    OSWDestroyPixmap(&mask);
	    return(NULL);
	}

	/* Create the X cursor */
	wcursor->cursor = XCreatePixmapCursor(
	    dpy,
	    pixmap,
	    mask,
	    &fg_xc,
	    &bg_xc,
	    hot_x, hot_y
	);
	/* Delete the loaded XPM pixmap and mask */
	OSWDestroyPixmap(&pixmap);
	OSWDestroyBitmap(&mask);
	/* Cursor create failed? */
	if(wcursor->cursor == None)
	{
	    free(wcursor);
	    return(NULL);
	}

	/* Set values to newly allocated WCursor */
	wcursor->x = hot_x;
	wcursor->y = hot_y;
	wcursor->width = xpmattr.width;
	wcursor->height = xpmattr.height;

	wcursor->depth = 1;	/* X requires that depth be 1 for cursors */

	if(c_fg != NULL)
	{
	    wcursor->color.r = c_fg->r;
	    wcursor->color.g = c_fg->g;
	    wcursor->color.b = c_fg->b;
	}

	return(wcursor);
#else
#warning WidgetCreateCursorFromFile() is not supported on this target platform
	return(NULL);
#endif
}

/*
 *	Creates a widget cursor from the specified XPM data.
 *
 *	Transparent pixels on the XPM data will be transparent on the
 *	cursor. The first color will be the foreground, the second color
 *	will be the background.
 */
WCursor *WidgetCreateCursorFromData(
	char **xpmdata,
	int hot_x, int hot_y,
	const WColorStruct *c_fg,
	const WColorStruct *c_bg
)
{
#if defined(HAVE_X) && defined(HAVE_LIBXPM)
	osw_gui_struct *gui = &osw_gui[0];
	int status;
	unsigned int	width, height,
			req_width, req_height;
	display_t *dpy = gui->display;
	bitmap_t mask;
	pixmap_t pixmap;
	win_t w = gui->root_win;
	WCursor *wcursor;
	XColor fg_xc, bg_xc;
	XpmAttributes xpmattr;

	if((dpy == NULL) || (w == 0) || (xpmdata == NULL))
	    return(NULL);

	/* Set up XPM attributes */
	memset(&xpmattr, 0x00, sizeof(XpmAttributes));
	xpmattr.closeness = WIDGET_XPM_DEF_COLOR_CLOSENESS;
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
		
	/* We need to load the XPM data to a bitmap in order to
	 * create a cursor, so the depth must be 1 bit
	 */
	xpmattr.depth = 1;

	/* Load the XPM data to a pixmap and mask */
	status = XpmCreatePixmapFromData(
	    dpy, w,
	    xpmdata,
	    &pixmap,
	    &mask,
	    &xpmattr
	);
	if(status != XpmSuccess)
	{
	    fprintf(
		stderr,
"%p: Unable to load XPM data.\n",
		xpmdata
	    );
	    return(NULL);
	}

	/* Get size of loaded pixmap and mask */
	width = xpmattr.width;
	height = xpmattr.height;

	/* Get the recommend cursor size */
	if(!XQueryBestCursor(
	    dpy, w,
	    width, height,
	    &req_width, &req_height
	))
	{
	    req_width = width;
	    req_height = height;
	}

	/* Need to reduce size of pixmap and mask? */
	if((width > req_width) || (height > req_height))
	{
	    bitmap_t tmp_mask;
	    pixmap_t tmp_pixmap;
	    gc_val_t gcv;
	    gc_t gc;

	    /* Create resized pixmap and mask */
	    OSWCreateBitmap(&tmp_mask, req_width, req_height);
	    OSWCreatePixmap(&tmp_pixmap, req_width, req_height);

	    /* Copy to the resized pixmap */
	    if(tmp_pixmap != None)
	    {
		XCopyArea(
		    dpy,
		    pixmap,		/* Src */
		    tmp_pixmap,		/* Tar */
		    gui->gc,
		    0, 0,
		    req_width, req_height,
		    0, 0
		);
		OSWDestroyPixmap(&pixmap);
		pixmap = tmp_pixmap;
	    }


	    /* Create gc for the mask and copy to the resized mask */
	    gcv.function = GXcopy;
	    gcv.plane_mask = 1;                 /* Monochrome */
	    gcv.foreground = gui->white_pix;
	    gcv.background = gui->black_pix;;
	    gc = XCreateGC(
		dpy,
		mask,
		GCFunction | GCPlaneMask |
		GCForeground | GCBackground,
		&gcv
	    );
	    if((tmp_mask != None) && (gc != None))
	    {
		XCopyArea(
		    dpy,
		    mask,		/* Source */
		    tmp_mask,		/* Target */
		    gc,
		    0, 0,
		    req_width, req_height,
		    0, 0
		);
		OSWDestroyBitmap(&mask);
		mask = tmp_mask;
	    }

	    XFreeGC(dpy, gc);
	}

	/* Set colors */
	if(c_fg != NULL)
	{
	    fg_xc.red	= ((u_int16_t)c_fg->r << 8) | (u_int16_t)c_fg->r;
	    fg_xc.green	= ((u_int16_t)c_fg->g << 8) | (u_int16_t)c_fg->g;
	    fg_xc.blue	= ((u_int16_t)c_fg->b << 8) | (u_int16_t)c_fg->b;
	}
	else
	{
	    fg_xc.red	= 0xffff;
	    fg_xc.green	= 0xffff;
	    fg_xc.blue	= 0xffff;
	}
	if(c_bg != NULL)
	{
	    bg_xc.red	= ((u_int16_t)c_bg->r << 8) | (u_int16_t)c_bg->r;
	    bg_xc.green	= ((u_int16_t)c_bg->g << 8) | (u_int16_t)c_bg->g;
	    bg_xc.blue	= ((u_int16_t)c_bg->b << 8) | (u_int16_t)c_bg->b;
	}
	else
	{
	    bg_xc.red	= 0x0000;
	    bg_xc.green	= 0x0000;
	    bg_xc.blue	= 0x0000;
	}

	/* Allocate memory for WCursor */
	wcursor = (WCursor *)calloc(1, sizeof(WCursor));
	if(wcursor == NULL)
	{
	    OSWDestroyPixmap(&pixmap);
	    OSWDestroyBitmap(&mask);
	    return(NULL);
	}

	/* Create cursor */
	wcursor->cursor = XCreatePixmapCursor(
	    dpy,
	    pixmap,
	    mask,
	    &fg_xc,
	    &bg_xc,
	    hot_x, hot_y
	);
	/* Delete the pixmap and mask */
	OSWDestroyPixmap(&pixmap);
	OSWDestroyBitmap(&mask);
	/* Cursor create failed? */
	if(wcursor->cursor == None)
	{
	    free(wcursor);
	    return(NULL);
	}

	/* Set values to newly allocated WCursor */
	wcursor->x = hot_x;
	wcursor->y = hot_y;
	wcursor->width = xpmattr.width;
	wcursor->height = xpmattr.height;

	wcursor->depth = 1;     /* X requires that depth be 1 for cursors */
  
	if(c_fg != NULL)
	{
	    wcursor->color.r = c_fg->r;
	    wcursor->color.g = c_fg->g;
	    wcursor->color.b = c_fg->b;
	}

	return(wcursor);
#else
#warning WidgetCreateCursorFromData() is not supported on this target platform
	return(NULL);
#endif
}


/*
 *	Sets a window to use the given WCursor.
 */
void WidgetSetWindowCursor(win_t w, WCursor *wcursor)
{
	if((w == 0) || (wcursor == NULL))
	    return;

	if(wcursor->cursor == None)
	    return;

	/* Set win_t w to use new WCursor */
	OSWSetWindowCursor(w, wcursor->cursor);
}

/*
 *	Destroys an allocated WCursor.
 */
void WidgetDestroyCursor(WCursor **wcursor)
{
	if(wcursor == NULL)
	    return;

	if(*wcursor == NULL)
	    return;

	/* Delete actual Cursor */
	OSWDestroyCursor(&((*wcursor)->cursor));

	free(*wcursor);
	*wcursor = NULL;
}

/*
 *	Moves window w to the center.
 */
void WidgetCenterWindow(win_t w, int win_rel)
{
	int x, y, pointer_x, pointer_y;
	unsigned int width, height;
	win_t parent;
	osw_gui_struct *gui = &osw_gui[0];

	if(w == None)
	    return;

	switch(win_rel)
	{
	  case WIDGET_WINDOW_RELATIVE_ROOT:
	    WidgetCenterWindowToWindow(w, gui->root_win);
	    break;

	  case WIDGET_WINDOW_RELATIVE_PARENT: 
	    parent = OSWGetWindowParent(w);
	    WidgetCenterWindowToWindow(w, parent);
	    break;

	  case WIDGET_WINDOW_RELATIVE_POINTER:
	    OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	    OSWGetPointerCoords(
		w,
		&pointer_x, &pointer_y,
		&x, &y
	    );

	    x = pointer_x - ((int)width / 2);
	    y = pointer_y - ((int)height / 2);

	    if((x + (int)width) > (int)gui->display_width)
		x = (int)gui->display_width - (int)width;
	    if((y + (int)height) > (int)gui->display_height)
		y = (int)gui->display_height - (int)height;
	    if(x < 0)
		x = 0;
	    if(y < 0)
		y = 0;

	    OSWMoveWindow(w, x, y);
	    break;
	}
}

/*
 *	Centers the window to the relative window.
 */
void WidgetCenterWindowToWindow(win_t w, win_t rel_w)
{
	int x, y, rel_x, rel_y;
	unsigned int width, height, rel_width, rel_height;
	osw_gui_struct *gui = &osw_gui[0];

	if((w == None) || (rel_w == None))
	    return;

	if(!OSWGetWindowRootOrigin(rel_w, &rel_x, &rel_y))
	    return;

	OSWGetWindowGeometry(w, NULL, NULL, &width, &height, NULL);
	OSWGetWindowGeometry(rel_w, NULL, NULL, &rel_width, &rel_height, NULL);

	x = rel_x + (((int)rel_width - (int)width) / 2);
	y = rel_y + (((int)rel_height - (int)height) / 2);
	if((x + (int)width) > (int)gui->display_width)
	    x = (int)gui->display_width - (int)width;
	if(x < 0)
	    x = 0;
	if((y + (int)height) > (int)gui->display_height)
	    y = (int)gui->display_height - (int)height;
	if(y < 0)
	    y = 0;

	OSWMoveWindow(w, x, y);
}


/*
 *	Copy/resizes the source image data to the target image data.
 *
 *	The depth of the two images must be identical.
 */
void WidgetResizeImageBuffer(
	unsigned int bits_per_pixel,
	u_int8_t *tar_buf, const u_int8_t *src_buf,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int tbpl,
	unsigned int src_width, unsigned int src_height,
	unsigned int sbpl
)
{
	const int bpp = bits_per_pixel / 8;
	u_int8_t *tar_ptr;
	const u_int8_t *src_ptr;

	int tar_x_col, tar_y_row;
	int src_x_col, src_y_row;	/* In units of 256 */

	/* In units of 256 */
	int src_dh_x_col_inc, src_dh_y_row_inc;
	int src_dh_width, src_dh_height;

	if((tar_buf == NULL) || (src_buf == NULL) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (src_width == 0) || (src_height == 0) ||
	   (bpp == 0)
	)
	    return;

	/* Calculate bytes per line values as needed */
	if(tbpl == 0)
	    tbpl = tar_width * bpp;
	if(sbpl == 0)
	    sbpl = src_width * bpp;

	src_dh_x_col_inc = (int)((int)src_width * 256) / (int)tar_width;  
	src_dh_y_row_inc = (int)((int)src_height * 256) / (int)tar_height;
	    
	src_dh_width = (int)src_width * 256;  
	src_dh_height = (int)src_height * 256;
	    
	tar_x_col = 0;
	tar_y_row = 0;
	src_x_col = 0;
	src_y_row = 0;

	/* Begin copying source buffer to target buffer */
	while((tar_y_row < (int)tar_height) &&
	      (src_y_row < src_dh_height)
	)
	{
	    /* Get buffer position */
	    src_ptr = src_buf +
		((src_y_row >> 8) * sbpl) +
		((src_x_col >> 8) * bpp);
	    tar_ptr = tar_buf +
		(tar_y_row * tbpl) +
		(tar_x_col * bpp);
	    memcpy(tar_ptr, src_ptr, bpp);

	    /* Increment column positions */
	    tar_x_col += 1;
	    src_x_col += src_dh_x_col_inc;
	
	    /* Go to next line? */
	    if((tar_x_col >= (int)tar_width) ||
	       (src_x_col >= src_dh_width)
	    )
	    {
		tar_x_col = 0;
		src_x_col = 0;
	 
		tar_y_row += 1;
		src_y_row += src_dh_y_row_inc;
		continue;
	    }
	}
}



/*
 *	Tiles an image_t to the drawable_t.
 */
void WidgetPutImageTile(
	drawable_t tar_d, image_t *src_img,
	unsigned int tar_width, unsigned int tar_height
)
{
#if defined(HAVE_X)
	int x, y;
	XImage *ximg;
	osw_gui_struct *gui = &osw_gui[0];

	if((tar_d == None) || (src_img == NULL) ||
	   (tar_width == 0) || (tar_height == 0)
	)
	    return;

	ximg = src_img->ximage;
	if(ximg == NULL)
	    return;

	x = 0;
	y = 0;
	while(y < (int)tar_height)
	{
	    XPutImage(
		gui->display,
		(drawable_t)tar_d,
		gui->gc,
		ximg,
		0, 0,
		x, y,
		ximg->width, ximg->height
	    );

	    x += (int)ximg->width;
	    if(x >= (int)tar_width)
	    {
		x = 0;
		y += (int)ximg->height;
	    }
	}
#else
#warning WidgetPutImageTile() is not supported on this target platform
#endif
}


/*
 *	Tiles a pixmap to a drawable_t.
 */
void WidgetPutPixmapTile(
	drawable_t tar_d, pixmap_t src_pm,
	unsigned int tar_width, unsigned int tar_height,
	unsigned int src_width, unsigned int src_height
)
{
	int x, y;
	osw_gui_struct *gui = &osw_gui[0];

	if((tar_d == 0) || (src_pm == 0) ||
	   (tar_width == 0) || (tar_height == 0) ||
	   (src_width == 0) || (src_height == 0) ||
	   (widget_global.is_init == 0)
	)
	    return;

	x = 0;
	y = 0;
	while(y < (int)tar_height)
	{
	    XCopyArea(
		gui->display,
		(drawable_t)src_pm,
	        (drawable_t)tar_d,
		gui->gc,
		0, 0,
		src_width, src_height,
		x, y
	    );

	    x += (int)src_width;
	    if(x >= (int)tar_width)
	    {
		x = 0;
		y += (int)src_height;
	    }
	}       
}


/*
 *	Draws a frame around the window.
 */
void WidgetFrameButton(
	win_t w, bool_t state,
	pixel_t fg_pix, pixel_t bg_pix  
)
{
	win_attr_t wattr;

	if((w == 0) || (widget_global.is_init == 0))
	    return;

	OSWGetWindowAttributes(w, &wattr);

	if(state)
	{
	    OSWSetForeground(bg_pix);
	    OSWDrawLine((drawable_t)w,
		0, 0,
		0, (int)wattr.height - 1
	    );
	    OSWDrawLine((drawable_t)w,
		0, 0,
		(int)wattr.width - 1, 0
	    );

	    OSWSetForeground(fg_pix);
	    OSWDrawLine((drawable_t)w,
		0, (int)wattr.height - 1,
		(int)wattr.width - 1, (int)wattr.height - 1
	    );
	    OSWDrawLine((drawable_t)w,
		(int)wattr.width - 1, (int)wattr.height - 1,
		(int)wattr.width - 1, 1  
	    );
	}
	else
	{
	    OSWSetForeground(fg_pix);
	    OSWDrawLine((drawable_t)w,
		0, 0,
		0, (int)wattr.height - 1
	    );
	    OSWDrawLine((drawable_t)w,
		0, 0,
		(int)wattr.width - 1, 0
	    );

	    OSWSetForeground(bg_pix);
	    OSWDrawLine((drawable_t)w,
		0, (int)wattr.height - 1,
		(int)wattr.width - 1, (int)wattr.height - 1
	    );
	    OSWDrawLine((drawable_t)w,
		(int)wattr.width - 1, (int)wattr.height - 1,
		(int)wattr.width - 1, 1
	    );
	}
}


/*
 *	Draws a frame around the pixmap, the specified width and
 *	height must be the size of the pixmap.
 */
void WidgetFrameButtonPixmap(
	pixmap_t pixmap,
	bool_t state,
	unsigned int width, unsigned int height,
	pixel_t fg_pix, pixel_t bg_pix
)
{
	if((pixmap == 0) || (width == 0) || (height == 0))
	    return;

	if(state)
	{
	    OSWSetForeground(bg_pix);
	    /* Left vertical */
	    OSWDrawLine((drawable_t)pixmap,
		0, 0,
		0, (int)height - 1
	    );
	    OSWDrawLine((drawable_t)pixmap,
		1, 1,
		1, (int)height - 2
	    );

	    /* Top horizontal */
	    OSWDrawLine((drawable_t)pixmap,
		0, 0,
		(int)width - 1, 0
	    );
	    OSWDrawLine((drawable_t)pixmap,
		1, 1,
		(int)width - 2, 1
	    );

	    OSWSetForeground(fg_pix);
	    /* Bottom horizontal */
	    OSWDrawLine((drawable_t)pixmap,
		1, (int)height - 1,
		(int)width - 1, (int)height - 1
	    );
	    OSWDrawLine((drawable_t)pixmap,
		2, (int)height - 2,
		(int)width - 2, (int)height - 2
	    );

	    /* Right vertical */
	    OSWDrawLine((drawable_t)pixmap,
		(int)width - 1, (int)height - 1,
		(int)width - 1, 1
	    );
	    OSWDrawLine((drawable_t)pixmap,
		(int)width - 2, (int)height - 2,
		(int)width - 2, 2
	    );
	}
	else
	{
	    OSWSetForeground(fg_pix);
	    /* Left vertical */
	    OSWDrawLine((drawable_t)pixmap,
		0, 0,
		0, (int)height - 1
	    );
	    OSWDrawLine((drawable_t)pixmap,
		1, 1,
		1, (int)height - 2
	    );

	    /* Top horizontal */
	    OSWDrawLine((drawable_t)pixmap,
		0, 0,
		(int)width - 1, 0
	    );
	    OSWDrawLine((drawable_t)pixmap,
		1, 1,
		(int)width - 2, 1
	    );

	    OSWSetForeground(bg_pix);
	    /* Bottom horizontal */
	    OSWDrawLine((drawable_t)pixmap,
		0, (int)height - 1,
		(int)width - 1, (int)height - 1
	    );
	    OSWDrawLine((drawable_t)pixmap,
		2, (int)height - 2,
		(int)width - 2, (int)height - 2
	    );

	    /* Right vertical */
	    OSWDrawLine((drawable_t)pixmap,
		(int)width - 1, (int)height - 1,
		(int)width - 1, 1
	    );
	    OSWDrawLine((drawable_t)pixmap,
		(int)width - 2, (int)height - 2,
		(int)width - 2, 2  
	    );
	}
}


/*
 *	Creates an image displaying the specified string.
 */
image_t *WidgetCreateImageText(
	const char *s, font_t *font,
	pixel_t fg_pix, pixel_t bg_pix,
	bool_t outline, pixel_t outline_pix
)
{
	int lbearing, rbearing, swidth, ascent, descent, font_height;
	unsigned int width, height;
	font_t *cur_font;
	image_t *img = NULL;
	pixmap_t pixmap;

	if((s == NULL) || (font == NULL))
	    return(NULL);

	/* Record current font */
	cur_font = OSWQueryCurrentFont();

	OSWGetTextBounds(
	    font, s, -1,
	    &lbearing, &rbearing, &swidth, &ascent, &descent
	);
	font_height = ascent + descent;

	if(outline)
	{
	    width = (unsigned int)(swidth + 4);
	    height = (unsigned int)(font_height + 4);
	    if(IS_NUM_ODD(width))
		width += 1;
	    if(IS_NUM_ODD(height))
		height += 1;

	    if(OSWCreatePixmap(&pixmap, width, height))
		return(NULL);

	    OSWClearPixmap(pixmap, width, height, bg_pix);

	    OSWSetFont(font);

	    /* Draw text outline */
	    OSWSetForeground(outline_pix);
	    OSWDrawString(
		pixmap,
		1 - lbearing,
		0 + ascent,
		s
	    );
	    OSWDrawString(
		pixmap,
		2 - lbearing,
		1 + ascent,
		s
	    );
	    OSWDrawString(   
		pixmap,
		1 - lbearing,
		2 + ascent,
		s
	    );
	    OSWDrawString(   
		pixmap,
		0 - lbearing,
		1 + ascent,
		s
	    );

	    /* Draw text foreground */
	    OSWSetForeground(fg_pix);
	    OSWDrawString(
		pixmap,
		1 - lbearing,
		1 + ascent,
		s
	    );

	    img = OSWGetImage(pixmap, 0, 0, width, height);
	    OSWDestroyPixmap(&pixmap);
	}
	else
	{
	    width = (unsigned int)swidth;
	    height = (unsigned int)font_height;
	    if(IS_NUM_ODD(width))
		width += 1;
	    if(IS_NUM_ODD(height))
		height += 1;

	    if(OSWCreatePixmap(&pixmap, width, height))
		return(NULL);

	    OSWClearPixmap(pixmap, width, height, bg_pix);

	    OSWSetFont(font);

	    /* Draw foreground */
	    OSWSetForeground(fg_pix);
	    OSWDrawString(
		pixmap,
		0 - lbearing,
		0 + ascent,
		s
	    );

	    img = OSWGetImage(pixmap, 0, 0, width, height);
	    OSWDestroyPixmap(&pixmap);
	}

	/* Set back the previous font */
	OSWSetFont(cur_font);

	return(img);
}


/*
 *	Copies the image data from the src_img to the tar_img.
 */
void WidgetCopyImage(image_t *tar_img, image_t *src_img)
{
	int	y, bpp, xob, len,
		tar_bpl, src_bpl;
	unsigned int width, height;
	u_int8_t *tar_data, *tar_ptr;
	const u_int8_t *src_data, *src_ptr;

	if((tar_img == NULL) || (src_img == NULL))
	    return;

	width = MIN((unsigned int)tar_img->width, (unsigned int)src_img->width);
	height = MIN((unsigned int)tar_img->height, (unsigned int)src_img->height);
	bpp = tar_img->bits_per_pixel / 8;
	xob = MAX(tar_img->xoffset, src_img->xoffset) * bpp;
	tar_bpl = tar_img->bytes_per_line;
	src_bpl = src_img->bytes_per_line;
	tar_data = (u_int8_t *)tar_img->data;
	src_data = (const u_int8_t *)src_img->data;   

	len = (width * bpp) - xob;
	if(len <= 0)
	    return;

	for(y = 0; y < (int)height; y++)
	{
	    tar_ptr = tar_data + (y * tar_bpl) + xob;
	    src_ptr = src_data + (y * src_bpl) + xob;
	    memcpy(tar_ptr, src_ptr, len);
	}
}

/*
 *	Copies the image data from the src_img to the tar_img.
 */
void WidgetCopyImageToShared(shared_image_t *tar_img, image_t *src_img)
{
	int     y, bpp, xob, len,
		tar_bpl, src_bpl;
	unsigned int width, height;
	u_int8_t *tar_data, *tar_ptr;
	const u_int8_t *src_data, *src_ptr;

	if((tar_img == NULL) || (src_img == NULL))
	    return;

	width = MIN((unsigned int)tar_img->width, (unsigned int)src_img->width);
	height = MIN((unsigned int)tar_img->height, (unsigned int)src_img->height);
	bpp = tar_img->bits_per_pixel / 8;
	xob = MAX(tar_img->xoffset, src_img->xoffset) * bpp;
	tar_bpl = tar_img->bytes_per_line;
	src_bpl = src_img->bytes_per_line;
	tar_data = (u_int8_t *)tar_img->data;
	src_data = (const u_int8_t *)src_img->data;
	 
	len = (width * bpp) - xob;
	if(len <= 0)
	    return;

	for(y = 0; y < (int)height; y++)
	{
	    tar_ptr = tar_data + (y * tar_bpl) + xob;
	    src_ptr = src_data + (y * src_bpl) + xob;
	    memcpy(tar_ptr, src_ptr, len);
	}                               
}

/*
 *	Puts image_t ximage onto drawable_t d with option to allow
 *	transparency.
 */
void WidgetPutImageNormal(
	drawable_t d,			/* Target */
	image_t *img,			/* Source */
	int tar_x, int tar_y,
	bool_t allow_transparency
)
{
	int	y,
		src_bpp, tar_bpp,
		src_bpl, tar_bpl,
		src_x_offset_bytes, tar_x_offset_bytes;
	unsigned int width, height;
	unsigned long pixel_mask;
	u_int8_t *tar_ptr, *tar_end;
	const u_int8_t *src_ptr, *src_end;
	image_t *tar_img;

	if((d == 0) || (img == NULL))
	    return;

	/* Get target XImage */
	tar_img = OSWGetImage(
	    d,
	    tar_x, tar_y,
	    img->width,
	    img->height
	);
	if(tar_img == NULL)
	    return;

	tar_bpp = tar_img->bits_per_pixel / 8;
	src_bpp = img->bits_per_pixel / 8;
	tar_x_offset_bytes = (int)tar_img->xoffset * tar_bpp;
	src_x_offset_bytes = (int)img->xoffset * src_bpp;
	tar_bpl = tar_img->bytes_per_line;
	src_bpl = img->bytes_per_line;
	width = MIN(img->width, tar_img->width);
	height = MIN(img->height, tar_img->height);
	pixel_mask = img->red_mask | img->green_mask | img->blue_mask;

	for(y = 0; y < (int)height; y++)
	{
	    tar_ptr = (u_int8_t *)tar_img->data + (y * tar_bpl);
	    tar_end = tar_ptr + (width * tar_bpp);
	    src_ptr = (const u_int8_t *)img->data + (y * src_bpl);
	    src_end = src_ptr + (width * src_bpp);
	    tar_ptr += tar_x_offset_bytes;
	    src_ptr += src_x_offset_bytes;
	    while((tar_ptr < tar_end) && (src_ptr < src_end))
	    {
		switch(src_bpp)
		{
		  case 1:
		    if(*(u_int8_t *)src_ptr & (u_int8_t)pixel_mask)
			*(u_int8_t *)tar_ptr = *(u_int8_t *)src_ptr;
		    break;
		  case 2:
		    if(*(u_int16_t *)src_ptr & (u_int16_t)pixel_mask)
			*(u_int16_t *)tar_ptr = *(u_int16_t *)src_ptr;
		    break;
		  case 3:
		    if(src_ptr[0] || src_ptr[1] || src_ptr[2])
			memcpy(tar_ptr, src_ptr, src_bpp);
		    break;
		  case 4:
		    if(*(u_int32_t *)src_ptr & (u_int32_t)pixel_mask)
			*(u_int32_t *)tar_ptr = *(u_int32_t *)src_ptr;
		    break;
		}
		tar_ptr += tar_bpp;
		src_ptr += src_bpp;
	    }
	}

	/* Put image back to d */
	OSWPutImageToDrawableSect(
	    tar_img, d,
	    tar_x, tar_y,
	    0, 0,
	    tar_img->width, tar_img->height
	);          

	/* Destroy the target img, it is not needed */
	OSWDestroyImage(&tar_img);
}


/*
 *	Puts the image onto the specified drawable with a shadow to
 *	simulate raised height.
 */
void WidgetPutImageRaised(
	drawable_t d, image_t *img,
	int tar_x, int tar_y,
	unsigned int altitude
)
{
	int	c, y, r, g, b, x_rel, y_rel,
		tar_bpp, src_bpp,
		tar_bpl, src_bpl,
		src_x_offset_bytes, tar_x_offset_bytes;
	unsigned int width, height;
	unsigned long pixel_mask;
	u_int8_t *tar_ptr, *tar_end;
	const u_int8_t *src_ptr, *src_end;
#ifdef HAVE_X
	XImage *ximg;
#endif
	image_t *tar_img;
	osw_gui_struct *gui = &osw_gui[0];

	if((d == 0) || (img == NULL))
	    return;

	/* Blit shadow */
	x_rel = (int)((float)altitude * 0.70710670f);
	y_rel = x_rel;

	/* Get target image */
	tar_img = OSWGetImage(
	    d, 
	    tar_x + x_rel, tar_y + y_rel,
	    img->width,
	    img->height
	);
	if(tar_img == NULL)
	{
	    WidgetPutImageNormal(
		d,
		img,
		tar_x, tar_y,
		True
	    );
	    return;
	}

	/* Blit shadow */
	tar_bpp = tar_img->bits_per_pixel / 8;
	src_bpp = img->bits_per_pixel / 8;
	tar_bpl = tar_img->bytes_per_line;
	src_bpl = img->bytes_per_line;
	width = MIN(img->width, tar_img->width);
	height = MIN(img->height, tar_img->height);
	tar_x_offset_bytes = (int)tar_img->xoffset * tar_bpp;
	src_x_offset_bytes = (int)img->xoffset * src_bpp;
	pixel_mask = img->red_mask | img->green_mask | img->blue_mask;

	for(y = 0; y < (int)height; y++)
	{
	    tar_ptr = (u_int8_t *)tar_img->data + (y * tar_bpl);
	    tar_end = tar_ptr + (width * tar_bpp);
	    src_ptr = (const u_int8_t *)img->data + (y * src_bpl);
	    src_end = src_ptr + (width * src_bpp);
	    tar_ptr += tar_x_offset_bytes;
	    src_ptr += src_x_offset_bytes;
	    while((tar_ptr < tar_end) && (src_ptr < src_end))
	    {
		switch(img->depth)
		{
		  case 8:
		    if(*(u_int8_t *)src_ptr & (u_int8_t)pixel_mask)
		    {
			c = (int)((*tar_ptr & 0x000000E0) >> 5);
			r = SUBMIN(c - 3, c);

			c = (int)((*tar_ptr & 0x0000001C) >> 2);
			g = SUBMIN(c - 3, c);

			c = (int)(*tar_ptr & 0x00000003);
			b = SUBMIN(c - 2, c);

			*tar_ptr = (u_int8_t)(
			    (r << 5) | (g << 2) | (b)
			);
		    }
		    break;
		  case 15:
		    if(*(u_int16_t *)src_ptr & (u_int16_t)pixel_mask)
		    {
			c = (int)(
			    ((*(u_int16_t *)tar_ptr) & 0x7C00) >> 10
			);
			r = SUBMIN(c - 3, c);

			c = (int)(
			    ((*(u_int16_t *)tar_ptr) & 0x03E0) >> 5
			);
			g = SUBMIN(c - 6, c);

			c = (int)(
			    (*(u_int16_t *)tar_ptr) & 0x001F
			);
			b = SUBMIN(c - 3, c);

			*(u_int16_t *)tar_ptr = (u_int16_t)(
			    (r << 10) | (g << 5) | (b)
			);
		    }
		    break;
		  case 16:
		    if(*(u_int16_t *)src_ptr & (u_int16_t)pixel_mask)
		    {
			c = (int)(
			    ((*(u_int16_t *)tar_ptr) & 0xF800) >> 11
			);
			r = SUBMIN(c - 3, c);

			c = (int)(
			    ((*(u_int16_t *)tar_ptr) & 0x07E0) >> 5
			);
			g = SUBMIN(c - 6, c);

			c = (int)(
			    (*(u_int16_t *)tar_ptr) & 0x001F
			);
			b = SUBMIN(c - 3, c);

			*(u_int16_t *)tar_ptr = (u_int16_t)(
			    (r << 11) | (g << 5) | (b)
			);
		    }     
		    break;
		  case 24:
		  case 32:
		    switch(tar_bpp)
		    {
		      case 3:
			if(src_ptr[0] || src_ptr[1] || src_ptr[2])
			{
			    c = (int)tar_ptr[2];
			    r = SUBMIN(c - 26, c);

			    c = (int)tar_ptr[1];
			    g = SUBMIN(c - 26, c);

			    c = (int)tar_ptr[0];
			    b = SUBMIN(c - 26, c);

			    tar_ptr[2] = r;
			    tar_ptr[1] = g;
			    tar_ptr[0] = b;
			}
			break;
		      case 4:
			if(*(u_int32_t *)src_ptr & (u_int32_t)pixel_mask)
		        {
			    c = (int)tar_ptr[2];
			    r = SUBMIN(c - 26, c);

			    c = (int)tar_ptr[1];
			    g = SUBMIN(c - 26, c);

			    c = (int)tar_ptr[0];
			    b = SUBMIN(c - 26, c);

			    tar_ptr[2] = r;
			    tar_ptr[1] = g;
			    tar_ptr[0] = b;
			}
			break;
		    }
		    break;
		}
		tar_ptr += tar_bpp;
		src_ptr += src_bpp;
	    }
	}

	/* Put the shadow image back to drawable */
#ifdef HAVE_X
	ximg = tar_img->ximage;
	if(ximg != NULL)
	    XPutImage(
	        gui->display, (Drawable)d, gui->gc,
		ximg,
		0, 0,
		tar_x + x_rel, tar_y + y_rel,
		ximg->width, ximg->height
	    );
#endif
	OSWDestroyImage(&tar_img);

	WidgetPutImageNormal(
	    d,
	    img,
	    tar_x, tar_y,
	    True
	);
}


/*
 *	Adjusts the rgb color level on the image by the given
 *	rgb coefficient values from 0 to 255 inclusive.
 *
 *	0 to 1 will reduce color value, 1 to 255 will increase color
 *	value.
 */
void WidgetAdjustImageGamma(
	image_t *img,
	float r, float g, float b
)       
{
	u_int8_t cr, cg, cb;
	int y, bpp, bpl, xob, depth;
	u_int8_t *data, *data_ptr, *data_end;
	unsigned int width, height;
	u_int8_t *data_ptr8;
	u_int16_t *data_ptr16;

	if(img == NULL)
	    return;

	data = (u_int8_t *)img->data;
	width = img->width;
	height = img->height;

	if((data == NULL) || (width == 0) || (height == 0))
	    return;

	/* Sanitize r, g, b */
	if(r < 0.0)
	    r = 0.0;
	if(g < 0.0)
	    g = 0.0;
	if(b < 0)
	    b = 0.0;

	bpp = img->bits_per_pixel / 8;
	bpl = img->bytes_per_line;
	xob = img->xoffset * bpp;
	depth = img->depth;

	for(y = 0; y < (int)height; y++)
	{
	    data_ptr = data + (y * bpl);
	    data_end = data_ptr + ((int)width * bpp);
	    data_ptr += xob;
	    while(data_ptr < data_end)
	    {
		switch(depth)
		{
		  case 8:
		    /* rrrg ggbb */
		    data_ptr8 = (u_int8_t *)data_ptr;
		    cr = (u_int8_t)MIN(
			(u_int32_t)(*data_ptr & 0xe0) * r,
			0x000000ff
		    );
		    cg = (u_int8_t)MIN(
			(u_int32_t)((*data_ptr & 0x1c) << 3) * g,
			0x000000ff
		    );
		    cb = (u_int8_t)MIN(
			(u_int32_t)((*data_ptr & 0x03) << 6) * b,
			0x000000ff
		    );
		    *data_ptr8 = PACK8TO8(cr, cg, cb);
		    break;
		  case 15:
		    /* arrr rrgg gggb bbbb */
		    data_ptr16 = (u_int16_t *)data_ptr;
		    cr = (u_int8_t)MIN(
			(u_int32_t)((*data_ptr16 & 0x7C00) >> 7) * r,
			0x000000ff
		    );
		    cg = (u_int8_t)MIN(
			(u_int32_t)((*data_ptr16 & 0x03E0) >> 2) * g,
			0x000000ff
		    );
		    cb = (u_int8_t)MIN(
			(u_int32_t)((*data_ptr16 & 0x001f) << 3) * b,
			0x000000ff
		    );
		    *data_ptr16 = PACK8TO15(cr, cg, cb);
		    break;
		  case 16:
		    /* rrrr rggg gggb bbbb */
		    data_ptr16 = (u_int16_t *)data_ptr;
		    cr = (u_int8_t)MIN(
			(u_int32_t)((*data_ptr16 & 0xf800) >> 8) * r,
			0x000000ff
		    );
		    cg = (u_int8_t)MIN(
			(u_int32_t)((*data_ptr16 & 0x07e0) >> 3) * g,
			0x000000ff
		    );
		    cb = (u_int8_t)MIN(
			(u_int32_t)((*data_ptr16 & 0x001f) << 3) * b,
			0x000000ff
		    );
		    *data_ptr16 = PACK8TO16(cr, cg, cb);
		    break;
		  case 24:
		  case 32:
		    switch(bpp)
		    {
		      case 3:
			data_ptr[0] = (u_int8_t)(data_ptr[0] * r);
			data_ptr[1] = (u_int8_t)(data_ptr[1] * g);
			data_ptr[2] = (u_int8_t)(data_ptr[2] * b);
			break;
		      case 4:
			data_ptr[0] = (u_int8_t)(data_ptr[0] * r);
			data_ptr[1] = (u_int8_t)(data_ptr[1] * g);
			data_ptr[2] = (u_int8_t)(data_ptr[2] * b);
/*			data_ptr[3] = data_ptr[3]; */
			break;
		    }
		    break; 
		}
		data_ptr += bpp;
	    }
	}
}
