#include <stdlib.h>
#include <string.h>

#include "../include/string.h"

#include "../include/widget.h"

#include "../include/oswkeyvals.h"


/* Callbacks */
static void CListResizeCB(void *widget, void *data);
static void CListDrawCB(void *widget, void *data);
static void CListSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
);

/* Get/Set Row/Item Text */
const char *CListGetItemText(
	clist_struct *clist, int row_num, int column_num
);
void CListSetItemText(
	clist_struct *clist, int row_num, int column_num,
	const char *text
);
void CListSetRowTexts(
	clist_struct *clist, int row_num,
	const char **text, int ntexts
);

/* Get/Set Row/Item Pixmap */
pixmap_t CListGetItemPixmap(
	clist_struct *clist, int row_num, int column_num,
	bitmap_t *mask_rtn
);
void CListSetItemPixmap(
	clist_struct *clist, int row_num, int column_num,
	pixmap_t pixmap, bitmap_t mask
);
void CListSetRowPixmaps(
	clist_struct *clist, int row_num,
	pixmap_t *pixmap, bitmap_t *mask, int npixmaps
);

/* Set Row/Item Attributes */
void CListSetItemAttributes(
	clist_struct *clist, int row_num, int column_num,
	unsigned int flags,
	font_t *font, pixel_t text_c
);
void CListSetRowAttributes(
	clist_struct *clist, int row_num, unsigned int flags,
	font_t *font, pixel_t text_c
);

/* Row Selecting */
int CListGetFirstSelectedRow(clist_struct *clist);
int CListGetLastSelectedRow(clist_struct *clist);
bool_t CListIsRowSelected(clist_struct *clist, const int row_num);
void CListSelectRow(clist_struct *clist, const int row_num);
void CListSelectAllRows(clist_struct *clist);
void CListUnselectRow(clist_struct *clist, const int row_num);
void CListUnselectAllRows(clist_struct *clist);
void CListInvertSelection(clist_struct *clist);

/* Scrolling */
int CListGetRowFromCoordinateY(clist_struct *clist, const int y);
int CListGetColumnFromCoordinateX(clist_struct *clist, const int x);
visibility_t CListGetRowVisibility(clist_struct *clist, const int row_num);
void CListMoveTo(clist_struct *clist, const int row_num, const float coeff);

/* Columns */
int CListColumnAppend(
	clist_struct *clist,
	const char *text,
	font_t *font,
	pixel_t text_c,
	unsigned int flags,
	int x_pos			/* In pixels */
);
int CListGetColumnPosition(clist_struct *clist, int column_num);
void CListSetColumnPosition(
	clist_struct *clist, int column_num,
	int x
);
void CListSetColumnWidth(
	clist_struct *clist, int column_num,
	unsigned int width
);
void CListColumnRemove(clist_struct *clist, int column_num);

/* Column Sizing */
static bool_t CListGetItemMinBounds(
	clist_struct *clist, clist_item_struct *item,
	unsigned int *width_rtn, unsigned int *height_rtn
);
unsigned int CListGetColumnWidthMin(clist_struct *clist, int column_num);
void CListAutosizeColumn(clist_struct *clist, int column_num);
void CListAutosizeColumns(clist_struct *clist);

/* Rows */
int CListRowInsert(clist_struct *clist, int row_num);
int CListRowAppend(clist_struct *clist);
static void CListDeleteItem(
	clist_struct *clist, clist_item_struct *item
);
void CListRowRemove(clist_struct *clist, int row_num);
void CListRowSwap(
	clist_struct *clist, const int row_num1, const int row_num2
);
void CListClear(clist_struct *clist);

/* Row Data */
void CListRowSetData(
	clist_struct *clist, int row_num,
	void *data, void (*destroy_cb)(void *)
);
void *CListRowGetData(clist_struct *clist, int row_num);
int CListRowGetByData(clist_struct *clist, void *data);

/* CList */
int CListInit(
	clist_struct *clist, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height
);
void CListResize(clist_struct *clist);
void CListQueueResize(clist_struct *clist);
void CListDraw(clist_struct *clist, int amount);
void CListQueueDraw(clist_struct *clist, int amount);
static void CListDrawResizeBar(
	clist_struct *clist, int pos,
	bool_t has_previous
);
int CListManage(clist_struct *clist, event_t *event);
void CListMap(clist_struct *clist);
void CListUnmap(clist_struct *clist);
void CListDestroy(clist_struct *clist);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sizes (in pixels):
 */
#define CLIST_DEF_WIDTH			200
#define CLIST_DEF_HEIGHT		100

#define CLIST_DEF_ROW_HEIGHT		18

#define CLIST_HEADING_HEIGHT		20

#define CLIST_FRAME_MARGIN		2


/*
 *	Resize callback.
 */
static void CListResizeCB(void *widget, void *data)
{
	clist_struct *clist = (clist_struct *)widget;
	if(clist == NULL)
	    return;

	CListResize(clist);
} 

/*
 *	Draw callback.
 */
static void CListDrawCB(void *widget, void *data)
{
	int draw_amount = (int)data;
	clist_struct *clist = (clist_struct *)widget;
	if(clist == NULL)
	    return;

	if(clist->map_state)
	    CListDraw(clist, draw_amount);
}

/*
 *	Scroll bar value changed callback.
 */
static void CListSBValueChangedCB(
	scroll_bar_struct *sb, float x, float y, void *data
)
{
	clist_struct *clist = (clist_struct *)data;
	if(clist == NULL)
	    return;

	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}


/*
 *	Gets the item's text.
 */
const char *CListGetItemText( 
	clist_struct *clist, int row_num, int column_num 
)
{
	clist_row_struct *row;
	clist_item_struct *item;

	if(clist == NULL)
	    return(NULL);

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return(NULL);

	item = ((column_num >= 0) && (column_num < row->total_items)) ?
	    row->item[column_num] : NULL;
	if(item == NULL)
	    return(NULL);

	return(item->text);
}

/*
 *	Sets the item's text.
 */
void CListSetItemText(
	clist_struct *clist, int row_num, int column_num,
	const char *text
)
{
	clist_row_struct *row;
	clist_item_struct *item;

	if(clist == NULL)
	    return;

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return;

	item = ((column_num >= 0) && (column_num < row->total_items)) ?
	    row->item[column_num] : NULL;
	if(item == NULL)
	    return;

	free(item->text);
	item->text = STRDUP(text);

	/* Redraw as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}

/*
 *	Sets the text on all the items on the specified row.
 */
void CListSetRowTexts(
	clist_struct *clist, int row_num,
	const char **text, int ntexts
)
{
	int i, m, items_updated;
	clist_row_struct *row;
	clist_item_struct *item;

	if(clist == NULL)
	    return;

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return;

	/* Iterate through items on the row */
	m = MIN(ntexts, row->total_items);
	items_updated = 0;
	for(i = 0; i < m; i++)
	{
	    item = row->item[i];
	    if(item == NULL)
		continue;

	    free(item->text);
	    item->text = STRDUP(text[i]);

	    items_updated++;
	}

	/* Redraw as needed */
	if(clist->map_state && (items_updated > 0))
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}


/*
 *	Gets the item's pixmap and mask.
 */
pixmap_t CListGetItemPixmap(
	clist_struct *clist, int row_num, int column_num,
	bitmap_t *mask_rtn
)
{
	clist_row_struct *row;
	clist_item_struct *item;

	if(mask_rtn != NULL)
	    *mask_rtn = None;

	if(clist == NULL)
	    return(None);

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return(None);

	item = ((column_num >= 0) && (column_num < row->total_items)) ?
	    row->item[column_num] : NULL;
	if(item == NULL)
	    return(None);

	if(mask_rtn != NULL)
	    *mask_rtn = item->mask;

	return(item->pixmap);
}

/*
 *	Sets the item's pixmap and mask.
 */
void CListSetItemPixmap(
	clist_struct *clist, int row_num, int column_num,
	pixmap_t pixmap, bitmap_t mask
)
{
	clist_row_struct *row;
	clist_item_struct *item;

	if(clist == NULL)
	    return;

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return;

	item = ((column_num >= 0) && (column_num < row->total_items)) ?
	    row->item[column_num] : NULL;
	if(item == NULL)
	    return;

	/* Do not delete the previous pixmap or mask, they are shared */
	item->pixmap = pixmap;
	item->mask = mask;

	/* Redraw as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}

/*
 *	Sets the pixmaps and masks for all the items on the row.
 */
void CListSetRowPixmaps(
	clist_struct *clist, int row_num,
	pixmap_t *pixmap, bitmap_t *mask, int npixmaps
)
{
	int i, m, items_updated;
	clist_row_struct *row;
	clist_item_struct *item;

	if(clist == NULL)
	    return;

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return;

	/* Iterate through items on the row */
	m = MIN(npixmaps, row->total_items);
	items_updated = 0;
	for(i = 0; i < m; i++)
	{
	    item = row->item[i];
	    if(item == NULL)
		continue;

	    /* Do not delete the previous pixmap or mask, they are
	     * shared
	     */
	    item->pixmap = pixmap[i];
	    item->mask = mask[i];

	    items_updated++;
	}

	/* Redraw as needed */
	if(clist->map_state && (items_updated > 0))
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}

/*
 *	Set's the item's flags, font, and text color.
 */
void CListSetItemAttributes(
	clist_struct *clist, int row_num, int column_num,
	unsigned int flags,
	font_t *font, pixel_t text_c
)
{
	clist_row_struct *row;
	clist_item_struct *item;

	if(clist == NULL)
	    return;

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return;

	item = ((column_num >= 0) && (column_num < row->total_items)) ?
	    row->item[column_num] : NULL;
	if(item == NULL)
	    return;

	item->font = font;
	item->text_c = text_c;

	/* Redraw as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}

/*
 *	Sets the flags, font, and text color for all the items on the
 *	row.
 */
void CListSetRowAttributes(
	clist_struct *clist, int row_num, unsigned int flags,
	font_t *font, pixel_t text_c     
)
{
	int i, m, items_updated;
	clist_row_struct *row;
	clist_item_struct *item;

	if(clist == NULL)
	    return;

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return;

	/* Iterate through items on the row */
	m = row->total_items;
	items_updated = 0;
	for(i = 0; i < m; i++)
	{
	    item = row->item[i];
	    if(item == NULL)
		continue;

	    item->font = font;
	    item->text_c = text_c;

	    items_updated++;
	}

	/* Redraw as needed */
	if(clist->map_state && (items_updated > 0))
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}


/*
 *	Returns the first selected row index.
 */
int CListGetFirstSelectedRow(clist_struct *clist)
{
	if(clist == NULL)
	    return(-1);

	return((clist->total_sel_rows > 0) ? clist->sel_row[0] : -1);
}

/*
 *	Returns the last selected row index.
 */
int CListGetLastSelectedRow(clist_struct *clist)
{
	int m;

	if(clist == NULL)
	    return(-1);

	m = clist->total_sel_rows;
	return((m > 0) ? clist->sel_row[m - 1] : -1);
}

/*
 *	Checks if the specified row is selected.
 */
bool_t CListIsRowSelected(clist_struct *clist, const int row_num)
{
	int i;

	if(clist == NULL)
	    return(False);

	if(clist->sel_row == NULL)
	    return(False);

	if((row_num < 0) || (row_num >= clist->total_rows))
	    return(False);

	for(i = 0; i < clist->total_sel_rows; i++)
	{
	    if(clist->sel_row[i] == row_num)
		return(True);
	}

	return(False);
}

/*
 *	Selects the row.
 *
 *	Calls the select callback and redraws as needed.
 */
void CListSelectRow(clist_struct *clist, const int row_num)
{
	int n;

	if(clist == NULL)
	    return;

	if((row_num < 0) || (row_num >= clist->total_rows))
	    return;

	/* Is row already selected? */
	if(CListIsRowSelected(clist, row_num))
	    return;

	/* Select this row */
	if(clist->total_sel_rows < 0)
	    clist->total_sel_rows = 0;
	n = clist->total_sel_rows;
	clist->total_sel_rows = n + 1;
	clist->sel_row = (int *)realloc(
	    clist->sel_row,
	    clist->total_sel_rows * sizeof(int)
	);
	if(clist->sel_row == NULL)
	{
	    clist->total_sel_rows = 0;
	    return;
	}

	clist->sel_row[n] = row_num;

	if(clist->select_cb != NULL)
	    clist->select_cb(
		clist,
		row_num,
		clist->select_data
	    );

	/* Redraw as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}

/*
 *	Selects all the rows.
 *
 *	Calls the select callback and redraws as needed.
 */
void CListSelectAllRows(clist_struct *clist)
{
	int i, n, rows_selected;

	if(clist == NULL)
	    return;

	rows_selected = 0;
	for(i = 0; i < clist->total_rows; i++)
	{
	    /* Is this row already selected? */
	    if(CListIsRowSelected(clist, i))
		continue;

	    /* Select this row */
	    n = clist->total_sel_rows;
	    clist->total_sel_rows = n + 1;
	    clist->sel_row = (int *)realloc(
		clist->sel_row,
		clist->total_sel_rows * sizeof(int)
	    );
	    if(clist->sel_row == NULL)
	    {
		clist->total_sel_rows = 0;
		break;
	    }
	    clist->sel_row[n] = i;

	    rows_selected++;

	    if(clist->select_cb != NULL)
		clist->select_cb(
		    clist,
		    i,
		    clist->select_data
		);
	}

	/* Redraw as needed */
	if(clist->map_state && (rows_selected > 0))
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}

/*
 *	Unselects the row.
 *
 *	Calls the unselect callback and redraws as needed.
 */
void CListUnselectRow(clist_struct *clist, const int row_num)
{
	int i, n, rows_unselected;

	if(clist == NULL)
	    return;

	if((row_num < 0) || (row_num >= clist->total_rows))
	    return;

	rows_unselected = 0;
	for(i = 0; i < clist->total_sel_rows; i++)
	{
	    if(clist->sel_row[i] == row_num)
	    {
		clist->total_sel_rows--;
		for(n = i; n < clist->total_sel_rows; n++)
		    clist->sel_row[n] = clist->sel_row[n + 1];
		if(clist->total_sel_rows > 0)
		{
		    clist->sel_row = (int *)realloc(
			clist->sel_row,
			clist->total_sel_rows * sizeof(int)
		    );
		    if(clist->sel_row == NULL)
		    {
			clist->total_sel_rows = 0;
			break;
		    }
		}
		else
		{
		    free(clist->sel_row);
		    clist->sel_row = NULL;
		}

		rows_unselected++;

		if(clist->unselect_cb != NULL)
		    clist->unselect_cb(
			clist,
			row_num,
			clist->unselect_data
		    );

		break;
	    }
	}

	/* Redraw as needed */
	if(clist->map_state && (rows_unselected > 0))
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}

/*
 *	Unselects all rows.
 *
 *	Calls the unselect callback and redraws as needed.
 */
void CListUnselectAllRows(clist_struct *clist)
{
	if(clist == NULL)
	    return;

	if(clist->unselect_cb != NULL)
	{
	    while(clist->total_sel_rows > 0)
	    {
		clist->total_sel_rows--;
		clist->unselect_cb(
		    clist,
		    clist->sel_row[clist->total_sel_rows],
		    clist->unselect_data
		);
	    }
	}
	free(clist->sel_row);
	clist->sel_row = NULL;
	clist->total_sel_rows = 0;

	/* Redraw as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}

/*
 *	Unselect all selected rows and select all unselected rows.
 */
void CListInvertSelection(clist_struct *clist)
{
	int i;

	if(clist == NULL)
	    return;

	for(i = 0; i < clist->total_rows; i++)
	{
	    /* If the row is selected then unselect it, otherwise
	     * select it
	     */
	    if(CListIsRowSelected(clist, i))
		CListUnselectRow(clist, i);
	    else
		CListSelectRow(clist, i);
	}

	/* Redraw as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
}


/*
 *	Returns the row that matches the specified Y coordinate.
 *
 *	Can return -1 on failed match.
 */
int CListGetRowFromCoordinateY(clist_struct *clist, const int y)
{
	int i;

	if(clist == NULL)
	    return(-1);

	i = (y + (int)SBarGetPositionY(clist->sb) - 2) /
	    (int)MAX(clist->row_height, 1);
	return(
	    ((i >= 0) && (i < clist->total_rows)) ? i : -1
	);
}

/*
 *	Returns the column that matches the specified X coordinate.
 *
 *	Can return -1 on failed match.
 */
int CListGetColumnFromCoordinateX(clist_struct *clist, const int x)
{
	int i;
	clist_column_struct *column;

	if(clist == NULL)
	    return(-1);

	if(x < 0)
	    return(-1);

	/* Iterate through the list of columns but skip the first one */
	for(i = 1; i < clist->total_columns; i++)
	{
	    column = clist->column[i];
	    if(column == NULL)
		continue;

	    if(x < column->x_pos)
	    {
		/* Return the previous column */
		return(i - 1);
	    }
	}

	return(clist->total_columns - 1);
}

/*
 *	Returns the row's visibility.
 *
 *	Can return VisibilityFullyObscured on failed match.
 */
visibility_t CListGetRowVisibility(clist_struct *clist, int row_num)
{
	int y, height, row_height;
	scroll_bar_struct *sb;

	if(clist == NULL)
	    return(VisibilityFullyObscured);

	if((row_num < 0) || (row_num >= clist->total_rows))
	    return(VisibilityFullyObscured); 

	row_height = MAX((int)clist->row_height, 1);
	sb = clist->sb;
	height = (int)sb->height - (2 * CLIST_FRAME_MARGIN);

	y = (row_num * row_height) - (int)SBarGetPositionY(sb);
	if(y < 0)
	    return(((y + row_height) > 0) ?
		VisibilityPartiallyObscured : VisibilityFullyObscured
	    );
	else if((y + row_height) > height)
	    return((y < height) ?
		VisibilityPartiallyObscured : VisibilityFullyObscured
	    );
	else
	    return(VisibilityUnobscured);
}

/*
 *	Scrolls to the specified row with the specified coefficient
 *	applied as the offset.
 */
void CListMoveTo(clist_struct *clist, const int row_num, const float coeff)
{
	int y, last_y, height, row_height;
	scroll_bar_struct *sb;

	if(clist == NULL)
	    return;

	if((row_num < 0) || (row_num >= clist->total_rows))
	    return;

	row_height = MAX((int)clist->row_height, 1);

	sb = clist->sb;
	height = (int)sb->height - (2 * CLIST_FRAME_MARGIN);
	last_y = (int)SBarGetPositionY(sb);

	/* Move to */
	y = (row_num * row_height) -
	    (int)(coeff * (height - row_height));
	if(y != last_y)
	    SBarSetPositionY(sb, (float)y);
}


/*
 *	Appends a new column to the list.
 */
int CListColumnAppend(
	clist_struct *clist,
	const char *text,
	font_t *font,
	pixel_t text_c,
	unsigned int flags,
	int x_pos		/* In pixels */
)
{
	int n;
	clist_column_struct *column;

	if(clist == NULL)
	    return(-2);

	n = MAX(clist->total_columns, 0);
	clist->total_columns = n + 1;
	clist->column = (clist_column_struct **)realloc(
	    clist->column,
	    clist->total_columns * sizeof(clist_column_struct *)
	);
	if(clist->column == NULL)
	{
	    clist->total_columns = 0;
	    return(-3);
	}

	clist->column[n] = column = (clist_column_struct *)calloc(
	    1, sizeof(clist_column_struct)
	);
	if(column == NULL)
	{
	    clist->total_columns = n;
	    return(-3);
	}

	/* Set heading values */
	column->flags = flags;
	column->text = STRDUP(text);
	column->font = font;
	column->text_c = text_c;
	column->x_pos = x_pos;
	column->x_offset = 0;
	column->y_offset = 0;

	/* Redraw as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);

	return(n);
}

/*
 *	Returns the position of the specified column.
 */
int CListGetColumnPosition(clist_struct *clist, int column_num)
{
	clist_column_struct *column;

	if(clist == NULL)
	    return(0);

	column = ((column_num >= 0) && (column_num < clist->total_columns)) ?
	    clist->column[column_num] : NULL;
	return((column != NULL) ? column->x_pos : 0);
}
/*
 *	Sets the column's position.
 */
void CListSetColumnPosition(
	clist_struct *clist, int column_num,
	int x
)
{
	clist_column_struct *column;

	if(clist == NULL)
	    return;

	column = ((column_num >= 0) && (column_num < clist->total_columns)) ?
	    clist->column[column_num] : NULL;
	if(column == NULL)
	    return;

	/* Change in position? */
	if(column->x_pos != x)
	{
	    /* Set new position */
	    column->x_pos = x;

	    if(clist->map_state)
		CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);

	    if(clist->column_resized_cb != NULL)
		clist->column_resized_cb(
		    clist,
		    column_num,
		    column->x_pos,
		    clist->column_resized_data
		);
	}
}

/*
 *	Sets the column's width.
 *
 *	This actually sets the next (column_num + 1) column's position.
 */
void CListSetColumnWidth(
	clist_struct *clist, int column_num, 
	unsigned int width                             
)
{
	int x;
	clist_column_struct *column;

	if((clist == NULL) || (column_num < 0))
	    return;

	/* Both this column and the next column must be valid */
	if((column_num + 1) >= clist->total_columns)
	    return;

	/* Get this column's position */
	x = CListGetColumnPosition(clist, column_num);

	/* Calculate the next column's position, which is this
	 * column's width
	 */
	x += (int)width;

	/* Get the next column */
	column_num++;
	column = clist->column[column_num];
	if(column == NULL)
	    return;

	/* Change in position? */
	if(column->x_pos != x)
	{
	    /* Set new position */
	    column->x_pos = x;

	    if(clist->map_state)
		CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);

	    if(clist->column_resized_cb != NULL)
		clist->column_resized_cb(
		    clist,
		    column_num,
		    column->x_pos,
		    clist->column_resized_data
		);
	}
}

/*
 *	Removes the column.
 */
void CListColumnRemove(clist_struct *clist, int column_num)
{
	clist_column_struct *column;

	if(clist == NULL)
	    return;

	column = ((column_num >- 0) && (column_num < clist->total_columns)) ?
	    clist->column[column_num] : NULL;
	if(column == NULL)
	    return;

	free(column->text);
	free(column);
	clist->column[column_num] = NULL;

	/* Redraw as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);
}


/*
 *	Returns the mimimum size of the item.
 */
static bool_t CListGetItemMinBounds(
	clist_struct *clist, clist_item_struct *item,
	unsigned int *width_rtn, unsigned int *height_rtn
)
{
	unsigned int width, height;

	if(width_rtn != NULL)
	    *width_rtn = 0;
	if(height_rtn != NULL)
	    *height_rtn = 0;

	if((clist == NULL) || (item == NULL))
	    return(False);

	width = 0;
	height = 0;

	if(item->pixmap != None)
	{
	    unsigned int icon_width, icon_height;
	    OSWGetPixmapGeometry(
		item->pixmap, &icon_width, &icon_height, NULL
	    );
	    width += icon_width + 2;
	    height += icon_height;
	}
	if(item->text != NULL)
	{
	    int swidth, ascent, descent, font_height;
	    OSWGetTextBounds(
		clist->font, item->text, -1,
		NULL, NULL, &swidth, &ascent, &descent
	    );
	    font_height = ascent + descent;
	    width += swidth;
	    if(font_height > (int)height)
		height = (unsigned int)font_height;
	}

	if(width_rtn != NULL)
	    *width_rtn = width;
	if(height_rtn != NULL)
	    *height_rtn = height;

	return(True);
}

/*
 *	Returns the minimum width in needed to display the column
 *	heading and each item on the column without obscurity.
 */
unsigned int CListGetColumnWidthMin(clist_struct *clist, int column_num)
{
	int i, nrows;
	unsigned int width, item_width;
	clist_column_struct *column;
	clist_row_struct *row, **rows;
	clist_item_struct *item;

	if(clist == NULL)
	    return(0);

	column = ((column_num >= 0) && (column_num < clist->total_columns)) ?
	    clist->column[column_num] : NULL;
	if(column == NULL)
	    return(0);

	width = 0;

	rows = clist->row;
	nrows = clist->total_rows;
	for(i = 0; i < nrows; i++)
	{
	    row = rows[i];
	    if(row == NULL)
		continue;

	    item = (column_num < row->total_items) ?
		row->item[column_num] : NULL;

	    if(!CListGetItemMinBounds(
		clist, item,
		&item_width, NULL
	    ))
		continue;

	    if(item_width > width)
		width = item_width;
	}

	/* Check column heading text */
	if(column->text != NULL)
	{
	    int swidth;
	    OSWGetTextBounds(
		clist->font, column->text, -1,
		NULL, NULL, &swidth, NULL, NULL
	    );
	    if(swidth > (int)width)
		width = (unsigned int)swidth;
	}

	if(width > 0)
	    width += (4 * CLIST_FRAME_MARGIN);

	return(width);
}

/*
 *	Sets the column's width to the minimum width needed in order
 *	to display the heading and all its items.
 */
void CListAutosizeColumn(clist_struct *clist, int column_num)
{
	unsigned int width;

	if(clist == NULL)
	    return;

	width = CListGetColumnWidthMin(clist, column_num);
	CListSetColumnWidth(clist, column_num, width);
}

/*
 *	Autosizes all columns.
 */
void CListAutosizeColumns(clist_struct *clist)
{
	int i, m;

	if(clist == NULL)
	    return;

	m = clist->total_columns - 1;
	for(i = 0; i < m; i++)
	    CListAutosizeColumn(clist, i);
}


/*
 *	Inserts a row.
 */
int CListRowInsert(clist_struct *clist, int row_num)
{
	int n;
	scroll_bar_struct *sb;
	clist_row_struct *row;
	widget_global_struct *wglobal = &widget_global;

	if(clist == NULL)
	    return(-1);
	
	if(clist->total_rows < 0)
	    clist->total_rows = 0;

	/* Append? */
	if(row_num < 0)
	{
	    /* Append */
	    n = clist->total_rows;
	    clist->total_rows = n + 1;
	    clist->row = (clist_row_struct **)realloc(
		clist->row,
		clist->total_rows * sizeof(clist_row_struct *)
	    );
	    if(clist->row == NULL)
	    {
		clist->total_rows = 0; 
		return(-3);
	    }
	}
	else
	{
	    /* Insert */
	    clist->total_rows++;
	    if(row_num >= clist->total_rows)
		row_num = clist->total_rows - 1;
	    clist->row = (clist_row_struct **)realloc(
		clist->row,
		clist->total_rows * sizeof(clist_row_struct *)
	    );
	    if(clist->row == NULL)
	    {
		clist->total_rows = 0;
		return(-3);
	    }

	    /* Shift pointers */
	    for(n = clist->total_rows - 1; n > row_num; n--)
	        clist->row[n] = clist->row[n - 1];

	    n = row_num;
	}

	/* Allocate a new row */
	clist->row[n] = row = (clist_row_struct *)calloc(
	    1, sizeof(clist_row_struct)
	);
	if(row == NULL)
	    return(-3);

	/* Allocate the row's items */
	row->total_items = clist->total_columns;
	if(row->total_items > 0)
	{
	    row->item = (clist_item_struct **)malloc(
		row->total_items * sizeof(clist_item_struct *)
	    );
	    if(row->item != NULL)
	    {
		int i;
		clist_item_struct *item;
		for(i = 0; i < row->total_items; i++)
		{
		    row->item[i] = item = (clist_item_struct *)calloc(
			1, sizeof(clist_item_struct)
		    );
		    if(item != NULL)
		    {
			clist_column_struct *column = (i < clist->total_columns) ?
			    clist->column[i] : NULL;
			item->flags = 0;
			if(column != NULL)
			{
			    if(column->flags & CLIST_COLUMN_FLAG_ALIGN_LEFT)
				item->flags |= CLIST_ITEM_FLAG_ALIGN_LEFT;
			    if(column->flags & CLIST_COLUMN_FLAG_ALIGN_CENTER)
				item->flags |= CLIST_ITEM_FLAG_ALIGN_CENTER;
			    if(column->flags & CLIST_COLUMN_FLAG_ALIGN_RIGHT)
				item->flags |= CLIST_ITEM_FLAG_ALIGN_RIGHT;
			    item->font = column->font;
			    item->x_offset = column->x_offset;
			    item->y_offset = column->y_offset;
			}
			item->text = NULL;
			item->pixmap = None;
			item->mask = None;
			item->text_c = wglobal->editable_text_pix;
		    }
		}
	    }
	    else
	    {
		row->total_items = 0;
	    }
	}
	else
	{
	    row->item = 0;
	}

	row->data = NULL;
	row->destroy_cb = NULL;

	sb = clist->sb;
	SBarQueueResize(
	    sb,
	    sb->width, sb->height,
	    sb->width,
	    (clist->total_rows * clist->row_height) +  
		(2 * CLIST_FRAME_MARGIN)
	);
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);

	return(n);
}

/*
 *	Appends a row.
 *
 *	Does not redraw the list.
 */
int CListRowAppend(clist_struct *clist)
{
	return(CListRowInsert(clist, -1));
}

/*
 *	Deletes the item.
 */
static void CListDeleteItem(
	clist_struct *clist, clist_item_struct *item
)
{
	if(item == NULL)
	    return;

	free(item->text);
	free(item);
}

/*
 *	Deletes the row.
 */
void CListRowRemove(clist_struct *clist, int row_num)
{
	int i;
	scroll_bar_struct *sb;
	clist_row_struct *row;

	if(clist == NULL)
	    return;

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return;

	/* Unselect this row if it is selected */
	if(CListIsRowSelected(clist, row_num))
	    CListUnselectRow(clist, row_num);

	/* Call the row data destroy callback as needed */
	if(row->destroy_cb != NULL)
	    row->destroy_cb(row->data);

	/* Delete this row */
	for(i = 0; i < row->total_items; i++)
	    CListDeleteItem(clist, row->item[i]);
	free(row->item);
	free(row);
	clist->row[row_num] = NULL;

	/* Shift row pointers */
	clist->total_rows--;
	if(clist->total_rows < 0)
	    clist->total_rows = 0;
	for(i = row_num; i < clist->total_rows; i++)
	    clist->row[i] = clist->row[i + 1];

	/* Reallocate row pointers */
	if(clist->total_rows > 0)
	{
	    clist->row = (clist_row_struct **)realloc(
		clist->row,
		clist->total_rows * sizeof(clist_row_struct *)
	    );
	    if(clist->row == NULL)
		clist->total_rows = 0;
	}
	else
	{
	    free(clist->row);
	    clist->row = NULL;
	    clist->total_rows = 0;
	}

	/* Update the focus row to point to a non NULL row */
	if(clist->focus_row == row_num)
	{
	    for(i = MIN(row_num, (clist->total_rows - 1)); i >= 0; i--)
	    {
		if(clist->row[i] != NULL)
		{
		    clist->focus_row = i;
		    break;
		}
	    }
	}

	/* Update the scroll bar */
	sb = clist->sb;
	SBarQueueResize(
	    sb,
	    sb->width, sb->height,
	    sb->width,
	    (clist->total_rows * clist->row_height) +
		(2 * CLIST_FRAME_MARGIN)
	);

	/* Redraw the list and scroll bars as needed */
	if(clist->map_state)
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);
}

/*
 *	Swaps the two rows.
 */
void CListRowSwap(
	clist_struct *clist, const int row_num1, const int row_num2
)
{
	bool_t row1_selected, row2_selected;
	int n;
	clist_row_struct *row1, *row2;

	if(clist == NULL)
	    return;

	n = clist->total_rows;
	if((row_num1 < 0) || (row_num1 >= n))
	    return;
	if((row_num2 < 0) || (row_num2 >= n))
	    return;

	/* Record the selected state of each row */
	row1_selected = CListIsRowSelected(clist, row_num1);
	row2_selected = CListIsRowSelected(clist, row_num2);

	/* Unselect the rows as needed before swapping */
	if(row1_selected)
	    CListUnselectRow(clist, row_num1);
	if(row2_selected)
	    CListUnselectRow(clist, row_num2);

	/* Check again after unselected since callbacks may have
	 * modified the rows
	 */
	n = clist->total_rows;
	if((row_num1 < 0) || (row_num1 >= n))    
	    return;
	if((row_num2 < 0) || (row_num2 >= n))
	    return;

	/* Get the rows */
	row1 = clist->row[row_num1];
	row2 = clist->row[row_num2];

	/* Swap the rows */
	clist->row[row_num1] = row2;
	clist->row[row_num2] = row1;

	CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);

	/* Reselect the rows as needed */
	if(row2_selected)
	    CListSelectRow(clist, row_num1);
	if(row1_selected)
	    CListSelectRow(clist, row_num2);
}

/*
 *	Deletes all rows.
 *
 *	This is more efficient than calling CListRowRemove() for
 *	each row.
 */
void CListClear(clist_struct *clist)
{
	int i, n, rows_removed = 0;
	clist_row_struct *row;
	scroll_bar_struct *sb;

	if(clist == NULL)
	    return;

	/* Unselect all rows first */
	CListUnselectAllRows(clist);

	clist->focus_row = 0;

	/* Delete all rows */
	for(i = 0; i < clist->total_rows; i++)
	{
	    row = clist->row[i];
	    if(row == NULL)
		continue;

	    /* Call the row data destroy callback as needed */
	    if(row->destroy_cb != NULL)
		row->destroy_cb(row->data);

	    /* Delete this row */
	    for(n = 0; n < row->total_items; n++)
		CListDeleteItem(clist, row->item[n]);
	    free(row->item);
	    free(row);
	    clist->row[i] = NULL;

	    rows_removed++;
	}
	free(clist->row);
	clist->row = NULL;
	clist->total_rows = 0;

	/* Update the scroll bar and move it to the origin */
	sb = clist->sb;
	SBarQueueResize(
	    sb,
	    sb->width, sb->height,
	    sb->width,
	    (clist->total_rows * clist->row_height) +
		(2 * CLIST_FRAME_MARGIN)
	);
	SBarSetPositionX(sb, 0.0f);
	SBarSetPositionY(sb, 0.0f);

	/* Redraw the list and scroll bars as needed */
	if(clist->map_state && (rows_removed > 0))
	    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);
}


/*
 *	Sets the row's data.
 */
void CListRowSetData(
	clist_struct *clist, int row_num,
	void *data, void (*destroy_cb)(void *)
)
{
	clist_row_struct *row;

	if(clist == NULL)
	    return;

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return;

	/* Call destroy callback for the previously set data (if any) */
	if(row->destroy_cb != NULL)
	    row->destroy_cb(
		row->data
	    );

	row->destroy_cb = destroy_cb;
	row->data = data;
}

/*
 *	Returns the row's data.
 */
void *CListRowGetData(clist_struct *clist, int row_num)
{
	clist_row_struct *row;

	if(clist == NULL)
	    return(NULL);

	row = ((row_num >= 0) && (row_num < clist->total_rows)) ?
	    clist->row[row_num] : NULL;
	if(row == NULL)
	    return(NULL);

	return(row->data);
}

/*
 *	Returns the row who's data matches the specified data or -1 on
 *	failed match.
 */
int CListRowGetByData(clist_struct *clist, void *data)
{
	int i, n;

	if(clist == NULL)
	    return(-1);

	n = clist->total_rows;
	for(i = 0; i < n; i++)
	{
	    if(CListRowGetData(clist, i) == data)
		return(i);
	}

	return(-1);
}


/*
 *	Initializes the clist.
 */
int CListInit(
	clist_struct *clist, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	win_t w;
	scroll_bar_struct *sb;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if((clist == NULL) || (parent == None))
	    return(-2);

	clist->type = WIDGET_TYPE_CLIST;
	clist->map_state = False;
	clist->has_focus = False;
	clist->disabled = False;
	clist->visibility_state = VisibilityFullyObscured;
	clist->x = x;
	clist->y = y;
	clist->width = (width == 0) ? CLIST_DEF_WIDTH : width;
	clist->height = (height == 0) ? CLIST_DEF_HEIGHT : height;
	clist->font = wglobal->std_font;
	clist->next = NULL;
	clist->prev = NULL;
	clist->data = NULL;

	clist->options = 0;
	clist->row_height = CLIST_DEF_ROW_HEIGHT;

	clist->resizing_column = -1;
	clist->pointer_over_column_resize = -1;
	clist->resizing_column_bar_pos = 0;

	clist->column = NULL;
	clist->total_columns = 0;

	clist->row = NULL;
	clist->total_rows = 0;

	clist->sel_row = NULL;
	clist->total_sel_rows = 0;
	clist->focus_row = 0;

	clist->button_pressed = 0;
	clist->button_pressed_row = -1;
	clist->button_pressed_time = 0l; 

	clist->activate_cb = NULL;
	clist->activate_data = NULL;
	clist->select_cb = NULL;
	clist->select_data = NULL;
	clist->unselect_cb = NULL;
	clist->unselect_data = NULL;
	clist->column_resized_cb = NULL;
	clist->column_resized_data = NULL;

	/* Toplevel */
	x = clist->x;
	y = clist->y;
	width = clist->width;
	height = clist->height;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	clist->toplevel = w;
	OSWSetWindowInput(
	    w,
	    KeyPressMask | KeyReleaseMask
	);
	parent = w;

#ifdef HAVE_X
	if(w != 0)
	{
	    gc_val_t v;
	    v.function = GXinvert;
	    v.foreground = gui->white_pix;
	    v.background = gui->black_pix;
	    v.line_width = 1;
	    v.line_style = LineSolid;
	    v.subwindow_mode = IncludeInferiors;
	    clist->invert_gc = XCreateGC(
		gui->display, w,
		GCFunction |
		GCForeground | GCBackground |
		GCLineWidth | GCLineStyle |
		GCSubwindowMode,
		&v
	    );
	}
#endif

	/* Heading */
	x = 0;
	y = 0;
	width = clist->width;
	height = CLIST_HEADING_HEIGHT;
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	clist->heading = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask |
	    EnterWindowMask | LeaveWindowMask
	);

	/* List */
	x = 0;
	y = CLIST_HEADING_HEIGHT;
	width = clist->width;
	height = (unsigned int)MAX(
	    (int)clist->height - CLIST_HEADING_HEIGHT,
	    0
	);
	OSWCreateWindow(
	    &w, parent,
	    x, y, width, height
	);
	clist->list = w;
	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask
	);

	/* List scroll bar */
	clist->sb = sb = SCROLL_BAR(malloc(sizeof(scroll_bar_struct)));
	SBarInit(
	    sb, w,
	    width, height,
	    width,
	    (clist->total_rows * clist->row_height) +
		(2 * CLIST_FRAME_MARGIN)
	);
	sb->value_changed_cb = CListSBValueChangedCB;
	sb->value_changed_data = clist;


	/* Add widget to regeristry */
	WidgetRegister(clist, WIDGET_TYPE_CLIST);

	return(0);
}

/*
 *	Resizes the clist.
 */
void CListResize(clist_struct *clist)
{
	int x, y;
	unsigned int width, height;
	win_t w;

	if(clist == NULL)
	    return;

	w = clist->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);

	clist->x = x;
	clist->y = y;

	if((clist->width == width) && (clist->height == height))
	    return;

	clist->width = width;
	clist->height = height;

	/* Heading */
	w = clist->heading;
	x = 0;
	y = 0;
	width = clist->width;
	height = CLIST_HEADING_HEIGHT;
	OSWMoveResizeWindow(w, x, y, width, height);
	OSWDestroyPixmap(&clist->heading_buf);

	/* List */
	w = clist->list;
	x = 0;
	y = CLIST_HEADING_HEIGHT;
	width = clist->width;
	height = (unsigned int)MAX(
	    (int)clist->height - CLIST_HEADING_HEIGHT,
	    0
	);
	OSWMoveResizeWindow(w, x, y, width, height);
	OSWDestroyPixmap(&clist->list_buf);

	/* List scroll bars */
	SBarResize(
	    clist->sb,
	    width, height,
	    width,
	    (clist->total_rows * clist->row_height) +
		(2 * CLIST_FRAME_MARGIN)
	);
}

/*
 *	Queues the clist to resize.
 */
void CListQueueResize(clist_struct *clist)
{
	if(clist == NULL)
	    return;

	WidgetQueueResizeAction(CListResizeCB, clist, NULL);
}

/*
 *	Draws the clist.
 */
void CListDraw(clist_struct *clist, int amount)
{
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(clist == NULL)
	    return;

	/* Recreate buffers as needed */
	if(clist->heading_buf == None)
	{
	    unsigned int width, height;
	    OSWGetWindowGeometry(
		clist->heading, NULL, NULL, &width, &height, NULL
	    );
	    OSWCreatePixmap(&clist->heading_buf, width, height);
	}
	if(clist->list_buf == None)
	{
	    unsigned int width, height;
	    OSWGetWindowGeometry(
		clist->list, NULL, NULL, &width, &height, NULL
	    );
	    OSWCreatePixmap(&clist->list_buf, width, height);
	}


	/* Draw column headings? */
	if((amount == CLIST_DRAW_AMOUNT_COMPLETE) ||
	   (amount == CLIST_DRAW_AMOUNT_HEADING)
	)
	{
	    int	i, x, x2, y2, column_width,
		lbearing, rbearing, swidth, ascent, descent,
		font_height;
	    unsigned int width, height;
	    const char *s;
	    font_t *font;
	    win_t w = clist->heading;
	    pixmap_t pixmap = clist->heading_buf;
	    clist_column_struct *column, *column2;
	    rectangle_t rect;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    if(wglobal->force_mono)
	    {
		OSWClearPixmap(
		    pixmap, width, height,
		    gui->white_pix
		);
		OSWSetForeground(gui->black_pix);
	    }
	    else
	    {
		WidgetPutImageTile(
		    pixmap,
		    wglobal->menu_bar_bg_img,
		    width, height
		);
		OSWSetForeground(wglobal->surface_shadow_pix);
	    }

	    for(i = 0; i < clist->total_columns; i++)
	    {
		column = clist->column[i];
		if(column == NULL)
		    continue;

		/* Get this column's position */
		x = column->x_pos;

		/* Get next column */
		column2 = ((i + 1) < clist->total_columns) ?
		    clist->column[i + 1] : NULL;

		/* Get this column's width */
		column_width = ((column2 != NULL) ?
		    column2->x_pos : (int)width) - x;
		if(column_width <= 0)
		    continue;

		/* Set font */
		font = column->font;
		OSWSetFont(font);

		/* Get text bounds */
		s = column->text;
		OSWGetTextBounds(
		    font, s, -1,
		    &lbearing, &rbearing, &swidth,
		    &ascent, &descent
		);
		font_height = ascent + descent;

		/* Set clip rectangle to draw string so that it does
		 * not bleed into the next column if this column is
		 * too small to fit the string
		 */
		rect.x = x;
		rect.y = 0;
		rect.width = (unsigned int)MAX(
		    (column_width - 1), 0
		);
		rect.height = height;
		OSWSetClipRectangle(&rect);

		/* Draw the heading text */
		if(column->flags & CLIST_COLUMN_FLAG_ALIGN_CENTER)
		{
		    x2 = x + ((column_width - swidth) / 2);
		    y2 = (((int)height - font_height) / 2);
		}
		else if(column->flags & CLIST_COLUMN_FLAG_ALIGN_RIGHT)
		{
		    if((i + 1) == clist->total_columns)
		    {
			x2 = x + column_width - swidth -
			    (2 * CLIST_FRAME_MARGIN) -
			    SCROLLBAR_YBAR_WIDTH;
			y2 = (((int)height - font_height) / 2);
		    }
		    else
		    {
			x2 = x + column_width - swidth -
			    (2 * CLIST_FRAME_MARGIN);
			y2 = (((int)height - font_height) / 2);
		    }
		}
		else	/* CLIST_COLUMN_FLAG_ALIGN_LEFT */
		{
		    x2 = x + (2 * CLIST_FRAME_MARGIN);
		    y2 = (((int)height - font_height) / 2);
		}
		OSWDrawString(
		    pixmap,
		    x2 + column->x_offset - lbearing,
		    y2 + column->y_offset + ascent,
		    s
		);

		/* Draw this column's deliminator line */
		if(i > 0)
		{
		    x2 = x;
		    y2 = (int)height;
		    OSWDrawLine(
			pixmap,
			x2, 0,
			x2, y2
		    );
		}
	    }
	    OSWSetClipRectangle(NULL);

	    OSWPutDrawableToWindow(w, pixmap);
	}

	/* Draw list? */ 
	if((amount == CLIST_DRAW_AMOUNT_COMPLETE) ||
	   (amount == CLIST_DRAW_AMOUNT_LIST)
	)
	{
	    bool_t is_selected;
	    int	i, n, m, x, y, x2,
		lbearing, rbearing, swidth, ascent, descent,
		font_height, icon_spacing,
		y_bar_width = 0,
		column_width,
		row_height = MAX((int)clist->row_height, 1),
		items_visible;
	    unsigned int width, height, icon_width, icon_height;
	    const char *s;
	    rectangle_t rect;
	    font_t *font;
	    win_t w = clist->list;
	    pixmap_t pixmap = clist->list_buf, icon_pixmap;
	    bitmap_t icon_mask;
	    clist_row_struct *row;
	    clist_item_struct *item;
	    scroll_bar_struct *sb = clist->sb;

	    OSWGetWindowGeometry(
		w, NULL, NULL, &width, &height, NULL
	    );

	    /* Get scroll bar sizes */
	    if((int)height < ((clist->total_rows * row_height) +
		(2 * CLIST_FRAME_MARGIN)
	    ))
		y_bar_width = SCROLLBAR_YBAR_WIDTH;
	    else
		y_bar_width = CLIST_FRAME_MARGIN;

	    /* Redraw scroll bars if redrawing complete */
	    if(amount == CLIST_DRAW_AMOUNT_COMPLETE)
		SBarDraw(sb);

	    /* Clear background */
	    OSWClearPixmap(
		pixmap, width, height,
		wglobal->force_mono ?
		    gui->black_pix :
		    wglobal->surface_editable_pix
	    );

	    /* Calculate starting positions */
	    x = 0;
	    y = 2 - ((int)SBarGetPositionY(sb) % row_height);

	    /* Calculate items visible */
	    items_visible = ((int)height - (2 * CLIST_FRAME_MARGIN)) /
		row_height;

	    /* Calculate starting row */
	    i = MAX((int)SBarGetPositionY(sb) / row_height, 0);

	    /* Iterate from the starting row to the last visible
	     * row
	     */
	    m = MIN(clist->total_rows, i + items_visible + 2);
	    while(i < m)
	    {
		row = clist->row[i];
		if(row == NULL)
		{
		    i++;
		    y += clist->row_height;
		    continue;
		}

		/* Is row selected? */
		is_selected = CListIsRowSelected(clist, i);
		if(is_selected)
		{
		    if(wglobal->force_mono)
			OSWSetForeground(gui->white_pix);
		    else
			OSWSetForeground(wglobal->surface_selected_pix);

		    OSWDrawSolidRectangle(
			pixmap,
			x, y,
			width - y_bar_width,
			clist->row_height
		    );
		}
		/* Is row and widget in focus? */
		else if((i == clist->focus_row) && clist->has_focus)
		{
		    OSWSetForeground(
			wglobal->force_mono ?
			    gui->white_pix :
			    wglobal->surface_selected_pix
		    );
		    OSWDrawRectangle(
			pixmap,
			x + 2, y,
			width - y_bar_width - 3,
			clist->row_height - 1
		    );
		} 

		/* Draw each item on this row */
		for(n = 0; n < row->total_items; n++)
		{
		    item = row->item[n];
		    if(item == NULL)
			continue;

		    if(is_selected)
		    {
			if(wglobal->force_mono)
			    OSWSetForeground(gui->black_pix);
			else
			    OSWSetForeground(wglobal->selected_text_pix);
		    }
		    else
		    {
			if(wglobal->force_mono)
			    OSWSetForeground(gui->white_pix);
 			else
			    OSWSetForeground(item->text_c);
		    }

		    /* Set font */
		    font = item->font;
		    OSWSetFont(font);

		    /* Get column position and width */
		    x2 = CListGetColumnPosition(clist, n);
		    if((n + 1) < row->total_items)
			column_width = CListGetColumnPosition(clist, n + 1) -
			    x2;
		    else
			column_width = (int)width - x2;

		    /* No column width worth drawing? */
		    if(column_width <= 0)
			continue;

		    /* Set clip rectangle to draw string so that it
		     * does not bleed into the next column if this
		     * column is too small to fit the string
		     */
		    rect.x = x2;
		    rect.y = 0;
		    rect.width = (unsigned int)MAX(
			(column_width - 1), 0
		    );
		    rect.height = height;
		    OSWSetClipRectangle(&rect);

		    /* Get the pixmap */
		    icon_pixmap = item->pixmap;
		    icon_mask = item->mask;
		    if(icon_pixmap != None)
		    {
			OSWGetPixmapGeometry(
			    icon_pixmap, &icon_width, &icon_height, NULL
			);
			icon_spacing = STRISEMPTY(item->text) ? 0 : 2;
		    }
		    else
		    {
			icon_width = 0;
			icon_height = 0;
			icon_spacing = 0;
		    }

		    /* Get the text */
		    s = item->text;
		    OSWGetTextBounds(
			font, s, -1,
			&lbearing, &rbearing, &swidth,
			&ascent, &descent
		    );
		    font_height = ascent + descent;

		    /* Draw the pixmap */
		    if(icon_pixmap != None)
		    {
			int x3, y3;
			if(item->flags & CLIST_ITEM_FLAG_ALIGN_CENTER)
			{
			    const int twidth = (int)icon_width +
				icon_spacing + swidth;
			    x3 = x2 + ((column_width - twidth) / 2);
			    y3 = y + ((row_height - (int)icon_height) / 2);
			}
			else if(item->flags & CLIST_ITEM_FLAG_ALIGN_RIGHT)
			{
			    const int twidth = (int)icon_width +
				icon_spacing + swidth;
			    if((n + 1) == clist->total_columns)
			    {
				x3 = x2 + column_width - twidth -
				    (2 * CLIST_FRAME_MARGIN) -
				    SCROLLBAR_YBAR_WIDTH;
				y3 = y + ((row_height - (int)icon_height) / 2);
			    }
			    else
			    {
				x3 = x2 + column_width - twidth -
				    (2 * CLIST_FRAME_MARGIN);
				y3 = y + ((row_height - (int)icon_height) / 2);
			    }
			}
			else	/* CLIST_ITEM_FLAG_ALIGN_LEFT */
			{
			    x3 = x2 + (2 * CLIST_FRAME_MARGIN);
			    y3 = y + ((row_height - (int)icon_height) / 2);
			}
			OSWSetClipMask(icon_mask); 
			OSWSetClipOrigin(x3, y3);
			OSWCopyDrawablesCoord(
			    pixmap, icon_pixmap,
			    x3 + item->x_offset,
			    y3 + item->y_offset,
			    icon_width, icon_height,
			    0, 0
			);
			OSWSetClipMask(None);
		    }

		    /* Draw the text */
		    if(s != NULL)
		    {
			int x3, y3;
			if(item->flags & CLIST_ITEM_FLAG_ALIGN_CENTER)
			{
			    const int twidth = (int)icon_width +
				icon_spacing + swidth;
			    x3 = x2 + ((column_width - twidth) / 2);
			    y3 = y + ((row_height - font_height) / 2);
			}
			else if(item->flags & CLIST_ITEM_FLAG_ALIGN_RIGHT)
			{
			    const int twidth = (int)icon_width +
				icon_spacing + swidth;
			    if((n + 1) == clist->total_columns)
			    {
				x3 = x2 + column_width - twidth -
				    (2 * CLIST_FRAME_MARGIN) -
				    SCROLLBAR_YBAR_WIDTH;
				y3 = y + ((row_height - font_height) / 2);
			    }
			    else
			    {
				x3 = x2 + column_width - twidth -
				    (2 * CLIST_FRAME_MARGIN);
				y3 = y + ((row_height - font_height) / 2);
			    }
			}
			else	/* CLIST_ITEM_FLAG_ALIGN_LEFT */
			{
			    x3 = x2 + (2 * CLIST_FRAME_MARGIN) +
				    (int)icon_width + icon_spacing;
			    y3 = y + ((row_height - font_height) / 2);
			}
			OSWDrawString(
			    pixmap,
			    x3 + item->x_offset - lbearing,
			    y3 + item->y_offset + ascent,
			    s
			);
		    }
		}
		OSWSetClipRectangle(NULL);

		i++;
		y += clist->row_height;
	    }

	    WidgetFrameButtonPixmap(
		pixmap,
		True,
		width,
		height,
		wglobal->force_mono ?
		    gui->white_pix :
		    wglobal->surface_highlight_pix,
		wglobal->force_mono ?
		    gui->white_pix :
		    wglobal->surface_shadow_pix
	    );


	    OSWPutDrawableToWindow(w, pixmap);
	}
}

/*
 *	Queues the clist to draw.
 */
void CListQueueDraw(clist_struct *clist, int amount)
{
	if(clist == NULL)
	    return;

	WidgetQueueDrawAction(CListDrawCB, clist, (void *)amount);
}

/*
 *	Draws the clist's resize bar.
 */
static void CListDrawResizeBar(
	clist_struct *clist, int pos,
	bool_t has_previous
)
{
#if defined(HAVE_X)
	int x, y;
	unsigned int width, height, depth;
	win_t w;
	gc_t gc;
	osw_gui_struct *gui = &osw_gui[0];

	if(clist == NULL)
	    return;

	w = clist->toplevel;
	gc = clist->invert_gc;
	if((w == None) || (gc == None))
	    return;

	OSWGetWindowGeometry(w, &x, &y, &width, &height, &depth);

	/* Need to clear previous draw? */
	if(has_previous)
	{
	    x = clist->resizing_column_bar_pos;
	    XDrawLine(
		gui->display, (Drawable)w, gc,
		x, 0,
		x, (int)height
	    );
	}

	x = pos;
	XDrawLine(
	    gui->display, (Drawable)w, gc,
	    x, 0,
	    x, (int)height
	);

	/* Update last drawn column bar position */
	clist->resizing_column_bar_pos = pos;
#endif
}

/*
 *	Manages the clist event.
 */
int CListManage(clist_struct *clist, event_t *event)
{
	int events_handled = 0;
	keycode_t keycode;
	key_state_t key_state;
	keyval_t keyval;
	win_t w;
	event_key_t *key;
	event_button_t *button;
	event_motion_t *motion;
	event_crossing_t *crossing;
	widget_global_struct *wglobal = &widget_global;

	if((clist == NULL) || (event == NULL))
	    return(events_handled);

	if(!clist->map_state && (event->type != MapNotify))
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == clist->heading)
	    {
		if(clist->map_state)
		    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_HEADING);
		events_handled++;
	    }
	    else if(w == clist->list)
	    {
		if(clist->map_state)
		    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == clist->heading)
	    {
		CListQueueDraw(clist, CLIST_DRAW_AMOUNT_HEADING);
		events_handled++;
	    }
	    else if(w == clist->list)
	    {
		CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
	    if(!clist->has_focus)
		break;

	    keycode = key->keycode;
	    key_state = key->state;
	    keyval = OSWGetKeyValFromKeyCode(keycode, key_state);

	    /* Activate */
	    if((keyval == OSW_KEY_Return) ||
	       (keyval == OSW_KEY_KP_Enter) ||
	       (keyval == OSW_KEY_ISO_Enter)
	    )
	    {
		if(clist->activate_cb != NULL)
		    clist->activate_cb(
			clist,
			clist->activate_data
		    );

		if(!WidgetIsRegistered(clist))
		{
		    events_handled++;
		    return(events_handled);
		}

		events_handled++;
	    }
	    /* Select/unselect */
	    else if(keyval == OSW_KEY_space)
	    {
		if(clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		{
		    if(CListIsRowSelected(clist, clist->focus_row))
			CListUnselectRow(clist, clist->focus_row);
		    else
			CListSelectRow(clist, clist->focus_row);
		}
		events_handled++;
	    }
	    /* Up */
	    else if((keyval == OSW_KEY_Up) ||
		    (keyval == OSW_KEY_KP_Up)
	    )
	    {
		const int prev_focus_row = clist->focus_row;
		if((key->state & OSW_KEY_STATE_SHIFT) &&
		   (clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		)
		{
		    clist->focus_row--;
		    if(clist->focus_row < 0)
			clist->focus_row = 0;
		    if(CListIsRowSelected(clist, clist->focus_row))
		    {
			if(clist->focus_row != prev_focus_row)
			    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    }
		    else
		    {
			CListSelectRow(clist, clist->focus_row);
		    }
		}
		else
		{
		    clist->focus_row--;
		    if(clist->focus_row < 0)
			clist->focus_row = 0;
		    if(clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		    {
			if(clist->focus_row != prev_focus_row)
			    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    }
		    else
		    {
			CListUnselectAllRows(clist);
			CListSelectRow(clist, clist->focus_row);
		    }
		}
		/* Scroll to row if not fully visible */ 
		if(CListGetRowVisibility(
		    clist, clist->focus_row) != VisibilityUnobscured
		)
		    CListMoveTo(clist, clist->focus_row, 0.0f);

		events_handled++;
	    }
	    /* Down */
	    else if((keyval == OSW_KEY_Down) ||
		    (keyval == OSW_KEY_KP_Down)
	    )
	    {
		const int prev_focus_row = clist->focus_row;
		if((key->state & OSW_KEY_STATE_SHIFT) &&
		   (clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		)
		{
		    clist->focus_row++;
		    if(clist->focus_row >= clist->total_rows)
			clist->focus_row = clist->total_rows - 1;
		    if(CListIsRowSelected(clist, clist->focus_row))
		    {
			if(clist->focus_row != prev_focus_row)
			    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    }
		    else
		    {
			CListSelectRow(clist, clist->focus_row);
		    }
		}
		else
		{
		    clist->focus_row++;
		    if(clist->focus_row >= clist->total_rows)
			clist->focus_row = clist->total_rows - 1;
		    if(clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT) 
		    {
			if(clist->focus_row != prev_focus_row)
			    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    }
		    else
		    {
			CListUnselectAllRows(clist);
			CListSelectRow(clist, clist->focus_row);
		    }
		}
		/* Scroll to row if not fully visible */ 
		if(CListGetRowVisibility(
		    clist, clist->focus_row) != VisibilityUnobscured
		)
		    CListMoveTo(clist, clist->focus_row, 1.0f);

		events_handled++;
	    }
	    /* Page Up */
	    else if((keyval == OSW_KEY_Page_Up) ||
		    (keyval == OSW_KEY_KP_Page_Up)
	    )
	    {
		const int	row_height = MAX((int)clist->row_height, 1),
				prev_focus_row = clist->focus_row;
		scroll_bar_struct *sb = clist->sb;
		int rows_visible = ((int)sb->height -
		    (2 * CLIST_FRAME_MARGIN)) / row_height;

		/* Scroll up a page */
		clist->focus_row -= MAX(rows_visible - 1, 1);
		if(clist->focus_row < 0)
		    clist->focus_row = 0;

		/* Select rows? */
		if((key->state & OSW_KEY_STATE_SHIFT) &&
		   (clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		)
		{
		    /* Select rows inbetween (inclusive) */
		    if(clist->focus_row > prev_focus_row)
		    {
			int i;
			for(i = prev_focus_row;
			    i <= clist->focus_row;
			    i++
			)
			    CListSelectRow(clist, i);
		    }
		    else if(clist->focus_row < prev_focus_row)
		    {
			int i;
			for(i = prev_focus_row;
			    i >= clist->focus_row;
			    i--
			)
			    CListSelectRow(clist, i);
		    }
		}
		else
		{
		    if(clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		    {
			if(clist->focus_row != prev_focus_row)
			    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    }
		    else
		    {
			CListUnselectAllRows(clist);
			CListSelectRow(clist, clist->focus_row);
		    }
		}
		/* Scroll to row if not fully visible */
		if(CListGetRowVisibility(
		    clist, clist->focus_row) != VisibilityUnobscured
		)
		    CListMoveTo(clist, clist->focus_row, 0.0f);

		events_handled++;
	    }
	    /* Page Down */
	    else if((keyval == OSW_KEY_Page_Down) ||
		    (keyval == OSW_KEY_KP_Page_Down)
	    )
	    {
		const int       row_height = MAX((int)clist->row_height, 1),
				prev_focus_row = clist->focus_row;
		scroll_bar_struct *sb = clist->sb;
		int rows_visible = ((int)sb->height - 
		    (2 * CLIST_FRAME_MARGIN)) / row_height;

		/* Scroll down a page */
		clist->focus_row += MAX(rows_visible - 1, 1);
		if(clist->focus_row >= clist->total_rows)
		    clist->focus_row = clist->total_rows - 1;

		/* Select rows? */
		if((key->state & OSW_KEY_STATE_SHIFT) &&
		   (clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		)
		{
		    /* Select rows inbetween (inclusive) */
		    if(clist->focus_row > prev_focus_row)
		    {
			int i;
			for(i = prev_focus_row;
			    i <= clist->focus_row;
			    i++
			)
			    CListSelectRow(clist, i);
		    }
		    else if(clist->focus_row < prev_focus_row)
		    {
			int i;
			for(i = prev_focus_row;
			    i >= clist->focus_row;
			    i--
			)
			    CListSelectRow(clist, i);
		    }
		}
		else
		{
		    if(clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		    {
			if(clist->focus_row != prev_focus_row)
			    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    }
		    else
		    {
			CListUnselectAllRows(clist);
			CListSelectRow(clist, clist->focus_row);
		    }
		}
		/* Scroll to row if not fully visible */
		if(CListGetRowVisibility(
		    clist, clist->focus_row) != VisibilityUnobscured
		)
		    CListMoveTo(clist, clist->focus_row, 1.0f);

		events_handled++;
	    }
	    /* Home */
	    else if((keyval == OSW_KEY_Home) ||
		    (keyval == OSW_KEY_KP_Home)
	    )
	    {
		const int prev_focus_row = clist->focus_row;

		/* Scroll to home */
		clist->focus_row = 0;

		/* Select rows? */
		if((key->state & OSW_KEY_STATE_SHIFT) &&
		   (clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		)
		{
		    /* Select rows inbetween (inclusive) */
		    if(clist->focus_row > prev_focus_row)
		    {
			int i;
			for(i = prev_focus_row;
			    i <= clist->focus_row;
			    i++
			)
			    CListSelectRow(clist, i);
		    }
		    else if(clist->focus_row < prev_focus_row)
		    {
			int i;
			for(i = prev_focus_row;
			    i >= clist->focus_row;
			    i--
			)
			    CListSelectRow(clist, i);
		    }
		}
		else
		{
		    if(clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		    {
			if(clist->focus_row != prev_focus_row)
			    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    }
		    else
		    {
			CListUnselectAllRows(clist);
			CListSelectRow(clist, clist->focus_row);
		    }
		}
		/* Scroll to row if not fully visible */
		if(CListGetRowVisibility(
		    clist, clist->focus_row) != VisibilityUnobscured
		)
		    CListMoveTo(clist, clist->focus_row, 0.0f);

		events_handled++;
 	    }
	    /* End */
	    else if((keyval == OSW_KEY_End) ||
		    (keyval == OSW_KEY_KP_End)
	    )
	    {
		const int prev_focus_row = clist->focus_row;

		/* Scroll to end */
		clist->focus_row = clist->total_rows - 1;
		if(clist->focus_row < 0)
		    clist->focus_row = 0;

		/* Select rows? */
		if((key->state & OSW_KEY_STATE_SHIFT) &&
		   (clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		)
		{
		    /* Select rows inbetween (inclusive) */
		    if(clist->focus_row > prev_focus_row)
		    {
			int i;
			for(i = prev_focus_row;
			    i <= clist->focus_row;
			    i++
			)
			    CListSelectRow(clist, i);
		    }
		    else if(clist->focus_row < prev_focus_row)
		    {
			int i;
			for(i = prev_focus_row;
			    i >= clist->focus_row;
			    i--
			)
			    CListSelectRow(clist, i);
		    }
		}
		else
		{
		    if(clist->options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		    {
			if(clist->focus_row != prev_focus_row)
			    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    }
		    else
		    {
			CListUnselectAllRows(clist);
			CListSelectRow(clist, clist->focus_row);
		    }
		}
		/* Scroll to row if not fully visible */
		if(CListGetRowVisibility(
		    clist, clist->focus_row) != VisibilityUnobscured
		)
		    CListMoveTo(clist, clist->focus_row, 1.0f);

		events_handled++;
 	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
	    if(!clist->has_focus)
		break;

	    break;

	  case ButtonPress:
	    button = &event->xbutton;

	    /* Toplevel */
	    if(w == clist->toplevel)
	    {
		clist->has_focus = True;
	    }
	    /* List window */
	    else if(w == clist->list)
	    {
		const int	button_num = button->button;

		clist->has_focus = True;
		clist->button_pressed = button_num;

		/* Select/unselect row */
		if(button_num == Button1)
		{
		    const bool_t	ctrl_key_state = (button->state & OSW_KEY_STATE_CTRL) ? True : False,
					shift_key_state = (button->state & OSW_KEY_STATE_SHIFT) ? True : False;
		    const int row_num = CListGetRowFromCoordinateY(
			clist, button->y
		    );
		    const unsigned int options = clist->options;
		    bool_t single_select;
		    time_t cur_time, last_time;

		    /* Allow multiple selections? */
		    if(options & CLIST_FLAG_ALLOW_MULTI_SELECT)
		    {
			if(ctrl_key_state || shift_key_state)
			    single_select = False;
			else
			    single_select = True;
		    }
		    else
		    {
		        single_select = True;
		    }

		    /* Set the new focus row */
		    clist->focus_row = row_num;

		    /* Select/unselect row? */
		    if(ctrl_key_state &&
		       (options & CLIST_FLAG_ALLOW_MULTI_SELECT) &&
		       (row_num > -1)
		    )
		    {
		        if(CListIsRowSelected(clist, row_num))
			    CListUnselectRow(clist, row_num);
		        else
			    CListSelectRow(clist, row_num);
		    }
		    /* Select a sequence of rows? */
		    else if(shift_key_state &&
		            (options & CLIST_FLAG_ALLOW_MULTI_SELECT) &&
			    (row_num > -1)
		    )
		    {
		        const int n = (clist->total_sel_rows > 0) ?
			    clist->sel_row[clist->total_sel_rows - 1] : 0;

			if((n >= 0) && (n < clist->total_rows))
			{
			    if(row_num == n)
			    {
				/* Same item, don't do anything */
			    }
			    else if(row_num > n)
			    {
				int j;
				for(j = row_num; j > n; j--)
				    CListSelectRow(clist, j);
			    }
			    else
			    {
				int j;
				for(j = row_num; j < n; j++)
				    CListSelectRow(clist, j);
			    }
			}
		    }
		    /* Regular select */
		    else if(row_num > -1)
		    {
			/* Clicked on one row when more than one row
			 * is selected and not selected multiple?
			 */
			if((options & CLIST_FLAG_ALLOW_MULTI_SELECT) &&
			    !ctrl_key_state && !shift_key_state &&
			    (clist->total_sel_rows > 1)
			)
			    CListUnselectAllRows(clist);

			if(!CListIsRowSelected(clist, row_num))
			{
			    if(clist->total_sel_rows > 0)
				CListUnselectAllRows(clist);
			    CListSelectRow(clist, row_num);
			}
		    }

		    /* Check for double click */
		    cur_time = OSWTimeMS();
		    last_time = clist->button_pressed_time;
		    if(((last_time + wglobal->double_click_int) >= cur_time) &&
		       (clist->button_pressed_row == row_num) &&
		       (clist->total_sel_rows > 0) &&
		       single_select
		    )
		    {
			if(clist->activate_cb != NULL)
			{
			    clist->activate_cb(
				clist,
				clist->activate_data
			    );
			    if(!WidgetIsRegistered(clist))
			    {
				events_handled++;
				return(events_handled);
			    }
			}

			clist->button_pressed_row = -1;
			clist->button_pressed_time = 0l;
		    }
		    else
		    {
			clist->button_pressed_row = row_num;
			clist->button_pressed_time = cur_time;
		    }

		    /* Scroll to row if not fully visible */
		    if(CListGetRowVisibility(clist, row_num) != VisibilityUnobscured)
			CListMoveTo(clist, row_num, 0.5f);

		    events_handled++;
		}
		/* Scroll up button pressed? */
		else if(button_num == Button4)
		{
		    scroll_bar_struct *sb = clist->sb;
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) - sb->y_step_inc
		    );
		    events_handled++;
		}
		/* Scroll down button pressed? */
		else if(button_num == Button5)
		{
		    scroll_bar_struct *sb = clist->sb;
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) + sb->y_step_inc
		    );
		    events_handled++;
		}
	    }
	    /* Heading */
	    else if(w == clist->heading)
	    {
		const int	button_num = button->button,
				i = clist->pointer_over_column_resize,
				x = (int)button->x;

		clist->has_focus = True;
		clist->button_pressed = button->button;

		/* Scroll button pressed? */
		if(button_num == Button4)
		{
		    scroll_bar_struct *sb = clist->sb;
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) - sb->y_step_inc
		    );
		    events_handled++;
		    break;
		}
		else if(button_num == Button5)
		{
		    scroll_bar_struct *sb = clist->sb;
		    SBarSetPositionY(
			sb,
			SBarGetPositionY(sb) + sb->y_step_inc
		    );
		    events_handled++;
		    break;
		}

		/* Clicked on a valid column? */
		if((i >= 0) && (i < clist->total_columns) &&
		   (i != 0)	/* Column 0 may not be resized */
		)
		{
		    if(clist->column[i] != NULL)
		    {
			WCursor *wcur = wglobal->h_split_wcur;
			clist->resizing_column = i;
			OSWGrabPointerTime(
			    clist->toplevel,
			    True,
			    ButtonReleaseMask | PointerMotionMask,
			    GrabModeAsync, GrabModeAsync,
			    None,
			    (wcur != NULL) ? wcur->cursor : None,
			    (time_t)button->time
			);
			CListDrawResizeBar(clist, x, False);
		    }
		}

		events_handled++;
	    }
	    else
	    {
		/* Unfocus the list as needed but do not report that
		 * the event was handled
		 */
		if(clist->has_focus)
		{
		    clist->has_focus = False;
		    if(clist->map_state)
			CListQueueDraw(clist, CLIST_DRAW_AMOUNT_COMPLETE);
		}
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    /* Toplevel */
	    if(w == clist->toplevel)
	    {
		if(clist->resizing_column > -1)
		{
		    /* Was in drag, so set new heading column
		     * position
		     */
		    const int i = clist->resizing_column;
		    int x = (int)button->x;

		    if((i >= 0) && (i < clist->total_columns))
		    {
			clist_column_struct *column;

			/* Clip left bound */
			if((i - 1) >= 0)
			{
			    column = clist->column[i - 1];
			    if(column != NULL)
			    {
				if(x < column->x_pos)
				    x = column->x_pos;
			    }
			}
			/* Clip right bound */
			if((i + 1) < clist->total_columns)
			{
			    column = clist->column[i + 1];
			    if(column != NULL)
			    {
				if(x >= column->x_pos)
				    x = column->x_pos - 1;
			    }
			}

			/* Set the column's new position */
			column = clist->column[i];
			if(column != NULL)
			    column->x_pos = x;
		    }

		    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_HEADING);
		    CListQueueDraw(clist, CLIST_DRAW_AMOUNT_LIST);
		    events_handled++;
		}

		clist->button_pressed = 0;
	    }
	    else if(w == clist->heading)
	    {
		clist->button_pressed = 0;
	    }
	    else if(w == clist->list)
	    {
		clist->button_pressed = 0;
	    }

	    /* Stop heading resize as needed */
	    if(clist->resizing_column > -1)
	    {
		int i = clist->resizing_column;

		clist->resizing_column = -1;
		clist->resizing_column_bar_pos = 0;
		OSWUngrabPointerTime((time_t)button->time);

		if((i >= 0) && (i < clist->total_columns))
		{
		    clist_column_struct *column = clist->column[i];
		    if(clist->column_resized_cb != NULL)
			clist->column_resized_cb(
			    clist,
			    i,
			    column->x_pos,
			    clist->column_resized_data
			);
		}
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    /* Heading bar */
	    if((w == clist->toplevel) ||
	       (w == clist->heading)
	    )
	    {
		/* Currently resizing a column? */
		if(clist->resizing_column > -1)
		{
		    const int i = clist->resizing_column;
		    int x = (int)motion->x;

		    if((i >= 0) && (i < clist->total_columns))
		    {
			clist_column_struct *column;

			/* Clip left bound */
			if((i - 1) >= 0)
			{
			    column = clist->column[i - 1];
			    if(column != NULL)
			    {
				if(x < column->x_pos)
				    x = column->x_pos;
			    }
			}
			/* Clip right bound */
			if((i + 1) < clist->total_columns)
			{
			    column = clist->column[i + 1];
			    if(column != NULL)
			    {
				if(x >= column->x_pos)
				    x = column->x_pos - 1;
			    }
			}
		    }

		    /* Redraw resize bar */
		    CListDrawResizeBar(clist, x, True);
		}
		else
		{
		    /* No column being resized, so check for pointer
		     * motion over a column and update cursor
		     */
		    int	margin = 3,
			i,
			x = (int)motion->x;
		    clist_column_struct *column;

		    for(i = 0; i < clist->total_columns; i++)
		    {
			column = clist->column[i];
		        if(column == NULL)
			    continue;

			if((x >= (column->x_pos - margin)) &&
			   (x < (column->x_pos + margin))
			)
			{
			    if(i != clist->pointer_over_column_resize)
			    {
				WidgetSetWindowCursor(
				    clist->heading,
				    wglobal->h_split_wcur
				);
				clist->pointer_over_column_resize = i;
			    }
			    break;
			}
		    }
		    if(i >= clist->total_columns)
		    {
			if(clist->pointer_over_column_resize > -1)
			{
			    OSWUnsetWindowCursor(clist->heading);
			    clist->pointer_over_column_resize = -1;
			}
		    }
		}
	    }
	    /* List window */
	    else if(w == clist->list)
	    {


	    }
	    break;

	  case EnterNotify:
	    crossing = &event->xcrossing;
	    if(w == clist->heading)
	    {
		int margin = 3,
		    i,
		    x = (int)crossing->x;
		clist_column_struct *column;

		for(i = 0; i < clist->total_columns; i++)
		{
		    column = clist->column[i];
		    if(column == NULL)
			continue;

		    if((x >= (column->x_pos - margin)) &&
		       (x < (column->x_pos + margin))
		    )
		    {
			if(i != clist->pointer_over_column_resize)
			{
			    WidgetSetWindowCursor(
				clist->heading,
				wglobal->h_split_wcur
			    );
			    clist->pointer_over_column_resize = i;
			}
			break;
		    }
		}
		events_handled++;
	    }
	    break;

	  case LeaveNotify:
	    crossing = &event->xcrossing;
	    if(w == clist->heading)
	    {
		if(clist->pointer_over_column_resize > -1)
		{
		    OSWUnsetWindowCursor(clist->heading);
		    clist->pointer_over_column_resize = -1;
		}
		events_handled++;
	    }
	    break;
	}

	if(events_handled == 0)
	    events_handled += WidgetManage(clist->sb, event);

	return(events_handled);
}

/*
 *	Maps the clist.
 */
void CListMap(clist_struct *clist)
{
	if(clist == NULL)
	    return;

	if(!clist->map_state)
	{
	    OSWMapRaised(clist->toplevel);
	    OSWMapWindow(clist->heading);
	    OSWMapWindow(clist->list);
	    clist->map_state = True;
	    WidgetQueueDraw(clist->sb);
	}
}

/*
 *	Unmaps the clist.
 */
void CListUnmap(clist_struct *clist)
{
	if(clist == NULL)
	    return; 

	OSWUnmapWindow(clist->toplevel);
	clist->map_state = False;

	OSWDestroyPixmap(&clist->heading_buf);
	OSWDestroyPixmap(&clist->list_buf);
}

/*
 *	Destroys the clist.
 */
void CListDestroy(clist_struct *clist)
{
	if(clist == NULL)
	    return;

	/* Delete all the rows and selections */
	CListClear(clist);

	/* Delete all the headings */
	if(clist->total_columns > 0)
	{
	    int i;
	    for(i = 0; i < clist->total_columns; i++)
		CListColumnRemove(clist, i);
	    free(clist->column);
	    clist->column = NULL;
	    clist->total_columns = 0;
	}

	WidgetDestroy(clist->sb);
	free(clist->sb);
	OSWDestroyPixmap(&clist->heading_buf);
	OSWDestroyWindow(&clist->heading);
	OSWDestroyPixmap(&clist->list_buf);
	OSWDestroyWindow(&clist->list);
	OSWDestroyWindow(&clist->toplevel);

#ifdef HAVE_X
	XFreeGC(osw_gui[0].display, clist->invert_gc);
	clist->invert_gc = None;
#endif

	clist->type = WIDGET_TYPE_UNKNOWN;
	clist->map_state = False;
	clist->has_focus = False;
	clist->disabled = False;
	clist->x = 0;
	clist->y = 0;
	clist->width = 0;
	clist->height = 0;
	clist->visibility_state = VisibilityFullyObscured;
	clist->font = NULL;
	clist->next = NULL;
	clist->prev = NULL;

	clist->options = 0;
	clist->row_height = CLIST_DEF_ROW_HEIGHT;

	clist->button_pressed = 0;
	clist->button_pressed_row = -1;
	clist->button_pressed_time = 0l;

	clist->activate_cb = NULL;
	clist->activate_data = NULL;
	clist->select_cb = NULL;
	clist->select_data = NULL;
	clist->unselect_cb = NULL;
	clist->unselect_data = NULL;
	clist->column_resized_cb = NULL;
	clist->column_resized_data = NULL;
}
