#include <stdlib.h>
#include <string.h>

#include <X11/xpm.h>

#include "../include/graphics.h"
#include "../include/osw-x.h"
#include "../include/widget.h"
#include "../include/tga.h"


image_t *WidgetLoadImageFromTgaFile(const char *path);
image_t *WidgetLoadImageFromTgaData(const u_int8_t *data);

pixmap_t WidgetLoadPixmapFromTgaFile(
	const char *path, bitmap_t *mask_rtn
);
pixmap_t WidgetLoadPixmapFromTgaData(
	u_int8_t *data, bitmap_t *mask_rtn
);

image_t *WidgetLoadImageFromXpmFile(const char *path);
image_t *WidgetLoadImageFromXpmData(char **data);

pixmap_t WidgetLoadPixmapFromXpmFile(
	const char *path, bitmap_t *mask_rtn
);
pixmap_t WidgetLoadPixmapFromXpmData(
	char **data, bitmap_t *mask_rtn
);


#define ATOI(s)		(((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)		(((s) != NULL) ? atol(s) : 0)
#define ATOF(s)		(((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)	(((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)	(((a) > (b)) ? (a) : (b))
#define MIN(a,b)	(((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)	(MIN(MAX((a),(l)),(h)))
#define STRLEN(s)	(((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)	(((s) != NULL) ? (*(s) == '\0') : 1)


static void WidgetCopyTgaDataToXImageData(
	image_t *img, tga_data_struct *td
)
{
	int	x_offset,
		src_bpp, tar_bpp,
		src_bpl, tar_bpl;
	unsigned int width, height;
	u_int8_t *tar_data;
	const u_int8_t *src_data;

	if((img == NULL) || (td == NULL))
	    return;

	tar_data = (u_int8_t *)img->data;
	src_data = (const u_int8_t *)td->data;
	x_offset = img->xoffset;
	width = MIN((unsigned int)img->width, (unsigned int)td->width);
	height = MIN((unsigned int)img->height, (unsigned int)td->height);
	tar_bpp = img->bits_per_pixel / 8;
	src_bpp = td->data_bits_per_pixel / 8;
	tar_bpl = img->bytes_per_line;
	src_bpl = (int)td->width * src_bpp;

	if((tar_data == NULL) || (src_data == NULL) ||
	   (width == 0) || (height == 0)
	)
	    return;

	/* Allocated pixel sizes the same? */
	if(tar_bpp == src_bpp)
	{
	    /* Copy line by line */
	    const int	bpp = tar_bpp,
			content_bpl = ((int)width - x_offset) * bpp;
	    int y;

	    if(content_bpl > 0)
	    {
		for(y = 0; y < (int)height; y++)
		{
		    memcpy(
			tar_data + (y * tar_bpl) + (x_offset * bpp),
			src_data + (y * src_bpl),
			content_bpl
		    );
		}
	    }
	}
	else
	{
	    /* Different pixel sizes, copy pixel by pixel */
	    const int min_bpp = MIN(tar_bpp, src_bpp);
	    int tx, ty, sx, sy;

	    for(sy = 0, ty = 0; ty < (int)height; sy++, ty++)
	    {
		for(sx = 0, tx = x_offset; tx < (int)width; sx++, tx++)
		{
		    memcpy(
			tar_data + (ty * tar_bpl) + (tx * tar_bpp),
			src_data + (sy * src_bpl) + (sx * src_bpp),
			min_bpp
		    );
		}
	    }
	}
}

/*
 *      Loads an image from a TGA file.
 *      The image's depth will be that of the GUI's.
 */
image_t *WidgetLoadImageFromTgaFile(const char *path)
{
	int status = 0;
	tga_data_struct *td;
	image_t *img = NULL;
	osw_gui_struct *gui = &osw_gui[0];

	if(path == NULL)
	    return(img);

	td = (tga_data_struct *)malloc(sizeof(tga_data_struct));
	if(td == NULL)
	    return(img);

	/* Load tga image from file */
	status = TgaReadFromFile(
	    path,
	    td,
	    TgaGetDepthForX(
		gui->bits_per_pixel,
		gui->depth
	    )
	);
	if(status != TgaSuccess)
	{
	    fprintf(stderr, "WidgetLoadImageFromTgaFile(): %s: ", path);
	    switch(status)
	    {
	      case TgaNoBuffers:
		fprintf(stderr, "No buffers.\n");
		break;

	      case TgaBadHeader:
		fprintf(stderr, "Bad header.\n");
		break;

	      case TgaBadValue:
		fprintf(stderr, "Bad value.\n");
		break;

	      case TgaNoFile:
		fprintf(stderr, "No such file.\n");
		break;

	      case TgaNoAccess:
		fprintf(stderr, "No access.\n");
		break;

	      default:
		fprintf(stderr, "\n");
		break;
	    }

	    TgaDestroyData(td);
	    free(td);

	    return(img);
	}

	/* Create GUI image */
	if(OSWCreateImage(&img, td->width, td->height))
	{
	    fprintf(
		stderr,
"%s: Error: Unable to allocate GUI image.\n",
		path
	    );
	}
	else
	{
	    WidgetCopyTgaDataToXImageData(img, td);
	}

	/* Destroy the tga data */
	TgaDestroyData(td);
	free(td);

	return(img);
}


/*
 *	Loads an image from TGA data in memory.
 *	The image's depth will be that of the GUI's.
 */
image_t *WidgetLoadImageFromTgaData(const u_int8_t *data)
{
	int status = 0;
	tga_data_struct *td;
	image_t *img = NULL;
	osw_gui_struct *gui = &osw_gui[0];

	if(data == NULL)
	    return(img);

	td = (tga_data_struct *)malloc(sizeof(tga_data_struct));
	if(td == NULL)
	    return(img);

	/* Load tga image from data */
	status = TgaReadFromData(
	    data,
	    td,
            TgaGetDepthForX(
                gui->bits_per_pixel,
                gui->depth
            )
	);
	if(status != TgaSuccess)
	{
	    fprintf(stderr, "WidgetLoadImageFromTgaData(): %p: ", data);
	    switch(status)
	    {
	      case TgaNoBuffers:
		fprintf(stderr, "No buffers.\n");
		break;

	      case TgaBadHeader:
		fprintf(stderr, "Bad header.\n");
		break;

	      case TgaBadValue:
		fprintf(stderr, "Bad value.\n");
		break;

	      case TgaNoAccess:
		fprintf(stderr, "No access.\n");
		break;

	      default:
		fprintf(stderr, "\n");
		break;
	    }

	    TgaDestroyData(td);
	    free(td);

	    return(img);
	}

	/* Create GUI image */
	if(OSWCreateImage(&img, td->width, td->height))
	{
	    fprintf(
		stderr,
"%p: Error: Unable to allocate GUI image.\n",
		data
	    );
	}
	else
	{
	    WidgetCopyTgaDataToXImageData(img, td);
	}

	/* Destroy the tga data */
	TgaDestroyData(td);
	free(td);

	return(img);
}


/*
 *      Loads a pixmap from a TGA file.
 *      The pixmap's depth will match that of the GUI's.
 */
pixmap_t WidgetLoadPixmapFromTgaFile(
	const char *path, bitmap_t *mask_rtn 
)
{
	image_t *img;
	pixmap_t pixmap;

	if(mask_rtn != NULL)
	    *mask_rtn = 0;

	img = WidgetLoadImageFromTgaFile(path);
	if(img == NULL)
	    return(0);

	pixmap = OSWCreatePixmapFromImage(img);
	if(mask_rtn != NULL)
	    *mask_rtn = OSWCreateMaskFromImage(img);
	OSWDestroyImage(&img);

	return(pixmap);
}


/*
 *      Loads a pixmap from TGA data in memory.
 *      The pixmap's depth will match that of the GUI's.
 */
pixmap_t WidgetLoadPixmapFromTgaData(
	u_int8_t *data, bitmap_t *mask_rtn
)
{
	image_t *img;
	pixmap_t pixmap = 0;

	if(mask_rtn != NULL)
	    *mask_rtn = 0;

	img = WidgetLoadImageFromTgaData(data);
	if(img == NULL)
	    return(pixmap);

	pixmap = OSWCreatePixmapFromImage(img);
	if(mask_rtn != NULL)
	    *mask_rtn = OSWCreateMaskFromImage(img);
	OSWDestroyImage(&img);

	return(pixmap);
}


/*
 *	Loads an image from an XPM file.
 *	The image's depth will match that of the GUI's.
 */
image_t *WidgetLoadImageFromXpmFile(const char *path)
{
#if defined(HAVE_X)
	int status;
	char *dpath;
	XImage *ximg, *imgmask;
	XpmAttributes xpmattr;
	osw_gui_struct *gui = &osw_gui[0];

	if((gui->root_win == 0) || (path == NULL))
	    return(0);

	/* Set XPM attributes */
	memset(&xpmattr, 0x00, sizeof(XpmAttributes));
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
	xpmattr.closeness = WIDGET_XPM_DEF_COLOR_CLOSENESS;
	xpmattr.depth = gui->depth;

	/* Load image from XPM file */
	dpath = strdup(path);
	status = XpmReadFileToImage(
	    gui->display,
	    dpath,
	    &ximg,
	    &imgmask,
	    &xpmattr
	);
	free(dpath);
	if(status != XpmSuccess)
	{
	    fprintf(
		stderr,
"%s: Unable to open XPM file.",
		path
	    );
	    return(0);
	}

	/* Destroy the mask image */
	if(imgmask != NULL)
	    XDestroyImage(imgmask);

	return(OSWCreateImageFromXImage(ximg));
#else
#warning WidgetLoadImageFromXpmFile() is not supported on this target platform
 	return(NULL);
#endif
}


/*
 *      Loads an image from XPM data in memory.
 *      The image's depth will match that of the GUI's.
 */
image_t *WidgetLoadImageFromXpmData(char **data)
{
#if defined(HAVE_X)
	int status;
	XImage *ximg, *ximgmask;
	XpmAttributes xpmattr;
	osw_gui_struct *gui = &osw_gui[0];

	if((gui->root_win == 0) || (data == NULL))
	    return(0);

	/* Set XPM attributes */   
	memset(&xpmattr, 0x00, sizeof(XpmAttributes));
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
	xpmattr.closeness = WIDGET_XPM_DEF_COLOR_CLOSENESS;
	xpmattr.depth = gui->depth;

	/* Load image from XPM data */
	status = XpmCreateImageFromData(
	    gui->display,
	    data,
	    &ximg,
	    &ximgmask,
	    &xpmattr
	);
	if(status != XpmSuccess)
	{
	    fprintf(
		stderr,
"%p: Unable to open XPM data.\n",
		data
	    );
	    return(0);
	}

	/* Destroy the mask image */
	if(ximgmask != NULL)
	    XDestroyImage(ximgmask);

	return(OSWCreateImageFromXImage(ximg));
#else
#warning WidgetLoadImageFromXpmData() is not supported on this target platform
	return(NULL);
#endif
}


/*
 *	Opens the pixmap from the XPM file.
 */
pixmap_t WidgetLoadPixmapFromXpmFile(
	const char *path, bitmap_t *mask_rtn
)
{
	int status;
	char *dpath;
	pixmap_t pixmap = None;
	bitmap_t mask = None;
	XpmAttributes xpmattr;
	osw_gui_struct *gui = &osw_gui[0];

	if(mask_rtn != NULL)
	    *mask_rtn = None;

	if((gui->root_win == None) || (path == NULL))
	    return(None);

	/* Set XPM attributes */
	memset(&xpmattr, 0x00, sizeof(XpmAttributes));
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
	xpmattr.closeness = WIDGET_XPM_DEF_COLOR_CLOSENESS;
	xpmattr.depth = gui->depth;

	/* Attempt to read the pixmap data from file */
	dpath = strdup(path);
	status = XpmReadFileToPixmap(
	    gui->display,
	    gui->root_win,
	    dpath,
	    &pixmap,
	    &mask,
	    &xpmattr
	);
	free(dpath);
	if(status != XpmSuccess)
	{
	    fprintf(
		stderr,
"%s: Unable to open XPM file.",
		path
	    );
	    return(None);
	}

	if(mask_rtn != NULL)
	    *mask_rtn = mask;
	else
	    OSWDestroyBitmap(&mask);

	return(pixmap);
}

/*
 *	Opens the pixmap from the XPM data.
 */
pixmap_t WidgetLoadPixmapFromXpmData(
	char **data, bitmap_t *mask_rtn
)
{
	int status;
        pixmap_t pixmap = None;
        bitmap_t mask = None;
	XpmAttributes xpmattr;
	osw_gui_struct *gui = &osw_gui[0];

        if(mask_rtn != NULL)
            *mask_rtn = None;

	if((gui->root_win == None) || (data == NULL))
	    return(0);

	/* Set XPM attributes */
	memset(&xpmattr, 0x00, sizeof(XpmAttributes));
	xpmattr.valuemask = XpmSize | XpmCloseness | XpmDepth;
	xpmattr.closeness = WIDGET_XPM_DEF_COLOR_CLOSENESS;
	xpmattr.depth = gui->depth;

	/* Attempt to read the pixmap data from file */
	status = XpmCreatePixmapFromData(
	    gui->display,
	    gui->root_win,
	    data,
	    &pixmap,
	    &mask,
	    &xpmattr
	);
	if(status != XpmSuccess)
	{
	    fprintf(
		stderr,
"%p: Unable to open XPM data.\n",
		data
	    );
	    return(None);
	}

        if(mask_rtn != NULL)
            *mask_rtn = mask;
        else
            OSWDestroyBitmap(&mask);

	return(pixmap);
}
