#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "../include/string.h"
#include "../include/widget.h"


static void MenuBarResizeCB(void *widget, void *data);
static void MenuBarDrawCB(void *widget, void *data);

int MenuBarIsItemAllocated(menu_bar_struct *mb, int n);
menu_struct *MenuBarGetMenuFromItem(menu_bar_struct *mb, int n);
bool_t MenuBarGetItemToggleState(
	menu_bar_struct *mb,
	int menu_num, int item_num
);
void MenuBarSetItemToggleState(
	menu_bar_struct *mb,
	int menu_num, int item_num,
	bool_t state
);

int MenuBarMatchItemByPos(menu_bar_struct *mb, int x, int y);

int MenuBarMenuAppend(
	menu_bar_struct *mb,
	const char *text,
	int x, int y,
	unsigned int width, unsigned int height
);
int MenuBarMenuItemInsert(
	menu_bar_struct *mb, int menu_num, int item_num,
	int type,		/* One of MENU_ITEM_* */
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	int id
);
int MenuBarMenuItemAppend(
	menu_bar_struct *mb, int menu_num,
	int type,		/* One of MENU_ITEM_* */
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	int id
);
void MenuBarDeleteItem(menu_bar_struct *mb, int n);

int MenuBarInit(
	menu_bar_struct *mb, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	void (*activate_cb)(menu_struct *, int, void *),
	void *activate_data
);
void MenuBarResize(menu_bar_struct *mb);
void MenuBarQueueResize(menu_bar_struct *mb);
void MenuBarDraw(menu_bar_struct *mb);
void MenuBarQueueDraw(menu_bar_struct *mb);
int MenuBarManage(menu_bar_struct *mb, event_t *event);
void MenuBarMap(menu_bar_struct *mb);
void MenuBarUnmap(menu_bar_struct *mb);
void MenuBarDestroy(menu_bar_struct *mb);


#define ATOI(s)         (((s) != NULL) ? atoi(s) : 0)
#define ATOL(s)         (((s) != NULL) ? atol(s) : 0)
#define ATOF(s)         (((s) != NULL) ? atof(s) : 0.0f)
#define STRDUP(s)       (((s) != NULL) ? strdup(s) : NULL)

#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))
#define STRLEN(s)       (((s) != NULL) ? strlen(s) : 0)
#define STRISEMPTY(s)   (((s) != NULL) ? (*(s) == '\0') : 1)


/*
 *	Sizes (in pixels):
 */
#define MB_ITEM_MARGIN_X		5
#define MB_FRAME_MARGIN			2

#define MB_DEF_WIDTH			256
#define MB_DEF_HEIGHT			(25 + (2 * MB_FRAME_MARGIN))


/*
 *	Resize callback.
 */
static void MenuBarResizeCB(void *widget, void *data)
{
	menu_bar_struct *mb = (menu_bar_struct *)widget;
	if(mb == NULL)
	    return;

	MenuBarResize(mb);
}

/*
 *	Draw callback.
 */
static void MenuBarDrawCB(void *widget, void *data)
{
	menu_bar_struct *mb = (menu_bar_struct *)widget;
	if(mb == NULL)
	    return;

	if(mb->map_state)
	    MenuBarDraw(mb);
}


/*
 *	Checks if the menu bar item n is valid and allocated.
 */
int MenuBarIsItemAllocated(menu_bar_struct *mb, int n)
{
	if(mb == NULL)
	    return(0);
	else if((n < 0) || (n >= mb->total_items))
	    return(0);
	else if(mb->item[n] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Returns a menu structure pointer belonging to the menu bar
 *	item n.  Can return NULL on error.
 */
menu_struct *MenuBarGetMenuFromItem(menu_bar_struct *mb, int n)
{
	if(MenuBarIsItemAllocated(mb, n))
	    return(mb->item[n]->menu);
	else
	    return(NULL);
}

/*
 *	Gets the toggle state of the menu item on the specified
 *	menu. Does not distinguish from errors.
 */
bool_t MenuBarGetItemToggleState(
	menu_bar_struct *mb,
	int menu_num, int item_num
) 
{
	menu_struct *menu;

	if(mb == NULL)
	    return(False);

	menu = MenuBarGetMenuFromItem(mb, menu_num);
	if(menu == NULL)
	    return(False);

	return(MenuGetItemState(menu, item_num));
}

/*
 *	Sets the toggle state of the menu item on the specified
 *      menu.
 */
void MenuBarSetItemToggleState(
	menu_bar_struct *mb,
	int menu_num, int item_num,
	bool_t state
)
{
	menu_struct *menu;

	if(mb == NULL)
	    return;

	menu = MenuBarGetMenuFromItem(mb, menu_num);
	if(menu == NULL)
	    return;

	return(MenuSetItemState(menu, item_num, state));
}

/*
 *	Matches a menu bar item by a given positionto be on the
 *	menu bar.
 */
int MenuBarMatchItemByPos(
	menu_bar_struct *mb, 
	int x, int y 
)  
{
	int i;
	menu_bar_item_struct *item;

	if(mb == NULL)
	    return(-1);

	for(i = 0; i < mb->total_items; i++)
	{
	    item = mb->item[i];
	    if(item == NULL)
		continue;

	    if((item->x <= x) &&
	       (item->y <= y) &&
	       ((item->x + (int)item->width) > x) &&
	       ((item->y + (int)item->height) > y)
	    )
		return(i);
	}

	return(-1);
}

   
/*
 *	Appends a menu to the menu bar.
 */
int MenuBarMenuAppend(
	menu_bar_struct *mb,
	const char *text,
	int x, int y,
	unsigned int width, unsigned int height
)
{
	int n, swidth;
	font_t *font;
	menu_struct *menu;
	menu_bar_item_struct *item;
	osw_gui_struct *gui = &osw_gui[0];

	if(mb == NULL)
	    return(-1);

	font = mb->font;

	if(text == NULL)
	    text = "(null)";

	if(mb->total_items < 0)
	    mb->total_items = 0;

	n = mb->total_items;
	mb->total_items = n + 1;
	mb->item = (menu_bar_item_struct **)realloc(
	    mb->item,
	    mb->total_items * sizeof(menu_bar_item_struct *)
	);
	if(mb->item == NULL)
	{
	    mb->total_items = 0;
	    return(-3);
	}

	mb->item[n] = item = (menu_bar_item_struct *)calloc(
	    1, sizeof(menu_bar_item_struct)
	);
	if(item == NULL)
	    return(-3); 

	item->text = STRDUP(text);
	item->x = x;
	item->y = y;

	OSWGetTextBounds(
	    font, text, -1,
	    NULL, NULL, &swidth, NULL, NULL
	);
	item->width = (width != 0) ?
	    width :
	    (swidth + (2 * MB_ITEM_MARGIN_X) + (2 * MB_FRAME_MARGIN));
	item->height = (height != 0) ?
	    height :
	    mb->height;

	/* Create menu */
	item->menu = menu = (menu_struct *)calloc(
	    1, sizeof(menu_struct)
	);
	MenuInit(
	    menu, gui->root_win,
	    mb->activate_cb, mb->activate_data
	);

	return(n);
}

/*
 *	Inserts a new menu item to the specified menu bar menu.
 */
int MenuBarMenuItemInsert(
	menu_bar_struct *mb, int menu_num, int item_num,
	int type,		/* One of MENU_ITEM_* */
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	int id
)
{
	menu_struct *menu;

	if(mb == NULL)
	    return(-1);

	if(menu_num < 0)
	{               
	    menu_bar_item_struct *item = (mb->total_items > 0) ?
		mb->item[mb->total_items - 1] : NULL;
	    menu = (item != NULL) ? item->menu : NULL;
	}                                             
	else
	{
	    menu = MenuBarGetMenuFromItem(mb, menu_num);
	}
	if(menu == NULL)
	    return(-1);

	return(MenuItemInsert(
	    menu, item_num, type, text, pixmap, mask, id
	));
}

/*
 *	Appends a new menu item to the specified menu bar menu.
 */
int MenuBarMenuItemAppend(
	menu_bar_struct *mb, int menu_num,
	int type,		/* One of MENU_ITEM_* */
	const char *text,
	pixmap_t pixmap, bitmap_t mask,
	int id
)
{
	menu_struct *menu;

	if(mb == NULL)
	    return(-1);

	if(menu_num < 0)
	{
	    menu_bar_item_struct *item = (mb->total_items > 0) ?
		mb->item[mb->total_items - 1] : NULL;
	    menu = (item != NULL) ? item->menu : NULL;
	}
	else
	{
	    menu = MenuBarGetMenuFromItem(mb, menu_num);
	}
	if(menu == NULL)
	    return(-1);

	return(MenuItemAppend(
	    menu, type, text, pixmap, mask, id
	));
}

/*
 *	Deletes the menu bar item and its menu.
 */
void MenuBarDeleteItem(menu_bar_struct *mb, int n)
{
	menu_bar_item_struct *item;

	/* Check if menu bar item structure is allocated */
	if(MenuBarIsItemAllocated(mb, n))
	    item = mb->item[n];
	else
	    return;

	/* Delete menu */
	if(item->menu != NULL)  
	{
	    WidgetDestroy(item->menu);
	    free(item->menu);
	}

	free(item->text);

	free(item);
	mb->item[n] = item = NULL;
}


/*
 *	Initializes the menu bar.
 */   
int MenuBarInit(
	menu_bar_struct *mb, win_t parent,
	int x, int y,
	unsigned int width, unsigned int height,
	void (*activate_cb)(menu_struct *, int, void *),
	void *activate_data
)
{
	unsigned int wwidth, wheight;
	win_t w;
	widget_global_struct *wglobal = &widget_global;

	if((mb == NULL) || (parent == None))
	    return(-2);

	OSWGetWindowGeometry(parent, NULL, NULL, &wwidth, &wheight, NULL);
	if(width == 0)
	    width = wwidth;
	if(width == 0)
	    width = MB_DEF_WIDTH;
	if(height == 0)
	    height = MB_DEF_HEIGHT;

	mb->type = WIDGET_TYPE_MENU_BAR;
	mb->map_state = False;
	mb->has_focus = False;
	mb->disabled = False;
	mb->visibility_state = VisibilityFullyObscured;
	mb->x = x;
	mb->y = y;
	mb->width = width;
	mb->height = height;
	mb->font = wglobal->menu_font;
	mb->next = NULL;
	mb->prev = NULL;
	mb->data = NULL;

	/* Create toplevel */
	OSWCreateWindow(
	    &w, parent,
	    mb->x, mb->y,
	    mb->width, mb->height
	);
	mb->toplevel = w;
	mb->toplevel_buf = 0;

	OSWSetWindowInput(
	    w,
	    StructureNotifyMask | ExposureMask |
	    KeyPressMask | KeyReleaseMask |
	    ButtonPressMask | ButtonReleaseMask |
	    PointerMotionMask |
	    VisibilityChangeMask
	);

	mb->activate_cb = activate_cb;
	mb->activate_data = activate_data;

	mb->sel_item = -1;

	/* Add this menu bar to the widget regeristry */
	WidgetRegister(mb, mb->type);

	return(0);
}

/*
 *	Resizes the menu bar.
 */
void MenuBarResize(menu_bar_struct *mb)
{
	int x, y;
	unsigned int width, height;
	win_t w;

	if(mb == NULL)
	    return;

	w = mb->toplevel;
	OSWGetWindowGeometry(w, &x, &y, &width, &height, NULL);

	mb->x = x;
	mb->y = x;
	mb->width = width;
	mb->height = height;

	OSWDestroyPixmap(&mb->toplevel_buf);
}

/*
 *	Queues the menu bar to resize.
 */
void MenuBarQueueResize(menu_bar_struct *mb)
{
	if(mb == NULL)
	    return;

	WidgetQueueResizeAction(MenuBarResizeCB, mb, NULL);
}

/*
 *	Draws the menu bar.
 */
void MenuBarDraw(menu_bar_struct *mb)
{
	int sel_item;
	unsigned int width, height;
	pixmap_t pixmap;
	win_t w;
	osw_gui_struct *gui = &osw_gui[0];
	widget_global_struct *wglobal = &widget_global;

	if(mb == NULL)
	    return;

	w = mb->toplevel;
	width = mb->width;
	height = mb->height;

	/* Recreate buffers as needed */
	if(mb->toplevel_buf == None)
	{
	    if(OSWCreatePixmap(&mb->toplevel_buf, width, height))
		return;
	}
	pixmap = mb->toplevel_buf;

	/* Draw the background */
	if(wglobal->force_mono)
	{
	    OSWClearPixmap(
		pixmap, width, height,
		gui->white_pix
	    );
	    WidgetFrameButtonPixmap(
		pixmap, False, width, height,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_highlight_pix,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_shadow_pix
	    );
	    OSWSetForeground(gui->black_pix);
	}
	else
	{
	    WidgetPutImageTile(
		pixmap, wglobal->menu_bar_bg_img, 
		width, height
	    );
	    WidgetFrameButtonPixmap(
		pixmap, False, width, height,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_highlight_pix,
		(wglobal->force_mono) ?
		    gui->white_pix :
		    wglobal->surface_shadow_pix
	    );
	    OSWSetForeground(wglobal->surface_shadow_pix);
	}

	/* Draw labels */
	sel_item = -1;
	if(mb->total_items > 0)
	{
	    int i, lbearing, rbearing, ascent, descent, font_height;
	    const char *text;
	    font_t *font;
	    menu_bar_item_struct *item; 

	    for(i = 0; i < mb->total_items; i++)
	    {
		item = mb->item[i];
		if(item == NULL)
		    continue;

		text = item->text;
	        if(text == NULL)
		    continue;

		font = mb->font;
		OSWSetFont(font);
	        OSWGetTextBounds(
		    font, item->text, -1,
		    &lbearing, &rbearing, NULL,
		    &ascent, &descent
		);
		font_height = ascent + descent;
		OSWDrawString(
		    pixmap,
		    item->x + MB_FRAME_MARGIN + MB_ITEM_MARGIN_X - lbearing,
		    (((int)item->height - font_height) / 2) + ascent,
		    text
		);

		if(mb->sel_item == i)
		    sel_item = i;
	    }

	    /* Draw frame around the selected item */
	    if((sel_item > -1) && !wglobal->force_mono)
	    {
		int x, y;
		unsigned int width, height;

		item = mb->item[sel_item];

		x = item->x + MB_FRAME_MARGIN;
		y = item->y + MB_FRAME_MARGIN;
		width = (unsigned int)MAX(
		    (item->width - (2 * MB_FRAME_MARGIN)),
		    1
		);
		height = (unsigned int)MAX(
		    (item->height - (2 * MB_FRAME_MARGIN)),
		    1
		);

		OSWSetForeground(wglobal->surface_highlight_pix);

		OSWDrawLine(
		    pixmap,
		    x, y + (int)height - 1,
		    x, y
	        );
	        OSWDrawLine(
		    pixmap,
		    x, y,
		    x + (int)width, y
		);

		OSWSetForeground(wglobal->surface_shadow_pix);

		OSWDrawLine(
		    pixmap,
		    x + (int)width, y + 1,
		    x + (int)width, y + (int)height - 1
		);
		OSWDrawLine(
		    pixmap,
		    x + 1, y + (int)height - 1,
		    x + (int)width, y + (int)height - 1
		);
	    }
	}
	OSWPutDrawableToWindow(w, pixmap);
}

/*
 *	Queues the menu bar to draw.
 */
void MenuBarQueueDraw(menu_bar_struct *mb)
{
	if(mb == NULL)
	    return;

	WidgetQueueDrawAction(MenuBarDrawCB, mb, NULL);
}

/*
 *	Manages the menu bar.
 */
int MenuBarManage(menu_bar_struct *mb, event_t *event)
{
	int i, events_handled = 0;
	win_t w;
	event_button_t *button;
	event_key_t *key;
	event_motion_t *motion;
	menu_struct *menu;
	menu_bar_item_struct *item;

	if((mb == NULL) || (event == NULL))
	    return(events_handled);

	if(!mb->map_state &&
	   (event->type != MapNotify)
	)
	    return(events_handled);

	w = event->xany.window;

	switch(event->type)
	{
	  case ConfigureNotify:
	    if(w == mb->toplevel)
	    {
		if(mb->map_state)
		    MenuBarQueueDraw(mb);
		events_handled++;
	    }
	    break;

	  case Expose:
	    if(w == mb->toplevel)
	    {
		MenuBarQueueDraw(mb);
		events_handled++;
	    }
	    break;

	  case KeyPress:
	    key = &event->xkey;
#if 0
	    if(!mb->has_focus)
		break;
#endif
	    /* Pass the key event unconditionally to each menu for hot
	     * key checks
	     */
	    for(i = 0; i < mb->total_items; i++)
	    {
		item = mb->item[i];
		if(item == NULL)
		    continue;

		menu = item->menu;
		if(menu == NULL)
		    continue;

		events_handled += WidgetManage(menu, event);
		if(events_handled > 0)
		    break;
	    }
	    break;

	  case KeyRelease:
	    key = &event->xkey;
#if 0
	    if(!mb->has_focus) 
		break;
#endif
            /* Pass the key event unconditionally to each menu for hot
	     * key checks
             */
	    for(i = 0; i < mb->total_items; i++)
	    {
		item = mb->item[i];
		if(item == NULL)
		    continue;

		menu = item->menu;
		if(menu == NULL)
		    continue;

		events_handled += WidgetManage(menu, event);
		if(events_handled > 0)
		    break;
	    }
	    break;

	  case ButtonPress:
	    button = &event->xbutton;
	    if(w == mb->toplevel)
	    {
		/* Get the menu item that was click on */
		i = MenuBarMatchItemByPos(
		    mb, button->x, button->y
		);
		if(i < 0)
		{
		    /* No menu item was clicked on, so unmap the
		     * selected menu item's menu
		     */
		    if(mb->sel_item > -1)
		    {
			menu = MenuBarGetMenuFromItem(mb, mb->sel_item);
			if(menu != NULL)
			    WidgetUnmap(menu);
			mb->sel_item = -1;
		    }
		    break;
		}

		/* Clicked on the selected menu item? */
		if(mb->sel_item == i)
		{
		    /* Unmap menu */
		    menu = MenuBarGetMenuFromItem(mb, mb->sel_item);
		    if(menu != NULL)
			WidgetUnmap(menu);
		    mb->sel_item = -1;
		}
		else
		{
		    /* Clicked on an unselected menu item, so map that
		     * menu item's menu
		     */
		    int x, y, root_x, root_y;
		    unsigned int width, height;

		    OSWGetWindowGeometry(
			mb->toplevel, NULL, NULL, &width, &height, NULL
		    );
 
		    mb->sel_item = i;
		    item = mb->item[i];

		    /* Map menu */
		    menu = MenuBarGetMenuFromItem(mb, i);
		    if(menu != NULL)
		    {
			OSWGetPointerCoords(
			    mb->toplevel,
			    &root_x, &root_y,
			    &x, &y
			);
			MenuMapCoordinates(
			    menu,
			    root_x - x + item->x,
			    root_y - y + (int)height
			);
		    }
		}
		MenuBarQueueDraw(mb);
		events_handled++;
	    }
	    else
	    {
		/* Clicked on another window, check if a menu is
		 * currently mapped
		 */
		if(mb->sel_item > -1)
		{
		    /* Send this event to the mapped menu */
		    menu = MenuBarGetMenuFromItem(mb, mb->sel_item);
		    if((menu != NULL) ? menu->map_state : False)
		    {
			events_handled += WidgetManage(menu, event);
			if(events_handled > 0)
			    break;
		    }
		}
	    }
	    break;

	  case ButtonRelease:
	    button = &event->xbutton;
	    if(w == mb->toplevel)
	    {
		MenuBarQueueDraw(mb);
		events_handled++;
	    }
	    else
	    {
	        /* If the button was released anywhere else then
		 * set the seleted item to -1 but do not report event
		 * handled
		 */
	        if(mb->sel_item > -1)
		{
		    mb->sel_item = -1;
		    MenuBarQueueDraw(mb);
		}
	    }
	    break;

	  case MotionNotify:
	    motion = &event->xmotion;
	    if(w == mb->toplevel)
	    {
		if(mb->sel_item < 0)
		    break;

		i = MenuBarMatchItemByPos(
		    mb, (int)motion->x, (int)motion->y
		);
		if(i < 0)
		    break;

		if(mb->sel_item != i)
		{
		    int x, y, root_x, root_y;
		    unsigned int width, height;
		    menu_bar_item_struct *item = mb->item[i];

		    OSWGetWindowGeometry(
			mb->toplevel,
			NULL, NULL, &width, &height, NULL
		    );

		    /* Unmap previous menu and map new menu */
		    menu = MenuBarGetMenuFromItem(mb, mb->sel_item);
		    if(menu != NULL)
			WidgetUnmap(menu);

		    menu = MenuBarGetMenuFromItem(mb, i);
		    if(menu != NULL)
		    {
		        OSWGetPointerCoords(
			    mb->toplevel,
			    &root_x, &root_y,
			    &x, &y
		        );
			MenuMapCoordinates(
			    menu,
			    root_x - x + item->x,
			    root_y - y + (int)height
			);
		    }

		    mb->sel_item = i;
		    MenuBarQueueDraw(mb);
		    events_handled++;
		}
	    } 
	    break;

	  case VisibilityNotify:
	    if(w == mb->toplevel)
	    {
		mb->visibility_state = event->xvisibility.state;
		events_handled++;
	    }
	    break;
	}

	/* Forward the event to each menu */
	if(events_handled == 0)
	{
	    for(i = 0; i < mb->total_items; i++)
	    {
		item = mb->item[i];
		if(item == NULL)
		    continue;

		events_handled += WidgetManage(item->menu, event);
		if(events_handled > 0)
		    break;
	    }
	}

	return(events_handled);
}

/*
 *	Maps the menu bar.
 */
void MenuBarMap(menu_bar_struct *mb)
{
	if(mb == NULL)
	    return;

	/* Map as needed */
	if(!mb->map_state)
	{
	    OSWMapRaised(mb->toplevel);
	    mb->map_state = True;
	}
}

/*
 *	Unmaps the menu bar.
 */
void MenuBarUnmap(menu_bar_struct *mb)
{
	int i;
	menu_bar_item_struct *item;

	if(mb == NULL)
	    return;

	/* Unmap all menus */
	for(i = 0; i < mb->total_items; i++)
	{
	    item = mb->item[i];
	    if(item == NULL)
		continue;

	    WidgetUnmap(item->menu);
	}
	mb->sel_item = -1;

	OSWUnmapWindow(mb->toplevel);
	mb->map_state = False;

	OSWDestroyPixmap(&mb->toplevel_buf);
}

/*
 *	Destroys the menu bar.
 */
void MenuBarDestroy(menu_bar_struct *mb)
{
	int i;

	if(mb == NULL)
	    return;

	/* Delete all the menu bar items */
	for(i = 0; i < mb->total_items; i++)
	    MenuBarDeleteItem(mb, i);
	free(mb->item);
	mb->item = NULL;
	mb->total_items = 0;

	OSWDestroyWindow(&mb->toplevel);
	OSWDestroyPixmap(&mb->toplevel_buf);

	mb->activate_cb = NULL;
	mb->activate_data = NULL;

	mb->type = WIDGET_TYPE_UNKNOWN;
	mb->map_state = False;
	mb->has_focus = False;
	mb->visibility_state = VisibilityFullyObscured;
	mb->x = 0;
	mb->y = 0;
	mb->width = 0;
	mb->height = 0;
	mb->font = NULL;
	mb->next = NULL;
	mb->prev = NULL;
}
